/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/*!
 * @file alarmmgr.h
 *
 * @brief File public header for the ALP Alarm Manager APIs.
 *
 *****************************************************************************/
#ifndef __ALP_ALARMMGR_H__
#define __ALP_ALARMMGR_H__

#include <glib.h>
#include <hiker/types.h>
#include <hiker/sysclass.h>

#ifdef __cplusplus
 extern "C"
 {
#endif

/** 
 * @addtogroup ALP_ALARMMGR Alarm Manager
 * @ingroup ALP_SYSTEM
 * @{
 */

#define ALP_ALARM_MAX_APPID_SIZE		64 //!< Maximum size of an application ID

/**
 * @name Errors
 * @{
 */

#define ALP_STATUS_ALARM_NO_MEMORY		(ALP_CLASS_ALARM | 0x00010000) //!< Unable to allocation memory to complete function call

#define ALP_STATUS_ALARM_BAD_PARAM		(ALP_CLASS_ALARM | 0x00020000) //!< Supplied application ID is invalid, either being NULL or exceeding ALP_ALARM_MAX_APPID_SIZE or requested alarm time has already passed

#define ALP_STATUS_ALARM_IPC			(ALP_CLASS_ALARM | 0x00030000) //!< IPC error occurred during function call

/** @}*/

/**
 * @brief Set or cancel an alarm for the given application
 *
 * @param[in] appID        Package name of the calling application.
 * @param[in] ref          Caller-defined value.
 * @param[in] alarmSeconds Specifies the date/time in seconds since
 *                         01/01/1970 UTC at which the alarm will be
 *                         triggered, or 0 to cancel the alarm specified
 *                         by ref if any
 *
 * @return ALP_STATUS_OK    if no error occurred during call
 * @return ALP_STATUS_ALARM_BAD_PARAM if invalid parameters are passed such
 *                          as a NULL for appId
 */

     alp_status_t alp_alm_set_alarm (const gchar *appID,
                                     guint32 ref,
                                     guint32 alarmSeconds);


/**
 * @brief Return the current setting for an alarm
 *
 * Returns the date and time in seconds since 01/01/1970 UTC for the
 * given application ID and alarm reference number.
 *
 * @param[in]  appID        Package name of the calling application.
 * @param[in]  ref          Caller-defined value.
 * @param[out] alarmSeconds Pointer to location to hold the time value set
 *                          for the specified alarm. The time value represents
 *                          the number of seconds since 01/01/1970 UTC;
 *                          if no alarm is active for the application,
 *                          0 is returned for the alarm seconds.
 *
 * @return ALP_STATUS_OK    if no error occurred during call
 * @return ALP_STATUS_ALARM_BAD_PARAM if invalid parameters are passed such
 *                          as a NULL for appId or alarmSeconds
 * @return ALP_STATUS_ALARM_IPC if an IPC error occurred during the call
 */

     alp_status_t alp_alm_get_alarm(const gchar *appID,
                                    guint32 ref,
                                    guint32 *alarmSeconds);

/** @} */

#ifdef __cplusplus
}       // extern "C"
#endif

#endif  /*  __ALP_ALARMMGR_H__ */

