/******************************************************************************
 *
 * Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the entire contents of this file.
 *
 * The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.
 *
 * Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
 * Rights Reserved.
 * 
 * Contributor(s): none.
 *
 *****************************************************************************/
/**
 * @file attnmgr_common.h
 *
 * @brief ALP Attention common component definitions
 *
 *****************************************************************************/
#ifndef __ALP_ATTN_COMMON_H__
#define __ALP_ATTN_COMMON_H__

#include <time.h>
#include <hiker/types.h>
#include <hiker/sysclass.h>

/**
 * @addtogroup ALP_ATTN_MGR
 */
/* @{ */

/* typedefs common for attention manager components */
typedef AlpLuid       AlpAttnId;        //!< type of an attn alert ID
typedef void         *AlpAttnAlertH;    //!< opaque type holding an alert object
typedef void         *AlpAttnPropertyH; //!< opaque type holding a property object

/**
 * @name Status subclasses
 * Definitions for the various attention components
 * @{
 */
#define ALP_STATUS_ATTN                     (ALP_CLASS_ATTN  | 0x00010000) //!< Status class for attention manager API
#define ALP_STATUS_ATTN_DML                 (ALP_CLASS_ATTN  | 0x00020000) //!< Status class for attention manager DML
#define ALP_STATUS_ATTN_IPC                 (ALP_CLASS_ATTN  | 0x00030000) //!< Status class for attention manager IPC
/* @} */

/**
 * @name Errors
 * @{
 */
#define ALP_STATUS_ATTN_BAD_PARAM           (ALP_STATUS_ATTN | 0x00000100) //!< An invalid parameter was passed to the function call
#define ALP_STATUS_ATTN_NO_MEMORY           (ALP_STATUS_ATTN | 0x00000200) //!< Unable to allocate enough memory to complete the function call
#define ALP_STATUS_ATTN_INTERNAL_ERROR      (ALP_STATUS_ATTN | 0x00000300) //!< An unexpected internal error occurred during the function call
#define ALP_STATUS_ATTN_NOT_FOUND           (ALP_STATUS_ATTN | 0x00000400) //!< The requested item was not present
/* @} */

/**
 * @name Priority level definitions
 * @{
 */

/** priority levels */
typedef enum {
    ALP_ATTN_PRIO_NONE = -1,  //!< No priority specified
    ALP_ATTN_PRIO_1 = 1,      //!< Highest priority
    ALP_ATTN_PRIO_2 = 2,
    ALP_ATTN_PRIO_3 = 3,
    ALP_ATTN_PRIO_4 = 4,
    ALP_ATTN_PRIO_5 = 5,      //!< Lowest priority
    ALP_ATTN_PRIO_NULL = 254, //!< priority to sink alerts
    ALP_ATTN_PRIO_TEST = 255, //!< priority to test bad param
} AlpAttnPriority;

#define ALP_ATTN_PRIO_MAX         ALP_ATTN_PRIO_1
#define ALP_ATTN_PRIO_MIN         ALP_ATTN_PRIO_5

#define ALP_ATTN_PRIO_RANGE_LOWER 0
#define ALP_ATTN_PRIO_RANGE_UPPER 255
/* @} */

/**
 * @name Attention Alert Notifications
 * @{
 */
#define ALP_NOTIFY_EVENT_ATTN_ALERT     "/alp/attn/alert/"
typedef struct _AlpNotifyEventAttnAlert AlpNotifyEventAttnAlert;
struct _AlpNotifyEventAttnAlert {
    uint32_t index;    //<! Alert index (currently zero)
    uint32_t flags;    //<! Flags associated with alert (currently zero)
    char     handle[]; //<! Handle of alert associated with event
};


/* @} */

/**
 * @name Common data structures
 * These are used by the various components of the Attention Manager.
 * @{
 */

/**
 * @brief Attention Alert Property
 *
 * This structure is meant to hold a property for an attention alert.
 * It currently consists of a name field to hold the property key and
 * a value field to hold the property contents.
 */
typedef struct _AlpAttnProp AlpAttnProp;
struct _AlpAttnProp {
    const char  *name;  //!< Key name of property 
    const void  *value; //!< Pointer to property contents
};

#if 0
/*
 * These structures will probably be hidden in one of the
 * prv include files once the their proper scope is determined.
 */
typedef struct _AlpAttnProperty AlpAttnProperty;
struct _AlpAttnProperty {
    int32_t           itemType;
    AlpLuid           itemId;
    AlpAttnProperty  *next;

    /* property specific data */
    AlpLuid           alertId;
    const char       *name;
    const char       *value;
};

typedef struct _AlpAttnAlert AlpAttnAlert;
struct _AlpAttnAlert {
    int32_t           itemType;
    AlpLuid           itemId;
    AlpAttnAlert     *next;

    /* alert specific data */
    time_t            timestamp;
    bool              pending;
    char             *source;
    char             *name;
    char             *handle;
    char             *interface;
    int32_t           priority;
    int32_t           duration;
    AlpAttnProperty  *plist;
};
#endif
/* @} */

/* @} */
#endif /* __ALP_ATTN_COMMON_H__ */
