/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/*!
 * @file bundlemgr.h
 * 
 * @brief Public Header File for the Bundle Manager APIs. 
 *
 ******************************************************************************/


#ifndef ALP_BUNDLEMGR_H_
#define ALP_BUNDLEMGR_H_

#include <hiker/types.h> // for alp_status_t
#include <glib.h> //for gboolean, GList, GArray, gchar
#include <glib-object.h> // for GClosure*

#ifdef __cplusplus
extern "C" {
#endif

/*! @addtogroup ALP_BUNDLEMGR Bundle Manager
 *  @ingroup ALP_SYSTEM
 *  @{
 *
 *  The Bundle Manager is a system framework designed to coordinate and control the installation,
 *  and removal of software components, including applications and shared libraries, for various
 *  target subsystems, including the ALP native application format.
 *
 *  It is intended that the Bundle Manager provide the mechanism to install and track applications
 *  for other executable systems, as a 68K compatiblity environment, or Java apps.
 *
 *  The Bundle Manager APIs operate on three levels:
 *
 * 	Bundle Indexes: every bundle present in the system has a unique
 * 	index number, and index values will not be re-used. These indexes
 *	are stored in variables of type AlpBundle. You'll use bundle indexes
 *	to iterate over available bundles, retrieve metadata, and generally
 *	work across multiple bundles.
 *	
 *	Bundle Refs: in order to operate on the contents of a bundle, it
 *	must be opened. This gives back a handle, stored in a variable of
 *	type AlpBundleRef. Many operations can only be performed on a bundle
 *	if it is open. (Generally, the contents of a bundle are only 
 *	accessible while it is open. Other information, such as metadata,
 *	can be accessed if it isn't open.) You'll use bundle refs to
 *	work within a particular bundle. Opening a bundle to get a ref is
 *	a potentially expensive operation (consuming time, memory, limited
 *	resources, etc.) so you want to do this only when you have to.
 *
 *	Bundle Names: each bundle has a supplied name of some sort. Bundles
 *	can be retrieved by name. However, there can be conflicts (bundles
 *	of the same name, perhaps installed on separate locations on the system).
 *	If there are conflicts, one bundle is always 'elected' to respond to
 *	the name, based on conflict resolution rules. The name of a bundle is
 *	text, and is intended as a symbolic reference. You'll use bundle names
 *	to retrieve a bundle by a symbolic name, probably as part of the
 *	your program's resource material.
 *	
 * There are some other concepts used by the Bundle Manager:
 *
 *	Schemes: each type of bundle is considered a distinct 'scheme'. The
 *	ALP native appliacation format is the 'bar' scheme.
 *
 *	Stores: each bundle is stored in a particular location (perhaps a folder
 *	in a file-system). A 'store' is a name for such a location.
 *
 *	Images: one of the characteristics of a bundle is that it can be put into
 *	a flattened format for transmitting off of the device, to a another device,
 *	"beamed" via IR, attached to e-mail, downloaded off the web, etc. An
 *	"image" is flat file containing a bundle.
 *
 *	Metadata: information about a bundle is called "metadata". This includes
 *	material retrieved from the bundle's header/manifest/etc., as well as
 *	dynamic information about the actual use of the bundle (ref-count,
 *	system paths, etc.) Metadata for a bundle can be retrieved without
 *	opening the bundle.
 *
 *	Properties: a property is generic (but structured) material associated
 *	with a bundle. In the case of ALP native bundles, the properties are
 *	retrieved from arbitray XML in the bundle's manifest. Properties are
 *	intended for customization by the developer, representing arbitray
 *	information about a bundle. Various ALP components may rely on particular
 *	properties, such as the @ref ALP_APPLICATION_MGR using the "application"
 *	property. Properties for a bundle can be retrieved without opening the
 *	bundle.
 *
 *	URLs: various naming concepts used by the Bundle Manager are designed to
 *	be represented as text, which can be fitted together into standard URL
 *	syntax. To give a quick example of some of the possibilities:
 *
 *	@code
 *		bar:com.access.apps.memo
 *		bar:com.access.apps.memo/ListView.xml
 *		bar:com.access.apps.memo/libalp_memo.so?main
 *		bar:com.access.apps.memo/^p/application/icon
 *	@endcode
 *
 *	These represent <tt>scheme:name/path?entry</tt>, as will be described
 *	in the API reference. Generally, you don't need to use URLs with
 *	the Bundle Manager APIs, however there are a few routines which benefit
 *	from them, and give a lot of flexibility if they are used.
 *
 * <p><b>ALP native bundles</b></p>
 *
 * ALP native bundles are represented as folders or bar files which contain a Manifest.xml at its root.
 *
 * The Manifest must have an outer-most &lt;manifest&gt; element, which can have these attributes:
 *
 * 	name: this attribute is required, and contains the name of the bundle. It should not start
 *	with "bar:", as that part of the name is implied.
 *
 *	version: this attribute is optional, and if present should be an integer. If not present,
 *	the bundle version is considered to be 0.
 *
 *	arch: this attribute is optional, and if present should be a string indicating which
 *	architecture this bundle is intended for.
 *
 *	desired_filename: this attribute is optional, and if present indicates the file name
 *	which should be used to convert the bundle to an image file. The name given will
 *	have the ".bar" extension appended to it.
 *
 *	searchpath: this attribute is optional. If present, it should be a string in the form
 *	of a set of colon separate path prefixes, potentially containing "^l" to indicate a
 *	locale fill-point. This affects how localized resources are searched for. If not present,
 *	it defaults to <tt>/rsc/^l/:/</tt>.
 *
 *	textdomain: this attribute is optional. If present, it defines the gettext omain applied
 *	by default to code running from this bundle. It should be a string potentially containing
 *	"^n" or "^N". The former is mapped to the bundle name without scheme ("com.companyname.appname"),
 *	while the latter is mapped to the complete bundle name ("bar:com.companyname.appname").
 *	If not present, this attribute defaults to <tt>bar-^n</tt>.
 *
 *	bindtextdomain: this attribute is optional. If present, it defines the directory the
 *	gettext domain is bound to. It should be a string which defines a path relative to the
 *	base of the bundle. If not present, this attribute defaults to <tt>rsc</tt>.
 *
 * The rest of the manifest should be in this form:
 * 
 * @code
 *  <manifest name="foo.bar" searchpath="/share/%L/:/">
 *   <a>
 *     <b>1</b>
 *     <c>2</c>
 *   </a>
 *   <a>
 *     <x></x>
 *   </a>
 *   <other>
 *      <somevalue/>
 *   </other>
 *  </manifest>
 * @endcode
 *
 *
 * Any elements within the manifest at the third level ("b", "c", and "x" in the above example) will be registered as
 * bundle "properties", specifically:
 *
 * @code
 *   property "a", #0, key "b" -> "1"
 *                     key "c" -> "2"
 *   property "a", #1, key "x" -> ""
 *   property "other", #0, key "somevalue" -> ""
 * @endcode
 *
 * Any second-level elements not containing third-level elements, or any elements otherwise not conforming, will
 * be ignored for the purpose of generating bundle properties. They won't prevent the bundle from being accepted,
 * however.
 *
 * Two properties have specific impact on the support for native bundles:
 *
 * @code
 *    <reference>
 *      <to>bundle-name</to>
 *      <optional/>
 *    </reference>
 * @endcode
 *
 * Any "reference" properties present will indicate that this bundle has a dependency on another bundle,
 * particularly regarding use of shared libraries exported by the other bundle. The other bundle will 
 * automatically be opened when this bundle is opened. If the "optional" element isn't present, then this
 * bundle cannot be opened if the one it is dependant on isn't present.
 *
 * @code
 *    <export>
 *       <library>somelib.so</library>
 *       <as>lib_somelib.so.3</as>
 *    </export>
 * @endcode
 *
 * Or
 *
 * @code
 *    <export>
 *       <classpath>lib/myclass.jar</classpath>
 *    </export>
 * @endcode
 *
 * The "library" syntax will cause a library exported from this bundle to be available whenever the
 * bundle is open. The optional "as" element changes the name used for the exported version of the library.
 *
 * The "classpath" syntax adds an entry (either a folder or .jar file) to the CLASSPATH environment variable
 * while the bundle is open.
 *
 * Either "export" element can be repeated to export multiple things.
 *
 *
 * The resource files associated with ALP native bundles map directly to the contents of the bundle's
 * folder. A path starting with a slash is a "fixed path", and matches only that exact file within the
 * bundle: <tt>/rsc/MyView.xml</tt>. A path starting without a slash is looked up through the locale
 * search path, which can translate "MyView.xml" into "/rsc/en_US/MyView.xml", as appropriate.
 *
 * Note that the use of a leading slash to distinguish fixed paths from localized paths means that you
 * might need to use two slashes in URL syntax:
 *
 * @code
 *   bar:com.access.apps.memo//my/filename
 * @endcode
 *
 * There is one "magic" resource syntax specifically supported by ALP native bundles: the
 * translation syntax is:
 *
 * @code
 *   ^t/Some string
 * @endcode
 *
 * This "path" will be retrieved by taking the part of the path after the slash, and feeding
 * it through gettext() to perform a translation to the current locale. The result of the
 * translation is considered the "contents" of the resource.
 *
 * In URL syntax:
 *
 * @code
 *   bar:com.access.apps.memo/^t/MemoPad
 * @endcode
 *
 */

// -----------------------------------------------------------------------

/*! @name Initialization
 *
 *  @{
 */

/*! @brief Initialize the bundle manager client library. Needed if you are not running as an ALP application
 *  The bundle manager needs to be initialized before you can do anything with it. ALP applications
 *  normally run in an environment where the bundle manager is already initialized.
 *
 *  Use this routine if you want to support running in environments without an initialized bundle manager.
 *
 *  If the bundle manager is already initialized (in any mode), this will increment a refcount. Thus,
 *  it's safe to use from arbitrary code.
 */
alp_status_t alp_bundle_init(void);

/*! @brief Shutdown (finalize) the bundle manager client library.
 *
 *  If you explicitly initialized the bundle manager with alp_bundle_init(), you should use
 *  alp_bundle_fini(). This routine will decrement a refcount, and then shut down the bundle manager.
 */
alp_status_t alp_bundle_fini(void);

//! @}

// -----------------------------------------------------------------------

/*! @name Data types
 *
 *  @{
 */

//! @brief Data type used to refer to a particular bundle
//! The AlpBundle type is a short integral type used to index available bundles.
//! An AlpBundle value will be good until (reset?). (If a bundle is deleted, the
//! AlpBundle value will not be immediately reused). Any routine called with an
//! invalid AlpBundle value will fail. Using an invalid AlpBundle value will not
//! cause any alp_bundle routine to crash.
//!
//! Do not attempt to manipulate AlpBundle values numerically: always use values returned from
//! alp_bundle_by_name() or alp_bundle_search().
//!
//! An AlpBundle may be "NULL", meaning that it is a not a valid bundle index. This
//! condition can be tested with alp_bundle_is_NULL().
//! @note This type is a structure primarily so other numeric values aren't accidentally
//! used as AlpBundle indexes. If you want to step over each available bundle, use
//! alp_bundle_search() with a set of NULL parameters to retrieve all valid bundle indexes.
typedef struct _AlpBundle { guint value; } AlpBundle;

//! @brief Data type used as handle to an open bundle
//! The AlpBundleRef type is a pointer to an opaque type, used as a handle to
//! open bundles.
typedef struct _AlpBundleRef * AlpBundleRef;

//! @brief Test to see if an AlpBundle value is NULL
//! @param[in] index AlpBundle index to test
//! @returns TRUE if the AlpBundle index is NULL, i.e., doesn't represent a bundle
static inline gboolean alp_bundle_is_NULL(AlpBundle index)
{
	return index.value == 0;
}

//! @brief Constant which is a NULL AlpBundle value
#define ALP_BUNDLE_NULL ((struct _AlpBundle){0})

#define ALP_BUNDLE_REF_NULL (NULL)
			   
//! Generic failure
#define ALP_STATUS_BUNDLEMGR_FAILURE (ALP_CLASS_BUNDLEMGR | 0x00010000)
			   
  
//! @}

// -----------------------------------------------------------------------

/*! @name Finding bundles
 *
 *  @{
 */

//! @brief Get bundle index for a given name
//! @param[in] name This argument should be in the normal form of a bundle ID, i.e., a URL such as "bar:com.access.apps.Contacts". The name may be a full URL including path, etc.: that material will
//! be ignored for purposes of finding a bundle by name.
//! @returns An AlpBundle index to a bundle with a matching name if one exists, or else
//! a NULL AlpBundle value.
AlpBundle alp_bundle_by_name(const char * name);

//! @brief Search for bundles that match a set of criteria, including name, store, properties, type.
//! This function returns zero, one, or more AlpBundle indexes for bundle which match a set of
//! criteria. This is the fundamental routine for finding AlpBundle identifiers if you can't use
//! alp_bundle_by_name(). If you want to retrieve all valid bundle indexes, use this routine with
//! a set of NULL parameters.
//! @param[in] name A bundle name (of form "type:id", such as "bar:com.access.apps.Contacts"). May be NULL
//!                 if you want to search for a bundle with any name. To search for a bundle of a particular
//!                 type, but not name, use the scheme parameter. The name may be a full URL including path,
//!                 etc.: that material will be ignored for purposes of finding a bundle by name.
//! @param[in] named	TRUE if only named bundles should be returned, FALSE if you don't want to restrict the results.
//! @param[in] scheme A scheme name, such as "bar", or "ghost". May be NULL if you don't want to restrict the
//!                   search results by scheme type.
//! @param[in] store A store name. May be NULL if you don't want to restrict the search results by store.
//! @param[in] property A property name. If supplied, the search will only return bundles which have a property
//!                     of the given name. If NULL, then the results will not be restricted by properties.
//! @returns A GArray of AlpBundle values. If there were no results, a GArray will be returned, which is empty.
//!          NULL will be retuned if there is an actual error. In order to obtain the results, use the g_array
//!          routines:
//!		@code
//!             int i;
//!             for (i=0;i > results->len;i++) {
//!                 alp_bundle_open(g_array_index(results, AlpBundle, i));
//!             }
//!		@endcode
//!
//!          to free, use g_array_free(results);
GArray * alp_bundle_search(const char * name, gboolean named, const char * scheme, const char * store, const char * property);

//! All of these return a GList of allocated strings.
//! @brief Return a list of all valid bundle types the system supports
//! @returns a GList* containing allocated gchar*'s of valid bundle schemes. Free this list with g_list_foreach(results, (GFunc)g_free, NULL); g_list_free(results);
GList * alp_bundle_list_schemes(void);
//! @brief Return a list of all stores current available in the system
//! @returns a GList* containing allocated gchar*'s of available stores. Free this list with g_list_foreach(results, (GFunc)g_free, NULL); g_list_free(results);
GList * alp_bundle_list_stores(void);
//! @brief Return a list of all bundle names in the system
//! @returns a GList* containing allocated gchar*'s of available bundle names. Free this list with g_list_foreach(results, (GFunc)g_free, NULL); g_list_free(results);
GList * alp_bundle_list_names(void);
//! @brief Return a list of all properties which are present in any bundle in the system
//! @returns a GList* containing allocated gchar*'s of all bundle properties present in the system. Free this list with g_list_foreach(results, (GFunc)g_free, NULL); g_list_free(results);
GList * alp_bundle_list_properties(void);

//! @}

// -----------------------------------------------------------------------

/*! @name Operations on Bundles
 *  @{
 */

//! @brief Move a bundle to an active store
//! This uses the bundle manager to move a registered bundle from one
//! store to another.
//! @param[in] bundle index of bundle to move
//! @param[in] store name of store to move it to
//! @param[in] flags	ALP_BUNDLE_FLAG flags regarding collisions
//! @returns AlpBundle index of the moved bundle. Do not assume any relationship between the original and
//! new index. They may differ, or be identical.
//!
//! @note this routine is not intended to be used to get a copy of
//! a bundle as a flattened image for sending to another system, placing
//! on a media card, etc. Moving a bundle to an external store may have
//! the effect of turning it into a file which can be transmitted, but the
//! alp_bundle_image_*() routines are to be preferred for this purpose.
//!
//! @note Since the bundle is being moved between stores, and the elected bundle for a
//! particular name depends on store, it is possible that if similarly named bundles exist
//! on several stores, the bundle responding to a name could change after the move.
AlpBundle alp_bundle_move_to_store(AlpBundle bundle, const char * store, int flags);

	  
//! @brief Copy a bundle to an active store
//! This uses the bundle manager to copy a registered bundle from one
//! store to another. The new copy will be registered along with the old
//! one -- only one will have its name registered, and which one will
//! depend on the particular stores.
//! @param[in] bundle index of bundle to copy
//! @param[in] store name of store to copy it to
//! @param[in] flags	ALP_BUNDLE_FLAG flags regarding collisions
//! @returns AlpBundle index of the copy.
//!
//! @note this routine is not intended to be used to get a copy of
//! a bundle as a flattened image for sending to another system, placing
//! on a media card, etc. Copying a bundle to an external store may have
//! the effect of turning it into a file which can be transmitted, but the
//! alp_bundle_image_*() routines are to be preferred for this purpose.
AlpBundle alp_bundle_copy_to_store(AlpBundle bundle, const char * store, int flags);

//! @brief Delete a bundle
//! @param[in] bundle index of bundle to delete
//! @param[in] everything 1 to delete all associated data, including RW folder, global settings, etc., 0 otherwise
//! @returns Non-zero value if the deletion was not successful
alp_status_t alp_bundle_delete(AlpBundle bundle, int everything);

// metadata includes name, refcount, version, misc data
GHashTable * alp_bundle_metadata(AlpBundle);

//! @brief Return a the value for a metadata key of a bundle
//! Given a bundle index and a key name, return the value of that
//! key as an allocated string, or NULL if there is no such key.
//! @param[in] index	A bundle index. May be invalid, in which case an error will be returned.
//! @param[in] key	A key string. Results are undefined if this is NULL.
//! @returns 		An allocated string containing the value of the metadata field for the
//!			given key. If the value is empty, an empty string will be returned. If
//!			the key is not present, NULL will be returned. When you are done with the
//!			result, free it with g_free().
gchar * alp_bundle_metadata_value(AlpBundle index, const char * key);

//! @brief Returns the path to the R/W data folder for a bundle
//!
//! This routine will automatically make any folders necessary. I.e.,
//! a call to alp_bundle_rw_pathname(pkg, "a/b/c", TRUE); will gaurantee
//! that the "a", and "a/b" folders exist, so that the "a/b/c" file can
//! be created if it does not already exist.
//! @param[in] index	
//! @param[in] file_path
//! @param[in] vivify	If TRUE, then the path will be created if it does not
//!                     already exist. If FALSE, NULL will be returned if the path
//!			does not exist. NULL may be returned for the empty path "",
//!			if the entire R/W data folder has never been accessed.
//! @returns An allocated string containg the desired path, assuming that the
//! AlpBundle was valid. NULL otherwise. Free the result with g_free() when you
//! are done with it.
gchar * alp_bundle_rw_pathname(AlpBundle index, const char * file_path, gboolean vivify);

//! One R/W sub-folder is automatically backed up to the desktop. If you
//! want data to be backed up, please use the ALP_BUNDLE_RW_SAVED
//! path:
//!
//!   path = alp_bundle_rw_pathname(pkg, ALP_BUNDLE_RW_SAVED "mydata");
#define ALP_BUNDLE_RW_SAVED	"/saved/"


//! @brief Delete some or all files within a bundles R/W data folder
//! As a convenience, you can easily delete all of the R/W data for
//! a bundle, or a subset of that data, using this routine. If the
//! supplied file_path is empty (""), then all data will be removed,
//! otherwise only data in a folder matching the supplied path.
//! @param[in] index
//! @param[in] folder_path
//! @param[in] itself	If FALSE then only the contents of the named folder will be removed,
//! 			and not the folder itself.
//! @returns Non-zero value if the deletion was not successfull
alp_status_t alp_bundle_rw_delete(AlpBundle index, const char * folder_path, gboolean itself);

GHashTable * alp_bundle_properties(AlpBundle index, const char * property, int which);

GList * alp_bundle_all_properties(AlpBundle index);

//! Given a bundle index and a key name, return the value of that
//! key as an allocated string, or NULL if there is no such key.
//! @param[in] index	A bundle index. May be invalid, in which case an error will be returned.
//! @param[in] property
//! @param[in] which
//! @param[in] key	A key string. Results are undefined if this is NULL.
//! @returns 		An allocated string containing the value of the property field for the
//!			given key. If the value is empty, an empty string will be returned. If
//!			the key is not present, NULL will be returned. When you are done with the
//!			result, free it with g_free().
gchar * alp_bundle_property_value(AlpBundle index, const char * property, int which, const char * key);

/*! @brief Identify open bundle containing a file
 * Given a POSIX file path, this will attempt to identify an open bundle which
 * matches that path. The supplied path can be the root of a folder, or any file
 * or folder name within that folder (existing or not).
 *
 * @param[in] filename
 * @returns   NULL bundle index if no appropriate bundle can be found.
 */
AlpBundle alp_bundle_locate_file(const char * filename);

/*! @brief Return a bundle index of the bundle containing the running code
 *  Returns a bundle index to the bundle which contains the code which invoked
 *  this function. Note that this function explicitly detects what routine it is
 *  being called from. Don't expect it to operate from a utility subroutine -- 
 *  consider using a macro.
 *
 *  This routine identifies the caller, and determines if that caller is in 
 *  library loaded from a bundle. If the immediate caller is not in a bundle,
 *  NULL will be returned.
 *
 *  @note This routine is only specified to operate with shared libraries inside
 *        bundles. Operation when invoked from an executable is unpredictable.
 *
 *  @returns       NULL bundle index if the code is not running from an identifable bundle,
 *                 otherwise a bundle index.
 */
AlpBundle alp_bundle_me();

/*! @brief Return a bundle index of the running application
 *  Returns a bundle index to the bundle which contains the launched ALP application.
 *
 *  @returns       NULL bundle index if the code is not running from an ALP application,
 *                 otherwise a valid bundle index.
 */
AlpBundle alp_bundle_application();

//! @}

// -----------------------------------------------------------------------

/*! @name Operations on Stores
 *  @{
 */

/*! @brief Retrieve information about a store.
 */
GHashTable * alp_bundle_store(const char * store);
 
/*! @brief Retrieve a particular information value about a store.
 */
gchar * alp_bundle_store_value(const char * store, const char * key);

//! @}

// -----------------------------------------------------------------------  

/*! @name Operations on BundleRefs
 *  @{
 *
 * 
 */

/*! @brief Open a bundle
 * Given a bundle index, open the bundle, making its contents available for use,
 * until it is closed.
 *
 * Note that opening a bundle will automatically make sure that any bundles it
 * requires are also opened -- if any requirements are not met that are not optional,
 * the open will fail.
 */
AlpBundleRef alp_bundle_open(AlpBundle idx);

/*! @brief Close an open bundle
 *  Close a bundleref. No further access through the bundleref should be
 *  attempted. If all references to a bundle are closed, the
 *  bundle manager may disconnect or otherwise make the bundle unavailable
 *  for operations. Pathnames to ro bundle files may not function once it is
 *  closed.
 *
 *  @param[in] ref Bundleref to close
 *  @returns       non-zero if failed. 
 */
alp_status_t alp_bundle_close(AlpBundleRef ref);
	   
/*! @brief Return a bundle index given a bundle ref
 *  Returns a bundle index for a supplied bundle ref.
 *
 *  @param[in]	ref	A bundle ref, or NULL.This routine
 *                                  has undefined behaviour if the ref is invalid.
 *  @returns       A bundle index. Will be NULL if the bundle ref is NULL. 
 */
AlpBundle alp_bundle_ref_bundle(AlpBundleRef ref);

/*! @brief Return a bundle ref to the running code
 *  Returns a bundle ref to the bundle which contains the code which invoked
 *  this function. Note that this function explicitly detects what routine it is
 *  being called from. Don't expect it to operate from a utility subroutine -- 
 *  consider using a macro
 *
 *  This routine identifies the caller, and determines if that caller is in 
 *  library loaded from a bundle. If the immediate caller is not in a bundle,
 *  NULL will be returned.
 *
 *  @note.	This routine will return the same AlpBundleRef for each invocation
 *		from the same bundle. This AlpBundleRef will be effectively global,
 *		and should not be closed with alp_bundle_ref_close(). As such, it
 *		"leaks" and will not be closed until the application exits. For this
 *		reason, if you want fine control over the lifetime of the bundle containing
 *		code (likely for a plugin), then code in such bundles should not use
 *		alp_bundle_ref_me(), and instead use alp_bundle_me(), and explictly
 *		get an AlpBundleRef via alp_bundle_open() -- that will also be more efficient,
 *		as the invocation of either alp_bundle_me() or alp_bundle_ref_me() is somewhat expensive,
 *		and the result ought to be cached if possible.
 *
 *  @returns       NULL bundle ref if the code is not running from an identifable bundle,
 *                 otherwise a bundle ref.
 */
AlpBundleRef alp_bundle_ref_me();

	   
/*! @brief Return a bundle ref to the running application 
 *  Returns a bundle ref to the bundle which the running application was launched from.
 *
 *  @note.	This routine will return the same AlpBundleRef for each invocation
 *		from the same instance of application. This AlpBundleRef will be effectively global,
 *		and should not be closed with alp_bundle_ref_close(). As such, it
 *		"leaks" and will not be closed until the application exits.
 *
 *  @returns       NULL bundle ref if the code is not running from a launched application process,
 *                 otherwise a bundle ref.
 */
AlpBundleRef alp_bundle_ref_application();

/*! @brief Open a GIOChannel to retrieve a read-only file from a bundle
 *
 *  Given a pathname referring to a file existing within a bundle, this routine will
 *  open it for reading, and seeking if possible, and return a GIOChannel to the caller.
 *
 *  If the pathname starts with a slash, then it is taken as a fixed path within the
 *  bundle. If it does not start with a slash, it will be looked up according to the
 *  bundle's search path.
 *
 * @param[in] ref	Valid AlpBundleRef
 * @param[in] path	Path to file within bundle
 * @returns		A GIOChannel * if the file could be found and opened, otherwise NULL.
 *			The result should be freed with  g_io_channel_unref(). 
 *
 * @note The returned channel includes a reference to the bundle the file comes from. This will
 * keep the bundle open even if alp_bundle_close(ref) is invoked -- the bundle will not be
 * closed until the GIOChannel* is also unref'd.
 */
GIOChannel * alp_bundle_ref_ro_channel(AlpBundleRef ref, const char * path);

	   
/*! @brief Retrieve a pathname for read-only file in a bundle
 *
 *  Given a pathname referring to a file existing within a bundle, this routine will
 *  convert it to an absolute path, suitable for use with normal C and POSIX calls.
 *
 *  If the pathname starts with a slash, then it is taken as a fixed path within the
 *  bundle. If it does not start with a slash, it will be looked up according to the
 *  bundle's search path.
 *
 * @param[in] ref	Valid AlpBundleRef
 * @param[in] path	Path to file within bundle
 * @returns		An allocated absolute pathname if the file could be located, otherwise NULL.
 *			The result should be freed with g_free(). 
 *
 * @note The returned pathname is only valid as long as the bundle is open. You should not invoke
 * alp_bundle_close(ref) until after you have finished using the file represented by the returned path.
 */
gchar * alp_bundle_ref_ro_pathname(AlpBundleRef ref, const char * path);

/*! @brief Retrieve a read-only file in a bundle as an allocated data block
 *
 *  Given a pathname referring to a file existing within a bundle, this routine will
 *  load the file contents into a newly allocated data block.
 *
 *  If the pathname starts with a slash, then it is taken as a fixed path within the
 *  bundle. If it does not start with a slash, it will be looked up according to the
 *  bundle's search path.
 *
 * @param[in] ref	Valid AlpBundleRef
 * @param[in] path	Path to file within bundle
 * @param[out] data	Pointer to gchar * which will be filled in with the allocation for the file contents.
 * @param[out] length	Pointer to size_t which will be filled in with the size of the file contents.
 * @returns		Non-zero value if it was not possible to retrieve the data
 *
 * When you are done with the data, you should free it with g_free().
 *
 */
alp_status_t alp_bundle_ref_ro_data(AlpBundleRef ref, const char * path, gchar ** data, size_t * length);


/*! @brief Retrieve list of resources within a bundle
 *
 * Given a pathname referring to a folder within a bundle, this routine will
 * return a list of all items in that folder inside the bundle.
 *
 * @param[in] ref	Valid AlpBundleRef
 * @param[in] path	Path to folder within bundle. For consistency, these paths should
 * 			always start with a "/", so they cannot be considered search-path-relative.
 *			Use the path "/" to retrieve items from the root of a bundle.
 * @param[in] recurse	TRUE to recurse all of the way down through sub-folders in the bundle, FALSE
 *			to only return items directly contained in the supplied folder path.
 * @returns		A GList of allocate gchar* pointers, containing full paths for each item
 *			in the bundle at that folder. Items which are themselves folders will end
 *			in a slash. When you are done with the results, free the list with
 *			g_list_foreach(results, (GFunc)g_free, NULL); g_list_free(results);
 */
GList * alp_bundle_ref_ro_directory(AlpBundleRef ref, const char * path, gboolean recurse);

/*! @brief Create a temporary file with the contents of a read-only file in a bundle
 *
 *  Given a pathname referring to a file existing within a bundle, this routine will
 *  return the name of a temporary file suitable for retrieving the contents.
 *
 *  If the pathname starts with a slash, then it is taken as a fixed path within the
 *  bundle. If it does not start with a slash, it will be looked up according to the
 *  bundle's search path.
 *
 * @param[in] ref	Valid AlpBundleRef
 * @param[in] path	Path to file within bundle
 * @param[in] keep	True if the tempfile is going to be used after the bundle is closed
 * @param[in] ext	Unsupported: must be NULL
 * @returns		An allocated absolute pathname if the file could be located, otherwise NULL.
 *			The result should be freed with alp_bundle_ref_ro_tempfile_free(). 
 *
 * @note The returned tempfile may only be valid as long as the bundle is open. You should not invoke
 * alp_bundle_close(ref) until after you have finished using the file represented by the returned path.
 *
 * This routine is distinguished from alp_bundle_ref_to_pathname() because it will always be able to
 * return a file name, regardless of the particular bundle type -- if necessary, it will create a 
 * temporary file and copy the resource into it. This is also done if keep is true, so that the resource
 * will be accessible after the bundle is closed.
 */
gchar * alp_bundle_ref_ro_tempfile(AlpBundleRef ref, const char * path, gboolean keep, const char * ext);

/*! @brief Releases any storage associated with a tempfile returned from alp_bundle_ref_ro_tempfile()
 *
 *  This routine should be used when you are done with a tempfile. It'll take care of
 *  releasing any storage (file, memory, heap, etc.) associated with a tempfile.
 *
 * @param[in] tempfile	A value returned from a call to alp_bundle_ref_ro_tempfile()
 * @returns		Sucess or failure
 */
alp_status_t alp_bundle_tempfile_release(gchar * tempfile);

//! Use this define as the path to retrieve the manifest from a bar bundle
#define ALP_BUNDLE_BAR_MANIFEST	"/Manifest.xml"

//! @}

// -----------------------------------------------------------------------

/*! @name Operations on Images
 *  @{
 */

/*! @brief Tracks data for callback regarding progression of install or archive of bundle images
 *  An object of type AlpBundleProgressiveData is used to indicate to the alp_bundle_image_install()
 *  and alp_bundle_image_archive() routines that they should periodically invoke a callback function to
 *  let their caller know that the image creation/transfer process is proceeding.
 *
 *  The structure contains some fields that should be filled in by the caller, and other fields which will
 *  be filled in by either of the image routines.
 *
 *  Note that the fields filled in are only good for the duration of the image routine execution, and
 *  should only be retrieved from within the callback function. After the image routine returns, the
 *  fields will be zeroed.
 *
 *  To use this structure, create an object of the right size, and set version to ALP_BUNDLE_PROGRESSIVE_DATA_VERSION.
 *
 */
	
typedef struct _AlpBundleProgressiveData AlpBundleProgressiveData;
struct _AlpBundleProgressiveData {
	//! Caller should set this to ALP_BUNDLE_PROGRESSIVE_DATA_VERSION, so that the image
	//! routines know what layout the caller is using.
	gint version;
	//! Caller should set this. The callback should return TRUE if it wants to
	//! abort the image routine.
	gboolean (*callback)(AlpBundleProgressiveData*);
	
	enum {
		ALP_BUNDLE_PROGRESSIVE_HAVE_BYTES_SO_FAR	= 1<<0,
		ALP_BUNDLE_PROGRESSIVE_HAVE_TOTAL_BYTES	= 1<<1,
		ALP_BUNDLE_PROGRESSIVE_HAVE_MAX_OVERHEAD	= 1<<2,
		ALP_BUNDLE_PROGRESSIVE_HAVE_METADATA		= 1<<3,
		ALP_BUNDLE_PROGRESSIVE_HAVE_PROPERTIES		= 1<<4,
		ALP_BUNDLE_PROGRESSIVE_HAVE_FILENAME		= 1<<5,
		ALP_BUNDLE_PROGRESSIVE_HAVE_MIMETYPE		= 1<<6,
		ALP_BUNDLE_PROGRESSIVE_HAVE_STORE		= 1<<7,
		ALP_BUNDLE_PROGRESSIVE_HAVE_OVERWRITING_PKGINDEX = 1<<8,
		ALP_BUNDLE_PROGRESSIVE_HAVE_PKGINDEX		= 1<<9,
		ALP_BUNDLE_PROGRESSIVE_HAVE_FINISHED		= 1<<10
	} ALP_BUNDLE_PROGRESSIVE_ENUM;
	
	//! Any ALP_BUNDLE_PROGRESSIVE_HAVE_... flags which are added to
	//! current_state in this invocation of the callback will also be listed
	//! in new_state, for the convenience of the callback.
	guint new_state;
	
	//! Any ALP_BUNDLE_PROGRESSIVE_HAVE_... flags which are set in this member
	//! indicate that the respective fields are valid.
	guint current_state;

	//! Valid only if ALP_BUNDLE_PROGRESSIVE_HAVE_BYTES_SO_FAR, 0 otherwise
	//! This field indicates how many bytes of data have been processed
	guint bytes_so_far;
	//! Valid only if ALP_BUNDLE_PROGRESSIVE_HAVE_TOTAL_BYTES, 0 otherwise
	//! This field indicates how many bytes in total are needed
	guint total_bytes;
	//! Valid only if ALP_BUNDLE_PROGRESSIVE_HAVE_OVERHEAD, 0 otherwise
	//! This field indicates how much space has been temporarily used on
	//! the system in order to process the stream. Any overhead space will
	//! be released before the image call returns.
	guint max_overhead;

	//! Valid only if ALP_BUNDLE_PROGRESSIVE_HAVE_METADATA, NULL otherwise
	GHashTable * metadata;
	//! Valid only if ALP_BUNDLE_PROGRESSIVE_HAVE_PROPERTIES, NULL otherwise
	GHashTable * properties;
	//! Valid only if ALP_BUNDLE_PROGRESSIVE_HAVE_FILENAME, NULL otherwise
	const char * recommended_filename;
	//! Valid only if ALP_BUNDLE_PROGRESSIVE_HAVE_MIMETYPE, NULL otherwise
	const char * recommended_mimetype;
	//! Valid only if ALP_BUNDLE_PROGRESSIVE_HAVE_STORE, NULL otherwise
	const char * store;
	//! Valid only if ALP_BUNDLE_PROGRESSIVE_HAVE_OVERWRITING_PKGINDEX, NULL otherwise
	AlpBundle overwriting_index;
	//! Valid only if ALP_BUNDLE_PROGRESSIVE_HAVE_PKGINDEX, NULL otherwise
	AlpBundle index;
	
	//! This field is intended for the caller to refer to any additional data
	//! that might be needed, for state, etc.
	void * userdata;
};

enum {
	//! Current version of progressive data structure. This will change if the
	//! structure is modified in future.
	ALP_BUNDLE_PROGRESSIVE_DATA_VERSION = 1
};


/*! @brief Verify that a bundle image is somewhat undamaged, and get its metainfo
 *
 * This routine will verify that an uninstalled/unregistered image of a bundle (i.e.,
 * the contents of a .jar, .bar, or .prc file) fulfills some basic consistency checks, and
 * will retrieve the static metainformation from that bundle.
 *
 * This routine operates on a GIOChannel, so the bundle image can be read from a file,
 * or from some other data-source, potentially a streaming source.
 *
 *  @param[in] provide_filename	If not NULL, this should be the filename that was supplied with the data source.
 *                              It's possible that the name is neither relevant nor needed, but it can be supplied.
 *  @param[in] provide_mimetype If not NULL, this should be the mimetype that was supplied with the data source.
 *                              It's possible that the mimetype is neither relevant nor needed, but it can be supplied.
 *  @param[in] image	A GIOChannel * referring to a channel which the bundle data should be read from.
 *                      This channel should at least support READ operations, and more efficient operation may
 *                      be possible if it supports SEEK. WRITE will never be used.
 *  @param[in] full_check Set this gbool to FALSE if you want to skip expensive tests (such as complete
 *                        file-system integrity checks). It is not recommended to do this if you don't
 *                        know the bundle previously passed full verification.
 *  @param[out] metadata If not NULL, then this pointer will be filled with a copy of the bundle metadata.
 *  @param[out] properties If not NULL, then this pointer will be filled with a copy of the bundle properties
 *  @param[in] progress If not NULL, this should be an AlpBundleProgressiveData object, filled in as described
 */
int alp_bundle_image_verify(const char * provide_filename, const char * provide_mimetype, GIOChannel * image, gboolean full_check, GHashTable ** metadata, GHashTable ** properties, AlpBundleProgressiveData * progress);


/*! @brief Install a bundle into the system from an I/O channel
 *  Given a data source specified by a GIOChannel, this routine will read a bundle from
 *  that source, and install it into a store in the system. This is the fundamental routine
 *  used to install new programs.
 *
 *  @param[in] provide_filename	If not NULL, this should be the filename that was supplied with the data source.
 *                              It's possible that the name is neither relevant nor needed, but it can be supplied.
 *  @param[in] provide_mimetype If not NULL, this should be the mimetype that was supplied with the data source.
 *                              It's possible that the mimetype is neither relevant nor needed, but it can be supplied.
 *  @param[in] image	A GIOChannel * referring to a channel which the bundle data should be read from.
 *                      This channel should at least support READ operations, and more efficient operation may
 *                      be possible if it supports SEEK. WRITE will never be used.
 *  @param[in] destination_store If not NULL, this should be the name of a store which the 
 *                               bundle should be placed on. If NULL, the system will pick
 *                               a default, normally the internal store.
 *  @param[in] flags	0, or ALP_BUNDLE_FLAG_OVERWRITE_DUPLICATES, if you want to allow erasure of an identically named bundle
 *  @param[out] metadata If not NULL, then this pointer will be filled with a copy of the bundle metadata.
 *  @param[out] properties If not NULL, then this pointer will be filled with a copy of the bundle properties
 *  @param[in] progress If not NULL, this should be an AlpBundleProgressiveData object.
 *  @returns  An AlpBundle index if installation was successful, otherwise a NULL index.
 */
AlpBundle alp_bundle_image_install(const char * provide_filename, const char * provide_mimetype, GIOChannel * image, const char * destination_store, int flags, GHashTable ** metadata, GHashTable ** properties,  AlpBundleProgressiveData * progress); 

/*! @brief Archive a bundle to an I/O channel
 *  Given a bundle, and a data sink specified by a GIOChannel, this routine will flatten
 *  a bundle into an image, and write it to that channel. This is the fundamental routine
 *  used to prepare bundles for sending outside the device. (e-mail, beaming, files, etc.)
 *
 *  @param[in] index	A valid AlpBundle index
 *  @param[out] desired_filename	If not NULL, this will be filled in with the filename that the bundle thinks it should have.
 *  @param[out] desired_mimetype If not NULL, this will be filled in with the mimetype that the bundle thinks it should have.
 *  @param[in] image	A GIOChannel * referring to a channel which the bundle data should be written to. 
 *                      This channel should at least support WRITE operations, and more efficient operation may
 *                      be possible if it supports SEEK, or READ, (SEEK being more important than READ).
 *			If NULL, this routine will attempt to fill in the desired_filename and desired_mimetype
 *			parameters, if supplied, but will not actually archive the bundle.
 *  @param[in] progress If not NULL, this should be an AlpBundleProgressiveData object.
 *  @returns Non-zero value if is was not possible to archive the bundle
 */
alp_status_t alp_bundle_image_archive(AlpBundle index, char ** desired_filename, char ** desired_mimetype, GIOChannel * image, AlpBundleProgressiveData * progress); 

//! @}

// -----------------------------------------------------------------------

/*! @name Acquisition operations
 *  @{
 *
 * These are all implemented as composites of other routines, really just conveniences.
 *
 * The URLs allowed by these routines may have a scheme and bundle name and/or a path and/or a
 * search entry for entrypoint selection.
 *
 * That is, you can say:
 *
 * @code
 *	bar:bundle.name/resource/path
 *	/resource/path
 *	bar:bundle.name
 *	bar:bundle.name/library?function
 * @endcode
 *
 * An additional path syntax is available from the acquire routines to use bundle properties:
 *
 * @code
 *	bar:bundle.name/^p/application/icon
 * @endcode
 *
 * This has the affect of retrieving the value of the <tt>icon</tt> key of
 * the <tt>application</tt> property of the <tt>bar:bundle.name</tt> bundle, and
 * using that value as the name of another resource.
 *
 */

/*! @brief Get a function pointer to a shared library via a URL
 *
 *  @param[in] url		A bundle URL or an absolute <tt>file:///</tt> URL. If a bundle name is used, then a
 *				shared library within a bundle can be opened. The file path indicates the
 *				name of a shared library: if absent, then a name automatically derived for
 *				the particular bundle type will be used, e.g.
 *
 *				@code
 *					bar:com.access.apps.calculator
 *				@endcode
 *				is the same as
 *				@code
 *					bar:com.access.apps.calculator/libalp_calculator.so
 *				@endcode
 *
 *				A symbol name can be supplied with '?' syntax:
 *
 *				@code
 *					bar:bundle.name/lib_mylib.so?OtherEntry
 *				@endcode
 *
 *				The default symbol name is <tt>main</tt>.
 *
 *				An alternate functionality is available: if the path names an executable
 *				instead of a shared library, the returned closure will invoke the
 *				executable as a child process.
 *
 *				URL may be NULL, if a relativeRef is supplied. If this is the case, all defaults
 *				will be used.
 *
 *  @param[in] relativeRef	If supplied, and the URL does not have an absolute bundle name then this bundle ref is
 *				used as the bundle to apply the file path to. If NULL, then relative URLs will fail.
 *				It is acceptable (and intended) that a relativeRef be supplied even if the URL
 *				is absolute.
 *
 *  @returns	A GClosure* object. If a shared library entrypoint was acquired, a marshaller is set up
 *		which assumes the function takes the parameters <tt>int arc, char **argv, char **envp</tt>,
 *		and returns an int. If an executable was acquired, a marshaller is set up to invoke
 *		the executable in a child process and return its status. In both cases, the marshaller
 *		expects three arguments (int, char**, char **), corresponding to argc, argv, and envp,
 *		and returns an integer status value.
 *
 *  @note	The returned GClosure* keeps a reference on the bundle containing the entrypoint. The bundle
 *		will be opened in the process of obtaining the GClosure (if it was not already open), and the
 *		bundle will remain open until the GClosure is freed with g_closure_unref().
 *
 *		The bundle references are managed in such a way that you do not need additional unrefs due
 *		to using this function. If you opened a bundle to get a ref to pass as the relativeRef parameter,
 *		you should close that ref after using this routine. If the relative ref is used by the closure,
 *		an additional refcount will be made.
 *
 * @par Example of usage:
 *
 * @code
 *	AlpBundleRef myRef = alp_bundle_open(myBundleName);
 *	GClosure * myRoutine = alp_bundle_acquire_entrypoint("/somelib?routine", myRef);
 *	alp_bundle_close(myRef);
 *	g_closure_invoke(myRoutine, ...);
 *	g_close_unref(myRoutine);
 * @endcode
 */
GClosure * alp_bundle_acquire_entrypoint(const char * url, AlpBundleRef relativeRef /* may be NULL */);

/*! @brief Get a pathname to a R/O file within a bundle via a URL
 *
 *  @param[in] url		A bundle URL or absolute <tt>file:///</tt> URL. If a bundle name is used, then a file
 *				within a bundle can be located. This my be NULL if a relativeRef is supplied. If NULL
 *				then the pathname to the base of the bundle will be returned.
 *  @param[out] openRef 	If a matching file is found in a bundle, then the bundle will be opened, and it's
 *				reference placed in *openRef. You should close this ref with alp_bundle_close() when
 * 				you are done with the pathname. (A ref will be returned, even if the supplied
 *				relativeRef is used to locate the resource. Always close the returned ref,
 *				regardless of whether a relativeRef is supplied as input. That will assure the
 *				correct refcounts.) The openRef might be NULL if a <tt>file:///</tt> URL is supplied.
 *  @param[in] relativeRef	If supplied, and the URL does not have an absolute bundle name then this bundle ref is
 *				used as the bundle to apply the file path to. If NULL, then relative URLs will fail.
 *				It is acceptable (and intended) that a relativeRef be supplied even if the URL
 *				is absolute.
 *  @returns	An allocated gchar* or NULL. Free with g_free() when you are done with it. Close the openRef with
 *		alp_bundle_close() as well. Do not rely on the pathname being accessible after alp_bundle_close()
 *		is invoked.
 *
 *  @par Example of usage:
 *
 *  @code
 *	AlpBundleRef myRef = alp_bundle_open(myBundleName);
 *	AlpBundleRef outRef;
 *	gchar * pathname = alp_bundle_acquire_ro_pathname("/somefile", &outRef, myRef);
 *	alp_bundle_close(myRef);
 *	FILE * f = fopen(pathname, "r");
 *	// process f
 *	flcose(f);
 *	alp_bundle_close(outRef);
 *  @endcode
 */
gchar * alp_bundle_acquire_ro_pathname(const char * url, AlpBundleRef* openRef, AlpBundleRef relativeRef /* may be NULL */);

/*! @brief Get an GIOChannel for a R/O file within a bundle via a URL
 *
 *  This routine returns the pathname to a bundle resource. This only operates for some bundle types,
 *  particularly ALP native bundles (bar). Other types may not support direct file access, preventing this
 *  routine from being used. In that case, please consider using other routines, such as acquire_ro_data
 *  or acquire_ro_tempfile.
 *
 *  @param[in] url		A bundle URL or absolute <tt>file:///</tt> URL. If a bundle name is used, then a file
 *				within a bundle can be located.
 *  @param[in] relativeRef	If supplied, and the URL does not have an absolute bundle name then this bundle ref is
 *				used as the bundle to apply the file path to. If NULL, then relative URLs will fail.
 *				It is acceptable (and intended) that a relativeRef be supplied even if the URL
 *				is absolute.
 *  @returns	A GIOChannel* or NULL. Free with g_io_channel_unref() when you are done with it.
 *
 *  @note	The returned GIOChannel* keeps a reference on the bundle containing the file. The bundle
 *		will be opened in the process of obtaining the GIOChannel (if it was not already open), and the
 *		bundle will remain open until the GIOChannel is freed with g_io_channel_unref().
 *
 *		The bundle references are managed in such a way that you do not need additional unrefs due
 *		to using this function. If you opened a bundle to get a ref to pass as the relativeRef parameter,
 *		you should close that ref after using this routine. If the relative ref is used by the channel,
 *		an additional refcount will be made.
 *
 *  @par Example of usage:
 *
 *  @code
 *	AlpBundleRef myRef = alp_bundle_open(myBundleName);
 *	AlpBundleRef outRef;
 *	GIOChannel * channel = alp_bundle_acquire_ro_channel("/somefile", myRef);
 *	alp_bundle_close(myRef);
 *	// process channel
 *	g_io_channel_unref(channel);
 *  @endcode
 */
GIOChannel * alp_bundle_acquire_ro_channel(const char * url, AlpBundleRef relativeRef/* may be NULL */);

/*! @brief Get a block of data from the contents of a R/O file within a bundle via a URL
 *
 *  @param[in] url		A bundle URL or absolute <tt>file:///</tt> URL. If a bundle name is used, then a file
 *				within a bundle can be located.
 *  @param[out]	data		A pointer to a gchar* should be supplied. It will be filled in with the
 *				allocated data block, if the acquisition is successful.
 *  @param[out] length		A pointer to a size_t should be supplied, or NULL. If supplied, it will be
 *				filled in with the length of the R/O file.
 *  @param[in] relativeRef	If supplied, and the URL does not have an absolute bundle name then this bundle ref is
 *				used as the bundle to apply the file path to. If NULL, then relative URLs will fail.
 *				It is acceptable (and intended) that a relativeRef be supplied even if the URL
 *				is absolute.
 *
 *  @returns	Non-zero status on a failure. On success, ALP_STATUS_OK (0) is returned, and the supplied
 *		*data argument will have been filled in with a newly allocated block containing the contents
 *		of the R/O file. There is guaranteed to be a NUL past its end, to make sure it is NUL
 *		terminated. When you are done with the data, free it with g_free().
 *
 *  @note	If you opened a bundle to get a ref to pass as the relativeRef parameter you may close
 *		that ref after using this routine.
 *
 *  @par Example of usage:
 *
 *  @code
 *	AlpBundleRef myRef = alp_bundle_open(myBundleName);
 *	AlpBundleRef outRef;
 *	gchar * data;
 *	alp_status_t result = alp_bundle_acquire_ro_data("/sometextfile", &data, NULL, myRef);
 *	// process data
 *	g_free(data);
 *  @endcode
 */
alp_status_t alp_bundle_acquire_ro_data(const char * url, gchar ** data, size_t * length, AlpBundleRef relativeRef /* may be NULL */);

/*! @brief Get a temporary file for the contents of a R/O file within a bundle via a URL
 *
 *  @param[in] url		A bundle URL or absolute <tt>file:///</tt> URL. If a bundle name is used, then a file
 *				within a bundle can be located.
 *  @param[out] openRef 	If a matching file is found in a bundle, then the bundle will be opened, and it's
 *				reference placed in *openRef. You should close this ref with alp_bundle_close() when
 * 				you are done with the pathname. (A ref will be returned, even if the supplied
 *				relativeRef is used to locate the resource. Always close the returned ref,
 *				regardless of whether a relativeRef is supplied as input. That will assure the
 *				correct refcounts.) The openRef might be NULL if a <tt>file:///</tt> URL is supplied.
 *  @param[in]	keep		If true, the temporary file is gauranteed to remain accessible after the
 *  				openRef bundle is closed.
 *  @param[in]	ext		Not used. This parameter must be NULL.
 *  @param[in] relativeRef	If supplied, and the URL does not have an absolute bundle name then this bundle ref is
 *				used as the bundle to apply the file path to. If NULL, then relative URLs will fail.
 *				It is acceptable (and intended) that a relativeRef be supplied even if the URL
 *				is absolute.
 *  @returns	An allocated gchar* or NULL. The result points to a file name, which can be used
 *		for an open command. The form or contents of the name is not specified. When you are
 *		done with the name, it must be freed with alp_bundle_tempfile_release(). If keep is not true,
 *		then the file should be closed and released before the openRef bundle is closed.
 *
 *  @note The resulting temporary file should be released with alp_bundle_tempfile_release() to make sure
 *	  resources are properly reclaimed. However, the temporary file will be removed if
 *	  the program exits without releasing it.
 *		
 *	  If you acquired the tempfile without the keep parameter set to TRUE, release the tempfile
 *	  before closing the outRef, otherwise the results are undefined. 
 *
 *  @par Example of usage:
 *
 *  @code
 *	AlpBundleRef myRef = alp_bundle_open(myBundleName);
 *	AlpBundleRef outRef;
 *	gchar * filename = alp_bundle_acquire_ro_tempfile("/somefile", &outRef, FALSE, NULL, myRef);
 *	alp_bundle_close(myRef);
 *	FILE * f = fopen(filename, "r");
 *	// process f
 *	flcose(f);
 *	alp_bundle_close(outRef);
 *  @endcode
 */
gchar * alp_bundle_acquire_ro_tempfile(const char * url, AlpBundleRef* openRef, gboolean keep, const char * ext, AlpBundleRef relativeRef /* may be NULL */);

/*! This structure is derived from GClosure, and permit manipulating addition state returned
 *  by alp_bundle_acquire_ro_entrypoint().
 */
 
typedef struct {
  // initial layout should match that of GCClosure
  GClosure	closure;
  gpointer	callback;
  int		struct_version;
  AlpBundleRef	ref;
  void *	dlhandle;
  gchar *	path;
  gchar *	sym;
  int		result_signal;
} AlpBundleEntrypointClosure;

#define ALP_BUNDLE_ENTRYPOINT_CLOSURE_VERSION 1

//! Retrieve the AlpBundleRef held by the closure, if any
#define ALP_BUNDLE_ENTRYPOINT_CLOSURE_REF(closure) (((AlpBundleEntrypointClosure*)(closure))->ref)

//! Retrieve the dlopen() handle held by the closure, if any
#define ALP_BUNDLE_ENTRYPOINT_CLOSURE_DLHANDLE(closure) (((AlpBundleEntrypointClosure*)(closure))->dlhandle)

//! Retrieve the dlsym() function pointer held by the closure, if any
#define ALP_BUNDLE_ENTRYPOINT_CLOSURE_CALLBACK(closure) (((AlpBundleEntrypointClosure*)(closure))->callback)

//! Retrieve the R/O library path used by setting up the closure, if any
#define ALP_BUNDLE_ENTRYPOINT_CLOSURE_PATH(closure) (((AlpBundleEntrypointClosure*)(closure))->path)

//! Retrieve the symbol name used by dlsym(), if any
#define ALP_BUNDLE_ENTRYPOINT_CLOSURE_SYM(closure) (((AlpBundleEntrypointClosure*)(closure))->sym)

//! After closure invocation, this will be zero on a successful invocation, less than zero for
//! some internal errors, positive if an executable was invoked and terminated with a signal
#define ALP_BUNDLE_ENTRYPOINT_CLOSURE_RESULT_SIGNAL(closure) (((AlpBundleEntrypointClosure*)(closure))->result_signal)

/*! @brief Invoke a GClosure with command-line parameters from an array, return result
 *
 * This routine is a convenience function to invoke a GClosure with a set of parameters
 * resembling the C 'main' entry sequence:
 *
 * @code
 *    int (closure)(int argc, char *argv[], char *envp[])
 * @endcode
 *
 * This routine takes argv and evnvp array pointers, packs up their contents and invokes the
 * given closure. The result of the closure invocation is returned as an integer.
 *
 * @param[in]	closure		A GClosure*, probably returned from alp_bundle_acquire_ro_entrypoint
 * @param[in]	argv		An array of char*'s, which must be terminated with a NULL
 * @param[in]	envp		An array of char*'s, which must be terminated with a NULL -- or a NULL
 *				pointer. If NULL is passed, this routine will automatically use
 *				the global environ variable. If you want to pass an empty environment,
 *				you should construct an envp consisting of just a NULL value.
 * @returns 	exit status
 */
int alp_bundle_entrypoint_closure_execve(GClosure * closure, char * const argv[], char * const envp[]);

/*! Invoke a GClosure with command-line parameters from function parameters, return result
 *
 * This routine is a convenience function to invoke a GClosure with a set of parameters
 * resembling the C 'main' entry sequence:
 *
 * @code
 *    int (closure)(int argc, char *argv[], char *envp[])
 * @endcode
 *
 * This routine takes an arbitrary number of C parameters, packs them up and invokes the
 * given closure. The result of the closure invocation is returned as an integer.
 *
 * @param[in]	closure		A GClosure*, probably returned from alp_bundle_acquire_ro_entrypoint
 * @param[in]	arg, ...	A set of strings to be passed to the closure. This list must be
 *				terminated with NULL.
 * @returns 	exit status
 */
int alp_bundle_entrypoint_closure_execl(GClosure * closure, const char * arg, ...);

//! @}
// -----------------------------------------------------------------------

/*! @name Notifications
 *  @{
 */

/*! Name of notification sent when a bundle has is in the process of being deleted.
 *  If a handler of this notification set bit 0 of the handled flag, the deletion
 *  will be cancelled.
 */
#define ALP_NOTIFY_EVENT_BUNDLE_DELETE "/alp/bundlemgr/delete_bundle"
typedef struct _AlpNotifyEventBundleDelete AlpNotifyEventBundleDelete;
struct _AlpNotifyEventBundleDelete {
    AlpBundle index;
    int flags;
    gboolean named;
    char name[];
};
 
/*! Name of notification sent when a bundle has is in the process of being moved.
 *  If a handler of this notification set bit 0 of the handled flag, the move
 *  will be cancelled.
 */
#define ALP_NOTIFY_EVENT_BUNDLE_MOVE "/alp/bundlemgr/move_bundle"
typedef struct _AlpNotifyEventBundleMove AlpNotifyEventBundleMove;
struct _AlpNotifyEventBundleMove {
    AlpBundle index;
    char store[];
};

/*! Name of notification sent when a bundle has is in the process of being copied.
 *  If a handler of this notification set bit 0 of the handled flag, the copy
 *  will be cancelled.
 */
#define ALP_NOTIFY_EVENT_BUNDLE_COPY "/alp/bundlemgr/copy_bundle"
typedef struct _AlpNotifyEventBundleCopy AlpNotifyEventBundleCopy;
struct _AlpNotifyEventBundleCopy {
    AlpBundle index;
    char store[];
};
 
/*! Name of notification sent when a bundle has been registered.
 */
#define ALP_NOTIFY_EVENT_BUNDLE_REGISTER "/alp/bundlemgr/register_bundle"
typedef struct _AlpNotifyEventBundleRegister AlpNotifyEventBundleRegister;
struct _AlpNotifyEventBundleRegister {
    AlpBundle index;
};

/*! Name of notification sent when a bundle has is in the process of being unregistered.
 */
#define ALP_NOTIFY_EVENT_BUNDLE_UNREGISTER "/alp/bundlemgr/unregister_bundle"
typedef struct _AlpNotifyEventBundleUnregister AlpNotifyEventBundleUnregister;
struct _AlpNotifyEventBundleUnregister {
    AlpBundle index;
};
 
/*! Name of notification sent when all bundle manager state has been lost
 */
#define ALP_NOTIFY_EVENT_BUNDLE_RESET_EVERYTHING "/alp/bundlemgr/reset_everything"
typedef struct _AlpNotifyEventBundleResetEverything AlpNotifyEventBundleResetEverything;
struct _AlpNotifyEventBundleResetEverything {
};

/*! Name of notification sent when new bundle name has appeared.
 */
#define ALP_NOTIFY_EVENT_BUNDLE_NAME_ADDED "/alp/bundlemgr/name_added"
typedef struct _AlpNotifyEventBundleNameAdded AlpNotifyEventBundleNameAdded;
struct _AlpNotifyEventBundleNameAdded {
    AlpBundle index;
    char name[];
};

/*! Name of notification sent when a bundle name has been removed.
 */
#define ALP_NOTIFY_EVENT_BUNDLE_NAME_REMOVED "/alp/bundlemgr/name_removed"
typedef struct _AlpNotifyEventBundleNameRemoved AlpNotifyEventBundleNameRemoved;
struct _AlpNotifyEventBundleNameRemoved {
    AlpBundle index;
    char name[];
};

/*! Name of notification sent when a bundle name has changed to refer to a
 *  different bundle.
 */
#define ALP_NOTIFY_EVENT_BUNDLE_NAME_CHANGED "/alp/bundlemgr/name_changed"
typedef struct _AlpNotifyEventBundleNameChanged AlpNotifyEventBundleNameChanged;
struct _AlpNotifyEventBundleNameChanged {
    AlpBundle old_index;
    AlpBundle new_index;
    char name[];
};
 
 
//! @} // endname notifications

// -----------------------------------------------------------------------
/*! @name Manifest Metadata
 *  These constants define the names of common manifest metadata property lists. 
 *  @{
 */

/// The "application" manifest metadata indicates that the bundle contains an app.
#define ALP_BUNDLE_PROPERTY_APPLICATION	"application"


//! @}

// -----------------------------------------------------------------------
/*! @name Bundle Types
 *  These constants define the schemes for various bundle types. 
 *  @{
 */

/// The native ALP bundle type, "bar"
#define ALP_BUNDLE_TYPE_BAR		"bar"


//! @}

// -----------------------------------------------------------------------
/*! @name Miscellaneous and localization assistance
 *  @{
 */

//! Flag to pass to some bundle manager routines to indicate all associated data should be removed
#define ALP_BUNDLE_FLAG_DELETE_EVERYTHING 1
//! Flag to pass to some bundle manager routines to cause an image install to overwrite a bundle of the same name, if it exists
#define ALP_BUNDLE_FLAG_OVERWRITE_DUPLICATE 128


/*! Given a bundle ref, this will retrieve its domain for gettext
 *
 * Some bundle types (ALP native bundles -- "bar") support gettext for localization.
 * For these types of bundles, you can retrieve a text domain.
 *
 * @param[in]	ref	Bundle reference to retrieve text domain for
 * @returns		A char* of the appropriate textdomain, or NULL. This value is
 *			owned by bundle manager, and should not be freed, or used after
 *			the bundle reference is closed.
 */
const char * alp_bundle_ref_gettextdomain(AlpBundleRef ref);

/*! Given a bundle ref, this will retrieve the bundle's name
 *
 * @param[in]	ref	Bundle reference to retrieve name for
 * @returns		A char* of the bundle name. This value is owned by bundle manager,
 *			and should not be freed, or used after the bundle reference is closed.
 */
const char * alp_bundle_ref_name(AlpBundleRef ref);

/*! Given a bundle index, this will retrieve the bundle's name
 *
 * @param[in]	index	Bundle index to retrieve text domain for
 * @returns		An allocated gchar* of the bundle name. This value should be freed with
 *			g_free() when you are done with it.
 */
gchar * alp_bundle_name(AlpBundle index);

/*! Given a bundle index, retrieve the matching saved global settings key
 *
 * Saved settings are global settings specific to a bundle which survive
 * even when the associated bundle is deleted. The settings are related to bundle
 * name, so any bundles of the same name share the same saved global settings.
 *
 * @param[in]	index	Bundle index to retrieve settings key for
 * @param[in]	key	Sub-key to use for specific settings. This may be empty or NULL, in which
 *			case the primary key for saved settings will be returned.
 * @returns		An allocated gchar* of the appropriate global settings key.
 *			This value should be freed with	g_free() when you are done with it.
 */
gchar * alp_bundle_saved_settings(AlpBundle index, const char * key);

/*! Given a bundle index, retrieve the matching unsaved global settings key
 *
 * Unsaved settings are global settings specific to a bundle which are deleted
 * when the associated bundle is deleted. The settings are related to bundle
 * name, so any bundles of the same name share the same unsaved global settings.
 *
 * @param[in]	index	Bundle index to retrieve settings key for
 * @param[in]	key	Sub-key to use for specific settings. This may be empty or NULL, in which
 *			case the primary key for unsaved settings will be returned.
 * @returns		An allocated gchar* of the appropriate global settings key.
 *			This value should be freed with	g_free() when you are done with it.
 */
gchar * alp_bundle_unsaved_settings(AlpBundle index, const char * key);


/*! Utility macro to load a glade XML file from a bundle
 *
 * @param[in]	url	URL to bundle including path to glade xml file
 * @param[in]	root	Name of root to start from, or NULL
 * @returns		A GladeXML *
 */
#define alp_bundle_acquire_glade_xml(url, root)	\
({							\
  AlpBundleRef _ref;					\
  AlpBundleRef _me = alp_bundle_ref_me();		\
  gchar * _pathname = alp_bundle_acquire_ro_pathname((url), &_ref, _me);	\
  GladeXML * _result = glade_xml_new_with_domain(_pathname, (root), alp_bundle_ref_gettextdomain(_ref));	\
  g_free(_pathname);					\
  alp_bundle_close(_ref);				\
  _result;						\
})

/*! Utility macro to retrieve a gettext localization
 *
 * @param[in]	ref	Bundle reference
 * @param[in]	text	String to localize
 * @param[in]	category	gettext category
 * @returns		A char* which should not be freed
 */
#define alp_bundle_ref_cgettext(ref, text, category)		\
  dcgettext(alp_bundle_ref_gettextdomain((ref)), (text), (category))	\

/*! Utility macro to retrieve a gettext localization
 *
 * @param[in]	ref	Bundle reference
 * @param[in]	text	String to localize
 * @returns		A char* which should not be freed
 */
#define alp_bundle_ref_gettext(ref, text)			\
  dgettext(alp_bundle_ref_gettextdomain((ref)), (text))	\


/*! Utility macro to retrieve a gettext localization from the current bundle
 *
 * You might want to say:
 *
 * @code
 * 	#define _(x) alp_bundle_gettext(x)
 * @endcode
 *
 * @param[in]	text	String to localize
 * @returns		A char* which should not be freed
 */
#define alp_bundle_gettext(text)				\
  dgettext(alp_bundle_ref_gettextdomain(alp_bundle_ref_me()), (text))

/*! Returns status of last failing alp_bundle routine.
 *
 * Some Bundle Manager routines cannot return a detailed status value.
 * Any Bundle Manager routine which fails (and returns and error value)
 * will set an error status value which can be retrieved via this function.
 *
 * Successfull operations do not clear this status -- it is only modified
 * by failing functions. This function should not be used except immediately
 * after a failing Bundle Manager routine.
 *
 * @returns	An alp_status_t value which won't be ALP_STATUS_OK
 */
alp_status_t alp_bundle_error(void);

//! @}


// -----------------------------------------------------------------------
/*! @name Scheme routines
 *  @{
 */

/*! @brief Return all data describing a scheme
 *  Given a scheme name, returns a hashtable containing keys and values describing the scheme.
 *
 *  As an example, the "bar" scheme has "extension" and "mimetype" keys. The value for the
 *  "extension" key is ".bar".
 
 *  @param[in] scheme	A scheme name. May be an invalid scheme name or NULL, in which case an error will be returned.
 *  @returns 		An GHashTable containing allocate gchar* keys and allocated gchar* values. If
 *			the scheme isn't valid, NULL will be returned.
 *
 *			When you are done with the result, free it with g_hash_table_destroy().
 */
GHashTable * alp_bundle_scheme_values(const char * scheme);

/*! @brief Return a the value for a scheme description key
 *  Given a scheme name and a key name, return the value of that
 *  key as an allocated string, or NULL if there is no such key.
 *  @param[in] scheme	A scheme name. May be invalid, or NULL, in which case an error will be returned.
 *  @param[in] key	A key string. Results are undefined if this is NULL.
 *  @returns 		An allocated string containing the value of the scheme field for the
 *			given key. If the value is empty, an empty string will be returned. If
 * 			the key is not present, NULL will be returned. When you are done with the
 * 			result, free it with g_free().
 */
gchar * alp_bundle_scheme_value(const char * scheme, const char * key);

/*! @brief Given a filename and/or mimetype, determine what bundle scheme is appropriate
 *
 * If there's enough information to make a guess about the type of bundle represented
 * by a filename and/or mimetype, this routine will return the matching scheme name.
 *
 * The scheme name returned may not be correct, and this does not gaurantee that
 * attempting to install such a named bundle with that scheme will succeed -- this is
 * strictly a front-line convenience routine. 
 *
 * @param[in]	filename	Name of file, or NULL
 * @param[in]	mimetype	Mimetype, or NULL
 * @returns			An allocated string containng the appropriate scheme name
 * 				if one appears to be appropriate. NULL otherwise. When you
 *				are done with the result, free it with g_free();
 */
gchar * alp_bundle_scheme_identify(const char * filename, const char * mimetype);

//! @}


//! @} // endgroup ALP_BUNDLEMGR

#ifdef __cplusplus
}
#endif


#endif /* ALP_BUNDLEMGR_H_ */

