/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/

/*!
 * @file exgmgrutils.h 
 *
 * @brief Public headers for Exchange Manager utilities
 *
 */

#ifndef EXGMGRUTILS_H_
#define EXGMGRUTILS_H_


#include <linux/types.h>
#include <hiker/types.h>
#include <hiker/exgmgr.h>

#ifdef __cplusplus
 extern "C"
 {
#endif

/*! @addtogroup ALP_EXCHANGE_MGR Exchange Manager
 *  @ingroup ALP_SYSTEM
 *  @{
 */

/*! @name Creating a 'store' request that also works through IrOBEX or BtOBEX
 *	Requests that are just used to send data are special because they must also work when destined to a non ALP device through Irda or Bluetooth.
 *  In theses cases, the transport will automatically convert the request to OBEX objects and use an OBEX transport to deliver the data to the destination.
 *  To create this special request, you must use the set of functions below.
 *  @{
 */
 
/*! @brief Create a request to send objects using OBEX

 This function creates a new request and sets the right verb to let the exchange transport know you want to send objects using OBEX.
 This lets you send one or more objects to any device which understands OBEX.
 Note that upon return, the request verb will have been set (don't set it yourself).

 @param[out] oRequest pointer to an AlpExgRequest variable that will receive the created request reference.
 
 @return
	#ALP_EXGMGR_ERR_NO_ERROR - success<br>
*/
alp_status_t alp_exg_utils_store_request_create (AlpExgRequest *oRequest);

/*! @brief add an object by file descriptor

 Use this function when you want to send only a single object and specify a file descriptor. If you elect to use this function
 you cannot use alp_exg_utils_store_request_add_object().
 Note that the fd belongs to the request if the function returns no error. It will be closed when the request is destroyed.

 @param[in] iRequest pointer to an AlpExgRequest variable that will receive the created request reference.
 @param[in] iFd the data file descriptor (belongs to the request when this function returns no error).
 @param[in] iMIMEType the data type (mandatory, e.g. "text/x-vcard")
 @param[in] iDescription description that will be passed through Obex (optional, pass NULL if no description)
 @param[in] iName an object name that will be passed through OBEX (mandatory). This name should end with an extension equivalent to the MIMEType (e.g. ".vcf").
     
 @return
	#ALP_EXGMGR_ERR_NO_ERROR - success<br>
*/
alp_status_t alp_exg_utils_store_request_set_data_fd (AlpExgRequest iRequest, int iFd, char *iMIMEType, char *iDescription, char *iName);

/*! @brief add an object by filename

 Use this function when you want to send one or several objects. Each object must be stored in its own file. If you elect to use this function
 you cannot use alp_exg_utils_store_request_set_data_fd().

 @param[in] iRequest pointer to an AlpExgRequest variable that will receive the created request reference.
 @param[in] iFd the data file descriptor (belongs to the request when this function returns no error).
 @param[in] iFilename the name of the file containing the object (mandatory)
 @param[in] iMIMEType the data type (mandatory, e.g. "text/x-vcard")
 @param[in] iDescription description that will be passed through Obex (optional, pass NULL if no description)
 @param[in] iName an object name that will be passed through OBEX (mandatory). This name should end with an extension equivalent to the MIMEType (e.g. ".vcf").
     
 @return
	#ALP_EXGMGR_ERR_NO_ERROR - success<br>
*/
alp_status_t alp_exg_utils_store_request_add_object (AlpExgRequest iRequest, char *iFilename, char *iMIMEType, char *iDescription, char *iName);

//! @}	//end name

/*! @name Handling 'store' requests
 *  ALP exchange manager can receive data from both ALP and non ALP devices. In order for an application to receive such data objects, one must register
 *  a handler for the verb ALP_EXGMGR_STORE_OBJECT_VERB and the data type he is interested in. In the handler, he must then use the set of functions below
 *  to extract the data objects.
 *  @{
 */

/*! @brief Check if a request is a 'store' request

 Use this function to determine whether the request you have received is a 'store' request. You may need to check that in case you register
 the same handler for multiple verbs.

 @param[in] iRequest pointer to an AlpExgRequest variable that will receive the created request reference.
 @param[in] oIsStoreRequest a pointer to an int. Upon return, the int value will be 1 if iRequest is a 'store' request.
 
 @return
	#ALP_EXGMGR_ERR_NO_ERROR - success<br>
*/
alp_status_t alp_exg_utils_is_store_request (AlpExgRequest iRequest, int *oIsStoreRequest);

/*! @brief Get the number of objects received

 Use this function to determine how many objects you have received.

 @param[in] iRequest pointer to an AlpExgRequest variable that will receive the created request reference.
 @param[in] oObjectCount a pointer to an int. Upon return, the int value will be the number of objects received.
 
 @return
	#ALP_EXGMGR_ERR_NO_ERROR - success<br>
*/
alp_status_t alp_exg_utils_store_request_get_object_count (AlpExgRequest iRequest, int *oObjectCount);

/*! @brief Get the object data file descriptor <DEPRECATED, use alp_exg_utils_store_request_get_object_fd()>

 If there is only a single object in the request, it may have been passed by file descriptor. Call this function to check and possibly retrieve the fd.
 Once you have got the fd, it belongs to you and you are responsible to close it when you are finished with it.
 If the fd returned is -1, then use alp_exg_utils_store_request_get_object() to retrieve the object filename.

 @param[in] iRequest pointer to an AlpExgRequest variable that will receive the created request reference.
 @param[in] oFd a pointer to an int. Upon return, it will contain the fd or -1 if there is no fd available.
 @param[in] oMIMEType a pointer to a buffer where to store the data type (can be NULL if you don't want this info)
 @param[in] iMIMETypeBufLen the size of the buffer
 @param[in] oDescription a pointer to a buffer where to store the description (can be NULL if you don't want this info)
 @param[in] iDescriptionBufLen the size of the buffer
 @param[in] oName a pointer to a buffer where to store the name (can be NULL if you don't want this info)
 @param[in] iNameBufLen the size of the buffer

 @return
	#ALP_EXGMGR_ERR_NO_ERROR - success<br>
*/
alp_status_t alp_exg_utils_store_request_get_data_fd (AlpExgRequest iRequest, int *oFd, char *oMIMEType, size_t iMIMETypeBufLen, char *oDescription, size_t iDescriptionBufLen, char *oName, size_t iNameBufLen);

/*! @brief Get info about one of the received objects.

 The file you get is yours. You are responsible to either move it to your private storage, or delete it when done.

 @param[in] iRequest pointer to an AlpExgRequest variable that will receive the created request reference.
 @param[in] oFilename a pointer to a buffer where to store the filename (can be NULL if you don't want this info)
 @param[in] iFilenameBufLen the size of the buffer
 @param[in] oMIMEType a pointer to a buffer where to store the data type (can be NULL if you don't want this info)
 @param[in] iMIMETypeBufLen the size of the buffer
 @param[in] oDescription a pointer to a buffer where to store the description (can be NULL if you don't want this info)
 @param[in] iDescriptionBufLen the size of the buffer
 @param[in] oName a pointer to a buffer where to store the name (can be NULL if you don't want this info)
 @param[in] iNameBufLen the size of the buffer

 @return
	#ALP_EXGMGR_ERR_NO_ERROR - success<br>
*/
alp_status_t alp_exg_utils_store_request_get_object (AlpExgRequest iRequest, int iObjectNum, char *oFilename, size_t iFilenameBufLen, char *oMIMEType, size_t iMIMETypeBufLen, char *oDescription, size_t iDescriptionBufLen, char *oName, size_t iNameBufLen);

/*! @brief Get info about one of the received objects with data from file descriptor

 The fd you get is yours. You are responsible to close it when done.

 @param[in] iRequest pointer to an AlpExgRequest variable that will receive the created request reference.
 @param[in] oFd a pointer to an int. Upon return, it will contain an fd to the data. Close it when you have read the data.
 @param[in] oMIMEType a pointer to a buffer where to store the data type (can be NULL if you don't want this info)
 @param[in] iMIMETypeBufLen the size of the buffer
 @param[in] oDescription a pointer to a buffer where to store the description (can be NULL if you don't want this info)
 @param[in] iDescriptionBufLen the size of the buffer
 @param[in] oName a pointer to a buffer where to store the name (can be NULL if you don't want this info)
 @param[in] iNameBufLen the size of the buffer

 @return
	#ALP_EXGMGR_ERR_NO_ERROR - success<br>
*/
alp_status_t alp_exg_utils_store_request_get_object_fd (AlpExgRequest iRequest, int iObjectNum, int *oFd, char *oMIMEType, size_t iMIMETypeBufLen, char *oDescription, size_t iDescriptionBufLen, char *oName, size_t iNameBufLen);

/*! @brief Select a send method

 Sending data can be done through Bluetooth, Irda, or any other local application as an attachment. This function implements
 a dialog UI that merges all the possible methods for a given type and automatically sets the verb, data type and transport
 of the request so that it will reach the right handler.

 @param[in] iRequest pointer to an AlpExgRequest that you have just created
 @param[in] iMIMEType the mime type of the data you want to send.

 @return
	#ALP_EXGMGR_ERR_NO_ERROR - success<br>
*/
alp_status_t alp_exg_utils_send_method_select (AlpExgRequest iRequest, char *iMIMEType);


//! @}	//end name

//! @} // endgroup ALP_EXCHANGEMGR

#ifdef __cplusplus
 }
#endif


#endif /*EXGMGRUTILS_H_*/
