/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/*!
 * @file <hiker/ipc.h>
 *
 * @brief SDK Header File for ALP simple interprocess communication (IPC) API
 *
 ******************************************************************************/
#ifndef ALP_IPC_H_
#define ALP_IPC_H_ 1

#include <glib.h>

#include <hiker/config.h>
#include <hiker/sysclass.h>
#include <hiker/types.h>

#define ALP_STATUS_IPC_ERROR	((alp_status_t)ALP_CLASS_IPC | 0x00010000)
#define ALP_STATUS_IPC_TIMEOUT	((alp_status_t)ALP_CLASS_IPC | 0x00020000)
#define ALP_STATUS_IPC_BROKEN_CONNECTION ((alp_status_t)ALP_CLASS_IPC | 0x00030000)
#define ALP_STATUS_IPC_NOT_CONNECTED ((alp_status_t)ALP_CLASS_IPC | 0x00040000)
#define ALP_STATUS_IPC_NO_SERVER ((alp_status_t)ALP_CLASS_IPC | 0x00050000)
#define ALP_STATUS_IPC_INVALID_RESPONSE ((alp_status_t)ALP_CLASS_IPC | 0x00060000)

#define ALP_IPC_RECEIVE_ANY -1

#ifdef __cplusplus
extern "C" {
#endif

/*! 
@addtogroup ALP_IPC ALP Abstract IPC
@ingroup ALP_SYSTEM
@{
*/

// 
// Opaque typedef's
//

/*!
@brief Opaque type for representing a particular IPC channel (i.e. server)
*/
typedef struct _AlpChannel AlpChannel;		

/*!
@brief Opaque type representing a connection between the server 
and a single client
*/
typedef struct _AlpConnection AlpConnection;

/*!
@brief Opaque type used to contain the data for a message
*/
typedef struct _AlpMessage AlpMessage;

/*!
@brief Settings for channel access mode
*/
typedef enum {
	ALP_IPC_CHANNEL_ACCESS_OWNER,
	ALP_IPC_CHANNEL_ACCESS_ALL,
} AlpIpcChannelAccessMode;

/*!
@brief Connection usage hints
*/
typedef enum {
	ALP_IPC_CONNECTION_USAGE_HINT_NONE = 0x0,
	ALP_IPC_CONNECTION_USAGE_HINT_SYNCH_ONLY = 0x1,
} AlpConnectionUsageHint;

//
// Callback function signatures
//

// typedef's for connect, disconnect, and message receive callback functions

/*!
@brief Signature for connect callback function
*/
typedef void (*AlpChannelConnectCB) (AlpConnection* connection, 
                                     gpointer cbData);

/*! 
@brief Signature for disconnect callback function
*/
typedef void (*AlpChannelDisconnectCB) (AlpConnection* connection, 
                                        gpointer cbData);

/*!
@brief Signature for message receive callback function
*/
typedef void (*AlpMessageReceiveCB) (AlpConnection* connection,
							  AlpMessage* message,
							  gpointer cbData);


/*!
@brief Creates an AlpChannel

Should only be called by the server for a given channel name

The process that makes the call (i.e. the server for the channel) 
is responsible for calling alp_ipc_channel_destroy to clean-up the 
channel and its connections when finished.
@param channelName A string containing the name of the channel to be created
@param accessMode Sets permissions for who may access the channel.  Possible 
values are ALP_IPC_CHANNEL_ACCESS_OWNER and ALP_IPC_CHANNEL_ACCESS_ALL.  
For ALP_IPC_CHANNEL_ACCESS_OWNER, only processes with the same owner as
the channel creator will be able to connect to the IPC channel.  For 
ALP_IPC_CHANNEL_ACCESS_ALL, any process can connect to the IPC channel.
@return A pointer to the AlpChannel for the server if it was created OK
@return NULL otherwise
*/
AlpChannel* 	
alp_ipc_channel_create(const gchar* channelName, AlpIpcChannelAccessMode accessMode);


/*!
@brief Destroys an AlpChannel

Should only be called by the process that creating the channel by calling
alp_ipc_channel_create()

Notifies clients that have registered a disconnect callback, then
destroys all memory and closes any file descriptors used internally 
by the channel
@param channel Pointer to the AlpChannel to be destroyed
@return ALP_STATUS_OK if no error
@return ALP_STATUS_IPC_ERROR if any error condition occurred
*/
alp_status_t
alp_ipc_channel_destroy(AlpChannel* channel);	


/*!
@brief Registers a callback for the server process whenever a client connects

Only one connect callback function can be registered at any time.  If a second call to register a callback is made, it replaces the first call.  Passing NULL as the callback function deletes the callback.

NOTE: If the server process wishes to send messages to the client
who's connecting, it must store the AlpConnection* passed
to it in the callback when the client connects.
@param channel Pointer to the AlpChannel the callback is being added to
@param callback Callback function to be invoked when a client connects
@param cbData Pointer to data for the callback function
@return ALP_STATUS_OK if no error
@return ALP_STATUS_IPC_ERROR if any error condition occurred
*/
alp_status_t
alp_ipc_channel_register_connect_callback(AlpChannel* channel,
								  AlpChannelConnectCB callback,
								  gpointer cbData);


/*!
@brief Registers a callback for the server process whenever a 
client disconnects or exits without disconnecting

Only one disconnect callback function can be registered at any time.  If a second call to register a callback is made, it replaces the first call.  Passing NULL as the callback function deletes the callback.

@param channel Pointer to the AlpChannel the callback is being added to
@param callback Callback function to be invoked when a client disconnects
@param cbData Pointer to data for the callback function
@return ALP_STATUS_OK if no error
@return ALP_STATUS_IPC_ERROR if any error condition occurred
*/
alp_status_t
alp_ipc_channel_register_disconnect_callback(AlpChannel* channel,
									AlpChannelDisconnectCB callback,
									gpointer cbData);

/*! 
@brief Requests a connection to a AlpChannel 

Can be called by one or more processes to connect as clients of the channel

Does not block, rather a callback is invoked when the connection is ready
@param channelName Name of the channel to connect to
@param usageHint Provides hint of how connection will be used.  The IPC implementation may use this hint to optimize how the connection is set-up.  The param is a bitwise OR of the AlpConnectionUsageHint enum values.  Currently, the only values are ALP_IPC_CONNECTION_USAGE_HINT_NONE (no flags set) and ALP_IPC_CONNECTION_USAGE_HINT_SYNCH_ONLY which indicates that the connection will only be used for synchronous message sending (i.e. no received callback is registered, all messages are received using only alp_ipc_connection_receive and alp_ipc_connection_send_with_response).  Use ALP_IPC_CONNECTION_USAGE_HINT_NONE unless you REALLY understand what you're doing!
@param callback Callback function to be invoked when the connection is ready
@param cbData Pointer to data for the callback function
@return A pointer to the connection if successful 
@return NULL if the connection cannot be created
*/
AlpConnection*
alp_ipc_channel_connect(const gchar* channelName, 
				    AlpConnectionUsageHint hint,
				    AlpChannelConnectCB callback, 
				    gpointer cbData);

/*! 
@brief Requests a connection to a AlpChannel 

Can be called by one or more processes to connect as clients of the channel

The call blocks until the connection is ready or until the number of milliseconds specified by the timeout parameter have elapsed.

The caller passes in an AlpConnection** to receive the pointer to the new connection.  A valid AlpConnection* is returned only if the function's return status is ALP_STATUS_OK.  Otherwise, the returned connection parameter is set to NULL.

The caller is responsible for destroying the returned connection with alp_ipc_connection_destory.

NOTE: If the server has crashed unexpectedly, there may be cases where
ALP_STATUS_IPC_ERROR is returned instead of ALP_STATUS_IPC_NO_SERVER.

@param channelName Name of the channel to connect to
@param usageHint Provides hint of how connection will be used.  The IPC implementation may use this hint to optimize how the connection is set-up.  The param is a bitwise OR of the AlpConnectionUsageHint enum values.  Currently, the only values are ALP_IPC_CONNECTION_USAGE_HINT_NONE (no flags set) and ALP_IPC_CONNECTION_USAGE_HINT_SYNCH_ONLY which indicates that the connection will only be used for synchronous message sending (i.e. no received callback is registered, all messages are received using only alp_ipc_connection_receive and alp_ipc_connection_send_with_response).  Use ALP_IPC_CONNECTION_USAGE_HINT_NONE unless you REALLY understand what you're doing!
@param timeout Maximum amount of time to block before returning in milliseconds.  Specifying a timeout value of 0 causes the function to never unblock until a message is received.  If the timeout is reached before a message is received, the function returns NULL.
@param connection Pointer to receive the AlpConnection*.  
@return ALP_STATUS_OK if the connection was created
@return ALP_STATUS_IPC_TIMEOUT if the timeout occurred
@return ALP_STATUS_IPC_NO_SERVER if the server is not running
@return ALP_STATUS_IPC_ERROR if any other condition occurred
@return NULL if the connection cannot be created or if the timeout occurred
*/
alp_status_t
alp_ipc_channel_connect_synch(const gchar* channelName, 
						AlpConnectionUsageHint usageHint,
						guint timeout,
						AlpConnection** connection);



/*! 
@brief Disconnects process from an AlpChannel 

This should only be called by the
client processes that connected to the channel using alp_ipc_channel_connect().

This will trigger any registered disconnect callback in the server 
process when any client disconnects.  

This also destroys all memory and closes any file descriptors
used internally by the connection.
@param connection Pointer to the AlpConnection to disconnect
@return ALP_STATUS_OK if no error
@return ALP_STATUS_IPC_ERROR if any error condition occurred
*/
alp_status_t
alp_ipc_channel_disconnect(AlpConnection* connection); 


/*!
@brief Assigns a pointer to a custom block of data to be stored with this AlpChannel

The user owns the block of memory referenced by the custom data pointer and must free it 
when it is no longer needed.
@param channel Pointer to the AlpChannel where the custom data pointer is to be stored
@param customData Pointer to the custom data to be stored with the AlpChannel
*/
void
alp_ipc_channel_set_custom_data(AlpChannel* channel,
						  gpointer customData);

/*!
@brief Retrieves the pointer to a custom block of data stored with this AlpChannel

The user owns the block of memory referenced by the custom data pointer and must free it 
when it is no longer needed.
@param channel Pointer to the AlpChannel where the custom data pointer is stored
@return Pointer to the custom data stored with this AlpChannel
@return NULL if no custom data has been set
*/
gpointer
alp_ipc_channel_get_custom_data(AlpChannel* channel);


/*!
@brief Registers to receive a callback when the server disconnects by destroying the channel or exiting

Only one disconnect callback function can be registered at any time.  If a second call to register a callback is made, it replaces the first call.  Passing NULL as the callback function deletes the callback.

@param connection Pointer to the AlpConnection the callback is being added to
@param callback Callback function to be invoked when the server disconnects
@param cbData Pointer to data for the callback function
@return ALP_STATUS_OK if no error
@return ALP_STATUS_IPC_ERROR if any error condition occurred
*/
alp_status_t
alp_ipc_connection_register_disconnect_callback(AlpConnection* connection, 
									   AlpChannelDisconnectCB callback, 
									   gpointer cbData);

/*!
@brief Returns the process ID of the peer process for this connection.

@param connection AlpConnection pointer
@return The process ID of the peer process for the connection
@return -1 if connection is NULL or if the connection is not connected
*/
gint
alp_ipc_connection_get_peer_pid(const AlpConnection* connection);

/*!
@brief Returns the user ID of the peer process for this connection.

@param connection AlpConnection pointer
@return The user ID of the peer process for the connection
@return -1 if connection is NULL or if the connection is not connected
*/
gint
alp_ipc_connection_get_peer_uid(const AlpConnection* connection);

/*!
@brief Returns the group ID of the peer process for this connection.

@param connection AlpConnection pointer
@return The group ID of the peer process for the connection
@return -1 if connection is NULL or if the connection is not connected
*/
gint
alp_ipc_connection_get_peer_gid(const AlpConnection* connection);


/*!
@brief Allocates a AlpMessage structure

@return A pointer to the allocated AlpMessage structure if successfull
@return NULL if the allocation fails
*/
AlpMessage* 
alp_ipc_message_create();

/*! @brief Cleans up and frees all memory associated with the message

@param message Pointer to the message to be destroyed
@return ALP_STATUS_OK if no error
@return ALP_STATUS_IPC_ERROR if any error condition occurred
*/
alp_status_t
alp_ipc_message_destroy(AlpMessage* message);


/*!
@brief Returns the message ID of the message

The message ID is a integer value that can be any non-negative value agreed 
upon by the clients and server of a particular channel.  Negative 
values are reserved for internal use by the IPC module.  It is set 
by the sender of the message in the call to alp_ipc_message_start_pack().
@param message Pointer to an AlpMessage
@return The message ID of the message
*/
gint 
alp_ipc_message_get_message_ID(AlpMessage* message);


/*!
@brief Sends a message to the other end of a connection

The send is asynchronous (i.e. non-blocking)
@param connection Pointer to the AlpConnection on which the message is to be sent
@param message Pointer to the AlpMessage to send
@return ALP_STATUS_OK if no error
@return ALP_STATUS_IPC_NOT_CONNECTED if the connection is not ready for use (usually because the other end of the connection has not created)
@return ALP_STATUS_IPC_ERROR for other error conditions
*/
alp_status_t
alp_ipc_connection_send(AlpConnection* connection, AlpMessage* message);

/*!
@brief Sends a message to the other end of a connection, then waits for a
response message.

The caller passes in an AlpMessage** to receive the pointer to the response message.  The caller is responsible for deleting both the original and the response AlpMessage*'s when finished.  The receiver is also responsible for deleting the AlpMessage* it receives and the AlpMessage* response is returns.  A valid AlpMessage* is returned only if the function's return status is ALP_STATUS_OK.  Otherwise, the returned message parameter is set to NULL.

alp_ipc_connection_send_with_response
blocks until a response message for the sent message has 
been received.  The function will return if the 
specified timeout interval is reached before a message is received.  
Any messages received while blocked that do not correspond to the 
requested message ID are automatically queued and delivered once 
alp_ipc_connection_send_with_response returns.

The receiver of the message should should a response to the message, using alp_message_response_pack_start() instead of alp_message_pack_start() when creating the message.

The caller is responsible for deleting the response AlpMessage struct 
returned by calling alp_ipc_message_destroy.

@param connection Pointer to the AlpConnection on which the message is to be sent
@param message Pointer to the AlpMessage to send
@param timeout Maximum amount of time to block before returning in milliseconds.  Specifying a timeout value of 0 causes the function to never unblock until a message is received.  If the timeout is reached before a message is received, the status code ALP_IPC_TIMEOUT is returned.
@param response Pointer to receive the response AlpMessage*.  
@return ALP_STATUS_OK if no error
@return ALP_STATUS_IPC_TIMEOUT if the timeout occurs before the message is received
@return ALP_STATUS_IPC_BROKEN_CONNECTION if the connection is broken before the timeout or a message is received
@return ALP_STATUS_IPC_NOT_CONNECTED if the connection is not ready for use (usually because the other end of the connection has not created)
@return ALP_STATUS_IPC_INVALID_RESPONSE if the connection receives a message that's intended to be a response to a different message
@return ALP_STATUS_IPC_ERROR for other error conditions
*/
alp_status_t
alp_ipc_connection_send_with_response(AlpConnection* connection, 
							   AlpMessage* message,
							   guint timeout, 
							   AlpMessage** response);

/*!
@brief Synchronously receive a message from the other end of a connection

The caller passes in an AlpMessage** to receive the pointer to the received message.  The caller is responsible for deleting the AlpMessage* when finished.  A valid AlpMessage* is returned only if the function's return status is ALP_STATUS_OK.  Otherwise, the returned message parameter is set to NULL.

The caller can specify to wait for a message that matches a particular message ID or can wait for any message to be received on the connection by setting the messageID parameter to ALP_IPC_RECEIVE_ANY.  

If you specify a messageID of ALP_IPC_RECEIVE_ANY, alp_ipc_connection_receive will unblock and return the first message received on the connection.  If you specify any other value for messageID, alp_ipc_connection_receive will block until a message with the specified messageID has been received.  In either case, the function will return if the specified timeout interval is reached before a message is received.  Any messages received while blocked that do not correspond to the requested message ID are automatically queued and delivered once alp_ipc_connection_receive returns.

The caller is responsible for deleting the AlpMessage struct returned by callingalp_ipc_message_destroy.

@param connection Pointer to the AlpConnection on which the message is to be received
@param messageID Specifies which message ID to wait for.  
@param timeout Maximum amount of time to block before returning in milliseconds.  Specifying a timeout value of 0 causes the function to never unblock until a message is received.  If the timeout is reached before a message is received, the status code ALP_IPC_TIMEOUT is returned.
@param message Pointer to receive the AlpMessage*.  
@return ALP_STATUS_OK if no error
@return ALP_STATUS_IPC_TIMEOUT if the timeout occurs before the message is received
@return ALP_STATUS_IPC_BROKEN_CONNECTION if the connection is broken before the timeout or a message is received
@return ALP_STATUS_IPC_NOT_CONNECTED if the connection is not ready for use (usually because the other end of the connection has not created)
@return ALP_STATUS_IPC_INVALID_RESPONSE if the connection receives a message that's intended to be a response to an alp_ipc_connection_send_with_response() call
@return ALP_STATUS_IPC_ERROR for other error conditions
*/
alp_status_t
alp_ipc_connection_receive(AlpConnection* connection, gint messageID, 
					  guint timeout, AlpMessage** message);


/*!
@brief Registers to receive an asynchronous callback when a message is received by the connection

Only one receive callback function can be registered at any time.  If a second call to register a callback is made, it replaces the first call.  Passing NULL as the callback function deletes the callback.

The cbData structure specified is passed to the callback along 
with a pointer to the AlpMessage.  The message ID and the format 
of the message are available when the callback is made, but the 
other data in the message must be unpacked using
the various alp_ipc_message_unpack_* functions.

The caller is responsible for deleting the AlpMessage structure returned by
calling alp_ipc_message_destroy();

Requires that a GMainLoop is running in the process
@param connection Pointer to the AlpConnection the callback is being added to
@param callback Callback function to be invoked when a message is received
@param cbData Pointer to data for the callback function
@return ALP_STATUS_OK if no error
@return ALP_STATUS_IPC_ERROR if any error condition occurred
*/
alp_status_t
alp_ipc_connection_register_receive_callback(AlpConnection* connection,
									AlpMessageReceiveCB callback, 
									gpointer cbData);

/*!
@brief Assigns a pointer to a custom block of data to be stored with this AlpConnection

The user owns the block of memory referenced by the custom data pointer and must free it 
when it is no longer needed.
@param connection Pointer to the AlpConnection where the custom data pointer is to be stored
@param customData Pointer to the custom data to be stored with the AlpConnection
*/
void
alp_ipc_connection_set_custom_data(AlpConnection* connection,
							gpointer customData);

/*!
@brief Retrieves the pointer to a custom block of data stored with this AlpConnection

The user owns the block of memory referenced by the custom data pointer and must free it 
when it is no longer needed.
@param connection Pointer to the AlpConnection where the custom data pointer is stored
@return Pointer to the custom data stored with this AlpConnection
@return NULL if no custom data has been set
*/
gpointer
alp_ipc_connection_get_custom_data(AlpConnection* connection);

/*! 
@brief Call to begin to pack a message into an existing AlpMessage struct

Sets the message ID and prepares internal buffers for packing the
message data.  If reusing an AlpMessage struct, calling 
alp_ipc_message_pack_start() causes any old message info to be
discarded.

Packing a message involves calling alp_ipc_message_pack_start, 
followed by zero or more calls to alp_ipc_message_pack_<type> where
type is the value type being added to the message (e.g. int32, 
string, etc.), followed by alp_ipc_message_pack_end.

@param message Pointer to the AlpMessage being packed
@param messageID An integer value that can be any non-negative value agreed upon by the clients and server of a particular channel.  Negative values are reserved for internal use by the IPC library.
@return ALP_STATUS_OK if no error
@return ALP_STATUS_IPC_ERROR if any error condition occurred
*/
alp_status_t
alp_ipc_message_pack_start(AlpMessage* message, gint messageID);

/*! 
@brief Call to begin to pack a response message into an existing AlpMessage struct

Sets the message ID and response identifier and prepares internal 
buffers for packing the
message data.  If reusing an AlpMessage struct, calling 
alp_ipc_response_pack_start() causes any old message info to be
discarded.

Packing a response message involves calling alp_ipc_response_pack_start, 
followed by zero or more calls to alp_ipc_message_pack_<type> where
type is the value type being added to the message (e.g. int32, 
string, etc.), followed by alp_ipc_message_pack_end.

@param message Pointer to the AlpMessage being packed
@param messageID An integer value that can be any non-negative value agreed upon by the clients and server of a particular channel.  Negative values are reserved for internal use by the IPC library.
@param messageToRespondTo Pointer to the AlpMessage being responded to
@return ALP_STATUS_OK if no error
@return ALP_STATUS_IPC_ERROR if any error condition occurred
*/
alp_status_t
alp_ipc_message_response_pack_start(AlpMessage* message, gint messageID,
							 AlpMessage* messageToRespondTo);


/*!
@brief Call to finish packing a message into an existing AlpMessage struct

@param message Pointer to the AlpMessage being packed
@return ALP_STATUS_OK if no error
@return ALP_STATUS_IPC_ERROR if any error condition occurred
*/
alp_status_t
alp_ipc_message_pack_end(AlpMessage* message);


/*!
@brief Adds an int32 value to the message

@param message Pointer to the AlpMessage being packed
@param val The integer value to be appended to the message
@return ALP_STATUS_OK if no error
@return ALP_STATUS_IPC_ERROR if any error condition occurred
*/
alp_status_t
alp_ipc_message_pack_int32(AlpMessage* message, gint32 val);


/*!
@brief Adds an uint32 value to the message

@param message Pointer to the AlpMessage being packed
@param val The unsigned integer value to be appended to the message
@return ALP_STATUS_OK if no error
@return ALP_STATUS_IPC_ERROR if any error condition occurred
*/
alp_status_t
alp_ipc_message_pack_uint32(AlpMessage* message, guint32 val);


/*!
@brief Adds a float value to the message

@param message Pointer to the AlpMessage being packed
@param val The floating point value to be appended to the message
@return ALP_STATUS_OK if no error
@return ALP_STATUS_IPC_ERROR if any error condition occurred
*/
alp_status_t
alp_ipc_message_pack_float(AlpMessage* message, gfloat val);



/*!
@brief Adds a null-terminated string value to the message.  

@param message Pointer to the AlpMessage being packed
@param val The string to be appended to the message.  If val is NULL, the receiver will get a NULL pointer when unpacking the message.
@return ALP_STATUS_OK if no error
@return ALP_STATUS_IPC_ERROR if any error condition occurred
*/
alp_status_t
alp_ipc_message_pack_string(AlpMessage* message, const gchar* val);


/*!
@brief Adds a array of bytes (i.e. unsigned chars) to the message

@param message Pointer to the AlpMessage being packed
@param byteArray Pointer to the byte array.  If byteArray is NULL, numElements will be set to zero.
@param numElements Number of elements to pack. 
@return ALP_STATUS_OK if no error
@return ALP_STATUS_IPC_ERROR if any error condition occurred
*/
alp_status_t
alp_ipc_message_pack_byte_array(AlpMessage* message, 
						  guchar* byteArray, 
						  guint numElements);


/*!
@brief Call to begin unpacking a message from an AlpMessage struct

Unpacking a message involves calling alp_ipc_message_unpack_start, 
followed by zero or more calls to alp_ipc_message_unpack_<type> where
type is the value type being added to the message (e.g. int32, 
string, etc.), followed by alp_ipc_message_unpack_end.

Messages can be unpacked more than once by calling 
alp_ipc_message_unpack_start() a second time to reset the
internal state

@param message Pointer to the AlpMessage to unpack
@return ALP_STATUS_OK if no error
@return ALP_STATUS_IPC_ERROR if any error condition occurred
*/
alp_status_t
alp_ipc_message_unpack_start(AlpMessage* message);


/*!
@brief Call to finish unpacking a message from an AlpMessage struct

@param message Pointer to the AlpMessage to unpack
@return ALP_STATUS_OK if no error
@return ALP_STATUS_IPC_ERROR if any error condition occurred
*/
alp_status_t
alp_ipc_message_unpack_end(AlpMessage* message);


/*!
@brief Unpacks an int32 value from the message

@param message Pointer to the AlpMessage to unpack
@param val Pointer to an integer to store the unpacked value in
@return ALP_STATUS_OK if no error
@return ALP_STATUS_IPC_ERROR if any error condition occurred
*/
alp_status_t
alp_ipc_message_unpack_int32(AlpMessage* message, gint32* val);



/*!
@brief Unpacks an uint32 value from the message

@param message Pointer to the AlpMessage to unpack
@param val Pointer to an unsigned integer to store the unpacked value in
@return ALP_STATUS_OK if no error
@return ALP_STATUS_IPC_ERROR if any error condition occurred
*/
alp_status_t
alp_ipc_message_unpack_uint32(AlpMessage* message, guint32* val);



/*!
@brief Unpacks a float value from the message

@param message Pointer to the AlpMessage to unpack
@param val Pointer to a float to store the unpacked value in
@return ALP_STATUS_OK if no error
@return ALP_STATUS_IPC_ERROR if any error condition occurred
*/
alp_status_t
alp_ipc_message_unpack_float(AlpMessage* message, gfloat* val);


/*!
@brief Unpacks a null-terminated string from the message

Storage for the unpacked string is allocated and must be freed 
by the caller with g_free()
@param message Pointer to the AlpMessage to unpack
@param val Pointer to an gchar* to store the unpacked value in.  If the sender packed a NULL string pointer, *val will be set to NULL.
@return ALP_STATUS_OK if no error
@return ALP_STATUS_IPC_ERROR if any error condition occurred
*/
alp_status_t
alp_ipc_message_unpack_string(AlpMessage* message, gchar** val);


/*!
@brief Unpacks an array of bytes (i.e. unsigned chars) from the message

Storage for the byte array is allocated and must be freed 
by the caller with g_free()
@param message Pointer to the AlpMessage to unpack
@param array Pointer to the guchar* that is allocated to store the array in.  If numElements is zero, the pointer (i.e. *array) will be set to NULL.
@param numElements Pointer to an integer containing the number of elements in the array.  
@return ALP_STATUS_OK if no error
@return ALP_STATUS_IPC_ERROR if any error condition occurred
*/
alp_status_t
alp_ipc_message_unpack_byte_array(AlpMessage* message, 
						    guchar** array, 
						    guint* numElements);

/*!
@}
*/


#ifdef __cplusplus
}	// extern "C"
#endif

#endif // #ifndef ALP_IPC_H_

