/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/

/*! 
 * @file notifymgr.h
 * 
 * @brief Notification manager public header file 
 * 
 ******************************************************************************/


#ifndef NOTIFY_MGR_H
#define NOTIFY_MGR_H

#include <hiker/types.h>
#include <hiker/sysclass.h>

#ifdef __cplusplus
extern "C" {
#endif

/*! 
@addtogroup ALP_NOTIFY Notification Manager
@ingroup ALP_SYSTEM
@{
*/

// -----------------------------------------------------------------------------
// notification types
// -----------------------------------------------------------------------------

//! notification type: register notification
#define ALP_NOTIFY_EVENT_NOTIFY_REGISTER		"/alp/notify/register/"

//! notification type: install notification (deprecated, please use ALP_NOTIFY_EVENT_NOTIFY_REGISTER)
#define ALP_NOTIFY_EVENT_NOTIFY_INSTALL		"/alp/notify/install/"

//! notification type: boot notification
#define ALP_NOTIFY_EVENT_NOTIFY_BOOT		"/alp/notify/boot/"

// -----------------------------------------------------------------------------
// maximum sizes
// -----------------------------------------------------------------------------

//! maximum size of the notification type
#define ALP_NOTIFY_MAX_NOTIFY_SIZE		128

//! maximum size of the notification specific data
#define ALP_NOTIFY_MAX_DETAILS_SIZE		256

//! maximum size of the appIDs
#define ALP_NOTIFY_MAX_APPID_SIZE		64

// -----------------------------------------------------------------------------
// argv parameters when receiving a notification
// -----------------------------------------------------------------------------

//! command line parameter containing encoded notification details AlpNotifyLaunch
#define ALP_NOTIFY_LAUNCH_DETAILS		"--alp-notify-details="

//! command line parameter containing notification type
#define ALP_NOTIFY_LAUNCH_TYPE			"--alp-notify-type="

// -----------------------------------------------------------------------------
// register priority and flags
// -----------------------------------------------------------------------------

//! register priority: high priority notification
#define ALP_NOTIFY_PRIORITY_HIGH		10

//! register priority: normal priority notification
#define ALP_NOTIFY_PRIORITY_NORMAL		0

//! register priority: low priority notification
#define ALP_NOTIFY_PRIORITY_LOW			-10

//! register flags: don't save app registration (don't persist across resets)
#define ALP_NOTIFY_REGISTER_FLAG_DONTSAVE	0x01

//! register flags: automatic unregister after first use
#define ALP_NOTIFY_REGISTER_FLAG_ONETIME	0x02

// -----------------------------------------------------------------------------
// broadcast queues and flags
// -----------------------------------------------------------------------------

//! broadcast queue: default broadcast queue
#define ALP_NOTIFY_QUEUE_DEFAULT		0x01

//! broadcast queue: status gadget broadcast queue
#define ALP_NOTIFY_QUEUE_STATUSGADGET		0x02

//! broadcast flags: added the broadcast to the front of the queue
#define ALP_NOTIFY_BROADCAST_FLAG_IMMEDIATE	0x01

//! broadcast flags: sync broadcast, block until apps have been notified
#define ALP_NOTIFY_BROADCAST_FLAG_SYNC		0x02

// -----------------------------------------------------------------------------
// callback return values
// -----------------------------------------------------------------------------

//! continue the current broadcast
#define ALP_NOTIFY_CALLBACK_CONTINUE		0x00

//! cancel the current broadcast
#define ALP_NOTIFY_CALLBACK_CANCEL		0x01

//! already called the completion routine
#define ALP_NOTIFY_CALLBACK_COMPLETED_ALREADY	0x02

// -----------------------------------------------------------------------------
// error codes
// -----------------------------------------------------------------------------

//! error code: out of memory
#define ALP_NOTIFY_ERR_OUT_OF_MEMORY		(ALP_CLASS_NOTIFY | 0x00010000)

//! error code: invalid callback
#define ALP_NOTIFY_ERR_INVALID_CALLBACK		(ALP_CLASS_NOTIFY | 0x00020000)

//! error code: invalid appID
#define ALP_NOTIFY_ERR_INVALID_APPID		(ALP_CLASS_NOTIFY | 0x00030000)

//! error code: invalid queue
#define ALP_NOTIFY_ERR_INVALID_QUEUE		(ALP_CLASS_NOTIFY | 0x00040000)

//! error code: invalid priority
#define ALP_NOTIFY_ERR_INVALID_PRIORITY		(ALP_CLASS_NOTIFY | 0x00050000)

//! error code: invalid notification type
#define ALP_NOTIFY_ERR_INVALID_NOTIFYTYPE	(ALP_CLASS_NOTIFY | 0x00060000)

//! error code: invalid length
#define ALP_NOTIFY_ERR_INVALID_LENGTH		(ALP_CLASS_NOTIFY | 0x00070000)

//! error code: invalid command
#define ALP_NOTIFY_ERR_INVALID_CMD		(ALP_CLASS_NOTIFY | 0x00080000)

//! error code: not registered
#define ALP_NOTIFY_ERR_NOT_REGISTERED		(ALP_CLASS_NOTIFY | 0x00090000)

//! error code: already registered
#define ALP_NOTIFY_ERR_ALREADY_REGISTERED	(ALP_CLASS_NOTIFY | 0x000a0000)

//! error code: too many registrations
#define ALP_NOTIFY_ERR_TOO_MANY_REGISTRATIONS	(ALP_CLASS_NOTIFY | 0x000b0000)

//! error code: too many active broadcasts
#define ALP_NOTIFY_ERR_TOO_MANY_BROADCASTS	(ALP_CLASS_NOTIFY | 0x000c0000)

//! error code: active broadcast
#define ALP_NOTIFY_ERR_ACTIVE_BROADCAST		(ALP_CLASS_NOTIFY | 0x000d0000)

//! error code: alp ipc
#define ALP_NOTIFY_ERR_IPC			(ALP_CLASS_NOTIFY | 0x000e0000)

//! error code: sync broadcast timed out
#define ALP_NOTIFY_ERR_TIMEOUT			(ALP_CLASS_NOTIFY | 0x000f0000)

//! error code: permission denied
#define ALP_NOTIFY_ERR_PERMISSION_DENIED	(ALP_CLASS_NOTIFY | 0x00100000)

// -----------------------------------------------------------------------------
// data structures
// -----------------------------------------------------------------------------

/*! notification launch data
*/
typedef struct _AlpNotifyLaunch AlpNotifyLaunch;
struct _AlpNotifyLaunch {
	int handled;
	int length;
	char details[ALP_NOTIFY_MAX_DETAILS_SIZE+1];
	char notifyType[ALP_NOTIFY_MAX_NOTIFY_SIZE+1];
	char appID[ALP_NOTIFY_MAX_APPID_SIZE+1];
};

// -----------------------------------------------------------------------------
// functions
// -----------------------------------------------------------------------------

/*! Notification handler callback
@param[in] notifyType the type of notification (card inserted, vfs mounted, etc)
@param[in] details notification specific data
@param[in] length size of notification specific data
@param[in,out] handled advisory handled flag. Notification handlers can set an advisory handled flag indicating that the notification has been handled. Setting the handled flags does not cancel notification but rather informs subsequent handlers that the notification has already been handled. The value of the handled flag is only valid for a single instance of the notification (that is the value is not passed on to the next broadcast of the same notification).
@param[in] appID application that broadcasted the notification
@param[in] userData pointer to user data supplied at the registration of the callback
@return ALP_NOTIFY_CALLBACK_CONTINUE continue
@return ALP_NOTIFY_CALLBACK_CANCEL cancel this broadcast
@return ALP_NOTIFY_CALLBACK_COMPLETED_ALREADY completion has already been called (don't call again)
*/
typedef int (AlpNotifyCallbackProc)(const char * notifyType, void * details, int length, int * handled, const char * appID, void * userData);

/*! Registers to receive the specified notification
@param[in] notifyType the type of notification (card inserted, vfs mounted, etc) (must be non-null)
@param[in] callbackProc function pointer to callback function for running process (must be non-null)
@param[in] userData     user supplied data pointer to be used by callback function (may be NULL)
@param[in] priority priority in which the receiver is called (default is ALP_NOTIFY_PRIORITY_NORMAL)
@param[in] flags (0 for default and see ALP_NOTIFY_REGISTER_FLAG_ONETIME to change the default behavior)

@return ALP_OK success
@return ALP_NOTIFY_ERR_OUT_OF_MEMORY out of memory to perform the register
@return ALP_NOTIFY_ERR_INVALID_CALLBACK invalid callbackProc
@return ALP_NOTIFY_ERR_INVALID_NOTIFYTYPE invalid notifyType
@return ALP_NOTIFY_ERR_INVALID_PRIORITY invalid priority
@return ALP_NOTIFY_ERR_ALREADY_REGISTERED specified notifyType is already registered
@return ALP_NOTIFY_ERR_TOO_MANY_REGISTRATIONS too many registered registrations
@return ALP_NOTIFY_ERR_IPC ipc error

@remarks The default behavior is that the registration is persistent until the process exists or the notification is unregistered (see flags to change the default behavior). The callback function is called to handle the notification.
*/
extern alp_status_t alp_notify_register_server(const char * notifyType, AlpNotifyCallbackProc * callbackProc, void *userData, int priority, int flags);

/*! Registers an application to receive the specified notification
@param[in] appID application to receive the notification (must be non-null)
@param[in] notifyType the type of notification (card inserted, vfs mounted, etc) (must be non-null)
@param[in] priority priority in which the receiver is called (default is ALP_NOTIFY_PRIORITY_NORMAL)
@param[in] flags (0 for default and see ALP_NOTIFY_REGISTER_DONTSAVE and ALP_NOTIFY_REGISTER_FLAG_ONETIME to change the default behavior)

@return ALP_OK success
@return ALP_NOTIFY_ERR_OUT_OF_MEMORY out of memory to perform the register
@return ALP_NOTIFY_ERR_INVALID_APPID invalid appID
@return ALP_NOTIFY_ERR_INVALID_NOTIFYTYPE invalid notifyType
@return ALP_NOTIFY_ERR_INVALID_PRIORITY invalid priority
@return ALP_NOTIFY_ERR_ALREADY_REGISTERED specified appID and notifyType is already registered
@return ALP_NOTIFY_ERR_TOO_MANY_REGISTRATIONS too many registered registrations
@return ALP_NOTIFY_ERR_IPC ipc error

@remarks The default behavior is that the registration is persistent until the device hard resets or the notification is unregistered (see flags to change the default behavior). The application is launch or relaunched to handle the notification.
*/
extern alp_status_t alp_notify_register(const char * appID, const char * notifyType, int priority, int flags);

/*! unregisters the specified notification
@param[in] appID application previously registered to receive the notification (must be non-null)
@param[in] notifyType the type of notification (card inserted, vfs mounted, etc) (must be non-null)
@param[in] flags reserved for future use

@return ALP_OK success
@return ALP_NOTIFY_ERR_OUT_OF_MEMORY out of memory to perform the unregister
@return ALP_NOTIFY_ERR_INVALID_APPID invalid appID
@return ALP_NOTIFY_ERR_INVALID_NOTIFYTYPE invalid notifyType
@return ALP_NOTIFY_ERR_NOT_REGISTERED specified appID and notifyType are not registered
@return ALP_NOTIFY_ERR_ACTIVE_BROADCAST failed because of active broadcast
@return ALP_NOTIFY_ERR_IPC ipc error
*/
extern alp_status_t alp_notify_unregister(const char * appID, const char * notifyType, int flags);

/*! unregisters the specified notification
@param[in] notifyType the type of notification (card inserted, vfs mounted, etc) (must be non-null)
@param[in] flags reserved for future use

@return ALP_OK success
@return ALP_NOTIFY_ERR_OUT_OF_MEMORY out of memory to perform the unregister
@return ALP_NOTIFY_ERR_INVALID_NOTIFYTYPE invalid notifyType
@return ALP_NOTIFY_ERR_NOT_REGISTERED specified notifyType is not registered
@return ALP_NOTIFY_ERR_ACTIVE_BROADCAST failed because of active broadcast
@return ALP_NOTIFY_ERR_IPC ipc error
*/
extern alp_status_t alp_notify_unregister_server(const char * notifyType, int flags);

/*! broadcasts the specified notification
@param[in] appID application that broadcasts the notification the notification (must be non-null)
@param[in] notifyType the type of notification (card inserted, vfs mounted, etc) (must be non-null)
@param[in] details notification specific data
@param[in] length size of notification specific data
@param[in] queue broadcast queue (use ALP_NOTIFY_QUEUE_DEFAULT as the default) (must be non-zero)
@param[in] flags (0 for default and see ALP_NOTIFY_BROADCAST_IMMEDIATE and ALP_NOTIFY_BROADCAST_FLAG_SYNC to change the default behavior)

@return ALP_OK success
@return ALP_NOTIFY_ERR_OUT_OF_MEMORY out of memory to perform the broadcast
@return ALP_NOTIFY_ERR_INVALID_APPID invalid appID
@return ALP_NOTIFY_ERR_INVALID_NOTIFYTYPE invalid notifyType
@return ALP_NOTIFY_ERR_INVALID_LENGTH invalid length
@return ALP_NOTIFY_ERR_INVALID_QUEUE invalid queue
@return ALP_NOTIFY_ERR_TOO_MANY_BROADCASTS too many active broadcasts
@return ALP_NOTIFY_ERR_IPC ipc error

@remarks The default behavior is to broadcast each queue in a serial FIFO order. Using the ALP_NOTIFY_BROADCAST_FLAG place the broadcast at the front of the queue. Different broadcast queues are broadcast in parallel.

@remarks Use ALP_NOTIFY_BROADCAST_FLAG_SYNC to block until the all registered clients have been notified. Also see the alp_notify_broadcast_sync() API to block with a specified timeout or to get the handled flag. The default behavior is to not block.
*/
extern alp_status_t alp_notify_broadcast(const char * appID, const char * notifyType, const void * details, int length, int queue, int flags);

/*! broadcasts the specified notification synchronously
@param[in] appID application that broadcasts the notification the notification (must be non-null)
@param[in] notifyType the type of notification (card inserted, vfs mounted, etc) (must be non-null)
@param[in] details notification specific data
@param[in] length size of notification specific data
@param[in] queue broadcast queue (use ALP_NOTIFY_QUEUE_DEFAULT as the default) (must be non-zero)
@param[in] flags (0 for default and see ALP_NOTIFY_BROADCAST_IMMEDIATE to change the default behavior)
@param[in] timeout timeout in seconds, use 0 for infinite timeout
@param[out] handled value of the handled flag

@return ALP_OK success
@return ALP_NOTIFY_ERR_OUT_OF_MEMORY out of memory to perform the broadcast
@return ALP_NOTIFY_ERR_INVALID_APPID invalid appID
@return ALP_NOTIFY_ERR_INVALID_NOTIFYTYPE invalid notifyType
@return ALP_NOTIFY_ERR_INVALID_LENGTH invalid length
@return ALP_NOTIFY_ERR_INVALID_QUEUE invalid queue
@return ALP_NOTIFY_ERR_TOO_MANY_BROADCASTS too many active broadcasts
@return ALP_NOTIFY_ERR_IPC ipc error
@return ALP_NOTIFY_ERR_TIMEOUT timed out

@remarks The default behavior is to broadcast each queue in a serial FIFO order. Using the ALP_NOTIFY_BROADCAST_FLAG place the broadcast at the front of the queue.   Different broadcast queues are broadcast in parallel.
*/
extern alp_status_t alp_notify_broadcast_sync(const char * appID, const char * notifyType, const void * details, int length, int queue, int flags, int timeout, int * handled);

/*! notification handler completion routine for applications
@param[in] appID application that broadcasts the notification the notification (may be null)
@param[in] handled handled flag
@param[in] status status of the notification

@return ALP_OK success
@return ALP_NOTIFY_ERR_OUT_OF_MEMORY out of memory to perform the completion
@return ALP_NOTIFY_ERR_INVALID_APPID invalid appID
@return ALP_NOTIFY_ERR_IPC ipc error

@remarks appID maybe be null to specific the current running application.
*/
extern alp_status_t alp_notify_done(const char * appID, int handled, int status);

/*! decodes argv details parameter
@param[in] src argv encoded AlpNotifyLaunch launch data
@param[out] dst decoded AlpNotifyLaunch launch data

@return number of bytes 
*/
extern int alp_notify_details(const char * src, AlpNotifyLaunch * dst);

//! @}

#ifdef __cplusplus
}       // extern "C"
#endif

#endif
