/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/*!
 * @file appmgr_pdk.h
 * @brief Header file for MAX Application Manager PDK APIs.
 *  Contains the PDK API for the application server and its associated
 *  client library.
 *****************************************************************************/

#ifndef __ALP_APP_MGR_PDK_H_
#define __ALP_APP_MGR_PDK_H_

#include <hiker/appmgr.h>

#ifdef __cplusplus
extern "C" {
#endif

/*! @addtogroup ALP_APPLICATION_MGR Application Manager
 *  @ingroup ALP_SYSTEM
 *  @{
 */

 /************************************************
	PDK.  Available to licensees only.
 ************************************************/


// Define trace classes:
#define ALP_APP_SERVER_TRACE	ALP_TRACE_DEFINE(ALP_CLASS_APPSERVER, 1)
#define ALP_APP_CLIENT_TRACE	ALP_TRACE_DEFINE(ALP_CLASS_APPSERVER, 2)
#define ALP_APP_TEST_TRACE	ALP_TRACE_DEFINE(ALP_CLASS_APPSERVER, 3)

 
 
typedef void (*AlpDiagnosticHandler) (const char* appID, const char* diagStr, void* cbData);
typedef void (*AlpThemeUpdateProc) (void);

typedef struct _AlpAppContext AlpAppContext;

/// More flexible PDK version of alp_app_launch.
/*!
This API has all the options.  It allows the caller to specify the context 
from which the launch should be executed, and the blocking behavior can also 
be controlled.  If NULL is passed for outPID, then this call will not block 
to wait for a response from appserver, ala alp_app_launch_async() (keep in mind
that the result code will be somewhat less useful in this case).

@param[in] context The AlpAppContext* to use to send the launch request.
@param[in] appID The name of the bundle containing the application to be launched.
@param[in] argc The number of strings in the argv array.
@param[in] argv An array of char* containing launch arguments for the application.
@param[out] outPID a pointer to pid_t to receive the app's process ID once it registers with the app server, or NULL to not wait for a response.

@return Yields ALP_STATUS_OK on success, or an error code if there is a problem sending the message to the app server.  If outPID is non-null, then this API will wait for a response from the appserver, and will also return an error if there was a problem launching the application.
 */
alp_status_t alp_app_launch_full(AlpAppContext* context, const char * appID, int argc,  char *const argv[], pid_t *outPID);

/// Sets up a default connection to the application server, & registers the app.
/*!

This API is generally called on behalf of apps during launch by the launchpad 
just before beginning application execution.  Various system services (not true applications)
may also call it (along with alp_app_destroy()) in order to "pose" as an application and
take advantage of various application based services (e.g., calling alp_app_launch() and 
receiving relaunch callbacks and exit requests).  In the latter case, the "appID" parameter
is not required to represent a valid bundle.

This call initializes the 2-way connection with the app server, and causes the app to be 
added to the server's list of active applications.  Many APIs require this connection 
in order to function properly (alp_app_launch(), alp_app_exit(), alp_app_register(),
alp_app_flush(), etc.).  If there is no default app context, then the newly allocated 
context will become the default.

The launch flags are used to describe the newly launched application's mode of 
execution.  These are a combination of the ALP_APP_LAUNCH_FLAG_* constants
from appmgr_pdk.h, and usually come from alp_app_parse_launch_flags(), which parses 
the application's argv array and returns the appropriate set of bits.  Note that
these bits describe the newly launched application's mode of execution ONLY.  
They have no relation to any other apps which might have launched this one, 
or which the new one may launch (i.e., they are not propagated between apps).

@param[in] appID The name of the bundle containing the application.
@param[in] flags Standard launch flags indicating the app's mode of execution.
@param[in] scheme The bundle type for received launch requests.  Used only when registering as a launchpad.
@return Returns a newly allocated AlpAppContext structure encapsulating the appserver connection & associated app state.
 */
AlpAppContext* alp_app_setup(const char *appID, guint32 flags, const char* scheme);

/// Unregisters app & closes the connection opened by alp_app_setup().
/*!
This API is the reverse of alp_app_setup().  It 
unregisters the context with the appserver, closes the IPC connection, 
and deallocates the AlpAppContext structure.  
It's generally called by launchpads on behalf of the launched application,
when it terminates.  In order for app termination to be considered 'graceful' 
by the appserver, this API must be called to close the connection.

@param[in] context The app context for the terminating application.
@param[in] result The app's exit status (e.g., as returned from main()).
 
 */
void alp_app_destroy(AlpAppContext *context, int result);

/// Close IPC connection & deallocate context without unregistering with appserver.
/*!
This obscure API is used by the native launchpad to close an app context's 
IPC connection (and deallocate the memory) without unregistering with the
appserver.  This is needed when forking in order to prevent the child process
from sharing the parent's IPC connection.

@param[in] context The app context whose server connection is to be closed, or NULL for the default context.
 */
void alp_app_abandon(AlpAppContext* context);

/// Sets the default application context.
/*!
The default application context is used internally by many of the app-level SDK APIs to 
encapsulate the connection to the app server, callback handlers for various app events, etc.

@param[in] context The new default application context.
 */
void alp_app_set_default_context(AlpAppContext* context);


/// Re-register a relaunched application with the appserver.
/*!
This API should only be called in response to a launch request from the app server (either
an initial launch, or a relaunch).  If it is called otherwise, it could result in pending 
launch requests being accidentally dropped from the appserver's internal queue (since they
are normally removed when the app registration is received).  Since the appserver client 
libraries pretty much take care of this, use of this API is mostly limited to testing.

This API is generally called (via alp_app_setup) on behalf of a newly launched application by 
the launchpad's setup code before calling into the application's main() entrypoint.  It is also 
called by the client-side libappserver code directly in the case of an application relaunch.
It must be called from the context of the registering application.

@param[in] context The app context to be registered, obtained by calling alp_app_setup().
@param[in] appID The name of the bundle containing the application.
@param[in] flags Standard launch flags indicating the app's mode of execution.
@param[in] scheme The bundle type for received launch requests.  Used only when registering as a launchpad.
 */
void alp_app_register(AlpAppContext *context, const char *appID, guint32 flags, const char* scheme);

/// Ensure appserver has handled any pending messages for the given context.
/*!
This API is used by unit tests to help with synchronization.  It sends a synchronous no-op message to the appserver, and waits for a reply.  This ensures that the appserver has received and handled any pending messages for the given context.

@param[in] context The app context to be registered.
@return Returns an alp_status_t indicating success or failure.
 */
alp_status_t alp_app_flush(AlpAppContext* context);


/// Parse default ALP launch arguments and return a bit flag representation.
/*!
This utility routine is used by launchpads, the app server, and the client-side code to 
parse the standard arguments passed to the app and set appropriate launch flags in the result,
which provide a more compact and easily accessible representation.  

@param[in] launchFlags The current launch flags.  Needed (for example) when neither ALP_APP_PRIMARY or ALP_APP_BACKGROUNDED is specified. 
@param[in] argc The number of strings in the argv array.
@param[in] argv An array of char* containing launch arguments for the application.

@return Returns a bitfield representing some of the standard parameters in the argv array.
 */
guint32 alp_app_parse_launch_flags(guint32 launchFlags, int argc, char *const argv[]);

/// Set a handler to receive diagnostic messages from appserver.
/*!
This API is used by diagnostic code to set a handler for diagnostic messages received from the app server.
Applications register to receive these messages by setting the ALP_APP_LAUNCH_FLAG_DIAGNOSTIC flag when registering.
Then, when they arrive, the client-side libappserver code will call this handler.

@param[in] context The app context for which to set the diagnostic handler.
@param[in] procP A pointer to the diagnostic handler to call.
@param[in] cbData A caller-defined pointer which is passed to the handler.
 */
void alp_app_set_diagnostic_handler(AlpAppContext *context, AlpDiagnosticHandler procP, void* cbData);


// Jump List.  
// The app server tracks jump list apps, and these APIs are used to access it.
// Currently there are no APIs to manipulate the jump list.

/// Get the current list of applications in the "jump list".
/*!
This call sends a request through the default application context, and returns the current 
contents of the Jump List.  The result is an array of char*, where each string is the bundle
ID of an application in the current jump list.  The 'isRunnings' array contains boolean values
indicating whether or not each of the apps in 'appIDs' was running when the data was packed up.
The result lists are variable length,
and sorted in order of launch (most recently launched first).  The jumplist application 
is expected to sort through the data and decide what is the appropriate subset to display.

Note that the jump list will not necessarily contain *all* running apps.  Only those that have not 
fallen off the list since being launched with ALP_APP_PRIMARY.  Apps that are never launched as primary
are not added to the jump list unless they specify "always" for the jump-list property in their manifest.

The result strings and arrays should be all be freed 
by the caller using g_free() (or alp_app_free_argv()).

@param[out] count Receives the number of elements in appIDs and isRunnings.
@param[out] appIDs Receives an array of char*, containing bundle IDs for the jump list apps.
@param[out] isRunnings Receives an array of gboolean, indicating whether or not each app is running.

@return Yields ALP_STATUS_OK on success, or an error code if there is a problem with the request.
 */
alp_status_t alp_app_get_jump_list(int *count, char***appIDs, gboolean **isRunnings);



// Miscellaneous utility routines.


/// Deallocate a dynamically allocated argc/argv style argument list.
/*!
Ignores null string pointers, so the caller can zero out any argv[] elements that should not be deallocated.

@param[in] argc The number of strings in the argv array.
@param[in] argv An array of char* containing launch arguments for the application.
 */
void alp_app_free_argv(int argc, char* argv[]);


/// Halt execution, fork a new gnome-terminal, and attach gdb to the app.
/*!
	This API is a vague equivalent of the old PalmOS DbgBreak().
	It's intended to make it slightly easier to debug apps until we 
	get the real debugging infrastructure in place.  Once that happens, 
	this API will probably disappear.
	
	This call will create a child process in which it opens a new terminal, runs GDB, 
	and attaches to the parent app.  Meanwhile, the parent sleeps for a few seconds
	to give the debugger a chance to catch up.
 */
void alp_app_debug_break(void);


/// Set the procedure used to update the app's theme when a theme update message is received.
/*!
	This API is necessary in order to prevent appserver from having to link to GTK+
	in order to update the theme.  The alternative would mean pulling GTK+ into many 
	system daemons which do not need it.  This routine is used by alp_max_gtk_init()
	to make sure that the proper theme update routine is called for MAX apps (and
	any other UI which uses GTK+ and calls alp_max_gtk_init().

@param[in] context The app context for which to set the theme update procedure. Pass NULL for the default context.
@param[in] proc A pointer to the prcedure to call when a theme update message is received.
 */
void alp_app_set_theme_update_proc(AlpAppContext *context, AlpThemeUpdateProc proc);

/// Obtain the "security ID" for the app most recently launched through this context.
/*!
	Intended for use by launchpads, specifically NLPD, so that it can fork a process & set security parameters 
	without having to create an extra thread before lowering priveleges.  This API is subject to change.

@param[in] context The app context from which to get the security ID.
@return Yields a security ID on success, or ALP_APP_DEFAULT_SECURITY_ID on failure.
 */
unsigned int alp_app_get_security_id(AlpAppContext *context);


#define ALP_APP_DEFAULT_SECURITY_ID	0xFFFFFFFF


#define ALP_APP_LAUNCHPAD_SCHEME_SIZE	16	// max bytes for a launchpad scheme.

/*
The launch flags are passed to alp_app_setup() and alp_app_register(), and are
used to describe an application's mode of execution.  They are usually obtained
from alp_app_parse_launch_flags(), which parses the application's argv array and 
returns the appropriate set of bits.  Note that these bits describe the application's
mode of execution ONLY.  They have no relation to any other apps which might have
launched this one, or which this one may launch.
 */
typedef enum
{
	ALP_APP_LAUNCH_FLAG_NONE		= 0,
	ALP_APP_LAUNCH_FLAG_DEBUG		= (1 << 0),	// app is being debugged.
	ALP_APP_LAUNCH_FLAG_PRIMARY	= (1 << 1),	// app is the primary "main ui" app
	ALP_APP_LAUNCH_FLAG_LAUNCHPAD	= (1 << 2),	// app is a launchpad/cookiecutter
	ALP_APP_LAUNCH_FLAG_DIAGNOSTIC	= (1 << 3),	// app should be sent diagnostic messages.
	ALP_APP_LAUNCH_FLAG_TRANSIENT	= (1 << 4),	// app should be closed when main app is switched.  
										// Causes transient server flag to be set.
										// DOLATER - this should probably be moved to server flags & driven 
										// from a manifest property when possible.  
	ALP_APP_LAUNCH_FLAG_ERROR	= (1 << 5)	// launchpad encountered an error. 
} AlpAppLaunchFlags;

/// These launch flags should always remain set for an 
#define ALP_APP_LAUNCH_FLAGS_STICKY	(ALP_APP_LAUNCH_FLAG_TRANSIENT | ALP_APP_LAUNCH_FLAG_DEBUG | ALP_APP_LAUNCH_FLAG_LAUNCHPAD)

/// This represents the default launchflags for an app.  Background run, no debugging, diagnostics, etc.
#define ALP_APP_LAUNCH_FLAGS_DEFAULT (ALP_APP_LAUNCH_FLAG_NONE)


// Diagnostic messages:
#define ALP_APP_DIAG_UNEXPECTED_EXIT "Application quit unexpectedly."
#define ALP_APP_DIAG_CLEAN_EXIT "Application quit cleanly."
#define ALP_APP_DIAG_MISSING_LAUNCPAD "Can't find launchpad for scheme, not launching."



// Appserver specific launch arguments:
#define ALP_APP_NO_DEFAULT_APP		"--no-default-app"
#define ALP_APP_NO_DEFAULTS			"--no-defaults"
#define ALP_APP_NO_DAEMON			"--no-daemon"
#define ALP_APP_DEFAULT_PRIMARY		"--default-primary"



//! @}
//! @}

#ifdef __cplusplus
}	// extern "C"
#endif

#endif // __ALP_APP_MGR_PDK_H_


