/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/**
 * @file apas.h
 *
 * @brief Contains helper functions for the Adjunct Package Attestation System
 *
 */

#ifndef _H_ALP_APAS
#define _H_ALP_APAS

#include <hiker/types.h>
#include <sys/types.h>
#include <hiker/sysclass.h>

#ifdef __cplusplus
extern "C" {
#endif

  /* Macro Definitions */

  /*! @brief Adjunct Package Attestation System constants */

  /*! error code : no error was detected. success. w00t!       */
#define ALP_STATUS_SPF_APAS_E_NOERR                  ( ALP_CLASS_SPF | 0x00 )

  /*! error code : a required parameter was set to null        */
#define ALP_STATUS_SPF_APAS_E_NULL_PARAM             ( ALP_CLASS_SPF | 0x01 )

  /*! error code : generic file error, see errno for more info */
#define ALP_STATUS_SPF_APAS_E_FILE                   ( ALP_CLASS_SPF | 0x02 )

  /*! error code : memory allocation error.                    */
#define ALP_STATUS_SPF_APAS_E_MALLOC                 ( ALP_CLASS_SPF | 0x03 )

  /*! error code : GID specified for new policy is in use */
#define ALP_STATUS_SPF_APAS_E_GID_IN_USE             ( ALP_CLASS_SPF | 0x04 )

  /*! error code : ID specified for new package is in use */
#define ALP_STATUS_SPF_APAS_E_ID_IN_USE              ( ALP_CLASS_SPF | 0x05 )

  /*! error code : UID specified for new package is in use */
#define ALP_STATUS_SPF_APAS_E_UID_IN_USE             ( ALP_CLASS_SPF | 0x06 )

  /*! error code : function unimplemented */
#define ALP_STATUS_SPF_APAS_E_UNIMPLEMENTED          ( ALP_CLASS_SPF | 0x07 )

  /*! error code : assertion unknown */
#define ALP_STATUS_SPF_APAS_E_UNKNOWN_ASSERTION      ( ALP_CLASS_SPF | 0x08 )

  /*! error code : policy unknown */
#define ALP_STATUS_SPF_APAS_E_UNKNOWN_POLICY         ( ALP_CLASS_SPF | 0x09 )

  /*! error code : package unknown */
#define ALP_STATUS_SPF_APAS_E_UNKNOWN_PACKAGE        ( ALP_CLASS_SPF | 0x0A )

  /*! error code : gid unknown */
#define ALP_STATUS_SPF_APAS_E_UNKNOWN_GID            ( ALP_CLASS_SPF | 0x0B )

  /*! error code : UI queue is full */
#define ALP_STATUS_SPF_APAS_E_UNKNOWN_FULL_QUEUE     ( ALP_CLASS_SPF | 0x0C )

  /*! error code : UI queue entry not found */
#define ALP_STATUS_SPF_APAS_E_NOT_FOUND              ( ALP_CLASS_SPF | 0x0D )

  /*! error code : unspecified error */
#define ALP_STATUS_SPF_APAS_E_UNSPECIFIED            ( ALP_CLASS_SPF | 0x0E )

  /*!
   * @brief This function "enrolls" a package into the appropriate security
   *        policy domain. It finds the correct policy by examining the
   *        certificate(s) used to sign the package and (assuming the
   *        certificates are valid) checks to see which policy the certificate
   *        is associated with (you should be able to find this information by
   *        examining the 'certref' elements in the 'policy' elements in
   *        /etc/default.policy). Upon completion, the packageid assigned to
   *        the package is written into the unsigned integer pointed to by
   *        out_packageid. Note that if the APAS system finds the package's
   *        metadata already in the system, it will return the packageid in
   *        the metadata instead of assigning a new one. It will also
   *        increment the value of the package metadata reference count by one.
   *
   * @param[in] path           the file path to the directory that contains
   *                           the security.xml file for the package
   * @param[out] out_packageid a pointer to an unsigned int into which the
   *                           call will write the package id
   *
   * @return ALP_STATUS_SPF_APAS_E_NOERR      success.
   * @return ALP_STATUS_SPF_APAS_E_NULL_PARAM either the path or out_packageid
   *                                          parameters were NULL.
   * @return ALP_STATUS_SPF_APAS_E_FILE       there was a problem opening the
   *                                          security.xml manifest for the
   *                                          package given at the specified
   *                                          path.
   * @return ALP_STATUS_SPF_APAS_E_MALLOC     there was an interim allocation
   *                                          error. this is very bad.
   *
   * @par Example
   * An example of enrolling a package with the alp_spf_apas_enroll() call.
   * @code
   * #define APP_E_NOERR 0
   * #define APP_E_USAGE 1
   * #define APP_E_OTHER 2
   * 
   * #include <stdio.h>
   * #include "hiker/prv/apas.h"
   * 
   * typedef unsigned int Err;
   * typedef struct {
   * char *path;
   * } App;
   * 
   * static Err cli_parse( App *app, int argc, char **argv );
   * static void usage_message_emit( char *command );
   * 
   * int main( int argc, char *argv[] ) {
   * Err err = APP_E_NOERR;
   * App app;
   * unsigned int packageid;
   * alp_status_t status;
   * 
   * do {
   * if( APP_E_NOERR != ( err = cli_parse( &app, argc, argv ) ) ) break;
   * 
   * if( ALP_STATUS_SPF_APAS_E_NOERR != (status = alp_spf_apas_enroll( app.path, &packageid ) ) ){
   * err = APP_E_OTHER;
   * printf("%%APAST-F-ENROLL; Error %08X while trying to enroll package %s.\n", (unsigned int) status, app.path );
   * break;
   * }
   * 
   * printf("%%APAST-S-ENROLL; The package at %s was successfully installed.\n", app.path );
   * 
   * } while( 0 );
   * 
   * if( APP_E_USAGE == err ) {
   * usage_message_emit(argv[0]);
   * }
   * 
   * return( err & 0xFF );
   * }
   * 
   * static Err cli_parse( App *app, int argc, char **argv ) {
   * Err err = APP_E_NOERR;
   * 
   * do {
   * if( argc < 2 ) {
   * err = APP_E_USAGE;
   * break;
   * }
   * 
   * app->path = argv[1];
   * 
   * } while( 0 );
   * 
   * return( err );
   * }
   * 
   * static void usage_message_emit( char *command ) {
   * printf("%%APAST-E-USAGE; Usage is:\n; %s <path>\n; path - path to a directory where we can find a security\n; manifest (security.xml) file.\n", command );
   * }
   * 
   */

  alp_status_t alp_spf_apas_enroll( char *path, unsigned int *out_packageid );

  /*!
   * @brief After a package has been enrolled, if you need to figure out what
   *        its packageid is based on it's security manifest, use this
   *        function. Given a path to a directory, it will look in that
   *        directory (and every directory above it) until it finds a file
   *        called security.xml. This will be the security manifest for the
   *        package. The manifest will be hashed and compared to all the hashes
   *        of the enrolled packages. When it finds a hit, we return the
   *        package id.
   *
   * @param[in] path           the file path to the directory that contains
   *                           the security.xml file for the package
   * @param[out] out_packageid a pointer to an unsigned int into which the
   *                           call will write the package id
   *
   * @return ALP_STATUS_SPF_APAS_E_NOERR      success.
   * @return ALP_STATUS_SPF_APAS_E_NULL_PARAM either the path or out_packageid
   *                                          parameters were NULL.
   * @return ALP_STATUS_SPF_APAS_E_FILE       there was a problem opening the
   *                                          security.xml manifest for the
   *                                          package given at the specified
   *                                          path.
   * @return ALP_STATUS_SPF_APAS_E_MALLOC     there was an interim allocation
   *                                          error. this is very bad.
   * @par Example
   * An example of getting a package's packageid from it's path with the
   * alp_spf_apas_path_get_packageid() call.
   * @code
   * #define APP_E_NOERR 0
   * #define APP_E_USAGE 1
   * #define APP_E_OTHER 2
   * 
   * #include <stdio.h>
   * #include "hiker/prv/apas.h"
   * 
   * typedef unsigned int Err;
   * typedef struct {
   * char *path;
   * } App;
   * 
   * static Err cli_parse( App *app, int argc, char **argv );
   * static void usage_message_emit( char *command );
   * 
   * int main( int argc, char *argv[] ) {
   * Err err = APP_E_NOERR;
   * App app;
   * unsigned int packageid;
   * alp_status_t status;
   * 
   * do {
   * if( APP_E_NOERR != ( err = cli_parse( &app, argc, argv ) ) ) break;
   * 
   * if( ALP_STATUS_SPF_APAS_E_NOERR != (status = alp_spf_apas_path_get_packageid( app.path, &packageid ) ) ){
   * err = APP_E_OTHER;
   * printf("%%APAST-F-PATHGP; Error %08X while trying to enroll package %s.\n", (unsigned int) status, app.path );
   * break;
   * }
   * 
   * printf("%%APAST-S-PATHGP; The package at %s is installed as package %u.\n", app.path, packageid );
   * 
   * } while( 0 );
   * 
   * if( APP_E_USAGE == err ) {
   * usage_message_emit(argv[0]);
   * }
   * 
   * return( err & 0xFF );
   * }
   * 
   * static Err cli_parse( App *app, int argc, char **argv ) {
   * Err err = APP_E_NOERR;
   * 
   * do {
   * if( argc < 2 ) {
   * err = APP_E_USAGE;
   * break;
   * }
   * 
   * app->path = argv[1];
   * 
   * } while( 0 );
   * 
   * return( err );
   * }
   * 
   * static void usage_message_emit( char *command ) {
   * printf("%%APAST-E-USAGE; Usage is:\n; %s <path>\n; path - path to a directory where we can find a security\n; manifest (security.xml) file.\n", command );
   * }
   */

  alp_status_t alp_spf_apas_path_get_packageid( char *path, unsigned int *out_packageid );

  /*!
   * @brief This function removes the package metadata the system (it does
   *        not remove the package itself.) Note however that this call will
   *        first decrement the reference count for the package metadata. If
   *        the reference count is not zero after being decrementing, the
   *        metadata will not be removed. If out_refcount is non-NULL, it the
   *        system will write the reference count (after decrementing) into
   *        the unsigned int pointed to by out_refcount.
   *
   * @param[in] packageid     the package id of the package whose metadata
   *                          you would like to remove
   * @param[out] out_refcount If this parameter is non-NULL, the unsigned int
   *                          to which it points will be overwritten with the
   *                          reference count after it is decremented.
   *
   * @return ALP_STATUS_SPF_APAS_E_NOERR  no errors. success. w00t!
   * @return ALP_STATUS_SPF_APAS_E_UNKNOWN_PACKAGE You provided a package id
   *                                      that is unknown to the system.
   * @return ALP_STATUS_SPF_APAS_E_MALLOC There was an interim allocation
   *                                      error.
   * @par Example
   * An example of removing a package's metadata with the
   * alp_spf_apas_packageid_remove() call.
   * @code
   * #define APP_E_NOERR 0
   * #define APP_E_USAGE 1
   * #define APP_E_OTHER 2
   * 
   * #include <stdio.h>
   * #include <stdlib.h>
   * #include <string.h>
   * #include "hiker/prv/apas.h"
   * 
   * typedef unsigned int Err;
   * typedef struct {
   * unsigned int packageid;
   * unsigned int tell_refcount;
   * } App;
   * 
   * static Err cli_parse( App *app, int argc, char **argv );
   * static void usage_message_emit( char *command );
   * 
   * int main( int argc, char *argv[] ) {
   * Err err = APP_E_NOERR;
   * App app;
   * alp_status_t status;
   * unsigned int refcount;
   * unsigned int *out_refcount;
   * do {
   * if( APP_E_NOERR != ( err = cli_parse( &app, argc, argv ) ) ) break;
   * 
   * if( 0 != app.tell_refcount ) {
   * out_refcount = &refcount;
   * } else {
   * out_refcount = NULL;
   * }
   * 
   * if( ALP_STATUS_SPF_APAS_E_NOERR != (status = alp_spf_apas_packageid_remove(app.packageid, out_refcount ) ) ) {
   * err = APP_E_OTHER;
   * printf("%%APAST-F-REMOVE; Error %08X while trying to remove package %u / %08X.\n", (unsigned int) status, app.packageid, (unsigned int) out_refcount );
   * break;
   * }
   * 
   * printf("%%APAST-S-REMOVE; Package %u was successfully removed.\n", app.packageid );
   * 
   * if( 0 != app.tell_refcount ) {
   * printf("%%APAST-I-REFCOUNT; The reference count is now %u.\n", refcount );
   * }
   * 
   * } while( 0 );
   * 
   * if( APP_E_USAGE == err ) {
   * usage_message_emit(argv[0]);
   * }
   * 
   * return( err & 0xFF );
   * }
   * 
   * static Err cli_parse( App *app, int argc, char **argv ) {
   * Err err = APP_E_NOERR;
   * 
   * do {
   * if( argc < 2 ) {
   * err = APP_E_USAGE;
   * break;
   * }
   * 
   * app->packageid = atoi( argv[1] );
   * 
   * if( argc > 2 ) {
   * if( !strncmp( "-t", argv[2], 2 ) ) {
   * app->tell_refcount = 1;
   * } else {
   * err = APP_E_USAGE;
   * break;
   * }
   * } else {
   * app->tell_refcount = 0;
   * }
   * 
   * } while( 0 );
   * 
   * return( err );
   * }
   * 
   * static void usage_message_emit( char *command ) {
   * printf("%%APAST-E-USAGE; Usage is:\n; %s <packageid> [-t]\n; <packageid> - the id of the package to remove\n; -t (optional) if specified, it will print the post-removal refcount.\n", command );
   * }
   */

  alp_status_t alp_spf_apas_packageid_remove( unsigned int packageid, unsigned int *out_refcount );

  /*!
   * @brief This function returns the uid associated with a package id. The
   *        SPF/APAS system assumes that each package has its own UID.
   *
   * @param[in] packageid     the package id of the package in which you're
   *                          interested
   * @param[out] out_uid      a pointer to an uid_t which will be overwritten
   *                          with the package's UID.
   *
   * @return ALP_STATUS_SPF_APAS_E_NOERR  no errors. success. w00t!
   * @return ALP_STATUS_SPF_APAS_E_NULL_PARAM The pointer to the uid_t provided
   *                                      was NULL.
   * @return ALP_STATUS_SPF_APAS_E_UNKNOWN_PACKAGE You provided a package id
   *                                      that is unknown to the system.
   * @return ALP_STATUS_SPF_APAS_E_MALLOC There was an interim allocation
   *                                      error.
   * @par Example
   * An example program that gets the uid, gid and refcount associated with a 
   * packageid.
   * @code
   * #define APP_E_NOERR 0
   * #define APP_E_USAGE 1
   * #define APP_E_OTHER 2
   * 
   * #include <stdio.h>
   * #include <stdlib.h>
   * #include <sys/types.h>
   * #include "hiker/prv/apas.h"
   * 
   * typedef unsigned int Err;
   * typedef struct {
   * unsigned int packageid;
   * } App;
   * 
   * static Err cli_parse( App *app, int argc, char **argv );
   * static void usage_message_emit( char *command );
   * 
   * int main( int argc, char *argv[] ) {
   * Err err = APP_E_NOERR;
   * App app;
   * alp_status_t status;
   * unsigned int refcount;
   * uid_t uid;
   * gid_t gid;
   * 
   * do {
   * if( APP_E_NOERR != ( err = cli_parse( &app, argc, argv ) ) ) break;
   * 
   * if( ALP_STATUS_SPF_APAS_E_NOERR != (status = alp_spf_apas_packageid_refcount_get( app.packageid, &refcount ) ) ){
   * err = APP_E_OTHER;
   * printf("%%APAST-F-PIDRCQ; Error %08X while trying to get the refcount for\n; package %u.\n", (unsigned int) status, app.packageid );
   * break;
   * }
   * 
   * printf("%%APAST-S-PIDRCQ; The refcount for packageid %u is %u.\n", app.packageid, refcount );
   * 
   * if( ALP_STATUS_SPF_APAS_E_NOERR != (status = alp_spf_apas_packageid_gid_get( app.packageid, &gid ) ) ){
   * err = APP_E_OTHER;
   * printf("%%APAST-F-PIDRCQ; Error %08X while trying to get the gid for package %u.\n", (unsigned int) status, app.packageid );
   * break;
   * }
   * 
   * printf("%%APAST-S-PIDRCQG; The gid for packageid %u is %u.\n", app.packageid, (unsigned int) gid );
   * 
   * if( ALP_STATUS_SPF_APAS_E_NOERR != (status = alp_spf_apas_packageid_uid_get( app.packageid, &gid ) ) ){
   * err = APP_E_OTHER;
   * printf("%%APAST-F-PIDRCQ; Error %08X while trying to get the uid for package %u.\n", (unsigned int) status, app.packageid );
   * break;
   * }
   * 
   * printf("%%APAST-S-PIDRCQG; The uid for packageid %u is %u.\n", app.packageid, (unsigned int) uid );
   * 
   * } while( 0 );
   * 
   * if( APP_E_USAGE == err ) {
   * usage_message_emit(argv[0]);
   * }
   * 
   * return( err & 0xFF );
   * }
   * 
   * static Err cli_parse( App *app, int argc, char **argv ) {
   * Err err = APP_E_NOERR;
   * 
   * do {
   * if( argc < 2 ) {
   * err = APP_E_USAGE;
   * break;
   * }
   * 
   * app->packageid = atoi( argv[1] );
   * 
   * } while( 0 );
   * 
   * return( err );
   * }
   * 
   * static void usage_message_emit( char *command ) {
   * printf("%%APAST-E-USAGE; Usage is:\n; %s <packageid>\n; packagid - the id of the package whose refcount will be queried.\n", command );
   * }
   */

  alp_status_t alp_spf_apas_packageid_uid_get( unsigned int packageid, uid_t *out_uid);

  /*!
   * @brief This function returns the gid associated with a package id. The
   *        SPF/APAS system assumes that each package is associated with a
   *        policyid through which it determines its GID.
   *
   * @param[in] packageid     the package id of the package in which you're
   *                          interested
   * @param[out] out_gid      a pointer to an uid_t which will be overwritten
   *                          with the package's GID.
   *
   * @return ALP_STATUS_SPF_APAS_E_NOERR  no errors. success. w00t!
   * @return ALP_STATUS_SPF_APAS_E_NULL_PARAM The pointer to the gid_t provided
   *                                      was NULL.
   * @return ALP_STATUS_SPF_APAS_E_UNKNOWN_PACKAGE You provided a package id
   *                                      that is unknown to the system.
   * @return ALP_STATUS_SPF_APAS_E_MALLOC There was an interim allocation
   *                                      error.
   */

  alp_status_t alp_spf_apas_packageid_gid_get( unsigned int packageid, gid_t *out_gid );

  /*!
   * @brief This function returns the sgids associated with a package id. The
   *        SPF/APAS system assumes that each package is associated with a
   *        policyid through which it determines its SGIDs.
   *
   * @param[in] packageid     the package id of the package in which you're
   *                          interested
   * @param[out] out_sgids    a pointer to an array of gid_t's representing the
   *                          sgids. The array is null-terminated
   * @param[out] out_sgids_length  the length of the out_sgids array
   *
   * @return ALP_STATUS_SPF_APAS_E_NOERR  no errors. success. w00t!
   * @return ALP_STATUS_SPF_APAS_E_NULL_PARAM The pointer to the gid_t provided
   *                                      was NULL.
   * @return ALP_STATUS_SPF_APAS_E_UNKNOWN_PACKAGE You provided a package id
   *                                      that is unknown to the system.
   * @return ALP_STATUS_SPF_APAS_E_MALLOC There was an interim allocation
   *                                      error.
   */

  alp_status_t alp_spf_apas_packageid_sgids_get( unsigned int packageid, gid_t *out_sgids[], unsigned int *out_sgids_length );


  /*!
   * @brief This function returns the reference count associated with a
   *        package id. 
   *
   * @param[in] packageid     the package id of the package in which you're
   *                          interested
   * @param[out] out_refcount a pointer to an unsigned int which will be
   *                          overwritten with the package's reference count.
   *
   * @return ALP_STATUS_SPF_APAS_E_NOERR  no errors. success. w00t!
   * @return ALP_STATUS_SPF_APAS_E_NULL_PARAM The pointer to the unsigned int
   *                                      provided was NULL.
   * @return ALP_STATUS_SPF_APAS_E_UNKNOWN_PACKAGE You provided a package id
   *                                      that is unknown to the system.
   * @return ALP_STATUS_SPF_APAS_E_MALLOC There was an interim allocation
   *                                      error.
   */

  alp_status_t alp_spf_apas_packageid_refcount_get( unsigned int packageid, unsigned int *out_refcount );

  /*!
   * @brief This function sets the reference count associated with a package
   *        id. 
   *
   * @param[in] packageid     the package id of the package in which you're
   *                          interested
   * @param[in] refcount      an unsigned int representing the value you want
   *                          the package id's refcount to be set to
   *
   * @return ALP_STATUS_SPF_APAS_E_NOERR  no errors. success. w00t!
   * @return ALP_STATUS_SPF_APAS_E_UNKNOWN_PACKAGE You provided a package id
   *                                      that is unknown to the system.
   * @return ALP_STATUS_SPF_APAS_E_MALLOC There was an interim allocation
   *                                      error.
   */

  alp_status_t alp_spf_apas_packageid_refcount_set( unsigned int packageid, unsigned int refcount );

  /*!
   * @brief This function increments the reference count associated with a
   *        package id. 
   *
   * @param[in] packageid     the package id of the package in which you're
   *                          interested
   *
   * @return ALP_STATUS_SPF_APAS_E_NOERR  no errors. success. w00t!
   * @return ALP_STATUS_SPF_APAS_E_UNKNOWN_PACKAGE You provided a package id
   *                                      that is unknown to the system.
   * @return ALP_STATUS_SPF_APAS_E_MALLOC There was an interim allocation
   *                                      error.
   */

  alp_status_t alp_spf_apas_packageid_refcount_increment( unsigned int packageid );

  /*!
   * @brief This function decrements the reference count associated with a
   *        package id. 
   *
   * @param[in] packageid     the package id of the package in which you're
   *                          interested
   *
   * @return ALP_STATUS_SPF_APAS_E_NOERR  no errors. success. w00t!
   * @return ALP_STATUS_SPF_APAS_E_UNKNOWN_PACKAGE You provided a package id
   *                                      that is unknown to the system.
   * @return ALP_STATUS_SPF_APAS_E_MALLOC There was an interim allocation
   *                                      error.
   */

  alp_status_t alp_spf_apas_packageid_refcount_decrement( unsigned int packageid );

  /*!
   * @brief This function returns the default package id used for non-signed
   *        apps
   *
   * @param[out] out_packageid  a pointer to an unsigned integer that will
   *                            receive the default package id.
   *
   * @return ALP_STATUS_SPF_APAS_E_NOERR  no errors. success.
   * @return ALP_STATUS_SPF_APAS_E_MALLOC There was an interim allocation
   *                                      error.
   */

  alp_status_t alp_spf_apas_packageid_default( unsigned int *out_packageid );

#ifdef __cplusplus
}
#endif

#endif
