/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/**
 * @file attnmgr_dml.h
 *
 * @brief Contains the SDK & PDK API for the ALP Attention DML.
 *
 */

#ifndef __ALP_ATTN_DML_H__
#define __ALP_ATTN_DML_H__

#include <time.h>

#include <hiker/types.h>
#include <hiker/sysclass.h>
#include <hiker/attnmgr_common.h>
#include <hiker/sal.h>

/**
 * @addtogroup ALP_ATTN_MGR_PRIVATE Private
 * @ingroup ALP_ATTN_MGR
 */
/**
 * @defgroup ALP_ATTN_MGR_DML Data Model Definitions and API
 * @ingroup ALP_ATTN_MGR_PRIVATE
 */
/* @{ */

/**
 * @name Error code definitions for attention data model
 * @{
 */
#define ALP_STATUS_ATTN_DML_BAD_PARAM       (ALP_STATUS_ATTN_DML | 0x00000100)
#define ALP_STATUS_ATTN_DML_NO_MEMORY       (ALP_STATUS_ATTN_DML | 0x00000200)
#define ALP_STATUS_ATTN_DML_INTERNAL_ERROR  (ALP_STATUS_ATTN_DML | 0x00000300)
#define ALP_STATUS_ATTN_DML_NO_RECORDS      (ALP_STATUS_ATTN_DML | 0x00000400)
#define ALP_STATUS_ATTN_DML_INVALID_LUID    (ALP_STATUS_ATTN_DML | 0x00000500)
#define ALP_STATUS_ATTN_DML_DATABASE_ERROR  (ALP_STATUS_ATTN_DML | 0x00000600)
#define ALP_STATUS_ATTN_DML_DUPLICATE       (ALP_STATUS_ATTN_DML | 0x00000700)
#define ALP_STATUS_ATTN_DML_CORRUPT         (ALP_STATUS_ATTN_DML | 0x00000800)
#define ALP_STATUS_ATTN_DML_EXISTS          (ALP_STATUS_ATTN_DML | 0x00000900)
/* @} */

/*! Used for alp_attn_dml_item_create() */
typedef enum {
    InvalidItem      = 0,
    AlertItem        = 1,
    PropertyItem     = 2,
    PropertyListItem = 3,
} AlpAttnDmlItemType;

/**
 * The enum below lists all the fields that can be get and set in any of the
 * AlpContactsDmlItemType's above. Note the enum value is used in the names of the edit
 * dialog items (quasi data-driven programming with no meta info avail)
 */
typedef enum {
    Source          = 1,   //!< String - Used in: Alerts - Cannot be NULL
    Name            = 2,   //!< String - Used in: Alerts, Properties - Cannot be NULL
    Handle          = 3,   //!< String - Used in: Alerts - Cannot be NULL
    Interface       = 4,   //!< String - Used in: Alerts
    Priority        = 5,   //!< Int32 - Used in: Alerts
    Duration        = 6,   //!< Int32 - Used in: Alerts
    Pending         = 7,   //!< Bool - Used in: Alerts
    Attributes      = 8,   //!< Bool - Used in: Alerts
    AlertLuid       = 9,   //!< Uint32 - Used in: Alerts, Properties
    Value           = 10,  //!< String - Used in: Properties
    Property        = 11,  //!< PropertyItem - Used in: Alerts (pseudo)
    PropLuid        = 12,  //!< Uint32 - Used in: Properties
    UserId          = 13,  //!< Int32 - Used in: Alerts
    GroupId         = 14,  //!< Int32 - Used in: Alerts
    Created         = 15,  //!< Int32 - Used in: Alerts
    Updated         = 16,  //!< Int32 - Used in: Alerts
    PropType        = 17,  //!< Uint32 - Used in: Props
} AlpAttnDmlColumnType;

/** @brief Used by alp_attn_dml_alerts_statement_begin() */
typedef enum {
    OrderByNone,        //!< Do not specify a sort order
    OrderByPriority,    //!< Order query by priority, highest first
    OrderByTimestamp,   //!< Order query by time, oldest first
    NumOrderByAttnTypes //!< Order query by alert type
} AlpAttnDmlOrderByType;

#ifdef __cplusplus
extern "C" {
#endif

    /**
     * @name Attention DML DataModel API
     * @{
     */
    /* Attention DML functions */
    alp_status_t  alp_attn_dml_open  (AlpDmlH *dmlH);
    alp_status_t  alp_attn_dml_close (AlpDmlH dmlH);

    alp_status_t  alp_attn_dml_set_filename (const char *directory,
                                            const char *filename);
    /**
     * @brief Return db pathname
     * 
     * @return char * containing pointing to the database pathname.
     *                Do not free() this string.
     */
    const char   *alp_attn_dml_get_filename (void);

    /**
     * @brief Check database integrity
     *
     * @param[in] dmlH Database DML handle
     * @return    ALP_STATUS_OK                integrity check succeeded
     * @return    ALP_STATUS_ATTN_DML_CORRUPT  integrity check failed
     */
    alp_status_t  alp_attn_dml_check_database (AlpDmlH dmlH);
    /*@}*/

    /**
     * @name Attention DML Item API
     * @{
     */
    alp_status_t alp_attn_dml_item_create     (AlpAttnDmlItemType itemType,
                                               AlpDmlItemH *itemH);

    alp_status_t alp_attn_dml_item_destroy    (AlpDmlItemH  itemH);

    alp_status_t alp_attn_dml_item_set_int32  (AlpDmlItemH  itemH,
                                               AlpAttnDmlColumnType column,
                                               int32_t     *value);

    alp_status_t alp_attn_dml_item_set_uint32 (AlpDmlItemH  itemH,
                                               AlpAttnDmlColumnType column,
                                               uint32_t    *value);

    alp_status_t alp_attn_dml_item_set_bool  (AlpDmlItemH  itemH,
                                               AlpAttnDmlColumnType column,
                                               bool     *value);

    alp_status_t alp_attn_dml_item_set_string (AlpDmlItemH  itemH,
                                               AlpAttnDmlColumnType column,
                                               const char  *value);

    alp_status_t alp_attn_dml_item_set_blob   (AlpDmlItemH  itemH,
                                               AlpAttnDmlColumnType column,
                                               void        *data,
                                               uint32_t     size);

    const char *alp_attn_dml_item_get_string  (AlpDmlItemH   itemH,
                                               AlpAttnDmlColumnType column);

    int32_t     alp_attn_dml_item_get_int32   (AlpDmlItemH   itemH,
                                               AlpAttnDmlColumnType column);

    uint32_t    alp_attn_dml_item_get_uint32  (AlpDmlItemH   itemH,
                                               AlpAttnDmlColumnType column);

    /**
     * miscellaneous setter / getter functions
     * These are for adding / getting properties associated with an alert.
     * There is not a db column associated with these but instead another
     * table (PROPERTIES).
     */
    alp_status_t alp_attn_dml_item_set_property_list (AlpDmlItemH itemH,
                                                      AlpDmlItemH propListH);
    alp_status_t alp_attn_dml_item_add_property     (AlpDmlItemH      itemH,
                                                     AlpDmlItemH      propH);
    /*@}*/

    /**
     * @name Attention DML Write API
     * @{
     */
    extern alp_status_t alp_attn_dml_insert (AlpDmlH      dmlH,
                                             AlpDmlItemH  itemH,
                                             AlpLuid     *itemLuid);
    
    extern alp_status_t alp_attn_dml_replace (AlpDmlH      dmlH,
                                              AlpDmlItemH  itemH);

    extern alp_status_t alp_attn_dml_update (AlpDmlH      dmlH,
                                             AlpDmlItemH  itemH);

    extern alp_status_t alp_attn_dml_delete (AlpDmlH      dmlH,
                                             AlpDmlItemH  itemH);

    extern alp_status_t alp_attn_dml_find   (AlpDmlH       dmlH,
                                             AlpDmlItemH   itemH,
                                             AlpLuid      *itemLuid);
    /* define more setters here */
    /*@}*/

    
    /**
     * @name Attention DML Read API
     * @{
     */
    alp_status_t alp_attn_dml_alert_statement_begin (AlpDmlH           dmlH,
                                                     AlpDmlStatementH *stmtH,
                                                     char             *source,
                                                     char             *name,
                                                     char             *handle,
                                                     AlpLuid           alertLuid,
                                                     AlpAttnDmlOrderByType orderBy,
                                                     uint32_t *numResults);

    alp_status_t alp_attn_dml_prop_statement_begin  (AlpDmlH           dmlH,
                                                     AlpDmlStatementH *stmtH,
                                                     /* more selection args */
                                                     AlpLuid           alertLuid,
                                                     AlpLuid           propLuid,
                                                     char             *property,
                                                     uint32_t *numResults);

    alp_status_t alp_attn_dml_statement_next        (AlpDmlStatementH   stmtH);

    alp_status_t alp_attn_dml_statement_end         (AlpDmlStatementH   stmtH);

    alp_status_t alp_attn_dml_statement_get_int32   (AlpDmlStatementH   stmtH,
                                                     AlpAttnDmlColumnType column,
                                                     int32_t           *value);

    alp_status_t alp_attn_dml_statement_get_uint32  (AlpDmlStatementH   stmtH,
                                                     AlpAttnDmlColumnType column,
                                                     uint32_t          *value);

    alp_status_t alp_attn_dml_statement_get_int64   (AlpDmlStatementH   stmtH,
                                                     AlpAttnDmlColumnType column,
                                                     int64_t           *value);

    alp_status_t alp_attn_dml_statement_get_string  (AlpDmlStatementH   stmtH,
                                                     AlpAttnDmlColumnType column,
                                                     char             **value);

    alp_status_t alp_attn_dml_statement_get_bool    (AlpDmlStatementH   stmtH,
                                                     AlpAttnDmlColumnType column,
                                                     bool              *value);

    alp_status_t alp_attn_dml_statement_get_timestamp(AlpDmlStatementH   stmtH,
                                                     AlpAttnDmlColumnType column,
                                                      time_t            *value);

    alp_status_t alp_attn_dml_statement_get_luid   (AlpDmlStatementH  stmtH,
                                                    AlpLuid          *itemLuid);

    /**
     * @brief  Given a statement handle that has been initialized
     *         and stepped using alp_attn_dml_statement_next() where
     *         an item has been found, this function can be used to
     *         retrieve any integer field from the item located at
     *         the current statement position.
     *
     * Note that unless copy is set to true, the value returned is only valid
     * until the dmlStmtH is closed with alp_attn_dml_statement_end() or 
     * stepped with alp_attn_dml_statement_next()
     *
     * @param [in] dmlStmtH Handle to a statement
     * @param [in] column   Any column type that exists for the item
     *                      type pointed to in the statement.
     * @param [in] copy     If true, then this routine mallocs memory for
     *                      value of size bytes.  It's the responsibility
     *                      of the caller to free it.
     * @param [out] value   Pointer to data to read out from a field of the
     *                      current item in the statement
     * @param [out] size    size of the value item, in bytes.
     *                      Note that if the copy parameter is true,
     *                      this is how much memory is allocated by
     *                      the routine.
     *
     * @return ALP_STATUS_ATTN_DML_* error
     */
    alp_status_t alp_attn_dml_statement_get_blob (AlpDmlStatementH stmtH,
                                                  AlpAttnDmlColumnType column,
                                                  bool       copy, 
                                                  void     **value,
                                                  uint32_t  *size);

    /**
     * @name Attention DML Transaction API
     * @{
     */
    /* Transaction APIs */
    extern alp_status_t alp_attn_dml_transaction_begin   (AlpDmlH dmlH,
                                                          bool exclusive);
    extern alp_status_t alp_attn_dml_transaction_end     (AlpDmlH dmlH);
    extern alp_status_t alp_attn_dml_transaction_rollback(AlpDmlH dmlH);
    /* @} */

    /* @} */

#ifdef __cplusplus
}
#endif

#endif /* __ALP_ATTN_DML_H__ */
