/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/

/***************************************************************************
//
//	File:			support/KeyedVector.h
//
//	Description:	A templatized key/value mapping.
//
//	Copyright 2001, Be Incorporated, All Rights Reserved.
//
***************************************************************************/

#ifndef _SUPPORT_ASSOCIATIVEVECTOR_H
#define _SUPPORT_ASSOCIATIVEVECTOR_H

#include <hiker/prv/support/Vector.h>
#include <hiker/prv/support/OrderedVector.h>

namespace alp {

/*--------------------------------------------------------*/
/*----- KeyedVector class -------------------------------*/

template<class KEY, class VALUE>
class KeyedVector
{
public:
							KeyedVector(const VALUE& undefined = VALUE());
							KeyedVector(const KeyedVector<KEY,VALUE>& o);
							KeyedVector(	const OrderedVector<KEY>& keys,
											const Vector<VALUE>& values,
											const VALUE& undefined = VALUE());
	virtual					~KeyedVector();

			KeyedVector<KEY,VALUE>&	operator=(const KeyedVector<KEY,VALUE>& o);

	/* Size stats */

			void			SetCapacity(size_t total_space);
			void			SetExtraCapacity(size_t extra_space);
			size_t			Capacity() const;

			size_t			CountItems() const;

	/* Value by Key */

			const VALUE&	ValueFor(const KEY& key, bool* found = NULL) const;
			VALUE&			EditValueFor(const KEY& key, bool* found = NULL);

	/* Value/Key by index */

			const KEY&		KeyAt(size_t i) const;
			const VALUE&	ValueAt(size_t i) const;
			VALUE&			EditValueAt(size_t i);

			const OrderedVector<KEY>&	KeyVector() const;
			const Vector<VALUE>&		ValueVector() const;
			Vector<VALUE>&				ValueVector();

	/* List manipulation */

			ssize_t			IndexOf(const KEY& key) const;
			bool			GetIndexOf(const KEY& key, size_t* index) const;

			ssize_t			AddItem(const KEY& key, const VALUE& value);

			void			RemoveItemsAt(size_t index, size_t count = 1);
			ssize_t			RemoveItemFor(const KEY& key);

			void			MakeEmpty();

			void			Swap(KeyedVector& o);

private:
			OrderedVector<KEY>	m_keys;
			Vector<VALUE>		m_values;
			VALUE				m_undefined;
			size_t				_reserved_data1;
			size_t				_reserved_data2;
};

// Type optimizations.
template<class KEY, class VALUE>
void Swap(KeyedVector<KEY, VALUE>& v1, KeyedVector<KEY, VALUE>& v2);

/*-------------------------------------------------------------*/
/*---- No user serviceable parts after this -------------------*/

template<class KEY, class VALUE> inline
KeyedVector<KEY,VALUE>::KeyedVector(const VALUE& undefined)
	:	m_undefined(undefined),
		_reserved_data1(0), _reserved_data2(0)
{
}

template<class KEY, class VALUE> inline
KeyedVector<KEY,VALUE>::KeyedVector(const KeyedVector<KEY,VALUE>& o)
	:	m_keys(o.m_keys), m_values(o.m_values), m_undefined(o.m_undefined),
		_reserved_data1(0), _reserved_data2(0)
{
}

template<class KEY, class VALUE> inline
KeyedVector<KEY,VALUE>::KeyedVector(const OrderedVector<KEY>& keys,
			const Vector<VALUE>& values, const VALUE& undefined)
	:	m_keys(keys), m_values(values), m_undefined(undefined),
		_reserved_data1(0), _reserved_data2(0)
{
}

template<class KEY, class VALUE> inline
KeyedVector<KEY,VALUE>::~KeyedVector()
{
}

template<class KEY, class VALUE> inline
KeyedVector<KEY,VALUE>& KeyedVector<KEY,VALUE>::operator=(const KeyedVector<KEY,VALUE>& o)
{
	m_keys = o.m_keys; m_values = o.m_values; m_undefined = o.m_undefined;
	return *this;
}

template<class KEY, class VALUE> inline
void KeyedVector<KEY,VALUE>::SetCapacity(size_t total_space)
{
	m_keys.SetCapacity(total_space); m_values.SetCapacity(total_space);
}

template<class KEY, class VALUE> inline
void KeyedVector<KEY,VALUE>::SetExtraCapacity(size_t extra_space)
{
	m_keys.SetExtraCapacity(extra_space); m_values.SetExtraCapacity(extra_space);
}

template<class KEY, class VALUE> inline
size_t KeyedVector<KEY,VALUE>::Capacity() const
{
	return m_keys.Capacity();
}

template<class KEY, class VALUE> inline
size_t KeyedVector<KEY,VALUE>::CountItems() const
{
	return m_keys.CountItems();
}

template<class KEY, class VALUE> inline
const VALUE& KeyedVector<KEY,VALUE>::ValueFor(const KEY& key, bool* found) const
{
	size_t index = 0;
	bool has = m_keys.GetIndexOf(key, &index);
	
	if( NULL != found ) *found = has;
	
	if( true == has )
	{
		return m_values.ItemAt(index);
	}
	else
	{
		return m_undefined;
	}
}

template<class KEY, class VALUE> inline
VALUE& KeyedVector<KEY,VALUE>::EditValueFor(const KEY& key, bool* found)
{
	size_t index = 0;
	bool has = m_keys.GetIndexOf(key, &index);
	
	if( NULL != found ) *found = has;
	
	if( true == *found )
	{
		return m_values.EditItemAt(index);
	}
	else
	{
		return m_undefined;
	}
}

template<class KEY, class VALUE> inline
const KEY& KeyedVector<KEY,VALUE>::KeyAt(size_t i) const
{
	return m_keys.ItemAt(i);
}

template<class KEY, class VALUE> inline
const VALUE& KeyedVector<KEY,VALUE>::ValueAt(size_t i) const
{
	return m_values.ItemAt(i);
}

template<class KEY, class VALUE> inline
VALUE& KeyedVector<KEY,VALUE>::EditValueAt(size_t i)
{
	return m_values.EditItemAt(i);
}

template<class KEY, class VALUE> inline
const OrderedVector<KEY>& KeyedVector<KEY,VALUE>::KeyVector() const
{
	return m_keys;
}

template<class KEY, class VALUE> inline
const Vector<VALUE>& KeyedVector<KEY,VALUE>::ValueVector() const
{
	return m_values;
}

template<class KEY, class VALUE> inline
Vector<VALUE>& KeyedVector<KEY,VALUE>::ValueVector()
{
	return m_values;
}

template<class KEY, class VALUE> inline
ssize_t KeyedVector<KEY,VALUE>::IndexOf(const KEY& key) const
{
	return m_keys.IndexOf(key);
}

template<class KEY, class VALUE> inline
bool KeyedVector<KEY,VALUE>::GetIndexOf(const KEY& key, size_t* index) const
{
	return m_keys.GetIndexOf(key, index);
}

template<class KEY, class VALUE> inline
ssize_t KeyedVector<KEY,VALUE>::AddItem(const KEY& key, const VALUE& value)
{
	bool added;
	const ssize_t pos = m_keys.AddItem(key, &added);
	if (added) {
		const ssize_t vpos = m_values.AddItemAt(value, pos);
		if (vpos < 0) m_keys.RemoveItemsAt(pos);
		return vpos;
	} else {
		m_values.ReplaceItemAt(value, pos);
		return pos;
	}
}

template<class KEY, class VALUE> inline
void KeyedVector<KEY,VALUE>::RemoveItemsAt(size_t index, size_t count)
{
	m_keys.RemoveItemsAt(index, count); m_values.RemoveItemsAt(index, count);
}

template<class KEY, class VALUE> inline
ssize_t KeyedVector<KEY,VALUE>::RemoveItemFor(const KEY& key)
{
	const ssize_t pos = m_keys.RemoveItemFor(key);
	if (pos >= 0) m_values.RemoveItemsAt(pos);
	return pos;
}

template<class KEY, class VALUE> inline
void KeyedVector<KEY,VALUE>::MakeEmpty()
{
	m_keys.MakeEmpty(); m_values.MakeEmpty();
}

template<class KEY, class VALUE> inline
void KeyedVector<KEY, VALUE>::Swap(KeyedVector<KEY, VALUE>& o)
{
	Swap(m_keys, o.m_keys);
	Swap(m_values, o.m_values);
	Swap(m_undefined, o.m_undefined);
}

template<class KEY, class VALUE> inline
void Swap(KeyedVector<KEY, VALUE>& v1, KeyedVector<KEY, VALUE>& v2)
{
	v1.Swap(v2);
}

}	// namespace alp

#endif
