/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/***************************************************************************
//
//	File:			support/OrderedVector.h
//
//	Description:	A Vector whose types have an intrinsic ordering.
//
//	Copyright 2001, Be Incorporated, All Rights Reserved.
//
***************************************************************************/

#ifndef _SUPPORT_ORDEREDVECTOR_H
#define _SUPPORT_ORDEREDVECTOR_H

#include <hiker/prv/support/TypeFuncs.h>
#include <hiker/prv/support/Vector.h>

namespace alp {

/*--------------------------------------------------------*/
/*----- AbstractOrderedVector abstract base class -------*/

class AbstractOrderedVector : public AbstractVector
{
public:
							AbstractOrderedVector(size_t element_size);
							AbstractOrderedVector(const AbstractVector& o);
							// WARNING: Your subclass must call MakeEmpty()
							// in its own destructor!
	virtual					~AbstractOrderedVector();

			AbstractOrderedVector& operator=(const AbstractOrderedVector& o);

			// Note that we inherit the assignment operator, so you can
			// assign a plain AbstractVector to an instance of this class.

			ssize_t			AddOrdered(const void* newElement, bool* added = NULL);

			ssize_t			OrderOf(const void* element) const;
			bool			GetOrderOf(const void* element, size_t* index) const;

			ssize_t			RemoveOrdered(const void* element);

			void			Swap(AbstractOrderedVector& o);

protected:
	virtual	int32_t			PerformCompare(const void* d1, const void* d2) const = 0;
	virtual	bool			PerformLessThan(const void* d1, const void* d2) const = 0;

private:
	virtual	alp_status_t	_ReservedUntypedOrderedVector1();
	virtual	alp_status_t	_ReservedUntypedOrderedVector2();
	virtual	alp_status_t	_ReservedUntypedOrderedVector3();
	virtual	alp_status_t	_ReservedUntypedOrderedVector4();
	virtual	alp_status_t	_ReservedUntypedOrderedVector5();
	virtual	alp_status_t	_ReservedUntypedOrderedVector6();

			int32_t			_reserved[2];
};

// Type optimizations.
void MoveBefore(AbstractOrderedVector* to, AbstractOrderedVector* from, size_t count = 1);
void MoveAfter(AbstractOrderedVector* to, AbstractOrderedVector* from, size_t count = 1);
void Swap(AbstractOrderedVector& v1, AbstractOrderedVector& v2);

/*--------------------------------------------------------*/
/*----- OrderedVector concrete class --------------------*/

template<class TYPE>
class OrderedVector : private AbstractOrderedVector
{
public:
							OrderedVector();
							OrderedVector(const OrderedVector<TYPE>& o);
	virtual					~OrderedVector();

			OrderedVector<TYPE>&	operator=(const OrderedVector<TYPE>& o);

	/* Size stats */

			void			SetCapacity(size_t total_space);
			void			SetExtraCapacity(size_t extra_space);
			size_t			Capacity() const;

			size_t			CountItems() const;

	/* Data access */

			const TYPE&		operator[](size_t i) const;
			const TYPE&		ItemAt(size_t i) const;

			ssize_t			IndexOf(const TYPE& item) const;
			bool			GetIndexOf(const TYPE& item, size_t* index) const;

	/* Array modification */

			ssize_t			AddItem(const TYPE& item, bool* added = NULL);

			void			MakeEmpty();
			void			RemoveItemsAt(size_t index, size_t count = 1);

			ssize_t			RemoveItemFor(const TYPE& item);

			void			Swap(OrderedVector<TYPE>& o);

protected:
	virtual	void			PerformConstruct(void* base, size_t count) const;
	virtual	void			PerformCopy(void* to, const void* from, size_t count) const;
	virtual	void			PerformDestroy(void* base, size_t count) const;

	virtual	void			PerformMoveBefore(void* to, void* from, size_t count) const;
	virtual	void			PerformMoveAfter(void* to, void* from, size_t count) const;

	virtual	void			PerformAssign(void* to, const void* from, size_t count) const;

	virtual	int32_t			PerformCompare(const void* d1, const void* d2) const;
	virtual	bool			PerformLessThan(const void* d1, const void* d2) const;
};

// Type optimizations.
template<class TYPE>
void MoveBefore(OrderedVector<TYPE>* to, OrderedVector<TYPE>* from, size_t count = 1);
template<class TYPE>
void MoveAfter(OrderedVector<TYPE>* to, OrderedVector<TYPE>* from, size_t count = 1);
template<class TYPE>
void Swap(OrderedVector<TYPE>& v1, OrderedVector<TYPE>& v2);

/*-------------------------------------------------------------*/
/*---- No user serviceable parts after this -------------------*/

inline void Swap(AbstractOrderedVector& v1, AbstractOrderedVector& v2)
{
	v1.Swap(v2);
}

/*-------------------------------------------------------------*/

template<class TYPE> inline
OrderedVector<TYPE>::OrderedVector()
	:	AbstractOrderedVector(sizeof(TYPE[2])/2)
{
}

template<class TYPE> inline
OrderedVector<TYPE>::OrderedVector(const OrderedVector<TYPE>& o)
	:	AbstractOrderedVector(o)
{
}

template<class TYPE> inline
OrderedVector<TYPE>::~OrderedVector()
{
	MakeEmpty();
}

template<class TYPE> inline
OrderedVector<TYPE>& OrderedVector<TYPE>::operator=(const OrderedVector<TYPE>& o)
{
	AbstractOrderedVector::operator=(o);
	return *this;
}

template<class TYPE> inline
void OrderedVector<TYPE>::SetCapacity(size_t total_space)
{
	AbstractVector::SetCapacity(total_space);
}

template<class TYPE> inline
void OrderedVector<TYPE>::SetExtraCapacity(size_t extra_space)
{
	AbstractVector::SetExtraCapacity(extra_space);
}

template<class TYPE> inline
size_t OrderedVector<TYPE>::Capacity() const
{
	return AbstractVector::Capacity();
}

template<class TYPE> inline
size_t OrderedVector<TYPE>::CountItems() const
{
	return AbstractVector::CountItems();
}

template<class TYPE> inline
const TYPE& OrderedVector<TYPE>::operator[](size_t i) const
{
	return *static_cast<const TYPE*>(At(i));
}

template<class TYPE> inline
const TYPE& OrderedVector<TYPE>::ItemAt(size_t i) const
{
	return *static_cast<const TYPE*>(At(i));
}

template<class TYPE> inline
ssize_t OrderedVector<TYPE>::AddItem(const TYPE& item, bool* added)
{
	return AddOrdered(&item, added);
}

template<class TYPE> inline
ssize_t OrderedVector<TYPE>::IndexOf(const TYPE& item) const
{
	return OrderOf(&item);
}

template<class TYPE> inline
bool OrderedVector<TYPE>::GetIndexOf(const TYPE& item, size_t* index) const
{
	return GetOrderOf(&item, index);
}

template<class TYPE> inline
void OrderedVector<TYPE>::MakeEmpty()
{
	AbstractOrderedVector::MakeEmpty();
}

template<class TYPE> inline
void OrderedVector<TYPE>::RemoveItemsAt(size_t index, size_t count)
{
	AbstractOrderedVector::RemoveItemsAt(index, count);
}

template<class TYPE> inline
ssize_t OrderedVector<TYPE>::RemoveItemFor(const TYPE& item)
{
	return RemoveOrdered(&item);
}

template<class TYPE> inline
void OrderedVector<TYPE>::Swap(OrderedVector<TYPE>& o)
{
	AbstractOrderedVector::Swap(o);
}

template<class TYPE>
void OrderedVector<TYPE>::PerformConstruct(void* base, size_t count) const
{
	Construct(static_cast<TYPE*>(base), count);
}

template<class TYPE>
void OrderedVector<TYPE>::PerformCopy(void* to, const void* from, size_t count) const
{
	Copy(static_cast<TYPE*>(to), static_cast<const TYPE*>(from), count);
}

template<class TYPE>
void OrderedVector<TYPE>::PerformDestroy(void* base, size_t count) const
{
	Destroy(static_cast<TYPE*>(base), count);
}

template<class TYPE>
void OrderedVector<TYPE>::PerformMoveBefore(void* to, void* from, size_t count) const
{
	MoveBefore(static_cast<TYPE*>(to), static_cast<TYPE*>(from), count);
}

template<class TYPE>
void OrderedVector<TYPE>::PerformMoveAfter(void* to, void* from, size_t count) const
{
	MoveAfter(static_cast<TYPE*>(to), static_cast<TYPE*>(from), count);
}

template<class TYPE>
void OrderedVector<TYPE>::PerformAssign(void* to, const void* from, size_t count) const
{
	Assign(static_cast<TYPE*>(to), static_cast<const TYPE*>(from), count);
}

template<class TYPE>
int32_t OrderedVector<TYPE>::PerformCompare(const void* d1, const void* d2) const
{
	return Compare(*static_cast<const TYPE*>(d1), *static_cast<const TYPE*>(d2));
}

template<class TYPE>
bool OrderedVector<TYPE>::PerformLessThan(const void* d1, const void* d2) const
{
	return LessThan(*static_cast<const TYPE*>(d1), *static_cast<const TYPE*>(d2));
}

/*-------------------------------------------------------------*/

template<class TYPE> inline
void MoveBefore(OrderedVector<TYPE>* to, OrderedVector<TYPE>* from, size_t count)
{
	MoveBefore(static_cast<AbstractOrderedVector*>(to), static_cast<AbstractOrderedVector*>(from), count);
}

template<class TYPE> inline
void MoveAfter(OrderedVector<TYPE>* to, OrderedVector<TYPE>* from, size_t count)
{
	MoveAfter(static_cast<AbstractOrderedVector*>(to), static_cast<AbstractOrderedVector*>(from), count);
}

template<class TYPE> inline
void Swap(OrderedVector<TYPE>& v1, OrderedVector<TYPE>& v2)
{
	v1.Swap(v2);
}

}	// namespace alp

#endif
