/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/***************************************************************************
//
//	File:			support/TypeFuncs.h
//
//	Description:	Templatized functions for various type operations.
//
//	Copyright 2001, Be Incorporated, All Rights Reserved.
//
***************************************************************************/

#ifndef _SUPPORT_TYPEFUNCS_H
#define _SUPPORT_TYPEFUNCS_H

//#include <support/SupportDefs.h>

#include <memory.h>
#include <new>
#include <stdint.h>

namespace alp {

/*--------------------------------------------------------*/
/*----- Basic operations on types ------------------------*/

template<class TYPE>
static inline void Construct(TYPE* base, size_t count = 1)
{
	while (--count != (size_t)-1) {
		new(base) TYPE();
		base++;
	}
}

template<class TYPE>
static inline void Destroy(TYPE* base, size_t count = 1)
{
	while (--count != (size_t)-1) {
		base->~TYPE();
		base++;
	}
}

template<class TYPE>
static inline void Copy(TYPE* to, const TYPE* from, size_t count = 1)
{
	while (--count != (size_t)-1) {
		new(to) TYPE(*from);
		to++;
		from++;
	}
}

template<class TYPE>
static inline void MoveBefore(	TYPE* to, TYPE* from, size_t count = 1)
{
	while (--count != (size_t)-1) {
		new(to) TYPE(*from);
		from->~TYPE();
		to++;
		from++;
	}
}

template<class TYPE>
static inline void MoveAfter(	TYPE* to, TYPE* from, size_t count = 1)
{
	to+=(count-1);
	from+=(count-1);
	while (--count != (size_t)-1) {
		new(to) TYPE(*from);
		from->~TYPE();
		to--;
		from--;
	}
}

template<class TYPE>
static inline void Assign(	TYPE* to, const TYPE* from, size_t count = 1)
{
	while (--count != (size_t)-1) {
		*to = *from;
		to++;
		from++;
	}
}

template<class TYPE>
static inline void Swap(TYPE& v1, TYPE& v2)
{
	TYPE tmp(v1); v1 = v2; v2 = tmp;
}

template<class TYPE>
static inline int32_t Compare(const TYPE& v1, const TYPE& v2)
{
	return (v1 < v2) ? -1 : ( (v2 < v1) ? 1 : 0 );
}

template<class TYPE>
static inline bool LessThan(const TYPE& v1, const TYPE& v2)
{
	return v1 < v2;
}

/*--------------------------------------------------------*/
/*----- Optimizations for all pointer types --------------*/

template<class TYPE>
static inline void Construct(TYPE** base, size_t count = 1)
	{ (void)base; (void)count; }
template<class TYPE>
static inline void Destroy(TYPE** base, size_t count = 1)
	{ (void)base; (void)count; }
template<class TYPE>
static inline void Copy(TYPE** to, TYPE* const * from, size_t count = 1)
	{ if (count == 1) *to = *from; else memcpy(to, from, sizeof(TYPE*)*count); }
template<class TYPE>
static inline void MoveBefore(TYPE** to, TYPE** from, size_t count = 1)
	{ if (count == 1) *to = *from; else memcpy(to, from, sizeof(TYPE*)*count); }
template<class TYPE>
static inline void MoveAfter(TYPE** to, TYPE** from, size_t count = 1)
	{ if (count == 1) *to = *from; else memmove(to, from, sizeof(TYPE*)*count); }
template<class TYPE>
static inline void Assign(TYPE** to, TYPE* const * from, size_t count = 1)
	{ if (count == 1) *to = *from; else memcpy(to, from, sizeof(TYPE*)*count); }

/*--------------------------------------------------------*/
/*----- Optimizations for basic data types ---------------*/

// Standard optimizations for types that don't contain internal or
// other pointers to themselves.
#define ALP_IMPLEMENT_SIMPLE_TYPE_FUNCS(TYPE)											\
static inline void MoveBefore(TYPE* to, TYPE* from, size_t count = 1)					\
	{ memcpy(to, from, (sizeof(TYPE[2])/2)*count); }									\
static inline void MoveAfter(TYPE* to, TYPE* from, size_t count = 1)					\
	{ memmove(to, from, (sizeof(TYPE[2])/2)*count); }									\

// Extreme optimizations for types whose constructor and destructor
// don't need to be called.
#define ALP_IMPLEMENT_BASIC_TYPE_FUNCS(TYPE)											\
static inline void Construct(TYPE* base, size_t count)									\
	{ (void)base; (void)count; }														\
static inline void Destroy(TYPE* base, size_t count)									\
	{ (void)base; (void)count; }														\
static inline void Copy(TYPE* to, const TYPE* from, size_t count = 1)					\
	{ if (count == 1) *to = *from; else memcpy(to, from, (sizeof(TYPE[2])/2)*count); }	\
static inline void MoveBefore(TYPE* to, TYPE* from, size_t count = 1)					\
	{ if (count == 1) *to = *from; else memcpy(to, from, (sizeof(TYPE[2])/2)*count); }	\
static inline void MoveAfter(TYPE* to, TYPE* from, size_t count = 1)					\
	{ if (count == 1) *to = *from; else memmove(to, from, (sizeof(TYPE[2])/2)*count); }	\
static inline void Assign(TYPE* to, const TYPE* from, size_t count = 1)				\
	{ if (count == 1) *to = *from; else memcpy(to, from, (sizeof(TYPE[2])/2)*count); }	\

ALP_IMPLEMENT_BASIC_TYPE_FUNCS(bool)
ALP_IMPLEMENT_BASIC_TYPE_FUNCS(int8_t)
ALP_IMPLEMENT_BASIC_TYPE_FUNCS(uint8_t)
ALP_IMPLEMENT_BASIC_TYPE_FUNCS(int16_t)
ALP_IMPLEMENT_BASIC_TYPE_FUNCS(uint16_t)
ALP_IMPLEMENT_BASIC_TYPE_FUNCS(int32_t)
ALP_IMPLEMENT_BASIC_TYPE_FUNCS(uint32_t)
ALP_IMPLEMENT_BASIC_TYPE_FUNCS(int64_t)
ALP_IMPLEMENT_BASIC_TYPE_FUNCS(uint64_t)
ALP_IMPLEMENT_BASIC_TYPE_FUNCS(float)
ALP_IMPLEMENT_BASIC_TYPE_FUNCS(double)

}	// namespace alp

#endif
