/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/***************************************************************************
//
//	File:			support/Vector.h
//
//	Description:	A simple array-like container class.  Implemented as
//					a general purpose abstract base-class AbstractVector,
//					with the concrete class Vector layered on top and
//					templatized on the array type.
//
//	Copyright 2001, Be Incorporated, All Rights Reserved.
//
***************************************************************************/

#ifndef _SUPPORT_VECTOR_H
#define _SUPPORT_VECTOR_H

#include <errno.h>
#include <stddef.h>
#include <stdint.h>

#include <hiker/types.h>

#include <hiker/prv/support/TypeFuncs.h>

typedef int32_t	ssize_t;

#ifndef SSIZE_MAX
#	define SSIZE_MAX INT32_MAX
#endif

namespace alp {

/*--------------------------------------------------------*/
/*----- AbstractVector abstract base class --------------*/

class AbstractVector
{
public:
							AbstractVector(size_t element_size);
							AbstractVector(const AbstractVector& o);
							// WARNING: Your subclass must call MakeEmpty()
							// in its own destructor!
	virtual					~AbstractVector();

			AbstractVector&	operator=(const AbstractVector& o);

	/* Size stats */

			void			SetCapacity(size_t total_space);
			void			SetExtraCapacity(size_t extra_space);
			size_t			Capacity() const;

			size_t			ItemSize() const;

			size_t			CountItems() const;

	/* Data access */

			const void*		At(size_t index) const;
			void*			EditAt(size_t index);

	/* Array modification */

			ssize_t			Add(const void* newElement);
			ssize_t			AddAt(const void* newElement, size_t index);

			ssize_t			AddVector(const AbstractVector& o);
			ssize_t			AddVectorAt(const AbstractVector& o, size_t index = SSIZE_MAX);

			void			MakeEmpty();
			void			RemoveItemsAt(size_t index, size_t count);

			void			Swap(AbstractVector& o);

protected:
	static char const * const gStaskOverFlow;

	virtual	void			PerformConstruct(void* base, size_t count) const = 0;
	virtual	void			PerformCopy(void* to, const void* from, size_t count) const = 0;
	virtual	void			PerformDestroy(void* base, size_t count) const = 0;

	virtual	void			PerformMoveBefore(void* to, void* from, size_t count) const = 0;
	virtual	void			PerformMoveAfter(void* to, void* from, size_t count) const = 0;

	virtual	void			PerformAssign(void* to, const void* from, size_t count) const = 0;

private:

	virtual	alp_status_t	_ReservedUntypedVector1();
	virtual	alp_status_t	_ReservedUntypedVector2();
	virtual	alp_status_t	_ReservedUntypedVector3();
	virtual	alp_status_t	_ReservedUntypedVector4();
	virtual	alp_status_t	_ReservedUntypedVector5();
	virtual	alp_status_t	_ReservedUntypedVector6();
	virtual	alp_status_t	_ReservedUntypedVector7();
	virtual	alp_status_t	_ReservedUntypedVector8();
	virtual	alp_status_t	_ReservedUntypedVector9();
	virtual	alp_status_t	_ReservedUntypedVector10();

			uint8_t*		grow(size_t amount, size_t factor=3, size_t pos=0xFFFFFFFF);
			uint8_t*		shrink(size_t amount, size_t factor=4, size_t pos=0xFFFFFFFF);
			const uint8_t*	data() const;
			uint8_t*		edit_data();

			const size_t	m_elementSize;
			const size_t	m_localSpace;
			size_t			m_avail;
			size_t			m_size;

			union {
				uint8_t*	heap;
				uint8_t		local[8];
			} m_data;

			int32_t			_reserved[2];
};

// Type optimizations.
void MoveBefore(AbstractVector* to, AbstractVector* from, size_t count = 1);
void MoveAfter(AbstractVector* to, AbstractVector* from, size_t count = 1);
void Swap(AbstractVector& v1, AbstractVector& v2);

/*--------------------------------------------------------*/
/*----- Vector concrete class ---------------------------*/

/** @brief A templatized array class for type safety for use as a small std::vector alternative. 
*/
template<class TYPE>
class Vector : private AbstractVector
{
public:
							Vector();
							Vector(const Vector<TYPE>& o);
	virtual					~Vector();

			Vector<TYPE>&	operator=(const Vector<TYPE>& o);

	/* Size stats */

			void			SetCapacity(size_t total_space);
			void			SetExtraCapacity(size_t extra_space);
			size_t			Capacity() const;

			size_t			CountItems() const;

	/* Data access */

			const TYPE&		operator[](size_t i) const;
			const TYPE&		ItemAt(size_t i) const;
			TYPE&			EditItemAt(size_t i);

	/* Array modification */

			ssize_t			AddItem();
			ssize_t			AddItem(const TYPE& item);
			ssize_t			AddItemAt(size_t index);
			ssize_t			AddItemAt(const TYPE& item, size_t index);

			ssize_t			ReplaceItemAt(const TYPE& item, size_t index);

			ssize_t			AddVector(const Vector<TYPE>& o);
			ssize_t			AddVectorAt(const Vector<TYPE>& o, size_t index);

			void			MakeEmpty();
			void			RemoveItemsAt(size_t index, size_t count = 1);

			void			Swap(Vector<TYPE>& o);

	/* Use as a stack */

			void			Push();
			void			Push(const TYPE& item);
			TYPE &			EditTop();
			const TYPE &	Top() const;
			void			Pop();

protected:
	virtual	void			PerformConstruct(void* base, size_t count) const;
	virtual	void			PerformCopy(void* to, const void* from, size_t count) const;
	virtual	void			PerformDestroy(void* base, size_t count) const;

	virtual	void			PerformMoveBefore(void* to, void* from, size_t count) const;
	virtual	void			PerformMoveAfter(void* to, void* from, size_t count) const;

	virtual	void			PerformAssign(void* to, const void* from, size_t count) const;
};

// Type optimizations.
template<class TYPE>
void MoveBefore(Vector<TYPE>* to, Vector<TYPE>* from, size_t count = 1);
template<class TYPE>
void MoveAfter(Vector<TYPE>* to, Vector<TYPE>* from, size_t count = 1);
template<class TYPE>
void Swap(Vector<TYPE>& v1, Vector<TYPE>& v2);

/*-------------------------------------------------------------*/
/*---- No user serviceable parts after this -------------------*/

inline void Swap(AbstractVector& v1, AbstractVector& v2)
{
	v1.Swap(v2);
}

/*-------------------------------------------------------------*/

template<class TYPE> inline
Vector<TYPE>::Vector()
	:	AbstractVector(sizeof(TYPE[2])/2)
{
}

template<class TYPE> inline
Vector<TYPE>::Vector(const Vector<TYPE>& o)
	:	AbstractVector(o)
{
}

template<class TYPE> inline
Vector<TYPE>::~Vector()
{
	MakeEmpty();
}

template<class TYPE> inline
Vector<TYPE>& Vector<TYPE>::operator=(const Vector<TYPE>& o)
{
	AbstractVector::operator=(o);
	return *this;
}

template<class TYPE> inline
void Vector<TYPE>::SetCapacity(size_t total_space)
{
	AbstractVector::SetCapacity(total_space);
}

template<class TYPE> inline
void Vector<TYPE>::SetExtraCapacity(size_t extra_space)
{
	AbstractVector::SetExtraCapacity(extra_space);
}

template<class TYPE> inline
size_t Vector<TYPE>::Capacity() const
{
	return AbstractVector::Capacity();
}

template<class TYPE> inline
size_t Vector<TYPE>::CountItems() const
{
	return AbstractVector::CountItems();
}

template<class TYPE> inline
const TYPE& Vector<TYPE>::operator[](size_t i) const
{
	return *static_cast<const TYPE*>(At(i));
}

template<class TYPE> inline
const TYPE& Vector<TYPE>::ItemAt(size_t i) const
{
	return *static_cast<const TYPE*>(At(i));
}

template<class TYPE> inline
TYPE& Vector<TYPE>::EditItemAt(size_t i)
{
	return *static_cast<TYPE*>(EditAt(i));
}

template<class TYPE> inline
ssize_t Vector<TYPE>::AddItem()
{
	return Add(NULL);
}

template<class TYPE> inline
ssize_t Vector<TYPE>::AddItem(const TYPE &item)
{
	return Add(&item);
}

template<class TYPE> inline
ssize_t Vector<TYPE>::AddItemAt(size_t index)
{
	return AddAt(NULL, index);
}

template<class TYPE> inline
ssize_t Vector<TYPE>::AddItemAt(const TYPE &item, size_t index)
{
	return AddAt(&item, index);
}

template<class TYPE> inline
ssize_t Vector<TYPE>::ReplaceItemAt(const TYPE& item, size_t index)
{
	TYPE* elem = static_cast<TYPE*>(EditAt(index));
	if (elem) {
		*elem = item;
		return index;
	}
	return ENOMEM;
}

template<class TYPE> inline
ssize_t Vector<TYPE>::AddVector(const Vector<TYPE>& o)
{
	return AbstractVector::AddVector(o);
}

template<class TYPE> inline
ssize_t Vector<TYPE>::AddVectorAt(const Vector<TYPE>& o, size_t index)
{
	return AbstractVector::AddVectorAt(o, index);
}

template<class TYPE> inline
void Vector<TYPE>::MakeEmpty()
{
	AbstractVector::MakeEmpty();
}

template<class TYPE> inline
void Vector<TYPE>::RemoveItemsAt(size_t index, size_t count)
{
	AbstractVector::RemoveItemsAt(index, count);
}

template<class TYPE> inline
void Vector<TYPE>::Swap(Vector<TYPE>& o)
{
	AbstractVector::Swap(o);
}

template<class TYPE> inline
void Vector<TYPE>::Push()
{
	AddItem();
}

template<class TYPE> inline
void Vector<TYPE>::Push(const TYPE& item)
{
	AddItem(item);
}

template<class TYPE> inline
TYPE& Vector<TYPE>::EditTop()
{
	const size_t count = CountItems();
//	if (count == 0) debugger(gStaskOverFlow);
	return EditItemAt(count-1);
}

template<class TYPE> inline
const TYPE& Vector<TYPE>::Top() const
{
	const size_t count = CountItems();
//	if (count == 0) debugger(gStaskOverFlow);
	return ItemAt(count-1);
}

template<class TYPE> inline
void Vector<TYPE>::Pop()
{
	const size_t count = CountItems();
//	if (count == 0) debugger(gStaskOverFlow);
	RemoveItemsAt(count-1);
}

template<class TYPE>
void Vector<TYPE>::PerformConstruct(void* base, size_t count) const
{
	Construct(static_cast<TYPE*>(base), count);
}

template<class TYPE>
void Vector<TYPE>::PerformCopy(void* to, const void* from, size_t count) const
{
	Copy(static_cast<TYPE*>(to), static_cast<const TYPE*>(from), count);
}

template<class TYPE>
void Vector<TYPE>::PerformDestroy(void* base, size_t count) const
{
	Destroy(static_cast<TYPE*>(base), count);
}

template<class TYPE>
void Vector<TYPE>::PerformMoveBefore(	void* to, void* from, size_t count) const
{
	MoveBefore(static_cast<TYPE*>(to), static_cast<TYPE*>(from), count);
}

template<class TYPE>
void Vector<TYPE>::PerformMoveAfter(	void* to, void* from, size_t count) const
{
	MoveAfter(static_cast<TYPE*>(to), static_cast<TYPE*>(from), count);
}

template<class TYPE>
void Vector<TYPE>::PerformAssign(	void* to, const void* from, size_t count) const
{
	Assign(static_cast<TYPE*>(to), static_cast<const TYPE*>(from), count);
}

/*-------------------------------------------------------------*/

template<class TYPE> inline
void MoveBefore(Vector<TYPE>* to, Vector<TYPE>* from, size_t count)
{
	MoveBefore(reinterpret_cast<AbstractVector*>(to), reinterpret_cast<AbstractVector*>(from), count);
}

template<class TYPE> inline
void MoveAfter(Vector<TYPE>* to, Vector<TYPE>* from, size_t count)
{
	MoveAfter(reinterpret_cast<AbstractVector*>(to), reinterpret_cast<AbstractVector*>(from), count);
}

template<class TYPE> inline
void Swap(Vector<TYPE>& v1, Vector<TYPE>& v2)
{
	v1.Swap(v2);
}

}	// namespace alp

#endif
