/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright � 2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/

/*!
 * @file unit_test.h
 * 
 * @brief Public header for the ALP Unit Tests. 
 * 
 ******************************************************************************/

#ifndef UNIT_TEST_H
#define UNIT_TEST_H

#ifndef WIN32
#include <dlfcn.h>
#include <stdarg.h>
#endif // WIN32

#include <cppunit/extensions/TestFactoryRegistry.h>
#include <cppunit/extensions/HelperMacros.h>

#ifdef CPPUNIT_MAKE_UNIQUE_NAME
#undef CPPUNIT_MAKE_UNIQUE_NAME
#define CPPUNIT_MAKE_UNIQUE_NAME( prefix ) CPPUNIT_JOIN( prefix##Type__, __LINE__ )
#endif

#define CPPUNIT_REGISTRY_TYPE_AUTO "Auto"
#define CPPUNIT_REGISTRY_TYPE_MANUAL "Manual"
#define CPPUNIT_REGISTRY_TYPE_STRESS "Stress"

CPPUNIT_REGISTRY_ADD_TO_DEFAULT( CPPUNIT_REGISTRY_TYPE_AUTO );
CPPUNIT_REGISTRY_ADD_TO_DEFAULT( CPPUNIT_REGISTRY_TYPE_MANUAL );
CPPUNIT_REGISTRY_ADD_TO_DEFAULT( CPPUNIT_REGISTRY_TYPE_STRESS );

#undef CPPUNIT_MAKE_UNIQUE_NAME
#define CPPUNIT_MAKE_UNIQUE_NAME( prefix ) CPPUNIT_JOIN( autoRegisterRegistry__, __LINE__ )

#define ALP_ROOT_TEST_SUITE_REGISTRATION( registryName, registryType ) \
	extern "C" int AddSuite( void ) \
	{\
		CPPUNIT_NS::TestFactoryRegistry::getRegistry( registryName );\
		return 0;\
	}\
	CPPUNIT_REGISTRY_ADD( registryName, registryType );
	
#define CPPUNIT_NOFORK_TEST( testMethod )	\
    CPPUNIT_TEST_SUITE_ADD_TEST(	\
        ( new CPPUNIT_NS::TestCaller<TestFixtureType>(	\
				context.getTestNameFor( #testMethod),	\
				&TestFixtureType::testMethod,	\
				context.makeFixture() ) ) )

#ifndef WIN32

#define MAX_TEST_TIME_IN_SECONDS 15

#define CPPUNIT_TIMED_TEST( testMethod, timeLimit )	\
	CPPUNIT_TEST_SUITE_ADD_TEST(	\
		( new CPPUNIT_NS::TimeOutTestCaller<TestFixtureType>(	\
				context.getTestNameFor( #testMethod ),	\
				&TestFixtureType::testMethod,	\
				context.makeFixture(),	\
				timeLimit ) ) )

#define CPPUNIT_TIMED_EXCEPTION_TEST( testMethod, timeLimit, ExceptionType )	\
	CPPUNIT_TEST_SUITE_ADD_TEST(	\
		(new CPPUNIT_NS::ExceptionTestCaseDecorator< ExceptionType >(	\
			new CPPUNIT_NS::TimeOutTestCaller< TestFixtureType >(	\
				context.getTestNameFor( #testMethod ),	\
				&TestFixtureType::testMethod,	\
				context.makeFixture(),	\
				timeLimit ) ) ) )

#ifdef CPPUNIT_TEST
#undef CPPUNIT_TEST
#endif
#define CPPUNIT_TEST( testMethod )	\
	CPPUNIT_TIMED_TEST( testMethod, MAX_TEST_TIME_IN_SECONDS )

#ifdef CPPUNIT_TEST_FAIL
#undef CPPUNIT_TEST_FAIL
#endif
#define CPPUNIT_TEST_FAIL( testMethod ) \
	CPPUNIT_TIMED_EXCEPTION_TEST( testMethod, MAX_TEST_TIME_IN_SECONDS, CPPUNIT_NS::Exception )

#ifdef CPPUNIT_TEST_EXCEPTION
#undef CPPUNIT_TEST_EXCEPTION
#endif
#define CPPUNIT_TEST_EXCEPTION( testMethod, ExceptionType )	\
	CPPUNIT_TIMED_EXCEPTION_TEST( testMethod, MAX_TEST_TIME_IN_SECONDS, ExceptionType )

#include <cppunit/Exception.h>
#include <cppunit/TestCase.h>
#include <pthread.h>
#include <signal.h>
#include <sys/errno.h>
#include <sys/shm.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <unistd.h>

#if CPPUNIT_USE_TYPEINFO_NAME
#  include <cppunit/extensions/TypeInfoHelper.h>
#endif

CPPUNIT_NS_BEGIN

class ExceptionData
{
public:
	ExceptionData() {};
	~ExceptionData() {};
	
	void setData( const char *sd )
	{
		m_isValid = true;
		strcpy( m_shortDescription, sd );
	};
	
	void setData( const char *sd, const char *fn, int ln )
	{
		m_isValid = true;
		strcpy( m_shortDescription, sd );
		strcpy( m_fileName, fn );
		m_lineNumber = ln;
	};
	
	void clearData()
	{
		m_isValid = false;
		strcpy( m_shortDescription, "" );
		strcpy( m_fileName, "" );
		m_lineNumber = 0;
	}

	bool m_isValid;
	char m_shortDescription[ 255 ];
	char m_fileName[ 255 ];
	int m_lineNumber;
};

static int shmid;
static ExceptionData *shmptr;

template <class Fixture>
class TimeOutTestCaller : public TestCase
{
	typedef void ( Fixture::*TestMethod )();
    
public:
	// Constructor for TestCaller. This constructor builds a new Fixture
	TimeOutTestCaller( std::string name, TestMethod test, double time ) :
		TestCase( name ), 
		m_exception( NULL ),
		m_ownFixture( true ),
		m_fixture( new Fixture() ),
		m_test( test ),
		m_time( time )
	{
		pthread_mutex_init( &m_test_mutex, NULL );
		pthread_cond_init( &m_test_cond, NULL );
		pthread_setcanceltype( PTHREAD_CANCEL_ASYNCHRONOUS, NULL );
	}

	// Constructor for TimeOutTestCaller. 
	TimeOutTestCaller( std::string name, TestMethod test, Fixture& fixture, double time ) :
		TestCase( name ), 
		m_exception( NULL ),
		m_ownFixture( false ),
		m_fixture( &fixture ),
		m_test( test ),
		m_time( time )
	{
		pthread_mutex_init( &m_test_mutex, NULL );
		pthread_cond_init( &m_test_cond, NULL );
		pthread_setcanceltype( PTHREAD_CANCEL_ASYNCHRONOUS, NULL );
	}

	// Constructor for TimeOutTestCaller.
	TimeOutTestCaller( std::string name, TestMethod test, Fixture* fixture, double time ) :
		TestCase( name ), 
		m_exception( NULL ),
		m_ownFixture( true ),
		m_fixture( fixture ),
		m_test( test ),
		m_time( time )
	{
		pthread_mutex_init( &m_test_mutex, NULL );
		pthread_cond_init( &m_test_cond, NULL );
		pthread_setcanceltype( PTHREAD_CANCEL_ASYNCHRONOUS, NULL );
	}
  
	~TimeOutTestCaller() 
	{
		pthread_cond_destroy( &m_test_cond );
		pthread_mutex_destroy( &m_test_mutex );
		
		if ( m_ownFixture )
			delete m_fixture;
		
		if ( m_exception )
			delete m_exception;
	}

	static void * waitThreadFunc( void *ptr )
	{
		TimeOutTestCaller<Fixture> *pThis = static_cast<TimeOutTestCaller<Fixture> *>( ptr );

		sleep( static_cast<unsigned int>( pThis->m_time ) );
		// test timed out
		kill( pThis->m_test_pid, SIGUSR1 );
		
		pthread_exit( NULL );
	}
	
	void runTest()
	{
		// get a shared memory segment
		if ( ( shmid = shmget( IPC_PRIVATE, sizeof( ExceptionData ), 0666 ) ) < 0 )
		{
			perror( "shmget" );
			exit( 1 );
		}
		shmptr = static_cast<ExceptionData *>( shmat( shmid, 0, 0 ) );
		if ( shmptr == reinterpret_cast<ExceptionData *>( -1 ) )
		{
			perror( "shmat" );
			exit( 1 );
		}
		shmptr->clearData();
		
		if ( ( m_test_pid = fork() ) == 0 )
		{
			// child process			
			try
			{
				// run the actual test
				m_fixture->setUp();
				( m_fixture->*m_test )();
	   		    m_fixture->tearDown();
			}
			catch ( Exception &e )
			{
				// caught CppUnit Exception
				shmptr->setData( e.what(), e.sourceLine().fileName().c_str(), e.sourceLine().lineNumber() );
		    }
		    catch ( std::exception &e )
		    {
				// caught std::exception
				shmptr->setData( e.what() );
		    }
		    catch ( ... )
		    {
		    	// caught unknown exception
				shmptr->setData( "uncaught unknown exception" );
		    }
		    
   		    // test finished
			exit( 0 );
		}
		else if ( m_test_pid > 0 )
		{
			// parent process
			
			// spawning a wait thread
			pthread_t waitThread;
			pthread_create( &waitThread, NULL, &waitThreadFunc, static_cast<void *>( this ) );
			
			int status = 0, exitCode = 0;
			
			// waiting for child process to finish
			pid_t pid = waitpid( m_test_pid, &status, 0 );
			if ( pid == -1 )
				perror( "wait error" );
			
			pthread_cancel( waitThread );
			
			if ( WIFEXITED( status ) )
			{
				// test exited with a code
				exitCode = WEXITSTATUS( status );
			}
			else if ( WIFSIGNALED( status ) )
			{
				// test exited with a signal
				exitCode = WTERMSIG( status );
			}
			
			char buf[255];
			if ( exitCode == SIGUSR1 )
			{
				sprintf( buf, "Test timed out after %i seconds", static_cast<unsigned int>( m_time ) );
				shmptr->setData( buf );
			}
			else if ( exitCode != 0 )
			{
				sprintf( buf, "Test crashed with exit code = %i", exitCode );
				shmptr->setData( buf );
			}
	
			// received process finished signal
			pthread_join( waitThread, NULL );
			
			int retVal = 0;
			waitpid( -1, &retVal, WNOHANG );
	
			if ( shmptr->m_isValid )
				m_exception = new Exception(
					Message( shmptr->m_shortDescription ),
					SourceLine( shmptr->m_fileName, shmptr->m_lineNumber ) );
			
			if ( shmctl( shmid, IPC_RMID, 0 ) < 0 )
				printf( "shmctl error" );
			
			if ( m_exception )
				throw *m_exception;
		}
		else
			printf( "fork error" );
	}

	void setUp()
	{
	}

	void tearDown()
	{
	}

	std::string toString() const
	{
		return "TimeOutTestCaller " + getName();
	}
	
private: 
	TimeOutTestCaller( const TimeOutTestCaller &other );
	TimeOutTestCaller &operator =( const TimeOutTestCaller &other );

	pid_t m_test_pid;
	pthread_mutex_t m_test_mutex;
	pthread_cond_t m_test_cond;
	Exception *m_exception;
	bool m_ownFixture;
	Fixture *m_fixture;
	TestMethod m_test;
	double m_time;
};

class BundleLoader
{
public:
	BundleLoader(const char * bundle1, ...)
	{
		bundlemgr_handle = dlopen("libalp_bundlemgr.so", RTLD_LAZY|RTLD_LOCAL);
		
		if (!bundlemgr_handle) {
			throw Exception(Message(dlerror()));
		}
		
		typedef int (*init_f_t)(void);
		init_f_t init_f = (init_f_t)dlsym(bundlemgr_handle, "alp_bundle_init");
		
		if (!init_f) {
			throw Exception(Message("Unable to find alp_bundle_init symbol"));
		}
		
		if (init_f() != 0) {
			throw Exception(Message("alp_bundle_init returned failure"));
		}
		
		struct index_t {
			unsigned int value;
		};
		
		typedef index_t (*find_by_name_f_t)(const char*);
		find_by_name_f_t find_by_name_f = (find_by_name_f_t)dlsym(bundlemgr_handle, "alp_bundle_by_name");
		
		if (!find_by_name_f) {
			throw Exception(Message("Unable to find alp_bundle_by_name"));
		}

		typedef void * (*open_f_t)(index_t);
		open_f_t open_f = (open_f_t)dlsym(bundlemgr_handle, "alp_bundle_open");
		
		if (!open_f) {
			throw Exception(Message("Unable to find alp_bundle_open"));
		}
		
		va_list ap;
		va_start(ap, bundle1);

		for (const char * bundle = bundle1; bundle; bundle = va_arg(ap, const char*)) {
		
			index_t index = find_by_name_f(bundle);

			if (index.value == 0) {
				throw Exception(Message("Unable to find bundle"));
			}
			void * ref = open_f(index);
			if (ref == 0) {
				throw Exception(Message("Unable to open bundle"));
			}
			
		}	
	}
	
	~BundleLoader()
	{
		typedef int (*fini_f_t)(void);
		fini_f_t fini_f = (fini_f_t)dlsym(bundlemgr_handle, "alp_bundle_fini");
		
		if (!fini_f) {
			throw "Unable to find alp_bundle_fini symbol";
		}
		
		if (fini_f() != 0) {
			throw "alp_bundle_fini returned failure";
		}
		
		dlclose(bundlemgr_handle);
	}

private:
	void * bundlemgr_handle;
		
};

CPPUNIT_NS_END
#endif // WIN32

#endif
