/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/

/*! @file volumesvcs.h
 *
 * @brief Volume Services public header file for ALP Volume Services Functions.
 *
 ********************************************************************************/ 

#ifndef VOLUMESVCS_H
#define VOLUMESVCS_H

#include <hiker/types.h>
#include <hiker/volumesvcs_types.h>

#ifdef __cplusplus
extern "C" {
#endif

/*!
 * @addtogroup ALP_VOLUMESVCS Volume Services
 * @ingroup ALP_SYSTEM
 * @{
 */ 
	
	
/*! @brief Get information on all pluggable devices and any mounted file systems on them
 *
 *  @param[out] fsinfo_array - A poniter to pointer to GArray of fsinfo structs, an array
 *              is allocated
 *
 *	@return ALP_STATUS_OK if data is returned
 *          ALP_STATUS_VOLUMESVCS_INVALID_PARAMETER if the udi parameter is NULL.
 *  @remarks it is up to the caller to free the GArray and the elements (if any) in it
 *  @code
 *  GArray *fsinfo_array;
 *  alp_volumesvcs_fsinfo( &fsinfo_array)
 *  ...
 *  alp-volumesvcs_fsinfo_free( fsinfo_array);
 *  @endcode
 */
extern alp_status_t alp_volumesvcs_fsinfo_array(GArray **fsinfo_array);

/*! @brief Allocates and returns an AlpVolumeSvcsFSinfo* matching the _udi parameter passed in.
 *
 *  @param [out] fsinfo This will be allocated if the return is ALP_STATUS_OK
 *         This must be freed with alp_volumesvcs_fsinfo_free().
 *  @param [in] udi   The UDI to match for.
 *  @return  ALP_STATUS_OK if the UDI is matched.
 *          ALP_STATUS_VOLUMESVCS_NOT_FOUND if there is no match for the UDI.
 *          ALP_STATUS_VOLUMESVCS_INVALID_PARAMETER if the udi parameter is NULL.
 *
 */
extern alp_status_t alp_volumesvcs_fsinfo_for_udi( AlpVolumeSvcsFSinfo **fsinfo, const gchar *udi);

/*! @brief Free the array of AlpVolumeSvcsFSinfo struct's
 *
 *  @param[in] fsinfo_array - A pointer to the GArray returned on a alp_volumesvcs_fsinfo_array
 *       	  call. A distinct call is needed here rather than g_array_free due to allowing
 *       	  freeing of the embeded structures not just the array elements and the array itself.
 *       	  Passing in NULL is allowed and is not an error.
 *
 *  @return  ALP_STATUS_OK always
 */
extern alp_status_t alp_volumesvcs_fsinfo_array_free( GArray *fsinfo_array);

/*! @brief Frees AlpVolumeSvcsFSinfo* allocated by alp_volumesvcs_fsinfo_for_udi().
 *
 *  @param [in] fsinfo - The pointer to free or NULL.
 *
 *  @return ALP_STATUS_OK is always returned.
 */

extern alp_status_t alp_volumesvcs_fsinfo_free( AlpVolumeSvcsFSinfo *fsinfo);

/*! @brief Mount a volume
 *
 *  @param[in] device_udi - the string returned in the device_udi filed of fsinfo that uniquely
 *            specifies a devices, the mount point is derived from the HAL info or the name
 *            associated with this entry via global settings.
 *
 *  @return ALP_STATUS_OK if the volume is (or WAS ALREADY) mounted.
 *          ALP_STATUS_VOLUMESVCS_NOT_FOUND if there is no match for the UDI.
 *          ALP_STATUS_VOLUMESVCS_INVALID_PARAMETER if the udi parameter is NULL.
 *
 * @remarks this may generated an ALP_NOTIFY_VOLUME_ADDED notification
*/
extern alp_status_t alp_volumesvcs_mount(const gchar *device_udi);

/*! @brief UNmount a volume
 *
 *  @param[in] device_udi - same as was specified to UNmount
 *
 *  @return ALP_STATUS_OK if the device is now unmounted (by this call or already being so)
 *          ALP_STATUS_VOLUMESVCS_NOT_FOUND if there is no match for the UDI.
 *          ALP_STATUS_VOLUMESVCS_INVALID_PARAMETER if the udi parameter is NULL.
 *
 * @remarks this may generate an ALP_NOTIFY_VOLUME_REMOVED notification
*/
extern alp_status_t alp_volumesvcs_umount( const gchar *device_udi );

/*! @brief return a list of filesystem types that may be used as input to alp_volumesvcs_makefs
 *  Other formats MAY be supported for mounting a pre-formated volume, but only these may
 *  be used to format a volume ON the ALP device.
 *
 *  @param[in] fstypes - address of a pointer to a GPtrArray
 *
 *  @return - ALP_STATUS_OK in all cases
 *
 *  @remarks the returned array may be freed wtih g_ptr_array_free( fstypes, TRUE )
 */
extern alp_status_t alp_volumesvcs_fstypes( GPtrArray **fstypes);

/*! @brief Initialzie a volume, Create a 1 part partition table and make a msdos file system in it
 *
 *	@param[in] device_udi - the hal UDI for the parent device (e.g. /dev/sdd) to format.
 *	@param[in] name - the user visible name (label) this device should have, this will be the
 *	           default mount point under /media
 *	@param[in] fstype - (must be amonng the strings returned by alp_volumesvcs_fstypes, typically
 *	           FAT16 or FAT32
 *
 *	@return ALP_STATUS_OK if it was found and unmounted
 *          ALP_STATUS_VOLUMESVCS_NOT_FOUND if there is no match for the UDI.
 *          ALP_STATUS_VOLUMESVCS_INVALID_PARAMETER if the udi parameter is NULL.or if the
 *                 format did not complete sucessfuly.
 *          ALP_STATUS_VOLUMESVCS_IN_USE if all invovled volumes can not be unmounted (or are
 *          		not already so).
 *
 *	@remarks Note there is no option to add a partition or have multiple partions, you just get
 *	a single partition table and a new fs in that single partition.  Multiple partition devices
 *	are supported, but they must be formated elsewhere.  If a multiple partitioned device is
 *	targeted by this call all data on all parttions is destroyed.  The newly formated volume
 *	is retunred mounted.
*/
extern alp_status_t alp_volumesvcs_makefs(const gchar *device_udi, const gchar *name, const gchar *fstype);

/*! @brief Mount a directory as if it were a file system (bind mounts)
 *  This entry will be assgined a udi which may be looked up in fsinfo and explicitly
 *  umounted but the entry can not be removed until the volumesvcsd server is restarted.
 *
 *	@param[in] name  - the user visible name (label) this device should have, this will be the
 *	                   default mount point under /media
 *	@param[in] path - the directory to mount on /media/\<name>
 *
 *	@return ALP_STATUS_OK if a new fs was successfully written
 *          ALP_STATUS_VOLUMESVCS_NOT_FOUND if there is no match for the UDI.
 *          ALP_STATUS_VOLUMESVCS_INVALID_PARAMETER if the udi parameter is NULL.or the
 *                   mount is not sucessful for any reason
 */
extern alp_status_t alp_volumesvcs_bind( const gchar *name, const gchar *path);

/*! @brief Reanme a volume.  The volume name (part of the vfat format) is changed.  The volume
 *  can be mounted or unmounted when the call is made.  If mounted it will be UNmounted
 *  as part of the call.  The renamed volume is always mounted on return from this call
 *  (barrying errors that prevent such).
 *
 *	@param[in] udi - the volume to rename
 *	@param[in] name  - the new user visible name (label) this device should have, this will be the
 *	                   default mount point under /media
 *
 *	@return ALP_STATUS_OK if a new fs was successfully written
 *          ALP_STATUS_VOLUMESVCS_NOT_FOUND if there is no match for the UDI.
 *          ALP_STATUS_VOLUMESVCS_INVALID_PARAMETER if the udi parameter is NULL.
 *          ALP_STATUS_VOLUMESVCS_IN_USE if the volume is mounted and for some reason
 *          		can not be unmounted
 */
extern alp_status_t alp_volumesvcs_rename_volume( const char *udi, const gchar *name);

/*!
 @}
*/


#ifdef __cplusplus
}       // extern "C"
#endif

#endif

