/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/

/*!
 * @file volumesvcs_types.h
 * 
 * @brief Volume Services public header file for types. 
 *
 *****************************************************************************/

#ifndef VOLUMESVCS_TYPES_H
#define VOLUMESVCS_TYPES_H

#include <glib.h>
#include <hiker/syspaths.h>

#ifdef __cplusplus
extern "C" {
#endif

/*!
 * @addtogroup ALP_VOLUMESVCS Volume Services
 * @ingroup ALP_SYSTEM
 * @{
 */
	
//! operation completed sucessfully
#define ALP_STATUS_VOLUMESVCS_OK                 ALP_STATUS_OK
//! bad input 
#define ALP_STATUS_VOLUMESVCS_INVALID_PARAMETER ( ALP_CLASS_VOLUMESVCS | 0x01)
//! target device or volume could not be found
#define ALP_STATUS_VOLUMESVCS_NOT_FOUND         ( ALP_CLASS_VOLUMESVCS | 0x02)
//! target device or bindmount could not be unmounted as it is in active use
#define ALP_STATUS_VOLUMESVCS_DEVICE_BUSY       ( ALP_CLASS_VOLUMESVCS | 0x03)

/*!
 * @remarks
 * Alp Notification events are just strings representing each event type
 * the following notifications are broadcast with
 * ALP_NOTIFY_PRIORITY_NORMAL
 * ALP_NOTIFY_QUEUE_DEFAULT
 */
//! a card has been inserted and the volumes on it mounted
#define ALP_NOTIFY_EVENT_VOLUME_ADDED "/alp/volumesvcs/volume_added"
//! a card with mounted volumes has been removed
#define ALP_NOTIFY_EVENT_VOLUME_REMOVED "/alp/volumesvcs/volume_removed"
//! a card has been inserted that is either blank or the file system on it ws not recognized
#define ALP_NOTIFY_EVENT_DEVICE_ADDED "/alp/volumesvcs/device_added"
//! a card was removed that did not contain mounted file systems
#define ALP_NOTIFY_EVENT_DEVICE_REMOVED "/alp/volumesvcs/device_removed"

/*! @remark
 * global settings names, setup by the device vendor to name the media slots
 * on the device e.g.
 *   /system/slots/0/name = "Expansion Card"
 *   /system/slots/0/hal_key = "info.parent"
 *   /system/slots/0/hal_value = "/org/freedesktop/Hal/devices/platform_pxa2xx_mci_mmc_card_rca1"
 * Which would cause all cards with the specifed parent to get a name of "Expansion Card"
 */
//! the head of the tree for slot names
#define ALP_SYSTEM_SLOTS_PREFIX "/dev/slots/"
//! the name to be used for this slot
#define ALP_SYSTEM_SLOTS_NAME "/name"
//! the /dev/name to match against (may be incomplete /dev/mmcblk matches /dev/mmcblk0...)
#define ALP_SYSTEM_SLOTS_DEV_PATH "/dev_path"

//! if this exists the volume is not exposed to users
#define ALP_VOLUMESVCS_HIDEVOL_PATH ALP_EXT_SYSPATH_HIDDEN_VOL

/*!
 * @remarks
 * The detail argument for Alp notifications for volume services are simply the character
 * strings that are the UDI for the event in question (found in the udi filed of an
 * AlpVolumesvcsFSinfo struct).  These typedefs are just for convienence.
 */
//! notification detail for a volume added event
typedef char* AlpNotifyEventVolumeAdded;
//! notification detail for a volume removed event
typedef char* AlpNotifyEventVolumeRemoved;
//! notification detail for a device added event
typedef char* AlpNotifyEventDeviceAdded;
//! notification detail for a device removed event
typedef char* AlpNotifyEventdeviceRemoved;

/*! @brief File System info block, AlpVolumesvcsFSinfo returns a GArray of these
 */
typedef struct _AlpVolumeSvcsFSinfo {
	    gchar *udi;                //!< present if a pluggable device holds this storage NULL otherwise
		gchar *dev_path;           //!< /dev path to this device if pluggable NULL otherwise
	    gchar *fs_path;            //!< present if the storage is accessible as a file system NULL otherwise
	    gchar *name;               //!< optional label on the volume (e.g.myphotos)
		gchar *slot;               //!< optional vendor supplied name of the slot (e.g. top)
		gchar *vendor_model;       //!< optional string of manufactuer & model of the device
		gchar *uuid;               //!< a string unique to this storage if pluggable NULL otherwise
	    gchar *card_type;          //!< only advisory but if possible will be USB, MMC, SF, etc.
	    guint64 total_space;       //!< total storage size in bytes (mounted volumes only)
	    guint64 used_space;        //!< used space in bytes (mounted volumes only)
	    guint64 total_files;       //!< total allowed files on this storage (mounted volumes only)
	    guint64 used_files;        //!< existing files on this storage (mounted volumes only)
		guint32 flags;             //!< flags for this entry as defined below
		guint index;               //!< index of this entry in the GArray as first received
		                           //!< i.e. user inserts or removes will render such inaccurate
} AlpVolumeSvcsFSinfo;

//! entry represents a bind mount not necessarily a real device
#define ALP_VOLUMESVCS_FLAGS_BINDMOUNT 0x1
//! entry represents a partition with or witout fs as opposed to an entire disk (disk are only shown if there are no partitions on them).
#define ALP_VOLUMESVCS_FLAGS_PARTITION 0x2
//! entry represents an entry to be hidden from users
#define ALP_VOLUMESVCS_FLAGS_HIDDEN 0x4

/*!
 * @remark
 * Volume Services (in combination with Bundle Manager) will perform certain actions based on
 * an ACCESS/card.xml file. Given contents like this:
 *
 * @code
 * <card>
 *   <OnInsert>
 *     <copy>bar:com.access.apps.bunny</copy>
 *     <launch argument="--alp-primary">bar:com.access.apps.xcalc</launch>
 *   </OnInsert>
 *   <OnRemove>
 *     <delete>bar:com.access.apps.bunny</delete>
 *     <launch>bar:fred</launch>
 *   </OnRemove>
 * </card>
 * @endcode
 *
 * @remarks
 * On insertion of such a card, 'bunny' will be copied into the internal store, and 'xcalc' (from any store)
 * will be launched. (If no argument is supplied, the argument for launching on insert or remove are
 * <tt>--alp-insert-card</tt>, or <tt>--alp-remove-card</tt>, respectively.
 * On removal of the card, bunny will be deleted from the internal store, and 'fred' will be launched.
 *
 */

/*!
 @}
 */

#ifdef __cplusplus
}       // extern "C"
#endif

#endif

