/********************************************************************

Copyright  2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/**
 * @file alarm_client_tests.cpp
 *
 * @brief Alarm Manager client API unit tests
 *
 *****************************************************************************/
#include <hiker/unit_test.h>
#include <cppunit/extensions/HelperMacros.h>

// faster CPPUNIT_ASSERT
#define ALP_CPPUNIT_ASSERT(x) if (!(x)) CPPUNIT_ASSERT(x)

#include <hiker/alarmmgr.h>

#include <signal.h>

class AlarmMgrTests : public CPPUNIT_NS::TestFixture
{
	CPPUNIT_TEST_SUITE(AlarmMgrTests);
//	CPPUNIT_TEST(StartupAlarmMgrServer);
	CPPUNIT_TEST(SetAlarmTest);
	CPPUNIT_TEST(UpdateAlarmTest);
	CPPUNIT_TEST(CancelAlarmTest);
	CPPUNIT_TEST(MultipleAlarmTest);
	CPPUNIT_TEST(GetAlarmTest);
	CPPUNIT_TEST(InvalidAppIdTest);
	CPPUNIT_TEST(InvalidAlarmTimeTest);
    CPPUNIT_TEST(GetAlarmNegativeTest);
	
	//CPPUNIT_TEST(TeardownAlarmMgrServer);
	CPPUNIT_TEST_SUITE_END();

protected:
	void StartupAlarmMgrServer();
	void TeardownAlarmMgrServer();
	void SetAlarmTest();
	void UpdateAlarmTest();
	void CancelAlarmTest();
	void MultipleAlarmTest();
	void GetAlarmTest();
	void InvalidAppIdTest();
	void InvalidAlarmTimeTest();
    void GetAlarmNegativeTest();
public:
			
private:
	static pid_t sAlarmMgrServerPID;	
};

pid_t AlarmMgrTests::sAlarmMgrServerPID = 0;

ALP_ROOT_TEST_SUITE_REGISTRATION("UI-006", CPPUNIT_REGISTRY_TYPE_AUTO);
CPPUNIT_TEST_SUITE_NAMED_REGISTRATION(AlarmMgrTests, "UI-006");



// copied from the appserver
void AlarmMgrTests::StartupAlarmMgrServer()
{
	char *top;
	char *argv[2];


	// First do some stuff that might fail, since we can't assert after forking.
	top = getenv("ALP_RUN_TOP");
	CPPUNIT_ASSERT(top != NULL);
	
	// Fork child process
	sAlarmMgrServerPID = fork();
	
	if(sAlarmMgrServerPID == 0)
	{
		// Child process, exec the alarmmgr server.
		int size = strlen(top);
		size += strlen("/opt/alp/bin/alarmmgrd") + 1;
	
		char *runme;
		runme = (char *)malloc(sizeof(char) * size);
		strcpy(runme, top);
		strcat(runme, "/opt/alp/bin/alarmmgrd");
	
		argv[0] = "alarmmgrd";
		argv[1] = NULL;
		execv(runme, argv);
		
		// execution should never reach here...
		CPPUNIT_ASSERT(false);

		// exit if the exev fails
		exit(0);
	} else {
		// Parent process. Sleep for a moment to give the server a chance to start up.
		sleep(2);
	}
	
}

void AlarmMgrTests::TeardownAlarmMgrServer()
{
	// remove alarmmgr server
	if (sAlarmMgrServerPID) kill(sAlarmMgrServerPID, SIGKILL);


}

/*********
 * @brief  API Tested: alp_alm_set_alarm, alp_alm_get_alarm
 *
 * @description Set an alarm for a certain time and verify alp_alm_get_alarm gets the correct time.
 *
 * ******/

void AlarmMgrTests::SetAlarmTest()
{
	alp_status_t err;
	time_t now;
        guint32	expected, alarm;

	// set 1st alarm
	now = time(NULL);
	expected = now + 20;
	printf ("The current local UTC time is %s, and also is %ld in seconds\n",
            asctime(localtime(&now)), (gulong)now);
	err = alp_alm_set_alarm("bar:com.access.apps.duck", 1, expected);
	CPPUNIT_ASSERT(err == ALP_STATUS_OK);

	err = alp_alm_get_alarm("bar:com.access.apps.duck", 1, &alarm);
	CPPUNIT_ASSERT(err == ALP_STATUS_OK);
	CPPUNIT_ASSERT(expected == alarm);

	// cancel alarm
	err = alp_alm_set_alarm("bar:com.access.apps.duck", 1, 0);
	CPPUNIT_ASSERT(err == ALP_STATUS_OK);	
}

/********
 * @brief API Tested: alp_alm_set_alarm, alp_get_alarm
 *
 * @description  Set an alarm and update it.  Verify that getting the alarm gets the correct, updated alarm. 
 *
 * ******/
void AlarmMgrTests::UpdateAlarmTest()
{
	alp_status_t err;
	time_t now = time(NULL);
	guint32 expected, alarm, updatedAlarm;

	// set alarm
	expected = now + 60;
	err = alp_alm_set_alarm("bar:com.access.apps.duck", 2, expected);
	CPPUNIT_ASSERT(err == ALP_STATUS_OK);

	// verify the current alarm
	err = alp_alm_get_alarm("bar:com.access.apps.duck", 2, &alarm);
	CPPUNIT_ASSERT(err == ALP_STATUS_OK);
	CPPUNIT_ASSERT(expected == alarm);
	
	// update alarm & verify
	updatedAlarm = now + 5;
	err = alp_alm_set_alarm("bar:com.access.apps.duck", 2, updatedAlarm);
	CPPUNIT_ASSERT(err == ALP_STATUS_OK);
	err = alp_alm_get_alarm("bar:com.access.apps.duck", 2, &alarm);
	CPPUNIT_ASSERT(err == ALP_STATUS_OK);
	CPPUNIT_ASSERT(updatedAlarm == alarm);

	// cancel alarm
	err = alp_alm_set_alarm("bar:com.access.apps.duck", 2, 0);
	CPPUNIT_ASSERT(err == ALP_STATUS_OK);	
}

/********
 * @brief API Tested: alp_alm_set_alarm, alp_get_alarm
 *
 * @description  Set an alarm and cancel it.  Verify that the alarm is canceled with alp_alm_get_alarm and verify that the
 * application is not launched.
 *
 * ******/
void AlarmMgrTests::CancelAlarmTest()
{
	alp_status_t err;
	time_t now = time (NULL);
	guint32 expected, alarm;

	// set alarm
	expected = now + 10;
	err = alp_alm_set_alarm("bar:com.access.apps.duck", 3, expected);
	CPPUNIT_ASSERT(err == ALP_STATUS_OK);

	err = alp_alm_get_alarm("bar:com.access.apps.duck", 3, &alarm);
	CPPUNIT_ASSERT(err == ALP_STATUS_OK);
	CPPUNIT_ASSERT(expected == alarm);

	// cancel alarm
	err = alp_alm_set_alarm("bar:com.access.apps.duck", 3, 0);
	CPPUNIT_ASSERT(err == ALP_STATUS_OK);	

	err = alp_alm_get_alarm("bar:com.access.apps.duck", 3, &alarm);
	CPPUNIT_ASSERT(err == ALP_STATUS_OK);	
	CPPUNIT_ASSERT(0 == alarm);
}

/**
 * @brief API Tested: alp_alm_set_alarm, alp_get_alarm
 *
 * Set two alarms, cancel one and update one.  Verify that the canceled
 * alarm does not proceed, verify that the updated alarm acts at the
 * appropriate time.
 *
 */
void AlarmMgrTests::MultipleAlarmTest()
{
	alp_status_t err;
	time_t now;
	guint32 expected, alarm;

    now = time(NULL);

	// set alarm
	err = alp_alm_set_alarm("bar:com.access.apps.duck", 4, now + 60);
	CPPUNIT_ASSERT(err == ALP_STATUS_OK);

	// set 2nd alarm
	err = alp_alm_set_alarm("bar:com.access.apps.duck", 5, now + 50);
	CPPUNIT_ASSERT(err == ALP_STATUS_OK);

	// update 2nd alarm
	err = alp_alm_set_alarm("bar:com.access.apps.duck", 5, now + 5);
	CPPUNIT_ASSERT(err == ALP_STATUS_OK);

	// cancel 1st alarm
	err = alp_alm_set_alarm("bar:com.access.apps.duck", 4, 0);
	CPPUNIT_ASSERT(err == ALP_STATUS_OK);	

    // cancel 2nd alarm
    err = alp_alm_set_alarm("bar:com.access.apps.duck", 5, 0);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
}

/**
 * @brief  API Tested: alp_alm_get_alarm
 *
 * Reset the alarm for ref1 and verify the correct time is returned by
 * alp_alm_get_alarm.
 *
 */
void AlarmMgrTests::GetAlarmTest()
{
	alp_status_t err;
	time_t now = time(NULL);
	guint32	alarm;	
	guint32 expected = now + 30; 

	// reset 1st alarm
	err = alp_alm_set_alarm("bar:com.access.apps.duck", 1, expected);
	CPPUNIT_ASSERT(err == ALP_STATUS_OK);
	
	// get 1st alarm
	err = alp_alm_get_alarm("bar:com.access.apps.duck", 1, &alarm);
	CPPUNIT_ASSERT(err == ALP_STATUS_OK);
	CPPUNIT_ASSERT(expected == alarm);

    // cancel alarm
	err = alp_alm_set_alarm("bar:com.access.apps.duck", 1, 0);
	CPPUNIT_ASSERT(err == ALP_STATUS_OK);	
}

/**
 * @brief  API Tested: alp_alm_get_alarm, alp_alm_set_alarm
 *
 * Attempt to set an alarm with an invalid appID.
 *
 */
void AlarmMgrTests::InvalidAppIdTest()
{
	alp_status_t err;
	time_t now = time(NULL);
	guint32 alarm, expected;
    gchar *appID;
    gint i;

    appID = g_new(gchar, ALP_ALARM_MAX_APPID_SIZE + 10);
    for (i = 0; i < ALP_ALARM_MAX_APPID_SIZE + 10 - 1; i++) {
        appID[i] = 'a';
    }
    appID[ALP_ALARM_MAX_APPID_SIZE + 10 - 1] = '\0';

	//set the alarm
	expected = now + 30;
	err = alp_alm_set_alarm(0, 1, expected);
	CPPUNIT_ASSERT(err == ALP_STATUS_ALARM_BAD_PARAM);

	err = alp_alm_set_alarm(appID, 1, expected);
	CPPUNIT_ASSERT(err == ALP_STATUS_ALARM_BAD_PARAM);

	//get the alarm
	err = alp_alm_get_alarm(0, 6, &alarm);
	CPPUNIT_ASSERT(err == ALP_STATUS_ALARM_BAD_PARAM);

	err = alp_alm_get_alarm(appID, 1, &alarm);
	CPPUNIT_ASSERT(err == ALP_STATUS_ALARM_BAD_PARAM);
}

/**
 * @brief  API Tested: alp_alm_set_alarm
 *
 * Attempt to set an alarm that has passed.
 *
 */
void AlarmMgrTests::InvalidAlarmTimeTest()
{
	alp_status_t err;
	time_t now = time(NULL);
	guint32 alarm, expected;

	//set the alarm
	expected = now - 30;
	err = alp_alm_set_alarm("bar:com.access.apps.test", 1, expected);
	CPPUNIT_ASSERT(err == ALP_STATUS_ALARM_BAD_PARAM);
}

/**
 * @brief  API Tested: alp_alm_get_alarm
 *
 * Attempt to get an alarm setting with invalid parameters
 *
 */
void AlarmMgrTests::GetAlarmNegativeTest()
{
	alp_status_t err;
	guint32 expected;

    /* check invalid appid (i.e. NULL) */
	err = alp_alm_get_alarm (NULL, 1, &expected);
	CPPUNIT_ASSERT(err == ALP_STATUS_ALARM_BAD_PARAM);

    /* check invalid alarm time pointer */
    err = alp_alm_get_alarm ("bar:com.access.apps.duck", 1, NULL);
	CPPUNIT_ASSERT(err == ALP_STATUS_ALARM_BAD_PARAM);
}
