/********************************************************************

Copyright  2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/*! @file appserverclient.c
	This file contains the client-side APIs & implementation for the MAX Application server.
*/

// Uncomment the next line to disable traces AT COMPILE TIME 
#define ALP_TRACE ALP_TRACE_OFF		//Enable/disable


#include <signal.h>
#include <hiker/pdk/appmgr_pdk.h>
#include <hiker/ipc.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <hiker/bundlemgr.h>
#include <stdlib.h>

#include <hiker/prv/appmgr_prv.h>

/*
	TODO:
	
	
	Cookie Cutter Stuff Done Elsewhere:
	Start up cookie cutter
	init IPC connection with server
	register as launchpad
	handle launch requests, set up environment for app using app client APIs
	Implement debug - block on SIGUSR1 
	Deal with calling actual app
	Send unregistration ("graceful exit") message after Main returns (this should use a private API)
	
	
	Possible new client-side APIs:	
	
	
	Consider Doing Later:
	Allocate app ID dynamically, and at the right size so we don't waste memory
	
*/


/*
	Private structures:
*/

typedef struct _AlpExitHandlerInfo AlpExitHandlerInfo;
struct _AlpExitHandlerInfo{
	AlpExitHandler	handler;
	gpointer		cbData;
};

struct _AlpAppContext{
	AlpDiagnosticHandler diagHandler;
	void 			*diagCBData;
	
	AlpRelaunchHandler relaunchHandler;
	gpointer		 relaunchCBData;
	GList		*exitHandlers;
	char			appID[PKG_ID_MAX_LEN];
	char 		*scheme;
	AlpConnection*	appServerConnection;
	GMainLoop		*blockingLoop; // FIXME: can this be removed?
	guint32		appLaunchFlags;
	
	gboolean		waitingForLaunchReply;
	guint		timeoutSourceID; // FIXME: can this be removed?
	alp_status_t	launchReplyStatus; // FIXME: can this be removed?
	pid_t		launchReplyPID; // FIXME: can this be removed?
	AlpThemeUpdateProc themeUpdateProc;
	unsigned int	lastSecurityID;
};




/*
	Prototypes
*/
static alp_status_t prv_send_async_request(AlpAppContext *ctxt, AlpMessage *msg);

static alp_status_t prv_app_launch_full(AlpAppContext* context, char *appID, int argc, char* argv[], pid_t *outAppPID);

static void prv_app_receive_cb(AlpConnection* connection, AlpMessage* msg, gpointer cbData);
//void PrvChannelConnectCB (AlpConnection* connection, gpointer cbData);
static void prv_execute_exit_request(AlpAppContext* context);

static alp_status_t prv_set_backgrounding(AlpAppContext *ctxt, AlpAppBackgroundingType newBackgrounding);

static alp_status_t prv_remove_exit_handler(AlpExitHandler procP, 
								    gpointer cbData, gboolean checkData);

alp_status_t prv_watcher_stuff(const char *appID, AlpWatcherRef *ioWatcherRef, gboolean *outAppRan, gboolean remove, gint blockMode);

/*
	Globals
*/
AlpAppContext *gDefaultAppContext = NULL;

/*
static AlpDiagnosticHandler gDiagHandler = NULL;
static void *gDiagCBData = NULL;
static AlpRelaunchHandler gRelaunchHandler = NULL;
static gpointer		 gRelaunchCBData = NULL;
static GList			*gExitHandlers = NULL;
static char			gAppID[PKG_ID_MAX_LEN];
static AlpConnection*	gAppServerConnection;
static GMainLoop		*gBlockingLoop = NULL;
static guint32			gAppLaunchFlags = ALP_APP_LAUNCH_FLAGS_DEFAULT;
*/



/* ***************************************************
	Public SDK APIs
   *************************************************** */



alp_status_t alp_app_launch(const char *appID, int argc,  char *const argv[], pid_t *outPID)
{
	pid_t dummyPID;
	if(outPID == NULL) outPID = &dummyPID;
	
	return alp_app_launch_full(gDefaultAppContext, appID, argc, argv, outPID);
}


alp_status_t alp_app_launch_async(const char *appID, int argc,  char *const argv[])
{
	return alp_app_launch_full(gDefaultAppContext, appID, argc, argv, NULL);
}



alp_status_t alp_app_launch_full(AlpAppContext* context, const char *appID, int argc,  char *const argv[], pid_t *outPID)
{
	alp_status_t err = ALP_OK;
	alp_status_t launchStatus;
	char *replyAppID;
	pid_t appPID;
	gint msgID;
	AlpMessage *msg, *reply;
	
	
	// Can't work with NULL context
	if(context == NULL) return ALP_STATUS_APP_NO_CONTEXT;
	
	// DOLATER - what sort of parameter verification should we do here?
	
	// Already waiting for a reply from some launch.  Only one at a time, please.
	if(context->timeoutSourceID != 0)
	{
		ALP_FAIL_NON_FATAL_IF(TRUE, "Already waiting for launch request!");
		return ALP_STATUS_APP_IPC; 
	}
	
	
	msg = alp_app_prv_make_launch_msg(appID, argc, argv, (outPID != NULL), ALP_APP_DEFAULT_SECURITY_ID);
	if(msg==NULL)
	{
		ALP_FAIL_NON_FATAL_IF(TRUE, "Can't pack message!\n", NULL);
		return ALP_STATUS_APP_IPC;
	}
	
	
	// Send launch request to the app server
//	err = alp_ipc_connection_send(context->appServerConnection, msg);
	reply = NULL;
	
	if(outPID == NULL) err = alp_ipc_connection_send(context->appServerConnection, msg);
	else
	{
		err = alp_ipc_connection_send_with_response(context->appServerConnection, 
				msg, 7000, &reply);
		
		if(err != ALP_STATUS_OK || reply == NULL)
		{
			ALP_FAIL_NON_FATAL_IF(err != ALP_STATUS_OK 
						&& err != ALP_STATUS_IPC_TIMEOUT, 
						"launch message failed to send, err=0x%lx\n", err);
			ALP_FAIL_NON_FATAL_IF(reply == NULL 
						&& err != ALP_STATUS_IPC_TIMEOUT, 
						"launch request failed - NULL reply.\n", NULL);
			err = ALP_STATUS_APP_IPC;
		}
		else
		{
			msgID = alp_ipc_message_get_message_ID(reply);
			if(msgID != ALP_APP_MSG_ID_LAUNCH_REPLY)
			{
				ALP_FAIL_NON_FATAL_IF(TRUE, "Got msg ID=%d instead of a launch reply.\n", msgID);
				err = ALP_STATUS_APP_IPC;
			}
			else
			{
		
				err = alp_app_prv_unpack_launch_reply_msg(reply, &replyAppID, &launchStatus, &appPID);
				g_free(replyAppID);	
				if(err == ALP_OK) err = launchStatus;
				
				if(outPID != NULL) *outPID = appPID;
			}
			
			alp_ipc_message_destroy(reply);
		}
	}
		
	
	alp_ipc_message_destroy(msg);
	
	return err;
}




alp_status_t alp_app_exit(const char *appID)
{
	return prv_send_async_request(gDefaultAppContext, alp_app_prv_make_exit_msg(appID));
}


alp_status_t alp_app_retire(const char *appID)
{
	return prv_send_async_request(gDefaultAppContext, alp_app_prv_make_retire_msg(appID));
}



alp_status_t alp_app_watcher_add(const char *appID, AlpWatcherRef *outWatcherRef, gboolean *outAppRan)
{
	if(outAppRan == NULL) return ALP_STATUS_APP_PARAM;
	*outWatcherRef = 0; // set the watcherRef to 0, so a new one will be allocated.
	return prv_watcher_stuff(appID, outWatcherRef, outAppRan, FALSE, ALP_APP_WATCHER_IGNORE);
}

alp_status_t alp_app_watcher_check(AlpWatcherRef watcherRef, gboolean *outAppRan)
{
	if(outAppRan == NULL) return ALP_STATUS_APP_PARAM;
	if(watcherRef == 0) return ALP_STATUS_APP_PARAM;
	return prv_watcher_stuff("", &watcherRef, outAppRan, FALSE, ALP_APP_WATCHER_IGNORE);
}

alp_status_t alp_app_watcher_remove(AlpWatcherRef watcherRef, gboolean *outAppRan)
{
	if(watcherRef == 0) return ALP_STATUS_APP_PARAM;
	return prv_watcher_stuff("", &watcherRef, outAppRan, TRUE, ALP_APP_WATCHER_IGNORE);
}

alp_status_t alp_app_watcher_block(AlpWatcherRef watcherRef, gboolean blockLaunch)
{
	gint blockMode;
	
	if(watcherRef == 0) return ALP_STATUS_APP_PARAM;
	
	if(blockLaunch) blockMode = ALP_APP_WATCHER_BLOCK;
	else blockMode = ALP_APP_WATCHER_UNBLOCK;
	
	return prv_watcher_stuff("", &watcherRef, NULL, FALSE, blockMode);
}



alp_status_t alp_app_current(char **outPkgID)
{

	alp_status_t err = ALP_OK;
	alp_status_t launchStatus;
	char *replyAppID;
	gint msgID;
	AlpMessage *msg, *reply;
	AlpAppContext *context = gDefaultAppContext;
	
	// Can't work with NULL context
	if(context == NULL) return ALP_STATUS_APP_NO_CONTEXT;
	
	// Need to be able to return result.
	if(outPkgID == NULL) return ALP_STATUS_APP_PARAM;
	
	// Pack up the request:
	msg = alp_app_prv_make_cur_app_msg(NULL, "");
	if(msg == NULL) return ALP_STATUS_APP_IPC;
	
	// Send request to the app server
	reply = NULL;
	err = alp_ipc_connection_send_with_response(context->appServerConnection, 
			msg,
			3000,
			&reply);
	alp_ipc_message_destroy(msg);
	
	ALP_FAIL_NON_FATAL_IF(reply == NULL, "Received NULL CurApp reply.");
	
	// Error check reply:
	if(err != ALP_STATUS_OK || reply == NULL)
		return ALP_STATUS_APP_IPC;
	
	msgID = alp_ipc_message_get_message_ID(reply);
	if(msgID != ALP_APP_MSG_ID_CUR_APP)
		return ALP_STATUS_APP_IPC;
	
	// Unpack response.
	err = alp_app_prv_unpack_cur_app_msg(reply, outPkgID);
	ALP_FAIL_NON_FATAL_IF(*outPkgID == NULL, "Received NULL appID.");
	
	alp_ipc_message_destroy(reply);
	
	return err;
}




alp_status_t alp_app_set_backgrounding(AlpAppBackgroundingType newBackgrounding)
{
	return prv_set_backgrounding(gDefaultAppContext, newBackgrounding);
}

alp_status_t alp_app_set_relaunch_handler(AlpRelaunchHandler procP, gpointer cbData)
{
	AlpAppContext *ctxt = gDefaultAppContext;
	if(gDefaultAppContext == NULL) return ALP_STATUS_APP_NO_CONTEXT;
	ctxt->relaunchHandler = procP;
	ctxt->relaunchCBData = cbData;
	
	return ALP_STATUS_OK;
}

alp_status_t alp_app_add_exit_handler(AlpExitHandler procP, gpointer cbData)
{
	AlpAppContext *ctxt = gDefaultAppContext;
	AlpExitHandlerInfo *info;

	if(gDefaultAppContext == NULL) return ALP_STATUS_APP_NO_CONTEXT;
	
	info = g_new0(AlpExitHandlerInfo, 1);
	info->handler = procP;
	info->cbData = cbData;
	
	ctxt->exitHandlers = g_list_prepend(ctxt->exitHandlers, info);
	
	return ALP_STATUS_OK;
}


alp_status_t alp_app_remove_exit_handler(AlpExitHandler procP)
{
	return prv_remove_exit_handler(procP, NULL, FALSE);
}

alp_status_t alp_app_remove_exit_handler_with_data(AlpExitHandler procP, gpointer cbData)
{
	return prv_remove_exit_handler(procP, cbData, TRUE);
}


/* ***************************************************
	PDK APIs - for use by launchpads/cookiecutters, etc,
   *************************************************** */

/*
	Initialize the IPC connection with the app server, and set it up for default client-app type message handling.
	Called by launchpads in the context of a newly forked application process, before calling the app's entrypoint.
*/

AlpAppContext* alp_app_setup(const char *appID, guint32 flags, const char* scheme)
{
	AlpConnection *connection;
	AlpAppContext *appContext;
	alp_status_t err;
	
	if(appID == NULL) return NULL;
	
	ALP_TM(TL(ALP_APP_CLIENT_TRACE, "Setting up app %s.", appID));
	
	// Connect to server, wait a max of 1 sec.
	err = alp_ipc_channel_connect_synch(ALP_APP_SERVER_CHANNEL, ALP_IPC_CONNECTION_USAGE_HINT_NONE, 7000, &connection);
	
	if(err != ALP_STATUS_OK) 
	{
		ALP_TM(TL(ALP_APP_CLIENT_TRACE, "Connect failed, err=0x%lx", (unsigned long)err));
		return NULL;
	}
	
	
	// Allocate & initialize new app context:
	appContext = g_new0(AlpAppContext, 1);
	appContext->appLaunchFlags = flags;
//	appContext->blockingLoop = g_main_loop_new(NULL, TRUE);
	appContext->appServerConnection = connection;
	err = alp_ipc_connection_register_receive_callback(connection, 
			prv_app_receive_cb, appContext);
	
	ALP_FAIL_FATAL_IF(err != ALP_STATUS_OK, 
					  "AppClientLib: Receive CB registration failed.\n", NULL);
		
	// Set it as the default context, if there is none currently:
	if(gDefaultAppContext == NULL) gDefaultAppContext = appContext;
	
	// Register it with appserver, per our parameters:
	alp_app_register(appContext, appID, flags, scheme);
	
	return appContext;
}




void alp_app_destroy(AlpAppContext* context, int result)
{
	alp_status_t err = ALP_OK;  
	
	if(context->appServerConnection == NULL) return;
	
	ALP_TM(TL(ALP_APP_CLIENT_TRACE, "%s: Unregistering with appserver.", context->appID));
	
	// Unregister with appserver:
	err = prv_send_async_request(context, alp_app_prv_make_unregister_msg(result));
	 
	ALP_FAIL_NON_FATAL_IF(err != ALP_STATUS_OK, 
					  "AppClientLib: unregistration message failed to send for %s.\n", context->appID);
		
	// Close IPC connection & deallocate context.
	alp_app_abandon(context);
	
	return;
}


void alp_app_abandon(AlpAppContext* context)
{
	AlpMessage *msg;
	alp_status_t status = ALP_OK;  
	
	if(context == NULL) context = gDefaultAppContext;
	if(context->appServerConnection == NULL) return;
		
	ALP_TM(TL(ALP_APP_CLIENT_TRACE, "%s: Disconnecting from appserver.", context->appID));
	
	// Close the channel
	status = alp_ipc_channel_disconnect(context->appServerConnection);
	ALP_FAIL_NON_FATAL_IF(status, "alp_ipc_channel_disconnect failed.\n", NULL);
	
	// Unset the default context to eliminate possible stale pointers.
	if(gDefaultAppContext == context) gDefaultAppContext = NULL;
	
	// Empty list of exit handlers:
	GList *node = context->exitHandlers;
	while(node != NULL)
	{
		g_free(node->data);
		node = g_list_remove(node, node->data);
	}
	
	// If timeout is installed, remove it.
	if(context->timeoutSourceID > 0)
		g_source_remove(context->timeoutSourceID);
	
	// If there's a scheme, deallocate it:
	if(context->scheme) 
		g_free(context->scheme);	
	
	// Deallocate context
	if(context->blockingLoop) 
		g_main_loop_unref(context->blockingLoop);
	
	g_free(context);
	
	return;
}

void alp_app_set_default_context(AlpAppContext* context)
{
	gDefaultAppContext = context;
	return;
}


void alp_app_register(AlpAppContext* context, const char *appID, guint32 flags, const char* scheme)
{
	pid_t appPID;
	AlpMessage *msg;
	alp_status_t err = ALP_OK;
	int schemelen;
	
	// Verify connection:
	if(context->appServerConnection == NULL)
	{
		ALP_FAIL_NON_FATAL_IF(TRUE, "No connection to app server.\n", NULL);
		return;
	}
	

	// Check sticky flags - are any being cleared?
	guint32 oldStickyFlags = context->appLaunchFlags & ALP_APP_LAUNCH_FLAGS_STICKY;
	guint32 newStickyFlags = flags & ALP_APP_LAUNCH_FLAGS_STICKY;
	
	ALP_FAIL_NON_FATAL_IF((newStickyFlags & oldStickyFlags) != oldStickyFlags, 
					"Can't clear sticky launch flags.\n");
	
	
	// Check scheme parameter, make sure it's in sync with the launchpad flag,
	// and any previously registered scheme:
	if(scheme == NULL) scheme = "";
	schemelen = strlen(scheme);
	
	if((flags & ALP_APP_LAUNCH_FLAG_LAUNCHPAD) && schemelen == 0)
	{
		ALP_FAIL_NON_FATAL_IF(TRUE, "Can't register launchpad without scheme.\n", NULL);
		flags &= ~ALP_APP_LAUNCH_FLAG_LAUNCHPAD;
	}
	else if((flags & ALP_APP_LAUNCH_FLAG_LAUNCHPAD) == 0 && schemelen > 0)
	{
		ALP_FAIL_NON_FATAL_IF(TRUE, "Can't register scheme for non-launchpad.\n", NULL);
		scheme = "";
	}
	
	if(context->scheme != NULL && 
		  (scheme == NULL || strcmp(context->scheme, scheme) != 0))
	{
		ALP_FAIL_NON_FATAL_IF(TRUE, "Can't reregister for different scheme.\n", NULL);
		return;
	}
	
	
	
	// Copy app ID into our internal buffer
	// In the case of prv_execute_relaunch, we're actually called with
	// appID==context->appID, and although harmless, this makes valgrind
	// complain, so we check for it and don't do the copy in that case.
	if(appID != context->appID)
	{
		strncpy(context->appID, appID, PKG_ID_MAX_LEN);
		if(context->appID[PKG_ID_MAX_LEN-1] != '\0')	// The app ID should never be allowed to be longer than will fit in our buffer.
		{
			ALP_FAIL_NON_FATAL_IF(TRUE, "AppClientLib: Registration from invalid app name: %s (max  of %d bytes).\n", appID, PKG_ID_MAX_LEN);
			context->appID[PKG_ID_MAX_LEN-1] = '\0';
		}
	}
	
	// If a scheme is specified, copy that too:
	if(context->scheme == NULL && scheme != NULL)
		context->scheme = g_strdup(scheme);
	
	// Save flags too:
	context->appLaunchFlags = flags;
	
	appPID = getpid();
	
	// Send the request:
	prv_send_async_request(context, alp_app_prv_make_register_msg(appID, appPID, flags, scheme));
	
	
	return;
}

alp_status_t alp_app_flush(AlpAppContext* context)
{
	alp_status_t err = ALP_OK;
	alp_status_t launchStatus;
	char *replyAppID;
	pid_t appPID;
	gint msgID;
	AlpMessage *msg, *reply;
	
	
	// Can't work with NULL context
	if(context == NULL) return ALP_STATUS_APP_NO_CONTEXT;
		
	msg = alp_app_prv_make_flush_msg(NULL);
	if(msg == NULL) return ALP_STATUS_APP_IPC;
	
	// Send flush to the app server
	reply = NULL;
	err = alp_ipc_connection_send_with_response(context->appServerConnection, 
			msg,
			3000,
			&reply);
	alp_ipc_message_destroy(msg);
	
	ALP_FAIL_NON_FATAL_IF(reply == NULL, "Received NULL flush reply.", NULL);
	
	// Error check reply:
	if(err != ALP_STATUS_OK || reply == NULL)
		return ALP_STATUS_APP_IPC;
	
	msgID = alp_ipc_message_get_message_ID(reply);
	if(msgID != ALP_APP_MSG_ID_FLUSH)
		return ALP_STATUS_APP_IPC;
		
	alp_ipc_message_destroy(reply);
	
	return err;
}


guint32 alp_app_parse_launch_flags(guint32 launchFlags, int argc, char *const argv[])
{
	const char*	str;
	int		i=0;
	
	while(i<argc)
	{
		str = argv[i];
		
		// If primary or backgrounded are specified, then we need to set the primary bit appropriately.
		if((strcmp(str, ALP_APP_PRIMARY)==0) ||
				  (strcmp(str, "-alp-primary") == 0)) /* FIXME: backwards compatibility */
				launchFlags |= ALP_APP_LAUNCH_FLAG_PRIMARY;
		else if(strcmp(str, ALP_APP_BACKGROUNDED)==0) 
				launchFlags &= ~ALP_APP_LAUNCH_FLAG_PRIMARY;
		
		// If the transient argument is specified, we set the transient bit.
		// Like debug, below, this is "sticky" and not ever cleared if we are relaunched.
		else if(strcmp(str, ALP_APP_TRANSIENT)==0)
				launchFlags |= ALP_APP_LAUNCH_FLAG_TRANSIENT;
		
		// If debugging, just set the debug bit.
		// This is never cleared so that a debugged app will always reregister with the 
		// debug bit set.  This ensures that any relaunches won't cause it to appear 
		// that it is no longer being debugged.
		else if((strcmp(str, ALP_APP_DEBUG)==0) ||
					  (strcmp(str, "-alp-debug") == 0)) /* FIXME: backwards compatibility */
				launchFlags |= ALP_APP_LAUNCH_FLAG_DEBUG;
		
		i++;
	}
	
	return launchFlags;
}


void alp_app_set_diagnostic_handler(AlpAppContext *ctxt, AlpDiagnosticHandler procP, void* cbData)
{
	ctxt->diagHandler = procP;
	ctxt->diagCBData = cbData;
}

/* ***************************************************
	Jump List APIs.
   *************************************************** */

alp_status_t alp_app_get_jump_list(int *count, char***appIDs, gboolean **isRunnings)
{
	AlpAppContext *ctxt = gDefaultAppContext;	
	alp_status_t err = ALP_OK;
	gint msgID;
	AlpMessage *msg, *reply;
	
	if(ctxt == NULL) return ALP_STATUS_APP_NO_CONTEXT;
	
	msg = alp_app_prv_make_jump_list_request_msg();
	if(msg == NULL) return ALP_STATUS_APP_IPC;
	
	// Send flush to the app server
	reply = NULL;
	err = alp_ipc_connection_send_with_response(
						ctxt->appServerConnection, 
						msg, 3000, &reply);
	alp_ipc_message_destroy(msg);
	
	ALP_FAIL_NON_FATAL_IF(reply == NULL, "Received NULL jumplist reply.");
	
	// Error check reply:
	if(err != ALP_STATUS_OK || reply == NULL)
		return ALP_STATUS_APP_IPC;
	
	msgID = alp_ipc_message_get_message_ID(reply);
	if(msgID != ALP_APP_MSG_ID_JUMP_LIST)
		return ALP_STATUS_APP_IPC;
	
	err = alp_app_prv_unpack_jump_list_msg(reply, count, appIDs, isRunnings);
	
	alp_ipc_message_destroy(reply);
	
	return err;
}





void alp_app_free_argv(int argc, char* argv[])
{	
	while(argc > 0)
	{
		argc--;
		if(argv[argc]) g_free(argv[argc]);
	}
	
	g_free(argv);
	
	return;
}


void alp_app_debug_break(void)
{
	char argstr[48];
	
	sprintf(argstr, "--command=gdb nlpd %d", getpid());	
	
	if(fork() == 0)
	{
		execl("/usr/bin/gnome-terminal", "gnome-terminal", argstr, (char*)NULL);
		g_error("Can't start up gnome-terminal!");
		exit(1);
	}
	else
	{
		// "wait" for debugger to attach.
		sleep(5);
	}
	
	return;
}


alp_status_t alp_app_theme_update(void)
{
	return prv_send_async_request(gDefaultAppContext,
				alp_app_prv_make_theme_update_msg());
}

void alp_app_set_theme_update_proc(AlpAppContext *context, AlpThemeUpdateProc proc)
{
	if(context != NULL) context->themeUpdateProc = proc;
	else if(gDefaultAppContext != NULL) gDefaultAppContext->themeUpdateProc = proc;
	
	return;
}


alp_status_t alp_app_reset_system(void)
{
	return prv_send_async_request(gDefaultAppContext,
							alp_app_prv_make_reset_msg());
}

unsigned int alp_app_get_security_id(AlpAppContext *context)
{
	return context->lastSecurityID;
}

/* ***************************************************
	Internal implementation
*************************************************** */

// Private implementation of the handler for a theme update message.
// alp_app_theme_update eventually results in all apps receiving one of 
// these messages, and this routine does the actual work of reloading the theme.
static void prv_do_theme_update(AlpAppContext *context)
{
	if(context->themeUpdateProc != NULL)
		context->themeUpdateProc();
}

/*
Private implementation bottleneck for all the watcher APIs.
This more complex API is used to implement the four simpler exposed APIs.
The purpose of the exposed APIs is basically to provide a more understandable
& intuitive API.  Meanwhile, the implementation can use a more convenient single message format
for all communication.


@param[in] appID The package ID the application to watch for, may be empty if the watcherRef is specified.
@param[in/out] ioWatcherRef Pointer to Ref for the watcher.  Use a ref set to 0 in order to allocate a new one.
@param[out] outAppRan Set to TRUE if the target application has run since the watcher was added.
@param[in] remove Pass TRUE if the indicated watcher should be deallocated.  Do not use in combination with *ioWatcherRef = 0.
@param[in] blockMode This is a tri-state value that indicates whether or not the watcher should block launches, or not, or remain unchanged in this respect.

@return Yields ALP_STATUS_OK on success.
*/
alp_status_t prv_watcher_stuff(const char *appID, AlpWatcherRef *ioWatcherRef, gboolean *outAppRan, gboolean remove, gint blockMode)
{
	alp_status_t err = ALP_OK;
	alp_status_t result;
	char *replyAppID = NULL;
	pid_t appPID;
	gint msgID;
	AlpMessage *msg, *reply;
	AlpAppContext *ctxt = gDefaultAppContext;	
	
	ALP_FAIL_NON_FATAL_IF(*ioWatcherRef == 0 && remove == TRUE, 
				   "Can't allocate & deallocate in one step!");
	if(*ioWatcherRef == 0 && remove == TRUE) return ALP_STATUS_APP_PARAM;
	
	// Can't work with NULL context
	if(ctxt == NULL) return ALP_STATUS_APP_NO_CONTEXT;
	
	// Compose request to appserver:
	msg = alp_app_prv_make_watcher_msg(NULL, appID, *ioWatcherRef, FALSE, remove, ALP_OK, blockMode);
	if(msg == NULL) return ALP_STATUS_APP_IPC;
	
	// Send request to the app server
	reply = NULL;
	err = alp_ipc_connection_send_with_response(ctxt->appServerConnection, 
			msg,
			3000,
			&reply);
	alp_ipc_message_destroy(msg);
	
	ALP_FAIL_NON_FATAL_IF(reply == NULL, "Received NULL watcher reply.", NULL);
	
	// Error check reply:
	if(err != ALP_STATUS_OK || reply == NULL)
		return ALP_STATUS_APP_IPC;
	
	msgID = alp_ipc_message_get_message_ID(reply);
	if(msgID != ALP_APP_MSG_ID_WATCHER)
		return ALP_STATUS_APP_IPC;
	
	
	err = alp_app_prv_unpack_watcher_msg(reply, &replyAppID, ioWatcherRef, outAppRan, &remove, &result, NULL);
	if(replyAppID) g_free(replyAppID);
	
	if(err == ALP_OK)
	{
		// if we unpack okay, then we want to return the requests result code.
		err = result;
	}
		
	
	
	alp_ipc_message_destroy(reply);
	
	
	
	return err;
}



/*
	Client side code for handling a relaunch request.
	If the app has registered a relaunch handler, then all we need to do is call it.
	Otherwise, we "fake" an exit request.  When we exit, the app server will notice that 
	we were pending a relaunch & launch us again with the new parameters.
*/
void prv_execute_relaunch(AlpAppContext* context, AlpMessage *msg)
{
	int argc, i;
	char ** argv;
	guint32 flags;
	gboolean relaunchingUs;
	gboolean waitForPID;
	gboolean debugLaunch;
	unsigned int secID;
	
	alp_status_t err = ALP_OK;
	
	ALP_TM(TL(ALP_APP_CLIENT_TRACE, "Client %s received relaunch.", context->appID));
	
	// Unpack arguments, do some sanity checking.
	err = alp_app_prv_unpack_launch_msg(msg, &argc, &argv, &waitForPID, &secID);
	
	if(err)
	{
		ALP_FAIL_NON_FATAL_IF(TRUE, "AppClient: received malformed relaunch request.\n", NULL);
		return;
	}
	
	flags = context->appLaunchFlags;
	relaunchingUs = (strcmp(argv[0], context->appID) == 0);
	if((flags & ALP_APP_LAUNCH_FLAG_LAUNCHPAD) == 0 && !relaunchingUs)
	{
		ALP_FAIL_NON_FATAL_IF(TRUE, "AppClient: received mislaid relaunch request for %s (we're %s).\n", argv[0], context->appID);
		return;
	}
	
	// Save security ID in case we need it:
	context->lastSecurityID = secID;
	
	// Update launch flags for new args
	flags = alp_app_parse_launch_flags(flags, argc, argv);
	
	// Check to see if this is a debug launch.
	// If so, we never pass it on to the app, we always want to start fresh.
	// So, we just treat it like an exit request (same as no relaunch handler)
	// and the appserver will launch a new instance of us when we exit.		
	debugLaunch = FALSE;
	for(i=0; i<argc; i++)
	{
		if(strcmp(argv[i], ALP_APP_DEBUG)==0) 
		{
			debugLaunch = TRUE;
			break;
		}
	}
	
	
	// If there is a relaunch handler, and we're not being launched by a debugger,
	// then we pass the launch arguments into the app.  
	if(context->relaunchHandler != NULL && 
		  (relaunchingUs == FALSE || debugLaunch == FALSE))
	{
		ALP_TM(TL(ALP_APP_CLIENT_TRACE, "%s: calling relaunch handler.", context->appID));
		// re-registration is necessary to update flags on server side, and clear pending launch status.
		// However, if this a launch request for a launchpad, then we don't want to register the new app, since the launchpad code is responsible handle that.
		if(relaunchingUs) alp_app_register(context, context->appID, flags, context->scheme);
		
		// Let the app know that the launch arguments have changed.
		context->relaunchHandler(argc, argv, context->relaunchCBData);
	}
	else
	{
		ALP_TM(TL(ALP_APP_CLIENT_TRACE, "%s: synthesizing exit request.", context->appID));
		// ... Otherwise, treat it like an exit request.
		prv_execute_exit_request(context);
	}
	
	alp_app_free_argv(argc, argv);
	
	ALP_TM(TL(ALP_APP_CLIENT_TRACE, "%s: relaunch complete.", context->appID));
	
	return;
}

/*
	Client-side routine for handling exit requests sent from the app server.
	If the application has registered any exit handlers, then we execute them, 
	and assume the app will terminate on its own.
	Otherwise, we explicitly unregister and then send ourselves a SIGTERM.  
	The unregister ensures that our termination will be considered clean, since
	it's a valid choice for the developer to just receive a sigterm instead.
	
	The app server will notice if we 
	do not exit within a few seconds, and will send us a SIGKILL.
*/
void prv_execute_exit_request(AlpAppContext* context)
{
	GList *list; 
	AlpExitHandlerInfo *info;
	int status;
	
	ALP_TM(TL(ALP_APP_CLIENT_TRACE, "%s: Exit request.", context->appID));
	
	if(context->exitHandlers != NULL)
	{
		list = context->exitHandlers;
		while(list != NULL)
		{
			// Obtain ptr directly to the handler info structure... 
			info = list->data;

			// ... execute the handler... 
			info->handler(info->cbData);
			
			// ... and check the next element in the list.
			list = g_list_next(list);
		}
	}
	else
	{
		// Send unregister message to appserver, so if our process terminates 
		prv_send_async_request(context, alp_app_prv_make_unregister_msg(0));
		
		// We don't call alp_app_destroy(), since that would close the IPC connection,
		// and disable the appserver's kill timer.  This way, the connection remains 
		// open until the process terminates, which means appserver will still notice 
		// and kill us if we hang.
		
		status = raise(SIGTERM);
		ALP_FAIL_NON_FATAL_IF(status != 0, "AppClient: raise(SIGTERM) failed.\n", NULL);
	}
	
	return;
}

/*
	Private version of alp_app_set_backgrounding(), which accepts a pointer to an app context.
*/
alp_status_t prv_set_backgrounding(AlpAppContext *ctxt, AlpAppBackgroundingType newBackgrounding)
{
	return prv_send_async_request(ctxt, alp_app_prv_make_backgrounding_msg(newBackgrounding));
}


// bottleneck routine for the remove exit handler APIs.
// This code handles both cases.
alp_status_t prv_remove_exit_handler(AlpExitHandler procP, 
							 gpointer cbData, gboolean checkData)
{
	AlpAppContext *ctxt = gDefaultAppContext;
	GList *list; 
	AlpExitHandlerInfo *info;
	
	if(gDefaultAppContext == NULL) return ALP_STATUS_APP_NO_CONTEXT;
	
	list = ctxt->exitHandlers;
	while(list != NULL)
	{
		// Obtain ptr directly to the handler info structure:
		info = list->data;
		
		// If the procPtr matches, 
		// and we're either not checking data, or the data matches too, 
		// then we have found our match and can remove the handler.
		if(info->handler == procP && (checkData == FALSE || info->cbData == cbData))
			break;
		
		list = g_list_next(list);
	}
	
	if(list == NULL) return ALP_STATUS_APP_NOT_FOUND;
	
	// Deallocate list item * info structure:
	info = list->data;
	ctxt->exitHandlers = g_list_remove(ctxt->exitHandlers, info);
	g_free(info);
	
	
	return ALP_STATUS_OK;
}



void prv_app_receive_cb(AlpConnection* connection, AlpMessage* msg, gpointer cbData)
{
	AlpAppContext *context = (AlpAppContext*)cbData;
	gint msgID;
	char *appID;
	char *diagStr;
	alp_status_t err = ALP_OK;
	alp_status_t launchStatus;
	pid_t appPID;
//	gint argc;
//	char *argv;
//	guint32 flags;
	
	msgID = alp_ipc_message_get_message_ID(msg);
	
	ALP_TM(TL(ALP_APP_CLIENT_TRACE, "Client %s: received msg ID %d", context->appID, msgID));
	
	// We received a message from the server.  Handle the request.
	switch(msgID)
	{
		
//	Should never be received by app client... only AppServer.
//		case ALP_MSG_ID_REGISTER:
//			break;
		
		case ALP_APP_MSG_ID_LAUNCH:
			prv_execute_relaunch(context, msg);
			break;
		
		case ALP_APP_MSG_ID_LAUNCH_REPLY:
			ALP_FAIL_NON_FATAL_IF(TRUE, "Received misplaced launch reply.\n", NULL);
			break;
			
		case ALP_APP_MSG_ID_EXIT:
				
			ALP_TM(TL(ALP_APP_CLIENT_TRACE, "Client %s received exit request.", context->appID));
			
			// Unpack arguments, do some sanity checking.			
			err = alp_app_prv_unpack_exit_msg(msg, &appID);
			if(err)
			{
				ALP_FAIL_NON_FATAL_IF(TRUE, "AppClient: received malformed exit request.\n", NULL);
				break;
			}
			
			ALP_FAIL_NON_FATAL_IF(appID != NULL && strlen(appID) > 0 && 
									strcmp(appID, context->appID) != 0, 
							  "AppClient: received mislaid exit request for %s (we're %s).\n", appID, context->appID);
			
			if(appID != NULL) g_free(appID);
			
			prv_execute_exit_request(context);
			
			break;
		
		case ALP_APP_MSG_ID_DIAGNOSTIC:
			
			if(context->diagHandler != NULL)
			{				
				if(alp_app_prv_unpack_diagnostic_msg(msg, &appID, &diagStr))
				{
					ALP_FAIL_NON_FATAL_IF(TRUE, 
									  "AppClient: received malformed diagnostic msg.\n", NULL);
					break;
				}
				
				context->diagHandler(appID, diagStr, context->diagCBData);
				g_free(appID);
				g_free(diagStr);
			}
						
			break;
			
		case ALP_APP_MSG_ID_THEME_UPDATE:
			prv_do_theme_update(context);
			break;
						
		default:
			// DOLATER: handle error
			ALP_FAIL_NON_FATAL_IF(TRUE, "AppClient: message %d received, but not handled.\n", msgID);
			break;
	}
	
	// DOLATER: Deallocate the message, since we're all done with it.
	alp_ipc_message_destroy(msg);
	
	return;
}


/*
	Utility routine to encapsulate the boiler-plate logic around sending a simple
	asynchronous request to the appserver.
	Also deallocates the message.
*/	
static alp_status_t prv_send_async_request(AlpAppContext *ctxt, AlpMessage *msg)
{
	alp_status_t err = ALP_STATUS_OK;
	
	if(ctxt == NULL || ctxt->appServerConnection == NULL) 
		return ALP_STATUS_APP_NO_CONTEXT;
	
	if(msg)
	{
		err = alp_ipc_connection_send(ctxt->appServerConnection, msg);
		ALP_FAIL_NON_FATAL_IF(err != ALP_OK, 
				  "AppClientLib: message ID %d failed to send for %s, error 0x%lx.\n",
				  alp_ipc_message_get_message_ID(msg), ctxt->appID, (long)err);
		
		alp_ipc_message_destroy(msg);
	}
	else err = ALP_STATUS_APP_IPC;
	
	return err;
}


