/********************************************************************

Copyright  2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/*! @file appservermsg.c
	This file contains the message-packing APIs for the MAX Application server.
 */

// Uncomment the next line to disable traces AT COMPILE TIME 
#define ALP_TRACE ALP_TRACE_OFF		//Enable/disable


#include <errno.h>

#include <hiker/pdk/appmgr_pdk.h>
#include <hiker/prv/appmgr_prv.h>
#include <hiker/ipc.h>
#include <string.h>
#include <sys/types.h>
#include <stdlib.h>

static AlpMessage* prv_make_msg_with_string(gint msgID, const char *str);
static alp_status_t prv_unpack_msg_with_string(AlpMessage* msg, gint msgID, char** str);
static AlpMessage* prv_make_empty_msg(guint msgID);

/* ***************************************************
	Message packing & unpacking utility routines.
   *************************************************** */


/*!
If appID is non-null, then it is packed into the message as argv[0], argc is incremented, and the remaining parameters follow.
If appID is null, then argc & argv are packed just as passed to this routine.

@param[in] msg The AlpMessage to pack data into.
@param[in] appID The name of the package containing the application.  Becomes argv[0].
@param[in] argc The number of strings in the argv array.
@param[in] argv An array of char* containing launch arguments for the application.

@return Yields ALP_STATUS_OK on success, or an error code if there is a problem packing the message.
 */
alp_status_t alp_app_prv_pack_argv(AlpMessage *msg, const char* appID, int argc,  char *const argv[])
{
	alp_status_t err = ALP_OK;
	int i;
	
	// If we need to prepend the app's package ID, then increment argc and do so.
	// This will appear as just another argv parameter when unpacked.
	if(appID != NULL)
	{
		err = alp_ipc_message_pack_int32(msg, argc+1);
		if(err == ALP_OK)
			alp_ipc_message_pack_string(msg, appID);
	}
	else err = alp_ipc_message_pack_int32(msg, argc);
	
	
	for(i=0; i<argc && err == ALP_OK; i++)
	{
		err = alp_ipc_message_pack_string(msg, argv[i]);
	}
	
	return err;
}


/*!

@param[in] msg The AlpMessage to pack data into.
@param[out] outArgc Receives the number of strings in the argv array.
@param[out] outArgv Receives a newly allocated array of char* containing launch arguments for the application.

@return Yields ALP_STATUS_OK on success, or an error code if there is a problem unpacking the message.
 */
alp_status_t alp_app_prv_unpack_argv(AlpMessage *msg, int *outArgc, char*** outArgv)
{
	alp_status_t err = ALP_OK;
	int i;
	int argc;
	char **argv;
	
	err = alp_ipc_message_unpack_int32(msg, &argc);
	if(err) return ALP_STATUS_APP_IPC;
	
	argv = g_new0(char*, argc+1);
	if(argv == NULL) return ALP_STATUS_APP_MEM;
	
	for(i=0; i<argc && err == ALP_OK; i++)
	{
		err = alp_ipc_message_unpack_string(msg, &argv[i]);
	}
	
	if(err == ALP_OK)
	{
		*outArgc = argc;
		*outArgv = argv;
		argv[argc] = NULL;
	}
	else
	{
		*outArgc = 0;
		*outArgv = NULL;
		g_free(argv);
	}
	
	if(err) err = ALP_STATUS_APP_MSG;
	return err;
}



/*!

@param[in] appID The name of the package containing the application to launch, or NULL if it's contained in argv[0].
@param[in] argc The number of strings in the argv array.
@param[in] argv An array of char* containing launch arguments for the application.

@return Returns the newly allocated message, or NULL on failure.
 */
AlpMessage* alp_app_prv_make_launch_msg(const char *appID, int argc,  char *const argv[], gboolean waitForPID, unsigned int securityID)
{
	AlpMessage *msg = alp_ipc_message_create();
	alp_status_t err = ALP_OK;
	
	if(msg == NULL) err = ALP_STATUS_APP_MEM;

	if(!err) err = alp_ipc_message_pack_start(msg, ALP_APP_MSG_ID_LAUNCH);
	if(!err) err = alp_app_prv_pack_argv(msg, appID, argc, argv);
	if(!err) err = alp_ipc_message_pack_int32(msg, (int32_t) waitForPID);
	if(!err) err = alp_ipc_message_pack_uint32(msg, (uint32_t) securityID);
	if(!err) err = alp_ipc_message_pack_end(msg);
	if(err)
	{
		ALP_FAIL_NON_FATAL_IF(TRUE, "pack failed!\n", NULL);
		if(msg != NULL) alp_ipc_message_destroy(msg);
		msg = NULL;
	}
	
	return msg;
}

/*!

@param[in] msg The AlpMessage to unpack data from.
@param[out] outArgc Receives the number of strings in the argv array.
@param[out] outArgv Receives a newly allocated array of char* containing launch arguments for the application.

@return Yields ALP_STATUS_OK on success, or an error code if there is a problem unpacking the message.
 */
alp_status_t alp_app_prv_unpack_launch_msg(AlpMessage* msg, int *argc, char*** argv, gboolean *waitForPID, unsigned int *outSecurityID)
{
	alp_status_t err = ALP_OK;
	int32_t waitp;
	
	if(alp_ipc_message_get_message_ID(msg) != ALP_APP_MSG_ID_LAUNCH)
		return ALP_STATUS_APP_MSG;
	
	err = alp_ipc_message_unpack_start(msg);
	if(!err) err = alp_app_prv_unpack_argv(msg, argc, argv);
	if(!err) err = alp_ipc_message_unpack_int32(msg, &waitp);
	if(!err) err = alp_ipc_message_unpack_uint32(msg, outSecurityID);
	if(!err) alp_ipc_message_unpack_end(msg);
	
	*waitForPID = (gboolean)waitp;
	
	if(err) err = ALP_STATUS_APP_MSG;
	return err;
}

/*!

@param[in] appID The name of the package containing the application to exit, or NULL.

@return Returns the newly allocated message, or NULL on failure.
 */
AlpMessage* alp_app_prv_make_exit_msg(const char *appID)
{
	return prv_make_msg_with_string(ALP_APP_MSG_ID_EXIT, appID);
}

/*!

@param[in] msg The AlpMessage to unpack data from.
@param[out] appID Receives the name of the package containing the application to exit.

@return Yields ALP_STATUS_OK on success, or an error code if there is a problem unpacking the message.
 */
alp_status_t alp_app_prv_unpack_exit_msg(AlpMessage* msg, char** appID)
{
	return prv_unpack_msg_with_string(msg, ALP_APP_MSG_ID_EXIT, appID);
}

/*!

@param[in] appID The name of the package containing the application to retire, or NULL

@return Returns the newly allocated message, or NULL on failure.
 */
AlpMessage* alp_app_prv_make_retire_msg(const char *appID)
{
	return prv_make_msg_with_string(ALP_APP_MSG_ID_RETIRE, appID);
}

/*!

@param[in] msg The AlpMessage to unpack data from.
@param[out] appID Receives the name of the package containing the application to retire.

@return Yields ALP_STATUS_OK on success, or an error code if there is a problem unpacking the message.
 */
alp_status_t alp_app_prv_unpack_retire_msg(AlpMessage* msg, char** appID)
{
	return prv_unpack_msg_with_string(msg, ALP_APP_MSG_ID_RETIRE, appID);
}



/*!

@param[in] appID The name of the package containing the application.
@param[in] appPID The process ID of the newly spawned app process.
@param[in] flags Standard launch flags indicating the app's mode of execution.
@param[in] lpScheme The package type for received launch requests.  Used only when registering as a launchpad.

@return Returns the newly allocated message, or NULL on failure.
 */
AlpMessage* alp_app_prv_make_register_msg(const char *appID, pid_t appPID, guint32 flags, const char *lpScheme)
{
	AlpMessage *msg;
	alp_status_t err = ALP_OK;
	
	if(lpScheme == NULL) lpScheme = "";
	
	msg = alp_ipc_message_create();
	if(msg == NULL) err = ALP_STATUS_APP_MEM;
	if(!err) err = alp_ipc_message_pack_start(msg, ALP_APP_MSG_ID_REGISTER);
	if(!err) err = alp_ipc_message_pack_string(msg, appID);
	if(!err) err = alp_ipc_message_pack_int32(msg, appPID);
	if(!err) err = alp_ipc_message_pack_uint32(msg, flags);
	if(!err) err = alp_ipc_message_pack_string(msg, lpScheme); // scheme for launchpads
	if(!err) err = alp_ipc_message_pack_end(msg);
	
	if(err)
	{
		ALP_FAIL_NON_FATAL_IF(TRUE, "pack failed!\n", NULL);
		msg = NULL;
	}
	
	return msg;
}

/*!

@param[in] msg The AlpMessage to unpack data from.
@param[out] appID The name of the package containing the application.
@param[out] appPID The process ID of the newly spawned app process.
@param[out] flags Standard launch flags indicating the app's mode of execution.
@param[out] lpScheme The package type for received launch requests.  Used only when registering as a launchpad.

@return Yields ALP_STATUS_OK on success, or an error code if there is a problem unpacking the message.
 */
alp_status_t alp_app_prv_unpack_register_msg(AlpMessage* msg, char** appID, pid_t *appPID, guint32 *flags, char**lpScheme)
{
	alp_status_t err = ALP_OK;
	
	if(alp_ipc_message_get_message_ID(msg) != ALP_APP_MSG_ID_REGISTER)
		return ALP_STATUS_APP_MSG;
	
	err = alp_ipc_message_unpack_start(msg);
	if(!err) err = alp_ipc_message_unpack_string(msg, appID);
	if(!err) err = alp_ipc_message_unpack_int32(msg, appPID);
	if(!err) err = alp_ipc_message_unpack_uint32(msg, flags);
	if(!err) err = alp_ipc_message_unpack_string(msg, lpScheme);
	if(!err) err = alp_ipc_message_unpack_end(msg);
	
	if(err) err = ALP_STATUS_APP_MSG;
	return err;
}



/*!

@param[in] result The app process's exit/result code.

@return Returns the newly allocated message, or NULL on failure.
 */
AlpMessage* alp_app_prv_make_unregister_msg(int result)
{
	AlpMessage *msg;
	alp_status_t err = ALP_OK;
		
	msg = alp_ipc_message_create();
	if(msg == NULL) err = ALP_STATUS_APP_MEM;
	if(!err) err = alp_ipc_message_pack_start(msg, ALP_APP_MSG_ID_UNREGISTER);
	if(!err) err = alp_ipc_message_pack_int32(msg, result);
	if(!err) err = alp_ipc_message_pack_end(msg);
	
	if(err)
	{
		ALP_FAIL_NON_FATAL_IF(TRUE, "pack failed!\n", NULL);
		if(msg != NULL) alp_ipc_message_destroy(msg);
		msg = NULL;
	}
	
	return msg;
}


/*!

@param[in] msg The AlpMessage to unpack data from.
@param[out] result Receives the app process's exit/result code.

@return Yields ALP_STATUS_OK on success, or an error code if there is a problem unpacking the message.
 */
alp_status_t alp_app_prv_unpack_unregister_msg(AlpMessage* msg, int *result)
{
	alp_status_t err = ALP_OK;
	
	if(alp_ipc_message_get_message_ID(msg) != ALP_APP_MSG_ID_UNREGISTER)
		return ALP_STATUS_APP_MSG;
	
	err = alp_ipc_message_unpack_start(msg);
	if(!err) err = alp_ipc_message_unpack_int32(msg, result);
	if(!err) err = alp_ipc_message_unpack_end(msg);
	
	if(err) err = ALP_STATUS_APP_MSG;
	return err;
}

/*!

@param[in] appID The name of the package containing the application to which the message pertains.
@param[in] str The diagnostic message to include.

@return Returns the newly allocated message, or NULL on failure.
 */
AlpMessage* alp_app_prv_make_diagnostic_msg(const char *appID, const char *str)
{
	AlpMessage *msg;
	alp_status_t err = ALP_OK;
		
	msg = alp_ipc_message_create();
	if(msg == NULL) err = ALP_STATUS_APP_MEM;
	if(!err) err = alp_ipc_message_pack_start(msg, ALP_APP_MSG_ID_DIAGNOSTIC);
	if(!err) err = alp_ipc_message_pack_string(msg, appID);
	if(!err) err = alp_ipc_message_pack_string(msg, str);
	if(!err) err = alp_ipc_message_pack_end(msg);
	
	if(err)
	{
		ALP_FAIL_NON_FATAL_IF(TRUE, "pack failed!\n", NULL);
		if(msg != NULL) alp_ipc_message_destroy(msg);
		msg = NULL;
	}
	
	return msg;
}

/*! 

@param[in] msg The AlpMessage to unpack data from.
@param[out] appID Receives the name of the package containing the application.
@param[out] str Receives the diagnostic message string.

@return Yields ALP_STATUS_OK on success, or an error code if there is a problem unpacking the message.
 */
alp_status_t alp_app_prv_unpack_diagnostic_msg(AlpMessage* msg, char **appID, char **str)
{
	alp_status_t err = ALP_OK;
	
	if(alp_ipc_message_get_message_ID(msg) != ALP_APP_MSG_ID_DIAGNOSTIC)
		return ALP_STATUS_APP_MSG;
	
	err = alp_ipc_message_unpack_start(msg);
	if(!err) err = alp_ipc_message_unpack_string(msg, appID);
	if(!err) err = alp_ipc_message_unpack_string(msg, str);
	if(!err) err = alp_ipc_message_unpack_end(msg);
	
	if(err) err = ALP_STATUS_APP_MSG;
	return err;
}

/*!

@param[in] replyTo The launch request message to reply to.
@param[in] appID The name of the package containing the application to which the message pertains.
@param[in] launchErr The error result code from the launch operation.
@param[in] appPID The process ID of the newly launched application, when applicable (<=0 otherwise).

@return Returns the newly allocated message, or NULL on failure.
 */
AlpMessage* alp_app_prv_make_launch_reply_msg(AlpMessage* replyTo, const char *appID, alp_status_t launchErr, pid_t appPID)
{
	AlpMessage *msg;
	alp_status_t err = ALP_OK;
	
	msg = alp_ipc_message_create();
	if(msg == NULL) err = ALP_STATUS_APP_MEM;
	if(!err) err = alp_ipc_message_response_pack_start(msg, ALP_APP_MSG_ID_LAUNCH_REPLY, replyTo);
	if(!err) err = alp_ipc_message_pack_string(msg, appID);
	if(!err) err = alp_ipc_message_pack_int32(msg, launchErr);
	if(!err) err = alp_ipc_message_pack_int32(msg, appPID);
	if(!err) err = alp_ipc_message_pack_end(msg);
	
	if(err)
	{
		ALP_FAIL_NON_FATAL_IF(TRUE, "pack failed!\n", NULL);
		if(msg != NULL) alp_ipc_message_destroy(msg);
		msg = NULL;
	}
		
	return msg;
}

/*! 

@param[in] msg The AlpMessage to unpack data from.
@param[out] appID Receives the name of the package containing the application.
@param[out] launchErr Receives the error result code from the launch operation.
@param[out] appPID Receives the process ID of the newly launched application, when applicable (<=0 otherwise).

@return Yields ALP_STATUS_OK on success, or an error code if there is a problem unpacking the message.
 */
alp_status_t alp_app_prv_unpack_launch_reply_msg(AlpMessage* msg, char **appID, alp_status_t *launchErr, pid_t *appPID)
{
	alp_status_t err = ALP_OK;
	
	if(alp_ipc_message_get_message_ID(msg) != ALP_APP_MSG_ID_LAUNCH_REPLY)
		return ALP_STATUS_APP_MSG;
	
	err = alp_ipc_message_unpack_start(msg);
	if(!err) err = alp_ipc_message_unpack_string(msg, appID);
	if(!err) err = alp_ipc_message_unpack_int32(msg, (gint32*)launchErr);
	if(!err) err = alp_ipc_message_unpack_int32(msg, appPID);
	if(!err) err = alp_ipc_message_unpack_end(msg);
	
	if(err) err = ALP_STATUS_APP_MSG;
	return err;
}


/*!

@param[in] newBG The new backgrounding setting for the app.

@return Returns the newly allocated message, or NULL on failure.
 */
AlpMessage* alp_app_prv_make_backgrounding_msg(AlpAppBackgroundingType newBG)
{
	AlpMessage *msg = alp_ipc_message_create();
	alp_status_t err = ALP_OK;
	
	if(msg == NULL) err = ALP_STATUS_APP_MEM;
	if(!err) err = alp_ipc_message_pack_start(msg,
	ALP_APP_MSG_ID_BACKGROUNDING);
	if(!err) err = alp_ipc_message_pack_int32(msg, (int32_t)newBG);
	if(!err) err = alp_ipc_message_pack_end(msg);
	
	if(err)
	{
		ALP_FAIL_NON_FATAL_IF(TRUE, "pack failed!\n", NULL);
		if(msg != NULL) alp_ipc_message_destroy(msg);
		msg = NULL;
	}
	
	return msg;
}

/*!

@param[in] msg The AlpMessage to unpack data from.
@param[out] outBG Receives the new backgrounding setting for the app.

@return Yields ALP_STATUS_OK on success, or an error code if there is a problem unpacking the message.
 */
alp_status_t alp_app_prv_unpack_backgrounding_msg(AlpMessage* msg, AlpAppBackgroundingType* outBG)
{
	alp_status_t err = ALP_OK;
	int32_t setting = 0;
	
	if(alp_ipc_message_get_message_ID(msg) != ALP_APP_MSG_ID_BACKGROUNDING)
		return ALP_STATUS_APP_MSG;
	
	err = alp_ipc_message_unpack_start(msg);
	if(!err) err = alp_ipc_message_unpack_int32(msg, &setting);
	if(!err) err = alp_ipc_message_unpack_end(msg);
	
	*outBG = (AlpAppBackgroundingType)setting;
	
	if(err) err = ALP_STATUS_APP_MSG;
	return err;
}


/*!

@param[in] replyTo The launch request message to reply to.
@return Returns the newly allocated message, or NULL on failure.
 */
AlpMessage* alp_app_prv_make_flush_msg(AlpMessage *replyTo)
{
	AlpMessage *msg;
	alp_status_t err = ALP_OK;
		
	msg = alp_ipc_message_create();
	if(msg == NULL) err = ALP_STATUS_APP_MEM;
	if(!err)
	{
		if(replyTo == NULL) err = alp_ipc_message_pack_start(msg,
		ALP_APP_MSG_ID_FLUSH);
		else err = alp_ipc_message_response_pack_start(msg,
				ALP_APP_MSG_ID_FLUSH, replyTo);
	}
	if(!err) err = alp_ipc_message_pack_end(msg);
	
	if(err)
	{
		ALP_FAIL_NON_FATAL_IF(TRUE, "pack failed!\n", NULL);
		if(msg != NULL) alp_ipc_message_destroy(msg);
		msg = NULL;
	}
	
	return msg;
}


/*!

@return Returns the newly allocated message, or NULL on failure.
 */
AlpMessage* alp_app_prv_make_jump_list_request_msg(void)
{
	AlpMessage *msg;
	alp_status_t err = ALP_OK;
		
	msg = alp_ipc_message_create();
	if(msg == NULL) err = ALP_STATUS_APP_MEM;
	if(!err) err = alp_ipc_message_pack_start(msg,
	ALP_APP_MSG_ID_JUMP_LIST_REQ);
	if(!err) err = alp_ipc_message_pack_end(msg);
	
	if(err)
	{
		ALP_FAIL_NON_FATAL_IF(TRUE, "pack failed!\n", NULL);
		if(msg != NULL) alp_ipc_message_destroy(msg);
		msg = NULL;
	}
	
	return msg;
}

/*!
@param[in] replyTo The AlpMessage containing the original jumplist request.
@param[in] count The number of strings in the appIDs array.
@param[in] appIDs An array of char* containing package ID strings for each app in the jump list.

@return Returns the newly allocated message, or NULL on failure.
 */
AlpMessage* alp_app_prv_make_jump_list_msg(AlpMessage *replyTo, int count,  char *const appIDs[], const gboolean *isRunnings)
{
	AlpMessage *msg;
	alp_status_t err = ALP_OK;
	int i;
		
	msg = alp_ipc_message_create();
	if(msg == NULL) err = ALP_STATUS_APP_MEM;
	if(!err) err = alp_ipc_message_response_pack_start(msg,
	ALP_APP_MSG_ID_JUMP_LIST, replyTo);
	
	if(!err) err = alp_app_prv_pack_argv(msg, NULL, count, appIDs);
	
	i=0;
	while(!err && i<count)
	{
		err = alp_ipc_message_pack_int32(msg, (int32_t)isRunnings[i]);
		i++;
	}
	
	if(!err) err = alp_ipc_message_pack_end(msg);
	
	if(err)
	{
		ALP_FAIL_NON_FATAL_IF(TRUE, "pack failed!\n", NULL);
		if(msg != NULL) alp_ipc_message_destroy(msg);
		msg = NULL;
	}
	
	return msg;
}

/*! 

@param[in] msg The AlpMessage to unpack data from.
@param[out] count Receives the length of the appIDs array.
@param[out] appIDs Receives a ptr to an array of char* containing package IDs of the apps in the jump list.

@return Yields ALP_STATUS_OK on success, or an error code if there is a problem unpacking the message.
 */
alp_status_t alp_app_prv_unpack_jump_list_msg(AlpMessage* msg, 
								  int *count, char***appIDs, gboolean **outIsRunnings)
{
	alp_status_t err = ALP_OK;
	int32_t temp;
	gboolean *isRunnings = NULL;
	int i;
	
	*appIDs = NULL;
	
	if(alp_ipc_message_get_message_ID(msg) != ALP_APP_MSG_ID_JUMP_LIST)
		return ALP_STATUS_APP_MSG;
	
	err = alp_ipc_message_unpack_start(msg);
	if(!err) err = alp_app_prv_unpack_argv(msg, count, appIDs);
	
	if(!err) isRunnings = g_new(gboolean, *count);
	
	i=0;
	while(!err && i < *count)
	{
		err = alp_ipc_message_unpack_int32(msg, &temp);
		isRunnings[i] = (gboolean)temp;
		i++;
	}
	
	
	
	if(!err) err = alp_ipc_message_unpack_end(msg);
	
	
	if(err)
	{
		err = ALP_STATUS_APP_MSG;
		if(isRunnings) g_free(isRunnings);
		if(*appIDs) 
		{
			alp_app_free_argv(*count, *appIDs);
			*appIDs = NULL;
		}
	}
	else *outIsRunnings = isRunnings;
	
	return err;
}



/*!
This message is used for both the initial request to appserver, and also to carry the reply.

@param[in] replyTo The AlpMessage to reply to, or NULL if this is a request to appserver.
@param[in] appID Package ID to store in the message.  Use "" for a request to the appserver.

@return Returns the newly allocated message, or NULL on failure.
 */
AlpMessage* alp_app_prv_make_cur_app_msg(AlpMessage *replyTo, const char *appID)
{
	AlpMessage *msg = alp_ipc_message_create();
	alp_status_t err = ALP_OK;
	
	// Handle NULL, just to be friendly.
	if(appID == NULL) appID = "";
	
	if(msg == NULL) err = ALP_STATUS_APP_MEM;
	if(!err) 
	{
		if(replyTo) err = alp_ipc_message_response_pack_start(msg,
								ALP_APP_MSG_ID_CUR_APP, replyTo);
		else err = alp_ipc_message_pack_start(msg, 
								ALP_APP_MSG_ID_CUR_APP);
	}
	
	if(!err) err = alp_ipc_message_pack_string(msg, appID);
	if(!err) err = alp_ipc_message_pack_end(msg);
	
	if(err)
	{
		ALP_FAIL_NON_FATAL_IF(TRUE, "pack failed!\n", NULL);
		if(msg != NULL) alp_ipc_message_destroy(msg);
		msg = NULL;
	}
	
	return msg;
}

/*!

@param[in] msg The AlpMessage to unpack data from.
@param[out] outPkgID Receives pkg ID stored in the message.

@return Yields ALP_STATUS_OK on success, or an error code if there is a problem unpacking the message.
 */
alp_status_t alp_app_prv_unpack_cur_app_msg(AlpMessage* msg, char** outPkgID)
{
	alp_status_t err = ALP_OK;
	
	if(alp_ipc_message_get_message_ID(msg) != ALP_APP_MSG_ID_CUR_APP)
		return ALP_STATUS_APP_MSG;
	
	err = alp_ipc_message_unpack_start(msg);
	if(!err) err = alp_ipc_message_unpack_string(msg, outPkgID);
	if(!err) err = alp_ipc_message_unpack_end(msg);
		
	if(err) err = ALP_STATUS_APP_MSG;
	
	return err;
}




/*!
This message is used for both the initial request to appserver, and also to carry the reply.

@param[in] replyTo The AlpMessage to reply to, or NULL if this is a request to appserver.
@param[in] appID Package ID to store in the message.
@param[in] watcherRef Watcher ID to store in message.
@param[in] wasRun Boolean wasRun value to store in message.
@param[in] remove Boolean "remove" value to store in message.

@return Returns the newly allocated message, or NULL on failure.
 */
AlpMessage* alp_app_prv_make_watcher_msg(AlpMessage *replyTo, const char *appID, AlpWatcherRef watcherRef, gboolean wasRun, gboolean remove, alp_status_t result, gint blockMode)
{
	AlpMessage *msg = alp_ipc_message_create();
	alp_status_t err = ALP_OK;
	
	if(msg == NULL) err = ALP_STATUS_APP_MEM;
	if(!err) 
	{
		if(replyTo) err = alp_ipc_message_response_pack_start(msg,
				ALP_APP_MSG_ID_WATCHER, replyTo);
		else err = alp_ipc_message_pack_start(msg, 
				ALP_APP_MSG_ID_WATCHER);
	}
	
	if(!err) err = alp_ipc_message_pack_string(msg, appID);
	if(!err) err = alp_ipc_message_pack_uint32(msg, watcherRef);
	if(!err) err = alp_ipc_message_pack_uint32(msg, (uint32_t)wasRun);
	if(!err) err = alp_ipc_message_pack_uint32(msg, (uint32_t)remove);
	if(!err) err = alp_ipc_message_pack_uint32(msg, (uint32_t)result);
	if(!err) err = alp_ipc_message_pack_int32(msg, (int32_t)blockMode);
	if(!err) err = alp_ipc_message_pack_end(msg);
	
	if(err)
	{
		ALP_FAIL_NON_FATAL_IF(TRUE, "pack failed!\n", NULL);
		if(msg != NULL) alp_ipc_message_destroy(msg);
		msg = NULL;
	}
	
	return msg;
}

/*!

@param[in] msg The AlpMessage to unpack data from.
@param[out] outPkgID Receives pkg ID stored in the message.

@return Yields ALP_STATUS_OK on success, or an error code if there is a problem unpacking the message.
 */
alp_status_t alp_app_prv_unpack_watcher_msg(AlpMessage* msg, char** outPkgID, AlpWatcherRef *outWatcherRef, gboolean *outWasRun, gboolean *outRemove, alp_status_t *outResult, gint *outBlockMode)
{
	alp_status_t err = ALP_OK;
	uint32_t tempWasRun, tempRemove, tempResult;
	int32_t tempBlockMode;
	
	if(alp_ipc_message_get_message_ID(msg) != ALP_APP_MSG_ID_WATCHER)
		return ALP_STATUS_APP_MSG;
	
	err = alp_ipc_message_unpack_start(msg);
	if(!err) err = alp_ipc_message_unpack_string(msg, outPkgID);
	if(!err) err = alp_ipc_message_unpack_uint32(msg, outWatcherRef);
	if(!err) err = alp_ipc_message_unpack_uint32(msg, &tempWasRun);
	if(!err) err = alp_ipc_message_unpack_uint32(msg, &tempRemove);
	if(!err) err = alp_ipc_message_unpack_uint32(msg, &tempResult);
	if(!err) err = alp_ipc_message_unpack_int32(msg, &tempBlockMode);
	if(!err) err = alp_ipc_message_unpack_end(msg);
	
	if(outWasRun) *outWasRun = tempWasRun;
	if(outRemove) *outRemove = tempRemove;
	if(outResult) *outResult = tempResult;
	if(outBlockMode) *outBlockMode = tempBlockMode;
	
	if(err) err = ALP_STATUS_APP_MSG;
	
	return err;
}


/*!

@return Returns the newly allocated message, or NULL on failure.
 */
AlpMessage* alp_app_prv_make_theme_update_msg(void)
{
	return prv_make_empty_msg(ALP_APP_MSG_ID_THEME_UPDATE);
}

/*!

@return Returns the newly allocated message, or NULL on failure.
 */
AlpMessage* alp_app_prv_make_reset_msg(void)
{
	return prv_make_empty_msg(ALP_APP_MSG_ID_RESET);
}


/*
utility routine used for messages which carry only an ID.
*/
static AlpMessage* prv_make_empty_msg(guint msgID)
{
	AlpMessage *msg;
	alp_status_t err = ALP_OK;
		
	msg = alp_ipc_message_create();
	if(msg == NULL) err = ALP_STATUS_APP_MEM;
	if(!err) err = alp_ipc_message_pack_start(msg, msgID);
	if(!err) err = alp_ipc_message_pack_end(msg);
	
	if(err)
	{
		ALP_FAIL_NON_FATAL_IF(TRUE, "pack failed!\n", NULL);
		if(msg != NULL) alp_ipc_message_destroy(msg);
		msg = NULL;
	}
	
	return msg;
}


/*
utility routine used for exit & retire messages.
*/
static AlpMessage* prv_make_msg_with_string(gint msgID, const char *str)
{
	AlpMessage *msg = alp_ipc_message_create();
	alp_status_t err = ALP_OK;
	
	if(msg == NULL) err = ALP_STATUS_APP_MEM;
	if(!err) err = alp_ipc_message_pack_start(msg, msgID);
	if(!err) err = alp_ipc_message_pack_string(msg, str);
	if(!err) err = alp_ipc_message_pack_end(msg);
	
	if(err)
	{
		ALP_FAIL_NON_FATAL_IF(TRUE, "pack failed!\n", NULL);
		msg = NULL;
	}
	
	return msg;
}

/*
	utility routine used for exit & retire messages. 
*/
static alp_status_t prv_unpack_msg_with_string(AlpMessage* msg, gint msgID, char** str)
{
	alp_status_t err = ALP_OK;
	
	if(alp_ipc_message_get_message_ID(msg) != msgID)
		return ALP_STATUS_APP_MSG;
	
	err = alp_ipc_message_unpack_start(msg);
	if(!err) err = alp_ipc_message_unpack_string(msg, str);
	if(!err) err = alp_ipc_message_unpack_end(msg);
	
	if(err) err = ALP_STATUS_APP_MSG;
	return err;
}
