/********************************************************************

Copyright  2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <hiker/appmgr.h>
#include <hiker/pdk/appmgr_pdk.h>
#include <hiker/bundlemgr.h>
#include <hiker/prv/bundlemgr_prv.h>


#define PRV_BUNDLE_ID_LEN 255
#define PRV_BUNDLE_EXPAND_STR "bar:com.access.apps."
static char sBundleID[PRV_BUNDLE_ID_LEN];


// Expands a short bundle name ("launcher") into a full-frm long one 
// used by all the APIs ("bar:com.access.apps.launcher");
// It does this very simply - if the ':' is not specified, it prepends "bar:com.access.apps."
const char * prv_expand_bundle_id(const char *shortName)
{
    if(shortName == NULL) return shortName;
    if(strchr(shortName, ':') == NULL && 
        strlen(shortName) + strlen(PRV_BUNDLE_EXPAND_STR) < PRV_BUNDLE_ID_LEN - 1)
    {
        strcpy(sBundleID, PRV_BUNDLE_EXPAND_STR);
        strcat(sBundleID, shortName);
        return sBundleID;
    }
    
    return shortName;
}

int cmd_launch(const char * appname, int argc, char *const argv[])
{
  appname = prv_expand_bundle_id(appname);
  alp_status_t result = alp_app_launch(appname, argc, argv, NULL);
  if (result != ALP_STATUS_OK) {
    fprintf(stderr, "Error launching: 0x%lx\n", (unsigned long)result);
    return 1;
  }
  
  return 0;
}

int cmd_debug(const char * appname, int argc, char *const argv[])
{
  pid_t pid;
  appname = prv_expand_bundle_id(appname);
  alp_status_t result = alp_app_launch(appname, argc, argv, &pid);
  if (result != ALP_STATUS_OK) {
    fprintf(stderr, "Error launching: 0x%lx\n", (unsigned long)result);
    return 1;
  } else {
    fprintf(stdout, "%d\n", pid);
    return 0;
  }
}

int cmd_retire(const char * appname)
{
  alp_status_t result = alp_app_retire(appname);
  appname = prv_expand_bundle_id(appname);
  if (result != ALP_STATUS_OK) {
    fprintf(stderr, "Error retiring: 0x%lx\n", (unsigned long)result);
    return 1;
  } else {
    return 0;
  }
}

int cmd_exit(const char * appname)
{
  appname = prv_expand_bundle_id(appname);
  alp_status_t result = alp_app_exit(appname);
  if (result != ALP_STATUS_OK) {
    fprintf(stderr, "Error exiting: 0x%lx\n", (unsigned long)result);
    return 1;
  } else {
    return 0;
  }
}

int cmd_list(void)
{
  if (alp_bundle_init() != ALP_STATUS_OK)
    return 1;

  GArray * apps = alp_bundle_search(NULL, TRUE, NULL, NULL, "application");
  if (!apps)
    return 1;

  unsigned int i;
  for (i=0; i<apps->len; i++) {
    AlpBundle package = g_array_index(apps, AlpBundle, i);

    if (alp_bundle_is_NULL(package))
      continue;

    // URL to pass to appserver
    gchar * package_name = alp_bundle_metadata_value(package, "name");

    // Name to display to user
    gchar * title = alp_bundle_property_value(package, "application", 0, "name");
    
    printf("%s\n", package_name);
    
    g_free(package_name);
    g_free(title);
  }
  
  g_array_free(apps, TRUE);

  return 0;
}

int cmd_wait(AlpAppContext ** context)
{
  
  while (*context == NULL) {
    struct timespec t;
    t.tv_sec = 0;
    t.tv_nsec = 1000000000 / 4;
    nanosleep(&t, NULL);

    *context = alp_app_setup("alp_app", ALP_APP_LAUNCH_FLAG_NONE, NULL);
    
    if (isatty(fileno(stdout))) {
      printf(".");
      fflush(stdout);
    }
  }

  return 0;
}

const char * argv0;

int usage(const char *bad_option)
{
  if (bad_option)
    printf("%s: bad option \"%s\"\n", argv0, bad_option);
  printf("Usage: %s [--OPTION ...]\n", argv0);
  printf(
    "Available Options:\n"
    "    --launch <appname> [launchArgs] Launch an app\n"
    "    --debug  <appname> [launchArgs] Launch an app for debugging; return pid\n"
    "    --exit   [<appname>]            Terminate a running application\n"
    "    --retire [<appname>]            Retire a primary application\n"
    "    --list (or --names)             List available apps\n"
    "    --wait                          Wait until appserver is available\n"
    "    --current                       Prints the name of the current primary application\n"
    "    --reboot                        Reboot the system\n"
    "    --help                          Show help options\n"
    "\n"
    "If <appname> does not include a ':', then \"bar:com.access.apps.\" \n"
    "will be automatically prepended.\n"
    "\n"
    "Available launchArgs:\n"
    "    --alp-primary, --alp-backgrounded, --alp-debug, --alp-display,\n"
    "    --alp-notify, --alp-exchange, --alp-alarm, --alp-alert\n"
    );
  return 2;
}

int main(int argc,  char *argv[])
{
  int result = 0;
  argv0 = argv[0];
  const char *opt;

  if (argc < 2)
    return usage(NULL);
  
  AlpAppContext *ctxt = alp_app_setup("alp_app", ALP_APP_LAUNCH_FLAG_NONE, NULL);
  
  opt = argv[1];

  if (*opt == '-') {
      opt++;
      // skip second dash 
      if (*opt == '-') opt++;

      if (strcmp(opt, "launch") == 0 && argc >= 3)
          result = cmd_launch(argv[2], argc-3, argv+3);
      else if (strcmp(opt, "debug") == 0 && argc >= 3)
          result = cmd_debug(argv[2], argc-3, argv+3);
      else if (strcmp(opt, "exit") == 0 && (argc == 2 || argc == 3))
          result = cmd_exit(argc==3 ? argv[2] : NULL);
      else if (strcmp(opt, "retire") == 0 && (argc == 2 || argc == 3))
          result = cmd_retire(argc==3 ? argv[2] : NULL);
      else if (strcmp(opt, "list") == 0 && argc == 2)
          result = cmd_list();
      else if (strcmp(opt, "names") == 0 && argc == 2)
          result = cmd_list();
      else if (strcmp(opt, "current") == 0 && argc == 2)
      {
          char *app = 0;
          alp_status_t status = alp_app_current(&app);
          if (status == ALP_STATUS_OK)
          {
              if( app != NULL) printf("current primary app is %s.\n", app);
              else printf("There is no current primary app.\n");
          }
          else fprintf(stderr, "Error querying: 0x%lx\n", (unsigned long)status);
          
          g_free(app);
          
          result = 0;
      }
	 else if (strcmp(opt, "reboot") == 0 && argc == 2)
		 alp_app_reset_system();
	 else if (strcmp(opt, "wait") == 0 && argc == 2)
			 result = cmd_wait(&ctxt);
      else if (strcmp(opt, "help") == 0)
          result = usage(NULL);
      else
          result = usage(argv[1]);
  } else {
      result = usage(argv[1]);
  }
  
  if(ctxt) alp_app_destroy(ctxt, 0);
  
  return result;
}


