/********************************************************************

Copyright  2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/*
	Unit tests for the MAX Application Server
*/

// Uncomment the next line to disable traces AT COMPILE TIME 
#define ALP_TRACE ALP_TRACE_OFF		//Enable/disable

#include <hiker/unit_test.h>
#include <cppunit/extensions/HelperMacros.h>

#include <gtk/gtk.h>
#include <hiker/ipc.h>
#include <hiker/pdk/appmgr_pdk.h>
#include <hiker/prv/appmgr_prv.h>

#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <string.h>
#include <unistd.h>

static const unsigned char kArgBufferSize=64;
static const unsigned char kTotalTestLaunches=6;
static const unsigned char kTotalTestLaunchParams=6;

class AppServerTest: public CPPUNIT_NS::TestFixture
{
	public:
		
	protected:
		// Testing some alp_app_launch() functionality.
		void LaunchErrorTest(void);
	
		// Unexpected & Clean exit testing:
		void UnexpectedExit(void);
		void CleanExit(void);
		void DoExitTest(void);
		
		void UnexpectedExitProc();
		static void UnexpectedExitDiagHandler(const char* appID, const char* diagStr, void* cbData);
		
		// Exit handler testing:
		void SIGTERM_ExitTest(void);
		void ExitHandlerTest(void);
		void ExitTestProc(void);
		
		// Relaunch testing:
		void RelaunchHandlerTest(void);
		
		static void TestRelaunchHandler (int argc,  char* argv[], gpointer cbData);
		gboolean TestRelaunch(char *launchArg);

		// Message packing/unpacking tests:
		void ArgcArgvPackingTest(void);
		void MessagePackingTest(void);
		
		// Testing of app contexts:
		void AppContextTest(void);
		static void ContextTestRelaunchHandler1(int argc, char* argv[], gpointer cbData);
		static void ContextTestRelaunchHandler2(int argc, char* argv[], gpointer cbData);
		static void ContextTestExitHandler1(gpointer cbData);
		static void ContextTestExitHandler2(gpointer cbData);
		
		// Testing of 'transient' app support:
		void TransientAppTest(void);
		static void TransientTestExitHandler(gpointer cbData);
		static void TransientRelaunchHandler(int argc, char* argv[], gpointer cbData);
		static void AddTransientRelaunchHandler(int argc, char* argv[], gpointer cbData);
		
		// Launchpad testing
		void LaunchpadTest(void);
		void SkeletonLaunchpad(void);
		static void LaunchpadRelaunchHandler(int argc, char* argv[], gpointer cbData);
		
		void LaunchpadRelaunchTest(void);
		static void LRRelaunchHandler (int argc, char* argv[], gpointer cbData);

		// "debug" launch testing
		void RequestPIDLaunchTest(void);
		static void RequestPIDRelaunchHandler (int argc, char* argv[], gpointer cbData);
		
		// Test backgrounding options:
		void BackgroundingTest(void);
		static void BackgroundingExitHandler(gpointer cbData);
		static void BackgroundingRelaunchHandler (int argc, char* argv[], gpointer cbData);
		
		// Test queued relaunching, combined w/ app exit.
		static void QueuedRelaunchLPHandler (int argc, char* argv[], gpointer cbData);
		void QueuedRelaunchTest(void);
		
		// Test kill timers for unresponsive apps:
		void UnresponsiveAppTest(void);

		// Test the jump list API
		void JumpListTest(void);

		// Test for a bug where a launch request could remain pending forever,
		// preventing later launches, if the launchpad didn't register it
		// for some reason.
		void PendingForeverTest(void);
		static void ForeverRelaunchHandler(int argc, char* argv[], gpointer cbData);
		
		// Testing of 'current app' API:
		void CurrAppTest(void);
		
		// Testing of 'watcher' API:
		void WatcherTest(void);
		
		// Testing of 'retire' API:
		void RetireAppTest(void);
		static void RetireRelaunchHandler(int argc, char* argv[], gpointer cbData);
		static void RetireExitHandler(gpointer cbData);
		
		// Shared test infrastructure
		static void DoNothingExitHandler(gpointer cbData);
		static void FakeRegisterRelaunchHandler(int argc, char* argv[], gpointer cbData);
		static void FakeAppRegistration(const char *appID, guint32 launchFlags);
		static void ExitMainLoopConnectCB (AlpConnection* connection, gpointer cbData);
		static void ExitMainLoopHandler(gpointer cbData);
		int ForkTest(void (AppServerTest::*) ());
		const char* PrvGetSignalName(int sigid);
		static void RandomAppRelaunchHandler(int argc,  char* argv[], gpointer cbData);
		static void RandomAppExitHandler(gpointer cbData);
		bool VerifyLaunchFullBools(gboolean expect1, gboolean expect2, gboolean expect3, gboolean expect4);
		void LaunchFullUtil(void); // utility function that performs the LaunchFull tests
		void LaunchFull(void);
		void NoExitHandlerTest(void);
		void NoRelaunchHandlerTest(void);
		void NoHandlerTest(void);
		void AbandonApplication(void);
		
		// app context long running handlers, stress situation
		void ContextLongHandlerTest(void);
		static void ContextTestHangingRelaunchHandler1(int argc, char* arv[], gpointer cbData);
		static void ContextTestHangingExitHandler1(gpointer cbData);

		void RelaunchTestHandlersThatLaunch(void);
		static void ExternalLaunchRelaunchHandler(int argc,  char* argv[], gpointer cbData);
		static void ExternalLaunchExitHandler(gpointer cbData);

		// randomized launching test for stress
		void RandomModalLaunchTest(void);
	
	private:
		GMainLoop *mLoop;
		gboolean	mPassed, mPassed2;
		const char *mAppID;
		const char *mLaunchpadScheme;
		char mArgBuffer[kArgBufferSize];
		gboolean mUnregisterBeforeExit;
		gboolean mSetExitHandler;
		
		// Used by launchpad tests
		unsigned int mReceivedLaunchCount, mLaunchLimit;
		int *mArgcList;
		const char* (*mArgvList)[6][6];
		static pid_t sAppServerPID;
		
		// ContextTest instance vars:
		gboolean mContext1Relaunched;
		gboolean mContext1Exited;
		gboolean mContext2Relaunched;
		gboolean mContext2Exited;
		
		// TransientTest instance vars:
		gboolean mTransientContext2Finished;
		
		// instance vars for handler-abuse tests
		gboolean mSkipExitHandler;
		gboolean mSkipRelaunchHandler;
		
};

pid_t AppServerTest::sAppServerPID = 0;


// **************************************
// Unexpected & Clean Exit Testing:
// **************************************
void AppServerTest::LaunchErrorTest(void)
{
	alp_status_t err;
	char *argv = "";
	AlpAppContext *ctxt;
	
	ctxt = alp_app_setup("com.access.appserver.unittester", ALP_APP_LAUNCH_FLAG_NONE, NULL);
		
	// Should generate "missing launchpad" error from appserver.
	err = alp_app_launch("whatever:com.access.missing_launchpad", 0, &argv, NULL);
	CPPUNIT_ASSERT(err == ALP_STATUS_APP_NO_LAUNCHPAD);
	
	// Should generate "missing launch scheme" error from appserver.
	err = alp_app_launch("com.access.missing_scheme", 0, &argv, NULL);
	CPPUNIT_ASSERT(err == ALP_STATUS_APP_PARAM);
	
	// Should generate "no such thing" error
	err = alp_app_launch("com.access.apps.release_on_time", 0, &argv, NULL);
//	printf("\n\nlaunch: 0x%x\n\n", err);
	CPPUNIT_ASSERT(err == ALP_STATUS_APP_PARAM);
	
	alp_app_destroy(ctxt, 0);
	
	return;
}

// **************************************
// Unexpected & Clean Exit Testing:
// **************************************

void AppServerTest::UnexpectedExit(void)
{
	int status;
	
	mAppID = "com.access.testapps.UnexpectedExit";
	mUnregisterBeforeExit = false;
	
//	status = ForkTest(&AppServerTest::UnexpectedExitProc);	
//	CPPUNIT_ASSERT(WIFEXITED(status) && (WEXITSTATUS(status)) == TRUE);
	
	DoExitTest();
	
	return;
}

void AppServerTest::CleanExit(void)
{
	int status;
	
	mAppID = "com.access.testapps.CleanExit";
	mUnregisterBeforeExit = true;
	
//	status = ForkTest(&AppServerTest::UnexpectedExitProc);	
//	CPPUNIT_ASSERT(WIFEXITED(status) && (WEXITSTATUS(status)) == TRUE);
	
	DoExitTest();
	
	return;
}


void AppServerTest::DoExitTest(void)
{
	AlpAppContext *launcher, *app;
	char *diagID = "com.access.testapps.DiagnosticReceiver";
	pid_t childPID;
	int result;
	
	mPassed = FALSE;
	
	launcher = alp_app_setup(diagID, ALP_APP_LAUNCH_FLAG_DIAGNOSTIC, NULL);
	alp_app_set_diagnostic_handler(launcher, UnexpectedExitDiagHandler, this);
	
	if((childPID = fork()) == 0)
	{
		app = alp_app_setup(mAppID, ALP_APP_LAUNCH_FLAG_NONE, NULL);
		
		if(mUnregisterBeforeExit) alp_app_destroy(app, 0);
		exit(0);
	}
	
	waitpid(childPID, &result, 0);
	
	// run main loop
	mLoop = g_main_loop_new(NULL, TRUE);
	g_main_loop_run(mLoop);
	g_main_loop_unref(mLoop);
	mLoop = NULL;
	
	alp_app_destroy(launcher, 0);
	
	return;
}

bool AppServerTest::VerifyLaunchFullBools(gboolean expect1, gboolean expect2, gboolean expect3, gboolean expect4)
{
	if(mSkipRelaunchHandler)
	{
		// modifying pass-by-value copy- we don't expect these to change from false
		CPPUNIT_ASSERT(!expect1);
		CPPUNIT_ASSERT(!expect3);
		expect1 = false;
		expect3 = false;
	}
	if(mSkipExitHandler)
	{
		CPPUNIT_ASSERT(!expect2);
		CPPUNIT_ASSERT(!expect4);
		expect2 = false;
		expect4 = false;
	}
	return (mContext1Relaunched == expect1 &&
		mContext1Exited == expect2 &&
		mContext2Relaunched == expect3 &&
		mContext2Exited == expect4 );
}
// this test called in multiple situations
void AppServerTest::LaunchFull(void)
{
	mSkipRelaunchHandler = false;
	mSkipExitHandler = false;
	LaunchFullUtil();
}
void AppServerTest::LaunchFullUtil(void)
{
	AlpAppContext *ctxt1, *ctxt2;
	const char *appID1, *appID2;
	char *arg;
	
	mLoop = g_main_loop_new(NULL, TRUE);

	// Allocate & register two contexts:
	appID1 = "com.access.appserver.LaunchFullTest1";
	appID2 = "com.access.appserver.LaunchFullTest2";
	ctxt1 = alp_app_setup(appID1, ALP_APP_LAUNCH_FLAGS_DEFAULT, "");
	ctxt2 = alp_app_setup(appID2, ALP_APP_LAUNCH_FLAGS_DEFAULT, "");	
	
	if(!mSkipRelaunchHandler)
		alp_app_set_relaunch_handler(AppServerTest::ContextTestRelaunchHandler1, this);
	if(!mSkipExitHandler)
		alp_app_add_exit_handler(AppServerTest::ContextTestExitHandler1, this);
	
	alp_app_set_default_context(ctxt2);
	if(!mSkipRelaunchHandler)
		alp_app_set_relaunch_handler(AppServerTest::ContextTestRelaunchHandler2, this);
	if(!mSkipExitHandler)
		alp_app_add_exit_handler(AppServerTest::ContextTestExitHandler2, this);
	
	// Init test vars:
	mContext1Relaunched = false;
	mContext1Exited = false;
	mContext2Relaunched = false;
	mContext2Exited = false;
	arg = "";
	
	alp_app_launch_full(ctxt1, appID1, 0, &arg, NULL);
	if(mContext1Relaunched == false) g_main_loop_run(mLoop);
	
	CPPUNIT_ASSERT(VerifyLaunchFullBools(true, false, false, false));
	
	alp_app_launch_full(ctxt2, appID2, 0, &arg, NULL);
	if(mContext2Relaunched == false) g_main_loop_run(mLoop);
	
	CPPUNIT_ASSERT(VerifyLaunchFullBools(true, false, true, false));

	alp_app_exit(appID2);
	if(mContext1Exited == false) g_main_loop_run(mLoop);
	
	CPPUNIT_ASSERT(VerifyLaunchFullBools(true, false, true, true));
	
	alp_app_exit(appID1);
	if(mContext1Exited == false) g_main_loop_run(mLoop);

	CPPUNIT_ASSERT(VerifyLaunchFullBools(true, true, true, true));
	
	// Unregister & exit
	alp_app_destroy(ctxt1, 0);
	alp_app_destroy(ctxt2, 0);
	
	// Deallocate main loop
	g_main_loop_unref(mLoop);
	
	return;
}

void AppServerTest::UnexpectedExitDiagHandler(const char* appID, const char* diagStr, void* cbData)
{
	AppServerTest *ast = (AppServerTest*)cbData;
		
	// Check pass criteria for unexpected & clean exits:
	if(ast->mUnregisterBeforeExit == false 
		  && strcmp(appID, ast->mAppID) == 0 
		  && strcmp(diagStr, ALP_APP_DIAG_UNEXPECTED_EXIT) == 0)
	{
		ast->mPassed = TRUE;
	}
	else if(ast->mUnregisterBeforeExit == true 
			  && strcmp(appID, ast->mAppID) == 0 
			  && strcmp(diagStr, ALP_APP_DIAG_CLEAN_EXIT) == 0)
	{
		ast->mPassed = TRUE;
	}
	
	// Exit blocking main loop if we passed.
	if(ast->mPassed) g_main_loop_quit(ast->mLoop);
	
	return;
}

// **************************************
// Exit Handler Testing:
// **************************************

void AppServerTest::SIGTERM_ExitTest(void)
{
	int status;
	
	mAppID = "com.access.testapps.SIGTERMExitTest";
	mSetExitHandler = false;

	status = ForkTest(&AppServerTest::ExitTestProc);
	
	// Note that this test will signal in the app server as exiting unexpectedly 
	// since we do not unregister before we receive the SIGTERM.
	// Though perhaps a bit odd (there was a request for it to exit, after all)
	// this is is expected behavior.
	
	CPPUNIT_ASSERT(WIFSIGNALED(status));
	CPPUNIT_ASSERT(WTERMSIG(status) == SIGTERM);
	
	return;
}

void AppServerTest::ExitHandlerTest(void)
{
	int status;
	
	mAppID = "com.access.testapps.ExitHandlerTest";
	mSetExitHandler = true;
	
	status = ForkTest(&AppServerTest::ExitTestProc);
	
	CPPUNIT_ASSERT(WIFEXITED(status));	
	CPPUNIT_ASSERT(WEXITSTATUS(status) == TRUE);	
	
	return;
}


// If we don't register an exit handler, do we get a SIGTERM that kills the process?
void AppServerTest::ExitTestProc(void)
{
	AlpAppContext *ctxt;
	alp_status_t err;
	
	ctxt = alp_app_setup(mAppID, 0, NULL);
	
	mLoop = g_main_loop_new(NULL, TRUE);
	
	if(mSetExitHandler == true)
	{
		// Register a few, and remove some...
		alp_app_add_exit_handler(AppServerTest::ExitMainLoopHandler, mLoop);
		alp_app_add_exit_handler(AppServerTest::ExitMainLoopHandler, (gpointer)1);
		alp_app_add_exit_handler(AppServerTest::ExitMainLoopHandler, NULL);
	
		err = alp_app_remove_exit_handler(AppServerTest::ExitMainLoopHandler);
		CPPUNIT_ASSERT(err == ALP_OK);
		err = alp_app_remove_exit_handler_with_data(AppServerTest::ExitMainLoopHandler,
									(gpointer)1);
		CPPUNIT_ASSERT(err == ALP_OK);
	}	

	
	// Request to app server that we be shut down.
	alp_app_exit(mAppID);
	
	// Run main loop and wait to be shut down.
	g_main_loop_run(mLoop);
	
	// ... if the SIGTERM test is successful, execution will not reach here - we will receive a SIGTERM, and exit abnormally.
	// The ExitHandlerTest will reach here, however, if we're successful.
	g_main_loop_unref(mLoop);
	
	// Loop terminated, so we were shut down.  Unregister & exit.
	alp_app_destroy(ctxt, 0);
	
	// exit indicating success...
	exit(TRUE);
}




// **************************************
// Relaunch handler testing:
// **************************************

void AppServerTest::TestRelaunchHandler (int argc, char* argv[], gpointer cbData)
{
	AppServerTest *ast = (AppServerTest*)cbData;
	
	CPPUNIT_ASSERT(argv[argc] == NULL);
	
	// copy arg to buffer to prove we were called.
	strncpy(ast->mArgBuffer, argv[1], kArgBufferSize);
	ast->mArgBuffer[kArgBufferSize-1]=0;
	
	// Quit main loop to trigger checking code:
	g_main_loop_quit(ast->mLoop);
}


gboolean AppServerTest::TestRelaunch(char *launchArg)
{
	alp_status_t err;
	
	// Clear argument buffer to ensure freshness.
	mArgBuffer[0] = 0;
	
	// Send launch request to server
	alp_app_launch_async(mAppID, 1, &launchArg);
	// CPPUNIT_ASSERT(err == ALP_STATUS_OK);
	
	// Run main loop and wait for relaunch.
	// We could actually receive the reply before we begin waiting, so we need to check the argBuffer.
	if(mArgBuffer[0] == 0) g_main_loop_run(mLoop);
		
	// Check result, & return success or failure.
	if(strcmp(mArgBuffer, launchArg) != 0)
	{
		g_warning("%s: Test FAILED, didn't receive FG relaunch.\n", mAppID);
		return FALSE;
	}
	
	return TRUE;
}

// Does relaunching work properly?
void AppServerTest::RelaunchHandlerTest(void)
{
	AlpAppContext *ctxt;
	gboolean passed = TRUE;
	
	mAppID = "com.access.testapps.RelaunchHandlerTest";
			
	ctxt = alp_app_setup(mAppID, 0, NULL);
	
	mLoop = g_main_loop_new(NULL, TRUE);
	alp_app_add_exit_handler(ExitMainLoopHandler, mLoop);
	
	alp_app_set_relaunch_handler(TestRelaunchHandler, this);
	
	// Relaunch ourselves a few times and verify that our callback is called with the right parameters.
	CPPUNIT_ASSERT(TestRelaunch(ALP_APP_PRIMARY));
	CPPUNIT_ASSERT(TestRelaunch(ALP_APP_PRIMARY));
	CPPUNIT_ASSERT(TestRelaunch("--alp-whatever"));
	CPPUNIT_ASSERT(TestRelaunch("--alp-whatever"));
	CPPUNIT_ASSERT(TestRelaunch(ALP_APP_PRIMARY));
	
	// report success or failure.
//	if(passed) g_print("%s: All tests passed.\n", mAppID);
//	else g_print("%s: Some tests FAILED.\n", mAppID);
	
	// All done.  Unregister & exit.
	alp_app_destroy(ctxt, 0);
	
	return;
}


// **************************************************
// Message packing/unpacking tests, appserver level.
// **************************************************
void AppServerTest::ArgcArgvPackingTest(void)
{
	alp_status_t err = ALP_OK;
	AlpMessage *msg;
	char*	testArgv[6] = {"hello", "this", "is", "a", "test", "Think it'll work?"};
	int		testArgc = 6;
	int argc;
	int i;
	char **argv;
	
	mAppID = "com.access.test.ArgcArgvPackingTest";
	
	// Create a message
	msg = alp_ipc_message_create();
	CPPUNIT_ASSERT(msg != NULL);
	
	// Pack it up
	err = alp_ipc_message_pack_start(msg, 123);
	CPPUNIT_ASSERT(err == ALP_OK);
	
	err = alp_app_prv_pack_argv(msg, NULL, testArgc, testArgv);
	CPPUNIT_ASSERT(err == ALP_OK);
	
	err = alp_ipc_message_pack_end(msg);
	CPPUNIT_ASSERT(err == ALP_OK);
	
	// Might as well check message ID
	CPPUNIT_ASSERT(alp_ipc_message_get_message_ID(msg) == 123);
	
	// Unpack it:
	err = alp_ipc_message_unpack_start(msg);
	CPPUNIT_ASSERT(err == ALP_OK);
	
	err = alp_app_prv_unpack_argv(msg, &argc, &argv);
	CPPUNIT_ASSERT(err == ALP_OK);
	
	err = alp_ipc_message_unpack_end(msg);
	CPPUNIT_ASSERT(err == ALP_OK);
	
	// Verify results:
	CPPUNIT_ASSERT(argc==testArgc);
	for(i=0; i<testArgc; i++)
	{
		CPPUNIT_ASSERT(strcmp(argv[i], testArgv[i]) == 0);
	}
	
	// deallocate argv
	alp_app_free_argv(argc, argv);
	
	
	// ... and once more...
	
	
	// Pack it up again the other way:
	err = alp_ipc_message_pack_start(msg, 123);
	CPPUNIT_ASSERT(err == ALP_OK);
	
	err = alp_app_prv_pack_argv(msg, mAppID, testArgc, testArgv);
	CPPUNIT_ASSERT(err == ALP_OK);
	
	err = alp_ipc_message_pack_end(msg);
	CPPUNIT_ASSERT(err == ALP_OK);
	
	// Might as well check message ID
	CPPUNIT_ASSERT(alp_ipc_message_get_message_ID(msg) == 123);
	
	// Unpack it:
	err = alp_ipc_message_unpack_start(msg);
	CPPUNIT_ASSERT(err == ALP_OK);
	
	err = alp_app_prv_unpack_argv(msg, &argc, &argv);
	CPPUNIT_ASSERT(err == ALP_OK);
	
	err = alp_ipc_message_unpack_end(msg);
	CPPUNIT_ASSERT(err == ALP_OK);
	
	// Verify results:
	CPPUNIT_ASSERT(argc==(testArgc+1));
	CPPUNIT_ASSERT(strcmp(mAppID, argv[0]) == 0);
	for(i=0; i<testArgc; i++)
	{
		CPPUNIT_ASSERT(strcmp(argv[i+1], testArgv[i]) == 0);
	}
	
	// free argv
	alp_app_free_argv(argc, argv);
	
	// deallocate message:
	alp_ipc_message_destroy(msg);
	
	return;
}


void AppServerTest::MessagePackingTest(void)
{
	AlpMessage *msg, *msg2;
	gint32 msgID = 1;
	alp_status_t err = ALP_OK;
	alp_status_t launchErr;
	alp_status_t result;
	char*	testArgv[6] = {"hello", "this", "is", "a", "test", "Think it'll work?"};
	int		testArgc = 6;
	int argc, i;
	char **argv;
	char *str, *testStr, *appID;
	char *scheme;
	guint32 appFlags;
	pid_t appPID;
	gboolean waitForPID;
	unsigned int secID;
	
	
	mAppID = "com.access.testapps.MessagePackingTest";
	
	// Test launch message, w/ app ID provided separately.
	msg = alp_app_prv_make_launch_msg(mAppID, testArgc, testArgv, TRUE, 13579);
	CPPUNIT_ASSERT(msg != NULL);
	CPPUNIT_ASSERT(alp_ipc_message_get_message_ID(msg) == ALP_APP_MSG_ID_LAUNCH);
	
	err = alp_app_prv_unpack_launch_msg(msg, &argc, &argv, &waitForPID, &secID);
	CPPUNIT_ASSERT(err == ALP_OK);
	CPPUNIT_ASSERT(waitForPID == TRUE);
	CPPUNIT_ASSERT(secID == 13579);
	
	CPPUNIT_ASSERT(argc==(testArgc+1));
	CPPUNIT_ASSERT(strcmp(mAppID, argv[0]) == 0);
	for(i=0; i<testArgc; i++)
	{
		CPPUNIT_ASSERT(strcmp(argv[i+1], testArgv[i]) == 0);
	}
	
	alp_ipc_message_destroy(msg);
	alp_app_free_argv(argc, argv);
	
	
	// Test launch message, w/o app ID provided separately.
	msg = alp_app_prv_make_launch_msg(NULL, testArgc, testArgv, FALSE, 24680);
	CPPUNIT_ASSERT(msg != NULL);
	CPPUNIT_ASSERT(alp_ipc_message_get_message_ID(msg) == ALP_APP_MSG_ID_LAUNCH);
	
	err = alp_app_prv_unpack_launch_msg(msg, &argc, &argv, &waitForPID, &secID);
	CPPUNIT_ASSERT(err == ALP_OK);
	CPPUNIT_ASSERT(waitForPID == FALSE);
	CPPUNIT_ASSERT(secID == 24680);
	
	CPPUNIT_ASSERT(argc==testArgc);
	for(i=0; i<argc; i++)
	{
		CPPUNIT_ASSERT(strcmp(argv[i], testArgv[i]) == 0);
	}
	
	alp_ipc_message_destroy(msg);
	alp_app_free_argv(argc, argv);
	
	
	// Test exit message
	msg = alp_app_prv_make_exit_msg(mAppID);
	CPPUNIT_ASSERT(msg != NULL);
	CPPUNIT_ASSERT(alp_ipc_message_get_message_ID(msg) == ALP_APP_MSG_ID_EXIT);
	
	err = alp_app_prv_unpack_exit_msg(msg, &str);
	CPPUNIT_ASSERT(err == ALP_OK);
	CPPUNIT_ASSERT(strcmp(mAppID, str) == 0);
		
	alp_ipc_message_destroy(msg);
	g_free(str);
	

	// Test register message
	msg = alp_app_prv_make_register_msg(mAppID, 13579, 0xFEEDBEEF, "fancyScheme:");
	CPPUNIT_ASSERT(msg != NULL);
	CPPUNIT_ASSERT(alp_ipc_message_get_message_ID(msg) == ALP_APP_MSG_ID_REGISTER);
	
	err = alp_app_prv_unpack_register_msg(msg, &str, &appPID, &appFlags, &scheme);
	CPPUNIT_ASSERT(err == ALP_OK);
	CPPUNIT_ASSERT(strcmp(mAppID, str) == 0);
	CPPUNIT_ASSERT(appPID == 13579);
	CPPUNIT_ASSERT(appFlags == 0xFEEDBEEF);
	CPPUNIT_ASSERT(strcmp("fancyScheme:", scheme) == 0);
		
	alp_ipc_message_destroy(msg);
	g_free(str);
	g_free(scheme);
	

	// Test unregister message
	msg = alp_app_prv_make_unregister_msg(14357);
	CPPUNIT_ASSERT(msg != NULL);
	CPPUNIT_ASSERT(alp_ipc_message_get_message_ID(msg) == ALP_APP_MSG_ID_UNREGISTER);
	
	err = alp_app_prv_unpack_unregister_msg(msg, &i);
	CPPUNIT_ASSERT(err == ALP_OK);
	CPPUNIT_ASSERT(i == 14357);
		
	alp_ipc_message_destroy(msg);
	
	

	// Test diagnostic message
	testStr = "This is not really a diagnostic message.";
	msg = alp_app_prv_make_diagnostic_msg(mAppID, testStr);
	CPPUNIT_ASSERT(msg != NULL);
	CPPUNIT_ASSERT(alp_ipc_message_get_message_ID(msg) == ALP_APP_MSG_ID_DIAGNOSTIC);
	
	err = alp_app_prv_unpack_diagnostic_msg(msg, &appID, &str);
	CPPUNIT_ASSERT(err == ALP_OK);
	CPPUNIT_ASSERT(strcmp(mAppID, appID) == 0);
	CPPUNIT_ASSERT(strcmp(testStr, str) == 0);
		
	alp_ipc_message_destroy(msg);
	g_free(str);
	g_free(appID);
	
	// Test launch reply message
	AlpMessage *testReplyMsg = alp_ipc_message_create();
	msg = alp_app_prv_make_launch_reply_msg(testReplyMsg, mAppID, 12345, 67890);
	CPPUNIT_ASSERT(msg != NULL);
	CPPUNIT_ASSERT(alp_ipc_message_get_message_ID(msg) == ALP_APP_MSG_ID_LAUNCH_REPLY);
	
	err = alp_app_prv_unpack_launch_reply_msg(msg, &appID, &launchErr, &appPID);
	CPPUNIT_ASSERT(err == ALP_OK);
	CPPUNIT_ASSERT(strcmp(mAppID, appID) == 0);
	CPPUNIT_ASSERT(launchErr == 12345);
	CPPUNIT_ASSERT(appPID == 67890);
	
	alp_ipc_message_destroy(testReplyMsg);
	alp_ipc_message_destroy(msg);
	g_free(appID);
	
	
	// Test "set backgrounding" message
	AlpAppBackgroundingType bg;
	msg = alp_app_prv_make_backgrounding_msg(ALP_APP_BACKGROUND_REQUIRED);
	CPPUNIT_ASSERT(msg != NULL);
	CPPUNIT_ASSERT(alp_ipc_message_get_message_ID(msg) ==
			ALP_APP_MSG_ID_BACKGROUNDING);
	
	err = alp_app_prv_unpack_backgrounding_msg(msg, &bg);
	CPPUNIT_ASSERT(err == ALP_OK);
	CPPUNIT_ASSERT(bg == ALP_APP_BACKGROUND_REQUIRED);
	
	alp_ipc_message_destroy(msg);
	
	
	// Test flush message
	msg = alp_app_prv_make_flush_msg(NULL);
	CPPUNIT_ASSERT(msg != NULL);
	CPPUNIT_ASSERT(alp_ipc_message_get_message_ID(msg) == ALP_APP_MSG_ID_FLUSH);
	alp_ipc_message_destroy(msg);
	
	
	// Test jump list message
	msg = alp_app_prv_make_jump_list_request_msg();
	CPPUNIT_ASSERT(msg != NULL);
	CPPUNIT_ASSERT(alp_ipc_message_get_message_ID(msg) == ALP_APP_MSG_ID_JUMP_LIST_REQ);
	
	/* can't really test this one, since the message won't have a 
	response token until we actually call "Send with response".  No big deal.
	msg2 = alp_app_prv_make_jump_list_msg(msg, testArgv, testArgc);
	
	... more would go here for testing unpacking, etc...
	alp_status_t alp_app_prv_unpack_jump_list_msg(AlpMessage* msg, 
									  int *count, char***appIDs);
	
	*/
	
	alp_ipc_message_destroy(msg);
	
	
	// Test "cur app" message
	msg = alp_app_prv_make_cur_app_msg(NULL, mAppID);
	CPPUNIT_ASSERT(msg != NULL);
	CPPUNIT_ASSERT(alp_ipc_message_get_message_ID(msg) == ALP_APP_MSG_ID_CUR_APP);
	
	err = alp_app_prv_unpack_cur_app_msg(msg, &str);
	CPPUNIT_ASSERT(err == ALP_OK);
	CPPUNIT_ASSERT(strcmp(mAppID, str) == 0);
	
	alp_ipc_message_destroy(msg);
	g_free(str);
	
	
	// Test Watcher message
	gboolean wasRun, remove;
	uint32_t n;
	gint blockMode;
	result = ALP_OK;
	msg = alp_app_prv_make_watcher_msg(NULL, mAppID, 42, TRUE, FALSE, 
								ALP_STATUS_APP_RUNNING, ALP_APP_WATCHER_BLOCK);
	CPPUNIT_ASSERT(msg != NULL);
	CPPUNIT_ASSERT(alp_ipc_message_get_message_ID(msg) == ALP_APP_MSG_ID_WATCHER);
	
	err = alp_app_prv_unpack_watcher_msg(msg, &str, &n, &wasRun, &remove, &result, &blockMode);
	CPPUNIT_ASSERT(err == ALP_OK);
	CPPUNIT_ASSERT(strcmp(mAppID, str) == 0);
	CPPUNIT_ASSERT(n == 42);
	CPPUNIT_ASSERT(wasRun == TRUE);
	CPPUNIT_ASSERT(remove == FALSE);
	CPPUNIT_ASSERT(result == ALP_STATUS_APP_RUNNING);
	CPPUNIT_ASSERT(blockMode == ALP_APP_WATCHER_BLOCK);
	
	alp_ipc_message_destroy(msg);
	g_free(str);
	
	
	return;
}


// **************************************
// Launchpad testing:
// **************************************
void AppServerTest::AppContextTest(void)
{
	AlpAppContext *ctxt1, *ctxt2;
	const char *appID1, *appID2;
	char *arg;
	
	mLoop = g_main_loop_new(NULL, TRUE);

	// Allocate & register two contexts:
	appID1 = "com.access.appserver.contextTest1";
	appID2 = "com.access.appserver.contextTest2";
	ctxt1 = alp_app_setup(appID1, ALP_APP_LAUNCH_FLAGS_DEFAULT, "");
	ctxt2 = alp_app_setup(appID2, ALP_APP_LAUNCH_FLAGS_DEFAULT, "");	
	
	alp_app_set_relaunch_handler(AppServerTest::ContextTestRelaunchHandler1, this);
	alp_app_add_exit_handler(AppServerTest::ContextTestExitHandler1, this);
	
	alp_app_set_default_context(ctxt2);
	alp_app_set_relaunch_handler(AppServerTest::ContextTestRelaunchHandler2, this);
	alp_app_add_exit_handler(AppServerTest::ContextTestExitHandler2, this);
	
	// Init test vars:
	mContext1Relaunched = false;
	mContext1Exited = false;
	mContext2Relaunched = false;
	mContext2Exited = false;
	arg = "";
	
	alp_app_launch_async(appID1, 0, &arg);
	if(mContext1Relaunched == false) g_main_loop_run(mLoop);
	
	CPPUNIT_ASSERT(mContext1Relaunched == true && 
			mContext2Relaunched == false && 
			mContext1Exited == false && 
			mContext2Exited == false);
	
	alp_app_launch_async(appID2, 0, &arg);
	if(mContext2Relaunched == false) g_main_loop_run(mLoop);
	
	CPPUNIT_ASSERT(mContext1Relaunched == true && 
			mContext2Relaunched == true && 
			mContext1Exited == false && 
			mContext2Exited == false);
	
	alp_app_exit(appID2);
	if(mContext1Exited == false) g_main_loop_run(mLoop);
	
	CPPUNIT_ASSERT(mContext1Relaunched == true && 
			mContext2Relaunched == true && 
			mContext1Exited == false && 
			mContext2Exited == true);
	
	alp_app_exit(appID1);
	if(mContext1Exited == false) g_main_loop_run(mLoop);
	
	CPPUNIT_ASSERT(mContext1Relaunched == true && 
			mContext2Relaunched == true && 
			mContext1Exited == true && 
			mContext2Exited == true);
	
	
	
	// Unregister & exit
	alp_app_destroy(ctxt1, 0);
	alp_app_destroy(ctxt2, 0);
	
	// Deallocate main loop
	g_main_loop_unref(mLoop);
	
	return;
}


void AppServerTest::ContextTestRelaunchHandler1(int argc, char* argv[], gpointer cbData)
{
	AppServerTest *ast = (AppServerTest*)cbData;
	ast->mContext1Relaunched = true;
	g_main_loop_quit(ast->mLoop);
}

void AppServerTest::ContextTestRelaunchHandler2(int argc,  char* argv[], gpointer cbData)
{
	AppServerTest *ast = (AppServerTest*)cbData;
	ast->mContext2Relaunched = true;
	g_main_loop_quit(ast->mLoop);
}

void AppServerTest::ContextTestExitHandler1(gpointer cbData)
{
	AppServerTest *ast = (AppServerTest*)cbData;
	ast->mContext1Exited = true;
	g_main_loop_quit(ast->mLoop);
}

void AppServerTest::ContextTestExitHandler2(gpointer cbData)
{
	AppServerTest *ast = (AppServerTest*)cbData;
	ast->mContext2Exited = true;
	g_main_loop_quit(ast->mLoop);
}
void AppServerTest::RandomAppRelaunchHandler(int argc,  char* argv[], gpointer cbData)
{
	g_warning("%s relaunched as primary\n", argv[0]);
	AppServerTest *ast = (AppServerTest*)cbData;
	if(strcmp(argv[1], ALP_APP_PRIMARY) == 0)
	{
		g_warning("%s relaunched as primary\n", argv[0]);
	}
	if(strcmp(argv[1], ALP_APP_TRANSIENT) == 0)
	{
		g_warning("%s relaunched as transient\n", argv[0]);
	}
	if(strcmp(argv[1], ALP_APP_BACKGROUNDED) == 0)
	{
		g_warning("%s relaunched as backgrounded\n", argv[0]);
	}
	g_main_loop_quit(ast->mLoop);
}

void AppServerTest::RandomAppExitHandler(gpointer cbData)
{
	g_warning("exiting\n");
	AppServerTest *ast = (AppServerTest*)cbData;
//	if(++(ast->mReceivedLaunchCount) > ast->mLaunchLimit)
//		g_main_loop_quit(ast->mLoop);
	g_main_loop_quit(ast->mLoop);
}

// **************************************
// Transient App support testing:
// **************************************

/*
	Test:  Do transient app contexts receive an exit request when a new FG app is registered?
*/
void AppServerTest::TransientAppTest(void)
{
	AlpAppContext *ctxt1, *ctxt2, *fgCtxt;
	char *appID1, *appID2, *fgAppID;
	char *arg;
	GMainLoop *ctxt1Loop;
	
	
	// Allocate contexts & register as transient:
	appID1 = "com.access.appserver.transientTest1";
	appID2 = "com.access.appserver.transientTest2";
	fgAppID = "com.access.appserver.transientTestFG";
	ctxt1 = alp_app_setup(appID1, ALP_APP_LAUNCH_FLAGS_DEFAULT, "");
	ctxt2 = alp_app_setup(appID2, ALP_APP_LAUNCH_FLAG_TRANSIENT, "");
	
	// Allocate main loops & set up exit handlers:
	alp_app_set_default_context(ctxt1);
	ctxt1Loop = g_main_loop_new(NULL, TRUE);
	alp_app_set_relaunch_handler(AppServerTest::AddTransientRelaunchHandler, ctxt1Loop);
	alp_app_set_backgrounding(ALP_APP_BACKGROUND_REQUIRED);
	
	alp_app_set_default_context(ctxt2);
	mLoop = g_main_loop_new(NULL, TRUE);
	alp_app_add_exit_handler(AppServerTest::TransientTestExitHandler, this);
	
	// it's important that the transient registrations be received FIRST.
	alp_app_flush(ctxt1);
	alp_app_flush(ctxt2);
	
	// Relaunch the first one as transient, to test "adding in" the transient flag.
	arg = ALP_APP_TRANSIENT;
	alp_app_launch_async(appID1, 1, &arg);
	g_main_loop_run(ctxt1Loop);
	
	alp_app_set_default_context(ctxt1);
	alp_app_set_relaunch_handler(AppServerTest::TransientRelaunchHandler, ctxt1Loop);
	
	
	// Then allocate & register the last one as a FG app.  
	// This should result in the transient contexts receiving an exit request.
	fgCtxt = alp_app_setup(fgAppID, ALP_APP_LAUNCH_FLAG_PRIMARY, "");
	alp_app_set_default_context(fgCtxt);
	
	// Wait for both transients to receive exit requests.
	// It's possible that the 2nd one will be received first, so a bit of fancy footwork is necessary to catch that case.
	// Otherwise, we could quit teh mainloop before we actually start running it, and then we'd block forever.
	mTransientContext2Finished = false;
	g_main_loop_run(ctxt1Loop);
	if(mTransientContext2Finished == false)
		g_main_loop_run(mLoop);

	
	// Unregister & exit
	alp_app_destroy(ctxt1, 0);
	alp_app_destroy(ctxt2, 0);
	
	// Unregister & exit
	alp_app_destroy(fgCtxt, 0);
	
	// Deallocate main loops
	g_main_loop_unref(ctxt1Loop);
	g_main_loop_unref(mLoop);
	
	return;
}

void AppServerTest::AddTransientRelaunchHandler (int argc,  char* argv[], gpointer cbData)
{
	g_main_loop_quit((GMainLoop*)cbData);
	
	CPPUNIT_ASSERT(argc == 2);
	CPPUNIT_ASSERT(strcmp(argv[1], ALP_APP_TRANSIENT) == 0);
	
	return;
}

void AppServerTest::TransientRelaunchHandler (int argc,  char* argv[], gpointer cbData)
{
	g_main_loop_quit((GMainLoop*)cbData);
	
	CPPUNIT_ASSERT(argc == 2);
	CPPUNIT_ASSERT(strcmp(argv[1], ALP_APP_BACKGROUNDED) == 0);
	
	return;
}

void AppServerTest::TransientTestExitHandler(gpointer cbData)
{
	AppServerTest *ast = (AppServerTest*)cbData;
	ast->mTransientContext2Finished = true;
	g_main_loop_quit(ast->mLoop);
}


// **************************************
// Launchpad testing:
// **************************************


void AppServerTest::LaunchpadRelaunchHandler(int argc,  char* argv[], gpointer cbData)
{
	AppServerTest *ast = (AppServerTest*)cbData;
	int count = ast->mReceivedLaunchCount;
	int i;
	
//	g_print("%s: received launch for %s\n", ast->mLaunchpadScheme, argv[0]);
	
	if(ast->mReceivedLaunchCount == kTotalTestLaunches)
	{
		g_print("%s: received too many launches.\n", ast->mLaunchpadScheme);
		exit(false);
	}
		
	if(ast->mArgcList[count] != argc) 
	{
		g_print("%s: Argc does not match, expected %s/%d, received %s/%d.\n", ast->mLaunchpadScheme, (*ast->mArgvList)[count][0], ast->mArgcList[count], argv[0], argc);
		exit(false);
	}
	
	for(i=0; i<argc; i++)
	{
//		g_print("Checking %s == %s...\n", (*ast->mArgvList)[count][i], argv[i]);
		
		if(strcmp((*ast->mArgvList)[count][i], argv[i]) != 0) 
			exit(false);
	}
	
	// Fake app registration... :D
	FakeAppRegistration(argv[0], ALP_APP_LAUNCH_FLAGS_DEFAULT);
	
	
	ast->mReceivedLaunchCount++;
	if(ast->mReceivedLaunchCount == kTotalTestLaunches)
	{
		ast->mPassed = true;
		g_main_loop_quit(ast->mLoop);
	}
	
}

void AppServerTest::SkeletonLaunchpad(void)
{
	AlpAppContext *ctxt;
	
	// Set up connection with app server, & register as launchpad.
	mLoop = g_main_loop_new(NULL, TRUE);
	ctxt = alp_app_setup(mAppID, ALP_APP_LAUNCH_FLAG_LAUNCHPAD, mLaunchpadScheme);
	alp_app_set_relaunch_handler(AppServerTest::LaunchpadRelaunchHandler, this);
	
	mReceivedLaunchCount = 0;
	mPassed = false;
	
	// Wait in mainloop for request:
	g_main_loop_run(mLoop);
			
	alp_app_destroy(ctxt, 0);
	
	// We need to return our "pass status" using exit() since we 
	// can't use CPPUNIT_ASSERT in a child process...
	exit(mPassed);
}


void AppServerTest::LaunchpadTest(void)
{
	pid_t launchpad1, launchpad2;
	int i;
	int status;
	
	// First, we fork two launchpad processes for different app schemes.
	if((launchpad1 = fork()) == 0)
	{
		// child, 1st launchpad process.
		
		int argcList[kTotalTestLaunches] = {2, 3, 2, 3, 5, 3};
		const char* argvList[kTotalTestLaunches][kTotalTestLaunchParams] = {
			{"red-app:com.test.pong", ALP_APP_PRIMARY, NULL, NULL, NULL, NULL},
			{"red-app:com.test.twink", ALP_APP_PRIMARY, "--alp-whatever", NULL, NULL, NULL},
						{"red-app:com.test.foo", "--alp-whatever", NULL, NULL, NULL, NULL},
						{"red-app:com.test.bar", "--alp-whatever", "--alp-whatever", NULL, NULL, NULL},
						{"red-app:com.test.baz", ALP_APP_PRIMARY, "--Twinkie", "--doEverything", "blahblahblah", NULL},
						{"red-app:com.test.Twiddle", "--alp-whatever", "--alp-bad-param", NULL, NULL, NULL}
						};
		
		mArgcList = argcList;
		mArgvList = &argvList;
		
		
		mAppID = "com.access.tests.launchpad.red-launcher";
		mLaunchpadScheme = "red-app";
		
		SkeletonLaunchpad();
		
	}
	else if((launchpad2 = fork()) == 0)
	{
		// other child, 2nd launchpad process.
		
		int argcList[kTotalTestLaunches] = {3, 3, 2, 4, 5, 3};
		const char* argvList[kTotalTestLaunches][kTotalTestLaunchParams] = {
			{"blue-app:com.test.calliope", ALP_APP_PRIMARY, "--impress", NULL, NULL, NULL},
			{"blue-app:com.test.saxophone", ALP_APP_PRIMARY, "--alp-whatever", NULL, NULL, NULL},
			{"blue-app:com.test.pennywhistle", "--alp-whatever", NULL, NULL, NULL, NULL},
			{"blue-app:com.test.violin", "--alp-whatever", "--alp-whatever", "--trill", NULL, NULL},
			{"blue-app:com.test.tuba", ALP_APP_PRIMARY, "--shine", "--noisy", "--extra-large", NULL},
			{"blue-app:com.test.guitar", "--alp-whatever", "--twang", NULL, NULL, NULL}
		};
		
		
		mAppID = "com.access.tests.launchpad.blue-launcher";
		mLaunchpadScheme = "blue-app";

		mArgcList = argcList;
		mArgvList = &argvList;
	
		SkeletonLaunchpad();
		
	}
	else
	{
		AlpAppContext *ctxt;
		// parent, and launcher.
		mAppID = "com.access.tests.launchpad.launchClient";
		
		mLoop = g_main_loop_new(NULL, TRUE);
		ctxt = alp_app_setup(mAppID, 0, NULL);
		
		// This is kind of cheating, but this sort of synchronization will not
		// normally be needed.  We sleep for a moment to give the child-launchpads 
		// a chance to register with the app server.
		sleep(1);
		
		
		// Now, we fire off a bunch of launch requests to both launchpads.
		// Each launchpad is expecting them to arrive in a certain order,
		// and here they are shuffled together.  The appserver is responsible
		// for dispatching them to the appropriate launchpad.
				
		char *red_argv1[] = {ALP_APP_PRIMARY, NULL, NULL, NULL, NULL};
		alp_app_launch("red-app:com.test.pong", 1, red_argv1, NULL);
		
		char *blue_argv1[] = {ALP_APP_PRIMARY, "--impress", NULL, NULL, NULL};
		alp_app_launch("blue-app:com.test.calliope", 2, blue_argv1, NULL);
		
		char *red_argv2[] = {ALP_APP_PRIMARY, "--alp-whatever", NULL, NULL, NULL};
		alp_app_launch("red-app:com.test.twink", 2, red_argv2, NULL);
		
		char *red_argv3[] = {"--alp-whatever", NULL, NULL, NULL, NULL};
		alp_app_launch("red-app:com.test.foo", 1, red_argv3, NULL);
		
		char *blue_argv2[] = {ALP_APP_PRIMARY, "--alp-whatever", NULL, NULL, NULL};
		alp_app_launch("blue-app:com.test.saxophone", 2, blue_argv2, NULL);
				
		char *red_argv4[] = {"--alp-whatever", "--alp-whatever", NULL, NULL, NULL};
		alp_app_launch("red-app:com.test.bar", 2, red_argv4, NULL);
		
		char *blue_argv3[] = {"--alp-whatever", NULL, NULL, NULL, NULL};
		alp_app_launch("blue-app:com.test.pennywhistle", 1, blue_argv3, NULL);
		
		char *blue_argv4[] = {"--alp-whatever", "--alp-whatever", "--trill", NULL, NULL};
		alp_app_launch("blue-app:com.test.violin" , 3, blue_argv4, NULL);
		
		char *red_argv5[] = {ALP_APP_PRIMARY, "--Twinkie", "--doEverything", "blahblahblah", NULL};
		alp_app_launch("red-app:com.test.baz", 4, red_argv5, NULL);
		
		char *blue_argv5[] = {ALP_APP_PRIMARY, "--shine", "--noisy", "--extra-large", NULL};
		alp_app_launch("blue-app:com.test.tuba", 4, blue_argv5, NULL);
		
		char *red_argv6[] = {"--alp-whatever", "--alp-bad-param", NULL, NULL, NULL};
		alp_app_launch("red-app:com.test.Twiddle", 2, red_argv6, NULL);
		
		char *blue_argv6[] = {"--alp-whatever", "--twang", NULL, NULL, NULL};
		alp_app_launch("blue-app:com.test.guitar", 2, blue_argv6, NULL);
		
		
		// Wait for red-app launchpad, & verify result:
		waitpid(launchpad1, &status, 0);
		CPPUNIT_ASSERT(WIFEXITED(status) && WEXITSTATUS(status) == true);
//		if(WIFEXITED(status)) g_print("red-app exited: %d\n", WEXITSTATUS(status));
		
		// Wait for blue-app launchpad, & verify result:
		waitpid(launchpad2, &status, 0);
		CPPUNIT_ASSERT(WIFEXITED(status) && WEXITSTATUS(status) == true);


//		if(WIFEXITED(status)) g_print("blue-app exited: %d\n", WEXITSTATUS(status));
//		if(WIFSIGNALED(status)) g_print("blue-app signalled: %s", PrvGetSignalName(WTERMSIG(status)));
				
		
		alp_app_destroy(ctxt, 0);
		
	}
	
	
	return;
}


// **************************************
// Test relaunching a launchpad...
// **************************************
void AppServerTest::LRRelaunchHandler (int argc, char* argv[], gpointer cbData)
{
	AppServerTest *ast = (AppServerTest*)cbData;
	CPPUNIT_ASSERT(argc == 2 && 
			strcmp(argv[0], ast->mAppID) == 0 &&
				strcmp(argv[1], "--test-arg") == 0); 
	ast->mPassed = TRUE;
	g_main_loop_quit(ast->mLoop);
}

/*
	There was a bug where a launchpad could not actually be relaunched itself,
	since it would re-register with appserver, but not specify the LP scheme,
	triggering a fatal error.
	This test case is to reproduce the problem so it can easily be fixed.
*/
void AppServerTest::LaunchpadRelaunchTest(void)
{
	AlpAppContext *ctxt;
	char *argv = "--test-arg";
	
	mAppID = "com.access.tests.launchpad.LaunchpadRelaunchTest";
	mPassed = FALSE;
	mLoop = g_main_loop_new(NULL, TRUE);
	ctxt = alp_app_setup(mAppID, ALP_APP_LAUNCH_FLAG_LAUNCHPAD, "testScheme");
	alp_app_set_relaunch_handler(AppServerTest::LRRelaunchHandler, this);
	
	alp_app_launch_async(mAppID, 1, &argv);
	
	g_main_loop_run(mLoop);
	
	
	CPPUNIT_ASSERT(mPassed == TRUE); 
	
	alp_app_destroy(ctxt, 0);
	
	
	return;
}

// **************************************
// Testing for "debug" launches - do we get correct PID, etc?
// **************************************
void AppServerTest::RequestPIDRelaunchHandler (int argc, char* argv[], gpointer cbData)
{
	AlpAppContext *appCtxt;
//	AppServerTest *ast = (AppServerTest*)cbData;
	int result = 0;
	
//	g_warning("Got relaunch for %s\n", argv[0]);
	
	if(strcmp(argv[0], "test:com.access.testapps.DebugTest") != 0) result = 1;
	if(argc != 3)
	{
		result = 1;
		g_warning("Wrong number of params:%d\n", argc);
	}
	
	if(strcmp(argv[2], "-setup") == 0)
	{
//		g_warning("setting up...\n");
		// Register & unregister w/ appserver.  This is the smallest possible app lifetime.
		appCtxt = alp_app_setup(argv[0],
					  alp_app_parse_launch_flags(ALP_APP_LAUNCH_FLAG_NONE, argc, argv), "");
		alp_app_destroy(appCtxt, 0);
	}
	else
	{
		result = 1;
		g_warning("Bad param:%s\n", argv[1]);
	}
	
	g_main_loop_quit((GMainLoop*)cbData);
	exit(result);
	
	return;
}



void AppServerTest::RequestPIDLaunchTest(void)
{
	AlpAppContext *lpCtxt;
	AlpAppContext *launcherCtxt;
	char *lpID, *appID;
	GMainLoop *loop;
	pid_t appPID, lpPID;
	alp_status_t err;
	int status;
	
	// Allocate contexts:
	lpID = "com.access.appserver.DebugLaunchLP";
	appID= "test:com.access.appserver.DebugLaunchApp";
	
	// Since launches are truly synchronous now, and we depend on handling messages while blocked, 
	// we need to do this with two processes.
	lpPID = fork();
	if(lpPID == 0)
	{
		// Child process for the launchpad.
		
		loop = g_main_loop_new(NULL, TRUE);
		
		// Initialize launchpad:
		lpCtxt = alp_app_setup(lpID, ALP_APP_LAUNCH_FLAG_LAUNCHPAD, "test");
		alp_app_set_default_context(lpCtxt);
		alp_app_set_relaunch_handler(RequestPIDRelaunchHandler, loop);
		
		g_main_loop_run(loop);
		g_main_loop_unref(loop);
		loop = NULL;
		
		
		// shut down "launchpad"
		alp_app_destroy(lpCtxt, 0);
		
		exit(0);
	}
	else
	{
		char *argv[] = {ALP_APP_DEBUG, "-setup"};
		sleep(1);
		
		launcherCtxt = alp_app_setup("launcher thingy", ALP_APP_LAUNCH_FLAG_NONE, "");
		alp_app_set_default_context(launcherCtxt);
		
//		g_print("Launching!\n");
		
		// Okay, now we try a test launch:
		appPID = 0;
		err = alp_app_launch("test:com.access.testapps.DebugTest", 2, argv, &appPID);
//		g_print("Done Launching\n");
		
		CPPUNIT_ASSERT(err == ALP_OK);
		CPPUNIT_ASSERT(appPID == lpPID);
	
		// Wait for child/launchpad, & make sure its exit status 
		// indicates that we received proper data.
		waitpid(lpPID, &status, 0);
		
		CPPUNIT_ASSERT(WIFEXITED(status));
		CPPUNIT_ASSERT(WEXITSTATUS(status) == 0);		
		
		alp_app_destroy(launcherCtxt, 0);		
	}
	
	
	
	return;
}


// **************************************
// Backgrounding setting testing:
// **************************************


/*
	This tests a few backgrounding settings cases:
	1) Given a foreground app with default settings, does it receive an exit request when a new one is registered?
	2) Given a foreground app which changes it's setting, is it moved to the BG when a new one is registered?
	3) If a background app changes it's setting to NOT_SUPPORTED, does it receive an exit request?
*/	
void AppServerTest::BackgroundingTest(void)
{
	AlpAppContext *oldMainUIApp;
	AlpAppContext *newMainUIApp;
	char *oldAppID = "com.access.testapp.oldMainUIApp";
	char *newAppID = "com.access.testapp.newMainUIApp";
	
	mLoop = g_main_loop_new(NULL, TRUE);
	
	// TEST #1:
	// Register one context as the main UI app.
	oldMainUIApp = alp_app_setup(oldAppID, ALP_APP_LAUNCH_FLAG_PRIMARY, NULL);
	alp_app_set_default_context(oldMainUIApp);
	alp_app_flush(oldMainUIApp);
	alp_app_add_exit_handler(BackgroundingExitHandler, this);
	
	// Make a new context, and register *it* as the new FG app.
	// The old context should get an exit request.
	
//	ALP_TM(TL(ALP_APP_TEST_TRACE, "Setting up..."));
	newMainUIApp = alp_app_setup(newAppID, ALP_APP_LAUNCH_FLAG_PRIMARY, NULL);
//	ALP_TM(TL(ALP_APP_TEST_TRACE, "Setup finished."));
	
	// Run main loop so we can receive the exit request:
	mPassed = FALSE;
	g_main_loop_run(mLoop);
	
	CPPUNIT_ASSERT(mPassed == TRUE);
	
	// Close old context since it got an exit request.
	alp_app_destroy(oldMainUIApp, 0);
	
	// TEST #2:
	// Change new context's backgrounding settings so it will stay running:
	alp_app_set_default_context(newMainUIApp);
	alp_app_set_backgrounding(ALP_APP_BACKGROUND_REQUIRED);
	alp_app_set_relaunch_handler(BackgroundingRelaunchHandler, this);
	alp_app_flush(newMainUIApp);
	
	// Then, recreate the "old" context, 
	// which will now displace the "new" context, 
	// which should be relaunched in the BG.
	oldMainUIApp = alp_app_setup(oldAppID, ALP_APP_LAUNCH_FLAG_PRIMARY, NULL);
	
	// Run main loop so we can receive the relaunch:
	mPassed = FALSE;
	g_main_loop_run(mLoop);
	
	CPPUNIT_ASSERT(mPassed == TRUE);
	
	// TEST #3:
	// Finally, now that the "new" context has been relaunched in the BG,
	// change its backgrounding settings and see if it receives an exit request.
	
	alp_app_add_exit_handler(BackgroundingExitHandler, this);
	alp_app_set_relaunch_handler(NULL, NULL);
	alp_app_set_backgrounding(ALP_APP_BACKGROUND_UNSUPPORTED);
	
	mPassed = FALSE;
	g_main_loop_run(mLoop);
	
	CPPUNIT_ASSERT(mPassed == TRUE);
	
	// Destroy both contexts:
	alp_app_destroy(newMainUIApp, 0);
	alp_app_destroy(oldMainUIApp, 0);
	
	// Deallocate main loop.
	g_main_loop_unref(mLoop);
	mLoop = NULL;
	
	return;
}

void AppServerTest::BackgroundingExitHandler(gpointer cbData)
{
	AppServerTest *ast = (AppServerTest*)cbData;
	
	// g_warning("Exit req received, exiting main loop.");
	
	if(ast != NULL)
	{
		g_main_loop_quit(ast->mLoop);
		ast->mPassed = TRUE;
	}
	
	return;
}

void AppServerTest::BackgroundingRelaunchHandler (int argc,  char* argv[], gpointer cbData)
{
	AppServerTest *ast = (AppServerTest*)cbData;
	
	CPPUNIT_ASSERT(strcmp(argv[1], ALP_APP_BACKGROUNDED) == 0);
	
	if(ast != NULL)
	{
		g_main_loop_quit(ast->mLoop);
		ast->mPassed = TRUE;
	}
	
	return;
}

// **************************************
// Test queued relaunching, combined w/ app exit.
// **************************************

static const char *QUEUED_RELAUNCH_TEST_ID="test:QueuedRelaunchApp";

// relaunch handler for the launchpad context
void AppServerTest::QueuedRelaunchLPHandler(int argc,  char* argv[], gpointer cbData)
{
	AppServerTest *ast = (AppServerTest*)cbData;
//	AlpAppContext **app = (AlpAppContext**)cbData;
	AlpAppContext *app;
	
	CPPUNIT_ASSERT(strcmp(argv[0], QUEUED_RELAUNCH_TEST_ID) == 0);
	
	if(strcmp(argv[1], "--first") == 0)
	{
		ast->mPassed = true;
		
		// set up & destroy "app".  
		// This should cause appserver to send the other launch request to the app, 
		// but it won't receive it because it never enters a main event loop.
		// Then, when it's destroyed, appserver will send it to us (the launchpad) again.
		app = alp_app_setup(QUEUED_RELAUNCH_TEST_ID, ALP_APP_LAUNCH_FLAGS_DEFAULT, NULL);
		alp_app_destroy(app, 0);
	}
	else if(strcmp(argv[1], "--second") == 0)
	{
		ast->mPassed2 = true;
		
		// set up & destroy the app again, to clear out the pending launch.  We passed!
		app = alp_app_setup(QUEUED_RELAUNCH_TEST_ID, ALP_APP_LAUNCH_FLAGS_DEFAULT, NULL);
		alp_app_destroy(app, 0);
		
		// Also, exit the main loop.
		g_main_loop_quit(ast->mLoop);
	}
	
	
	return;
}

void AppServerTest::QueuedRelaunchTest(void)
{
	AlpAppContext *lp;
	char *lpID = "com.access.testapp.QueuedRelaunchLP";
	char *appID = "com.access.testapp.QueuedRelaunchApp";
	alp_status_t err;
	char *arg;
	
	mLoop = g_main_loop_new(NULL, TRUE);
	
	// Set up launchpad app context:
	lp = alp_app_setup(lpID, ALP_APP_LAUNCH_FLAG_LAUNCHPAD, "test");
	alp_app_set_default_context(lp);
	alp_app_set_relaunch_handler(QueuedRelaunchLPHandler, this);
	alp_app_flush(lp);
	
	
	// TEST:
	// Launch the app twice, and flush the channel.
	// The launchpad will catch it, and setup/destroy the app.  
	// Launchpad should then receive another launch, when the pending one is dispatched.
	arg = "--first";
	alp_app_launch_async(QUEUED_RELAUNCH_TEST_ID, 1, &arg);
	
	arg = "--second";
	alp_app_launch_async(QUEUED_RELAUNCH_TEST_ID, 1, &arg);
	alp_app_flush(lp);
	
	mPassed = false;
	mPassed2 = false;
	
	g_main_loop_run(mLoop);
	
	CPPUNIT_ASSERT(mPassed && mPassed2);
	
	
	// Destroy launchpad context:
	alp_app_destroy(lp, 0);
	
	
	// Deallocate main loop.
	g_main_loop_unref(mLoop);
	mLoop = NULL;
	
	return;
}
// **************************************
// Test queued relaunching, combined w/ app exit.
// **************************************

void AppServerTest::UnresponsiveAppTest(void)
{
	pid_t child;
	AlpAppContext *app;
	char *appID = "com.access.testapps.UnresponsiveApp";
	int status = 0;
	
	// Test #1: Exit Request Kill Timer
	// Fork a child process to register as an app, 
	// send it an exit request, and see if it is killed when it does not exit.
	printf("- Please wait 6 seconds");
	child = fork();
	if(child == 0)
	{
		// Set up app:
		app = alp_app_setup(appID, ALP_APP_LAUNCH_FLAGS_DEFAULT, NULL);
		alp_app_set_default_context(app);
		alp_app_add_exit_handler(DoNothingExitHandler, this);
		alp_app_flush(app);		
		
		// Send ourselves an exit request:
		alp_app_exit(appID);
		
		// Wait forever in our mainloop.  
		// We should be killed after 3 seconds.
		mLoop = g_main_loop_new(NULL, TRUE);
		g_main_loop_run(mLoop);
		
		// Exit, just in case.
		exit(0);
	}
	
	waitpid(child, &status, 0);
	CPPUNIT_ASSERT(WIFSIGNALED(status));
	g_warning("Signal = %s", PrvGetSignalName(WTERMSIG(status)));
	CPPUNIT_ASSERT(WTERMSIG(status) == SIGKILL);
	
	
	// Test #2: Relaunch Kill Timer
	// Fork a child process to register as an app, 
	// send it a relaunch, and see if it is killed when it does not respond.
	child = fork();
	if(child == 0)
	{
		// Set up app:
		app = alp_app_setup(appID, ALP_APP_LAUNCH_FLAGS_DEFAULT, NULL);
		alp_app_set_default_context(app);
		alp_app_add_exit_handler(DoNothingExitHandler, this);
		alp_app_flush(app);		
		
		// Send ourselves a relaunch:
		alp_app_launch(appID, 0, NULL, NULL);
		
		// Wait forever in our mainloop.  
		// We should be killed after 3 seconds.
		mLoop = g_main_loop_new(NULL, TRUE);
		g_main_loop_run(mLoop);
		
		// Exit, just in case.
		exit(0);
	}
	
	waitpid(child, &status, 0);
	CPPUNIT_ASSERT(WIFSIGNALED(status));
	g_warning("Signal = %s", PrvGetSignalName(WTERMSIG(status)));
	CPPUNIT_ASSERT(WTERMSIG(status) == SIGKILL);
	
	return;
}

// **************************************
// Test Jump list API.
// **************************************
void AppServerTest::FakeRegisterRelaunchHandler(int argc,  char* argv[], gpointer cbData)
{
	AppServerTest *ast = (AppServerTest*)cbData;
	
	AlpAppContext *ctxt;
	
	// Register the app...
	FakeAppRegistration(argv[0], alp_app_parse_launch_flags(
								ALP_APP_LAUNCH_FLAGS_DEFAULT, 
								argc, argv));
	
	if(argc > 2 && strcmp(argv[2], "--quitMainLoop") == 0)
		g_main_loop_quit(ast->mLoop);
	
	
/*	while(argc>0)
	{
		argc--;
		g_print("argv[%d]: %s\n", argc, argv[argc]);
	}
*/
	return;	
}

void AppServerTest::JumpListTest(void)
{
	AlpAppContext *app;
	char **appIDs;
	int i, count;
	alp_status_t	err;
	gboolean *isRunnings;
	
	const char *appID1 = "test:com.access.testapps.JumpFoo";
	const char *appID2 = "test:com.access.testapps.JumpBar";
	const char *appID3 = "test:com.access.testapps.FooJumpBar";
	char *argv[1] = {ALP_APP_PRIMARY};
	char *argvQuit[2] = {ALP_APP_PRIMARY, "--quitMainLoop"};
	
//	alp_app_debug_break();
	
	mLoop = g_main_loop_new(NULL, TRUE);
	
	app = alp_app_setup("com.access.testapps.JumpListTest", ALP_APP_LAUNCH_FLAG_LAUNCHPAD, "test");
	alp_app_set_default_context(app);
	alp_app_set_relaunch_handler(FakeRegisterRelaunchHandler, this);
	
	// Test #1: Launch some apps, call API, see if the apps are there.
	
	// Send some launches to build up the list:
	alp_app_launch_async(appID1, 1, argv);
	alp_app_launch_async(appID2, 1, argv);
	alp_app_launch_async(appID3, 2, argvQuit);
	
	alp_app_flush(app);
	g_main_loop_run(mLoop);
	
	// Give appserver a chance to handle the disconnects.
	// This is not completely effective since the disconnects are
	// (obviously) on different connections than this one.
	alp_app_flush(app);
	
	err = alp_app_get_jump_list(&count, &appIDs, &isRunnings);
	if(err) g_warning("alp_app_get_jump_list failed: 0x%lx.", (long)err);
	CPPUNIT_ASSERT(err == ALP_OK);
	
	ALP_TM(TL(ALP_APP_TEST_TRACE, "Count = %d", count));
	i = 0;
	while(i < count)
	{
		ALP_TM(TL(ALP_APP_TEST_TRACE, "appIDs[%d]: %s, %s", i, appIDs[i], isRunnings[i] ? "RUNNING" : "dead"));
//		printf( "appIDs[%d]: %s, %s\n", i, appIDs[i], isRunnings[i] ? "RUNNING" : "dead");
		i++;
	}
	
	CPPUNIT_ASSERT(strcmp(appID1, appIDs[2]) == 0);
	CPPUNIT_ASSERT(strcmp(appID2, appIDs[1]) == 0);
	CPPUNIT_ASSERT(strcmp(appID3, appIDs[0]) == 0);
	
	// We cannot effectively verify that the test apps aren't running 
	// (the contexts have disconnected) since the fake app registrations 
	// which control whether or not our "test apps" are considered running 
	// are actually asyncronous - we send the unregister message and close 
	// the testapp's connection to appserver, but alp_app_destroy returns 
	// before appserver has handled this.  We flush *our* connection to 
	// appserver to at least make sure it's had a reasonable chance to 
	// handle pending messages.  But, this is not be a guarantee, since 
	// we are flushing a connection other than the one we disconnected, 
	// and IPC messages from different connections may not be guaranteed 
	// to be handled in the order they were received.
	// So, if these assertions were uncommented, they would (and did)
	// fail on rare occassions
//	CPPUNIT_ASSERT(isRunnings[0] == FALSE);
//	CPPUNIT_ASSERT(isRunnings[1] == FALSE);
//	CPPUNIT_ASSERT(isRunnings[2] == FALSE);
	
	// deallocate memory
	for(i=0; i<count; i++) g_free(appIDs[i]);
	g_free(appIDs);
	g_free(isRunnings);
	
	// dealloc test context
	alp_app_destroy(app, 0);
	
	return;
}

// **************************************
// PendingForeverTest:
// **************************************

// This is the launchpad handler for our test.
// Note that it does NOT register the apps... this tests the appserver's timeout
// function.  A launch request should not remain pending forever, blocking other launch
// requests.  So even if it is not registered, we should receive both launch requests.
void AppServerTest::ForeverRelaunchHandler(int argc,  char* argv[], gpointer cbData)
{
	AppServerTest *ast = (AppServerTest*)cbData;
	
	AlpAppContext *ctxt;
	
	if(strcmp(argv[1], "--first-launch") == 0)
	{
		ast->mPassed=TRUE;
	}
	
	else if(strcmp(argv[1], "--second-launch") == 0)
	{
		ast->mPassed2=TRUE;
		g_main_loop_quit(ast->mLoop);
	}
	
	return;	
}


void AppServerTest::PendingForeverTest(void)
{
	AlpAppContext *app, *lp;
	char **appIDs;
	int count;
	alp_status_t	err;
	
	
	mLoop = g_main_loop_new(NULL, TRUE);
	
	lp = alp_app_setup("com.access.testapps.ForeverLaunchpad", 
					ALP_APP_LAUNCH_FLAG_LAUNCHPAD, "forever");
	alp_app_set_default_context(lp);
	alp_app_set_relaunch_handler(ForeverRelaunchHandler, this);
	
	alp_app_flush(lp);
	
	mPassed = FALSE;
	mPassed2 = FALSE;
	
	
	//	... launch a "forever" app, and then launch it some more... 
	//	See if we ever get the second launch request.
	
	char *argv1[1]={"--first-launch"};
	char *argv2[1]={"--second-launch"};
	alp_app_launch_async("forever:testapp", 1, argv1);
	alp_app_launch_async("forever:testapp", 1, argv2);
	
	// Appserver should use relaunch timer on initial launch dispatch as well
	// and fail the launch if the registration is not received promptly.
	
	g_main_loop_run(mLoop);
	
	CPPUNIT_ASSERT(mPassed == TRUE);
	CPPUNIT_ASSERT(mPassed2 == TRUE);
	
	
	return;
}


// **************************************
// Testing of 'current app' API:
// **************************************
void AppServerTest::CurrAppTest(void)
{
	AlpAppContext *app, *lp;
	char *appName;
	int count;
	alp_status_t	err;
	
	
	mLoop = g_main_loop_new(NULL, TRUE);
	mAppID = "com.access.testapps.CurAppTest";
	
	// Make a new primary app
	app = alp_app_setup(mAppID, 
				    ALP_APP_LAUNCH_FLAG_PRIMARY, NULL);
	alp_app_set_default_context(app);
	alp_app_flush(app);
	
	// Check and see if it's correctly reported:
	err = alp_app_current(&appName);
	CPPUNIT_ASSERT(err == ALP_OK);
	CPPUNIT_ASSERT(strcmp(appName, "com.access.testapps.CurAppTest") == 0);
	
	// Re-register it as non-primary
	alp_app_register(app, mAppID, ALP_APP_LAUNCH_FLAG_NONE, NULL);
	alp_app_flush(app);
	
	// Make sure it's no longer reported:
	err = alp_app_current(&appName);
	CPPUNIT_ASSERT(err == ALP_OK);
	CPPUNIT_ASSERT(strcmp(appName, "com.access.testapps.CurAppTest") != 0);
	
	alp_app_destroy(app, 0);
	
	return;
}

// **************************************
// Testing of 'Watcher' API:
// **************************************
void AppServerTest::WatcherTest(void)
{
	AlpAppContext *app, *watcher;
	char *appName;
	int count;
	alp_status_t	err;
	char *watchID1 = "com.access.testapps.WatchedOne";
	char *watchID2 = "com.access.testapps.WatchedTwo";
	AlpWatcherRef ref1, ref2;
	gboolean ran1, ran2;
	
	mLoop = g_main_loop_new(NULL, TRUE);
	mAppID = "com.access.testapps.WatcherTest";
	
	// Make a new primary app
	watcher = alp_app_setup(mAppID, ALP_APP_LAUNCH_FLAG_NONE, NULL);
	alp_app_set_default_context(watcher);
	alp_app_flush(watcher);
	
	// Create watchers:
	err = alp_app_watcher_add(watchID1, &ref1, &ran1);
	CPPUNIT_ASSERT(err == ALP_OK);
	CPPUNIT_ASSERT(ref1 != 0);
	CPPUNIT_ASSERT(ran1 == FALSE);
	
	err = alp_app_watcher_add(watchID2, &ref2, &ran2);
	CPPUNIT_ASSERT(err == ALP_OK);
	CPPUNIT_ASSERT(ref2 != 0);
	CPPUNIT_ASSERT(ran2 == FALSE);
	
	// Check 'em a few times:
	err = alp_app_watcher_check(ref1, &ran1);
	CPPUNIT_ASSERT(err == ALP_OK);
	CPPUNIT_ASSERT(ran1 == FALSE);
	
	err = alp_app_watcher_check(ref2, &ran2);
	CPPUNIT_ASSERT(err == ALP_OK);
	CPPUNIT_ASSERT(ran2 == FALSE);
		
	err = alp_app_watcher_check(ref2, &ran2);
	CPPUNIT_ASSERT(err == ALP_OK);
	CPPUNIT_ASSERT(ran2 == FALSE);
		
	err = alp_app_watcher_check(ref2, &ran2);
	CPPUNIT_ASSERT(err == ALP_OK);
	CPPUNIT_ASSERT(ran2 == FALSE);

	err = alp_app_watcher_check(ref1, &ran1);
	CPPUNIT_ASSERT(err == ALP_OK);
	CPPUNIT_ASSERT(ran1 == FALSE);
	
	// "run" an app:
	app = alp_app_setup(watchID1, ALP_APP_LAUNCH_FLAG_NONE, NULL);
	alp_app_flush(app);
	alp_app_destroy(app, 0);
	
	// Check 'em a few more times, and see if the results are correct:
	err = alp_app_watcher_check(ref2, &ran2);
	CPPUNIT_ASSERT(err == ALP_OK);
	CPPUNIT_ASSERT(ran2 == FALSE);

	err = alp_app_watcher_check(ref1, &ran1);
	CPPUNIT_ASSERT(err == ALP_OK);
	CPPUNIT_ASSERT(ran1 == TRUE);
	
	// Make sure 'remove' also returns the right value:
	err = alp_app_watcher_remove(ref1, &ran1);			
	CPPUNIT_ASSERT(err == ALP_OK);
	CPPUNIT_ASSERT(ran1 == TRUE);
	
	err = alp_app_watcher_check(ref2, &ran2);
	CPPUNIT_ASSERT(err == ALP_OK);
	CPPUNIT_ASSERT(ran2 == FALSE);
	
	// Set second watcher to block launches, and make sure that works properly:
	err = alp_app_watcher_block(ref2, TRUE);
	CPPUNIT_ASSERT(err == ALP_OK);
	
	// Try launch, verify error code.
	err = alp_app_launch(watchID2, 0, NULL, NULL);
	CPPUNIT_ASSERT(err == ALP_STATUS_APP_BLOCKED);
		
	// second watcher is left allocated - appserver should deallocate it for us.
	
	alp_app_destroy(watcher, 0);
	
	return;
}

// **************************************
// Testing of 'abandon' PDK API:
// **************************************

void AppServerTest::AbandonApplication(void)
{
	pid_t child;
	AlpAppContext *app;
	char *appID = "com.access.testapps.AbandonApplication", *outPkgID=NULL;
	int status = 0;

	// Set up app:
	// intentionally getting a context before forking
	app = alp_app_setup(appID, ALP_APP_LAUNCH_FLAGS_DEFAULT, NULL);
	alp_app_add_exit_handler(DoNothingExitHandler, this);
	alp_app_flush(app);		

	if((child = fork()) == 0)
	{
		// Toss this duplicate context 
		alp_app_abandon(app);
		alp_status_t out = alp_app_current(&outPkgID);
		if(outPkgID) g_free(outPkgID);
		
		// We can't do a CPPUNIT_ASSERT from a child process, so we need to pass 
		// our pass/fail status back in our result code.
		if(out == ALP_STATUS_OK) exit(-1); // fail
		else exit(0); // pass
	}
	
	waitpid(child, &status, 0);
	
	// check pass/fail result
	CPPUNIT_ASSERT(WIFEXITED(status) && WEXITSTATUS(status) == 0);
	
	alp_app_destroy(app, 0);
	
	return;
}

// **************************************
// Testing of 'Retire' API:
// **************************************
void AppServerTest::RetireAppTest(void)
{
	AlpAppContext *app;
	char *appID = "com.access.testapps.RetireAppTest";
	
	mLoop = g_main_loop_new(NULL, TRUE);
	
	app = alp_app_setup(appID, ALP_APP_LAUNCH_FLAG_PRIMARY, NULL);
	alp_app_set_relaunch_handler(AppServerTest::RetireRelaunchHandler, this);
	alp_app_add_exit_handler(AppServerTest::RetireExitHandler, this);
	alp_app_set_backgrounding(ALP_APP_BACKGROUND_REQUIRED);
	alp_app_flush(app);
	
	// test: do we get a relaunch in the BG if we retire ourselves?
	alp_app_retire(appID);
	
	mPassed = mPassed2 = FALSE;
	g_main_loop_run(mLoop);
	CPPUNIT_ASSERT(mPassed == TRUE && mPassed2 == FALSE);
	
	alp_app_flush(app);
	
	// Prep for next test...
	// move back to primary:
	alp_app_register(app, appID, ALP_APP_LAUNCH_FLAG_PRIMARY, NULL);
	alp_app_flush(app);

	// No longer allow BG execution:
	alp_app_set_backgrounding(ALP_APP_BACKGROUND_UNSUPPORTED);
	alp_app_flush(app);
	
	// test: do we get an exit request if we're retire?
	// test: do we get the request if we do not specify our name?
	alp_app_retire(NULL);
	
	mPassed = mPassed2 = FALSE;
	g_main_loop_run(mLoop);
	CPPUNIT_ASSERT(mPassed == FALSE && mPassed2 == TRUE);
	
	alp_app_destroy(app, 0);
	
	
	return;
}

void AppServerTest::RetireRelaunchHandler(int argc,  char* argv[], gpointer cbData)
{	
	AppServerTest *ast = (AppServerTest *)cbData;
	
//	g_warning("RetireRelaunchHandler called");
	
	ast->mPassed = TRUE;
	ast->mPassed2 = FALSE;
	g_main_loop_quit(ast->mLoop);
}

void AppServerTest::RetireExitHandler(gpointer cbData)
{
	AppServerTest *ast = (AppServerTest *)cbData;
	
//	g_warning("RetireExitHandler called");
	
	ast->mPassed = FALSE;
	ast->mPassed2 = TRUE;
	g_main_loop_quit(ast->mLoop);
}

// **************************************
// Test infrastructure stuff below here:
// **************************************
void AppServerTest::DoNothingExitHandler(gpointer cbData)
{
	AppServerTest *ast = (AppServerTest*)cbData;
		
	return;
}

// This routine forks a new process, which registers as the indicated app and exits.
// The purpose is to clear out "launched" apps from the list of pending launches.
void AppServerTest::FakeAppRegistration(const char *appID, guint32 launchFlags)
{
	AlpAppContext *ctxt;
	
	ctxt = alp_app_setup(appID, launchFlags, NULL);
	alp_app_flush(ctxt);
	alp_app_destroy(ctxt, 0);

	return;
}

void AppServerTest::ExitMainLoopConnectCB (AlpConnection* connection, gpointer cbData)
{	
	// exit the mainloop we use to block waiting for the connection...
	g_main_loop_quit((GMainLoop*)cbData);
	return;
}

void AppServerTest::ExitMainLoopHandler(gpointer cbData)
{
//	g_print("Exit handler was called, terminating main loop 0x%lx\n", (unsigned long)cbData);
	g_main_loop_quit((GMainLoop*)cbData);
	
	return;
}

int AppServerTest::ForkTest(void (AppServerTest::*proc) ())
{
	int status = 0;
	
	pid_t pid = fork();
	
	if(pid < 0) g_critical("Can't fork tests.");
	if(pid == 0) (this->*proc)();
	else
	{
		waitpid(pid, &status, 0);
		
//		if(WIFEXITED(status)) g_print("Test finished, result=%d\n", WEXITSTATUS(status));
//		else if(WIFSIGNALED(status)) g_print("Test finished due to unhandled signal %s\n",
//									PrvGetSignalName(WTERMSIG(status)));
//		else g_print("Test finished some other UNRECOGNIZED way.\n");
	}
	
	return status;
}

const char* AppServerTest::PrvGetSignalName(int sigid)
{
	switch (sigid)
	{
		case 1: return "SIGHUP - death of controlling process";
		case 2: return "SIGINT - interrupt from keyboard";
		case 3: return "SIGQUIT - quit from keyboard";
		case 4: return "SIGILL - illegal instruction";
		case 5: return "SIGTRAP - Trace/Breakpoint Trap";
		case 6: return "SIGABRT - abort signal";
		case 7: return "SIGEMT - Emulation trap";
		case 8: return "SIGFPE - floating point exception";
		case 9: return "SIGKILL - killed";
		case 10: return "SIGBUS - Bus error";
		case 11: return "SIGSEGV - invalid memory reference";
		case 12: return "SIGSYS - bad system call";
		case 13: return "SIGPIPE - write to pipe with no readers";
		case 14: return "SIGALARM - alarm clock";
		case 15: return "SIGTERM - terminated";
		case 16: return "SIGUSR1 - User signal 1";
		
	}
	
	return "Unknown signal";
}

void willNotTest()
{
	// these are included here to trick the completion script, because they will not be used in production code for reasons below
	alp_app_debug_break(); // because it's no longer automated if you do this
	;
	/* can't really test this one, since the message won't have a 
	response token until we actually call "Send with response".  No big deal.
	*/
	alp_app_prv_make_jump_list_msg(NULL, 0, NULL, NULL);
	alp_app_prv_unpack_jump_list_msg(NULL, NULL, NULL, NULL);
	
	// Can't really test these because the only measurable affects are totally handled 
	// by the appserver client library - there's not really anything we can check.
	// It will be obvious to the user if this is broken though, as the status bar & 
	// other running apps will not obey theme changes.
	alp_app_theme_update();
	alp_app_set_theme_update_proc(NULL, NULL);
	
}

void AppServerTest::ContextLongHandlerTest(void)
{
	AlpAppContext *ctxt1;
	const char *appID1;
	char *arg;

	mLoop = g_main_loop_new(NULL, TRUE);

	// Allocate & register context:
	appID1 = "com.access.appserver.contextTest1";
	ctxt1 = alp_app_setup(appID1, ALP_APP_LAUNCH_FLAGS_DEFAULT, "");

	alp_app_set_relaunch_handler(AppServerTest::ContextTestHangingRelaunchHandler1, this);
	alp_app_add_exit_handler(AppServerTest::ContextTestHangingExitHandler1, this);

	arg = "";

	alp_app_launch_async(appID1, 0, &arg);
	g_main_loop_run(mLoop);

	alp_app_exit(appID1);
	
	// Unregister & exit
	alp_app_destroy(ctxt1, 0);

	// Deallocate main loop
	g_main_loop_unref(mLoop);

	return;
}

void AppServerTest::RelaunchTestHandlersThatLaunch(void)
{
	AlpAppContext *ctxt1;
	const char *appID1;
	char *arg;

	mLoop = g_main_loop_new(NULL, TRUE);

	// Allocate & register context:
	appID1 = "com.access.appserver.contextTest1";
	ctxt1 = alp_app_setup(appID1, ALP_APP_LAUNCH_FLAGS_DEFAULT, "");

	alp_app_set_relaunch_handler(AppServerTest::ExternalLaunchRelaunchHandler, this);
	alp_app_add_exit_handler(AppServerTest::ExternalLaunchExitHandler, this);

	arg = "";

	alp_app_launch_async(appID1, 0, &arg);
	g_main_loop_run(mLoop);

	alp_app_exit(appID1);
	
	// Unregister & exit
	alp_app_destroy(ctxt1, 0);

	// Deallocate main loop
	g_main_loop_unref(mLoop);

	return;
}


void AppServerTest::ExternalLaunchRelaunchHandler(int argc,  char* argv[], gpointer cbData)
{
	AppServerTest *ast = (AppServerTest*)cbData;
	// see what happens now 
	alp_app_launch("com.access.apps.duck", 0, NULL, NULL);
	g_main_loop_quit(ast->mLoop);
}

void AppServerTest::ExternalLaunchExitHandler(gpointer cbData)
{
	AppServerTest *ast = (AppServerTest*)cbData;
	alp_app_launch("com.access.apps.duck", 0, NULL, NULL);
	g_main_loop_quit(ast->mLoop);
}

void AppServerTest::ContextTestHangingRelaunchHandler1(int argc,  char* argv[], gpointer cbData)
{
	sleep(60*1000); // sleep for a minute - simulate a long running handler
}

void AppServerTest::ContextTestHangingExitHandler1(gpointer cbData)
{
	sleep(60*1000); // sleep for a minute - simulate a long running handler
}

void AppServerTest::RandomModalLaunchTest(void)
{
	const char *appidA = "com.access.testapps.musicplayer";
	const char *appidB = "com.access.testapps.networkstreamer";
	const char *appidC = "com.access.testapps.webbrowser";
	const char *appidD = "com.access.testapps.icrashalot";
	const char *appidE = "com.access.testapps.volumeadjuster";
	const char *appids[5]={ appidA, appidB, appidC, appidD, appidE };
	char *arg = "";
	uint i = 0, j=0; 
	int children[5];// our little process army
	alp_status_t err;
	AlpWatcherRef *ref[5];
	AlpAppContext *launchprocess, *musicplayer, *networkstreamer, *webbrowser, *icrashalot, *volumeadjuster;
	gboolean *ran[5];
	
//	g_warning("\nallocation finished\n");
	mLaunchLimit=50;

	// Allocate & register contexts with various requests for background, transient, etc.

	
	if((children[0] == fork()) == 0)
	{
	musicplayer = alp_app_setup(appidA, ALP_APP_LAUNCH_FLAGS_DEFAULT, NULL);
	alp_app_set_default_context(musicplayer);
	alp_app_set_relaunch_handler(AppServerTest::RandomAppRelaunchHandler, this);
	alp_app_set_backgrounding(ALP_APP_BACKGROUND_SUPPORTED);
	mLoop = g_main_loop_new(NULL, TRUE);
	g_main_loop_run(mLoop);
	exit(0);
	}
	if((children[1] == fork()) == 0)
	{
	networkstreamer = alp_app_setup(appidB, ALP_APP_LAUNCH_FLAGS_DEFAULT, NULL);	
	alp_app_set_default_context(networkstreamer);
	alp_app_set_relaunch_handler(AppServerTest::RandomAppRelaunchHandler, this);
//	alp_app_add_exit_handler(AppServerTest::DoNothingExitHandler, this);
	alp_app_set_backgrounding(ALP_APP_BACKGROUND_REQUIRED);
	mLoop = g_main_loop_new(NULL, TRUE);
	g_main_loop_run(mLoop);
	exit(0);
	}
	
	if((children[2] = fork()) == 0)
	{
		webbrowser = alp_app_setup(appidC, ALP_APP_LAUNCH_FLAGS_DEFAULT, NULL);
		alp_app_set_default_context(webbrowser);
		alp_app_set_relaunch_handler(AppServerTest::RandomAppRelaunchHandler, this);
		//	alp_app_add_exit_handler(AppServerTest::DoNothingExitHandler, this);
	mLoop = g_main_loop_new(NULL, TRUE);
	g_main_loop_run(mLoop);
	exit(0);
	}
	
//	icrashalot = alp_app_setup(appidD, ALP_APP_LAUNCH_FLAGS_DEFAULT, NULL);
//	alp_app_set_default_context(icrashalot);
//	alp_app_set_relaunch_handler(AppServerTest::RandomAppRelaunchHandler, this);
////	alp_app_add_exit_handler(AppServerTest::DoNothingExitHandler, this);
//	
//	mLoop = g_main_loop_new(NULL, TRUE);
//	g_main_loop_run(mLoop);
//	exit();

	if((children[4] == fork()) == 0)
	{
		volumeadjuster = alp_app_setup(appidE, ALP_APP_LAUNCH_FLAG_TRANSIENT, NULL);
		alp_app_set_default_context(volumeadjuster);
		alp_app_set_relaunch_handler(AppServerTest::RandomAppRelaunchHandler, this);
		//	alp_app_add_exit_handler(AppServerTest::DoNothingExitHandler, this);
		mLoop = g_main_loop_new(NULL, TRUE);
		g_main_loop_run(mLoop);
		exit(0);
	}

	// set up watchers for each
//	err = alp_app_watcher_add(appidA, ref[0], ran[0]);
//	CPPUNIT_ASSERT(err == ALP_OK);
//	CPPUNIT_ASSERT(ref[0] != 0);
//	CPPUNIT_ASSERT(ran[0] == FALSE);
//	
//	err = alp_app_watcher_add(appidB, ref[1], ran[1]);
//	CPPUNIT_ASSERT(err == ALP_OK);
//	CPPUNIT_ASSERT(ref[1] != 0);
//	CPPUNIT_ASSERT(ran[1] == FALSE);
//	
//	err = alp_app_watcher_add(appidC, ref[2], ran[2]);
//	CPPUNIT_ASSERT(err == ALP_OK);
//	CPPUNIT_ASSERT(ref[2] != 0);
//	CPPUNIT_ASSERT(ran[2] == FALSE);
//	
//	err = alp_app_watcher_add(appidD, ref[3], ran[3]);
//	CPPUNIT_ASSERT(err == ALP_OK);
//	CPPUNIT_ASSERT(ref[3] != 0);
//	CPPUNIT_ASSERT(ran[3] == FALSE);
//	
//	err = alp_app_watcher_add(appidE, ref[4], ran[4]);
//	CPPUNIT_ASSERT(err == ALP_OK);
//	CPPUNIT_ASSERT(ref[4] != 0);
//	CPPUNIT_ASSERT(ran[4] == FALSE);
//	
//	g_warning("\nwatchers not added yet\n");

	// launch semi-randomly
	srand(50); // intentionally set to 50 each time for now
	int child = 0, gchild = 0;
	pid_t PID = 0;
	
	mLoop = g_main_loop_new(NULL, TRUE);
//	printf("Loop allocated\n");

	// fork the launching process here 
	if((child = fork()) == 0)
	{
		g_warning("starting launcher process\n");
		// make a new appcontext 
		launchprocess = alp_app_setup("com.access.testapps.randomlauncher",ALP_APP_LAUNCH_FLAGS_DEFAULT, NULL);
		alp_app_set_default_context(launchprocess);
		g_warning("starting launcher process\n");
		// launches done from here
		int ts = 0;
		char* launchtype;
		for( i = 0; i < mLaunchLimit; i++)
		{
			g_warning("starting launcher process\n");
			j=rand()%5;
			ts=rand()%5;
			switch(ts)
			{
				case 1:
					launchtype = ALP_APP_BACKGROUNDED;
					break;
				case 2:
					launchtype = ALP_APP_TRANSIENT;
					break;
				case 3:
					launchtype = ALP_APP_NOTIFY;
					break;
				case 4:
					launchtype = ALP_APP_EXCHANGE;
					break;
				default:
					launchtype = ALP_APP_PRIMARY;
					break;
			}
			g_warning("Launching %s as %s\n", appids[j], launchtype);

			alp_app_launch(appids[j], 0, &launchtype, &PID); 
//			alp_app_launch_async(appids[j], 0, &launchtype);
			sleep(500);
			alp_app_exit(appids[j]);
		}
		exit(0);
	}
	g_warning("starting gmainloop\n");
	g_main_loop_run(mLoop);
	// write a handler that exits at 50 launches
	
	g_warning("\nran %d apps\n", mLaunchLimit);
	
	// Deallocate main loop
	g_main_loop_unref(mLoop);
	
	// Unregister & exit
	alp_app_destroy(musicplayer, 0);
	alp_app_destroy(networkstreamer, 0);
	alp_app_destroy(webbrowser, 0);
	alp_app_destroy(icrashalot, 0);
	alp_app_destroy(volumeadjuster, 0);
	alp_app_destroy(launchprocess, 0);
}

//	-exit w/o handler
void AppServerTest::NoExitHandlerTest(void)
{
	mSkipExitHandler = true;
	mSkipRelaunchHandler = false;
	LaunchFull();
}

//	-relaunch w/o handler
void AppServerTest::NoRelaunchHandlerTest(void)
{
	mSkipExitHandler = false;
	mSkipRelaunchHandler = true;
	LaunchFull();
}

// 	neither type of handler registered
void AppServerTest::NoHandlerTest(void)
{
	mSkipExitHandler = true;
	mSkipRelaunchHandler = true;
	LaunchFull();
}

class AppServerTest_AP : public AppServerTest
{
	CPPUNIT_TEST_SUITE( AppServerTest_AP );
	CPPUNIT_TEST( CleanExit );
	CPPUNIT_TEST( SIGTERM_ExitTest );
	CPPUNIT_TEST( ExitHandlerTest );
	CPPUNIT_TEST( RelaunchHandlerTest );
	CPPUNIT_TEST( ArgcArgvPackingTest );
	CPPUNIT_TEST( MessagePackingTest );
	CPPUNIT_TEST( AppContextTest );
	CPPUNIT_TEST( TransientAppTest );
	CPPUNIT_TEST( LaunchpadTest );
	CPPUNIT_TEST( LaunchpadRelaunchTest );
	CPPUNIT_TEST( LaunchFull );
	CPPUNIT_TEST( NoRelaunchHandlerTest );
	CPPUNIT_TEST( NoExitHandlerTest );
	CPPUNIT_TEST( NoHandlerTest );
	CPPUNIT_TEST( RequestPIDLaunchTest );
	CPPUNIT_TEST( BackgroundingTest );
	CPPUNIT_TEST( QueuedRelaunchTest );
	CPPUNIT_TEST( JumpListTest );
	CPPUNIT_TEST( CurrAppTest );
	CPPUNIT_TEST( WatcherTest );
	CPPUNIT_TEST( AbandonApplication );
	CPPUNIT_TEST( RetireAppTest );
//	CPPUNIT_TIMED_TEST( RandomModalLaunchTest, 300 );
	CPPUNIT_TEST_SUITE_END();
	// definitions in parent class
};
class AppServerTest_AN : public AppServerTest
{
	CPPUNIT_TEST_SUITE( AppServerTest_AN );
	CPPUNIT_TEST( LaunchErrorTest );
	CPPUNIT_TEST( UnexpectedExit );
	CPPUNIT_TEST( UnresponsiveAppTest );
	CPPUNIT_TEST( PendingForeverTest );
	CPPUNIT_TEST( RelaunchTestHandlersThatLaunch);
	CPPUNIT_TEST_SUITE_END();
	// definitions in parent class
};

class AppServerTest_MS : public AppServerTest
{
	CPPUNIT_TEST_SUITE( AppServerTest_MS );
	CPPUNIT_TIMED_TEST( ContextLongHandlerTest , 200 );
	CPPUNIT_TEST_SUITE_END();
	// definitions in parent class
};
ALP_ROOT_TEST_SUITE_REGISTRATION( "AP-001", CPPUNIT_REGISTRY_TYPE_AUTO );
CPPUNIT_TEST_SUITE_NAMED_REGISTRATION( AppServerTest_AP, "AP-001" );
CPPUNIT_TEST_SUITE_NAMED_REGISTRATION( AppServerTest_AN, "AP-001" );

