/******************************************************************************
 *
 * Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the entire contents of this file.
 *
 * The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.
 *
 * Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
 * Rights Reserved.
 * 
 * Contributor(s): none.
 *
 *****************************************************************************/
/**
 * @file attn.c
 *
 * @brief ALP attention manager API implementation
 *
 *****************************************************************************/
#include <stdarg.h>
#include <unistd.h>
#include <string.h>

#include <glib.h>

#include <hiker/types.h>
#include <hiker/bundlemgr.h>
#include <hiker/ipc.h>
#include <hiker/attnmgr.h>
#include <hiker/prv/attnmgr_traces.h>
#include <hiker/prv/attnmgr_ipc.h>

#ifndef ALP_ATTN_DIRECT_DB
#define ALP_ATTN_DIRECT_DB false
#endif

#if ALP_ATTN_DIRECT_DB
#include <hiker/prv/attnmgr_dml.h>
#endif

#define ALP_ATTN_MAX_CALLBACK_ARGS 30

static const char *sAlpAttnCallbackChannelName;
static AlpChannel *sAlpAttnCallbackChannel;
static GList      *sAlpAttnCallbackList;

typedef struct _AlpAttnCallbackInfo AlpAttnCallbackInfo;
struct _AlpAttnCallbackInfo {
    char          *sourceId;  // app-id of posting application / service 
    char          *alertType; // alert type name (i.e. "alarm")
    char          *handle;    // app supplied opaque string handle
    AlpAttnAlertCB callback;  // call back to invoke
    void          *user_data; // user supplied data to pass to the callback
};

static void
alp_prv_attn_receive_cb (AlpConnection *connection,
                         AlpMessage    *message,
                         gpointer       cbData)
{
    alp_status_t            err = ALP_STATUS_OK;
    AlpAttnCallbackInfo    *callbackInfo = NULL;
    gchar                  *sourceId  = NULL;
    gchar                  *alertType = NULL;
    gchar                  *handle    = NULL;
    guint                   argc;
    char                  **argv = NULL;
    gint                    msgId;
    GList                  *list      = NULL;    
    guint                   i;

    msgId = alp_ipc_message_get_message_ID (message);
	ALP_ATLX (API, "Received message ID %x from ...", msgId);

    /* unpack the message */
    err = alp_ipc_message_unpack_start (message);
    if (!err) err = alp_ipc_message_unpack_string (message, &sourceId);
    if (!err) err = alp_ipc_message_unpack_string (message, &alertType);
    if (!err) err = alp_ipc_message_unpack_string (message, &handle);
    if (!err) err = alp_ipc_message_unpack_uint32 (message, &argc);

    if (err) {  /* should drop message - leaves callback hanging */
        ALP_ATLX (API,
                  "ipc failure while unpacking callback header (err = %x)",
                  err);
        goto cleanup;
    }

    /* sanity check the argc, limit the number of accepted args */
    if (argc > ALP_ATTN_MAX_CALLBACK_ARGS) {
        ALP_ATLX(
            API,
            "received bogus callback args (%d), more than the allowed (%d)",
            argc,
            ALP_ATTN_MAX_CALLBACK_ARGS);
        alp_ipc_message_unpack_end(message);
        goto cleanup;
    }

    argv = NULL;
    if (argc) {
        argv = g_new (char *, argc);
        if (argv == NULL) {
            ALP_ATLX(API,
                     "No memory for array of %d args. "
                     "Callback not called for alert %s,%s,%s",
                     argc, sourceId, alertType, handle);
            alp_ipc_message_unpack_end(message);
            goto cleanup;
        }
        for (i = 0; i < argc; i++) {
            gchar *arg;
            err = alp_ipc_message_unpack_string (message, &arg);
            if (err) {
                ALP_ATLX(API, "error unpacking string from message: %x", err);
                alp_ipc_message_unpack_end(message);
                goto cleanup;
            }
            argv[i] = arg;
        }
    }

    /* search thru callback info list for the proper callback */
    callbackInfo = NULL;
    list = g_list_first (sAlpAttnCallbackList);
    while (list) {
        AlpAttnCallbackInfo *info;

        info = (AlpAttnCallbackInfo *) list->data;
        if ((strcmp(info->sourceId, sourceId) == 0) &&
            (strcmp(info->alertType, alertType) == 0) &&
            (strcmp(info->handle, handle) == 0)) {
            
            callbackInfo = info;

            sAlpAttnCallbackList = g_list_remove_link (sAlpAttnCallbackList, list);
            list->data = NULL;
            g_list_free_1 (list);
            break;
        }
        list = g_list_next (list);
    }

    if (callbackInfo && callbackInfo->callback) {
        ALP_ATLX(API, "calling callback %p for %s,%s,%s with %d args:",
                 callbackInfo->callback,
                 callbackInfo->sourceId,
                 callbackInfo->alertType,
                 callbackInfo->handle,
                 argc);
        for (i = 0; i < argc; i++) {
            ALP_ATLX(API,"    argv[%d] = %s", i, argv[i]);
        }
        (*callbackInfo->callback)(sourceId, alertType, handle,
                                  argc, argv,
                                  callbackInfo->user_data);
    }

cleanup:
    /* destroy callback info */
    if (callbackInfo) {
        /* remove from list */
        g_free ((gchar *)callbackInfo->sourceId);
        g_free ((gchar *)callbackInfo->alertType);
        g_free ((gchar *)callbackInfo->handle);
        g_free (callbackInfo);

    }

    g_free (sourceId);
    g_free (alertType);
    g_free (handle);
    g_free (argv);

    /* destroy message */
    if (message) {
        alp_ipc_message_destroy (message);
    }

    /* if no pending callbacks expected, destroy channel */
}

static void
alp_prv_attn_connect_cb (AlpConnection *connection,
                         gpointer cbData)
{
    alp_status_t err;

    /* register receive callback */
	err =
        alp_ipc_connection_register_receive_callback (connection,
                                                      alp_prv_attn_receive_cb,
                                                      cbData);
}

static void
alp_prv_attn_disconnect_cb (AlpConnection *connection,
                            gpointer       cbData)
{
    alp_status_t err;

    /* TODO: what to do here ? */
}

static void
alp_prv_attn_setup_callback (const char *sourceId,
                             const char *alertType,
                             const char *handle,
                             AlpAttnAlertCB callback,
                             void       *user_data)
{
    alp_status_t         err;
    AlpAttnCallbackInfo *callbackInfo;

    /* create callback channel name */
    if (sAlpAttnCallbackChannelName == NULL) {
        sAlpAttnCallbackChannelName =
            g_strdup_printf(ALP_ATTN_CALLBACK_CHANNEL_FORMAT, getpid());
        if (sAlpAttnCallbackChannelName == NULL) {
            ALP_ATLX(API, "no memory to store callback channel name");
            return;
        }
    }

    /* setup callback channel */
    if (sAlpAttnCallbackChannel == NULL) {
        sAlpAttnCallbackChannel =
            alp_ipc_channel_create (sAlpAttnCallbackChannelName,
                                    ALP_IPC_CHANNEL_ACCESS_ALL);
        if (sAlpAttnCallbackChannel == NULL) {
            ALP_FAIL_FATAL_IF(TRUE, "Callback channel creation failed.");
            return;
        }
    }

    callbackInfo = g_new0 (AlpAttnCallbackInfo, 1);
    if (callbackInfo == NULL) {
        ALP_ATLX(API, "no memory for callback info %s,%s,%s",
                 sourceId, alertType, handle);
        goto cleanup;
    }
    callbackInfo->sourceId  = g_strdup (sourceId);
    callbackInfo->alertType = g_strdup (alertType);
    callbackInfo->handle    = g_strdup (handle);
    callbackInfo->callback  = callback;
    callbackInfo->user_data = user_data;

    sAlpAttnCallbackList = g_list_prepend (sAlpAttnCallbackList,
                                           callbackInfo);
    callbackInfo = NULL;

	alp_ipc_channel_register_connect_callback(sAlpAttnCallbackChannel,
                                              alp_prv_attn_connect_cb,
                                              NULL);
	alp_ipc_channel_register_disconnect_callback(sAlpAttnCallbackChannel,
                                                 alp_prv_attn_disconnect_cb,
                                                 NULL);
cleanup:
    if (callbackInfo) {
        g_free(callbackInfo->sourceId);
        g_free(callbackInfo->alertType);
        g_free(callbackInfo->handle);
        g_free(callbackInfo);
    }

    /* check if callback list is empty, if so destroy connection */
    if (g_list_first (sAlpAttnCallbackList) == NULL) {
        ;
    }
}

alp_status_t
alp_attn_alert_post (const char  *source,
                     const char  *name,
                     const char  *handle,
                     const char  *interface,
                     int32_t      priority,
                     int32_t      duration,
                     uint32_t     nprops,
                     AlpAttnProp *properties[])
{
    alp_status_t  err  = ALP_STATUS_OK;
    alp_status_t  status;
    AlpMessage   *msg  = NULL;
    AlpMessage   *resp = NULL;
    AlpAttnId     alertId;

    if (source == NULL || name == NULL || handle == NULL ||
        priority > ALP_ATTN_PRIO_RANGE_UPPER) {
        ALP_ATLX(API, "bad parameters  %s,%s,%s prio %d",
                 source, name, handle, priority);
        return ALP_STATUS_ATTN_BAD_PARAM;
    }
    if (nprops && properties == NULL) {
        ALP_ATLX(API, "bad prop parameters  %d %x",
                 nprops, (unsigned)properties);
        return ALP_STATUS_ATTN_BAD_PARAM;
    }

    err = alp_attn_ipc_setup();
    if (err == ALP_STATUS_OK) {
        msg = alp_attn_ipc_pack_alert_post_msg (
            source,
            name,
            handle,
            interface,
            priority,
            duration,
            nprops,
            properties,
            NULL);

        if (msg) {
            err = alp_attn_ipc_send_with_response (msg, &resp);
            if (err == ALP_STATUS_OK) {
                /* decode response and get alert id */
                err = alp_attn_ipc_unpack_alert_post_reply (resp,
                                                            &status,
                                                            &alertId);
                if (err == ALP_STATUS_OK) {
                    err = status;
                }
                /* done with response, now free it */
                alp_ipc_message_destroy (resp);
            }
            alp_ipc_message_destroy (msg);
        }
        alp_attn_ipc_teardown();
    }
    if (err) {
        ALP_ATLX(API, "returning err %x alert %s,%s,%s id %d",
                 err, source, name, handle, alertId);
    }

    return err;
}

alp_status_t
alp_attn_alert_post_with_callback (
    const char  *source,
    const char  *name,
    const char  *handle,
    const char  *interface,
    int32_t      priority,
    int32_t      duration,
    uint32_t     nprops,
    AlpAttnProp *properties[],
    AlpAttnAlertCB callback,
    void        *user_data)
{
    alp_status_t  err  = ALP_STATUS_OK;
    alp_status_t  status;
    AlpMessage   *msg  = NULL;
    AlpMessage   *resp = NULL;
    AlpAttnId     alertId;

    if (source == NULL || name == NULL || handle == NULL ||
        priority > ALP_ATTN_PRIO_RANGE_UPPER) {
        ALP_ATLX(API, "bad parameters  %s,%s,%s prio %d",
                 source, name, handle, priority);
        return ALP_STATUS_ATTN_BAD_PARAM;
    }
    if (nprops && properties == NULL) {
        ALP_ATLX(API, "bad prop parameters  %d %x",
                 nprops, (unsigned)properties);
        return ALP_STATUS_ATTN_BAD_PARAM;
    }
    if (callback == NULL) {
        ALP_ATLX(API, "bad callback %x", (unsigned)callback);
        return ALP_STATUS_ATTN_BAD_PARAM;
    }

    alp_prv_attn_setup_callback (source, name, handle, callback, user_data);

    err = alp_attn_ipc_setup();
    if (err == ALP_STATUS_OK) {
        msg = alp_attn_ipc_pack_alert_post_msg (
            source,
            name,
            handle,
            interface,
            priority,
            duration,
            nprops,
            properties,
            sAlpAttnCallbackChannelName);

        if (msg) {
            err = alp_attn_ipc_send_with_response (msg, &resp);
            if (err == ALP_STATUS_OK) {
                /* decode response and get alert id */
                err = alp_attn_ipc_unpack_alert_post_reply (resp,
                                                            &status,
                                                            &alertId);
                if (err == ALP_STATUS_OK) {
                    err = status;
                }
                /* done with response, now free it */
                alp_ipc_message_destroy (resp);
            }
            alp_ipc_message_destroy (msg);
        }
        alp_attn_ipc_teardown();
    }
    if (err) {
        ALP_ATLX(API, "returning err %x alert %s,%s,%s id %d",
                 err, source, name, handle, alertId);
    }

    return err;
}

alp_status_t
alp_attn_alert_update (const char  *source,
                       const char  *name,
                       const char  *handle,
                       uint32_t     nprops,
                       AlpAttnProp *properties[])
{
    alp_status_t  err  = ALP_STATUS_OK;
    alp_status_t  status;
    AlpMessage   *msg  = NULL;
    AlpMessage   *resp = NULL;

    if (source == NULL || name == NULL || handle == NULL ||
        nprops == 0 || properties == NULL) {
        ALP_ATLX(API, "bad parameters: %s:%s:%s %d %x",
                 source, name, handle, nprops, (unsigned) properties);
        return ALP_STATUS_ATTN_BAD_PARAM;
    }
    
    err = alp_attn_ipc_setup();
    if (err == ALP_STATUS_OK) {
        msg = alp_attn_ipc_pack_alert_update_msg (source, name, handle,
                                                  nprops, properties);

        if (msg) {
            err = alp_attn_ipc_send_with_response (msg, &resp);
            if (err == ALP_STATUS_OK) {
                /* decode response */
                err = alp_attn_ipc_unpack_alert_update_reply (resp, &status);
                if (err == ALP_STATUS_OK) {
                    err = status;
                }
                /* done with response, now free it */
                alp_ipc_message_destroy (resp);
            }
            alp_ipc_message_destroy (msg);
        }
        alp_attn_ipc_teardown();
    }
    if (err) {
        ALP_ATLX(API, "returning err %x for alert %s,%s,%s",
                 err, source, name, handle);
    }

    return err;
}

alp_status_t
alp_attn_alert_delete (const char  *source,
                       const char  *name,
                       const char  *handle)
{
    alp_status_t  err  = ALP_STATUS_OK;
    alp_status_t  status;
    AlpMessage   *msg  = NULL;
    AlpMessage   *resp = NULL;


    if (source == NULL || name == NULL || handle == NULL) {
        ALP_ATLX(API, "bad parameters: %s,%s,%s", source, name, handle);
        return ALP_STATUS_ATTN_BAD_PARAM;
    }

    err = alp_attn_ipc_setup();
    if (err == ALP_STATUS_OK) {
        msg = alp_attn_ipc_pack_alert_delete_msg (source, name, handle);

        if (msg) {
            err = alp_attn_ipc_send_with_response (msg, &resp);
            if (err == ALP_STATUS_OK) {
                /* decode response and get alert id */
                err = alp_attn_ipc_unpack_alert_delete_reply (resp, &status);
                if (err == ALP_STATUS_OK) {
                    err = status;
                }
                /* done with response, now free it */
                alp_ipc_message_destroy (resp);
            }
            alp_ipc_message_destroy (msg);
        }
        alp_attn_ipc_teardown();
    }
    if (err) {
        ALP_ATLX(API, "returning err %x for alert %s,%s,%s",
                 err, source, name, handle);
    }

    return err;
}
