/******************************************************************************
 *
 * Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the entire contents of this file.
 *
 * The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.
 *
 * Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
 * Rights Reserved.
 * 
 * Contributor(s): none.
 *
 *****************************************************************************/
/**
 * @file attn_api_tests.cpp
 *
 * @brief ALP attention manager API unit tests
 *
 *****************************************************************************/

#include <glib.h>
#include <unistd.h>
#include <signal.h>
#include <paths.h>
#include <errno.h>
#include <wait.h>
#include <string.h>

#include <hiker/types.h>
#include <hiker/syspaths.h>
#include <hiker/traces.h>
#include <hiker/unit_test.h>

#include <hiker/attnmgr.h>
#include <hiker/prv/attnmgr_traces.h>
#include <hiker/prv/attnmgr_dml.h>
#include <hiker/prv/attnmgr_ipc.h>

#include "attn_dml_prv.h"

#ifndef ALP_ATTN_TRACE_SQL
#define ALP_ATTN_TRACE_SQL false   /* set true to trace SQL for debug */
#endif

#ifndef ALP_ATTN_SAVE_DB
#define ALP_ATTN_SAVE_DB false
#endif

#ifndef ALP_ATTN_TEST_USE_PIDFILE
#define ALP_ATTN_TEST_USE_PIDFILE 0
#endif

#define CPPUNIT_ASSERT_FMT( X, fmt, ...) {\
	char cppunit_tmp[256]; \
	sprintf(cppunit_tmp,fmt,__VA_ARGS__); \
	CPPUNIT_ASSERT_MESSAGE( cppunit_tmp, X ); }

#define CPPUNIT_ERR( got, shouldbe ) CPPUNIT_ASSERT_FMT( got == shouldbe, "expected alp_status_t:%0X but got %0X",shouldbe,got)

typedef struct _AlpAttnAlertId AlpAttnAlertId;
struct _AlpAttnAlertId {
    const char *source;
    const char *name;
    const char *handle;
};

class AlpAttnTest: public CPPUNIT_NS::TestFixture
{
    CPPUNIT_TEST_SUITE( AlpAttnTest );
    CPPUNIT_TEST( post_alert_negative );
    CPPUNIT_TEST( update_alert_negative );
    CPPUNIT_TEST( delete_alert_negative );
    CPPUNIT_TEST( post_delete_alert_negative );
    CPPUNIT_TEST( post_alert );
    CPPUNIT_TEST( update_alert );
    CPPUNIT_TEST( delete_alert );
    CPPUNIT_TEST( post_alert_with_callback );
    CPPUNIT_TEST( post_alert_with_callback_negative );
    CPPUNIT_TEST( post_multiple_alert_test );
    CPPUNIT_TEST( update_alert_test2 );
    CPPUNIT_TEST_SUITE_END();

public:
    void setUp();         // overriding TestFixture method
    void tearDown();      // overriding TestFixture method

protected:
    void post_alert_negative();
    void update_alert_negative();
    void delete_alert_negative();
    void post_delete_alert_negative();
    void post_alert();
    void update_alert();
    void delete_alert();
    void post_alert_with_callback();
    void post_alert_with_callback_negative();
    void post_multiple_alert_test();
    void update_alert_test2();
    
    void initialize();
    void start_daemon();
    void stop_daemon(pid_t pid);
    void init_database();

private:
    AlpDmlH   dmlH;
};

ALP_ROOT_TEST_SUITE_REGISTRATION( "UI-007", CPPUNIT_REGISTRY_TYPE_AUTO );
CPPUNIT_TEST_SUITE_NAMED_REGISTRATION( AlpAttnTest, "UI-007" );

// some globals ...
static bool  sInitialized = false;
static pid_t sMyPid     = 0;
static pid_t sDaemonPid = 0;
static char *sDaemonPidFilename;
static char *sDaemonChannel;
static char *sDbFilename;

// local helper functions

static void
prv_sql_trace (void *data, const char *sql)
{
    ALP_ATLX(DML,"<%s>", sql);
}

#define  ALP_ATTNMGRD_CMDNAME  "attnmgrd"

static void
prv_start_daemon_process (void)
{
    pid_t   pid;
	char   *argv[12];
    char  **p;
    int     i;
    const gchar *db_pathname;

	if ((pid = fork()) == 0) { /* child process */
		char *cmd;
	
        db_pathname = alp_attn_dml_get_filename ();
		cmd = ALP_SYSPATH_BIN ALP_ATTNMGRD_CMDNAME;
        i = 0;
		argv[i++] = ALP_ATTNMGRD_CMDNAME;
		argv[i++] = "--no-daemon";
        argv[i++] = "--test";
        argv[i++] = "--debug";
        argv[i++] = "--channel";
        argv[i++] = sDaemonChannel;
        argv[i++] = "--database";
        argv[i++] = (char *) db_pathname;
#if ALP_ATTN_TEST_USE_PIDFILE == 1
        argv[i++] = "--pidfile";
        argv[i++] = sDaemonPidFilename;
#endif
        argv[i++] = NULL;

        p = argv;
        i = 0;
        while (*p) {
            ALP_ATLX(TEST, "    argv[%d] = %s", i, *p);
            i++;
            p++;
        }
		if (execv(cmd, argv) < 0) {
            ALP_ATLX(TEST, "error occured during exec: %d - %s",
                     errno, strerror (errno));
        } else {
            ALP_ATLX(TEST, "this shouldn't happen - exec return with out error");
        }
		exit(0);
	} else {
        usleep (1000000);
	}
    ALP_ATLX(TEST, "daemon started pid %d", pid);
    sDaemonPid = pid;
	return;
}

// class methods

void
AlpAttnTest::initialize(void)
{
    char *filename;

    ALP_ATLX(TEST,"initialized called");
    sMyPid = getpid();
#if ALP_ATTN_TEST_USE_PIDFILE == 1
    sDaemonPidFilename = g_strdup_printf( _PATH_VARRUN "attnmgr-%d", sMyPid);
#endif
    sDaemonChannel = g_strdup_printf( "Attention-Manager-%d", sMyPid);
    filename = g_strdup_printf("AttnMgr-%d.db", sMyPid);
    alp_attn_ipc_init (sDaemonChannel);
    alp_attn_dml_set_filename (ALP_SYSPATH_HOME, filename);
    g_free (filename);
    sDbFilename = g_strdup (alp_attn_dml_get_filename());
    sInitialized = true;
}
void
AlpAttnTest::setUp()
{
    alp_status_t err;

    dmlH = NULL;

    if (sInitialized == false) {
        initialize();
    }

    /* init ipc and daemon, startup daemon */
    start_daemon();
}

void
AlpAttnTest::tearDown()
{
    alp_status_t err;

    /* stop daemon and clean up after it */
    if (sDaemonPid) {
        /* shutdown daemon */
        stop_daemon(sDaemonPid);
        sDaemonPid = 0;

        /* clean up /var/run file */
        if (sDaemonPidFilename) {
            unlink (sDaemonPidFilename);
        }
    }

    /* close database */
    if (dmlH) {
        err = alp_attn_dml_close (dmlH);
        if (err) ALP_ATLX(TEST,"db close failed: %x", err);
    }
    dmlH = 0;
    
    /* remove database after each test */
    {
        FILE *fp;
        int retval;
#if 0        
        ALP_ATLX(TEST, "checking if %s can be opened", dbname);
        if ((fp = fopen (dbname, "r"))) {
            fclose(fp);
            retval = unlink(dbname);
            ALP_ATLX(TEST, "unlinked %s: errno %x %s",
                     dbname, retval < 0 ? errno : 0,
                     retval < 0 ? strerror(errno) : "");
            CPPUNIT_ASSERT (retval == 0);
        }
#else
        retval = unlink(sDbFilename);
        ALP_ATLX(TEST, "unlinking %s retval %d", sDbFilename, retval);
#endif
        g_free (sDbFilename);
        sDbFilename = NULL;
    }
}

void
AlpAttnTest::start_daemon()
{
    FILE  *fp;
    gchar *cmd;
    pid_t  pid;

    prv_start_daemon_process();

#if ALP_ATTN_TEST_USE_PIDFILE == 1
    if ((fp = fopen(sDaemonPidFilename, "r"))) {
        int n;

        n = fscanf(fp, "%u", &pid);
        if (n == 1) {
            ALP_ATLX(TEST, "sDaemon %d found pid %d", sDaemonPid, pid);
            sDaemonPid = pid;
        }
        CPPUNIT_ASSERT (n == 1);
    } else {
        ALP_ATLX(TEST, "sDaemon %d didn't find pid file %s\n",
                 sDaemonPid, sDaemonPidFilename);
        CPPUNIT_ASSERT (FALSE); // unable to open pid file
    }
#endif
}

void
AlpAttnTest::stop_daemon(pid_t pid)
{
    int status;

    if (pid) {
        ALP_ATLX(TEST, "killing daemon at pid %d", pid);
        kill (pid, SIGTERM);
    }

    waitpid (pid, &status, 0);
    ALP_ATLX(TEST, "received status %x for pid %d", status, pid);
}

/**
 * @brief  API tested: alp_attn_alert_post
 *
 * @description  called with invalid parameters
 * 
 */
void AlpAttnTest::post_alert_negative()
{
    alp_status_t    err;

    AlpAttnProp  prop1 =  { "title", "alarm" };
    AlpAttnProp  prop2 =  { "title2", "alarm2" };
    AlpAttnProp  prop3 =  { "title3", "alarm3" };
    AlpAttnProp *properties[] = { &prop1, &prop2, &prop3 };

    /* test bad parameter handling of post */
    err = alp_attn_alert_post (
        NULL, "test", "test", "bar:com.access.apps.attngen.test.xml",
        1, 0, 0, NULL);
    CPPUNIT_ASSERT (err == ALP_STATUS_ATTN_BAD_PARAM);

    err = alp_attn_alert_post (
        "test", NULL,"test", "bar:com.access.apps.attngen.test.xml",
        1, 0, 0, NULL);
    CPPUNIT_ASSERT (err == ALP_STATUS_ATTN_BAD_PARAM);

    err = alp_attn_alert_post (
        "test", "test", NULL, "bar:com.access.apps.attngen.test.xml",
        1, 0, 0, NULL);
    CPPUNIT_ASSERT (err == ALP_STATUS_ATTN_BAD_PARAM);

    err = alp_attn_alert_post (
        "test", "test", "test", "bar:com.access.apps.attngen.test.xml",
        6, 0, 0, NULL);
    CPPUNIT_ASSERT (err == ALP_STATUS_ATTN_BAD_PARAM);

    err = alp_attn_alert_post (
        "test", "test", "test", "bar:com.access.apps.attngen.test.xml",
        1, 0, 1, NULL);
    CPPUNIT_ASSERT (err == ALP_STATUS_ATTN_BAD_PARAM);
    
    err = alp_attn_alert_post (
        "test", "test", "test", "bar:com.access.apps.attngen.test.xml",
         -3, 0, 1, NULL);
    CPPUNIT_ASSERT (err == ALP_STATUS_ATTN_BAD_PARAM);
}

/**
 * @brief  API tested: alp_attn_alert_update
 *
 * @description  called with invalid parameters
 * 
 */
void AlpAttnTest::update_alert_negative()
{

    alp_status_t    err;

    AlpAttnProp  prop0 =  { "test", "test" };
    AlpAttnProp  prop1 =  { "title", "alarm" };
    AlpAttnProp *props[] = { &prop0, &prop1 };


    /* test bad parameter handling of update */
    err = alp_attn_alert_update (NULL, "test", "test", 0, NULL);
    CPPUNIT_ASSERT (err == ALP_STATUS_ATTN_BAD_PARAM);

    err = alp_attn_alert_update ("test", NULL, "test", 0, NULL);
    CPPUNIT_ASSERT (err == ALP_STATUS_ATTN_BAD_PARAM);

    err = alp_attn_alert_update ("test", "test", NULL, 0, NULL);
    CPPUNIT_ASSERT (err == ALP_STATUS_ATTN_BAD_PARAM);

    err = alp_attn_alert_update ("test", "test", "test", 1, NULL);
    CPPUNIT_ASSERT (err == ALP_STATUS_ATTN_BAD_PARAM);

    err = alp_attn_alert_update ("test", "test", "test", 0, props);
    CPPUNIT_ASSERT (err == ALP_STATUS_ATTN_BAD_PARAM);
}

/**
 * @brief  API tested: alp_attn_alert_delete
 *
 * @description  called with invalid parameters
 * 
 */
void AlpAttnTest::delete_alert_negative()
{

    alp_status_t    err;


    /* test bad parameter handling of delete */
    err = alp_attn_alert_delete (NULL, "test", "test");
    CPPUNIT_ASSERT (err == ALP_STATUS_ATTN_BAD_PARAM);

    err = alp_attn_alert_delete ("test", NULL, "test");
    CPPUNIT_ASSERT (err == ALP_STATUS_ATTN_BAD_PARAM);

    err = alp_attn_alert_delete ("test", "test", NULL);
    CPPUNIT_ASSERT (err == ALP_STATUS_ATTN_BAD_PARAM);
}

/**
 * @brief   API tested: alp_attn_alert_delete
 *
 * @description  Attempt to post an alert then
 * 		 delete an alert with similar but
 * 		 incorrect tuple.  
 *
 */
void AlpAttnTest::post_delete_alert_negative()
{
    alp_status_t err;
    AlpAttnProp  prop = { "titleA", "alarmA" };
    AlpAttnProp *props1[] = { &prop };
    uint32_t nprops = sizeof(props1) / sizeof(props1[0]);

    // post an alert
    err = alp_attn_alert_post (
	"test1", "test", "posted_alert", "bar:com.access.apps.attngen.test.xml", 
	1, 0, nprops, props1 );
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    err = alp_attn_alert_delete ("test1", "test", "invalid_param");
    ALP_ATLX(TEST, "delete non-existent record: err = %x", err);
    CPPUNIT_ASSERT(err == ALP_STATUS_ATTN_NOT_FOUND);

    // delete the posted alert
    err = alp_attn_alert_delete ("test1", "test", "posted_alert");
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
}

/**
 * @brief   API Tested: alp_attn_alert_post()
 *
 * @description	 post an alert and verify that it was posted
 *
 */
void
AlpAttnTest::post_alert()
{
    alp_status_t    err = ALP_STATUS_OK;
    int32_t         count;
    AlpAttnId       id;
    AlpAttnAlertId  alert = {
        "bar:com.access.apps.attngen",
        "test",
        "12345"
    };

    /* open database */
    err = alp_attn_dml_open (&dmlH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

#if ALP_ATTN_TRACE_SQL == TRUE
    /* setup sqlite trace callback */
    sqlite3_trace ((sqlite3 *)dmlH, prv_sql_trace, NULL);
#endif

    ALP_ATLX(TEST,"channel set to %s", alp_attn_ipc_get_channel());
    /* verify that db is empty */
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnAlertsTableName ";");
    ALP_ATLX(TEST, "err = %d count = %d", err, count);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 0);

    // post an event, assert success

    ALP_ATLX(TEST, "starting tests in post alert");

    err = alp_attn_alert_post (
        alert.source,
        alert.name,
        alert.handle,
        "bar:com.access.apps.attngen.test.xml",
        1,
        0,
        0,
        NULL);

    if (err) {
        ALP_ATLX(TEST, "insert failed with err %x", err);
    }
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    // verify record is in db
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnAlertsTableName " WHERE "
                                kAttnAlertNameSource " = %Q AND "
                                kAttnAlertNameName   " = %Q AND "
                                kAttnAlertNameHandle " = %Q;",
                                alert.source, alert.name, alert.handle);
    ALP_ATLX(TEST,
             "Verify alert matching %s,%s,%s is there: count = %d err = %x",
             alert.source, alert.name, alert.handle, count, err);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 1);

    // check alert by reading it back
    // verify contents of posted event

    // test that posting a dup alert gets accepted
    err = alp_attn_alert_post (
        alert.source,
        alert.name,
        alert.handle,
        "bar:com.access.apps.attngen.test.xml",
        2,
        0,
        0,
        NULL);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    // verify that the record was changed
    int32_t prio;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &prio, 
                                "SELECT Priority FROM "
                                kAttnAlertsTableName " WHERE "
                                kAttnAlertNameSource " = %Q AND "
                                kAttnAlertNameName   " = %Q AND "
                                kAttnAlertNameHandle " = %Q;",
                                alert.source, alert.name, alert.handle);
    ALP_ATLX(TEST,
             "Verify alert matching %s,%s,%s is has prio 2 == %d err = %x",
             alert.source, alert.name, alert.handle, prio, err);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(prio == 2);

    // verify that only one record with this tuple is in db
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnAlertsTableName " WHERE "
                                kAttnAlertNameSource " = %Q AND "
                                kAttnAlertNameName   " = %Q AND "
                                kAttnAlertNameHandle " = %Q;",
                                alert.source, alert.name, alert.handle);
    ALP_ATLX(TEST,
             "Verify alert matching id %s,%s,%s is there: count = %d err = %x",
             alert.source, alert.name, alert.handle, count, err);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 1);
}

/**
 * @brief  API Tested: alp_attn_alert_update()
 *
 * @description	 post an alert and verify update the
 * 		 alert, verify it was updated
 *
 */
void
AlpAttnTest::update_alert()
{
    alp_status_t    err = ALP_STATUS_OK;
    int32_t         count;
    int32_t         id;
    AlpAttnAlertId  alert = {
        "bar:com.access.apps.attngen",
        "test",
        "12345",
    };

    AlpAttnProp  prop0 = { "icon", "icons/Mail.png" };
    AlpAttnProp  prop1 = { "title", "New message" };
    AlpAttnProp  prop2 = { "image", "some great url that gets me a contact photo" };
    AlpAttnProp *properties[] = { &prop0, &prop1, &prop2 };
    uint32_t nprops = sizeof(properties) / sizeof(properties[0]);

    /* open database */
    err = alp_attn_dml_open (&dmlH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

#if ALP_ATTN_TRACE_SQL == TRUE
    /* setup sqlite trace callback */
    sqlite3_trace ((sqlite3 *)dmlH, prv_sql_trace, NULL);
#endif

    /* verify that db is empty */
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnAlertsTableName ";");
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 0);

    // post an event, assert success

    err = alp_attn_alert_post (
        alert.source, alert.name, alert.handle,
        "bar:com.access.apps.attngen.test.xml",
        1,
        0,
        0,
        NULL);

    if (err) {
        ALP_ATL("insert failed with err %x", err);
    }
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    /* verify record matching return luid is in db */
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnAlertsTableName " WHERE "
                                kAttnAlertNameSource " = %Q AND "
                                kAttnAlertNameName   " = %Q AND "
                                kAttnAlertNameHandle " = %Q;",
                                alert.source, alert.name, alert.handle);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 1);

    /* get alert id to look up associated properties */
    id = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &id, 
                                "SELECT "
                                kAttnAlertNameAlertLuid " FROM "
                                kAttnAlertsTableName " WHERE "
                                kAttnAlertNameSource " = %Q AND "
                                kAttnAlertNameName   " = %Q AND "
                                kAttnAlertNameHandle " = %Q;",
                                alert.source, alert.name, alert.handle);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    CPPUNIT_ASSERT (id > 0);

    /* verify that no property records with id are in the db */
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnPropsTableName " WHERE "
                                kAttnPropNameAlertLuid " = %d;",
                                id);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 0);

    /* update alert */
    err = alp_attn_alert_update (alert.source, alert.name, alert.handle,
                                 nprops, properties);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    
    /* verify that property records with that id are in the db */
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnPropsTableName " WHERE "
                                kAttnAlertNameAlertLuid " = %d;",
                                id);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == (int) nprops);

    /* later test for duplicate props and other corner cases */
}

/**
 * @brief  API Tested: alp_attn_alert_delete();
 *
 * @description	 post an alert and delete it, verify
 * 
 */
void
AlpAttnTest::delete_alert()
{
    alp_status_t err = ALP_STATUS_OK;
    AlpAttnId    id;
    int32_t      count;
    AlpAttnAlertId  alert = {
        "bar:com.access.apps.attngen",
        "test",
        "12345",
    };

    /* open database */
    err = alp_attn_dml_open (&dmlH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

#if ALP_ATTN_TRACE_SQL == TRUE
    /* setup sqlite trace callback */
    sqlite3_trace ((sqlite3 *)dmlH, prv_sql_trace, NULL);
#endif

    /* verify that db is empty */
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnAlertsTableName ";");
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 0);


    // post an event, assert success

    err = alp_attn_alert_post (
        alert.source, alert.name, alert.handle,
        "bar:com.access.apps.attngen.test.xml",
        1,
        0,
        0,
        NULL);
    if (err) {
        ALP_ATL("insert failed with err %x", err);
    }
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    /* verify record matching return luid is in db */
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnAlertsTableName " WHERE "
                                kAttnAlertNameSource " = %Q AND "
                                kAttnAlertNameName   " = %Q AND "
                                kAttnAlertNameHandle " = %Q;",
                                alert.source, alert.name, alert.handle);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 1);

    /* delete alert by alert id */
    err = alp_attn_alert_delete (alert.source, alert.name, alert.handle);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    
    /* verify that no records with that luid are in the db */
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnAlertsTableName " WHERE "
                                kAttnAlertNameSource " = %Q AND "
                                kAttnAlertNameName   " = %Q AND "
                                kAttnAlertNameHandle " = %Q;",
                                alert.source, alert.name, alert.handle);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 0);
}

#define TEST_STRING "ABCxyz TEST TEST 123"

static GMainLoop *sAlpAttnApiTestLoop;
static bool       sAlpAttnApiTestCallbackReceived = false;
static bool       sAlpAttnApiTestCallbackReceivedOkay = false;
static AlpAttnAlertId  sAttnTestAlert = {
        "bar:com.access.tests.attn_api_tests",
        "test",
        "12345"
    };

static void
alp_prv_attn_test_alert_post_callback(const char *sourceId,
                                      const char *alertType,
                                      const char *handle,
                                      int         argc,
                                      char       *argv[],
                                      void       *user_data)
{
    ALP_ATLX(TEST, "Received callback on alert action for %s,%s,%s with %d args",
             sourceId, alertType, handle, argc);
    sAlpAttnApiTestCallbackReceived = true;

    /* validate arguments */
    if ((strcmp(sAttnTestAlert.source, sourceId) == 0) &&
        (strcmp(sAttnTestAlert.name, alertType) == 0) &&
        (strcmp(sAttnTestAlert.handle, handle) == 0)) {
        ALP_ATLX(TEST, "Callback alert id verified");
        if (user_data && strcmp((const char *)user_data, TEST_STRING) == 0) {
            ALP_ATLX(TEST, "Callback alert user data verified");
            sAlpAttnApiTestCallbackReceivedOkay = true;
        }
    }
    if (sAlpAttnApiTestLoop) {
        ALP_ATLX(TEST, "we are in a main loop so let's exit it");
        g_main_loop_quit (sAlpAttnApiTestLoop);
    }
}

gint
alp_prv_attn_test_callback_timeout (gpointer data)
{
    GMainLoop *loop = (GMainLoop *) data;
    ALP_ATLX(TEST, "timeout: stopping main loop");
    g_main_loop_quit (loop);
    return FALSE;
}

/**
 * @brief  API Tested: alp_attn_alert_post_with_callback() 
 *
 * @description	 
 *
 */
void
AlpAttnTest::post_alert_with_callback(void)
{
    alp_status_t    err = ALP_STATUS_OK;
    int32_t         count;
    gboolean        events_dispatched;
    AlpAttnId       id;
    GMainLoop       *loop;
    char *test_string = TEST_STRING;

    loop = g_main_loop_new (NULL, FALSE);
    CPPUNIT_ASSERT (loop != NULL);
    sAlpAttnApiTestLoop = loop;

    /* open database */
    err = alp_attn_dml_open (&dmlH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

#if ALP_ATTN_TRACE_SQL == TRUE
    /* setup sqlite trace callback */
    sqlite3_trace ((sqlite3 *)dmlH, prv_sql_trace, NULL);
#endif

    ALP_ATLX(API,"channel set to %s", alp_attn_ipc_get_channel());
    /* verify that db is empty */
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnAlertsTableName ";");
    ALP_ATLX(TEST, "err = %d count = %d", err, count);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 0);

    // post an event, assert success

    ALP_ATLX(TEST, "starting tests in post alert");

    err = alp_attn_alert_post_with_callback (
        sAttnTestAlert.source, /* source app-id */
        sAttnTestAlert.name,   /* alert type name */
        sAttnTestAlert.handle, /* handle */
        "bar:com.access.apps.attngen.test.xml",  /* interface */
        1,            /* priority */
        0,            /* duration */
        0,            /* nprops   */
        NULL,         /* props    */
        alp_prv_attn_test_alert_post_callback,
        test_string);

    if (err) {
        ALP_ATLX(TEST, "insert failed with err %x", err);
    }
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    g_timeout_add (5000, alp_prv_attn_test_callback_timeout, loop);
    ALP_ATLX(TEST, "entering main loop");
    g_main_loop_run (loop);
    ALP_ATLX(TEST, " exited main loop - continuing");
    CPPUNIT_ASSERT (sAlpAttnApiTestCallbackReceived == true);
    CPPUNIT_ASSERT (sAlpAttnApiTestCallbackReceivedOkay == true);

    // verify record is in db
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnAlertsTableName " WHERE "
                                kAttnAlertNameSource " = %Q AND "
                                kAttnAlertNameName   " = %Q AND "
                                kAttnAlertNameHandle " = %Q;",
                                sAttnTestAlert.source,
                                sAttnTestAlert.name,
                                sAttnTestAlert.handle);
    ALP_ATLX(TEST,
             "Verify alert matching %s,%s,%s is there: count = %d err = %x",
             sAttnTestAlert.source,
             sAttnTestAlert.name,
             sAttnTestAlert.handle,
             count,
             err);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 1);
}

/**
 * @brief  API Tested: alp_attn_alert_post_with_callback()
 *
 * @description	 Called with invalid parameters
 *
 */
void AlpAttnTest::post_alert_with_callback_negative()
{
    alp_status_t   err = ALP_STATUS_OK;
    int32_t         count;
    gboolean        events_dispatched;
    AlpAttnId       id;
    GMainLoop       *loop;
    char *test_string = TEST_STRING;

    err = alp_attn_alert_post_with_callback (
        NULL, /* source app-id */
        sAttnTestAlert.name,   /* alert type name */
        sAttnTestAlert.handle, /* handle */
        "bar:com.access.apps.attngen.test.xml",  /* interface */
        1,            /* priority */
        0,            /* duration */
        0,            /* nprops   */
        NULL,         /* props    */
        alp_prv_attn_test_alert_post_callback,
        test_string);
    CPPUNIT_ASSERT(err == ALP_STATUS_ATTN_BAD_PARAM);

    err = alp_attn_alert_post_with_callback (
        sAttnTestAlert.source, NULL,
        sAttnTestAlert.handle, "bar:com.access.apps.attngen.test.xml",
        1, 0, 0, NULL, alp_prv_attn_test_alert_post_callback,
        test_string);
    CPPUNIT_ASSERT(err == ALP_STATUS_ATTN_BAD_PARAM);
    
    err = alp_attn_alert_post_with_callback (
        sAttnTestAlert.source, sAttnTestAlert.name,
        NULL, "bar:com.access.apps.attngen.test.xml",
        1, 0, 0, NULL, alp_prv_attn_test_alert_post_callback,
        test_string);
    CPPUNIT_ASSERT(err == ALP_STATUS_ATTN_BAD_PARAM);
    
    err = alp_attn_alert_post_with_callback (
        sAttnTestAlert.source, sAttnTestAlert.name,
        sAttnTestAlert.handle, "bar:com.access.apps.attngen.test.xml",
        6, 0, 0, NULL, alp_prv_attn_test_alert_post_callback,
        test_string);
    CPPUNIT_ASSERT(err == ALP_STATUS_ATTN_BAD_PARAM);
    
    err = alp_attn_alert_post_with_callback (
        sAttnTestAlert.source, sAttnTestAlert.name,
        sAttnTestAlert.handle, "bar:com.access.apps.attngen.test.xml",
        1, 0, 1, NULL, alp_prv_attn_test_alert_post_callback,
        test_string);
    CPPUNIT_ASSERT(err == ALP_STATUS_ATTN_BAD_PARAM);
    
    err = alp_attn_alert_post_with_callback (
        sAttnTestAlert.source, sAttnTestAlert.name,
        sAttnTestAlert.handle, "bar:com.access.apps.attngen.test.xml",
        -3, 0, 0, NULL, alp_prv_attn_test_alert_post_callback,
        test_string);
    CPPUNIT_ASSERT(err == ALP_STATUS_ATTN_BAD_PARAM);
    
};

/**
 * @brief  API tested: alp_attn_alert_post
 *
 * @description Posting multiple alerts and verifying that they're in
 *              the database, they don't generate errors and alerts
 *              can be accessed
 * 
 */
void AlpAttnTest::post_multiple_alert_test()
{
    alp_status_t    err = ALP_STATUS_OK;
    int32_t         count;
    int32_t         id;
    AlpAttnAlertId  alert1 =
        { "bar:com.access.apps.test1", "test2a", "abcdef" };
    AlpAttnAlertId  alert2 =
        { "bar:com.access.apps.test2", "test2b", "ghijkl" };

    AlpAttnProp  prop0 = { "title", "New message" };
    AlpAttnProp  prop1 = { "image", "contact photo" };
    AlpAttnProp *properties[] = { &prop0, &prop1 };
    uint32_t nprops = sizeof(properties) / sizeof(properties[0]);
    
    /* open database */
    err = alp_attn_dml_open (&dmlH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

#if ALP_ATTN_TRACE_SQL == TRUE
    /* setup sqlite trace callback */
    sqlite3_trace ((sqlite3 *)dmlH, prv_sql_trace, NULL);
#endif

    /* verify that db is empty */
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count,
		   		 "SELECT COUNT(*) FROM "
                                 kAttnAlertsTableName ";");
    CPPUNIT_ERR(err, ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 0);

    // post a couple of alerts 
    // printf("posting event1:  source = %s, name = %s, handle = %s\n", alert1.source, alert1.name, alert1.handle);

    // adding first alert
        err = alp_attn_alert_post (
        alert1.source,
        alert1.name,
        alert1.handle,
        "bar:com.access.apps.test1.test.xml",
        1,
        5,
        0,
        NULL);

        CPPUNIT_ERR (err, ALP_STATUS_OK);

    // verify record is in db
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnAlertsTableName " WHERE "
                                kAttnAlertNameSource " = %Q AND "
                                kAttnAlertNameName   " = %Q AND "
                                kAttnAlertNameHandle " = %Q;",
                                alert1.source, alert1.name, alert1.handle);

    // printf("Verify alert matching %s,%s,%s is there: count = %d err = %x\n",
    //         alert1.source, alert1.name, alert1.handle, count, err);
    CPPUNIT_ERR(err, ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 1);

    // printf("posting event2:  source = %s, name = %s, handle = %s\n", alert2.source, alert2.name, alert2.handle);
    // adding second alert
        err = alp_attn_alert_post (
        alert2.source,
        alert2.name,
        alert2.handle,
        "bar:com.access.apps.test2.test.xml",
        1,
        3,
        0,
        NULL);

        CPPUNIT_ERR (err, ALP_STATUS_OK);

    // verify 2nd record is in db
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnAlertsTableName " WHERE "
                                kAttnAlertNameSource " = %Q AND "
                                kAttnAlertNameName   " = %Q AND "
                                kAttnAlertNameHandle " = %Q;",
                                alert2.source, alert2.name, alert2.handle);

    //printf("Verify alert matching %s,%s,%s is there: count = %d err = %x\n",
    //         alert2.source, alert2.name, alert2.handle, count, err);

    CPPUNIT_ERR(err, ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 1);

    // check alert1 by reading it back and adding properties
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnAlertsTableName " WHERE "
                                kAttnAlertNameSource " = %Q AND "
                                kAttnAlertNameName   " = %Q AND "
                                kAttnAlertNameHandle " = %Q;",
                                alert1.source, alert1.name, alert1.handle);
    CPPUNIT_ERR(err, ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 1);

    // get alert id to look up associated properties
    id = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &id, 
                                "SELECT "
                                kAttnAlertNameAlertLuid " FROM "
                                kAttnAlertsTableName " WHERE "
                                kAttnAlertNameSource " = %Q AND "
                                kAttnAlertNameName   " = %Q AND "
                                kAttnAlertNameHandle " = %Q;",
                                alert1.source, alert1.name, alert1.handle);
    CPPUNIT_ERR(err, ALP_STATUS_OK);
    CPPUNIT_ASSERT (id > 0);

    // verify that no property records with this id are in the db
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnPropsTableName " WHERE "
                                kAttnPropNameAlertLuid " = %d;",
                                id);
    CPPUNIT_ERR(err, ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 0);
    // printf("Verifing no property records exist with %s,%s,%s[%d] is there: count = %d err = %x\n",
    //         alert1.source, alert1.name, alert1.handle, id, count, err);

    // update alert
    err = alp_attn_alert_update (alert1.source, alert1.name, alert1.handle, nprops, properties);
    CPPUNIT_ERR(err, ALP_STATUS_OK);
    
    // verify that property records with this id are in the db
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnPropsTableName " WHERE "
                                kAttnAlertNameAlertLuid " = %d;",
                                id);
    CPPUNIT_ERR(err, ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 2);
    // printf("Verifing property records with id %s,%s,%s is there: count = %d err = %x\n",
    //         alert1.source, alert1.name, alert1.handle, count, err);

    // verify contents of posted events
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count,
                                "SELECT COUNT(*) FROM "
                                kAttnAlertsTableName " WHERE "
                                kAttnAlertNameSource " = %Q AND "
                                kAttnAlertNameName   " = %Q AND "
                                kAttnAlertNameHandle " = %Q;",
                                alert1.source, alert1.name, alert1.handle);
    CPPUNIT_ERR(err, ALP_STATUS_OK);
    // printf("Verify alert matching id %s,%s,%s is there: count = %d err = %x\n",
    //         alert1.source, alert1.name, alert1.handle, count, err);
    
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count,
                                "SELECT COUNT(*) FROM "
                                kAttnAlertsTableName " WHERE "
                                kAttnAlertNameSource " = %Q AND "
                                kAttnAlertNameName   " = %Q AND "
                                kAttnAlertNameHandle " = %Q;",
                                alert2.source, alert2.name, alert2.handle);
    CPPUNIT_ERR(err, ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 1);
    // printf("Verify alert matching id %s,%s,%s is there: count = %d err = %x\n",
    //         alert2.source, alert2.name, alert2.handle, count, err);
    
}


/**
 * @brief  API tested: alp_attn_alert_update
 *
 * @description Posting multiple alerts and updating alert2, then verifying
 *              alert2 is updated, while the rest remain unmodified.
 * 
 */
void AlpAttnTest::update_alert_test2()
{
    alp_status_t    err = ALP_STATUS_OK;
    int32_t         count;
    int32_t         id, id1, id3;
    AlpAttnAlertId  alert1 = { "bar:com.access.apps.attn1gen", "test1", "12345", };
    AlpAttnAlertId  alert2 = { "bar:com.access.apps.attn2gen", "test2", "678", };
    AlpAttnAlertId  alert3 = { "bar:com.access.apps.attn3gen", "test3", "9", };

    AlpAttnProp  prop0 = { "icon", "icons/Mail.png" };
    AlpAttnProp  prop1 = { "title", "New message" };
    AlpAttnProp  prop2 = { "image", "test message" };
    AlpAttnProp *properties[] = { &prop0, &prop1, &prop2 };
    uint32_t nprops = sizeof(properties) / sizeof(properties[0]);

    // open database
    err = alp_attn_dml_open (&dmlH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

#if ALP_ATTN_TRACE_SQL == TRUE
    /* setup sqlite trace callback */
    sqlite3_trace ((sqlite3 *)dmlH, prv_sql_trace, NULL);
#endif

    // verify that db is empty
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnAlertsTableName ";");
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 0);

    // post a few events
    err = alp_attn_alert_post (alert1.source, alert1.name, alert1.handle,"bar:com.access.apps.attn1gen.test.xml",1,0,0,NULL);
    CPPUNIT_ERR (err, ALP_STATUS_OK);

    err = alp_attn_alert_post (alert2.source, alert2.name, alert2.handle,"bar:com.access.apps.attn2gen.test.xml",1,0,0,NULL);
    CPPUNIT_ERR (err, ALP_STATUS_OK);

    err = alp_attn_alert_post (alert3.source, alert3.name, alert3.handle,"bar:com.access.apps.attn3gen.test.xml",1,0,0,NULL);
    CPPUNIT_ERR (err, ALP_STATUS_OK);

    // verify record matching return luid is in db
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnAlertsTableName " WHERE "
                                kAttnAlertNameSource " = %Q AND "
                                kAttnAlertNameName   " = %Q AND "
                                kAttnAlertNameHandle " = %Q;",
                                alert2.source, alert2.name, alert2.handle);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 1);

    // get alert id to look up associated properties
    id = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &id, 
                                "SELECT "
                                kAttnAlertNameAlertLuid " FROM "
                                kAttnAlertsTableName " WHERE "
                                kAttnAlertNameSource " = %Q AND "
                                kAttnAlertNameName   " = %Q AND "
                                kAttnAlertNameHandle " = %Q;",
                                alert2.source, alert2.name, alert2.handle);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    CPPUNIT_ASSERT (id > 0);

    // verify that no property records with id are in the db
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnPropsTableName " WHERE "
                                kAttnPropNameAlertLuid " = %d;",
                                id);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 0);

    // update alert
    err = alp_attn_alert_update (alert2.source, alert2.name, alert2.handle, nprops, properties);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    
    // verify that property records with that id are in the db
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnPropsTableName " WHERE "
                                kAttnAlertNameAlertLuid " = %d;",
                                id);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == (int) nprops);

    // verify that no property records are associated with alert1
    id1 = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &id1, 
                                "SELECT "
                                kAttnAlertNameAlertLuid " FROM "
                                kAttnAlertsTableName " WHERE "
                                kAttnAlertNameSource " = %Q AND "
                                kAttnAlertNameName   " = %Q AND "
                                kAttnAlertNameHandle " = %Q;",
                                alert1.source, alert1.name, alert1.handle);
    CPPUNIT_ERR(err, ALP_STATUS_OK);
    CPPUNIT_ASSERT(id > 0);

    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnPropsTableName " WHERE "
                                kAttnAlertNameAlertLuid " = %d;",
                                id1);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 0);

    // verify that no property records are associated with alert3
    id3 = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &id3, 
                                "SELECT "
                                kAttnAlertNameAlertLuid " FROM "
                                kAttnAlertsTableName " WHERE "
                                kAttnAlertNameSource " = %Q AND "
                                kAttnAlertNameName   " = %Q AND "
                                kAttnAlertNameHandle " = %Q;",
                                alert3.source, alert3.name, alert3.handle);
    CPPUNIT_ERR(err, ALP_STATUS_OK);
    CPPUNIT_ASSERT(id > 0);

    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnPropsTableName " WHERE "
                                kAttnAlertNameAlertLuid " = %d;",
                                id3);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 0);
}
