/********************************************************************

Copyright  2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/**
 * @file attn_dml.c
 *
 * @brief ALP attention manager data model implementation
 *
 ******************************************************************************/

#include <hiker/types.h>
#include <hiker/prv/attnmgr_traces.h>
#include <hiker/prv/attnmgr_dml.h>

#include "attn_dml_prv.h"

#include <glib.h>

/********************************************************************************/
static const gchar *alp_attn_db_path = NULL;

alp_status_t
alp_attn_dml_set_filename (const char *directory, const char *database)
{
    /* Now form path to database */
    ALP_ATLX(DML, "called with dir [%s] file [%s] (current %s)",
             directory ? directory : "<NULL>",
             database ? database : "<NULL>",
             alp_attn_db_path ? alp_attn_db_path : "<NULL>");

    if (directory == NULL) {
        directory = kAttnDefDBDirectory;
        ALP_ATLX(DML,"directory has been defaulted to %s", directory);
    }

    if (database == NULL) {
        database = kAttnDefDBFileName;
        ALP_ATLX(DML,"database has been defaulted to %s",
                 database);
    }
    ALP_ATLX(DML, "after checking: dir [%s] file [%s]", directory, database);

    /* verify directory exists */
    if (!g_file_test (directory, (G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR))) {
        ALP_ATLX(DML, "Directory %s does not exist or is not a directory",
                 directory);
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    /* verify database name doesn't include directories */
    if (g_strrstr(database, "/")) {
        ALP_ATLX(DML, "Database filename %s is not a simple filename",
                 database);
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    if (alp_attn_db_path) {
        ALP_ATLX(DML,
                 "attention dml db path has already been initialized to %s",
                 alp_attn_db_path);
        return ALP_STATUS_ATTN_DML_EXISTS;   /* already set, ignore request  */
    }

    /* Now form path to database */
    alp_attn_db_path = g_build_path (G_DIR_SEPARATOR_S,
                                     directory,
                                     database,
                                     NULL);
    ALP_ATLX(DML, "db path set to %s", alp_attn_db_path);

    return ALP_STATUS_OK;
}

const char *
alp_attn_dml_get_filename(void)
{
    return alp_attn_db_path;
}

/*
 * Only open the database. The attention manager daemon will be responsible
 * for creating and initializing the database.
 */
alp_status_t
alp_attn_dml_open (AlpDmlH *dmlH)
{
    alp_status_t err;
    FILE *fp;
    uint32_t length;

    /* Validate parameter */
    if (!dmlH) {
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    /*
     * If the DB pathname hasn't been initialized, initialize it
     * here to the default directory and default db name.
     */
    if (alp_attn_db_path == NULL) {
        err = alp_attn_dml_set_filename (NULL, NULL);
        if (err) {
            return err;
        }
    }

    /* Check if database exists and has non-zero length */
    if ((fp = fopen (alp_attn_db_path, "r")) == NULL) {
        length = 0;
    } else {
	    /* measure length of file in bytes */
	    fseek(fp, 0, SEEK_END);
	    length = ftell(fp);
	    fclose(fp);
    }

    if (length == 0) {
        return ALP_STATUS_ATTN_DML_INTERNAL_ERROR;
    }

    /* Call the database open call in the SAL Library. */
    ALP_ATLX(DML, "Opening database %s", alp_attn_db_path);
    err = alp_sal_database_open((AlpSALHandle)dmlH, alp_attn_db_path);
    if (err) {
            ALP_ATLX(DML,
                     "Opening database %s error %d",
                     alp_attn_db_path, err);
            return alp_prv_attn_dml_convert_sal_error (err);
    }

    return ALP_STATUS_OK;
}

alp_status_t
alp_attn_dml_close(AlpDmlH dmlH)
{
	const char  *errMsg = NULL;
    alp_status_t err;
    const gchar *filename;

    /* disassociate database filename */
    filename = alp_attn_db_path;
    alp_attn_db_path = NULL;
    g_free ((gchar *) filename);

    /* Call the database close call in the SAL Library. */
    err = alp_sal_database_close((AlpSALHandle)dmlH);
    if (err) {
        errMsg = alp_sal_get_error_message ((AlpSALHandle)dmlH);
        ALP_ATLX(DML, "failed with err %x: %s", err, errMsg);
        return alp_prv_attn_dml_convert_sal_error (err);
    }

    return err; /* ALP_STATUS_OK */
}


alp_status_t
alp_attn_dml_transaction_begin (AlpDmlH dmlH, bool exclusive)
{

    alp_status_t err;

    /* Call the transaction begin call in the SAL Library. */
    err = alp_sal_transaction_begin((AlpSALHandle)dmlH, exclusive);
    if (err) {
        ALP_ATLX(DML,"begin failed %s err %x",
                 alp_sal_get_error_message ((AlpSALHandle) dmlH), err);
        return alp_prv_attn_dml_convert_sal_error (err);
    }

    return err; /* ALP_STATUS_OK */
}

alp_status_t
alp_attn_dml_transaction_rollback (AlpDmlH dmlH)
{

    alp_status_t err;

    /* Call the database open call in the SAL Library. */
    err = alp_sal_transaction_rollback ((AlpSALHandle) dmlH);
    if (err) {
        ALP_ATLX(DML,"rollback failed %s err %x",
                 alp_sal_get_error_message ((AlpSALHandle) dmlH), err);
        return alp_prv_attn_dml_convert_sal_error (err);
    }

    return err; /* ALP_STATUS_OK */
}

alp_status_t
alp_attn_dml_transaction_end (AlpDmlH dmlH)
{

    alp_status_t err;

    /* Call the database open call in the SAL Library. */
    err = alp_sal_transaction_commit ((AlpSALHandle) dmlH);
    if (err) {
        ALP_ATLX(DML,"end failed %s err %x",
                 alp_sal_get_error_message ((AlpSALHandle) dmlH), err);
        return alp_prv_attn_dml_convert_sal_error (err);
    }

    return err; /* ALP_STATUS_OK */
}

alp_status_t
alp_attn_dml_check_database (AlpDmlH dmlH)
{
    alp_status_t err;
    AlpSALStatement         *stmt = NULL;
    char                    *result;
    char                    *sql;

        
    /* perform integrity check */
    sql = "PRAGMA integrity_check;";
    err = alp_sal_statement_initialize ((AlpSALHandle)dmlH, &stmt, sql);
    if (err) {
        const char *errMsg = alp_sal_get_error_message ((AlpSALHandle)dmlH);
        ALP_ATLX(DML, "statement_init: sql %s : %x - %s", sql, err, errMsg);
        return alp_prv_attn_dml_convert_sal_error (err);
    }

    /* Check for the result */
    err = alp_sal_statement_next_result (stmt);
    if (err) {
        const char *errMsg = alp_sal_get_error_message ((AlpSALHandle)dmlH);
        ALP_ATLX(DML, "next statement: %x - %s", err, errMsg);
        alp_sal_statement_destroy (stmt);
        return alp_prv_attn_dml_convert_sal_error (err);
    }
    result = alp_sal_statement_get_text (stmt, 0, false);
    if (result == NULL || strcmp(result, "ok") != 0) {
            ALP_ATLX (DML,
                      "integrity check failed: reason [%s]",
                      result ? result : "<NULL>");
            alp_sal_statement_destroy (stmt);
            return ALP_STATUS_ATTN_DML_CORRUPT;
    }
    alp_sal_statement_destroy (stmt);

    /* check database version */
    /* TODO */

    return ALP_STATUS_OK;
}
