/********************************************************************

Copyright  2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/**
 * @file attn_dml_item.c
 *
 * @brief ALP attention manager data model implementation
 *
 *****************************************************************************/
#include <unistd.h>

#include <glib-object.h>

#include <hiker/types.h>
#include <hiker/prv/attnmgr_traces.h>
#include <hiker/prv/attnmgr_dml.h>

#include "attn_dml_prv.h"

alp_status_t
alp_attn_dml_item_create (AlpAttnDmlItemType itemType,
                          AlpDmlItemH *itemH)
{
    AlpAttnAlertType     *alertP    = NULL;
    AlpAttnPropType      *propP     = NULL;
    AlpAttnPropListType  *propListP = NULL;

    ALP_ATLX(DML, "create %d item", itemType);
    /* Validate parameter */
    if (!itemH) {
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    /* Create appropriate item */
    switch (itemType) {
    case AlertItem:
        /* Allocate structure */
        alertP = (AlpAttnAlertType *) malloc(sizeof(AlpAttnAlertType));
        if (!alertP) {
            return ALP_STATUS_ATTN_DML_NO_MEMORY;
        }
        
        /* Clear structure */
        memset(alertP, 0, sizeof(AlpAttnAlertType));

        alertP->itemType = AlertItem;
        *itemH = (AlpDmlItemH) alertP;
        ALP_ATLX(DML,"create alert item at %x size %d",
                 (unsigned) alertP, sizeof(AlpAttnAlertType));
        break;

    case PropertyItem:
        /* Allocate structure */
        propP = (AlpAttnPropType *)malloc(sizeof(AlpAttnPropType));
        if (!propP) {
            return ALP_STATUS_ATTN_DML_NO_MEMORY;
        }
        
        /* Clear structure */
        memset(propP, 0, sizeof(AlpAttnPropType));

        propP->itemType = PropertyItem;
        *itemH = (AlpDmlItemH) propP;
        ALP_ATLX(DML,"create prop item at %x size %d",
                 (unsigned) propP, sizeof(AlpAttnPropType));
        break;

    case PropertyListItem:
        /* Allocate structure */
        propListP = (AlpAttnPropListType *)malloc(sizeof(AlpAttnPropListType));
        if (!propListP) {
            return ALP_STATUS_ATTN_DML_NO_MEMORY;
        }
        
        /* Clear structure */
        memset(propListP, 0, sizeof(AlpAttnPropListType));

        propListP->itemType = PropertyListItem;
        *itemH = (AlpDmlItemH) propListP;
        ALP_ATLX(DML,"create prop list item at %x size %d",
                 (unsigned) propListP, sizeof(AlpAttnPropListType));
        break;

    default:
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    return ALP_STATUS_OK;
}

static void
prv_attn_dml_property_free (AlpAttnPropType *property)
{
    if (property->name) {
        free (property->name);
    }
    if (property->value) {
        free (property->value);
    }
    free (property);
}

alp_status_t
alp_attn_dml_item_destroy (AlpDmlItemH itemH)
{
    AlpAttnAlertType    *alertP = (AlpAttnAlertType *) itemH;
    AlpAttnPropType     *propP;
    AlpAttnPropListType *propListP;
    AlpAttnDmlItemType   itemType;
    AlpAttnPropType     *property;

    /* Validate parameter */
    if (!itemH) {
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }
    itemType = alertP->itemType;

    /* Create appropriate item */
    switch (itemType) {
    case AlertItem:
        if (alertP->source) {
            free(alertP->source);
        }
        if (alertP->name) {
            free(alertP->name);
        }
        if (alertP->handle) {
            free(alertP->handle);
        }

        /*
         * Handle the associated property list
         */
        property = alertP->plist;

        /* detach property list at this point */
        alertP->plist = NULL;

        /* free property list if attached */
        if (property) {
            alp_attn_dml_item_destroy (property);
        }
        free(itemH);

        break;

    case PropertyItem:
        propP = (AlpAttnPropType *) itemH;
        property = propP;
        while (property) {
            AlpAttnPropType *next = property->next;
            ALP_ATLX(DML,"delete prop from prop: prop %x next %x",
                     (unsigned) property, (unsigned) next);
            ALP_FAIL_FATAL_IF (property->itemType != PropertyItem,
                               "list item is not a property item: type = %d",
                               property->itemType);
            prv_attn_dml_property_free (property);
            property = next;    /* advance to next property if present */
        }
        break;

    case PropertyListItem:
        propListP = (AlpAttnPropListType *) itemH;
        if (propListP->source) {
            free(propListP->source);
        }
        if (propListP->name) {
            free(propListP->name);
        }
        if (propListP->handle) {
            free(propListP->handle);
        }
        property = propListP->list;
        while (property) {
            AlpAttnPropType *next = property->next;
            prv_attn_dml_property_free (property);
            property = next;
        }
        free (propListP);
        break;

    default:
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }


    return ALP_STATUS_OK;
}

alp_status_t
alp_attn_dml_item_set_int32 (AlpDmlItemH itemH,
                             AlpAttnDmlColumnType column,
                             int32_t *value)
{
    AlpAttnAlertType   *alertP = (AlpAttnAlertType *) itemH;
    AlpAttnPropType    *propP  = (AlpAttnPropType  *) itemH;
    AlpAttnDmlItemType  itemType;

    /* Validate parameter */
    if (!itemH) {
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }
    itemType = alertP->itemType;

    switch (itemType) {
    case AlertItem:
        switch (column) {
        case Priority:
            if (value) {
                alertP->priority = *value;
            } else {
                alertP->priority = kAttnDefPriority;
            }
            break;

        case Duration:
            if (value) {
                alertP->duration = *value;
            } else {
                alertP->duration = kAttnDefDuration;
            }
            break;
            
        case UserId:
            if (value) {
                alertP->userId = *value;
            } else {
                alertP->userId = kAttnDefUserId;
            }
            break;
            
        case GroupId:
            if (value) {
                alertP->groupId = *value;
            } else {
                alertP->groupId = kAttnDefGroupId;
            }
            break;
            
        case Pending:
            if (value) {
                alertP->pending = *value;
            } else {
                alertP->pending = kAttnDefPending;
            }
            break;

        case Created:
            if (value) {
                alertP->created = *value;
            } else {
                alertP->created = kAttnDefTimestamp;
            }
            break;

        case Updated:
            if (value) {
                alertP->updated = *value;
            } else {
                alertP->updated = kAttnDefTimestamp;
            }
            break;

        default:
            return ALP_STATUS_ATTN_DML_BAD_PARAM;
        }
        break;

    case PropertyItem:
        switch (column) {
        default:
            return ALP_STATUS_ATTN_DML_BAD_PARAM;
        }
        break;

    default:
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    return ALP_STATUS_OK;
}

alp_status_t
alp_attn_dml_item_set_uint32 (AlpDmlItemH itemH,
                              AlpAttnDmlColumnType column,
                              uint32_t *value)
{
    AlpAttnAlertType    *alertP     = (AlpAttnAlertType    *) itemH;
    AlpAttnPropType     *propP      = (AlpAttnPropType     *) itemH;
    AlpAttnPropListType *propListP  = (AlpAttnPropListType *) itemH;
    AlpAttnDmlItemType  itemType;

    /* Validate parameter */
    if (!itemH) {
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }
    itemType = alertP->itemType;

    switch (itemType) {
    case AlertItem:
        switch (column) {
        case AlertLuid:
            if (value) {
                alertP->itemLuid = *value;
            } else {
                alertP->itemLuid = kAttnDefAlertLuid;
            }
            break;

        case Attributes:
            if (value) {
                alertP->attributes = *value;
            } else {
                alertP->attributes = kAttnDefAttributes;
            }
            break;

        default:
            return ALP_STATUS_ATTN_DML_BAD_PARAM;
        }
        break;

    case PropertyItem:
        switch (column) {
        case PropLuid:
            if (value) {
                propP->itemLuid = *value;
            } else {
                propP->itemLuid = kAttnDefAlertLuid;
            }
            break;

        case AlertLuid:
            if (value) {
                propP->alertLuid = *value;
            } else {
                propP->alertLuid = kAttnDefAlertLuid;
            }
            break;

        case PropType:
            if (value) {
                propP->type = *value;
            } else {
                propP->type = kAttnDefPropType;
            }
            break;

        default:
            return ALP_STATUS_ATTN_DML_BAD_PARAM;
        }
        break;

    case PropertyListItem:
        switch (column) {
        case AlertLuid:
            if (value) {
                propListP->itemLuid = *value;
            } else {
                propListP->itemLuid = kAttnDefAlertLuid;
            }
            break;

        default:
            return ALP_STATUS_ATTN_DML_BAD_PARAM;
        }
        break;

    default:
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    return ALP_STATUS_OK;
}

alp_status_t
alp_attn_dml_item_set_bool (AlpDmlItemH itemH,
                            AlpAttnDmlColumnType column,
                            bool *value)
{
    AlpAttnAlertType   *alertP = (AlpAttnAlertType *) itemH;
    AlpAttnDmlItemType  itemType;

    /* Validate parameter */
    if (!itemH) {
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }
    itemType = alertP->itemType;

    switch (itemType) {
    case AlertItem:
        switch (column) {
        case Pending:
            if (value) {
                alertP->pending = *value;
            } else {
                alertP->pending = kAttnDefPending;
            }
            break;

        default:
            return ALP_STATUS_ATTN_DML_BAD_PARAM;
        }
        break;

    default:
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    return ALP_STATUS_OK;
}

alp_status_t
alp_attn_dml_item_set_string (AlpDmlItemH itemH,
                              AlpAttnDmlColumnType column,
                              const char *value)
{
    AlpAttnAlertType   *alertP = (AlpAttnAlertType *) itemH;
    AlpAttnPropType    *propP;
    AlpAttnPropListType *propListP;
    AlpAttnDmlItemType  itemType;

    /* Validate parameter */
    if (!itemH) {
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }
    itemType = alertP->itemType;

    switch (itemType) {
    case AlertItem:
        switch (column) {
        case Source:
            if (alertP->source) {
                free (alertP->source);
            }
            if (value) {
                alertP->source = strdup(value);
            } else {
                alertP->source = NULL;
            }
            break;

        case Name:
            if (alertP->name) {
                free (alertP->name);
            }
            if (value) {
                alertP->name = strdup(value);
            } else {
                alertP->name = NULL;
            }
            break;

        case Handle:
            if (alertP->handle) {
                free (alertP->handle);
            }
            if (value) {
                alertP->handle = strdup(value);
            } else {
                alertP->handle = NULL;
            }
            break;

        default: 
            return ALP_STATUS_ATTN_DML_BAD_PARAM;
        }
        break;

    case PropertyItem:
        propP = (AlpAttnPropType *) itemH;
        switch (column) {
        case Name:
            if (propP->name) {
                free (propP->name);
            }
            if (value) {
                propP->name = strdup(value);
            } else {
                propP->name = NULL;
            }
            break;

        case Value:
            /* convenience function for setting string blobs */
            if (propP->value) {
                free (propP->value);
            }
            propP->size = 0;
            propP->type = G_TYPE_NONE;
            propP->value = NULL;
            /* if new value to set */
            if (value) {
                uint32_t size;

                size = strlen (value) + 1;
                propP->value = malloc (size);

                if (propP->value) {
                    memcpy (propP->value, value, size);
                    propP->size = size;
                    propP->type = G_TYPE_STRING;
                    ALP_ATLX(DML,"%d: string to blob %s, size %d type %x",
                             __LINE__, propP->value, propP->size, propP->type);
                }
            }
            break;

        default:
            return ALP_STATUS_ATTN_DML_BAD_PARAM;
        }
        break;
       
    case PropertyListItem:
        propListP = (AlpAttnPropListType *) itemH;
        switch (column) {
        case Source:
            if (propListP->source) {
                free (propListP->source);
            }
            if (value) {
                propListP->source = strdup(value);
            } else {
                propListP->source = NULL;
            }
            break;

        case Name:
            if (propListP->name) {
                free (propListP->name);
            }
            if (value) {
                propListP->name = strdup(value);
            } else {
                propListP->name = NULL;
            }
            break;

        case Handle:
            if (propListP->handle) {
                free (propListP->handle);
            }
            if (value) {
                propListP->handle = strdup(value);
            } else {
                propListP->handle = NULL;
            }
            break;

        default:
            return ALP_STATUS_ATTN_DML_BAD_PARAM;
        }
        break;

    default:
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    return ALP_STATUS_OK;
}

alp_status_t
alp_attn_dml_item_set_blob (AlpDmlItemH itemH,
                            AlpAttnDmlColumnType column,
                            void *data,
                            uint32_t size)
{
    AlpAttnAlertType   *alertP = (AlpAttnAlertType *) itemH;
    AlpAttnPropType    *propP;
    AlpAttnDmlItemType  itemType;

    /* Validate parameter */
    if (!itemH) {
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }
    itemType = alertP->itemType;

    switch (itemType) {
    case PropertyItem:
        propP = (AlpAttnPropType *) itemH;
        switch (column) {
        case Value:
            if (propP->value) {
                free (propP->value);
            }
            propP->size  = 0;
            propP->value = NULL;
            if (data && size > 0) {
                propP->value = malloc (size);
                if (propP->value) {
                    memcpy (propP->value, data, size);
                    propP->size = size;
                }
            }
            break;

        default:
            return ALP_STATUS_ATTN_DML_BAD_PARAM;
        }
        break;
    default:
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    return ALP_STATUS_OK;
}

/**
 * @brief set the property list of an alert item
 *
 * @param[in] itemH an alert item
 * @param[in] propListH a property list item containing a linked list of
 */
alp_status_t
alp_attn_dml_item_set_property_list (AlpDmlItemH itemH,
                                     AlpDmlItemH propListH)
{
    AlpAttnAlertType    *alertP       = (AlpAttnAlertType    *) itemH;
    AlpAttnPropListType *propListP    = (AlpAttnPropListType *) propListH;

    switch (alertP->itemType) {
    case AlertItem:
        if (propListP->itemType != PropertyListItem)
            return ALP_STATUS_ATTN_DML_BAD_PARAM; /* not a property list */
        if (alertP->plist) {
            /* if there already exists a property list, destroy it */
            alp_attn_dml_item_destroy (alertP->plist);
        }
        /*
         * Detach the linked list of properties from the property list
         * and assign them to the alert item's property list.
         */
        alertP->pcount = propListP->count;
        alertP->plist  = propListP->list;
        propListP->list = NULL;
        break;

    default:
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }
    return ALP_STATUS_OK;
}

alp_status_t
alp_attn_dml_item_add_property (AlpDmlItemH itemH,
                                AlpDmlItemH propH)
{
    AlpAttnAlertType    *alertP       = (AlpAttnAlertType    *) itemH;
    AlpAttnPropListType *propListP    = (AlpAttnPropListType *) itemH;
    AlpAttnPropType     *propP        = (AlpAttnPropType     *) propH;
    AlpAttnPropType     *property;

    if (itemH == NULL || propH == NULL || propP->itemType != PropertyItem) {
        ALP_ATLX(DML, "bad parameters: itemH %x propH %x", itemH, propH);
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }
        
    switch (alertP->itemType) {
    case AlertItem:
        /*
         * Insert one or more linked properties to the top of the
         * alert item's property list.
         */
        while (propP) {
            property = propP->next;
            propP->next = alertP->plist;
            alertP->plist = propP;
            alertP->pcount++;
            propP = property;
        }
        break;

    case PropertyListItem:
        propP->next = propListP->list;
        propListP->list = propP;
        propListP->count++;
        break;

    default:
        ALP_ATLX(DML, "bad parameters: unknown item type %d",
                 alertP->itemType);
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }
    return ALP_STATUS_OK;
}

const char *
alp_attn_dml_item_get_string (AlpDmlItemH itemH, AlpAttnDmlColumnType column)
{
    AlpAttnBaseType   *baseP = (AlpAttnBaseType *) itemH;
    AlpAttnAlertType  *alertP;
    AlpAttnPropType   *propP;

    if (!itemH) {
        return NULL;
    }
    switch (baseP->itemType) {
    case AlertItem:
        alertP = (AlpAttnAlertType *) itemH;
        switch (column) {
        case Source:    return alertP->source;
        case Name:      return alertP->name;
        case Handle:    return alertP->handle;
        default:
            break;
        }
    case PropertyItem:
        propP = (AlpAttnPropType *) itemH;
        switch (column) {
        case Name:      return propP->name;
        default:
            break;
        }
    default:
        break;
    }
    return NULL;
}

int32_t
alp_attn_dml_item_get_int32 (AlpDmlItemH itemH, AlpAttnDmlColumnType column)
{
    AlpAttnBaseType   *baseP = (AlpAttnBaseType *) itemH;
    AlpAttnAlertType  *alertP;

    if (!itemH) {
        return 0;
    }

    switch (baseP->itemType) {
    case AlertItem:
        alertP = (AlpAttnAlertType *) itemH;
        switch (column) {
        case Priority:    return alertP->priority;
        case Duration:    return alertP->duration;
        case Attributes:  return alertP->attributes;
        default:
            break;
        }
    default:
        break;
    }
    return 0;
}

uint32_t
alp_attn_dml_item_get_uint32 (AlpDmlItemH itemH, AlpAttnDmlColumnType column)
{
    AlpAttnBaseType   *baseP = (AlpAttnBaseType *) itemH;
    AlpAttnAlertType  *alertP;
    AlpAttnPropType   *propP;

    if (!itemH) {
        return 0;
    }

    switch (baseP->itemType) {
    case AlertItem:
        alertP = (AlpAttnAlertType *) itemH;
        switch (column) {
        case AlertLuid:   return alertP->itemLuid;
        case Attributes:  return alertP->attributes;
        default:
            break;
        }
    case PropertyItem:
        propP = (AlpAttnPropType *) itemH;
        switch (column) {
        case PropLuid:    return propP->itemLuid;
        case AlertLuid:   return propP->alertLuid;
        case PropType:    return propP->type;
        default:
            break;
        }

    default:
        break;
    }
    return 0;
}
