/********************************************************************

Copyright  2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/**
 * @file attn_dml_prv.c
 *
 * @brief ALP attention manager DML private functions
 *
 *****************************************************************************/
#include <unistd.h>
#include <errno.h>
#include <string.h>

#include <glib.h>
#include <glib/gstdio.h>

#include <hiker/types.h>
#include <hiker/prv/attnmgr_traces.h>
#include <hiker/prv/attnmgr_dml.h>

#include "attn_dml_prv.h"

/**
 * @brief Translate SAL errors to Attention error codes.
 */
alp_status_t
alp_prv_attn_dml_convert_sal_error (alp_status_t salError)
{
    ALP_ATLX(DML, " %x", salError);
    switch (salError) {
        case ALP_STATUS_SAL_BADPARAM:
            return ALP_STATUS_ATTN_DML_BAD_PARAM;
        case ALP_STATUS_SAL_NOMEM:
            return ALP_STATUS_ATTN_DML_NO_MEMORY;
        case ALP_STATUS_SAL_NORECORDS:
            return ALP_STATUS_ATTN_DML_NO_RECORDS;
        case ALP_STATUS_SAL_BUSY:
        case ALP_STATUS_SAL_LOCKED:
        case ALP_STATUS_SAL_READONLY:
        case ALP_STATUS_SAL_IOERR:
        case ALP_STATUS_SAL_FULL:
        case ALP_STATUS_SAL_CANTOPEN:
        case ALP_STATUS_SAL_TOOBIG:
        case ALP_STATUS_SAL_AUTH:
        case ALP_STATUS_SAL_ERROR:
        case ALP_STATUS_SAL_NOLFS:
        case ALP_STATUS_SAL_PROTOCOL:
        case ALP_STATUS_SAL_EMPTY:
            return ALP_STATUS_ATTN_DML_DATABASE_ERROR;
        case ALP_STATUS_SAL_INTERRUPT:
        case ALP_STATUS_SAL_MISMATCH:
        case ALP_STATUS_SAL_MISUSE:
        case ALP_STATUS_SAL_SCHEMA:
        case ALP_STATUS_SAL_NOTFOUND:
        case ALP_STATUS_SAL_INTERNAL:
        case ALP_STATUS_SAL_PERM:
        case ALP_STATUS_SAL_ABORT:
        case ALP_STATUS_SAL_UNKNOWN:
        case ALP_STATUS_SAL_DMLEXISTS:
            return ALP_STATUS_ATTN_DML_INTERNAL_ERROR;
        case ALP_STATUS_SAL_CONSTRAINT:
            return ALP_STATUS_ATTN_DML_DUPLICATE;
        case ALP_STATUS_SAL_CORRUPT:
            return ALP_STATUS_ATTN_DML_CORRUPT;
    }

    return ALP_STATUS_OK;
}

char *
alp_prv_attn_dml_get_order_string (AlpAttnDmlOrderByType orderBy)
{
	static char *clauses[] = {
		"",
		"Priority, Created ASC",
		"Created, Priority DESC",
	};

	if (orderBy < NumOrderByAttnTypes) {
		return clauses[orderBy];
	}

	return NULL;
}

/**
 * @brief Verify that the specified LUID exists as an alertLUID
 * in the ALERTS table.
 *
 * @param [in] an AlpDmlH handle for the data model
 * @param [in] an AlpLuid for the alertLuid to be verified.
 *
 * @return alp_status_t
 */
alp_status_t
alp_prv_attn_dml_validate_luid (AlpDmlH dmlH, 
                                AlpAttnDmlItemType itemType,
                                AlpLuid luid)
{
	int32_t      count;
    char        *countClause;
	alp_status_t err;

    ALP_ATLX (DML,"called with dmlH %x type %d luid %d",
              (unsigned) dmlH, itemType, luid);
	/* Validate parameter */
	if (!dmlH) {
		return ALP_STATUS_ATTN_DML_BAD_PARAM;
	}

    switch (itemType) {
    case AlertItem:
    case PropertyListItem:
        countClause =
            "SELECT COUNT(*) FROM "
            kAttnAlertsTableName " WHERE "
            kAttnAlertNameAlertLuid " = %d;";
        break;
    case PropertyItem:
        countClause =
            "SELECT COUNT(*) FROM "
            kAttnPropsTableName " WHERE "
            kAttnPropNamePropLuid " = %d;";
        break;
    default:
		return ALP_STATUS_ATTN_DML_BAD_PARAM;
	}

    /* Validate entityID */
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, countClause,luid);
    if (err) {
        return alp_prv_attn_dml_convert_sal_error (err);
    }

	if (!count) {
		return ALP_STATUS_ATTN_DML_INVALID_LUID;
    }

	return ALP_STATUS_OK;
}

/**
 * @brief Create attention database.
 *
 * Called by the database initialize routine to (re) create database
 * should it not exist or be corrupted. 
 *
 * @param[in] db_pathname desired pathname of database
 */
alp_status_t
alp_prv_attn_dml_create_database (const char *db_pathname)
{
    alp_status_t  err;
    char         *schema_pathname;

    /* Get the pathname to the schema file. */
    
    schema_pathname = g_build_path (G_DIR_SEPARATOR_S,
                                    kAttnResourceDirectory,
                                    kAttnSchemaResourceFileName,
                                    NULL);

    /* Allow SAL to create the Attention Manager database with schema. */
    err =  alp_sal_database_create(db_pathname, schema_pathname);
    if (err) {
        ALP_ATLX (DAEMON,
                  "failed to create database %s with schema %s: error %x",
                 db_pathname, schema_pathname, err);
        g_free (schema_pathname);
        return alp_prv_attn_dml_convert_sal_error (err);
    }
    g_free (schema_pathname);

    return err; /* ALP_STATUS_OK */
}

/**
 * @brief Initialize attention manager database.
 *        Called by the attention manager daemon to verify the existence
 *        and integrity of the database.
 *
 * @param[in] db_pathname desired pathname of database
 *
 * @return ALP_STATUS_OK if database exists and is valid
 */
alp_status_t
alp_prv_attn_dml_init_database (const char *db_pathname)
{
    alp_status_t err;
    struct stat  file_stat;

    /* test for database existence */
    if (g_file_test (db_pathname, G_FILE_TEST_IS_REGULAR)) {
        if (g_stat (db_pathname, &file_stat) < 0) {
            ALP_ATLX (DAEMON, "unable to stat %s: %d %s",
                      db_pathname, errno, strerror(errno));
        } else {
            if (file_stat.st_size > 0) {
                ALP_ATLX(DAEMON, "%s db valid", db_pathname);
                /* database should be okay to use */
                return ALP_STATUS_OK;
            }
        }
    } else {
        ALP_ATLX (DAEMON, "file %s is not a regular file or does not exist",
                  db_pathname);
    }

    /*
     * If file exists, it is either not a regular file or size 0
     * so remove it.
     */
    g_remove (db_pathname);

    /* create datamodel */
    err = alp_prv_attn_dml_create_database (db_pathname);

    return err;
}
