/********************************************************************

Copyright  2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/**
 * @file attn_dml_read.c
 *
 * @brief ALP attention manager data model read implementation
 *
 *****************************************************************************/
#define _XOPEN_SOURCE

#include <time.h>
#include <unistd.h>

#include <hiker/types.h>
#include <hiker/prv/attnmgr_traces.h>
#include <hiker/prv/attnmgr_dml.h>

#include "attn_dml_prv.h"

/**
 * @brief Creates and initializes a new AlpDmlStatementH that can be used to 
 * iterate over a set of AlertItem's in the Attention data model.
 *
 * After calling this routine, you must call alp_attn_dml_statement_next()
 * to advance the statement to the first object in the result set before
 * you can call any of the alp_attn_dml_statement_get_* API to retrieve
 * information from an object.
 *
 * It is important to keep track of which alp_attn_dml_*_statement_begin
 * function you use to initialize which AlpDmlStatementH, because that will
 * dictate which objects are included in the statement's result set, and what
 * AlpAttnDmlColumnType's you can use to fetch values from items in the
 * statement.
 *
 * If there are no records found that match the search critereon, numResults,
 * if it was supplied will be set to 0, ALP_STATUS_ATTN_DML_NO_RECORDS
 * will be returned, and the stmtH parameter will not be initialized
 * (so the caller should not call alp_attn_dml_statement_next (or end)
 * on it.

 * The AlpDmlH handle must have been initialized by a previously opened.
 *
 * @param [in]  an AlpDmlH structure which is an AlpSALHandle
 * @param [out] a pointer to an AlpDmlStatementH, which is an AlpSALStatement
 * @param [in]  result order specification
 * @param [out] an integer indicating the number of items found
 *
 * @see alp_attn_dml_statement_next
 * @see alp_attn_dml_statement_end
 *
 * @return alp_status_t
 */
alp_status_t
alp_attn_dml_alert_statement_begin (
    AlpDmlH                  dmlH,
    AlpDmlStatementH        *stmtH,
    char                    *source,
    char                    *name,
    char                    *handle,
    AlpLuid                  alertLuid,
    AlpAttnDmlOrderByType    orderBy,
    uint32_t                *numResults)
{
    AlpSALStatement         *salStmt = NULL;
    AlpAttnDmlStatementType *attnStmtP = NULL;
    char                    *sql;
    char                    *orderByClause = NULL;
    char                    *whereClause = NULL;
    char                    *firstClause = NULL;
    char                    *countClause = NULL;
    alp_status_t             err;

    /* Validate parameters */
    if (!dmlH || !stmtH) {
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    /* Allocate new statement handle */
    attnStmtP =
        (AlpAttnDmlStatementType *)malloc(sizeof(AlpAttnDmlStatementType));
    if (!attnStmtP) {
        return ALP_STATUS_ATTN_DML_NO_MEMORY;
    }
    attnStmtP->itemType = AlertItem;

    /* Build order by clause */
    if (orderBy != OrderByNone) {
        err = alp_sal_string_build(&orderByClause, " ORDER BY %s", 
                                   alp_prv_attn_dml_get_order_string(orderBy));
    } else {
        err = alp_sal_string_build(&orderByClause, "");
    }
    if (err) goto failure;

    /* Build count statement */
    err = alp_sal_string_build(&countClause,
                               "SELECT COUNT(*) FROM " kAttnAlertsTableName);
    if (err) goto failure;

    /* Build result set clause */
    err = alp_sal_string_build(&firstClause, 
                               "SELECT "
                               kAttnAlertFields
                               " FROM "
                               kAttnAlertsTableName);
    if (err) goto failure;

    /* Build where clause */
    if (alertLuid) {
        err = alp_sal_string_build(
            &whereClause, " WHERE " kAttnAlertNameAlertLuid " =  %d",
            alertLuid);
    } else if (source || name || handle) {
        char *sourceClause = NULL;
        char *nameClause = NULL;
        char *handleClause = NULL;
            
        err = ALP_STATUS_OK;
        if (source) {
            err = alp_sal_string_build (
                &sourceClause, "%s = %Q", kAttnAlertNameSource, source);
        }
        if (name && err == ALP_STATUS_OK) {
            err = alp_sal_string_build (
                &nameClause, "%s = %Q", kAttnAlertNameName, name);
        }
        if (handle && err == ALP_STATUS_OK) {
            err = alp_sal_string_build (
                &handleClause, "%s = %Q", kAttnAlertNameHandle, handle);
        }
        if (err == ALP_STATUS_OK) {
            err = alp_sal_string_build (
                &whereClause, " WHERE %s%s%s%s%s",
                sourceClause,
                (sourceClause && (nameClause || handleClause)) ? " AND " : "",
                nameClause,
                (nameClause && handleClause) ? " AND " : "",
                handleClause);
        }
        if (sourceClause) alp_sal_string_destroy (sourceClause);
        if (nameClause)   alp_sal_string_destroy (nameClause);
        if (handleClause) alp_sal_string_destroy (handleClause);
    } else {
        err = alp_sal_string_build(&whereClause, "");
    }
    if (err) goto failure;

    ALP_ATLX(DML, "first = <%s> count = <%s> order = <%s> where = <%s>",
             firstClause, countClause, orderByClause, whereClause);

    /* Count results */
    if (numResults) {
        err = alp_sal_string_build(&sql, "%q%s;", countClause, whereClause);
        if (err) {
            const char *errMsg = alp_sal_get_error_message((AlpSALHandle)dmlH);
            ALP_ATLX(DML, "count sql string build: %s", errMsg);
            goto failure;
        }

        err = alp_sal_sql_get_value((AlpSALHandle)dmlH, numResults, sql);
        if (err) {
            const char *errMsg = alp_sal_get_error_message((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"count sql get results: %s", errMsg);
            ALP_ATLX(DML,"count sql <%s>", sql);
            goto failure;
        }

        if (*numResults == 0) {
            /* Convert no results to ALP_ATTN_CONTACTS_DML_NO_RECORDS */
            err = ALP_STATUS_SAL_NORECORDS;
            goto failure;
        }
    }

    /* Prepare statement */
    err = alp_sal_string_build(&sql, "%q%q%s;",
                               firstClause, orderByClause, whereClause);
    if (err) {
        const char *errMsg = alp_sal_get_error_message((AlpSALHandle)dmlH);
        ALP_ATLX(DML, "query string build: %s", errMsg);
        goto failure;
    }
    /* Initialize a statement with the final SQL */
    err = alp_sal_statement_initialize((AlpSALHandle)dmlH, &salStmt, sql);
    if (err)  {
        const char *errMsg = alp_sal_get_error_message((AlpSALHandle)dmlH);
        ALP_ATLX(DML, "statement_init: %s", errMsg);
        ALP_ATLX(DML, "statement sql <%s>", sql);
        goto failure;
    }

    attnStmtP->salStmtH = salStmt;
    *stmtH = (AlpDmlStatementH) attnStmtP;

    goto success;

failure:
    if (attnStmtP) {
        free (attnStmtP);
    }

success:
    if (countClause) {
        alp_sal_string_destroy (countClause);
    }
    if (firstClause) {
        alp_sal_string_destroy (firstClause);
    }
    if (sql) {
        alp_sal_string_destroy (sql);
    }
    if (orderByClause) {
        alp_sal_string_destroy (orderByClause);
    }
    if (whereClause) {
        alp_sal_string_destroy (whereClause);
    }
    
    return alp_prv_attn_dml_convert_sal_error (err);
}

/**
 * @brief Creates and initializes a new AlpDmlStatementH that can be used to 
 * iterate over a set of PropertyItem's in the Attention data model.
 *
 * After calling this routine, you must call alp_attn_dml_statement_next()
 * to advance the statement to the first object in the result set before
 * you can call any of the alp_attn_dml_statement_get_* API to retrieve
 * information from an object.
 *
 * It is important to keep track of which alp_attn_dml_*_statement_begin
 * function you use to initialize which AlpDmlStatementH, because that will
 * dictate which objects are included in the statement's result set, and what
 * AlpAttnDmlColumnType's you can use to fetch values from items in the
 * statement.
 *
 * If there are no records found that match the search critereon, numResults,
 * if it was supplied will be set to 0, ALP_STATUS_ATTN_DML_NO_RECORDS
 * will be returned, and the stmtH parameter will not be initialized
 * (so the caller should not call alp_attn_dml_statement_next (or end)
 * on it.

 * The AlpDmlH handle must have been initialized by a previously opened.
 *
 * @param [in]  an AlpDmlH structure which is an AlpSALHandle
 * @param [out] a pointer to an AlpDmlStatementH, which is an AlpSALStatement
 * @param [in]  an optional alertLuid to limit the result set.
 * @param [in]  an optional propLuid to limit the result set.
 * @param [in]  an optional property spec to limit the result set.
 * @param [out] an integer indicating the number of items found
 *
 * @see alp_attn_dml_statement_next
 * @see alp_attn_dml_statement_end
 *
 * @return alp_status_t
 */
alp_status_t
alp_attn_dml_prop_statement_begin (
    AlpDmlH                  dmlH,
    AlpDmlStatementH        *stmtH,
    /*  more selection args */
    AlpLuid                  alertLuid,
    AlpLuid                  propLuid,
    char                    *property,
    uint32_t                *numResults)
{
    AlpSALStatement         *salStmt         = NULL;
    AlpAttnDmlStatementType *attnStmtP       = NULL;
    char                    *sql;
    char                    *limitClause     = NULL;
    char                    *propertyClause  = NULL;
    char                    *alertLuidClause = NULL;
    char                    *propLuidClause  = NULL;
    char                    *firstClause     = NULL;
    char                    *countClause     = NULL;
    alp_status_t             err;
    bool                     do_limits          = false;

    /* Validate parameters */
    if (!dmlH || !stmtH) {
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    /* Allocate new statement handle */
    attnStmtP =
        (AlpAttnDmlStatementType *)malloc(sizeof(AlpAttnDmlStatementType));
    if (!attnStmtP) {
        return ALP_STATUS_ATTN_DML_NO_MEMORY;
    }
    attnStmtP->itemType = PropertyItem;

    /* Build limit clause */
    if (property) {
        err = alp_sal_string_build (&propertyClause,
                                    kAttnPropNameName " = %Q",
                                    property);
        if (err) goto failure;
        do_limits = true;
    }
    if (alertLuid) {
        err = alp_sal_string_build (&alertLuidClause,
                                    kAttnPropNameAlertLuid " = %d",
                                    alertLuid);
        if (err) goto failure;
        do_limits = true;
    }
    if (propLuid) {
        err = alp_sal_string_build (&propLuidClause,
                                    kAttnPropNamePropLuid " = %d",
                                    propLuid);
        if (err) goto failure;
        do_limits = true;
    }
    if (do_limits) {
        err = alp_sal_string_build (
            &limitClause,
            " WHERE %s%s%s%s%s",
            propertyClause,
            (propertyClause && (alertLuidClause || propLuidClause)) ? " AND " : "",
            alertLuidClause,
            (alertLuidClause && propLuidClause) ? " AND " : "",
            propLuidClause);
        if (err) goto failure;
    }

    ALP_ATLX(DML, "limit clause <%s>", limitClause);

    /* Build first segment of statement */
    err = alp_sal_string_build(&countClause,
                               "SELECT COUNT(*) FROM " kAttnPropsTableName);
    if (err) goto failure;

    err = alp_sal_string_build(&firstClause, 
                               "SELECT "
                               kAttnPropFields
                               " FROM "
                               kAttnPropsTableName);
    if (err) goto failure;

    /* Count results */
    if (numResults) {
        err = alp_sal_string_build(&sql, "%s%s;", countClause, limitClause);
        if (err) goto failure;

        ALP_ATLX(DML,"count sql <%s>", sql);
        err = alp_sal_sql_get_value((AlpSALHandle)dmlH, numResults, sql);
        if (err) {
            const char *errMsg = alp_sal_get_error_message((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"get count: %s", errMsg);
            goto failure;
        }

        if (*numResults == 0) {
            /* Convert no results to ALP_ATTN_CONTACTS_DML_NO_RECORDS */
            err = ALP_STATUS_SAL_NORECORDS;
            goto failure;
        }
    }

    /* Prepare statement */
    err = alp_sal_string_build(&sql, "%s%s;", firstClause, limitClause);
    if (err) {
        const char *errMsg = alp_sal_get_error_message((AlpSALHandle)dmlH);
        ALP_ATLX(DML, "build final sql: %s", errMsg);
        goto failure;
    }

    ALP_ATLX(DML, "final sql <%s>", sql);

    /* Initialize a statement with the final SQL */
    err = alp_sal_statement_initialize((AlpSALHandle)dmlH, &salStmt, sql);
    if (err) {
        const char *errMsg = alp_sal_get_error_message((AlpSALHandle)dmlH);
        ALP_ATLX(DML, "stmt init:  %s", errMsg);
        goto failure;
    }

    attnStmtP->salStmtH = salStmt;
    *stmtH = (AlpDmlStatementH) attnStmtP;

    goto success;

failure:
    ALP_ATLX(DML, "failure: %x", err);
    if (attnStmtP) {
        free (attnStmtP);
    }

success:
    if (countClause) {
        alp_sal_string_destroy (countClause);
    }
    if (firstClause) {
        alp_sal_string_destroy (firstClause);
    }
    if (sql) {
        alp_sal_string_destroy (sql);
    }
    if (limitClause) {
        alp_sal_string_destroy (limitClause);
    }
    if (propertyClause) {
        alp_sal_string_destroy (propertyClause);
    }
    if (alertLuidClause) {
        alp_sal_string_destroy (alertLuidClause);
    }
    if (propLuidClause) {
        alp_sal_string_destroy (propLuidClause);
    }

    return alp_prv_attn_dml_convert_sal_error (err);
}

/**
 * @brief Given a statement handle initialized with
 * alp_attn_dml_*_statement_init() API, this routine will step the 
 * statement forward to point to the next item in the sequence.  Note that
 * you must call this API in order to get to the first element in the list.
 *
 * @param [in] stmtH Statement handle that has been initialized with
 *     alp_attn_dml_*_statement_init() API
 *
 * @return alp_status_t
 *
 * @see alp_attn_dml_statement_begin
 * @see alp_attn_dml_statement_end
 */
alp_status_t
alp_attn_dml_statement_next (AlpDmlStatementH stmtH)
{
    AlpAttnDmlStatementType *attnStmtP = (AlpAttnDmlStatementType *) stmtH;
    alp_status_t err;

    /* Validate parameters */
    if (!attnStmtP) {
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    /* Check if there's another result */
    err = alp_sal_statement_next_result(attnStmtP->salStmtH);
    if (err) {
        const char *errMsg;

        errMsg = alp_sal_get_error_message(alp_sal_statement_get_handle(stmtH));
        ALP_ATLX(DML, "next result: %s", errMsg);
        return alp_prv_attn_dml_convert_sal_error (err);
    }

    return ALP_STATUS_OK;
}

/**
 *  @brief Close/end the selection statement
 *
 *  stmtH must be from a previous call to begin
 *
 *  @param [in] a pointer to an AlpDmlStatementH, which is an AlpSALStatement
 *
 *  @return alp_status_t
 *
 *  @see alp_attn_dml_statement_begin
 *  @see alp_attn_dml_statement_next
 *
 */
alp_status_t
alp_attn_dml_statement_end (AlpDmlStatementH stmtH)
{
    AlpAttnDmlStatementType *attnStmtP = 
        (AlpAttnDmlStatementType *) stmtH;
    alp_status_t err;

    ALP_ATLX(DML, "about to end stmt %x", (unsigned int) stmtH);
    if (!attnStmtP) {
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }
    err = alp_sal_statement_destroy (attnStmtP->salStmtH);
    if (err) {
        return alp_prv_attn_dml_convert_sal_error (err);
    }

    free (attnStmtP);

    return ALP_STATUS_OK;
}

/**
 *  @brief Get a int32_t from the current statement record
 *  statementH must be from a previous call to begin
 *
 *  @param [in] an AlpDmlStatementH, which is an AlpSALStatement
 *  @param [in] an AlpAttnColumnType to specify the field to get the value from
 *  @param [out] the int32_t value of the field
 *
 *  @return alp_status_t
 */
alp_status_t
alp_attn_dml_statement_get_int32 (AlpDmlStatementH      stmtH,
                                  AlpAttnDmlColumnType  column,
                                  int32_t              *value)
{
    AlpAttnDmlStatementType *attnStmtP = (AlpAttnDmlStatementType *) stmtH;
    AlpAttnDmlItemType itemType;
    AlpSALStatement salStmtH;

    /* Validate parameters */
    if (!stmtH || !value) {
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    itemType = attnStmtP->itemType;
    salStmtH = attnStmtP->salStmtH;

    switch (itemType) {
    case AlertItem:
        switch (column) {
        case Priority:
            *value = alp_sal_statement_get_int (salStmtH, kAttnAlertColPriority);
            break;

        case Duration:
            *value = alp_sal_statement_get_int (salStmtH, kAttnAlertColDuration);
            break;

        case Attributes:
            *value = alp_sal_statement_get_int (salStmtH, kAttnAlertColAttributes);
            break;

        case UserId:
            *value = alp_sal_statement_get_int (salStmtH, kAttnAlertColUserId);
            break;

        case GroupId:
            *value = alp_sal_statement_get_int (salStmtH, kAttnAlertColGroupId);
            break;

        default:
            return ALP_STATUS_ATTN_DML_BAD_PARAM;
        }
        break;

    default:
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    return ALP_STATUS_OK;
}

/**
 *  @brief Get a uint32_t from the current statement record
 *  statementH must be from a previous call to begin
 *
 *  @param [in] stmtH  an AlpDmlStatementH, which is an AlpSALStatement
 *  @param [in] column an AlpAttnColumnType to specify the field to get
 *                     the value from
 *  @param [out] value the uint32_t value of the field
 *
 *  @return alp_status_t
 */
alp_status_t
alp_attn_dml_statement_get_uint32 (AlpDmlStatementH      stmtH,
                                   AlpAttnDmlColumnType  column,
                                   uint32_t             *value)
{
    AlpAttnDmlStatementType *attnStmtP = (AlpAttnDmlStatementType *) stmtH;
    AlpAttnDmlItemType itemType;
    AlpSALStatement salStmtH;

    /* Validate parameters */
    if (!stmtH || !value) {
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    itemType = attnStmtP->itemType;
    salStmtH = attnStmtP->salStmtH;

    switch (itemType) {
    case AlertItem:
        switch (column) {
        case Attributes:
            *value =
                (uint32_t) alp_sal_statement_get_int (salStmtH,
                                                      kAttnAlertColAttributes);
            break;

        default:
            return ALP_STATUS_ATTN_DML_BAD_PARAM;
        }
        break;

    case PropertyItem:
        switch (column) {
        case PropType:
            *value =
                (uint32_t) alp_sal_statement_get_int (salStmtH,
                                                      kAttnPropColType);
            break;

        default:
            return ALP_STATUS_ATTN_DML_BAD_PARAM;
        }
        break;

    default:
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    return ALP_STATUS_OK;
}

/**
 *  @brief Get a int64_t from the current statement record
 *  statementH must be from a previous call to begin
 *
 *  @param [in] an AlpDmlStatementH, which is an AlpSALStatement
 *  @param [in] an AlpAttnColumnType to specify the field to get the value from
 *  @param [out] the int64_t value of the field
 *
 *  @return alp_status_t
 */
alp_status_t
alp_attn_dml_statement_get_int64 (AlpDmlStatementH      stmtH,
                                  AlpAttnDmlColumnType  column,
                                  int64_t              *value)
{
    AlpAttnDmlStatementType *attnStmtP = (AlpAttnDmlStatementType *) stmtH;
    AlpAttnDmlItemType itemType;
    AlpSALStatement salStmtH;

    /* Validate parameters */
    if (!stmtH || !value) {
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    itemType = attnStmtP->itemType;
    salStmtH = attnStmtP->salStmtH;

    switch (itemType) {
    case PropertyItem:
        switch (column) {
        case AlertLuid:
            *value = alp_sal_statement_get_int (salStmtH,
                                                kAttnPropColAlertLuid);
        default:
            return ALP_STATUS_ATTN_DML_BAD_PARAM;
        }
        break;

    default:
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    return ALP_STATUS_OK;
}

/**
 *  @brief Get a bool from the current statement record
 *  statementH must be from a previous call to begin
 *
 *  @param [in] an AlpDmlStatementH, which is an AlpSALStatement
 *  @param [in] an AlpAttnColumnType to specify the field to get the value from
 *  @param [out] the bool value of the field
 *
 *  @return alp_status_t
 */
alp_status_t
alp_attn_dml_statement_get_bool (AlpDmlStatementH      stmtH,
                                 AlpAttnDmlColumnType  column,
                                 bool                 *value)
{
    AlpAttnDmlStatementType *attnStmtP = (AlpAttnDmlStatementType *) stmtH;
    AlpAttnDmlItemType itemType;
    AlpSALStatement salStmtH;

    /* Validate parameters */
    if (!stmtH || !value) {
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    itemType = attnStmtP->itemType;
    salStmtH = attnStmtP->salStmtH;

    switch (itemType) {
    case AlertItem:
        switch (column) {
        case Pending:
            if (alp_sal_statement_get_int (salStmtH, kAttnAlertColPending))
                *value = true;
            else
                *value = false;
            break;

        default:
            return ALP_STATUS_ATTN_DML_BAD_PARAM;
        }
        break;

    default:
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    return ALP_STATUS_OK;
}

/**
 *  @ brief Get a text string from the current statement record
 *  statementH must be from a previous call to begin
 *
 *  @param [in] an AlpDmlStatementH, which is an AlpSALStatement
 *  @param [in] an AlpAttnColumnType to specify the field to get the value from
 *  @param [out] the char* (string) value of the field
 *
 *  @return alp_status_t
 */
alp_status_t
alp_attn_dml_statement_get_string (AlpDmlStatementH       stmtH,
                                   AlpAttnDmlColumnType   column,
                                   char                 **value)
{
    AlpAttnDmlStatementType *attnStmtP = (AlpAttnDmlStatementType *) stmtH;
    AlpAttnDmlItemType itemType;
    AlpSALStatement salStmtH;

    /* Validate parameters */
    if (!stmtH || !value) {
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }
    itemType = attnStmtP->itemType;
    salStmtH = attnStmtP->salStmtH;

    switch (itemType) {
    case AlertItem:
        switch (column) {
        case Source:
            *value = alp_sal_statement_get_text (salStmtH, kAttnAlertColSource, true);
            break;

        case Name:
            *value = alp_sal_statement_get_text (salStmtH, kAttnAlertColName, true);
            break;

        case Handle:
            *value = alp_sal_statement_get_text (salStmtH, kAttnAlertColHandle, true);
            break;

        case Created:  /* get string version of the created timestamp */
            *value = alp_sal_statement_get_text (salStmtH, kAttnAlertColCreated, true);
            break;
            
        case Updated:  /* get string version of the updated timestamp */
            *value = alp_sal_statement_get_text (salStmtH, kAttnAlertColUpdated, true);
            break;
            
        default:
            return ALP_STATUS_ATTN_DML_BAD_PARAM;
        }
        break;

    case PropertyItem:
        switch (column) {
        case Name:
            *value = alp_sal_statement_get_text (salStmtH, kAttnPropColName, true);
            break;

        default:
            return ALP_STATUS_ATTN_DML_BAD_PARAM;
        }
        break;

    default:
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    return ALP_STATUS_OK;
}

/**
 *  @ brief Get a blob from the current statement record
 *  statementH must be from a previous call to begin
 *
 *  @param [in] stmtH  an AlpDmlStatementH, which is an AlpSALStatement
 *  @param [in] column an AlpAttnColumnType to specify the field to
 *                     get the value from
 *  @param [out] value the char* (string) value of the field
 *
 *  @return alp_status_t
 */
alp_status_t
alp_attn_dml_statement_get_blob (AlpDmlStatementH       stmtH,
                                 AlpAttnDmlColumnType   column,
                                 bool                   copy,
                                 void                 **value,
                                 uint32_t              *size)
{
    AlpAttnDmlStatementType *attnStmtP = (AlpAttnDmlStatementType *) stmtH;
    AlpAttnDmlItemType itemType;
    AlpSALStatement salStmtH;

    /* Validate parameters */
    if (!stmtH || !value) {
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }
    itemType = attnStmtP->itemType;
    salStmtH = attnStmtP->salStmtH;

    switch (itemType) {
    case PropertyItem:
        switch (column) {
        case Value:
            *value =
                alp_sal_statement_get_blob (salStmtH, kAttnPropColValue, size, true);
            break;

        default:
            return ALP_STATUS_ATTN_DML_BAD_PARAM;
        }
        break;

    default:
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    return ALP_STATUS_OK;
}

/**
 *  @brief Get the timestamp from the current statement record
 *  statementH must be from a previous call to begin
 *
 *  @param [in] an AlpDmlStatementH, which is an AlpSALStatement
 *  @param [in] an AlpAttnColumnType to specify the field to get the value from
 *  @param [out] the timestamp value from the record
 *
 *  @return alp_status_t
 */
alp_status_t
alp_attn_dml_statement_get_timestamp (AlpDmlStatementH      stmtH,
                                      AlpAttnDmlColumnType  column,
                                      time_t               *value)
{
    AlpAttnDmlStatementType *attnStmtP = (AlpAttnDmlStatementType *) stmtH;
    AlpAttnDmlItemType       itemType;
    AlpSALStatement          salStmtH;
    char                    *timeString;
    char                    *result;
    struct tm                tm;
    time_t                   t;

    /* Validate parameters */
    if (!stmtH || !value) {
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    itemType = attnStmtP->itemType;
    salStmtH = attnStmtP->salStmtH;

    switch (itemType) {
    case AlertItem:
        switch (column) {
        case kAttnAlertColCreated:
            timeString = alp_sal_statement_get_text (salStmtH,
                                                     kAttnAlertColCreated,
                                                     false);
            /* convert time string to int */
            result = strptime (timeString, kAttnTimestampFormat, &tm);
            if (result == NULL) {
                /* conversion failed */
                return ALP_STATUS_ATTN_DML_CORRUPT;
            }
            t = mktime(&tm);
            ALP_ATLX(DML,
                     "created timestamp string <%s> seconds since epoch %d",
                     timeString, t);
            if (t == (time_t) -1) {
                return ALP_STATUS_ATTN_DML_CORRUPT;
            }
            *value = t;
            break;

        case kAttnAlertColUpdated:
            timeString = alp_sal_statement_get_text (salStmtH,
                                                     kAttnAlertColUpdated,
                                                     false);
            /* convert time string to int */
            result = strptime (timeString, kAttnTimestampFormat, &tm);
            if (result == NULL) {
                /* conversion failed */
                return ALP_STATUS_ATTN_DML_CORRUPT;
            }
            t = mktime(&tm);
            ALP_ATLX(DML,
                     "updated timestamp string <%s> seconds since epoch %d",
                     timeString, t);
            if (t == (time_t) -1) {
                return ALP_STATUS_ATTN_DML_CORRUPT;
            }
            *value = t;
            break;

        default:
            return ALP_STATUS_ATTN_DML_BAD_PARAM;
        }
        break;

    default:
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    return ALP_STATUS_OK;
}

/**
 *  @brief Get the luid from the current statement record
 *
 *  stmtH must be from a previous call to begin
 *
 *  @param [in] an AlpDmlStatementH, which is an AlpSALStatement
 *  @param [out] the item luid value from the record
 *
 *  @return alp_status_t
 */
alp_status_t
alp_attn_dml_statement_get_luid (
    AlpDmlStatementH  stmtH,
    AlpLuid          *itemLuid)
{
    AlpAttnDmlStatementType *attnStmtP;
    AlpSALStatement          salStmtH;
    AlpAttnDmlItemType       itemType;

    if (!stmtH || !itemLuid) {
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    attnStmtP = (AlpAttnDmlStatementType *) stmtH;
    salStmtH  = attnStmtP->salStmtH;
    itemType  = attnStmtP->itemType;

    switch (itemType) {
    case AlertItem:
        *itemLuid = (AlpLuid) alp_sal_statement_get_int (salStmtH,
                                                         kAttnAlertColAlertLuid);
        break;
        
    case PropertyItem:    
        if (!itemLuid) return ALP_STATUS_ATTN_DML_BAD_PARAM;
        *itemLuid  = (AlpLuid) alp_sal_statement_get_int(salStmtH,
                                                         kAttnPropColPropLuid);
        break;

    default:
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    return ALP_STATUS_OK;
}
