/********************************************************************

Copyright  2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/**
 * @file attn_dml_write.c
 *
 * @brief ALP attention manager data model write implementation
 *
 *****************************************************************************/
#include <unistd.h>

#include <hiker/types.h>
#include <hiker/prv/attnmgr_traces.h>
#include <hiker/prv/attnmgr_dml.h>

#include "attn_dml_prv.h"

alp_status_t
alp_attn_dml_insert(AlpDmlH dmlH,
                    AlpDmlItemH itemH,
                    AlpLuid *luid)
{
    AlpAttnAlertType   *alertP   = (AlpAttnAlertType *) itemH;
    AlpAttnPropType    *propP    = (AlpAttnPropType *)  itemH;
    AlpAttnDmlItemType  itemType;
    alp_status_t        err;
    AlpDmlStatementH   *stmtH;
    const char         *errMsg = NULL;
    AlpLuid             itemLuid;
    AlpLuid             alertLuid;

    ALP_ATLX(DML,"called with dmlH %x itemH %x itemType %d &id = %x",
             (unsigned) dmlH, (unsigned) itemH, alertP ? alertP->itemType : 0, luid);

    /* Validate params */
    if (!dmlH || !itemH) {
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    itemType = alertP->itemType;
    switch (itemType) {
    case AlertItem:
        /*
         * check if record exists prior to insert since it is hard
         * to determine if insert failed due to existing record or
         * some other error (need to do more research on this).
         */
        itemLuid = 0;
        err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &itemLuid, 
                                    "SELECT "
                                    kAttnAlertNameAlertLuid " FROM "
                                    kAttnAlertsTableName " WHERE "
                                    kAttnAlertNameSource " = %Q AND "
                                    kAttnAlertNameName   " = %Q AND "
                                    kAttnAlertNameHandle " = %Q;",
                                    alertP->source,
                                    alertP->name,
                                    alertP->handle);
        if (err == ALP_STATUS_OK && itemLuid != 0) {
            ALP_ATLX(DML, "record %s,%s,%s already exists at %d",
                     alertP->source, alertP->name, alertP->handle, itemLuid);
            return ALP_STATUS_ATTN_DML_EXISTS;
        }

        /* Create sql statement */
        err = alp_sal_statement_initialize(
            (AlpSALHandle)dmlH, &stmtH,
            "INSERT INTO "
            kAttnAlertsTableName
            " ("
            kAttnAlertNameAlertLuid  ", "
            kAttnAlertNameSource     ", "
            kAttnAlertNameName       ", "
            kAttnAlertNameHandle     ", "
            kAttnAlertNamePriority   ", "
            kAttnAlertNameDuration   ", "
            kAttnAlertNamePending    ", "
            kAttnAlertNameAttributes ", "
            kAttnAlertNameUserId     ", "
            kAttnAlertNameGroupId
            ") VALUES "
            "(null, ?, ?, ?, ?, ?, ?, ?, ?, ?);");
        
        if (err) {
            errMsg = alp_sal_get_error_message ((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"ALERT STMT INIT: ERRMSG: %s code %x",
                     errMsg, sqlite3_errcode ((sqlite3 *) dmlH));
            return alp_prv_attn_dml_convert_sal_error (err);
        }

        /* Bind parameters to statement */
        err = alp_sal_statement_set_text(stmtH, -1, alertP->source, 0);
        if (err) {
            errMsg = alp_sal_get_error_message ((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"BIND Source: %s: ERRMSG: %s", alertP->source, errMsg);
            return alp_prv_attn_dml_convert_sal_error (err);
        }
        err = alp_sal_statement_set_text(stmtH, -1, alertP->name, 0);
        if (err) {
            errMsg = alp_sal_get_error_message ((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"BIND Name: %s: ERRMSG: %s", alertP->name, errMsg);
            return alp_prv_attn_dml_convert_sal_error (err);
        }
        err = alp_sal_statement_set_text(stmtH, -1, alertP->handle, 0);
        if (err) {
            errMsg = alp_sal_get_error_message ((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"BIND Handle: %s: ERRMSG: %s", alertP->handle, errMsg);
            return alp_prv_attn_dml_convert_sal_error (err);
        }
        err = alp_sal_statement_set_int(stmtH, -1, alertP->priority);
        if (err) {
            errMsg = alp_sal_get_error_message ((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"BIND Priority: ERRMSG: %s", errMsg);
            return alp_prv_attn_dml_convert_sal_error (err);
        }
        err = alp_sal_statement_set_int(stmtH, -1, alertP->duration);
        if (err) {
            errMsg = alp_sal_get_error_message ((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"BIND Duration: ERRMSG: %s\n", errMsg);
            return alp_prv_attn_dml_convert_sal_error (err);
        }
        err = alp_sal_statement_set_int(stmtH, -1, alertP->pending);
        if (err) {
            errMsg = alp_sal_get_error_message ((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"BIND Pending: ERRMSG: %s\n", errMsg);
            return alp_prv_attn_dml_convert_sal_error (err);
        }
        err = alp_sal_statement_set_int(stmtH, -1, alertP->attributes);
        if (err) {
            errMsg = alp_sal_get_error_message ((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"BIND Attributes: ERRMSG: %s\n", errMsg);
            return alp_prv_attn_dml_convert_sal_error (err);
        }
        err = alp_sal_statement_set_int(stmtH, -1, alertP->userId);
        if (err) {
            errMsg = alp_sal_get_error_message ((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"BIND UserId: ERRMSG: %s\n", errMsg);
            return alp_prv_attn_dml_convert_sal_error (err);
        }
        err = alp_sal_statement_set_int(stmtH, -1, alertP->groupId);
        if (err) {
            errMsg = alp_sal_get_error_message ((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"BIND GroupId: ERRMSG: %s\n", errMsg);
            return alp_prv_attn_dml_convert_sal_error (err);
        }

        /* Execute insert into ALERTS table */
        err = alp_sal_statement_execute(stmtH);
        if (err) {
            errMsg = alp_sal_get_error_message ((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"EXEC ALERT STMT: ERRMSG: %s code %x",
                     errMsg, sqlite3_errcode ((sqlite3 *) dmlH));
            alp_sal_statement_destroy(stmtH);
            return alp_prv_attn_dml_convert_sal_error (err);
        }
        
        /* Destroy statement */
        err = alp_sal_statement_destroy(stmtH);
        if (err) {
            errMsg = alp_sal_get_error_message ((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"DESTROY ALERT STMT: ERRMSG: %s", errMsg);
            return alp_prv_attn_dml_convert_sal_error (err);
        }

        /*
         * Read back itemLuid from new row. Need to set 
         * using system provided AlpLuid
         */
        err = alp_sal_last_insert_rowid ((AlpSALHandle) dmlH, &itemLuid);
        if (err) {
            errMsg = alp_sal_get_error_message ((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"ALERT: LAST ROWID: ERRMSG: %s", errMsg);
            return alp_prv_attn_dml_convert_sal_error (err);
        }
        alertP->itemLuid = itemLuid; /* fill in the incore item */

        /* write property list out to PROPERTY Table if attached */
        propP = alertP->plist;
        while (propP) {
            ALP_FAIL_FATAL_IF (propP->itemType != PropertyItem,
                               "list item is not a property item: type = %d",
                               propP->itemType);
            propP->alertLuid = itemLuid;
            alp_attn_dml_insert (dmlH, propP, NULL);
            propP = propP->next;
        }
        
        if (luid) *luid   = itemLuid; /* return LUID if wanted */
        break;

    case PropertyItem:
        /* Validate alertLuid */
        alertLuid = propP->alertLuid;
        err = alp_prv_attn_dml_validate_luid(dmlH, AlertItem, alertLuid);
        if (err) {
            return err;
        }

        /* Insert Address into table */
#if 0
        err = alp_sal_sql_execute((AlpSALHandle)dmlH, 
                                  "INSERT INTO "
                                  kAttnPropsTableName
                                  " VALUES("
                                  "NULL, %d, %Q, %Q);", 
                                  alertLuid, propP->name, propP->value);

        if (err) {
            errMsg = alp_sal_get_error_message((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"EXEC PROP DIRECT: ERRMSG: %s", errMsg);
            return alp_prv_attn_dml_convert_sal_error (err);
        }
#else
        /* try this by using a statement instead of direct execute */
        err = alp_sal_statement_initialize((AlpSALHandle)dmlH, &stmtH,
                                           "INSERT INTO "
                                           kAttnPropsTableName    " ("
                                           kAttnPropNamePropLuid  ", "
                                           kAttnPropNameAlertLuid ", "
                                           kAttnPropNameName      ", "
                                           kAttnPropNameType      ", "
                                           kAttnPropNameValue     ")"
                                           "VALUES (null, ?, ?, ?, ?);");
        
        if (err) {
            errMsg = alp_sal_get_error_message ((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"INIT PROP STMT: ERRMSG: %s", errMsg);
            return alp_prv_attn_dml_convert_sal_error (err);
        }
    
        /* Bind parameters to statement */
        err = alp_sal_statement_set_int (stmtH, -1, alertLuid);
        if (err) { 
            errMsg = alp_sal_get_error_message ((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"BIND AlertLUID: ERRMSG: %s", errMsg);
            return alp_prv_attn_dml_convert_sal_error (err);
        }
        err = alp_sal_statement_set_text(stmtH, -1, propP->name, NULL);
        if (err) { 
            errMsg = alp_sal_get_error_message ((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"BIND Name: ERRMSG: %s", errMsg);
            return alp_prv_attn_dml_convert_sal_error (err);
        }
        err = alp_sal_statement_set_int(stmtH, -1, propP->type);
        if (err) { 
            errMsg = alp_sal_get_error_message ((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"BIND Type: ERRMSG: %s", errMsg);
            return alp_prv_attn_dml_convert_sal_error (err);
        }
        err = alp_sal_statement_set_blob(stmtH, -1,
                                         propP->value,
                                         propP->size,
                                         false);
        if (err) { 
            errMsg = alp_sal_get_error_message ((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"BIND Value: ERRMSG: %s", errMsg);
            return alp_prv_attn_dml_convert_sal_error (err);
        }

        /* Execute insert into PROPERTIES table */
        err = alp_sal_statement_execute(stmtH);
        if (err) {
            errMsg = alp_sal_get_error_message ((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"EXEC PROP STMT: ERRMSG: %s", errMsg);
            alp_sal_statement_destroy(stmtH);
            return alp_prv_attn_dml_convert_sal_error (err);
        }
        
        /* Destroy statement */
        err = alp_sal_statement_destroy(stmtH);
        if (err) {
            errMsg = alp_sal_get_error_message ((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"DESTROY PROP STMT: ERRMSG: %s", errMsg);
            return alp_prv_attn_dml_convert_sal_error (err);
        }
#endif

        /* Read back itemLuid from new row. */
        err = alp_sal_last_insert_rowid ((AlpSALHandle) dmlH,
                                         &itemLuid);
        if (err) {
            errMsg = alp_sal_get_error_message ((AlpSALHandle) dmlH);
            ALP_ATLX(DML,"PROP: LAST ROWID: ERRSG: %s", errMsg);
            return alp_prv_attn_dml_convert_sal_error (err);
        }
        propP->itemLuid = itemLuid; /* fill in the incore item */
        if (luid) *luid = itemLuid; /* return LUID if wanted */

        break;

    default:
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    return ALP_STATUS_OK;
}

alp_status_t
alp_attn_dml_update (AlpDmlH     dmlH,
                     AlpDmlItemH itemH)
{
    AlpAttnAlertType    *alertP = (AlpAttnAlertType *) itemH;
    AlpAttnPropType     *propP;
    AlpAttnPropListType *propListP;
    AlpAttnPropType     *property;
    AlpAttnDmlItemType   itemType;
    alp_status_t         err;
    AlpDmlStatementH    *stmtH;
    uint32_t             count = 0;
    AlpLuid              itemLuid;
    bool                 inTransaction = true;
    const char          *errMsg = NULL;
    char                *sql = NULL;

    /* Validate params */
    if (!dmlH || !itemH) {
        ALP_ATLX(DML,"bad parameters: dmlH %x itemH %x",
                 (unsigned int) dmlH,
                 (unsigned int) itemH);
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    itemLuid = 0;
    itemType = alertP->itemType;
    switch (itemType) {
    case AlertItem:
        /* Verify alert record exists */
        if (alertP->source == NULL ||
            alertP->name == NULL ||
            alertP->handle == NULL) {
            ALP_ATLX(DML,"bad parameters:  %s|%s|%s",
                     alertP->source, alertP->name, alertP->handle);
            return ALP_STATUS_ATTN_DML_BAD_PARAM;
        }
        /* check if there is a record matching this item */
        err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &itemLuid, 
                                    "SELECT "
                                    kAttnAlertNameAlertLuid " FROM "
                                    kAttnAlertsTableName " WHERE "
                                    kAttnAlertNameSource " = %Q AND "
                                    kAttnAlertNameName   " = %Q AND "
                                    kAttnAlertNameHandle " = %Q;",
                                    alertP->source,
                                    alertP->name,
                                    alertP->handle);
        if (err) {
            /*
             * need to find the error that will be given if the
             * sql doesn't match and return the appropriate ATTN_DML error.
             * (NO_RECORDS) seems appropriate.
             * Was returning _INVALID_LUID but probably need to replace that
             * error code with something else.
             */
            errMsg = alp_sal_get_error_message((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"find %s,%s,%s failed: %s",
                     alertP->source, alertP->name, alertP->handle, errMsg);
            return alp_prv_attn_dml_convert_sal_error (err);
        }
        if (itemLuid == 0) {
            ALP_ATLX(DML,"update of %s,%s,%s had invalid luid %d",
                     alertP->source, alertP->name, alertP->handle, itemLuid);
            return ALP_STATUS_ATTN_DML_INVALID_LUID;
        } else {
            ALP_ATLX(DML,"found item %s,%s,%s id %d to update",
                     alertP->source, alertP->name, alertP->handle,
                     itemLuid);
        }

        /* set the itemLuid to the value of the record being updated */
        alertP->itemLuid = itemLuid;

        /* Try to begin a transcation */
        ALP_ATLX(DML,"begin transaction (id %d)", alertP->itemLuid);
        err = alp_attn_dml_transaction_begin (dmlH, true);
        ALP_ATLX(DML,"begin transaction err %x", err);
        if (err) {
            errMsg = alp_sal_get_error_message((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"begin transaction: %s: not a problem", errMsg);
            inTransaction = false;
        }

        ALP_ATLX(DML,"initializing update %x", err);
        /* Construct statement to update ALERTS table */
        err = alp_sal_statement_initialize ((AlpSALHandle)dmlH, &stmtH, 
                                            "UPDATE "
                                            kAttnAlertsTableName
                                            " SET "
                                            kAttnAlertNamePriority   " = ?, "
                                            kAttnAlertNameDuration   " = ?, "
                                            kAttnAlertNamePending    " = ?, "
                                            kAttnAlertNameAttributes " = ?, "
                                            kAttnAlertNameUserId     " = ?, "
                                            kAttnAlertNameGroupId    " = ?, "
                                            kAttnAlertNameUpdated
                                            " = DATETIME( \"NOW\" )"
                                            " WHERE "
                                            kAttnAlertNameSource " = ? AND "
                                            kAttnAlertNameName   " = ? AND "
                                            kAttnAlertNameHandle " = ?;");
        ALP_ATLX(DML,"initialized update %x", err);
        if (err) {
            ALP_ATLX(DML,"statement_init: err %x %s", err,
                     alp_sal_get_error_message((AlpSALHandle)dmlH));
            if (inTransaction) {
                alp_attn_dml_transaction_rollback(dmlH);
            }
            err =  alp_prv_attn_dml_convert_sal_error (err);
            break;
        }

        /* items being updated */
        err = alp_sal_statement_set_int (stmtH, -1, alertP->priority);
        err = alp_sal_statement_set_int (stmtH, -1, alertP->duration);
        err = alp_sal_statement_set_int (stmtH, -1, alertP->pending);
        err = alp_sal_statement_set_int (stmtH, -1, alertP->attributes);
        err = alp_sal_statement_set_int (stmtH, -1, alertP->userId);
        err = alp_sal_statement_set_int (stmtH, -1, alertP->groupId);
        /* for the record that matches this tuple */
        err = alp_sal_statement_set_text(stmtH, -1, alertP->source, 0);
        err = alp_sal_statement_set_text(stmtH, -1, alertP->name, 0);
        err = alp_sal_statement_set_text(stmtH, -1, alertP->handle, 0);

        /* Execute update to ALERT and check for success */
        err = alp_sal_statement_execute(stmtH);
        if (err) {
            ALP_ATLX(DML,"statment_exec: %s",
                     alp_sal_get_error_message((AlpSALHandle)dmlH));
            if (inTransaction) {
                alp_attn_dml_transaction_rollback(dmlH);
            }
            err =  alp_prv_attn_dml_convert_sal_error (err);
            break;
        }

        /* Destroy statement used to insert to ALERTS table */
        err = alp_sal_statement_destroy(stmtH);
        if (err) {
            errMsg = alp_sal_get_error_message((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"statment_destroy: %s", errMsg);
            if (inTransaction) {
                alp_attn_dml_transaction_rollback(dmlH);
            }
            err = alp_prv_attn_dml_convert_sal_error (err);
            break;
        }

        alertP->itemLuid = itemLuid; /* fill in the incore item */

        /* Commit the transaction if everything was successful. */
        if (inTransaction) {
            err = alp_attn_dml_transaction_end(dmlH);
        }
        break;

    case PropertyItem:
        propP = (AlpAttnPropType *) itemH;
        /* Verify record to update exists */
        ALP_ATLX(DML,"validating luid %d", propP->itemLuid);
        err = alp_prv_attn_dml_validate_luid(dmlH, itemType, propP->itemLuid);
        if (err) {
            ALP_ATLX(DML,"validate property luid %d: %x",
                     propP->itemLuid, err);
            return err;
        }

        /* Construct statement to update PROPERTIES table */
        err = alp_sal_statement_initialize((AlpSALHandle)dmlH, &stmtH, 
                                           "UPDATE "
                                           kAttnPropsTableName
                                           " SET "
                                           kAttnPropNameName  " = ?, "
                                           kAttnPropNameType  " = ?, "
                                           kAttnPropNameValue " = ?"
                                           " WHERE "
                                           kAttnPropNamePropLuid " = ?;");
        if (err) {
            errMsg = alp_sal_get_error_message((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"statement init: %s", errMsg);
            err =  alp_prv_attn_dml_convert_sal_error (err);
            break;
        }

        err = alp_sal_statement_set_text(stmtH, -1, propP->name, false);
        err = alp_sal_statement_set_int (stmtH, -1, propP->type);
        err = alp_sal_statement_set_blob(stmtH, -1, propP->value, propP->size, false);

        err = alp_sal_statement_set_int(stmtH, -1, propP->itemLuid);

        /* Execute update to CONTACTS and check for success */
        err = alp_sal_statement_execute(stmtH);
        if (err) {
            errMsg = alp_sal_get_error_message((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"statement exec: %s err %x", errMsg, err);
            err = alp_prv_attn_dml_convert_sal_error (err);
            break;
        }

        /* Destroy statement used to insert to PROPERTIES table */
        err = alp_sal_statement_destroy(stmtH);
        if (err) {
            ALP_ATLX(DML,"destroy statement: %s err %x",
                     alp_sal_get_error_message ((AlpSALHandle) dmlH),
                     err);
            err = alp_prv_attn_dml_convert_sal_error (err);
        }
        break;

    case PropertyListItem:
        propListP = (AlpAttnPropListType *) itemH;
        /* Verify alert record exists for property list */
        if (propListP->source == NULL ||
            propListP->name == NULL ||
            propListP->handle == NULL) {
            ALP_ATLX(DML,"bad parameters:  %s|%s|%s",
                     propListP->source, propListP->name, propListP->handle);
            return ALP_STATUS_ATTN_DML_BAD_PARAM;
        }
        /* check if there is a record matching this item */
        err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &itemLuid, 
                                    "SELECT "
                                    kAttnAlertNameAlertLuid " FROM "
                                    kAttnAlertsTableName " WHERE "
                                    kAttnAlertNameSource " = %Q AND "
                                    kAttnAlertNameName   " = %Q AND "
                                    kAttnAlertNameHandle " = %Q;",
                                    propListP->source,
                                    propListP->name,
                                    propListP->handle);
        
        if (err) {
            /*
             * need to find the error that will be given if the
             * sql doesn't match and return the appropriate ATTN_DML error.
             * (NO_RECORDS) seems appropriate.
             * Was returning _INVALID_LUID but probably need to replace that
             * error code with something else.
             */
            errMsg = alp_sal_get_error_message((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"begin transaction: %s", errMsg);
            return alp_prv_attn_dml_convert_sal_error (err);
        }
        if (itemLuid == 0) {
            ALP_ATLX(DML,"update item id %d invalid", itemLuid);
            return ALP_STATUS_ATTN_DML_INVALID_LUID;
        }

        propListP->itemLuid = itemLuid;

        /* Iterate over the list of properties */
        property = propListP->list;

        /* Try to begin a transcation */
        err = alp_attn_dml_transaction_begin (dmlH, true);
        if (err) {
            errMsg = alp_sal_get_error_message((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"begin transaction: %s", errMsg);
            inTransaction = false;
        }

        while (property) {
            ALP_FAIL_FATAL_IF (property->itemType != PropertyItem,
                               "list item is not a property item: type = %d",
                               property->itemType);

            if (property->name == NULL) {
                /* skip to next item in list */
                ALP_ATLX(DML, "bogus property containing NULL name, skip");
                property = property->next;
                continue;
            }

            property->alertLuid = propListP->itemLuid;

            /* first check if property already in database */
            err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                        "SELECT COUNT(*) FROM "
                                        kAttnPropsTableName " WHERE "
                                        kAttnPropNameAlertLuid " = %d and "
                                        kAttnPropNameName " = %Q;",
                                        property->alertLuid,
                						property->name);
            if (err) {
                ALP_ATLX(DML,"count stmt: err %x", err);
                break;
            }

            if (count) {
                /* property is in the database */
                if (property->value) { 
                    /* if value set update property to the specified value */
                    err = alp_sal_statement_initialize(
                        (AlpSALHandle)dmlH, &stmtH, "UPDATE "
                        kAttnPropsTableName
                        " SET "
                        kAttnPropNameType  " = ?, "
                        kAttnPropNameValue " = ?"
                        " WHERE "
                        kAttnPropNameAlertLuid " = ? AND "
                        kAttnPropNameName " = ?;");
                    if (err) break;
                    err = alp_sal_statement_set_int(stmtH, -1, property->type);
                    if (err) break;
                    err = alp_sal_statement_set_blob(stmtH, -1,
                                                     property->value,
                                                     property->size,
                                                     false);
                    if (err) break;
                    err = alp_sal_statement_set_int(stmtH, -1, property->alertLuid);
                    if (err) break;
                    err = alp_sal_statement_set_text(stmtH, -1, property->name, 0);
                    if (err) break;

                    /* Execute update to CONTACTS and check for success */
                    err = alp_sal_statement_execute(stmtH);
                    if (err) break;

                    /* Destroy statement used to insert to PROPERTIES table */
                    err = alp_sal_statement_destroy(stmtH);
                } else {
                    /* value is NULL so remove the property */
                    err = alp_sal_sql_execute((AlpSALHandle)dmlH, 
                                              "DELETE FROM "
                                              kAttnPropsTableName
                                              " WHERE "
                                              kAttnPropNameAlertLuid
                                              " = %d AND "
                                              kAttnPropNameName " = %Q;",
                                              property->alertLuid,
                                              property->name);
                }
            } else {
                /* not in the database, so add if value is not NULL */
                if (property->value) {
                    err = alp_attn_dml_insert (dmlH, property, NULL);
                    if (err) {
                        if (inTransaction) {
                            alp_attn_dml_transaction_rollback (dmlH);
                        }
                        return err; /* return since err already translated */
                    }
                }
            }
            if (err) break;

            /* go to the next property in the list. */
            property = property->next;
        }
        if (err) {
            errMsg = alp_sal_get_error_message((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"update alert %d prop %s: %s",
                     property ? property->alertLuid : 0,
                     property ? property->name : "n/a",
                     errMsg);
            if (inTransaction) {
                alp_attn_dml_transaction_rollback (dmlH);
            }
            err = alp_prv_attn_dml_convert_sal_error (err);
            break;
        }

        err = alp_sal_string_build (&sql,
                                    "UPDATE " kAttnAlertsTableName
                                    " SET "    kAttnAlertNameUpdated
                                    " = DATETIME( \"NOW\" )"
                                    " WHERE " kAttnAlertNameAlertLuid
                                    " = %d;", propListP->itemLuid);
        err = alp_sal_sql_execute ((AlpSALHandle)dmlH, sql);
        if (err) {
            errMsg = alp_sal_get_error_message((AlpSALHandle)dmlH);
            ALP_ATLX(DML,
                     "update timestamp alert %d: sql [%s] err %x errmsg %s",
                     propListP->itemLuid, sql, err, errMsg);
            if (inTransaction) {
                alp_attn_dml_transaction_rollback (dmlH);
            }
            err = alp_prv_attn_dml_convert_sal_error (err);
            if (sql) alp_sal_string_destroy (sql);
            break;
        }
        if (sql) alp_sal_string_destroy (sql);
        if (inTransaction) {
            err = alp_attn_dml_transaction_end (dmlH);
        }
        break;

    default:
        err = ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    return err;
}

alp_status_t
alp_attn_dml_replace (AlpDmlH     dmlH,
                      AlpDmlItemH itemH)
{
    AlpAttnAlertType    *alertP = (AlpAttnAlertType *) itemH;
    AlpAttnPropType     *propP;
    AlpAttnPropListType *propListP;
    AlpAttnPropType     *property;
    AlpAttnDmlItemType   itemType;
    alp_status_t         err;
    AlpDmlStatementH    *stmtH;
    uint32_t             count = 0;
    AlpLuid              itemLuid;
    bool                 inTransaction = true;
    const char          *errMsg = NULL;
    char                *sql = NULL;

    /* Validate params */
    if (!dmlH || !itemH) {
        ALP_ATLX(DML,"bad parameters: dmlH %x itemH %x",
                 (unsigned int) dmlH,
                 (unsigned int) itemH);
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    itemLuid = 0;
    itemType = alertP->itemType;
    switch (itemType) {
    case AlertItem:
        /* Verify alert record exists */
        if (alertP->source == NULL ||
            alertP->name == NULL ||
            alertP->handle == NULL) {
            ALP_ATLX(DML,"bad parameters:  %s|%s|%s",
                     alertP->source, alertP->name, alertP->handle);
            return ALP_STATUS_ATTN_DML_BAD_PARAM;
        }
        /* check if there is a record matching this item */
        err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &itemLuid, 
                                    "SELECT "
                                    kAttnAlertNameAlertLuid " FROM "
                                    kAttnAlertsTableName " WHERE "
                                    kAttnAlertNameSource " = %Q AND "
                                    kAttnAlertNameName   " = %Q AND "
                                    kAttnAlertNameHandle " = %Q;",
                                    alertP->source,
                                    alertP->name,
                                    alertP->handle);
        if (err) {
            /*
             * need to find the error that will be given if the
             * sql doesn't match and return the appropriate ATTN_DML error.
             * (NO_RECORDS) seems appropriate.
             * Was returning _INVALID_LUID but probably need to replace that
             * error code with something else.
             */
            errMsg = alp_sal_get_error_message((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"find %s,%s,%s failed: %s",
                     alertP->source, alertP->name, alertP->handle, errMsg);
            return alp_prv_attn_dml_convert_sal_error (err);
        }
        if (itemLuid == 0) {
            ALP_ATLX(DML,"replace of %s,%s,%s had invalid luid %d",
                     alertP->source, alertP->name, alertP->handle, itemLuid);
            return ALP_STATUS_ATTN_DML_INVALID_LUID;
        } else {
            ALP_ATLX(DML,"found item %s,%s,%s id %d to replace",
                     alertP->source, alertP->name, alertP->handle,
                     itemLuid);
        }

        /* set the itemLuid to the value of the record being replaced */
        alertP->itemLuid = itemLuid;

        /* Try to begin a transcation */
        ALP_ATLX(DML,"begin transaction (id %d)", alertP->itemLuid);
        err = alp_attn_dml_transaction_begin (dmlH, true);
        ALP_ATLX(DML,"begin transaction err %x", err);
        if (err) {
            errMsg = alp_sal_get_error_message((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"begin transaction: %s: not a problem", errMsg);
            inTransaction = false;
        }

        ALP_ATLX(DML,"initializing replace stmt %x", err);
        /* Construct statement to update ALERTS table */
        err = alp_sal_statement_initialize ((AlpSALHandle)dmlH, &stmtH, 
                                            "UPDATE "
                                            kAttnAlertsTableName
                                            " SET "
                                            kAttnAlertNamePriority   " = ?, "
                                            kAttnAlertNameDuration   " = ?, "
                                            kAttnAlertNamePending    " = ?, "
                                            kAttnAlertNameAttributes " = ?, "
                                            kAttnAlertNameUserId     " = ?, "
                                            kAttnAlertNameGroupId    " = ?, "
                                            kAttnAlertNameCreated
                                            " = DATETIME( \"NOW\" ), "
                                            kAttnAlertNameUpdated
                                            " = DATETIME( \"NOW\" )"
                                            " WHERE "
                                            kAttnAlertNameSource " = ? AND "
                                            kAttnAlertNameName   " = ? AND "
                                            kAttnAlertNameHandle " = ?;");
        ALP_ATLX(DML,"initialized replaced %x", err);
        if (err) {
            ALP_ATLX(DML,"statement_init: err %x %s", err,
                     alp_sal_get_error_message((AlpSALHandle)dmlH));
            if (inTransaction) {
                alp_attn_dml_transaction_rollback(dmlH);
            }
            err =  alp_prv_attn_dml_convert_sal_error (err);
            break;
        }

        /* items being replaced */
        err = alp_sal_statement_set_int (stmtH, -1, alertP->priority);
        err = alp_sal_statement_set_int (stmtH, -1, alertP->duration);
        err = alp_sal_statement_set_int (stmtH, -1, alertP->pending);
        err = alp_sal_statement_set_int (stmtH, -1, alertP->attributes);
        err = alp_sal_statement_set_int (stmtH, -1, alertP->userId);
        err = alp_sal_statement_set_int (stmtH, -1, alertP->groupId);
        /* for the record that matches this tuple */
        err = alp_sal_statement_set_text(stmtH, -1, alertP->source, 0);
        err = alp_sal_statement_set_text(stmtH, -1, alertP->name, 0);
        err = alp_sal_statement_set_text(stmtH, -1, alertP->handle, 0);

        /* Execute update to ALERT and check for success */
        err = alp_sal_statement_execute(stmtH);
        if (err) {
            ALP_ATLX(DML,"statment_exec: %s",
                     alp_sal_get_error_message((AlpSALHandle)dmlH));
            if (inTransaction) {
                alp_attn_dml_transaction_rollback(dmlH);
            }
            err =  alp_prv_attn_dml_convert_sal_error (err);
            break;
        }

        /* Destroy statement used to update the ALERTS table */
        err = alp_sal_statement_destroy(stmtH);
        if (err) {
            errMsg = alp_sal_get_error_message((AlpSALHandle)dmlH);
            ALP_ATLX(DML,"statment_destroy: %s", errMsg);
            if (inTransaction) {
                alp_attn_dml_transaction_rollback(dmlH);
            }
            err = alp_prv_attn_dml_convert_sal_error (err);
            break;
        }

        alertP->itemLuid = itemLuid; /* fill in the incore item */

        /* remove old properties associated with alert */
        err = alp_sal_sql_execute((AlpSALHandle)dmlH, 
                                  "DELETE FROM "
                                  kAttnPropsTableName
                                  " WHERE "
                                  kAttnPropNameAlertLuid " = %d;",
                                  itemLuid);

        ALP_ATLX (DML,"deleting old properties with alert luid %d err %x",
                  itemLuid, err);

        /* write property list out to PROPERTY Table if attached */
        propP = alertP->plist;
        while (propP) {
            ALP_FAIL_FATAL_IF (propP->itemType != PropertyItem,
                               "list item is not a property item: type = %d",
                               propP->itemType);
            propP->alertLuid = itemLuid;
            alp_attn_dml_insert (dmlH, propP, NULL);
            propP = propP->next;
        }

        /* Commit the transaction if everything was successful. */
        if (inTransaction) {
            err = alp_attn_dml_transaction_end(dmlH);
        }
        break;

    default:
        err = ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    return err;
}

alp_status_t
alp_attn_dml_delete (AlpDmlH     dmlH,
                     AlpDmlItemH itemH)
{
    AlpAttnAlertType   *alertP = (AlpAttnAlertType *) itemH;
    AlpAttnPropType    *propP;
    alp_status_t        err;
    AlpDmlStatementH   *stmtH;
    AlpAttnDmlItemType  itemType;
    AlpLuid             itemLuid;
    bool                inTransaction = true;
    const char         *errMsg = NULL;

    /* Validate params */
    if (!dmlH || !itemH) {
        ALP_ATLX(DML,"bad parameters: dml %x itemH %x",
                 dmlH, itemH);
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    itemType = alertP->itemType;

    /* verify item is valid record */
    itemLuid = 0;
    switch (itemType) {
    case AlertItem:
        /* check parameters for alert item */
        if (alertP->source == NULL ||
            alertP->name == NULL ||
            alertP->handle == NULL) {
            ALP_ATLX (DML, "bad parameters: alert %s|%s|%s",
                      alertP->source, alertP->name, alertP->handle);
            return ALP_STATUS_ATTN_DML_BAD_PARAM;
        }
        err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &itemLuid,
                                    "SELECT " kAttnAlertNameAlertLuid " FROM "
                                    kAttnAlertsTableName
                                    " WHERE "
                                    kAttnAlertNameSource " = %Q AND "
                                    kAttnAlertNameName   " = %Q AND "
                                    kAttnAlertNameHandle " = %Q;",
                                    alertP->source,
                                    alertP->name,
                                    alertP->handle);
        if (err) {
            ALP_ATLX(DML, "didn't find item %s|%s|%s msg %s err %x",
                     alertP->source, alertP->name, alertP->handle,
                     alp_sal_get_error_message ((AlpSALHandle)dmlH), err);
        }
        break;

    case PropertyItem:
        propP = (AlpAttnPropType *) itemH;
        if (propP->alertLuid == 0 || propP->name == NULL) {
            ALP_ATLX (DML, "bad parameters: prop alert id %d name %s",
                      propP->alertLuid, propP->name);
            return ALP_STATUS_ATTN_DML_BAD_PARAM;
        }
        err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &itemLuid, 
                                    "SELECT " kAttnPropNamePropLuid " FROM "
                                    kAttnPropsTableName
                                    " WHERE "
                                    kAttnPropNameAlertLuid " = %d AND "
                                    kAttnPropNameName      " = %Q;", 
                                    propP->alertLuid,
                                    propP->name);
        break;

    default:
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }
    if (err) {
        /* TODO: fix errcode for no records */
        ALP_ATLX(DML, "delete alert item id %d msg %s err %x", itemLuid,
                 alp_sal_get_error_message ((AlpSALHandle)dmlH), err);
        return alp_prv_attn_dml_convert_sal_error (err);
    }
    if (itemLuid == 0) {
        ALP_ATLX(DML,"delete item id %d invalid", itemLuid);
        return ALP_STATUS_ATTN_DML_INVALID_LUID;
    }

    switch (itemType) {
    case AlertItem:
        /* Open a transaction */
        err = alp_attn_dml_transaction_begin(dmlH, true);
        if (err) {
            ALP_ATLX(DML,"begin trans failed err %x msg %s",
                     err, alp_sal_get_error_message ((AlpSALHandle)dmlH));
           /* XXX Need to block */
            inTransaction = false;
        }
        /* Delete the alert */
        err = alp_sal_sql_execute((AlpSALHandle)dmlH, 
                                  "DELETE FROM "
                                  kAttnAlertsTableName
                                  " WHERE "
                                  kAttnAlertNameAlertLuid " = %d;",
                                  itemLuid);
        if (err) {
            ALP_ATLX(DML,"delete alert %s|%s|%s[%d] sql failed with err %x",
                     alertP->source, alertP->name, alertP->handle,
                     itemLuid, err);
            err = alp_prv_attn_dml_convert_sal_error (err);
            if (inTransaction) {
                alp_attn_dml_transaction_rollback(dmlH);
            }
            return err;
        }

        /* Delete all properties associated with alert */
        err = alp_sal_sql_execute((AlpSALHandle)dmlH, 
                                  "DELETE FROM "
                                  kAttnPropsTableName
                                  " WHERE "
                                  kAttnPropNameAlertLuid
                                  " = %d;",
                                  itemLuid);
        if (err) {
            ALP_ATLX(DML,
                     "delete properties for alert %d sql failed: %s err %x",
                     itemLuid,
                     alp_sal_get_error_message ((AlpSALHandle)dmlH),
                     err);
            err = alp_prv_attn_dml_convert_sal_error (err);
            if (inTransaction) {
                alp_attn_dml_transaction_rollback(dmlH);
            }
            return err;
        }

        /* End transaction */
        if (inTransaction) {
            err = alp_attn_dml_transaction_end(dmlH);
        }
        break;

    case PropertyItem:
        /* Delete property */
        err = alp_sal_sql_execute((AlpSALHandle)dmlH, 
                                  "DELETE FROM "
                                  kAttnPropsTableName
                                  " WHERE "
                                  kAttnPropNamePropLuid
                                  " = %d;",
                                  itemLuid);
        if (err) {
            return alp_prv_attn_dml_convert_sal_error (err);
        }
        break;

    default:
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    return ALP_STATUS_OK;
}

alp_status_t
alp_attn_dml_find (AlpDmlH      dmlH,
                   AlpDmlItemH  itemH,
                   AlpLuid     *itemLuid)
{
    alp_status_t         err;
    AlpDmlStatementH    *stmtH;
    AlpAttnAlertType    *alertP   = (AlpAttnAlertType *)itemH;
    AlpAttnPropType     *propP;
    AlpAttnPropListType *propListP;
    uint32_t             count = 0;
    bool                 inTransaction = true;
    AlpLuid              luid = 0;
    const char          *errMsg = NULL;
    char                *whereClause = NULL;

    /* Validate params */
    if (!dmlH || !itemH || !itemLuid) {
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    /* Validate itemLuid */
    switch (alertP->itemType) {
    case AlertItem:
        err = alp_sal_string_build (
            &whereClause, " WHERE %s = %Q AND %s = %Q AND %s = %Q;", 
            kAttnAlertNameSource, alertP->source,
            kAttnAlertNameName,   alertP->name,
            kAttnAlertNameHandle, alertP->handle);

        if (whereClause) {
            err = alp_sal_sql_get_value(
                (AlpSALHandle)dmlH, &count,
                "SELECT COUNT(*) FROM " kAttnAlertsTableName "%s;",
                whereClause);
            if (err) {
                ALP_ATLX(DML, "select = <%s> where = <%s> err = %x",
                         "SELECT COUNT(*) FROM " kAttnAlertsTableName "%s;",
                         whereClause, err);
            }
        }
        if (count == 1) {
            err = alp_sal_sql_get_value(
                (AlpSALHandle)dmlH, &luid,
                "SELECT " kAttnAlertNameAlertLuid
                " FROM " kAttnAlertsTableName "%s;",
                whereClause);
            if (err) {
                ALP_ATLX(DML, "select = <%s> where = <%s> err = %x",
                         "SELECT " kAttnAlertNameAlertLuid
                         " FROM " kAttnAlertsTableName "%s;",
                         whereClause, err);
            }
        }
        break;

    case PropertyItem:
        propP = (AlpAttnPropType *) itemH;
        err = alp_sal_string_build (
            &whereClause, " WHERE %s = %d AND %s = %Q;",
            kAttnPropNameAlertLuid, propP->alertLuid,
            kAttnPropNameName,      propP->name);
        if (whereClause) {
            err = alp_sal_sql_get_value(
                (AlpSALHandle)dmlH, &count, 
                "SELECT COUNT(*) FROM " kAttnPropsTableName "%s;",
                whereClause);
            if (err) {
                ALP_ATLX(DML, "select = <%s> where = <%s> err = %x",
                         "SELECT COUNT(*) FROM " kAttnPropsTableName "%s;",
                         whereClause, err);
            }
        }
        if (count == 1) {
            err = alp_sal_sql_get_value(
                (AlpSALHandle)dmlH, &luid, 
                "SELECT " kAttnPropNamePropLuid
                " FROM " kAttnPropsTableName "%s;",
                whereClause);
            if (err) {
                ALP_ATLX(DML, "select = <%s> where = <%s> err = %x",
                         "SELECT " kAttnPropNamePropLuid
                         " FROM " kAttnPropsTableName "%s;",
                         whereClause, err);
            }
        }
        break;

    case PropertyListItem:
        propListP = (AlpAttnPropListType *) itemH;
        err = alp_sal_string_build (
            &whereClause, " WHERE %s = %Q AND %s = %Q AND %s = %Q;", 
            kAttnAlertNameSource, propListP->source,
            kAttnAlertNameName,   propListP->name,
            kAttnAlertNameHandle, propListP->handle);

        if (whereClause) {
            err = alp_sal_sql_get_value(
                (AlpSALHandle)dmlH, &count,
                "SELECT COUNT(*) FROM " kAttnAlertsTableName "%s;",
                whereClause);
            if (err) {
                ALP_ATLX(DML, "select = <%s> where = <%s> err = %x",
                         "SELECT COUNT(*) FROM " kAttnAlertsTableName "%s;",
                         whereClause, err);
            }
        }
        if (count == 1) {
            err = alp_sal_sql_get_value(
                (AlpSALHandle)dmlH, &luid,
                "SELECT " kAttnAlertNameAlertLuid
                " FROM " kAttnAlertsTableName "%s;",
                whereClause);
            if (err) {
                ALP_ATLX(DML, "select = <%s> where = <%s> err = %x",
                         "SELECT " kAttnAlertNameAlertLuid
                         " FROM " kAttnAlertsTableName "%s;",
                         whereClause, err);
            }
        }
        break;

    default:
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }
    ALP_ATLX(DML, "err = %x count = %d itemLuid = %d", err, count, luid);

    if (whereClause) alp_sal_string_destroy (whereClause);

    if (err) {
        err = alp_prv_attn_dml_convert_sal_error (err);
    } else {
        *itemLuid = luid;
    }

    return err;
}
