/********************************************************************

Copyright  2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/**
 * @file attn_dml_tests.c
 *
 * @brief ALP attention manager DML unit tests
 *
 ******************************************************************************/

#include <stdlib.h>
#include <time.h>

#include <glib.h>  /* used for FALSE, TRUE */
#include <glib-object.h>

#include <hiker/syspaths.h>
#include <hiker/types.h>
#include <hiker/unit_test.h>

#include <hiker/prv/attnmgr_traces.h>
#include <hiker/prv/attnmgr_alert.h>
#include <hiker/prv/attnmgr_dml.h>

#include "attn_dml_prv.h"

#ifndef ALP_ATTN_TRACE_SQL
#define ALP_ATTN_TRACE_SQL   true   /* set true to trace SQL for debug */
#endif
#ifndef ALP_ATTN_DML_SAVE_DB
#define ALP_ATTN_DML_SAVE_DB false  /* set true to save db for debug */
#endif


#define STRINGIFY(x)  #x
#define TOSTRING(x)   STRINGIFY(x)

#define ARRAY_SIZE(x)  (sizeof(x)/sizeof((x)[0]))


class AttnDMLTest: public CPPUNIT_NS::TestFixture
{
    CPPUNIT_TEST_SUITE( AttnDMLTest );

    CPPUNIT_TEST( create_destroy_items );
    CPPUNIT_TEST( set_int32 );
    CPPUNIT_TEST( set_string );
    CPPUNIT_TEST( open_close_db );
    CPPUNIT_TEST( insert_alert_item );
    CPPUNIT_TEST( update_alert_item );
    CPPUNIT_TEST( replace_alert_item );
    CPPUNIT_TEST( delete_alert_items );
    CPPUNIT_TEST( insert_prop_item );
    CPPUNIT_TEST( update_prop_item );
    CPPUNIT_TEST( update_prop_list );
    CPPUNIT_TEST( delete_prop_items );
    CPPUNIT_TEST( add_get_property );
    CPPUNIT_TEST( insert_compound_item );
    
    CPPUNIT_TEST_SUITE_END();

public:
    void setUp();         // overriding TestFixture method
    void tearDown();      // overriding TestFixture method

protected:
    void create_destroy_items();
    void set_int32();
    void set_string();
    void open_close_db();
    void insert_alert_item();
    void update_alert_item();
    void replace_alert_item();
    void delete_alert_items();
    void insert_prop_item();
    void update_prop_item();
    void update_prop_list();
    void delete_prop_items();
    void add_get_property();
    void insert_compound_item();

    void initialize();

private:
    AlpDmlH          dmlH;
    AlpDmlStatementH stmtH;
    gchar           *db_pathname;
};

ALP_ROOT_TEST_SUITE_REGISTRATION( "UI-007", CPPUNIT_REGISTRY_TYPE_AUTO );
CPPUNIT_TEST_SUITE_NAMED_REGISTRATION( AttnDMLTest, "UI-007" );

// local globals
static bool sInitialized = false;

// local helper functions

#if ALP_ATTN_TRACE_SQL == true
void
attn_sql_trace (void *data, const char *sql)
{
    ALP_ATLX(DML,"<%s>", sql);
}
#endif

int32_t
Rand (int32_t min, int32_t max)
{
    int32_t next;

    next = rand() % (max - min + 1) + min;

    return next;
}

typedef struct _RandomString RandomString;
struct _RandomString {
    char *string;
    bool  used;
};


RandomString company_strings[] = { { "com.access.apps", false },
                                   { "com.palm.apps", false },
                                   { "com.foo.apps", false },
                                   { "net.homeip.brainey.apps", false }
};
#define NUM_COMPANY_STRINGS (ARRAY_SIZE(company_strings))

RandomString application_strings[] = { { "phone", false },
                                       { "calendar", false },
                                       { "contacts", false },
                                       { "tasks", false },
                                       { "memo", false },
                                       { "browser", false },
                                       { "duck", false },
                                       { "bunny", false },
                                       { "sheep", false },
                                       { "mfp", false },
                                       { "radio", false },
                                       { "tv", false },
                                       { "thumbscan", false },
                                       { "email", false },
                                       { "im", false },
                                       { "web", false },
                                       { "aim", false },
                                       { "irc", false },
                                       { "sms", false }
};
#define NUM_APP_STRINGS (ARRAY_SIZE(application_strings))

RandomString alert_strings[] = { { "incoming-call", false },
                                 { "alarm", false },
                                 { "missed-call", false },
                                 { "chat-request", false }
};
#define NUM_ALERT_STRINGS (ARRAY_SIZE(alert_strings))

RandomString handle_strings[] = { { "14085551781", false },
                                  { "123145", false },
                                  { "15555551234", false },
                                  { "122134@person", false }
};
#define NUM_HANDLE_STRINGS (ARRAY_SIZE(handle_strings))

RandomString property_strings[] = { { "icon", false },
                                    { "title", false },
                                    { "image", false },
                                    { "sound", false },
                                    { "text-1", false },
                                    { "text-2", false },
                                    { "text-3", false },
                                    { "text-4", false },
                                    { "text-5", false },
                                    { "buttons", false },
                                    { "button-ok", false },
                                    { "button-answer", false },
                                    { "button-ignore", false },
                                    { "button-goto", false },
                                    { "button-snooze", false },
                                    { "action-ok", false },
                                    { "action-answer", false },
                                    { "action-ignore", false },
                                    { "action-goto", false },
                                    { "action-snooze", false }
};
#define NUM_PROPERTY_STRINGS (ARRAY_SIZE(property_strings))

struct RandomStrings {
    RandomString *strings;
    unsigned      count;
} ListOfRandomStrings[] = {
    { company_strings, NUM_COMPANY_STRINGS },
    { application_strings, NUM_APP_STRINGS },
    { alert_strings, NUM_ALERT_STRINGS },
    { handle_strings, NUM_HANDLE_STRINGS },
    { property_strings, NUM_PROPERTY_STRINGS }
};
#define NUM_RANDOM_STRINGS (ARRAY_SIZE(ListOfRandomStrings))

char *
GetRandomString (RandomString *strings, int count)
{
    int index;
    int sindex;
    int max;
    RandomString *rstrP;

    max    = count - 1;
    index  = Rand(0, max);
    rstrP  = &strings[index];
    sindex = index;

    while (rstrP->used) {
        sindex = (sindex + 1) % count;
        if (sindex == index) /* reached original try, time to give up */
            return "all-strings-used";
        rstrP = &strings[sindex];
    }
    rstrP->used = TRUE;

    return rstrP->string;
}

void
ResetRandomStrings(void)
{
    unsigned int group;
    unsigned int item;
    unsigned int count;
    RandomString *rstrP;

    for (group = 0; group < NUM_RANDOM_STRINGS; group++) {
        rstrP = ListOfRandomStrings[group].strings;
        count = ListOfRandomStrings[group].count;
        for (item = 0; item < count; item++) {
            rstrP[item].used = false;
        }
    }
}

char *
RandomSource()
{
    static char buff[64];
    sprintf (buff, "%s.%s",
             GetRandomString(company_strings, NUM_COMPANY_STRINGS),
             GetRandomString(application_strings, NUM_APP_STRINGS));

    return buff;
}

char *
RandomName()
{
    return GetRandomString(alert_strings, NUM_ALERT_STRINGS);
}

char *
RandomHandle()
{
    return GetRandomString(handle_strings, NUM_HANDLE_STRINGS);
}

int32_t
RandomPriority()
{
    return (Rand(0,5) + 1);
}

int32_t
RandomDuration()
{
    return Rand(0,123456);
}

uint32_t
RandomAttributes()
{
    return Rand(0,0x7fffffff);
}

char *
RandomPropName()
{
    return GetRandomString(property_strings, NUM_PROPERTY_STRINGS);

}

char *
RandomPropValue()
{
    static char *values[] = { "Incoming call",
                              "Call from John Smith(W)",
                              "gtk-stock-cancel",
                              "bar:com.access.apps.xyz/rsc/image1",
                              "gtk-stock-okay",
                              "alp-stock-snooze",
                              "alp-stock-goto",
                              "3",
                              "bar:com.access.apps.phone/rsc/incoming-call.glade",
                              "$source/rsc/missed-call.glade"};
    return values[Rand(0,ARRAY_SIZE(values)-1)];
}

/**
 * @brief Print out a timestamp with a label. 
 */
void
PrintDate(FILE *out, char *label, int s, int indent)
{
    int i;

    if (!label) {
        return;
    }

    for (i=0; i < indent; i++) {
        fprintf(out, "\t");
    }

    fprintf(out, "%s: %d\n", label, s);
}


/**
 * @brief Print out an integer with a label
 */
void
PrintInt(FILE *out, char *label, int s, int indent)
{
    int i;

    if (!label) {
        return;
    }

    for (i=0;i<indent;i++) {
        fprintf(out, "\t");
    }

    fprintf(out, "%s: %d\n", label, s);
}


/**
 * @brief Print out an integer in hex with a label
 */
void
PrintHex(FILE *out, char *label, int s, int indent)
{
    int i;

    if (!label) {
        return;
    }

    for (i=0;i<indent;i++) {
        fprintf(out, "\t");
    }

    fprintf(out, "%s: %x\n", label, s);
}


/**
 * @brief Print out a boolean with a label
 */
void
PrintBoolean(FILE *out, char *label, bool b, int indent)
{
    int i;

    if (!label) {
        return;
    }

    for (i=0;i<indent;i++) {
        fprintf(out, "\t");
    }

    fprintf(out, "%s: %s\n", label, b ? "true" : "false");
}

/**
 * @brief Print out a string field with a label
 */
void
PrintString(FILE *out, char *label, char *s, int indent)
{
    int i;

    if (!label || !s) {
        return;
    }

    for (i=0;i<indent;i++) {
        fprintf(out, "\t");
    }

    fprintf(out, "%s: %s\n", label, s);
}

/**
 * @brief Print out a string field with a label
 */
void
PrintProperty(FILE *out, char *label, GType type, void *s, uint32_t size, int indent)
{
    int i;

    if (!label || !s) {
        return;
    }

    for (i=0;i<indent;i++) {
        fprintf(out, "\t");
    }

    fprintf(out, "%s: Type=%d Size=%d:", label, type, size);
    switch (type) {
    case G_TYPE_STRING:
        fprintf(out, "%*s\n", size, (char *)s);
        break;
    default:
        fprintf(out, "\n");
        break;
    }
}

/**
 * @brief Print out the fields of an item with labels
 */
void
PrintItem (FILE *out, AlpDmlItemH itemH)
{
    AlpAttnAlertType *alertP = (AlpAttnAlertType *) itemH;
    AlpAttnPropType  *propP  = (AlpAttnPropType  *) itemH;
    alp_status_t      err;

    if (itemH == NULL) {
        fprintf(out, "\tInvalid Item Handle:=============\n");
        return;
    }

    switch (alertP->itemType) {
    case AlertItem:
        fprintf(out, "\tALERT:==========================\n");
        PrintInt(out, "Alert Luid", alertP->itemLuid, 1);
        PrintString(out, "Source", alertP->source, 1);
        PrintString(out, "Name", alertP->name, 1);
        PrintString(out, "Handle", alertP->handle, 1);
        PrintInt(out, "Priority", alertP->priority, 1);
        PrintInt(out, "Duration", alertP->duration, 1);
        PrintInt(out, "User Id", alertP->userId, 1);
        PrintInt(out, "Group Id", alertP->groupId, 1);
        PrintBoolean(out, "Pending", alertP->pending, 1); 
        PrintHex(out, "Attributes", alertP->attributes, 1); 
        PrintDate(out, "Created", alertP->created, 1);
        PrintDate(out, "Updated", alertP->updated, 1);
        break;

    case PropertyItem:
        fprintf(out, "\tProperty:========================\n");
        PrintInt(out, "Prop Luid", propP->itemLuid, 1);
        PrintInt(out, "Alert Luid", propP->alertLuid, 1);
        PrintString(out, "Name", propP->name, 1);
        PrintProperty(out, "Value", propP->type, propP->value, propP->size, 1);
        break;

    default:
        fprintf(out, "\tUnknown item type %d:===========\n", alertP->itemType);
        break;
    }
}

/**
 * @brief create a new data model item as specified by itemType
 *
 * @param [in]  itemType the attention dml item to create
 * @param [out] itemH    a pointer to store the handle of the created item
 *
 * @return alp_status_t
 */
alp_status_t
NewItem (AlpAttnDmlItemType itemType, AlpDmlItemH *itemH)
{
    alp_status_t      err;
    AlpAttnAlertType *alertP = NULL;
    AlpAttnPropType  *propP  = NULL;
    GType             type;
    int32_t           value;
    uint32_t          valueu;
    char             *str;
    
    err = alp_attn_dml_item_create (itemType, itemH);
    if (err) {
        ALP_ATLX(DML,"%d: failed to create item type %d err %x",
                 __LINE__, itemType, err);
        return err;
    }

    /* Fill in item */
    switch (itemType) {
    case AlertItem:
        alertP = (AlpAttnAlertType *) (*itemH);

        /* Set values */
        err = alp_attn_dml_item_set_string (alertP, Source, RandomSource());
        if (err) {
        ALP_ATLX(DML,"%d: failed to create item type %d err %x",
                 __LINE__, itemType, err);
            return err;
        }
        err = alp_attn_dml_item_set_string (alertP, Name, RandomName());
        if (err) {
        ALP_ATLX(DML,"%d: failed to create item type %d err %x",
                 __LINE__, itemType, err);
            return err;
        }
        err = alp_attn_dml_item_set_string (alertP, Handle, RandomHandle());
        if (err) {
        ALP_ATLX(DML,"%d: failed to create item type %d err %x",
                 __LINE__, itemType, err);
            return err;
        }
        value = RandomPriority();
        err = alp_attn_dml_item_set_int32 (alertP, Priority, &value);
        if (err) {
        ALP_ATLX(DML,"%d: failed to create item type %d err %x",
                 __LINE__, itemType, err);
            return err;
        }
        value = RandomDuration();
        err = alp_attn_dml_item_set_int32 (alertP, Duration, &value);
        if (err) {
        ALP_ATLX(DML,"%d: failed to create item type %d err %x",
                 __LINE__, itemType, err);
            return err;
        }
        value = true;
        err = alp_attn_dml_item_set_int32 (alertP, Pending, &value);
        if (err) {
            ALP_ATLX(DML,"%d: failed to create item type %d err %x",
                     __LINE__, itemType, err);
            return err;
        }
        valueu = RandomAttributes();
        err = alp_attn_dml_item_set_uint32 (alertP, Attributes, &valueu);
        if (err) {
        ALP_ATLX(DML, "%d: failed to create item type %d err %x",
                 __LINE__, itemType, err);
            return err;
        }
        break;

    case PropertyItem:
        propP = (AlpAttnPropType *) (*itemH);
        str = RandomPropName();
        err = alp_attn_dml_item_set_string (propP, Name, str);
        if (err) {
        ALP_ATLX(DML,"%d: failed to create item type %d err %x",
                 __LINE__, itemType, err);
            return err;
        }
        type = G_TYPE_STRING;
        err = alp_attn_dml_item_set_uint32 (propP, PropType,
                                            (uint32_t *)&type);
        if (err) {
            ALP_ATLX(DML,"failed to set prop type to %x err %x",
                     type, err);
            return err;
        }
        ALP_ATLX(DML,"type set to %x", propP->type);
        str = RandomPropValue();
        if (str) {
            ALP_ATLX(DML,
                     TOSTRING(__LINE__)
                     ": create prop item type %x str %s",
                     type, str);
            err = alp_attn_dml_item_set_blob (propP, Value, str, strlen(str)+1);
            ALP_ATLX(DML,
                     TOSTRING(__LINE__)
                     ": blob added type %x str %s size %d  err %x",
                     type, str, strlen(str)+1, err);
        } else {
            ALP_ATLX(DML, "RandomPropValue() returned NULL");
        }
        ALP_ATLX(DML,"type %x size %d value %x",
                 propP->type, propP->size, propP->value);
        if (err) {
            ALP_ATLX(DML,
                     TOSTRING(__LINE__) ": failed to create item type %d err %x",
                     itemType, err);
            return err;
        }
        break;

    default:
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    return ALP_STATUS_OK;
}

static alp_status_t
DupItem (AlpDmlItemH  itemH, AlpDmlItemH *dupH)
{
    alp_status_t        err;
    AlpAttnDmlItemType  itemType;
    AlpAttnBaseType    *itemP = (AlpAttnBaseType *) itemH;
    AlpAttnAlertType   *alertSrcP = NULL;
    AlpAttnAlertType   *alertDestP = NULL;
    AlpAttnPropType    *propSrcP  = NULL;
    AlpAttnPropType    *propDestP  = NULL;
    GType               type;
    int32_t             value;
    uint32_t            valueu;
    uint32_t            size;
    char               *str;
    void               *data;
    
    if (dupH == NULL) return ALP_STATUS_ATTN_DML_BAD_PARAM;

    itemType = itemP->itemType;
    err = alp_attn_dml_item_create (itemType, dupH);
    if (err) {
        ALP_ATLX(DML,"%d: failed to create item type %d err %x",
                 __LINE__, itemType, err);
        return err;
    }

    /* Fill in item */
    switch (itemType) {
    case AlertItem:
        alertSrcP  = (AlpAttnAlertType *) itemH;
        alertDestP = (AlpAttnAlertType *) (*dupH);

        /* Set values */
        err = alp_attn_dml_item_set_string (alertDestP, Source, alertSrcP->source);
        if (err) {
        ALP_ATLX(DML,"%d: failed to create item type %d err %x",
                 __LINE__, itemType, err);
            return err;
        }
        err = alp_attn_dml_item_set_string (alertDestP, Name, alertSrcP->name);
        if (err) {
        ALP_ATLX(DML,"%d: failed to create item type %d err %x",
                 __LINE__, itemType, err);
            return err;
        }
        err = alp_attn_dml_item_set_string (alertDestP, Handle, alertSrcP->handle);
        if (err) {
        ALP_ATLX(DML,"%d: failed to create item type %d err %x",
                 __LINE__, itemType, err);
            return err;
        }

        value = alertSrcP->priority;
        err = alp_attn_dml_item_set_int32 (alertDestP, Priority, &value);
        if (err) {
        ALP_ATLX(DML,"%d: failed to create item type %d err %x",
                 __LINE__, itemType, err);
            return err;
        }
        value = alertSrcP->duration;
        err = alp_attn_dml_item_set_int32 (alertDestP, Duration, &value);
        if (err) {
        ALP_ATLX(DML,"%d: failed to create item type %d err %x",
                 __LINE__, itemType, err);
            return err;
        }
        value = alertSrcP->pending;
        err = alp_attn_dml_item_set_int32 (alertDestP, Pending, &value);
        if (err) {
        ALP_ATLX(DML,"%d: failed to create item type %d err %x",
                 __LINE__, itemType, err);
            return err;
        }
        valueu = alertSrcP->attributes;
        err = alp_attn_dml_item_set_uint32 (alertDestP, Attributes, &valueu);
        if (err) {
        ALP_ATLX(DML, "%d: failed to create item type %d err %x",
                 __LINE__, itemType, err);
            return err;
        }
        break;

    case PropertyItem:
        propSrcP  = (AlpAttnPropType *) itemH;
        propDestP = (AlpAttnPropType *) (*dupH);

        err = alp_attn_dml_item_set_string (propDestP, Name, propSrcP->name);
        if (err) {
        ALP_ATLX(DML,"%d: failed to create item type %d err %x",
                 __LINE__, itemType, err);
            return err;
        }
        type = propSrcP->type;
        err = alp_attn_dml_item_set_uint32 (propDestP, PropType,
                                            (uint32_t *)&type);
        if (err) {
            ALP_ATLX(DML,"failed to set prop type to %x err %x",
                     type, err);
            return err;
        }

        size = propSrcP->size;
        ALP_ATLX(DML,"size set to %x", propDestP->size);

        data = propSrcP->value;
        if (data && size) {
            ALP_ATLX(DML,
                     TOSTRING(__LINE__)
                     ": create prop item type %x str %s",
                     type, str);
            data = g_memdup (propSrcP->value, size);
            err = alp_attn_dml_item_set_blob (propDestP, Value, data, size);
            ALP_ATLX(DML,
                     TOSTRING(__LINE__)
                     ": blob added type %x str %s size %d  err %x",
                     type, str, strlen(str)+1, err);
        } else {
            ALP_ATLX(DML,
                     "either data pointer is null (%x) or length is zero (%d)",
                     data, size);
            err = alp_attn_dml_item_set_blob (propDestP, Value, NULL, 0);
        }
        ALP_ATLX(DML,"name %s type %x size %d value %x",
                 propDestP->name, propDestP->type,
                 propDestP->size, propDestP->value);
        if (err) {
            ALP_ATLX(DML,
                     TOSTRING(__LINE__) ": failed to create item type %d err %x",
                     itemType, err);
            return err;
        }
        break;

    default:
        return ALP_STATUS_ATTN_DML_BAD_PARAM;
    }

    return ALP_STATUS_OK;
}


/**
 * bool MyStrCmp(char *a, char *b)
 */
bool
MyStrCmp (char *a, char *b)
{
    if (!a && b) {
        return false;
    }
    if (a && !b) {
        return false;
    }
    if (!a && !b) {
        return true;
    }

    return !strcmp(a, b);
}

/**
 * bool CompareItems(AlpDmlItemH aH, AlpDmlStatementH bH)
 * True if a == b
 */
bool
CompareItems (AlpDmlItemH itemH, AlpDmlStatementH stmtH)
{
    bool              match  = true;
    AlpAttnAlertType *alertP = (AlpAttnAlertType *) itemH;
    AlpAttnPropType  *propP  = (AlpAttnPropType  *) itemH;
    char             *s;
    int32_t           value;
    int64_t           value64;
    uint32_t          valueu;
    bool              valueb;
    void             *pvalue;
    uint32_t          ptype;
    uint32_t          psize;
    AlpLuid           alertLuid;
    AlpLuid           propLuid;
    alp_status_t      err;

    if (!itemH && !stmtH) {
        return true;
    }

    ALP_ATLX(DML,"itemType %d", alertP->itemType);
    switch (alertP->itemType) {
    case AlertItem:
        /* Compare AlertLuids */
        alertLuid = 0;
        err = alp_attn_dml_statement_get_luid (stmtH, &alertLuid);
        ALP_ATLX(DML, "alertLuid %d == %d",
                 alertP->itemLuid, alertLuid);
        if (alertP->itemLuid != alertLuid) return false;
        
        /* Validate alert source */
        err = alp_attn_dml_statement_get_string (stmtH, Source, &s);
        ALP_ATLX(DML,"source <%s> == <%s>", alertP->source, s);
        if (!MyStrCmp(alertP->source, s)) {
            free (s);
            return false;
        }
        free (s);

        /* Validate alert name */
        err = alp_attn_dml_statement_get_string (stmtH, Name, &s);
        ALP_ATLX(DML,"name <%s> == <%s>", alertP->name, s);
        if (!MyStrCmp(alertP->name, s)) {
            free (s);
            return false;
        }
        free (s);

        /* Validate alert handle */
        err = alp_attn_dml_statement_get_string (stmtH, Handle, &s);
        ALP_ATLX(DML,"handle <%s> == <%s>", alertP->handle, s);
        if (!MyStrCmp(alertP->handle, s)) {
            free (s);
            return false;
        }
        free (s);

        /* Validate priority */
        err = alp_attn_dml_statement_get_int32 (stmtH, Priority, &value);
        ALP_ATLX(DML,"priority %d == %d", alertP->priority, value);
        if (alertP->priority != value) return false;

        /* Validate duration */
        err = alp_attn_dml_statement_get_int32 (stmtH, Duration, &value);
        ALP_ATLX(DML,"duration %d == %d", alertP->duration, value);
        if (alertP->duration != value) return false;

        /* Validate pending */
        err = alp_attn_dml_statement_get_bool (stmtH, Pending, &valueb);
        ALP_ATLX(DML,"pending %d == %d", alertP->pending, valueb);
        if (alertP->pending != valueb) return false;

        /* Validate attributes */
        err = alp_attn_dml_statement_get_uint32 (stmtH, Attributes, &valueu);
        ALP_ATLX(DML,"attributes %x == %x", alertP->attributes, valueu);
        if (alertP->attributes != valueu) return false;

        break;

    case PropertyItem:
        /* Validate PropLuid */
        propLuid  = 0;
        err = alp_attn_dml_statement_get_luid (stmtH, &propLuid);
        ALP_ATLX(DML,"propLuid %d == %d", propP->itemLuid, propLuid);
        if (propP->itemLuid  != propLuid)  return false;
        
        /* Validate AlertLuid */
        err = alp_attn_dml_statement_get_int64 (stmtH, AlertLuid, &value64);
        ALP_ATLX(DML,"alertLuid %d == %d", propP->alertLuid, value64);
        if (propP->alertLuid != (uint32_t) value64) return false;

        /* Validate property name */
        err = alp_attn_dml_statement_get_string (stmtH, Name, &s);
        ALP_ATLX(DML,"prop name <%s> == <%s>", propP->name, s);
        if (!MyStrCmp(propP->name, s)) {
            free (s);
            return false;
        }
        free (s);

        /* Validate property value */
        err = alp_attn_dml_statement_get_uint32 (stmtH, PropType, &ptype);

        err = alp_attn_dml_statement_get_blob (stmtH, Value, true,
                                               &pvalue, &psize);
        ALP_ATLX(DML,"prop value T:%x Sz:%d == T:%x Sz:%d",
                 propP->type, propP->size, ptype, psize);
        if (propP->type != ptype ||
            propP->size != psize ||
            memcmp(propP->value, pvalue, psize) != 0) {
            free (pvalue);
            return false;
        }
        free (pvalue);

        break;

    default:
        return false;
    }
        
    return true;
}


//
// AttnDMLTest methods
//

void
AttnDMLTest::initialize(void)
{
    pid_t         pid = getpid();
    gchar       *dbname;

    dbname = g_strdup_printf("AttnTests-%d.db", pid);
    
    alp_attn_dml_set_filename (ALP_SYSPATH_HOME, dbname);
    g_free (dbname);
    db_pathname = g_strdup(alp_attn_dml_get_filename());
    ALP_ATLX (DML, "set db filename to %s", db_pathname);
    alp_prv_attn_dml_init_database (db_pathname);
    sInitialized = true;
}

void
AttnDMLTest::setUp(void)
{
    dmlH = NULL;
    stmtH = NULL;
    db_pathname = NULL;
    initialize();
    ResetRandomStrings();
}

void
AttnDMLTest::tearDown(void)
{
    alp_status_t err;

    /* terminate a pending statement */
    if (stmtH) {
        err = alp_attn_dml_statement_end (stmtH);
        if (err) ALP_ATLX(DML,"statement end failed: %x", err);
    }
    stmtH = NULL;

    /* close the db */
    if (dmlH) {
        err = alp_attn_dml_close (dmlH);
        if (err) ALP_ATLX(DML,"db close failed: %x", err);
    }
    dmlH = NULL;

#if ALP_ATTN_DML_SAVE_DB == false
    /* remove db after each test */
    {
        FILE *fp;

        if ((fp = fopen (db_pathname, "r"))) {
            fclose(fp);
            err = unlink(db_pathname);
            ALP_ATLX(TEST, "unlink %s returns %x", db_pathname, err);
            CPPUNIT_ASSERT (err == 0);
        } else {
            ALP_ATLX(TEST, "unable to open %s", db_pathname);
        }
    }
    g_free (db_pathname);
    db_pathname = NULL;
#endif
}

void
AttnDMLTest::create_destroy_items()
{
    AlpDmlItemH         itemH;
    AlpAttnAlertType   *alertP   = NULL;
    AlpAttnPropType    *propP    = NULL;
    AlpAttnDmlItemType  itemType;
    int32_t             value;
    alp_status_t        err;

    /* test invalid item create */
    itemType = InvalidItem;
    err = alp_attn_dml_item_create (itemType, &itemH);
    CPPUNIT_ASSERT (err == ALP_STATUS_ATTN_DML_BAD_PARAM);

    /* test alert item create */
    itemType = AlertItem;
    err = alp_attn_dml_item_create (itemType, &itemH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    alertP = (AlpAttnAlertType *) itemH;
    CPPUNIT_ASSERT (alertP->itemType == AlertItem);
    err = alp_attn_dml_item_destroy (itemH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    
    /* test property item create */
    itemType = PropertyItem;
    err = alp_attn_dml_item_create (itemType, &itemH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    alertP = (AlpAttnAlertType *) itemH;
    CPPUNIT_ASSERT (alertP->itemType == PropertyItem);
    err = alp_attn_dml_item_destroy (itemH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    /* test property list item create */
    itemType = PropertyListItem;
    err = alp_attn_dml_item_create (itemType, &itemH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    alertP = (AlpAttnAlertType *) itemH;
    CPPUNIT_ASSERT (alertP->itemType == PropertyListItem);
    err = alp_attn_dml_item_destroy (itemH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
}

void
AttnDMLTest::set_int32()
{
    AlpDmlItemH          itemH;
    AlpAttnAlertType    *alertP    = NULL;
    AlpAttnPropType     *propP     = NULL;
    AlpAttnPropListType *propListP = NULL;
    AlpAttnDmlItemType   itemType;
    int32_t              value;
    uint32_t             valueu;
    alp_status_t         err;

    ALP_ATLX(DML,"create alert item...");
    /* test alert item create */
    itemType = AlertItem;
    err = alp_attn_dml_item_create (itemType, &itemH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    alertP = (AlpAttnAlertType *) itemH;
    CPPUNIT_ASSERT (alertP->itemType == AlertItem);

    ALP_ATLX(DML,"set priority...");
    value = 3;
    err = alp_attn_dml_item_set_int32 (alertP, Priority, &value);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    CPPUNIT_ASSERT (alertP->priority == 3);

    ALP_ATLX(DML,"set priority again...");
    value = 1;
    err = alp_attn_dml_item_set_int32 (alertP, Priority, &value);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    ALP_ATLX(DML,"set duration...");
    CPPUNIT_ASSERT (alertP->duration == 0);
    value = 1232456;
    err = alp_attn_dml_item_set_int32 (alertP, Duration, &value);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    
    ALP_ATLX(DML,"set pending...");
    CPPUNIT_ASSERT (alertP->pending == false);
    value = true;
    err = alp_attn_dml_item_set_int32 (alertP, Pending, &value);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    ALP_ATLX(DML,"set attributes...");
    CPPUNIT_ASSERT (alertP->attributes == 0);
    valueu = 0x6221701;
    err = alp_attn_dml_item_set_uint32 (alertP, Attributes, &valueu);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    ALP_ATLX(DML,"set user id...");
    CPPUNIT_ASSERT (alertP->userId == 0);
    value = 4678;
    err = alp_attn_dml_item_set_int32 (alertP, UserId, &value);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    ALP_ATLX(DML,"set group id...");
    CPPUNIT_ASSERT (alertP->groupId == 0);
    value = 1001;
    err = alp_attn_dml_item_set_int32 (alertP, GroupId, &value);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    ALP_ATLX(DML,"set udpated timestamp...");
    CPPUNIT_ASSERT (alertP->created == 0);
    value = 1492;
    err = alp_attn_dml_item_set_int32 (alertP, Created, &value);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    ALP_ATLX(DML,"set updated timestamp...");
    CPPUNIT_ASSERT (alertP->updated == 0);
    value = 1701;
    err = alp_attn_dml_item_set_int32 (alertP, Updated, &value);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    /* verify sets */
    CPPUNIT_ASSERT (alertP->priority == 1);
    CPPUNIT_ASSERT (alertP->duration == 1232456);
    CPPUNIT_ASSERT (alertP->userId   == 4678);
    CPPUNIT_ASSERT (alertP->groupId  == 1001);
    CPPUNIT_ASSERT (alertP->pending  == true);
    CPPUNIT_ASSERT (alertP->attributes  == 0x6221701);
    CPPUNIT_ASSERT (alertP->created == 1492);
    CPPUNIT_ASSERT (alertP->updated == 1701);

    /* test calls with a null arg now */
    ALP_ATLX(DML,"reset priority...");
    err = alp_attn_dml_item_set_int32 (alertP, Priority, NULL);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    CPPUNIT_ASSERT (alertP->priority == kAttnDefPriority);

    ALP_ATLX(DML,"reset duration...");
    err = alp_attn_dml_item_set_int32 (alertP, Duration, NULL);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    CPPUNIT_ASSERT (alertP->duration == kAttnDefDuration);

    ALP_ATLX(DML,"reset pending...");
    err = alp_attn_dml_item_set_int32 (alertP, Pending, NULL);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    CPPUNIT_ASSERT (alertP->pending == kAttnDefPending);

    ALP_ATLX(DML,"reset attributes...");
    err = alp_attn_dml_item_set_uint32 (alertP, Attributes, NULL);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    CPPUNIT_ASSERT (alertP->attributes == kAttnDefAttributes);

    ALP_ATLX(DML,"reset user id...");
    err = alp_attn_dml_item_set_int32 (alertP, UserId, NULL);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    CPPUNIT_ASSERT (alertP->userId == kAttnDefUserId);

    ALP_ATLX(DML,"reset group id...");
    err = alp_attn_dml_item_set_int32 (alertP, GroupId, NULL);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    CPPUNIT_ASSERT (alertP->groupId == kAttnDefGroupId);

    ALP_ATLX(DML,"reset created timestamp...");
    err = alp_attn_dml_item_set_int32 (alertP, Created, NULL);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    CPPUNIT_ASSERT (alertP->created == kAttnDefTimestamp);

    ALP_ATLX(DML,"reset updated timestamp...");
    err = alp_attn_dml_item_set_int32 (alertP, Updated, NULL);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    CPPUNIT_ASSERT (alertP->updated == kAttnDefTimestamp);

    /* destroy alert item */
    ALP_ATLX(DML,"destroy alert...");
    err = alp_attn_dml_item_destroy (itemH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    
    /* test property item create */
    ALP_ATLX(DML,"create property item...");
    itemType = PropertyItem;
    err = alp_attn_dml_item_create (itemType, &itemH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    propP = (AlpAttnPropType *) itemH;
    CPPUNIT_ASSERT (propP->itemType == PropertyItem);

    ALP_ATLX(DML,"set properties.alertluid...");
    CPPUNIT_ASSERT (propP->alertLuid == 0);
    valueu = 0x86621701;
    err = alp_attn_dml_item_set_uint32 (propP, AlertLuid, &valueu);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    CPPUNIT_ASSERT (propP->alertLuid == 0x86621701);

    err = alp_attn_dml_item_set_uint32 (propP, AlertLuid, NULL);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    CPPUNIT_ASSERT (propP->alertLuid == kAttnDefAlertLuid);

    /* destroy property item */
    ALP_ATLX(DML,"destroy property...");
    err = alp_attn_dml_item_destroy (itemH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    /* test property list item create */
    ALP_ATLX(DML,"create property list item...");
    itemType = PropertyListItem;
    err = alp_attn_dml_item_create (itemType, &itemH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    propListP = (AlpAttnPropListType *) itemH;
    CPPUNIT_ASSERT (propListP->itemType == PropertyListItem);

    ALP_ATLX(DML,"set property list alertluid...");
    CPPUNIT_ASSERT (propListP->itemLuid == 0);
    valueu = 0x86621701;
    err = alp_attn_dml_item_set_uint32 (propListP, AlertLuid, &valueu);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    CPPUNIT_ASSERT (propListP->itemLuid == 0x86621701);

    err = alp_attn_dml_item_set_uint32 (propListP, AlertLuid, NULL);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    CPPUNIT_ASSERT (propListP->itemLuid == kAttnDefAlertLuid);

    /* destroy property list item */
    ALP_ATLX(DML,"destroy property...");
    err = alp_attn_dml_item_destroy (itemH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
}

void
AttnDMLTest::set_string()
{
    AlpDmlItemH         itemH;
    AlpAttnAlertType   *alertP   = NULL;
    AlpAttnPropType    *propP    = NULL;
    AlpAttnDmlItemType  itemType;
    char               *value;
    char               *source, *name, *handle, *prop_value;
    GType               type;
    alp_status_t        err;

    source    = NULL;
    name      = NULL;
    handle    = NULL;
    
    /* test alert item create */
    itemType = AlertItem;
    err = alp_attn_dml_item_create (itemType, &itemH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    alertP = (AlpAttnAlertType *) itemH;
    CPPUNIT_ASSERT (alertP->itemType == AlertItem);

    /* Source */
    value = "Foo";
    err = alp_attn_dml_item_set_string (alertP, Source, value);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    CPPUNIT_ASSERT (strcmp(alertP->source, "Foo") == 0);

    value = RandomSource();
    source = strdup(value);
    err = alp_attn_dml_item_set_string (alertP, Source, value);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    /* Alert name */
    value = RandomName();
    name = strdup(value);
    err = alp_attn_dml_item_set_string (alertP, Name, value);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    /* Handle */
    value = RandomHandle();
    handle = strdup(value);
    err = alp_attn_dml_item_set_string (alertP, Handle, value);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    CPPUNIT_ASSERT (strcmp(alertP->source,    source) == 0);
    CPPUNIT_ASSERT (strcmp(alertP->name,      name) == 0);
    CPPUNIT_ASSERT (strcmp(alertP->handle,    handle) == 0);

    /* free up saved strings */
    if (source)     free(source);
    if (name)       free(name);
    if (handle)     free(handle);

    name = NULL;
    prop_value = NULL;
    
    /* test calls with a null arg now */
    err = alp_attn_dml_item_set_string (alertP, Source, NULL);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    CPPUNIT_ASSERT (alertP->source == NULL);

    err = alp_attn_dml_item_set_string (alertP, Name, NULL);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    CPPUNIT_ASSERT (alertP->name == NULL);

    err = alp_attn_dml_item_set_string (alertP, Handle, NULL);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    CPPUNIT_ASSERT (alertP->handle == NULL);

    /* destroy alert item */
    err = alp_attn_dml_item_destroy (itemH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    
    /* test property item create */
    itemType = PropertyItem;
    err = alp_attn_dml_item_create (itemType, &itemH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    propP = (AlpAttnPropType *) itemH;
    CPPUNIT_ASSERT (propP->itemType == PropertyItem);

    /* Property Name */
    err = alp_attn_dml_item_set_string (propP, Name, "Foo");
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    CPPUNIT_ASSERT (strcmp(propP->name, "Foo") == 0);

    value = RandomPropName();
    name = strdup(value);
    err = alp_attn_dml_item_set_string (propP, Name, value);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    /* Property value */
    value = RandomPropValue();
    prop_value = strdup(value);
    err = alp_attn_dml_item_set_string (propP, Value, value);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    /* verify sets */
    CPPUNIT_ASSERT (strcmp(propP->name,  name) == 0);
    CPPUNIT_ASSERT (propP->type == G_TYPE_STRING);
    CPPUNIT_ASSERT (propP->size == (strlen (prop_value) + 1));
    CPPUNIT_ASSERT (memcmp (propP->value, prop_value, propP->size) == 0);

    /* test calls with a null arg now */
    err = alp_attn_dml_item_set_string (propP, Name, NULL);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    CPPUNIT_ASSERT (propP->name == NULL);

    err = alp_attn_dml_item_set_string (propP, Value, NULL);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    CPPUNIT_ASSERT (propP->value == NULL);

    /* free up saved strings */
    if (name)       free(name);
    if (prop_value) free(prop_value);


    /* destroy property item */
    err = alp_attn_dml_item_destroy (itemH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
}

void
AttnDMLTest::open_close_db()
{
    alp_status_t      err;
    AlpDmlH           dmlH;
    FILE             *fp;
    uint32_t          length;
    char             *dbname;
    char             *saved_dbname = NULL;

    // save a copy to verify with later
    dbname = (char *) alp_attn_dml_get_filename();
    CPPUNIT_ASSERT (dbname != NULL);
    saved_dbname = g_strdup (dbname);

    // test bogus directory name
    err = alp_attn_dml_set_filename ("/foo", NULL);
    CPPUNIT_ASSERT (err == ALP_STATUS_ATTN_DML_BAD_PARAM);
    
    // test bogus filename
    err = alp_attn_dml_set_filename (NULL, "foo/bar");
    CPPUNIT_ASSERT (err == ALP_STATUS_ATTN_DML_BAD_PARAM);
    
    // test already set
    err = alp_attn_dml_set_filename (".", "foo");
    ALP_ATLX(DML,"./foo returns %x", err);
    CPPUNIT_ASSERT (err == ALP_STATUS_ATTN_DML_EXISTS);

    dbname = g_strdup(alp_attn_dml_get_filename());
    CPPUNIT_ASSERT (strcmp(dbname, saved_dbname) == 0);
    g_free (saved_dbname);

    // open database
    err = alp_attn_dml_open (&dmlH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    
    // verify it has been created and has length > 0
    if ((fp = fopen (dbname, "r")) == NULL) {
        length = 0;
    } else {
        fseek(fp, 0, SEEK_END);
        length = ftell(fp);
        fclose(fp);
    }
    CPPUNIT_ASSERT (length > 0);

    /* close the db */
    if (dmlH) {
        err = alp_attn_dml_close (dmlH);
        if (err) ALP_ATLX(DML,"db close failed: %x", err);
    }
    dmlH = NULL;
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    (void) unlink (dbname);
    g_free (dbname);
}

void
AttnDMLTest::insert_alert_item()
{
    alp_status_t      err;
    AlpDmlItemH       itemH;
    AlpAttnAlertType *alertP;
    AlpLuid           itemLuid;
    int32_t           value;
    uint32_t          valueu;
    int32_t           count;
    
    /* seed the random generator */
    srand (time(NULL));

    /* should verify database is empty first */
    err = alp_attn_dml_item_create (AlertItem, &itemH);
    ALP_ATLX(DML, "create return with err %x", err);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);

    alertP = (AlpAttnAlertType *) itemH;

    /* populate item */
    err = alp_attn_dml_item_set_string (itemH, Source, RandomSource());
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    err = alp_attn_dml_item_set_string (itemH, Name, RandomName());
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    err = alp_attn_dml_item_set_string (itemH, Handle, RandomHandle());
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    value = RandomDuration();
    err = alp_attn_dml_item_set_int32 (itemH, Duration, &value);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    value = RandomPriority();
    err = alp_attn_dml_item_set_int32 (itemH, Priority, &value);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    valueu = 0;
    err = alp_attn_dml_item_set_uint32 (itemH, Attributes, &valueu);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);

    /* open database */
    err = alp_attn_dml_open (&dmlH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

#if ALP_ATTN_TRACE_SQL == TRUE
    /* setup sqlite trace callback */
    sqlite3_trace ((sqlite3 *)dmlH, attn_sql_trace, NULL);
#endif

    /* test insert fail modes */
    err = alp_attn_dml_insert (NULL, itemH, &itemLuid);
    CPPUNIT_ASSERT (err == ALP_STATUS_ATTN_DML_BAD_PARAM);
    err = alp_attn_dml_insert (dmlH, NULL, &itemLuid);
    CPPUNIT_ASSERT (err == ALP_STATUS_ATTN_DML_BAD_PARAM);

    /* insert item into data base */
    err = alp_attn_dml_insert (dmlH, itemH, &itemLuid);
    ALP_ATLX(DML,"insert item %x", err);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(itemLuid > 0);

    /*  verify item is present */
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnAlertsTableName " WHERE "
                                kAttnAlertNameAlertLuid " = %d;",
                                alertP->itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 1);

    /* test insert duplicate alert */
     err = alp_attn_dml_insert (dmlH, itemH, &itemLuid);
     ALP_ATLX(TEST,"dup insert should fail %x", err);
     // CPPUNIT_ASSERT(err == ALP_STATUS_ATTN_DML_DUPLICATE);
     // CPPUNIT_ASSERT(err == ALP_STATUS_ATTN_DATABASE_ERROR);
     CPPUNIT_ASSERT(err == ALP_STATUS_ATTN_DML_EXISTS);

    /* destory alert item */
    err = alp_attn_dml_item_destroy (alertP);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
}

void
AttnDMLTest::update_alert_item (void)
{
    AlpDmlItemH       insertAlertH;
    AlpDmlItemH       updateAlertH;
    AlpAttnAlertType *alertP;
    AlpLuid           luid;
    alp_status_t      err;
    int32_t           value;
    uint32_t          valueu;
    uint32_t          numResults = 0;
    bool              match = false;

    /* Construct a new AlertItem */
    err = NewItem(AlertItem, &insertAlertH);
    ALP_ATLX(DML,"New item returned %x", err);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    /* Open Data Model */
    err = alp_attn_dml_open(&dmlH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

#if ALP_ATTN_TRACE_SQL == TRUE
    /* setup sqlite trace callback */
    sqlite3_trace ((sqlite3 *)dmlH, attn_sql_trace, NULL);
#endif

    /* Insert alert */
    err = alp_attn_dml_insert(dmlH, insertAlertH, &luid);
    if (err) ALP_ATLX(DML, "%d err %x", __LINE__, err);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    alertP = (AlpAttnAlertType *) insertAlertH;

    /* Update alert */
    err = alp_attn_dml_item_create (AlertItem, &updateAlertH);
    if (err) ALP_ATLX(DML, "%d err %x", __LINE__, err);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    err = alp_attn_dml_item_set_string (updateAlertH, Source, alertP->source);
    if (err) ALP_ATLX(DML, "%d err %x", __LINE__, err);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    err = alp_attn_dml_item_set_string (updateAlertH, Name,   alertP->name);
    if (err) ALP_ATLX(DML, "%d err %x", __LINE__, err);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    err = alp_attn_dml_item_set_string (updateAlertH, Handle, alertP->handle);
    if (err) ALP_ATLX(DML, "%d err %x", __LINE__, err);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    value = 2;
    err = alp_attn_dml_item_set_int32 (updateAlertH, Priority, &value);
    if (err) ALP_ATLX(DML, "%d err %x", __LINE__, err);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    value = 2000;
    err = alp_attn_dml_item_set_int32 (updateAlertH, Duration, &value);
    if (err) ALP_ATLX(DML, "%d err %x", __LINE__, err);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    value = (int) false;
    err = alp_attn_dml_item_set_int32 (updateAlertH, Pending, &value);
    if (err) ALP_ATLX(DML, "%d err %x", __LINE__, err);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    
    valueu = RandomAttributes();
    err = alp_attn_dml_item_set_uint32 (updateAlertH, Attributes, &valueu);
    if (err) ALP_ATLX(DML, "%d err %x", __LINE__, err);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    
    /* Update database with changed alert */
    err = alp_attn_dml_update (dmlH, updateAlertH);
    if (err) ALP_ATLX(DML, "%d err %x", __LINE__, err);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    /* Initialize selection statement to read alert */
    ALP_ATLX(DML,"beginning alert statement");
    err = alp_attn_dml_alert_statement_begin(
        dmlH, 
        &stmtH,
        NULL, NULL, NULL,
        luid,
        OrderByNone,
        &numResults);
    ALP_ATLX(DML,"begin returns %d results", numResults);
    if (err) ALP_ATLX(DML, "%d err %x", __LINE__, err);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(numResults == 1);

    /* Step to first result */
    err = alp_attn_dml_statement_next (stmtH);
    if (err) ALP_ATLX(DML, "%d err %x", __LINE__, err);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);

    // set so that it matches
    valueu = luid;
    err = alp_attn_dml_item_set_uint32 (updateAlertH, AlertLuid, &valueu);
    if (err) ALP_ATLX(DML, "%d err %x", __LINE__, err);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    /* Compare alert we created and item in current statement selection */
    CPPUNIT_ASSERT( CompareItems(updateAlertH, stmtH) == true );

    /* Destroy alert item */
    err = alp_attn_dml_item_destroy (insertAlertH);
    if (err) ALP_ATLX(DML, "%d err %x", __LINE__, err);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    err = alp_attn_dml_item_destroy (updateAlertH);
    if (err) ALP_ATLX(DML, "%d err %x", __LINE__, err);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
}

void
AttnDMLTest::replace_alert_item (void)
{
    AlpDmlItemH       itemH;
    AlpDmlItemH       replaceItemH;
    AlpAttnAlertType *alertP;
    AlpLuid           luid;
    alp_status_t      err;
    int32_t           value;
    uint32_t          valueu;
    uint32_t          numResults = 0;
    bool              match = false;

    char             *source = "bar:mil.starfleet.ops.cmd";
    char             *name   = "self-destruct";
    char             *handle = "ncc-1701";
    
    AlpAttnProp       prop1  = {"destruct-sequence-1", "code-1-1A" };
    AlpAttnProp       prop2  = {"destruct-sequence-2", "code-1-1A-2B" };
    AlpAttnProp       prop3  = {"destruct-sequence-3", "code-1B-2B-3" };
    AlpAttnProp       prop4  = {"final-code", "0-0-0-destruct-0" };
    AlpAttnProp       prop5  = {"abort-destruct-sequence", "code-1-2-3-continuity" };

    /* Construct a new AlertItem */
    err = alp_attn_dml_item_create (AlertItem, &itemH);
    ALP_ATLX(DML,"create alert item returned %x", err);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    /* populate alert item */
    err = alp_attn_dml_item_set_string (itemH, Source, source);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    err = alp_attn_dml_item_set_string (itemH, Name, name);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    err = alp_attn_dml_item_set_string (itemH, Handle, handle);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    value = 1;
    err = alp_attn_dml_item_set_int32 (itemH, Priority, &value);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    value = 30;
    err = alp_attn_dml_item_set_int32 (itemH, Duration, &value);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    valueu =
        ALP_ATTN_ATTR_TYPE_ATTN |
        ALP_ATTN_ATTR_DISPLAY_ALERT |
        ALP_ATTN_ATTR_DISPLAY_LIST;
    err = alp_attn_dml_item_set_uint32 (itemH, Attributes, &valueu);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);

    /* Open Data Model */
    err = alp_attn_dml_open(&dmlH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

#if ALP_ATTN_TRACE_SQL == TRUE
    /* setup sqlite trace callback */
    sqlite3_trace ((sqlite3 *)dmlH, attn_sql_trace, NULL);
#endif

    err = alp_attn_dml_item_set_int32 (itemH, Duration, &value);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    valueu =
        ALP_ATTN_ATTR_TYPE_ATTN |
        ALP_ATTN_ATTR_DISPLAY_ALERT |
        ALP_ATTN_ATTR_DISPLAY_LIST;
    err = alp_attn_dml_item_set_uint32 (itemH, Attributes, &valueu);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);

    /* Insert alert */
    err = alp_attn_dml_insert (dmlH, itemH, &luid);
    if (err) ALP_ATLX(DML, "%d err %x", __LINE__, err);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    alertP = (AlpAttnAlertType *) itemH;


    /* replace alert */
    err = DupItem (itemH, &replaceItemH);
    if (err) ALP_ATLX(DML, "%d err %x", __LINE__, err);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    value = 2;
    err = alp_attn_dml_item_set_int32 (replaceItemH, Priority, &value);
    if (err) ALP_ATLX(DML, "%d err %x", __LINE__, err);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    value = 2000;
    err = alp_attn_dml_item_set_int32 (replaceItemH, Duration, &value);
    if (err) ALP_ATLX(DML, "%d err %x", __LINE__, err);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    
    /* Replace alert in database with changed alert */
    err = alp_attn_dml_replace (dmlH, replaceItemH);
    if (err) ALP_ATLX(DML, "%d err %x", __LINE__, err);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    /* Initialize selection statement to read alert */
    ALP_ATLX(DML,"beginning alert statement");
    err = alp_attn_dml_alert_statement_begin(
        dmlH, 
        &stmtH,
        NULL, NULL, NULL,
        luid,
        OrderByNone,
        &numResults);
    ALP_ATLX(DML,"begin returns %d results", numResults);
    if (err) ALP_ATLX(DML, "%d err %x", __LINE__, err);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(numResults == 1);

    /* Step to first result */
    err = alp_attn_dml_statement_next (stmtH);
    if (err) ALP_ATLX(DML, "%d err %x", __LINE__, err);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);

#if 0
    // not necessary any more because both update and replace will update id
    // set so that it matches
    valueu = luid;
    err = alp_attn_dml_item_set_uint32 (updateAlertH, AlertLuid, &valueu);
    if (err) ALP_ATLX(DML, "%d err %x", __LINE__, err);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
#endif

    /* Compare alert we created and item in current statement selection */
    CPPUNIT_ASSERT( CompareItems(replaceItemH, stmtH) == true );

    /* Destroy alert item */
    err = alp_attn_dml_item_destroy (itemH);
    if (err) ALP_ATLX(DML, "%d err %x", __LINE__, err);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    err = alp_attn_dml_item_destroy (replaceItemH);
    if (err) ALP_ATLX(DML, "%d err %x", __LINE__, err);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
}

void
AttnDMLTest::delete_alert_items()
{
    alp_status_t      err;
    AlpDmlItemH       itemH;
    AlpAttnAlertType *alert1, *alert2;
    AlpLuid           itemLuid;
    AlpLuid           alertLuid1;
    AlpLuid           alertLuid2;
    int32_t           value;
    int32_t           count;
    
    err = alp_attn_dml_open (&dmlH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

#if ALP_ATTN_TRACE_SQL == true
    /* setup sqlite trace callback */
    sqlite3_trace ((sqlite3 *)dmlH, attn_sql_trace, NULL);
#endif

#if  ALP_ATTN_DML_SAVE_DB == false
    /* verify db empty to start */
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM " kAttnAlertsTableName ";");
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    CPPUNIT_ASSERT (count == 0);
#endif

    /* create 1st item */
    err = NewItem (AlertItem, &itemH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    alert1 = (AlpAttnAlertType *) itemH;

    /* create 2nd item */
    err = NewItem (AlertItem, &itemH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    alert2 = (AlpAttnAlertType *) itemH;

    /* insert 2nd item */
    err = alp_attn_dml_insert (dmlH, alert2, &alertLuid2);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(itemLuid > 0);

    /* insert 1st item */
    err = alp_attn_dml_insert (dmlH, alert1, &alertLuid1);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(itemLuid > 0);

    /* verify items are in database */
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnAlertsTableName " WHERE "
                                kAttnAlertNameAlertLuid " = %d;",
                                alert1->itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 1);
    
#if ALP_ATTN_DML_SAVE_DB == false
    /* verify db has only two alerts */
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM " kAttnAlertsTableName ";");
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    CPPUNIT_ASSERT (count == 2);
#endif

    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnAlertsTableName " WHERE "
                                kAttnAlertNameAlertLuid " = %d;",
                                alert2->itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 1);
    
    /* delete 2nd item from database */
    err = alp_attn_dml_delete (dmlH, alert2);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    /* verify 2nd item has been removed */
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnAlertsTableName " WHERE "
                                kAttnAlertNameAlertLuid " = %d;",
                                alert2->itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 0);

    /* verify 1st item has not been removed */
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnAlertsTableName " WHERE "
                                kAttnAlertNameAlertLuid " = %d;",
                                alert1->itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 1);

    /* delete 1st item from database */
    err = alp_attn_dml_delete (dmlH, alert1);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    
    /* verify 1st item has been removed */
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnAlertsTableName " WHERE "
                                kAttnAlertNameAlertLuid " = %d;",
                                alert1->itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 0);

    /* verify that the db is also empty */
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM " kAttnAlertsTableName ";");
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
#if ALP_ATTN_DML_SAVE_DB == false
    CPPUNIT_ASSERT(count == 0);
#endif    
    
    /* destory alert items */
    err = alp_attn_dml_item_destroy (alert1);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    err = alp_attn_dml_item_destroy (alert2);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
}

void
AttnDMLTest::insert_prop_item()
{
    alp_status_t      err;
    AlpDmlItemH       itemH;
    AlpAttnAlertType *alertP;
    AlpAttnPropType  *prop1, *prop2, *prop3;
    AlpLuid           itemLuid;
    int32_t           value;
    int32_t           count;
    
    /* should verify database is empty first */
    err = NewItem (AlertItem, &itemH);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);

    alertP = (AlpAttnAlertType *) itemH;

    /* open database */
    err = alp_attn_dml_open (&dmlH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

#if ALP_ATTN_TRACE_SQL == TRUE
    /* setup sqlite trace callback */
    sqlite3_trace ((sqlite3 *)dmlH, attn_sql_trace, NULL);
#endif

    /* insert item into data base */
    err = alp_attn_dml_insert (dmlH, itemH, &itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(itemLuid > 0);


    /*  verify item is present */
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM " kAttnAlertsTableName
                                " WHERE "
                                kAttnAlertNameAlertLuid " = %d;",
                                alertP->itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 1);

    /* now add some properties */
    err = NewItem (PropertyItem, &itemH);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    prop1 = (AlpAttnPropType *) itemH;

    err = NewItem (PropertyItem, &itemH);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    prop2 = (AlpAttnPropType *) itemH;

    err = NewItem (PropertyItem, &itemH);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    prop3 = (AlpAttnPropType *) itemH;

    err = alp_attn_dml_item_set_uint32 (prop2, AlertLuid, &alertP->itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    err = alp_attn_dml_insert (dmlH, prop2, &itemLuid);
    ALP_ATLX(DML,"prop2 insert err = %x", err);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(itemLuid > 0);

    err = alp_attn_dml_item_set_uint32 (prop3, AlertLuid, &alertP->itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    err = alp_attn_dml_insert (dmlH, prop3, &itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(itemLuid > 0);

    err = alp_attn_dml_item_set_uint32 (prop1, AlertLuid, &alertP->itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    err = alp_attn_dml_insert (dmlH, prop1, &itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(itemLuid > 0);

    /* verify that they are present */
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnPropsTableName " WHERE "
                                kAttnPropNamePropLuid " = %d;",
                                prop1->itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 1);

    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnPropsTableName " WHERE "
                                kAttnPropNamePropLuid " = %d;",
                                prop2->itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 1);

    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                "PROPERTIES WHERE PropLUID = %d;",
                                prop3->itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 1);

    /* test insert duplicate property */
    itemLuid = alertP->itemLuid;
    err = alp_attn_dml_insert (dmlH, prop3, &itemLuid);
//    CPPUNIT_ASSERT(err == ALP_STATUS_ATTN_DML_DUPLICATE);
    CPPUNIT_ASSERT(err == ALP_STATUS_ATTN_DML_DATABASE_ERROR);
    
    /* destroy items */
    err = alp_attn_dml_item_destroy (alertP);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);

    err = alp_attn_dml_item_destroy (prop1);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);

    err = alp_attn_dml_item_destroy (prop2);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);

    err = alp_attn_dml_item_destroy (prop3);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
}

void
AttnDMLTest::update_prop_item()
{
    AlpDmlItemH          itemH;
    AlpAttnAlertType    *alertP;
    AlpAttnPropType     *propP;
    AlpLuid              alertLuid;
    AlpLuid              itemLuid;
    alp_status_t         err;
    uint32_t             numResults = 0;
    char                *property_id;

    /* Construct a new AlertItem */
    err = NewItem(AlertItem, &itemH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    alertP = (AlpAttnAlertType *) itemH;

    /* Open Data Model */
    err = alp_attn_dml_open(&dmlH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

#if ALP_ATTN_TRACE_SQL == TRUE
    /* setup sqlite trace callback */
    sqlite3_trace ((sqlite3 *)dmlH, attn_sql_trace, NULL);
#endif

    /* Insert alert */
    err = alp_attn_dml_insert(dmlH, itemH, &itemLuid);
    ALP_ATLX(DML,"alert insert: %d", err);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    alertLuid = itemLuid;
    ALP_ATLX(DML,"alert inserted with luid = %d", itemLuid);

    /* Construct a property item */
    err = NewItem (PropertyItem, &itemH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    propP = (AlpAttnPropType *) itemH;
    property_id = strdup (propP->name);

    err = alp_attn_dml_item_set_uint32 (propP, AlertLuid, &alertLuid);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    CPPUNIT_ASSERT (propP->alertLuid == alertLuid);

    /* Insert property */
    err = alp_attn_dml_insert (dmlH, itemH, &itemLuid);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    ALP_ATLX(DML,"property inserted with luid = %d", itemLuid);
    
    /* Change local property */
    err = alp_attn_dml_item_set_string (
        itemH,
        Name,
        "location");
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    err = alp_attn_dml_item_set_string (
        itemH,
        Value,
        "on the unfasionable end of the western spiral arm");
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    /* Update database with changed property */
    err = alp_attn_dml_update (dmlH, itemH);
    ALP_ATLX(DML,"updating property: err = %x", err);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    /* Initialize selection statement to read alert */
    ALP_ATLX(DML,"beginning alert statement: alert %d prop %d id %s",
            alertLuid, itemLuid, property_id);
    err = alp_attn_dml_prop_statement_begin(
        dmlH, 
        &stmtH,
        alertLuid,
        itemLuid,
        "location",
        &numResults);
    ALP_ATLX(DML,"begin returns %d results %x err", numResults, err);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(numResults == 1);

    /* Step to first result */
    err = alp_attn_dml_statement_next (stmtH);
    ALP_ATLX(DML,"next statement: %x", err);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);

    /* Compare alert we created and item in current statement selection */
    CPPUNIT_ASSERT( CompareItems(itemH, stmtH) == true );

    /* Destroy items */
    err = alp_attn_dml_item_destroy (alertP);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);

    err = alp_attn_dml_item_destroy (propP);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);

    free (property_id);
}

void
AttnDMLTest::update_prop_list()
{
    AlpDmlItemH          itemH;
    AlpAttnAlertType    *alertP;
    AlpAttnPropListType *propListP;
    AlpAttnPropType     *prop1, *prop2, *prop3, *prop4, *prop5;
    AlpLuid              alertLuid;
    AlpLuid              itemLuid;
    alp_status_t         err;
    int32_t              count;
    uint32_t             numResults = 0;
    char                *property_name;
    void                *property_value;
    char                *property_string;
    uint32_t             property_type;
    uint32_t             property_size;

    /* Construct a new AlertItem */
    err = NewItem(AlertItem, &itemH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    alertP = (AlpAttnAlertType *) itemH;

    /* Open Data Model */
    err = alp_attn_dml_open(&dmlH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

#if ALP_ATTN_TRACE_SQL == TRUE
    /* setup sqlite trace callback */
    sqlite3_trace ((sqlite3 *)dmlH, attn_sql_trace, NULL);
#endif

    /* Insert alert */
    err = alp_attn_dml_insert(dmlH, itemH, &alertLuid);
    ALP_ATLX(DML,"alert insert: %d", err);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    CPPUNIT_ASSERT (alertLuid > 0);
    ALP_ATLX(DML,"alert inserted with luid = %d", alertLuid);

    /* Create property list */
    err = alp_attn_dml_item_create (PropertyListItem, &itemH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    propListP = (AlpAttnPropListType *) itemH;
    err = alp_attn_dml_item_set_uint32 (itemH, AlertLuid, &alertLuid);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    CPPUNIT_ASSERT (propListP->itemLuid == alertLuid);

    err = alp_attn_dml_item_set_string (itemH, Source, alertP->source);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    err = alp_attn_dml_item_set_string (itemH, Name,   alertP->name);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    err = alp_attn_dml_item_set_string (itemH, Handle, alertP->handle);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    
    /* Construct property items */
    err = NewItem (PropertyItem, &itemH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    prop1 = (AlpAttnPropType *) itemH;
    err = alp_attn_dml_item_add_property (propListP, prop1);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    property_name = strdup (prop1->name);

    err = NewItem (PropertyItem, &itemH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    prop2 = (AlpAttnPropType *) itemH;
    err = alp_attn_dml_item_add_property (propListP, prop2);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    err = NewItem (PropertyItem, &itemH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    prop3 = (AlpAttnPropType *) itemH;
    err = alp_attn_dml_item_add_property (propListP, prop3);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    err = NewItem (PropertyItem, &itemH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    prop4 = (AlpAttnPropType *) itemH;
    err = alp_attn_dml_item_add_property (propListP, prop4);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    err = NewItem (PropertyItem, &itemH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    prop5 = (AlpAttnPropType *) itemH;
    err = alp_attn_dml_item_add_property (propListP, prop5);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    /* verify that the existing property list is empty */
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnPropsTableName " WHERE "
                                kAttnPropNameAlertLuid " = %d;",
                                alertLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 0);

    /* Insert property */
    err = alp_attn_dml_update (dmlH, propListP);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    ALP_ATLX(DML,"alert id %d property list updated", alertLuid);
    
    /* verify that all properties in list were added to alert's properties */
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnPropsTableName " WHERE "
                                kAttnPropNameAlertLuid " = %d;",
                                alertLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 5);

    
    /* Change items  in property list */

    /* change prop1 to add new prop "location" */
    err = alp_attn_dml_item_set_string (prop1, Name, "location");
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    err = alp_attn_dml_item_set_string (
        prop1,
        Value,
        "on the unfasionable end of the western spiral arm");
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    /* mess up prop2 so it has a NULL name and is skipped */
    err = alp_attn_dml_item_set_string (prop2, Name, NULL);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    
    /* delete prop3 from db by setting value to NULL */
    err = alp_attn_dml_item_set_string (prop3, Value, NULL);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    /* change prop4 value so it actually updates the property */
    err = alp_attn_dml_item_set_string (prop4, Value, "18886221701");
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    /* delete prop5 from db by setting value to NULL */
    err = alp_attn_dml_item_set_string (prop5, Value, NULL);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    /* Update database with changed property */
    err = alp_attn_dml_update (dmlH, propListP);
    ALP_ATLX(DML,"updating property list for alert %d: err = %x", alertLuid, err);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    /* verify count is 4 (add one item, delete 2, update one) */
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnPropsTableName " WHERE "
                                kAttnPropNameAlertLuid " = %d;",
                                alertLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 4);

    /* verify that original prop1 is still present */
    count = -1;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnPropsTableName " WHERE "
                                kAttnPropNameName " = %Q AND "
                                kAttnPropNameAlertLuid " = %d;",
                                property_name,
                                alertLuid);
    ALP_ATLX(DML, "check orig alert %d prop1 %s is present: count %d err %x",
             alertLuid, property_name, count, err);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 1);

    /* verify that prop3 is no longer in the db */
    count = -1;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnPropsTableName " WHERE "
                                kAttnPropNamePropLuid " = %d;",
                                prop3->itemLuid);
    ALP_ATLX(DML, "check alert %d prop3 %s id %d has been removed: count %d err %x",
             alertLuid, prop3->name, prop3->itemLuid, count, err);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 0);

    /* verify that prop5 is no longer in the db */
    count = -1;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnPropsTableName " WHERE "
                                kAttnPropNamePropLuid " = %d;",
                                prop5->itemLuid);
    ALP_ATLX(DML, "check alert %d prop5 %s id %d has been removed: count %d err %x",
             alertLuid, prop5->name, prop5->itemLuid, count, err);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 0);

    /*
     * verify that added "location" prop is present and
     * it's value is correct
     */
    ALP_ATLX(DML,"prop statement begin: alert %d prop %s",
            alertLuid, prop1->name);
    err = alp_attn_dml_prop_statement_begin(
        dmlH, 
        &stmtH,
        alertLuid,
        0,
        prop1->name,
        &numResults);
    ALP_ATLX(DML,"begin returns %d results %x err", numResults, err);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(numResults == 1);

/* Step to first result */
    err = alp_attn_dml_statement_next (stmtH);
    ALP_ATLX(DML,"next statement: %x", err);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);

    /* Compare value of added item */
    err = alp_attn_dml_statement_get_uint32 (stmtH, PropType, &property_type);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    CPPUNIT_ASSERT (property_type == G_TYPE_STRING);
    err = alp_attn_dml_statement_get_blob (stmtH, Value, true,
                                           &property_value, &property_size);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    CPPUNIT_ASSERT (property_size == (strlen((const char *)prop1->value) + 1));
    CPPUNIT_ASSERT (memcmp(prop1->value, property_value, property_size) == 0);
    free (property_value);

    /* destroy statement */
    err = alp_attn_dml_statement_end (stmtH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    stmtH = NULL;

    /* verify the updated prop4 is present and it's value is correct */
    ALP_ATLX(DML,"prop statement begin: alert %d prop %s",
            alertLuid, prop4->name);
    err = alp_attn_dml_prop_statement_begin(
        dmlH, 
        &stmtH,
        alertLuid,
        0,
        prop4->name,
        &numResults);
    ALP_ATLX(DML,"begin returns num results %d err %x", numResults, err);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(numResults == 1);

    /* Step to first result */
    err = alp_attn_dml_statement_next (stmtH);
    ALP_ATLX(DML,"next statement: %x", err);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);

    /* Compare value of added item */
    err = alp_attn_dml_statement_get_uint32 (stmtH, PropType, &property_type);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    ALP_ATLX(DML,"prop types %x == %x (G_STRING_TYPE = %x)",
             prop4->type, property_type, G_TYPE_STRING);
    CPPUNIT_ASSERT (prop4->type == property_type);
    property_value = NULL;
    property_size = 0;
    err = alp_attn_dml_statement_get_blob (stmtH, Value, true,
                                           &property_value, &property_size);
    ALP_ATLX(DML,"prop value <%*s>:%d == <%*s>:%d",
             prop4->size, prop4->value, prop4->size,
             property_size, property_value, property_size);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    CPPUNIT_ASSERT (prop4->size == property_size);
    CPPUNIT_ASSERT (memcmp(prop4->value, property_value, property_size) == 0);
    free (property_value);

    /* Destroy statement */
    err = alp_attn_dml_statement_end (stmtH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    stmtH = NULL;

    /* Destroy items */
    err = alp_attn_dml_item_destroy (alertP);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);

    err = alp_attn_dml_item_destroy (propListP);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);

    free (property_name);
}

void
AttnDMLTest::delete_prop_items (void)
{
    alp_status_t      err;
    AlpDmlItemH       itemH;
    AlpAttnAlertType *alert1, *alert2;
    AlpAttnPropType  *prop1, *prop2, *prop3, *prop4, *prop5;
    AlpLuid           itemLuid, copyPropLuid1;
    AlpLuid           propLuid1, propLuid2, propLuid3, propLuid4, propLuid5;
    int32_t           value;
    int32_t           count;
    
    /* create some alerts */
    err = NewItem (AlertItem, &itemH);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    alert1 = (AlpAttnAlertType *) itemH;

    err = NewItem (AlertItem, &itemH);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    alert2 = (AlpAttnAlertType *) itemH;

    /* open database */
    err = alp_attn_dml_open (&dmlH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

#if ALP_ATTN_TRACE_SQL == TRUE
    /* setup sqlite trace callback */
    sqlite3_trace ((sqlite3 *)dmlH, attn_sql_trace, NULL);
#endif

    /* insert alert items into data base */
    err = alp_attn_dml_insert (dmlH, alert1, &itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(itemLuid > 0);

    err = alp_attn_dml_insert (dmlH, alert2, &itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(itemLuid > 0);


    /*  verify alert items are present */
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnAlertsTableName " WHERE "
                                kAttnAlertNameAlertLuid " = %d;",
                                alert1->itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 1);

    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnAlertsTableName " WHERE "
                                kAttnAlertNameAlertLuid " = %d;",
                                alert2->itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 1);

    /* now add some properties */
    err = NewItem (PropertyItem, &itemH);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    prop1 = (AlpAttnPropType *) itemH;

    err = NewItem (PropertyItem, &itemH);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    prop2 = (AlpAttnPropType *) itemH;

    err = NewItem (PropertyItem, &itemH);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    prop3 = (AlpAttnPropType *) itemH;

    err = NewItem (PropertyItem, &itemH);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    prop4 = (AlpAttnPropType *) itemH;

    err = NewItem (PropertyItem, &itemH);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    prop5 = (AlpAttnPropType *) itemH;

    err = alp_attn_dml_item_set_uint32 (prop4, AlertLuid, &alert1->itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    err = alp_attn_dml_insert (dmlH, prop4, &propLuid4);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(itemLuid > 0);

    err = alp_attn_dml_item_set_uint32 (prop5, AlertLuid, &alert1->itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    err = alp_attn_dml_insert (dmlH, prop5, &propLuid5);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(itemLuid > 0);

    err = alp_attn_dml_item_set_uint32 (prop2, AlertLuid, &alert2->itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    err = alp_attn_dml_insert (dmlH, prop2, &propLuid2);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(itemLuid > 0);

    err = alp_attn_dml_item_set_uint32 (prop3, AlertLuid, &alert2->itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    err = alp_attn_dml_insert (dmlH, prop3, &propLuid3);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(itemLuid > 0);

    err = alp_attn_dml_item_set_uint32 (prop1, AlertLuid, &alert2->itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    err = alp_attn_dml_insert (dmlH, prop1, &propLuid1);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(itemLuid > 0);

    /*
     * reuse prop item 1 as an alert 1 property
     * this will modify prop1's alertLuid and propLuid
     * so we make a copy of it's propLuid prior to the insert.
     */
    err = alp_attn_dml_item_set_uint32 (prop1, AlertLuid, &alert1->itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    err = alp_attn_dml_insert (dmlH, prop1, &copyPropLuid1);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(copyPropLuid1 > 0);

    /* verify that they are present */
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnPropsTableName " WHERE "
                                kAttnPropNamePropLuid " = %d;",
                                copyPropLuid1);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 1);

    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnPropsTableName " WHERE "
                                kAttnPropNamePropLuid " = %d;",
                                prop2->itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 1);

    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnPropsTableName " WHERE "
                                kAttnPropNamePropLuid " = %d;",
                                prop3->itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 1);

    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnPropsTableName " WHERE "
                                kAttnPropNamePropLuid " = %d;",
                                prop4->itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 1);

    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnPropsTableName " WHERE "
                                kAttnPropNamePropLuid " = %d;",
                                prop5->itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 1);

    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnPropsTableName " WHERE "
                                kAttnPropNamePropLuid " = %d;",
                                prop1->itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 1);

    /* find luid for prop2 attached to alert item 1 */
    err = alp_attn_dml_find (dmlH, prop1, &itemLuid);
    ALP_ATLX(DML,"find prop1 alert1 id %d prop %s %d,%d err %x",
             alert1->itemLuid, prop1->name, itemLuid, copyPropLuid1, err);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    CPPUNIT_ASSERT (itemLuid == copyPropLuid1);

    /* find luid for prop1 attached to alert item 2 */
    err = alp_attn_dml_item_set_uint32 (prop1, AlertLuid, &alert2->itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    err = alp_attn_dml_find (dmlH, prop1, &itemLuid);
    ALP_ATLX(DML,"find prop1 alert2 id %d prop %s %d,%d err %x",
             alert1->itemLuid, prop1->name, itemLuid, propLuid1, err);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    CPPUNIT_ASSERT (itemLuid == propLuid1);

    /* delete a property */
    err = alp_attn_dml_delete (dmlH, prop4);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    /* verify it has been deleted */
    count = -1;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnPropsTableName " WHERE "
                                kAttnPropNamePropLuid " = %d;",
                                prop4->itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 0);

    /* verify remaining properties still present */
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnPropsTableName " WHERE "
                                kAttnPropNameAlertLuid " in ( %d, %d);",
                                alert1->itemLuid,
                                alert2->itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 5);

    /* verify count of remaining properties for associated alert */
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnPropsTableName " WHERE "
                                kAttnPropNameAlertLuid " = %d;",
                                alert1->itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 2);
    
    /* delete the associated alert */
    err = alp_attn_dml_delete (dmlH, alert1);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    /* verify all remaining properties associate have been deleted */
    count = -1;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnPropsTableName " WHERE "
                                kAttnPropNameAlertLuid " = %d;",
                                itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 0);

    /* verify property belong to alert2 are still present */
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnPropsTableName " WHERE "
                                kAttnPropNameAlertLuid " = %d;",
                                alert2->itemLuid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 3);

    /* destroy items */
    err = alp_attn_dml_item_destroy (alert1);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    err = alp_attn_dml_item_destroy (alert2);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);

    err = alp_attn_dml_item_destroy (prop1);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    err = alp_attn_dml_item_destroy (prop2);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    err = alp_attn_dml_item_destroy (prop3);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    err = alp_attn_dml_item_destroy (prop4);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    err = alp_attn_dml_item_destroy (prop5);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
}

void
AttnDMLTest::add_get_property (void)
{
    AlpDmlItemH       itemH;
    AlpDmlItemH       prop1H, prop2H, prop3H;
    AlpAttnAlertType *alertP;
    AlpAttnPropType  *propP;
    alp_status_t      err;

    /* Construct a new AlertItem */
    err = NewItem(AlertItem, &itemH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

#if 0
    /* test get property return errors */
    err = alp_attn_dml_item_get_property (NULL, "foo", &prop1H);
    CPPUNIT_ASSERT (err == ALP_STATUS_ATTN_DML_BAD_PARAM);
    err = alp_attn_dml_item_get_property (itemH, NULL, &prop1H);
    CPPUNIT_ASSERT (err == ALP_STATUS_ATTN_DML_BAD_PARAM);
    err = alp_attn_dml_item_get_property (itemH, "", &prop1H);
    CPPUNIT_ASSERT (err == ALP_STATUS_ATTN_DML_BAD_PARAM);
    err = alp_attn_dml_item_get_property (itemH, "foo", NULL);
    CPPUNIT_ASSERT (err == ALP_STATUS_ATTN_DML_BAD_PARAM);
#endif

    /* test add property return errors */
    err = alp_attn_dml_item_add_property (NULL, prop1H);
    CPPUNIT_ASSERT (err == ALP_STATUS_ATTN_DML_BAD_PARAM);
    err = alp_attn_dml_item_add_property (itemH, NULL);
    CPPUNIT_ASSERT (err == ALP_STATUS_ATTN_DML_BAD_PARAM);
    

    /* test add property now */
    err = NewItem(PropertyItem, &prop1H);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    err = alp_attn_dml_item_add_property (itemH, prop1H);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    /* verify prop1H on alertItem property list */
    alertP = (AlpAttnAlertType *) itemH;
    CPPUNIT_ASSERT (alertP->plist == (AlpAttnPropType *) prop1H);
    CPPUNIT_ASSERT (alertP->plist->next == NULL);

    /* now add two more properties */
    err = NewItem(PropertyItem, &prop2H);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    err = alp_attn_dml_item_add_property (itemH, prop2H);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    err = NewItem(PropertyItem, &prop3H);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    err = alp_attn_dml_item_add_property (itemH, prop3H);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    /* verify they are on property list */
    alertP = (AlpAttnAlertType *) itemH;
    propP = alertP->plist;
    CPPUNIT_ASSERT (propP == (AlpAttnPropType *) prop3H);
    propP = propP->next;
    CPPUNIT_ASSERT (propP != NULL);
    CPPUNIT_ASSERT (propP == (AlpAttnPropType *) prop2H);
    propP = propP->next;
    CPPUNIT_ASSERT (propP != NULL);
    CPPUNIT_ASSERT (propP == (AlpAttnPropType *) prop1H);
    propP = propP->next;
    CPPUNIT_ASSERT (propP == NULL);

    /* clean up */
    err = alp_attn_dml_item_destroy (itemH); // this will delete property item too
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
}

void
AttnDMLTest::insert_compound_item (void)
{
    AlpDmlItemH       itemH;
    AlpDmlItemH       prop1H, prop2H, prop3H;
    AlpAttnAlertType *alertP;
    AlpAttnPropType  *propP;
    AlpLuid           luid;
    alp_status_t      err;
    int32_t           count;

    /* Construct a new AlertItem */
    err = NewItem(AlertItem, &itemH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    /* test add property now */
    err = NewItem(PropertyItem, &prop1H);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    err = alp_attn_dml_item_add_property (itemH, prop1H);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    ALP_ATLX(DML, "added 1st prop %s to alert [%s|%s|%s]",
             ((AlpAttnPropType *)prop1H)->name,
             ((AlpAttnAlertType *)itemH)->source,
             ((AlpAttnAlertType *)itemH)->name,
             ((AlpAttnAlertType *)itemH)->handle);

    /* verify prop1H on alertItem property list */
    alertP = (AlpAttnAlertType *) itemH;
    CPPUNIT_ASSERT (alertP->plist == (AlpAttnPropType *) prop1H);
    CPPUNIT_ASSERT (alertP->plist->next == NULL);

    /* now add two more properties */
    err = NewItem(PropertyItem, &prop2H);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    ALP_ATLX(DML, "added 2nd prop %s to alert [%s|%s|%s]",
             ((AlpAttnPropType *)prop2H)->name,
             ((AlpAttnAlertType *)itemH)->source,
             ((AlpAttnAlertType *)itemH)->name,
             ((AlpAttnAlertType *)itemH)->handle);

    err = alp_attn_dml_item_add_property (itemH, prop2H);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    err = NewItem(PropertyItem, &prop3H);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
    ALP_ATLX(DML, "added 3rd prop %s to alert [%s|%s|%s]",
             ((AlpAttnPropType *)prop3H)->name,
             ((AlpAttnAlertType *)itemH)->source,
             ((AlpAttnAlertType *)itemH)->name,
             ((AlpAttnAlertType *)itemH)->handle);
    err = alp_attn_dml_item_add_property (itemH, prop3H);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    /* open database */
    err = alp_attn_dml_open (&dmlH);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

#if ALP_ATTN_TRACE_SQL == TRUE
    /* setup sqlite trace callback */
    sqlite3_trace ((sqlite3 *)dmlH, attn_sql_trace, NULL);
#endif

    /* now insert item */
    err = alp_attn_dml_insert (dmlH, itemH, &luid);
    ALP_ATLX(DML, "inserted luid %d", luid);
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);

    /* verify alert item insertion */
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnAlertsTableName
                                " WHERE "
                                kAttnAlertNameAlertLuid
                                " = %d;",
                                luid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 1);
    
    /* verify property items insertion */
    count = 0;
    err = alp_sal_sql_get_value((AlpSALHandle)dmlH, &count, 
                                "SELECT COUNT(*) FROM "
                                kAttnPropsTableName
                                " WHERE "
                                kAttnPropNameAlertLuid
                                " = %d;",
                                luid);
    CPPUNIT_ASSERT(err == ALP_STATUS_OK);
    CPPUNIT_ASSERT(count == 3);

    /* clean up */
    err = alp_attn_dml_item_destroy (itemH); // this will delete property item too
    CPPUNIT_ASSERT (err == ALP_STATUS_OK);
}
