/********************************************************************

Copyright  2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/**
 * @file attn_ipc_data.c
 *
 * @brief ALP attention manager IPC data routines
 *
 *****************************************************************************/

#include <stdarg.h>
#include <glib.h>
#include <glib-object.h>

#include <hiker/types.h>
#include <hiker/prv/attnmgr_traces.h>
#include <hiker/prv/attnmgr_ipc.h>
#include <hiker/prv/attnmgr_dml.h> /* pick up data structs for alerts & props */

#include "attn_ipc_prv.h"

/***** PACKING ROUTINES *******/

/**
 * @brief Pack alert properties into message
 * @param[in] msg message to which properties will be added to
 * @param[in] nprops     number of properties in array
 * @param[in] properties array of properties
 * @param[in] channel    optional channel name
 * @return ALP_STATUS_OK if packing was successful
 */
alp_status_t
alp_attn_ipc_pack_properties (AlpMessage  *msg,
                              guint32      nprops,
                              AlpAttnProp *properties[],
                              const char  *channel)
{
	alp_status_t   err = ALP_OK;
    AlpAttnProp  **property;
    guint32        i;

    /* sanity checks */
    if (msg == NULL) {
        ALP_ATLX (IPC, "bad msg parameter");
        return ALP_STATUS_ATTN_IPC_BAD_PARAM;
    }
    if (nprops > ALP_ATTN_IPC_MAX_PROPERTIES) {
        ALP_ATLX (IPC, "bad nprops parameter, exceeds MAX_PROPERTIES");
        return ALP_STATUS_ATTN_IPC_BAD_PARAM;
    }
    if (nprops > 0 && properties == NULL) {
        ALP_ATLX (IPC, "bad nprops or properties parameter");
        return ALP_STATUS_ATTN_IPC_BAD_PARAM;
    }

    err = alp_ipc_message_pack_uint32 (msg, channel ? nprops + 1: nprops);
    if (err) {
        ALP_ATLX(IPC, "pack property count %d failed: err %x", nprops, err);
        return err;
    }

    property = properties;
    for (i = 0; i < nprops; i++) {
        const char *pname  = (*property)->name;
        const void *pvalue = (*property)->value;
        guint       psize  = 0;
        GType       ptype  = G_TYPE_STRING;

        if (pvalue) {
            /* count null byte terminator */
            psize = strlen((char *)(*property)->value) + 1;
        }

        /* first the property name */
        err = alp_ipc_message_pack_string(msg, pname);
        if (err) {
            ALP_ATLX(IPC, "pack property name %s returns %x", pname, err);
            return err;
        }
        /* second the property type */
        err = alp_ipc_message_pack_uint32 (msg, ptype);
        if (err) {
            ALP_ATLX(IPC, "pack property type %x returns %x", ptype, err);
            return err;
        }

        /* third the property value size and data */
        err = alp_ipc_message_pack_byte_array (msg, (guchar *)pvalue, psize);
        if (err) {
            ALP_ATLX(IPC, "packing of properties returns err = %x", err);
            return err;
        }
        property++;             /* advance to next property */
	}
    if (channel) {
        if (!err) {
            err = alp_ipc_message_pack_string (
                msg,
                ALP_ATTN_CALLBACK_CHANNEL_PROPERTY);
        }
        if (err) ALP_ATLX(IPC,
                          "pack property name  %s returns %x",
                          ALP_ATTN_CALLBACK_CHANNEL_PROPERTY,
                          err);
        if (!err) err = alp_ipc_message_pack_uint32 (msg, G_TYPE_STRING);
        if (!err) err = alp_ipc_message_pack_byte_array (msg,
                                                         (guchar *)channel,
                                                         strlen(channel)+1);
        if (err) ALP_ATLX(IPC, "pack value %s returns %x", channel, err);
    }
	
	return err;
}

/**
 * @brief Create post alert IPC message
 * @param[in] source alert source string
 * @param[in] name alert name string
 * @param[in] handle alert handle string
 * @param[in] priority integer priority of alert range 1 to 5
 * @param[in] duration integer duration of alert in number of seconds the
 * alert should exist after being posted, 0 means indefinitely.
 * @param[in] nprops number of properties
 * @param[in] properties array of properties
 * @return pointer to message buffer if successful
 * @return NULL if packing of the message buffer failed
 */
AlpMessage *
alp_attn_ipc_pack_alert_post_msg (const char  *source,
                                  const char  *name,
                                  const char  *handle,
                                  const char  *interface, /* TODO: ignore for the momemnt */
                                  gint32       priority,
                                  gint32       duration,
                                  guint32      nprops,
                                  AlpAttnProp *properties[],
                                  const char  *channel)
{
    AlpMessage      *msg;
    alp_status_t     err = ALP_STATUS_OK;

    msg = alp_ipc_message_create();
    if (msg == NULL) {
        err = ALP_STATUS_ATTN_IPC_NO_MEMORY;
        ALP_ATLX(IPC, "no memory on message create");
    }

    if (!err) err = alp_ipc_message_pack_start (msg, ALP_ATTN_MSG_ID_POST);
    if (!err) err = alp_ipc_message_pack_string (msg, source);
    if (!err) err = alp_ipc_message_pack_string (msg, name);
    if (!err) err = alp_ipc_message_pack_string (msg, handle);
    if (!err) err = alp_ipc_message_pack_int32 (msg, priority);
    if (!err) err = alp_ipc_message_pack_int32 (msg, duration);
    if (!err) err = alp_attn_ipc_pack_properties (msg, nprops,  properties, channel);
    if (!err) err = alp_ipc_message_pack_end(msg);

    if (err != ALP_STATUS_OK) {
        ALP_FAIL_NON_FATAL_IF (TRUE, "attn ipc post pack failed!");
        if (msg) alp_ipc_message_destroy (msg);
        msg = NULL;
    }
    return msg;
}

/**
 * @brief Create alert post IPC reply message
 * @param[in] replyTo original message being replied to
 * @param[in] status
 * @param[in] id the identifier of the posted alert
 * @return pointer to message buffer if successful
 * @return NULL if packing of the message buffer failed
 */
AlpMessage *
alp_attn_ipc_pack_alert_post_reply (AlpMessage *replyTo, guint32 status, guint32  id)
{
    AlpMessage      *msg;
    alp_status_t     err = ALP_STATUS_OK;

    msg = alp_ipc_message_create();
    if (msg == NULL) {
        err = ALP_STATUS_ATTN_IPC_NO_MEMORY;
        ALP_ATLX(IPC, "no memory on message create");
    }

    if (!err) {
        err = alp_ipc_message_response_pack_start (msg,
                                                   ALP_ATTN_MSG_ID_POST_REPLY,
                                                   replyTo);
    }
    if (!err) err = alp_ipc_message_pack_uint32 (msg, status);
    if (!err) err = alp_ipc_message_pack_uint32 (msg, id);
    if (!err) err = alp_ipc_message_pack_end(msg);

    if (err) {
        ALP_FAIL_NON_FATAL_IF (TRUE, "attn ipc post reply pack failed!");
        if (msg) alp_ipc_message_destroy (msg);
        msg = NULL;
    }

    return msg;
}

/**
 * @brief Create alert update IPC message
 * @param[in] identifier for alert to be updated
 * @param[in] nprops number of properties
 * @param[in] properties array of properties
 * @return pointer to message buffer if successful
 * @return NULL if packing of the message buffer failed
 */
AlpMessage *
alp_attn_ipc_pack_alert_update_msg (const char  *source,
                                    const char  *name,
                                    const char  *handle,
                                    guint32      nprops,
                                    AlpAttnProp *properties[])
{
    AlpMessage      *msg;
    alp_status_t     err = ALP_STATUS_OK;

    msg = alp_ipc_message_create();
    if (msg == NULL) {
        err = ALP_STATUS_ATTN_IPC_NO_MEMORY;
        ALP_ATLX(IPC, "no memory on message create");
    }

    if (!err) err = alp_ipc_message_pack_start (msg, ALP_ATTN_MSG_ID_UPDATE);
    if (!err) err = alp_ipc_message_pack_string (msg, source);
    if (!err) err = alp_ipc_message_pack_string (msg, name);
    if (!err) err = alp_ipc_message_pack_string (msg, handle);
    if (!err) err = alp_attn_ipc_pack_properties (msg,nprops,properties,NULL);
    if (!err) err = alp_ipc_message_pack_end (msg);

    if (err != ALP_STATUS_OK) {
        ALP_FAIL_NON_FATAL_IF (TRUE, "attn ipc post pack failed!");
        if (msg) alp_ipc_message_destroy (msg);
        msg = NULL;
    }

    return msg;
}

/**
 * @brief Create alert update IPC reply
 * @param[in] replyTo original message being replied to
 * @param[in] status
 * @return pointer to message buffer if successful
 * @return NULL if packing of the message buffer failed
 */
AlpMessage *
alp_attn_ipc_pack_alert_update_reply (AlpMessage *replyTo, guint32  status)
{
    AlpMessage      *msg;
    alp_status_t     err = ALP_STATUS_OK;

    msg = alp_ipc_message_create();
    if (msg == NULL) {
        err = ALP_STATUS_ATTN_IPC_NO_MEMORY;
        ALP_ATLX(IPC, "no memory on message create");
    }

    if (!err) {
        err =
            alp_ipc_message_response_pack_start (msg,
                                                 ALP_ATTN_MSG_ID_UPDATE_REPLY,
                                                 replyTo);
    }
    if (!err) err = alp_ipc_message_pack_uint32 (msg, status);
    if (!err) err = alp_ipc_message_pack_end(msg);

    if (err) {
        ALP_FAIL_NON_FATAL_IF (TRUE, "attn ipc post reply pack failed!");
        if (msg) alp_ipc_message_destroy (msg);
        msg = NULL;
    }

    return msg;
}

/**
 * @brief Create alert delete IPC message
 * @param[in] identifier for alert to be updated
 * @return pointer to message buffer if successful
 * @return NULL if packing of the message buffer failed
 */
AlpMessage *
alp_attn_ipc_pack_alert_delete_msg (const char  *source,
                                    const char  *name,
                                    const char  *handle)
{
    AlpMessage      *msg;
    alp_status_t     err = ALP_STATUS_OK;

    msg = alp_ipc_message_create();
    if (msg == NULL) {
        err = ALP_STATUS_ATTN_IPC_NO_MEMORY;
        ALP_ATLX(IPC, "no memory on message create");
    }

    if (!err) {
        err = alp_ipc_message_pack_start (msg, ALP_ATTN_MSG_ID_DELETE);
    }
    if (!err) err = alp_ipc_message_pack_string (msg, source);
    if (!err) err = alp_ipc_message_pack_string (msg, name);
    if (!err) err = alp_ipc_message_pack_string (msg, handle);
    if (!err) err = alp_ipc_message_pack_end(msg);

    if (err) {
        ALP_FAIL_NON_FATAL_IF (TRUE,
                               "pack failed: alert %s|%s|%s err %x",
                               source,
                               name,
                               handle,
                               err);
        if (msg) alp_ipc_message_destroy (msg);
        msg = NULL;
    }

    return msg;
}

/**
 * @brief Create alert delete IPC message
 * @param[in] replyTo original message being replied to
 * @param[in] status
 * @return pointer to message buffer if successful
 * @return NULL if packing of the message buffer failed
 */
AlpMessage *
alp_attn_ipc_pack_alert_delete_reply (AlpMessage *replyTo, guint32  status)
{
    AlpMessage      *msg;
    alp_status_t     err = ALP_STATUS_OK;

    msg = alp_ipc_message_create();
    if (msg == NULL) {
        err = ALP_STATUS_ATTN_IPC_NO_MEMORY;
        ALP_ATLX(IPC, "no memory on message create");
    }

    if (!err) {
        err =
            alp_ipc_message_response_pack_start (msg,
                                                 ALP_ATTN_MSG_ID_DELETE_REPLY,
                                                 replyTo);
    }
    if (!err) err = alp_ipc_message_pack_uint32 (msg, status);
    if (!err) err = alp_ipc_message_pack_end(msg);

    if (err) {
        ALP_FAIL_NON_FATAL_IF (TRUE, "attn ipc delete reply pack failed!");
        if (msg) alp_ipc_message_destroy (msg);
        msg = NULL;
    }
    return msg;
}

/**
 * @brief Create status IPC reply message
 * @param[in] replyTo original message being replied to
 * @param[in] srv_status server status string
 * @return pointer to message buffer if successful
 * @return NULL if packing of the message buffer failed
 */
AlpMessage *
alp_attn_ipc_pack_status_reply (AlpMessage *replyTo, guint32 status, gchar *srv_status)
{
    AlpMessage      *msg;
    alp_status_t     err = ALP_STATUS_OK;

    msg = alp_ipc_message_create();
    if (msg == NULL) {
        err = ALP_STATUS_ATTN_IPC_NO_MEMORY;
        ALP_ATLX(IPC, "no memory on message create");
    }

    if (!err) {
        err =
            alp_ipc_message_response_pack_start (msg,
                                                 ALP_ATTN_MSG_ID_STATUS_REPLY,
                                                 replyTo);
    }
    if (!err) err = alp_ipc_message_pack_uint32 (msg, status);
    if (!err) err = alp_ipc_message_pack_string (msg, srv_status);
    if (!err) err = alp_ipc_message_pack_end(msg);

    if (err) {
        ALP_FAIL_NON_FATAL_IF (TRUE, "attn ipc post reply pack failed!");
        if (msg) alp_ipc_message_destroy (msg);
        msg = NULL;
    }
    return msg;
}

/***** UNPACKING ROUTINES *******/
alp_status_t
alp_attn_ipc_unpack_properties (AlpMessage  *message,
                                guint32     *nprops,
                                AlpDmlItemH *propListH)
{
    alp_status_t err;
    guint32      count;
    guint32      i;
    gchar  *pname;
    guchar *pvalue;
    guint32 psize;
    guint32 ptype;
    AlpDmlItemH  prop = NULL;
    AlpDmlItemH  last_prop = NULL;
    AlpDmlItemH  prop_list = NULL;

    ALP_ATLX(IPC, "unpack properties msg %x &nprops %x propListH %x",
             message, nprops, propListH);
    if (message == NULL || nprops == NULL || propListH == NULL)
        return ALP_STATUS_ATTN_IPC_BAD_PARAM;

    err = alp_ipc_message_unpack_uint32 (message, &count);
    if (err) {
        ALP_ATLX(IPC, "unpack nprops err %x count %d", err, count);
        return ALP_STATUS_ATTN_IPC_INTERNAL;
    }

    if (count) {
        alp_attn_dml_item_create (PropertyListItem, &prop_list);

        for (i = 0; i < count; i++) {
            /* unpack from the IPC message */
            err = alp_ipc_message_unpack_string (message, &pname);
            if (!err)
                err = alp_ipc_message_unpack_uint32 (message, &ptype);
            if (!err)
                err = alp_ipc_message_unpack_byte_array (message,
                                                         &pvalue, &psize);

            /* create a property item to store the data in */
            if (!err)
                err = alp_attn_dml_item_create (PropertyItem, &prop);
            if (!err)
                err = alp_attn_dml_item_set_string (prop, Name, pname);
            if (!err)
                err = alp_attn_dml_item_set_uint32 (prop, PropType, &ptype);
            if (!err)
                err = alp_attn_dml_item_set_blob (prop, Value, pvalue, psize);
            g_free (pname);
            g_free (pvalue);

            /* add the new property to the property list */
            if (!err) err = alp_attn_dml_item_add_property (prop_list, prop);
            if (err) break;
        }
    }

    if (err == ALP_STATUS_OK) {
        *propListH = prop_list;
        *nprops = count;
    } else {
        /* free any items on prop list */
        if (prop_list) {
            /*
             *  Delete the property list and all the properties on it.
             */
            (void) alp_attn_dml_item_destroy (prop_list);
        }
        /* translate error code to an IPC specific code */
    }
    return err;
}

alp_status_t
alp_attn_ipc_unpack_alert_post_msg (AlpMessage  *message,
                                    AlpDmlItemH *itemH)
{
    alp_status_t err       = ALP_STATUS_OK;
    gchar       *source    = NULL;
    gchar       *name      = NULL;
    gchar       *handle    = NULL;
    gint32       priority;
    gint32       duration;
    guint32      nprops;
    AlpDmlItemH  alertH    = NULL;
    AlpDmlItemH  propListH = NULL;

    if (message == NULL || itemH == NULL) {
        err = ALP_STATUS_ATTN_BAD_PARAM;
        ALP_ATLX(IPC, "bad parameters: %x", err);
        return err;
    }

    /* unpack alert msg data */
    if (!err) err = alp_ipc_message_unpack_start (message);

    if (!err) err = alp_ipc_message_unpack_string (message, &source);
    if (!err) err = alp_ipc_message_unpack_string (message, &name);
    if (!err) err = alp_ipc_message_unpack_string (message, &handle);
    if (!err) err = alp_ipc_message_unpack_int32 (message, &priority);
    if (!err) err = alp_ipc_message_unpack_int32 (message, &duration);
    if (!err) err = alp_attn_ipc_unpack_properties (message, &nprops, &propListH);

    if (!err) err = alp_ipc_message_unpack_end(message);

    /* create an AlertItem and store the data in it */
    if (!err) err = alp_attn_dml_item_create (AlertItem, &alertH);
    if (!err) err = alp_attn_dml_item_set_string (alertH, Source, source);
    if (!err) err = alp_attn_dml_item_set_string (alertH, Name, name);
    if (!err) err = alp_attn_dml_item_set_string (alertH, Handle, handle);
    if (!err) err = alp_attn_dml_item_set_int32  (alertH, Priority, &priority);
    if (!err) err = alp_attn_dml_item_set_int32  (alertH, Duration,  &duration);
    ALP_ATLX(IPC, "add duration %d: %x", duration, err);

    if (propListH) {  /* if properties present, attach to the alert item */
        if (!err) err = alp_attn_dml_item_set_property_list (alertH, propListH);
        ALP_ATLX(IPC, "set prop list: %x", err);
        alp_attn_dml_item_destroy (propListH);
    }

    g_free(source);
    g_free(name);
    g_free(handle);

    if (err == ALP_STATUS_OK) {
        *itemH = alertH;
    } else {
        if (alertH) {
            alp_attn_dml_item_destroy (alertH);
        } else if (propListH) {
            alp_attn_dml_item_destroy (propListH);
        }
        ALP_FAIL_NON_FATAL_IF (TRUE,
                               "unpack of alert post msg failed: %x!",
                               err);
    }

    return err;
}

alp_status_t
alp_attn_ipc_unpack_alert_post_reply (AlpMessage   *response,
                                      alp_status_t *status,
                                      AlpAttnId    *alertId)
{
    alp_status_t  err;
    uint32_t      id;
    uint32_t      err_status;

    if (response == NULL) {
        return ALP_STATUS_ATTN_BAD_PARAM;
    }

    err = alp_ipc_message_unpack_start (response);
    if (!err) err = alp_ipc_message_unpack_uint32 (response, &err_status);
    if (!err) err = alp_ipc_message_unpack_uint32 (response, &id);
    if (!err) err = alp_ipc_message_unpack_end(response);

    if (err == ALP_STATUS_OK) {
        *status  = (alp_status_t) err_status;
        if (alertId) {
            *alertId = (AlpAttnId) id;
        }
    } else {
        ALP_FAIL_NON_FATAL_IF (TRUE, "attn ipc unpost pack resp failed!");
    }

    return err;
}

alp_status_t
alp_attn_ipc_unpack_alert_update_msg (AlpMessage    *message,
                                      AlpDmlItemH   *itemH)
{
    alp_status_t err       = ALP_STATUS_OK;
    gchar       *source    = NULL;
    gchar       *name      = NULL;
    gchar       *handle    = NULL;
    guint32      nprops;
    AlpDmlItemH  propListH = NULL;

    if (message == NULL || itemH == NULL) {
        return ALP_STATUS_ATTN_BAD_PARAM;
    }

    if (!err) err = alp_ipc_message_unpack_start (message);

    if (!err) err = alp_ipc_message_unpack_string (message, &source);
    if (!err) err = alp_ipc_message_unpack_string (message, &name);
    if (!err) err = alp_ipc_message_unpack_string (message, &handle);
    if (!err) err = alp_attn_ipc_unpack_properties (message, &nprops,
                                                    &propListH);

    if (!err) err = alp_ipc_message_unpack_end(message);

    if (!err) err = alp_attn_dml_item_set_string (propListH, Source, source);
    if (!err) err = alp_attn_dml_item_set_string (propListH, Name,   name);
    if (!err) err = alp_attn_dml_item_set_string (propListH, Handle, handle);

    g_free(source);
    g_free(name);
    g_free(handle);

    if (err == ALP_STATUS_OK) {
        *itemH   = propListH;
    } else {
        if (propListH) {
            alp_attn_dml_item_destroy (propListH);
        }
        ALP_FAIL_NON_FATAL_IF (TRUE, "attn ipc unpack update msg failed!");
    }

    return err;
}

alp_status_t
alp_attn_ipc_unpack_alert_update_reply (AlpMessage    *response,
                                        alp_status_t  *status)
{
    alp_status_t  err;
    uint32_t      err_status;

    if (response == NULL || status == NULL) {
        return ALP_STATUS_ATTN_BAD_PARAM;
    }

    err = alp_ipc_message_unpack_start (response);
    if (!err) err = alp_ipc_message_unpack_uint32 (response, &err_status);
    if (!err) err = alp_ipc_message_unpack_end(response);

    if (err == ALP_STATUS_OK) {
        *status  = (alp_status_t) err_status;
    } else {
        ALP_FAIL_NON_FATAL_IF (TRUE, "attn ipc unpack update reply failed!");
    }

    return err;
}

alp_status_t
alp_attn_ipc_unpack_alert_delete_msg (AlpMessage    *message,
                                      AlpDmlItemH   *itemH)
{
    alp_status_t  err;
    gchar        *source    = NULL;
    gchar        *name      = NULL;
    gchar        *handle    = NULL;
    AlpDmlItemH  alertH    = NULL;

    if (message == NULL || itemH == NULL) {
        return ALP_STATUS_ATTN_BAD_PARAM;
    }

    err = alp_ipc_message_unpack_start (message);

    if (!err) err = alp_ipc_message_unpack_string (message, &source);
    if (!err) err = alp_ipc_message_unpack_string (message, &name);
    if (!err) err = alp_ipc_message_unpack_string (message, &handle);

    if (!err) err = alp_ipc_message_unpack_end(message);

    if (!err) err = alp_attn_dml_item_create (AlertItem, &alertH);
    if (!err) err = alp_attn_dml_item_set_string (alertH, Source, source);
    if (!err) err = alp_attn_dml_item_set_string (alertH, Name,   name);
    if (!err) err = alp_attn_dml_item_set_string (alertH, Handle, handle);

    g_free(source);
    g_free(name);
    g_free(handle);

    if (err == ALP_STATUS_OK) {
        *itemH = alertH;
    } else {
        if (alertH) {
            alp_attn_dml_item_destroy (alertH);
        }
        ALP_FAIL_NON_FATAL_IF (TRUE, "attn ipc unpack delete msg failed!");
    }

    return err;
}

alp_status_t
alp_attn_ipc_unpack_alert_delete_reply (AlpMessage    *response,
                                        alp_status_t  *status)
{
    alp_status_t  err;
    uint32_t      err_status;

    if (response == NULL || status == NULL) {
        return ALP_STATUS_ATTN_BAD_PARAM;
    }

    err = alp_ipc_message_unpack_start (response);
    if (!err) err = alp_ipc_message_unpack_uint32 (response, &err_status);
    if (!err) err = alp_ipc_message_unpack_end(response);

    if (err == ALP_STATUS_OK) {
        *status  = (alp_status_t) err_status;
    } else {
        ALP_FAIL_NON_FATAL_IF (TRUE, "attn ipc unpack delete reply failed!");
    }

    return err;
}

alp_status_t
alp_attn_ipc_unpack_status_reply (AlpMessage  *response,
                                  gchar      **srv_status)
{
    alp_status_t  err;
    gchar        *str;

    if (response == NULL || srv_status == NULL) {
        return ALP_STATUS_ATTN_BAD_PARAM;
    }

    err = alp_ipc_message_unpack_start (response);
    if (!err) err = alp_ipc_message_unpack_string (response, &str);
    if (!err) err = alp_ipc_message_unpack_end(response);

    if (err == ALP_STATUS_OK) {
        *srv_status = str;
    } else {
        ALP_FAIL_NON_FATAL_IF (TRUE, "attn ipc unpost pack resp failed!");
    }

    return err;
}

static void
alp_prv_attn_callback_connect_cb (AlpConnection *connection,
                                  gpointer cbData)
{
    alp_status_t  err;
    AlpMessage   *msg = (AlpMessage *) cbData;

    ALP_ATLX(API, "connection ready (%x) - sending callback msg (%x)",
             (unsigned) connection, (unsigned) msg);

    err = alp_ipc_connection_send (connection, msg);
    if (err) {
        ALP_ATLX(API, "send failed: %x", err);
    }
}

alp_status_t
alp_attn_ipc_send_callback (const char *channel,
                            const char *source,
                            const char *name,
                            const char *handle,
                            uint32_t    argc,
                            char       *argv[])
{
    alp_status_t   err = ALP_STATUS_OK;
    AlpMessage *msg;
    AlpConnection *connection;
    uint32_t i;
    char **p;

    /* pack callback */
    msg = alp_ipc_message_create();
    if (msg == NULL) {
        err = ALP_STATUS_ATTN_IPC_NO_MEMORY;
        ALP_ATLX(IPC, "no memory on message create");
        return err;
    }

    ALP_ATLX(IPC, "packing callback into msg %x", (unsigned) msg);
    err = alp_ipc_message_pack_start (msg, 0x6221701);
    if (!err) err = alp_ipc_message_pack_string (msg, source);
    if (!err) err = alp_ipc_message_pack_string (msg, name);
    if (!err) err = alp_ipc_message_pack_string (msg, handle);
    if (!err) err = alp_ipc_message_pack_uint32 (msg, argc);

    p = argv;
    if (p) {
        for (i = 0; *p && i < argc; i++) {
            if (!err) err = alp_ipc_message_pack_string (msg, *p);
            p++;
        }
    }

    if (!err) err = alp_ipc_message_pack_end (msg);

    if (err) {
        alp_ipc_message_destroy (msg);
        ALP_ATLX(IPC, "packing callback message failed with err %x", err);
        return err;
    }

    /* open connection */
    connection = alp_ipc_channel_connect (channel,
                                          ALP_IPC_CONNECTION_USAGE_HINT_NONE,
                                          alp_prv_attn_callback_connect_cb,
                                          msg);
    if (connection == NULL) {
        ALP_ATLX(IPC, "failed to connect to callback channel <%s>",
                 channel);
    } else {
        ALP_ATLX(IPC, "connection established (%x) for channel <%s>",
                 (unsigned) connection, channel);
    }
    return err;
}
