/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
#include <hiker/bundlemgr.h>
#include <hiker/prv/bundlemgr_prv.h>
#include <hiker/ipc.h>
#include <string.h>
#include <stdlib.h>
#include <sqlite3.h>

#include "pm_traces.h"
#include "pm_globals.h"
#include "pm_ipc.h"
#include "pm_sql.h"

AlpBundle alp_bundle_by_name(const char * name)
{
	AlpBundle r = ALP_BUNDLE_NULL;

	S_STMT s;
	char buf[PATH_MAX];
	
	if (!name) {
		set_error(ALP_STATUS_BUNDLEMGR_FAILURE);
		return ALP_BUNDLE_NULL;
	}

	// trim trailing path or '?' off of name, to reduce it to simplest "scheme:id" URL form
	int len = strcspn(name, "/?");
	if (name[len] != '\0') {
		memcpy(buf, name, len);
		buf[len] = '\0';
		name = buf;
	}

	S_PREPARE(s, "SELECT containers.package_id FROM containers INNER JOIN names ON (containers.package_id=names.package_id) WHERE names.name = ?1");
	S_BIND_TEXT(s, 1, name);
	S_BEGIN_SELECT(s) {
		r.value = S_COL_INT(s, 0);
	}
	S_END_SELECT(s);
	
	return r;
}

GArray* alp_bundle_search(	const char * name,
				gboolean  	named,
				const char *  	scheme,
				const char *  	store,
				const char *  	property) 
{
	GArray * result = g_array_new(TRUE, FALSE, sizeof(AlpBundle));
	
	if (!result) {
		set_error(ALP_STATUS_BUNDLEMGR_FAILURE);
		return NULL;
	}

	if (name) {
		char namebuf[PATH_MAX];
		// trim trailing path or '?' off of name, to reduce it to simplest "scheme:id" URL form
		int len = strcspn(name, "/?");	
		if (name[len] != '\0') {
			memcpy(namebuf, name, len);
			namebuf[len] = '\0';
			name = namebuf;
		}
	}

	S_STMT s;

	// note that wildcarding is built into the SQL select statement, instead of having variant statements, or constructing
	// the statement dynamically. Once we pre-prepare the statements, we should compare the speed of the single statement to
	// purpose-built statements.
	S_PREPARE(s, "SELECT DISTINCT containers.package_id "
	             " FROM containers LEFT OUTER JOIN names ON (containers.package_id=names.package_id) "
	             " LEFT OUTER JOIN properties ON (containers.package_id=properties.package_id) "
	             " LEFT OUTER JOIN schemes ON (containers.scheme_id=schemes.scheme_id) "
	             " LEFT OUTER JOIN stores ON (containers.store_id=stores.store_id) "
	             " WHERE (?1 ISNULL OR names.name = ?1) AND (?2 ISNULL OR names.name NOTNULL) "
	             "  AND (?3 ISNULL OR schemes.scheme_name = ?3) AND "
	             "  (?4 ISNULL OR stores.store_name = ?4) AND (?5 ISNULL OR properties.property = ?5)");
	if (name)
		S_BIND_TEXT(s, 1, name);
	if (named)
		S_BIND_INT(s, 2, 1);
	if (scheme)
		S_BIND_TEXT(s, 3, scheme);
	if (store)
		S_BIND_TEXT(s, 4, store);
	if (property)
		S_BIND_TEXT(s, 5, property);
		
	S_BEGIN_SELECT(s) {
		AlpBundle v;
		v.value = S_COL_INT(s, 0);
		g_array_append_val(result, v);
	}
	S_END_SELECT(s);
	
	return result;
}

GList * alp_bundle_list_names()
{
	GList * results = NULL;
	
	S_STMT s;
	S_PREPARE(s, "SELECT names.name FROM names ORDER BY name");
	S_BEGIN_SELECT(s) {
		results = g_list_append(results, g_strdup(S_COL_TEXT(s, 0)));
	}
	S_END_SELECT(s);

	return results;	
}

GList * alp_bundle_list_schemes()
{
	GList * results = NULL;
	
	S_STMT s;
	S_PREPARE(s, "SELECT schemes.scheme_name FROM schemes ORDER BY scheme_id");
	S_BEGIN_SELECT(s) {
		results = g_list_append(results, g_strdup(S_COL_TEXT(s, 0)));
	}
	S_END_SELECT(s);

	return results;	
}

GList * alp_bundle_list_stores()
{
	GList * results = NULL;
	
	S_STMT s;
	S_PREPARE(s, "SELECT stores.store_name FROM stores ORDER BY store_id");
	S_BEGIN_SELECT(s) {
		results = g_list_append(results, g_strdup(S_COL_TEXT(s, 0)));
	}
	S_END_SELECT(s);

	return results;	
}

GList * alp_bundle_list_properties()
{
	GList * results = NULL;
	
	S_STMT s;
	S_PREPARE(s, "SELECT DISTINCT properties.property FROM properties");
	S_BEGIN_SELECT(s) {
		results = g_list_append(results, g_strdup(S_COL_TEXT(s, 0)));
	}
	S_END_SELECT(s);

	return results;	
}


