/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <hiker/bundlemgr.h>
#include <hiker/prv/bundlemgr_prv.h>

AlpBundle find_bundle(const char * bundle)
{
  // first see if we were given a bundle index -- a non-zero value
  AlpBundle p = {atoi(bundle)};
  
  // next try looking up bundle by name
  if (!p.value) {
    p = alp_bundle_by_name(bundle);
  }

  // then try finding a bundle mounted from a given filename
  if (!p.value) {
    GArray * all;
    
    all = alp_bundle_search(NULL, FALSE, NULL, NULL, NULL);
    
    unsigned int i;
    for (i=0; i < all->len; i++) {
      gchar * prop = alp_bundle_metadata_value(g_array_index(all, AlpBundle, i), "bar_filename");
      if (prop && strcmp(prop, bundle) == 0) {
        p = g_array_index(all, AlpBundle, i);
        g_free(prop);
        break;
      }
      g_free(prop);
    }
    
    g_array_free(all, TRUE);
  }
  
  return p;
}

int cmd_names(void)
{
  alp_bundle_init();

  GArray * bundles = alp_bundle_search(NULL, TRUE, NULL, NULL, NULL);
  unsigned int i;
  
  if (!bundles) {
    printf("Error retrieving named bundles\n");
    return 1;
  }
  
  for (i=0;i < bundles->len ; i++) {
    gchar * name = alp_bundle_metadata_value(g_array_index(bundles, AlpBundle, i), "name");
    printf("%s\n", name);
    g_free(name);
  }
  g_array_free(bundles, TRUE);
  return 0;
}

int cmd_archive(const char * bundle, const char * file)
{
  alp_bundle_init();

  AlpBundle p = find_bundle(bundle);

  if (!p.value) {
    printf("Error locating bundle to archive\n");
    return 1;
  }
  
  GIOChannel * channel = g_io_channel_new_file(file, "w", NULL);

  if (!channel) {
    fprintf(stderr, "Error opening output file \"%s\": %s\n", file, strerror(errno));
    return 1;
  }

  g_io_channel_set_encoding(channel, NULL, NULL);

  alp_status_t result = alp_bundle_image_archive(p, NULL, NULL, channel, NULL);
  
  if (result != ALP_STATUS_OK) {
    printf("Error producing archive image\n");
    return 1;
  }
  
  g_io_channel_unref(channel);
  
  return 0;
}

int cmd_install(const char * file)
{
  alp_bundle_init();

  GIOChannel * channel = g_io_channel_new_file(file, "r", NULL);
  
  if (!channel) {
    fprintf(stderr, "Error opening input file \"%s\": %s\n", file, strerror(errno));
    return 1;
  }

  g_io_channel_set_encoding(channel, NULL, NULL);

  // Note that provide_filename argument is not used to read the contents of the
  // image to install: that comes from the channel. provide_filename is going giving
  // a name to deduce information from.
  AlpBundle result = alp_bundle_image_install(file /* provide_filename */, NULL, channel, NULL, ALP_BUNDLE_FLAG_OVERWRITE_DUPLICATE, NULL, NULL, NULL);
  
  if (alp_bundle_is_NULL(result)) {
    fprintf(stderr, "Error installing image\n");
    return 1;
  }
  
  g_io_channel_unref(channel);
  
  return 0;
}

int cmd_move_to_store(const char * bundle, const char * store)
{
  alp_bundle_init();

  AlpBundle p = find_bundle(bundle);

  if (!p.value) {
    printf("Error locating bundle to move\n");
    return 1;
  }

  AlpBundle result = alp_bundle_move_to_store(p, store, ALP_BUNDLE_FLAG_OVERWRITE_DUPLICATE);
  
  if (alp_bundle_is_NULL(result)) {
    printf("Error moving bundle\n");
    return 1;
  }
  
  return 0;
}

int cmd_copy_to_store(const char * bundle, const char * store)
{
  alp_bundle_init();

  AlpBundle p = find_bundle(bundle);
  
  if (!p.value) {
    printf("Error locating bundle to copy\n");
    return 1;
  }

  AlpBundle result = alp_bundle_copy_to_store(p, store, ALP_BUNDLE_FLAG_OVERWRITE_DUPLICATE);
  
  if (alp_bundle_is_NULL(result)) {
    printf("Error copying bundle\n");
    return 1;
  }
  
  return 0;
}

int cmd_ro_path(const char * bundle, const char * path)
{
  alp_bundle_init();

  AlpBundle p = find_bundle(bundle);

  if (!p.value) {
    fprintf(stderr, "Error locating bundle to get ro path for\n");
    return 1;
  }
  
  AlpBundleRef ref = alp_bundle_open(p);
  
  if (!ref) {
    fprintf(stderr, "Error opening bundle\n");
    return 1;
  }
  
  gchar * result = alp_bundle_ref_ro_pathname(ref, path);
  
  printf("%s", result);
  
  g_free(result);
  
  alp_bundle_close(ref);
  
  return 0;
}

int cmd_rw_path(const char * bundle, const char * path)
{
  alp_bundle_init();

  AlpBundle p = find_bundle(bundle);

  if (!p.value) {
    fprintf(stderr, "Error locating bundle to get rw path for\n");
    return 1;
  }
  
  gchar * result = alp_bundle_rw_pathname(p, path, TRUE);
  
  printf("%s", result);
  
  g_free(result);
  
  return 0;
}


int cmd_register(const char * file, const char * store, char * scheme)
{
  alp_bundle_init();
  
  AlpBundle result = alp_prv_bundle_register_bundle(file, scheme ?: "bar", store ?: "internal", 0);
  
  if (alp_bundle_is_NULL(result)) {
    fprintf(stderr, "Error registering bundle\n");
    return 1;
  }
  
  return 0;
}

int cmd_delete(const char * bundle, gboolean everything)
{
  alp_bundle_init();

  AlpBundle p = find_bundle(bundle);
  
  if (!p.value) {
    printf("Error locating bundle to delete\n");
    return 1;
  }
  
  alp_status_t result = alp_bundle_delete(p, everything);
  
  if (result != ALP_STATUS_OK) {
    printf("Error deleting bundle\n");
    return 1;
  }
  
  return 0;
}

int cmd_incref(const char * bundle)
{
  alp_bundle_init();

  AlpBundle p = find_bundle(bundle);

  if (!p.value) {
    printf("Error locating bundle to incref\n");
    return 1;
  }
  
  alp_status_t result = alp_prv_bundle_incref(p, 0);
  
  if (result != ALP_STATUS_OK) {
    printf("Error increfing bundle\n");
    return 1;
  }
  
  return 0;
}

int cmd_decref(const char * bundle)
{
  alp_bundle_init();

  AlpBundle p = find_bundle(bundle);
  
  if (!p.value) {
    printf("Error locating bundle to decref\n");
    return 1;
  }
  
  alp_status_t result = alp_prv_bundle_decref(p, 0);
  
  if (result != ALP_STATUS_OK) {
    printf("Error decrefing bundle\n");
    return 1;
  }
  
  return 0;
}

int cmd_unregister(const char * bundle)
{
  alp_bundle_init();

  AlpBundle p = find_bundle(bundle);
  
  if (!p.value) {
    printf("Error locating bundle to unregister\n");
    return 1;
  }
  
  alp_status_t result = alp_prv_bundle_unregister_bundle(p, 0);
  
  if (result != ALP_STATUS_OK) {
    printf("Error unregistering bundle\n");
    return 1;
  }
  
  return 0;
}

int cmd_reregister(const char * bundle)
{
  alp_bundle_init();

  AlpBundle p = find_bundle(bundle);
  
  if (!p.value) {
    printf("Error locating bundle to reregister\n");
    return 1;
  }
 
  gchar * bar_filename = alp_bundle_metadata_value(p, "bar_filename");
  gchar * scheme_name = alp_bundle_metadata_value(p, "scheme_name");
  gchar * store_name = alp_bundle_metadata_value(p, "store_name");
  
  if (!bar_filename) {
    printf("Unable to record underlying filename for bundle to reregister\n");
    return 1;
  }

  if (!scheme_name) {
    printf("Unable to record scheme for bundle to reregister\n");
    return 1;
  }

  if (!store_name) {
    printf("Unable to record store for bundle to reregister\n");
    return 1;
  }
  
  alp_status_t result = alp_prv_bundle_unregister_bundle(p, 0);
  
  if (result != ALP_STATUS_OK) {
    printf("Error reregistering bundle -- cannot unregister bundle\n");
    return 1;
  }
  
  AlpBundle p2 = alp_prv_bundle_register_bundle(bar_filename, scheme_name, store_name, 0);

  if (alp_bundle_is_NULL(p2)) {
    printf("Error reregistering bundle -- cannot register bundle\n");
    return 1;
  }
  
  
  return 0;
}

static void helper(gpointer key, gpointer value, gpointer data)
{
  printf("%s: %s\n", (char*)key, (char*)value);
}

int cmd_metadata(const char * bundle)
{
  alp_bundle_init();

  AlpBundle p = find_bundle(bundle);
  
  if (!p.value) {
    printf("Error locating bundle to display metadata for\n");
    return 1;
  }

  GHashTable * result = alp_bundle_metadata(p);
  
  if (result == NULL) {
    printf("Error displaying metadata for bundle\n");
    return 1;
  }
  
  g_hash_table_foreach(result, helper, NULL);
  
  return 0;
}

int cmd_dump()
{
  alp_bundle_init();

  alp_prv_bundle_dump();
  
  return 0;
}

int cmd_scan(const char * store)
{
  alp_bundle_init();

  alp_status_t result = alp_prv_bundle_scan_store(store, 0, 0);
  
  if (result != ALP_STATUS_OK) {
    printf("Error scanning store\n");
  }
  
  return 0;
}

int cmd_add_store(const char * store, const char ** arguments)
{
  alp_bundle_init();

  alp_status_t result = alp_prv_bundle_add_store(store, arguments);
  
  if (result != ALP_STATUS_OK) {
    printf("Error adding store\n");
  }
  
  return 0;
}

int cmd_delete_store(const char * store)
{
  alp_bundle_init();

  alp_status_t result = alp_prv_bundle_delete_store(store);
  
  if (result != ALP_STATUS_OK) {
    printf("Error deleting store\n");
  }
  
  return 0;
}

int cmd_list_stores(void)
{
  alp_bundle_init();
  
  GList * stores = alp_bundle_list_stores();
  while (stores) {
    printf("%s\n", (char*)stores->data);
    stores = g_list_next(stores);
  }
  
  return 0;
}

void dump_hash(gpointer key, gpointer value, gpointer data)
{
  printf("%s: %s\n", (char*)key, (char*)value);
}

int cmd_store_info(const char * store)
{
  alp_bundle_init();
  
  GHashTable * info = alp_bundle_store(store);

  if (!info) {
    printf("Error getting store info\n");
    return 1;
  }
  
  g_hash_table_foreach(info, dump_hash, NULL);
  
  return 0;
}

int cmd_wipe()
{
  alp_status_t result;

  alp_bundle_init();
  
  result = alp_prv_bundle_wipe();

  if (result != ALP_STATUS_OK) {
    printf("Error wiping bundle manager\n");
    return 1;
  }
  
  return 0;
}

int cmd_rescan_volumes()
{
  alp_status_t result;

  alp_bundle_init();
  
  result = alp_prv_bundle_rescan_volumes();

  if (result != ALP_STATUS_OK) {
    printf("Error rescanning volumes\n");
    return 1;
  }
  
  return 0;
}

int cmd_hold_notifications(gboolean hold)
{
  alp_status_t result;

  alp_bundle_init();
  
  result = alp_prv_bundle_hold_notifications(hold);

  if (result != ALP_STATUS_OK) {
    printf("Error setting notification hold\n");
    return 1;
  }
  
  return 0;
}

int cmd_present(char * argv[], int argc)
{
  alp_status_t result;

  alp_bundle_init();
  
  int i;
  int app = 0;
  
  if ((argc > 0) && (
        (strcmp(argv[0], "--app") == 0) || 
        (strcmp(argv[0], "-app") == 0)
      )) {
    app = 1;
    argc--;
    argv++;
  }
  
  for (i=0;i<argc;i++) {
    GArray * ary = alp_bundle_search(argv[i], TRUE, NULL /* scheme */, NULL /* store */, app ? "application" : NULL);

    if (ary && ary->len > 0) {
      printf("%s", argv[i]);
      g_array_free(ary, TRUE);
      return 0;
    }
    
    if (ary)
      g_array_free(ary, TRUE);
  }
  
  return 1;
}

int cmd_run_with(char ** argv)
{
  alp_bundle_init();
  
  while (argv[0] && argv[1] && ((strcmp(argv[0], "-run-with") == 0) || (strcmp(argv[0], "--run-with") == 0) )) {
    AlpBundle b = alp_bundle_by_name(argv[1]);
    
    if (alp_bundle_is_NULL(b)) {
      fprintf(stderr, "Unable to find bundle %s\n", argv[1]);
      exit(1);
    }
    AlpBundleRef r = alp_bundle_open(b);
    if (!r) {
      fprintf(stderr, "Unable to open bundle %s\n", argv[1]);
      exit(1);
    }
    argv += 2;
  }
  
  if (fork() == 0) {
    int result = execvp(argv[0], argv);
  
    if (result == -1) {
      fprintf(stderr, "Unable to run program %s: %s\n", argv[0], strerror(errno));
      exit(2);
    }
    
    abort();
  } else {
    int result;
    wait(&result);
    
    if (WIFSIGNALED(result)) {
      kill(getpid(), WTERMSIG(result));
    } else {
      return WEXITSTATUS(result);
    }
  }
  
  
  /* NOT REACHED */
  
  return 0;
}


int cmd_wait()
{
  alp_status_t result;
  
  result = alp_prv_bundle_init(FALSE, NULL, NULL, NULL, NULL, NULL, TRUE);
  
  while (result != ALP_STATUS_OK) {
    struct timespec t;
    t.tv_sec = 0;
    t.tv_nsec = 1000000000 / 4;
    nanosleep(&t, NULL);

    result = alp_prv_bundle_init(FALSE, NULL, NULL, NULL, NULL, NULL, TRUE);

    if (isatty(fileno(stdout))) {
      printf(".");
      fflush(stdout);
    }
  }
  
  return 0;
}



char * argv0;

int usage()
{
  printf("Usage: %s [--OPTION ...]\n", argv0);
  printf("Available Options:\n");
  printf("    --list (or --names)          # Show all bundle names\n");
  printf("    --archive <bundle> <file>   # Archive bundle (by index or name) to file\n");
  printf("    --install <file>            # Create bundle from file\n");
  printf("    --ro_path <bundle> [<path>] # Returns a R/O path for bundle resource\n");
  printf("    --rw_path <bundle> [<path>] # Returns a R/W path for bundle data\n");
  printf("    --delete <bundle>           # Delete a bundle\n");
  printf("    --delete-all <bundle>       # Delete a bundle and associated data\n");
  printf("    --register <file> [<store>] # Register a file or folder with bundle manager\n");
  printf("    --unregister <bundle>       # Unregister a bundle\n");
  printf("    --reregister <bundle>       # Unregister and then re-register a bundle\n");
  printf("    --metadata <bundle>         # List metadata for bundle\n");
  printf("    --dump                      # Dump out bundle db\n");
  printf("    --scan <store>              # Scan/rescan entire store, reg/unreg as needed\n");
  printf("    --copy-to-store <bundle> <store> # Copy existing bundle to a different store\n");
  printf("    --move-to-store <bundle> <store> # Move existing bundle to a different store\n");
  printf("    --add-store <store-name> <arg-name> <arg-value> ... # Add a new store\n");
  printf("    --delete-store <store-name> # Delete a store\n");
  printf("    --stores                    # List stores\n");
  printf("    --store-info <store-name>   # Retrieve information about a store\n");
  printf("    --wait                      # Wait for bundle manager server to be available\n");
  printf("    --incref <bundle>           # Open the bundle, pinning it\n");
  printf("    --decref <bundle>           # Close the bundle, possibly unpinning it\n");
  printf("    --wipe                      # Remove all bundle info; doesn't touch bundles\n");
  printf("    --hold-notifications        # Tell bundlemgrd to not send out notifications\n");
  printf("    --release-notifications     # Send notifications again, and send out any pending\n");
  printf("    --run-with <bundle> [--run-with <bundle> ...] program <args...> # Run program while bundles are open\n");
  printf("    --present [--app] <bundle> [ <bundle> ... ] # Print out name of first bundle which is present\n");
  return 2;
}

int main(int argc, char *argv[])
{
  argv0 = argv[0];
  
  if (argc < 2)
    return usage();
    
  if (argv[1][0] == '-' && argv[1][1] == '-')
    argv[1]++; // ignore first dash on two-dash options
    
  if (strcmp(argv[1], "-names") == 0 && argc == 2)
    return cmd_names();
  else if (strcmp(argv[1], "-list") == 0 && argc == 2)
    return cmd_names();
  else if (strcmp(argv[1], "-archive") == 0 && argc == 4)
    return cmd_archive(argv[2], argv[3]);
  else if (strcmp(argv[1], "-install") == 0 && argc == 3)
    return cmd_install(argv[2]);
  else if (strcmp(argv[1], "-delete") == 0 && argc == 3)
    return cmd_delete(argv[2], 0);
  else if (strcmp(argv[1], "-delete-all") == 0 && argc == 3)
    return cmd_delete(argv[2], 1);
  else if (strcmp(argv[1], "-copy-to-store") == 0 && argc == 4)
    return cmd_copy_to_store(argv[2], argv[3]);
  else if (strcmp(argv[1], "-move-to-store") == 0 && argc == 4)
    return cmd_move_to_store(argv[2], argv[3]);
  else if (strcmp(argv[1], "-ro-path") == 0 && (argc == 3 || argc == 4))
    return cmd_ro_path(argv[2], argv[3] ?: "");
  else if (strcmp(argv[1], "-rw-path") == 0 && (argc == 3 || argc == 4))
    return cmd_rw_path(argv[2], argv[3] ?: "");
  else if ((strcmp(argv[1], "-incref") == 0 || strcmp(argv[1], "-pin") == 0) && argc == 3)
    return cmd_incref(argv[2]);
  else if ((strcmp(argv[1], "-decref") == 0 || strcmp(argv[1], "-unpin") == 0) && argc == 3)
    return cmd_decref(argv[2]);
  else if (strcmp(argv[1], "-register") == 0 && (argc == 5 || argc == 4 || argc == 3))
    return cmd_register(argv[2], (argc>3) ? argv[3] : NULL, (argc>4) ? argv[4] : NULL);
  else if (strcmp(argv[1], "-unregister") == 0 && argc == 3)
    return cmd_unregister(argv[2]);
  else if (strcmp(argv[1], "-reregister") == 0 && argc == 3)
    return cmd_reregister(argv[2]);
  else if (strcmp(argv[1], "-metadata") == 0 && argc == 3)
    return cmd_metadata(argv[2]);
  else if (strcmp(argv[1], "-dump") == 0 && argc == 2)
    return cmd_dump();
  else if ((strcmp(argv[1], "-scan") == 0 || strcmp(argv[1], "-rescan") == 0) && argc == 3)
    return cmd_scan(argv[2]);
  else if (strcmp(argv[1], "-add-store") == 0 && argc > 3)
    return cmd_add_store(argv[2], (const char**)argv+3);
  else if (strcmp(argv[1], "-delete-store") == 0 && argc == 3)
    return cmd_delete_store(argv[2]);
  else if (strcmp(argv[1], "-stores") == 0 && argc == 2)
    return cmd_list_stores();
  else if (strcmp(argv[1], "-store-info") == 0 && argc == 3)
    return cmd_store_info(argv[2]);
  else if (strcmp(argv[1], "-wipe") == 0 && argc == 2)
    return cmd_wipe();
  else if (strcmp(argv[1], "-wait") == 0 && argc == 2)
    return cmd_wait();
  else if (strcmp(argv[1], "-rescan-volumes") == 0 && argc == 2)
    return cmd_rescan_volumes();
  else if (strcmp(argv[1], "-hold-notifications") == 0 && argc == 2)
    return cmd_hold_notifications(1);
  else if (strcmp(argv[1], "-release-notifications") == 0 && argc == 2)
    return cmd_hold_notifications(0);
  else if (strcmp(argv[1], "-run-with") == 0 && argc > 3)
    return cmd_run_with(argv+1);
  else if (strcmp(argv[1], "-present") == 0 && argc >= 3)
    return cmd_present(argv+2,argc-2);
  else
    return usage();
}


