/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
#include <hiker/bundlemgr.h>
#include <hiker/prv/bundlemgr_prv.h>
#include <hiker/notifymgr.h>
#include <hiker/ipc.h>
#include <string.h>
#include <stdlib.h>
#include <sqlite3.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <glib/gthread.h>

#include "pm_traces.h"
#include "pm_globals.h"
#include "pm_ipc.h"
#include "pm_sql.h"
#include "pm_perms.h"
#include "pm_msg.h"
#include "ghost_support.h"

sqlite3 * db;

int pm_status = STATUS_UNINIT;
int pm_open_count = 0;

AlpChannel * pm_server_channel;
AlpConnection * pm_client_connection;
GHashTable * gConnectionTable;
gboolean gNotify;
AlpGlobalSettingsContext * gGSContext;

const char * gLiveDB;

const char * gBundleWorkPath;
const char * gBundleRWPath;

static void setup_live_metadata(const char * live_db_filename, int erase, int create);
static void create_view(void);

static void free_it(gpointer data)
{
	g_array_free((GArray*)data, TRUE);
}

alp_status_t alp_bundle_init(void)
{
	if (pm_open_count == 0) {
		// propagate potential error
		return alp_prv_bundle_init(FALSE, NULL, NULL, NULL, NULL, NULL, TRUE);
	} else if (pm_open_count > 0) {
		pm_open_count++;
		return ALP_STATUS_OK;
	} else {
		ALP_TMTL("Attempted to open package many afer closing too many times");
		return set_error(ALP_STATUS_BUNDLEMGR_FAILURE);
	}
}

alp_status_t alp_bundle_fini(void)
{
	if (pm_open_count > 1) {
		pm_open_count--;
		return ALP_STATUS_OK;
	} else if (pm_open_count == 1) {
		// propagate potential error
		return alp_prv_bundle_fini();
	} else {
		ALP_TMTL("Attempted to close package manager too many times");
		return set_error(ALP_STATUS_BUNDLEMGR_FAILURE);
	}
}

alp_status_t alp_prv_bundle_init_playground(gboolean server, const char * playground_name, gboolean notify)
{
	char * dir = g_strdup_printf("/tmp/bundle-%s", playground_name);
	mkdir_p(dir, 0666);
	char * database = g_strdup_printf("%s/database.db", dir);
	char * live_database = g_strdup_printf("%s/live_database.db", dir);
	char * ipc_channel = g_strdup_printf("alp-bundlemgr-private-%s", playground_name);
	char * work_path = g_strdup_printf("%s/work/", dir);
	char * rw_path = g_strdup_printf("%s/rw/", dir);

	alp_status_t result = alp_prv_bundle_init(server, database, live_database, ipc_channel, work_path, rw_path, notify);
fprintf("result is %d\n", result);//???
	g_free(database);
	g_free(ipc_channel);
	g_free(work_path);
	g_free(rw_path);
	g_free(dir);

	return set_error(result);
}

alp_status_t alp_prv_bundle_init(gboolean server, const char * database, const char * live_database, const char * ipc_channel, const char * work_path, const char * rw_path, gboolean notify)
{
	char db_path[PATH_MAX];
	
	if (pm_status != STATUS_UNINIT) {
		ALP_TMTL("Error: already initialized");
		return set_error(ALP_STATUS_BUNDLEMGR_FAILURE);
	}
	
	if (!work_path)
		work_path = DEFAULT_BUNDLE_WORK_PATH;
	
	if (!rw_path)
		rw_path = DEFAULT_BUNDLE_RW_PATH;
	
	if (!ipc_channel)
		ipc_channel = DEFAULT_IPC_CHANNEL;
	
	if (!database) 
		database = DEFAULT_DATABASE;
	
	if (!live_database)
		live_database = DEFAULT_LIVE_DATABASE;
	
	gNotify = notify;
	gBundleWorkPath = g_strdup(work_path);
	gBundleRWPath = g_strdup(rw_path);
	gLiveDB = g_strdup(live_database);

	
	strcpy(db_path, getenv("ALP_RUN_TOP") ?: "");
	strcat(db_path, "/");
	strcat(db_path, database);
	
	char * r;
	if ((r = strrchr(db_path, '/'))) {
		*r = '\0';
		mkdir_p(db_path, 0777);
		*r = '/';
	}

	//ALP_TMTL("Opening database '%s'", db_path);

	db = NULL;
	int result = sqlite3_open(db_path, &db);
	if (result != SQLITE_OK) {
		ALP_TMTL("Cannot open database '%s': %s", db_path, sqlite3_errmsg(db));
		goto failure;
	}
	
	// TEMPORARY views and tables should be stored in unshared memory,
	// instead of putting them in /var/tmp
	S_EXEC("PRAGMA temp_store = MEMORY");
	
	// If there is database contention, continue trying up to this maximum time.
	// Database contention is not expected to last any significant time, so the full
	// range of this timeout is not expected to be seen in normal use.
	
	sqlite3_busy_timeout(db, SQLITE_RETRY_MS);

	if (server) {
		
		//ALP_TMTL("Opening channel '%s'", ipc_channel);

		if (!g_thread_supported()) {
			// Odd, bundlemgrd should have initialised the thread package		

			g_thread_init(NULL);
			ALP_TMTL("Thread support required for server-mode bundle manager");
			if (!g_thread_supported()) {
				ALP_TMTL("Thread support not available");
				goto failure;
			}
		}
		
		int check_result = 0;
	
		S_STMT s;
		S_PREPARE(s, "PRAGMA integrity_check;");
		if (S_STEP_ROW(s)) {
			const char * check_result_str = S_COL_TEXT(s,0);
			check_result = check_result_str && (strcmp(check_result_str, "ok") == 0);
			if (!check_result)
				ALP_TMTL("Integrity_check returned failure: %s", check_result_str);
		} else {
			ALP_TMTL("Very odd, integrity check failed to give a response");
		}
		S_FINALIZE(s);
		
		if (!check_result) {
			//If the database was corrupt, fail. Someone external will
			//need to wipe the database so we can set up a fresh copy
			goto failure;
		}
		
		int db_version = -1;
		if (S_PREPARE_P(s, "SELECT db_version FROM general;")) {
			if (S_STEP_ROW(s)) {
				db_version = S_COL_INT(s, 0);
			}
			S_FINALIZE(s);
		} else {
			db_version = -1;
		}
		
		if (db_version == -1) {
			// Empty database
			alp_prv_bundle_setup_schema();

		} else if (db_version != PM_CURRENT_DB_VERSION) {
			// Incorrect schema, fail. Someone external will need to
			// wipe the database so we can set up a fresh copy
			goto failure;
		}
		
		// Construct the separate database for live metadata. This
		// database can be in volatile storage (/tmp) as it does not
		// need to survive reboot. However, it is shared between processes,
		// so it does need to be in a file, and cannot be private memory.
		setup_live_metadata(gLiveDB, 1, 1);

		// A master view is created for metadata -- this is constructed
		// in a temporary table so it has a view across the live_metadata and
		// other tables, in separate databases. As this view is in private
		// memory, clients will need to construct their own view (does this
		// lose some degree of caching/sharing?)
		create_view();

		gGHost = ghost_available();
		
		if (alp_global_settings_open(&gGSContext) != ALP_STATUS_OK) {
			ALP_TMTL("Unable to open global settings context");
			goto failure;
		}

		if (gGSContext) {
			// Setup up /bundles key in global settings if it doesn't already exist
			alp_global_settings_set_key_directory(gGSContext, "/bundles");
			alp_global_settings_set_mode(gGSContext, "/bundles", 0775); // user rxw, group rwx, other rx
		}
		
		pm_server_channel = alp_ipc_channel_create(ipc_channel, ALP_IPC_CHANNEL_ACCESS_ALL);
		
		if (!pm_server_channel) {
			ALP_TMTL("Unable to establish IPC channel for server");
			goto failure;
		}
		
		alp_ipc_channel_register_connect_callback(pm_server_channel, handle_connect, NULL);
		alp_ipc_channel_register_disconnect_callback(pm_server_channel, handle_disconnect, NULL);
		
		register_volsvc_hooks();
		
		pm_status = STATUS_SERVER;
		
	} else { // !server, thus client

		int db_version = -1;
		
		S_STMT s;
		if (S_PREPARE_P(s, "SELECT db_version FROM general;")) {
			if (S_STEP_ROW(s)) {
				db_version = S_COL_INT(s, 0);
			}
			S_FINALIZE(s);
		}
		
		if (db_version != PM_CURRENT_DB_VERSION) {
			// Incorrect schema or empty DB on the client, fail.
			// Someone external will need to wipe the database
			// so we can set up a fresh copy

			goto failure;
		}
		
		setup_live_metadata(gLiveDB, 0, 0);

		create_view();

		pm_client_connection = NULL;
		
		//ALP_TMTL("Opening connection to channel '%s'", ipc_channel);
		
		alp_ipc_channel_connect_synch(ipc_channel, ALP_IPC_CONNECTION_USAGE_HINT_NONE, 0, &pm_client_connection);
	
		if (!pm_client_connection) {
			ALP_TMTL("Unable to establish IPC connection to server");
			goto failure;
		}
		
		pm_status = STATUS_CLIENT;
	
	}
	
	pm_open_count = 1;
	
	return 0; // success

failure:
	if (db) {
		int result = sqlite3_close(db);
		if (result != SQLITE_OK) {
			ALP_TMTL("Error during failing close of db: %s", sqlite3_errmsg(db));
		}
		db = NULL;
	}
	if (pm_server_channel) {
		alp_ipc_channel_destroy(pm_server_channel);
		pm_server_channel = NULL;
	}	
	if (pm_client_connection) {
		alp_ipc_channel_disconnect(pm_client_connection);
		pm_client_connection = NULL;
	}
	if (gBundleWorkPath) {
		g_free((char*)gBundleWorkPath);
		gBundleWorkPath = NULL;
	}
	if (gBundleRWPath) {
		g_free((char*)gBundleRWPath);
		gBundleRWPath = NULL;
	}
	if (gLiveDB) {
		g_free((char*)gLiveDB);
		gLiveDB = NULL;
	}
	
	pm_status = STATUS_UNINIT;

	ALP_TMTL("Failing!");
		
	return set_error(ALP_STATUS_BUNDLEMGR_FAILURE);
		
}

alp_status_t alp_prv_bundle_fini(void)
{
	if (pm_status == STATUS_SERVER) {
		//ALP_TMTL("Shutting down server");
	
		// flush any pending notifications
		destroy_notify_pool();
	
		alp_ipc_channel_destroy(pm_server_channel);
		pm_server_channel = NULL;
		
		int result = sqlite3_close(db);
		if (result != SQLITE_OK) {
			ALP_TMTL("Error during server close of db: %s", sqlite3_errmsg(db));
		}
		db = NULL;
		
		wipe_locales();
		
		alp_prv_package_clear_refs();
		
		alp_prv_bundle_clear_local_effects();

		if (gGSContext)
			alp_global_settings_close(gGSContext);

		reset_hold();
		
		pm_open_count = 0;
		
		gGHost = 0;
		
	} else if (pm_status == STATUS_CLIENT) {
		//ALP_TMTL("Shutting down client");
	
		alp_ipc_channel_disconnect(pm_client_connection);
		pm_client_connection = NULL;
		int result = sqlite3_close(db);
		if (result != SQLITE_OK) {
			ALP_TMTL("Error during client close of db: %s", sqlite3_errmsg(db));
		}
		db = NULL;
		
		wipe_locales();
		
		alp_prv_package_clear_refs();

		alp_prv_bundle_clear_local_effects();

		reset_hold();
		
		pm_open_count = 0;
		
	} else if (pm_status != STATUS_UNINIT) {
		ALP_TMTL("Error, trying to shut down package manager in odd state");
		
		return set_error(ALP_STATUS_BUNDLEMGR_FAILURE);
	}
	
	pm_status = STATUS_UNINIT;
	return 0; 
}

alp_status_t alp_prv_bundle_setup_schema(void)
{
	S_EXEC_W( "BEGIN; \n"
		"PRAGMA auto_vacuum = 1;\n"
		"CREATE TABLE general(db_version INTEGER);\n"
		"CREATE TABLE schemes(scheme_id INTEGER PRIMARY KEY, scheme_name TEXT UNIQUE, mimetype TEXT, extension TEXT); \n"
		"CREATE TABLE stores(store_id INTEGER PRIMARY KEY, "
		                    "store_name TEXT UNIQUE, "
		                    "bar_folder TEXT, "
		                    "bar_prefer_archive BOOLEAN, "
		                    "ghost_storage INTEGER, "
		                    "ghost_path TEXT, "
		                    "jam_store BOOLEAN, "
				    		"register_policy TEXT, "
				    		"unregister_policy TEXT, "
				    		"delete_policy TEXT, "
				    		"install_policy TEXT, "
				    		"scan_policy TEXT, "
		                    "wipe_by_default BOOLEAN, "
		                    "vs_udi TEXT, "
		                    "install_priority INTEGER DEFAULT 0, "
		                    "is_rom BOOLEAN); \n"
		// Containers has redundant columns 'name', and 'register_notify'. These are to be removed from
		// common_metadata once the *_metadata tables are folder into containers.
		"CREATE TABLE containers(active BOOLEAN, package_id INTEGER UNIQUE PRIMARY KEY AUTOINCREMENT, name TEXT, name_version DEFAULT 0, register_notify INTEGER, scheme_id INTEGER, store_id INTEGER, open_foldername TEXT, archive_filename TEXT, stub BOOLEAN, did_register_notify BOOLEAN, did_unregister_notify BOOLEAN, security_handle INTEGER); \n"
		"CREATE INDEX containers_idx_1 ON containers (name, active, name_version, store_id, scheme_id, package_id); \n"
		"CREATE TABLE notifications(seq INTEGER PRIMARY KEY, key TEXT, data BLOB); \n"
		"CREATE TABLE names(name TEXT UNIQUE PRIMARY KEY, package_id INTEGER, notified BOOLEAN, register_notify INTEGER); \n"
		"CREATE TABLE common_metadata(package_id INTEGER, name TEXT, open BOOLEAN, recycle BOOLEAN, register_notify INTEGER); \n"
		"CREATE TABLE bar_metadata(package_id INTEGER, bar_version TEXT, bar_searchpath TEXT, bar_filename TEXT, bar_textdomain_name TEXT, bar_textdomain_dir TEXT); \n"
		"CREATE TABLE ghost_metadata(package_id INTEGER, ghost_db_storage INTEGER,	\
			ghost_db_name TEXT,			\
			ghost_db_attributes INTEGER,		\
			ghost_db_version INTEGER,		\
			ghost_db_creation_date INTEGER,		\
			ghost_db_modification_date INTEGER,	\
			ghost_db_last_backup_data INTEGER,	\
			ghost_db_modification_number INTEGER,	\
			ghost_db_type INTEGER,			\
			ghost_db_creator INTEGER,		\
			ghost_db_type_str TEXT,			\
			ghost_db_creator_str TEXT		\
			); \n"
		"CREATE TABLE java_metadata(package_id INTEGER, \
			midlet_name TEXT, \
			midlet_id INTEGER \
			); \n"	
		"CREATE TABLE bar_exports(package_id INTEGER, path TEXT, name TEXT, kind TEXT); \n"
		"CREATE TABLE bar_references (package_id INTEGER, refer_to TEXT, optional BOOLEAN); \n"
		"CREATE TABLE properties(package_id INTEGER, sequence INTEGER, property TEXT, key TEXT, value BLOB); \n"	
		"INSERT INTO schemes VALUES (NULL,'bar','x-application/access-bar','.bar'); \n"
		"INSERT INTO schemes VALUES (NULL,'GHost','x-application/access-ghost','.prc'); \n"
		"INSERT INTO schemes VALUES (NULL,'java','x-application/access-java','.jar'); \n"
		"COMMIT; \n"
		"VACUUM; \n"
	);
	
	S_STMT s;
	S_PREPARE(s, "INSERT INTO general (db_version) VALUES (?1)");
	S_BIND_INT(s, 1, PM_CURRENT_DB_VERSION);
	S_DO(s);

	if (gNotify) {
		alp_status_t err = alp_notify_broadcast(NOTIFY_NAME, ALP_NOTIFY_EVENT_BUNDLE_RESET_EVERYTHING, NULL, 0, ALP_NOTIFY_QUEUE_DEFAULT, 0);
		if (err != ALP_STATUS_OK) {
			ALP_TMTL("Broadcast of bundlemgr reset failed");
		}
	} 
	
	return ALP_STATUS_OK;
}

static void setup_live_metadata(const char * live_db_filename, int erase, int create)
{
	char * buf = sqlite3_mprintf("ATTACH DATABASE '%q' AS shared_temp", live_db_filename);
	
	// Treat live_db_filename as a pathname, and do a mkdir_p() on its
	// directory, if any.
	
	char dirbuf[PATH_MAX];
	strcpy(dirbuf, live_db_filename);
	char * c;
	
	if ((c = strrchr(dirbuf, '/'))) {
		// trim off file-name
		*c = '\0';
		
		// FIXME: consider permissions
		mkdir_p(dirbuf, 0777);
	}

	if (erase) {
		unlink(live_db_filename);
	}
	S_EXEC(buf);

	if (create) {
		S_EXEC("CREATE TABLE shared_temp.live_metadata (package_id INTEGER, refcount INTEGER, \
			gettextdomain TEXT, real_folder TEXT, resolved_real_folder TEXT)");
	}
	
	sqlite3_free(buf);
}

static void create_view(void)
{
	S_EXEC("CREATE TEMPORARY VIEW merged_metadata AS SELECT containers.package_id as package_id,schemes.scheme_name as scheme_name,containers.security_handle AS security_handle,"
		             "stores.store_name as store_name,bar_metadata.bar_version as bar_version,bar_metadata.bar_filename as bar_filename,common_metadata.name as name,open,"
		             "live_metadata.refcount as refcount,live_metadata.gettextdomain as gettextdomain,live_metadata.real_folder as real_folder,"
		             "bar_metadata.bar_textdomain_dir as bar_textdomain_dir,bar_metadata.bar_textdomain_name as bar_textdomain_name,live_metadata.resolved_real_folder as resolved_real_folder,common_metadata.recycle as recycle FROM "
		             " containers LEFT OUTER JOIN schemes ON (containers.scheme_id=schemes.scheme_id) LEFT OUTER JOIN stores ON (containers.store_id=stores.store_id) LEFT OUTER JOIN bar_metadata ON (containers.package_id=bar_metadata.package_id) LEFT OUTER JOIN common_metadata ON (containers.package_id=common_metadata.package_id) LEFT OUTER JOIN live_metadata ON (containers.package_id=live_metadata.package_id); ");
}

alp_status_t alp_prv_bundle_wipe(void)
{
	if (pm_status == STATUS_CLIENT) {
		AlpMessage * message, *response;
			
		message = alp_ipc_message_create();
		alp_ipc_message_pack_start(message, MSG_PACKAGE_WIPE);
		alp_ipc_message_pack_end(message);
		if (alp_ipc_connection_send_with_response(pm_client_connection, message, 0, &response)) {
			alp_ipc_message_destroy(message);
			ALP_TMTL("some failure of IPC");
			// some failure of IPC
			return set_error(ALP_STATUS_BUNDLEMGR_FAILURE);
		}
		alp_ipc_message_destroy(message);

		alp_status_t result;
		alp_ipc_message_unpack_start(response);
		alp_ipc_message_unpack_int32(response, &result);
		alp_ipc_message_unpack_end(response);
		alp_ipc_message_destroy(response);
		
		return set_error(result);
	}
	else if (pm_status != STATUS_SERVER)
		return set_error(ALP_STATUS_BUNDLEMGR_FAILURE);
	
	// You need SQLite 3.3 for the 'DROP TABLE IF EXISTS' syntax,
	// so I have to do this manually
	S_EXEC_W("DROP VIEW merged_metadata;");

	S_EXEC_W("DROP TABLE general;");
	S_EXEC_W("DROP TABLE schemes;");
	S_EXEC_W("DROP TABLE stores;");
	S_EXEC_W("DROP TABLE containers;");
	S_EXEC_W("DROP TABLE notifications;");
	S_EXEC_W("DROP TABLE common_metadata;");
	S_EXEC_W("DROP TABLE bar_metadata;");
	S_EXEC_W("DROP TABLE ghost_metadata;");
	S_EXEC_W("DROP TABLE java_metadata;");
	S_EXEC_W("DROP TABLE bar_references;");
	S_EXEC_W("DROP TABLE bar_exports;");
	S_EXEC_W("DROP TABLE properties;");
	S_EXEC_W("DROP TABLE names;");
	S_EXEC_W("DELETE FROM shared_temp.live_metadata;");
	
	alp_prv_bundle_setup_schema();
	
	//setup_live_metadata(gLiveDB, 0, 1);

	create_view();
	

	return ALP_STATUS_OK;
}

void handle_client_alp_prv_package_wipe(AlpConnection * connection, AlpMessage * message, gpointer cbData)
{
	alp_ipc_message_unpack_start(message);
	alp_ipc_message_unpack_end(message);
			
	alp_status_t result;

	if (has_permission(connection, PERM_WIPE_POLICY)) {
		result = alp_prv_bundle_wipe();
	} else {
		result = ALP_STATUS_BUNDLEMGR_FAILURE;
	}
	
	AlpMessage * response = alp_ipc_message_create();
	alp_ipc_message_response_pack_start(response, MSG_PACKAGE_WIPE_RESPONSE, message);
	alp_ipc_message_pack_int32(response, result); 
	alp_ipc_message_pack_end(response);
	alp_ipc_connection_send(connection, response);
	alp_ipc_message_destroy(response);

	alp_ipc_message_destroy(message);
}


int alp_prv_bundle_state(void)
{
	return pm_status;
}
