/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <sys/types.h>
#include <dirent.h>
#include <stdio.h>

#include "pm_globals.h"
#include "pm_traces.h"

int mkdir_p(const char * path, int mode)
{
  char buf[PATH_MAX];

  strcpy(buf, path);
  
  int result = mkdir(buf, mode);
    
  // success? Nothing more to do.
  if (result == 0)
    return result;
    
  // Already there? Nothing more to do
  if (result == -1 && errno == EEXIST)
    return result;

  // Something in the middle was missing?
  if (result == -1 && errno == ENOENT) {
  
    // find the last entry
    char * slash = strrchr(buf, '/');
    
    // wasn't one? We can't do much
    if (!slash)
      return result;
      
    // chop it off
    *slash = '\0';
    
    // and try on the chopped version
    result = mkdir_p(buf, mode);
    
    // if that failed, give up  
    if (result == -1)
      return result;
    
    // put back the last bit, and try again
    *slash = '/';

    result = mkdir(buf, mode);      
  }
  
  return result;
}

int rmdir_contents(const char * path)
{
	GList * dirs = NULL, * removal = NULL;

	gboolean first = TRUE;

	dirs = g_list_append(dirs, g_strdup(path));

	while (dirs) {

		char buf[PATH_MAX];
	
		strcpy(buf, (char*)dirs->data);
	
		DIR * dir = opendir(buf);
		
		g_free(dirs->data);

		dirs = g_list_delete_link(dirs, dirs);

		if (!dir)
			break;  
  
		char * end = buf + strlen(buf);

		*end++ = '/';
		struct dirent * d;
		struct stat stbuf;
	
		while ((d = readdir(dir))) {

			if (d->d_name[0] == '.' && (d->d_name[1] == '\0' || (d->d_name[1] == '.' && d->d_name[2] == '\0')))
				continue;

			strcpy(end, d->d_name);

			if (stat(buf, &stbuf)==0 && S_ISDIR(stbuf.st_mode)) {
				dirs = g_list_prepend(dirs, g_strdup(buf));
			}
			removal = g_list_prepend(removal, g_strdup(buf));
		}
	
		closedir(dir);
	}
	
	GList * iter;
	int result;
	for (iter=removal;iter;iter = g_list_next(iter)) {
		if ((result = unlink((char*)iter->data)) == -1 && errno == EISDIR) {
			result = rmdir((char*)iter->data);
		}
		if (result == -1) {
			g_list_foreach(iter, (GFunc)g_free, NULL);
			g_list_free(removal);
			return result;
		}

		g_free(iter->data);
	}
	g_list_free(removal);

	return 0;
}

// realpath() only operates on completely extant paths, and has some very
// bad allocation bugs. This routine is safer, and will operate on paths
// that don't fully exist, giving up and passing through unchanged elements
// which can't be read.

gchar * hypotheticalpath(const char * path)
{
	typedef struct fragment_t {
		const char * str;
		int len;
		int pos;
		int str_alloc;
		int fragment_alloc;
		
		struct fragment_t * next;
	} fragment_t;
	
	const int debug = 0;
	
	if (debug)
		ALP_TMTL("path={%s}", path);
	
	if (!path)
		return NULL;

	fragment_t head = {0,0,0,0,0, 0};
	fragment_t cwd = {0,0,0,0,0, 0};
	
	head.str = path;
	head.len = strlen(path);

	fragment_t * fragments = &head;
	
	if (path[0] != '/') {
		cwd.str = getcwd(NULL, 0);
		cwd.len = strlen(cwd.str);
		cwd.pos = 0;
		cwd.str_alloc = 1;
		cwd.fragment_alloc = 0;
		cwd.next = fragments;
		fragments = &cwd;
	}

	GString * buffer = g_string_new("");
	
	int known_invalid_path = 0;
	
	while (fragments) {
	
		fragment_t * f = fragments;

		if (debug)
			ALP_TMTL("Fragment str={%s}, at pos %d of %d", f->str, f->pos, f->len);

		while (f->pos < f->len && f->str[f->pos] == '/') {
			f->pos++;
		}
		
		int segment_len;
		char * slash = memchr(f->str+f->pos, '/', f->len-f->pos);
		if (!slash)
			segment_len = f->len-f->pos;
		else
			segment_len = slash-(f->str+f->pos);

		
		int check_link = 0;
		
		if (segment_len == 1 && f->str[f->pos] == '.') {
			// process .
			// no change to buffer
		} else if (!known_invalid_path && segment_len == 2 && f->str[f->pos] == '.' && f->str[f->pos+1] == '.') {
			// process ..
			
			// since the buffer will consist of completely resolved names at
			// this point, it's safe to "back up" by simply stripping off
			// the last path element, as we know a lookup of .. in that
			// folder would equate to the expected parent -- this isn't true
			// if the path has unknown components
			
			char * last_slash = strrchr(buffer->str, '/');
			if (last_slash) {
				g_string_truncate(buffer, last_slash - buffer->str);
			} else {
				g_string_truncate(buffer, 0);
			}

		} else if (segment_len > 0) {
			g_string_append_c(buffer, '/');
			g_string_append_len(buffer, f->str+f->pos, segment_len);
			
			check_link = 1;
		}
		
		if (debug)
			ALP_TMTL("Buffer is [%s]", buffer->str);

		f->pos += segment_len;
		if (f->pos >= f->len) {
			if (f->str_alloc)
				free((void*)f->str);
			
			fragments = f->next;
			
			if (f->fragment_alloc)
				free(f);
		}
		
		if (check_link) {
			char symbuf[PATH_MAX+2];

			if (debug) 
				ALP_TMTL("known_invalid_path=%d, buffer={%s}", known_invalid_path, buffer->str);

			int count;
			if (!known_invalid_path) {
				if (((count=readlink(buffer->str, symbuf, sizeof(symbuf))) >= 0)) {
					if (debug)
						ALP_TMTL("readlink gave count %d", count);

					// overly clever: allocate the fragment structure and
					// the strdup buffer in one go. Note that only the
					// fragment should be free'd -- it'll take the string
					// with it
					
					fragment_t * symfrag = malloc(sizeof(fragment_t) + count + 1);
					symfrag->str = (char*)(symfrag+1);
					memcpy((void*)symfrag->str, symbuf, count);

					symfrag->len = count;
					symfrag->pos = 0;
					symfrag->str_alloc = 0;
					symfrag->fragment_alloc = 1;
					symfrag->next = fragments;

					// FIXME: should only be needed for debug
					if (debug)
						((char*)symfrag->str)[count] = '\0';
				
					if (symfrag->str[0] == '/') {
						// Absolute symlink, start over
						g_string_truncate(buffer, 0);
					}
					
					if (debug)
						ALP_TMTL("Adding fragment for symlink {%s}", symfrag->str);
					
					fragments = symfrag;
				} else {
					if (debug)
						ALP_TMTL("Readlink failed, %s", strerror(errno));
						
					// Assume that any errors other than EINVAL will
					// still apply for children of this folder, so there's
					// no reason to repeat testing for symlinks
					if (errno != EINVAL)
						known_invalid_path = 1;
				}
			}
		}
	}
	
	if (debug)
		ALP_TMTL("result={%s}", buffer->str);
	
	//FIXME: remove this when I'm comfortable with the newer routine
	if (1) {
		char buf[PATH_MAX] = {0};
		realpath(path, buf);
		
		g_string_append_c(buffer, '\0');
		if (strcmp(buf, buffer->str) != 0) {
			ALP_TMTL("difference from realpath. It gave {%s}, we gave {%s}",
				buf,
				buffer->str);
		}
	}
	return g_string_free(buffer, FALSE);
}

int alp_prv_folder_empty(const char * path, int status_on_missing)
{	
	DIR * dir = opendir(path);
	
	if (!dir) {
		return status_on_missing; 
	}
	
	if (dir) {
		struct dirent *	d;

		while ((d = readdir(dir))) {
			if (d->d_name[0] == '.' && d->d_name[1] == '\0')
				continue;
			if (d->d_name[0] == '.' && d->d_name[1] == '.' && d->d_name[2] == '\0')
				continue;
			
			closedir(dir);

			return 0; // something there, thus not empty
			
		}
		closedir(dir);
	}
	
	return 1; // nothing there, thus empty
	
}
