/********************************************************************
 
Copyright  2006, ACCESS Systems Americas, Inc. All rights reserved.
 
The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/
 
Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.
 
The Original Code is the entire contents of this file.
 
The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.
 
Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.
 
Contributor(s): none.
 
********************************************************************/
/*!
 * @file global_settings_server.c
 *
 * @brief the implementation of the Global Settings service APIs.
 *
 *****************************************************************************/
#include <pthread.h>
#include <assert.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/poll.h>
#include <time.h>
#include <pwd.h>
#include <grp.h>
#include <glib.h>
/*#include <dbus/dbus-glib.h>
 */
#include "hiker/traces.h"
#include "hiker/fail.h"
#include "hiker/global_settings.h"
#include "hiker/syspaths.h"
#include "hiker/ipc.h"
#include "hiker/notifymgr.h"
#include "hiker/pdk/sysutils_pdk.h"

/*#include "hiker/security.h"
*/
#include "gs_imp.h"

#include "global_settings_server.h"

#include "serialize.h"
#include "policy_strings.h"

#include "sqlfs.h"

#include "sqlite3.h"


#define DEFAULT_PREFS_DB ALP_SYSPATH_RO_DATA "global_settings/prefs"


#define NUM_WORKER 3

static GThreadPool* thread_pools[NUM_WORKER] = { 0, };

time_t thread_last_check_time[NUM_WORKER] = { 0, }, thread_now[NUM_WORKER] ;

static  void thread_context_periodic_check(int thread_id);

#define CONTEXT_HASH_LOCK  pthread_mutex_lock(&context_hash_lock);
#define CONTEXT_HASH_UNLOCK  pthread_mutex_unlock(&context_hash_lock);

static pthread_mutex_t context_hash_lock = PTHREAD_MUTEX_INITIALIZER ;
static int context_operations_in_progress = 0;
static pthread_cond_t no_context_operation_cond = PTHREAD_COND_INITIALIZER ;

static const char * const default_database_name = ALP_SYSPATH_PREFS;  /* default ALP location */

static GHashTable *context_table = 0;
static GMainLoop *loop;

static int notify_mgr_working = 0;


static int next_connection_serial_no(void)
{
    static int connection_serial_no = 1;

    static pthread_mutex_t connection_serial_no_lock = PTHREAD_MUTEX_INITIALIZER ;
    int i;
    pthread_mutex_lock(&connection_serial_no_lock);
    i =  (connection_serial_no)++;
    pthread_mutex_unlock(&connection_serial_no_lock);
    return i;
}

static int mkdirp(const char *path, mode_t mode)  /* return 0 on success and -1 on error, with errno set */
{
    /* emulate the behavior of mkdir -p */
    char saved_pwd[PATH_MAX], buf[PATH_MAX];
    char *tail;
    int r = 0, i;
    errno = 0;
    if (path[0] == 0)
        return 0;
    getcwd(saved_pwd, sizeof(saved_pwd));
    if (path[0] == '/') /* absolute path */
    {
        r = chdir("/");
        strncpy(buf, path + 1, sizeof(buf));
    }
    else
        strncpy(buf, path, sizeof(buf));
    tail = strchr(buf, '/');
    if (tail)
        *tail = 0;
    if (!r)
    {
        i = is_dir(buf);
        if (i == 0)
            r = mkdir(buf, mode);
        else if (i == -1)
        {
            errno = EEXIST;
            r = -1;
        }
    }
    if (tail && !r)
    {
        r = chdir(buf);
        if (!r)
            r = mkdirp(tail + 1, mode);
    }
    chdir(saved_pwd);
    return r;

}


static void send_key_change_notification(const char *key)
{
    int handled;
    static int prefix_length = 0;
    char *buf = 0;
    alp_status_t status;
    if (notify_mgr_working == 0)
        return;
    if (prefix_length == 0)
        prefix_length = strlen(ALP_NOTIFY_EVENT_GLOBAL_SETTINGS_KEY_CHANGE);
    buf = calloc(prefix_length + strlen(key) + 3, sizeof(char));

    sprintf(buf, "%s%s", ALP_NOTIFY_EVENT_GLOBAL_SETTINGS_KEY_CHANGE, key);
    status = alp_notify_broadcast/*_sync*/(GLOBAL_SETTINGS_APP_ID, buf, key, strlen(key) + 1, ALP_NOTIFY_QUEUE_DEFAULT, 0 /*, 10, &handled*/);
    if (status != ALP_OK)
        ALP_TM(T(ALP_CLASS_GLOBAL_SETTINGS, "failed to notify changes!\n"));

    free(buf);
}

static alp_status_t
gs_svr_proc_set( gs_svr_context *context,
                 const gchar *key,
                 const AlpGlobalSettingsValue *value);
static alp_status_t
gs_svr_proc_get( gs_svr_context *context,
                 const gchar *key,
                 AlpGlobalSettingsValue *value);
static alp_status_t
gs_svr_set_key_string_value(gs_svr_context * context,
                            const gchar *key,
                            const gchar *value)
{
    AlpGlobalSettingsValue gsvalue = GLOBAL_SETTINGS_VALUE_INITIALIZER;
    alp_status_t status = alp_global_settings_value_set_string(&gsvalue, value);
    if (status != ALP_STATUS_OK)
        return status;
    status = gs_svr_proc_set(context, key, &gsvalue);
    global_settings_value_clean(&gsvalue);
    return status;

}

static alp_status_t
gs_svr_get_key_string_value(gs_svr_context *context,
                            const gchar *key,

                            gchar *value,
                            guint value_size)
{
    AlpGlobalSettingsValue gsvalue = GLOBAL_SETTINGS_VALUE_INITIALIZER;
    alp_status_t status = gs_svr_proc_get(context, key, &gsvalue);
    if (status != ALP_STATUS_OK)
        return status;
    status = alp_global_settings_value_get_string(&gsvalue, value, value_size);
    global_settings_value_clean(&gsvalue);
    return status;

}


static alp_status_t
gs_svr_set_key_bool_value(gs_svr_context *context,
                          const gchar *key,
                          gboolean value)
{
    AlpGlobalSettingsValue gsvalue = GLOBAL_SETTINGS_VALUE_INITIALIZER;
    alp_status_t status = alp_global_settings_value_set_bool(&gsvalue, value);
    if (status != ALP_STATUS_OK)
        return status;
    status = gs_svr_proc_set(context, key, &gsvalue);
    global_settings_value_clean(&gsvalue);
    return status;

}

static alp_status_t
gs_svr_get_key_bool_value(gs_svr_context *context,
                          const gchar *key,
                          gboolean *value)
{
    AlpGlobalSettingsValue gsvalue = GLOBAL_SETTINGS_VALUE_INITIALIZER;
    alp_status_t status = gs_svr_proc_get(context, key, &gsvalue);
    if (status != ALP_STATUS_OK)
        return status;
    status = alp_global_settings_value_get_bool(&gsvalue, value);
    global_settings_value_clean(&gsvalue);
    return status;

}



static alp_status_t
gs_svr_set_key_int_value(gs_svr_context *context,
                         const gchar *key,
                         int value)
{
    AlpGlobalSettingsValue gsvalue = GLOBAL_SETTINGS_VALUE_INITIALIZER;
    alp_status_t status = alp_global_settings_value_set_int(&gsvalue, value);
    if (status != ALP_STATUS_OK)
        return status;
    status = gs_svr_proc_set(context, key, &gsvalue);
    global_settings_value_clean(&gsvalue);
    return status;

}

static alp_status_t
gs_svr_get_key_int_value(gs_svr_context *context,
                         const gchar *key,
                         int *value)
{
    AlpGlobalSettingsValue gsvalue = GLOBAL_SETTINGS_VALUE_INITIALIZER;
    alp_status_t status = gs_svr_proc_get(context, key, &gsvalue);
    if (status != ALP_STATUS_OK)
        return status;
    status = alp_global_settings_value_get_int(&gsvalue, value);
    global_settings_value_clean(&gsvalue);
    return status;

}


static alp_status_t
gs_svr_set_key_float_value(gs_svr_context *context,
                           const gchar *key,
                           double value)
{
    AlpGlobalSettingsValue gsvalue = GLOBAL_SETTINGS_VALUE_INITIALIZER;
    alp_status_t status = alp_global_settings_value_set_float(&gsvalue, value);
    if (status != ALP_STATUS_OK)
        return status;
    status = gs_svr_proc_set(context, key, &gsvalue);
    global_settings_value_clean(&gsvalue);
    return status;

}

static alp_status_t
gs_svr_get_key_float_value(gs_svr_context *context,
                           const gchar *key,
                           double *value)
{
    AlpGlobalSettingsValue gsvalue = GLOBAL_SETTINGS_VALUE_INITIALIZER;
    alp_status_t status = gs_svr_proc_get(context, key, &gsvalue);
    if (status != ALP_STATUS_OK)
        return status;
    status = alp_global_settings_value_get_float(&gsvalue, value);
    global_settings_value_clean(&gsvalue);
    return status;

}

static alp_status_t
gs_svr_get_key_blob_value(gs_svr_context *context,
                          const char *key,
                          char  *value,
                          int buffer_size,
                          int * value_size)
{
    AlpGlobalSettingsValue gsvalue = GLOBAL_SETTINGS_VALUE_INITIALIZER;
    alp_status_t status = gs_svr_proc_get(context, key, &gsvalue);
    if (status != ALP_STATUS_OK)
        return status;
    if (gsvalue.type !=  ALP_GLOBAL_SETTINGS_TYPE_BLOB)
        status =  ALP_STATUS_GLOBAL_SETTINGS_INCORRECT_TYPE ;
    else if ((int)(gsvalue.blob_value.len) > buffer_size)
        status = ALP_STATUS_GLOBAL_SETTINGS_BUFFER_TOO_SMALL ;
    else
    {
        memcpy(value, gsvalue.blob_value.data, gsvalue.blob_value.len );
        *value_size = gsvalue.blob_value.len;
    }
    global_settings_value_clean(&gsvalue);
    return status;

}


static alp_status_t
gs_svr_set_key_blob_value(gs_svr_context *context,
                          const char *key,
                          const char *value,
                          int value_size)
{
    AlpGlobalSettingsValue gsvalue = GLOBAL_SETTINGS_VALUE_INITIALIZER;
    alp_status_t status ;
    gsvalue.type = ALP_GLOBAL_SETTINGS_TYPE_BLOB;
    gsvalue.blob_value.data = (char*) value;
    /* note we use blob_value as a carrier to eliminate the need of an extra local copy */
    gsvalue.blob_value.len = value_size;
    status = gs_svr_proc_set(context, key, &gsvalue);
    /* so we do not free the value here */
    return status;

}

static alp_status_t
gs_svr_set_key_directory(gs_svr_context *context,
                         const char *key)
{
    AlpGlobalSettingsValue gsvalue = GLOBAL_SETTINGS_VALUE_INITIALIZER;
    alp_status_t status ;
    gsvalue.type = ALP_GLOBAL_SETTINGS_TYPE_DIR;
    status = gs_svr_proc_set(context, key, &gsvalue);
    /* so we do not free the value here */
    return status;

}


#define BEGIN sqlfs_begin_transaction(context->sqlfs);
#define COMPLETE(r) sqlfs_complete_transaction(context->sqlfs, (r));


alp_status_t
gs_svr_init(const char *db_file)
{

    sqlfs_init(db_file);
    return ALP_STATUS_OK;

}

#if 0
static void * event_loop_thread_func(void *dummy)
{
    while (1)
    {
        //if  (g_main_context_acquire(g_main_context_default()) )
        {
            //while (g_main_context_pending(g_main_context_default()))
            g_main_context_iteration(g_main_context_default(), FALSE);
            //g_main_context_release(g_main_context_default());
        }

    }

}
#endif

alp_status_t
gs_svr_open(const char *db_file, gs_svr_context **context)
{
    GError *err = NULL;
    int i;
    ALP_FAIL_FATAL_IF(context == 0, "incorrect argument", "");


    *context = calloc(1, sizeof(**context));
    ALP_FAIL_FATAL_IF(*context == 0, "out of memory", "");

    i = sqlfs_open(db_file, &((*context)->sqlfs));
    ALP_FAIL_FATAL_IF(i == 0, "cannot open preference database", "");


#if 0
    (*context)->callback_table = g_hash_table_new_full(g_str_hash, g_str_equal, NULL, (GDestroyNotify) free_callback);
    ALP_FAIL_FATAL_IF((*context)->callback_table == 0, "Cannot allocate a hash table!", "");

    (*context)->any_callback_table = g_hash_table_new_full(g_str_hash, g_str_equal, NULL, (GDestroyNotify) free_callback);
    ALP_FAIL_FATAL_IF((*context)->any_callback_table == 0, "Cannot allocate a hash table!", "");
#endif
    /*(*context)->own_thread = separate_thread;
    if (separate_thread)
    {
        pthread_create(&((*context)->event_loop_thread), NULL, event_loop_thread_func, (void*) *context);
    }*/

    (*context)->sqlfs->default_mode = S_IXUSR | S_IRUSR | S_IWUSR ; /* default */
    (*context)->notification_on = 1;
    (*context)->cwd = 0;


    /* cannot return error */


    return ALP_STATUS_OK;
}

alp_status_t
gs_svr_close(gs_svr_context *context)
{
    CHECK_CONTEXT

    if (context->transaction_level > 0)
    {
        sqlfs_break_transaction(context->sqlfs);

    }


    sqlfs_close(context->sqlfs);

    /*if (context->own_thread)
    {
        pthread_join((context->event_loop_thread), NULL);
    }*/
#if 0
    g_hash_table_destroy(context->callback_table);
    context->callback_table = 0;
    g_hash_table_destroy(context->any_callback_table);
    context->any_callback_table = 0;

#endif
    if (context->cwd)
        g_free(context->cwd);

    free(context);
    return ALP_STATUS_OK;
}

static alp_status_t
gs_svr_proc_set_default_mode(gs_svr_context *context, mode_t mode)
{

    context->sqlfs->default_mode = mode;
    return ALP_STATUS_OK;
}

static alp_status_t
gs_svr_proc_get_default_mode(gs_svr_context *context, mode_t *mode)
{
    *mode = context->sqlfs->default_mode;
    return ALP_STATUS_OK;
}



alp_status_t
gs_svr_proc_begin_transaction(gs_svr_context *context)
{


    int i = sqlfs_begin_transaction(context->sqlfs);
    if (i == 2)
        return ALP_STATUS_GLOBAL_SETTINGS_BUSY;
    /*fprintf(stderr, "begin transaction!\n");*///???
    context->transaction_level++;
    if (i == 0)
        return ALP_STATUS_GLOBAL_SETTINGS_ERROR;
    return ALP_STATUS_OK;
}


alp_status_t
gs_svr_proc_complete_transaction(gs_svr_context *context)
{

    int i = sqlfs_complete_transaction(context->sqlfs, 1);
    if (i == 2)
        return ALP_STATUS_GLOBAL_SETTINGS_BUSY;
    /*fprintf(stderr, "complete transaction!\n");*///???
    context->transaction_level--;
    if (i == 0)
        return ALP_STATUS_GLOBAL_SETTINGS_ERROR;
    return ALP_STATUS_OK;


}

alp_status_t
gs_svr_proc_cancel_transaction(gs_svr_context *context)
{

    int i = sqlfs_complete_transaction(context->sqlfs, 0);
    if (i == 2)
        return ALP_STATUS_GLOBAL_SETTINGS_BUSY;

    /*fprintf(stderr, "cancel transaction!\n");*///???
    context->transaction_level--;

    if (i == 0)
        return ALP_STATUS_GLOBAL_SETTINGS_ERROR;
    return ALP_STATUS_OK;



}

static alp_status_t
gs_svr_proc_set_mode(gs_svr_context *context, const char *key, mode_t mode)
{

    int i = sqlfs_proc_chmod(context->sqlfs, key, mode);

    if (i == -EACCES)
        return ALP_STATUS_GLOBAL_SETTINGS_PERMISSION_DENIED;
    else
        if (i == -ENOENT)
            return ALP_STATUS_GLOBAL_SETTINGS_KEY_NOT_FOUND;
    if (i < 0)
        return ALP_STATUS_GLOBAL_SETTINGS_ERROR;
    return ALP_STATUS_OK;
}

static alp_status_t
gs_svr_proc_get_mode(gs_svr_context *context, const char *key, mode_t *mode)
{
    struct stat st;
    int i = sqlfs_proc_getattr(context->sqlfs, key, &st);
    if (i == -EACCES)
        return ALP_STATUS_GLOBAL_SETTINGS_PERMISSION_DENIED;
    else if (i == -ENOENT)
        return ALP_STATUS_GLOBAL_SETTINGS_KEY_NOT_FOUND;
    if (i < 0)
        return ALP_STATUS_GLOBAL_SETTINGS_ERROR;
    *mode = st.st_mode;

    return ALP_STATUS_OK;
}

static alp_status_t
gs_svr_proc_set_uid(gs_svr_context *context, const char *key, uid_t id)
{
    key_attr attr = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
    int i, result = ALP_STATUS_OK;
    BEGIN
    i = sqlfs_get_attr(context->sqlfs, key, &attr);
    if (i == 0)
        result = ALP_STATUS_GLOBAL_SETTINGS_ERROR;
    else if (i == -1)
        result = ALP_STATUS_GLOBAL_SETTINGS_KEY_NOT_FOUND;
    else if (i == -2)
        result = ALP_STATUS_GLOBAL_SETTINGS_PERMISSION_DENIED;
    else if (i == 2)
        result = ALP_STATUS_GLOBAL_SETTINGS_BUSY;
    else
    {
        attr.uid = id;
        i = sqlfs_set_attr(context->sqlfs, key, &attr);
        if (i == 0)
            result = ALP_STATUS_GLOBAL_SETTINGS_ERROR;
        else if (i == 2)
            result = ALP_STATUS_GLOBAL_SETTINGS_BUSY;
    }
    clean_attr(&attr);
    COMPLETE(1)
    return result;

}

static alp_status_t
gs_svr_proc_get_uid(gs_svr_context *context, const char *key, uid_t *id)
{
    key_attr attr = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
    int i, result = ALP_STATUS_OK;

    i = sqlfs_get_attr(context->sqlfs, key, &attr);
    if (i == 0)
        result = ALP_STATUS_GLOBAL_SETTINGS_ERROR;
    else if (i == -1)
        result = ALP_STATUS_GLOBAL_SETTINGS_KEY_NOT_FOUND;
    else if (i == -2)
        result = ALP_STATUS_GLOBAL_SETTINGS_PERMISSION_DENIED;
    else if (i == 2)
        result = ALP_STATUS_GLOBAL_SETTINGS_BUSY;
    else
    {
        *id = attr.uid ;
    }
    clean_attr(&attr);

    return result;

}


static alp_status_t
gs_svr_proc_set_gid(gs_svr_context *context, const char *key, gid_t id)
{
    key_attr attr = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
    int i, result = ALP_STATUS_OK;
    BEGIN
    i = sqlfs_get_attr(context->sqlfs, key, &attr);
    if (i == 0)
        result = ALP_STATUS_GLOBAL_SETTINGS_ERROR;
    else if (i == -1)
        result = ALP_STATUS_GLOBAL_SETTINGS_KEY_NOT_FOUND;
    else if (i == -2)
        result = ALP_STATUS_GLOBAL_SETTINGS_PERMISSION_DENIED;
    else if (i == 2)
        result = ALP_STATUS_GLOBAL_SETTINGS_BUSY;
    else
    {
        attr.gid = id;
        i = sqlfs_set_attr(context->sqlfs, key, &attr);
        if (i == 0)
            result = ALP_STATUS_GLOBAL_SETTINGS_ERROR;
        else if (i == 2)
            result = ALP_STATUS_GLOBAL_SETTINGS_BUSY;
    }
    clean_attr(&attr);
    COMPLETE(1)
    return result;

}

static alp_status_t
gs_svr_proc_get_gid(gs_svr_context *context, const char *key, gid_t *id)
{
    key_attr attr = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
    int i, result = ALP_STATUS_OK;

    i = sqlfs_get_attr(context->sqlfs, key, &attr);
    if (i == 0)
        result = ALP_STATUS_GLOBAL_SETTINGS_ERROR;
    else if (i == -1)
        result = ALP_STATUS_GLOBAL_SETTINGS_KEY_NOT_FOUND;
    else if (i == -2)
        result = ALP_STATUS_GLOBAL_SETTINGS_PERMISSION_DENIED;
    else if (i == 2)
        result = ALP_STATUS_GLOBAL_SETTINGS_BUSY;
    else
    {
        *id = attr.gid;
    }
    clean_attr(&attr);

    return result;

}

static alp_status_t
gs_svr_proc_set_notification_on_off(gs_svr_context *context, int on_off)
{

    context->notification_on = on_off;
    return ALP_STATUS_OK;
}

static alp_status_t
gs_svr_proc_set( gs_svr_context *context,
                 const gchar *key,
                 const AlpGlobalSettingsValue *value)
{
    key_value kvalue = { 0, 0, 0 };

    char buf[100];

    int r, i, num;
    GList *e;
    const AlpGlobalSettingsValue *value2 ;
    serialize_buf sbuf = { 0, 0, 0};
    CHECK_CONTEXT
    CHECK_ARG(key)
    CHECK_ARG(value)
    /*fprintf(stderr, "set key %s\n", key); *///???    debug use
    kvalue.data = buf; /* default */

    switch(value->type)
    {

    case ALP_GLOBAL_SETTINGS_TYPE_INVALID:
        ALP_TM(T(ALP_CLASS_GLOBAL_SETTINGS, "invalid value type"));
        return ALP_STATUS_GLOBAL_SETTINGS_VALUE_INVALID ;
    case ALP_GLOBAL_SETTINGS_TYPE_STRING:

        kvalue.data = value->string_value->str;
        kvalue.size = strlen(kvalue.data) + 1;
        break;

    case ALP_GLOBAL_SETTINGS_TYPE_INT:
        sprintf(kvalue.data, "%d", value->int_value);
        kvalue.size = strlen(kvalue.data) + 1;
        break;

    case ALP_GLOBAL_SETTINGS_TYPE_FLOAT:

        sprintf(kvalue.data, "%lf", value->float_value);
        kvalue.size = strlen(kvalue.data) + 1;
        break;

    case ALP_GLOBAL_SETTINGS_TYPE_BOOL:
        snprintf(kvalue.data, sizeof(buf), "%s", gs_bool_to_string(value->bool_value));
        kvalue.size = strlen(kvalue.data) + 1;
        break;

    case ALP_GLOBAL_SETTINGS_TYPE_LIST:
        serialize_gs_value(&sbuf, value);
        kvalue.data = sbuf.data;
        kvalue.size = sbuf.length;

        break;

    case ALP_GLOBAL_SETTINGS_TYPE_DIR:
        kvalue.size = 0;
        break;
    case ALP_GLOBAL_SETTINGS_TYPE_BLOB:
        kvalue.data = value->blob_value.data;
        kvalue.size = value->blob_value.len;
        break;

    }
    if (value->type == ALP_GLOBAL_SETTINGS_TYPE_DIR)
        r = sqlfs_proc_mkdir(context->sqlfs, key, context->sqlfs->default_mode);
    else
    {

        r = sqlfs_proc_write(context->sqlfs, key, kvalue.data, kvalue.size, 0, 0);
        if (r >= 0)
            r = sqlfs_proc_truncate(context->sqlfs, key, r);
    }

    if (r >= 0)
    {
        i = sqlfs_set_type(context->sqlfs, key, gs_type_to_string(value->type));
        if (i == 2)
            r = -EBUSY;
        else if (i == 0)
            r = -EIO;
    }

    /*fprintf(stderr, "set key %s result %d\n", key, r); *///??? debug use

    if (r == -EEXIST)
        return ALP_STATUS_GLOBAL_SETTINGS_KEY_EXISTS;
    else if ((r == -EPERM) || (r == -EACCES))
        return ALP_STATUS_GLOBAL_SETTINGS_PERMISSION_DENIED;
    else if (r == -ENOENT)
        return ALP_STATUS_GLOBAL_SETTINGS_KEY_NOT_FOUND;
    else if (r == -EBUSY)
        return ALP_STATUS_GLOBAL_SETTINGS_BUSY;
    else
        if (r < 0)
            return ALP_STATUS_GLOBAL_SETTINGS_ERROR;

    if (context->notification_on)
        send_key_change_notification(key);

    if (value->type == ALP_GLOBAL_SETTINGS_TYPE_LIST)
    {
        clean_serialize_buf(&sbuf);
    }

    return ALP_STATUS_OK;
}

static alp_status_t
gs_svr_proc_get( gs_svr_context *context,
                 const gchar *key,
                 AlpGlobalSettingsValue *value)
{
    key_attr  kattr = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
    key_value kvalue = { 0, 0, 0 };
    int i, result = ALP_STATUS_OK;
    CHECK_CONTEXT
    CHECK_ARG(key)
    CHECK_ARG(value)


    global_settings_value_clean(value);

    BEGIN
    if ((i = sqlfs_get_attr(context->sqlfs, key, &kattr)) <= 0)
    {
        COMPLETE(1)
        if (i == -1)
        {
            char tmp[PATH_MAX];

            snprintf(tmp,  sizeof(tmp), "%s.default", key);
            if (sqlfs_proc_access(context->sqlfs, tmp, R_OK) == 0)
                return gs_svr_proc_get(context, tmp, value);
            else
                return ALP_STATUS_GLOBAL_SETTINGS_KEY_NOT_FOUND;
        }
        else if (i == -2)
            return ALP_STATUS_GLOBAL_SETTINGS_PERMISSION_DENIED;
        else if (i == 2)
            return ALP_STATUS_GLOBAL_SETTINGS_BUSY;
        else return ALP_STATUS_GLOBAL_SETTINGS_ERROR;
    }
    else if (strcmp(kattr.type, TYPE_DIR))
    {
        i = sqlfs_get_value(context->sqlfs, key, &kvalue, 0, 0);

        if (i == 0)
        {
            result = ALP_STATUS_GLOBAL_SETTINGS_ERROR;
        }
        else if (i == 2)
            return ALP_STATUS_GLOBAL_SETTINGS_BUSY;
        else if (i == -1)
            result = ALP_STATUS_GLOBAL_SETTINGS_KEY_NOT_FOUND;
    }

    if (result == ALP_STATUS_OK)
    {
        if (!strcmp(kattr.type, TYPE_STRING))
        {
            value->type = ALP_GLOBAL_SETTINGS_TYPE_STRING;

            gstring_copy(&(value->string_value), kvalue.data);
        }
        else if (!strcmp(kattr.type, TYPE_BLOB))
        {
            value->type = ALP_GLOBAL_SETTINGS_TYPE_BLOB;
            value->blob_value.data = kvalue.data;
            value->blob_value.len = kvalue.size;
            memset(&kvalue, 0, sizeof(kvalue));
        }
        else if (!strcmp(kattr.type, TYPE_INT))
        {
            value->type = ALP_GLOBAL_SETTINGS_TYPE_INT;
            sscanf(kvalue.data, "%d", &(value->int_value));
        }
        else if (!strcmp(kattr.type, TYPE_DOUBLE))
        {
            value->type = ALP_GLOBAL_SETTINGS_TYPE_FLOAT;
            sscanf(kvalue.data, "%lf", &(value->float_value));
        }
        else if (!strcmp(kattr.type, TYPE_BOOL))
        {

            value->type = ALP_GLOBAL_SETTINGS_TYPE_BOOL;
            if (!strcasecmp(kvalue.data, "FALSE"))
                value->bool_value = FALSE;
            else
                value->bool_value = TRUE;
        }
        else if (!strcmp(kattr.type, TYPE_LIST))
        {
            value->type = ALP_GLOBAL_SETTINGS_TYPE_LIST;
            serialize_buf sbuf = {0, 0, 0};
            sbuf.data = kvalue.data;
            sbuf.length = kvalue.size;
            deserialize_gs_value(&sbuf, value);

        }
        else if (!strcmp(kattr.type, TYPE_DIR))
        {
            value->type = ALP_GLOBAL_SETTINGS_TYPE_DIR;

        }
        else
        {
            ALP_TM(T(ALP_CLASS_GLOBAL_SETTINGS, "unsupported value type"));
            result = ALP_STATUS_GLOBAL_SETTINGS_ERROR;
        }
    }
    clean_attr(&kattr);
    clean_value(&kvalue);

    COMPLETE(1)
    return result;
}



static alp_status_t
gs_svr_proc_get_type( gs_svr_context *context,
                      const gchar *key,
                      AlpGlobalSettingsValueType *type)
{
    key_attr  kattr = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

    int i, result = ALP_STATUS_OK;
    CHECK_CONTEXT
    CHECK_ARG(key)

    BEGIN
    if ((i = sqlfs_get_attr(context->sqlfs, key, &kattr)) <= 0)
    {
        COMPLETE(1)
        if (i == -1)
            return ALP_STATUS_GLOBAL_SETTINGS_KEY_NOT_FOUND;
        else if (i == -2)
            return ALP_STATUS_GLOBAL_SETTINGS_PERMISSION_DENIED;
        else if (i == 2)
            return ALP_STATUS_GLOBAL_SETTINGS_BUSY;
        else return ALP_STATUS_GLOBAL_SETTINGS_ERROR;
    }

    if (result == ALP_STATUS_OK)
    {
        if (!strcmp(kattr.type, TYPE_STRING))
        {
            *type = ALP_GLOBAL_SETTINGS_TYPE_STRING;

        }
        else if (!strcmp(kattr.type, TYPE_BLOB))
        {
            *type = ALP_GLOBAL_SETTINGS_TYPE_BLOB;

        }
        else if (!strcmp(kattr.type, TYPE_INT))
        {
            *type = ALP_GLOBAL_SETTINGS_TYPE_INT;

        }
        else if (!strcmp(kattr.type, TYPE_DOUBLE))
        {
            *type = ALP_GLOBAL_SETTINGS_TYPE_FLOAT;

        }
        else if (!strcmp(kattr.type, TYPE_BOOL))
        {

            *type = ALP_GLOBAL_SETTINGS_TYPE_BOOL;

        }
        else if (!strcmp(kattr.type, TYPE_LIST))
        {
            *type = ALP_GLOBAL_SETTINGS_TYPE_LIST;


        }
        else if (!strcmp(kattr.type, TYPE_DIR))
        {
            *type = ALP_GLOBAL_SETTINGS_TYPE_DIR;

        }
        else
        {
            ALP_TM(T(ALP_CLASS_GLOBAL_SETTINGS, "unsupported value type"));
            result = ALP_STATUS_GLOBAL_SETTINGS_ERROR;
        }
    }
    clean_attr(&kattr);


    COMPLETE(1)
    return result;
}


static alp_status_t
gs_svr_proc_get_value_size( gs_svr_context *context,
                            const gchar *key,
                            size_t *size)
{
    key_attr  kattr = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

    int i, result = ALP_STATUS_OK;
    CHECK_CONTEXT
    CHECK_ARG(key)

    BEGIN
    if ((i = sqlfs_get_attr(context->sqlfs, key, &kattr)) <= 0)
    {
        COMPLETE(1)
        if (i == -1)
        {
            char tmp[PATH_MAX];

            snprintf(tmp, sizeof(tmp), "%s.default", key);
            if (sqlfs_proc_access(context->sqlfs, tmp, R_OK) == 0)
                return gs_svr_proc_get_value_size(context, tmp, size);
            else
                return ALP_STATUS_GLOBAL_SETTINGS_KEY_NOT_FOUND;
        }
        else if (i == -2)
            return ALP_STATUS_GLOBAL_SETTINGS_PERMISSION_DENIED;
        else if (i == 2)
            return ALP_STATUS_GLOBAL_SETTINGS_BUSY;
        else return ALP_STATUS_GLOBAL_SETTINGS_ERROR;
    }

    if (result == ALP_STATUS_OK)
    {
        *size = kattr.size;
    }
    clean_attr(&kattr);


    COMPLETE(1)
    return result;
}



static alp_status_t
gs_svr_proc_delete( gs_svr_context *context, const gchar *key)
{
    int i, r;
    int result = ALP_STATUS_OK;
    CHECK_CONTEXT
    CHECK_ARG(key)
    BEGIN
    if ((i = sqlfs_is_dir(context->sqlfs, key)), (i == 2))
    {
        COMPLETE(1)
        return ALP_STATUS_GLOBAL_SETTINGS_BUSY;
    }
    else if (i == 1)
    {
        r = sqlfs_proc_rmdir(context->sqlfs,  key);
        if (r == -EBUSY)
        {
            result = ALP_STATUS_GLOBAL_SETTINGS_BUSY;
        }
        else if (r < 0)
            result = ALP_STATUS_GLOBAL_SETTINGS_ERROR;
        COMPLETE(1)
        return result;

    }

    r = sqlfs_proc_unlink(context->sqlfs, key);
    if (r == -ENOENT)
    {
        result = ALP_STATUS_GLOBAL_SETTINGS_KEY_NOT_FOUND;
    }
    else if (r == -EBUSY)
    {
        result = ALP_STATUS_GLOBAL_SETTINGS_BUSY;
    }
    else if (r < 0)
        result = ALP_STATUS_GLOBAL_SETTINGS_ERROR;
    if (result == ALP_STATUS_OK)
        if (context->notification_on)
            send_key_change_notification(key);

    COMPLETE(1)
    return result;
}


alp_status_t
gs_svr_proc_delete_tree(gs_svr_context *context, const gchar *key)
{

    int r;
    int result = ALP_STATUS_OK;
    CHECK_CONTEXT
    CHECK_ARG(key)
    r = sqlfs_del_tree(context->sqlfs, key);
    if (r == -ENOENT)
    {
        result = ALP_STATUS_GLOBAL_SETTINGS_KEY_NOT_FOUND;
    }
    else if (r < 0)
        result = ALP_STATUS_GLOBAL_SETTINGS_ERROR;
    if (result == ALP_STATUS_OK)
        if (context->notification_on)
            send_key_change_notification(key);
    return result;

}


static alp_status_t
gs_svr_proc_delete_tree_keeping_defaults(gs_svr_context *context, const gchar *key)
{

    int r;
    int result = ALP_STATUS_OK;
    CHECK_CONTEXT
    CHECK_ARG(key)
    r = sqlfs_del_tree_with_exclusion(context->sqlfs, key, "*.default");
    if (r == -ENOENT)
    {
        result = ALP_STATUS_GLOBAL_SETTINGS_KEY_NOT_FOUND;
    }
    else if (r < 0)
        result = ALP_STATUS_GLOBAL_SETTINGS_ERROR;
    if (result == ALP_STATUS_OK)
        if (context->notification_on)
            send_key_change_notification(key);
    return result;

}


#if 0
static void gconf_callback(GConfClient *gclient, guint id, GConfEntry *gentry, gpointer user_data)
{
    AlpGlobalSettingsEntry  entry = { 0, { 0, { 0 }}  };
    struct _Callback * c = (struct _Callback*) user_data ;
    ALP_FAIL_FATAL_IF(c == 0, "Callback table management failed!", "");

    if ((c->exact_key == 0) || (!strcmp(gentry->key, c->key->str)))
    {
        GConfEntry_to_global_settings_entry(gentry, &entry);
        c->func(c->context, &entry);
        global_settings_entry_clean(&entry);
    }
}
#endif
#if 0
static alp_status_t
gs_svr_keychange_callback_register( gs_svr_context *context,
                                    const gchar *key,
                                    const AlpGlobalSettingsCallbackcallback)
{
    struct _Callback *c = calloc(1, sizeof(*c));
    GError *err = NULL;

    ALP_FAIL_FATAL_IF(c == 0, "out of memory", "");

    CHECK_CONTEXT
    CHECK_ARG(key)
    c->context = context;
    c->exact_key = 1;
    c->key = g_string_new(key);
    c->func = callback;
    c->id = gconf_client_notify_add(context->gclient, key, gconf_callback, c, NULL, &err);
    if (err != NULL)
    {
        ALP_TM(T(ALP_CLASS_GLOBAL_SETTINGS, "%s", (err)->message));
        g_error_free(err);
        g_string_free(c->key, TRUE);
        free(c);
        return ALP_STATUS_GLOBAL_SETTINGS_ERROR;
    }
    ALP_FAIL_FATAL_IF(c->id == 0, "Invalud callback ID", "");
    g_hash_table_insert(context->callback_table, c->key->str, c);
    gconf_client_add_dir(c->context->gclient, key, GCONF_CLIENT_PRELOAD_ONELEVEL, &err);
    CHECK_ERROR_RETURN(err,  ALP_STATUS_GLOBAL_SETTINGS_ERROR);

    return ALP_STATUS_OK;
}

static alp_status_t
gs_svr_keychange_callback_unregister( gs_svr_context *context,
                                      const gchar *key)
{
    struct _Callback *c = 0;
    gchar *rkey;

    gboolean b;
    CHECK_CONTEXT
    CHECK_ARG(key)

    b = g_hash_table_lookup_extended(context->callback_table, key, (gpointer) &rkey, (gpointer)&c);
    if (b == FALSE)
    {
        return ALP_STATUS_GLOBAL_SETTINGS_KEY_NOT_FOUND;
    }
    gconf_client_notify_remove(context->gclient, c->id);

    g_string_free(c->key, TRUE);
    free(c);

    return ALP_STATUS_OK;
}

static alp_status_t
gs_svr_anykeychange_callback_register( gs_svr_context *context,
                                       const gchar *dir_path,
                                       const AlpGlobalSettingsCallbackcallback)
{
    struct _Callback *c = calloc(1, sizeof(*c));
    GError *err = NULL;

    ALP_FAIL_FATAL_IF(c == 0, "out of memory", "");

    CHECK_CONTEXT
    CHECK_ARG(dir_path)
    c->context = context;
    c->exact_key = 0;
    c->key = g_string_new(dir_path);
    c->func = callback;
    c->id = gconf_client_notify_add(context->gclient, dir_path, gconf_callback, (gpointer) c, NULL, &err);
    if (err != NULL)
    {
        ALP_TM(T(ALP_CLASS_GLOBAL_SETTINGS, "%s", (err)->message));
        g_error_free(err);
        g_string_free(c->key, TRUE);
        free(c);
        return ALP_STATUS_GLOBAL_SETTINGS_ERROR;
    }
    g_hash_table_insert(context->callback_table, c->key->str, c);
    gconf_client_add_dir(c->context->gclient, dir_path, GCONF_CLIENT_PRELOAD_ONELEVEL, &err);
    CHECK_ERROR_RETURN(err,  ALP_STATUS_GLOBAL_SETTINGS_ERROR);

    return ALP_STATUS_OK;
}

static alp_status_t
gs_svr_anykeyschange_callback_unregister( gs_svr_context *context,
        const gchar *dir_path)
{
    struct _Callback *c = 0;
    gchar *rkey;
    gboolean b;
    CHECK_CONTEXT
    CHECK_ARG(dir_path)

    b = g_hash_table_lookup_extended(context->any_callback_table, dir_path, (gpointer) &rkey, (gpointer) &c);
    if (b == FALSE)
    {
        return ALP_STATUS_GLOBAL_SETTINGS_KEY_NOT_FOUND;
    }
    gconf_client_notify_remove(context->gclient, c->id);

    g_string_free(c->key, TRUE);
    free(c);
    return ALP_STATUS_OK;
}
#endif



static int fill_dir_entry_func(void *buf, const char *name, const struct stat *stbuf, off_t off)
{
    GList **list = (GList **) buf;
    if (strcmp(name, ".") && strcmp(name, ".."))
        *list = g_list_prepend(*list, (gpointer) strdup(name));
    return 0;
}

static alp_status_t
gs_svr_proc_findkeys_start( gs_svr_context *context, const gchar *dir_path, AlpGlobalSettingsEntryIterator **iterator)
{
    GError *err = 0;
    GList *list = 0;
    int i, result = ALP_STATUS_OK;
    CHECK_CONTEXT

    i = sqlfs_proc_readdir(context->sqlfs, dir_path, &list, fill_dir_entry_func, 0, 0);
    if (i == -ENOENT)
        return ALP_STATUS_GLOBAL_SETTINGS_KEY_NOT_FOUND;
    else
        if (i == -ENOTDIR)
            return ALP_STATUS_GLOBAL_SETTINGS_VALUE_NOT_DIR;
        else
            if (i < 0)
                return ALP_STATUS_GLOBAL_SETTINGS_ERROR;
    list = g_list_reverse(list);
    *iterator = calloc(1, sizeof(**iterator));
    (*iterator)->entries = (*iterator)->current = list;
    while ((*iterator)->current && (!strcmp((*iterator)->current->data, ".") || !strcmp((*iterator)->current->data, "..")))
        (*iterator)->current = (*iterator)->current->next;
    if ((*iterator)->current == NULL)
    {
        result = ALP_STATUS_GLOBAL_SETTINGS_ENTRY_ITERATOR_AT_END;
    }
    (*iterator)->base_path = strdup(dir_path);

    return result;
}

static alp_status_t
gs_svr_proc_list_keys( gs_svr_context *context, const gchar *pattern, AlpGlobalSettingsEntryIterator **iterator)
{
    GError *err = 0;
    GList *list = 0;
    int i, result = ALP_STATUS_OK;
    CHECK_CONTEXT

    i = sqlfs_list_keys(context->sqlfs, pattern, &list, fill_dir_entry_func);
    if (i == -ENOENT)
        return ALP_STATUS_GLOBAL_SETTINGS_KEY_NOT_FOUND;
    else
        if (i == -ENOTDIR)
            return ALP_STATUS_GLOBAL_SETTINGS_VALUE_NOT_DIR;
        else
            if (i < 0)
                return ALP_STATUS_GLOBAL_SETTINGS_ERROR;
    list = g_list_reverse(list);
    *iterator = calloc(1, sizeof(**iterator));
    (*iterator)->entries = (*iterator)->current = list;
    while ((*iterator)->current && (!strcmp((*iterator)->current->data, ".") || !strcmp((*iterator)->current->data, "..")))
        (*iterator)->current = (*iterator)->current->next;
    if ((*iterator)->current == NULL)
    {
        result = ALP_STATUS_GLOBAL_SETTINGS_ENTRY_ITERATOR_AT_END;
    }
    (*iterator)->base_path = strdup("/");

    return result;
}


/* beginning of the installation of mass preference installation from an XML file */

typedef struct keyvalue
{
    char *key;
    char *text;
    char *value;
    char *type;
    char *value_type;
    char *mode;
    char *uid;
    char *gid;
    int in_valid_element;

}
keyvalue_t;



static const char *type_string = "type", *name_string = "name", *value_string = "value",
                                 *key_string = "key", *mode_string = "mode", *keyvalue_string ="keyvalue",
                                                                     *uid_string = "uid", *gid_string = "gid";

static void error_handle(GMarkupParseContext *context,
                         GError              *error,
                         gpointer             user_data)
{
    ALP_TM(T(ALP_CLASS_GLOBAL_SETTINGS, "%s\n", error->message));
}


static  void   start_element(GMarkupParseContext *context, const gchar *element_name, const gchar **attribute_names,
                             const gchar        **attribute_values, gpointer             user_data, GError             **error)
{
    gs_svr_context *svr_context = (gs_svr_context *) user_data;
    keyvalue_t *top = g_list_nth_data(svr_context->node_stack, 0);
    keyvalue_t *kv;
    int i;
    if (!strcmp(element_name, "keyvalue") || !strcmp(element_name, "default_mode") ||  !strcmp(element_name, "default_uid") || !strcmp(element_name, "default_gid") )
    {

        kv = calloc(1, sizeof(*kv));
        svr_context->node_stack = g_list_prepend(svr_context->node_stack, (gpointer) kv);
        if (strcmp(element_name, "keyvalue"))
        {
            top = g_list_nth_data(svr_context->node_stack, 0);
            top->in_valid_element = 1;
        }

    }

    else

        if (top)
        {
            if (!strcmp(element_name, type_string))
            {
                top->in_valid_element = 1;
            }
            else if (!strcmp(element_name, key_string))
            {
                top->in_valid_element = 1;
            }
            else if (!strcmp(element_name, mode_string))
            {
                top->in_valid_element = 1;
            }
            else if (!strcmp(element_name, uid_string))
            {
                top->in_valid_element = 1;
            }
            else if (!strcmp(element_name, gid_string))
            {
                top->in_valid_element = 1;
            }
            else
                if (!strcmp(element_name, value_string))
                {
                    top->in_valid_element = 1;
                    for (i = 0; attribute_names[i]; i++)
                    {
                        if (!strcmp(attribute_names[i], type_string))
                        {
                            top->value_type = strdup(attribute_values[i]);
                            break;
                        }
                    }
                }
        }
}

static void install_keyvalue(gs_svr_context *context, const keyvalue_t *kv, GMarkupParseContext *gp);

static void end_element(GMarkupParseContext *context, const gchar *element_name, gpointer user_data, GError             **error)
{
    gs_svr_context *svr_context = (gs_svr_context *) user_data;
    keyvalue_t *top = g_list_nth_data(svr_context->node_stack, 0);
    if (!strcmp(element_name, "keyvalue"))
    {
        ALP_FAIL_FATAL_IF(top == 0, "Bad XML preference file", "");
        install_keyvalue((gs_svr_context*) user_data, top, context);
        if (top->key)
            free(top->key);
        if (top->value)
            free(top->value);
        if (top->text)
            free(top->text);
        if (top->type)
            free(top->type);
        if (top->mode)
            free(top->mode);
        if (top->uid)
            free(top->uid);
        if (top->gid)
            free(top->gid);
        if (top->value_type)
            free(top->value_type);
        svr_context->node_stack = g_list_remove(svr_context->node_stack, top);
        free(top);

        return;
    }
    else if (!strcmp(element_name, "default_mode"))
    {
        int i, i2;
        i = sscanf(top->text, "%i", &i2);
        if (i == 1)
        {
            svr_context->sqlfs->default_mode = (mode_t)  i2;
        }
        if (top->text)
            free(top->text);
        svr_context->node_stack = g_list_remove(svr_context->node_stack, top);
        free(top);

        return;
    }
    else if (!strcmp(element_name, "default_uid"))
    {
        svr_context->default_uid_in_import = top->text;
        svr_context->sqlfs->default_xattrs[1].value = top->text;
        svr_context->sqlfs->default_xattrs[1].value_size = strlen(top->text) + 1;

        svr_context->node_stack = g_list_remove(svr_context->node_stack, top);
        free(top);

        return;
    }
    else if (!strcmp(element_name, "default_gid"))
    {
        svr_context->default_gid_in_import = top->text;
        svr_context->sqlfs->default_xattrs[0].value = top->text;
        svr_context->sqlfs->default_xattrs[0].value_size = strlen(top->text) + 1;

        svr_context->node_stack = g_list_remove(svr_context->node_stack, top);
        free(top);

        return;
    }
    else
        if (top)
        {
            if (!strcmp(element_name, type_string))
            {
                top->type = top->text;
            }
            else if (!strcmp(element_name, value_string))
            {
                top->value = top->text;
            }
            else if (!strcmp(element_name, key_string))
            {
                char absolute_key[PATH_MAX];
                top->key = top->text;
                form_absolute_path(svr_context->cwd, top->key, absolute_key);
                free(top->key);
                top->key = strdup(absolute_key);
            }
            else if (!strcmp(element_name, mode_string))
            {
                top->mode = top->text;
            }
            else if (!strcmp(element_name, gid_string))
            {
                top->gid = top->text;
            }
            else if (!strcmp(element_name, uid_string))
            {
                top->uid = top->text;
            }
            else
            {
                ALP_TM(T(ALP_CLASS_GLOBAL_SETTINGS,"Unsupported XML node  %s\n", element_name));
                if (top->text)
                {
                    free(top->text);
                }
            }

            top->text = 0;
            top->in_valid_element = 0;
        }

}


static void text_func(GMarkupParseContext *context,
                      const gchar         *text,
                      gsize                text_len,
                      gpointer             user_data,
                      GError             **error)
{
    gs_svr_context *svr_context = (gs_svr_context *) user_data;
    keyvalue_t *top = g_list_nth_data(svr_context->node_stack, 0);
    if (top && top->in_valid_element)
    {
        top->text = calloc(text_len + 1, sizeof(char));
        memcpy(top->text, text, text_len);

    }

}


static int hex_to_blob(const char *hex, char **blob, size_t *length)
{
    int str_len = strlen(hex);
    const char *c, *end = hex + str_len;
    unsigned int b;
    int i, j, n, r = 1;
    *blob = malloc(sizeof(char) * str_len);

    ALP_FAIL_FATAL_IF(*blob == 0, "Out of memory!", "");

    for (c = hex, i = 0; c < end; i++)
    {
        j = sscanf(c, "%2x%n", &b, &n);
        ALP_FAIL_NON_FATAL_IF(j < 1, "Invalid hex string in XML file! imported failed \n", "");
        if (j < 1)
        {
            r = 0;
            free(*blob);
            *blob = 0;
            *length = 0;
            break;
        }
        *(unsigned char*) (*blob + i) = (unsigned char) b;
        c += n;
    }
    *length = (size_t) i;



    return r;

}

static void install_keyvalue(gs_svr_context *context, const keyvalue_t *kv,
                             GMarkupParseContext *gp)
{
    alp_status_t status = ~ALP_STATUS_OK;
    int i;
    context->last_op_time = time(NULL);
    if (kv->key && kv->type)
    {
        if (kv->value && (!strcmp(kv->type, "boolean") || !strcmp(kv->type, "bool")))
        {
            if (!strcasecmp(kv->value, "true"))
                status = gs_svr_set_key_bool_value(context, kv->key, TRUE);
            else if (!strcasecmp(kv->value, "false"))
                status = gs_svr_set_key_bool_value(context, kv->key, FALSE);
            else ALP_TM(T(ALP_CLASS_GLOBAL_SETTINGS,"Illegal boolean value %s\n", kv->value));
        }
        else if (kv->value && !strcmp(kv->type, "float"))
        {
            double f;
            i = sscanf(kv->value, "%lf", &f);
            if (i != 1)
            {
                ALP_TM(T(ALP_CLASS_GLOBAL_SETTINGS,"Illegal float value %s\n", kv->value));
            }
            else
                status = gs_svr_set_key_float_value(context, kv->key, f);
        }
        else if (kv->value && !strcmp(kv->type, "int"))
        {
            int v;
            unsigned int u;
            if (kv->value[0] == '\'')
            {
                char *t = kv->value + 1;
                u = 0;
                while (*t != '\'')
                {
                    u = (u << 8) | *t;
                    t++;
                }
                v = (int) u;
                i = 1;
            }
            else
                if ((kv->value[0] == '0') && (tolower(kv->value[1]) == 'x'))
                    i = sscanf(kv->value, "%x", &u), v = (int) u;
                else if  (kv->value[0] == '0')
                    i = sscanf(kv->value, "%o", &u), v = (int) u;
                else
                    i = sscanf(kv->value, "%d", &v);
            if (i != 1)
            {
                ALP_TM(T(ALP_CLASS_GLOBAL_SETTINGS,"Illegal int value %s\n", kv->value));
            }
            else
                status = gs_svr_set_key_int_value(context, kv->key, v);
        }
        else if (kv->value && !strcmp(kv->type, "string"))
        {
            status = gs_svr_set_key_string_value(context, kv->key, kv->value);
        }
        else if (kv->value && !strcmp(kv->type, "blob"))
        {
            int v;
            unsigned int u;
            if (kv->value_type && !strcmp(kv->value_type, "int"))
            {
                if (kv->value[0] == '\'')
                {
                    char *t = kv->value + 1;
                    u = 0;
                    while (*t != '\'')
                    {
                        u = (u << 8) | *t;
                        t++;
                    }
                    v = (int) u;
                    i = 1;
                }
                else
                    if ((kv->value[0] == '0') && (tolower(kv->value[1]) == 'x'))
                        i = sscanf(kv->value, "%x", &u), v = (int) u;
                    else if  (kv->value[0] == '0')
                        i = sscanf(kv->value, "%o", &u), v = (int) u;
                    else
                        i = sscanf(kv->value, "%d", &v);
                if (i != 1)
                    ALP_TM(T(ALP_CLASS_GLOBAL_SETTINGS,"Illegal int value %s\n", kv->value));
                else
                    status = gs_svr_set_key_blob_value(context, kv->key, (char*) &v, sizeof(v));
            }
            else
                if (kv->value_type && !strcmp(kv->value_type, "string"))
                {
                    status = gs_svr_set_key_blob_value(context, kv->key, kv->value, strlen(kv->value) + 1);  /* ensure the terminating "0" is stored as well */
                }
                else if (kv->value_type && !strcmp(kv->value_type, "hex"))
                {
                    char *blob = 0;
                    int blob_size;

                    i = hex_to_blob(kv->value, &blob, &blob_size);
                    if (i == 1)
                    {
                        status = gs_svr_set_key_blob_value(context, kv->key, blob, blob_size);
                    }
                    if (blob)
                        free(blob);

                }

                else if (kv->value_type == 0)
                {
                    ALP_TM(T(ALP_CLASS_GLOBAL_SETTINGS,"binary initial blob value not supported yet\n"));
                }
                else
                    ALP_TM(T(ALP_CLASS_GLOBAL_SETTINGS,"Unsupported blob value %s\n", kv->value));



        }
        else if (!strcmp(kv->type, "dir"))
        {
            status = gs_svr_set_key_directory(context, kv->key);
        }

        if ((status == ALP_STATUS_OK) && (kv->mode && kv->key))
        {
            mode_t mode;
            int i2;
            i = sscanf(kv->mode, "%i", &i2);
            if (i == 1)
            {
                status = gs_svr_proc_set_mode(context, kv->key, (mode_t)  i2);

            }
            else
                ALP_TM(T(ALP_CLASS_GLOBAL_SETTINGS,"Unsupported mode value %s\n", kv->mode));


        }

        if ((status == ALP_STATUS_OK) && (kv->key && (kv->uid || context->default_uid_in_import)))
        {
            const char *uid = kv->uid ? kv->uid : context->default_uid_in_import;
            status = sqlfs_proc_setxattr(context->sqlfs, kv->key, uid_string, uid, strlen(uid) + 1, 0);

        }

        if ((status == ALP_STATUS_OK) && (kv->key && (kv->gid || context->default_gid_in_import)))
        {
            const char *gid = kv->gid ? kv->gid : context->default_gid_in_import;
            status = sqlfs_proc_setxattr(context->sqlfs, kv->key, gid_string, gid, strlen(gid) + 1, 0);

        }
        if (status != ALP_STATUS_OK)
        {
            gint line, pos;
            g_markup_parse_context_get_position(gp, &line, &pos);
            ALP_TM(T(ALP_CLASS_GLOBAL_SETTINGS,"Bad XML node at line %d char %d, key %s not inserted, due to error %s\n",
                     line, pos, kv->key, status_to_string(status)));
        }
    }
}


alp_status_t
gs_svr_proc_import_from_file(gs_svr_context *context,
                             const char *file)
{
    int i;
    FILE *f = 0;
    alp_status_t status = ALP_STATUS_OK;

    static const char *type_string = "type";
    GMarkupParser parse;
    GMarkupParseContext *parse_context = 0;
    GError *error = 0;
    mode_t saved_default_mode ;
    xattr_t xattrs[2] = { { gid_string, 0, 0}, { uid_string, 0, 0 }};

    static const int buf_size = 128 * 1024;
    char *buf = 0;
    size_t n_read ;

    memset(&parse, 0, sizeof(parse));
    parse.start_element = start_element;
    parse.end_element = end_element;
    parse.text = text_func;
    parse.error = error_handle;

    context->node_stack = 0;
    context->default_uid_in_import = 0;
    context->default_gid_in_import = 0;

    context->sqlfs->default_xattrs = xattrs;
    context->sqlfs->num_default_xattrs = sizeof(xattrs) / sizeof(xattrs[0]);
    saved_default_mode = context->sqlfs->default_mode;
    parse_context = g_markup_parse_context_new(&parse, 0, (gpointer) context, NULL);

    f  = fopen(file, "r");
    if (!f)
    {
        ALP_TM(T(ALP_CLASS_GLOBAL_SETTINGS,"Cannot open %s, %s\n", file, strerror(errno)));

        status = ALP_STATUS_GLOBAL_SETTINGS_ERROR;
    }
    else
    {
        buf = malloc(buf_size * sizeof(char));

        if (!buf)
        {
            ALP_TM(T(ALP_CLASS_GLOBAL_SETTINGS,"Out of memory \n"));
            status = ALP_STATUS_GLOBAL_SETTINGS_ERROR;
        }
        if (status == ALP_STATUS_OK)
            while (1)
            {
                n_read = fread(buf, sizeof(buf[0]), sizeof(char) * buf_size, f);
                if (n_read > 0)
                {

                    if (g_markup_parse_context_parse(parse_context, buf, n_read, &error) == FALSE)
                    {

                        ALP_TM(T(ALP_CLASS_GLOBAL_SETTINGS,"error parsing %s, %s\n", file, error->message));
                        status = ALP_STATUS_GLOBAL_SETTINGS_ERROR;
                        break;
                    }
                }
                if (feof(f) || ferror(f))
                    break;
            }


        fclose(f);
        if (buf)
            free(buf);
    }
    if (status != ALP_STATUS_OK)
    {
        gint line, pos;
        g_markup_parse_context_get_position(parse_context, &line, &pos);
        ALP_TM(T(ALP_CLASS_GLOBAL_SETTINGS,"Bad XML node at line %d char %d due to error %s\n",
                 line, pos,  status_to_string(status)));
    }

    if (g_markup_parse_context_end_parse(parse_context, &error) == FALSE)
    {
        ALP_TM(T(ALP_CLASS_GLOBAL_SETTINGS,"error ending parsing %s, %s\n", file, error->message));

    }
    g_markup_parse_context_free(parse_context);

    if (context->default_uid_in_import) free(context->default_uid_in_import);
    if (context->default_gid_in_import) free(context->default_gid_in_import);
    context->default_uid_in_import = 0;
    context->default_gid_in_import = 0;
    context->sqlfs->default_xattrs = 0;
    context->sqlfs->num_default_xattrs = 0;
    context->sqlfs->default_mode = saved_default_mode;
    return status;

}


/* end of the installation of mass preference installation from an XML file */

static void context_do_close(gpointer value)
{

    gs_svr_context *context = (gs_svr_context*) value;
    gs_svr_close(context);
}

static gboolean context_to_close(gpointer key, gpointer value, gpointer user_data)
{
    int thread_id = (int) user_data;
    gs_svr_context *context = (gs_svr_context*) value;
    if ((context->thread_id == thread_id) && (context->to_close))
    {

        return TRUE;
    }
    return FALSE;
}

static int to_quit = 0;
struct conn_msg
{
    AlpConnection *conn;
    AlpMessage *msg;
} ;

static void thread_func(gpointer data, gpointer user_data)
{
#define CONTEXT_OPERATION_EXIT\
    CONTEXT_HASH_LOCK \
    context_operations_in_progress--;\
    if (context_operations_in_progress == 0) pthread_cond_signal(&no_context_operation_cond); \
    CONTEXT_HASH_UNLOCK

    int thread_id = (int) user_data;
    int quit_flag = 0;
    gs_svr_context *context = 0;

    CONTEXT_HASH_LOCK
    context_operations_in_progress++;

    CONTEXT_HASH_UNLOCK

    if (data != (gpointer) -1)
    {

        struct conn_msg *conn_msg = (struct conn_msg *) data;
        AlpMessage *msg = conn_msg->msg, *resp;
        AlpChannel *channel = (AlpChannel*) user_data;
        char *cmd;
        char *key;
        char *file;
        char *env_switch = 0;
        int32_t msg_id;
        gint pid; /* process id of the client process */

        ALP_GLOBAL_SETTINGS_RPC_PROCS proc;
        uint32_t ver, m;
        mode_t mt;
        uid_t uid;
        gid_t gid;
        size_t size;
        AlpGlobalSettingsValueType type;

        alp_status_t status;
        AlpGlobalSettingsValue *value = 0;
        AlpGlobalSettingsEntryIterator *iterator = 0;

        ALP_FAIL_FATAL_IF(msg == 0, "zero pointer: Illegal message!", "");
        pid = alp_ipc_connection_get_peer_pid(conn_msg->conn);

        if (pid == (gint) -1)
        {
            ALP_TM(T(ALP_CLASS_GLOBAL_SETTINGS, "Peer has no pid!", ""));
            alp_ipc_message_destroy(msg);
            free(conn_msg);
            CONTEXT_OPERATION_EXIT
            return ;
        }

        uid = alp_ipc_connection_get_peer_uid(conn_msg->conn);
        gid = alp_ipc_connection_get_peer_gid(conn_msg->conn);

        CONTEXT_HASH_LOCK

        context = (gs_svr_context*) g_hash_table_lookup(context_table, (gconstpointer)
                  alp_ipc_connection_get_custom_data(conn_msg->conn));
        CONTEXT_HASH_UNLOCK



        msg_id = alp_ipc_message_get_message_ID(msg);
        alp_ipc_message_unpack_start(msg);

        alp_ipc_message_unpack_uint32(msg, &ver);
        resp =   alp_ipc_message_create();
        alp_ipc_message_response_pack_start(resp, msg_id, msg);
        alp_ipc_message_pack_uint32(resp, ALP_GLOBAL_SETTINGS_VERSION);
        if (ver != ALP_GLOBAL_SETTINGS_VERSION)
        {
            alp_ipc_message_unpack_end(msg);
            alp_ipc_message_pack_uint32(resp, -1);
            alp_ipc_message_pack_string(resp, "Illegal version");
            alp_ipc_message_pack_end(resp);
            alp_ipc_connection_send(conn_msg->conn, resp);
            alp_ipc_message_destroy(resp);
            alp_ipc_message_destroy(msg);
            free(conn_msg);
            CONTEXT_OPERATION_EXIT
            return ;
        }
        alp_ipc_message_unpack_uint32(msg, &m);
        proc = (ALP_GLOBAL_SETTINGS_RPC_PROCS) m;
#if 0
        /* security */

        if ((env_switch = getenv( "NLPD_USE_LSM" )) && !strcmp("TRUE", env_switch))
        {
            static pthread_mutex_t spf_lock = PTHREAD_MUTEX_INITIALIZER ;
            char * assertion_value = 0;
            const char *policy_name = policy_strings_get(proc);
            char *t = 0;
            /*fprintf(stderr, "policy name is %s\n", policy_name);*///???

            if (policy_name)
            {
                pthread_mutex_lock(&spf_lock);
                status = alp_spf_query_open();
                if (status == ALP_STATUS_SPF_E_NOERR )
                {

                    alp_spf_query_set_subject_pid(pid);
                    t = strdup(policy_name);
                    status = alp_spf_query_get_assertion_string(t /*policy_name*/, strlen(policy_name), &assertion_value);
                    free(t);
                    if ((status == ALP_STATUS_SPF_E_NOERR) && assertion_value && strcmp(assertion_value, "true"))
                    {
                        alp_ipc_message_unpack_end(msg);
                        alp_ipc_message_pack_uint32(resp, (uint32_t) ALP_STATUS_GLOBAL_SETTINGS_PERMISSION_DENIED);

                        alp_ipc_message_pack_end(resp);
                        alp_ipc_connection_send(conn_msg->conn, resp);
                        alp_ipc_message_destroy(resp);
                        alp_ipc_message_destroy(msg);
                        free(conn_msg);
                        g_free(assertion_value);
                        alp_spf_query_close();
                        pthread_mutex_unlock(&spf_lock);
                        CONTEXT_OPERATION_EXIT
                        return ;
                    }

                }
                if (assertion_value)
                    g_free(assertion_value);
                alp_spf_query_close();
                pthread_mutex_unlock(&spf_lock);
            }
        }
        /* security */
#endif
        if (proc != PROC_open)
        {
            if ((context == NULL) && (proc != PROC_ping) && (proc != PROC_server_quit))
            {

                alp_ipc_message_unpack_end(msg);
                alp_ipc_message_pack_uint32(resp, -1);
                alp_ipc_message_pack_string(resp, "Client has no open connection to Global Settings!");
                alp_ipc_message_pack_end(resp);
                alp_ipc_connection_send(conn_msg->conn, resp);
                alp_ipc_message_destroy(resp);
                alp_ipc_message_destroy(msg);
                free(conn_msg);
                CONTEXT_OPERATION_EXIT
                return ;


            }
        }
        else if (context != NULL)
        {
            alp_ipc_message_unpack_end(msg);
            alp_ipc_message_pack_uint32(resp, -1);
            alp_ipc_message_pack_string(resp, "Client already has open connection to Global Settings!");
            alp_ipc_message_pack_end(resp);
            alp_ipc_connection_send(conn_msg->conn, resp);
            alp_ipc_message_destroy(resp);
            alp_ipc_message_destroy(msg);
            free(conn_msg);
            CONTEXT_OPERATION_EXIT
            return ;

        }

        switch(proc)
        {

        case PROC_open:
            status = gs_svr_open(0, &context);
            ALP_FAIL_FATAL_IF(!(status == ALP_STATUS_OK), "", "");
            ALP_FAIL_FATAL_IF(!(context != 0), "", "");
            context->sqlfs->uid = uid;
            context->sqlfs->gid = gid;

            context->pid = pid;
            context->thread_id = thread_id;
            context->to_close = 0;
            alp_ipc_connection_set_custom_data(conn_msg->conn, (gpointer) next_connection_serial_no());
            CONTEXT_HASH_LOCK
            g_hash_table_insert(context_table, (gpointer) alp_ipc_connection_get_custom_data(conn_msg->conn), (gpointer) context);
            CONTEXT_HASH_UNLOCK
            alp_ipc_message_pack_uint32(resp, (uint32_t) status);

            break;

        case PROC_close:
            context->to_close = 1;

            status = ALP_STATUS_OK;
            alp_ipc_message_pack_uint32(resp, (uint32_t) status);
            break;

        case PROC_begin_transaction:
            status = gs_svr_proc_begin_transaction(context);
            alp_ipc_message_pack_uint32(resp, (uint32_t) status);
            break;
        case PROC_complete_transaction:
            status = gs_svr_proc_complete_transaction(context);
            alp_ipc_message_pack_uint32(resp, (uint32_t) status);
            break;
        case PROC_cancel_transaction:
            status = gs_svr_proc_cancel_transaction(context);
            alp_ipc_message_pack_uint32(resp, (uint32_t) status);
            break;
        case PROC_set_default_mode:
            alp_ipc_message_unpack_uint32(msg, &m);
            status = gs_svr_proc_set_default_mode(context, m);
            alp_ipc_message_pack_uint32(resp, (uint32_t) status);
            break;
        case PROC_get_default_mode:
            status = gs_svr_proc_get_default_mode(context, &mt);
            alp_ipc_message_pack_uint32(resp, (uint32_t) status);
            if (status == ALP_STATUS_OK)
            {
                m = mt;
                alp_ipc_message_pack_uint32(resp, m);
            }
            break;
        case PROC_set_mode:
            alp_ipc_message_unpack_string(msg, &key);
            alp_ipc_message_unpack_uint32(msg, &m);
            status = gs_svr_proc_set_mode(context, key, m);
            alp_ipc_message_pack_uint32(resp, (uint32_t) status);
            g_free(key);
            break;
        case PROC_get_mode:
            alp_ipc_message_unpack_string(msg, &key);
            status = gs_svr_proc_get_mode(context, key, &mt);
            alp_ipc_message_pack_uint32(resp, (uint32_t) status);
            if (status == ALP_STATUS_OK)
            {
                m = mt;
                alp_ipc_message_pack_uint32(resp, m);
            }
            g_free(key);
            break;

        case PROC_set_uid:
            alp_ipc_message_unpack_string(msg, &key);
            alp_ipc_message_unpack_uint32(msg, &m);
            status = gs_svr_proc_set_uid(context, key, (uid_t) m);
            alp_ipc_message_pack_uint32(resp, (uint32_t) status);
            g_free(key);
            break;
        case PROC_get_uid:
            alp_ipc_message_unpack_string(msg, &key);
            status = gs_svr_proc_get_uid(context, key, &uid);
            alp_ipc_message_pack_uint32(resp, (uint32_t) status);
            if (status == ALP_STATUS_OK)
            {
                m = uid;
                alp_ipc_message_pack_uint32(resp, m);
            }
            g_free(key);
            break;
        case PROC_set_gid:
            alp_ipc_message_unpack_string(msg, &key);
            alp_ipc_message_unpack_uint32(msg, &m);
            status = gs_svr_proc_set_gid(context, key, (gid_t) m);
            alp_ipc_message_pack_uint32(resp, (uint32_t) status);
            g_free(key);
            break;
        case PROC_get_gid:
            alp_ipc_message_unpack_string(msg, &key);
            status = gs_svr_proc_get_gid(context, key, &gid);
            alp_ipc_message_pack_uint32(resp, (uint32_t) status);
            if (status == ALP_STATUS_OK)
            {
                m = gid;
                alp_ipc_message_pack_uint32(resp, m);
            }
            g_free(key);
            break;

        case PROC_set:
            alp_ipc_message_unpack_string(msg, &key);
            value =  alp_global_settings_value_alloc();
            gs_unpack_value(msg, value);
            status = gs_svr_proc_set(context, key, value);
            alp_ipc_message_pack_uint32(resp, (uint32_t) status);
            g_free(key);
            alp_global_settings_value_free(value);
            break;


        case PROC_get:
            alp_ipc_message_unpack_string(msg, &key);
            value =  alp_global_settings_value_alloc();
            status = gs_svr_proc_get(context, key, value);
            alp_ipc_message_pack_uint32(resp, (uint32_t) status);
            gs_pack_value(resp, value);
            g_free(key);
            alp_global_settings_value_free(value);
            break;

        case PROC_delete:
            alp_ipc_message_unpack_string(msg, &key);

            status = gs_svr_proc_delete(context, key);
            alp_ipc_message_pack_uint32(resp, (uint32_t) status);
            g_free(key);
            break;

        case PROC_delete_tree:
            alp_ipc_message_unpack_string(msg, &key);

            status = gs_svr_proc_delete_tree(context, key);
            alp_ipc_message_pack_uint32(resp, (uint32_t) status);
            g_free(key);
            break;


        case PROC_delete_tree_keeping_defaults:
            alp_ipc_message_unpack_string(msg, &key);

            status = gs_svr_proc_delete_tree_keeping_defaults(context, key);
            alp_ipc_message_pack_uint32(resp, (uint32_t) status);
            g_free(key);
            break;

        case PROC_get_type:
            alp_ipc_message_unpack_string(msg, &key);
            status = gs_svr_proc_get_type(context, key, &type);
            alp_ipc_message_pack_uint32(resp, (uint32_t) status);
            if (status == ALP_STATUS_OK)
            {
                m = (int32_t) type;
                alp_ipc_message_pack_uint32(resp, m);
            }
            g_free(key);
            break;

        case PROC_get_value_size:
            alp_ipc_message_unpack_string(msg, &key);
            status = gs_svr_proc_get_value_size(context, key, &size);
            alp_ipc_message_pack_uint32(resp, (uint32_t) status);
            if (status == ALP_STATUS_OK)
            {
                m = (int32_t) size;
                alp_ipc_message_pack_uint32(resp, m);
            }
            g_free(key);
            break;

        case PROC_findkeys_start:

            alp_ipc_message_unpack_string(msg, &key);

            status = gs_svr_proc_findkeys_start(context, key, &iterator);
            alp_ipc_message_pack_uint32(resp, (uint32_t) status);
            if (status ==  ALP_STATUS_OK)
            {

                gs_svr_pack_entry_iterator(resp, iterator);
            }
            g_free(key);
            if (iterator)
                alp_global_settings_entry_iterator_free(iterator);
            break;



        case PROC_list_keys:

            alp_ipc_message_unpack_string(msg, &key);

            status = gs_svr_proc_list_keys(context, key, &iterator);
            alp_ipc_message_pack_uint32(resp, (uint32_t) status);

            if (status ==  ALP_STATUS_OK)
            {


                gs_svr_pack_entry_iterator(resp, iterator);

            }
            g_free(key);
            if (iterator)
                alp_global_settings_entry_iterator_free(iterator);
            break;


        case PROC_set_notification_on_off:
            alp_ipc_message_unpack_uint32(msg, &m);
            status = gs_svr_proc_set_notification_on_off(context, m);
            alp_ipc_message_pack_uint32(resp, (uint32_t) status);
            break;

        case PROC_import_from_file:
            alp_ipc_message_unpack_string(msg, &file);
            if (context->cwd)
                g_free(context->cwd);

            alp_ipc_message_unpack_string(msg, &(context->cwd));
            status = gs_svr_proc_import_from_file(context, file);
            alp_ipc_message_pack_uint32(resp, (uint32_t) status);
            g_free(file);
            break;


        case PROC_server_quit:
            status = ALP_STATUS_OK;
            alp_ipc_message_pack_uint32(resp, (uint32_t) status);
            quit_flag = 1;
            break;

        case PROC_ping:
            status = ALP_STATUS_OK;
            alp_ipc_message_pack_uint32(resp, (uint32_t) status);

            break;
        default:
            ALP_FAIL_FATAL_IF(1, "Illegal procedure!", "");
        }
        alp_ipc_message_unpack_end(msg);

        alp_ipc_message_pack_end(resp);
        alp_ipc_connection_send(conn_msg->conn, resp);
        alp_ipc_message_destroy(resp);
        alp_ipc_message_destroy(msg);
        free(conn_msg);
    }
    thread_now[thread_id] = time(NULL);
    if (context)
        context->last_op_time = thread_now[thread_id];
    if (thread_now[thread_id] - thread_last_check_time[thread_id] > TRANSACTION_CHECK_INTERVAL)
    {
        thread_context_periodic_check(thread_id);
        thread_last_check_time[thread_id] = thread_now[thread_id];
    }

    CONTEXT_HASH_LOCK
    g_hash_table_foreach_remove(context_table, context_to_close, (gpointer) thread_id);

    CONTEXT_HASH_UNLOCK


    CONTEXT_OPERATION_EXIT
    if (quit_flag == 1)
    {
        g_main_loop_quit(loop);
        g_main_context_wakeup(g_main_loop_get_context(loop));
        return;
    }
}






static void receive_msg_func(AlpConnection* connection,
                             AlpMessage* message,
                             gpointer cbData)
{
    struct conn_msg *conn_msg = malloc(sizeof(struct conn_msg));
    gint pid = alp_ipc_connection_get_peer_pid(connection);
    int thread_id = (int)((long) (connection) % NUM_WORKER);

    conn_msg->conn = connection;
    conn_msg->msg = message;
    g_thread_pool_push(thread_pools[thread_id], conn_msg, (gpointer) thread_id);

}

static void disconnected_func(AlpConnection *connection, gpointer cbData)
{
    gs_svr_context *context;
    gint pid = alp_ipc_connection_get_peer_pid(connection);


    CONTEXT_HASH_LOCK
    while (context_operations_in_progress > 0)
        pthread_cond_wait(&no_context_operation_cond, &context_hash_lock);
    fprintf(stderr, "\nPid %d disconnected\n", pid);//???

    context = (gs_svr_context*) g_hash_table_lookup(context_table, (gconstpointer)
              alp_ipc_connection_get_custom_data(connection));
    if (context)
    {
        context->to_close = 1;
    }
    CONTEXT_HASH_UNLOCK

}


static void connected_func(AlpConnection *connection, gpointer cbData)
{

    alp_ipc_connection_register_receive_callback(connection, receive_msg_func, cbData);
    alp_ipc_connection_register_disconnect_callback(connection, disconnected_func, cbData);


}



static void transaction_check(gpointer key, gpointer value,  gpointer user_data)
{

    gs_svr_context *context = (gs_svr_context*) value;
    int thread_id = (int) user_data;
    if ((context->to_close == 0) && (context->thread_id == thread_id) && (context->transaction_level > 0))
        if (thread_now[thread_id] - context->last_op_time > TRANSACTION_CHECK_INTERVAL)
        {
            fprintf(stderr, "breaking a connection for client with pid %d\n", context->pid);//???
            sqlfs_break_transaction(context->sqlfs);

        }


}


static void thread_context_periodic_check(int thread_id)
{


    CONTEXT_HASH_LOCK
    g_hash_table_foreach(context_table, transaction_check, (gpointer) thread_id);
    CONTEXT_HASH_UNLOCK

}


static gboolean threads_ping(gpointer data)
{
    int i;
    for (i = 0; i < NUM_WORKER; i++)
        g_thread_pool_push(thread_pools[i], (gpointer) -1, (gpointer) i);
    return TRUE;
}

static void ensure_default_db_existence(const char *db_name);

int AlpGlobalSettingsServerMain(int argc, char *argv[])
{
    unsigned int i;

    AlpChannel *channel;
    alp_status_t status;
    const char *db_name = default_database_name;
    if (argc == 2)
        db_name = argv[1];
    else
        if (argc > 2)
        {
            fprintf(stderr, "Usage: %s [database_name]\n", argv[0]);
            exit(1);
        }
    if (!g_thread_supported())
    {
        g_thread_init(NULL);
        /*dbus_g_thread_init();*/
    }

    context_table = g_hash_table_new_full(g_direct_hash, g_direct_equal, NULL, context_do_close);
    ALP_FAIL_FATAL_IF((context_table == 0), "", "");

    notify_mgr_working = 1;

    loop = g_main_loop_new(NULL, TRUE);
    for (i = 0; i < sizeof(thread_pools) / sizeof(thread_pools[0]); i++)
    {
        thread_pools[i] = g_thread_pool_new(thread_func, (gpointer) i, 1, TRUE, 0);
        ALP_FAIL_FATAL_IF(thread_pools[i] == 0, "cannot create the thread pool", "");
    }

    {
        char tmp[PATH_MAX], *s, *s2;


        strncpy(tmp, db_name, sizeof(tmp));
        normalize_path(tmp);
        s2 = strdup(tmp);
        s = strrchr(tmp, '/');
        if (s)
        {
            *s = 0;
            mkdirp(tmp, S_IRUSR |  S_IWUSR |  S_IXUSR);
        }
        ensure_default_db_existence(s2);
        free(s2);

    }
    status = gs_svr_init(db_name);
    ALP_FAIL_FATAL_IF(!(status == ALP_STATUS_OK), "", "");

    channel = alp_ipc_channel_create(GLOBAL_SETTINGS_CHANNEL_NAME, ALP_IPC_CHANNEL_ACCESS_ALL);
    ALP_FAIL_FATAL_IF(channel == 0, "cannot create the IPC channel", "");

    alp_ipc_channel_register_connect_callback(channel, connected_func, (gpointer) NULL);
    alp_ipc_channel_register_disconnect_callback(channel, disconnected_func, (gpointer) NULL);

    g_timeout_add(5 * 1000, threads_ping, (gpointer) 0);

    alp_daemon_release(0);

    g_main_loop_run(loop);
    for (i = 0; i < sizeof(thread_pools) / sizeof(thread_pools[0]); i++)
    {
        g_thread_pool_free(thread_pools[i], TRUE, TRUE);
    }
    g_main_loop_unref(loop);
    alp_ipc_channel_destroy(channel);

    return 0;
}

alp_status_t alp_global_settings_server_wait_started(void)
{
    static const int wait_time = 10 ; /* seconds */
    AlpConnection *connection;
    AlpMessage *msg, *resp;
    time_t start, now;
    alp_status_t status = ~ALP_STATUS_OK;
    uint32_t i;

    time(&start);


    while (ALP_STATUS_OK != status)
    {
        sleep(1);
        time(&now);
        if (now > start + wait_time)
        {
            status = ALP_STATUS_GLOBAL_SETTINGS_ERROR;
            break;
        }
        connection = 0;
        msg = resp = 0;
        alp_ipc_channel_connect_synch(GLOBAL_SETTINGS_CHANNEL_NAME, ALP_IPC_CONNECTION_USAGE_HINT_NONE, GLOBAL_SETTINGS_TIMEOUT,  &(connection));
        if (connection == 0)
        {
            continue;
        }
        gs_start_pack(&msg, PROC_ping);

        alp_ipc_message_pack_end(msg);
        status = alp_ipc_connection_send_with_response(connection, msg, GLOBAL_SETTINGS_TIMEOUT, &resp);
        if (status == ALP_STATUS_OK)
        {
            alp_ipc_message_unpack_start(resp);
            alp_ipc_message_unpack_uint32(resp, &i);
            if (i != ALP_GLOBAL_SETTINGS_VERSION)
                status = ALP_STATUS_GLOBAL_SETTINGS_VERSION_MISMATCH;
            else
            {
                alp_ipc_message_unpack_uint32(resp, &i);
                status = i;
            }
            alp_ipc_message_unpack_end(resp);
        }
        alp_ipc_message_destroy(msg);
        if (resp)
            alp_ipc_message_destroy(resp);
        alp_ipc_channel_disconnect(connection);
    }
    return status;


}


alp_status_t alp_global_settings_server_running(void)
{
    static const int wait_time = 10 ; /* seconds */
    AlpConnection *connection;
    AlpMessage *msg, *resp;
    alp_status_t status;
    uint32_t i;

    connection = 0;
    msg = resp = 0;
    status = alp_ipc_channel_connect_synch(GLOBAL_SETTINGS_CHANNEL_NAME, ALP_IPC_CONNECTION_USAGE_HINT_NONE, GLOBAL_SETTINGS_TIMEOUT,  &(connection));
    if (status != ALP_STATUS_OK)
        return status;
    gs_start_pack(&msg, PROC_ping);

    alp_ipc_message_pack_end(msg);
    status = alp_ipc_connection_send_with_response(connection, msg, GLOBAL_SETTINGS_TIMEOUT, &resp);
    if (status == ALP_STATUS_OK)
    {
        alp_ipc_message_unpack_start(resp);
        alp_ipc_message_unpack_uint32(resp, &i);
        if (i != ALP_GLOBAL_SETTINGS_VERSION)
            status = ALP_STATUS_GLOBAL_SETTINGS_VERSION_MISMATCH;
        else
        {
            alp_ipc_message_unpack_uint32(resp, &i);
            status = i;
        }
        alp_ipc_message_unpack_end(resp);
    }
    alp_ipc_message_destroy(msg);
    if (resp)
        alp_ipc_message_destroy(resp);
    alp_ipc_channel_disconnect(connection);

    return status;


}


alp_status_t alp_global_settings_server_stop(void)
{
    AlpConnection *connection;
    AlpMessage *msg = 0, *resp = 0;
    alp_status_t status;
    uint32_t i;

    status = alp_ipc_channel_connect_synch(GLOBAL_SETTINGS_CHANNEL_NAME, ALP_IPC_CONNECTION_USAGE_HINT_NONE, GLOBAL_SETTINGS_TIMEOUT,  &(connection));
    if (status != ALP_STATUS_OK)
    {
        return status;/* server not found */
    }

    gs_start_pack(&msg, PROC_server_quit);

    alp_ipc_message_pack_end(msg);
    status = alp_ipc_connection_send_with_response(connection, msg, GLOBAL_SETTINGS_TIMEOUT, &resp);
    if (status == ALP_STATUS_OK)
    {
        alp_ipc_message_unpack_start(resp);
        alp_ipc_message_unpack_uint32(resp, &i);
        if (i != ALP_GLOBAL_SETTINGS_VERSION)
            status = ALP_STATUS_GLOBAL_SETTINGS_VERSION_MISMATCH;
        else
        {
            alp_ipc_message_unpack_uint32(resp, &i);
            status = i;
        }
        alp_ipc_message_unpack_end(resp);
    }
    alp_ipc_message_destroy(msg);
    if (resp)
        alp_ipc_message_destroy(resp);

    alp_ipc_channel_disconnect(connection);
    return status;

}

alp_status_t alp_global_settings_server_clean_data(void)
{
    AlpConnection *connection;
    AlpMessage *msg = 0, *resp = 0;
    alp_status_t status;
    uint32_t i;

    status = alp_ipc_channel_connect_synch(GLOBAL_SETTINGS_CHANNEL_NAME, ALP_IPC_CONNECTION_USAGE_HINT_NONE, GLOBAL_SETTINGS_TIMEOUT,  &(connection));
    if (status != ALP_STATUS_OK)
    {
        return status;/* server not found */
    }

    gs_start_pack(&msg, PROC_delete_tree);
    alp_ipc_message_pack_string(msg, "/");
    alp_ipc_message_pack_end(msg);
    status = alp_ipc_connection_send_with_response(connection, msg, GLOBAL_SETTINGS_TIMEOUT, &resp);
    if (status == ALP_STATUS_OK)
    {
        alp_ipc_message_unpack_start(resp);
        alp_ipc_message_unpack_uint32(resp, &i);
        if (i != ALP_GLOBAL_SETTINGS_VERSION)
            status = ALP_STATUS_GLOBAL_SETTINGS_VERSION_MISMATCH;
        else
        {
            alp_ipc_message_unpack_uint32(resp, &i);
            status = i;
        }
        alp_ipc_message_unpack_end(resp);
    }
    alp_ipc_message_destroy(msg);
    if (resp)
        alp_ipc_message_destroy(resp);

    alp_ipc_channel_disconnect(connection);
    return status;

}

static int cp(const char *src_file, const char *dst_file)
{
    static const int BLOCK_SIZE = 1024 * 1024;
    FILE *f, *f2;
    char *buf;
    size_t read_size;
    fprintf(stderr, "cp %s to %s\n", src_file, dst_file);//???
    f = fopen(src_file, "rb");
    if (!f)
        return 0;
    buf = calloc(BLOCK_SIZE, sizeof(char));
    ALP_FAIL_FATAL_IF(buf == 0, "Out of memory", "");

    f2 = fopen(dst_file, "wb");
    if (!f2)
    {
        fclose(f);
        free(buf);
        return 0;
    }
    while ((read_size = fread(buf, sizeof(char), BLOCK_SIZE, f)), (read_size > 0))
    {
        fwrite(buf, sizeof(char), read_size, f2);
    }
    fclose(f2);
    fclose(f);
    free(buf);
    return 1;


}

static void set_undefined_uid_gid_func(gpointer data, gpointer user_data)
{
    gs_svr_context *context = (gs_svr_context *) user_data;
    const char *key = (const char *) data;
    char tmp[200], tmp2[200];
    alp_status_t status;
    int result = sqlfs_proc_getxattr(context->sqlfs, key, uid_string, tmp, sizeof(tmp));
    int i;

    if (result == 0)
    {
        struct passwd passwd, *r;
        i = getpwnam_r(tmp, &passwd, tmp2, sizeof(tmp2), &r);
        if (r != 0)
        {

            status = gs_svr_proc_set_uid(context, key, passwd.pw_uid);
            ALP_FAIL_FATAL_IF(status != ALP_STATUS_OK, "Failed to set key uid?", "", "");
            if (status == ALP_STATUS_OK)
                sqlfs_proc_removexattr(context->sqlfs, key, uid_string);
        }
        else ALP_TM(T(ALP_CLASS_GLOBAL_SETTINGS, "failed to get the uid for user name %s", tmp));
    }
    result = sqlfs_proc_getxattr(context->sqlfs, key, gid_string, tmp, sizeof(tmp));

    if (result == 0)
    {
        struct group group, *r;
        i = getgrnam_r(tmp, &group, tmp2, sizeof(tmp2), &r);
        if (r != 0)
        {
            status = gs_svr_proc_set_gid(context, key, group.gr_gid);
            ALP_FAIL_FATAL_IF(status != ALP_STATUS_OK, "Failed to set key gid?", "", "");
            if (status == ALP_STATUS_OK)
                sqlfs_proc_removexattr(context->sqlfs, key, gid_string);
        }
        else ALP_TM(T(ALP_CLASS_GLOBAL_SETTINGS, "failed to get the gid for group %s", tmp));
    }


}


static void set_undefined_uids_gids(gs_svr_context *context)
{

    AlpGlobalSettingsEntryIterator *iterator = 0;

    alp_status_t status = gs_svr_proc_list_keys(context, "*", &iterator);
    ALP_FAIL_FATAL_IF(status != ALP_STATUS_OK, "failed to get key lists!", "", "");

    g_list_foreach(iterator->entries, set_undefined_uid_gid_func, (gpointer)context);


    alp_global_settings_entry_iterator_free(iterator);
}


static void ensure_default_db_existence(const char *db_name)
{

    struct stat st;

    int i;
    i = stat(db_name, &st);
    if ((i == -1) && (ENOENT == errno))
    {

        cp( DEFAULT_PREFS_DB, db_name);

        gs_svr_context *context = 0;
        gs_svr_open(db_name, &context);
        ALP_FAIL_FATAL_IF(context == 0, "", "");
        gs_svr_proc_begin_transaction(context);
        set_undefined_uids_gids(context);
        gs_svr_proc_complete_transaction(context);
        gs_svr_close(context);

    }

}


