/********************************************************************

Copyright  2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/*!
 * @file serialize.c
 *
 * @brief fucntions for data type serialization in
 *       the implementation of the Global Settings server.
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include "hiker/traces.h"
#include "hiker/fail.h"


#include "serialize.h"


serialize_buf * make_serialize_buf(void)
{
    serialize_buf *buf = calloc(1, sizeof(serialize_buf));
    ALP_FAIL_FATAL_IF(buf == 0, "Out of memory!", "");
    
    return buf;
}

void clean_serialize_buf(serialize_buf *buf)
{
    if (buf)
    {
        if (buf->data)
            free(buf->data);
        buf->data = 0;
        buf->current = 0;
        buf->length = 0;
    }
}


void free_serialize_buf(serialize_buf *buf)
{
    if (buf)
    {
        if (buf->data)
            free(buf->data);
            
        free(buf);
    }
    
}

static __inline__ int serialize_data(serialize_buf *buf, const void *data, int size)
{
    
    buf->data = realloc(buf->data,  buf->length + size);
    buf->current = buf->data + buf->length;
    memcpy(buf->current, data, size);
    buf->length += size;
    return 1;

}


static __inline__ int serialize_int32(serialize_buf *buf,  int32_t i)
{
    return serialize_data(buf, &i, sizeof(i));
}

int serialize_bool(serialize_buf *buf, gboolean b)
{
    int32_t i = (int32_t) b;
    return serialize_int32(buf, i);
}

    
int serialize_int(serialize_buf *buf, int i)
{
   
    return serialize_int32(buf, (int32_t) i);
}

int serialize_float(serialize_buf *buf, double f)
{
    return serialize_data(buf, &f, sizeof(f));
}

int serialize_string(serialize_buf *buf, const char *s)
{
    return serialize_int32(buf, (int32_t) strlen(s)) && serialize_data(buf, s, strlen(s));
}

int serialize_blob(serialize_buf *buf, const char *data, int length)
{
    return serialize_int32(buf, (int32_t) length) && serialize_data(buf, data, length);
}

int serialize_gs_value(serialize_buf *buf, const AlpGlobalSettingsValue *value)
{
    int32_t i, num;
    double fl;
    int r = 1;
    GList *element;
    serialize_int32(buf, (int32_t) value->type);
    switch(value->type)
    {
    
    case ALP_GLOBAL_SETTINGS_TYPE_INVALID:
        break;
    case ALP_GLOBAL_SETTINGS_TYPE_STRING:
        r = serialize_string(buf, value->string_value->str);
        break;
    case ALP_GLOBAL_SETTINGS_TYPE_INT:
        i = value->int_value;
        r = serialize_int32(buf, i);
        break;
    case ALP_GLOBAL_SETTINGS_TYPE_FLOAT:
        fl = value->float_value;
        r = serialize_float(buf, fl);
        break;
    case ALP_GLOBAL_SETTINGS_TYPE_BOOL:
        i = value->bool_value;
        r = serialize_int32(buf, i);
        break;
    case ALP_GLOBAL_SETTINGS_TYPE_DIR:
        break;
    case ALP_GLOBAL_SETTINGS_TYPE_BLOB:
        r = serialize_blob(buf, value->blob_value.data, value->blob_value.len);
        break;
    case ALP_GLOBAL_SETTINGS_TYPE_LIST:
        num = g_list_length(value->list_value);
        r = serialize_int32(buf, num);        
        for (element = value->list_value; element; element = element->next)
        {
            r = serialize_gs_value(buf, (AlpGlobalSettingsValue*) (element->data));
        }
        break;   
    default:
        ALP_TM(T(ALP_CLASS_GLOBAL_SETTINGS,  "Illegal Global Settings value type"));
        r = 0;
    }
    return r;
}


static __inline__ int deserialize_data(serialize_buf *buf, void *data, int size)
{
    
    if (buf->current == 0)
        buf->current = buf->data;
    memcpy(data, buf->current, size);
    buf->current += size;
    return 1;

}


static __inline__ int deserialize_int32(serialize_buf *buf,  int32_t *i)
{
    return deserialize_data(buf, i, sizeof(i));
}

int deserialize_bool(serialize_buf *buf, gboolean *b)
{
    int32_t i ;
    int r = deserialize_int32(buf, &i);
    if (r)
        *b = (gboolean) i;
    return r;
}

    
int deserialize_int(serialize_buf *buf, int *i)
{
    int32_t i2 ;
    int r = deserialize_int32(buf, &i2);
    if (r)
        *i = (int) i2;
    return r; 
}

int deserialize_float(serialize_buf *buf, double *f)
{
    return deserialize_data(buf, f, sizeof(*f));
}

int deserialize_string(serialize_buf *buf, char **s, int *length)
{
    int32_t l;
    int r;
    r = deserialize_int32(buf, &l);
    if (r && l)
    {
        *s = calloc(l + 1, sizeof(char));
        ALP_FAIL_FATAL_IF(*s == 0, "Out of memory!", "");
        r = deserialize_data(buf, *s, l);
    }
         
    return r;
}

int deserialize_blob(serialize_buf *buf, char  **data, int *length)
{
    int32_t l;
    int r;
    r = deserialize_int32(buf, &l);
    if (r && l)
    {
        *data = malloc(sizeof(char) * l);
        ALP_FAIL_FATAL_IF(*data == 0, "Out of memory!", "");
        r = deserialize_data(buf, *data, l);
    }
         
    return r;
}

int deserialize_gs_value(serialize_buf *buf, AlpGlobalSettingsValue *value)
{
    int32_t t;
    int r;
    int len;
    int32_t i, num;
    char *s, *data;
    AlpGlobalSettingsValue *value2;
    r = deserialize_int32(buf, &t);
    value->type = (AlpGlobalSettingsValueType) t;
    switch (value->type)
    {
    case ALP_GLOBAL_SETTINGS_TYPE_INVALID:
        break;
    case ALP_GLOBAL_SETTINGS_TYPE_STRING:
        r = deserialize_string(buf, &s, &len);
        if (r)
            value->string_value = g_string_new(s);
        free(s); 
        break;
    case ALP_GLOBAL_SETTINGS_TYPE_INT:
        r = deserialize_int(buf, &(value->int_value));
        break;
    case ALP_GLOBAL_SETTINGS_TYPE_FLOAT:
        r = deserialize_float(buf, &(value->float_value));
        break;
    case ALP_GLOBAL_SETTINGS_TYPE_BOOL:
        r = deserialize_int32(buf, &i);
        if (r)
            value->bool_value = i;
        break;
        
    case ALP_GLOBAL_SETTINGS_TYPE_DIR:
        break;
        
    case ALP_GLOBAL_SETTINGS_TYPE_BLOB:
        r = deserialize_blob(buf, &(data), &len);
        
        if (r)
        {
            value->blob_value.len = len;
            value->blob_value.data = data;
        }
        break;
        
    case ALP_GLOBAL_SETTINGS_TYPE_LIST:
        r = deserialize_int32(buf, &num);
        if (r)
        {
            for (i = 0; i < num; i++)
            {
                value2 = alp_global_settings_value_alloc();
                ALP_FAIL_FATAL_IF(value2 == 0, "out of memory", "");  
                deserialize_gs_value(buf, value2);
                value->list_value = g_list_prepend(value->list_value,  (gpointer) value2);
            }
            value->list_value = g_list_reverse(value->list_value);
        }
        break;
        
    default:
        ALP_TM(T(ALP_CLASS_GLOBAL_SETTINGS,  "Illegal Global Settings value type"));
        r = 0;
    }
    return r;
}

