/********************************************************************

Copyright  2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/*!
 * @file alp_appprefs.c
 *
 * @brief the implementation of the Global Settings service APIs.
 *
 *****************************************************************************/
#include <pthread.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <glib.h>
#include "hiker/syspaths.h"
#include "hiker/traces.h"
#include "hiker/fail.h"
#include "hiker/global_settings.h"

#include "gs_imp.h"

#include "sqlfs.h"



static const char *database_name = ALP_SYSPATH_PREFS;  /* default ALP location */

void free_glist(GList *list)
{
    GList *iterator = list;
    for (;iterator; iterator = iterator->next)
    {
        alp_global_settings_value_free((AlpGlobalSettingsValue *)(iterator->data));
    }
    g_list_free(list);
}

void gstring_copy(GString **gs, const gchar *src)
{

    *gs = g_string_new((src));

}

void global_settings_value_clean(AlpGlobalSettingsValue *value)
{
    switch(value->type)
    {
    case ALP_GLOBAL_SETTINGS_TYPE_BLOB:
        free(value->blob_value.data);
        break;

    case ALP_GLOBAL_SETTINGS_TYPE_STRING:
        g_string_free((value->string_value), TRUE);
        break;

    case ALP_GLOBAL_SETTINGS_TYPE_LIST:
        free_glist((value->list_value));
        break;
    default:
        break;
    }
    memset(value, 0, sizeof(*value));
}

void global_settings_entry_clean(AlpGlobalSettingsEntry *entry)
{
    if (entry->key)
        g_string_free((entry->key), TRUE);
    (entry->key) = 0;

    global_settings_value_clean(&(entry->value));
}


void global_settings_value_copy(const AlpGlobalSettingsValue *src, AlpGlobalSettingsValue *des)
{
    GList *tmp = 0;
    CHECK_ARG_VOID_RETURN(src)
    CHECK_ARG_VOID_RETURN(des)
    global_settings_value_clean(des);
    des->type = src->type;
    switch(des->type)
    {
    case ALP_GLOBAL_SETTINGS_TYPE_STRING:
        gstring_copy(&(des->string_value), src->string_value->str);
        break;

    case ALP_GLOBAL_SETTINGS_TYPE_LIST:
        tmp = g_list_copy((src->list_value));
        (des->list_value) = tmp;
        break;

    case  ALP_GLOBAL_SETTINGS_TYPE_BLOB:
        des->blob_value.len = src->blob_value.len;
        des->blob_value.data = malloc(des->blob_value.len);
        ALP_FAIL_FATAL_IF(des->blob_value.data == 0, "out of memory", "");
        memcpy(des->blob_value.data, src->blob_value.data, des->blob_value.len);
        break;

    default:
        *des = *src;
        break;
    }
}



void global_settings_list_copy(const GList *src, GList **dst);

AlpGlobalSettingsValue * global_settings_value_duplicate(const AlpGlobalSettingsValue *src)
{
    AlpGlobalSettingsValue *tmp = 0;
    if (src == 0)
    {
        ALP_TM(T(ALP_CLASS_GLOBAL_SETTINGS, "zero value for src??"));
        return 0;
    }

    tmp = calloc(1, sizeof(*tmp));
    ALP_FAIL_FATAL_IF(tmp == 0, "out of memory", "");
    tmp->type = src->type;
    switch(tmp->type)
    {
    case ALP_GLOBAL_SETTINGS_TYPE_STRING:
        tmp->string_value = g_string_new(src->string_value->str);
        break;

    case ALP_GLOBAL_SETTINGS_TYPE_LIST:
        tmp->list_value = NULL;
        global_settings_list_copy(src->list_value, &(tmp->list_value));
        break;

    case  ALP_GLOBAL_SETTINGS_TYPE_BLOB:
        tmp->blob_value.len = src->blob_value.len;
        tmp->blob_value.data = malloc(src->blob_value.len);
        ALP_FAIL_FATAL_IF(tmp->blob_value.data == 0, "out of memory", "");
        memcpy(tmp->blob_value.data, src->blob_value.data, src->blob_value.len);
        break;

    default:
        *tmp = *src;
        break;
    }
    return tmp;
}


void global_settings_list_copy(const GList *src, GList **dst)
{
    AlpGlobalSettingsValue *tmp = 0;
    const GList *current;

    ALP_FAIL_FATAL_IF(g_list_length(*dst) != 0, "destination list not empty", "");

    for (current = src; current != NULL; current = current->next)
    {
        tmp = global_settings_value_duplicate((const AlpGlobalSettingsValue *)(current->data));
        *dst = g_list_prepend(*dst, (gpointer) tmp);
    }
    *dst = g_list_reverse(*dst);
}


AlpGlobalSettingsEntry *alp_global_settings_entry_alloc(void)
{
    AlpGlobalSettingsEntry *entry = calloc(1, sizeof(*entry));
    return entry;

}

void alp_global_settings_entry_free(AlpGlobalSettingsEntry *e)
{
    global_settings_entry_clean(e);
    free(e);
}


AlpGlobalSettingsValue *alp_global_settings_value_alloc(void)
{
    AlpGlobalSettingsValue *value = calloc(1, sizeof(*value));
    return value;

}

void alp_global_settings_value_clean(AlpGlobalSettingsValue *v)
{
    global_settings_value_clean(v);

}


void alp_global_settings_value_free(AlpGlobalSettingsValue *v)
{
    global_settings_value_clean(v);
    free(v);
}

void alp_global_settings_value_list_free(GList *l)
{
    free_glist(l);
}


void alp_global_settings_entry_iterator_free(AlpGlobalSettingsEntryIterator *iterator)
{
    GList *tmp;
    for (tmp = iterator->entries; tmp != NULL; tmp = tmp->next)
        free(tmp->data);
    g_list_free(iterator->entries);
    free(iterator->base_path);
    free(iterator);

}


alp_status_t
alp_global_settings_entry_get_key_value(const AlpGlobalSettingsEntry  *entry,
                                gchar *key,
                                guint key_size,
                                AlpGlobalSettingsValue *value)
{
    CHECK_ARG(entry);
    if (key)
    {
        if (key_size < entry->key->len)
        {
            ALP_TM(T(ALP_CLASS_GLOBAL_SETTINGS, "buffer not big enough"));
            return ALP_STATUS_GLOBAL_SETTINGS_BUFFER_TOO_SMALL;
        }            
        strncpy(key, entry->key->str, key_size);
    }
    if (value)
    {
        global_settings_value_clean(value);
        global_settings_value_copy(&(entry->value), value);    
    }
    
    return ALP_STATUS_OK;
}

alp_status_t
alp_global_settings_entry_set_key_value(AlpGlobalSettingsEntry  *entry,
                                const gchar *key,
                                const AlpGlobalSettingsValue *value)
{
    CHECK_ARG(entry);
    CHECK_ARG(key)
    global_settings_entry_clean(entry);
  
    gstring_copy(&(entry->key), key);
    if (value)
    {
        global_settings_value_copy(value, &(entry->value));    
    }
    
    return ALP_STATUS_OK;
}




alp_status_t
alp_global_settings_value_set_string(AlpGlobalSettingsValue *value,
                             const gchar *string)
{
    CHECK_ARG(value);
    CHECK_ARG(string)
    global_settings_value_clean(value);
    value->type = ALP_GLOBAL_SETTINGS_TYPE_STRING;
   
    gstring_copy(&(value->string_value), string);
    return ALP_STATUS_OK;
}

alp_status_t
alp_global_settings_value_set_int(AlpGlobalSettingsValue *value,
                          gint i)
{
    CHECK_ARG(value);
   
    global_settings_value_clean(value);
    value->type = ALP_GLOBAL_SETTINGS_TYPE_INT;
    value->int_value = i;
    return ALP_STATUS_OK;
}

alp_status_t
alp_global_settings_value_set_float(AlpGlobalSettingsValue *value,
                            gdouble f)
{
    CHECK_ARG(value);
   
    global_settings_value_clean(value);
    value->type = ALP_GLOBAL_SETTINGS_TYPE_FLOAT;
    value->float_value = f;
    return ALP_STATUS_OK;
}


alp_status_t
alp_global_settings_value_set_bool(AlpGlobalSettingsValue *value,
                           gboolean b)
{
    CHECK_ARG(value);
   
    global_settings_value_clean(value);
    value->type = ALP_GLOBAL_SETTINGS_TYPE_BOOL;
    value->bool_value = b;
    return ALP_STATUS_OK;
}


alp_status_t
alp_global_settings_value_set_list(AlpGlobalSettingsValue *value,
                           const GList *list)
{
    CHECK_ARG(value);
   
    global_settings_value_clean(value);
    value->type = ALP_GLOBAL_SETTINGS_TYPE_LIST;
    
    value->list_value = NULL;
    
    global_settings_list_copy(list, &(value->list_value));
    return ALP_STATUS_OK;
}


alp_status_t
alp_global_settings_value_get_string(AlpGlobalSettingsValue *value,
                             gchar *string,
                             guint string_size)
{
    CHECK_ARG(value);
   
    CHECK_TYPE(value, ALP_GLOBAL_SETTINGS_TYPE_STRING);
    if (string_size < value->string_value->len)
    {
        ALP_TM(T(ALP_CLASS_GLOBAL_SETTINGS, "buffer not big enough"));
        return ALP_STATUS_GLOBAL_SETTINGS_BUFFER_TOO_SMALL;
    }            
    strncpy(string, value->string_value->str, string_size);
    
    return ALP_STATUS_OK;
}


alp_status_t
alp_global_settings_value_get_int(AlpGlobalSettingsValue *value,
                          gint *i)
{
    CHECK_ARG(value);
   
    CHECK_TYPE(value, ALP_GLOBAL_SETTINGS_TYPE_INT);
    *i = value->int_value;
    return ALP_STATUS_OK;
}

alp_status_t
alp_global_settings_value_get_float(AlpGlobalSettingsValue *value,
                            gdouble *f)
{
    CHECK_ARG(value);
   
    CHECK_TYPE(value, ALP_GLOBAL_SETTINGS_TYPE_FLOAT);
    *f = value->float_value;
    return ALP_STATUS_OK;
}


alp_status_t
alp_global_settings_value_get_bool(AlpGlobalSettingsValue *value,
                           gboolean *b)
{
    CHECK_ARG(value);
   
    CHECK_TYPE(value, ALP_GLOBAL_SETTINGS_TYPE_BOOL);
    *b = value->bool_value;
    return ALP_STATUS_OK;
}

alp_status_t
alp_global_settings_value_get_list(AlpGlobalSettingsValue *value,
                           GList **list)
{
    CHECK_ARG(value);
   
    CHECK_TYPE(value, ALP_GLOBAL_SETTINGS_TYPE_LIST);
    if (*list)
        free_glist(*list);
    *list = NULL;
   
    global_settings_list_copy(value->list_value, list);
    
 
    return ALP_STATUS_OK;
}

const char * gs_type_to_string(AlpGlobalSettingsValueType t)
{
    switch(t)
    {
    case ALP_GLOBAL_SETTINGS_TYPE_INVALID:
        return TYPE_NULL;
    case ALP_GLOBAL_SETTINGS_TYPE_STRING:
        return TYPE_STRING;
    case ALP_GLOBAL_SETTINGS_TYPE_INT:
        return TYPE_INT;
    case ALP_GLOBAL_SETTINGS_TYPE_FLOAT:
        return TYPE_DOUBLE;
    case ALP_GLOBAL_SETTINGS_TYPE_BOOL:
        return TYPE_BOOL;
    case ALP_GLOBAL_SETTINGS_TYPE_DIR:
        return TYPE_DIR;
    case ALP_GLOBAL_SETTINGS_TYPE_BLOB:
        return TYPE_BLOB;
    case ALP_GLOBAL_SETTINGS_TYPE_LIST:
        return TYPE_LIST;
    default:
        ;
    };
    return 0;
}

const char *gs_bool_to_string(gboolean b)
{
    switch(b)
    {
    case FALSE:
        return "FALSE";
    default:
        return "TRUE";
    }
}



int gs_unpack_value(AlpMessage* msg, AlpGlobalSettingsValue *value)
{
    uint32_t type;
    int32_t num;
    int i;
    double fl;
    char *str;
    GList *element;
    guchar *data;
    guint len;
    AlpGlobalSettingsValue *value2 = 0;
    global_settings_value_clean(value);
    alp_ipc_message_unpack_uint32(msg, &type);
    value->type = type;
    switch (value->type)
    {
    case ALP_GLOBAL_SETTINGS_TYPE_INVALID:
        break;
    case ALP_GLOBAL_SETTINGS_TYPE_STRING:
        alp_ipc_message_unpack_string(msg, &str);
        value->string_value = g_string_new(str);
        g_free(str);
        break;
    case ALP_GLOBAL_SETTINGS_TYPE_INT:
        alp_ipc_message_unpack_int32(msg, &i);
        value->int_value = i;
        break;
    case ALP_GLOBAL_SETTINGS_TYPE_FLOAT:
        alp_ipc_message_unpack_byte_array(msg, &data, &len);
        ALP_FAIL_FATAL_IF(len != sizeof(fl), "float handling error", "");
        memcpy(&fl, data, len);
        value->float_value = fl;
        g_free(data);
        break;
    case ALP_GLOBAL_SETTINGS_TYPE_BOOL:
        alp_ipc_message_unpack_int32(msg, &i);
        value->bool_value = i;
        break;
    case ALP_GLOBAL_SETTINGS_TYPE_DIR:
        break;
    case ALP_GLOBAL_SETTINGS_TYPE_BLOB:
        alp_ipc_message_unpack_byte_array(msg, &data, &len);
        value->blob_value.data = calloc(len, sizeof(char));
        ALP_FAIL_FATAL_IF(value->blob_value.data == 0, "out of memory", "");
        memcpy(value->blob_value.data, data, len);
        value->blob_value.len = len;
        g_free(data);
        break;

    case ALP_GLOBAL_SETTINGS_TYPE_LIST:
        alp_ipc_message_unpack_int32(msg, &num);
        for (i = 0; i < num; i++)
        {
            value2 = alp_global_settings_value_alloc();
            ALP_FAIL_FATAL_IF(value2 == 0, "out of memory", "");
            gs_unpack_value(msg, value2);
            value->list_value = g_list_prepend(value->list_value,  (gpointer) value2);
        }
        value->list_value = g_list_reverse(value->list_value);
        break;

    default:
        ;
    }
    return 1;
}

int gs_pack_value(AlpMessage* msg, const AlpGlobalSettingsValue *value)
{
    uint32_t type;
    int32_t num;
    int i;
    double fl;
    GList *element = 0;
    type = value->type;
    alp_ipc_message_pack_uint32(msg, type);
    switch(value->type)
    {

    case ALP_GLOBAL_SETTINGS_TYPE_INVALID:
        break;
    case ALP_GLOBAL_SETTINGS_TYPE_STRING:
        alp_ipc_message_pack_string(msg, value->string_value->str);
        break;
    case ALP_GLOBAL_SETTINGS_TYPE_INT:
        i = value->int_value;
        alp_ipc_message_pack_int32(msg, i);
        break;
    case ALP_GLOBAL_SETTINGS_TYPE_FLOAT:
        fl = value->float_value;
        alp_ipc_message_pack_byte_array(msg, (void*) &fl, sizeof(fl));
        break;
    case ALP_GLOBAL_SETTINGS_TYPE_BOOL:
        i = value->bool_value;
        alp_ipc_message_pack_int32(msg, i);
        break;
    case ALP_GLOBAL_SETTINGS_TYPE_DIR:
        break;
    case ALP_GLOBAL_SETTINGS_TYPE_BLOB:
        alp_ipc_message_pack_byte_array(msg, value->blob_value.data, value->blob_value.len);
        break;
    case ALP_GLOBAL_SETTINGS_TYPE_LIST:
        num = g_list_length(value->list_value);
        alp_ipc_message_pack_int32(msg, num);

        for (element = value->list_value; element; element = element->next)
        {
            gs_pack_value(msg, (AlpGlobalSettingsValue*) (element->data));

        }
        break;
    default:
        ;

    }
    return 1;
}

void gs_start_pack(AlpMessage** msg, ALP_GLOBAL_SETTINGS_RPC_PROCS proc_id)
{
    static int msg_id = 0;
    static pthread_mutex_t msg_id_lock = PTHREAD_MUTEX_INITIALIZER;
    (*msg) = alp_ipc_message_create();
    ALP_FAIL_FATAL_IF((*msg) == 0, "out of memory", "");
    pthread_mutex_lock(&msg_id_lock );
    alp_ipc_message_pack_start((*msg), msg_id);
    msg_id++;
    pthread_mutex_unlock(&msg_id_lock );
    alp_ipc_message_pack_uint32((*msg), ALP_GLOBAL_SETTINGS_VERSION);
    alp_ipc_message_pack_uint32((*msg), (uint32_t) proc_id);
}


int gs_svr_pack_entry_iterator(AlpMessage* msg, const AlpGlobalSettingsEntryIterator * iterator)
{
    int num = g_list_length(iterator->entries);
    GList *it = iterator->entries;
    alp_ipc_message_pack_uint32(msg, num);
    alp_ipc_message_pack_string(msg, iterator->base_path);
    for (; it; it = it->next)
    {

        alp_ipc_message_pack_string(msg, it->data);
    }

    return 1;
}

int gs_svr_unpack_entry_iterator(AlpMessage* msg, AlpGlobalSettingsEntryIterator * iterator)
{
    /* assume iterator now is empty */
    uint32_t i, num;
    char *str = 0;
    alp_ipc_message_unpack_uint32(msg, &num);
    alp_ipc_message_unpack_string(msg, &str);

    iterator->base_path = strdup(str);
    g_free(str);
    for (i = 0; i < num; i++)
    {
        alp_ipc_message_unpack_string(msg, &str);


        iterator->entries = g_list_prepend(iterator->entries, strdup(str));
        g_free(str);
    }
    iterator->current = iterator->entries = g_list_reverse(iterator->entries);
    return 1;
}


int is_dir(const char *path)
{
    struct stat st;
    int i;
    i = stat(path, &st);
    if (i == 0)
    {
        if (S_ISDIR(st.st_mode))
            return 1;
        return -1;
    }
    return 0;
}

void normalize_path(char *path)
{
    char *t = path;
    char *t2;
    
    while (1)
    {
        t = strchr(t, '/');
        if (t == 0)
            break;
        t++;    
        t2 = t;
        if (*t2 == '/')
        {
            for (t2++ ; *t2 == '/'; t2++)
                ;
            memmove(t, t2, strlen(t2) + 1);            
        }
    } ;

}



int form_absolute_path(const char *cwd, const char *key, char result[PATH_MAX])
{
    if ((cwd == 0) || (strlen(cwd) == 0))
        cwd = "/";
    if ((key[0] == '/')  /* absolute path */ ||
        (!strcmp(cwd, "/") && (key[0] == '*')))
        strncpy(result, key, sizeof(result[0]) * PATH_MAX);
    else
        snprintf(result, sizeof(result[0]) * PATH_MAX, "%s/%s", cwd, key);
     
    normalize_path(result);
    return 1;
}

const char *status_to_string(int status)
{
    switch (status)
    {
    case ALP_STATUS_OK:
        return "no error";
    case  ALP_STATUS_GLOBAL_SETTINGS_ERROR:
        return "Global Settings error";
    case  ALP_STATUS_GLOBAL_SETTINGS_KEY_NOT_FOUND :
        return "key not found";
    case  ALP_STATUS_GLOBAL_SETTINGS_VALUE_NOT_FOUND:
        return "key value not found";
    case  ALP_STATUS_GLOBAL_SETTINGS_ENTRY_ITERATOR_AT_END:
        return "directory enty interator at end";
    case  ALP_STATUS_GLOBAL_SETTINGS_ENTRY_ITERATOR_INVALID:
        return "directory entry iterator not valid";
    case  ALP_STATUS_GLOBAL_SETTINGS_VALUE_NOT_DIR:
        return "key value is not directory";
    case  ALP_STATUS_GLOBAL_SETTINGS_INCORRECT_TYPE:
        return "key value is not of the correct type";
    case  ALP_STATUS_GLOBAL_SETTINGS_ENTRY_INVALID:
        return "key value entry is not valid";
    case  ALP_STATUS_GLOBAL_SETTINGS_VALUE_INVALID:
        return "key value is not valid";
    case  ALP_STATUS_GLOBAL_SETTINGS_ARGUMENT_INVALID:
        return "argument is not valid";
    case  ALP_STATUS_GLOBAL_SETTINGS_BUFFER_TOO_SMALL:
        return "buffer provided is too small";
    case  ALP_STATUS_GLOBAL_SETTINGS_PERMISSION_DENIED:
        return "operation permission is denied";
    case  ALP_STATUS_GLOBAL_SETTINGS_KEY_EXISTS :
        return "key already exists";
    case  ALP_STATUS_GLOBAL_SETTINGS_VERSION_MISMATCH:
        return "Global Settings service version mismatches";
    default:
        return "unknown Global Settings status code";
    }
    return "";
}
