/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/*!
 * @file alp_ipc.c
 *
 * @brief Contains implementation of public API functions for the IPC library
 *
 *****************************************************************************/

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>

#include <fcntl.h>
#include <errno.h>
#include <unistd.h>

#include <string.h>

#include <stdlib.h>
#include <stdarg.h>

#include <gnet-2.0/pack.h>

#include <hiker/ipc.h>
#include <hiker/traces.h>

#include "alp_ipc_internal.h"

AlpChannel* 	
alp_ipc_channel_create(const gchar* channelName, 
				   AlpIpcChannelAccessMode accessMode)
{
	int ret_val;
	GError* gErr = NULL;

	if ((channelName == NULL) || (strlen(channelName) == 0))
		return NULL;

	AlpChannel* newChannel = prv_channel_create();

	if (newChannel == NULL)
		return NULL;

	prv_server_socket_name(channelName, newChannel->socketName); 

	// Create the first named - pipe 
	prv_make_dirs(newChannel->socketName);

	// Clean up any old files
	prv_clean_up_old_socket(channelName, newChannel->socketName);

	newChannel->socketFD = socket(PF_UNIX, SOCK_STREAM, 0);

	if (newChannel->socketFD < 0){
		TRACE_IPC("Error creating server socket %s\n",
				channelName);

		prv_channel_destroy(newChannel);

		return NULL;
	}

	ret_val = fcntl(newChannel->socketFD, F_SETFL, O_NONBLOCK);

	if (ret_val == -1) {
		TRACE_IPC("Fcntl error for server socket %s\n", channelName);
		
		prv_channel_destroy(newChannel);

		return NULL;
	}

	struct sockaddr_un addr;
	addr.sun_family = AF_UNIX;

	if (newChannel->socketName->len >= sizeof(addr.sun_path)) {
		TRACE_IPC("Socket name %s is too long (max is %d)\n",
				newChannel->socketName->str, sizeof(addr.sun_path));

		prv_channel_destroy(newChannel);

		return NULL;
	}
	strcpy(addr.sun_path, newChannel->socketName->str);

	ret_val = bind(newChannel->socketFD, 
				(const struct sockaddr *)&addr, 
				sizeof(addr));

	if (ret_val == -1) {
		TRACE_IPC("Error binding server socket %s\n", channelName);
		
		prv_channel_destroy(newChannel);

		return NULL;
	}

	mode_t permissions = 0;
	switch (accessMode) {
	case ALP_IPC_CHANNEL_ACCESS_ALL:
		permissions = S_IRWXU | S_IRWXG | S_IRWXO;
		break;
	case ALP_IPC_CHANNEL_ACCESS_OWNER:
	default:
		permissions = S_IRWXU;
		break;
	}

	mode_t oldUmask = umask(0);
	chmod(newChannel->socketName->str, permissions);
	umask(oldUmask);  // restore original umask

	// call "listen" to watch for new connections
	ret_val = listen(newChannel->socketFD, 4);  // 4 is max number of
	                                            // processes that can
                                                 // be queued to connect
                                                 // 4 seems like a reasonable
                                                 // number

	if (ret_val == -1) {
		TRACE_IPC("Error listening on server socket %s\n", channelName);
		
		prv_channel_destroy(newChannel);

		return NULL;
	}

	// create a GIO Channel to listen to the socket
	newChannel->gioChannel = g_io_channel_unix_new(newChannel->socketFD);

	if (newChannel->gioChannel == NULL) {
		TRACE_IPC("Error creating the server GIOChannel\n", NULL);
	
		prv_channel_destroy(newChannel);
	
		return NULL;
	}

	g_io_channel_set_encoding(newChannel->gioChannel, NULL, &gErr);

	// add a input and hangup watch on the pipe from the client
	newChannel->inSource = g_io_add_watch(newChannel->gioChannel, 
								   G_IO_IN,
								   prv_server_accept_CB, 
								   (gpointer)newChannel);


	return newChannel;
}

alp_status_t
alp_ipc_channel_destroy(AlpChannel* channel)
{ 
	prv_channel_destroy(channel);

	return ALP_STATUS_OK;
}	

alp_status_t
alp_ipc_channel_register_connect_callback(AlpChannel* channel,
								  AlpChannelConnectCB callback,
								  gpointer cbData)
{
	ConnectCBInfo* cbInfo = NULL;

	if (channel == NULL) {
		TRACE_IPC("Can't register channel connect callback\n", NULL); 
		return(ALP_STATUS_IPC_ERROR);  
	}

	// free any existing callback
	if (channel->connectCB != NULL)
		free((void *)channel->connectCB);

	if (callback != NULL) {
		cbInfo = (ConnectCBInfo *)malloc(sizeof(ConnectCBInfo));

		if (cbInfo == NULL) 
			return ALP_STATUS_IPC_ERROR;
    
		cbInfo->callback = callback;
		cbInfo->cbData = cbData;
	}

	channel->connectCB = cbInfo;

	return ALP_STATUS_OK;
}

alp_status_t
alp_ipc_channel_register_disconnect_callback(AlpChannel* channel,
									AlpChannelDisconnectCB callback,
									gpointer cbData)
{
	DisconnectCBInfo* cbInfo = NULL;

	if (channel == NULL) {
		TRACE_IPC("Can't register channel disconnect callback\n", NULL);
		return(ALP_STATUS_IPC_ERROR);
	}

	// free any existing callback
	if (channel->disconnectCB != NULL)
		free((void *)channel->disconnectCB);

	if (callback != NULL) {
		cbInfo = (DisconnectCBInfo *)malloc(sizeof(DisconnectCBInfo));

		if (cbInfo == NULL) 
			return ALP_STATUS_IPC_ERROR;
    
		cbInfo->callback = callback;
		cbInfo->cbData = cbData;
	}

	channel->disconnectCB = cbInfo;

	return ALP_STATUS_OK;
}

AlpConnection*
alp_ipc_channel_connect(const gchar* channelName, 
				    AlpConnectionUsageHint usageHint,
				    AlpChannelConnectCB callback, 
				    gpointer cbData)
{
	int ret_val;
	AlpConnection* newConnection = NULL;
	ConnectCBInfo* cbInfo = NULL;
	GError* gErr = NULL;

	if ((channelName == NULL) || (strlen(channelName) == 0))
		return NULL;

	newConnection = prv_connection_create(usageHint);

	if (newConnection == NULL)
		return NULL;

	if (callback != NULL) {
		cbInfo = (ConnectCBInfo *)malloc(sizeof(ConnectCBInfo));

		if (cbInfo == NULL) {
			prv_connection_destroy(newConnection);
			return NULL;
		}
    
		cbInfo->callback = callback;
		cbInfo->cbData = cbData;
	}

	newConnection->connectCB = cbInfo;

	newConnection->socketFD = socket(PF_UNIX, SOCK_STREAM, 0);

	if (newConnection->socketFD < 0){
		TRACE_IPC("Error creating client socket %s\n",
				channelName);

		prv_connection_destroy(newConnection);

		return NULL;
	}

	ret_val = fcntl(newConnection->socketFD, F_SETFL, O_NONBLOCK);
	if (ret_val != -1) ret_val = fcntl(newConnection->socketFD, F_SETFD, FD_CLOEXEC);

	if (ret_val == -1) {
		TRACE_IPC("Fcntl error for client socket %s\n", channelName);
		
		prv_connection_destroy(newConnection);

		return NULL;
	}

	newConnection->status = ALP_IPC_CONNECTION_CREATED;
	newConnection->amServer = FALSE;

	// set up the GIOChannel for the socket
	newConnection->gioChannel = 
		g_io_channel_unix_new(newConnection->socketFD);

	if (newConnection->gioChannel == NULL) {
		TRACE_IPC("Error creating the GIOChannel for the client\n",
				NULL); 

		return FALSE;
	}

	g_io_channel_set_encoding(newConnection->gioChannel, NULL, &gErr);

	ret_val = prv_try_to_connect(newConnection, channelName, FALSE, -1, NULL);

	if (ret_val == -1) {
		TRACE_IPC("Client error connecting to server socket %s\n",
				channelName);
		prv_connection_destroy(newConnection);
		return NULL;
	}

	return newConnection;
}

alp_status_t
alp_ipc_channel_connect_synch(const gchar* channelName, 
						AlpConnectionUsageHint usageHint,
						guint timeout,
						AlpConnection**  connection)
{
	int ret_val;
	AlpConnection* newConnection = NULL;
	ConnectCBInfo* cbInfo = NULL;
	GError* gErr = NULL;

	alp_status_t status = ALP_STATUS_OK;

	if ((channelName == NULL) || (strlen(channelName) == 0))
		return ALP_STATUS_IPC_ERROR;

	newConnection = prv_connection_create(usageHint);

	if (newConnection == NULL)
		return ALP_STATUS_IPC_ERROR;

	newConnection->socketFD = socket(PF_UNIX, SOCK_STREAM, 0);
	if (newConnection->socketFD < 0){
		TRACE_IPC("Error creating client socket %s\n",
				channelName);

		prv_connection_destroy(newConnection);

		return ALP_STATUS_IPC_ERROR;
	}

	ret_val = fcntl(newConnection->socketFD, F_SETFL, O_NONBLOCK);
	if (ret_val != -1) ret_val = fcntl(newConnection->socketFD, F_SETFD, FD_CLOEXEC);

	if (ret_val == -1) {
		TRACE_IPC("Fcntl error for client socket %s\n", channelName);
		
		prv_connection_destroy(newConnection);

		return ALP_STATUS_IPC_ERROR;
	}

	newConnection->status = ALP_IPC_CONNECTION_CREATED;
	newConnection->amServer = FALSE;

	// set up the GIOChannel for the socket
	newConnection->gioChannel = 
		g_io_channel_unix_new(newConnection->socketFD);

	//****	newConnection->gioChannel = 
	//		g_io_channel_unix_new(newConnection->socketFD);

	if (newConnection->gioChannel == NULL) {
		TRACE_IPC("Error creating the GIOChannel for the client\n",
				NULL); 

		return ALP_STATUS_IPC_ERROR;
	}

	g_io_channel_set_encoding(newConnection->gioChannel, NULL, &gErr);

	ret_val = prv_try_to_connect(newConnection, channelName, FALSE, 
						    (gint)timeout,
						    &status);

	if (ret_val == -1) {
		TRACE_IPC("Client error connecting to server socket %s\n",
				channelName);
		prv_connection_destroy(newConnection);
		return ALP_STATUS_IPC_ERROR;
	}

	*connection = newConnection;

	return status;
}

alp_status_t
alp_ipc_channel_disconnect(AlpConnection* connection) 
{
	prv_connection_destroy(connection);

	return ALP_STATUS_OK;
}

void
alp_ipc_channel_set_custom_data(AlpChannel* channel,
						  gpointer customData)
{
	if (channel == NULL) {
		return;
	}

	channel->customData = customData;
}


gpointer
alp_ipc_channel_get_custom_data(AlpChannel* channel)
{
	if (channel == NULL)
		return NULL;

	return channel->customData;
}

alp_status_t
alp_ipc_connection_register_disconnect_callback(AlpConnection* connection, 
									   AlpChannelDisconnectCB callback, 
									   gpointer cbData)
{
	DisconnectCBInfo* cbInfo = NULL;

	if (connection == NULL) {
		TRACE_IPC("Can't register connection disconnect callback\n", NULL);
		return(ALP_STATUS_IPC_ERROR);
	}

	// free any existing callback
	if (connection->disconnectCB != NULL)
		free((void *)connection->disconnectCB);

	if (callback != NULL) {
      
		cbInfo = (DisconnectCBInfo *)malloc(sizeof(DisconnectCBInfo));

		if (cbInfo == NULL) 
			return ALP_STATUS_IPC_ERROR;
    
		cbInfo->callback = callback;
		cbInfo->cbData = cbData;
	}

	connection->disconnectCB = cbInfo;

	return ALP_STATUS_OK;
}

gint
alp_ipc_connection_get_peer_pid(const AlpConnection* connection) 
{
	if (connection == NULL)
		return -1;

	if (connection->status != ALP_IPC_CONNECTION_READY)
		return -1;

	return connection->pid;
}

gint
alp_ipc_connection_get_peer_gid(const AlpConnection* connection) 
{
    struct ucred credentials;
    socklen_t cred_len;
    int ret_val;
	
    if (connection == NULL)
		return -1;

	if (connection->status != ALP_IPC_CONNECTION_READY)
		return -1;

    cred_len = sizeof(credentials);
    ret_val = getsockopt(connection->socketFD, SOL_SOCKET, SO_PEERCRED, &credentials, &cred_len);

    if (ret_val == -1) {
            TRACE_IPC("Error retrieving peer credentials %s\n", strerror(errno));
            return -1;
    } 

    return credentials.gid;
       
}

gint
alp_ipc_connection_get_peer_uid(const AlpConnection* connection) 
{
    struct ucred credentials;
    socklen_t cred_len;
    int ret_val;
	
    if (connection == NULL)
		return -1;

	if (connection->status != ALP_IPC_CONNECTION_READY)
		return -1;

    cred_len = sizeof(credentials);
    ret_val = getsockopt(connection->socketFD, SOL_SOCKET, SO_PEERCRED, &credentials, &cred_len);

    if (ret_val == -1) {
            TRACE_IPC("Error retrieving peer credentials %s\n", strerror(errno));
            return -1;
    } 

    return credentials.uid;
}

AlpMessage* 
alp_ipc_message_create()
{
	return prv_message_create();
}

alp_status_t
alp_ipc_message_destroy(AlpMessage* msg)
{
	prv_message_destroy(msg);

	return ALP_STATUS_OK;
}

gint 
alp_ipc_message_get_message_ID(AlpMessage* msg)
{
	if (msg)
		return msg->header.msgID;

	return ALP_IPC_MSG_UNKNOWN;
}

static void
prv_check_g_io_channel_status(GIOStatus status, GError* error) 
{
	GIOChannelError gioErr;

	if (status != G_IO_STATUS_NORMAL) {
		//		TRACE_IPC("g_io_channel_write_chars got non-normal status %d", 
		//			   status);

		switch (status) {
		case G_IO_STATUS_ERROR:
			TRACE_IPC("-- status was G_IO_STATUS_ERROR", NULL);
			break;
		case G_IO_STATUS_EOF:
			TRACE_IPC("-- status was G_IO_STATUS_EOF", NULL);
			break;
		case G_IO_STATUS_AGAIN:
			//			TRACE_IPC("-- status was G_IO_STATUS_AGAIN", NULL);
			break;
		default:
			TRACE_IPC("-- unknown g_io_channel_write_chars status", NULL);
			break;
		}
	}

	if (status == G_IO_STATUS_ERROR) {
		TRACE_IPC("-- GError was %s\n", error->message);
	}
}

alp_status_t
alp_ipc_connection_send(AlpConnection* connection, AlpMessage* message)
{
	struct sigaction sigActionStruct;
	struct sigaction oldHandler;

	GIOStatus status;
	guint      numWritten;
	GError*   gErr = NULL;

	gboolean alreadyQueued = FALSE;

	if (connection == NULL) {
		TRACE_IPC("Can't send message\n", NULL);
		return ALP_STATUS_IPC_ERROR;
	}

	if (connection->status != ALP_IPC_CONNECTION_READY) {
		return ALP_STATUS_IPC_NOT_CONNECTED;
	}

	alreadyQueued = (message->numBytesSent > 0);

	pthread_mutex_lock(&connection->writeMutex);
	if (alreadyQueued || (connection->okToWrite == TRUE)) {
		prv_install_SIGPIPE_handler(&oldHandler);

		status = g_io_channel_write_chars(connection->gioChannel,
								    message->bufPtr+message->numBytesSent, 
								    message->bufSize-message->numBytesSent,
								    &numWritten, &gErr);

		prv_check_g_io_channel_status(status, gErr);

		message->numBytesSent += numWritten;

		if (message->bufSize != message->numBytesSent) {
			// message was only partially written, so queue it
			// to be sent later

			if (!alreadyQueued)
				prv_queue_sent_message(connection, message);
		} else if (status != G_IO_STATUS_NORMAL) {
			prv_restore_SIGPIPE_handler(&oldHandler);
			pthread_mutex_unlock(&connection->writeMutex); 
			return ALP_STATUS_IPC_ERROR;
		}

		if (numWritten > 0)
			status = g_io_channel_flush(connection->gioChannel, &gErr);
			
		prv_restore_SIGPIPE_handler(&oldHandler);
	  
		prv_check_g_io_channel_status(status, gErr);
	  
		if (status == G_IO_STATUS_ERROR) {
			pthread_mutex_unlock(&connection->writeMutex);
			return ALP_STATUS_IPC_ERROR;
		}
	} else {
		prv_queue_sent_message(connection, message);
	}

	pthread_mutex_unlock(&connection->writeMutex); 
    
	return ALP_STATUS_OK;
}

alp_status_t
alp_ipc_connection_send_with_response(AlpConnection* connection, 
							   AlpMessage* message,
							   guint timeout, 
							   AlpMessage** response)
{
	alp_status_t status;
  
	if (connection == NULL) {
		TRACE_IPC("Can't send message\n", NULL);
		return ALP_STATUS_IPC_ERROR;
	}

	if (connection->status != ALP_IPC_CONNECTION_READY) {
		return ALP_STATUS_IPC_NOT_CONNECTED;
	}

	status = prv_send_with_response(connection, 
									message,
									timeout, 
									response);

	return status;
}

alp_status_t
alp_ipc_connection_receive(AlpConnection* connection, 
					  gint messageID,
					  guint timeout, 
					  AlpMessage** message)
{
	alp_status_t status;

	if (connection == NULL) {
		TRACE_IPC("Can't receive message\n", NULL);
		return ALP_STATUS_IPC_ERROR;
	}

	if (connection->status != ALP_IPC_CONNECTION_READY) {
		return ALP_STATUS_IPC_NOT_CONNECTED;
	}

	status = prv_wait_for_synch_receive(connection, 
									   messageID,
									   timeout, 
									   message);

	return status;
}

alp_status_t
alp_ipc_connection_register_receive_callback(AlpConnection* connection,
									AlpMessageReceiveCB callback, 
									gpointer cbData)
{
	ReceiveCBInfo* cbInfo = NULL;

	if (connection == NULL) {
		TRACE_IPC("Can't register receive callback\n", NULL);
		return(ALP_STATUS_IPC_ERROR);
	}

	if (connection->hint & ALP_IPC_CONNECTION_USAGE_HINT_SYNCH_ONLY) {
		TRACE_IPC("WARNING! Asynchronous receive callback registered for connection with synch only usage hint\n", NULL);
	}

	// free any existing callback
	if (connection->receiveCB != NULL)
		free((void *)connection->receiveCB);

	if (callback != NULL) {
		cbInfo = (ReceiveCBInfo *)malloc(sizeof(ReceiveCBInfo));

		if (cbInfo == NULL) 
			return ALP_STATUS_IPC_ERROR;
    
		cbInfo->callback = callback;
		cbInfo->cbData = cbData;
	}

	connection->receiveCB = cbInfo;

	return ALP_STATUS_OK;
}

void
alp_ipc_connection_set_custom_data(AlpConnection* connection,
							gpointer customData)
{
	if (connection == NULL) {
		return;
	}

	connection->customData = customData;
}


gpointer
alp_ipc_connection_get_custom_data(AlpConnection* connection)
{
	if (connection == NULL)
		return NULL;

	return connection->customData;
}

alp_status_t
alp_ipc_message_pack_start(AlpMessage* msg, gint msgID)
{
	msg->header.msgVersion = ALP_IPC_MSG_VERSION;
  
	msg->header.msgID = msgID;
	msg->header.sentToken = 0;
	msg->header.responseToken = 0;
	msg->headerSize   = 0;

	msg->msgSize = 0;
	msg->headerPtr = NULL;   // this is set in alp_message_pack_end
	msg->msgPtr = msg->bufPtr + 2 * sizeof(guint);
	msg->currMsgPtr = msg->msgPtr;

	msg->remainingSize = msg->allocBufSize - 2 * sizeof(guint);

	g_string_assign(msg->header.msgFormat, "!");

	msg->numBytesSent = 0;
  
	return ALP_STATUS_OK;
}

alp_status_t
alp_ipc_message_response_pack_start(AlpMessage* msg, gint msgID, AlpMessage* msgToRespondTo)
{
	msg->header.msgVersion = ALP_IPC_MSG_VERSION;

	msg->header.msgID = msgID;
	msg->header.responseToken = msgToRespondTo->header.sentToken;
	msg->header.sentToken = 0;
	msg->headerSize   = 0;

	msg->msgSize = 0;
	msg->headerPtr = NULL;   // this is set in alp_message_pack_end
	msg->msgPtr = msg->bufPtr + 2 * sizeof(guint);
	msg->currMsgPtr = msg->msgPtr;

	msg->remainingSize = msg->allocBufSize - 2 * sizeof(guint);

	msg->numBytesSent = 0;

	g_string_assign(msg->header.msgFormat, "!");
  
	return ALP_STATUS_OK;
}

alp_status_t
alp_ipc_message_pack_end(AlpMessage* msg)
{
	prv_finish_packing(msg);

	return ALP_STATUS_OK;
}

alp_status_t
alp_ipc_message_pack_int32(AlpMessage* msg, gint32 val)
{
	guint valSize;

	g_string_append_c(msg->header.msgFormat, 'i');
  
	valSize = gnet_calcsize("!i", val);
	if (msg->remainingSize < valSize) {
		if (prv_realloc_message_buffer(msg, valSize) == FALSE)
			return ALP_STATUS_IPC_ERROR;
	}
	if (valSize > msg->remainingSize)
		return ALP_STATUS_IPC_ERROR;

	msg->msgSize += valSize;
	msg->remainingSize -= valSize;

	gnet_pack("!i", msg->currMsgPtr, valSize, val);
	msg->currMsgPtr += valSize;
    
	return ALP_STATUS_OK;
}

alp_status_t
alp_ipc_message_pack_uint32(AlpMessage* msg, guint32 val)
{
	guint valSize;

	g_string_append_c(msg->header.msgFormat, 'I');
  
	valSize = gnet_calcsize("!I", val);
	if (msg->remainingSize < valSize) {
		if (prv_realloc_message_buffer(msg, valSize) == FALSE)
      
			return ALP_STATUS_IPC_ERROR;
	}
	if (valSize > msg->remainingSize)
		return ALP_STATUS_IPC_ERROR;

	msg->msgSize += valSize;
	msg->remainingSize -= valSize;

	gnet_pack("!I", msg->currMsgPtr, valSize, val);
	msg->currMsgPtr += valSize;
    
	return ALP_STATUS_OK;
}

alp_status_t
alp_ipc_message_pack_float(AlpMessage* msg, gfloat val)
{
	guint valSize;

	g_string_append_c(msg->header.msgFormat, 'f');
  
	valSize = gnet_calcsize("!f", val);
	if (msg->remainingSize < valSize) {
		if (prv_realloc_message_buffer(msg, valSize) == FALSE)
      
			return ALP_STATUS_IPC_ERROR;
	}
	if (valSize > msg->remainingSize)
		return ALP_STATUS_IPC_ERROR;

	msg->msgSize += valSize;
	msg->remainingSize -= valSize;

	gnet_pack("!f", msg->currMsgPtr, valSize, val);
	msg->currMsgPtr += valSize;
    
	return ALP_STATUS_OK;
}

alp_status_t
alp_ipc_message_pack_string(AlpMessage* msg, const gchar* val)
{
	guint valSize;

	if (val == NULL) {
		g_string_append_c(msg->header.msgFormat, 'n');
		valSize = 0;
	} else {
		g_string_append_c(msg->header.msgFormat, 's');
		valSize = gnet_calcsize("!s", val);

		if (msg->remainingSize < valSize) {
			if (prv_realloc_message_buffer(msg, valSize) == FALSE)
				return ALP_STATUS_IPC_ERROR;
		}

		if (valSize > msg->remainingSize)
			return ALP_STATUS_IPC_ERROR;

		msg->msgSize += valSize;
		msg->remainingSize -= valSize;

		if (val != NULL)
			gnet_pack("!s", msg->currMsgPtr, valSize, val);

		msg->currMsgPtr += valSize;
	}

	return ALP_STATUS_OK;
}

alp_status_t
alp_ipc_message_pack_byte_array(AlpMessage* msg, guchar* array, guint numElements)
{
	guint valSize;

	if (array == NULL)
		numElements = 0;

	g_string_append_c(msg->header.msgFormat, 'b'); 

	if (numElements > 0) 
		valSize = gnet_calcsize("!Ir", numElements, array, numElements);
	else
		valSize = gnet_calcsize("!I", 0);

	if (msg->remainingSize < valSize) {
		if (prv_realloc_message_buffer(msg, valSize) == FALSE)
			return ALP_STATUS_IPC_ERROR;
	}

	if (valSize > msg->remainingSize)
		return ALP_STATUS_IPC_ERROR;

	msg->msgSize += valSize;
	msg->remainingSize -= valSize;

	if (numElements > 0) 
		gnet_pack("!Ir", msg->currMsgPtr, valSize, numElements, array, numElements);
	else
		gnet_pack("!I", msg->currMsgPtr, valSize, 0);

	msg->currMsgPtr += valSize;

	return ALP_STATUS_OK;
}

alp_status_t
alp_ipc_message_unpack_start(AlpMessage* msg)
{
	msg->remainingSize = msg->msgSize;
	msg->formatIndex = 1;
	msg->currMsgPtr = msg->msgPtr;

	return ALP_STATUS_OK;
}

alp_status_t
alp_ipc_message_unpack_end(AlpMessage* msg)
{
	return ALP_STATUS_OK;
}

alp_status_t
alp_ipc_message_unpack_int32(AlpMessage* msg, gint32* val)
{ 
	guint numBytes;

	if (msg->formatIndex >= msg->header.msgFormat->len)
		return ALP_STATUS_IPC_ERROR;

	if (msg->header.msgFormat->str[msg->formatIndex++] != 'i')
		return ALP_STATUS_IPC_ERROR;

	numBytes = gnet_unpack("!i", msg->currMsgPtr, msg->remainingSize, val);

	if (numBytes > msg->remainingSize)
		return ALP_STATUS_IPC_ERROR;

	msg->currMsgPtr += numBytes;
	msg->remainingSize -= numBytes;

	return ALP_STATUS_OK;
}


alp_status_t
alp_ipc_message_unpack_uint32(AlpMessage* msg, guint32* val)
{ 
	guint numBytes;

	if (msg->formatIndex >= msg->header.msgFormat->len)
		return ALP_STATUS_IPC_ERROR;

	if (msg->header.msgFormat->str[msg->formatIndex++] != 'I')
		return ALP_STATUS_IPC_ERROR;

	numBytes = gnet_unpack("!I", msg->currMsgPtr, msg->remainingSize, val);

	if (numBytes > msg->remainingSize)
		return ALP_STATUS_IPC_ERROR;

	msg->currMsgPtr += numBytes;
	msg->remainingSize -= numBytes;

	return ALP_STATUS_OK;
}


alp_status_t
alp_ipc_message_unpack_float(AlpMessage* msg, gfloat* val)
{ 
	guint numBytes;

	if (msg->formatIndex >= msg->header.msgFormat->len)
		return ALP_STATUS_IPC_ERROR;

	if (msg->header.msgFormat->str[msg->formatIndex++] != 'f')
		return ALP_STATUS_IPC_ERROR;

	numBytes = gnet_unpack("!f", msg->currMsgPtr, msg->remainingSize, val);

	if (numBytes > msg->remainingSize)
		return ALP_STATUS_IPC_ERROR;

	msg->currMsgPtr += numBytes;
	msg->remainingSize -= numBytes;

	return ALP_STATUS_OK;
}

alp_status_t
alp_ipc_message_unpack_string(AlpMessage* msg, gchar** val)
{
	guint numBytes;
	gchar formatChar ;

	if (msg->formatIndex >= msg->header.msgFormat->len)
		return ALP_STATUS_IPC_ERROR;

	formatChar = msg->header.msgFormat->str[msg->formatIndex++];
	if (formatChar == 's') {
		numBytes = gnet_unpack("!s", msg->currMsgPtr, msg->remainingSize, val);

		if (numBytes > msg->remainingSize)
			return ALP_STATUS_IPC_ERROR;

		msg->currMsgPtr += numBytes;
		msg->remainingSize -= numBytes;
	
	} else if (formatChar == 'n') {
		*val = NULL;
	} else {
		return ALP_STATUS_IPC_ERROR;
	}

	return ALP_STATUS_OK;
}

alp_status_t
alp_ipc_message_unpack_byte_array(AlpMessage* msg, guchar** array, guint* numElements)
{
	guint numBytes;

	if (msg->formatIndex >= msg->header.msgFormat->len)
		return ALP_STATUS_IPC_ERROR;

	if (msg->header.msgFormat->str[msg->formatIndex++] != 'b') 
		return ALP_STATUS_IPC_ERROR;

	numBytes = gnet_unpack("!I", msg->currMsgPtr, msg->remainingSize, 
					   numElements);

	if (numBytes > msg->remainingSize)
		return ALP_STATUS_IPC_ERROR;

	msg->currMsgPtr += numBytes;
	msg->remainingSize -= numBytes;

	if ((*numElements) > 0) {
		numBytes = gnet_unpack("r", msg->currMsgPtr, 
						   msg->remainingSize, array, *numElements);

		if (numBytes > msg->remainingSize)
			return ALP_STATUS_IPC_ERROR;

		msg->currMsgPtr += numBytes;
		msg->remainingSize -= numBytes;
	} else {
		*array = NULL;
	}

	return ALP_STATUS_OK;
}

