/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/*!
 * @file alp_ipc_internal.c
 *
 * @brief Contains implementation of functions used internally by 
 * the IPC library
 *
 *****************************************************************************/

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>

#include <fcntl.h>
#include <errno.h>
#include <unistd.h>

#include <stdlib.h>
#include <string.h>

#include <pthread.h>
#include <dirent.h>

#include <gnet-2.0/pack.h>

#include <hiker/ipc.h>
#include <hiker/traces.h>

#include "alp_ipc_internal.h"

static gchar gMSG_HEADER_GNET_FORMAT[] = "!iiiis";

AlpChannel*
prv_channel_create()
{
	AlpChannel* channel = (AlpChannel *)malloc(sizeof(AlpChannel));

	if (channel == NULL) {
		TRACE_IPC("Can't create channel\n", NULL);
		return NULL;
	}

	channel->socketName = g_string_new("");
	channel->socketFD = -1;
	channel->gioChannel = NULL;	

	channel->connectCB = NULL;
	channel->disconnectCB = NULL;

	channel->connectionTable = g_hash_table_new(g_direct_hash, 
									    g_direct_equal);

	channel->inSource = 0;

	channel->customData = NULL;

	return channel;
}

void
prv_channel_destroy(AlpChannel* channel)
{
	GError*   gError = NULL;

	if (channel == NULL) {
		TRACE_IPC("Can't destroy channel\n", NULL);
		return;
	}

	if (channel->connectCB != NULL) {
		free((void *)channel->connectCB);
	}

	if (channel->disconnectCB != NULL) {
		free((void *)channel->disconnectCB);
	}

	struct sigaction oldHandler;
	prv_install_SIGPIPE_handler(&oldHandler);

	// using local sockets
	if (channel->gioChannel) {
		g_io_channel_shutdown(channel->gioChannel, 
						  TRUE, &gError);

		g_io_channel_unref(channel->gioChannel);
	}  

	prv_restore_SIGPIPE_handler(&oldHandler);

	if (channel->socketFD >= 0) {
		close(channel->socketFD);
	}

	unlink(channel->socketName->str);

	// clean up the connection table
	if (channel->connectionTable) {
		g_hash_table_foreach_remove(channel->connectionTable,
							   prv_connection_table_remove_CB,
							   NULL);
	
		g_hash_table_destroy(channel->connectionTable);
	}

	g_string_free(channel->socketName, TRUE);

	if (channel->inSource) {
		g_source_remove(channel->inSource);
	}

	free((void *)channel);
}

AlpConnection*
prv_connection_create(AlpConnectionUsageHint hint)
{
	AlpConnection* connection = (AlpConnection *)malloc(sizeof(AlpConnection));
	if (connection == NULL) {
		TRACE_IPC("Can't create connection\n", NULL);
		return NULL;
	}

	// using local sockets
	connection->socketFD = -1;
	connection->gioChannel = NULL;  

	connection->status = ALP_IPC_CONNECTION_CREATED;

	connection->amServer = FALSE;

	connection->inSource = 0;
	connection->outSource = 0;
	connection->okToWrite = TRUE;

	connection->connectCB = NULL;
	connection->disconnectCB = NULL;
	connection->receiveCB = NULL;

	connection->receiveContext = NULL;
	connection->receiveMainLoop = NULL;

	connection->receiveTimeoutSrc = NULL;
	connection->receiveIOWatchSrc = NULL;

	connection->responseTokenCounter = 1;

	connection->receiveMessageQueue = NULL;

	connection->sentMessageQueue = NULL;

	connection->partialMessage = NULL;
	connection->partialMessageRemainingSize = 0;
	connection->partialMessageBufPtr = NULL;

	pthread_mutex_init(&connection->writeMutex, NULL);

	connection->customData = NULL;

	connection->hint = hint;

	return connection;
}

void
prv_connection_destroy(AlpConnection* connection)
{
	GError*   gError = NULL;
	AlpMessage* tmpMessage = NULL;
	guint     i;

	struct sigaction oldHandler;

	if (connection == NULL) {
		TRACE_IPC("Can't destroy connection\n", NULL);
		return;
	}

	if (connection->connectCB != NULL) {
		free((void *)connection->connectCB);
	}

	if (connection->disconnectCB != NULL) {
		free((void *)connection->disconnectCB);
	}

	if (connection->receiveCB != NULL) {
		free((void *)connection->receiveCB);
	}

	// shut down the toServer pipe
	if (connection->gioChannel) {
		prv_install_SIGPIPE_handler(&oldHandler);
		g_io_channel_shutdown(connection->gioChannel, TRUE, &gError);
		prv_restore_SIGPIPE_handler(&oldHandler);

		g_io_channel_unref(connection->gioChannel);
	}  
    
	if (connection->socketFD >= 0) {
		close(connection->socketFD);
	}

	if (connection->inSource)
		g_source_remove(connection->inSource);

	if (connection->outSource)
		g_source_remove(connection->outSource);

	if (connection->receiveTimeoutSrc != NULL) {
		g_source_destroy(connection->receiveTimeoutSrc);
		g_source_unref(connection->receiveTimeoutSrc);
	}

	if (connection->receiveIOWatchSrc != NULL) {
		g_source_destroy(connection->receiveIOWatchSrc);
		g_source_unref(connection->receiveIOWatchSrc);
	}

	if (connection->receiveMainLoop != NULL) {
		if (g_main_loop_is_running(connection->receiveMainLoop)) {
			g_main_loop_quit(connection->receiveMainLoop);
		}
		g_main_loop_unref(connection->receiveMainLoop);
	}

	if (connection->receiveContext != NULL) {
		g_main_context_unref(connection->receiveContext);
	}

	if (connection->receiveMessageQueue) {
		pthread_mutex_lock(&connection->receiveMessageQueueMutex);
		if (connection->receiveMessageQueue != NULL) {
			for (i=0; i<connection->receiveMessageQueue->len; i++) {
				tmpMessage = g_array_index(connection->receiveMessageQueue, AlpMessage*, i);
				if (tmpMessage)
					alp_ipc_message_destroy(tmpMessage);
			}
			g_array_free(connection->receiveMessageQueue, TRUE);
		}
			
		connection->receiveMessageQueue = NULL;
		pthread_mutex_destroy(&connection->receiveMessageQueueMutex);
	}

	// don't allow any one else to write
	pthread_mutex_lock(&connection->writeMutex);
	if (connection->sentMessageQueue) {
		pthread_mutex_lock(&connection->sentMessageQueueMutex);
		connection->okToWrite = FALSE;
		if (connection->sentMessageQueue != NULL) {
			for (i=0; i<connection->sentMessageQueue->len; i++) {
				tmpMessage = g_array_index(connection->sentMessageQueue, AlpMessage*, i);
				if (tmpMessage)
					alp_ipc_message_destroy(tmpMessage);
			}
			g_array_free(connection->sentMessageQueue, TRUE);
		}
			
		connection->sentMessageQueue = NULL;
		pthread_mutex_destroy(&connection->sentMessageQueueMutex);
	}
	pthread_mutex_unlock(&connection->writeMutex);

	pthread_mutex_destroy(&connection->writeMutex);

	free((void *)connection);
}

AlpMessage*
prv_message_create()
{
	AlpMessage* message = (AlpMessage *)malloc(sizeof(AlpMessage));

	if (message == NULL) {
		TRACE_IPC("Can't create message\n", NULL);
		return NULL;
	}

	message->header.msgVersion = ALP_IPC_MSG_VERSION;
	message->header.msgID = ALP_IPC_MSG_UNKNOWN;
	message->header.msgFormat =  
		g_string_sized_new(DEFAULT_FORMAT_STRING_LENGTH);
	if (message->header.msgFormat == NULL)
		return NULL;

	message->header.responseToken = 0;

	message->bufPtr = (gchar *)malloc(sizeof(gchar) * DEFAULT_BUFFER_SIZE);
	if (message->bufPtr == NULL)
		return NULL;
	message->bufSize = 0;
	message->headerPtr = NULL;
	message->headerSize = 0;
	message->msgPtr = NULL;
	message->msgSize = 0;

	message->allocBufSize = DEFAULT_BUFFER_SIZE;

	message->numBytesSent = 0;

	return message;
}

AlpMessage*
prv_message_clone(AlpMessage* msg)
{
	AlpMessage* newMsg = prv_message_create();

	newMsg->header.msgVersion = msg->header.msgVersion;
	newMsg->header.msgID = msg->header.msgID;
	newMsg->header.sentToken = msg->header.sentToken;
	newMsg->header.responseToken = msg->header.responseToken;
	newMsg->header.msgFormat = g_string_new(msg->header.msgFormat->str);

	newMsg->bufPtr = 
		(gchar *)malloc(sizeof(gchar) * msg->bufSize);

	memcpy(newMsg->bufPtr, msg->bufPtr, msg->bufSize);
	newMsg->bufSize = msg->bufSize;

	newMsg->allocBufSize = msg->bufSize;
	newMsg->remainingSize = 0;

	newMsg->headerSize = msg->headerSize;

	newMsg->msgSize = msg->msgSize;

	newMsg->msgPtr    = newMsg->bufPtr + 2 * sizeof(guint);
	newMsg->headerPtr = newMsg->msgPtr + newMsg->msgSize;

	newMsg->numBytesSent = msg->numBytesSent;

	return newMsg;
}

void
prv_message_destroy(AlpMessage* message)
{
	if (message == NULL) {
		TRACE_IPC("Can't destroy message\n", NULL);
		return;
	}

	if (message->header.msgFormat != NULL)
		g_string_free(message->header.msgFormat, TRUE);

	if (message->bufPtr != NULL) 
		free((void *)message->bufPtr);

	free((void *)message);
}

GIOStatus
prv_get_msg_header(AlpConnection * connection, AlpMessage** msg, GIOChannel* source) {
	gchar   rcvBuf[16];
	GError* gErr = NULL;
	guint	numRead;
	GIOStatus   status;
	gchar*   tmpStr;

	guint	numBytesUnpacked;
	guint	 remainingSize;

	guint	i;

	AlpMessage* newMsg = NULL;
	*msg = NULL;

	if ((connection != NULL) && (connection->partialMessage != NULL)) {
		remainingSize = connection->partialMessageRemainingSize;
		status = g_io_channel_read_chars(source, 
								   connection->partialMessageBufPtr,
								   connection->partialMessageRemainingSize,
								   &numRead, 
								   &gErr);

		if (status != G_IO_STATUS_NORMAL)
			return status;

		if (connection->partialMessageRemainingSize != numRead) {
			// still didn't get all of the message
			connection->partialMessageRemainingSize -= numRead;
			connection->partialMessageBufPtr += numRead;

			*msg = NULL;
			return status;
		} else {
			// the rest of the message was received OK
			*msg = connection->partialMessage;
			newMsg = connection->partialMessage;
			connection->partialMessage = NULL;
			connection->partialMessageRemainingSize = 0;
			connection->partialMessageBufPtr = NULL;
		}
		
	} else {

		// first, get the headerSize, a single int stored in the first
		// four bytes of the message
		status = g_io_channel_read_chars(source, rcvBuf, 
								   2*sizeof(guint),
								   &numRead, &gErr);

		// return if there's nothing to read
		if (status != G_IO_STATUS_NORMAL) 
			return status;
	
		newMsg = alp_ipc_message_create();
		*msg = newMsg;

		newMsg->headerSize = *((guint *)rcvBuf);
		newMsg->msgSize = *((guint *)(rcvBuf+4));
		newMsg->bufSize = 2 * sizeof(guint) + newMsg->headerSize + newMsg->msgSize;

		// now allocate the message buffer
		// get the rest of the message	

		if (newMsg->allocBufSize < newMsg->bufSize ) {
			newMsg->allocBufSize = newMsg->bufSize;
			if (newMsg->bufPtr != NULL) {
				free((void *)newMsg->bufPtr);
				newMsg->bufPtr = NULL;
			}
			newMsg->bufPtr = 
				(gchar *)malloc(sizeof(gchar) * newMsg->allocBufSize);

			if (newMsg->bufPtr == NULL) {
				TRACE_IPC("Can't allocate message buffer\n", NULL);
				return G_IO_STATUS_ERROR;
			}
		}
		newMsg->msgPtr	= newMsg->bufPtr + 2 * sizeof(guint);
		newMsg->headerPtr = newMsg->msgPtr + newMsg->msgSize;

		// stick the eight bytes already
		// read in the first four bytes of the message buff
		for (i=0; i<2 * sizeof(guint); i++) {
			newMsg->bufPtr[i] = rcvBuf[i];
		}

		remainingSize = newMsg->msgSize + newMsg->headerSize;
		status = g_io_channel_read_chars(source, newMsg->msgPtr,
								   remainingSize, &numRead, &gErr);

		if (status != G_IO_STATUS_NORMAL)
			return status;		// ... and read the rest of the message, including the header and message


		if (remainingSize != numRead) {
			// didn't get all of the message, so set up the partial message
			if (connection == NULL)
				return G_IO_STATUS_ERROR;

			connection->partialMessage = newMsg;
			connection->partialMessageRemainingSize = remainingSize - numRead;
			connection->partialMessageBufPtr = newMsg->msgPtr + numRead;

			*msg = NULL;

			return status;
		}
	}

	numBytesUnpacked = gnet_unpack(gMSG_HEADER_GNET_FORMAT, 
							 newMsg->headerPtr,
							 newMsg->headerSize,
							 &newMsg->header.msgVersion,
							 &newMsg->header.msgID,
							 &newMsg->header.sentToken,
							 &newMsg->header.responseToken,
							 &tmpStr);
//	TRACE_IPC("UNPACKED MESSAGE rt %d\n", newMsg->header.responseToken);		
	if (numBytesUnpacked != newMsg->headerSize) {
		TRACE_IPC("Error unpacking message\n", NULL);
		return G_IO_STATUS_ERROR;
	}
		

	if (newMsg->header.msgFormat != NULL)
		g_string_free(newMsg->header.msgFormat, TRUE);
	newMsg->header.msgFormat = g_string_new(tmpStr);
	g_free(tmpStr);
	 
	return status;
}


void
prv_invoke_message_callback(AlpConnection* connection, AlpMessage* message)
{
	if (connection->receiveCB) {
		connection->receiveCB->callback(connection, message, 
								  connection->receiveCB->cbData);
	}
}

static
void
prv_handle_message_for_server(AlpChannel* channel,
						AlpConnection* connection, 
						AlpMessage* msg)
{
	// the message should never be a response to
	// a call to alp_ipc_connection_send_with_response()
	if (msg->header.responseToken != 0) {
		TRACE_IPC("Invalid Response\n", NULL);

		// get rid of the message and return
		alp_ipc_message_destroy(msg);
		return;
	}

	if (msg->header.msgID >= 0) {
		// if it's a user message (messageID >=0), then
		// call any message received callbacks that are registered
		if (connection)
			prv_invoke_message_callback(connection, msg);
		else
			// get rid of the message 
			alp_ipc_message_destroy(msg);
	} else {
		// else, it's an internal message, so handle it
		switch (msg->header.msgID) {
		default:
			//			g_print("UNKNOWN INTERNAL MESSAGE RECEIVED -- ID %d\n",
			//				   msg->header.msgID);

			TRACE_IPC("Unrecognized Internal Message\n", 
					NULL);

			break;
		}
				
		// get rid of the message 
		alp_ipc_message_destroy(msg);
	}
}

gboolean
prv_server_received_input_hangup_CB(GIOChannel* source, GIOCondition condition, gpointer data)
{
	AlpMessage* msg = NULL;

	GIOStatus gioStatus;
	guint	 headerSize;
	guint	 msgSize;
	guint	 bufSize;

	AlpChannel* channel = (AlpChannel *) data;

	AlpConnection* connection = 
		(AlpConnection *)g_hash_table_lookup(channel->connectionTable,
									  (gpointer)source);

	// before handling this message, send any queued messages
	if (connection && 
	    connection->receiveMessageQueue &&
	    connection->receiveMessageQueue->len > 0) {
		prv_send_queued_messages(connection);
	}

	// first handle any data on the pipe
	if (condition & G_IO_IN) { 
		while (1) {
			// read all input
			// get the message header
			gioStatus = prv_get_msg_header(connection, &msg, source);
			/*
			  if (gioStatus == G_IO_STATUS_NORMAL)
			  g_print("GIO NORMAL\n");
			  else if (gioStatus == G_IO_STATUS_ERROR)
			  g_print("GIO ERROR\n");
			  else if (gioStatus == G_IO_STATUS_AGAIN)
			  g_print("GIO AGAIN\n");
			  else if (gioStatus == G_IO_STATUS_EOF)
			  g_print("GIO EOF\n");
			*/

			if (gioStatus != G_IO_STATUS_NORMAL) {
				if (msg != NULL)
					alp_ipc_message_destroy(msg);
				
				break;
			}

			// It's OK for G_IO_STATUS_NORMAL to be returned and
			// for msg to be NULL.  In this case, the msg was
			// only partially received

			if (msg != NULL) {
				// the connection should be non-NULL
				if (connection != NULL) {
					prv_handle_message_for_server(channel, 
											connection, msg);
				} else {
					TRACE_IPC("No connection exists for message\n",
							NULL);
				}
			}

		}
	} 

	// then handle any hangups
	if (condition & G_IO_HUP) {
		if (connection == NULL) {
			TRACE_IPC("NULL connection for hangup\n", NULL);
			return TRUE;
		}

		// call any disconnect callback that has been registered
		if (channel->disconnectCB) {
			channel->disconnectCB->callback(connection, 
									  channel->disconnectCB->cbData);
		}

		if (g_hash_table_remove(channel->connectionTable, 
						    connection->gioChannel) == FALSE) {

			TRACE_IPC("Could not find connection for hangup\n",
					NULL);
		} 

		prv_connection_destroy(connection);

		return FALSE; // this removes the GSource for the hangup callback
	}

	return TRUE;
}


static void
prv_handle_message_for_client(AlpConnection* connection, AlpMessage* msg)
{
	// the message should never be a response to
	// a call to alp_ipc_connection_send_with_response()
	if (msg->header.responseToken != 0) {
		TRACE_IPC("Invalid Response\n", NULL);
//		TRACE_IPC("MESSAGE HEADER %d RESPONSE TOKEN %d CONNECTION %d\n",
//				msg->header.msgID,
//				msg->header.responseToken,
//				connection->socketFD);
		// get rid of the message and return
		alp_ipc_message_destroy(msg);
		
		return;
	}

	if (msg->header.msgID >= 0) {
		prv_invoke_message_callback(connection, msg);
	} else {
	
		switch (msg->header.msgID) {
		default:
			TRACE_IPC("Unknown internal message\n", NULL);
			break;
		}

		// get rid of the message 
		alp_ipc_message_destroy(msg);
	}
}

gboolean
prv_client_received_input_hangup_CB(GIOChannel* source, GIOCondition condition, gpointer data)
{
	AlpMessage* msg = NULL;
	GIOStatus gioStatus;

	AlpConnection* connection = (AlpConnection *) data;

	if (connection == NULL) {
		TRACE_IPC("Can't register receive callback\n", NULL);
		return TRUE;
	}

	// before handling this message, send any queued messages
	if (connection->receiveMessageQueue &&
	    connection->receiveMessageQueue->len > 0) {
		prv_send_queued_messages(connection);
	}
	
	// first handle any data on the pipe
	if (condition & G_IO_IN ) {
		while (1) {
		
			// read all input
			// get the message header
			gioStatus = prv_get_msg_header(connection, &msg, source);
			/*
			  if (gioStatus == G_IO_STATUS_NORMAL)
			  g_print("GIO NORMAL\n");
			  else if (gioStatus == G_IO_STATUS_ERROR)
			  g_print("GIO ERROR\n");
			  else if (gioStatus == G_IO_STATUS_AGAIN)
			  g_print("GIO AGAIN\n");
			  else if (gioStatus == G_IO_STATUS_EOF)
			  g_print("GIO EOF\n");
			*/
			if (gioStatus != G_IO_STATUS_NORMAL) {
				if (msg != NULL)
					alp_ipc_message_destroy(msg);
				
				break;
			}

			// It's OK for G_IO_STATUS_NORMAL to be returned and
			// for msg to be NULL.  In this case, the msg was
			// only partially received
							  
			if (msg != NULL)
				prv_handle_message_for_client(connection, msg);
		}
	} 

	// then handle any hangups
	if (condition & G_IO_HUP) {
		
		// call any disconnect callback that has been registered
		if (connection->disconnectCB) {
			connection->disconnectCB->callback(connection, 
										connection->disconnectCB->cbData);
		}

		prv_connection_destroy(connection);
		
		return FALSE;  // this removes the GSource
	}

	return TRUE;
}

gboolean
prv_received_output_hangup_CB(GIOChannel* source, GIOCondition condition, gpointer data)
{
	AlpMessage* msg = NULL;
	GIOStatus gioStatus;
	gboolean retVal = FALSE;
	guint mm;

	AlpConnection* connection = (AlpConnection *) data;

	if (connection == NULL) {
		TRACE_IPC("Can't register receive callback\n", NULL);
		return FALSE;
	}

	// first handle any data on the pipe
	if (condition & G_IO_OUT) { 
		pthread_mutex_lock(&connection->sentMessageQueueMutex);

		for (mm = 0; mm<connection->sentMessageQueue->len; mm++) {
			msg = g_array_index(connection->sentMessageQueue, AlpMessage *, mm);

			if (msg == NULL)
				continue;

			alp_ipc_connection_send(connection, msg);

			if (msg->numBytesSent == msg->bufSize) {
				// the queue has cloned copies of the original
				// messages, so should always destroy them
				alp_ipc_message_destroy(msg);
				g_array_remove_index(connection->sentMessageQueue,
								 mm);
			} else {
				break;
			}
		}
		if (connection->sentMessageQueue->len > 0) {
			retVal = TRUE; // return TRUE to not cancel this callback
		} else {
			connection->okToWrite = TRUE;  // all queued messages are 
			// sent  no need to queue
			// message anymore
		}
		pthread_mutex_unlock(&connection->sentMessageQueueMutex);

	} 
	return retVal;
}

gboolean
prv_server_accept_CB(GIOChannel* source, GIOCondition condition, gpointer data)
{
	AlpChannel* channel = (AlpChannel *)data;

	GError* gErr = NULL;

	int ret_val;

	if (channel == NULL)
		return FALSE;

	if (condition & G_IO_IN) {
		// call accept
		struct sockaddr_un client_addr;

		socklen_t client_addr_len = sizeof(client_addr);
		
		int clientFD = accept(channel->socketFD, 
						  (struct sockaddr *)&client_addr, 
						  &client_addr_len);

		if (clientFD < 0) {
			TRACE_IPC("Accept failed for server %s\n",  
					channel->socketName->str);

			return TRUE;
		} 

		// set up the server end of the connection
		AlpConnection* connection = prv_connection_create(ALP_IPC_CONNECTION_USAGE_HINT_NONE);

		connection->socketFD = clientFD;
		connection->amServer = TRUE;

		// make the socket be non-blocking
		ret_val = fcntl(connection->socketFD, F_SETFL, O_NONBLOCK);

		if (ret_val == -1) {
			TRACE_IPC("Fcntl error for server socket %s\n", channel->socketName->str);
		
			prv_connection_destroy(connection);
			
			return TRUE;
		}

		// set up the GIOChannels for the toServer pipe
		connection->gioChannel = g_io_channel_unix_new(connection->socketFD);

		if (connection->gioChannel == NULL) {
			TRACE_IPC("Error creating the server GIOChannel for connection\n", NULL);
	
			prv_connection_destroy(connection);
			
			return TRUE;
		}

		g_io_channel_set_encoding(connection->gioChannel, NULL, &gErr);
		
		// add a input and hangup watch on the pipe from the client
		connection->inSource = g_io_add_watch(connection->gioChannel, 
									   G_IO_IN | G_IO_HUP, 
									   prv_server_received_input_hangup_CB,
									   (gpointer)channel);

		struct ucred credentials;

		socklen_t cred_len = sizeof(credentials);
		ret_val = getsockopt(connection->socketFD, SOL_SOCKET, SO_PEERCRED, &credentials, &cred_len);

		if (ret_val == -1) {
			TRACE_IPC("Error retrieving peer credentials %s\n", strerror(errno));
		} else {
			connection->pid = credentials.pid;

//			g_print("PID IS %d \n", 
//				   connection->pid);
			
			connection->status = ALP_IPC_CONNECTION_READY;
	
			// stick the connection into the channel's connection hash table
			g_hash_table_insert(channel->connectionTable, 
							connection->gioChannel,
							(gpointer)connection);

			// invoke any registered connection callback
			if ((channel->connectCB) && (connection != NULL)) {
				channel->connectCB->callback(connection, 
									    channel->connectCB->cbData);
			}		
		}
	}

	return TRUE;
}


gboolean 
prv_connection_table_remove_CB(gpointer key, gpointer value, gpointer userData)
{
	prv_connection_destroy((AlpConnection *)value);
	return(TRUE);
}


gboolean
prv_finish_packing(AlpMessage* message)
{
	guint* iPtr = NULL;

	message->headerSize = gnet_calcsize(gMSG_HEADER_GNET_FORMAT, 
								 message->header.msgVersion,
								 message->header.msgID,
								 message->header.sentToken,
								 message->header.responseToken,
								 message->header.msgFormat->str);

	message->bufSize = 2 * sizeof(guint) + message->headerSize + message->msgSize;

	if (message->bufSize > message->allocBufSize) {
		if (prv_realloc_message_buffer(message,
								 message->bufSize - message->allocBufSize + message->remainingSize) == FALSE)
	  
			return FALSE;

		if (message->bufPtr == NULL)
			return FALSE;
		if (message->bufSize > message->allocBufSize)
			return FALSE;
	}

	iPtr = (guint *)message->bufPtr;
	*iPtr = message->headerSize;

	iPtr = (guint *)(message->bufPtr + 4);
	*iPtr = message->msgSize;

	message->headerPtr = message->bufPtr + 2 * sizeof(guint) + message->msgSize;
//	TRACE_IPC("PACKING MESSAGE rt %d\n",
//			message->header.responseToken);
	// pack the header
	gnet_pack(gMSG_HEADER_GNET_FORMAT, 
			message->headerPtr,
			message->headerSize,
			message->header.msgVersion,
			message->header.msgID,
			message->header.sentToken,
			message->header.responseToken,
			message->header.msgFormat->str);
  
	return TRUE;
}

gboolean
prv_repack_header(AlpMessage* message)
{
//	TRACE_IPC("REPACKING MESSAGE rt %d\n",
//			message->header.responseToken);
	// pack the header
	gnet_pack(gMSG_HEADER_GNET_FORMAT, 
			message->headerPtr,
			message->headerSize,
			message->header.msgVersion,
			message->header.msgID,
			message->header.sentToken,
			message->header.responseToken,
			message->header.msgFormat->str);
  
	return TRUE;
}

gboolean
prv_realloc_message_buffer(AlpMessage* message, guint numBytesNeeded) 
{
	guint newSize = 0;

	while (newSize + message->remainingSize < numBytesNeeded)
		newSize += DEFAULT_BUFFER_SIZE;

	message->bufPtr = realloc(message->bufPtr, 
						 message->allocBufSize + newSize);
	
	if (message->bufPtr == NULL) {
		TRACE_IPC("Can't realloc message buffer\n", NULL);
		return FALSE;
	}
  
	message->allocBufSize  += newSize;
	message->remainingSize += newSize;

	message->msgPtr = message->bufPtr + 2 * sizeof(guint);
	message->currMsgPtr = message->msgPtr + message->msgSize;
  
	return TRUE;
}

static void
prv_SIGPIPE_handler(int sigNumber, siginfo_t* siginfo, void* context)
{
	TRACE_IPC("HANDLING SIGPIPE", NULL);
}

void
prv_install_SIGPIPE_handler(struct sigaction* oldHandler)
{
	struct sigaction sigActionStruct;

	// install a SIGPIPE handler
	sigActionStruct.sa_sigaction = prv_SIGPIPE_handler;
	sigemptyset(&sigActionStruct.sa_mask);

	// SA_RESETHAND means restore the previous handler
	// after the handler executes
	sigActionStruct.sa_flags = SA_SIGINFO | SA_RESETHAND;
	sigActionStruct.sa_restorer = NULL;

	sigaction(SIGPIPE, &sigActionStruct, oldHandler);
}

void
prv_restore_SIGPIPE_handler(struct sigaction* oldHandler)
{
	sigaction(SIGPIPE, oldHandler, NULL);
}

void
prv_install_g_io_out_callback(AlpConnection* connection)
{
	// using local sockets
	// add a output and hangup watch on the pipe from the client
	connection->outSource = 
		g_io_add_watch(connection->gioChannel, 
					G_IO_OUT, 
					prv_received_output_hangup_CB, 
					(gpointer)connection);
}

alp_status_t
prv_queue_received_message(AlpConnection* connection, AlpMessage* msg)
{
	gboolean isFirstQueuedMessage = FALSE;

	if (connection == NULL)
		return ALP_STATUS_IPC_ERROR;

	if (msg == NULL)
		return ALP_STATUS_OK;

	if (connection->receiveMessageQueue == NULL) {
		if (pthread_mutex_init(&connection->receiveMessageQueueMutex, NULL) != 0)
			return ALP_STATUS_IPC_ERROR;

		pthread_mutex_lock(&connection->receiveMessageQueueMutex);

		connection->receiveMessageQueue = g_array_new(FALSE, TRUE, sizeof(AlpMessage*));
	} else {
		pthread_mutex_lock(&connection->receiveMessageQueueMutex);
	}

	if (connection->receiveMessageQueue->len == 0)
		isFirstQueuedMessage = TRUE;

	g_array_append_val(connection->receiveMessageQueue, msg);
	//	g_print("QUEUING MESSAGE ID %d\n", msg->header.msgID);
	//	g_print("THERE ARE %d QUEUED MESSAGES\n",
	//		   connection->receiveMessageQueue->len);
	pthread_mutex_unlock(&connection->receiveMessageQueueMutex);

	if (isFirstQueuedMessage &&
		(connection->hint & ALP_IPC_CONNECTION_USAGE_HINT_SYNCH_ONLY) == 0)
	{
		// add an idle callback to send the queued messages
		g_idle_add_full(G_PRIORITY_HIGH_IDLE, 
					 prv_idle_send_queued_messages_CB, 
					 (gpointer)connection, 
					 NULL);
	}

	return ALP_STATUS_OK;
}

gboolean 
prv_idle_send_queued_messages_CB(gpointer data)
{
	AlpConnection* connection = (AlpConnection *)data;
	//	g_print("IDLE CB IDLE CB IDLE CB IDLE CB IDLE CB\n");
	if (connection != NULL)
		prv_send_queued_messages(connection);

	return FALSE;  // returning false removes this GSource
}

alp_status_t
prv_send_queued_messages(AlpConnection* connection)
{
	guint mm;
	AlpMessage* msg;

	if (connection == NULL)
		return ALP_STATUS_IPC_ERROR;

	if ((connection->receiveMessageQueue == NULL) ||
	    (connection->receiveMessageQueue->len == 0)) 
		return ALP_STATUS_OK;

	pthread_mutex_lock(&connection->receiveMessageQueueMutex);

	for (mm = 0; mm<connection->receiveMessageQueue->len; mm++) {
		msg = g_array_index(connection->receiveMessageQueue, AlpMessage *, mm);

		if (msg == NULL)
			continue;
		
		//		g_print("SENDING DEFERRED MESSAGE %d\n", msg->header.msgID);
		
		if (connection->amServer) 
			prv_handle_message_for_server(NULL, connection, msg);
		else
			prv_handle_message_for_client(connection, msg);
	}

	g_array_remove_range(connection->receiveMessageQueue,
					 0, 
					 connection->receiveMessageQueue->len);

	pthread_mutex_unlock(&connection->receiveMessageQueueMutex);

	return ALP_STATUS_OK;
}

alp_status_t
prv_check_for_queued_message(AlpConnection* connection,
					    gint msgID,
					    AlpMessage** message)
{
	guint mm;
	AlpMessage* msg;

	if (connection == NULL)
		return ALP_STATUS_IPC_ERROR;

	if ((connection->receiveMessageQueue == NULL) ||
	    (connection->receiveMessageQueue->len == 0)) 
		return ALP_STATUS_OK;

	pthread_mutex_lock(&connection->receiveMessageQueueMutex);

	for (mm = 0; mm<connection->receiveMessageQueue->len; mm++) {
		msg = g_array_index(connection->receiveMessageQueue, AlpMessage *, mm);

		if (msg == NULL)
			continue;

		if ((msgID == ALP_IPC_RECEIVE_ANY) ||
		    (msgID == msg->header.msgID)) {
			*message = msg;
			g_array_remove_index(connection->receiveMessageQueue,
							 mm);
			break;
		}
	}

	pthread_mutex_unlock(&connection->receiveMessageQueueMutex);

	return ALP_STATUS_OK;
}

alp_status_t
prv_queue_sent_message(AlpConnection* connection, AlpMessage* msg)
{
	AlpMessage* clonedMessage = NULL;

	if (connection == NULL)
		return ALP_STATUS_IPC_ERROR;

	if (msg == NULL)
		return ALP_STATUS_OK;

	if (connection->sentMessageQueue == NULL) {
		if (pthread_mutex_init(&connection->sentMessageQueueMutex, NULL) != 0)
			return ALP_STATUS_IPC_ERROR;

		pthread_mutex_lock(&connection->sentMessageQueueMutex);

		connection->sentMessageQueue = g_array_new(FALSE, TRUE, sizeof(AlpMessage*));
	} else {
		pthread_mutex_lock(&connection->sentMessageQueueMutex);
	}

	// need to clone the message, because the app will typically
	// destroy or reuse the original
	clonedMessage = prv_message_clone(msg);

	g_array_append_val(connection->sentMessageQueue, clonedMessage);

	// install the G_IO_OUT callback if this is the first message queued
	if (connection->sentMessageQueue->len == 1)
		prv_install_g_io_out_callback(connection);

	connection->okToWrite = FALSE;

	pthread_mutex_unlock(&connection->sentMessageQueueMutex);

	return ALP_STATUS_OK;
}


//
// FUNCTIONS FOR DOING SYNCHRONOUS MESSAGE RECEIVE
//

gboolean
prv_receive_func_timeout_CB(gpointer data)
{
	SynchReceiveData* receiveData = NULL;

	if (data == NULL)
		return FALSE;

	receiveData = (SynchReceiveData*)data;

	AlpConnection* connection = receiveData->connection;

	if (connection == NULL)
		return FALSE;

	if (connection->receiveMainLoop == NULL)
		return FALSE;

	receiveData->status = ALP_STATUS_IPC_TIMEOUT;

	// destroy the IOWatch source
	if (connection->receiveIOWatchSrc != NULL) {
		g_source_destroy(connection->receiveIOWatchSrc);
		g_source_unref(connection->receiveIOWatchSrc);
		connection->receiveIOWatchSrc = NULL;
	}

	g_main_loop_quit(connection->receiveMainLoop);

	return FALSE;
}

gboolean
prv_receive_func_input_hangup_CB(GIOChannel* source, GIOCondition condition, gpointer data)
{
	AlpMessage* msg = NULL;

	SynchReceiveData* receiveData = NULL;

	if (data == NULL)
		return FALSE;

	GIOStatus gioStatus;

	bool retVal = FALSE;

	receiveData = (SynchReceiveData*)data;

	if (receiveData == NULL) {
		TRACE_IPC("Synch Receive Error\n", NULL);
		return FALSE;
	}

	receiveData->status = ALP_STATUS_OK;

	GIOChannel* channel = NULL;
	AlpConnection* connection = receiveData->connection;

	if (connection == NULL)
		return FALSE;

	// using local sockets
	channel = connection->gioChannel;

	gboolean receivedTheRightMsg = FALSE;

	// First send if there's anything to send
	// this can happen if a send with response could
	// not send all of its message in one chunk
	//
	// The connection->okToWrite flag indicates that
	// there is queued data to send
	//
	if ((condition & G_IO_OUT) && (!connection->okToWrite)) {
		if ((connection->sentMessageQueue) &&
		    (connection->sentMessageQueue->len > 0)) {
			prv_received_output_hangup_CB(source, condition, connection);	
		}
	}

	// first handle any data on the pipe
	if (condition & G_IO_IN) { 
		while (1) {
			// get the message header
			gioStatus = prv_get_msg_header(connection, &msg, source); 
			/*
			  if (gioStatus == G_IO_STATUS_NORMAL)
			  g_print("GIO NORMAL\n");
			  else if (gioStatus == G_IO_STATUS_ERROR)
			  g_print("GIO ERROR\n");
			  else if (gioStatus == G_IO_STATUS_AGAIN)
			  g_print("GIO AGAIN\n");
			  else if (gioStatus == G_IO_STATUS_EOF)
			  g_print("GIO EOF\n");
			*/
			if (gioStatus != G_IO_STATUS_NORMAL) {
				if (msg != NULL) {
					alp_ipc_message_destroy(msg);
				}
				break;
			}

			// It's OK for G_IO_STATUS_NORMAL to be returned and
			// for msg to be NULL.  In this case, the msg was
			// only partially received

               if (msg != NULL) {
				// if I'm looking for a response 
				if (receiveData->sentToken > 0) {
					if (receiveData->sentToken == msg->header.responseToken) {
						//						g_print("GOT DESIRED MESSAGE\n");
						// response to a send with response
						
						*receiveData->message = msg;
						receivedTheRightMsg = TRUE;
						retVal = FALSE;
					} else if (msg->header.responseToken == 0) {
						//						g_print("QUEUING MESSAGE -- NOT DESIRED MESSAGE\n");
						if (prv_queue_received_message(connection, msg) == 
						    ALP_STATUS_OK) {

							if (receivedTheRightMsg)  
								retVal = FALSE;
							else 
								retVal = TRUE; // don't cancel the IOWatch callback
						}
					} else {
						// if we received a response token and it
						// doesn't match the token we're waiting for
						// then that's probably an error

						//						g_print("QUEUING MESSAGE -- NOT A RESPONSE\n");

						receiveData->status = ALP_STATUS_IPC_INVALID_RESPONSE;
						// in this case, toss the message, but
						// don't cancel the IOWatch callback
						alp_ipc_message_destroy(msg);
						retVal = TRUE;
					}
				} else if (receiveData->sentToken == 0) {
					// not looking for a response
					if ((receiveData->messageID == ALP_IPC_RECEIVE_ANY) ||
					    (receiveData->messageID == msg->header.msgID)) {
						if (receivedTheRightMsg == FALSE) {
							// return first message
							*receiveData->message = msg;
							receivedTheRightMsg = TRUE;
						} else {
							// queue any others
							prv_queue_received_message(connection, msg);
						}
						retVal = FALSE;
					} else {
						//						g_print("QUEUING MESSAGE -- NOT DESIRED MESSAGE\n");
						if (prv_queue_received_message(connection, msg) == 
						    ALP_STATUS_OK) {
							retVal = TRUE; // don't cancel the IOWatch callback
						}
					}
				} else {
					// msg->header.responseToken should never be 
					// less than zero
					TRACE_IPC("Response token should never be < 0\n",
							NULL);

					retVal = FALSE;
				}
			}
		} 
	}
	
	// then handle any hangups
	if (condition & G_IO_HUP) {
		receiveData->status = ALP_STATUS_IPC_BROKEN_CONNECTION;
	}

	// It's OK to have a status of ALP_STATUS_OK and
	// no message received, but in this case we need
	// to return TRUE so we don't break out of the
	// receive main loop
	// 
	// One way this can happen is if there's a large
	// message that's only been partially received.
	//
	// Another way this can happen is if this is a
	// send with response's receive and the send
	// part could not send the entire message in
	// one chunk
	//
	if ((receiveData->status == ALP_STATUS_OK) && 
	    (*receiveData->message == NULL)) {
		retVal = TRUE;
	}

	// destroy the timeout source
	if (retVal == FALSE) {
		if (connection->receiveTimeoutSrc != NULL) {
			g_source_destroy(connection->receiveTimeoutSrc);
			g_source_unref(connection->receiveTimeoutSrc);
			connection->receiveTimeoutSrc = NULL;
		}
//		TRACE_IPC("BREAKING FROM RECEIVE LOOP\n", NULL);
		g_main_loop_quit(connection->receiveMainLoop);
	}
	
	return retVal;
}


alp_status_t
prv_receive_func(SynchReceiveData *receiveData)
{
	GMainContext* receiveContext = NULL;
	GMainLoop*	receiveLoop = NULL;

	if (receiveData == NULL)
		return ALP_STATUS_IPC_ERROR;

	if (receiveData->connection == NULL)
		return ALP_STATUS_IPC_ERROR;

	AlpConnection* connection = receiveData->connection;

	// create a new g_main_context and a new g_main_loop
	if (connection->receiveContext == NULL) {
		connection->receiveContext = g_main_context_new();
		if (connection->receiveContext == NULL)
			return ALP_STATUS_IPC_ERROR;
	}

	if (connection->receiveMainLoop == NULL) {
		connection->receiveMainLoop = g_main_loop_new(connection->receiveContext, 
											 TRUE);
	}

	if (connection->receiveMainLoop != NULL) {

		// add a timeout handler
		if (receiveData->timeout > 0) {
			connection->receiveTimeoutSrc = 
				g_timeout_source_new(receiveData->timeout);

			g_source_set_callback(connection->receiveTimeoutSrc, 
							  prv_receive_func_timeout_CB, 
							  (gpointer)receiveData,
							  NULL);
			g_source_attach(connection->receiveTimeoutSrc, 
						 connection->receiveContext);
		}

		// remove existing input/hangup handler
		gboolean replaceInSource = FALSE;
		if (connection->inSource != 0) { 
			replaceInSource = TRUE;
			g_source_remove(connection->inSource);

			connection->inSource = 0;
		}

		// add a receive message handler
		connection->receiveIOWatchSrc = 
			g_io_create_watch(connection->gioChannel,
						   G_IO_IN | G_IO_HUP); 

		g_source_set_callback (connection->receiveIOWatchSrc,
						   (GSourceFunc)prv_receive_func_input_hangup_CB,
						   (gpointer)receiveData,
						   NULL);

		g_source_attach(connection->receiveIOWatchSrc, 
					 connection->receiveContext);

		g_main_loop_run(connection->receiveMainLoop);

		if (connection->receiveTimeoutSrc != NULL) {
			g_source_destroy(connection->receiveTimeoutSrc);
			g_source_unref(connection->receiveTimeoutSrc);
			connection->receiveTimeoutSrc = NULL;
		} 

		if (connection->receiveIOWatchSrc != NULL) {
			g_source_destroy(connection->receiveIOWatchSrc);
			g_source_unref(connection->receiveIOWatchSrc);
			connection->receiveIOWatchSrc = NULL;
		}

		if (replaceInSource) {
			// if connection is used for synch comm only, only watch for
			// a hangup
			if ((connection->hint & 
				ALP_IPC_CONNECTION_USAGE_HINT_SYNCH_ONLY) == 0) {
				connection->inSource = 
					g_io_add_watch(connection->gioChannel,
								G_IO_IN | G_IO_HUP, 
								prv_client_received_input_hangup_CB, 
								(gpointer)connection);
			} else {
				connection->inSource = 
					g_io_add_watch(connection->gioChannel,
								G_IO_HUP, 
								prv_client_received_input_hangup_CB, 
								(gpointer)connection);
			}
		}
	}

	return receiveData->status;
}

alp_status_t
prv_wait_for_synch_receive(AlpConnection* connection, 
						    gint messageID,
						    guint timeout, 
						    AlpMessage** message)
{
	if (connection == NULL)
		return ALP_STATUS_IPC_NOT_CONNECTED;

	if ((connection->receiveMessageQueue) &&
	    (connection->receiveMessageQueue->len > 0)) {
		// check if there are any deferred messages of the
		// correct type before blocking 
		prv_check_for_queued_message(connection, messageID, message);

		if (*message != NULL)
			return ALP_STATUS_OK;
	}

	SynchReceiveData receiveData;
	alp_status_t status;

	receiveData.connection = connection; 
	receiveData.messageID = messageID;
	receiveData.sentToken = 0;
	receiveData.timeout = timeout;
	receiveData.message = message;

	*message = NULL;

	status = prv_receive_func(&receiveData);

	if ((status != ALP_STATUS_OK) &&
	    (status != ALP_STATUS_IPC_TIMEOUT) &&
	    (status != ALP_STATUS_IPC_BROKEN_CONNECTION) &&
	    (status != ALP_STATUS_IPC_INVALID_RESPONSE) &&
	    (status != ALP_STATUS_IPC_ERROR)) {
		TRACE_IPC("Unexpected status for synch receive\n",
				NULL);
	}

	return status;
}

//
// FUNCTIONS FOR DOING SEND WITH RESPONSE
//


alp_status_t
prv_send_with_response(AlpConnection* connection,
					AlpMessage*    messageToSend,
					guint          timeout,
					AlpMessage**   responseMessage)
{
	alp_status_t  status = ALP_STATUS_OK;

	GMainContext* receiveContext = NULL;
	GMainLoop*	receiveLoop = NULL;
  
	SynchReceiveData receiveData;

	if (connection == NULL)
		return ALP_STATUS_IPC_NOT_CONNECTED;

	if (messageToSend == NULL)
		return ALP_STATUS_IPC_ERROR;

	if (messageToSend == NULL)
		return ALP_STATUS_IPC_ERROR;

	// insert the response token into the message being sent
	messageToSend->header.sentToken = (connection->pid << 16) +
		connection->responseTokenCounter;
	++connection->responseTokenCounter;
	if (connection->responseTokenCounter == 65535) {
		connection->responseTokenCounter = 1;
	}
	
	messageToSend->header.responseToken = 0;

	prv_repack_header(messageToSend);

	// now mimic the synch receive

	receiveData.connection = connection; 
	receiveData.messageID = ALP_IPC_RECEIVE_ANY;
	receiveData.sentToken = messageToSend->header.sentToken;
	receiveData.timeout = timeout;
	receiveData.message = responseMessage;

	// create a new g_main_context and a new g_main_loop
	if (connection->receiveContext == NULL) {
		connection->receiveContext = g_main_context_new();
		if (connection->receiveContext == NULL)
			return ALP_STATUS_IPC_ERROR;
  	}

	if (connection->receiveMainLoop == NULL) {
		connection->receiveMainLoop = g_main_loop_new(connection->receiveContext, 
											 TRUE);
	}

	if (connection->receiveMainLoop != NULL) {

		// add a timeout handler
		if (receiveData.timeout > 0) {
			connection->receiveTimeoutSrc = 
				g_timeout_source_new(receiveData.timeout);

			g_source_set_callback(connection->receiveTimeoutSrc, 
							  prv_receive_func_timeout_CB, 
							  (gpointer)&receiveData,
							  NULL);
			g_source_attach(connection->receiveTimeoutSrc, 
						 connection->receiveContext);
		}

		// remove existing input/hangup handler
		gboolean replaceInSource = FALSE;
		if (connection->inSource != 0) { 
			replaceInSource = TRUE;
			g_source_remove(connection->inSource);

			connection->inSource = 0;
		}

		// WAITED TO SEND THE MESSAGE AFTER ANY EXISTING G_IO_IN SOURCES
		// WERE REMOVED TO MAKE SURE ONE OF THEM DOESN'T INADVERTANTLY
		// GRAB THE RESPONSE DUE TO A RACE CONDITION
		status = alp_ipc_connection_send(connection, messageToSend);

		if (status != ALP_STATUS_OK)
			return status;
		
		if (connection->okToWrite == 0) {
			// this means the message was only partially sent, so
			// we need to watch G_IO_OUT to find out when
			// the socket is available
			connection->receiveIOWatchSrc =
				g_io_create_watch(connection->gioChannel,
							   G_IO_OUT | G_IO_IN | G_IO_HUP); 

		} else {
			// add a receive message handler
			connection->receiveIOWatchSrc =
				g_io_create_watch(connection->gioChannel,
							   G_IO_IN | G_IO_HUP); 
		}

		g_source_set_callback (connection->receiveIOWatchSrc,
						   (GSourceFunc)prv_receive_func_input_hangup_CB,
						   (gpointer)&receiveData,
						   NULL);

		g_source_attach(connection->receiveIOWatchSrc, 
					 connection->receiveContext);

		g_main_loop_run(connection->receiveMainLoop);

		if (connection->receiveTimeoutSrc != NULL) {
			g_source_destroy(connection->receiveTimeoutSrc);
			g_source_unref(connection->receiveTimeoutSrc);
			connection->receiveTimeoutSrc = NULL;
		} 

		if (connection->receiveIOWatchSrc != NULL) {
			g_source_destroy(connection->receiveIOWatchSrc);
			g_source_unref(connection->receiveIOWatchSrc);
			connection->receiveIOWatchSrc = NULL;
		}

		if (replaceInSource) {
			// if connection is used for synch comm only, only watch for
			// a hangup
			if ((connection->hint & 
				ALP_IPC_CONNECTION_USAGE_HINT_SYNCH_ONLY) == 0) {
				connection->inSource = 
					g_io_add_watch(connection->gioChannel,
								G_IO_IN | G_IO_HUP, 
								prv_client_received_input_hangup_CB, 
								(gpointer)connection);
			} else {
				connection->inSource = 
					g_io_add_watch(connection->gioChannel,
								G_IO_HUP, 
								prv_client_received_input_hangup_CB, 
								(gpointer)connection);
			}
		}
	}

	status = receiveData.status;

	if ((status != ALP_STATUS_OK) &&
	    (status != ALP_STATUS_IPC_TIMEOUT) &&
	    (status != ALP_STATUS_IPC_BROKEN_CONNECTION) &&
	    (status != ALP_STATUS_IPC_INVALID_RESPONSE) &&
	    (status != ALP_STATUS_IPC_ERROR)) {
		TRACE_IPC("Unexpected status for send with response\n",
				NULL);
	}

	return status;
}

static inline gboolean 
prv_status_error_pack(alp_status_t status, AlpMessage* message)
{
	if (status != ALP_STATUS_OK) { 
		alp_ipc_message_destroy(message); 
		return FALSE; 
	} 

	return TRUE;
} 

static inline gboolean
prv_status_error_unpack(alp_status_t status)
{
	if (status != ALP_STATUS_OK) 
		return FALSE; 

	return TRUE;
}


void
prv_server_socket_name(const gchar* channelName, GString* socketFileName)
{
	if (socketFileName != NULL)
		g_string_printf(socketFileName, "%s%s%s", 
					getenv("ALP_RUN_TOP") ?: "",
					 DEFAULT_CHANNEL_DIRECTORY,
					 channelName);
	else
		TRACE_IPC("No socket name specified\n", NULL);
}

// makes the directories in the path
void
prv_make_dirs(GString* path)
{
	// this could be made more efficient, but this is a simple way to 
	// do it and it won't be used often

	gchar** splitStr = g_strsplit(path->str, "/", 16);
	//	gchar*  currChar = NULL;
	gint	currIndex = 0;
	gint	numDirs = 0;
	gint	i;

	GString* dirName = g_string_new_len("", path->len);
	g_string_assign(dirName, "/");

	if (splitStr == NULL)
		return;

	// count the number of dirs

	while (splitStr[currIndex] != NULL) {   
		currIndex++;
		numDirs++;
	}
	numDirs -= 1;

	for (i=0; i<numDirs; i++) {
		g_string_append(dirName, splitStr[i]);

		mkdir(dirName->str, 0755);

		if (i > 0)
			g_string_append(dirName, "/");

	}

	g_string_free(dirName, TRUE);

	g_strfreev(splitStr);
}

// deletes any old socket for this channel that may have been leftover
// from a non-standard shutdown
void
prv_clean_up_old_socket(const gchar* channelName, const GString* socketName)
{
	char dirname[PATH_MAX];
	
	sprintf(dirname, "%s%s", getenv("ALP_RUN_TOP") ?: "", DEFAULT_CHANNEL_DIRECTORY);
	
	DIR* dd = opendir(dirname);
	struct dirent* dirEntry = NULL;

	if (channelName == NULL) {
		TRACE_IPC("No channel name specified\n", NULL);
		return;
	}

	if (dd == NULL) {
		TRACE_IPC("Channel directory doesn't exist\n", NULL);
		return;
	}

	dirEntry = readdir(dd);
  
	while (dirEntry != NULL) {
		if (strcmp(channelName, dirEntry->d_name) == 0) {
			unlink(socketName->str);

			break;
		}

		dirEntry = readdir(dd);
	}

	closedir(dd);
}

gboolean
prv_set_up_client_connection(AlpConnection* newConnection)
{
	gint ret_val;

	// add a input and hangup watch on the pipe from the server
	// if connection is used for synch comm only, only watch for
	// a hangup
	if ((newConnection->hint & ALP_IPC_CONNECTION_USAGE_HINT_SYNCH_ONLY) == 0) {
		newConnection->inSource = g_io_add_watch(newConnection->gioChannel,
										 G_IO_IN | G_IO_HUP, 
										 prv_client_received_input_hangup_CB, 
										 (gpointer)newConnection);
	} else { 
		newConnection->inSource = g_io_add_watch(newConnection->gioChannel,
										 G_IO_HUP, 
										 prv_client_received_input_hangup_CB, 
										 (gpointer)newConnection);
	}
	
	struct ucred credentials;

	socklen_t cred_len = sizeof(credentials);
	ret_val = getsockopt(newConnection->socketFD, SOL_SOCKET, SO_PEERCRED, &credentials, &cred_len);

	if (ret_val == -1) {
		TRACE_IPC("Error retrieving peer credentials %s\n", strerror(errno));
	} else {
		newConnection->pid = credentials.pid;

//		g_print("PID IS %d\n", 
//			   newConnection->pid );

		newConnection->status = ALP_IPC_CONNECTION_READY;

		// invoke any registered connection callback
		if (newConnection->connectCB) {
			newConnection->connectCB->callback(newConnection, 
										newConnection->connectCB->cbData);
		}
	}

	return TRUE;
}

gboolean
prv_wait_for_connect_CB(GIOChannel* source, 
				    GIOCondition condition, gpointer data)
{
	AlpConnection* connection = (AlpConnection *)data;
	gint ret_val;
	gint soErrorVal;
	gint soErrorValLen = sizeof(gint);

	if (connection == NULL) {
		TRACE_IPC("Error connecting to channel\n", NULL);
		return FALSE;     // always return FALSE to cancel the callback
	}

	ret_val = getsockopt(connection->socketFD, SOL_SOCKET, SO_ERROR,
					 (void *)&soErrorVal, (socklen_t *)&soErrorValLen);

	if ((ret_val == 0) && (soErrorVal == 0)) {
		prv_set_up_client_connection(connection);
	} else {
		TRACE_IPC("Error connecting to channel\n", NULL);
	}

	return FALSE;       // always return FALSE to cancel the callback
}

gboolean
prv_synch_connect_timeout_CB(gpointer data )
{
	AlpConnection* connection = (AlpConnection *)data;

	if (connection == NULL)
		return FALSE;

	if (connection->receiveMainLoop == NULL)
		return FALSE;

	// destroy the IOWatch source
	if (connection->receiveIOWatchSrc != NULL) {
		g_source_destroy(connection->receiveIOWatchSrc);
		g_source_unref(connection->receiveIOWatchSrc);
		connection->receiveIOWatchSrc = NULL;
	}

	g_main_loop_quit(connection->receiveMainLoop);

	return FALSE;
}

gboolean
prv_synch_connect_ready_CB(gpointer data )
{
	AlpConnection* connection = (AlpConnection *)data;
	gint ret_val;
	gint soErrorVal;
	gint soErrorValLen = sizeof(gint);

	if (connection == NULL) {
		TRACE_IPC("Error connecting to channel\n", NULL);
		return FALSE;     // always return FALSE to cancel the callback
	}

	ret_val = getsockopt(connection->socketFD, SOL_SOCKET, SO_ERROR,
					 (void *)&soErrorVal, (socklen_t *)&soErrorValLen);

	if ((ret_val == 0) && (soErrorVal == 0)) {
		prv_set_up_client_connection(connection);
	} else {
		TRACE_IPC("Error connecting to channel\n", NULL);
	}

	// destroy the timeout source
	if (connection->receiveTimeoutSrc != NULL) {
		g_source_destroy(connection->receiveTimeoutSrc);
		g_source_unref(connection->receiveTimeoutSrc);
		connection->receiveTimeoutSrc = NULL;
	}

	g_main_loop_quit(connection->receiveMainLoop);

	return FALSE;
}

alp_status_t
prv_do_synch_connect_loop(AlpConnection* connection, gint timeout) 
{
	alp_status_t status = ALP_STATUS_OK;

	// create a new g_main_context and a new g_main_loop
	if (connection->receiveContext == NULL) {
		connection->receiveContext = g_main_context_new();
		if (connection->receiveContext == NULL)
			return ALP_STATUS_IPC_ERROR;
  	}

	if (connection->receiveMainLoop == NULL) {
		connection->receiveMainLoop = g_main_loop_new(connection->receiveContext, 
											 TRUE);
	}

	if (connection->receiveMainLoop != NULL) {
		// add a timeout handler
		if (timeout > 0) {
			connection->receiveTimeoutSrc = 
				g_timeout_source_new(timeout);

			g_source_set_callback(connection->receiveTimeoutSrc, 
							  prv_synch_connect_timeout_CB, 
							  (gpointer)connection,
							  NULL);
			g_source_attach(connection->receiveTimeoutSrc, 
						 connection->receiveContext);
		}

		// remove existing input/hangup handler
		gboolean replaceInSource = FALSE;
		if (connection->inSource != 0) { 
			replaceInSource = TRUE;
			g_source_remove(connection->inSource);

			connection->inSource = 0;
		}

		// add a receive message handler
		connection->receiveIOWatchSrc = 
			g_io_create_watch(connection->gioChannel,
						   G_IO_OUT);

		g_source_set_callback (connection->receiveIOWatchSrc,
						   prv_synch_connect_ready_CB,
						   (gpointer)connection,
						   NULL);

		g_source_attach(connection->receiveIOWatchSrc, 
					 connection->receiveContext);

		g_main_loop_run(connection->receiveMainLoop);

		if (connection->receiveTimeoutSrc != NULL) {
			status = ALP_STATUS_IPC_TIMEOUT;

			g_source_destroy(connection->receiveTimeoutSrc);
			g_source_unref(connection->receiveTimeoutSrc);
			connection->receiveTimeoutSrc = NULL;
		} 

		if (connection->receiveIOWatchSrc != NULL) {
			g_source_destroy(connection->receiveIOWatchSrc);
			g_source_unref(connection->receiveIOWatchSrc);
			connection->receiveIOWatchSrc = NULL;
		}

		if (replaceInSource) {
			// if connection is used for synch comm only, only watch for
			// a hangup
			if ((connection->hint & 
				ALP_IPC_CONNECTION_USAGE_HINT_SYNCH_ONLY) == 0) {
				connection->inSource = 
					g_io_add_watch(connection->gioChannel,
								G_IO_IN | G_IO_HUP, 
								prv_client_received_input_hangup_CB, 
								(gpointer)connection);
			} else {
				connection->inSource = 
					g_io_add_watch(connection->gioChannel,
								G_IO_HUP, 
								prv_client_received_input_hangup_CB, 
								(gpointer)connection);
			}
		}
	}

	return status;
}

gint
prv_try_to_connect(AlpConnection* connection, 
			    const gchar* channelName,
			    gboolean isRetry,
			    gint timeout,
			    alp_status_t* status) 
{
	gint ret_val;
	struct sockaddr_un addr;
	addr.sun_family = AF_UNIX;

	GString* socketName = g_string_new("");
	prv_server_socket_name(channelName, socketName);

	if (socketName->len >= sizeof(addr.sun_path)) {
		TRACE_IPC("Socket name %s is too long (max is %d)\n",
				socketName->str, sizeof(addr.sun_path));
                g_string_free(socketName, TRUE);
		return -1;
	}
	strcpy(addr.sun_path, socketName->str);

	g_string_free(socketName, TRUE);

	// call connect
	ret_val = connect(connection->socketFD, 			
				   (const struct sockaddr *)&addr, 
				   sizeof(addr));
//	TRACE_IPC("NEW SYNC CONNECTION TO CHANNEL %s IS %d\n",
//			socketName->str, connection->socketFD);
	if (ret_val == 0) {
		// connection happened right away
		prv_set_up_client_connection(connection);
	} else if (ret_val == -1) {
		switch (errno) {
		case EINPROGRESS:
			if (timeout == -1) {
				// set up a G_IO_OUT callback to tell me when the socket
				// can be written
				g_io_add_watch(connection->gioChannel, 
							G_IO_OUT, 
							prv_wait_for_connect_CB,
							(gpointer)connection);
			} else {
				// doing a sync connect, so need to
				// setup a new main loop with a G_IO_OUT watch
				// and a Timeout

				if (status == NULL) {
					TRACE_IPC("prv_try_to_connect called with an invalid parameter",
							NULL);
				}

				*status = prv_do_synch_connect_loop(connection, 
											 timeout);
			}

			break;

		case EAGAIN:
			// retry if you get an egain
			if (!isRetry)
				ret_val = prv_try_to_connect(connection, 
									    channelName, 
									    TRUE,
									    timeout,
									    status);

			break;
			
		default:
			break;
		}
	}

	return ret_val;
}

