/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/*!
 * @file alp_ipc_internal.h
 *
 * @brief File containing definitions of functions and data structures 
 * internal to the IPC library.
 *
 *****************************************************************************/

#ifndef ALP_IPC_INTERNAL_H_
#define ALP_IPC_INTERNAL_H_ 1

#include <pthread.h>
#include <signal.h>

#include <hiker/ipc.h>

#include <hiker/config_defines.h>

/* Uncomment the next line to disable traces at COMPILE TIME */
// #define ALP_TRACE ALP_TRACE_OFF

#include <hiker/traces.h>

// simple trace macro
#define TRACE_IPC(str, ...) ALP_TM(TL(ALP_CLASS_IPC, "%s(): " str, __PRETTY_FUNCTION__, __VA_ARGS__))

#define ALP_IPC_MSG_VERSION 1

typedef struct _ConnectCBInfo ConnectCBInfo;
struct _ConnectCBInfo {
	AlpChannelConnectCB callback;    
	gpointer           cbData;
};

typedef struct _DisconnectCBInfo DisconnectCBInfo;
struct _DisconnectCBInfo {
	AlpChannelDisconnectCB callback;    
	gpointer           cbData;
};

typedef struct _ReceiveCBInfo ReceiveCBInfo;
struct _ReceiveCBInfo {
	AlpMessageReceiveCB callback;    
	gpointer           cbData;
};

typedef enum _AlpConnectionStatus AlpConnectionStatus;
enum _AlpConnectionStatus {
	ALP_IPC_CONNECTION_CREATED,
	ALP_IPC_CONNECTION_READY
};

typedef enum _AlpIPCInternalMessageID AlpIPCInternalMessageID;
// NOTE: THE PUBLIC HEADER FILE DEFINES ALP_IPC_RECEIVE_ANY TO BE -1
// THE INTERNAL MESSAGE IDs NEED TO NOT USE THAT VALUE
enum _AlpIPCInternalMessageID {
	ALP_IPC_MSG_UNKNOWN = -2,
};

typedef struct _AlpMessageHeader AlpMessageHeader;
struct _AlpMessageHeader {
	gint      msgVersion;           // 1 for now

	gint      msgID;
	gint      sentToken;
	gint      responseToken;
	GString*  msgFormat;
};

//
// Opaque structure definitions (typedefs are include/alp/ipc.h)
//

struct _AlpChannel {

	// use local sockets
     GString*   socketName;

     int        socketFD;

	GIOChannel *gioChannel; 

	ConnectCBInfo*    connectCB;
	DisconnectCBInfo* disconnectCB;
 
	GHashTable *connectionTable;

	guint     inSource;

	gpointer  customData;              // pointer to a user-specified block of
	                                   // data that is to be associated with this AlpChannel
};

struct _AlpConnection {

	// using local sockets

	int       socketFD;

	GIOChannel* gioChannel;

	AlpConnectionStatus status;
  
	gboolean  amServer;
	AlpConnection* connection;     // only set if amServer is TRUE
  
	guint     inSource;
	guint     outSource;

	gboolean  okToWrite;

	ConnectCBInfo*    connectCB;
	DisconnectCBInfo* disconnectCB;
	ReceiveCBInfo*    receiveCB;

	// for synchronous message receipt
	GMainContext*  receiveContext;
	GMainLoop*     receiveMainLoop;

	GSource*       receiveTimeoutSrc;
	GSource*       receiveIOWatchSrc;

	guint          responseTokenCounter;

	GArray*  receiveMessageQueue;
	pthread_mutex_t receiveMessageQueueMutex;

	gint           pid;

	AlpMessage*    partialMessage;
	guint          partialMessageRemainingSize;
	gchar*         partialMessageBufPtr;

	GArray*         sentMessageQueue;
	pthread_mutex_t sentMessageQueueMutex;

	pthread_mutex_t writeMutex;

	gpointer  customData;              // pointer to a user-specified block of
	                                   // data that is to be associated with this AlpConnection

	AlpConnectionUsageHint hint;
};

// messages have three parts, an integer containing the
// size of the buffer that was sent, followed by 
// a GNet-packed message header, followed by the
// GNet-packed message data
struct _AlpMessage {
	AlpMessageHeader header;

	gchar*   bufPtr;
	guint    bufSize;  // total buffer size including
	// the header, message and prepended buffer size

	gchar*   headerPtr; // pointer to the character in the buffer
	// where the header data starts
	guint    headerSize;
  
	gchar*   msgPtr;    // pointer to the character in the buffer
	// where the message data starts
	guint    msgSize;
 
	guint    remainingSize;
	guint    allocBufSize;   // allocated size can be different
	// from the used bufSize

	guint    formatIndex;    // used for unpacking
	gchar*   currMsgPtr;     // used for packing and unpacking

	guint numBytesSent;
};

// 
// Internal structure definitions
//
typedef struct _SynchReceiveData SynchReceiveData;
struct _SynchReceiveData {
	AlpConnection* connection;
	gint           messageID;
	gint           sentToken;
	guint          timeout;
	AlpMessage**   message;

	alp_status_t   status;
};

#define DEFAULT_CHANNEL_DIRECTORY "/tmp/"

#define DEFAULT_FORMAT_STRING_LENGTH 32
#define DEFAULT_BUFFER_SIZE          1024

//
// Internal functions
//

AlpChannel* 
prv_channel_create();

void        
prv_channel_destroy(AlpChannel* channel);

AlpConnection* 
prv_connection_create(AlpConnectionUsageHint hint);

void           
prv_connection_destroy(AlpConnection* connection);

AlpMessage*    
prv_message_create();

void           
prv_message_destroy(AlpMessage* message);

GIOStatus
prv_get_msg_header(AlpConnection* connection, AlpMessage** msg, GIOChannel* source);

void
prv_invoke_messsage_callback(AlpConnection* connection, AlpMessage* message);

gboolean
prv_server_accept_CB(GIOChannel* source, GIOCondition condition, gpointer data);
gboolean
prv_server_received_input_hangup_CB(GIOChannel* source, GIOCondition condition, gpointer data);

gboolean
prv_client_received_input_hangup_CB(GIOChannel* source, GIOCondition condition, gpointer data);

gboolean
prv_received_output_hangup_CB(GIOChannel* source, GIOCondition condition, gpointer data);

gboolean 
prv_connection_table_remove_CB(gpointer key, gpointer value, gpointer userData);

gboolean
prv_finish_packing(AlpMessage* message);

gboolean
prv_repack_header(AlpMessage* message);

gboolean
prv_realloc_message_buffer(AlpMessage* message, guint numBytesNeeded);

//
// Functions for doing synchronous message receive in a separate thread
//
gboolean
prv_receive_func_timeout_CB(gpointer data);

gboolean
prv_receive_func_input_hangup_CB(GIOChannel* source, 
						   GIOCondition condition, 
						   gpointer data);

alp_status_t
prv_receive_func(SynchReceiveData* arg);

alp_status_t
prv_send_with_response(AlpConnection* connection,
					AlpMessage*    messageToSend,
					guint          timeout,
					AlpMessage**   responseMessage);

alp_status_t
prv_wait_for_synch_receive(AlpConnection* connection, 
						    gint messageID,
						    guint timeout, 
						    AlpMessage** message);

alp_status_t
prv_queue_received_message(AlpConnection* connection, AlpMessage* message);

gboolean 
prv_idle_send_queued_messages_CB(gpointer data);

alp_status_t
prv_send_queued_messages(AlpConnection* connection);

alp_status_t
prv_check_for_queued_message(AlpConnection* connection,
					    gint msgID,
					    AlpMessage** message);

alp_status_t
prv_queue_sent_message(AlpConnection* connection, AlpMessage* message);

void
prv_install_SIGPIPE_handler(struct sigaction* oldHandler);

void
prv_restore_SIGPIPE_handler(struct sigaction* oldHandler);

void
prv_install_g_io_out_callback(AlpConnection* connection);

void
prv_make_dirs(GString* path);

void
prv_clean_up_old_socket(const gchar* channelName, const GString* socketName);

void
prv_server_socket_name(const gchar* channelName, GString* socketFileName);

gboolean
prv_set_up_client_connection(AlpConnection* newConnection);

gboolean
prv_wait_for_connect_CB(GIOChannel* source, 
				    GIOCondition condition, gpointer data);

gint
prv_try_to_connect(AlpConnection* connection, 
			    const gchar* channelName,
			    gboolean isRetry,
			    gint timeout,
			    alp_status_t* status);

#endif // #ifndef ALP_IPC_INTERNAL_H_

