/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
// Includes ------------------------------------------------------------------
#include <hiker/prv/IpcMessage.h>
#include <hiker/prv/support/String.h>

namespace alp {

namespace ipc {

// Typedefs ------------------------------------------------------------------
template<class T>
class shared_ptr
{
	public:
		typedef void destroyer(T*);

		shared_ptr() : fShareCount(NULL) {;}
		shared_ptr(const shared_ptr<T>& rhs) { *this = rhs; }
		shared_ptr(T* data, destroyer d = default_destroy)
		{
			share_count<T>::create(data, fShareCount);
		}
		~shared_ptr()
		{
			if (fShareCount)
				fShareCount->release(fShareCount);
		}

		shared_ptr<T>& operator=(const shared_ptr<T>& rhs)
		{
			if (this != &rhs)
			{
				if (fShareCount)
					fShareCount->release(fShareCount);
				rhs.fShareCount->acquire(fShareCount);
			}
			return *this;
		}

		shared_ptr<T>& operator=(T* data)
		{
			if (fShareCount)
			{
				if (data != fShareCount->data())
				{
					fShareCount->release(fShareCount);
					share_count<T>::create(data, fShareCount);
				}
			}
			else
			{
				share_count<T>::create(data, fShareCount);
			}
			return *this;
		}

		T* get()
		{
			if (fShareCount)
				return fShareCount->data();
			return NULL;
		}

	private:
		void default_destroy(T* data) { delete data; }
		template<class U>
		class share_count
		{
			public:
				static void create(U* data, share_count<U>*& ref)
				{
					share_count* temp = new share_count(data);
					temp->acquire(ref);
				}

				void acquire(share_count<U>*& ref)
				{
					ref = this;
					++count;
				}
				void release(share_count<U>*& ref)
				{
					ref = NULL;
					--count;
					if (count <= 0)
						delete this;
				}

				U*		data()			{ return fData; }
				int32_t	count() const	{ return fCount; }

			private:
				share_count(U* data)
					:	fCount(0), fData(data)
				{
				}
				~share_count()
				{
					if (data)
					{
						delete fData;
						fData = NULL;
					}
				}

				int32_t	fCount;
				U*		fData;
		};

		share_count<T>*	fShareCount;
};


class Message::MessageSharedPtr
{
	public:
		MessageSharedPtr() : fShareCount(NULL) {;}
		MessageSharedPtr(const MessageSharedPtr& rhs) : fShareCount(NULL) { *this = rhs; }
		MessageSharedPtr(AlpMessage* message);
		~MessageSharedPtr();

		MessageSharedPtr& operator=(const MessageSharedPtr& rhs);
		MessageSharedPtr& operator=(AlpMessage* msg);

		AlpMessage* get();

	private:
		class ShareCount
		{
			public:
				static void Create(AlpMessage* msg, ShareCount*& ref);

				void	Acquire(ShareCount*& ref);
				void	Release(ShareCount*& ref);

				AlpMessage*	Data()			{ return message; }
				int32_t		Count() const	{ return count; }

			private:
				ShareCount(AlpMessage* message);
				~ShareCount();

				int32_t	count;
				AlpMessage*	message;
		};

		ShareCount*	fShareCount;
};


//----------------------------------------------------------------------------
Message::Message()
	:	fMessage(NULL),
		fMode(Message::none),
		fStatus(ALP_STATUS_IPC_ERROR),
		fWhat(0)
{
}


Message::Message(AlpMessage* msg)
	:	fMessage(new MessageSharedPtr(msg)),
		fMode(Message::receive),
		fStatus(ALP_OK)
{
	fWhat = alp_ipc_message_get_message_ID(GetMessage());
	fStatus = alp_ipc_message_unpack_start(GetMessage());
}


Message::Message(int what)
	:	fMessage(new MessageSharedPtr),
		fMode(Message::send),
		fStatus(ALP_OK),
		fWhat(what)
{
	*fMessage = alp_ipc_message_create();
	if (!fMessage)
		fStatus = ALP_STATUS_IPC_ERROR;
	else
		fStatus = alp_ipc_message_pack_start(GetMessage(), What());
}


Message::Message(const Message& rhs)
	:	fMessage(new MessageSharedPtr(*(rhs.fMessage))),
		fMode(rhs.Mode()),
		fStatus(rhs.Status()),
		fWhat(rhs.What())
{
}


Message::~Message()
{
	if (fMessage)
		delete fMessage;
}


Message&	Message::operator=(const Message& rhs)
{
	if (this != &rhs)
	{
		if (rhs.fMessage)
		{
			if (!fMessage)
			{
				fMessage = new MessageSharedPtr(*(rhs.fMessage));
			}
			else
			{
		*fMessage	= *rhs.fMessage;
			}
		}

		fMode		= rhs.fMode;
		fStatus		= rhs.fStatus;
		fWhat		= rhs.fWhat;
	}
	return *this;
}


alp_status_t Message::Send(AlpConnection* channel)
{
	alp_status_t err = ALP_STATUS_IPC_ERROR;
	if (Mode() == Message::send && GetMessage() && !Status() && channel)
	{
		err = alp_ipc_message_pack_end(GetMessage());
		if (!err)
			err = alp_ipc_connection_send(channel, GetMessage());
	}
	return err;
}


alp_status_t Message::AddInt32(int32_t i)
{
	if (Mode() == Message::send && GetMessage() && !Status())
		return alp_ipc_message_pack_int32(GetMessage(), i);
	return ALP_STATUS_IPC_ERROR;
}


alp_status_t Message::AddUInt32(uint32_t i)
{
	if (Mode() == Message::send && GetMessage() && !Status())
		return alp_ipc_message_pack_uint32(GetMessage(), i);
	return ALP_STATUS_IPC_ERROR;
}


alp_status_t Message::AddBool(bool b)
{
	if (Mode() == Message::send && GetMessage() && !Status())
		return alp_ipc_message_pack_int32(GetMessage(), b);
	return ALP_STATUS_IPC_ERROR;
}


alp_status_t Message::AddFloat(float f)
{
	if (Mode() == Message::send && GetMessage() && !Status())
		return alp_ipc_message_pack_float(GetMessage(), f);
	return ALP_STATUS_IPC_ERROR;
}


alp_status_t Message::AddString(const char* s)
{
	if (Mode() == Message::send && GetMessage() && !Status())
		return alp_ipc_message_pack_string(GetMessage(), s);
	return ALP_STATUS_IPC_ERROR;
}


alp_status_t Message::AddString(const std::string& s)
{
	if (Mode() == Message::send && GetMessage() && !Status())
		return alp_ipc_message_pack_string(GetMessage(), s.c_str());
	return ALP_STATUS_IPC_ERROR;
}

alp_status_t Message::AddString(const alp::string& s)
{
	return AddString(s.c_str());
}

alp_status_t Message::AddByteArray(const unsigned char* a, uint32_t len)
{
	if (Mode() == Message::send && GetMessage() && !Status())
		return alp_ipc_message_pack_byte_array(GetMessage(), (guchar*)a, len);
	return ALP_STATUS_IPC_ERROR;
}


alp_status_t Message::GetInt32(int32_t& i)
{
	if (Mode() == Message::receive && GetMessage() && !Status())
		return alp_ipc_message_unpack_int32(GetMessage(), &i);
	return ALP_STATUS_IPC_ERROR;
}


alp_status_t Message::GetUInt32(uint32_t& i)
{
	if (Mode() == Message::receive && GetMessage() && !Status())
	{
		alp_status_t err = alp_ipc_message_unpack_uint32(GetMessage(), &i);
		return err;
	}
	return ALP_STATUS_IPC_ERROR;
}


alp_status_t Message::GetBool(bool& b)
{
	if (Mode() == Message::receive && GetMessage() && !Status())
	{
		gint32 temp;
		alp_status_t err = alp_ipc_message_unpack_int32(GetMessage(), &temp);
		b = (bool)temp;
		return err;
	}
	return ALP_STATUS_IPC_ERROR;
}


alp_status_t Message::GetFloat(float& f)
{
	if (Mode() == Message::receive && GetMessage() && !Status())
		return alp_ipc_message_unpack_float(GetMessage(), &f);
	return ALP_STATUS_IPC_ERROR;
}


alp_status_t Message::GetString(char*& s)
{
	if (Mode() == Message::receive && GetMessage() && !Status())
		return alp_ipc_message_unpack_string(GetMessage(), &s);
	return ALP_STATUS_IPC_ERROR;
}


alp_status_t Message::GetString(std::string& s)
{
	if (Mode() == Message::receive && GetMessage() && !Status())
	{
		char* temp = NULL;
		alp_status_t err = alp_ipc_message_unpack_string(GetMessage(), &temp);
		if (!err)
			s = temp;
		if (temp)
			g_free(temp);
		return err;
	}
	return ALP_STATUS_IPC_ERROR;
}

alp_status_t Message::GetString(alp::string& s)
{
	char* str;
	
	alp_status_t result = GetString(str);
	if( ALP_STATUS_OK == result )
	{
		s = str;
		g_free(str);
	}
	
	return result;
}


alp_status_t Message::GetByteArray(unsigned char*& a, uint32_t& len)
{
	if (Mode() == Message::receive && GetMessage() && !Status())
		return alp_ipc_message_unpack_byte_array(GetMessage(), &a, &len);
	return ALP_STATUS_IPC_ERROR;
}


AlpMessage* Message::GetMessage()
{
	return fMessage->get();
}


Message& operator<<(Message& ipc, const int32_t i)
{
	ipc.AddInt32(i);
	return ipc;
}


Message& operator<<(Message& ipc, const uint32_t i)
{
	ipc.AddUInt32(i);
	return ipc;
}


Message& operator<<(Message& ipc, const bool b)
{
	ipc.AddBool(b);
	return ipc;
}


Message& operator<<(Message& ipc, const float f)
{
	ipc.AddFloat(f);
	return ipc;
}


Message& operator<<(Message& ipc, const char* s)
{
	ipc.AddString(s);
	return ipc;
}


Message& operator<<(Message& ipc, const std::string& s)
{
	ipc.AddString(s);
	return ipc;
}


Message& operator<<(Message& ipc, const Array& a)
{
	ipc.AddByteArray(a.array, a.len);
	return ipc;
}


Message& operator>>(Message& ipc, int32_t& i)
{
	ipc.GetInt32(i);
	return ipc;
}


Message& operator>>(Message& ipc, uint32_t& i)
{
	ipc.GetUInt32(i);
	return ipc;
}


Message& operator>>(Message& ipc, bool& b)
{
	ipc.GetBool(b);
	return ipc;
}


Message& operator>>(Message& ipc, float& f)
{
	ipc.GetFloat(f);
	return ipc;
}


Message& operator>>(Message& ipc, char*& s)
{
	ipc.GetString(s);
	return ipc;
}


Message& operator>>(Message& ipc, std::string& s)
{
	ipc.GetString(s);
	return ipc;
}


Message& operator>>(Message& ipc, Array& a)
{
	ipc.GetByteArray(a.array, a.len);
	return ipc;
}


//----------------------------------------------------------------------------

Message::MessageSharedPtr::MessageSharedPtr(AlpMessage* message)
{
	ShareCount::Create(message, fShareCount);
}


Message::MessageSharedPtr::~MessageSharedPtr()
{
	if (fShareCount)
		fShareCount->Release(fShareCount);
}


Message::MessageSharedPtr& Message::MessageSharedPtr::operator=(const Message::MessageSharedPtr& rhs)
{
	if (this != &rhs)
	{
		if (fShareCount)
			fShareCount->Release(fShareCount);
		rhs.fShareCount->Acquire(fShareCount);
	}

	return *this;
}


Message::MessageSharedPtr& Message::MessageSharedPtr::operator=(AlpMessage* msg)
{
	if (fShareCount)
	{
		if (msg != fShareCount->Data())
		{
			fShareCount->Release(fShareCount);
			ShareCount::Create(msg, fShareCount);
		}
	}
	else
	{
		ShareCount::Create(msg, fShareCount);
	}

	return *this;
}


AlpMessage* Message::MessageSharedPtr::get()
{
	if (fShareCount)
		return fShareCount->Data();
	return NULL;
}


//----------------------------------------------------------------------------

void Message::MessageSharedPtr::ShareCount::Create(AlpMessage* msg, ShareCount*& ref)
{
	ShareCount* temp = new ShareCount(msg);
	temp->Acquire(ref);
}


void Message::MessageSharedPtr::ShareCount::Acquire(ShareCount*& ref)
{
	ref = this;
	++count;
}


void Message::MessageSharedPtr::ShareCount::Release(ShareCount*& ref)
{
	ref = NULL;
	--count;
	if (count <= 0)
	{
		delete this;
	}
}


Message::MessageSharedPtr::ShareCount::ShareCount(AlpMessage* msg)
	:	count(0),
		message(NULL)
{
	message = msg;
}


Message::MessageSharedPtr::ShareCount::~ShareCount()
{
	if (message)
	{
		alp_ipc_message_destroy(message);
		message = NULL;
	}
}

}	// namespace ipc

}	// namespace Alp
