/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
// Includes ------------------------------------------------------------------
#include <hiker/fail.h>

#include <hiker/prv/IpcMessage.h>
#include <hiker/prv/IpcServer.h>


//----------------------------------------------------------------------------
namespace alp {

namespace ipc {

Server::Server(const char* name, AlpIpcChannelAccessMode mode, int flags)
	:	fServerChannel(NULL),
		fStatus(ALP_OK),
		fFlags(flags),
		fCurrentConnection(NULL)
{
	Init(name, mode);
}


Server::Server(const std::string& name, AlpIpcChannelAccessMode mode, int flags)
	:	fServerChannel(NULL),
		fStatus(ALP_OK),
		fFlags(flags),
		fCurrentConnection(NULL)
{
	Init(name.c_str(), mode);
}


Server::~Server()
{
	if (fServerChannel)
		alp_ipc_channel_destroy(fServerChannel);
}


alp_status_t Server::Run()
{
	GMainLoop *loop = g_main_loop_new(NULL, false);
	if (!Status())
	{
		g_main_loop_run(loop);
		return 0;
	}

	return Status();
}


alp_status_t Server::Status() const
{
	return fStatus;
}


int Server::Flags() const
{
	return fFlags;
}


alp_status_t Server::Send(Message& msg)
{
	alp_status_t result = ALP_STATUS_IPC_NOT_CONNECTED;
	
	if( NULL != fCurrentConnection )
	{
		result = msg.Send(fCurrentConnection);
	}
	
	return result;
}

uid_t
Server::GetConnectionUID() const
{
	return alp_ipc_connection_get_peer_uid(fCurrentConnection);
}

gid_t
Server::GetConnectionGID() const
{
	return alp_ipc_connection_get_peer_gid(fCurrentConnection);
}

pid_t
Server::GetConnectionPID() const
{
	return alp_ipc_connection_get_peer_pid(fCurrentConnection);
}

void Server::Init(const char* name, AlpIpcChannelAccessMode mode)
{
	fServerChannel = alp_ipc_channel_create(name, mode);
	if (!fServerChannel)
	{
		ALP_FAIL_FATAL_IF(TRUE, "Channel creation failed.\n", NULL);

		return;
	}

	if (alp_ipc_channel_register_connect_callback(fServerChannel,
												  (AlpChannelConnectCB)ChannelConnectCB,
												  gpointer(this)))
	{
		// TODO: handle error better???
		ALP_FAIL_FATAL_IF(TRUE, "Channel connect callback registration failed.\n", NULL);
		return;
	}

	if (fFlags & cfDisconnect)
	{
		if (alp_ipc_channel_register_disconnect_callback(fServerChannel,
														 (AlpChannelDisconnectCB)ChannelDisconnectCB,
														 gpointer(this)))
		{
			// TODO: handle error better???
			ALP_FAIL_FATAL_IF(TRUE, "Channel disconnect callback registration failed.\n", NULL);
			return;
		}
	}
}


void Server::ChannelConnectCB(AlpConnection* connection, Server* server)
{
	if (server->fFlags & cfConnect)
		server->OnChannelConnect(connection);

	if (server->fFlags & cfReceive)
	{
		if (alp_ipc_connection_register_receive_callback(connection,
														 (AlpMessageReceiveCB)MessageReceiveCB,
														 (gpointer)server))
		{
			// TODO: handle error better???
			ALP_FAIL_FATAL_IF(TRUE, "Channel message receive callback registration failed.\n", NULL);
		}
	}
}


void Server::ChannelDisconnectCB(AlpConnection* connection, Server* server)
{
	server->OnChannelDisconnect(connection);
}


void Server::MessageReceiveCB(AlpConnection* connection, AlpMessage* rawMsg, Server* server)
{
	Message message(rawMsg);

	// We had a bogus unpack start; bail out
	if (message.Status())
		return;

	server->fCurrentConnection = connection;
	server->OnMessageReceived(message);
	server->fCurrentConnection = NULL;
}

}	// namespace ipc

}	// namespace Alp
