/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/*
  Unit tests for the MAX Abstract IPC library
*/


#include <hiker/unit_test.h>
#include <cppunit/extensions/HelperMacros.h>

#include <gtk/gtk.h>
#include <hiker/ipc.h>

#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <errno.h>

#include "alp_ipc_unit_test.h"

class IPCTest: public CPPUNIT_NS::TestFixture
{
	CPPUNIT_TEST_SUITE( IPCTest );
	
	CPPUNIT_TEST( IPCConnectTest );

	CPPUNIT_TEST( IPCMessageTest );
	
	CPPUNIT_TEST_SUITE_END();
	
public:
	void setUp();         // overriding TestFixture method
	void tearDown();      // overriding TestFixture method
		
protected:
	// Suite-level setup & teardown.  Not really unit tests...
	void IPCConnectTest(void);

	void IPCMessageTest(void);

private:

	void send_hello_message(AlpConnection* connection);

	void send_hello_with_response(AlpConnection* connection, gboolean doTimeout);

	void send_long_message_with_response(AlpConnection* connection);

	void send_hello_with_deferred(AlpConnection* connection);

	void send_hello_response_with_deferred(AlpConnection* connection);

	void send_each_type_message(AlpConnection* connection);

	void send_long_format_string_message(AlpConnection* connection);

	void send_long_message(AlpConnection* connection);

	void send_shutdown_message(AlpConnection* connection);

	static void client_receive_message_CB(AlpConnection* connection,
								   AlpMessage* message, gpointer cbData);

	static void client_connect_CB(AlpConnection* connection, gpointer cbData);

	static gboolean timeout_CB(gpointer cbData);

	void use_custom_connection_data(AlpConnection* connection);
	void use_custom_channel_data(AlpChannel* connection);

	AlpConnection* mySynchConnection;
	AlpConnection* myAsynchConnection;

	GMainLoop*  myClientMainLoop;

	guint     myTimeoutSrcID;
	guint     myAsynchReceivedCount;

	guint     myAsynchExpectedCount;

	gint      myAsynchExpectedMsgID;

	gboolean  myCalledClientConnectCB;
};


ALP_ROOT_TEST_SUITE_REGISTRATION( "OS-045", CPPUNIT_REGISTRY_TYPE_AUTO );
CPPUNIT_TEST_SUITE_NAMED_REGISTRATION( IPCTest, "OS-045" );

void IPCTest::setUp()
{
	mySynchConnection = NULL;
	myAsynchConnection = NULL;

	myClientMainLoop = NULL;

	myTimeoutSrcID = 0;
	myAsynchReceivedCount = 0;

	myAsynchExpectedCount = 0;

	myAsynchExpectedMsgID = 0;

	myCalledClientConnectCB = FALSE;
}

void IPCTest::tearDown()
{

}


// **************************************
// Suite-level Setup & Teardown:
// **************************************
void IPCTest::IPCConnectTest(void)
{
	pthread_t serverThread;

	char *top;
	char *argv[5];

	alp_status_t status;

	mySynchConnection = NULL;
	myAsynchConnection = NULL;

	myCalledClientConnectCB = FALSE;
  
	myClientMainLoop = g_main_loop_new(NULL, TRUE);

	// First do some stuff that might fail, since we can't assert after forking.
	top = getenv("ALP_RUN_TOP");
	CPPUNIT_ASSERT(top != NULL);

	// first, try connect to the IPC_TEST channel
	// this should fail, since the server is not running
	myAsynchConnection = alp_ipc_channel_connect(UNIT_TEST_CHANNEL_NAME_CONNECT, 
										ALP_IPC_CONNECTION_USAGE_HINT_NONE,
										NULL, NULL);

	CPPUNIT_ASSERT(myAsynchConnection == NULL);

	// now try the same with the synchronous connect method
	// this should time out after 100 ms
	status = alp_ipc_channel_connect_synch(UNIT_TEST_CHANNEL_NAME_CONNECT, 
								    ALP_IPC_CONNECTION_USAGE_HINT_NONE,
								    100,
								    &mySynchConnection);

	CPPUNIT_ASSERT(status != ALP_STATUS_OK);
	CPPUNIT_ASSERT(mySynchConnection == NULL);

	int sockets[2];

	CPPUNIT_ASSERT(socketpair(AF_UNIX, SOCK_STREAM, 0, &sockets[0]) == 0);

	pid_t parentPID = getpid();
	pid_t newPID = fork();

	if (parentPID != getpid()) {
	
		close(sockets[0]);
		
		// the forked process exec's alpUnitTestServer
		int size = strlen(top);
		size += strlen("/opt/alp/bin/ipcUnitTestServer") + 1;
		char *runme;
		runme = (char *)malloc(sizeof(char) * size);

		strcpy(runme, top);
		strcat(runme, "/opt/alp/bin/ipcUnitTestServer");

		char pipename[32];
		argv[0] = "ipcUnitTestServer";
		argv[1] = "-parentFD";
		sprintf(pipename, "%d", sockets[1]);
		argv[2] = pipename;
		argv[3] = "-connectTest";
		argv[4] = NULL;

		execv(runme, argv);

		exit(0);   // make sure you don't fall through to code below
	}

	close(sockets[1]);
	
	char buf;
	// wait until server thread is started
	CPPUNIT_ASSERT(read(sockets[0], &buf, 1) == 1);

	// try an asynch connection
	myAsynchConnection = alp_ipc_channel_connect(UNIT_TEST_CHANNEL_NAME_CONNECT, 								    
										ALP_IPC_CONNECTION_USAGE_HINT_NONE,
										IPCTest::client_connect_CB, this);

	CPPUNIT_ASSERT(myAsynchConnection != NULL);

	if (myCalledClientConnectCB == FALSE) {
		// register a timeout callback to occur if the asynch connection
		// doesn't happen in 5 seconds
		this->myTimeoutSrcID = g_timeout_add(5000, timeout_CB, (gpointer)this);
		CPPUNIT_ASSERT(this->myTimeoutSrcID > 0);
	
		// this loop will be exited in the timeout and connect callbacks
		g_main_loop_run(myClientMainLoop); 
	}

	// get the server's pid using the peer queries
	gint serverPID = alp_ipc_connection_get_peer_pid(myAsynchConnection);
	CPPUNIT_ASSERT(serverPID = newPID);

	// call the get gid and uid APIs
	gint serverGID = alp_ipc_connection_get_peer_gid(myAsynchConnection);
	gint serverUID = alp_ipc_connection_get_peer_uid(myAsynchConnection);

	CPPUNIT_ASSERT(serverGID == (gint)getgid());
	CPPUNIT_ASSERT(serverUID == (gint)getuid());

	// check that we can use the custom data fields
	//use_custom_channel_data(myAsynchConnection);

	alp_ipc_channel_disconnect(myAsynchConnection);

	// try connect to the IPC_TEST channel using the synchronous method
	status = alp_ipc_channel_connect_synch(UNIT_TEST_CHANNEL_NAME_CONNECT, 
								    ALP_IPC_CONNECTION_USAGE_HINT_NONE,
								    5000,
								    &mySynchConnection);

	CPPUNIT_ASSERT(mySynchConnection != NULL);
	// see if we can use custom channel data

	send_shutdown_message(mySynchConnection);
	
	// the server might end on the HUP instead of the shutdown message, but that
	// won't alter the test behaviour. We'd need an ack on the pipe if we wanted
	// the shutdown to be verified.
	close(sockets[0]);

	return;

}

void IPCTest::IPCMessageTest(void)
{
	pthread_t serverThread;

	char *top;
	alp_status_t status;
	char *argv[5];

	mySynchConnection = NULL;
	myAsynchConnection = NULL;

	myClientMainLoop = g_main_loop_new(NULL, TRUE);

	AlpMessage* message = NULL;

	// First do some stuff that might fail, since we can't assert after forking.
	top = getenv("ALP_RUN_TOP");
	CPPUNIT_ASSERT(top != NULL);


	int sockets[2];
	
	CPPUNIT_ASSERT(socketpair(AF_UNIX, SOCK_STREAM, 0, &sockets[0]) == 0);

	pid_t parentPID = getpid();
	pid_t newPID = fork();

	if (parentPID != getpid()) {

		close(sockets[0]);

		// the forked process exec's alpUnitTestServer
		int size = strlen(top);
		size += strlen("/opt/alp/bin/ipcUnitTestServer") + 1;
		char *runme;
		runme = (char *)malloc(sizeof(char) * size);

		strcpy(runme, top);
		strcat(runme, "/opt/alp/bin/ipcUnitTestServer");

		argv[0] = "ipcUnitTestServer";
		char pipename[32];
		argv[1] = "-parentFD";
		sprintf(pipename, "%d", sockets[1]);
		argv[2] = pipename;
		argv[3] = "-messageTest";
		argv[4] = NULL;
		execv(runme, argv);

		exit(0);   // make sure you don't fall through to code below
	}

	//  fprintf(stderr, "NEW PID %d\n", newPID);

	close(sockets[1]);
	
	char buf;
	// wait until server thread is started
	CPPUNIT_ASSERT(read(sockets[0], &buf, 1) == 1);

	///////////////////////////////////////////////////////////////////
	//
	// Send an assortment of edge case messages using the asynchronous
	// receive API
	//
	///////////////////////////////////////////////////////////////////

	myCalledClientConnectCB = FALSE;

	// connect to the IPC_TEST channel using the synchronous method
	myAsynchConnection = alp_ipc_channel_connect(UNIT_TEST_CHANNEL_NAME_MESSAGE,
										ALP_IPC_CONNECTION_USAGE_HINT_NONE,
										IPCTest::client_connect_CB, this);

	CPPUNIT_ASSERT(myAsynchConnection != NULL);

	if (myCalledClientConnectCB == FALSE) {
		// register a timeout callback to occur if the asynch connection
		// doesn't happen in 5 seconds
		this->myTimeoutSrcID = g_timeout_add(5000, timeout_CB, (gpointer)this);
		CPPUNIT_ASSERT(this->myTimeoutSrcID > 0);

		// this loop will be exited in the timeout and connect callbacks
		g_main_loop_run(myClientMainLoop); 
	}

	// at this point, there's a good connection

	myAsynchReceivedCount = 0;
	myAsynchExpectedCount = 0;
	myAsynchExpectedMsgID = UNIT_TEST_MESSAGE_SERVER_OK;

	// send a Hello message to the server 
	send_hello_message(myAsynchConnection);

	// send a message with a packed parameter of each type to the 
	// server and wait for a reply
	send_each_type_message(myAsynchConnection);

	// send a message with a long format string (and consequently a large number
	// of parameters to the server and wait for a reply
	send_long_format_string_message(myAsynchConnection);

	// send a message containing a long string and wait for a reply
	send_long_message(myAsynchConnection);

	// now enter a GTK main loop and wait to receive 4 OK replies
	// to the four message just sent

	// register a timeout callback to occur if all the replies are not
	// received in 5 seconds
	this->myTimeoutSrcID = g_timeout_add(5000, timeout_CB, (gpointer)this);
	CPPUNIT_ASSERT(this->myTimeoutSrcID > 0);

	myAsynchReceivedCount = 0;
	myAsynchExpectedCount = 4;
	myAsynchExpectedMsgID = UNIT_TEST_MESSAGE_SERVER_OK;

	// this loop will be exited in the timeout and receive callbacks
	g_main_loop_run(myClientMainLoop); 

	///////////////////////////////////////////////////////////////////
	//
	// Send an assortment of edge case messages using the synchronous
	// receive API
	//
	///////////////////////////////////////////////////////////////////

	// connect to the IPC_TEST channel using the synchronous method
	status = alp_ipc_channel_connect_synch(UNIT_TEST_CHANNEL_NAME_MESSAGE, 
								    ALP_IPC_CONNECTION_USAGE_HINT_SYNCH_ONLY,
								    5000,
								    &mySynchConnection);

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	CPPUNIT_ASSERT(mySynchConnection != NULL);

	// send a Hello message to the server and wait for a reply
	send_hello_message(mySynchConnection);

	status = alp_ipc_connection_receive(mySynchConnection, ALP_IPC_RECEIVE_ANY,
								 1000, &message);

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	CPPUNIT_ASSERT(alp_ipc_message_get_message_ID(message) == UNIT_TEST_MESSAGE_SERVER_OK);

	alp_ipc_message_destroy(message);

	// send a message with a packed parameter of each type to the 
	// server and wait for a reply
	send_each_type_message(mySynchConnection);

	status = alp_ipc_connection_receive(mySynchConnection, ALP_IPC_RECEIVE_ANY,
								 1000, &message);
	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	CPPUNIT_ASSERT(alp_ipc_message_get_message_ID(message) == UNIT_TEST_MESSAGE_SERVER_OK);

	alp_ipc_message_destroy(message);

	// send a message with a long format string (and consequently a large number
	// of parameters to the server and wait for a reply
	send_long_format_string_message(mySynchConnection);

	status = alp_ipc_connection_receive(mySynchConnection, ALP_IPC_RECEIVE_ANY,
								 1000, &message);
	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	CPPUNIT_ASSERT(alp_ipc_message_get_message_ID(message) == UNIT_TEST_MESSAGE_SERVER_OK);

	alp_ipc_message_destroy(message);
	/*
	// send a message containing a long string and wait for a reply
	send_long_message(mySynchConnection);

	status = alp_ipc_connection_receive(mySynchConnection, ALP_IPC_RECEIVE_ANY,
								 1000, &message);
	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	CPPUNIT_ASSERT(alp_ipc_message_get_message_ID(message) == UNIT_TEST_MESSAGE_SERVER_OK);

	alp_ipc_message_destroy(message);
	*/
	///////////////////////////////////////////////////////////////////
	//
	// test send with response
	//	
	///////////////////////////////////////////////////////////////////
	send_hello_with_response(mySynchConnection, FALSE);
	send_hello_with_response(mySynchConnection, TRUE);

	send_long_message_with_response(mySynchConnection);

	///////////////////////////////////////////////////////////////////
	//
	// test message queuing
	//
	///////////////////////////////////////////////////////////////////

	// first with out of order message

	send_hello_with_deferred(mySynchConnection);

	status = alp_ipc_connection_receive(mySynchConnection, 
								 UNIT_TEST_MESSAGE_SERVER_OK,
								 1000, &message);
	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	CPPUNIT_ASSERT(alp_ipc_message_get_message_ID(message) == UNIT_TEST_MESSAGE_SERVER_OK);

	alp_ipc_message_destroy(message);

	// now there should still be an deferred message that was queued
	// that needs to be received
	status = alp_ipc_connection_receive(mySynchConnection, 
								 UNIT_TEST_MESSAGE_SERVER_DEFERRED,
								 1000, &message);
	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	CPPUNIT_ASSERT(alp_ipc_message_get_message_ID(message) == UNIT_TEST_MESSAGE_SERVER_DEFERRED);

	alp_ipc_message_destroy(message);

	// next, with a message sent before response

	send_hello_response_with_deferred(mySynchConnection);

	// now there should still be an deferred message that was queued
	// that needs to be received
	status = alp_ipc_connection_receive(mySynchConnection, 
								 UNIT_TEST_MESSAGE_SERVER_DEFERRED,
								 1000, &message);
	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	CPPUNIT_ASSERT(alp_ipc_message_get_message_ID(message) == UNIT_TEST_MESSAGE_SERVER_DEFERRED);

	alp_ipc_message_destroy(message);

	// finally, try it with the async connection

	send_hello_response_with_deferred(myAsynchConnection);

	// now enter a GTK main loop and wait to receive 4 OK replies
	// to the four message just sent

	// register a timeout callback to occur if all the replies are not
	// received in 5 seconds
	this->myTimeoutSrcID = g_timeout_add(5000, timeout_CB, (gpointer)this);
	CPPUNIT_ASSERT(this->myTimeoutSrcID > 0);

	myAsynchReceivedCount = 0;
	myAsynchExpectedCount = 1;
	myAsynchExpectedMsgID = UNIT_TEST_MESSAGE_SERVER_DEFERRED;

	// this loop will be exited in the timeout and receive callbacks
	g_main_loop_run(myClientMainLoop); 

	alp_ipc_channel_disconnect(myAsynchConnection);

	send_shutdown_message(mySynchConnection);

	// the server might end on the HUP instead of the shutdown message, but that
	// won't alter the test behaviour. We'd need an ack on the pipe if we wanted
	// the shutdown to be verified.
	close(sockets[0]);

	return;
}

void
IPCTest::client_connect_CB(AlpConnection* connection, gpointer cbData)
{
	IPCTest* ipcTest = (IPCTest *)cbData;

	CPPUNIT_ASSERT(connection != NULL);

	alp_ipc_connection_register_receive_callback(connection,
										IPCTest::client_receive_message_CB,
										cbData);

	ipcTest->myCalledClientConnectCB = TRUE;

	if (ipcTest->myTimeoutSrcID) {
		g_source_remove(ipcTest->myTimeoutSrcID);

		g_main_loop_quit(ipcTest->myClientMainLoop);  
	}
}

gboolean
IPCTest::timeout_CB(gpointer cbData)
{
	IPCTest* ipcTest = (IPCTest *)cbData;

	CPPUNIT_ASSERT(ipcTest != NULL);

	CPPUNIT_ASSERT(false);  // should never timeout

	g_main_loop_quit(ipcTest->myClientMainLoop);  

	return FALSE;
}


void 
IPCTest::client_receive_message_CB(AlpConnection* connection,
							AlpMessage* message, gpointer cbData)
{
	IPCTest* ipcTest = (IPCTest *)cbData;
	gboolean receivedError = FALSE;

	if (alp_ipc_message_get_message_ID(message) == 
	    ipcTest->myAsynchExpectedMsgID) {
		ipcTest->myAsynchReceivedCount++;
	} else if (alp_ipc_message_get_message_ID(message) == 
			 UNIT_TEST_MESSAGE_SERVER_ERROR) {
		receivedError = TRUE;
		CPPUNIT_ASSERT(false);
	}
	else {
		CPPUNIT_ASSERT(false);
	}

	if ((ipcTest->myAsynchReceivedCount == ipcTest->myAsynchExpectedCount) || 
	    receivedError) {
		CPPUNIT_ASSERT(ipcTest->myTimeoutSrcID > 0);
		g_source_remove(ipcTest->myTimeoutSrcID);

		g_main_loop_quit(ipcTest->myClientMainLoop);      
	}						 
}

void
IPCTest::send_hello_message(AlpConnection* connection)
{
	gchar msgStr[] = "Hello from client ";
	gchar* msgStrPtr = msgStr;

	alp_status_t status;

	CPPUNIT_ASSERT(connection != NULL);
	if (connection == NULL)
		return;

	AlpMessage* message = alp_ipc_message_create();

	CPPUNIT_ASSERT(message != NULL);
	if (message == NULL)
		return;

	status = alp_ipc_message_pack_start(message, 
								 UNIT_TEST_MESSAGE_HELLO);

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	if (status == ALP_STATUS_OK)
		status = alp_ipc_message_pack_string(message, msgStrPtr);

	// add a NULL string
	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	if (status == ALP_STATUS_OK)
		status = alp_ipc_message_pack_string(message, NULL);

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	if (status == ALP_STATUS_OK)
		status = alp_ipc_message_pack_int32(message, getpid());

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	if (status == ALP_STATUS_OK)
		status = alp_ipc_message_pack_end(message);

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	if (status == ALP_STATUS_OK) 
		status = alp_ipc_connection_send(connection, message); 

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	alp_ipc_message_destroy(message);
}

void
IPCTest::send_hello_with_deferred(AlpConnection* connection)
{
	gchar msgStr[] = "Hello from client ";
	gchar* msgStrPtr = msgStr;

	alp_status_t status;

	CPPUNIT_ASSERT(connection != NULL);
	if (connection == NULL)
		return;

	AlpMessage* message = alp_ipc_message_create();

	CPPUNIT_ASSERT(message != NULL);
	if (message == NULL)
		return;

	status = alp_ipc_message_pack_start(message, 
								 UNIT_TEST_MESSAGE_HELLO_SYNCH_WITH_EXTRA_MESSAGE);

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	if (status == ALP_STATUS_OK)
		status = alp_ipc_message_pack_string(message, msgStrPtr);

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	if (status == ALP_STATUS_OK)
		status = alp_ipc_message_pack_int32(message, getpid());

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	if (status == ALP_STATUS_OK)
		status = alp_ipc_message_pack_end(message);

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	if (status == ALP_STATUS_OK) 
		status = alp_ipc_connection_send(connection, message); 

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	alp_ipc_message_destroy(message);
}

void
IPCTest::send_hello_response_with_deferred(AlpConnection* connection)
{
	gchar msgStr[] = "Hello from client ";
	gchar* msgStrPtr = msgStr;

	gint  timeout = 100;

	alp_status_t status;

	CPPUNIT_ASSERT(connection != NULL);
	if (connection == NULL)
		return;

	AlpMessage* message = alp_ipc_message_create();
	AlpMessage* responseMessage = NULL;

	CPPUNIT_ASSERT(message != NULL);
	if (message == NULL)
		return;

	status = alp_ipc_message_pack_start(message, 
								 UNIT_TEST_MESSAGE_HELLO_RESPONSE_WITH_EXTRA_MESSAGE);

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	if (status == ALP_STATUS_OK)
		status = alp_ipc_message_pack_string(message, msgStrPtr);

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	if (status == ALP_STATUS_OK)
		status = alp_ipc_message_pack_int32(message, getpid());

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	if (status == ALP_STATUS_OK)
		status = alp_ipc_message_pack_end(message);

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	if (status == ALP_STATUS_OK) 
		status = alp_ipc_connection_send_with_response(connection, message, timeout, &responseMessage); 

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);

	CPPUNIT_ASSERT(responseMessage != NULL);
    
	CPPUNIT_ASSERT(alp_ipc_message_get_message_ID(responseMessage) == 
				UNIT_TEST_MESSAGE_SERVER_OK);

	if (responseMessage)
		alp_ipc_message_destroy(responseMessage);

	alp_ipc_message_destroy(message);
}

void
IPCTest::send_hello_with_response(AlpConnection* connection, gboolean doTimeout)
{
	gchar msgStr[] = "Hello from client ";
	gchar* msgStrPtr = msgStr;

	gint  timeout = 100;

	alp_status_t status;

	CPPUNIT_ASSERT(connection != NULL);
	if (connection == NULL)
		return;

	AlpMessage* message = alp_ipc_message_create();
	AlpMessage* responseMessage = NULL;

	CPPUNIT_ASSERT(message != NULL);
	if (message == NULL)
		return;

	if (doTimeout == FALSE) 
		status = alp_ipc_message_pack_start(message, 
									 UNIT_TEST_MESSAGE_HELLO_WITH_RESPONSE);
	else
		status = alp_ipc_message_pack_start(message, 
									 UNIT_TEST_MESSAGE_HELLO_WITH_RESPONSE_TIMEOUT);

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	if (status == ALP_STATUS_OK)
		status = alp_ipc_message_pack_string(message, msgStrPtr);

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	if (status == ALP_STATUS_OK)
		status = alp_ipc_message_pack_int32(message, getpid());

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	if (status == ALP_STATUS_OK)
		status = alp_ipc_message_pack_end(message);

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	if (status == ALP_STATUS_OK) 
		status = alp_ipc_connection_send_with_response(connection, message, timeout, &responseMessage); 

	if (doTimeout == TRUE) {
		CPPUNIT_ASSERT(status == ALP_STATUS_IPC_TIMEOUT);
	} else {
		CPPUNIT_ASSERT(status == ALP_STATUS_OK);

		CPPUNIT_ASSERT(responseMessage != NULL);
    
		CPPUNIT_ASSERT(alp_ipc_message_get_message_ID(responseMessage) == 
					UNIT_TEST_MESSAGE_SERVER_OK);
	}

	if (responseMessage)
		alp_ipc_message_destroy(responseMessage);

	alp_ipc_message_destroy(message);
}

void
IPCTest::send_long_message_with_response(AlpConnection* connection)
{
	alp_status_t status;
	guint i;
	GString* longString = NULL;

	CPPUNIT_ASSERT(connection != NULL);
	if (connection == NULL)
		return;

	AlpMessage* message = alp_ipc_message_create();

	CPPUNIT_ASSERT(message != NULL);
	if (message == NULL)
		return;

	status = alp_ipc_message_pack_start(message, UNIT_TEST_MESSAGE_LONG_MESSAGE_WITH_RESPONSE);

	longString = g_string_sized_new(gLongMessageLength + 1);
	CPPUNIT_ASSERT(longString != NULL);

	for (i=0; i<gLongMessageLength / 10; i++) {
		g_string_append(longString, "0123456789");
	}

	CPPUNIT_ASSERT(longString->len == gLongMessageLength);

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	if (status == ALP_STATUS_OK) {
		status = alp_ipc_message_pack_string(message, longString->str);
	}

	g_string_free(longString, TRUE);

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	if (status == ALP_STATUS_OK)
		status = alp_ipc_message_pack_end(message);

	AlpMessage* responseMessage = NULL;

	CPPUNIT_ASSERT(message != NULL);
	if (message == NULL)
		return;

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);

	if (status == ALP_STATUS_OK) {
		status = alp_ipc_connection_send_with_response(connection, message, 2000, &responseMessage); 

		CPPUNIT_ASSERT(status == ALP_STATUS_OK);

		CPPUNIT_ASSERT(responseMessage != NULL);
	
		CPPUNIT_ASSERT(alp_ipc_message_get_message_ID(responseMessage) == 
					UNIT_TEST_MESSAGE_SERVER_OK);
	}

	if (responseMessage)
		alp_ipc_message_destroy(responseMessage);

	alp_ipc_message_destroy(message);
}

void
IPCTest::send_each_type_message(AlpConnection* connection)
{
	// send a message with one parameter of each type
	gchar msgStr[] = "-23";
	gchar* msgStrPtr = msgStr;
	guchar array23Bytes[] = { 23, 22, 21, 20, 19, 18, 17, 16, 15, 14, 13, 12, 11,
						 10, 9, 8, 7, 6, 5, 4, 3, 2, 1 };
	alp_status_t status;

	CPPUNIT_ASSERT(connection != NULL);
	if (connection == NULL)
		return;

	AlpMessage* message = alp_ipc_message_create();

	CPPUNIT_ASSERT(message != NULL);
	if (message == NULL)
		return;

	status = alp_ipc_message_pack_start(message, UNIT_TEST_MESSAGE_PACK_EACH_TYPE);

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	if (status == ALP_STATUS_OK)
		status = alp_ipc_message_pack_int32(message, -23);

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	if (status == ALP_STATUS_OK)
		status = alp_ipc_message_pack_uint32(message, 23);

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	if (status == ALP_STATUS_OK)
		status = alp_ipc_message_pack_float(message, -23.0);

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	if (status == ALP_STATUS_OK)
		status = alp_ipc_message_pack_string(message, msgStrPtr);

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	if (status == ALP_STATUS_OK)
		status = alp_ipc_message_pack_byte_array(message, array23Bytes, 23);

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	if (status == ALP_STATUS_OK)
		status = alp_ipc_message_pack_end(message);

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	if (status == ALP_STATUS_OK) 
		status = alp_ipc_connection_send(connection, message); 

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	alp_ipc_message_destroy(message);
}

void
IPCTest::send_long_format_string_message(AlpConnection* connection)
{
	alp_status_t status;
	int i;

	CPPUNIT_ASSERT(connection != NULL);
	if (connection == NULL)
		return;

	AlpMessage* message = alp_ipc_message_create();

	CPPUNIT_ASSERT(message != NULL);
	if (message == NULL)
		return;

	status = alp_ipc_message_pack_start(message, UNIT_TEST_MESSAGE_LONG_FORMAT_STRING);
	CPPUNIT_ASSERT(status == ALP_STATUS_OK);

	if (status == ALP_STATUS_OK) {
		for (i=0; i<128; i++) {
			CPPUNIT_ASSERT(status == ALP_STATUS_OK);
			if (status == ALP_STATUS_OK)
				status = alp_ipc_message_pack_int32(message, i);
			else
				break;
		}
	}

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	if (status == ALP_STATUS_OK)
		status = alp_ipc_message_pack_end(message);

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	if (status == ALP_STATUS_OK) 
		status = alp_ipc_connection_send(connection, message); 

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	alp_ipc_message_destroy(message);
}

void
IPCTest::send_long_message(AlpConnection* connection)
{
	alp_status_t status;
	guint i;
	GString* longString = NULL;

	CPPUNIT_ASSERT(connection != NULL);
	if (connection == NULL)
		return;

	AlpMessage* message = alp_ipc_message_create();

	CPPUNIT_ASSERT(message != NULL);
	if (message == NULL)
		return;

	status = alp_ipc_message_pack_start(message, UNIT_TEST_MESSAGE_LONG_MESSAGE);

	longString = g_string_sized_new(gLongMessageLength + 1);
	CPPUNIT_ASSERT(longString != NULL);

	for (i=0; i<gLongMessageLength / 10; i++) {
		g_string_append(longString, "0123456789");
	}

	CPPUNIT_ASSERT(longString->len == gLongMessageLength);

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	if (status == ALP_STATUS_OK) {
		status = alp_ipc_message_pack_string(message, longString->str);
	}

	g_string_free(longString, TRUE);

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	if (status == ALP_STATUS_OK)
		status = alp_ipc_message_pack_end(message);

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	if (status == ALP_STATUS_OK) 
		status = alp_ipc_connection_send(connection, message); 

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	alp_ipc_message_destroy(message);
}

void
IPCTest::send_shutdown_message(AlpConnection* connection)
{
	alp_status_t status;

	CPPUNIT_ASSERT(connection != NULL);
	if (connection == NULL)
		return;

	AlpMessage* message = alp_ipc_message_create();

	CPPUNIT_ASSERT(message != NULL);
	if (message == NULL)
		return;

	status = alp_ipc_message_pack_start(message, 
								 UNIT_TEST_MESSAGE_SHUTDOWN_SERVER);
    
	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	if (status == ALP_STATUS_OK)
		status = alp_ipc_message_pack_end(message);

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	if (status == ALP_STATUS_OK) 
		status = alp_ipc_connection_send(connection, message); 

	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
	alp_ipc_message_destroy(message);
}

void IPCTest::use_custom_connection_data(AlpConnection* connection)
{
	alp_status_t status;
	char *data = "yay";

	CPPUNIT_ASSERT(connection != NULL);
	if (connection == NULL)
		return;

	alp_ipc_connection_set_custom_data(connection, data);
	CPPUNIT_ASSERT(alp_ipc_connection_get_custom_data(connection) == data);
}

void IPCTest::use_custom_channel_data(AlpChannel* connection)
{
	alp_status_t status;
	char *data = "yay";
	AlpChannel* me = alp_ipc_channel_create("thisIsATest",ALP_IPC_CHANNEL_ACCESS_ALL);

	alp_ipc_channel_set_custom_data(me, data);

	CPPUNIT_ASSERT(alp_ipc_channel_get_custom_data(me) == data);
	status = alp_ipc_channel_destroy(me);
	CPPUNIT_ASSERT(status == ALP_STATUS_OK);
}
