/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
#include <hiker/notifymgr.h>
#include <hiker/prv/notifymgr_prv.h>
#include <hiker/fail.h>

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>

extern int alp_notify_client_send(void * buf, int length, int sync, int timeout, int * handled);
extern void * alp_notify_wait_for_callbacks(void *);

static int sNotifyClientPID = 0;

// get the pid of the private notifymgr client thread
static int alp_prv_notify_getpid(void)
{
	// check if the private notifymgr client thread has been created
	if (!sNotifyClientPID) {
		struct timespec time;

		// start the private notifymgr client thread to receive callbacks
		pthread_t thread;
		pthread_create(&thread,NULL,alp_notify_wait_for_callbacks,(void *)&sNotifyClientPID);

		// wait for the private thread to be initialized
		time.tv_sec = 0;
		time.tv_nsec = 1000000000 / 50;
		while (!sNotifyClientPID) nanosleep(&time,NULL);
	}

	// return the pid of the private notifymgr client thread
	return sNotifyClientPID;
}

alp_status_t alp_notify_register(const char * appID, const char * notifyType, int priority, int flags)
{
	AlpNotifyCmdRegister request;

#ifdef ALP_NOTIFY_DEBUG
	printf("alp_notify_register\n");
#endif
	// check for errors
	if (!appID || strlen(appID) > ALP_NOTIFY_MAX_APPID_SIZE)
		return ALP_NOTIFY_ERR_INVALID_APPID;
	if (!notifyType || strlen(notifyType) > ALP_NOTIFY_MAX_NOTIFY_SIZE)
		return ALP_NOTIFY_ERR_INVALID_NOTIFYTYPE;

	// fill out the message header
	memset(&request,0,sizeof(AlpNotifyCmdRegister));
	request.cmd = ALP_NOTIFY_CMD_REGISTER;
	request.size = sizeof(AlpNotifyCmdRegister);

	// fill out the message body
	strcpy(request.notifyType,notifyType);
	request.callback = 0;
    request.userData = 0;
	request.priority = priority;
	request.flags = flags;
	request.pid = 1;
	strcpy(request.appID,appID);
	request.zero = 0;

	// send the message and wait for a response
	return alp_notify_client_send(&request,request.size,0,0,0);
}

alp_status_t alp_notify_register_server(const char * notifyType, AlpNotifyCallbackProc * callbackProc, void * userData, int priority, int flags)
{
	AlpNotifyCmdRegister request;

#ifdef ALP_NOTIFY_DEBUG
	printf("alp_notify_register_server\n");
#endif
	// check for errors
	if (!callbackProc)
		return ALP_NOTIFY_ERR_INVALID_CALLBACK;
	if (!notifyType || strlen(notifyType) > ALP_NOTIFY_MAX_NOTIFY_SIZE)
		return ALP_NOTIFY_ERR_INVALID_NOTIFYTYPE;

	// fill out the message header
	memset(&request,0,sizeof(AlpNotifyCmdRegister));
	request.cmd = ALP_NOTIFY_CMD_REGISTER;
	request.size = sizeof(AlpNotifyCmdRegister);

	// fill out the message body
	strcpy(request.notifyType,notifyType);
	request.callback = (void *)callbackProc;
    request.userData = userData;
	request.priority = priority;
	request.flags = flags;
	request.pid = alp_prv_notify_getpid();
	sprintf(request.appID,"process-%d",alp_prv_notify_getpid());
	request.zero = 0;

	// send the message and wait for a response
	return alp_notify_client_send(&request,request.size,0,0,0);
}

alp_status_t alp_notify_unregister(const char * appID, const char * notifyType, int flags)
{
	AlpNotifyCmdUnregister request;

#ifdef ALP_NOTIFY_DEBUG
	printf("alp_notify_unregister\n");
#endif
	// check for errors
	if (!appID || strlen(appID) > ALP_NOTIFY_MAX_APPID_SIZE)
		return ALP_NOTIFY_ERR_INVALID_APPID;
	if (!notifyType || strlen(notifyType) > ALP_NOTIFY_MAX_NOTIFY_SIZE)
		return ALP_NOTIFY_ERR_INVALID_NOTIFYTYPE;

	// fill out the message header
	memset(&request,0,sizeof(AlpNotifyCmdUnregister));
	request.cmd = ALP_NOTIFY_CMD_UNREGISTER;
	request.size = sizeof(AlpNotifyCmdUnregister);

	// fill out the message body
	strcpy(request.notifyType,notifyType);
	request.flags = flags;
	strcpy(request.appID,appID);
	request.zero = 0;

	// send the message and wait for a response
	return alp_notify_client_send(&request,request.size,0,0,0);
}

alp_status_t alp_notify_unregister_server(const char * notifyType, int flags)
{
	AlpNotifyCmdUnregister request;

#ifdef ALP_NOTIFY_DEBUG
	printf("alp_notify_unregister_server\n");
#endif
	// check for errors
	if (!notifyType || strlen(notifyType) > ALP_NOTIFY_MAX_NOTIFY_SIZE)
		return ALP_NOTIFY_ERR_INVALID_NOTIFYTYPE;

	// fill out the message header
	memset(&request,0,sizeof(AlpNotifyCmdUnregister));
	request.cmd = ALP_NOTIFY_CMD_UNREGISTER;
	request.size = sizeof(AlpNotifyCmdUnregister);

	// fill out the message body
	strcpy(request.notifyType,notifyType);
	request.flags = flags;
	sprintf(request.appID,"process-%d",alp_prv_notify_getpid());
	request.zero = 0;

	// send the message and wait for a response
	return alp_notify_client_send(&request,request.size,0,0,0);
}

static alp_status_t alp_prv_notify_broadcast(const char * appID, const char * notifyType, const void * details, int length, int queue, int flags, int timeout, int * handled)
{
	AlpNotifyCmdBroadcast request;
	alp_status_t err;

#ifdef ALP_NOTIFY_DEBUG
	printf("alp_prv_notify_broadcast\n");
#endif
	// check for errors
	if (!appID || strlen(appID) > ALP_NOTIFY_MAX_APPID_SIZE)
		return ALP_NOTIFY_ERR_INVALID_APPID;
	if (!notifyType || strlen(notifyType) > ALP_NOTIFY_MAX_NOTIFY_SIZE)
		return ALP_NOTIFY_ERR_INVALID_NOTIFYTYPE;
	if (!queue)
		return ALP_NOTIFY_ERR_INVALID_QUEUE;
	if (length > ALP_NOTIFY_MAX_DETAILS_SIZE || length < 0)
		return ALP_NOTIFY_ERR_INVALID_LENGTH;

	// fill out the message header
	memset(&request,0,sizeof(AlpNotifyCmdBroadcast));
	request.cmd = ALP_NOTIFY_CMD_BROADCAST;
	request.size = sizeof(AlpNotifyCmdBroadcast);

	// fill out the message body
	strcpy(request.notifyType,notifyType);
	request.length = length;
	request.flags = flags;
	request.queue = queue;
	strcpy(request.appID,appID);
	if (length) memmove(request.details,details,length);
	request.zero = 0;

	// send the message and wait for a response (or two for sync broadcasts)
	err = alp_notify_client_send(&request,request.size,flags & ALP_NOTIFY_BROADCAST_FLAG_SYNC ? 1 : 0,timeout,handled);

	// not registered shouldn't be returned as an error
	if(err == ALP_NOTIFY_ERR_NOT_REGISTERED) {
       		err = ALP_OK;
		if (handled) *handled = 0;
	}

	// return status
	return err;
}

alp_status_t alp_notify_broadcast(const char * appID, const char * notifyType, const void * details, int length, int queue, int flags)
{
	return alp_prv_notify_broadcast(appID, notifyType, details, length, queue, flags, 0, 0);
}

alp_status_t alp_notify_broadcast_sync(const char * appID, const char * notifyType, const void * details, int length, int queue, int flags, int timeout, int * handled)
{
	return alp_prv_notify_broadcast(appID, notifyType, details, length, queue, flags | ALP_NOTIFY_BROADCAST_FLAG_SYNC, timeout, handled);
}

alp_status_t alp_notify_done(const char * appID, int handled, int status)
{
	AlpNotifyCmdNotificationDone request;

#ifdef ALP_NOTIFY_DEBUG
	printf("alp_notify_done\n");
#endif
	// check for errors
	if (!appID || strlen(appID) > ALP_NOTIFY_MAX_APPID_SIZE)
		return ALP_NOTIFY_ERR_INVALID_APPID;

	// fill out the message header
	memset(&request,0,sizeof(AlpNotifyCmdNotificationDone));
	request.cmd = ALP_NOTIFY_CMD_NOTIFY_DONE;
	request.size = sizeof(AlpNotifyCmdNotificationDone);

	// fill out the message body
	request.handled = handled;
	request.status = status;
	strcpy(request.appID,appID);
	request.zero = 0;

	// send the message and wait for a response
	return alp_notify_client_send(&request,request.size,0,0,0);
}

static void alp_prv_notify_client_do_callback(AlpNotifyCmdNotification * request, int length)
{
	AlpNotifyCallbackProc * callback;
	int status;

	callback = (AlpNotifyCallbackProc *)request->callback;
	status = (callback)(request->notifyType,request->details,request->length,&request->handled,request->appID,request->userData);
	if (!(status & ALP_NOTIFY_CALLBACK_COMPLETED_ALREADY))
		alp_notify_done(request->appID2,request->handled,status);
}

void alp_notify_client_mesg(int * buf, int length)
{
	switch(*buf) {
		case ALP_NOTIFY_CMD_NOTIFY:
			ALP_FAIL_NON_FATAL_IF(length != sizeof(AlpNotifyCmdNotification),"notification callback message is wrong size\n");
			if (length == sizeof(AlpNotifyCmdNotification)) alp_prv_notify_client_do_callback((AlpNotifyCmdNotification *)buf,length);
		break;
		default:
		break;
	}
	ALP_FAIL_NON_FATAL_IF(*buf != ALP_NOTIFY_CMD_NOTIFY,"received unknown message type\n");
}

// Decoding (not from RFC 3548). Assumes standard ASCII
static const unsigned char sDecodeBase64[] = {
	/*          0   1   2   3   4   5   6   7   8   9 */
	/*  00 */  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	/*  10 */  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	/*  20 */  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	/*  30 */  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	/*  40 */  -1, -1, -1, 62, -1, -1, -1, 63, 52, 53,

	/*          0   1   2   3   4   5   6   7   8   9 */
	/*  50 */  54, 55, 56, 57, 58, 59, 60, 61, -1, -1,
	/*  60 */  -1,  0, -1, -1, -1,  0,  1,  2,  3,  4,
	/*  70 */   5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
	/*  80 */  15, 16, 17, 18, 19, 20, 21, 22, 23, 24,
	/*  90 */  25, -1, -1, -1, -1, -1, -1, 26, 27, 28,
	/* 100 */  29, 30, 31, 32, 33, 34, 35, 36, 37, 38,

	/*          0   1   2   3   4   5   6   7   8   9 */
	/* 110 */  39, 40, 41, 42, 43, 44, 45, 46, 47, 48,
	/* 120 */  49, 50, 51, -1, -1, -1, -1, -1, -1, -1,
	/* 130 */  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	/* 140 */  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	/* 150 */  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,

	/*          0   1   2   3   4   5   6   7   8   9 */
	/* 160 */  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	/* 170 */  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	/* 180 */  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	/* 190 */  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	/* 200 */  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,

	/*          0   1   2   3   4   5   6   7   8   9 */
	/* 210 */  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	/* 220 */  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	/* 230 */  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	/* 240 */  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	/* 250 */  -1, -1, -1, -1, -1, -1
};

// Decodes four bytes into one, two, or three bytes
// Assumes length is four bytes.
// Returns the number of decoded bytes
static int alp_prv_notify_base64_decode_block(unsigned char src[], unsigned char dst[])
{
	// first character
	dst[0] = (sDecodeBase64[src[0]] << 2) | ((sDecodeBase64[src[1]] & 0x30) >> 4);
	if (src[2] == '=') return 1;

	// second character
	dst[1] = ((sDecodeBase64[src[1]] & 0x0F) << 4) | ((sDecodeBase64[src[2]] & 0x3C) >> 2);
	if (src[3] == '=') return 2;

	// third character
	dst[2] = ((sDecodeBase64[src[2]] & 0x03) << 6) | (sDecodeBase64[src[3]] & 0x3F);
	return 3;
}

// Decode a buffer of data
// Caller must allocate the dst buffer
// Returns the actual number of decoded bytes
static int alp_prv_notify_base64_decode(unsigned char * src, unsigned char * dst, int len)
{
	int count = 0;
	while (len > 0) {
		count += alp_prv_notify_base64_decode_block(src,dst);
	        src += 4;
		dst += 3;
		len -= 4;
	}

	return count;
}

int alp_notify_details(const char * src, AlpNotifyLaunch * dst)
{
	return alp_prv_notify_base64_decode((unsigned char *)src + strlen(ALP_NOTIFY_LAUNCH_DETAILS),(unsigned char *)dst,strlen(src) - strlen(ALP_NOTIFY_LAUNCH_DETAILS));
}
