/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
#include <hiker/unit_test.h>
#include <cppunit/extensions/HelperMacros.h>

// faster CPPUNIT_ASSERT
#define ALP_CPPUNIT_ASSERT(x) if (!(x)) CPPUNIT_ASSERT(x)

// more detailed err message
#define CPPUNIT_ASSERT_FMT( X, fmt, ...) {\
	char cppunit_tmp[256]; \
	sprintf(cppunit_tmp,fmt,__VA_ARGS__); \
	CPPUNIT_ASSERT_MESSAGE( cppunit_tmp, X ); }

#define CPPUNIT_ERR( got, shouldbe ) CPPUNIT_ASSERT_FMT( got == shouldbe, "expected alp_status_t:%0X but got %0X",shouldbe,got)


#include <hiker/notifymgr.h>

#include <signal.h>

class NotifyMgrTests : public CPPUNIT_NS::TestFixture
{
	CPPUNIT_TEST_SUITE(NotifyMgrTests);
	// CPPUNIT_TEST(StartupNotifyMgrServer);

	CPPUNIT_TEST(ServerInitPositiveTest);
	CPPUNIT_TEST(ServerAsyncBroadcastPositiveTest);
	CPPUNIT_TEST(ServerSyncBroadcastPositiveTest);
	CPPUNIT_TEST(ServerAsyncBroadcastDataPositiveTest);
	CPPUNIT_TEST(ServerSyncBroadcastDataPositiveTest);
	CPPUNIT_TEST(ServerBroadcastFilterPositiveTest);
	CPPUNIT_TEST(ServerSyncBroadcastPriorityPositiveTest);

	CPPUNIT_TEST(AppRegisterPositiveTest);
	CPPUNIT_TEST(AppDonePositiveTest);

	CPPUNIT_TEST(BroadcastNegativeTest);
	CPPUNIT_TEST(RegisterNegativeTest);
	CPPUNIT_TEST(UnregisterNegativeTest);
	CPPUNIT_TEST(NotifyDoneNegativeTest);

	CPPUNIT_TEST(Base64PositiveTest1);
	CPPUNIT_TEST(Base64PositiveTest2);
	CPPUNIT_TEST(Base64PositiveTest3);

	// CPPUNIT_TEST(TeardownNotifyMgrServer);
	CPPUNIT_TEST_SUITE_END();

	protected:
		void StartupNotifyMgrServer();
		void ServerInitPositiveTest();
		void ServerAsyncBroadcastPositiveTest();
		void ServerSyncBroadcastPositiveTest();
		void ServerAsyncBroadcastDataPositiveTest();
		void ServerSyncBroadcastDataPositiveTest();
		void ServerBroadcastFilterPositiveTest();
		void ServerSyncBroadcastPriorityPositiveTest();

		void AppRegisterPositiveTest();
		void AppDonePositiveTest();

		void BroadcastNegativeTest();
		void RegisterNegativeTest();
		void UnregisterNegativeTest();
		void NotifyDoneNegativeTest();

		void Base64PositiveTest1();
		void Base64PositiveTest2();
		void Base64PositiveTest3();
		void TeardownNotifyMgrServer();
	public:

	private:
		static pid_t sNotifyMgrServerPID;
};

pid_t NotifyMgrTests::sNotifyMgrServerPID = 0;
static int sNotifyMgrCallback, sNotifyMgrDataCheck;
static void * sNotifyMgrUserData;

ALP_ROOT_TEST_SUITE_REGISTRATION("OS-007", CPPUNIT_REGISTRY_TYPE_AUTO);
CPPUNIT_TEST_SUITE_NAMED_REGISTRATION(NotifyMgrTests, "OS-007");

// copied from the appserver
void NotifyMgrTests::StartupNotifyMgrServer()
{
	char *top;
	char *argv[2];

	// remove any previous settings
	unlink("/tmp/notifymgr.db");

	// First do some stuff that might fail, since we can't assert after forking.
	top = getenv("ALP_RUN_TOP");
	CPPUNIT_ASSERT(top != NULL);
	
	// Fork child process
	sNotifyMgrServerPID = fork();
	
	if(sNotifyMgrServerPID == 0)
	{
		// Child process, exec the notifymgr server.
		int size = strlen(top);
		size += strlen("/opt/alp/bin/notifymgr") + 1;
	
		char *runme;
		runme = (char *)malloc(sizeof(char) * size);
		strcpy(runme, top);
		strcat(runme, "/opt/alp/bin/notifymgr");
	
		argv[0] = "notifymgr";
		argv[1] = NULL;
		execv(runme, argv);
		
		// execution should never reach here...
		CPPUNIT_ASSERT(false);

		// exit if the exev fails
		exit(0);
	} else {
		// Parent process. Sleep for a moment to give the server a chance to start up.
		sleep(2);
	}
}

void NotifyMgrTests::TeardownNotifyMgrServer()
{
	// remove notifymgr server
	if (sNotifyMgrServerPID) kill(sNotifyMgrServerPID, SIGKILL);

	// remove any previous settings
	unlink("/tmp/notifymgr.db");
}

// -------------------------------------------------------------------------------------------------
// common utils
// -------------------------------------------------------------------------------------------------

void delay(int div)
{
	struct timespec time;

	time.tv_sec = 0;
	time.tv_nsec = 1000000000 / div;
	nanosleep(&time,NULL);
}

// -------------------------------------------------------------------------------------------------
// fast callback
// -------------------------------------------------------------------------------------------------

int callback_fast(const char * notifyType, void * details, int length, int * handled, const char * appID, void * userData)
{
	++*handled;
	sNotifyMgrCallback++;
	return 0;
}

/**
 * @brief API tested: alp_notify_register_server().
 *
 * @description check the first time alp_notify_register_server is called.
 */
void NotifyMgrTests::ServerInitPositiveTest()
{
	alp_status_t err;
	int handled = 0;

	// initialize
	sNotifyMgrCallback = 0;

	// register the callback
	err = alp_notify_register_server("notify0",callback_fast,NULL,0,0);
	CPPUNIT_ASSERT(err == ALP_OK);

	// broadcast
	err = alp_notify_broadcast_sync("unit test","notify0",0,0,ALP_NOTIFY_QUEUE_DEFAULT,0,0,&handled);
	CPPUNIT_ASSERT(err == ALP_OK);
	CPPUNIT_ASSERT(handled == 1);

	// unregister the callback
	err = alp_notify_unregister_server("notify0",0);
	CPPUNIT_ASSERT(err == ALP_OK);

	// make sure the callback was called
	CPPUNIT_ASSERT(sNotifyMgrCallback == 1);
}

/**
 * @brief API tested: alp_notify_register_server(), alp_notify_broadcast().
 *
 * @description check async broadcasts to the server.
 */
void NotifyMgrTests::ServerAsyncBroadcastPositiveTest()
{
	alp_status_t err;
	int i;

	// initialize
	sNotifyMgrCallback = 0;

	// register the callback
	err = alp_notify_register_server("notify1",callback_fast,NULL,0,0);
	CPPUNIT_ASSERT(err == ALP_OK);

	// broadcast ten times
	for (i = 0 ; i < 10 ; i++) {
		err = alp_notify_broadcast("unit test","notify1",0,0,ALP_NOTIFY_QUEUE_DEFAULT,0);
		CPPUNIT_ASSERT(err == ALP_OK);
		delay(4);
	}

	// unregister the callback
	do {
		delay(4);
		err = alp_notify_unregister_server("notify1",0);
		CPPUNIT_ASSERT(err == ALP_OK || err == ALP_NOTIFY_ERR_ACTIVE_BROADCAST);
	} while (err == ALP_NOTIFY_ERR_ACTIVE_BROADCAST);

	// make sure the callback was called
	CPPUNIT_ASSERT(sNotifyMgrCallback == 10);
}

/**
 * @brief API tested: alp_notify_register_server(), alp_notify_broadcast_sync().
 *
 * @description check sync broadcasts to the server
 */
void NotifyMgrTests::ServerSyncBroadcastPositiveTest()
{
	alp_status_t err;
	int handled, i;

	// initialize
	sNotifyMgrCallback = 0;

	// register the callback
	err = alp_notify_register_server("notify2",callback_fast,NULL,0,0);
	CPPUNIT_ASSERT(err == ALP_OK);

	// sync broadcast ten times
	for (i = 0 ; i < 10 ; i++) {
		handled = 0;
		err = alp_notify_broadcast_sync("unit test","notify2",0,0,ALP_NOTIFY_QUEUE_DEFAULT,0,0,&handled);
		CPPUNIT_ASSERT(err == ALP_OK);
		CPPUNIT_ASSERT(handled == 1);
	}

	// unregister the callback (don't need to wait)
	err = alp_notify_unregister_server("notify2",0);
	CPPUNIT_ASSERT(err == ALP_OK);

	// make sure the callback was called
	CPPUNIT_ASSERT(sNotifyMgrCallback == 10);
};

/**
 * @brief API tested: alp_notify_register_server(), alp_notify_broadcast_sync().
 *
 * @description check broadcasts filtering
 */
void NotifyMgrTests::ServerBroadcastFilterPositiveTest()
{
	alp_status_t err;

	// initialize
	sNotifyMgrCallback = 0;

	// register the callback
	err = alp_notify_register_server("/test/cppunit/",callback_fast,NULL,0,0);
	CPPUNIT_ASSERT(err == ALP_OK);

	// prefix match
	err = alp_notify_broadcast("unit test","/test/cppunit/notify0",0,0,ALP_NOTIFY_QUEUE_DEFAULT,0);
	CPPUNIT_ASSERT(err == ALP_OK);

	// exact match
	err = alp_notify_broadcast("unit test","/test/cppunit/",0,0,ALP_NOTIFY_QUEUE_DEFAULT,0);
	CPPUNIT_ASSERT(err == ALP_OK);

	// no match
	err = alp_notify_broadcast("unit test","/test/",0,0,ALP_NOTIFY_QUEUE_DEFAULT,0);
	CPPUNIT_ASSERT(err == ALP_OK);
	delay(4);

	// unregister
	do {
		delay(4);
		err = alp_notify_unregister_server("/test/cppunit/",0);
		CPPUNIT_ASSERT(err == ALP_OK || err == ALP_NOTIFY_ERR_ACTIVE_BROADCAST);
	} while (err == ALP_NOTIFY_ERR_ACTIVE_BROADCAST);

	// make sure the callback was called
	CPPUNIT_ASSERT(sNotifyMgrCallback == 2);
};

// -------------------------------------------------------------------------------------------------
// data checking callback
// -------------------------------------------------------------------------------------------------

int callback_data(const char * notifyType, void * details, int length, int * handled, const char * appID,void * userData)
{
	++*handled;
	if (length == 3 && !strncmp("one",(char *)details,3)) sNotifyMgrCallback++;
    sNotifyMgrUserData = userData;
	return 0;
}


/**
 * @brief API tested: alp_notify_register_server(), alp_notify_broadcast().
 *
 * @description check async broadcasts to the server with data
 */
void NotifyMgrTests::ServerAsyncBroadcastDataPositiveTest()
{
	alp_status_t err;
	int i;

    char test_data[5];

	// initialize
	sNotifyMgrCallback = 0;
    sNotifyMgrUserData = 0;

	// register the callback
	err = alp_notify_register_server("notify3",callback_data,test_data,0,0);
	CPPUNIT_ASSERT(err == ALP_OK);

	// broadcast three times
	for (i = 0 ; i < 3 ; i++) {
		err = alp_notify_broadcast("unit test","notify3",(void *)"one",3,ALP_NOTIFY_QUEUE_DEFAULT,0);
		CPPUNIT_ASSERT(err == ALP_OK);
		delay(4);
        CPPUNIT_ASSERT(sNotifyMgrUserData == test_data);
        sNotifyMgrUserData = 0;
	}

	// unregister the callback
	do {
		delay(4);
		err = alp_notify_unregister_server("notify3",0);
		CPPUNIT_ASSERT(err == ALP_OK || err == ALP_NOTIFY_ERR_ACTIVE_BROADCAST);
	} while (err == ALP_NOTIFY_ERR_ACTIVE_BROADCAST);

	// make sure the callback was called
	CPPUNIT_ASSERT(sNotifyMgrCallback == 3);
}

/**
 * @brief API tested: alp_notify_register_server(), alp_notify_broadcast_sync().
 *
 * @description check broadcasts to the server with data
 */
void NotifyMgrTests::ServerSyncBroadcastDataPositiveTest()
{
	alp_status_t err;
	int handled, i;

	// initialize
	sNotifyMgrCallback = 0;

	// register the callback
	err = alp_notify_register_server("notify4",callback_data,NULL,0,0);
	CPPUNIT_ASSERT(err == ALP_OK);

	// sync broadcast three times
	for (i = 0 ; i < 3 ; i++) {
		handled = 0;
		err = alp_notify_broadcast_sync("unit test","notify4",(void *)"one",3,ALP_NOTIFY_QUEUE_DEFAULT,0,0,&handled);
		CPPUNIT_ASSERT(err == ALP_OK);
		CPPUNIT_ASSERT(handled == 1);
	}

	// unregister the callback
	err = alp_notify_unregister_server("notify4",0);
	CPPUNIT_ASSERT(err == ALP_OK);

	// make sure the callback was called
	CPPUNIT_ASSERT(sNotifyMgrCallback == 3);
};

// -------------------------------------------------------------------------------------------------
// priority checking callback
// -------------------------------------------------------------------------------------------------

int callback_low(const char * notifyType, void * details, int length, int * handled, const char * appID, void * userData)
{
	printf("callback_low\n");
	++*handled;
	sNotifyMgrCallback *= 3;
	return 0;
}

int callback_hi(const char * notifyType, void * details, int length, int * handled, const char * appID, void * userData)
{
	printf("callback_hi\n");
	++*handled;
	sNotifyMgrCallback += 1;
	return 0;
}

/**
 * @brief API tested: alp_notify_register_server(), alp_notify_broadcast_sync().
 *
 * @description check broadcasts to the server with registration priority
 */
void NotifyMgrTests::ServerSyncBroadcastPriorityPositiveTest()
{
	alp_status_t err;
	int handled;

	// initialize
	sNotifyMgrCallback = 0;

	// register the callbacks
	err = alp_notify_register_server("/notify15 low",callback_low,NULL,ALP_NOTIFY_PRIORITY_LOW,0);
	CPPUNIT_ASSERT(err == ALP_OK);

	err = alp_notify_register_server("/notify15 hi",callback_hi,NULL,ALP_NOTIFY_PRIORITY_HIGH,0);
	CPPUNIT_ASSERT(err == ALP_OK);

	// broadcast
	err = alp_notify_broadcast_sync("unit test","/notify15",(void *)"one",3,ALP_NOTIFY_QUEUE_DEFAULT,0,0,&handled);
	CPPUNIT_ASSERT(err == ALP_OK);

	// unregister the callback
	err = alp_notify_unregister_server("/notify15 low",0);
	CPPUNIT_ASSERT(err == ALP_OK);

	err = alp_notify_unregister_server("/notify15 hi",0);
	CPPUNIT_ASSERT(err == ALP_OK);

	// make sure the callback was called in the correct order
	CPPUNIT_ASSERT(sNotifyMgrCallback == 3);

	// make sure the two callback was called
	CPPUNIT_ASSERT(handled == 2);
}

// -------------------------------------------------------------------------------------------------
// app tests
// -------------------------------------------------------------------------------------------------

/**
 * @brief API tested: alp_notify_register().
 *
 * @description register an application
 */
void NotifyMgrTests::AppRegisterPositiveTest()
{
	alp_status_t err;

	// register app
	err = alp_notify_register("bar:com.access.apps.attn_alert","notify5",0,0);
	CPPUNIT_ASSERT(err == ALP_OK);

	// unregister app
	err = alp_notify_unregister("bar:com.access.apps.attn_alert","notify5",0);
	CPPUNIT_ASSERT(err == ALP_OK);
}

/**
 * @brief API tested: alp_notify_done().
 *
 * @description 
 */
void NotifyMgrTests::AppDonePositiveTest()
{
	alp_status_t err;

	// notification done
	err = alp_notify_done("bar:com.access.apps.simple_launcher",0,0);
	CPPUNIT_ASSERT(err == ALP_OK);
}

// -------------------------------------------------------------------------------------------------
// negative tests
// -------------------------------------------------------------------------------------------------

/**
 * @brief Negative test for alp_notify_broadcast() and alp_notify_broadcast_sync().
 *
 * @description register an app with an invalid queue
 */
void NotifyMgrTests::BroadcastNegativeTest()
{
	alp_status_t err;
	int handled = 0;

	// invalid appID
	err = alp_notify_broadcast(0,"notify6",0,0,ALP_NOTIFY_QUEUE_DEFAULT,0);
	CPPUNIT_ASSERT(err == ALP_NOTIFY_ERR_INVALID_APPID);

	// invalid notifytype
	err = alp_notify_broadcast("unit test",NULL,0,0,ALP_NOTIFY_QUEUE_DEFAULT,0);
	CPPUNIT_ASSERT(err == ALP_NOTIFY_ERR_INVALID_NOTIFYTYPE);

	// invalid queue
	err = alp_notify_broadcast("unit test","notify7",0,0,0,0);
	CPPUNIT_ASSERT(err == ALP_NOTIFY_ERR_INVALID_QUEUE);

	// invalid length
	err = alp_notify_broadcast("unit test","notify7",(void *)"details",-3,ALP_NOTIFY_QUEUE_DEFAULT,0);
	CPPUNIT_ERR(err, ALP_NOTIFY_ERR_INVALID_LENGTH);

	// invalid appID
	err = alp_notify_broadcast_sync(0,"notify6",0,0,ALP_NOTIFY_QUEUE_DEFAULT,0,0,&handled);
	CPPUNIT_ASSERT(err == ALP_NOTIFY_ERR_INVALID_APPID);

	// invalid notifytype
	err = alp_notify_broadcast_sync("unit test",NULL,0,0,ALP_NOTIFY_QUEUE_DEFAULT,0,0,&handled);
	CPPUNIT_ASSERT(err == ALP_NOTIFY_ERR_INVALID_NOTIFYTYPE);

	// invalid queue
	err = alp_notify_broadcast_sync("unit test","notify7",0,0,0,0,0,&handled);
	CPPUNIT_ASSERT(err == ALP_NOTIFY_ERR_INVALID_QUEUE);

	// invalid length
	err = alp_notify_broadcast_sync("unit test","notify7",(void *)"details",-3,
	      ALP_NOTIFY_QUEUE_DEFAULT,0,0,&handled);
	CPPUNIT_ERR(err, ALP_NOTIFY_ERR_INVALID_LENGTH);
}


/**
 * @brief Negative Tests for alp_notify_register_server() and alp_notify_register().
 * 
 * @description	Register to receive a notification, then register to receive it again.
 * 		expected return ALP_NOTIFY_ERR_ALREADY_REGISTERED
 */
void NotifyMgrTests::RegisterNegativeTest()
{
	alp_status_t err;

	// invalid appID
	err = alp_notify_register(0,"notify8",0,0);
	CPPUNIT_ASSERT(err == ALP_NOTIFY_ERR_INVALID_APPID);

	// invalid callback
	err = alp_notify_register_server("notify9",0,NULL,0,0);
	CPPUNIT_ASSERT(err == ALP_NOTIFY_ERR_INVALID_CALLBACK);

	// invalid notifytype 
	err = alp_notify_register_server(NULL,callback_fast,NULL,0,0);
	CPPUNIT_ASSERT(err == ALP_NOTIFY_ERR_INVALID_NOTIFYTYPE);

	err = alp_notify_register("bar:com.access.apps.simple_launcher",NULL,0,0);
	CPPUNIT_ASSERT(err == ALP_NOTIFY_ERR_INVALID_NOTIFYTYPE);

	// invalid priority
//	err = alp_notify_register_server("notify9",callback_fast,NULL,-100,0);
//	CPPUNIT_ERR(err, ALP_NOTIFY_ERR_INVALID_PRIORITY);

//	err = alp_notify_register("bar:com.access.apps.simple_launcher","notify8",100,0);
//	CPPUNIT_ASSERT(err == ALP_NOTIFY_ERR_INVALID_PRIORITY);

	// register for a callback
	err = alp_notify_register_server("notify10", callback_fast, NULL, 0, 0);
	CPPUNIT_ASSERT(err == ALP_OK);

	err = alp_notify_register_server("notify10", callback_fast, NULL, 0, 0);
	CPPUNIT_ASSERT(err == ALP_NOTIFY_ERR_ALREADY_REGISTERED);

	// unregister server
	err = alp_notify_unregister_server("notify10", 0);
	CPPUNIT_ASSERT(err == ALP_OK);

	// register app
	err = alp_notify_register("bar:com.access.apps.attn_alert","notify11",0,0);
	CPPUNIT_ASSERT(err == ALP_OK);

	// register app again
	err = alp_notify_register("bar:com.access.apps.attn_alert","notify11",0,0);
	CPPUNIT_ASSERT(err == ALP_NOTIFY_ERR_ALREADY_REGISTERED);

	// unregister app
	err = alp_notify_unregister("bar:com.access.apps.attn_alert","notify11",0);
	CPPUNIT_ASSERT(err == ALP_OK);
}


/**
 * @brief Negative Tests for alp_notify_unregister_server() and alp_notify_unregister().
 * 
 * @description	Unregister from receiving a notification without registering for it first.
 * 		expected return ALP_NOTIFY_ERR_NOT_REGISTERED
 */
void NotifyMgrTests::UnregisterNegativeTest()
{
	alp_status_t err;

	// invalid appID
	err = alp_notify_unregister(0,"notify12",0);
	CPPUNIT_ASSERT(err == ALP_NOTIFY_ERR_INVALID_APPID);

	// invalid notifytype 
	err = alp_notify_unregister("bar:com.access.apps.simple_launcher",NULL,0);
	CPPUNIT_ASSERT(err == ALP_NOTIFY_ERR_INVALID_NOTIFYTYPE);

	err = alp_notify_unregister_server(NULL,0);
	CPPUNIT_ASSERT(err == ALP_NOTIFY_ERR_INVALID_NOTIFYTYPE);

	// unregister server
	err = alp_notify_unregister_server("notify13", 0);
	CPPUNIT_ASSERT(err == ALP_NOTIFY_ERR_NOT_REGISTERED);

	// unregister app
	err = alp_notify_unregister("bar:com.access.apps.simple_launcher","notify14",0);
	CPPUNIT_ASSERT(err == ALP_NOTIFY_ERR_NOT_REGISTERED);
}

/**
 * @brief Negative Tests for alp_notify_done().
 *
 * @description Call the api with an invalid appID.
 * 		expected return ALP_NOTIFY_ERR_INVALID_APPID
 */
void NotifyMgrTests::NotifyDoneNegativeTest()
{
	alp_status_t err;

	err = alp_notify_done(0,0,0);
	CPPUNIT_ASSERT(err == ALP_NOTIFY_ERR_INVALID_APPID);

}
// -------------------------------------------------------------------------------------------------
// base 64 tests
// -------------------------------------------------------------------------------------------------

// Encoding from RFC 3548 among others 
static const unsigned char encodeBase64[]="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

// Encodes the one, two, or third bytes into four bytes (padded if necessary)
// Assumes length is at least one byte.
void base64_encode_block(unsigned char src[], unsigned char dst[], int length)
{
	// first character
	dst[0] = encodeBase64[ src[0] >> 2 ];

	// second character
	if (length > 1)
		dst[1] = encodeBase64[ (src[0] & 0x03) << 4 | (src[1] & 0xf0) >> 4 ];
	else
		dst[1] = encodeBase64[ (src[0] & 0x03) << 4 ];

	// third character
	if (length > 1)
		dst[2] = encodeBase64[ (src[1] & 0x0f) << 2 | (src[2] & 0xc0) >> 6 ];
	else
		dst[2] = '=';

	// fourth character
	if (length > 2)
		dst[3] = encodeBase64[ src[2] & 0x3f ];
	else
		dst[3] = '=';
}

// Encode a buffer of data
// Caller must allocate the dst buffer
void base64_encode(unsigned char * src, unsigned char * dst, int len)
{
	while (len > 0) {
		base64_encode_block(src,dst,len < 3 ? len : 3);
	        src += 3;
		dst += 4;
		len -= 3;
	}
}

#define ALP_NOTIFY_TESTS 40

void NotifyMgrTests::Base64PositiveTest1()
{
	int i,j = 0;
	char argv[ALP_NOTIFY_TESTS];

	memset(argv,0,ALP_NOTIFY_TESTS);
	strcpy(argv,ALP_NOTIFY_LAUNCH_DETAILS);

	for (i = 0 ; i < 256 ; i++) {
		base64_encode((unsigned char *)&i,(unsigned char *)&argv[strlen(ALP_NOTIFY_LAUNCH_DETAILS)],1);
		alp_notify_details(argv,(AlpNotifyLaunch *)&j);
		ALP_CPPUNIT_ASSERT(i == j);
	}
}
void NotifyMgrTests::Base64PositiveTest2()
{
	int i,j = 0;
	char argv[ALP_NOTIFY_TESTS];

	memset(argv,0,ALP_NOTIFY_TESTS);
	strcpy(argv,ALP_NOTIFY_LAUNCH_DETAILS);

	for (i = 0 ; i < 256*256 ; i++) {
		base64_encode((unsigned char *)&i,(unsigned char *)&argv[strlen(ALP_NOTIFY_LAUNCH_DETAILS)],2);
		alp_notify_details(argv,(AlpNotifyLaunch *)&j);
		ALP_CPPUNIT_ASSERT(i == j);
	}
}
void NotifyMgrTests::Base64PositiveTest3()
{
	int i,j = 0;
	char argv[ALP_NOTIFY_TESTS];

	memset(argv,0,ALP_NOTIFY_TESTS);
	strcpy(argv,ALP_NOTIFY_LAUNCH_DETAILS);

	// maybe every seventh for better performance
	for (i = 0 ; i < 256*256*256 ; i += 7) {
		base64_encode((unsigned char *)&i,(unsigned char *)&argv[strlen(ALP_NOTIFY_LAUNCH_DETAILS)],3);
		alp_notify_details(argv,(AlpNotifyLaunch *)&j);
		ALP_CPPUNIT_ASSERT(i == j);
	}
}
