/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/*! @SAL.cpp
 *
 */
#include <stdarg.h>
#include <hiker/sal.h>
#include "sal_prv.h"


/*! Opens a given database handle.  The database should be closed when
 * no longer needed with alp_sal_database_close()
 *
 * @param [out] sal Address of a pointer to the AlpSALHandle
 * @param [in] dbName Filename of the SQL database
 *
 * @return alp_status_t
 *
 * @see alp_sal_database_close
 */
alp_status_t alp_sal_database_open(AlpSALHandle *sal, const char *dbName)
{
	int32_t sqlErr;

	if (!sal) {
		return ALP_STATUS_SAL_BADPARAM;
	}

	sqlErr = sqlite3_open(dbName, (sqlite3 **)sal);
	if (sqlErr) {
        return alp_prv_sal_translate_sqlite_err(sqlErr);
	}

	/* Rollback in case database was previously left open inside an incomplete
	 * transaction.  Error code can be ignored. */
	sqlErr = alp_sal_transaction_rollback(sal);

	return ALP_STATUS_OK;
}


/*! Closes a given database handle.
 *
 * @param [in] sal The AlpSALHandle to close
 *
 * @return alp_status_t
 *
 * @see alp_sal_database_open
 */
alp_status_t alp_sal_database_close(AlpSALHandle sal)
{
	int32_t sqlErr;

	if (!sal) {
		return ALP_STATUS_SAL_BADPARAM;
	}

	sqlErr = sqlite3_close((sqlite3 *)sal);

    return alp_prv_sal_translate_sqlite_err(sqlErr);
}


/*! Creates a database and initializes its schema.
 *
 * @param [in] dbFileName Path to file to store the created database
 *
 * @return alp_status_t
 */
alp_status_t alp_sal_database_create(const char *dbFileName, const char *schemaFileName)
{
	alp_status_t salErr;
	AlpSALHandle sal;
	FILE *fp;
	int32_t length = 0;
	const char *errMsg;

	/* If dbFileName exists, return */
	if (alp_prv_sal_file_exists(dbFileName)) {
	    return ALP_STATUS_SAL_DMLEXISTS;
	}

	/* Open database (creating it, if it didn't exist) */
	salErr = alp_sal_database_open(&sal, dbFileName);

	if (salErr) {
		/* sqlite3 docs say we need to close the db even when the open errs */
		alp_sal_database_close(sal);
		return salErr;
	}

	/* Execute schema script */
	if ((fp = fopen(schemaFileName, "r")) == NULL) {
		return ALP_STATUS_SAL_INTERNAL;
	}
	salErr = alp_sal_execute_sql_script(sal, fp);
	if (salErr) {
		errMsg = alp_sal_get_error_message(sal);
		fclose(fp);
		return salErr;
	}
	fclose(fp);

	/* Close database */
	salErr = alp_sal_database_close(sal);

	return salErr;
}


/*! Executes "COMMIT TRANSACTION;" on the given Database.
 *
 * @param[in] sal Database to execute against.
 *
 * @return alp_status_t
 */
alp_status_t alp_sal_transaction_commit(AlpSALHandle sal)
{
	char *sql, *errMsg = NULL;
	int32_t sqlErr;

    sql = sqlite3_mprintf("COMMIT TRANSACTION;");
	if (!sql) {
	    return ALP_STATUS_SAL_NOMEM;
	}
    sqlErr = sqlite3_exec((sqlite3 *)sal, sql, NULL, NULL, &errMsg);
	if (sqlErr && *errMsg) {
		free(errMsg);
	}
	sqlite3_free(sql);
    if (sqlErr) {
        return alp_prv_sal_translate_sqlite_err(sqlErr);
    }

	return ALP_STATUS_OK;
}


/*! Executes "BEGIN TRANSACTION;" on the given Database.
 *
 * @param[in] sal Database to execute against.
 * @param[in] exclusive flag indicating whether to use
 *     "BEGIN EXCLUSIVE TRANSACTION;"
 *
 * @return alp_status_t
 */
alp_status_t alp_sal_transaction_begin(AlpSALHandle sal, char exclusive)
{
	char *sql = NULL, *errMsg = NULL;
	int sqlErr;

    sql = sqlite3_mprintf("BEGIN %qTRANSACTION;",
		exclusive ? "EXCLUSIVE " : "");
	if (!sql) {
	    return ALP_STATUS_SAL_NOMEM;
	}
    sqlErr = sqlite3_exec((sqlite3 *)sal, sql, NULL, NULL, &errMsg);
	if (sqlErr && *errMsg) {
		free(errMsg);
	}
	sqlite3_free(sql);
    if (sqlErr) {
        return alp_prv_sal_translate_sqlite_err(sqlErr);
    }

	return ALP_STATUS_OK;
}


/*! Executes "ROLLBACK TRANSACTION;" on the given Database.
 *
 * @param [in] sal Database to execute against.
 *
 * @return alp_status_t
 */
alp_status_t alp_sal_transaction_rollback(AlpSALHandle sal)
{
	char *sql = NULL, *errMsg = NULL;
	int sqlErr;

	sql = sqlite3_mprintf("ROLLBACK TRANSACTION;");
	if (!sql) {
		return ALP_STATUS_SAL_NOMEM;
	}
	sqlErr = sqlite3_exec((sqlite3 *)sal, sql, NULL, NULL, &errMsg);
	if (sqlErr && *errMsg) {
		free(errMsg);
	}
	sqlite3_free(sql);

    if (sqlErr) {
        return alp_prv_sal_translate_sqlite_err(sqlErr);
    }

	return ALP_STATUS_OK;
}


/*! Given a SQL statement, usually of the form "SELECT COUNT(*) FROM..."
 * this function will prepare the statement from the SQL string and given
 * host variables, if any, then execute the statement and read back the first
 * column of the first row returned in the result set and assign that to value.
 *
 * @param [in] sal The database to execute the selection statement against
 * @param [out] value Will obtain the result from the selection.
 * @param [in] va_list Host parameters, if any to plug into sql statement
 *
 * @return alp_status_t
 */
alp_status_t alp_sal_sql_get_value(AlpSALHandle sal, int32_t *value,
	const char *sql, ...)
{
	va_list ap;
	char *dummy;
	int sqlErr;
	sqlite3_stmt *countStmt;
	char *combinedSQL;
	const char *errMsg = NULL;

	if (!sql) {
		return ALP_STATUS_OK;
	}
	va_start(ap, sql);
	combinedSQL = sqlite3_vmprintf(sql, ap);
	va_end(ap);
	errMsg = alp_sal_get_error_message(sal);
	if (!combinedSQL) {
		return ALP_STATUS_SAL_NOMEM;
	}

    sqlErr = sqlite3_prepare((sqlite3 *)sal, combinedSQL,
		strlen(combinedSQL), &countStmt, (const char **) &dummy);
	if (sqlErr) {
	    errMsg = alp_sal_get_error_message(sal);
	    sqlite3_free(combinedSQL);
        return alp_prv_sal_translate_sqlite_err(sqlErr);
	}
	sqlite3_free(combinedSQL);
	sqlErr = sqlite3_step(countStmt);
    if (sqlErr != SQLITE_ROW) {
	    errMsg = alp_sal_get_error_message(sal);
	    sqlite3_finalize(countStmt);
        return alp_prv_sal_translate_sqlite_err(sqlErr);
	}
	*value = (long)sqlite3_column_int(countStmt, 0);
	sqlite3_finalize(countStmt);

    return ALP_STATUS_OK;
}


/*! Returns string describing last error generated by Database
 *
 * @param [in] sal database to read error string from
 *
 * @return pointer to string owned by SQL engine representing last error
 *     message generated
 */
const char *alp_sal_get_error_message(AlpSALHandle sal)
{
	return sqlite3_errmsg((sqlite3 *)sal);
}


/*!
* Executes a sql command given in a string buffer with a variable list of
* host parameter arguments.  After plugging in the host variables to the
* sql string, the sql is executed and the buffer created by this routine is
* freed.
*
* In addition to the subsitution variables supported by printf, this routine
* also supports the variable: "%Q".  When using %Q, the string passed will
* get single quote escaped, or will get "NULL" if the string is null.  If
* the string contains a single quote, it will be escaped (doubled) to form ''
*
* @param [in] sal AlpSALHandle of database to execute against
* @param [in] sql SQL string
* @param [in] variables va_list of host variables to plug into SQL string
*
* @return alp_status_t
*/
alp_status_t alp_sal_sql_execute(AlpSALHandle sal, const char *sql, ...)
{
    va_list ap;
	int32_t sqlErr;
	char *combinedSQL;
	char *errMsg = NULL;

	if (!sal || !sql) {
		return ALP_STATUS_SAL_BADPARAM;
	}

    va_start(ap,sql);
	combinedSQL = sqlite3_vmprintf(sql, ap);
    va_end(ap);
	if (!combinedSQL) {
		return ALP_STATUS_SAL_NOMEM;
	}
	sqlErr = sqlite3_exec((sqlite3 *)sal, combinedSQL, NULL, NULL, &errMsg);
	if (sqlErr && *errMsg) {
		free(errMsg);
	}
	sqlite3_free(combinedSQL);

    return alp_prv_sal_translate_sqlite_err(sqlErr);
}


/*! Creates a copy of the given SQL string returned in the first parameter,
 * that has been resized and had it's substitution variables filled in.
 *
 * The SQL string can embed any of the standard C variable types (ie, "%d",
 * "%s", etc) and it also provides a new variable type: "%q" that should be
 * used for any string substitution which may have a single quotation mark.
 * Single quotation marks are escaped (doubled) when substitution is
 * performed on them.
 *
 * The caller should free the string returned in the first parameter by calling
 * alp_sal_string_destroy
 *
 * @param [out] combined The resulting SQL statement with all variables
 *     substituted in.
 * @param [in] sql Any string (usually a SQL statement) that has substitution
 *     variables such as "%d", "%s" and the special "%q" described above
 * @param [in] varargs Host variables that will be substituted into the
 *     sql string.
 *
 * @see SALStringDestroy
 *
 * @return alp_status_t
 */
alp_status_t alp_sal_string_build(char **combined, const char *sql, ...)
{
    va_list ap;

	if (!combined) {
		return ALP_STATUS_SAL_BADPARAM;
	}

    va_start(ap,sql);
	*combined = sqlite3_vmprintf(sql, ap);
    va_end(ap);
	if (!(*combined)) {
		return ALP_STATUS_SAL_NOMEM;
	}

	return ALP_STATUS_OK;
}


/*! Frees a string created by SALBuildString.  On most platforms (except
 * notably Windows) it should be safe to just call "free" instead of
 * SALStringDestroy(), so this API may be depricated
 *
 * @param [in] s String to free
 *
 * @see SALBuildString
 */
void alp_sal_string_destroy(char *s)
{
	if (s) {
	    sqlite3_free(s);
	}
}


/*! Reads in all the text from the given file descriptor for a file that
 * contains SQL commands.  Once the file is read into memory, it's
 * the executed and the buffer is freed.  This is usful for initializing
 * a table or schema using a resource file.
 *
 * @param [in] sal AlpSALHandle of database to execute against
 * @param [in] fd Open file descriptor of a text file containing SQL commands
 *
 * @return alp_status_t
 */
alp_status_t alp_sal_execute_sql_script(AlpSALHandle sal, FILE *fd)
{
	int32_t len, sqlErr;
	char *buff;
	char *errMsg = NULL;

	/* Figure out length of resource file and read it into memory */
	fseek(fd, 0, SEEK_END);
	len = ftell(fd);
	fseek(fd, 0, SEEK_SET);
	buff = (char *)malloc(len+1);
	if (!buff) {
		return ALP_STATUS_SAL_NOMEM;
	}
	buff[len] = '\0';
	len = fread(buff, len, 1, fd);

	/* Execute SQL statements in the resource file */
	sqlErr = sqlite3_exec((sqlite3 *)sal, buff, NULL, NULL, &errMsg);
	if (sqlErr && errMsg) {
		free(errMsg);
	}

	/* Free resource file buffer */
	free(buff);

	return alp_prv_sal_translate_sqlite_err(sqlErr);
}


/*! Assigns rowID to be the last auto-generated primary key created by an
 * SQL INSERT statement
 *
 * @param [in] sal AlpSALHandle of database to execute against
 * @param [out] rowID Primary Key generated by last INSERT statement
 *
 * @return alp_status_t
 */
alp_status_t alp_sal_last_insert_rowid(AlpSALHandle sal, uint32_t *rowID)
{
    /* Validate params */
	if (!sal || !rowID) {
		return ALP_STATUS_SAL_BADPARAM;
	}

	*rowID = sqlite3_last_insert_rowid((sqlite3 *)sal);

	return ALP_STATUS_OK;
}

