/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/*! @sal_statement.c
 */
#include <hiker/sal.h>
#include "sal_prv.h"


/*! Mallocs and initializes a AlpSALStatement pointed to by the AlpSALStatement 
 * handle parameter.  The AlpSALStatement is only valid for the lifetime of the 
 * SALHandle.  Once a SALHandle is closed or destroyed, or the AlpSALStatement
 * is no longer, needed, it should be destroyed with AlpSALStatementDestroy.
 * 
 * After initializing a AlpSALStatement, the user should call the appropriate
 * AlpSALStatementSet API's, in order, once for each host variable that needs
 * to be bound to a variable in the given sql string (ie, variables of the
 * form "FieldName = ?").
 * 
 * @param [in] sal Handle to datamodel statement will execute against
 * @param [out] stmt A newly allocated AlpSALStatement that will contain a copy
 *       of the given string.
 * @param [in] sql A sql statement possibly containing substitution variables
 *       ("X = ?")
 *
 * @return alp_status_t
 */
alp_status_t alp_sal_statement_initialize(AlpSALHandle sal, 
	AlpSALStatement **stmt, const char *sql)
{
	int32_t sqlErr;
	char *dummy;
	const char *errMsg = NULL;

    if (!stmt || !sal) {
		return ALP_STATUS_SAL_BADPARAM;
	}

	/* Alloc space for AlpSALStatement */
	*stmt = (void *)malloc(sizeof(_AlpSALStatement));
	if (!(*stmt)) {
		return ALP_STATUS_SAL_NOMEM;
	}
	memset(*stmt, 0, sizeof(_AlpSALStatement));

	((_AlpSALStatement *)(*stmt))->bindPosition = 1;
	((_AlpSALStatement *)(*stmt))->sal = sal;

	/* Prepare the statement */
    sqlErr = sqlite3_prepare((sqlite3 *)sal, sql, strlen(sql), 
		&(((_AlpSALStatement *)(*stmt))->stmt), (const char **) &dummy);
    errMsg = alp_sal_get_error_message((sqlite3 *)sal);

    return alp_prv_sal_translate_sqlite_err(sqlErr);
}


/*! Given a AlpSALStatement and a host text variable, we bind the variable to
 * the statement.  
 * 
 * @param [in] stmt to bind the host variable to.
 *       of the given string.
 * @param [in] pos Position of the substitution variable in the 
 *     AlpSALStatement's SQL command (ie "?").  Position indicies start at 
 *     1 for the first variable.  Pass a -1 will cause the AlpSALStatement 
 *     to use and increment an internal counter which is useful if you plan 
 *     to set all substitution variables in order.
 * @param [in] s A text host varabile that will be bound into the statement
 * @param [in] copy true if the Data Model should make it's own copy of the
 *       data provided.
 *
 * @return alp_status_t
 */
alp_status_t alp_sal_statement_set_text(AlpSALStatement *stmt, int16_t pos, 
	const char *s, bool copy)
{
	int32_t sqlErr;
	int16_t index;
	const char *errMsg = NULL;

	if (!stmt) {
		return ALP_STATUS_SAL_BADPARAM;
	}

	index = pos < 0 ? ((_AlpSALStatement *)stmt)->bindPosition++ : pos;

    sqlErr = sqlite3_bind_text(
		(sqlite3_stmt *)((_AlpSALStatement *)stmt)->stmt, index,
	    (const char *)s, s ? strlen(s) : 0, copy==0 ? SQLITE_STATIC : 
		SQLITE_TRANSIENT);
	errMsg = sqlite3_errmsg((sqlite3 *)((_AlpSALStatement *)stmt)->sal);

    return alp_prv_sal_translate_sqlite_err(sqlErr);
}

/*! Given a AlpSALStatement and a host integer variable, we bind the variable to
 * the statement.  
 * 
 * @param [in] stmt to bind the host variable to.
 *       of the given string.
 * @param [in] pos Position of the substitution variable in the 
 *     AlpSALStatement's SQL command (ie "?").  Position indicies start at 
 *     1 for the first variable.  Pass a -1 will cause the AlpSALStatement 
 *     to use and increment an internal counter which is useful if you 
 *     plan to set all substitution variables in order.
 * @param [in] n A host varabiale that will be bound into the statement
 *
 * @return alp_status_t
 */
alp_status_t alp_sal_statement_set_int(AlpSALStatement *stmt, int16_t pos, 
	int32_t n)
{
	int32_t sqlErr;
	int16_t index;
	const char *errMsg = NULL;

	if (!stmt) {
		return ALP_STATUS_SAL_BADPARAM;
	}

	index = pos < 0 ? ((_AlpSALStatement *)stmt)->bindPosition++ : pos;

    sqlErr = sqlite3_bind_int(
		(sqlite3_stmt *)((_AlpSALStatement *)stmt)->stmt, index, n);
	errMsg = sqlite3_errmsg((sqlite3 *)((_AlpSALStatement *)stmt)->sal);

    return alp_prv_sal_translate_sqlite_err(sqlErr);
}


/*! Given a AlpSALStatement and a host integer variable, we bind the 
 *     variable to the statement.  
 * 
 * @param [in] stmt to bind the host variable to.
 *       of the given string.
 * @param [in] pos Position of the substitution variable in the 
 *     AlpSALStatement's SQL command (ie "?").  Position indicies start at 
 *     1 for the first variable.  Pass a -1 will cause the AlpSALStatement 
 *     to use and increment an internal counter which is useful if you 
 *     plan to set all substitution variables in order.
 * @param [in] n A host varabiale that will be bound into the statement
 *
 * @return alp_status_t
 */
alp_status_t alp_sal_statement_set_int64(AlpSALStatement *stmt, int16_t pos, 
	int64_t n)
{
	int32_t sqlErr;
	int16_t index;
	const char *errMsg = NULL;

	if (!stmt) {
		return ALP_STATUS_SAL_BADPARAM;
	}

	index = pos < 0 ? ((_AlpSALStatement *)stmt)->bindPosition++ : pos;

    sqlErr = sqlite3_bind_int64(
		(sqlite3_stmt *)((_AlpSALStatement *)stmt)->stmt, index, n);
	errMsg = sqlite3_errmsg((sqlite3 *)((_AlpSALStatement *)stmt)->sal);

    return alp_prv_sal_translate_sqlite_err(sqlErr);
}


/*! Given a AlpSALStatement and a host blob variable, we bind the variable to
 * the statement.  
 * 
 * @param [in] stmt to bind the host variable to.
 *       of the given string.
 * @param [in] pos Position of the substitution variable in the 
 *     AlpSALStatement's SQL command (ie "?").  Position indicies start at 
 *     1 for the first variable.  Pass a -1 will cause the AlpSALStatement 
 *     to use and increment an internal counter which is useful if you 
 *     plan to set all substitution variables in order.
 * @param [in] blob A host varabile that will be bound into the statement
 * @param [in] size Number of bytes in blob
 * @param [in] copy true if the Data Model should make it's own copy of the
 *       data provided.
 *
 * @return alp_status_t
 */
alp_status_t alp_sal_statement_set_blob(AlpSALStatement *stmt, int16_t pos, 
	void *blob, int32_t size, bool copy)
{
	int32_t sqlErr;
	int16_t index;

	if (!stmt) {
		return ALP_STATUS_SAL_BADPARAM;
	}

	index = pos < 0 ? ((_AlpSALStatement *)stmt)->bindPosition++ : pos;

    sqlErr = sqlite3_bind_blob(
		(sqlite3_stmt *)((_AlpSALStatement *)stmt)->stmt, index,
		blob, size, copy==0 ? SQLITE_STATIC : SQLITE_TRANSIENT);

    return alp_prv_sal_translate_sqlite_err(sqlErr);
}

/*! Given a AlpSALStatement and a host float variable, we bind the variable to
 * the statement.  
 * 
 * @param [in] stmt to bind the host variable to.
 *       of the given string.
 * @param [in] pos Position of the substitution variable in the 
 *     AlpSALStatement's SQL command (ie "?").  Position indicies start at 
 *     1 for the first variable.  Pass a -1 will cause the AlpSALStatement 
 *     to use and increment an internal counter which is useful if you 
 *     plan to set all substitution variables in order.
 * @param [in] f A host varabile that will be bound into the statement
 *
 * @return alp_status_t
 */
alp_status_t alp_sal_statement_set_float(AlpSALStatement *stmt, 
	int16_t pos, float f)
{
	int32_t sqlErr;
	int16_t index;

	if (!stmt) {
		return ALP_STATUS_SAL_BADPARAM;
	}

	index = pos < 0 ? ((_AlpSALStatement *)stmt)->bindPosition++ : pos;

	if (!stmt) {
		return ALP_STATUS_SAL_BADPARAM;
	}

    sqlErr = sqlite3_bind_double(
		(sqlite3_stmt *)((_AlpSALStatement *)stmt)->stmt, index,
		(double)f);

    return alp_prv_sal_translate_sqlite_err(sqlErr);
}

/*! Executes a AlpSALStatement that has been initalized with 
 * alp_sal_statement_initialize, and bound to any host parameters required by 
 * its SQL statement using AlpSALStatementSet commands. 
 *
 * @param [in] stmt AlpSALStatement to execute.
 *
 * @return alp_status_t
 */
alp_status_t alp_sal_statement_execute(AlpSALStatement *stmt)
{
	int32_t sqlErr;
    
	if (!stmt) {
		return ALP_STATUS_SAL_BADPARAM;
	}

	sqlErr = sqlite3_step(((_AlpSALStatement *)stmt)->stmt);

    return alp_prv_sal_translate_sqlite_err(sqlErr);
}


/*! Destroys a AlpSALStatement.  Should always be called after you are done using
 * a AlpSALStatement.
 *
 * @param [in] stmt The AlpSALStatement to destroy
 *
 * @return alp_status_t
 */
alp_status_t alp_sal_statement_destroy(AlpSALStatement *stmt)
{
	int32_t sqlErr;

	if (!stmt) {
		return ALP_STATUS_SAL_BADPARAM;
	}

	sqlErr = sqlite3_finalize(((_AlpSALStatement *)stmt)->stmt);

	if (!sqlErr) {
		free(stmt);
	}

    return alp_prv_sal_translate_sqlite_err(sqlErr);
}


/*! Given a prepared statement containing a single SELECT statement, 
 * any substitution variable it contains bound to the appropriate host 
 * variables, this routine will advance to the next result matching the query.
 *
 * Although this function will execute any statement without producing an
 * error, it is important for the caller to only use this function to 
 * advance to the next result in a selection statement to ensure that 
 * notifications are handled properly
 *
 * @param [in] stmt containing SELECT statement with substitution variables
 *     bound to the appropriate host variables
 *
 * @return alp_status_t: 
 *    ALP_STATUS_SAL_NORECORDS if no records match the select statement or
 *    the statement has already found the last matching record.
 *    ALP_STATUS_OK, if a matching row is found
 */
alp_status_t alp_sal_statement_next_result(AlpSALStatement *stmt)
{
	int32_t sqlErr;

	if (!stmt) {
		return ALP_STATUS_SAL_BADPARAM;
	}

	((_AlpSALStatement *)stmt)->bindPosition = 0;

	sqlErr = sqlite3_step(((_AlpSALStatement *)stmt)->stmt);
	if (sqlErr == SQLITE_ROW) {
		return ALP_STATUS_OK;
	}
	if (sqlErr == SQLITE_DONE) {
		return ALP_STATUS_SAL_NORECORDS;
	}

    return alp_prv_sal_translate_sqlite_err(sqlErr);
}


/*! Retrieves integer value from column of current result in stmt.
 *
 * @param [in] stmt AlpSALStatement handle for a query statement that has 
 *     been properly prepared and stepped with AlpSALStatementNextResult()
 * @param [in] pos column to retrieve from current row (0 is the first row).
 *     Passing -1 causes the statement to use and increment an internal 
 *     counter; useful if you want to retrieve all column values in the order
 *     they were specified in the SQL statement 
 *
 * @return integer in indicated column of current row of stmt
 */
int32_t  alp_sal_statement_get_int(AlpSALStatement *stmt, int pos)
{
	int16_t index;
	int32_t i;

	index = pos < 0 ? ((_AlpSALStatement *)stmt)->bindPosition++ : pos;

    i = sqlite3_column_int(((_AlpSALStatement *)stmt)->stmt, index);

    return i;
}


/*! Retrieves 64 bit integer value from column of current result in stmt.
 *
 * @param [in] stmt AlpSALStatement handle for a query statement that has 
 *     been properly prepared and stepped with AlpSALStatementNextResult()
 * @param [in] pos column to retrieve from current row (0 is the first row).
 *     Passing -1 causes the statement to use and increment an internal 
 *     counter; useful if you want to retrieve all column values in the order
 *     they were specified in the SQL statement 
 *
 * @return integer in indicated column of current row of stmt
 */
int64_t  alp_sal_statement_get_int64(AlpSALStatement *stmt, int pos)
{
	int16_t index;
	int32_t i;

	index = pos < 0 ? ((_AlpSALStatement *)stmt)->bindPosition++ : pos;

    i = sqlite3_column_int64(((_AlpSALStatement *)stmt)->stmt, index);

    return i;
}


/*! Retrieves text value from column of current result in stmt.
 *
 * @param [in] stmt AlpSALStatement handle for a query statement that has 
 *     been properly prepared and stepped with SALStatementNextResult()
 * @param [in] pos column to retrieve from current row (0 is the first row).
 *     Passing -1 causes the statement to use and increment an internal 
 *     counter; useful if you want to retrieve all column values in the order
 *     they were specified in the SQL statement 
 * @param [in] copy true if a copy of the string should be made before
 *     assigning to the text parameter
 *
 * @return string (or copy) in indicated column of current row of stmt
 */
char *alp_sal_statement_get_text(AlpSALStatement *stmt, int pos, bool copy)
{
	const unsigned char *s = NULL;
	int16_t index;

	if (!stmt) {
		return NULL;
	}

	index = pos < 0 ? ((_AlpSALStatement *)stmt)->bindPosition++ : pos;

    s = sqlite3_column_text(((_AlpSALStatement *)stmt)->stmt, index);

	if (s && copy) {
		return strdup(s);
	}

	return (char *)s;
}

/*! Retrieves blob value from column of current result in stmt.
 *
 * @param [in] stmt AlpSALStatement handle for a query statement that has 
 *     been properly prepared and stepped with SALStatementNextResult()
 * @param [in] pos column to retrieve from current row (0 is the first row).
 *     Passing -1 causes the statement to use and increment an internal 
 *     counter; useful if you want to retrieve all column values in the order
 *     they were specified in the SQL statement 
 * @param [in] copy true if a copy of the blob should be made before
 *     assigning to the text parameter
 *
 * @return pointer to data in indicated column of current row of stmt, or
 *     NULL if statement parameter was invalid
 */
void *alp_sal_statement_get_blob(AlpSALStatement *stmt, int pos,
					int32_t *size, bool copy)
{
	void *dataCopy = NULL;
	const void *data = NULL;
	int index;

	if (!stmt) {
		return NULL;
	}

	index = pos < 0 ? ((_AlpSALStatement *)stmt)->bindPosition++ : pos;

	*size = sqlite3_column_bytes(((_AlpSALStatement *)stmt)->stmt, index);
    data = sqlite3_column_blob(((_AlpSALStatement *)stmt)->stmt, index);

	if (*size && copy) {
        dataCopy = (void *)malloc(*size);
		memcpy(dataCopy, data, *size);
		return dataCopy;
	}

	return (void *)data;
}

/*! Retrieves float value from column of current result in stmt.
 *
 * @param [in] stmt AlpSALStatement handle for a query statement that has 
 *     been properly prepared and stepped with SALStatementNextResult()
 * @param [in] pos column to retrieve from current row (0 is the first row).
 *     Passing -1 causes the statement to use and increment an internal 
 *     counter; useful if you want to retrieve all column values in the order
 *     they were specified in the SQL statement 
 *
 * @return float in indicated column of current row of stmt
 */
float alp_sal_statement_get_float(AlpSALStatement *stmt, int pos)
					
{
	int16_t index;
	double d;

	index = pos < 0 ? ((_AlpSALStatement *)stmt)->bindPosition++ : pos;

	/* SQLite has internal type of float, but only a column accessor for 
	 * double! */
    d = sqlite3_column_double(((_AlpSALStatement *)stmt)->stmt, index);

    return (float)d;
}

/*! Checks for NULL value in column of current result in stmt.
 *
 * @param [in] stmt AlpSALStatement handle for a query statement that has 
 *     been properly prepared and stepped with SALStatementNextResult()
 * @param [in] pos column to retrieve from current row (0 is the first row).
 *     Passing -1 causes the statement to use and increment an internal 
 *     counter; useful if you want to retrieve all column values in the order
 *     they were specified in the SQL statement 
 *
 * @return bool true if the value is NULL
 */
bool alp_sal_statement_is_null(AlpSALStatement *stmt, int pos)
					
{
	int16_t index;
	int type;

	index = pos < 0 ? ((_AlpSALStatement *)stmt)->bindPosition++ : pos;

	type = sqlite3_column_type(((_AlpSALStatement *)stmt)->stmt, index);

	return (type == SQLITE_NULL);
}


/*! Resets a SQL query in a AlpSALStatement so it's ready to be re-executed.  
 * Any variables set in the statement retain their values, but 
 * SALStatementNextResult will return the first result in the query again.
 *
 * @param [in] stmt AlpSALStatement handle for a query statement that has been
 *     properly 
 * 
 * @return alp_status_t
 */
alp_status_t alp_sal_statement_reset(AlpSALStatement *stmt)
{
	if (!stmt) {
		return ALP_STATUS_SAL_BADPARAM;
	}

    return alp_prv_sal_translate_sqlite_err(
		sqlite3_reset(((_AlpSALStatement *)stmt)->stmt));
}


/*! Returns the AlpSALHandle bound to the statement, or NULL, if the 
 * stmt handle is invalid */
AlpSALHandle alp_sal_statement_get_handle(AlpSALStatement *stmt)
{
	if (!stmt) {
		return NULL;
	}

	return ((_AlpSALStatement *)stmt)->sal;
}
