/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
#define ST_E_NOERR 0
#define ST_E_USAGE 1
#define ST_E_STAT  2
#define ST_E_UNDEF 3
#define ST_E_ALLOC 4
#define ST_E_FILE  5

#define ST_C_UPDATE 0
#define ST_C_QUERY  1
#define ST_C_PID    2

#include <hiker/security.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>

typedef unsigned int Err;

typedef struct {
  char *policy;
  char *assertion;
} Query;

typedef struct {
  char *update_file_name;
  char *update_contents;
  unsigned int update_contents_length;
} Update;

typedef union {
  Query query;
  Update update;
} Context;

typedef struct {
  unsigned int command;
  unsigned int flags;
  Context context;
  pid_t pid;
} App;

static Err st_parse_command_line( App *app, int argc, char *argv[] );
static int st_find_in_list( char *list[], int size, char *trial );
static unsigned int debug( App *app );
static Err st_get_update_contents( App *app );

static char *commands[] = {
  "update",
  "query",
  "pid"
};

int main( int argc, char *argv[] ) {
  static App app;
  Err error;
  alp_status_t spf_error;
  char * assertion_value = "-";

  do {
    if( ( error = st_parse_command_line( &app, argc, argv ) ) != ST_E_NOERR ) {
      break;
    }

    switch( app.command ) {
    case ST_C_UPDATE:
      debug( &app ) && printf( "Calling 'update' with %s.\n", app.context.update.update_file_name );
      if( ( error = st_get_update_contents( &app ) ) != ST_E_NOERR ) {
	break;
      }
      debug( &app ) && printf( "Calling alp_spf_update_open()\n");
      if( ( spf_error = alp_spf_update_open( ) ) != ALP_STATUS_SPF_E_NOERR ) break;
    
      debug( &app ) && printf( "Calling alp_spf_update() with:\n%*.*s", app.context.update.update_contents_length, app.context.update.update_contents_length, app.context.update.update_contents );
      spf_error = alp_spf_update( app.context.update.update_contents, app.context.update.update_contents_length );
      debug( &app ) && printf( "Call to alp_spf_update() returned %X, errno = %d\n", spf_error, errno );
      if( spf_error != ALP_STATUS_SPF_E_NOERR ) break;
      debug( &app ) && printf( "Calling alp_spf_update_close()\n" );
      if( ( spf_error = alp_spf_update_close( ) ) != ALP_STATUS_SPF_E_NOERR ) break;
      break;

    case ST_C_PID:

      debug( &app ) && printf( "Calling alp_spf_query_open()\n" );
      spf_error = alp_spf_query_open();
      if( spf_error != ALP_STATUS_SPF_E_NOERR ) break;


      debug( &app ) && printf( "Calling alp_spf_query_set_subject_pid()\n" );
      spf_error = alp_spf_query_set_subject_pid( app.pid );
      if( spf_error != ALP_STATUS_SPF_E_NOERR ) break;

      debug( &app ) && printf( "Calling 'query' with %s on policy %s.\n", app.context.query.assertion, app.context.query.policy );
      spf_error = alp_spf_query_get_assertion_string( app.context.query.assertion, (unsigned int) strlen( app.context.query.assertion ), & assertion_value );
      debug( &app ) && printf( "query returned %X, errno = %d\n", spf_error, errno );
      printf( "(assertion,value)=(%s,%s)\n", app.context.query.assertion, assertion_value );
      if( spf_error != ALP_STATUS_SPF_E_NOERR ) break;

      debug( &app ) && printf( "Calling alp_spf_query_close()\n" );
      spf_error = alp_spf_query_close();
      if( spf_error == ALP_STATUS_SPF_E_NOERR ) break;

    case ST_C_QUERY:
      debug( &app ) && printf( "Calling alp_spf_query_open()\n" );
      spf_error = alp_spf_query_open();
      if( spf_error != ALP_STATUS_SPF_E_NOERR ) break;

      debug( &app ) && printf( "Calling 'query' with %s on policy %s.\n", app.context.query.assertion, app.context.query.policy );
      spf_error = alp_spf_query_get_assertion_string_from_policy( app.context.query.policy, app.context.query.assertion, (unsigned int) strlen( app.context.query.assertion ), & assertion_value );
      debug( &app ) && printf( "query returned %X, errno = %d\n", spf_error, errno );
      printf( "(policy,assertion,value)=(%s,%s,%s)\n", app.context.query.policy, app.context.query.assertion, assertion_value );
      if( spf_error != ALP_STATUS_SPF_E_NOERR ) break;

      debug( &app ) && printf( "Calling alp_spf_query_close()\n" );
      spf_error = alp_spf_query_close();
      if( spf_error == ALP_STATUS_SPF_E_NOERR ) break;

      break;
    default:
      error = ST_E_UNDEF;
      break;
    }
  } while( 0 );

  if( error != ST_E_NOERR ) {
    printf( "Exiting with Error %d. Errno = %d.\n", error, errno );
  }

  return( error );
}

static int st_find_in_list( char *list[], int size, char *trial ) {
  int rv = -1;
  int i;

  for( i=0; i < size; i++ ) {
    if( ! strcmp( trial, list[ i ] ) ) {
      rv = i;
      break;
    }
  }

  return( rv );
}

static unsigned int debug( App *app ) {
  return( ( app->flags & 0x01 ) == 0x01 );
}

static Err st_parse_command_line( App *app, int argc, char *argv[] ) {
  int count = 1;
  Err rv = ST_E_NOERR;
  int command;

  do {
    if( count >= argc ) {
      rv = ST_E_USAGE;
      break;
    }

    if( ! strcmp( "debug", argv[ count ] ) ) {
      app->flags = 0x01;
      count ++;
    }

    if( count >= argc ) {
      rv = ST_E_USAGE;
      break;
    }

    command = st_find_in_list( commands, 3, argv[ count ] );
    app->command = command;
    count ++;

    switch( command ) {
    case ST_C_UPDATE:
      if( count >= argc ) {
	rv = ST_E_USAGE;
	break;
      }
      app->context.update.update_file_name = argv[ count ];
      break; 

    case ST_C_QUERY:
      if( count >= argc ) {
	rv = ST_E_USAGE;
	break;
      }

      app->context.query.assertion = argv[ count ];
      count++;

      if( count >= argc ) {
	rv = ST_E_USAGE;
	break;
      }

      app->context.query.policy = argv[ count ];

      break;

    case ST_C_PID:

      if( count >= argc ) {
	rv = ST_E_USAGE;
	break;
      }

      app->pid = atoi( argv[ count ] );
      count++;

      if( count >= argc ) {
	rv = ST_E_USAGE;
	break;
      }

      app->context.query.assertion = argv[ count ];

      break;

    default:
      rv = ST_E_USAGE;
      break;
    }
  } while( 0 );

  return( rv );
}

static Err st_get_update_contents( App *app ) {
  Err rv = ST_E_NOERR;
  struct stat buf;
  char *contents;
  char *file_name;
  unsigned int size;
  FILE *fp;

  do {

    file_name = app->context.update.update_file_name;
    if( stat( file_name, &buf ) ) {
      rv = ST_E_STAT;
      break;
    }

    size = (unsigned int) buf.st_size;

    if( ( contents = (char *)malloc( size + 1 ) ) == (char *) NULL ) {
      rv = ST_E_ALLOC;
      break;
    }

    app->context.update.update_contents = contents;
    app->context.update.update_contents_length = size;

    if( ( fp = fopen( file_name, "r" ) ) == (FILE *)NULL ) {
      rv = ST_E_FILE;
      break;
    }

    fread( contents, size, 1, fp );
    fclose( fp );

    contents[ size ] = '\0';

  } while( 0 );

  return( rv );
}
