/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/*!
 * @file <traces_conf_util.c>
 *
 * @brief C file with for accessing config file for traces
 *
 ******************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "traces_conf_util.h"



#define ALP_LIST_REPLACE_DUPS 0x00 		/* Overwrite duplicate values on insert */
#define ALP_LIST_ALLOW_DUPS   0x01 		/* Allow duplicate data entries to exist in the list */


/* ============== all Typedefs */ 
/* For File analysis   */
typedef struct _AlpConfFile	AlpConfFile;

typedef struct _AlpPListEntry AlpPListEntry;
typedef struct _AlpPList AlpPList;

typedef int (*alpEntryCompare)(const void *a, const void *b);
typedef	void *	(*alpEntryCreate)	(void *data);
typedef	void	(*alpEntryDestroy)	(void *data);

/* Section/entry list */
typedef struct _AlpConfSection AlpConfSection;
typedef struct _AlpConfEntry AlpConfEntry;


/* ========== The structure */
struct _AlpConfSection
{
        char            *name;
	AlpPList          *entry_list;
} ;

/* key list */
struct _AlpConfEntry
{
        char    *key;
        char    *value;
};

/* For config file */
struct _AlpConfFile
{
	int fdes;	/* File descriptor */
	char *buf;	/* Data buffer */
	char *p;	/* Current position in buffer */
	char *e;	/* Points one char past end of valid data in buffer */
	int bufsize;	/* Current buffer size */
};

/* Entries linked list structure */
struct _AlpPListEntry {
	void		     *data;    		/* Ptr to data for this node */
	struct _AlpPListEntry *next; 	/* Ptr to next nodes in the list */
	struct _AlpPListEntry *prev;    /* Ptrr to previous node in the list */
};


 struct _AlpPList {
	AlpPListEntry	*head;         			/* Pointer to the first node in the list */
	AlpPListEntry 	*tail;        			/* Pointer to the last node in the list */
	alpEntryCompare	compare_func;  /* Function used to compare data entries */
	alpEntryCreate	create_func;   	/* Function used to copy data entries */
	alpEntryDestroy	destroy_func;  	/* Function used to destroy data entries */
	int		options;       				/* List options */
	int		entries;       				/* The current number of list entries (ie. list length) */
} ;


/* These globals are for current (opened) configuration file */
static char   		*sconf_path = NULL ;
static  int        	sconf_writeable = 0 ;
static  AlpPList   	*sconf_section_list = NULL ;


/* Prototypes for file processing fct */
static AlpConfFile * prv_conf_get_file_struct(int fdes);
static char * prv_conf_read_line(AlpConfFile *file);
static void prv_conf_free_file_struct(AlpConfFile *file);

static AlpPList *prv_conf_list_create (alpEntryCompare compare_func,
			  alpEntryCreate create_func,  alpEntryDestroy destroy_func,
			  int options);

static int prv_conf_list_insert (AlpPList *list, void *data);
static AlpPListEntry *prv_conf_list_head	(AlpPList *list);
static AlpPListEntry *prv_conf_list_tail (AlpPList *list);
static AlpPListEntry *prv_conf_list_next (AlpPListEntry *entry);
static AlpPListEntry *prv_conf_list_prev(AlpPListEntry *entry);
static void *prv_conf_list_get_data (AlpPListEntry *entry);
static AlpPListEntry *prv_conf_list_search	(AlpPList *list, void *data);
static AlpPListEntry *prv_conf_list_search_alt(AlpPList *list, void *data,  alpEntryCompare compare_func);
static void prv_conf_list_delete_entry	(AlpPList *list, AlpPListEntry *entry);
static void prv_conf_list_free		(AlpPList *list);
static int prv_conf_list_count		(AlpPList *list);


/* Our private functions and vars */
static AlpPList * prv_scanfile_for_sections(const char *path);
static void prv_parse_line(char *line, char **varname, char **varvalue);
static int prv_section_compare(const void *a, const void *b);
static void *prv_section_create(void *data);
static void prv_section_destroy(void *data);
static int prv_entry_compare(const void *a, const void *b);
static void *prv_entry_create(void *data);
static void prv_entry_destroy(void *data);
static int prv_make_path(const char *path, mode_t mode);


/* ===============================================  */
/* API functions													*/
/* ===============================================  */

/* --------------------------------------------- */
/* alp_prv_conf_read_string()						*/
/* --------------------------------------------- */
char * prv_conf_read_string(const char *section,  const char *key, const char *xdefault)
{
	AlpPListEntry	*s_entry, *e_entry;
	AlpConfSection	*s_data;
	AlpConfEntry	e_srch, *e_data;

	/* Find the requested [Section] */
	s_entry = prv_conf_list_search(sconf_section_list, (void*)section);
	if(s_entry == NULL)
		goto do_default;
	s_data = prv_conf_list_get_data(s_entry);

	/* Locate the requested Key */
	e_srch.key = (char*)key;
	e_entry = prv_conf_list_search(s_data->entry_list, &e_srch);
	if(e_entry == NULL)
		goto do_default;
	e_data = prv_conf_list_get_data(e_entry);

	/* Duplicate the resulting value and return it to the caller */
	return strdup(e_data->value);

do_default:
	/* Any failure causes a branch to here to return */
	/* a default value if provided by the caller.	 */
	return strdup(xdefault);
}


/* --------------------------------------------- */
/* 											*/
/* --------------------------------------------- */
int prv_conf_read_int(const char *section, const char *key, int xdefault)
{
	char	tmp[20], *tmp2;
	int	value;

	sprintf(tmp, "%d", xdefault);
	tmp2 = prv_conf_read_string(section, key, tmp);
	value = atoi(tmp2);
	free(tmp2);

	return value;
}

/* --------------------------------------------- 
* alp_prv_conf_read_list(): read a key which is a list into a real list
 --------------------------------------------- */
int prv_conf_read_list(const char *section, const char *key,
		   int *argcp, char ***argvp)
{
	int	index, rc;
	char	*p, *s1, *s2;
	char	*str, *tmp, *tmp2;
	char	saw_space=0, saw_backspace;

	str = prv_conf_read_string(section, key, NULL);

	if (str != NULL) {
		rc = 0;
		for (*argcp = 1, p = str; *p != '\0'; p++)
		{
			if (*p == '\\' && *(p+1))
			{
				p++;
				if (saw_space) 
				{
					*argcp += 1;
					saw_space = 0;
				}
			} else if (*p == ' ')
			{
				saw_space = 1;
			} else if (saw_space)
			{
				*argcp += 1;
				saw_space = 0;
			}
		}

		*argvp = (char **) malloc((*argcp + 1) * sizeof(char *));
		
		/* Manually null-terminate list. */
		(*argvp)[*argcp] = NULL;


		p = str;
		index = 0;
		do {
			tmp = p;

			for (saw_backspace = 0;
			     *p != '\0' && (saw_backspace ? 1 : (*p != ' '));
			     p++)
			{
				if (*p == '\\')
					saw_backspace = 1;
				else
					saw_backspace = 0;
			}

			tmp2 = malloc(p-tmp+1);
			(*argvp)[index] = strncpy(tmp2, tmp, p - tmp);
			tmp2[p-tmp] = '\0';
			s1 = s2 = (*argvp)[index++];

			while (*s1)
			{
				if (*s1 == '\\')
					s1++;
				if (!*s1) break;
				*s2++ = *s1++;
			}
			*s2 = '\0';

			while (*p && *p == ' ')
				p++;
		} while (*p);

		free(str);
	} else
	{
		rc = -1;
		*argcp = 0;
		*argvp = NULL;
	}

	return rc;
}

/* --------------------------------------------- */
/* --------------------------------------------- */
int prv_conf_get_sections( int *argcp, char ***argvp)
{
	AlpPListEntry	*s_entry;
	AlpConfSection	*s_data;
	int count=0;
	char **sections=NULL;

	s_entry = prv_conf_list_head(sconf_section_list);
	while(s_entry) {
		s_data = prv_conf_list_get_data(s_entry);
		sections = realloc(sections, ++count * sizeof(char *));
		sections[count-1] = strdup(s_data->name);
		s_entry = prv_conf_list_next(s_entry);
	}

	*argcp = count;
	*argvp = sections;

	return 0;
}

/* --------------------------------------------- */
/* --------------------------------------------- */
int prv_conf_get_keys(const char *section, int *argcp, char ***argvp)
{
	AlpPListEntry	*s_entry, *e_entry;
	AlpConfSection	*s_data;
	AlpConfEntry	*e_data;
	int count=0;
	char **keys=NULL;

	/* Find the requested [Section] */
	s_entry = prv_conf_list_search(sconf_section_list, (void*)section);
	if(s_entry == NULL)
		return -1;
	s_data = prv_conf_list_get_data(s_entry);

	e_entry = prv_conf_list_head(s_data->entry_list);
	while(e_entry) {
		e_data = prv_conf_list_get_data(e_entry);
		keys = realloc(keys, ++count * sizeof(char *));
		keys[count-1] = strdup(e_data->key);
		e_entry = prv_conf_list_next(e_entry);
	}

	*argcp = count;
	*argvp = keys;

	return 0;
}


/* ==== APIs for Writing ========== */

/* --------------------------------------------- */
/* alp_prv_conf_write_string	()						*/
/* --------------------------------------------- */
int prv_conf_write_string(const char *section,  const char *key, const char *value)
{
	AlpPListEntry	*s_entry;
	AlpConfSection	*s_data;
	AlpConfEntry	e_data;

	/* Is this confio writeable? */
	if(!sconf_writeable) 
	{		
		return -1;
	}

	/* Find the requested [Section] */
	s_entry = prv_conf_list_search(sconf_section_list, (void*)section);
	if(s_entry == NULL) {
		/* We need to create a new [Section] */
		if(prv_conf_list_insert(sconf_section_list, (void*)section) < 0) {
			return -1;
		}
		s_entry = prv_conf_list_search(sconf_section_list, (void*)section);
	}
	s_data = prv_conf_list_get_data(s_entry);

	/* Insert the new value into the [Section]'s list */
	e_data.key = (char*)key;
	e_data.value = (char*)value;
	if(prv_conf_list_insert(s_data->entry_list, &e_data) < 0) {	  
		return -1;
	}

	return 0;
}


/* --------------------------------------------- */
/* 	alp_prv_conf_write_int()						*/
/* --------------------------------------------- */
int prv_conf_write_int(const char *section, const char *key, int value)
{
	char	tmp[20];

	sprintf(tmp, "%d", value);
	return prv_conf_write_string(section, key, tmp);
}



/* --------------------------------------------- */
/* alp_prv_conf_write_list()											*/
/* --------------------------------------------- */
int prv_conf_write_list(const char *section, const char *key, int argc, char **argv)
{
	int	i;
	char	buf[1023];
	char	*src, *dst, *eob;

	dst = buf;
	eob = buf+1023;
	for(i=0; i<argc; i++) {
		src = argv[i];
		while(*src != '\0') {
			if(*src == ' ') {
				if(dst >= eob)
					return -1;
				*dst++ = '\\';
			}
			if(dst >= eob)
				return -1;
			*dst++ = *src++;
		}
		if(dst >= eob)
			return -1;
		*dst++ = ' ';
	}

	dst--;
	*dst = '\0';

	return prv_conf_write_string(section, key, buf);
}


/* ==== APIs for Removing  ========== */

/* --------------------------------------------- */
/* alp_prv_conf_delete_section()					*/
/* --------------------------------------------- */
int prv_conf_delete_section( const char *section)
{
	AlpPListEntry	*s_entry;

	/* Is this confio writeable? */
	if(!sconf_writeable) {
		return -1;
	}

	/* Find the requested [Section] */
	s_entry = prv_conf_list_search(sconf_section_list, (void*)section);
	if(s_entry == NULL)
		return 1;

	/* Since the entry list will automatically be destroyed, all */
	/* we need to do is remove this section entry */
	prv_conf_list_delete_entry(sconf_section_list, s_entry);

	return 0;
}


/* --------------------------------------------- */
/* alp_prv_conf_delete_key() 						*/
/* --------------------------------------------- */
int prv_conf_delete_key( const char *section, const char *key)
{
	AlpPListEntry	*s_entry, *e_entry;
	AlpConfSection	*s_data;
	AlpConfEntry	e_data;


	/* Is this confio writeable? */
	if(!sconf_writeable) {
		return -1;
	}

	/* Find the requested [Section] */
	s_entry = prv_conf_list_search(sconf_section_list, (void*)section);
	if(s_entry == NULL)
		return 1;
	s_data = prv_conf_list_get_data(s_entry);

	/* Find the requested Key */
	e_data.key = (char*)key;
	e_entry = prv_conf_list_search(s_data->entry_list, &e_data);
	if(e_entry == NULL)
		return 1;

	/* Remove the key's list entry */
	prv_conf_list_delete_entry(s_data->entry_list, e_entry);

	return 0;
}


/* ========= Various file APIS==============	*/
/* --------------------------------------------- */
/*  alp_prv_conf_flush()							*/
/* --------------------------------------------- */
int prv_conf_flush(void)
{
	AlpPListEntry	*s_entry, *e_entry;
	AlpConfSection	*s_data;
	AlpConfEntry	*e_data;
	FILE		*c_file;
	int		firstline=1;

	/* Is this confio writeable? */
	if(!sconf_writeable) {
		return -1;
	}

	/* Open our configuration file for writing */
	if((c_file = fopen(sconf_path, "w")) == NULL)
	{
		return -1;
	}

	/* Now step through each section, writing each entry from it's list */
	s_entry = prv_conf_list_head(sconf_section_list);
	while(s_entry)
	{
		s_data = prv_conf_list_get_data(s_entry);

		/* Output a line for our [SectionID] */
		if(firstline)
		{
			fprintf(c_file, "[%s]\n", s_data->name);
			firstline = 0;
		} else
			fprintf(c_file, "\n[%s]\n", s_data->name);

		/* Bounce through the key entry list */
		e_entry = prv_conf_list_head(s_data->entry_list);
		while(e_entry)
		{
			e_data = prv_conf_list_get_data(e_entry);

			/* Output a line for our Key=Value */
			fprintf(c_file, "%s=%s\n", e_data->key,  e_data->value);

			e_entry = prv_conf_list_next(e_entry);
		}

		s_entry = prv_conf_list_next(s_entry);
	}

	fclose(c_file);

	return 0;
}


/* --------------------------------------------- */
 /* alp_prv_conf_close()							*/
/* --------------------------------------------- */
void prv_conf_close(void)
{
	prv_conf_flush();
	prv_conf_list_free(sconf_section_list);
	free(sconf_path);
	sconf_section_list = NULL;
	sconf_path = NULL;
	
}

/* -----------------------------------------------------------
alp_prv_conf_getfile: search for the requested file, and open it with settings 
 ----------------------------------------------------------- */
int prv_conf_getfile(const char *path, const ALPConfType options)
{
	static int	next_handle=0;

	AlpPList		*section_list;
	int		opt_create, opt_rdonly, opt_rdwr;
	int		t_file;

	/* Check for insane options */
	opt_create = ((options & CONF_CREATE) == CONF_CREATE);
	opt_rdonly = ((options & CONF_RDONLY) == CONF_RDONLY);
	opt_rdwr = ((options & CONF_RDWR) == CONF_RDWR);

	if((opt_rdonly && (opt_rdwr || opt_create)) ||  (!opt_rdonly && !opt_rdwr))
	 {
//		fprintf(stderr,"prv_conf_getfile: Invalid options\n");
		return -1;
	}	

	/* Check to see if the file exists */
	if(access(path, F_OK))
	{
		/* Create the file if requested and we can */
		if(opt_create)
		{
			prv_make_path(path, 0700);
//			fprintf(stderr,"TRACES: Config file creation: %s\n", path);
			t_file = open(path, O_RDWR | O_CREAT | O_EXCL,  S_IRUSR | S_IWUSR);
			if(t_file != -1)
				close(t_file);
			else 
			{
//				fprintf(stderr,"prv_conf_getfile:Unable to create file %s\n", path);
				return -1;
			}
		}
	}

	/* Check read or writablity of file */
	if(opt_rdonly && access(path, R_OK)) {
//		fprintf(stderr,"prv_conf_getfile:Unable to read file %s\n", path);
		return -1;
	}
	if(opt_rdwr && access(path, R_OK | W_OK)) {
//		fprintf(stderr,"prv_conf_getfile:Unable to read or write file %s\n", path);
		return -1;
	}

	/* DOLATER: Check if exists */
	/* If exists, change settings */
	if (sconf_section_list)
	{
		if(opt_rdwr && !sconf_writeable)	
			sconf_writeable = opt_rdwr;

		/* Return existing handle */
		return 1;
	}

	/* Go do the dirty work and give us a section_list */
	section_list = prv_scanfile_for_sections(path);
	if(!section_list)
		/* There was an error :( */
		return -1;

	/* Store our global settings */
	sconf_path = strdup(path);
	sconf_writeable = opt_rdwr;
	sconf_section_list = section_list;

	return 1;
}



/* =================================================== */
/* ------------------------------------------------------------------ */
/*	Parse the configuration file to fill the sections							*/
/* ------------------------------------------------------------------ */
static AlpPList * prv_scanfile_for_sections(const char *path)
{
	int		c_file;
	AlpConfFile		*c_struct;
	char		*line;
	char		*varname, *varvalue;
	int		linenum = 0;
	AlpPList		*s_list;
	AlpPListEntry	*s_entry;
	AlpConfSection	*s_data=NULL;
	AlpConfEntry	*e_data=NULL;
	struct stat		st;

	/* Create the section list */
	s_list = prv_conf_list_create(prv_section_compare,prv_section_create, prv_section_destroy,  ALP_LIST_REPLACE_DUPS);
	if(!s_list)
		return NULL;

	/* Check for file mode */
	if((stat(path, &st)) || (!S_ISREG(st.st_mode)))
	{
		fprintf(stderr,"Incorrect file type (%s)\n", path);
		return NULL;
	}

	/* Open the input config file */
	if((c_file = open(path, O_RDONLY)) == -1)
	{
		/* This should be impossible now, due to checks made earlier? */
		fprintf(stderr,"Unable to read file %s\n", path);
		return NULL;
	}

	/* Get a AlpConfFile struct */
	c_struct = prv_conf_get_file_struct(c_file);

	/* Setup some temp storage to use */
	e_data = malloc(sizeof(AlpConfEntry));
	varvalue = NULL;

	/* Read individual lines and pass them off to be parsed */
	while((line = prv_conf_read_line(c_struct)) != NULL)
	{
		linenum++;
		prv_parse_line(line, &varname, &varvalue);
		if(varname == NULL)
		{
			free(line);
			continue;	/* Blank line or comment */
		}
		if(varvalue == NULL)
		{
			/* Might be a [SectionID] */
			if(varname[0] == '['  && varname[strlen(varname)-1] == ']')
			{
				/* It is, so setup a new section entry */
				varname[strlen(varname)-1] = '\0';
				varname++;
				if(prv_conf_list_insert(s_list, varname) < 0)
					fprintf(stderr,"prv_scanfile_for_sections: list insert error\n");
				s_entry = prv_conf_list_search(s_list, varname);
				s_data = prv_conf_list_get_data(s_entry);
			} else
				fprintf(stderr,"prv_scanfile_for_sections:Syntax error, %s at line %d)\n", path, linenum);
			
			free(line);
			continue;
		}

		/* We have a valid varname/varvalue, add them to entry list */
		if(!s_data)
		{
			/* We haven't seen a [SectionID] yet :( */
			fprintf(stderr,"prv_scanfile_for_sections:Syntax error, %s at line %d\n", path, linenum);
			free(line);
			continue;
		}
		e_data->key = varname;
		e_data->value = varvalue;
		if(prv_conf_list_insert(s_data->entry_list, e_data) < 0)
			fprintf(stderr,"prv_scanfile_for_sections: list insert error\n");
		free(line);
	}

	/* Cleanup after ourselves */
	free(e_data);
	prv_conf_free_file_struct(c_struct);
	close(c_file);

	return s_list;
}


/* ------------------------------------------------------------------ */
/*																	*/
/* ------------------------------------------------------------------ */
static void prv_parse_line(char *p, char **varname, char **varvalue)
{
	char csave, *psave, *sol;

	/* Save start-of-line in sol */
	sol = p;

	*varname = NULL;
	/* Skip over whitespace */
	while((*p == ' ' || *p == '\t' || *p == '\n') && *p != '\0')
		p++;
	if(*p == '\0' || *p == '#')
		return;		/* The line is a comment */

	*varname = p;

	*varvalue = NULL;
	/* Skip until we find an equals sign */
	while(*p != '=' && *p != '\n' && *p != '\0')
		p++;
	csave = *p;
	psave = p;

	if(*p == '=')
	{
		/* Found '=', now backspace to remove trailing space */
		do
		{
			p--;
		} while(p >= sol && (*p == ' ' || *p == '\t' || *p == '\n'));
		p++;
	}
	*p = '\0';
	p = psave;
	p++;

	if(**varname == '\0' || p == sol)
	{
		*varname = NULL;
		return;
	}

	if(csave == '\n' || csave == '\0')
		return;		/* There is no argument */

	/* There appears to be an argument, skip to the start of it */
	while((*p == ' ' || *p == '\t' || *p == '\n' || *p == '=')&& *p != '\0')
		p++;
	if(*p == '\0')
		return;		/* Argument is missing */

	/* There definitely is an argument */
	*varvalue = p;

	/* Terminate it ... */
	while(*p != '\n' && *p != '\0')
		p++;
	/* Found EOL, now backspace over whitespace to remove trailing space */
	p--;
	while(*p == ' ' || *p == '\t' || *p == '\n')
		p--;
	p++;
	/* Finally terminate it with a NUL */
	*p = '\0';	/* Might have already been the NUL, but who cares? */
}

/* ==== prv section fct ==========*/
/* ------------------------------------------------------------------ */
/*																	*/
/* ------------------------------------------------------------------ */
static int prv_section_compare(const void *a, const void *b)
{
	const AlpConfSection *s_a = a;
	const char *s_b = b;

//	return strcmp(s_a->name, s_b);	//Fix Ogu 060801
	return strcasecmp(s_a->name, s_b);

}

/* ------------------------------------------------------------------ */
/*																	*/
/* ------------------------------------------------------------------ */
static void *prv_section_create(void *data)
{
	/* Note that this function is a little odd since the data passed
	   is not expected to be the full struct, but just the name str.
	   See the comment in section_compare. */
	AlpConfSection	*dst;

	dst = malloc(sizeof(AlpConfSection));

	/* Copy the section name and create an entry list */
	dst->name = strdup(data);
	dst->entry_list = prv_conf_list_create(prv_entry_compare,
				       prv_entry_create,  prv_entry_destroy,
				       ALP_LIST_REPLACE_DUPS);
	if(!dst->entry_list) {
		free(dst->name);
		free(dst);
		dst = NULL;
	}

	return dst;
}

/* ------------------------------------------------------------------ */
/*																	*/
/* ------------------------------------------------------------------ */
static void prv_section_destroy(void *data)
{
	AlpConfSection	*s_data;

	s_data = data;
	free(s_data->name);
	prv_conf_list_free(s_data->entry_list);
	free(s_data);
}


/* ------------------------------------------------------------------ */
/*																	*/
/* ------------------------------------------------------------------ */
static int prv_entry_compare(const void *a, const void *b)
{
	const AlpConfEntry *e_a = a, *e_b = b;

//	return strcmp(e_a->key, e_b->key);   //Fix Ogu 060801
	return strcasecmp(e_a->key, e_b->key);
}

static void *prv_entry_create(void *data)
{
	AlpConfEntry	*src, *dst;

	src = data;
	dst = malloc(sizeof(AlpConfEntry));

	/* Copy the key and value data */
	dst->key = strdup(src->key);
	dst->value = strdup(src->value);

	return dst;
}

/* ------------------------------------------------------------------ */
/*																	*/
/* ------------------------------------------------------------------ */
static void prv_entry_destroy(void *data)
{
	AlpConfEntry	*e_data;

	e_data = data;
	free(e_data->key);
	free(e_data->value);
	free(e_data);
}

/* ------------------------------------------------------------------ */
/*																	*/
/* ------------------------------------------------------------------ */
static int prv_make_path(const char *full, mode_t mode)
{
	struct stat stats;
	size_t len = strlen(full) + 1;
	char copy_buf[len], path[len];
	char *copy = copy_buf;

	strcpy(copy, full);
	path[0] = '\0';

	/* Skip preceding / */
	if (copy[0] == '/')
		copy++;

	do {
		char *next = strchr(copy, '/');

		/* If this is the last token, it's the file name - break */
		if (!next) break;
		
		*next = '\0';

		/* While there's still stuff left, it's a directory */
		strcat(strcat(path, "/"), copy);
		if (mkdir(path, mode) < 0
		    && (stat(path, &stats) < 0 || !S_ISDIR(stats.st_mode))) {
			return -1;
		}

		copy = next + 1;
	} while (1);

	return 0;
}




/* ========================================= */
/* List management functions								*/
/* ========================================= */
static AlpPList* prv_conf_list_create(alpEntryCompare compare_func,
			 alpEntryCreate create_func, alpEntryDestroy destroy_func,
			 int options)
{
	AlpPList *new;

	if(compare_func == NULL && !(options & ALP_LIST_ALLOW_DUPS)) {
		/* debug warn - "Cannot replace dups w/o a compare func" */
	}

	new = malloc(sizeof(AlpPList));
	new->head = NULL;
	new->tail = NULL;
	new->compare_func = compare_func;
	new->create_func = create_func;
	new->destroy_func = destroy_func;
	new->options = options;
	new->entries = 0;

	return new;
}


/* -----------------------------------------------------
 * prv_conf_list_insert()
 *	Inserts an entry into a list, copying the data entry if an entry
 *	creation function is known and placing the data in-order if a
 *	comparison function is registered.
 *
 *	Returns 0 normally
 *		or 1 if an entry was replaced
 *		or -1 on error
 ---------------------------------------------------*/
static int prv_conf_list_insert(AlpPList *list, void *data)
{
	AlpPListEntry *new, *p, *q;
	int compare=-1;
	int rc=0;

	if(!list) {
		/* debug warn - "NULL list passed to prv_conf_list_insert" */
		return -1;
	}
	if(!data) {
		/* debug warn - "NULL data passed to prv_conf_list_insert" */
		return -1;
	}

	/* Setup the new entry */
	new = malloc(sizeof(AlpPListEntry));
	if(!new) {
		/* debug FATAL - "malloc() failure in prv_conf_list_insert" */
		return -1;
	}
	new->prev = NULL;
	new->next = NULL;
	if(list->create_func)
		new->data = (*list->create_func) (data);
	else
		new->data = data;

	/* Find where to stuff it */
	if(list->compare_func)
	{
		/* Search for the first slot that has higher-order data */
		p = list->head;
		q = NULL;
		while(p) {
			compare = (*list->compare_func) (p->data, data);
			if(compare >= 0)
				break;
			q = p;
			p = p->next;
		}
		/* p should point AFTER the insertion point, q before */
		if(compare == 0
		   && !(list->options & ALP_LIST_ALLOW_DUPS)) {
			/* The matching case, replace existing entry in place */
			if(q)
				q->next = new;
			else
				list->head = new;
			new->next = p->next;
			new->prev = q;
			if(p->next)
				p->next->prev = new;
			else
				list->tail = new;
			if(list->destroy_func)
				(*list->destroy_func) (p->data);
			free(p);
			rc = 1;
		} else {
			/* Slightly easier, just insert a new value */
			if(q)
				q->next = new;
			else
				list->head = new;
			new->next = p;
			new->prev = q;
			if(p)
				p->prev = new;
			else
				list->tail = new;
		}
	} else {
		/* Pretty easy if not doing in-order list */
		p = list->tail;
		if(p)
			p->next = new;
		new->next = NULL;
		new->prev = p;
		list->tail = new;
		if(!list->head)
			list->head = new;
	}

	list->entries++;

	return rc;
}

/* ------------------------------------------------------- 
 * prv_conf_list_head/tail/next/prev()
 * prv_conf_list_get_data()
 * prv_conf_list_count()
 *	These functions simply return a pointer from a list or entry
 *	They aren't strictly necessary, but should make the user code
 *	a bit easier to read and maintain.
 * -------------------------------------------------------*/
static AlpPListEntry *prv_conf_list_head(AlpPList *list)
{
	if(!list) {
		/* debug warn - "NULL list passed to prv_conf_list_head" */
		return NULL;
	}
	return list->head;
}

/* ------------------------------------------------------- */
/* -------------------------------------------------------*/
static AlpPListEntry *prv_conf_list_tail(AlpPList *list)
{
	if(!list) {
		/* debug warn - "NULL list passed to prv_conf_list_tail" */
		return NULL;
	}
	return list->tail;
}

/* ------------------------------------------------------- */
/* -------------------------------------------------------*/
static AlpPListEntry *prv_conf_list_next(AlpPListEntry *entry)
{
	if(!entry) {
		/* debug warn - "NULL entry passed to prv_conf_list_next" */
		return NULL;
	}
	return entry->next;
}

/* ------------------------------------------------------- */
/* -------------------------------------------------------*/
static AlpPListEntry *prv_conf_list_prev(AlpPListEntry *entry)
{
	if(!entry) {
		/* debug warn - "NULL entry passed to prv_conf_list_prev" */
		return NULL;
	}
	return entry->prev;
}

/* ------------------------------------------------------- */
/* -------------------------------------------------------*/
static void * prv_conf_list_get_data(AlpPListEntry *entry)
{
	if(!entry) {
		/* debug warn - "NULL entry passed to prv_conf_list_get_data" */
		return NULL;
	}
	return entry->data;
}

/* ------------------------------------------------------- */
/* -------------------------------------------------------*/
static int prv_conf_list_count(AlpPList *list)
{
	return list->entries;
}


/* -------------------------------------------------------
 *	This function searches the list for a value using the
 *	user-supplied comparison function.  It will terminate early
 *	if it passes where the value should be in the list.
 *
 *	Returns pointer to list entry which contains value
 *		or NULL on failure
 * -------------------------------------------------------*/
static AlpPListEntry * prv_conf_list_search(AlpPList *pList, void *data)
{
	AlpPListEntry *p;
	int result;

	/* Check for parameter */
	if ((!pList) || (!data) )
	{
		return NULL;
	}

	if(pList->compare_func == NULL)
	{
		return NULL;
	}

	p = pList->head;
	while(p) {
		result = (*pList->compare_func) (p->data, data);
		if(result == 0)
			return p;
		if(result > 0)
			return NULL;
		p = p->next;
	}

	return NULL;
}


/* -------------------------------------------------------
 * prv_conf_list_search_alt()
 *	This function searches the list for a value using the search
 *	function specified by compare_func.  Note that the search cannot
 *	terminate early unless it finds the search target as we cannot
 *	assume that the list is in order for this comparison function.
 *
 *	Returns pointer to list entry which contains value
 *		or NULL on failure
 ------------------------------------------------------- */
static AlpPListEntry * prv_conf_list_search_alt(AlpPList *pList, void *data,
					      alpEntryCompare compare_func)
{
	AlpPListEntry *p;

	/* Check for parameter */
	if ((!pList) || (!data) ||(compare_func == NULL) )
	{
		return NULL;
	}

	p = pList->head;
	while(p) {
		if((compare_func) (p->data, data) == 0)
			return p;
		p = p->next;
	}

	return NULL;
}



/* -------------------------------------------------------
 *	Removes an entry from a list, calling a destructor if registered
 -------------------------------------------------------*/
static void prv_conf_list_delete_entry(AlpPList *pList, AlpPListEntry *pEntry)
{
	/* Check for parameter */
	if ((!pList) || (!pEntry))
	{
		return;
	}
	
	if(pEntry->prev) {
		pEntry->prev->next = pEntry->next;
		if(pEntry->next)
			pEntry->next->prev = pEntry->prev;
		else
			pList->tail = pEntry->prev;
	} else {
		pList->head = pEntry->next;
		if(pEntry->next)
			pEntry->next->prev = NULL;
		else
			pList->tail = NULL;
	}

	if(pList->destroy_func)
		(*pList->destroy_func) (pEntry->data);

	free(pEntry);

	pList->entries--;
}


/* -------------------------------------------------------
 *	prv_conf_list_free()
 *	Eliminates all list entries and deallocates the list structure
 -------------------------------------------------------*/
static void prv_conf_list_free(AlpPList *pList)
{
	AlpPListEntry *p, *q;


	/* Check for parameter */
	if(!pList)
		return;

	p = pList->head;
	while(p) {
		q = p->next;
		if(pList->destroy_func)
			(*pList->destroy_func) (p->data);
		free(p);
		p = q;
	}

	free(pList);
}

/* ===================================================== */

#define START_BUF_SIZE 1024
#define INC_BUF_SIZE   512



/* ----------------------------------------------------*/
/* ----------------------------------------------------*/
static AlpConfFile * prv_conf_get_file_struct(int fdes)
{
	AlpConfFile *new;
	int bytes_read;

	/* Setup a new AlpConfFile struct */
	new = malloc(sizeof(AlpConfFile));
	new->fdes = fdes;

	/* Allocate and prefill the data buffer */
	new->bufsize = START_BUF_SIZE;
	new->p = new->buf = malloc(new->bufsize);
	bytes_read = read(fdes, new->p, new->bufsize);
	new->e = new->p + bytes_read;

	return new;
}

/* ----------------------------------------------------*/
/* ----------------------------------------------------*/
static void prv_conf_free_file_struct(AlpConfFile *file)
{
	free(file->buf);
	free(file);
}


/* ----------------------------------------------------*/
/* ----------------------------------------------------*/
static char * _fill_buffer(char *low_byte, AlpConfFile *file)
{
	int off, len;

	if(low_byte == file->buf) {
		/* Buffer expansion required to fit data line */
		off = file->p - file->buf;	/* Save p/e offsets */

		file->bufsize += INC_BUF_SIZE;	/* Make a bigger buffer */
		file->buf = realloc(file->buf, file->bufsize);

		file->p = file->e = file->buf + off; /* Setup new p/e offsets */
	} else {
		/* Relocate bytes to low end of buffer to make space */
		if((len = file->e - low_byte) > 0)
			memmove(file->buf, low_byte, len);
		file->p = file->buf + len;
		file->e = file->p;
	}
	low_byte = file->buf;		/* low is now at start */

	/* Now that we've made space, fill the buffer */
	len = file->bufsize - (file->e - low_byte);
	if((len = read(file->fdes, file->e, len)) >= 0)
		file->e += len;

	return low_byte;
}


/* ----------------------------------------------------*/
/* ----------------------------------------------------*/
static char * prv_conf_read_line(AlpConfFile *file)
{
	char *data_line;
	/* int d_off; */

	data_line = file->p;
	while(1) {
		if(file->p == file->e) {
			data_line = _fill_buffer(data_line, file);
			/* EOF Check */
			if(file->p == file->e) {
				/* Must guarantee NULL term */
				/* We KNOW that p is a valid mem location */
				/* because we just tried to expand the buf */
				*file->p = '\0';
				break;
			}
		}
		if(*file->p == '\n') {
			*file->p = '\0';
			file->p++;
			break;
		}
		file->p++;
	}

	if(data_line == file->p)
		return NULL;
	else
		return strdup(data_line);
}

