/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/*!
 * @file <traces_helper.c>
 *
 * @brief C helper file for traces library. Provides support for formatting buffers.
 *
 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>

#include "traces_helper.h"
#include "traces_init.h"


#define kChunkSize	16384
#define kDefaultMemSize	256

/*	------------------------------------------------------------------- */
TTracer*				gTracer;
CRITICAL_SECTION	gSendCS;
double				dummy;

unsigned char defaultBuffer[kDefaultMemSize];


/*	-------------------------------------------------------------------
Message format:

vvXXXXXXXXXXXXXX
vv Version of trace protocol (2 bytes: Minor, Major)

================================================
For version 1.0 (0001XX...XX)
XX...XX = NNNNp0p1...pn
NNNN : total size of the message (including vvNNNN) (4 bytes)
pi :	Message parameters
		Format:	cnndd..dd
				c	: Parameter opcode (1 byte)
				n	: nn parameter data len (excluding cnn)
				d..d: parameter data

Parameters valid opcodes:
		1 : Attributes: xxxxt
			xxxx = module<<16 + code - 32 bits
			t = kind (0 = normal, 1 = EOT, 2 = binary) - 8 bits
		2 : Indentation level: i (8 bits)
		3 :	Process ID
		4 : Thread ID
		5 : Trace Data
		6 :	Time Stamp.

		128 : App Name.

		0 : End of message. Every message must end with this parameter
			As parameter data len is 0, the message MUST end with 3 0 bytes

User parameters must be >=128 (bit 7 set). This will prevent collision with
future ObjectSet parameters.
Take care that 16 bits and 32 bits values are transmitted in network byte order
(big-endian) over the network.

-----------------------------------------------------------------------
Create data will not append the trailing null parameter.
That will allow to subclass it, call the base class implementation,
and then append more parameters. The final null parameter is added by 
Send just before calling SendChunk.
Tough, the total size of the message is already set to take into account
the trailing 3 bytes that will be added so you will have to only add your
parameter size to the message size.

-----------------------------------------------------------------------
Version matching is done on the major only so increment the major
when a change made this incompatible with the reporter.
----------------------------------------------------------------------- */



/*	------------------------------------------------------------------- */
static void Trace_PrepareBuffer(TTraceSendParams* aParams, unsigned char *pBuffer, size_t bufferSize)
{
	unsigned char		theByte;
	unsigned short	theU16;
	unsigned long		theU32;
	
	#define ADD_BYTE(b)	theByte = (b); *theWorkPtr++ = theByte;
	#define ADD_U16(s)	theU16 = (s); *theWorkPtr++ = (unsigned char)(theU16 >> 8); *theWorkPtr++ = (unsigned char)theU16;
	#define ADD_U32(l)	theU32 = (l); *theWorkPtr++ = (unsigned char)(theU32 >> 24); *theWorkPtr++ = (unsigned char)(theU32 >> 16);\
						*theWorkPtr++ = (unsigned char)(theU32 >> 8); *theWorkPtr++ = (unsigned char)theU32;
		
	unsigned char		*theWorkPtr, *theMsgPtr;
	size_t			theWorkSize, theMsgLen;

	
	//save the correct message parameters:
	theMsgPtr = aParams->itsData;		
	theMsgLen = aParams->itsDataLen;		//Should be fixed!!!

	//Set new values:
	aParams->itsData = pBuffer;
	aParams->itsDataLen = bufferSize;
	
	theWorkPtr 	= aParams->itsData ;
	theWorkSize 	= aParams->itsDataLen ; 

	// Version (2)
	ADD_U16(TRACE_VERSION);

	// Reserve place for length (4)
	ADD_U32(aParams->itsDataLen);

	// Attributes (8)
	ADD_BYTE (TraceOpcode_Attributes); ADD_U16(5);
	ADD_U32(aParams->itsIdent);
	ADD_BYTE (aParams->itsKind);

	// Indentation level (4)
	ADD_BYTE (TraceOpcode_IndentLevel); ADD_U16(1);
	ADD_BYTE (aParams->itsIndentLevel);

	// Time stamp (7)
	ADD_BYTE(TraceOpcode_TimeStamp); ADD_U16(4);
	// So in the tracer we can assume that a null value means value not present.
	// The bias is 1 millisec only if the system timer is 0 (one millec every 49.7 days)
	if (aParams->itsTimeStamp == 0)
		aParams->itsTimeStamp++;
	ADD_U32(aParams->itsTimeStamp);

	// Thread ID and Process ID (7)
	ADD_BYTE(TraceOpcode_ProcessID); ADD_U16(4);
	ADD_U32(getpid());

	// (7)
	ADD_BYTE(TraceOpcode_ThreadID); ADD_U16(4);
	ADD_U32(pthread_self());

	// (3+++)
	if (gTracer->itsfFirstTrace)
	{
		size_t theLen = strlen(sAppName) + 1;
		ADD_BYTE(TraceOpcode_AppName); ADD_U16(theLen);
		memcpy(theWorkPtr, sAppName, theLen);
		theWorkPtr += theLen;
	}

	// (3)
	// Trace Data. The trailing 0 will be transmitted over the network...
	ADD_BYTE(TraceOpcode_Data); ADD_U16(theMsgLen);

	//And the remaining part: the message:
	memcpy(theWorkPtr, theMsgPtr, theMsgLen);
	theWorkPtr += theMsgLen;

	memset(theWorkPtr, 0, 3);
}

/*	------------------------------------------------------------------- */


void Trace_Send(TTraceSendParams* aParams)
{
	size_t		theSize, theOrgSize;
	unsigned char		*theData, *theOrgData;
	TraceKind	theKind;
	
	unsigned char		*pBuffer = NULL;
	size_t			bufferSize = 0;
	size_t			iHeaderSize = 0;


	EnterCriticalSection(&gSendCS);

	if (gTracer->itsOutputKind == OutputKind_HumanReadable)
	{
		gTracer->itsSendProc(gTracer, aParams);
		goto Trace_Send_end;
	}

	theSize = theOrgSize = aParams->itsDataLen;
	theData = theOrgData = aParams->itsData;
	theKind = aParams->itsKind;
	
	while (theSize != 0)
	{
		if (theSize > kChunkSize)
		{
			aParams->itsDataLen = kChunkSize;
			// If the trace is a EOT, only the last chunk must be sent this way
			if (theKind == TraceKind_EOL)
				aParams->itsKind = TraceKind_Normal;
			theSize -= kChunkSize;
		}
		else
		{
			aParams->itsDataLen = theSize;
			// Reset the original kind in case it has been changed.
			aParams->itsKind = theKind;
			theSize = 0;
		}

		aParams->itsData = theData;
		theData += aParams->itsDataLen;

		//Prepare the buffer allocation:
		iHeaderSize =  2 + 4 + 8 + 4 + 7 + 7 + 7 + 3 + 3 ;
		if (gTracer->itsfFirstTrace)
			iHeaderSize += 4 + strlen(sAppName);

		//Final size!
		bufferSize = iHeaderSize+ aParams->itsDataLen;

		//And allocate memory !
		pBuffer = (unsigned char*)malloc(bufferSize);
		if (pBuffer == 0)
		{
			fprintf(stderr,"Traces: Memory allocation error in Trace_Send() !!!\n");
			if (aParams->itsDataLen >= (kDefaultMemSize-iHeaderSize))	//msg too big to be displayed, trunc it !
			{
				aParams->itsDataLen = kDefaultMemSize-iHeaderSize-1;	//Not sure for -1 ;)
			}
			
			Trace_PrepareBuffer(aParams, defaultBuffer, kDefaultMemSize);			
		}
		else 	/* Call with mem buffer */ 
			Trace_PrepareBuffer(aParams, pBuffer, bufferSize);
				
		gTracer->itsSendProc(gTracer, aParams);

		if (pBuffer)
			free (pBuffer);
		
/*		if (aParams->itsData)
			free(aParams->itsData);
*/
		
	}
	aParams->itsData = theOrgData;
	aParams->itsDataLen = theOrgSize;

Trace_Send_end:
	LeaveCriticalSection(&gSendCS);
}

/*	------------------------------------------------------------------- */
void Trace_Close(TTracer* aTracer)
{
	aTracer->itsCloseProc(aTracer);

	
	if (aTracer->itsfFirstTrace == 314)	/* This generates a warning */
	{
		// If floating point is not used inside the DLL, floating point support may be
		// removed from the code by the linker, causing failure of printf-style calls
		// involving fp types. See description for the MS C runtime library error R6002.
		dummy = 3.14;
	}

	aTracer->itsfFirstTrace = 1;		/* For True */
}

/*	------------------------------------------------------------------- */
void Trace_SendText(TTraceSendParams* aParams, const char* aFormat,  va_list anArgs)
{
	int				theLen;
	unsigned char	buffer[SPRINTF_BUFFER_SIZE];
	struct timeval tp;

	aParams->itsData = (unsigned char*) buffer;
	aParams->itsIndentLevel = 0;
	gettimeofday(&tp, NULL);
	aParams->itsTimeStamp = (unsigned long) (tp.tv_sec * 1000) + (tp.tv_usec / 1000);

	theLen = vsnprintf((char*)aParams->itsData, SPRINTF_BUFFER_SIZE, aFormat, anArgs);
	
	if (theLen == -1)
	{
		theLen = SPRINTF_BUFFER_SIZE;
		
		buffer[SPRINTF_BUFFER_SIZE-1] = ']';
		buffer[SPRINTF_BUFFER_SIZE-2] =
		buffer[SPRINTF_BUFFER_SIZE-3] =
		buffer[SPRINTF_BUFFER_SIZE-4] = '.';
		buffer[SPRINTF_BUFFER_SIZE-5] = '[';
	}
	
	if (theLen > 0)
	{
		aParams->itsDataLen = (size_t)theLen;
		Trace_Send(aParams);
	}
}
