/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
#include <hiker/unit_test.h>
#include <cppunit/extensions/HelperMacros.h>

// faster CPPUNIT_ASSERT
#define ALP_CPPUNIT_ASSERT(x) if (!(x)) CPPUNIT_ASSERT(x)

#include <hiker/volumesvcs.h>
#include <hiker/notifymgr.h>
#include <pthread.h>
#include <time.h>

#define CPPUNIT_ASSERT_FMT( X, fmt, ...) {\
	char cppunit_tmp[256]; \
	sprintf(cppunit_tmp,fmt,__VA_ARGS__); \
	CPPUNIT_ASSERT_MESSAGE( cppunit_tmp, X ); }

#define CPPUNIT_ERR( got, shouldbe ) CPPUNIT_ASSERT_FMT( got == shouldbe, "expected alp_status_t:%0X but got %0X",shouldbe,got)

#define CB_ASSERT( X, S, U) \
		CPPUNIT_ASSERT_FMT( X,"%s  cb: %d, da: %d, dr: %d va: %d, vr: %d"\
			"\nda_udi: %s\ndr_udi: %s\nva_udi: %s\nvr_udi: %s\ntarget udi: %s",\
			   S, cb_cnt, da_cnt, dr_cnt, va_cnt, vr_cnt,da_udi,dr_udi,va_udi,vr_udi, U);

/*
 * allow a longtime but break early if the condition occurs.
 */
#define NOTIFY_TIMEOUT 60
#define SLEEP_WHILE(exp) {int n=NOTIFY_TIMEOUT;while ( exp && n--) sleep(1);}

#define TESTTIMEOUT 60
#define BIND_TARGET "/tmp"
class VolumeSvcsTests : public CPPUNIT_NS::TestFixture
{
	CPPUNIT_TEST_SUITE(VolumeSvcsTests);
	CPPUNIT_TEST(InfoArrayTest);
	CPPUNIT_TEST(FstypesTest);
	CPPUNIT_TIMED_TEST(BindTest, TESTTIMEOUT);
	CPPUNIT_TIMED_TEST(MountUmountTest,TESTTIMEOUT );
	CPPUNIT_TIMED_TEST(MountUmountHWTest, TESTTIMEOUT);
	CPPUNIT_TIMED_TEST(RenameTest, TESTTIMEOUT);
	CPPUNIT_TIMED_TEST(RenameHWTest, TESTTIMEOUT);
	CPPUNIT_TEST(InfoTest);
	CPPUNIT_TEST(InfoHWTest);
	CPPUNIT_TIMED_TEST(MakeFsTest, TESTTIMEOUT);
	CPPUNIT_TEST(MakeFsAndRenameTest);
	CPPUNIT_TEST(FsInfoArray_NegativeTests);
	CPPUNIT_TEST(FsTypes_NegativeTests);
	CPPUNIT_TEST(MountUmount_NegativeTests);
	CPPUNIT_TEST(Bind_NegativeTests);
	CPPUNIT_TEST(Rename_NegativeTests);
	CPPUNIT_TEST_SUITE_END();

	protected:
		void InfoArrayTest();
		void FstypesTest();
		void BindTest();
		void MountUmountTest();
		void MountUmountHWTest();
		void RenameTest();
		void RenameHWTest();
		void InfoTest();
		void InfoHWTest();
		void MakeFsTest();
		void MakeFsAndRenameTest();
		void FsInfoArray_NegativeTests();
		void FsTypes_NegativeTests();
		void MountUmount_NegativeTests();
		void Bind_NegativeTests();
		void Rename_NegativeTests();
	public:
		void setUp();
		void tearDown();
	private:
		bool hardwareCheck();
		void bindTestCheck(AlpVolumeSvcsFSinfo *e);
		void infoTestCheck(AlpVolumeSvcsFSinfo *fsinfo, AlpVolumeSvcsFSinfo *e);

};

ALP_ROOT_TEST_SUITE_REGISTRATION("OS-021", CPPUNIT_REGISTRY_TYPE_AUTO);
CPPUNIT_TEST_SUITE_NAMED_REGISTRATION(VolumeSvcsTests, "OS-021");

static int va_cnt=0;
static int vr_cnt=0;
static int da_cnt=0;
static int dr_cnt=0;
static int cb_cnt=0;
static char *va_udi=NULL;
static char *vr_udi=NULL;
static char *da_udi=NULL;
static char *dr_udi=NULL;

int va_cb(const char * notifyType, void * details, int length, int * handled,
          const char * appID, void * userData)
{
	va_cnt++;
	cb_cnt++;
	if (va_udi)
		g_free(va_udi);
	va_udi = strdup( (char*)details );
	return 0;
}


int vr_cb(const char * notifyType, void * details, int length, int * handled,
          const char * appID, void * userData)
{
	vr_cnt++;
	cb_cnt++;
	if (vr_udi)
		g_free(vr_udi);
	vr_udi = strdup( (char*)details );
	return 0;
}

int da_cb(const char * notifyType, void * details, int length, int * handled,
          const char * appID, void * userData)
{
	da_cnt++;
	cb_cnt++;
	if (da_udi)
		g_free(da_udi);
	da_udi = strdup( (char*)details );
	return 0;
}

int dr_cb(const char * notifyType, void * details, int length, int * handled,
          const char * appID, void * userData)
{
	dr_cnt++;
	cb_cnt++;
	if (dr_udi)
		g_free(dr_udi);
	dr_udi = strdup( (char*)details );
	return 0;
}

void reset()
{
	cb_cnt = da_cnt = dr_cnt = va_cnt = vr_cnt = 0;
	if (dr_udi)
		g_free(dr_udi);
	if (da_udi)
		g_free(da_udi);
	if (vr_udi)
		g_free(vr_udi);
	if (va_udi)
		g_free(va_udi);
	va_udi = vr_udi = da_udi = dr_udi = NULL;

}

/*
 * umount (and thus rename and makefs) can return a new device busy error so
 * the test has to allow for that.  We just retry.  Typically it is due to bundlemgr
 * poking around on the device or bindmount looking for apps (remember the system
 * is typically up and running for these tests)
 */
alp_status_t doUmount( const char *udi )
{
	int timeout = 90; // it can take up to 1 min+ on very slow devices
	alp_status_t err = ALP_OK;

	while ( timeout-- > 0 &&
			(err=alp_volumesvcs_umount(udi)) == ALP_STATUS_VOLUMESVCS_DEVICE_BUSY )
		sleep(1);
	return err;
}

alp_status_t doRename( const char *udi, const char *name )
{
	int timeout = 90; // it can take up to 1 min+ on very slow devices
	alp_status_t err = ALP_OK;

	while ( timeout-- > 0 &&
			(err=alp_volumesvcs_rename_volume(udi, name)) == ALP_STATUS_VOLUMESVCS_DEVICE_BUSY )
		sleep(1);
	return err;
}



void VolumeSvcsTests::setUp()
{
	alp_status_t err = ALP_OK;

	err = alp_notify_register_server(ALP_NOTIFY_EVENT_VOLUME_ADDED, va_cb, NULL,0,0);
	CPPUNIT_ERR(err, ALP_OK);
	err = alp_notify_register_server(ALP_NOTIFY_EVENT_VOLUME_REMOVED, vr_cb, NULL,0,0);
	CPPUNIT_ERR(err, ALP_OK);
	err = alp_notify_register_server(ALP_NOTIFY_EVENT_DEVICE_ADDED , da_cb, NULL,0,0);
	CPPUNIT_ERR(err, ALP_OK);
	err = alp_notify_register_server(ALP_NOTIFY_EVENT_DEVICE_REMOVED, dr_cb, NULL,0,0);
	CPPUNIT_ERR(err, ALP_OK);
}

void VolumeSvcsTests::tearDown()
{
	alp_status_t err;

	err = alp_notify_unregister_server(ALP_NOTIFY_EVENT_VOLUME_ADDED, 0);
	CPPUNIT_ERR(err, ALP_OK);
	err = alp_notify_unregister_server(ALP_NOTIFY_EVENT_VOLUME_REMOVED, 0);
	CPPUNIT_ERR(err, ALP_OK);
	err = alp_notify_unregister_server(ALP_NOTIFY_EVENT_DEVICE_ADDED, 0);
	CPPUNIT_ERR(err, ALP_OK);
	err = alp_notify_unregister_server(ALP_NOTIFY_EVENT_DEVICE_REMOVED, 0);
	CPPUNIT_ERR(err, ALP_OK);

}

void VolumeSvcsTests::InfoArrayTest()
{
	alp_status_t err;
	GArray *fsinfo;
	AlpVolumeSvcsFSinfo *e;

	err = alp_volumesvcs_fsinfo_array( &fsinfo );
	CPPUNIT_ERR(err, ALP_OK);

	// not really much we can look at here as we do not know what is supposed to be there
	err = alp_volumesvcs_fsinfo_array_free( fsinfo );
	CPPUNIT_ERR(err, ALP_OK);
}

void VolumeSvcsTests::FstypesTest()
{
	alp_status_t err = ALP_OK;
	char *p = NULL;
	bool fat16, fat32;
	guint i;
	GPtrArray *fstypes;

	err = alp_volumesvcs_fstypes( &fstypes );
	CPPUNIT_ERR(err, ALP_OK);
	/*
	 * check to be sure FAT16 and FAT32 are there
	 */
	fat16 = fat32 = FALSE;
	for (i=0; i<fstypes->len; i++) {
		p = (char*)g_ptr_array_index( fstypes, i);
		if (!strcasecmp( "FAT16", p))
			fat16 = TRUE;
		else if (!strcasecmp( "FAT32", p))
			fat32 = TRUE;
	}
	CPPUNIT_ASSERT_FMT( fat16 && fat32, "got unepxected filesystem type %s",p );

}

static AlpVolumeSvcsFSinfo *findBindMount( GArray *fsinfo)
{
	guint i;
	AlpVolumeSvcsFSinfo *e;

	for (i=0; i<fsinfo->len; i++) {
		e = &g_array_index( fsinfo, AlpVolumeSvcsFSinfo, i);
		if (e->flags & ALP_VOLUMESVCS_FLAGS_BINDMOUNT) {
			return e;
		}
	}
	return NULL;
}

static AlpVolumeSvcsFSinfo *getInfo( GArray *fsinfo)
{
	guint i;
	AlpVolumeSvcsFSinfo *e;

	for (i=0; i<fsinfo->len; i++)
	{
		e = &g_array_index( fsinfo, AlpVolumeSvcsFSinfo, i);
		if (e->flags & ALP_VOLUMESVCS_FLAGS_PARTITION) {
			return e;
		}
	}
	return NULL;
}

void VolumeSvcsTests::InfoTest()
{
	alp_status_t err;
	GArray *fsinfo_array;
	AlpVolumeSvcsFSinfo *fsinfo, *e;
	char tmpname[256];

	err = alp_volumesvcs_fsinfo_array( &fsinfo_array );
	CPPUNIT_ERR(err, ALP_OK);
	/*
	 * don't add a new one if we already have one (for multiple tests runs, since
	 * currently restarting volumesvcsd is the only way to delete them
	 */
	if ((e = findBindMount(fsinfo_array)) == NULL) {
		sprintf( tmpname, "vs_test.%d",(int)time(NULL));
		err = alp_volumesvcs_bind( tmpname, BIND_TARGET);
		CPPUNIT_ERR(err, ALP_OK);
		err = alp_volumesvcs_fsinfo_array_free( fsinfo_array );
		CPPUNIT_ERR( err, ALP_OK);
		err = alp_volumesvcs_fsinfo_array( &fsinfo_array );
		CPPUNIT_ERR(err, ALP_OK);
		e = findBindMount(fsinfo_array);
		CPPUNIT_ASSERT( e != NULL);
	}

	err = alp_volumesvcs_fsinfo_for_udi( &fsinfo, e->udi);
	CPPUNIT_ERR(err, ALP_OK);
	
	// check that the 2 arrays match
	infoTestCheck( fsinfo, e);
	
	err = alp_volumesvcs_fsinfo_free( fsinfo );
	CPPUNIT_ERR( err, ALP_OK);
	err = alp_volumesvcs_fsinfo_array_free( fsinfo_array );
	CPPUNIT_ERR( err, ALP_OK);
}

void VolumeSvcsTests::InfoHWTest()
{
	alp_status_t err;
	GArray *fsinfo_array;
	AlpVolumeSvcsFSinfo *fsinfo, *e;

	// This test is run on hardware only
	if (hardwareCheck()) {
		err = alp_volumesvcs_fsinfo_array( &fsinfo_array );
		CPPUNIT_ERR(err, ALP_OK);

		e = getInfo(fsinfo_array);
		CPPUNIT_ASSERT(e !=NULL);
		
		err = alp_volumesvcs_fsinfo_for_udi( &fsinfo, e->udi);
		CPPUNIT_ERR(err, ALP_OK);
		
		// check that the 2 arrays match
		infoTestCheck( fsinfo, e);
		
		// free the arrays
		err = alp_volumesvcs_fsinfo_free( fsinfo );
		CPPUNIT_ERR( err, ALP_OK);
		err = alp_volumesvcs_fsinfo_array_free( fsinfo_array );
		CPPUNIT_ERR( err, ALP_OK);
		
	}
} 

void VolumeSvcsTests::infoTestCheck(AlpVolumeSvcsFSinfo *fsinfo, AlpVolumeSvcsFSinfo *e)
{
	//check that the two arrays match
	CPPUNIT_ASSERT(!strcasecmp(fsinfo->udi, e->udi));
	CPPUNIT_ASSERT(!strcasecmp(fsinfo->udi, e->udi));
	CPPUNIT_ASSERT(!strcasecmp(fsinfo->card_type, e->card_type));
	CPPUNIT_ASSERT(fsinfo->flags == e->flags);
#define CPP_STR(x) CPPUNIT_ASSERT( (fsinfo->x == NULL && e->x == NULL) || \
		strcasecmp(fsinfo->x, e->x) == 0)
	CPP_STR(fs_path);
	CPP_STR(dev_path);
	CPP_STR(name);
	CPP_STR(slot);
	CPP_STR(vendor_model);
	CPP_STR(uuid);
#undef CPP_STR
	CPPUNIT_ASSERT( fsinfo->total_space == e->total_space);
	CPPUNIT_ASSERT( fsinfo->used_space == e->used_space);
	CPPUNIT_ASSERT( fsinfo->total_files == e->total_files);
	CPPUNIT_ASSERT( fsinfo->used_files == e->used_files);
}

void VolumeSvcsTests::BindTest()
{
	alp_status_t err;
	AlpVolumeSvcsFSinfo *e;
	char tmpname[100];
	char expectedpath[100];
	GArray *fsinfo;

	err = alp_volumesvcs_fsinfo_array( &fsinfo );
	CPPUNIT_ERR(err, ALP_OK);
	/*
	 * don't add a new one if we already have one (for multiple tests runs, since
	 * restarting volumesvcsd is the only way to delete them
	 */
	if ((e = findBindMount(fsinfo)) != NULL) {
		bindTestCheck(e);
		return;
	}
	reset();

	sprintf( tmpname, "vs_test.%d",(int)time(NULL));
	sprintf( expectedpath, "/media/%s", tmpname);
		
	err = alp_volumesvcs_bind( tmpname, BIND_TARGET);
	CPPUNIT_ERR(err, ALP_OK);

	err = alp_volumesvcs_fsinfo_array_free( fsinfo );
	CPPUNIT_ERR(err, ALP_OK);
	err = alp_volumesvcs_fsinfo_array( &fsinfo );
	CPPUNIT_ERR(err, ALP_OK);
	e = findBindMount(fsinfo);

	// checking for a valid array and valid returns upon initial binding
	CPPUNIT_ASSERT(!strcasecmp(e->fs_path, expectedpath)); 
	CPPUNIT_ASSERT(e->flags == ALP_VOLUMESVCS_FLAGS_BINDMOUNT);
	bindTestCheck(e);
	
	SLEEP_WHILE((cb_cnt == 0));
	CB_ASSERT( cb_cnt == 1, "cb != 1", e->udi);
	CB_ASSERT( va_cnt == 1, "va != 1", e->udi);
	CPPUNIT_ASSERT( va_udi != NULL );
	CPPUNIT_ASSERT_FMT( !strcasecmp(e->udi, va_udi ), "va_udi: %s != e->udi: %s",va_udi, e->udi);

	err = alp_volumesvcs_fsinfo_array_free( fsinfo );
	CPPUNIT_ERR(err, ALP_OK);
}

void VolumeSvcsTests::bindTestCheck(AlpVolumeSvcsFSinfo *e)
{
	alp_status_t err;

	CPPUNIT_ASSERT( e != NULL );
	CPPUNIT_ASSERT(g_file_test( e->fs_path, G_FILE_TEST_IS_DIR ));
	CPPUNIT_ASSERT( e->fs_path != NULL);
	CPPUNIT_ASSERT(!strcasecmp(e->slot,"Internal"));
		
}

void VolumeSvcsTests::RenameTest()
{
	alp_status_t err;
	char tmpname[100];
	char oldudi[100];
	char oldname[100];
	AlpVolumeSvcsFSinfo *e, *entry;
	GArray *fsinfo;
	guint i;

	err = alp_volumesvcs_fsinfo_array( &fsinfo );
	CPPUNIT_ERR(err, ALP_OK);
	/*
	 * sort of silly since rename is for msdos volumes but its in so might as well test it
	 * note no 11 char limit on the name here
	 */
	if ((e = findBindMount(fsinfo)) == NULL) {
		reset();
		sprintf( tmpname, "vs_test.%d",(int)time(NULL));
		err = alp_volumesvcs_bind( tmpname, BIND_TARGET);
		CPPUNIT_ERR(err, ALP_OK);
		err = alp_volumesvcs_fsinfo_array_free(fsinfo);
		CPPUNIT_ERR(err, ALP_OK);
		err = alp_volumesvcs_fsinfo_array( &fsinfo );
		CPPUNIT_ERR(err, ALP_OK);
		e = findBindMount(fsinfo);
		CPPUNIT_ASSERT(e != NULL);
		SLEEP_WHILE((cb_cnt == 0));
		CB_ASSERT( cb_cnt == 1, "cb != 1", e->udi);
		CB_ASSERT( va_cnt == 1, "va != 1", e->udi);
		CPPUNIT_ASSERT( va_udi != NULL );
		CPPUNIT_ASSERT_FMT( !strcasecmp(e->udi, va_udi ), "va_udi: %s != e->udi: %s",va_udi, e->udi);
	}

	if (e->fs_path == NULL) {
		reset();
		err = alp_volumesvcs_mount( e->udi );
		CPPUNIT_ERR(err, ALP_OK);
		SLEEP_WHILE((cb_cnt == 0));
		CB_ASSERT( cb_cnt == 1, "cb != 1", e->udi);
		CB_ASSERT( va_cnt == 1, "va != 1", e->udi);
	}

	strcpy( oldname, e->name );
	sprintf( tmpname, "%s.rename", e->name );
	reset();
	err = doRename( e->udi, tmpname );
	CPPUNIT_ERR(err, ALP_OK);
	SLEEP_WHILE((cb_cnt < 2));
	CB_ASSERT( vr_cnt == 1, "va != 1", e->udi);
	CB_ASSERT( va_cnt == 1, "va != 1", e->udi);
	err = alp_volumesvcs_fsinfo_for_udi( &entry, e->udi);
	CPPUNIT_ERR(err, ALP_OK);
	CPPUNIT_ASSERT( entry->fs_path != NULL);

	reset();
	err = doUmount( entry->udi );
	CPPUNIT_ERR(err, ALP_OK);
	SLEEP_WHILE((cb_cnt == 0));
	CB_ASSERT( cb_cnt == 1, "cb != 1", entry->udi);
	CB_ASSERT( vr_cnt == 1, "va != 1", entry->udi);

	reset();
	err = doRename( entry->udi, oldname );
	CPPUNIT_ERR(err, ALP_OK);
	SLEEP_WHILE((cb_cnt == 0));
	CB_ASSERT( cb_cnt == 1, "cb != 1", entry->udi);
	CB_ASSERT( va_cnt == 1, "va != 1", entry->udi);

	err = alp_volumesvcs_fsinfo_array_free( fsinfo );
	CPPUNIT_ERR(err, ALP_OK);
	err = alp_volumesvcs_fsinfo_free( entry );
	CPPUNIT_ERR(err, ALP_OK);
}

void VolumeSvcsTests::RenameHWTest() //volume mounted test
{
	alp_status_t err;
	char tmpname[100];
	char oldname[100];
	AlpVolumeSvcsFSinfo *volume;	
	AlpVolumeSvcsFSinfo *newEntry;
	GArray *fsinfo;
	guint i;

	if ( hardwareCheck()) {
		err = alp_volumesvcs_fsinfo_array( &fsinfo );
		CPPUNIT_ERR(err, ALP_OK);

		volume = getInfo(fsinfo);
		CPPUNIT_ASSERT(volume != NULL);
		
		// mount the volume if not already done
		if (volume->fs_path == NULL) {
			reset();
			err = alp_volumesvcs_mount( volume->udi );
			CPPUNIT_ERR(err, ALP_OK);
			CPPUNIT_ASSERT_MESSAGE("Bug 140169 filed, err != ALP_OK", err == ALP_OK);
			SLEEP_WHILE((cb_cnt == 0));
			CB_ASSERT( cb_cnt == 1, "cb != 1",volume->udi);
			CB_ASSERT( va_cnt == 1, "va != 1",volume->udi);
		}
	
		strcpy( oldname, volume->name );
		/*
		 * note using a fixed name is bad since if the test fails the card might be
		 * left with that fixed name and then we wouldn't know if the next rename worked
		 * so geneate a max 11 char new name (might be less here)
		 */
		sprintf( tmpname, "REN_%d", ((int)time(NULL)) % 100000 );

		reset();
		err = doRename( volume->udi, tmpname );
		CPPUNIT_ERR(err, ALP_OK);
		SLEEP_WHILE((cb_cnt < 2));
		CB_ASSERT( cb_cnt == 2, "cb != 2", volume->udi);
		CB_ASSERT( vr_cnt == 1, "vr != 1", volume->udi);
		CB_ASSERT( va_cnt == 1, "va != 1", volume->udi);
		
        err = alp_volumesvcs_fsinfo_for_udi( &newEntry, volume->udi );	
		CPPUNIT_ERR(err, ALP_OK);

		CPPUNIT_ASSERT( newEntry != NULL);
		CPPUNIT_ASSERT( newEntry->udi != NULL);
		CPPUNIT_ASSERT( strcasecmp(newEntry->udi, volume->udi) == 0);
		CPPUNIT_ASSERT_FMT( newEntry->fs_path != NULL, "udi %s\n fs_path %s\n name %s\n",
			   newEntry->udi, newEntry->fs_path, newEntry->name	);
		CPPUNIT_ASSERT( strcasecmp(newEntry->name, tmpname) == 0);
		
		err = alp_volumesvcs_fsinfo_free( newEntry );
		CPPUNIT_ERR(err, ALP_OK);

		reset();
		err = doUmount( volume->udi );
		CPPUNIT_ERR(err, ALP_OK);
		SLEEP_WHILE((cb_cnt == 0));
		CB_ASSERT( cb_cnt == 1, "cb != 1", volume->udi);
		CB_ASSERT( vr_cnt == 1, "vr != 1", volume->udi);

		// should be able to rename volume (back to original) even if it is unmounted and
		// it should mount automatically after rename is done.  only one event this time,
		// the mount
		reset();
		err = doRename( volume->udi, oldname );
		CPPUNIT_ERR(err, ALP_OK);
		SLEEP_WHILE((cb_cnt == 0));
		CB_ASSERT( cb_cnt == 1, "cb != 1", volume->udi);
		CB_ASSERT( va_cnt == 1, "va != 1", volume->udi);

		// check the volume name one last time
		err = alp_volumesvcs_fsinfo_for_udi( &newEntry, volume->udi);
		CPPUNIT_ERR(err, ALP_OK);

		CPPUNIT_ASSERT( newEntry != NULL);
		CPPUNIT_ASSERT( newEntry->udi != NULL);
		CPPUNIT_ASSERT( strcasecmp(newEntry->udi, volume->udi) == 0);
		CPPUNIT_ASSERT_FMT( newEntry->fs_path != NULL, "udi %s\n fs_path %s\n name %s\n",
			   newEntry->udi, newEntry->fs_path, newEntry->name	);
		CPPUNIT_ASSERT_FMT( strcasecmp(newEntry->name, oldname) == 0,
				"udi: %s, expected name: %s got: %s", newEntry->udi, oldname, newEntry->name);

		err = alp_volumesvcs_fsinfo_free( newEntry );
		CPPUNIT_ERR(err, ALP_OK);
		err = alp_volumesvcs_fsinfo_array_free( fsinfo );
		CPPUNIT_ERR(err, ALP_OK);

	}
} 

void VolumeSvcsTests::MountUmountTest()
{
	alp_status_t err;
	char tmpname[100];
	AlpVolumeSvcsFSinfo *e;
	GArray *fsinfo;

	err = alp_volumesvcs_fsinfo_array( &fsinfo );
	CPPUNIT_ERR(err, ALP_OK);

	if ((e = findBindMount(fsinfo)) == NULL) {
		reset();

		sprintf( tmpname, "vs_test.%d",(int)time(NULL));
		err = alp_volumesvcs_bind( tmpname, BIND_TARGET);
		CPPUNIT_ERR(err, ALP_OK);
		err = alp_volumesvcs_fsinfo_array_free(fsinfo);
		CPPUNIT_ERR(err, ALP_OK);
		err = alp_volumesvcs_fsinfo_array( &fsinfo );
		CPPUNIT_ERR(err, ALP_OK);
		e = findBindMount(fsinfo);
		CPPUNIT_ASSERT(e != NULL);

		SLEEP_WHILE((cb_cnt == 0));
		CB_ASSERT( cb_cnt == 1, "cb != 1", e->udi);
		CB_ASSERT( va_cnt == 1, "va != 1", e->udi);
		CPPUNIT_ASSERT( va_udi != NULL );
		CPPUNIT_ASSERT_FMT( !strcasecmp(e->udi, va_udi ), "va_udi: %s != e->udi: %s",va_udi, e->udi);
	}

	if (e->fs_path) {
		/*
		 * its already mounted so start by unmounting
		 */
		reset();
		err = doUmount( e->udi );
		CPPUNIT_ERR(err, ALP_OK);
		CPPUNIT_ASSERT_MESSAGE("Bug 140172 filed, !g_file_test fails",
				!g_file_test( e->fs_path, G_FILE_TEST_IS_DIR ));
		SLEEP_WHILE((cb_cnt == 0));
		CB_ASSERT( cb_cnt == 1, "cb != 1", e->udi);
		CB_ASSERT( vr_cnt == 1, "vr != 1", e->udi);
		CPPUNIT_ASSERT( vr_udi != NULL );
		CPPUNIT_ASSERT_FMT( !strcasecmp(e->udi, vr_udi ), "vr_udi: %s != e->udi: %s",vr_udi, e->udi);
	}
	/*
	 * mount it and make sure it appears as a dir at least
	 */
	reset();
	err = alp_volumesvcs_mount( e->udi );
	CPPUNIT_ERR(err, ALP_OK);
	SLEEP_WHILE((cb_cnt == 0));
	CB_ASSERT( cb_cnt == 1, "cb != 1", e->udi);
	CB_ASSERT( va_cnt == 1, "va != 1", e->udi);
	CPPUNIT_ASSERT( va_udi != NULL );
	CPPUNIT_ASSERT_FMT( !strcasecmp(e->udi, va_udi ), "va_udi: %s != e->udi: %s",va_udi, e->udi);
	CPPUNIT_ASSERT( e != NULL );
	CPPUNIT_ASSERT( e->fs_path != NULL);
	CPPUNIT_ASSERT( g_file_test( e->fs_path, G_FILE_TEST_IS_DIR ));

	/*
	 * unmount it and make sure the dir goes away
	 */
	reset();
	err = doUmount( e->udi );
	CPPUNIT_ERR(err, ALP_OK);
	CPPUNIT_ASSERT( !g_file_test( e->fs_path, G_FILE_TEST_IS_DIR ));
	SLEEP_WHILE((cb_cnt == 0));
	CB_ASSERT( cb_cnt == 1, "cb != 1", e->udi);
	CB_ASSERT( vr_cnt == 1, "vr != 1", e->udi);
	CPPUNIT_ASSERT_FMT( !strcasecmp(e->udi, vr_udi ), "vr_udi: %s != e->udi: %s",vr_udi, e->udi);

	err = alp_volumesvcs_fsinfo_array_free( fsinfo );
	CPPUNIT_ERR(err, ALP_OK);
}

void VolumeSvcsTests::MountUmountHWTest()
{
	alp_status_t err;
	char tmpname[100];
	AlpVolumeSvcsFSinfo *e;
	GArray *fsinfo;

	if ( hardwareCheck()) {
		err = alp_volumesvcs_fsinfo_array( &fsinfo );
		CPPUNIT_ERR(err, ALP_OK);

		e = getInfo(fsinfo);
		CPPUNIT_ASSERT(e != NULL);

		if (e != NULL) {
			// start by unmounting if mounted already
			if (e->fs_path) {
				reset();
				err = doUmount( e->udi );
				SLEEP_WHILE((cb_cnt == 0));
				CB_ASSERT( cb_cnt == 1, "cb != 1", e->udi);
				CB_ASSERT( vr_cnt == 1, "vr != 1", e->udi);
				CPPUNIT_ERR(err, ALP_OK);
				CPPUNIT_ASSERT( !g_file_test( e->fs_path, G_FILE_TEST_IS_DIR ));
			}

			// mount the volume and make sure it's a directory at least
			reset();
			err = alp_volumesvcs_mount( e->udi );
			SLEEP_WHILE((cb_cnt == 0));
			CB_ASSERT( cb_cnt == 1, "cb != 1", e->udi);
			CB_ASSERT( va_cnt == 1, "va != 1", e->udi);
			CPPUNIT_ERR(err, ALP_OK);
			CPPUNIT_ASSERT_MESSAGE("Bug 140169 filed, err != ALP_OK", err == ALP_OK);
			
			err = alp_volumesvcs_fsinfo_array_free(fsinfo);
			CPPUNIT_ERR(err, ALP_OK);
			err = alp_volumesvcs_fsinfo_array( &fsinfo );
			CPPUNIT_ERR(err, ALP_OK);
			
			e = getInfo(fsinfo);
			CPPUNIT_ASSERT( e != NULL );
			CPPUNIT_ASSERT( e->fs_path != NULL);
			CPPUNIT_ASSERT( g_file_test( e->fs_path, G_FILE_TEST_IS_DIR ));

			// unmount it and make sure the dir goes away
			reset();
			err = doUmount( e->udi );
			SLEEP_WHILE((cb_cnt == 0));
			CB_ASSERT( cb_cnt == 1, "cb != 1", e->udi);
			CB_ASSERT( vr_cnt == 1, "vr != 1", e->udi);
			CPPUNIT_ERR(err, ALP_OK);
			CPPUNIT_ASSERT( !g_file_test( e->fs_path, G_FILE_TEST_IS_DIR ));

			if (fsinfo != NULL) {
				err = alp_volumesvcs_fsinfo_array_free( fsinfo );
				CPPUNIT_ERR(err, ALP_OK);
			}
		}
	}
}

void VolumeSvcsTests::MakeFsTest()
{
	alp_status_t err;
	char udi[100];
	char newname[100];
	guint i;
	GPtrArray *fstypes;
	AlpVolumeSvcsFSinfo *e;
	GArray *fsinfo;
	int timeout;

	// make sure we're running on hardware, this test can not be run on the simulator
	if (hardwareCheck()) {
		// get the fsinfo array 
		err = alp_volumesvcs_fsinfo_array( &fsinfo );
		CPPUNIT_ERR(err, ALP_OK);
	
		// create a name for volume (11 chars remember)
		sprintf( newname, "mkfs_%d",(int)time(NULL) % 100000);

		// get the udi
		e = getInfo(fsinfo);
		CPPUNIT_ASSERT(e != NULL);

		// call makefs
		if (e->udi != NULL) {
			if (e->fs_path == NULL) {
				reset();
				err = alp_volumesvcs_mount( e->udi );
				CPPUNIT_ERR(err, ALP_OK);
				SLEEP_WHILE((cb_cnt == 0));
				CB_ASSERT( cb_cnt == 1, "cb != 1", e->udi);
				CB_ASSERT( va_cnt == 1, "va != 1", e->udi);
			}
			reset();
			timeout = 90;
			while (timeout-- && (err = alp_volumesvcs_makefs( e->udi, newname, "FAT32" )) ==
					ALP_STATUS_VOLUMESVCS_DEVICE_BUSY) {
				sleep(1);
			}
			CPPUNIT_ASSERT_FMT( err == ALP_OK,
					"expected ALP_OK, got %x on makefs udi: %s name: %s FAT32",
					err, e->udi, newname);
			CPPUNIT_ERR(err, ALP_OK);
			CPPUNIT_ASSERT_MESSAGE("Bug 140173 filed, err != ALP_OK", err == ALP_OK);
			SLEEP_WHILE((cb_cnt < 2));
			CB_ASSERT( cb_cnt == 2, "cb != 2", "");
			CB_ASSERT( vr_cnt == 1, "vr != 1", e->udi);
			// free the old get the new and look for a volume
			err = alp_volumesvcs_fsinfo_array_free( fsinfo );
			CPPUNIT_ERR(err, ALP_OK);
			err = alp_volumesvcs_fsinfo_array( &fsinfo );
			CPPUNIT_ERR(err, ALP_OK);
			e = getInfo(fsinfo);
			CPPUNIT_ASSERT(e != NULL);
			CB_ASSERT( va_cnt == 1, "va != 1", e->udi);
			CPPUNIT_ASSERT_FMT( g_file_test( e->fs_path, G_FILE_TEST_IS_DIR ),
					"fs_path: %s is not a dir", e->fs_path);
			CPPUNIT_ERR(err, ALP_OK);
			CPPUNIT_ASSERT_FMT(!strcasecmp(e->name, newname),
					"new name %s does not match expected name %s",
					e->name, newname);	
		}
			
		// release the fsinfo array
		err = alp_volumesvcs_fsinfo_array_free( fsinfo );
		CPPUNIT_ERR(err, ALP_OK);
	}

}

void VolumeSvcsTests::MakeFsAndRenameTest()
{
	// make sure we're on hardware
	// get fs types available
	// create a name
	// get udi
	// call makefs
	// veryt return
	// check that volume was mounted
	// rename the volume
	// verify the volume name was changed
	// save a file to the new volume
	// see if we can get the volume
	// unmount
	// release fsinfo array if necessary
}

void VolumeSvcsTests::FsInfoArray_NegativeTests()
{
	alp_status_t err;
	
	// fsinfo_array with NULL pointer 	
	err = alp_volumesvcs_fsinfo_array( NULL );
	CPPUNIT_ERR(err, ALP_STATUS_VOLUMESVCS_INVALID_PARAMETER);
}

void VolumeSvcsTests::FsTypes_NegativeTests()
{
	alp_status_t err;
	
	// fstypes with an NULL param 
	err = alp_volumesvcs_fstypes( NULL );
	CPPUNIT_ERR(err, ALP_STATUS_VOLUMESVCS_INVALID_PARAMETER);
}

void VolumeSvcsTests::MountUmount_NegativeTests()
{
	alp_status_t err;

	// umount with an invalid uid
	err = doUmount( "this can not be a udi");
	CPPUNIT_ERR(err, ALP_STATUS_VOLUMESVCS_NOT_FOUND );

	// umount with a NULL udi
	err = doUmount( NULL );
    	CPPUNIT_ERR(err, ALP_STATUS_VOLUMESVCS_INVALID_PARAMETER);

	// mount with an invalid udi
	err = alp_volumesvcs_mount( "this can not be a udi");
	CPPUNIT_ERR(err, ALP_STATUS_VOLUMESVCS_NOT_FOUND);

	// mount with a NULL udi
	err = alp_volumesvcs_mount( NULL );
    	CPPUNIT_ERR(err, ALP_STATUS_VOLUMESVCS_INVALID_PARAMETER);
}

void VolumeSvcsTests::Bind_NegativeTests()
{
	alp_status_t err;

	// bind with a NULL udi	
	err = alp_volumesvcs_bind( NULL, "string");
	CPPUNIT_ERR( err, ALP_STATUS_VOLUMESVCS_INVALID_PARAMETER);

	// bind with a NULL path
	err = alp_volumesvcs_bind( "string", NULL );
	CPPUNIT_ERR( err, ALP_STATUS_VOLUMESVCS_INVALID_PARAMETER);

}

void VolumeSvcsTests::Rename_NegativeTests()
{
	alp_status_t err;
	char tmpname[100];
	AlpVolumeSvcsFSinfo *e;
	GArray *fsinfo;

	err = doRename( "this can not be a udi", "name");
	CPPUNIT_ERR(err, ALP_STATUS_VOLUMESVCS_NOT_FOUND );

	err = doRename( NULL, "name" );
    	CPPUNIT_ERR(err, ALP_STATUS_VOLUMESVCS_INVALID_PARAMETER);

	err = alp_volumesvcs_fsinfo_array( &fsinfo );
	CPPUNIT_ERR(err, ALP_OK);

	// verify the volume has been mounted 
	if ((e = findBindMount(fsinfo)) != NULL)
       	{
		err = doRename( e->udi, NULL );
		CPPUNIT_ERR(err, ALP_STATUS_VOLUMESVCS_INVALID_PARAMETER);
	}

	err = alp_volumesvcs_fsinfo_array_free( fsinfo );
	CPPUNIT_ERR(err, ALP_OK);
}

bool VolumeSvcsTests::hardwareCheck()
{
	char str [80];
	FILE * pFile;
	bool hwCheck = false;

	pFile = fopen ("/proc/cpuinfo","r");

	while (	fscanf (pFile, "%s", str) != EOF)
        {
	        if (!strcasecmp(str, "Hardware"))
                {
                   hwCheck = true;
                   break;
                }
       }

        fclose (pFile);

        return hwCheck;
}
