/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/*!
 *   @file volumesvcs_client.c the client side of RPC to the server
 *   @addtogroup ALP_VOLUMESVCS Volume Services
 *   @{
 */

#include <stdio.h>
#include <unistd.h>
#include <hiker/volumesvcs.h>
#include <hiker/prv/volumesvcs_prv.h>
#include <hiker/ipc.h>
#include <hiker/fail.h>
#include <string.h>

/*! @brief get info all all pluggable devices and volumes
 *  this makes an RPC call to the volume services daemon
 *  @param[in] fsinfo_array is a pointer to the GArray of AlpVolumeSvcsFSinfo structs returned
 *              a distinct call is needed here rather than g_array_free due to allowing freeing
 *              of the embeded structures not just the array elements and the array itself.
 *  @return - ALP_STATUS_OK, an GArray was returned (*fsinfo_array->length might == 0)
 */

alp_status_t alp_volumesvcs_fsinfo_array( GArray **fsinfo_array)
{
	alp_status_t as=ALP_STATUS_OK;
	alp_status_t ras=ALP_STATUS_OK;
	AlpMessage *msg=NULL;
	AlpMessage *rsp=NULL;
	AlpConnection *con=NULL;	
	guint i,len;
	gint cmd;
	AlpVolumeSvcsFSinfo *e;

	if (fsinfo_array == NULL)
		return ALP_STATUS_VOLUMESVCS_INVALID_PARAMETER;
	as = alp_ipc_channel_connect_synch(ALP_VOLUMESVCS_SERVER_CHANNEL,ALP_IPC_CONNECTION_USAGE_HINT_NONE,0,&con);
	if (as != ALP_STATUS_OK || (msg=alp_ipc_message_create()) == NULL) {
		ras = as;
	} else {
		PACKSTART( as, msg, ALP_VOLUMESVCS_CMD_FSINFO, error );
		PACKEND( as, msg, error );
		as = alp_ipc_connection_send_with_response( con, msg, 0, &rsp);
		if (as != ALP_STATUS_OK) {
			ras = as;
		} else {
			ras = alp_ipc_message_get_message_ID( rsp );
			if (ras == ALP_STATUS_OK ) {
				UNPACKSTART( as, rsp, error );
				UNPACKU32( as, rsp, &len, error );
				*fsinfo_array = g_array_sized_new( FALSE, TRUE, sizeof(AlpVolumeSvcsFSinfo), len);
				g_array_set_size( *fsinfo_array, len );
				for (i=0; i<len; i++) {
					guchar *p;
					guint l;
					e = &g_array_index( *fsinfo_array, AlpVolumeSvcsFSinfo, i);
					UNPACKSTR( as, rsp, &e->udi, error );
					UNPACKSTR( as, rsp, &e->dev_path, error );
					UNPACKSTR( as, rsp, &e->fs_path, error );
					UNPACKSTR( as, rsp, &e->name, error );
					UNPACKSTR( as, rsp, &e->slot, error );
					UNPACKSTR( as, rsp, &e->vendor_model, error );
					UNPACKSTR( as, rsp, &e->uuid, error );
					UNPACKSTR( as, rsp, &e->card_type, error );
					UNPACKBYTES( as, rsp, &p, &l, error);
					memcpy( &e->total_space, p, l);
					e->index = i;
					e->flags &= ~ALP_VOLUMESVCS_FLAGS_PRIVATE_MASK; // hide private flags
					g_free(p);
				}
			}
		}
	}
error:
	if (rsp)
		alp_ipc_message_destroy(rsp);
	if (msg)
		alp_ipc_message_destroy(msg);
	alp_ipc_channel_disconnect(con);

	return ras;
}


/*! @brief Free the fsinfo array struct
 *  this is processed locally and free's the user struct allocatd by the RPC wrapper
 *
 *  @param[in] fsinfo_array is a pointer to the GArray returned on a alp_volumesvcs_fsinfo call
 *             a distinct call is needed here rather than g_array_free due to allowing freeing
 *             of the embeded structures not just the array elements and the array itself.
 *             Passing in NULL is ok
 *  @return - ALP_STATUS_OK in all cases (a bad pointer will result in the program aborting)
 */
alp_status_t alp_volumesvcs_fsinfo_array_free( GArray *fsinfo_array)
{
	guint i;
	AlpVolumeSvcsFSinfo *fs;

	if (fsinfo_array) {
		for (i=0; i<fsinfo_array->len; i++) {
			fs = &g_array_index(fsinfo_array, AlpVolumeSvcsFSinfo, i);
			g_free( fs->udi );
			g_free( fs->card_type );
			g_free( fs->name );
			g_free( fs->fs_path );
			g_free( fs->dev_path );
			g_free( fs->vendor_model );
			g_free( fs->uuid );
		}
		g_array_free( fsinfo_array, TRUE);
	}

	return ALP_STATUS_OK;
}

/*! @brief Allocates and returns an AlpVolumeSvcsFSinfo* matching the udi parameter passed in.
 *
 *  @param [out] _info This will be allocated if the udi parameter matches a currently avilable
 *                    volume, otherwise it will remain unchaged.
 *                    This must be freed with alp_volumesvcs_fsinfo_free().
 *
 *  @param [in] udi   The UDI to match for.
 *
 *  @return  ALP_STATUS_OK if the UDI is matched.
 *          ALP_STATUS_VOLUMESVCS_NOT_FOUND if there is no match for the UDI.
 *          ALP_STATUS_VOLUMESVCS_INVALID_PARAMETER if the udi parameter is NULL.
 *
 */
alp_status_t alp_volumesvcs_fsinfo_for_udi(AlpVolumeSvcsFSinfo **info, const gchar *udi)
{
	alp_status_t result;
	GArray *fsinfo_array;
	guint i;
	AlpVolumeSvcsFSinfo *entry = NULL;
		
	if (udi == NULL || (strlen(udi) == 0)) {
		result = ALP_STATUS_VOLUMESVCS_INVALID_PARAMETER;
	} else if ( (result = alp_volumesvcs_fsinfo_array(&fsinfo_array)) == ALP_STATUS_OK) {
		for (i=0; i<fsinfo_array->len; i++) {   
			entry = &g_array_index( fsinfo_array, AlpVolumeSvcsFSinfo, i);
			if ((entry->udi != NULL) && !strcmp(udi, entry->udi) ) {
				*info = g_new0(AlpVolumeSvcsFSinfo, 1);
				(*info)->udi = g_strdup(entry->udi);
				(*info)->dev_path = g_strdup(entry->dev_path);
				(*info)->fs_path = g_strdup(entry->fs_path);
				(*info)->name = g_strdup(entry->name);
				(*info)->slot = g_strdup(entry->slot);
				(*info)->vendor_model = g_strdup(entry->vendor_model);
				(*info)->uuid = g_strdup(entry->uuid);
				(*info)->card_type = g_strdup(entry->card_type);
				(*info)->total_space = entry->total_space;
				(*info)->used_space = entry->used_space;
				(*info)->total_files = entry->total_files;
				(*info)->used_files = entry->used_files;
				(*info)->flags = entry->flags;
				(*info)->index = entry->index;
				break;
			}
		}
		if (i == fsinfo_array->len) {
			result = ALP_STATUS_VOLUMESVCS_NOT_FOUND;
		}
		(void)alp_volumesvcs_fsinfo_array_free(fsinfo_array);
	}

	return result;
}

/*! @brief Frees AlpVolumeSvcsFSinfo* allocated by alp_volumesvcs_fsinfo_for_udi().
 *
 *  @param [in] _info The pointer to free.
 *
 *  @return ALP_STATUS_OK is always returned, even when _info is NULL.
 *
 */
alp_status_t alp_volumesvcs_fsinfo_free(AlpVolumeSvcsFSinfo* info)
{
	if( info != NULL ) {
		g_free(info->udi);
		g_free(info->dev_path);
		g_free(info->fs_path);
		g_free(info->name);
		g_free(info->slot);
		g_free(info->vendor_model);
		g_free(info->uuid);
		g_free(info->card_type);
		g_free(info);
	}
	return ALP_STATUS_OK;
}

/*! @brief Mount a device
 *  @param[in] udi - refers to an entry in fsinfo to mount
 *
 *  @return - ALP_STATUS_OK, the volume is mounted (might have already been mounted)
 */

alp_status_t alp_volumesvcs_mount( const gchar *udi )
{
	alp_status_t as=ALP_STATUS_OK;
	alp_status_t ras=ALP_STATUS_OK;
	AlpMessage *msg=NULL;
	AlpMessage *rsp=NULL;
	AlpConnection *con=NULL;	
	guint i,len;
	gint cmd;

	if (udi == NULL || (strlen(udi) == 0))
		return ALP_STATUS_VOLUMESVCS_INVALID_PARAMETER;
	as = alp_ipc_channel_connect_synch(ALP_VOLUMESVCS_SERVER_CHANNEL,ALP_IPC_CONNECTION_USAGE_HINT_NONE,0,&con);
	if (as != ALP_STATUS_OK || (msg=alp_ipc_message_create()) == NULL) {
		ras = as;
	} else {
		PACKSTART( as, msg, ALP_VOLUMESVCS_CMD_MOUNT, error );
		PACKSTR( as, msg, udi, error );
		PACKEND( as, msg, error );
		as = alp_ipc_connection_send_with_response( con, msg, 0,&rsp);
		if (as == ALP_STATUS_OK) {
			ras = alp_ipc_message_get_message_ID( rsp );
		} else {
			ras = as;
		}
	}
error:
	if (rsp)
		alp_ipc_message_destroy(rsp);
	if (msg)
		alp_ipc_message_destroy(msg);
	alp_ipc_channel_disconnect( con );
    return ras;
}

/*! @brief UnMount a device
 *  @param[in] udi - entry in fsinfo to umount
 *  @return - ALP_STATUS_OK, the referenced volume is (or was) unmounted
 */
alp_status_t alp_volumesvcs_umount( const gchar *udi )
{
	alp_status_t as=ALP_STATUS_OK;
	alp_status_t ras=ALP_STATUS_OK;
	AlpMessage *msg=NULL;
	AlpMessage *rsp=NULL;
	AlpConnection *con=NULL;	
	guint i,len;
	gint cmd;

	if (udi == NULL || (strlen(udi) == 0))
		return ALP_STATUS_VOLUMESVCS_INVALID_PARAMETER;
	as = alp_ipc_channel_connect_synch(ALP_VOLUMESVCS_SERVER_CHANNEL,ALP_IPC_CONNECTION_USAGE_HINT_NONE,0,&con);
	if (as != ALP_STATUS_OK || (msg=alp_ipc_message_create()) == NULL) {
		ras = as;
	} else {
		PACKSTART( as, msg, ALP_VOLUMESVCS_CMD_UMOUNT, error );
		PACKSTR( as, msg, udi, error );
		PACKEND( as, msg, error );
		as = alp_ipc_connection_send_with_response( con, msg, 0,&rsp);
		if (as == ALP_STATUS_OK) {
			ras = alp_ipc_message_get_message_ID( rsp );
		}
	}
error:
	if (rsp)
		alp_ipc_message_destroy(rsp);
	if (msg)
		alp_ipc_message_destroy(msg);
	alp_ipc_channel_disconnect( con );

    return ras;
}

/*! @brief Mount a dir as if it was a device
 *  @param[in] name - the mountpoint name under /media (will be created as usual)
 *  @param[in] path - the path to the dir to mount
 *
 *  @return - ALP_STATUS_OK mount succeeded (or is already mounted)
 */

alp_status_t alp_volumesvcs_bind( const gchar *name, const char *path )
{
	alp_status_t as=ALP_STATUS_OK;
	alp_status_t ras=ALP_STATUS_OK;
	AlpMessage *msg=NULL;
	AlpMessage *rsp=NULL;
	AlpConnection *con=NULL;	
	guint i,len;
	gint cmd;

	if (name == NULL || (strlen(name) == 0) || path == NULL ||
			!g_file_test( path, G_FILE_TEST_IS_DIR ))
		return ALP_STATUS_VOLUMESVCS_INVALID_PARAMETER;
	as = alp_ipc_channel_connect_synch(ALP_VOLUMESVCS_SERVER_CHANNEL,ALP_IPC_CONNECTION_USAGE_HINT_NONE,0,&con);
	if (as != ALP_STATUS_OK || (msg=alp_ipc_message_create()) == NULL) {
		ras = as;
	} else {
		PACKSTART( as, msg, ALP_VOLUMESVCS_CMD_BIND, error );
		PACKSTR( as, msg, name, error );
		PACKSTR( as, msg, path, error );
		PACKEND( as, msg, error );
		as = alp_ipc_connection_send_with_response( con, msg, 0,&rsp);
		if (as == ALP_STATUS_OK) {
			ras = alp_ipc_message_get_message_ID( rsp );
		}
	}
error:
	if (rsp)
		alp_ipc_message_destroy(rsp);
	if (msg)
		alp_ipc_message_destroy(msg);
	alp_ipc_channel_disconnect( con );

    return ras;
}

/*! @brief rename a volume
 *  @param[in] udi  - the volume to rename
 *  @param[in] name - what to rename it too
 *
 *  @return - ALP_STATUS_OK mount succeeded (or is already mounted)
 */

alp_status_t alp_volumesvcs_rename_volume( const gchar *udi, const char *name )
{
	alp_status_t as=ALP_STATUS_OK;
	alp_status_t ras=ALP_STATUS_OK;
	AlpMessage *msg=NULL;
	AlpMessage *rsp=NULL;
	AlpConnection *con=NULL;	

	if (name == NULL || udi == NULL || (strlen(udi) == 0) || (strlen(name) == 0))
		return ALP_STATUS_VOLUMESVCS_INVALID_PARAMETER;
	as = alp_ipc_channel_connect_synch(ALP_VOLUMESVCS_SERVER_CHANNEL,ALP_IPC_CONNECTION_USAGE_HINT_NONE,0,&con);
	if (as != ALP_STATUS_OK || (msg=alp_ipc_message_create()) == NULL) {
		ras = as;
	} else {
		PACKSTART( as, msg, ALP_VOLUMESVCS_CMD_RENAME, error );
		PACKSTR( as, msg, udi, error );
		PACKSTR( as, msg, name, error );
		PACKEND( as, msg, error );
		as = alp_ipc_connection_send_with_response( con, msg, 0,&rsp);
		if (as == ALP_STATUS_OK) {
			ras = alp_ipc_message_get_message_ID( rsp );
		}
	}
error:
	if (rsp)
		alp_ipc_message_destroy(rsp);
	if (msg)
		alp_ipc_message_destroy(msg);
	alp_ipc_channel_disconnect( con );

    return ras;
}

/*! @brief return a list of supported filesystem types as character strings, one of these
 *                 strings must be passed in to alp_volumesvcs_makefs
 *  @param[in] fstypes - address of a pointer to a GPtrArray
 *
 *  @return - ALP_STATUS_OK in all cases
 *  @remarks the returned array may be freed wtih g_ptr_array_free( fstypes, TRUE )
 */
extern alp_status_t alp_volumesvcs_fstypes( GPtrArray **fstypes)
{
	alp_status_t as=ALP_STATUS_OK;
	alp_status_t ras=ALP_STATUS_OK;
	AlpMessage *msg=NULL;
	AlpMessage *rsp=NULL;
	AlpConnection *con=NULL;	
	guint i,len;
	gint cmd;

	if (fstypes == NULL)
		return ALP_STATUS_VOLUMESVCS_INVALID_PARAMETER;
	as = alp_ipc_channel_connect_synch(ALP_VOLUMESVCS_SERVER_CHANNEL,ALP_IPC_CONNECTION_USAGE_HINT_NONE,0,&con);
	if (as != ALP_STATUS_OK || (msg=alp_ipc_message_create()) == NULL) {
		ras = as;
	} else {
		PACKSTART( as, msg, ALP_VOLUMESVCS_CMD_FSTYPES, error );
		PACKEND( as, msg, error );
		as = alp_ipc_connection_send_with_response( con, msg, 0, &rsp);
		if (as != ALP_STATUS_OK) {
			ras = as;
		} else {
			ras = alp_ipc_message_get_message_ID( rsp );
			if (ras == ALP_STATUS_OK ) {
				UNPACKSTART( as, rsp, error );
				UNPACKU32( as, rsp, &len, error );
				*fstypes = g_ptr_array_sized_new(len);
				for (i=0; i<len; i++) {
					gchar *p;
					UNPACKSTR( as, rsp, &p, error );
					g_ptr_array_add( *fstypes, p);
				}
			}
		}
	}
error:
	if (rsp)
		alp_ipc_message_destroy(rsp);
	if (msg)
		alp_ipc_message_destroy(msg);
	alp_ipc_channel_disconnect(con);

	return ras;
}

/*! @brief Initialzie a volume (create a 1 part partition table, and mkfs a FS in it).
 *
 *  @param[in] device_udi the hal UDI for the parent device (e.g. /dev/sdd) to format.
 *  @param[in] name the user visible name (label) this device should have, this will be
 *             the default mount point under /media
 *  @param[in] fstype (must be amonng the strings returned by alp_volumesvcs_fstypes, typically
 *             FAT16 or FAT32

 *  @return ALP_STATUS_OK if it was found and unmounted
 *  @remarks Note there is no option to add a partition or have multiple partions, you just
 *             get a single partition table and a newfs in that single partition.  Multiple
 *             partition devices are supported, but they must be formated elsewhere.
*/
alp_status_t alp_volumesvcs_makefs( const gchar *udi, const gchar *name, const gchar* fstype )
{
	alp_status_t as=ALP_STATUS_OK;
	alp_status_t ras=ALP_STATUS_OK;
	AlpMessage *msg=NULL;
	AlpMessage *rsp=NULL;
	AlpConnection *con=NULL;	
	guint i,len;
	gint cmd;

	if (udi == NULL || name == NULL || fstype == NULL)
		return ALP_STATUS_VOLUMESVCS_INVALID_PARAMETER;
	as = alp_ipc_channel_connect_synch(ALP_VOLUMESVCS_SERVER_CHANNEL,ALP_IPC_CONNECTION_USAGE_HINT_NONE,0,&con);
	if (as != ALP_STATUS_OK || (msg=alp_ipc_message_create()) == NULL) {
		ras = as;
	} else {
		PACKSTART( as, msg, ALP_VOLUMESVCS_CMD_MAKEFS, error );
		PACKSTR( as, msg, udi, error );
		PACKSTR( as, msg, name, error );
		PACKSTR( as, msg, fstype, error );
		PACKEND( as, msg, error );
		as = alp_ipc_connection_send_with_response( con, msg, 0,&rsp);
		if (as == ALP_STATUS_OK) {
			ras = alp_ipc_message_get_message_ID( rsp );
		} else {
			ras = as;
		}
	}
error:
	if (rsp)
		alp_ipc_message_destroy(rsp);
	if (msg)
		alp_ipc_message_destroy(msg);
	alp_ipc_channel_disconnect( con );
    return ras;
}
/*! @} */
