/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright � 2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/

#include <stdio.h>
#include <string.h>

#include <hiker/types.h>
#include <hiker/bundlemgr.h>
#include <hiker/global_settings.h>
#include <glib.h>

#include "demo_global.h"


char*	gFavoriteColorStr = NULL;
int		gFavoriteNum = 0;


alp_status_t read_globals(AlpGlobalSettingsContext* context);
alp_status_t write_globals(AlpGlobalSettingsContext* context);

/* initialize_globals will use the hiker global settings manager.
 * It will first try to read existing global values. If they have not
 * been stored, we will write them out based on a file we get the
 * bundle manager to find for us. */
void initialize_globals()
{
	AlpGlobalSettingsContext* context;
	alp_status_t status;

	/* Early exit if global settings just not working. */
	if (alp_global_settings_init() != ALP_STATUS_OK)
		return;
	if (alp_global_settings_open(&context) != ALP_STATUS_OK)
		return;

	status = read_globals(context);
	
	/* If we couldn't read globals, this may be the first time we
	 * have launched. So create default globals. */
	if (status != ALP_STATUS_OK)
	{
		status = write_globals(context);

		/* If write worked, refresh our app globals. */
		if (status != ALP_STATUS_OK)
			read_globals(context);
	}

	/* Only come here after settings opened. */
	alp_global_settings_close(context);
}


/* read_globals will also set them for the application. */
alp_status_t read_globals(AlpGlobalSettingsContext* context)
{
	/* Init, open, and close of global setting manager performed by calling function. */
	char tempBuf[100];
	alp_status_t status;

	/* Check for an existing favorite color. For now, assume
	 * if it is there that the favorite number is set too. */
	status = alp_global_settings_get_key_string_value(context, FAV_COLOR_KEY, 
													tempBuf, sizeof(tempBuf));
	if (status == ALP_STATUS_OK)
	{
		/* Make a persistent copy for the rest of the app life. But to cut
		 * down on memory thrashing, only do so for new or changed value. */
		if (gFavoriteColorStr == NULL ||
			strcmp(gFavoriteColorStr, tempBuf))
		{
			/* Prevent memory leak if we had previous color. */
			if (gFavoriteColorStr != NULL)
				g_free(gFavoriteColorStr);
				
			gFavoriteColorStr = g_malloc(strlen(tempBuf) + 1);
			if (gFavoriteColorStr != NULL) 
				strcpy(gFavoriteColorStr, tempBuf);
		}

		/* Only test for favorite number if we already have favorite color. */
		status = alp_global_settings_get_key_int_value(context, FAV_NUM_KEY, 
														&gFavoriteNum);
	}
	
	return status;
}


/* write_globals will get defaults from /data/defaults.xml. */
alp_status_t write_globals(AlpGlobalSettingsContext* context)
{
	/* Init, open, and close of global setting manager performed by calling function. */
	char *defaultsFilename = "/data/default.xml";
	char *defaultsPath;
	alp_status_t status;

	/* We need a transaction if we are going to do writing. */
	status = alp_global_settings_begin_transaction(context);
	if (status == ALP_STATUS_OK)
	{
		/* Our default.xml specifies access permissions, so no need for default mode. */
//		if (alp_global_settings_set_default_mode(context, S_IRUSR | S_IWUSR) != ALP_STATUS_OK)
//			goto Error;

		/* Bundle mgr finds path to our defaults. */
		defaultsPath = alp_bundle_ref_ro_pathname(alp_bundle_ref_me(), defaultsFilename);
		if (defaultsPath != NULL)
		{
			/* Actually write out preferences. */
			status = alp_global_settings_import_from_file(context, defaultsPath);
			g_free(defaultsPath);
		}
		else
			/* or set status to show mem error. */
			status = ALP_STATUS_GLOBAL_SETTINGS_BUFFER_TOO_SMALL;

		if (status == ALP_STATUS_OK)
			/* Successful completion of writing new globals. */
			alp_global_settings_complete_transaction(context);
		else
			alp_global_settings_cancel_transaction(context);
	}

	return status;
}


void finalize_globals()
{	
	if (gFavoriteColorStr != NULL)
		g_free(gFavoriteColorStr);
}


int alp_main(int argc, char** argv)
{
	initialize_globals();
	
	printf("My favorite color is %s and my favorite number is %d.\n", 
			gFavoriteColorStr, gFavoriteNum);
			
	finalize_globals();
	
	return 0;
}

