/******************************************************************************
 *
 * Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the entire contents of this file.
 *
 * The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.
 *
 * Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
 * Rights Reserved.
 * 
 * Contributor(s): none.
 *
 *****************************************************************************/
/**
 * @file alarm_server_database.c
 *
 * @brief Alarm Manager server database implementation
 *
 *****************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>

#include <sqlite3.h>

#include <glib.h>
#include <glib/gstdio.h>

#include <hiker/syspaths.h>
#include <hiker/appmgr.h>
#include <hiker/pdk/appmgr_pdk.h>

#include <hiker/alarmmgr.h>
#include <hiker/prv/alarmmgr_prv.h>

#include "alarmmgr_server_prv.h"

#define ALP_ALARM_DATABASE_DIR   ALP_SYSPATH_SYSTEM_CACHE "alarmmgr"
#define ALP_ALARM_DATABASE_FILE  "alarm_mgr.db"
#define ALP_ALARM_DATABASE_NAME  ALP_ALARM_DATABASE_DIR "/" ALP_ALARM_DATABASE_FILE
#define ALP_ALARM_DATABASE_MODE  (S_ISGID|S_ISVTX|S_IRWXU|S_IRWXG)

static int count_callback(void *some_entry, int ncol, char **cols, char **azColName);
int alp_prv_alarm_database_remove_triggered_alarms(void);
static int alp_prv_alarm_get_first_alarm_callback(void *some_entry, int ncol, char **cols, char **azColName);
int alp_prv_alarm_get_first_alarm(AlpAlmEntryType *entry);
int alp_prv_alarm_database_get_triggered_alarms(void);
static int alp_prv_alarm_sqlite_get_callback(void *some_entry, int ncol, char **cols, char **azColName);
static int alp_prv_alarm_sqlite_get(const gchar *appID, guint32 ref, AlpAlmEntryType *entry);
static int alp_prv_alarm_sqlite_update(const gchar *appID, guint32 ref, guint32 alarmSeconds);
static int alp_prv_alarm_sqlite_delete(const gchar *appID, guint32 ref);


int alp_alarm_database_init(void)
{
	sqlite3 *db;
	int rc;
	char *errmsg = NULL;
	char *dbname = ALP_ALARM_DATABASE_NAME;
    char *directory = ALP_ALARM_DATABASE_DIR;
	
    /* make sure database directory exists */
    if (!g_file_test (directory, G_FILE_TEST_IS_DIR)) {
        ALP_ATL("directory %s does not exit or is not directory",
                directory);

        /* check if exists and if so, attempt to remove non-directory */
        if (g_file_test (directory, G_FILE_TEST_EXISTS)) {
            if (g_remove (directory) < 0) {
                ALP_ATL ("unable to remove %s: %d %s",
                         directory, errno, strerror(errno));
                return -1;
            }
        }

        /* create directory */
        if (g_mkdir_with_parents (directory, ALP_ALARM_DATABASE_MODE) < 0) {
            /* unable to create directory */
            ALP_ATL ("unable to create %s", directory);
            return -1;
        }
    }

	rc = sqlite3_open(dbname,&db);
	if (rc != SQLITE_OK) {
		ALP_ATL ("sqlite3_open err %d", rc);
		goto Done;
	}
	//Note :Should not use the abbreviation "INT" 
	//when working with the PRIMARY KEY. Must use "INTEGER" 
	//for the primary key to update.
	rc = sqlite3_exec(
        db,
        "BEGIN;\n"
        "CREATE TABLE alarm ("
        "AlarmLUID INTEGER PRIMARY KEY, "
        "appID TEXT, "
        "ref INT, "
        "alarmSeconds INT, "
        "triggered INT, "
        "notified INT);\n"
        "CREATE TABLE VERSION (Version integer primary key default 1);"
        "INSERT INTO VERSION VALUES(1);\n"
        "COMMIT;\n",
        0, 0, &errmsg);
	
	if (rc != SQLITE_OK) {
		ALP_ATL ("sqlite3_exec err %d", rc);
		goto Done;
	}	

Done:
    if (errmsg) {
        ALP_ATL ("sqlite3_exec error mesg: %s", errmsg);
        sqlite3_free(errmsg);
    }
	sqlite3_close(db);

	return rc == SQLITE_OK ? ALP_STATUS_OK : rc;
}

// set an alarm
int alp_alarm_database_set(AlpAlmEntryType *entry)
{
	ALP_ATL ("app %s ref %u seconds %d",
             entry ? entry->appID : "NULL",
             entry ? entry->ref : 0,
             entry ? entry->alarmSeconds : 0);

    if (entry == NULL) {
        ALP_ATL ("passed NULL event");
        return ALP_STATUS_ALARM_BAD_PARAM;
    }

    if ((entry->alarmSeconds + 1) < (guint) time(NULL)) { /* 1 sec grace */
        return ALP_STATUS_ALARM_BAD_PARAM;
    }

	// add or update the alarm to the database
	return alp_prv_alarm_sqlite_update(entry->appID, entry->ref, entry->alarmSeconds);
}

// cancel an alarm
int alp_alarm_database_cancel(AlpAlmEntryType *entry)
{
	ALP_ATL ("app %s ref %u", entry->appID, entry->ref);

	// cancel the alarm from the database
	return alp_prv_alarm_sqlite_delete (entry->appID, entry->ref);
}

// get an alarm
int alp_alarm_database_get(AlpAlmEntryType *entry)
{
    int rc;

    if (entry == NULL) {
        ALP_ATL ("NULL alarm entry");
        return 0;
    }

	ALP_ATL ("%s %d", entry->appID, entry->ref);

	// get the alarm from the database
    rc = alp_prv_alarm_sqlite_get(entry->appID, entry->ref, entry);

    return rc;
}

static int alp_prv_alarm_get_first_alarm_callback(void *some_entry, int ncol, char **cols, char **azColName)
{
	AlpAlmEntryType *entry;
    int i;

	entry = (AlpAlmEntryType *)(some_entry);

	ALP_ATL ("first timeout: entry %x", entry);
    if (entry == NULL) return 0;

    for (i = 0; i < ncol; i++) {
        ALP_ATL ("    col[%d] (%s) = %s", i, azColName[i], cols[i]);
    }

	//entry->cmd = 0;
	entry->size = sizeof(*entry);
	entry->ref = (guint32) atol(cols[2]);
	entry->alarmSeconds = (guint32) atol(cols[3]);
	strcpy(entry->appID, cols[1]);
	//entry->triggered = 0;
	//entry->notified = 0;

	ALP_ATL ("entry->appID = %s", entry->appID);
	ALP_ATL ("entry->ref = %u", entry->ref);
	ALP_ATL ("entry->alarmSeconds = %u", entry->alarmSeconds);

    return 0;
}

int alp_prv_alarm_get_first_alarm (AlpAlmEntryType *entry)
{
	sqlite3 *db;
	int rc;
	char *errmsg = NULL;
	char *str    = NULL;
	time_t now;
	char *dbname = ALP_ALARM_DATABASE_NAME;

	rc = sqlite3_open(dbname,&db);
	if (rc != SQLITE_OK) {
		ALP_ATL ("sqlite3_open err %d", rc);
		goto Done;
	}
	now = time(NULL);
	str = sqlite3_mprintf("SELECT * FROM alarm "
                          "WHERE alarmSeconds > '%ld' "
                          "ORDER BY alarmSeconds, AlarmLUID LIMIT 1",
                          (gulong)now);

    ALP_ATL ("get first sql = <%s>", str);
	
	rc = sqlite3_exec(db, str, alp_prv_alarm_get_first_alarm_callback, entry, &errmsg);
	if (rc != SQLITE_OK) {
		ALP_ATL ("sqlite3_exec err %d", rc);
        goto Done;
	}

Done:
    if (errmsg) {
        ALP_ATL ("sqlite3_exe erro mesg: %s", errmsg);
        sqlite3_free(errmsg);
    }
    if (str)
        sqlite3_free (str);
	sqlite3_close(db);

	return rc == SQLITE_OK ? ALP_STATUS_OK : rc;
}

//remove alarms in the past
int alp_prv_alarm_database_remove_triggered_alarms(void)
{
	sqlite3 *db;
	int rc;
	char *errmsg = NULL;
	char *str    = NULL;
	time_t now;
	char *dbname = ALP_ALARM_DATABASE_NAME;

	ALP_ATL ("removing alarms");

	rc = sqlite3_open(dbname,&db);
	if (rc != SQLITE_OK) {
		ALP_ATL ("sqlite3_open err %d", rc);
		goto Done;
	}
	now = time(NULL);
	str = sqlite3_mprintf("DELETE FROM alarm "
                          "WHERE alarmSeconds <= '%ld'",
                          (gulong)now);
    ALP_ATL ("remove SQL = <%s>", str);
	
	rc = sqlite3_exec(db, str, 0, 0, &errmsg);
	if (rc != SQLITE_OK) {
		ALP_ATL ("sqlite3_exec err %d", rc);
        goto Done;
	}

Done:
    if (errmsg) {
        ALP_ATL ("sqlite3_exec error mesg %s", errmsg);
        sqlite3_free(errmsg);
    }
    if (str)
        sqlite3_free (str);
	sqlite3_close(db);

	return rc == SQLITE_OK ? ALP_STATUS_OK : rc;
}

//get alarms in the past
int alp_prv_alarm_database_get_triggered_alarms(void)
{
	sqlite3 *db;
    sqlite3_stmt *stmt = NULL;
	int rc;
	char *errmsg = NULL;
	char *str    = NULL;
	time_t now;
	char *dbname = ALP_ALARM_DATABASE_NAME;

	rc = sqlite3_open(dbname, &db);
	if (rc != SQLITE_OK) {
		ALP_ATL ("sqlite3_open err %d", rc);
		goto Done;
	}
	now = time(NULL);
	str = sqlite3_mprintf("SELECT appID, ref FROM alarm "
                          "WHERE alarmSeconds <= '%ld' "
                          "ORDER BY alarmSeconds, AlarmLUID",
                          (gulong)now);

    /* prepare statement with given SQL */
    rc = sqlite3_prepare (db, str, -1, &stmt, NULL);
    if (rc != SQLITE_OK) {
        ALP_ATL("sqlite3_prepare err %d", rc);
        goto Done;
    }

    if (stmt) {
        char           buf[50];
        AlpAppContext *context = NULL;

        /* setup launch context */
        sprintf(buf,"com.access.services.alarm.pid=%d",getpid());
        context = alp_app_setup(buf, ALP_APP_LAUNCH_FLAG_NONE, NULL);

        /* start processing statement */
        rc = sqlite3_step (stmt);
        while (rc == SQLITE_ROW) {
            char    *app_id;
            char     ref_buf[12];
            guint32  ref;
            char    *argv[3];

            app_id = (char *)  sqlite3_column_text (stmt, 0); // appID
            ref    = (guint32) sqlite3_column_int  (stmt, 1); // ref
            snprintf (ref_buf, sizeof(ref_buf), "%u", ref);

            /* prepare the launch arguments */
            argv[0] = ALP_APP_ALARM;
            argv[1] = ref_buf;
            argv[2] = NULL;     // for safety, don't include in arg count
            ALP_ATL ("****** rc = %d app = %s args=<%s,%s> ref=%u,%s ******",
                     rc, app_id, argv[0], argv[1], ref, ref_buf);

            /* launch the app for the triggered alarm */
            alp_app_launch_async(app_id, 2, argv);

            /* step to next triggered row */
            rc = sqlite3_step (stmt);
        }
        ALP_ATL("end of loop rc = %d", rc);

        if (rc == SQLITE_ERROR) {
            rc = sqlite3_reset (stmt);
            ALP_ATL("Goofy rc = %d", rc);
        }

        /* tear down launch context */
        if (context) alp_app_destroy(context,0);
        
        /* finish statement and free it */
        sqlite3_finalize (stmt);
    }

Done:
    if (errmsg) {
        ALP_ATL ("sqlite3_exec: %s", errmsg);
        sqlite3_free(errmsg);
    }
    if (str)
        sqlite3_free (str);
	sqlite3_close(db);

	return rc == SQLITE_OK ? ALP_STATUS_OK : rc;
}

//callback to handle each row from "SELECT * FROM alarm WHERE (appID == '%q' AND ref == %u')", appID, ref"
//where each row is launched through the app manager
static int alp_prv_alarm_sqlite_get_callback(void *some_entry, int ncol, char **cols, char **azColName)
{
	AlpAlmEntryType *entry;


	entry = (AlpAlmEntryType *)(some_entry);

    if (entry == NULL) {
        ALP_ATL ("passed invalid NULL entry");
        return 0;
    }
	//entry->cmd = 0;
	entry->size         = sizeof(*entry);
	entry->ref          = (guint32) atol(cols[2]);
	entry->alarmSeconds = (guint32) atol(cols[3]);
	strcpy(entry->appID, cols[1]);
	//entry->triggered = 0;
	//entry->notified = 0;
	ALP_ATL ("entry->appID = %s", entry->appID);
	ALP_ATL ("entry->ref = %u", entry->ref);
	ALP_ATL ("entry->alarmSeconds = %d", entry->alarmSeconds);

    return 0;
}

static int alp_prv_alarm_sqlite_get(const gchar *appID, guint32 ref, AlpAlmEntryType *entry)
{
	sqlite3 *db;
	int rc;
	char *errmsg = NULL;
	char *str    = NULL;
	guint counter = 0;
	char *dbname = ALP_ALARM_DATABASE_NAME;

    ALP_ATL ("app %s ref %d", appID, ref);
	
	rc = sqlite3_open(dbname,&db);
	if (rc != SQLITE_OK) {
		ALP_ATL ("sqlite3_open err %d", rc);
		goto Done;
	}

    /* make ref query signed since sqlite handles INT only as signed */
	str = sqlite3_mprintf("SELECT * FROM alarm "
                          "WHERE (appID == '%q' AND ref == '%d')",
                          appID, ref);
	rc = sqlite3_exec(db, str, count_callback, &counter, &errmsg);
	if (rc != SQLITE_OK) {
		ALP_ATL ("sqlite3_exec select err %d", rc);
		goto Done;
	}

	if (counter == 0) {
		entry->ref = ref;
		entry->alarmSeconds = 0;
		strcpy(entry->appID, appID);
        ALP_ATL("alarm entry not found ? ");
	}
	else {
		rc = sqlite3_exec(db, str, alp_prv_alarm_sqlite_get_callback, entry, &errmsg);
		if (rc != SQLITE_OK) {
			ALP_ATL ("sqlite3_exec err %d", rc);
			goto Done;
		}
	}

Done:
	if (errmsg) {
		ALP_ATL ("sqlite3_exec error mesg: %s", errmsg);
		sqlite3_free(errmsg);
	}
    if (str)
        sqlite3_free (str);
	sqlite3_close(db);

	return rc == SQLITE_OK ? ALP_STATUS_OK : rc;
}

static int count_callback(void *some_entry, int ncol, char **cols, char **azColName)
{
	guint *count;

	count = (guint *)(some_entry);

	*count += 1;

      	return 0;
}

static int alp_prv_alarm_sqlite_update(const gchar *appID, guint32 ref, guint32 alarmSeconds)
{
	sqlite3 *db;
	int rc;
	char *errmsg = NULL;
	char *str    = NULL;
	guint counter = 0;
	char *dbname = ALP_ALARM_DATABASE_NAME;

	ALP_ATL ("%s %d to %d seconds", appID, ref, alarmSeconds);	

	rc = sqlite3_open(dbname, &db);
	if (rc != SQLITE_OK) {
		ALP_ATL ("sqlite3_open err %d", rc);
		goto Done;
	}

	/*
	 * check to see if an alarm with appID and ref already exists
	 * if so, update it
	 * if not, create it
	 */
	str = sqlite3_mprintf("SELECT * FROM alarm "
                          "WHERE (appID == '%q' AND ref == '%d')",
                          appID, ref);
    ALP_ATL ("count sql = <%s>", str);
	rc = sqlite3_exec(db, str, count_callback, &counter, &errmsg);
	if (rc != SQLITE_OK) {
		ALP_ATL ("sqlite3_exec select err %d", rc);
		goto Done;
	}

    if (str) {
        sqlite3_free (str);
        str = NULL;
    }

	if (counter == 0) {
		str = sqlite3_mprintf("INSERT INTO alarm (appID, ref, alarmSeconds) "
                              "VALUES ('%q', %d, %d)",
                              appID, ref, alarmSeconds);
        ALP_ATL ("insert sql = <%s>", str);
	} else {
		str = sqlite3_mprintf("UPDATE alarm SET alarmSeconds  = '%d' "
                              "WHERE (appID == '%q' AND ref == '%d')",
                              alarmSeconds, appID, ref);
        ALP_ATL ("update sql = <%s>", str);
	}

	rc = sqlite3_exec(db, str, 0, 0, &errmsg);
	if (rc != SQLITE_OK) {
        ALP_ATL ("sqlite3_exec err %d", rc);
		goto Done;
	}

Done:
	if (errmsg) {
		ALP_ATL ("sqlite3_exec error mesg: %s", errmsg);
		sqlite3_free(errmsg);
	}
    if (str)
        sqlite3_free (str);
	sqlite3_close(db);

	return rc == SQLITE_OK ? ALP_STATUS_OK : rc;
}

static int alp_prv_alarm_sqlite_delete(const gchar *appID, guint32 ref)
{
	sqlite3 *db;
	int rc;
	char *errmsg = NULL;
	char *str    = NULL;
	guint counter = 0;
	char *dbname = ALP_ALARM_DATABASE_NAME;

	ALP_ATL ("%s %ds", appID, ref);	

	rc = sqlite3_open(dbname,&db);
	if (rc != SQLITE_OK) {
		ALP_ATL ("sqlite3_open err %d", rc);
		goto Done;
	}

	/*
	 * check to see if an alarm with appID and ref already exists
	 * if so, delete it
	 */
	str = sqlite3_mprintf("SELECT * FROM alarm "
                          "WHERE (appID == '%q' AND ref == '%d')",
                          appID, ref);
    ALP_ATL ("count sql = <%s>", str);
	rc = sqlite3_exec(db, str, count_callback, &counter, &errmsg);
	if (rc != SQLITE_OK) {
		ALP_ATL ("sqlite3_exec select err %d", rc);
		goto Done;
	}

    if (str) {
        sqlite3_free (str);
        str = NULL;
    }

	if (counter) {
		str = sqlite3_mprintf("DELETE FROM alarm "
                              "WHERE appID = '%q' and ref = %d",
                              appID, ref);
        ALP_ATL ("delete sql = <%s>", str);
        rc = sqlite3_exec(db, str, 0, 0,&errmsg);
        if (rc != SQLITE_OK) {
            ALP_ATL ("sqlite3_exec err %d", rc);
            goto Done;
        }
	} else {
        ALP_ATL("delete failed - alarm record not found");
    }

Done:
	if (errmsg) {
		ALP_ATL ("sqlite3_exec error mesg: %s", errmsg);
		sqlite3_free(errmsg);
	}
    if (str)
        sqlite3_free(str);
	sqlite3_close(db);

	return rc == SQLITE_OK ? ALP_STATUS_OK : rc;
}
