/******************************************************************************
 *
 * Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the entire contents of this file.
 *
 * The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.
 *
 * Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
 * Rights Reserved.
 *
 * Contributor(s): none.
 * 
 *f***************************************************************************/
/**
 * @file alarm_server_main.c
 *
 * @brief Alarm Manager server implementation
 *
 *****************************************************************************/
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>

#include <glib.h>

#include <hiker/datetime.h>
#include <hiker/ipc.h>
#include <hiker/notifymgr.h>
#include <alp/power.h>
#include <hiker/pdk/sysutils_pdk.h>

#include <hiker/alarmmgr.h>
#include <hiker/prv/alarmmgr_prv.h>

#include "alarmmgr_server_prv.h"

#define BANNER "AlarmManager : "

static guint TimeoutID = 0;
static GMainLoop *loop;

static void alp_prv_alarm_server_connection(AlpConnection *connection, gpointer data);
static void alp_prv_alarm_server_disconnection(AlpConnection *connection, gpointer data);
static void alp_prv_alarm_server_receive(AlpConnection *connection, AlpMessage *msg, gpointer data);
static gboolean alp_prv_alarm_daemon_timeout_handler (gpointer iData);

static gboolean do_detach = TRUE;

static void alp_prv_alarm_set_next_alarm_timeout(void)
{
	time_t start, now;
	AlpAlmEntryType	entry;
	alp_status_t err;

	// delete current g_timeout
	if (TimeoutID) {
        ALP_ATL ("remove source for timeout %d", TimeoutID);
        g_source_remove(TimeoutID);
    }

	//launch app through app-serverfor the alarms in the past
	alp_prv_alarm_database_get_triggered_alarms();

	//remove the alarms in the past	
	alp_prv_alarm_database_remove_triggered_alarms();

	memset(&entry,0,sizeof(AlpAlmEntryType));
	entry.size = sizeof(AlpAlmEntryType);

	// Find first alarm in the future
	err = alp_prv_alarm_get_first_alarm(&entry);
	
	if (entry.alarmSeconds)
	{
        guint id;

        ALP_ATL ("scheduling power wakeup for %*s",
                 24, ctime((time_t *)&entry.alarmSeconds));
        err = alp_power_request_wakeup ((time_t)entry.alarmSeconds,
                                        ALP_POWER_FULL_WAKE_STATE,
                                        NULL,
                                        NULL,
                                        &id);
        if (err != ALP_STATUS_OK) {
            if (err == ALP_STATUS_POWER_TIME_PASSED) {
                /*
                 * The alarm time is now so make sure the device is
                 * fully awake before we fire it
                 */
                ALP_ATL ("already up - forcing full wake");
                alp_power_force_full_wake();
            } else {
                /* something bad happened... */
                ALP_ATL ("unhandled err from power wakeup request %x", err);
            }
        } else {
            ALP_ATL ("power wakeup request scheduled - err %x id %x",
                     err, id);
        }

		// Set new g_timeout	
		now = time (NULL);
		start = entry.alarmSeconds - now;
		TimeoutID = g_timeout_add(start*1000,
                                  alp_prv_alarm_daemon_timeout_handler,
                                  NULL);
		
        ALP_ATL ("timeout %d added at %d for %s ref %d alarm %d (in %d seconds)",
                 TimeoutID,
                 now,
                 entry.appID,
                 entry.ref,
                 entry.alarmSeconds,
                 start);
	}
	else
	{
		ALP_ATL ("no future alarm %s %d %d\n",
                 entry.appID,
                 entry.ref,
                 entry.alarmSeconds);
	}	
}

static int
alp_prv_alarm_daemon_datetime_change_handler (
    const char *notify_type,
    void       *details,
    int         length,
    int        *handled,
    const char *app_id,
    void       *user_data)
{
    time_t  current;

    current = time(NULL);
    ALP_ATL ("Alarm notified of date time change %s from %s at %*s (%ld seconds)",
             notify_type, app_id, 24, ctime (&current), (gulong)current );

    alp_prv_alarm_set_next_alarm_timeout();

    return 0;
}

static gboolean alp_prv_alarm_daemon_timeout_handler (gpointer iData)
{	
	(void)iData;
 	time_t current;
 	current = time(NULL);
	
	ALP_ATL ("Alarm expired at %*s (%ld seconds)",
             24, ctime (&current), (gulong)current);

	alp_prv_alarm_set_next_alarm_timeout();	

 	// return FALSE to have the source removed.
	return FALSE;
		
}

static void
alp_prv_alarm_server_connection(AlpConnection *connection, gpointer data)
{
	alp_status_t err;
	ALP_ATL ("connection %x data %x", connection, data);

	err = alp_ipc_connection_register_receive_callback(connection, alp_prv_alarm_server_receive, connection);
}

static void
alp_prv_alarm_server_disconnection(AlpConnection *connection, gpointer data)
{
	ALP_ATL ("connection %x data %x",  connection, data);
}



static void
alp_prv_alarm_server_receive(AlpConnection *connection, AlpMessage *msg, gpointer data)
{
	AlpAlmEntryType *entry;
	guint len;
	guint status;
	int  cmd;
	alp_status_t err;

	ALP_ATL ("connection %x msg %x data %x", connection, msg, data);

	// unpack the message from the client
	alp_ipc_message_unpack_start(msg);
	alp_ipc_message_unpack_byte_array(msg, (guchar**)&entry, &len);
	alp_ipc_message_unpack_end(msg);

	// process the message
	cmd = entry->cmd;
	switch(cmd) {
    case ALP_ALARM_CMD_SET:
        status = alp_alarm_database_set(entry);
        if (status == ALP_STATUS_OK){
            ALP_ATL ("timeout %s %u %d stored @ %d, now set it",
                     entry->appID,
                     entry->ref,
                     entry->alarmSeconds,
                     time(NULL));
            alp_prv_alarm_set_next_alarm_timeout();
        }
		break;

    case ALP_ALARM_CMD_CANCEL:
        status = alp_alarm_database_cancel(entry);
        if (status == ALP_STATUS_OK){
            ALP_ATL ("deleted timeout %s %u %d",
                     entry->appID,
                     entry->ref,
                     entry->alarmSeconds);
            alp_prv_alarm_set_next_alarm_timeout();
        }
		break;

    case ALP_ALARM_CMD_GET:
        status = alp_alarm_database_get(entry);
        ALP_ATL ("get timeout %s %u = %d",
                 entry->appID,
                 entry->ref,
                 entry->alarmSeconds);
		break;

    default:
        ALP_ATL ("unknown command: %d", cmd);
        goto cleanup;
        break;
	}

	/* create and pack a response message */
	AlpMessage *reply = NULL;

	reply = alp_ipc_message_create();
	if (reply) {
		alp_ipc_message_response_pack_start (reply, 0, msg);
		alp_ipc_message_pack_uint32 (reply, status);

		if (cmd == ALP_ALARM_CMD_GET) {
			alp_ipc_message_pack_byte_array(reply, (guchar*)entry, len);
		}
		alp_ipc_message_pack_end(reply);

		// send the status back to the client
		err = alp_ipc_connection_send(connection, reply);
	} else {
		ALP_ATL ("no memory on reply response create");
	}

	if (err != ALP_STATUS_OK) {
		ALP_ATL("alp_ipc_connection_send err=%d\n", err);
	}
	
cleanup:
	// deallocate the message
	alp_ipc_message_destroy(msg);
	if (reply) alp_ipc_message_destroy(reply);

	g_free(entry);
}


static
int
alp_alarm_server_main(int argc, char *argv[])
{
	alp_status_t err;

	AlpChannel    *alp_channel;
	
	ALP_ATL (BANNER "starting...");

	// Create a child daemon process for alarmmgr initialization
    if (do_detach)
        alp_daemon_start(FALSE, TRUE);
	
	loop = g_main_loop_new(NULL, FALSE);

	/* initialize database here */
	alp_alarm_database_init();

	/* create a channel for the client library to reach us */
    ALP_ATL ("opening ipc channel %s", ALP_ALARM_MGR_SERVER_CHANNEL);
	alp_channel = alp_ipc_channel_create(ALP_ALARM_MGR_SERVER_CHANNEL,
                                         ALP_IPC_CHANNEL_ACCESS_ALL);
	if (!alp_channel) {
		ALP_ATL (BANNER "failed to create channel %s",
                 ALP_ALARM_MGR_SERVER_CHANNEL);
		goto exit;
	}

	err = alp_ipc_channel_register_connect_callback(alp_channel, alp_prv_alarm_server_connection, NULL);
	err = alp_ipc_channel_register_disconnect_callback(alp_channel, alp_prv_alarm_server_disconnection, NULL);

    alp_notify_register_server (ALP_NOTIFY_EVENT_DATETIME_CHANGE,
                                alp_prv_alarm_daemon_datetime_change_handler,
                                NULL,
                                0, 0);
	alp_prv_alarm_set_next_alarm_timeout();

	// Now we're all ready, allow boot sequence to continue.
    if (do_detach)
        alp_daemon_release(0);
	
	
	g_main_loop_run(loop);

	g_main_loop_unref(loop);

	/* get rid of the channel */
	alp_ipc_channel_destroy(alp_channel);


exit:
	
	ALP_ATL (BANNER "bailing out");
	
	return 0;
}


int
main(int argc, char *argv[])
{
	return alp_alarm_server_main(argc, argv);
}


/* vim: set ts=4 number sw=4 ai cin sm :miv */
