/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/*! @file appserver.c
	This file contains the MAX Application Server implementation.
 */


// Uncomment the next line to disable traces AT COMPILE TIME 
#define ALP_TRACE ALP_TRACE_OFF		//Enable/disable


// Source for the MAX Application Server

#include <errno.h>
#include <time.h>
#include <signal.h>
#include <string.h>
#include <limits.h> // for PATH_MAX
#include <sys/types.h>
#include <sys/wait.h>
#include <hiker/pdk/appmgr_pdk.h>
#include <hiker/prv/appmgr_prv.h>
#include <hiker/pdk/sysutils_pdk.h>
#include <hiker/ipc.h>
#include <hiker/traces.h>
#include <hiker/bundlemgr.h>
#include <hiker/prv/bundlemgr_prv.h>
#include <hiker/prv/appserver_lock.h>
#include <hiker/fail.h>
#include <hiker/config.h>
#include <hiker/prv/nlpd_bits.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/XF86keysym.h>
#include <X11/keysym.h>



#define APPSERVERLOCK "/var/lock/.alp_appserverd_lock"

/*
	More things we could test write unit tests for:
	More detailed use cases for adding & removing exit handlers.
	launch & exit
	exit w/o handler
	exit when ignoring SIGTERM
	relaunch handler
	relaunch w/o handler
	memory leaks
*/




// It's kind of important that the kill timer be shorter than the launch timeout.
// Otherwise, when we relaunch a hung app, the launch could fail (and be discarded)
// before the app is killed & restarted.
#define PRV_KILL_TIMER_SECS		3	// kill timer duration in seconds.
#define PRV_LAUNCH_TIMEOUT_SECS	5	// timeout duration in seconds before an initial
 								// registration must be received after a launch
 								// is dispatched to a launchpad.

// Constants for the 'killTimers' array.
#define PRV_EXIT_TIMER		0
#define PRV_RELAUNCH_TIMER	1
#define PRV_NUM_KILL_TIMERS	2	// one for relaunch, one for exit.

// Constants for jump list:
#define PRV_JUMPLIST_NUM_RECENT	10


struct RunningAppInfo;
typedef struct RunningAppInfo RunningAppInfo;

struct PendingLaunch;
typedef struct PendingLaunch PendingLaunch;

struct AppWatcher;
typedef struct AppWatcher AppWatcher;

// Structures
struct RunningAppInfo
{
	AlpConnection	*connection;	
	char			appID[PKG_ID_MAX_LEN];
	pid_t		appPID;
	guint32		launchFlags;		// flags that indicate the running state of the app.  
							// These are only ever changed by the app calling alp_app_register(), 
							// to ensure their correctness.
	guint32		serverFlags;// flags for keeping track of internal server state.
	
	AlpAppBackgroundingType backgrounding;
	char			waitingFor[PKG_ID_MAX_LEN]; // App ID we are waiting for, used when ALP_APP_SERVER_FLAG_WAITING_FOR_REPLY is set.
	
	char		launchpadScheme[ALP_APP_LAUNCHPAD_SCHEME_SIZE];
	AlpMessage *replyMessage; // Contains the message to reply to, if ALP_APP_SERVER_FLAG_WAITING_FOR_REPLY is set.
	
	guint	killTimer[PRV_NUM_KILL_TIMERS]; // If non-zero, the event source ID 
						// of the kill timer that will terminate the app due 
	 					// to not exiting or re-registering after a relaunch.
	
};

struct PendingLaunch
{
	int		argc;	// Set for pending launch requests which have not registered yet.
	char**	argv;	// These are used to re-launch the app in the event that it is
	 					// already running and has not registered a relaunch handler
	 					// (because it will quit in response to the relaunch request).
	
	guint	launchTimer;
	
	RunningAppInfo *requester; // the app that requested the launch.
};


struct AppWatcher
{
	char			appID[PKG_ID_MAX_LEN];
	gboolean		wasRun;
	gboolean		blockLaunch;
	AlpWatcherRef	ref;
	
	RunningAppInfo *owner; // the app that requested the watcher.
};




// prototypes:
void prv_check_default_app(void);
void prv_free_running_app_info(RunningAppInfo *app);
void prv_client_connect_cb(AlpConnection* connection, gpointer cbData);
void prv_client_disconnect_cb(AlpConnection* connection, gpointer cbData);
void prv_receive_data_cb(AlpConnection* connection, AlpMessage* message, gpointer cbData);
void prv_handle_launch(RunningAppInfo *requesterInfo, AlpMessage *msg);
RunningAppInfo* prv_find_app_by_id(char *appID);
//gboolean PrvSearchArgs(char *searchParam, int argc, char* argv[]);
void prv_handle_exit_request(AlpMessage *exitMsg);
void prv_send_launch_reply(RunningAppInfo *dest, char *appID, alp_status_t launchErr, pid_t appPID);
PendingLaunch* prv_find_pending_launch(char *appID);
alp_status_t prv_send_to_launchpad(char *appID, AlpMessage *msg);
void prv_do_next_pending_launch(char *appID);
void prv_remove_pending_launch(PendingLaunch *pendLaunch, alp_status_t err, pid_t appPID);
AppWatcher* prv_find_watcher(AlpWatcherRef ref);



// Globals
static GList *gAppList = NULL;			// list of currently running applications.

static char *gDefaultPrimary = NULL; // If non-NULL, this default app is run when there is no primary app.

// list of pending launch requests.
// When the server receives a valid launch request, it is stored in this list until it is 
// acknowledged by the app (by sending us a registration for it).  The list order is important.
// The first element for a given app is the launch request for which we're currently expecting
// a registration.  Any additional elements for the same app are more launch requests for it,
// which have not yet been sent out.  As soon as the first one is ackknowledged, we move on to 
// the next.
static GList *gPendingAppLaunches = NULL;

static GList *gAppWatchers = NULL;
static AlpWatcherRef gLastAppWatcherRef = 0;

static GList *gJumpList = NULL;

static RunningAppInfo *gCurrApp;	// current primary application.
static AlpChannel* gServerChannel;

static gboolean gDisableDefaultApp;


static gboolean gLaunchedPrimary; // set to TRUE the first time we receive a request to 
				// launch a primary app.  Until this happens, we will not launch the 
				// default automatically.  This prevents other stuff in the boot 
				// process from kicking off the default app launch early.

KeyCode gHomeKeyCode = 0;
KeyCode gHomePageKeyCode = 0;

/*
void debugging_free(gpointer ptr)
{
	unsigned long *lp = ptr;
	unsigned long cookie = 0xCAFEBABE;
	if(*lp == cookie) raise(10);
	
	*lp = cookie;
	return;
}
#define g_free(x) debugging_free(x)
*/


/*
	Server Implementation
*/




/*
	General implementation for all kill timers (relaunch & exit).
	This is called when a kill timer actually expires, and we need
	to kill a non-responsive app.
*/
gboolean prv_handle_kill_timer(RunningAppInfo* appInfo, gint which)
{
	// appInfo should always be valid, since the 
	// event source is removed if the appInfo is deallocated.
	ALP_FAIL_NON_FATAL_IF(appInfo == NULL, "NULL appInfo.\n");
	if(appInfo == NULL) return false;
	
	// clear the event source ID - our kill timers never repeat.
	appInfo->killTimer[which] = 0;
	
	
	if(appInfo->appPID <= 0)
	{
		// Not much we can do in this case - we don't know the PID of the app.
		// This generally should not happen, since alp_app_setup() also 
		// registers with the PID.  But... in this case we will simply let
		// it go.  The app can keep running.
		
		ALP_FAIL_NON_FATAL_IF(TRUE, "Kill timer failed.\n");
	}
	else
	{
		ALP_TM(TL(ALP_APP_SERVER_TRACE, "Kill timer expired - signalling  %s", appInfo->appID));
		// Send a kill signal to the process.
		// We should get the disconnect message shortly thereafter.
		kill(appInfo->appPID, SIGKILL);
	}
	
	
	return FALSE; // remove/deallocate the timer event source
}

gboolean prv_exit_timer_proc(gpointer cbData)
{
	return prv_handle_kill_timer((RunningAppInfo*)cbData, PRV_EXIT_TIMER);
}

gboolean prv_relaunch_timer_proc(gpointer cbData)
{
	return prv_handle_kill_timer((RunningAppInfo*)cbData, PRV_RELAUNCH_TIMER);
}

// Called to clear a relaunch or exit kill timer.
void prv_clear_kill_timer(RunningAppInfo *appInfo, int which)
{
	ALP_FAIL_NON_FATAL_IF(appInfo == NULL, "NULL appInfo.\n");
	if(appInfo->killTimer[which] == 0) return;
	
	gboolean found = g_source_remove(appInfo->killTimer[which]);
	ALP_FAIL_NON_FATAL_IF(found == FALSE, "Kill timer removal failed.\n");
	appInfo->killTimer[which] = 0;
	
	ALP_TM(TL(ALP_APP_SERVER_TRACE, "Cleared kill-timer %d on %s",which, appInfo->appID));
	
	return;
}

// Call to begin a relaunch or exit kill timer.
void prv_set_kill_timer(RunningAppInfo *appInfo, int which)
{
	ALP_FAIL_NON_FATAL_IF(appInfo == NULL, "NULL appInfo.\n");	
	ALP_FAIL_NON_FATAL_IF(which != PRV_EXIT_TIMER && which != PRV_RELAUNCH_TIMER, 
					  "Bad kill timer.\n");	
	
	// We never set kill timers on apps that are being debugged.
	// Wait, yes we do - the debugger needs it to be killed if it is relaunched.
//	if(appInfo->launchFlags & ALP_APP_LAUNCH_FLAG_DEBUG)
//		return;
	
	// Clear the old one if it's currently running.
	if(appInfo->killTimer[which] != 0) prv_clear_kill_timer(appInfo, which);
	
	// Sadly, we have to use separate callbacks in order to tell which timer went off.
	GSourceFunc timerProc;
	if(which == PRV_EXIT_TIMER) timerProc = prv_exit_timer_proc;
	else timerProc = prv_relaunch_timer_proc;
	
	// Install new timeout.
	appInfo->killTimer[which] = g_timeout_add(PRV_KILL_TIMER_SECS*1000,
									  timerProc, appInfo);
	
	ALP_TM(TL(ALP_APP_SERVER_TRACE, "Set kill-timer %d on %s", which, appInfo->appID));
	
	return;
}


/*
	A pending launch request timed out.
	Remove it, send an error to the caller, and move on to the next one (if any).
*/
gboolean prv_launch_timeout_proc(gpointer cbData)
{
	PendingLaunch* pend = (PendingLaunch*)cbData;
	char *appID = g_strdup(pend->argv[0]);
	
	ALP_TM(TL(ALP_APP_SERVER_TRACE, "Launch timeout for %s expired! Failing.", 
		  appID));
	
	// Clear timer variable so prv_remove_pending_launch() doesn't try to deallocate it.
	pend->launchTimer = 0;
	
	// Remove the pending launch from our list, deallocate it, send reply to requester.
	prv_remove_pending_launch(pend, ALP_STATUS_APP_LAUNCHPAD, -1);
	
	// If there are more pending requests for this app, dispatch the next one now.
	prv_do_next_pending_launch(appID);
	
	// Since this timed-out launch might have been for a new primary app,
	// we need to check it if there isn't already a primary app.
	if(gCurrApp == NULL) prv_check_default_app();
	
	// Free our saved copy of the appID.
	g_free(appID);
	
	return FALSE; // remove/deallocate the timer event source
}

// Set up a launch timeout for a pending launch request
void prv_set_launch_timeout(PendingLaunch* pend)
{
	
	ALP_FAIL_NON_FATAL_IF(pend == NULL, "NULL pending launch.");
	ALP_FAIL_NON_FATAL_IF(pend->launchTimer != 0, "Launch timeout already pending.");
	
	pend->launchTimer = g_timeout_add(PRV_LAUNCH_TIMEOUT_SECS*1000,
							    prv_launch_timeout_proc, pend);
	
	ALP_TM(TL(ALP_APP_SERVER_TRACE, "Set launch timeout %d on %s", 
		  			pend->launchTimer, pend->argv[0]));
	
	return;
}

// Clear the launch timeout for the pending launch request.
void prv_clear_launch_timeout(PendingLaunch* pend)
{
	ALP_FAIL_NON_FATAL_IF(pend == NULL, "NULL pending launch.");
	if(pend->launchTimer == 0) return;
	
	gboolean found = g_source_remove(pend->launchTimer);
	ALP_FAIL_NON_FATAL_IF(found == FALSE, "Launch timeout removal failed.\n");
	
	ALP_TM(TL(ALP_APP_SERVER_TRACE, "Cleared launch timeout %d on %s",
		  pend->launchTimer, pend->argv[0]));
	
	pend->launchTimer = 0;
	
	return;
}

/*
 Search a standard argument list for the given argument.
 Returns TRUE if argument is found, and FALSE otherwise.
*/
gboolean prv_search_args(char *searchArg, int argc, char* argv[])
{
	int i=0;
 
	while(i<argc)
	{
		if(strcmp(searchArg, argv[i]) == 0)
			return TRUE;
		i++;
	}
 
	return FALSE;
}


void prv_print_pending_list(void)
{
	GList *app = gPendingAppLaunches;
	PendingLaunch *pendLaunch;
	int i=0;
	
	g_print("===== Pending Launches =====\n");
	while(app != NULL)
	{
		pendLaunch = (PendingLaunch*)app->data;
		g_print("     %s\n", pendLaunch->argv[0]);
		app = g_list_next(app);
		i++;
	}
	
	g_print("===== %d Total ===== \n", i);
	

	return;
}


void prv_broadcast_diagnostic_msg(char *appID, char *str)
{
	AlpMessage *msg;
	GList *app;
	RunningAppInfo* appInfo;
	alp_status_t err = ALP_OK;
	
	ALP_TM(TL(ALP_APP_SERVER_TRACE, "Diagnostic message: %s, %s",appID, str));
	
	msg = alp_app_prv_make_diagnostic_msg(appID, str);
	
	app = gAppList;
	
	while(app != NULL)
	{
		appInfo = ((RunningAppInfo*)app->data);
		
		if(appInfo->launchFlags & ALP_APP_LAUNCH_FLAG_DIAGNOSTIC)
		{
			alp_ipc_connection_send(appInfo->connection, msg);
		}
		
		app = g_list_next(app);
	}
	
	alp_ipc_message_destroy(msg);

	return;
}


// A GCompareFunc that just calls strcmp.  
// We could actually use a pointer to strcmp for this, but we'd lose some 
// type safety if the definition of GCompareFunc ever changed...
gint prv_compare_strcmp(gconstpointer a, gconstpointer b)
{ 
	return strcmp(a,b);
}



/*
	Called when the primary application exits or is backgrounded without a replacement already selected.
	We may need to launch a new primary app, so this routine checks the pending launch queue to see if it's necessary,
	and if so it runs the "default" app (home screen or something).
*/
void prv_check_default_app(void)
{
	AlpMessage *msg;
	char* argv = ALP_APP_PRIMARY;
	GList *pendNode;
	PendingLaunch *pend, *lastPrimary;
	gboolean lastIsPrimary = FALSE;
	
	if(gDisableDefaultApp == TRUE || gLaunchedPrimary == FALSE) return;
	
	// If there is no default primary, then we have no work to do.
	if(gDefaultPrimary == NULL) return;
	
	// First check pending launch queue to see if we're in the process of launching a new primary already.
	// Since there could potentially be lots of launches in here, we look thoroughly:
	
	// 1: Find the last launch in the queue which is --alp-primary, and see what app it is.
	pendNode = gPendingAppLaunches;
	lastPrimary = NULL;
	while(pendNode != NULL)
	{
		pend = (PendingLaunch*)pendNode->data;
		
		if(prv_search_args(ALP_APP_PRIMARY, pend->argc, pend->argv))
			lastPrimary = pend;
		
		pendNode = g_list_next(pendNode);
	}
	
	// 2: Look for all pending launches for that app, and make sure 
	// there isn't an --alp-backgrounded after the --alp-primary.
	if(lastPrimary != NULL)
	{
		lastIsPrimary = TRUE;
		pendNode = gPendingAppLaunches;
		
		while(pendNode != NULL)
		{
			pend = (PendingLaunch*)pendNode->data;
			
			// If this is a pending launch for the app we are looking for...
			if(strcmp(lastPrimary->argv[0], pend->argv[0]) == 0 )
			{
				// ... then check if it's primary or backgrounded, and set our boolean accordingly.
				if(prv_search_args(ALP_APP_PRIMARY, pend->argc, pend->argv))
					lastIsPrimary = TRUE;
				else if(prv_search_args(ALP_APP_PRIMARY, pend->argc, pend->argv))
					lastIsPrimary = FALSE;
			}
			
			pendNode = g_list_next(pendNode);
		}
		
		// If the last launch for the last primarily launched app is a primary launch,
		// then there's no real point in adding another primary launch to the queue.
		if(lastIsPrimary == TRUE)
		{
			ALP_TM(TL(ALP_APP_SERVER_TRACE, "Primary app gone, launch for %s is pending.",lastPrimary->argv[0]));
			return;
		}
	}
	
	
	ALP_TM(TL(ALP_APP_SERVER_TRACE, "Primary app gone, launching default %s.",gDefaultPrimary));
	msg = alp_app_prv_make_launch_msg(gDefaultPrimary, 1, &argv, FALSE, ALP_APP_DEFAULT_SECURITY_ID);
	prv_handle_launch(NULL, msg);
	msg = NULL; // prv_handle_launch disposes of it.
	
	// Note that it is not necessary to quit transient apps, etc., 
	// here since that will happen automatically when the new FG app registers.
	
	
	return;
}
		




/*
	Search a standard argument list for the given argument.
	Returns TRUE if argument is found, and FALSE otherwise.
*/
/*
gboolean PrvSearchArgs(char *searchArg, int argc, char* argv[])
{
	int i=0;
	
	while(i<argc)
	{
		if(strcmp(searchArg, argv[i]) == 0)
			return TRUE;
		i++;
	}
	
	return FALSE;
}
*/


/*
	Given an GList of RunningAppInfo's and an appID, return a pointer to the 
	app's RunningAppInfo structure from our list of currently running applications.
	If the application is not currently running, then this routine returns NULL.
*/
RunningAppInfo* prv_find_app_by_id(char *appID)
{
	GList *app = gAppList;
	
	while(app != NULL)
	{
		if(strcmp(((RunningAppInfo*)app->data)->appID, appID) == 0)
			return app->data;
		
		app = g_list_next(app);
	}
	
	return NULL;
}


// Ditto, for pending launches.
PendingLaunch* prv_find_pending_launch(char *appID)
{
	GList *app = gPendingAppLaunches;
	
	while(app != NULL)
	{
		if(strcmp(((PendingLaunch*)app->data)->argv[0], appID) == 0)
			return app->data;
		
		app = g_list_next(app);
	}
	
	return NULL;
}


/*
	Remove a pending launch request from the global list.
	This call removes the pending request, deallocates all memory associated with it,
	and also sends a launch reply if the requester is still waiting for it.
	Note that it is not the only place a launch reply is sent - there are various other 
	error cases where a launch reply is sent when there is no pending launch in the
 	queue yet.
	
	This is used in cases of both success and failure - whenever the pending launch is
 	no longer pending.  Success or failure is indicated by the passed status code.
*/
void prv_remove_pending_launch(PendingLaunch *pendLaunch, alp_status_t err, pid_t appPID)
{
	char *appID = pendLaunch->argv[0];
	
	// Remove pending launch from the list.
	gPendingAppLaunches = g_list_remove(gPendingAppLaunches, pendLaunch);
		
	// If anyone is still waiting for a reply for this launch, we send it here.
	// This happens when the PID is requested in the call to alp_app_launch().
	if(pendLaunch->requester)
		prv_send_launch_reply(pendLaunch->requester, appID, err, appPID);
	
	// Clear timeout, if any:
	prv_clear_launch_timeout(pendLaunch);
	
	// Free launch arguments & structure:
	alp_app_free_argv(pendLaunch->argc, pendLaunch->argv);
	g_free(pendLaunch);
	pendLaunch = NULL;
	
	return;
}


/*
	Search the active watcher list for one matching the given ID.
*/
AppWatcher* prv_watcher_find(AlpWatcherRef ref)
{
	GList *node = gAppWatchers;
	
	while(node != NULL)
	{
		if(((AppWatcher*)node->data)->ref == ref)
			return node->data;
		
		node = g_list_next(node);
	}
	
	return NULL;
}


/*
	Remove all watchers in the active list that are owned by the indicated app.
	This is used when an app exits in order to clean out any watchers it may have requested.
*/
void prv_watcher_remove_by_owner(RunningAppInfo *owner)
{
	GList *node = gAppWatchers;
	GList *nextNode;
	
	while(node != NULL)
	{
		AppWatcher *watcher = (AppWatcher*)node->data;
		nextNode = g_list_next(node);
		
		ALP_FAIL_NON_FATAL_IF(watcher->owner == NULL, "Watcher with null owner!");
		if(watcher->owner == NULL || watcher->owner == owner)
		{
			gAppWatchers = g_list_remove_link (gAppWatchers, node);
			g_list_free(node);
			g_free(watcher);
			ALP_TM(TL(ALP_APP_SERVER_TRACE, "Removed watcher %d owned by %s",
				  watcher->ref, owner->appID));
		}
		
		node = nextNode;
	}
	
	return;
}

/*
	Called whenever a new application registers.
	This searches the list of active watchers, and sets the "wasRun" flag in any that are
	looking for the newly run app.
*/
void prv_watcher_notify(char* appID)
{
	GList *node = gAppWatchers;
	
	while(node != NULL)
	{
		AppWatcher *watcher = (AppWatcher*)node->data;
		if(strcmp(watcher->appID, appID) == 0)
			watcher->wasRun = TRUE;
		
		node = g_list_next(node);
	}
		
	return;
}



/*
	When a client connects, we need to create a new entry for it in the list of "running apps".
*/
void prv_client_connect_cb(AlpConnection* connection, gpointer cbData)
{
	RunningAppInfo *appInfo;
	alp_status_t err;
	
	ALP_TM(TL(ALP_APP_SERVER_TRACE, "Received connect from 0x%lx", (long unsigned int)connection));
	
	
	
	// We have a new client, so allocate a new RunningAppInfo structure, 
	// add it to the list, and set it as the data pointer to be passed to the receive callback.
	appInfo = g_new0(RunningAppInfo, 1);
	gAppList = g_list_prepend(gAppList, appInfo);
	
	// Fill in the client's connection object now.  Until the launchpad registers the app
	// via alp_app_setup(), this structure will be mostly empty, and we will not know which
	// application it represents.
	appInfo->connection = connection;
	
	// DOLATER: The connect operation (incl. callback) is synchronous, right?  
	// So the client can't send us data before I have a chance to install the callback?
	
	// Register a data receive callback, and pass the appInfo pointer as the data 
	// so we can easily get to all the sender info when we receive a message.
	err = alp_ipc_connection_register_receive_callback(connection, prv_receive_data_cb, (gpointer*)appInfo);
	 
	ALP_FAIL_NON_FATAL_IF(err != ALP_STATUS_OK, 
					  "register_receive_callback() failed.\n", NULL);
	
	return;
}


/*
	Dispatch the given launch request for the given app.
	This routine will either forward the message if the app is already running,
	or otherwise it will send it to the appropriate launchpad.
	If a matching launchpad cannot be found, or if no scheme is specified, then an error is returned.
*/
alp_status_t prv_dispatch_launch(int argc, char* argv[])
{
	char *appID = argv[0];
	RunningAppInfo *appInfo = prv_find_app_by_id(appID);
	alp_status_t err;
	AlpMessage* msg = NULL;
	
	
	// Is the app already running?
	// Then we send the launch request to it directly.
	if(appInfo != NULL)
	{
		ALP_TM(TL(ALP_APP_SERVER_TRACE, "App %s is already running, forwarding request...", appID));
					
		// Send a relaunch message to the app.
		// If the app doesn't handle this, the client-side code will turn it into an exit request.
		// When the app exits, we will see that it was pending a launch, and relaunch it with the saved parameters.
		
		// Happily, we can just forward the message:
		// (This will fail if (for example) the app has exited already).
		msg = alp_app_prv_make_launch_msg(NULL, argc, argv, FALSE, ALP_APP_DEFAULT_SECURITY_ID);
		err = alp_ipc_connection_send(appInfo->connection, msg);
				
		// If the app has already exited, we get an error back.
		// Unfortunately the error codes can't currently tell us why the send failed.
		// So, we assume that it's because the connection is now closed, and we will soon 
		// be receiving a disconnect callback.  
		// We therefore ignore the error:
		// This will ensure the launch request remains in the pending queue, and when 
		// the disconnect callback arrives, it will be dispatched again.
		if(err != ALP_OK)
		{
			ALP_TM(TL(ALP_APP_SERVER_TRACE, "Error sending launch to %s... ignoring.", appID));
			err = ALP_OK;
		}
		
		// The app may fail to receive/respond to the relaunch request 
		// (if, for example, it does not return to its main event loop).
		// To handle that case, we set up a kill timer for it.  If we 
		// do not receive the registration within a few seconds, a 
		// kill signal will be sent to the app's process.  When we get the 
		// disconnect callback, it will be launched again.
		prv_set_kill_timer(appInfo, PRV_RELAUNCH_TIMER);
		
		
		// Even though the application is already running, it will either exit or 
		// re-register when it receives the re-launch parameters, so we do not retire the current app at this time.
		// We wait for the re-registration, and if the launchflag PRIMARY is set, then we retire it.
		
		// For the same reason, we don't necessarily sent the launch reply now either - if the PID 
		// is requested, we have to wait for the app to re-register following the relaunch.
	}
	else
	{
		// The app is NOT already running, so we send a fresh launch request to the launchpad.
		ALP_TM(TL(ALP_APP_SERVER_TRACE, "Launching %s via launchpad...", appID));
			
		// Since it's a new launch, we need to look up & pass the security ID.
		AlpBundle bndl = alp_bundle_by_name(argv[0]);
		msg = alp_app_prv_make_launch_msg(NULL, argc, argv, FALSE,
								    alp_get_security_id(bndl));
		
		// Pass the message to the cookie cutter to launch the new app.
		err = prv_send_to_launchpad(appID, msg);

		// We may get an error if there is no launchpad scheme, or if 
		// there is no launchpad for the specified scheme.
		
		// We don't deal with app retirement here, it happens when the new app registers with us as a FG app.	
	}
	
	if(msg) alp_ipc_message_destroy(msg);
	
	return err;
}


/*
	Searches the pending launches list for the next launch request for the indicated app.
	Then, if there is one, it dispatches it (either launching the app, or forwarding the request to
	the existing instance).
*/
void prv_do_next_pending_launch(char *appID)
{
	PendingLaunch *pendLaunch;
	alp_status_t err;
	
	pendLaunch = prv_find_pending_launch(appID);
	if(pendLaunch == NULL) return;
	
	ALP_TM(TL(ALP_APP_SERVER_TRACE, "Dispatching pending launch for %s...: argv[1]=%s", appID, 
				pendLaunch->argc>1 ? pendLaunch->argv[1] : "NULL"));
	
	// We might still have the original launch request message available, 
	// but it's safer to make a new one because it often won't be available.
	
	err = prv_dispatch_launch(pendLaunch->argc, pendLaunch->argv);
	
	// If there was an error dispatching the launch, for whatever reason, 
	// we need to cancel the pending launch request.
	// This subroutine takes care of sending a launch reply if the requester is still waiting.
	if(err != ALP_OK)
	{
		prv_remove_pending_launch(pendLaunch, err, -1);	
	}
	
	// If there *wasn't* an error, then the launch remains 
	// in the pending list until we receive a registration for it.
	
	return;
}

/*
	Remove the indicated element from gAppList, and deallocate the RunningAppInfo structure associated with it.
*/
void prv_free_running_app_info(RunningAppInfo *app)
{
	
	// If this is the "current app", then clear it out - the pointer will no longer be valid.
	if(app == gCurrApp)
		gCurrApp = NULL;
	
	if(app->replyMessage)
		alp_ipc_message_destroy(app->replyMessage);
	
	
	g_free(app);
	
	return;
}



alp_status_t prv_send_to_launchpad(char *appID, AlpMessage *msg)
{
	GList *app;
	char *scheme = g_strdup(appID);
	RunningAppInfo *appInfo;
	
	// Find the launchpad "scheme", the first part of the string.
	// This is likely to change if we move to numeric app IDs, and 
	// will (in that case) be replaced with a PkgMgr call to do the same thing.
	// This logic looks weird because strtok stores state between calls.
	// The first call should return 'scheme', the pointer to the first token.
	// If the second call returns NULL, then it means there is either no separator 
	// char in the string, or nothing following the separator.	
	if(strtok(scheme, ":") != scheme || strtok(NULL, ":") == NULL)
	{
		ALP_TM(TL(ALP_APP_SERVER_TRACE, "appID with no scheme, not launching %s", appID));
		return ALP_STATUS_APP_PARAM;
	}
	
	
	// Search through list of running apps for a launchpad that can handle this app scheme:
	for(app = gAppList; app != NULL; app = g_list_next(app))
	{
		appInfo = app->data;
		
		if(appInfo->launchFlags & ALP_APP_LAUNCH_FLAG_LAUNCHPAD &&
			strcmp(scheme, appInfo->launchpadScheme) == 0)
			break;
	}
	
	// If we didn't find one, send a diagnostic message and return.
	if(app == NULL)
	{
		ALP_TM(TL(ALP_APP_SERVER_TRACE, "Can't find launchpad for scheme '%s', not launching.", scheme));

		// DOLATER - reenable this diagnostic message when we learn if it's safe to ignore SIGPIPE when we send data to a process that has exited.
//		prv_broadcast_diagnostic_msg(appID, ALP_DIAG_MISSING_LAUNCPAD);
		return ALP_STATUS_APP_NO_LAUNCHPAD;
	}
	
	// Otherwise, we send the launch request on to the launchpad:
	if(alp_ipc_connection_send(appInfo->connection, msg) != ALP_OK)
	{
		ALP_TM(TL(ALP_APP_SERVER_TRACE, "Can't send request to launchpad, %s", scheme));
		return ALP_STATUS_APP_IPC;
	}
//	else g_print("Sending fork msg for %s to %s.\n", appID, appInfo->appID);
		
	
	return ALP_STATUS_OK;
}



/*
	"Retire" the indicated app, either relaunching it in the background, or requesting that it terminate.
	If it is marked "transient", it is probably already running in the background, so we send it a "transient close" argument when relaunching it.
*/
void prv_retire_app(RunningAppInfo *appInfo)
{
	gpointer val;
	AlpMessage *msg;
	
	// If there is no app, then we're done.  
	// Nothing else to do except wait for the new one to register.
	if(appInfo == NULL) return;
	
	ALP_TM(TL(ALP_APP_SERVER_TRACE, "Retiring app: %s", appInfo->appID));

	// We should only ever retire the primary app, or a transient app
	if((appInfo->launchFlags & ALP_APP_LAUNCH_FLAG_PRIMARY) == 0 && 
		(appInfo->launchFlags & ALP_APP_LAUNCH_FLAG_TRANSIENT) == 0 )
	{
		ALP_FAIL_NON_FATAL_IF(TRUE, "Warning; Can't retire ordinary bg app %s.\n", appInfo->appID);
		return;
	}
	
	if(appInfo->appID[0] == '\0')
	{
		ALP_FAIL_NON_FATAL_IF(TRUE, "Warning; Can't retire unregistered app %s.\n", appInfo->appID);
		return;	
	}
	
	// Check app configuration - does it need to continue running in the background?
	if(appInfo->backgrounding == ALP_APP_BACKGROUND_REQUIRED)
	{
		char *argvPrimary[1] = {ALP_APP_BACKGROUNDED};
		// ... in that case, we relaunch it in the background.	
		
		// we now send the "backgrounded" argument to both primary & transient apps.
		// We used to have "--alp-close-transient" for the transient case.
		msg = alp_app_prv_make_launch_msg(appInfo->appID, 1, argvPrimary, FALSE, ALP_APP_DEFAULT_SECURITY_ID);
		
		prv_handle_launch(NULL, msg);
		msg = NULL; // prv_handle_launch disposes of it.
	}
	else
	{
		// Otherwise, request that it exit.
		ALP_TM(TL(ALP_APP_SERVER_TRACE, "Sending exit request to %s.", appInfo->appID));
		
		msg = alp_app_prv_make_exit_msg(appInfo->appID);		
		prv_handle_exit_request(msg);
		alp_ipc_message_destroy(msg);
	}
	
	// If we just retired the current app, then mark it so.
	if(appInfo == gCurrApp) gCurrApp = NULL;
	
		
	
	
	return;
}

// Retires all currently running apps marked 'transient'.
void prv_retire_transient_apps(void)
{
	GList *app;
	AlpMessage *msg;
	RunningAppInfo *appInfo;
	
	ALP_TM(TL(ALP_APP_SERVER_TRACE, "Retiring all transient apps."));

	// Now, we need to run through the list of running apps and quit any transient ones.
	app = gAppList;
	while(app != NULL)
	{
		appInfo = (RunningAppInfo*)app->data;
		
		
		
		// If the app is flagged as transient... 
		// ... but NOT if it's also primary (either a beginning or ending primary)
		if((appInfo->launchFlags & ALP_APP_LAUNCH_FLAG_TRANSIENT) &&
				  (appInfo->launchFlags & ALP_APP_LAUNCH_FLAG_PRIMARY) == 0)
		{
			// Then, retire the app.
			ALP_TM(TL(ALP_APP_SERVER_TRACE, "Retiring transient app %s.", appInfo->appID));
			
			prv_retire_app(appInfo);
		}
//		else ALP_TM(TL(ALP_APP_SERVER_TRACE, "Not retiring non-transient app %s.", appInfo->appID));
		
		app = g_list_next(app);
	}
	
	return;
}


/*
	There is a bunch of special stuff we do the first time a new app registers.
	This is where we do it.
*/
void prv_do_initial_app_registration(RunningAppInfo *appInfo, char *appID, 
							   pid_t appPID, guint32 launchFlags, char *scheme)
{
	GList *pendingEntry;
	char *pendingName;
	
	// Save the new app's PID.
	appInfo->appPID = appPID;
		
		// The app ID should never be allowed to be longer than will fit in our buffer:
	strncpy(appInfo->appID, appID, PKG_ID_MAX_LEN);
	if(appInfo->appID[PKG_ID_MAX_LEN-1] != '\0')
	{
		ALP_FAIL_NON_FATAL_IF(TRUE, "Registration from invalid app name: %s (max  of %d bytes).\n", appID, PKG_ID_MAX_LEN);
		
		appInfo->appID[PKG_ID_MAX_LEN-1] = '\0';
	}
	
		// copy the launchpad scheme into the RunningAppInfo afterwards.
	strncpy(appInfo->launchpadScheme, scheme, ALP_APP_LAUNCHPAD_SCHEME_SIZE);
	if(appInfo->launchpadScheme[ALP_APP_LAUNCHPAD_SCHEME_SIZE-1] != '\0')
	{
		ALP_FAIL_NON_FATAL_IF(TRUE, "launchpad registration with invalid scheme: %s\n", scheme);
		appInfo->launchpadScheme[ALP_APP_LAUNCHPAD_SCHEME_SIZE-1] = '\0';
	}
		
	// Initialize default backgrounding setting.
	appInfo->backgrounding = ALP_APP_BACKGROUND_DEFAULT;
		
		
	// Need to copy launchflags too, since they have not been initialized yet.
	appInfo->launchFlags = launchFlags;
	
	
	// ***
	// * For MAX applications (in bar packages), we check the manifest 
	// * properties for various settings:
	// ***
	if(strncmp(ALP_BUNDLE_TYPE_BAR ":", appID, strlen(ALP_BUNDLE_TYPE_BAR)+1) == 0)
	{
		// Use PkgMgr or AIM to get a copy of the apps properties 
		// & note the initial backgrounding settings & transient pref.
		
		ALP_TM(TL(ALP_APP_SERVER_TRACE, "Looking for appPkg(%s)", appID));
		
		AlpBundle appPkg = alp_bundle_by_name(appID);
		char *propVal = NULL;
			
		ALP_TM(TL(ALP_APP_SERVER_TRACE, "Looking for appPkg(%s)=%d", appID, appPkg.value));

		if(!alp_bundle_is_NULL(appPkg))
		{
			
			// If transient property is true, set the appropriate launch flag.
			propVal = alp_bundle_property_value(appPkg, 
					ALP_BUNDLE_PROPERTY_APPLICATION, 0, ALP_APP_PROP_TRANSIENT);
		
			if(propVal)
			{
				if(strcmp(propVal, "true") == 0)
				{
					appInfo->launchFlags |= ALP_APP_LAUNCH_FLAG_TRANSIENT;
					ALP_TM(TL(ALP_APP_SERVER_TRACE, "Setting %s transient from manifest", appID));
				}
				
				g_free(propVal);
			}
			
			
			// If backgrounding property is set, obey it.
			propVal = alp_bundle_property_value(appPkg, 
					ALP_BUNDLE_PROPERTY_APPLICATION, 0, ALP_APP_PROP_BACKGROUNDING);
			
			if(propVal)
			{
				if(strcmp(propVal, ALP_APP_PROP_BG_UNSUPPORTED) == 0)
					appInfo->backgrounding = ALP_APP_BACKGROUND_UNSUPPORTED;
				else if(strcmp(propVal, ALP_APP_PROP_BG_SUPPORTED) == 0)
					appInfo->backgrounding = ALP_APP_BACKGROUND_SUPPORTED;
				else if(strcmp(propVal, ALP_APP_PROP_BG_PREFERRED) == 0)
					appInfo->backgrounding = ALP_APP_BACKGROUND_PREFERRED;
				else if(strcmp(propVal, ALP_APP_PROP_BG_REQUIRED) == 0)
					appInfo->backgrounding = ALP_APP_BACKGROUND_REQUIRED;
				
				ALP_TM(TL(ALP_APP_SERVER_TRACE, 
						"Set %s backgrounding to '%s' from manifest.", 
						appID, propVal));
				
				g_free(propVal);
			}
			
			
			// If "jump list" property is set, obey it.
			propVal = alp_bundle_property_value(appPkg, 
					ALP_BUNDLE_PROPERTY_APPLICATION, 0, ALP_APP_PROP_JUMP_LIST);
			
			if(propVal)
			{
				if(strcmp(propVal, "never") == 0)
					appInfo->serverFlags |= ALP_APP_SERVER_FLAG_NO_JUMP_LIST;
				
				else if(strcmp(propVal, "always") == 0)
					appInfo->serverFlags |= ALP_APP_SERVER_FLAG_ALWAYS_JUMP_LIST;
				
				ALP_TM(TL(ALP_APP_SERVER_TRACE, 
						"Set %s jump-list to '%s' from manifest.", 
						appID, propVal));
				
				g_free(propVal);
			}
			
			// If "hidden" property is set, also don't display it in the jump list.
			// This is actually no longer the case - prefs panels will be hidden,
			// and need to appear in the jump list.  So, hidden apps not wishing to 
			// appear in the jump list must also specify jump-list=never
			/*
			propVal = alp_bundle_property_value(appPkg, 
					ALP_BUNDLE_PROPERTY_APPLICATION, 0, ALP_APP_PROP_HIDDEN);
			
			if(propVal)
			{
				if(strcmp(propVal, "true") == 0)
					appInfo->serverFlags |= ALP_APP_SERVER_FLAG_NO_JUMP_LIST;
				
				ALP_TM(TL(ALP_APP_SERVER_TRACE, 
						"Set %s hidden to '%s' from manifest.", 
						appID, propVal));
				
				g_free(propVal);
			}
			*/
		}
	}
	
	// If the launchpad flag is set, then we now have a new launchpad.
	// We used to check here if this was the *native* launchpad just starting up,
	// and take the opportunity to launch the default app.  Now the startup app is 
	// run directly from the boot script, so we don't have much special to do here anymore.
#if ALP_TRACE == ALP_TRACE_ON
	if(launchFlags & ALP_APP_LAUNCH_FLAG_LAUNCHPAD)
	{		
		ALP_TM(TL(ALP_APP_SERVER_TRACE, "New launchpad registered for %s packages.", appInfo->launchpadScheme));
	}
#endif
	
	// Notify any watchers that the new app is running:
	prv_watcher_notify(appID);
	
	return;
}



// Send a launch reply to the indicated destination.
void prv_send_launch_reply(RunningAppInfo *dest, char *appID, alp_status_t launchErr, pid_t appPID)
{
	GList *waitingApp = gAppList;
	AlpMessage *msg;
	alp_status_t err;
	
	if(dest == NULL) 
	{
		ALP_FAIL_NON_FATAL_IF(dest == NULL, "prv_send_launch_reply: NULL destination.");
		return;
	}
	
	if(dest->replyMessage == NULL) 
	{
		ALP_FAIL_NON_FATAL_IF(TRUE, "prv_send_launch_reply: NULL reply message.");
		return;
	}
	
	
	err = ALP_OK;
	msg = alp_app_prv_make_launch_reply_msg(dest->replyMessage, appID, launchErr, appPID);
	if(msg == NULL) err = ALP_STATUS_APP_IPC;
	else err = alp_ipc_connection_send(dest->connection, msg);

	// Send may fail if (for example) the requester's process has died.	
	
#if ALP_TRACE == ALP_TRACE_ON
	if(err == ALP_OK) ALP_TM(TL(ALP_APP_SERVER_TRACE, "Sent launch reply 0x%lx to %s (0x%lx).", (long)launchErr, dest->appID, (unsigned long)dest->connection));
	else ALP_TM(TL(ALP_APP_SERVER_TRACE, "Launch reply to %s (0x%lx) failed to send! Error = 0x%lx", dest->appID, (unsigned long)dest->connection, (long)err));
#endif	
	
	
	alp_ipc_message_destroy(msg);
	alp_ipc_message_destroy(dest->replyMessage);
	dest->replyMessage = NULL;
	
	// Clear "waiting flag" and waitingFor string.
	dest->serverFlags &= ~ALP_APP_SERVER_FLAG_WAITING_FOR_REPLY;
	strcpy(dest->waitingFor, "");
	
	return;
}

/*******************************************************
	Jump List Management
 *******************************************************/

// Pack up the current contents of the jump list, and send it to the requester.
void prv_send_jumplist_reply(RunningAppInfo *senderInfo, AlpMessage *replyTo)
{
	AlpMessage *msg;
	guint len = g_list_length(gJumpList);
	char **appIDs = g_new(char*, len);
	gboolean *isRunnings = g_new(gboolean, len);
	guint i;
	GList *element;
	
	// Build the necessary arrays for the jump list data:
	i=0;
	element = gJumpList;
	while(element != NULL && i < len)
	{
		appIDs[i] = element->data;		
		
		if(prv_find_app_by_id(appIDs[i]) != NULL)
			isRunnings[i] = TRUE;
		else isRunnings[i] = FALSE;
		
		i++;
		element = g_list_next(element);
	}
	
	
	// All we need to do is send a reply to the sender:
	msg = alp_app_prv_make_jump_list_msg(replyTo, i, appIDs, isRunnings);
	alp_ipc_connection_send(senderInfo->connection, msg);
	alp_ipc_message_destroy(msg);
	
	g_free(appIDs);
	g_free(isRunnings);
	
	return;
}

// Called to record the fact that a new app was launched as primary.
// This routine accumulates the app launch into the jump list.
// Depending on the jumplist tracking algorithm, this may or may 
// not result in any changes to the list.
void prv_jumplist_accumulate(char *appID)
{
	// Don't ever add the default application to the jump list.
	// We switch back to it frequently.
	if(gDefaultPrimary != NULL && strcmp(appID, gDefaultPrimary) == 0)
	{
		return;
	}
		   
	ALP_TM(TL(ALP_APP_SERVER_TRACE, "Adding %s to jump list...", appID));
	
	GList *app = gJumpList;

	// Remove it from the list if it is already there, so we don't
	// get duplicates.  We'll put it back at the top of the list next
	while(app != NULL)
	{
		char *item;
		item = (char *)(app->data);
		if(strcmp(item, appID) == 0)
		{
			gJumpList = g_list_remove(gJumpList, item);
			g_free(item);
			break;
		}
		app = g_list_next(app);
	}

	// Add it to the front of the list
	gJumpList = g_list_prepend(gJumpList, g_strdup(appID));
	
	// Make sure list does not exceed max size 
	// (include enough space for all running apps, plus some number of recent ones);
	if(g_list_length(gJumpList) > (PRV_JUMPLIST_NUM_RECENT + g_list_length(gAppList)))
	{
		// Remove the least recently launched non-running app from the jumplist.
		// Start at the end, and go back towards the beginning until we find one that is not running.
		
		app = g_list_last(gJumpList);
		while(app != NULL && prv_find_app_by_id(app->data) != NULL) 
			app = g_list_previous(app);
		
		if(app)
		{
			char* last = app->data;
			gJumpList = g_list_remove(gJumpList, last);
			g_free(last);
		}
	}
	
	return;
}


/*******************************************************
	Message & Connection Handling
 *******************************************************/
// Loop through all running apps, and send them a theme update message.
// Their appmgr client-side code will reload the theme for their process.
static void prv_handle_theme_update(RunningAppInfo *senderInfo, AlpMessage *msg)
{
	GList *app = gAppList;
	RunningAppInfo *info;
	
	ALP_TM(TL(ALP_APP_SERVER_TRACE, "Received theme update message from %s.  Forwarding to all running apps.", senderInfo->appID));
	
	while(app != NULL)
	{
		info = (RunningAppInfo*)app->data;
		
//		ALP_TM(TL(ALP_APP_SERVER_TRACE, "Sending theme update to %s.", info->appID));
		msg = alp_app_prv_make_theme_update_msg();
		if(info->connection != 0) alp_ipc_connection_send(info->connection, msg);
		alp_ipc_message_destroy(msg);
		
		app = g_list_next(app);
	}
	
	return;
}


/*
	Handle requests from the watcher APIs.
*/
void prv_handle_watcher(RunningAppInfo *requesterInfo, AlpMessage *msg)
{
	char *targetApp;
	AlpWatcherRef ref;
	gboolean wasRun, remove;
	alp_status_t err;
	AppWatcher *watcher = NULL;
	AlpMessage *reply;
	alp_status_t result;
	gint blockMode;
	
	err = alp_app_prv_unpack_watcher_msg(msg, &targetApp, &ref, &wasRun, &remove, &result, &blockMode);
	if(err == ALP_OK)
	{
		// If ref was not specified, then we need to allocate a new watcher.
		if(ref == 0) 
		{
			// Watchers are used infrequently enough that we 
			// just assume gLastWatcherRef will never overflow.
			// Code could be added to safely recycle watcher IDs, 
			// but it's not needed for now (or probably ever).
			gLastAppWatcherRef++;
			ref = gLastAppWatcherRef;
			
			// Allocate & fill out the watcher
			watcher = g_new0(AppWatcher, 1);
			if(watcher == NULL) err = ALP_STATUS_APP_MEM;
			else
			{
				strncpy(watcher->appID, targetApp, PKG_ID_MAX_LEN);		
				watcher->appID[PKG_ID_MAX_LEN-1] = '\0';
				watcher->wasRun = FALSE;
				watcher->ref = ref;
				watcher->owner = requesterInfo;
				watcher->blockLaunch = FALSE;
				
				// Add it to the list.
				gAppWatchers = g_list_prepend(gAppWatchers, watcher);
			}
		}
		else // Otherwise, find the appropriate watcher from our list:
		{
			watcher = prv_watcher_find(ref);
			
			// Verify the owner is making the call:
			if(watcher->owner != requesterInfo) watcher = NULL;
			
			// Send back appropriate error if the ref is bad.
			if(watcher == NULL) err = ALP_STATUS_APP_NOT_FOUND;
		}
		
		// ... search to see if watcher's app is running, update "wasRun".
		if(watcher && prv_find_app_by_id(watcher->appID) != NULL)
		{
			watcher->wasRun = TRUE;
		}
		
		// ... update launch blocking behavior, if necessary.
		if(blockMode == ALP_APP_WATCHER_BLOCK) watcher->blockLaunch = TRUE;
		else if(blockMode == ALP_APP_WATCHER_UNBLOCK) watcher->blockLaunch = FALSE;
		// else it should be ALP_APP_WATCHER_IGNORE, so we leave it unchanged.
		
		
		g_free(targetApp); // dealloc string we unpacked from the message
		
	} // end if unpack succeeded
	
	
	// Compose & send the reply message before possibly deallocating the watcher.
	if(watcher) reply = alp_app_prv_make_watcher_msg(msg, watcher->appID, 
							watcher->ref, watcher->wasRun, FALSE, err, ALP_APP_WATCHER_IGNORE);
	else reply = alp_app_prv_make_watcher_msg(msg, "", 
							  0, FALSE, FALSE, err, ALP_APP_WATCHER_IGNORE);
	
	alp_ipc_connection_send(requesterInfo->connection, reply);
	alp_ipc_message_destroy(reply);
	
	
	// If 'remove' is set, then deallocate the watcher.
	if(watcher && remove == TRUE)
	{
		gAppWatchers = g_list_remove(gAppWatchers, watcher);
		g_free(watcher);
		watcher = NULL;
	}
	
	return;
}


/*
	Handle dispatch of exit request message from a client.
	This call basically just forwards the request to the appropriate client,
	and sets up a "kill timer" to kill the app if it does not exit within a few seconds.
*/
void prv_handle_exit_request(AlpMessage *exitMsg)
{
	RunningAppInfo *appInfo = NULL;
	char *appID;
	alp_status_t err = ALP_OK;
	
	if(exitMsg == NULL)
	{
		ALP_FAIL_NON_FATAL_IF(TRUE, "prv_handle_exit_request received null exitMsg\n", NULL);
		return;
	}
	
	
	// Unpack message:
	err = alp_app_prv_unpack_exit_msg(exitMsg, &appID);
	
	// Verify that message format matches our expectations:
	if(err)
	{
		ALP_FAIL_NON_FATAL_IF(TRUE, "received malformed MSG_ID_EXIT.\n", NULL);
		return;
	}
	
	ALP_TM(TL(ALP_APP_SERVER_TRACE, "Exit request received for %s, forwarding...", appID));
	
	// Re-send the exit request to the appropriate app.
	if(appID == NULL || strlen(appID) == 0)
	{
		// If request was alp_app_exit(NULL), exit the current app
		// But only if there *is* a current app...
		if(gCurrApp != NULL) appInfo = prv_find_app_by_id(gCurrApp->appID);
	}
	else
	{
		// else exit the requested app
		appInfo = prv_find_app_by_id(appID);
	}

	if(appInfo != NULL)
	{		
		
		// This may fail if (for example), the app has already exited/crashed/etc.
		// Therefore, we don't check the error result.
		// We could check it, and only allow the specific error code for 
		// "other side gone", but I don't think we *get* a specific error 
		// code for that yet. 
		alp_ipc_connection_send(appInfo->connection, exitMsg);
		
		prv_set_kill_timer(appInfo, PRV_EXIT_TIMER);
	}
	else
	{
		// DOLATER: Error case - app is not running.
		// We'd like to return a status value here or something, 
		// but it's a message-request sort of call rather than a true RPC, so we can't.
	}
	
	// Deallocate the appID.
	// Message is deallocated by our caller.
	g_free(appID);
	
	return;
}



/*
	Handle dispatch of retire request message from a client.
	This call basically just retires the current primary app.
*/
void prv_handle_retire_request(AlpMessage *retireMsg)
{
	RunningAppInfo *appInfo;
	char *appID;
	alp_status_t err = ALP_OK;
	
	if(retireMsg == NULL)
	{
		ALP_FAIL_NON_FATAL_IF(TRUE, "prv_handle_retire_request received null retireMsg\n", NULL);
		return;
	}
	
	
	// Unpack message:
	err = alp_app_prv_unpack_retire_msg(retireMsg, &appID);
	
	// Verify that message format matches our expectations:
	if(err)
	{
		ALP_FAIL_NON_FATAL_IF(TRUE, "received malformed MSG_ID_RETIRE.\n", NULL);
		return;
	}
	
	ALP_TM(TL(ALP_APP_SERVER_TRACE, "retire request received for %s.", appID));
	
	
	// we only do anything if it's the current primary app.
	// Otherwise, it's already retired.
	if(gCurrApp != NULL && 
		  (appID == NULL || strlen(appID) == 0 || strcmp(appID, gCurrApp->appID) == 0))
	{
		prv_retire_app(gCurrApp);
		if(gCurrApp == NULL) prv_check_default_app();
	}
		
	// Deallocate the appID.
	// Message is deallocated by our caller.
	g_free(appID);
	
	return;
}

/*
	Handle "set backgrounding" message from a client.
	This call changes the current backgrounding setting for the requesting app.
	It may also result in an exit request being sent to the application
	(if, for example, it's running in the background and sets its preference to ALP_BACKGROUND_UNSUPPORTED).
*/
void prv_handle_set_backgrounding(RunningAppInfo *appInfo, AlpMessage *msg)
{
	AlpAppBackgroundingType bg;
	alp_status_t err = ALP_OK;
	
	if(msg == NULL)
	{
		ALP_FAIL_NON_FATAL_IF(TRUE, "received null msg\n", NULL);
		return;
	}
	
	
	// Unpack message:
	err = alp_app_prv_unpack_backgrounding_msg(msg, &bg);
	
	// Verify that message format matches our expectations:
	if(err)
	{
		ALP_FAIL_NON_FATAL_IF(TRUE, "received malformed backgrounding msg.\n", NULL);
		return;
	}
	
	
	// If app is running in the background, and no longer requires it, 
	// then we ask it to exit.  Eventually, we may allow it to keep running here.
	if(bg < ALP_APP_BACKGROUND_REQUIRED && 
		  (appInfo->launchFlags & ALP_APP_LAUNCH_FLAG_PRIMARY) == 0)
	{
		AlpMessage *exitMsg = alp_app_prv_make_exit_msg(appInfo->appID);		
		prv_handle_exit_request(exitMsg);
		alp_ipc_message_destroy(exitMsg);
	}
	
	
	// Update the backgrounding setting:
	appInfo->backgrounding = bg;
	
	
	return;
}


/*
	This is the bottleneck for all incoming launch requests.  
	Debug, primary, already running, etc. - they are all different forms of launch.
	
*/
void prv_handle_launch(RunningAppInfo *requesterInfo, AlpMessage *msg)
{
	char *appID;
	int argc;
	char **argv = NULL;
	RunningAppInfo *appInfo = NULL;
	PendingLaunch	*pendLaunch;
	gboolean pending;
//	char *findResult;
	alp_status_t err = ALP_OK;
	gboolean waitForPID;
	pid_t appPID;
	GList *node;
	AppWatcher *watcher;
	unsigned int	secID;
	
	// Unpack message:
	err = alp_app_prv_unpack_launch_msg(msg, &argc, &argv, &waitForPID, &secID);
	appID = argv[0];
		
	// if caller is not waiting for a response, null out the requester info.
	if(waitForPID == FALSE) requesterInfo = NULL;
	
	// Since otherwise a launch request is synchronous, 
	// we must properly prepare to send a response to the requester.
	if(requesterInfo)
	{
		ALP_FAIL_NON_FATAL_IF(requesterInfo->serverFlags & 
				ALP_APP_SERVER_FLAG_WAITING_FOR_REPLY, 
				"Client %s is already waiting on reply.\n", 
				requesterInfo->appID);
		
		if(requesterInfo->replyMessage != NULL) 
				alp_ipc_message_destroy(requesterInfo->replyMessage);
			
			// Mark the launch requester as "waiting for a reply".
			// We also note what app it's waiting for.  Then, 
			// (in the case of a debug launch) when the new app registers,
			// we will send the reply to the requester.
		requesterInfo->serverFlags |= ALP_APP_SERVER_FLAG_WAITING_FOR_REPLY;
		
		strncpy(requesterInfo->waitingFor, appID, PKG_ID_MAX_LEN);
		requesterInfo->waitingFor[PKG_ID_MAX_LEN-1] = '\0';
		requesterInfo->replyMessage = msg;
	}
	
	// If there was a problem unpacking the message, 
	// we skip to the end and send an error back to the requester immediately, 
	// and we're done.  The launch will not proceed.
	ALP_FAIL_NON_FATAL_IF(err, "received malformed MSG_ID_FORK_APP.\n", NULL);
	if(err) err = ALP_STATUS_APP_MSG; // translate error for caller.
	
	
	
	if(err == ALP_OK)
	{
		// Okay so far, launch will proceed...
		
		ALP_TM(TL(ALP_APP_SERVER_TRACE, "Launching App %s... argv[1]=%s", appID, argc>1?argv[1]:"NULL"));
		
		// If this is our first launch request for a primary app,
		// then we set gLaunchedPrimary.  This enables the code
		// that automatically switches back to the default primary
		// when needed.  Until then, there will be no primary app.
		if(gLaunchedPrimary == FALSE && 
				 prv_search_args(ALP_APP_PRIMARY, argc, argv))
		{
			gLaunchedPrimary = TRUE;
		}
		
		
		// First, we need to determine what state the app is in.
		// Is it running normally?  
		appInfo = prv_find_app_by_id(appID);
		
		// Are there any existing pending (re)launch requests for this app?
		// The app might not have connected with us or registered yet, so the 'gPendingAppLaunches' 
		// list is how we remember them until that happens.  
		// Currently, you can't do anything to an app in this state.
		// If it's pending, then we just queue up the launch request for later dispatch.
		pending = (prv_find_pending_launch(appID) != NULL);
		
		
		// We disallow the launch if the app is NOT running, and there is a watcher blocking it.
		if(appInfo == NULL)
		{
			node = gAppWatchers;
			while(node != NULL) // search list for any blocking watchers
			{
				watcher = (AppWatcher*)node->data;
				if(strcmp(watcher->appID, appID) == 0 && watcher->blockLaunch == TRUE)
				{
					err = ALP_STATUS_APP_BLOCKED;
					break;
				}
				node = g_list_next(node);
			}
		}
		
		
		// If we wanted to be pedantic, we could catch other 
		// "error cases" for invalid launch arguments here.
		
		
		if(err == ALP_OK)
		{
			// Still okay, launch will proceed...
			
			// If the app already has a launch pending, then we can't deal with the new one immediately.
			// In that case, we just add it to the pending list (below), and it will be dealt with later.
			// Otherwise, we dispatch the launch request (either to a launchpad or to an existing 
			// app instance).
			if(pending == FALSE)
			{
				// No launch pending - dispatch the launch request appropriately.
				err = prv_dispatch_launch(argc, argv);
			}
			
			ALP_TM(TL(ALP_APP_SERVER_TRACE, "Launching dispatch result: 0x%lx, reply will be sent to 0x%lx...", (long)err, (long)requesterInfo));
			
#if ALP_TRACE == ALP_TRACE_ON
			if(pending == true) ALP_TM(TL(ALP_APP_SERVER_TRACE, "Launch already pending for %s, queueing...", appID));
#endif
			
		}
	}
	
	// If there was no requester, then this launch request was generated internally by the 
	// appserver, and there is no need for a reply.  So we deallocate the message here 
	// instead of in prv_send_launch_reply.  We can't do this above, since the message is 
	// needed in the case where we dispatch it.
	if(requesterInfo == NULL)
	{
		alp_ipc_message_destroy(msg);
	}
	
	// If there was an error, then we send the reply immediately.
	// Otherwise, then we have to wait for app registration so we can send the proper PID back.
	// Note that if requesterInfo == NULL, this isn't a "real" launch request, just an internal call 
	// (to launch the default UI app, for example) so we send no reply.
	if(requesterInfo != NULL && err != ALP_OK)
	{
		if(appInfo != NULL) appPID = appInfo->appPID;
		else appPID = -1;
		
		prv_send_launch_reply(requesterInfo, appID, err, appPID);
		requesterInfo = NULL; //clear requester info so we don't generate another reply later.
	}
	
	// Okay, if we get this far, and there are no errors, then the launch will proceed 
	// and we need to add the launch request to the list of pending launches.
	// Note that if this app has no other launches already pending, then this
	// one was already sent to it above. However, it still needs to be added to the queue so
	// we can send a reply.  The launch request will remain there until we know that it 
	// was received by the app (i.e., the app registers with us). 
	// We save the appID & launch arguments, since they may be needed later.
	if(err == ALP_OK)
	{	
		pendLaunch = g_new0(PendingLaunch, 1);
		pendLaunch->argc = argc;
		pendLaunch->argv = argv;
		pendLaunch->requester = requesterInfo;
		
		gPendingAppLaunches = g_list_append(gPendingAppLaunches, pendLaunch);
		
		prv_set_launch_timeout(pendLaunch);
	}
	else
	{
		// There was an error; argc & argv aren't being saved, so we'd better deallocate them.
		if(argv != NULL) alp_app_free_argv(argc, argv);
	}
	
	
	return;
}




/*
	This routine handles application registration on the server side.
	This includes both the initial registration (mostly move to a subroutine),
	and any re-registration the app may do (in the case of a relaunch, for example).
*/
void prv_handle_app_register(RunningAppInfo *senderInfo, AlpMessage *msg)
{
	char *appID;
	pid_t appPID;
	guint32 launchFlags;
	char *scheme;
	PendingLaunch *pendLaunch;
	
	alp_status_t err = ALP_OK;
	
	// Unpack message:
	err = alp_app_prv_unpack_register_msg(msg, &appID, &appPID, &launchFlags, &scheme);
	
	if(err)
	{
		ALP_FAIL_NON_FATAL_IF(TRUE, "received malformed MSG_ID_REGISTER.\n", NULL);
		return;
	}
	
	ALP_TM(TL(ALP_APP_SERVER_TRACE, "Register received from %s: 0x%lx", appID, (unsigned long)launchFlags));
	
#if ALP_TRACE == ALP_TRACE_ON
	
	if(launchFlags & ALP_APP_LAUNCH_FLAG_PRIMARY)
	{
		if(gCurrApp != senderInfo) ALP_TM(TL(ALP_APP_SERVER_TRACE, "Primary app will change from %s to %s.", gCurrApp ? gCurrApp->appID : "", appID));
		else  ALP_TM(TL(ALP_APP_SERVER_TRACE, "Primary app will remain %s.", appID));
	}
#endif
	
	// Since we received a registration from this app, there is probably a launch request 
	// waiting in the pending queue.  The only exceptions should be when a system service
	// (not a real app) uses alp_app_setup() to initialize a connection to appserver so 
	// it can use various application-based services.  If there is one, we need to remove it,
	// and deallocate it.
	// Note that we are just assuming here that when we get a registration, it is in response to
	// the most recent launch request that was sent to the app.  If alp_app_register() is called 
	// directly, it will mess this up.  That's why alp_app_register is documented to be called only
	// in response to a launch request.  It would probably be better to 
	// allocate an ID to each incoming launch request, and match registrations up 
	// with them... maybe we'll have time for it in v2.  Meanwhile, as long as 
	// alp_app_register() is used appropriately, this should not cause problems.
	pendLaunch = prv_find_pending_launch(appID);
	if(pendLaunch)
	{
		prv_remove_pending_launch(pendLaunch, ALP_OK, appPID);
	}
	
#if ALP_TRACE == ALP_TRACE_ON
	if(pendLaunch == NULL)  ALP_TM(TL(ALP_APP_SERVER_TRACE, 
								"Registration from non-pending app %s\n", appID));
#endif
	
	// Clear pending 'relaunch' kill timer, if any.
	prv_clear_kill_timer(senderInfo, PRV_RELAUNCH_TIMER);
	
	
	
	// Now we can fill out the rest of the app's RunningAppInfo, if we haven't already.
	// If the information has been filled out once, we do not allow it to be changed by re-registration.
	if(senderInfo->appPID == 0) 
	{
		prv_do_initial_app_registration(senderInfo, appID, appPID, launchFlags, scheme);
		
		// We need to make sure launchFlags is set correctly in this case as 
		// well as the 'else' below.  Since prv_do_initial_app_registration()
		// may set additional bits based on manifest properties, we re-read it here.
		launchFlags = senderInfo->launchFlags;
	}
	else // App is re-registering:
	{
		// For now, all we do is verify some error cases.
		// If we get invalid data, we ignore the registration.
		// DOLATER: If we were feeling strict, we could kill the app in thse cases.
		
		// Is the app trying to re-register with a different package URL?
		if(strcmp(senderInfo->appID, appID) != 0)
		{
			ALP_FAIL_NON_FATAL_IF(TRUE, 
							  "re-registration with new app name: %s (expected %s).\n",
							  appID, senderInfo->appID);
			goto FreeStuffAndReturn;
			return;
		}
		
		// Is the app trying to re-register with a different PID?
		if(senderInfo->appPID != appPID)
		{
			ALP_FAIL_NON_FATAL_IF(TRUE, 
							  "re-registration with new PID: %d (expected %d).\n", 
							  appPID, senderInfo->appPID);
			goto FreeStuffAndReturn;
			return;
		}
		
		// Is the app trying to re-register with a different launchpad scheme?
		if(strcmp(senderInfo->launchpadScheme, scheme) != 0)
		{
			ALP_FAIL_NON_FATAL_IF(TRUE,
							  "re-registration with new launchhpad scheme: %s (was %s).\n", 
							  scheme, senderInfo->launchpadScheme);
			goto FreeStuffAndReturn;
			return;
		}
		
		// Is it trying to register with different "sticky" launch flags?
		// This is now an ignorable offense, since Transient may be specified in the manifest only.
		// Then, of course the client-side launch flags won't indicate its transiency.
		// And when the app is relaunched, of course it will send flags without that bit set.
		// The alternative to ignoring this problem would be to insert ALP_APP_TRANSIENT into
		// the launch args for any transient app being launched without it, but that seems like
		// more trouble than it's worth.  So, instead, we'll just fix up the bits.
				
		//ALP_FAIL_NON_FATAL_IF(((senderInfo->launchFlags & ALP_APP_LAUNCH_FLAGS_STICKY) != 
		//		(launchFlags & ALP_APP_LAUNCH_FLAGS_STICKY)), "Can't change sticky launch flags.\n");
		
		// 'Or' in any existing sticky launch flags to make sure they're not overwritten.
		launchFlags |= (senderInfo->launchFlags & ALP_APP_LAUNCH_FLAGS_STICKY);
		
		// Otherwise, re-registration looks okay.  Continue.
		
	}
	
	// Update jumplist to note that this app was launched.
	// We add it to the jump list if it's  marked "always jump-list", 
	// or if it's the new primary app.
	if( (senderInfo->serverFlags & ALP_APP_SERVER_FLAG_ALWAYS_JUMP_LIST) || 
		((launchFlags & ALP_APP_LAUNCH_FLAG_PRIMARY) && gCurrApp != senderInfo))
	{
		// ... but NOT if it's marked "never jump-list"
		if((senderInfo->serverFlags & ALP_APP_SERVER_FLAG_NO_JUMP_LIST) == 0)
			prv_jumplist_accumulate(appID);
	}
	
	// Note that you can't currently *clear* the transient flag by *not* 
	// registering as transient.
	// This keeps the manifest property working properly.
	
	// Update saved launchFlags in the appInfo structure.
	// This copy is necessary in the case where the app is re-registering with new launchflags.
	senderInfo->launchFlags = launchFlags;
	
	
	// This is where we handle app retirement.  alp_app_register() is effectively 
	// also the call used to "take over the foreground" (although apps wishing to do this
	// should use alp_app_launch() to relaunch themselves as primary).  This notifies us 
	// that the app is now the new primary app, and so we will retire the current one.
	
	// If it's being launched as the new primary app, and it's not the current one, 
	// then we need to retire the current app.
	if(launchFlags & ALP_APP_LAUNCH_FLAG_PRIMARY && gCurrApp != senderInfo)
	{
		// First, exit all the transient apps.
		prv_retire_transient_apps();
		
		// Then, retire the current primary app.
		prv_retire_app(gCurrApp);
		gCurrApp = senderInfo;
		
	}
	
		
	// Are there any more pending launches for this app?  If so, dispatch the next one!
	ALP_TM(TL(ALP_APP_SERVER_TRACE, "Registration complete, checking for more pending launches from %s.", 
		  appID));
	
	// If there are any more pending launches for this app waiting to be dispatched, do the next one.
	prv_do_next_pending_launch(appID);
	
	// Finally, if this app is moving to the background, and it *is* the current primary, 
	// then we need to find a new primary app!
	if(((launchFlags & ALP_APP_LAUNCH_FLAG_PRIMARY) == 0) && gCurrApp == senderInfo)
	{
		gCurrApp = NULL;
		prv_check_default_app();
	}
	
	
FreeStuffAndReturn:
	// Deallocate unpacked strings.
	if(appID) g_free(appID);
	if(scheme) g_free(scheme);
	
	return;
}

/*
	This routine is called whenever we receive a IPC request from a client.
	It was set up in the connect callback such that the cbData parameter 
	points to the RunningAppInfo structure for the client that sent the request.
*/
void prv_receive_data_cb(AlpConnection* connection, AlpMessage* message, gpointer cbData)
{
	RunningAppInfo *senderInfo = cbData; // DOLATER - remove cast once cbData type is fixed.
	gint msgID;
	AlpMessage *reply;
		
	msgID = alp_ipc_message_get_message_ID(message);
	
	ALP_TM(TL(ALP_APP_SERVER_TRACE, "Received message ID %d from %s.", msgID, senderInfo->appID));

	// Sanity check: make sure 'requesterInfo' is in our list of running apps.
	// This should be guaranteed, but we should at least check it on debug builds.
	// DOLATER: Make this a warning on release builds.
	ALP_FAIL_NON_FATAL_IF(g_list_find(gAppList, senderInfo) == NULL, 
					  " message %d received from unknown client.\n", msgID);
	
	
	// We received a message from a client.  Handle the request.
	switch(msgID)
	{
		case ALP_APP_MSG_ID_LAUNCH:
		
			prv_handle_launch(senderInfo, message);
			message = NULL; // prv_handle_launch may need to save the message, so we give up responsibility for it.
			break;
		
		// Handle registration message from newly launched app.
		case ALP_APP_MSG_ID_REGISTER: 
			
			prv_handle_app_register(senderInfo, message);
			break;
			
		case ALP_APP_MSG_ID_UNREGISTER: 
			
			// The message contains the app's result code.  We could unpack it... 
			// but we don't have anything to use the result code for anyways, so we don't bother.
			ALP_TM(TL(ALP_APP_SERVER_TRACE, "Unregister msg received from %s.", senderInfo->appID));
			
			// DOLATER: It might be nice to set a kill timer here, if there isn't one already running.
			
						
			senderInfo->serverFlags |= ALP_APP_SERVER_FLAG_EXIT_CLEAN;
			
			// Also stop sending messages to the client:
			senderInfo->launchFlags &= ~ALP_APP_LAUNCH_FLAG_DIAGNOSTIC;
			senderInfo->launchFlags &= ~ALP_APP_LAUNCH_FLAG_LAUNCHPAD;
			
			// DOLATER: Should we clear some server flags too, like WAITING_FOR_REPLY?
			
			break;
					
		case ALP_APP_MSG_ID_EXIT: // Handle an API call to request that this app exit.
			prv_handle_exit_request(message);
			break;
		
		case ALP_APP_MSG_ID_RETIRE: // Handle an API call to request that this app retire.
			prv_handle_retire_request(message);
			break;
		
			// Handle an API call to request a change to the app's bacngrounding setting.
		case ALP_APP_MSG_ID_BACKGROUNDING: 			
			prv_handle_set_backgrounding(senderInfo, message);
			break;
		
		case ALP_APP_MSG_ID_FLUSH: 			
			// All we need to do is send a reply to the sender:
			reply = alp_app_prv_make_flush_msg(message);
			alp_ipc_connection_send(senderInfo->connection, reply);
			alp_ipc_message_destroy(reply);
			
			break;
		
		case ALP_APP_MSG_ID_JUMP_LIST_REQ: 			
			prv_send_jumplist_reply(senderInfo, message);
			
			break;
		
		case ALP_APP_MSG_ID_CUR_APP: 			
			
			// All we need to do is send a brief reply to the sender, 
			// with the appID of the current primary app (if any).
			if(gCurrApp != NULL) reply = alp_app_prv_make_cur_app_msg(message, gCurrApp->appID);
			else reply = alp_app_prv_make_cur_app_msg(message, "");
			
			alp_ipc_connection_send(senderInfo->connection, reply);
			alp_ipc_message_destroy(reply);
			
			break;
		
		case ALP_APP_MSG_ID_WATCHER:
			prv_handle_watcher(senderInfo, message);
			break;
		
		case ALP_APP_MSG_ID_THEME_UPDATE:
			prv_handle_theme_update(senderInfo, message);
			break;
		
		case ALP_APP_MSG_ID_RESET:
			// Reboot the system!  Wheeee!
			system("reboot");
			break;
			
		default:
			// DOLATER: handle error
			ALP_FAIL_NON_FATAL_IF(TRUE, "message %d received, but not handled.\n", msgID);
			
			break;
	}
	
	// Deallocate message, if nobody else has taken care of it yet.
	if(message) alp_ipc_message_destroy(message);
	
	return;
}


/*
	For now, when a client disconnects all we do is remove the app from our list.
	We assume that the app will terminate itself gracefully, and are content to forget 
	about it's existence.
*/
void prv_client_disconnect_cb(AlpConnection* connection, gpointer cbData)
{
	GList *element;
	PendingLaunch *pendLaunch;
	
	ALP_TM(TL(ALP_APP_SERVER_TRACE, "Received disconnect from 0x%lx", (long unsigned int)connection));
		
	
	// The client disconnected, so remove the app from our list.
	// DOLATER: do we want to be paranoid and start up a watchdog timer here to kill the app if it doesn't exit?
	
	// DOLATER: Nice feature here would be to have the disconnect CB get the "receive data" CB 
	// data pointer too, since this is connection-specific. Then we could use g_list_remove() directly.
	// g_list_remove(gAppList, );
	
	element = gAppList;
	while(element != NULL)
	{
		if(((RunningAppInfo*)element->data)->connection == connection)
			break;
		
		element = g_list_next(element);
	}
	

	
	
	if(element != NULL)
	{
		RunningAppInfo *appInfo = element->data;
		
		ALP_TM(TL(ALP_APP_SERVER_TRACE, "Disconnect received from %s (%d)", appInfo->appID, appInfo->appPID));
		
		// Remove it from the list, since the app is no longer running.
		// This ensures the app is removed before we might (for example) try to broadcast a diagnostic message to it or something.
		gAppList = g_list_remove(gAppList, appInfo);
		
		// Remove any active watchers owned by this app:
		prv_watcher_remove_by_owner(appInfo);
		
		// If there are any references to it in the pending launch list (i.e., if it was 
		// waiting for a launch reply), then we need to clear those out:
		element = gPendingAppLaunches;
		while(element != NULL)
		{
			pendLaunch = (PendingLaunch*)element->data;
			if(pendLaunch->requester == appInfo)
				pendLaunch->requester = NULL;
		
			element = g_list_next(element);
		}
		
		// Clear any pending kill timers:
		prv_clear_kill_timer(appInfo, PRV_EXIT_TIMER);
		prv_clear_kill_timer(appInfo, PRV_RELAUNCH_TIMER);
		
		
		// Did the app properly unregister before exiting?
		// We could do something fancy if it didn't.
		if((appInfo->serverFlags & ALP_APP_SERVER_FLAG_EXIT_CLEAN) == 0)
		{
			ALP_TM(TL(ALP_APP_SERVER_TRACE, "Application %s exited unexpectedly.\n", appInfo->appID));
			prv_broadcast_diagnostic_msg(appInfo->appID, ALP_APP_DIAG_UNEXPECTED_EXIT);
			// DOLATER: What do we want to do about this?  Probably nothing for Vanilla.
		}
		else prv_broadcast_diagnostic_msg(appInfo->appID, ALP_APP_DIAG_CLEAN_EXIT);
		
		
		// Are there any pending launches for this app?  If so, dispatch the next one!
		ALP_TM(TL(ALP_APP_SERVER_TRACE, "App %s disconnected, checking for pending relaunches.", appInfo->appID));
		prv_do_next_pending_launch(appInfo->appID);
				
		prv_free_running_app_info(appInfo);
	}
	else
	{
		// DOLATER: handle the error
		// This should never happen, it means that a client disconnected, 
		// and we did not have an entry for it in our list of running applications.
		// That's a no-no, since it means an app was running, and we didn't know about it.
		ALP_FAIL_NON_FATAL_IF(TRUE, "disconnect message from unknown client.\n", NULL);
	}
	
	
	// If we no longer have a current app, then we need to launch another one...
	if(gCurrApp == NULL)
		prv_check_default_app();
	
	return;
}



/*******************************************************
	Startup & Takedown
 *******************************************************/

void prv_init_appserver(int argc, char *argv[])
{
	gboolean noDefaults = FALSE;
	int i;
	
	//
	// Make sure we are the only copy running.
	//
	if (alp_prv_acquirelock(APPSERVERLOCK) == FALSE)
	{
	//	ALP_FAIL_FATAL_IF( 1 == 2, "Whoops we were unable to aquire the lockfile.  Are two copies running?");
		ALP_TM(TL(ALP_APP_SERVER_TRACE, "Unable to aquire the lockfile.  Is appserverd already running?"));
		exit(0);
	}
	
	
	// Check if default app is specified in launch arguments:
	// Since the default app argument can be ignored if there is nothing 
	// following it, we don't need to search the last launch argument.	
	// This loop condition also guarantees that when we find it, there will be 
	// at least one more argument in the argv array, so we only have to make sure it's 
	// not another command line option.
	ALP_FAIL_FATAL_IF(gDefaultPrimary != NULL, "Globals improperly initialized.\n");
	for(i=1; i<(argc-1); i++) 
	{
		if(strcmp(ALP_APP_DEFAULT_PRIMARY, argv[i]) == 0 && 
			argv[i+1][0] != '-')
		{
			gDefaultPrimary = argv[i+1];
			break;
		}
	}
	
	
	// We need to keep track of the list of running apps.
	gAppList = NULL;
	
	gServerChannel = alp_ipc_channel_create(ALP_APP_SERVER_CHANNEL,
									ALP_IPC_CHANNEL_ACCESS_ALL);
	if(gServerChannel == NULL)
	{
		// DOLATER: handle error
		ALP_FAIL_FATAL_IF(TRUE, "Channel creation failed.\n");
		return;
	}
	
	// DOLATER: Tiny race condition here? A client could connect after the create, 
	// but before we have a connect callback installed?
	alp_ipc_channel_register_connect_callback(gServerChannel, prv_client_connect_cb, NULL);
	alp_ipc_channel_register_disconnect_callback(gServerChannel, prv_client_disconnect_cb, NULL);
	
	
	if(prv_search_args(ALP_APP_NO_DEFAULTS, argc, argv))
		noDefaults = TRUE;
	
	
	if(noDefaults || prv_search_args(ALP_APP_NO_DEFAULT_APP, argc, argv))
		gDisableDefaultApp = TRUE;
	else	gDisableDefaultApp = FALSE;
	
	gLaunchedPrimary = FALSE;
	
	return;
}

void prv_takedown_appserver(void)
{
	
	alp_ipc_channel_destroy(gServerChannel);
	
	// DOLATER: Probably should deallocate the app list too.
	// On the other hand, the app server is not meant to ever be taken down...
	
	return;
}

void prv_other_init(void)
{
	alp_bundle_init();
}

void prv_other_takedown(void)
{
	alp_bundle_fini();
}

int main(int argc, char *argv[])
{
	GMainLoop *mainLoop;
	
	printf("$BOOT_MODE = %s\n", getenv("BOOT_MODE"));
	
	// Move into background, allowing boot script to continue.
	// g_warning("Appserver old PID = %d, daemonizing...", getpid());
	if(prv_search_args(ALP_APP_NO_DAEMON, argc, argv) == FALSE)
		alp_daemon_start(FALSE, TRUE);
	// g_warning("Appserver daemon PID = %d.", getpid());
	ALP_TM(TL(ALP_APP_SERVER_TRACE, "App Server starting up..."));
	
	g_thread_init (NULL);
	
	mainLoop = g_main_loop_new(NULL, FALSE);
	prv_init_appserver(argc, argv);
	prv_other_init();
	
	// Allow boot sequence to continue.
	alp_daemon_release(0);
	
	// run main loop - now we will start handling requests.
	g_main_loop_run(mainLoop);
	
	
	ALP_TM(TL(ALP_APP_SERVER_TRACE, "App Server shutting down..."));

	prv_other_takedown();
	
	prv_takedown_appserver();
	
	return 0;
}


