/******************************************************************************
 *
 * Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the entire contents of this file.
 *
 * The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.
 *
 * Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
 * Rights Reserved.
 * 
 * Contributor(s): none.
 *
 *****************************************************************************/
/**
 * @file attnmgrd_alert.c
 *
 * @brief ALP attention manager process alert
 *
 *****************************************************************************/
#include <sys/types.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <assert.h>
#include <errno.h>
#include <string.h>
#include <signal.h>

#include <glib.h>

#include <hiker/types.h>
#include <hiker/appmgr.h>
#include <hiker/bundlemgr.h>
#include <hiker/notifymgr.h>

#include <hiker/attnmgr.h>
#include <hiker/prv/attnmgr_traces.h>
#include <hiker/prv/attnmgr_alert.h>
#include <hiker/prv/attnmgr_dml.h>
#include <hiker/prv/attnmgr_ipc.h>

#include "attnmgrd_prv.h"

#include "attn_dml_prv.h"

typedef enum {
    ALP_ATTN_CMD_POST = 0,
    ALP_ATTN_CMD_UPDATE,
    ALP_ATTN_CMD_DELETE
} AlpAttnCommand;

static alp_status_t
alp_prv_attn_convert_dml_error (alp_status_t dmlError)
{
    ALP_ATLX(DML, " %x", dmlError);
    switch (dmlError) {
    case ALP_STATUS_OK:
        return ALP_STATUS_OK;

    case ALP_STATUS_ATTN_DML_BAD_PARAM:
        return ALP_STATUS_ATTN_BAD_PARAM;

    case ALP_STATUS_ATTN_DML_NO_MEMORY:
        return ALP_STATUS_ATTN_NO_MEMORY;

    case ALP_STATUS_ATTN_DML_INVALID_LUID:
        return ALP_STATUS_ATTN_NOT_FOUND;

    case ALP_STATUS_ATTN_DML_NO_RECORDS:
    case ALP_STATUS_ATTN_DML_INTERNAL_ERROR:
    case ALP_STATUS_ATTN_DML_DATABASE_ERROR:
    case ALP_STATUS_ATTN_DML_CORRUPT:
    default:
        return ALP_STATUS_ATTN_INTERNAL_ERROR;
    }
}

static alp_status_t
alp_prv_attn_populate_property_defaults (AlpDmlItemH itemH)
{
    alp_status_t         err;
    AlpAttnBaseType     *baseP  = (AlpAttnBaseType *) itemH;
    AlpAttnAlertType    *alertP;
    AlpAttnPropListType *propListP;
    AlpAttnPropType     *propP;
    AlpAttnPropType     *listP;
    AlpDmlItemH          iconH = NULL;
    AlpAttnDmlItemType   itemType;
    char                *source;
    char                *name;
    char                *handle;
    uint32_t             nprops;

    if (itemH == NULL)
        return ALP_STATUS_ATTN_BAD_PARAM;

    itemType = baseP->itemType;
    switch (itemType) {
    case AlertItem:
        alertP = (AlpAttnAlertType *) itemH;
        source = alertP->source;
        name   = alertP->name;
        handle = alertP->handle;
        nprops = alertP->pcount;
        listP  = alertP->plist;
        break;

    case PropertyListItem:
        propListP = (AlpAttnPropListType *) itemH;
        source    = propListP->source;
        name      = propListP->name;
        handle    = propListP->handle;
        nprops    = propListP->count;
        listP     = propListP->list;
        break;

    case PropertyItem:
    default:
        ALP_ATLX(DAEMON, "bad or unsupported item type: %d",
                 (unsigned) baseP->itemType);
        return ALP_STATUS_ATTN_BAD_PARAM;
    }

    return ALP_STATUS_OK;
}

static void
alp_prv_attn_ui_launch_item (AlpAttnCommand cmd, AlpDmlItemH   itemH)
{
    alp_status_t         err;
    char                *ui_app_id;
    char                *alert_id = NULL;
    AlpAttnBaseType     *baseP  = (AlpAttnBaseType *) itemH;
    AlpAttnAlertType    *alertP;
    AlpAttnPropListType *propListP;
    AlpAttnPropType     *propP;
    AlpAttnPropType     *listP;
    char               **argv = NULL;
    int                  argc;
    pid_t                pid;
    uint32_t             nargs;
    uint32_t             nprops;
    uint32_t             i;

    /* validate args */
    if (itemH == NULL) {
        ALP_ATLX(DAEMON, "null (%x) bad parameter", (unsigned) itemH);
        return;    /* TODO: later will return an error code */
    }

    switch (baseP->itemType) {
    case AlertItem:
        alertP = (AlpAttnAlertType *) itemH;
        nprops = alertP->pcount;
        listP  = alertP->plist;
        break;

    case PropertyListItem:
        propListP = (AlpAttnPropListType *) itemH;
        nprops    = propListP->count;
        listP     = propListP->list;
        break;

    case PropertyItem:
    default:
        ALP_ATLX(DAEMON, "bad or unsupported item type: %d",
                 (unsigned) baseP->itemType);
        return;
    }

    /* fill in ui app id */

    ui_app_id = ALP_ATTNMGRD_APPID_ALERT;

    nargs = 5;
    argv = g_new(char *, nargs + 1);
    if (argv == NULL) {
        ALP_ATLX(DAEMON, "no memory ?? g_new(char *, %d)", nargs + 1);
        return;
    }

    ALP_ATLX(DAEMON, "building arg array for %d args [%d props]",
             nargs, nprops);
    i = 0;
    switch (cmd) {
    case ALP_ATTN_CMD_POST:
        ALP_ATLX(DAEMON, "post command: nargs %d", nargs);

        alert_id =  g_strdup_printf ("%s,%s,%s",
                                     alertP->source,
                                     alertP->name,
                                     alertP->handle);
        argv[i++] = "--post";
        argv[i++] = alert_id;

        ALP_ATLX(DAEMON, "post command done: %i args", i);
        break;

    case ALP_ATTN_CMD_UPDATE:
        ALP_ATLX(DAEMON, "update command: nargs %d", nargs);
        alert_id =  g_strdup_printf ("%s,%s,%s",
                                     propListP->source,
                                     propListP->name,
                                     propListP->handle);
        argv[i++] = "--update";
        argv[i++] = alert_id;

        ALP_ATLX(DAEMON, "post command done: %i args", i);
        break;

    case ALP_ATTN_CMD_DELETE:
        ALP_ATLX(DAEMON, "delete command: nargs %d", nargs);
        alert_id =  g_strdup_printf ("%s,%s,%s",
                                     alertP->source,
                                     alertP->name,
                                     alertP->handle);
        argv[i++] = "--delete";
        argv[i++] = alert_id;

        ALP_ATLX(DAEMON, "delete command done: %i args", i);
        break;

    default:
        ALP_ATLX(DAEMON, "unknown command: %d", cmd);
        goto cleanup;
    }    
    argv[i] = NULL;    /* for safety, terminate argv with NULL */
    argc = i;

    ALP_ATLX(DAEMON, "launching %s with argc %d nprops %d",
             ui_app_id, argc, nprops);
    /* launch ui */
    err = alp_app_launch (ui_app_id, argc, (char **)argv, &pid);

    ALP_ATLX(DAEMON, "launch %s argc %d nprops %d pid %d err %x",
             ui_app_id, argc, nprops, pid, err);

cleanup:
    g_free (alert_id);
    g_free (argv);
}

/*
 * Return the last component of the source app id to use as the feature name
 * Assumes that this is reasonable for all app ids ... may not be.
 */
static gchar *
GetFeatureName (AlpDmlItemH itemH)
{
    AlpAttnBaseType     *baseP  = (AlpAttnBaseType *) itemH;
    AlpAttnAlertType    *alertP;
    AlpAttnPropListType *propListP;
    gchar               *source;
    gchar               *feature;
    
    switch (baseP->itemType) {
    case AlertItem:
        alertP = (AlpAttnAlertType *) itemH;
        source = alertP->source;
        break;
    case PropertyListItem:
        propListP = (AlpAttnPropListType *) itemH;
        source    = propListP->source;
        break;
    case PropertyItem:
    default:
        return NULL;
    }

    feature = rindex(source, '.');
    if (feature) feature++;
    return feature;
}

static gchar *
GetAlertTypeName (AlpDmlItemH itemH)
{
    AlpAttnBaseType     *baseP  = (AlpAttnBaseType *) itemH;
    AlpAttnAlertType    *alertP;
    AlpAttnPropListType *propListP;
    gchar               *name;

    switch (baseP->itemType) {
    case AlertItem:
        alertP = (AlpAttnAlertType *) itemH;
        name   = alertP->name;
        break;
    case PropertyListItem:
        propListP = (AlpAttnPropListType *) itemH;
        name      = propListP->name;
        break;
    case PropertyItem:
    default:
        return NULL;
    }

    return name;
}

static gchar *
GetHandle (AlpDmlItemH itemH)
{
    AlpAttnBaseType     *baseP  = (AlpAttnBaseType *) itemH;
    AlpAttnAlertType    *alertP;
    AlpAttnPropListType *propListP;
    gchar               *handle;
    
    switch (baseP->itemType) {
    case AlertItem:
        alertP = (AlpAttnAlertType *) itemH;
        handle = alertP->handle;
        break;
    case PropertyListItem:
        propListP = (AlpAttnPropListType *) itemH;
        handle = propListP->handle;
        break;
    case PropertyItem:
    default:
        return NULL;
    }

    return handle;
}

static gchar *
GetAlertId (AlpDmlItemH itemH)
{
    AlpAttnBaseType     *baseP  = (AlpAttnBaseType *) itemH;
    AlpAttnAlertType    *alertP;
    AlpAttnPropListType *propListP;
    gchar               *source;
    gchar               *name;
    gchar               *handle;

    switch (baseP->itemType) {
    case AlertItem:
        alertP = (AlpAttnAlertType *) itemH;
        source = alertP->source;
        name   = alertP->name;
        handle = alertP->handle;
        break;
    case PropertyListItem:
        propListP = (AlpAttnPropListType *) itemH;
        source    = propListP->source;
        name      = propListP->name;
        handle    = propListP->handle;
        break;
    case PropertyItem:
    default:
        return NULL;
    }

    return g_strconcat (source, ",", name, ",", handle, NULL);
}

void
SendNotifyBroadcast (AlpDmlItemH itemH, gchar *type)
{
    alp_status_t  err;
    AlpNotifyEventAttnAlert *detail;
    gint          detail_len;
    gchar        *handle;
    gint          handle_len;
    gchar        *feature_name;
    gchar        *alert_type_name;
    gchar        *notify_type;

    /*
     * If alert type name does not begin with a '/', assume this is an old
     * style alert type name. In order to send a notification for this alert
     * type name, will need to concatenate the last component of the source
     * app id with the alert type name to form the desired notification type.
     */
    feature_name    = GetFeatureName (itemH);
    alert_type_name  = GetAlertTypeName (itemH);
    handle      = GetHandle (itemH);

    /* build notify type */
    if (alert_type_name && alert_type_name[0] == '/') {
        /* /alp/alert/alp/<alert type>/<action> */
        notify_type = g_strdup_printf ("%s%s/%s",
                                       ALP_NOTIFY_EVENT_ATTN_ALERT,
                                       &alert_type_name[1],
                                       type);
    } else {
        /* /alp/alert/alp/<feature>/<alert type>/<action> */
        notify_type = g_strdup_printf ("%salp/%s/%s/%s",
                                       ALP_NOTIFY_EVENT_ATTN_ALERT,
                                       feature_name,
                                       alert_type_name,
                                       type);
    }

    /* build notification detail */
    detail_len = sizeof(AlpNotifyEventAttnAlert);
    ALP_ATLX(APPS,"emtpy detail len = %d", detail_len);
    handle_len = strlen(handle);
    detail_len += (handle_len + 1);
    ALP_ATLX(APPS,"detail with handle len = %d", detail_len);
    detail = g_malloc0(detail_len);
    if (detail) {
        strncpy (detail->handle, handle, handle_len);
        detail->handle[handle_len] = '\0';
    }
    err = alp_notify_broadcast (ALP_ATTNMGRD_NOTIFYID_SELF,
                                notify_type,
                                detail, detail_len, ALP_NOTIFY_QUEUE_DEFAULT, 0);
    g_free (detail);
    g_free (notify_type);
}


static void
LaunchAlertDialog (void)
{
    alp_status_t         err;
    static pid_t         pid = 0;
    const char          *argv[1] = { NULL };

    if (pid) {
        /* make sure display alert process is still running */
        if (kill (pid, 0) < 0) {
            ALP_ATLX(DAEMON, "attn ui not running, restart. [errno=%d,%s]",
                     errno, strerror (errno));
            pid = 0;
        }
    }
    if (pid == 0) { /* start display alert process */
        /* launch ui */
        err = alp_app_launch (ALP_ATTNMGRD_APPID_ALERT,
                              0, (char **)argv,
                              &pid);
        ALP_ATLX(DAEMON, "attn ui started pid %d", pid);
    }
    return;
}

void
alp_attnmgrd_process_post_alert (AlpConnection *connection,
                                 AlpMessage    *message,
                                 AlpDmlItemH    itemH,
                                 gint           procId,
                                 gint           userId,
                                 gint           groupId)
{
    alp_status_t    err;
    AlpAttnId       id;
    AlpMessage     *reply;
    const char     *source;
    const char     *name;
    const char     *handle;
    uint32_t        priority;
    uint32_t        display_priority;
    uint32_t        attributes;
    int32_t         pending = true;


    /* validate priority level and set attributes based on it */
    err = ALP_STATUS_OK;
    priority = alp_attn_dml_item_get_int32 (itemH, Priority);
    switch (priority) {
    case 1:
        attributes =
            ALP_ATTN_ATTR_TYPE_ATTN |
            ALP_ATTN_ATTR_DISPLAY_ALERT |
            ALP_ATTN_ATTR_DISPLAY_SECONDARY |
            ALP_ATTN_ATTR_DELETE_ON_BOOT |
            ALP_ATTN_ATTR_SOUND |
            ALP_ATTN_ATTR_VIBRATE;
        display_priority = 1;
        break;

    case 2:
        attributes =
            ALP_ATTN_ATTR_TYPE_ATTN |
            ALP_ATTN_ATTR_DISPLAY_ALERT |
            ALP_ATTN_ATTR_DISPLAY_LIST |
            ALP_ATTN_ATTR_DISPLAY_SECONDARY |
            ALP_ATTN_ATTR_NAVIGATE_AWAY |
            ALP_ATTN_ATTR_SOUND |
            ALP_ATTN_ATTR_VIBRATE;
        display_priority = 3;
        break;

    case 3:    /* default prio for no dialog, list only */
        attributes =
            ALP_ATTN_ATTR_TYPE_ATTN |
            ALP_ATTN_ATTR_DISPLAY_LIST |
            ALP_ATTN_ATTR_SOUND |
            ALP_ATTN_ATTR_VIBRATE;
        display_priority = 3;
        break;

    case 4:    /* ?? */
        attributes =
            ALP_ATTN_ATTR_TYPE_ATTN |
            ALP_ATTN_ATTR_DISPLAY_ALERT |
            ALP_ATTN_ATTR_DELETE_ON_BOOT |
            ALP_ATTN_ATTR_SOUND |
            ALP_ATTN_ATTR_VIBRATE;
        display_priority = 4;
        break;

    case 5:    /* default prio for no dialog, list only - state count alerts */
        attributes =
            ALP_ATTN_ATTR_TYPE_ATTN |
            ALP_ATTN_ATTR_DISPLAY_LIST |
            ALP_ATTN_ATTR_SOUND |
            ALP_ATTN_ATTR_VIBRATE;
        display_priority = 5;
        break;

    case 17:   /* default progress alert level */
        attributes =
#if 1
            ALP_ATTN_ATTR_TYPE_ATTN |
#else
            ALP_ATTN_ATTR_TYPE_PROGRESS |
#endif
            ALP_ATTN_ATTR_DISPLAY_ALERT |
            ALP_ATTN_ATTR_DISPLAY_LIST |
            ALP_ATTN_ATTR_SOUND |
            ALP_ATTN_ATTR_VIBRATE;
        display_priority = 2;
        break;

    case 18:
        attributes =
#if 1
            ALP_ATTN_ATTR_TYPE_ATTN |
#else
            ALP_ATTN_ATTR_TYPE_ASYNC |
#endif
            ALP_ATTN_ATTR_DISPLAY_ALERT |
            ALP_ATTN_ATTR_SOUND |
            ALP_ATTN_ATTR_VIBRATE;
        display_priority = 2;
        break;

    case 254: /* null test priority - get stored in db */
        attributes =
            ALP_ATTN_ATTR_TYPE_TEST; /* don't display UI */
        display_priority = 1;
        break;
            
    case 255: /* invalid priority for bad param test */
    default:
        err = ALP_STATUS_ATTN_BAD_PARAM;
        break;
    }
    if (err == ALP_STATUS_OK) {
        (void) alp_attn_dml_item_set_uint32 (itemH, Attributes, &attributes);
        (void) alp_attn_dml_item_set_uint32 (itemH, Priority, &display_priority);
    }

    /* move this stuff to the client side of api for processing */
    (void) alp_prv_attn_populate_property_defaults (itemH);
    (void) alp_attn_dml_item_set_int32 (itemH, Pending, &pending);

    if (err == ALP_STATUS_OK) {
        err = alp_attn_dml_transaction_begin (gAlpAttnMgrDDmlH, true);
        if (err == ALP_STATUS_OK) {
            err = alp_attn_dml_insert (gAlpAttnMgrDDmlH, itemH, &id);
            ALP_ATLX(DAEMON, "insert of record returned id %d err %x",
                     id, err);
            if (err) {
                if (err == ALP_STATUS_ATTN_DML_EXISTS) {
                    AlpAttnAlertType *alertP = (AlpAttnAlertType *) itemH;
                    ALP_ATLX(DAEMON,
                             "alert %s,%s,%s exists, trying to update instead",
                             alertP->source,
                             alertP->name,
                             alertP->handle);
                    /*
                     * if alert has already been posted to this source,name,handle
                     * update it with the newly posted one.
                     */
                    err = alp_attn_dml_replace (gAlpAttnMgrDDmlH, itemH);
                    if (err == ALP_STATUS_OK) {
                        id = alp_attn_dml_item_get_uint32 (itemH, AlertLuid);
                        err = alp_attn_dml_transaction_end (gAlpAttnMgrDDmlH);
                    } else {
                        alp_attn_dml_transaction_rollback (gAlpAttnMgrDDmlH);
                    }
                } else {
                    alp_attn_dml_transaction_rollback (gAlpAttnMgrDDmlH);
                }
            } else {
                err = alp_attn_dml_transaction_end (gAlpAttnMgrDDmlH);
            }
        }
        err = alp_prv_attn_convert_dml_error (err);
    }

    /* need to return id back to caller */
    reply = alp_attn_ipc_pack_alert_post_reply (message, err, id);
    if (reply)
        alp_ipc_connection_send (connection, reply);

    if (err == ALP_STATUS_OK) {
        if (attributes & ALP_ATTN_ATTR_DISPLAY_ALERT) {
            if (gAlpAttnMgrDTest) {
                AlpAttnAlertType *alertP = (AlpAttnAlertType *) itemH;
                AlpAttnPropType  *propP;
                ALP_ATLX(IPC, "test mode: search for callback channel");

                propP = alertP->plist;
                while (propP) {
                    if (strcmp(propP->name,
                               ALP_ATTN_CALLBACK_CHANNEL_PROPERTY) == 0) {
                        ALP_ATLX(IPC,
                                 "callback channel found: sending "
                                 "callback to %s with %s|%s|%s %d",
                                 propP->value,
                                 alertP->source,
                                 alertP->name,
                                 alertP->handle,
                                 0);
                        alp_attn_ipc_send_callback (propP->value,
                                                    alertP->source,
                                                    alertP->name,
                                                    alertP->handle,
                                                    0,
                                                    NULL);
                        break;
                    }
                    propP = propP->next;
                }
            }
            /* launch ui app if necessary */
            LaunchAlertDialog();
        }
        /* broadcast db change notification */
        SendNotifyBroadcast (itemH, "post");
    }
}

void
alp_attnmgrd_process_update_alert (AlpConnection *connection,
                                   AlpMessage    *message,
                                   AlpDmlItemH    itemH)
{
    alp_status_t    err;
    AlpMessage     *reply;
    int32_t         pending = true;

    (void) alp_attn_dml_item_set_int32 (itemH, Pending, &pending);
    err = alp_attn_dml_transaction_begin (gAlpAttnMgrDDmlH, true);
    if (err == ALP_STATUS_OK) {
        err = alp_attn_dml_update (gAlpAttnMgrDDmlH, itemH);
        ALP_ATLX(DAEMON,"alert id %d update returned err %x", 0 /*id*/, err);
        if (err) {
            alp_attn_dml_transaction_rollback (gAlpAttnMgrDDmlH);
        } else {
            err = alp_attn_dml_transaction_end (gAlpAttnMgrDDmlH);
#if 0
            if (attributes & ALP_ATTN_ATTR_DISPLAY_ALERT) {
                LaunchAlertDialog();
            }
#endif
            SendNotifyBroadcast (itemH, "update");
        }
    }
    err = alp_prv_attn_convert_dml_error (err);

    /* need to return id back to caller */
    reply = alp_attn_ipc_pack_alert_update_reply (message, err);
    if (reply) {
        err = alp_ipc_connection_send (connection, reply);
    } else {
        err = ALP_STATUS_ATTN_IPC_INTERNAL;
    }
}

void
alp_attnmgrd_process_delete_alert (AlpConnection *connection,
                                   AlpMessage    *message,
                                   AlpDmlItemH    itemH)
{
    alp_status_t    err;
    AlpMessage     *reply;

    err = alp_attn_dml_transaction_begin (gAlpAttnMgrDDmlH, true);
    ALP_ATLX(DAEMON,"transaction begin err %x", err);
    if (err == ALP_STATUS_OK) {
        err = alp_attn_dml_delete (gAlpAttnMgrDDmlH, itemH);
        if (err) {
            ALP_ATLX(DAEMON,"delete of alert err %x", err);
            alp_attn_dml_transaction_rollback (gAlpAttnMgrDDmlH);
        } else {
            err = alp_attn_dml_transaction_end (gAlpAttnMgrDDmlH);
            ALP_ATLX(DAEMON,"transaction end err %x", err);
            SendNotifyBroadcast (itemH, "delete");
        }
    }
    err = alp_prv_attn_convert_dml_error (err);

    /* need to return id back to caller */
    reply = alp_attn_ipc_pack_alert_delete_reply (message, err);
    if (reply) {
        err = alp_ipc_connection_send (connection, reply);
    } else {
        err = ALP_STATUS_ATTN_IPC_INTERNAL;
    }
}
