/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/**
 * @file attnmgrd_database.c
 *
 * @brief ALP attention manager database
 *
 *****************************************************************************/

#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <pwd.h>
#include <grp.h>

#include <glib.h>
#include <glib/gstdio.h>

#include <hiker/types.h>

#include <hiker/prv/attnmgr_dml.h>
#include <hiker/prv/attnmgr_traces.h>

#include "attn_dml_prv.h"

#ifndef ALP_ATTN_TRACE_SQL
#define ALP_ATTN_TRACE_SQL   true   /* set true to trace SQL for debug */
#endif

#if 0         /* temporary work around for BUG152691 */
#define kAttnDBDirectoryMode (S_ISGID|S_ISVTX|S_IRWXU|S_IRWXG)
#define kAttnDBFileMode (S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP)
#define kAttnDBOwner "alert"
#define kAttnDBGroup "alert"
#else
#define kAttnDBDirectoryMode (S_IRWXU|S_IRWXG|S_IRWXO)
#define kAttnDBFileMode (S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH)
#define kAttnDBOwner "root"
#define kAttnDBGroup "daemon"
#endif

AlpDmlH gAlpAttnMgrDDmlH;

/* initialize database (i.e. create it) */

/*
 * not sure how much of this is necessary since a good portion will be done in
 * attnmgr_dml.
 */

static void
prv_attn_sql_trace (void *data, const char *sql)
{
    ALP_ATLX(DML,"<%s>", sql);
}

alp_status_t
alp_attnmgrd_db_setup (const char *directory, const char *filename, bool debug)
{
    alp_status_t err;
    struct stat  file_stat;
    const gchar *db_pathname;
    bool         has_retried_initialization = false;
    struct passwd *pw;
    struct group  *grp;
    uid_t        user_id = 0;
    gid_t        group_id = 0;

    pw = getpwnam (kAttnDBOwner);
    if (pw) {
        user_id = pw->pw_uid;
    }

    grp = getgrnam (kAttnDBGroup);
    if (grp) {
        group_id = grp->gr_gid;
    }

    if (directory == NULL) {
        directory = kAttnDefDBDirectory;
        ALP_ATLX(DAEMON, "defaulting directory to %s", directory);
    }

    if (filename  == NULL) {
        filename  = kAttnDefDBFileName;
        ALP_ATLX(DAEMON, "defaulting database filename to %s", filename);
    }

    /* verify database directory exists with proper permissions */
    if (!g_file_test (directory, G_FILE_TEST_IS_DIR)) {
        ALP_ATLX(DAEMON,
                 "directory %s does not exit or is not directory",
                 directory);

        if (g_file_test (directory, G_FILE_TEST_EXISTS)) {
            /* remove case non-directory is present */
            if (g_remove (directory) < 0) {
                ALP_ATLX(DAEMON, "unable to remove %s: %d %s",
                         directory, errno, strerror(errno));
                return ALP_STATUS_ATTN_DML_INTERNAL_ERROR;
            } else {
                ALP_ATLX(DAEMON, "%s removed", directory);
            }
        }

        /* create directory */
        if (g_mkdir_with_parents (directory, kAttnDBDirectoryMode) < 0) {
            /* unable to create directory */
            ALP_ATLX (DAEMON, "unable to create %s", directory);
            return ALP_STATUS_ATTN_DML_INTERNAL_ERROR;
        } else {
            if (g_stat (directory, &file_stat) < 0) {
                ALP_ATLX(DAEMON, "unable to stat %s: %d %s",
                         errno, strerror(errno));
                return ALP_STATUS_ATTN_DML_INTERNAL_ERROR;
            } else {
                ALP_ATLX(DAEMON, "%s created with mode %o user %d group %d",
                         directory,
                         file_stat.st_mode,
                         file_stat.st_uid,
                         file_stat.st_gid);
            }
        }
    }
    chown (directory, user_id, group_id);
    chmod (directory, kAttnDBDirectoryMode);

retry_init:
    alp_attn_dml_set_filename (directory, filename);
    db_pathname = alp_attn_dml_get_filename();

    /* initialize database - verify its existence and integrity */
    err = alp_prv_attn_dml_init_database (db_pathname);
    if (err) {
        ALP_ATLX(DAEMON,
                 "unable to initialize database %s: err %x",
                 db_pathname, err);
        return err;
    }
    chown (db_pathname, user_id, group_id);
    chmod (db_pathname, kAttnDBFileMode);

    /* open database */
    err = alp_attn_dml_open (&gAlpAttnMgrDDmlH);
    if (err) {
        ALP_ATLX(DAEMON, "failed to open database: err %x", err);
        return err;
    }

    /* check integrity of database */
    err = alp_attn_dml_check_database (gAlpAttnMgrDDmlH);
    if (err) {
        ALP_ATLX(DAEMON,"integrity check failed - recreating db now: err %x",
                 err);
        
        if (!has_retried_initialization) {
            g_remove (db_pathname);
            has_retried_initialization = true;
            alp_attn_dml_close (gAlpAttnMgrDDmlH);
            goto retry_init;
        }

        return err;
    }
#if ALP_ATTN_TRACE_SQL == true
    /* setup sqlite trace callback */
    if (debug) {
        sqlite3_trace ((sqlite3 *)gAlpAttnMgrDDmlH, prv_attn_sql_trace, NULL);
    }
#endif /* ALP_ATTN_TRACE_SQL == true */
    return ALP_STATUS_OK;
}

void
alp_attnmgrd_db_teardown (void)
{

}
