/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/**
 * @file attnmgrd_ipc.c
 *
 * @brief ALP attention manager IPC handler
 *
 *****************************************************************************/

/*
 * okay - there will be an attnmgr_ipc or maybe just an attnmgr library
 * which will contain client API to access (get,getnext,clear,clearall)
 * the database.
 * In addition there will be a attn library which provides client API to
 * post/update/delete to the database.
 * So what goes on here should be the server side of those API calls.
 * Question - where will all this come together ? attnmgrd_nexus.c obviously!
 */
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include <glib.h>

#include <hiker/types.h>
#include <hiker/appmgr.h>
#include <hiker/bundlemgr.h>

#include <hiker/attnmgr.h>
#include <hiker/prv/attnmgr_traces.h>
#include <hiker/prv/attnmgr_dml.h>
#include <hiker/prv/attnmgr_ipc.h>

#include "attnmgrd_prv.h"

#include "attn_dml_prv.h"

static AlpChannel *gServerChannel;
static GList      *gClientList;

typedef struct _AttnClientInfo AttnClientInfo;
struct _AttnClientInfo {
    gint             pid;
	AlpConnection	*connection;
	AlpMessage      *replyMessage; // Contains the message to reply to
                                   // if ALP_APP_SERVER_FLAG_WAITING_FOR_REPLY is set.
};

void
prv_free_client_info (AttnClientInfo *info)
{
	if(info && info->replyMessage) {
		alp_ipc_message_destroy(info->replyMessage);
    }
	
	g_free (info);
	
	return;
}

void
prv_receive_data_cb (AlpConnection *connection, AlpMessage *message, gpointer cbData)
{
	AttnClientInfo *senderInfo = cbData;
    AlpDmlItemH     itemH;
    AlpAttnId       id;
	AlpMessage     *reply;
	AlpMessage     *flushReply;
	gint            msgId;
    alp_status_t    err;
    gint            procId;
    gint            userId;
    gint            groupId;

	msgId   = alp_ipc_message_get_message_ID (message);
    procId  = alp_ipc_connection_get_peer_pid (connection);
    userId  = alp_ipc_connection_get_peer_uid (connection);
    groupId = alp_ipc_connection_get_peer_gid (connection);

	ALP_ATLX (IPC_SRV,
             "Received message ID %d from pid %d uid %d gid %d",
              msgId, procId, userId, groupId);

    /*
     * Sanity check: make sure 'requesterInfo' is in our list of clients.
	 * This should be guaranteed, but we should at least check it on debug builds.
     * DOLATER: Make this a warning on release builds.
     */
	ALP_FAIL_NON_FATAL_IF(g_list_find(gClientList, senderInfo) == NULL, 
                          " message %d received from unknown client (pid %d).\n",
                          msgId, senderInfo->pid);
	
	
	// We received a message from a client.  Handle the request.
	switch(msgId) {
    case ALP_ATTN_MSG_ID_STATUS:
        ALP_ATLX (IPC_SRV, "received status request");
        /*
         * compose status message
         * eventually this will get real data and also take any included string
         * in the request and echo back the first 80 characters of it.
         */
        gchar *srv_status =
            "server running with 4 alerts total, 2 pending notification";
        err = ALP_STATUS_OK;
        reply = alp_attn_ipc_pack_status_reply (message, err, srv_status);
        if (reply) {
            err = alp_ipc_connection_send (senderInfo->connection, reply);
        } else {
            err = ALP_STATUS_ATTN_IPC_INTERNAL;
        }
        break;

    case ALP_ATTN_MSG_ID_POST:
        ALP_ATLX (IPC_SRV, "received alert post IPC");
        err = alp_attn_ipc_unpack_alert_post_msg (message, &itemH);
        ALP_ATLX (IPC_SRV, "unpack post returned err %x", err);
        if (err == ALP_STATUS_OK) {
            alp_attnmgrd_process_post_alert (senderInfo->connection,
                                             message,
                                             itemH, procId, userId, groupId );
        }
        break;

    case ALP_ATTN_MSG_ID_UPDATE: 
        ALP_ATLX (IPC_SRV, "received alert update IPC");
        err = alp_attn_ipc_unpack_alert_update_msg (message, &itemH);
        ALP_ATLX (IPC_SRV, "unpack update returned err %x", err);
        if (err == ALP_STATUS_OK) {
            alp_attnmgrd_process_update_alert (senderInfo->connection,
                                               message,
                                               itemH);
        }
        break;

    case ALP_ATTN_MSG_ID_DELETE: 
        ALP_ATLX (IPC_SRV, "received delete event IPC");
        err = alp_attn_ipc_unpack_alert_delete_msg (message, &itemH);
        if (err == ALP_STATUS_OK) {
            alp_attnmgrd_process_delete_alert (senderInfo->connection,
                                               message,
                                               itemH);
        } else {
            ALP_ATLX (IPC_SRV, "unpack delete alert returned err %x", err);
        }
        break;

    default:
        // DOLATER: handle error
        ALP_FAIL_NON_FATAL_IF (TRUE, "message %d received, but not handled.\n", msgId);
        break;
	}
	
	// Deallocate message, if nobody else has taken care of it yet.
	if (message) {
        alp_ipc_message_destroy (message);
    }
	
	return;
}

static void
prv_client_connect_cb (AlpConnection *connection, gpointer cbData)
{
	AttnClientInfo *clientInfo;
	alp_status_t err;
	
	ALP_ATLX(IPC_SRV, "Received connect from 0x%lx",
              (long unsigned int)connection);
	
	// We have a new client, so allocate a new RunningAppInfo structure, 
	// add it to the list, and set it as the data pointer to be passed to the receive callback.
	clientInfo  = g_new0(AttnClientInfo, 1);

	gClientList = g_list_prepend(gClientList, clientInfo);
	
	// Fill in the client's connection object now.  Until the launchpad registers the app
	// via alp_app_setup(), this structure will be mostly empty, and we will not know which
	// application it represents.
	clientInfo->connection = connection;
	clientInfo->pid        = alp_ipc_connection_get_peer_pid (connection);

	// DOLATER: The connect operation (incl. callback) is synchronous, right?  
	// So the client can't send us data before I have a chance to install the callback?
	
	// Register a data receive callback, and pass the appInfo pointer as the data 
	// so we can easily get to all the sender info when we receive a message.
	err = alp_ipc_connection_register_receive_callback (connection,
                                                        prv_receive_data_cb,
                                                        (gpointer*)clientInfo);
	 
	ALP_FAIL_NON_FATAL_IF(err != ALP_STATUS_OK, 
					  "register_receive_callback() failed.\n", NULL);
	
	return;
}

void
prv_client_disconnect_cb (AlpConnection *connection, gpointer cbData)
{
	GList *element;
	
	ALP_ATLX(IPC_SRV, "Received disconnect from 0x%lx",
             (long unsigned int)connection);
		
	/*
	 * The client disconnected, so remove the app from our list.
     * DOLATER: do we want to be paranoid and start up a watchdog timer
     * here to kill the app if it doesn't exit?
	 * DOLATER: Nice feature here would be to have the disconnect CB get
	 * the "receive data" CB data pointer too, since this is
	 * connection-specific. Then we could use g_list_remove() directly.
	 */

    //g_list_remove(gAppList, );
	
	element = gClientList;
	while (element != NULL) {
		if (((AttnClientInfo*)element->data)->connection == connection)
			break;
		
		element = g_list_next(element);
	}
	
	if (element != NULL) {
		AttnClientInfo *clientInfo = element->data;
		
		ALP_ATLX(IPC_SRV, "Disconnect received from pid %d", clientInfo->pid);
		
		/*
		 * Remove it from the list, since the app is no longer running.
		 * This ensures the app is removed before we might (for example)
		 * try to broadcast a diagnostic message to it or something.
		 */
		gClientList = g_list_remove(gClientList, clientInfo);
		
		prv_free_client_info (clientInfo);
	} else {
        /*
		 * DOLATER: handle the error
		 * This should never happen, it means that a client disconnected,
		 * and we did not have an entry for it in our list of running applications.
		 * That's a no-no, since it means an app was running, and we didn't know about it.
		 */
		ALP_FAIL_NON_FATAL_IF(TRUE, "disconnect message from unknown client (pid %d).\n",
                              alp_ipc_connection_get_peer_pid (connection));
	}

	return;
}

void
alp_attnmgrd_ipc_setup (const char *channel)
{
    alp_status_t err;
    gClientList = NULL;

    err = alp_attn_ipc_init (channel);
    if (err != ALP_STATUS_OK) {
		ALP_FAIL_FATAL_IF(TRUE, "Channel assignment failed.\n");
        return;
    }
    gServerChannel = alp_ipc_channel_create(alp_attn_ipc_get_channel(),
                                            ALP_IPC_CHANNEL_ACCESS_ALL);	
	if(gServerChannel == NULL) {
		// DOLATER: handle error
		ALP_FAIL_FATAL_IF(TRUE, "Channel creation failed.\n", NULL);
		return;
	}
	
	// DOLATER: Tiny race condition here? A client could connect after the create, 
	// but before we have a connect callback installed?
	alp_ipc_channel_register_connect_callback(gServerChannel,
                                              prv_client_connect_cb,
                                              NULL);
	alp_ipc_channel_register_disconnect_callback(gServerChannel,
                                                 prv_client_disconnect_cb,
                                                 NULL);
}

void
alp_attnmgrd_ipc_teardown (void)
{
    AlpChannel *server;

    server = gServerChannel;
    gServerChannel = NULL;

    /* close all open connections */

    /* destroy server channel */
    alp_ipc_channel_destroy (server);

    alp_attn_ipc_init(NULL);    /* reset channel to NULL */
}
