/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/**
 * @file attnmgrd_main.c
 *
 * @brief ALP attention manager daemon implementation
 *
 *****************************************************************************/

#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <glib.h>
#include <libgen.h>

#include <hiker/types.h>
#include <hiker/pdk/appmgr_pdk.h>
#include <hiker/bundlemgr.h>
#include <hiker/notifymgr.h>
#include <hiker/pdk/sysutils_pdk.h>

#include <hiker/prv/attnmgr_traces.h>

#include "attnmgrd_prv.h"
/*
 * usage:
 *    --debug, -d:     enable debug output
 *    --version, -v:   print version number only
 *    --database, -D:  attention database to use with this attention manager daemon
 *    --channel, -c:   channel on which clients may connect
 *    --no-daemon, -n: don't detach to become a daemon
 *    --pidfile, -P:   name of file to store process ID
 *    --test:          don't applaunch attn UI
 */
#include <stdlib.h>
#include <getopt.h>

#define ALP_ATTNMGRD_VERSION "attnmgrd 0.1"

bool gAlpAttnMgrDDebug = false;
bool gAlpAttnMgrDTest  = false;
bool gAlpAttnMgrDAbort = false;

static void
usage(char *progname)
{
    g_print("usage: %s {options}\n"
            "    --help, -h                print this usage message\n"
            "    --debug, -d               run daemon in debug mode\n "
            "    --test                    run daemon in test mode\n"
            "    --version, -v             print version of daemon\n"
            "    --database, -D  filename  set database to filename\n"
            "    --db-filename   filename  set the filename portion of dbname\n"
            "    --db-directory  dirname   set the dirname portion of dbname\n"
            "    --channel, -c   channel   set ipc channel to channel\n"
            "    --no-daemon, -n           don't run in daemon mode\n"
            "    --pidfile, -P   pid-file  set pid filename to pid-file\n",
            progname);
}

int
main (int argc, char * argv[])
{
    alp_status_t err;
    GMainLoop  *loop;
    int         c;
    int         digit_optind = 0;
    const char *db_directory = NULL; /* free after use */
    const char *db_filename  = NULL;  /* free after use */
    const char *ipc_channel  = NULL;
    const char *pid_filename = ALP_ATTNMGRD_DEF_PIDFILE;
    bool        do_detach    = true;
    pid_t       my_pid;
    FILE       *fp;
    AlpAppContext *context;

    /* replace this with glib arg parser at some point */
    /* process loqcal args */
    while (1) {
        int this_option_optind = optind ? optind : 1;
        int option_index = 0;
        static struct option long_options[] = {
            {"help", 0, 0, 0},
            {"debug", 0, 0, 0},
            {"version", 0, 0, 0},
            {"database", 1, 0, 0},
            {"db-filename", 1, 0, 0},
            {"db-directory", 1, 0, 0},
            {"channel", 1, 0, 0},
            {"no-daemon", 0, 0, 0},
            {"pidfile", 1, 0, 0},
            {"test", 0, 0, 0},
            {"abort", 0, 0, 0},
            {0, 0, 0, 0}
        };

        c = getopt_long (argc, argv, "hdvc:nP:",
                         long_options, &option_index);
        if (c == -1)
            break;

        switch (c) {
        case 0:
            switch (option_index) {
            case 0:             /* help */
                usage(argv[0]);
                exit(0);
                break;

            case 1:
                gAlpAttnMgrDDebug = true;
                ALP_ATLX(DAEMON, "enabling debug mode");
                break;

            case 2:
                g_print ("%s\n", ALP_ATTNMGRD_VERSION);
                exit(0);
                break;

            case 3:
                ALP_ATLX(DAEMON, "database name [%s]", optarg);
                if (optarg) {
                    db_directory = g_path_get_dirname(optarg);
                    db_filename  = g_path_get_basename(optarg);
                }
                ALP_ATLX(DAEMON, "setting database name to %s/%s",
                         db_directory, db_filename);
                break;

            case 4:
                ALP_ATLX(DAEMON, "database filename [%s]", optarg);
                if (optarg) {
                    db_filename = g_strdup(optarg);
                }
                ALP_ATLX(DAEMON, "setting database name to %s",
                         db_filename);
                break;

            case 5:
                ALP_ATLX(DAEMON, "database directoy [%s]", optarg);
                if (optarg) {
                    int dirname_len = strlen(optarg);
                    if (dirname_len > 1 &&
                        optarg[dirname_len - 1] == '/') {
                        /* strip trailing directory separator */
                        optarg[dirname_len - 1] =  '\0';
                    }
                    db_directory = optarg;
                }
                ALP_ATLX(DAEMON, "setting database directory to %s",
                         db_directory);
                break;

            case 6:
                if (optarg) {
                    ipc_channel = optarg;
                }
                ALP_ATLX(DAEMON, "setting ipc channel name to %s",ipc_channel);
                break;

            case 7:
                /* no detach flag here */
                ALP_ATLX(DAEMON, "don't run in daemon mode");
                do_detach = false;
                break;

            case 8:
                /* pid file argument */
                if (optarg)
                    pid_filename = optarg;
                ALP_ATLX(DAEMON, "pid-file set to %s", pid_filename);
                break;

            case 9:
                /* test mode */
                gAlpAttnMgrDTest = true;
                ALP_ATLX(DAEMON, "test mode flag set to %s",
                         gAlpAttnMgrDTest ? "true" : "false");
                break;

            case 10:
                /* abort test */
                gAlpAttnMgrDAbort = true;
                ALP_ATLX(DAEMON, "abort test flag set to %s",
                         gAlpAttnMgrDAbort ? "true" : "false");
                break;

            default:
                ALP_ATLX(DAEMON, "unknow option index %d", option_index);
                break;
            }
            break;

        case 'h':
            usage(argv[0]);
            exit(0);
            break;

        case 'd':
            gAlpAttnMgrDDebug = true;
            ALP_ATLX(DAEMON, "enabling debug mode");
            break;

        case 'v':
            g_print ("%s\n", ALP_ATTNMGRD_VERSION);
            exit(0);
            break;

        case 'c':
            if (optarg) {
                ipc_channel = optarg;
            }
            ALP_ATLX(DAEMON, "setting ipc channel name to %s",ipc_channel);
            break;

        case 'n':
            ALP_ATLX(DAEMON, "don't run in daemon mode");
            do_detach = false;
            break;

        case 'P':
            /* pid file argument */
            if (optarg)
                pid_filename = optarg;
            ALP_ATLX(DAEMON, "pid-file set to %s", pid_filename);
            break;

        case '?':
            usage (argv[0]);
            exit(1);
            break;
        }

    }

    if (optind < argc) {
        g_print ("non-option ARGV-elements: ");
        while (optind < argc) {
            g_print ("%s ", argv[optind++]);
        }
        g_print ("\n");
    }

    /* do all that good daemon setup stuff */
    if (do_detach) {
        if (alp_daemon_start (false, true) < 0) {
            ALP_ATLX(DAEMON, "daemon detach failed: %s", strerror(errno));
        }
        ALP_ATLX(DAEMON, "daemon detached");
    }

    /* write out the pid to the pid-file */

    /* create main glib loop */
    loop = g_main_loop_new(NULL, FALSE);

    /* setup the database */
    err = alp_attnmgrd_db_setup(db_directory, db_filename, gAlpAttnMgrDDebug);
    if (err || gAlpAttnMgrDAbort) {
        ALP_ATLX (DAEMON, "db setup failed : err %x", err);
        /* should do something - log an error ? update status */
        return 1;
    }

    /* setup IPC handling */
    alp_attnmgrd_ipc_setup(ipc_channel);
    
    /* setup app server context */
    if (gAlpAttnMgrDTest == false) {
        context = alp_app_setup ("bar:com.access.servers.attnmgr", 0, NULL);
    }

    /* setup bundle manager */
    alp_bundle_init();

    /* get pid and write out to pid file */
    {
        char *pid_buf;

        my_pid = getpid();
        if ((fp = fopen(pid_filename, "w"))) {
            fprintf(fp, "%d\n", my_pid);
            fclose (fp);
        }
    }                

    /* notify parent we are ready and it's okay to exit */
    if (do_detach) {
        alp_daemon_release (0);
    }

    /* sending notification that attn mgr is up and running */
    {
        AlpNotifyEventAttnAlert detail;
        memset ((void *) &detail, 0, sizeof(detail));
        alp_notify_broadcast (ALP_ATTNMGRD_NOTIFYID_SELF,
                              ALP_NOTIFY_EVENT_ATTN_ALERT,
                              &detail, sizeof(detail),
                              ALP_NOTIFY_QUEUE_DEFAULT, 0);
    }
    
    /* enter the main loop */
    g_main_loop_run (loop);

    /* clean up */
    alp_attnmgrd_ipc_teardown();
    alp_attnmgrd_db_teardown();
    alp_app_abandon (context);

    /* restart should occur if we get here for some reason */


    /* remove pid file */


    /*  exit */
	return 0;
}
