/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/*! @file exgmgrd.c
 *
 * @brief Exchange Manager Daemon
 *
*/

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include <dlfcn.h>
#include <fcntl.h>

#include <sys/un.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <glib.h>
#include <glib/gmain.h>
#include <errno.h>

#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <hiker/syspaths.h>
#include <hiker/sysclass.h>
#include <hiker/traces.h>
#include <hiker/appmgr.h>
#include <hiker/pdk/appmgr_pdk.h>
#include <hiker/bundlemgr.h>
#include <hiker/notifymgr.h>
#include <hiker/pdk/sysutils_pdk.h>

#include "exgmgrd.h"

#define PRV_EXG_CACHE_DATA_FOLDER "exgmgr/"
#define PRV_HANDLER_CONFIG_DIR ALP_SYSPATH_SYSTEM_CACHE PRV_EXG_CACHE_DATA_FOLDER
#define PRV_HANDLER_CONFIG_FILE PRV_HANDLER_CONFIG_DIR"regcache.data"
#define PRV_HANDLER_CONFIG_FILE_TMP PRV_HANDLER_CONFIG_DIR"regcache.tmp"

// This contains the list of registered handlers
static AlpExgMgrdHandlerInfoType *g_alp_exg_daemon_handler_info_list_head;
static AlpExgMgrdHandlerInfoType *g_alp_exg_daemon_handler_info_list_tail;

// This contains the list of incoming requests pending dispatching
static AlpExgMgrdDispatchInfoType *g_alp_exg_daemon_dispatch_info_list_head;
static AlpExgMgrdDispatchInfoType *g_alp_exg_daemon_dispatch_info_list_tail;

// g_alp_exg_daemon_socket_listener is an AFUNIX listener used to accept connections from the exgmgr lib (client or handler API).
static int g_alp_exg_daemon_socket_listener;

// this handles the GLib event source to manage the client or handler connections.
static GPollFD g_alp_exg_daemon_GPollFd;
static GSourceFuncs g_alp_exg_daemon_event_source_funcs;
static GSource *g_alp_exg_daemon_event_source;
static guint g_alp_exg_daemon_event_source_id;

#define ALP_EXG_DAEMON_MAX_CONNECTIONS 10
static AlpExgDaemonCxnState g_alp_exg_daemon_cxn_state[ALP_EXG_DAEMON_MAX_CONNECTIONS];

static int g_alp_exg_daemon_transports_count;
static AlpExgMgrdTransportInfoType *g_alp_exg_daemon_transports;

// unique id current value
static int gNextGuidValue;

static GMainLoop *g_alp_exg_daemon_main_loop;

/* -------------------------------------------------------------------------------------------
 *  */
static void alp_prv_exg_daemon_SIGTERM_Handler (int status)
{
	ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : SIGTERM received"));
	g_main_loop_quit (g_alp_exg_daemon_main_loop);
}

/* -------------------------------------------------------------------------------------------
 * alp_prv_exg_daemon_dequeue_dispatch_entry */
static void alp_prv_exg_daemon_dequeue_dispatch_entry (AlpExgMgrdDispatchInfoType *iDispatchInfoP)
{
	// free the tmpdatafile if we use one
	if (iDispatchInfoP->itsTmpDatafile != NULL)
	{
		unlink (iDispatchInfoP->itsTmpDatafile);
		free (iDispatchInfoP->itsTmpDatafile);
		iDispatchInfoP->itsTmpDatafile = NULL;
	}

	if (iDispatchInfoP->itsPrev == NULL)
		g_alp_exg_daemon_dispatch_info_list_head = iDispatchInfoP->itsNext;
	else
		iDispatchInfoP->itsPrev->itsNext = iDispatchInfoP->itsNext;

	if (iDispatchInfoP->itsNext == NULL)
		g_alp_exg_daemon_dispatch_info_list_tail = iDispatchInfoP->itsPrev;
	else
		iDispatchInfoP->itsNext->itsPrev = iDispatchInfoP->itsPrev;
}

/* -------------------------------------------------------------------------------------------
 * alp_prv_exg_daemon_dequeue_handler_entry */
static void alp_prv_exg_daemon_dequeue_handler_entry (AlpExgMgrdHandlerInfoType *iHandlerInfoP)
{
	if (iHandlerInfoP->itsPrev == NULL)
		g_alp_exg_daemon_handler_info_list_head = iHandlerInfoP->itsNext;
	else
		iHandlerInfoP->itsPrev->itsNext = iHandlerInfoP->itsNext;

	if (iHandlerInfoP->itsNext == NULL)
		g_alp_exg_daemon_handler_info_list_tail = iHandlerInfoP->itsPrev;
	else
		iHandlerInfoP->itsNext->itsPrev = iHandlerInfoP->itsPrev;
}

/* -------------------------------------------------------------------------------------------
 * alp_prv_exg_daemon_find_handler */
static AlpExgMgrdHandlerInfoType *alp_prv_exg_daemon_find_handler (char *iVerb, char *iMIMEType, AlpExgMgrdHandlerInfoType *iStartHandlerInfoP)
{
	while (iStartHandlerInfoP)
	{
		if (strcmp (iStartHandlerInfoP->itsVerb, iVerb) == 0)
			if (strcasecmp (iStartHandlerInfoP->itsMIMEType, iMIMEType) == 0)
				return iStartHandlerInfoP;
				
		iStartHandlerInfoP = iStartHandlerInfoP->itsNext;
	}
	
	return NULL;
}

/* -------------------------------------------------------------------------------------------
 * alp_prv_exg_daemon_add_handler */
static alp_status_t alp_prv_exg_daemon_add_handler (char *iVerb, char *iLocalizedVerb, char *iMIMEType, int iInvokeType, char *iInvokeData, int iLocalOnly,
													int iSerializeMe, int iMustBeUnique, int iAcceptGenericInvocation)
{
	AlpExgMgrdHandlerInfoType *theHandlerInfoP = g_alp_exg_daemon_handler_info_list_head;
	
	// Check all handlers for these verb/MIMEType and see if they conflict
	while ((theHandlerInfoP = alp_prv_exg_daemon_find_handler (iVerb, iMIMEType, theHandlerInfoP)) != NULL)
	{
		// If the same bundle/notif has registered this handler, return ALP_EXG_MGR_ERR_HANDLER_ALREADY_REGISTERED_ERROR.
		// To be considered identical, the found handler must be of the same invoke type, and same bundle/notif type
		if (theHandlerInfoP->itsInvokeType == iInvokeType)
			switch (theHandlerInfoP->itsInvokeType)
			{
				case ALP_EXG_DAEMON_HANDLER_INVOKE_BY_BUNDLE:
				{
					if (strcmp (theHandlerInfoP->params.byBundle.itsBundleName, iInvokeData) == 0)
						return ALP_EXG_MGR_ERR_HANDLER_ALREADY_REGISTERED_ERROR;	// this is already registered for this bundlename
				} break;
				
				case ALP_EXG_DAEMON_HANDLER_INVOKE_BY_NOTIFICATION:
				{
					if (strcmp (theHandlerInfoP->params.byNotification.itsNotificationType, iInvokeData) == 0)
						return ALP_EXG_MGR_ERR_HANDLER_ALREADY_REGISTERED_ERROR;	// this is already registered for this Notification Type
				} break;
			}
			
		// Either it's not the same invoke type, or it's not the same bundle/Notif. Check uniqueness if needed
		if (iMustBeUnique || theHandlerInfoP->itsMustBeUnique)
			return ALP_EXG_MGR_ERR_HANDLER_ALREADY_EXISTS_ERROR;
			
		// continue to search starting at the next handler.
		theHandlerInfoP = theHandlerInfoP->itsNext;
	}
	
	// OK, there are no conflicting handlers already in the list.
	// Allocate a new handler entry and fill it...
	theHandlerInfoP = (AlpExgMgrdHandlerInfoType *) malloc (sizeof(AlpExgMgrdHandlerInfoType));
	if (theHandlerInfoP == NULL)
		return ALP_EXGMGR_ERR_OUT_OF_MEMORY_ERROR;
		
	theHandlerInfoP->itsVerb = (char *) malloc (strlen(iVerb) + 1);
	if (theHandlerInfoP->itsVerb == NULL)
		goto freeHandlerP;
	strcpy (theHandlerInfoP->itsVerb, iVerb);

	if (iLocalizedVerb[0])
	{
		theHandlerInfoP->itsLocalizedVerb = (char *) malloc (strlen(iLocalizedVerb) + 1);
		if (theHandlerInfoP->itsLocalizedVerb == NULL)
			goto freeVerb;
		strcpy (theHandlerInfoP->itsLocalizedVerb, iLocalizedVerb);
	}
	else
		theHandlerInfoP->itsLocalizedVerb = NULL;
				
	theHandlerInfoP->itsMIMEType = (char *) malloc (strlen(iMIMEType) + 1);
	if (theHandlerInfoP->itsMIMEType == NULL)
		goto freeLocalizedVerb;
	strcpy (theHandlerInfoP->itsMIMEType, iMIMEType);
	
	theHandlerInfoP->itsInvokeType = iInvokeType;
	switch (iInvokeType)
	{
		case ALP_EXG_DAEMON_HANDLER_INVOKE_BY_BUNDLE:
		{
			theHandlerInfoP->params.byBundle.itsBundleName = (char *) malloc (strlen(iInvokeData) + 1);
			if (theHandlerInfoP->params.byBundle.itsBundleName == NULL)
				goto freeMIMEType;
			strcpy (theHandlerInfoP->params.byBundle.itsBundleName, iInvokeData);
		} break;
		
		case ALP_EXG_DAEMON_HANDLER_INVOKE_BY_NOTIFICATION:
		{
			theHandlerInfoP->params.byNotification.itsNotificationType = (char *) malloc (strlen(iInvokeData) + 1);
			if (theHandlerInfoP->params.byNotification.itsNotificationType == NULL)
				goto freeMIMEType;
			strcpy (theHandlerInfoP->params.byNotification.itsNotificationType, iInvokeData);
		} break;
	}

	theHandlerInfoP->itsLocalOnly = iLocalOnly;
	theHandlerInfoP->itsSerializeMe = iSerializeMe;
	theHandlerInfoP->itsMustBeUnique = iMustBeUnique;
	theHandlerInfoP->itsAcceptGenericInvocation = iAcceptGenericInvocation;

	// Always insert at the end for now... This may change when we apply priorities to multiple handlers.
	theHandlerInfoP->itsNext = NULL;
	if (g_alp_exg_daemon_handler_info_list_head == NULL)
	{
		g_alp_exg_daemon_handler_info_list_head = theHandlerInfoP;
		theHandlerInfoP->itsPrev = NULL;
	}
	else
	{
		g_alp_exg_daemon_handler_info_list_tail->itsNext = theHandlerInfoP;
		theHandlerInfoP->itsPrev = g_alp_exg_daemon_handler_info_list_tail;
	}
	g_alp_exg_daemon_handler_info_list_tail = theHandlerInfoP;

	return 0;
	
freeMIMEType:	
	free (theHandlerInfoP->itsMIMEType);

freeLocalizedVerb:
	if (theHandlerInfoP->itsLocalizedVerb)
		free (theHandlerInfoP->itsLocalizedVerb);

freeVerb:	
	free (theHandlerInfoP->itsVerb);

freeHandlerP:
	free (theHandlerInfoP);
	
	return ALP_EXGMGR_ERR_OUT_OF_MEMORY_ERROR;
}


/* -------------------------------------------------------------------------------------------
 * prv_write_string */
static alp_status_t prv_write_string (int iFd, char *iString)
{
	size_t theLen = strlen (iString) + 1;
	if (write (iFd, &theLen, sizeof (size_t)) > 0)
		if (write (iFd, iString, theLen) > 0)
			return 0;
		
	ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr : prv_write_string: cannot write string to file, error = 0x%lx", errno));
	return ALP_EXG_MGR_ERR_WRITE_ERROR;
}

/* -------------------------------------------------------------------------------------------
 * prv_write_int */
static alp_status_t prv_write_int (int iFd, int iInt)
{
	if (write (iFd, &iInt, sizeof (int)) > 0)
		return 0;
		
	ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr : prv_write_string: cannot write int to file, error = 0x%lx", errno));
	return ALP_EXG_MGR_ERR_WRITE_ERROR;
}

/* -------------------------------------------------------------------------------------------
 * prv_save_handlers_config */
static alp_status_t prv_save_handlers_config (char *iFilename)
{
	alp_status_t theErr = 0;
	int theFd;
	AlpExgMgrdHandlerInfoType *theHandlerInfoP = g_alp_exg_daemon_handler_info_list_head;
	
	theFd = open (iFilename, O_RDWR | O_CREAT | O_EXCL, 0640);
	if (theFd < 0)
	{
		ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : Cannot open persistent data file for write"));
		return -1;
	}
	
	while (theHandlerInfoP)
	{
		// save only the bundle registrations
		if (theHandlerInfoP->itsInvokeType == ALP_EXG_DAEMON_HANDLER_INVOKE_BY_BUNDLE)
		{
			theErr = prv_write_string (theFd, theHandlerInfoP->itsVerb);
			if (theErr == 0)
				theErr = prv_write_string (theFd, theHandlerInfoP->itsLocalizedVerb ? theHandlerInfoP->itsLocalizedVerb : "");
			if (theErr == 0)
				theErr = prv_write_string (theFd, theHandlerInfoP->itsMIMEType);
			if (theErr == 0)
				theErr = prv_write_string (theFd, theHandlerInfoP->params.byBundle.itsBundleName);
			if (theErr == 0)
				theErr = prv_write_int (theFd, theHandlerInfoP->itsLocalOnly);
			if (theErr == 0)
				theErr = prv_write_int (theFd, theHandlerInfoP->itsSerializeMe);
			if (theErr == 0)
				theErr = prv_write_int (theFd, theHandlerInfoP->itsMustBeUnique);
			if (theErr == 0)
				theErr = prv_write_int (theFd, theHandlerInfoP->itsAcceptGenericInvocation);
			if (theErr)
				break;
		}
				
		theHandlerInfoP = theHandlerInfoP->itsNext;
	}
	
	// append a null string at the end (first string cannot be null)
	if (theErr == 0)
		theErr = prv_write_string (theFd, "");
	
	if (close (theFd) < 0)
		if (theErr == 0)
			theErr = -1;

	if (theErr)
		unlink (iFilename);

	return theErr;
}

alp_status_t prv_update_registration_cache ()
{
	alp_status_t theErr;

	// Make sure directory exist
	if (access(PRV_HANDLER_CONFIG_DIR, R_OK) != 0)
		if (mkdir (PRV_HANDLER_CONFIG_DIR, 0777) < 0)
			return ALP_EXGMGR_ERR_REG_CACHE_ERROR;
	
	// delete the old cache file (file may not exist)
	(void) unlink (PRV_HANDLER_CONFIG_FILE);
	// Update the cache file. First store the cache in a tmp file
	theErr = prv_save_handlers_config (PRV_HANDLER_CONFIG_FILE_TMP);

	// then rename the tmp file to the cache file (this is atomic and guarantees the cache file is valid)
	if (theErr == 0)
		if (rename (PRV_HANDLER_CONFIG_FILE_TMP, PRV_HANDLER_CONFIG_FILE) < 0)
			theErr = ALP_EXGMGR_ERR_REG_CACHE_ERROR;
			
	return theErr;
}

/* -------------------------------------------------------------------------------------------
 * prv_read_string */
alp_status_t prv_read_string (int iFd, char **oString)
{
	// read string len (including final null)
	size_t theLen;
	if (read (iFd, &theLen, sizeof(size_t)) > 0)
	{
		*oString = (char *) malloc (theLen);
		if (*oString == NULL)
			return ALP_EXGMGR_ERR_OUT_OF_MEMORY_ERROR;
		if (read (iFd, *oString, theLen) > 0)
			return 0;
		else
			free (*oString);
	}
	return ALP_EXG_MGR_ERR_READ_ERROR;
}

/* -------------------------------------------------------------------------------------------
 * prv_read_int */
alp_status_t prv_read_int (int iFd, int *oInt)
{
	if (read (iFd, oInt, sizeof(int)) > 0)
		return 0;
	return ALP_EXG_MGR_ERR_READ_ERROR;
}

/* -------------------------------------------------------------------------------------------
 * prv_load_handlers_config */
static alp_status_t prv_load_handlers_config (char *iFilename)
{
	int theFd;
	alp_status_t theErr;
	char *theVerb;
	char *theLocalizedVerb;
	char *theMIMEType;
	char *theBundleName;
	int theLocalOnly;
	int theSerializeMe;
	int theMustBeUnique;
	int theAcceptGenericInvocation;
					
	theFd = open (iFilename, O_RDONLY);
	if (theFd < 0)
	{
		ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : Cannot open persistent data file for read"));
		return -1;
	}
	
	while (1)
	{
		theVerb = theLocalizedVerb = theMIMEType = theBundleName = NULL;
		
		// Read first string (verb). If empty string, then we're done.
		theErr = prv_read_string (theFd, &theVerb);
		if (theErr == 0)
			if (*theVerb == 0)
			{
				free (theVerb);
				break;
			}
		
		if (theErr == 0)
			theErr = prv_read_string (theFd, &theLocalizedVerb);

		if (theErr == 0)
			theErr = prv_read_string (theFd, &theMIMEType);
			
		if (theErr == 0)
			theErr = prv_read_string (theFd, &theBundleName);			
				
		if (theErr == 0)
			theErr = prv_read_int (theFd, &theLocalOnly);		

		if (theErr == 0)
			theErr = prv_read_int (theFd, &theSerializeMe);		
		
		if (theErr == 0)
			theErr = prv_read_int (theFd, &theMustBeUnique);		
		
		if (theErr == 0)
			theErr = prv_read_int (theFd, &theAcceptGenericInvocation);		
				
		if (theErr == 0)
			theErr = alp_prv_exg_daemon_add_handler (theVerb, theLocalizedVerb, theMIMEType, ALP_EXG_DAEMON_HANDLER_INVOKE_BY_BUNDLE, theBundleName, 
													theLocalOnly, theSerializeMe, theMustBeUnique, theAcceptGenericInvocation);
		if (theVerb)
			free (theVerb);
		if (theLocalizedVerb)
			free (theLocalizedVerb);
		if (theMIMEType)
			free (theMIMEType);
		if (theBundleName)
			free (theBundleName);	
				
		if (theErr)
			break;	
	}
	
	close (theFd);
	
	return theErr;
}

/* -------------------------------------------------------------------------------------------
 * alp_prv_exg_daemon_remove_handler */
static alp_status_t alp_prv_exg_daemon_remove_handler (char *iVerb, char *iMIMEType, char *iBundleName)
{
	AlpExgMgrdHandlerInfoType *theHandlerInfoP = g_alp_exg_daemon_handler_info_list_head;
	
	// Check all handlers for these verb/MIMEType.
	while ((theHandlerInfoP = alp_prv_exg_daemon_find_handler (iVerb, iMIMEType, theHandlerInfoP)) != NULL)
	{
		if (theHandlerInfoP->itsInvokeType == ALP_EXG_DAEMON_HANDLER_INVOKE_BY_BUNDLE)
			if (strcmp (theHandlerInfoP->params.byBundle.itsBundleName, iBundleName) == 0)
			{
				alp_prv_exg_daemon_dequeue_handler_entry (theHandlerInfoP);
				free (theHandlerInfoP->itsVerb);
				if (theHandlerInfoP->itsLocalizedVerb)
					free (theHandlerInfoP->itsLocalizedVerb);
				free (theHandlerInfoP->itsMIMEType);
				free (theHandlerInfoP->params.byBundle.itsBundleName);
				free (theHandlerInfoP);
				
				return 0;
			}
			
		// continue to search starting at the next handler.
		theHandlerInfoP = theHandlerInfoP->itsNext;
	}

	return ALP_EXG_MGR_ERR_HANDLER_NOT_FOUND_ERROR;
}

/* -------------------------------------------------------------------------------------------
 * alp_prv_exg_daemon_remove_handler_notification */
static alp_status_t alp_prv_exg_daemon_remove_handler_notification (char *iVerb, char *iMIMEType, char *iNotificationType)
{
	AlpExgMgrdHandlerInfoType *theHandlerInfoP = g_alp_exg_daemon_handler_info_list_head;
	
	// Check all handlers for these verb/MIMEType.
	while ((theHandlerInfoP = alp_prv_exg_daemon_find_handler (iVerb, iMIMEType, theHandlerInfoP)) != NULL)
	{
		if (theHandlerInfoP->itsInvokeType == ALP_EXG_DAEMON_HANDLER_INVOKE_BY_NOTIFICATION)
			if (strcmp (theHandlerInfoP->params.byNotification.itsNotificationType, iNotificationType) == 0)
			{
				alp_prv_exg_daemon_dequeue_handler_entry (theHandlerInfoP);
				free (theHandlerInfoP->itsVerb);
				if (theHandlerInfoP->itsLocalizedVerb)
					free (theHandlerInfoP->itsLocalizedVerb);
				free (theHandlerInfoP->itsMIMEType);
				free (theHandlerInfoP->params.byNotification.itsNotificationType);
				free (theHandlerInfoP);
				
				return 0;
			}
			
		// continue to search starting at the next handler.
		theHandlerInfoP = theHandlerInfoP->itsNext;
	}

	return ALP_EXG_MGR_ERR_HANDLER_NOT_FOUND_ERROR;
}

/* -------------------------------------------------------------------------------------------
 * alp_prv_exg_daemon_check_handler_registration */
static alp_status_t alp_prv_exg_daemon_check_handler_registration (char *iVerb, char *iMIMEType)
{
	if (g_alp_exg_daemon_handler_info_list_head)
		if (alp_prv_exg_daemon_find_handler (iVerb, iMIMEType, g_alp_exg_daemon_handler_info_list_head) != NULL)
			return 0;

	return ALP_EXG_MGR_ERR_HANDLER_NOT_FOUND_ERROR;
}

/* -------------------------------------------------------------------------------------------
 * alp_prv_exg_daemon_query_generic_verbs */
static alp_status_t alp_prv_exg_daemon_query_generic_verbs (char *iMIMEType, size_t *oDataSize, char **oData)
{
	// Allocates *oData to contain the number of entries (int), followed by a null terminated verb and a verb description for each entry.
	// If it fails, nothing is allocated and *oData is set to NULL. oDataSize is set to the size of the buffer.
	
	AlpExgMgrdHandlerInfoType *theHandlerInfoP = g_alp_exg_daemon_handler_info_list_head;
	size_t theUsedSize;
	
	// initial size of data buffer
	*oDataSize = 1024;
	*oData = malloc (*oDataSize);
	if (*oData == NULL)
		return ALP_EXGMGR_ERR_OUT_OF_MEMORY_ERROR;
		
	// zero entries.
	* (int *) *oData = 0;
	theUsedSize = sizeof (int);

	// go through all handlers
	while (theHandlerInfoP)
	{
		if (theHandlerInfoP->itsAcceptGenericInvocation)
			if (strcasecmp (theHandlerInfoP->itsMIMEType, iMIMEType) == 0)
			{
				size_t theVerbLen;
				size_t theDescLen;
	
				// calculate the needed buffer space. If localized verb was not given, return the verb instead.
				theVerbLen = strlen (theHandlerInfoP->itsVerb) + 1;
				if (theHandlerInfoP->itsLocalizedVerb)
					theDescLen = strlen (theHandlerInfoP->itsLocalizedVerb) + 1;
				else
					theDescLen = strlen (theHandlerInfoP->itsVerb) + 1;
				
				// Check for space in the buffer, and realloc if necessary.	
				if (theUsedSize + theVerbLen + theDescLen > *oDataSize)
				{
					*oDataSize += 1024;
					*oData = realloc (*oData, *oDataSize);
					if (*oData == NULL)
						return ALP_EXGMGR_ERR_OUT_OF_MEMORY_ERROR;
				}
				
				// Copy verb and localized verb.
				strcpy (*oData + theUsedSize, theHandlerInfoP->itsVerb);
				if (theHandlerInfoP->itsLocalizedVerb)
					strcpy (*oData + theUsedSize + theVerbLen, theHandlerInfoP->itsLocalizedVerb);
				else
					strcpy (*oData + theUsedSize + theVerbLen, theHandlerInfoP->itsVerb);
				theUsedSize += theVerbLen + theDescLen;
				
				// one more entry
				* (int *) *oData += 1;
			}
				
		theHandlerInfoP = theHandlerInfoP->itsNext;
	}
	return 0;	
}

/* -------------------------------------------------------------------------------------------
 * alp_prv_exg_daemon_terminate_dispatcher : terminate all pending dispatch request immediately by sending an abort result to their transports */
static void alp_prv_exg_daemon_terminate_dispatcher ()
{
	// TODO: This is supposed to terminate immediately incoming requests in the queue that have not been dispatched yet.
	// If we don't do anything, the clients will get a broken connection and return this to their apps.
}

/* -------------------------------------------------------------------------------------------
 * alp_prv_exg_daemon_handler_timeout */
static gboolean alp_prv_exg_daemon_handler_timeout (gpointer iData)
{
	AlpExgMgrdDispatchInfoType *theDispatchInfoP = (AlpExgMgrdDispatchInfoType *) iData;

	ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : alp_prv_exg_daemon_handler_timeout : HANDLER DID NOT RESPOND TO INVOCATION. Terminate dispatch and return request to transport with result = TIMEOUT"));
	
	// The handler has not connected. Abort dispatch with timeout.
	g_alp_exg_daemon_transports[theDispatchInfoP->itsTransportId].itsTransportDispatchResultFunc (theDispatchInfoP->itsRequest, ALP_EXG_MGR_ERR_HANDLER_TIMEOUT_ERROR);
	// and remove from this queue
	alp_prv_exg_daemon_dequeue_dispatch_entry (theDispatchInfoP);
	// no need to remove the timeout source, this will be done by returning FALSE from this function.
	free (theDispatchInfoP);
	
	// return FALSE to have the source removed.	
	return FALSE;
}

/* -------------------------------------------------------------------------------------------
 * alp_prv_exg_daemon_run_dispatcher */
static alp_status_t alp_prv_exg_daemon_run_dispatcher ()
{
	// This function is called when a new pending request was added, or a handler signaled execution completion.
	// Go through all pending request, and dispatch those who don't conflict with the ones that are currently executing.
	
	//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : alp_prv_exg_daemon_run_dispatcher called"));
	
	AlpExgMgrdDispatchInfoType *theDispatchInfoP = g_alp_exg_daemon_dispatch_info_list_head;
	while (theDispatchInfoP)
	{
		// itsDispatchInProgress is set when the handler is invoked,
		if (theDispatchInfoP->itsDispatchInProgress == 0)
		{
			alp_status_t theErr;
			char theVerb[ALP_EXGMGR_VERB_MAXLEN + 1];
			char theMIMEType[ALP_EXGMGR_MIMETYPE_MAXLEN + 1];
			AlpExgMgrdHandlerInfoType *theHandlerInfoP;
			
			// If this request already executed a handler, search from the following handler. Otherwise, search from the start.
			if (theDispatchInfoP->itsLastHandlerInfoP == NULL)
				theHandlerInfoP = g_alp_exg_daemon_handler_info_list_head;
			else
				theHandlerInfoP = theDispatchInfoP->itsLastHandlerInfoP->itsNext;	// OK if this is NULL

			theErr = alp_exg_request_get_verb (theDispatchInfoP->itsRequest, theVerb, sizeof(theVerb));
			if (theErr == 0)
				theErr = alp_exg_request_get_data_type (theDispatchInfoP->itsRequest, theMIMEType, sizeof(theMIMEType));
			if (theErr == 0)
			{
				theHandlerInfoP = alp_prv_exg_daemon_find_handler (theVerb, theMIMEType, theHandlerInfoP);
				if (theHandlerInfoP == NULL)
				{
					// No handler found. If this is the first handler search for this request, return ALP_EXG_MGR_ERR_HANDLER_NOT_FOUND_ERROR
					// but if at least one handler already was already invoked, then this is a ALP_EXG_MGR_ERR_HANDLER_DID_NOT_CONSUME_ERROR.
					if (theDispatchInfoP->itsLastHandlerInfoP == NULL)
						theErr = ALP_EXG_MGR_ERR_HANDLER_NOT_FOUND_ERROR;
					else
						theErr = ALP_EXG_MGR_ERR_HANDLER_DID_NOT_CONSUME_ERROR;
				}
				else
				{
					if (theHandlerInfoP->itsMustBeUnique == 0)
					{
						// Handler may not be unique. If this is the first handler search for this request, save the data to a tmp file to
						// be able to give it again to each handler invocation, and save the current handler to search for the 'next' from there.
						if (theDispatchInfoP->itsLastHandlerInfoP == NULL)
						{
							int theFd;
							theErr = alp_exg_request_get_data_fd (theDispatchInfoP->itsRequest, &theFd);
							if (theErr == 0 && theFd >= 0)
							{
								theErr = alp_exg_prv_fd_to_file (theFd, "/tmp/exgrq", &theDispatchInfoP->itsTmpDatafile);
								if (theErr == 0)
								{
									theFd = open (theDispatchInfoP->itsTmpDatafile, O_RDONLY);
									if (theFd < 0)
										theErr = ALP_EXGMGR_ERR_CANNOT_OPEN_TMP_FILE_ERROR;
									else
									{
										theErr = alp_exg_request_set_data_fd (theDispatchInfoP->itsRequest, theFd);
										if (theErr)
											close (theFd);
									}
								}
							}						
						}
						// save current handler, so we'll search from the following next time
						theDispatchInfoP->itsLastHandlerInfoP = theHandlerInfoP;
					}
				}
			}
			
			if (theErr == 0 && theHandlerInfoP->itsLocalOnly == 1)
			{
				//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : alp_prv_exg_daemon_run_dispatcher : Found a handler with iLocalOnly"));
				if (strcmp (g_alp_exg_daemon_transports[theDispatchInfoP->itsTransportId].itsNameId, "local") != 0)
				{
					// this handler accepts request from the local device only
					theErr = ALP_EXG_MGR_ERR_HANDLER_IS_LOCAL_ONLY_ERROR;
				}
			}
			
			if (theErr == 0 && theHandlerInfoP->itsSerializeMe == 1)
			{
				// check that there are no currently executing handlers with itsSerializeMe also set.
				// TODO : FOR NOW, DON'T CHECK AND JUST GO AHEAD
				
			}

			if (theErr == 0)
			{
				// start a timeout source to protect against the handler never executing the request
				theDispatchInfoP->itsTimeoutSourceId = g_timeout_add (20000, alp_prv_exg_daemon_handler_timeout, (gpointer) theDispatchInfoP);
				if (theDispatchInfoP->itsTimeoutSourceId == 0)
					theErr = ALP_EXG_MGR_ERR_CANNOT_CREATE_GSOURCE_ERROR;
			}

			if (theErr == 0)
			{
				// We are going to invoke the handler. mark the entry in progress right now. 
				// Note that there can't be a race condition because we will not detect the handler connection message until we have
				// returned from here.
				theDispatchInfoP->itsDispatchInProgress = 1;
				
				// invoke handler.
				switch (theHandlerInfoP->itsInvokeType)
				{
					case ALP_EXG_DAEMON_HANDLER_INVOKE_BY_BUNDLE:
					{
						char *argv[3];
						char theRequestIdStr[15];
					
						argv[0] = ALP_APP_EXCHANGE;
						sprintf (theRequestIdStr, "%p", (void *) theDispatchInfoP->itsRequest);
						argv[1] = theRequestIdStr;
						argv[2] = ALP_APP_TRANSIENT;
						ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : alp_prv_exg_daemon_run_dispatcher : invoking %s to handle %s / %s (regId = %s)", theHandlerInfoP->params.byBundle.itsBundleName,
																																			theVerb, theMIMEType, theRequestIdStr));
						theErr = alp_app_launch(theHandlerInfoP->params.byBundle.itsBundleName, 3, argv, NULL);
						if (theErr)
							ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : alp_prv_exg_daemon_run_dispatcher : alp_app_launch ERROR 0x%lx", theErr));
					} break;
					
					case ALP_EXG_DAEMON_HANDLER_INVOKE_BY_NOTIFICATION:
					{
						ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : alp_prv_exg_daemon_run_dispatcher : broadcasting %s to handle %s / %s (regId = 0x%lx)", theHandlerInfoP->params.byNotification.itsNotificationType,
																																			theVerb, theMIMEType, theDispatchInfoP->itsRequest));
						theErr = alp_notify_broadcast("alpexgmgrd", theHandlerInfoP->params.byNotification.itsNotificationType, &theDispatchInfoP->itsRequest, sizeof(theDispatchInfoP->itsRequest), ALP_NOTIFY_QUEUE_DEFAULT, 0);
						if (theErr)
							ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : alp_prv_exg_daemon_run_dispatcher : alp_notify_broadcast ERROR 0x%lx", theErr));
					} break;
				}
			}
			
			if (theErr)
			{
				AlpExgMgrdDispatchInfoType *theTempDispatchInfoP;
				// It was not possible to dispatch, so reset the dispatch in progress flag
				theDispatchInfoP->itsDispatchInProgress = 0;
				
				ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : alp_prv_exg_daemon_run_dispatcher : could not dispatch eligible entry (error = 0x%lx), notify transport with immediate completion", theErr));
				// cannot dispatch this entry. Complete immediately. This transfer the request ownership back to the transport.
				g_alp_exg_daemon_transports[theDispatchInfoP->itsTransportId].itsTransportDispatchResultFunc (theDispatchInfoP->itsRequest, theErr);
				theTempDispatchInfoP = theDispatchInfoP;
				// move to next dispatchInfoP for next loop
				theDispatchInfoP = theDispatchInfoP->itsNext;
				// and remove from this queue
				alp_prv_exg_daemon_dequeue_dispatch_entry (theTempDispatchInfoP);
				if (theTempDispatchInfoP->itsTimeoutSourceId != 0)
					g_source_remove (theTempDispatchInfoP->itsTimeoutSourceId);
				free (theTempDispatchInfoP);
			}
			else
			{
				//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : alp_prv_exg_daemon_run_dispatcher : Handler invoked, waiting for handler to connect..."));
				theDispatchInfoP = theDispatchInfoP->itsNext;
			}
		}
		else
			theDispatchInfoP = theDispatchInfoP->itsNext;
	}

	// return an error only in case of unrecoverable error. Errors in dispatch processing are sent to the transport
	// in charge of the currently dispatched request.
	return 0;
}

/* -------------------------------------------------------------------------------------------
 * alp_prv_exg_transport_event_notify */
alp_status_t alp_prv_exg_transport_event_notify (int iTransportId, int iTransportEvent, void *iTransportEventData, size_t iTransportEventDataSize)
{
	// This is called back by a transport module that has an incoming request ready, or a transport unrecoverable error occured.
	// The function may return a non zero result if it was not possible to queue the request, or run_dispatcher failed with unrecoverable error.
	// NOTE : if the request is enqueued and run_dispatcher fails to dispatch it because of a 'normal' error, 
	// then the transport was already notified with the result, and this function returns no error.
	
	alp_status_t theErr = 0;
	
	switch (iTransportEvent)
	{
		case ALP_EXG_TRANSPORT_EVENT_INCOMING_REQUEST:
		{
			// Allocate a AlpExgMgrdDispatchInfoType, and link it to g_alp_exg_daemon_dispatch_info_list_head
			// if cannot allocate, return the error to the transport who will reply an out of memory to the client.
			// otherwise, the request is now ours until we return it to the transport.

			//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : Transport '%s' reported incoming request", g_alp_exg_daemon_transports[iTransportId].itsNameId));

			AlpExgMgrdDispatchInfoType *theDispatchInfoP = (AlpExgMgrdDispatchInfoType *) malloc (sizeof(AlpExgMgrdDispatchInfoType));
			if (theDispatchInfoP == NULL)
				return ALP_EXGMGR_ERR_OUT_OF_MEMORY_ERROR;

			theDispatchInfoP->itsDispatchInProgress = 0;
			theDispatchInfoP->itsTimeoutSourceId = 0;
			theDispatchInfoP->itsRequest = (AlpExgRequest) iTransportEventData;
			theDispatchInfoP->itsTransportId = iTransportId;
			theDispatchInfoP->itsLastHandlerInfoP = NULL;
			theDispatchInfoP->itsTmpDatafile = NULL;
			
			// append at the end
			theDispatchInfoP->itsNext = NULL;
			if (g_alp_exg_daemon_dispatch_info_list_head == NULL)
			{
				g_alp_exg_daemon_dispatch_info_list_head = theDispatchInfoP;
				theDispatchInfoP->itsPrev = NULL;
			}
			else
			{
				g_alp_exg_daemon_dispatch_info_list_tail->itsNext = theDispatchInfoP;
				theDispatchInfoP->itsPrev = g_alp_exg_daemon_dispatch_info_list_tail;
			}
			g_alp_exg_daemon_dispatch_info_list_tail = theDispatchInfoP;
			
			// make sure dispatch is initiated
			theErr = alp_prv_exg_daemon_run_dispatcher ();
			if (theErr)
			{
				// cannot initiate dispatch : this is a fatal error for the daemon
				// HOW SHOULD I HANDLE THAT?
				// The problem with that is if this happens, then it is very probable
				// that we won't be able to dispatch anything anymore....
				// We need a way to notify the whole system about this failure and maybe require a reboot?
				
				// In all cases, dequeue the request
				// and return an error to the transport which will be sent back to client.
				alp_prv_exg_daemon_dequeue_dispatch_entry (theDispatchInfoP);
				free (theDispatchInfoP);
				
				ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_prv_exg_transport_event_notify: alp_prv_exg_daemon_run_dispatcher FAILED."));
			}
			break;
		}
		break;
		
		case ALP_EXG_TRANSPORT_EVENT_FATAL_ERROR:	
		{
			// TODO:
			ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : Transport '%s' reported ALP_EXG_TRANSPORT_EVENT_FATAL_ERROR", g_alp_exg_daemon_transports[iTransportId].itsNameId));
			theErr = ALP_EXGMGR_ERR_NO_ERROR;
		}
		break;
	}
	
	return theErr;
}

/* -------------------------------------------------------------------------------------------
 * alp_prv_exg_daemon_cancel_request. */
static alp_status_t alp_prv_exg_daemon_cancel_request (int iGuid)
{
	// See if the request is waiting in the queue
	// NOTE : this would be for local request only. This is not likely to be used because Cancel would be done in the handler.
	//		  so don't do it for now

	// Bluetooth or Irda can be canceled

	// loop through the transports.
	// If the transport does not recognizes the guid, it returns ALP_EXG_MGR_ERR_NO_SUCH_REQUEST_ERROR
	// Otherwise, it tries to cancel and return the result.
	alp_status_t theErr;
	int theIndex;
	for (theIndex = 0; theIndex < g_alp_exg_daemon_transports_count; theIndex++)
	{
		theErr = g_alp_exg_daemon_transports[theIndex].itsTransportCancelFunc(iGuid);
		if (theErr == 0)
			return 0;
		else if (theErr != ALP_EXG_MGR_ERR_NO_SUCH_REQUEST_ERROR)
			return theErr;
	}
	
	return ALP_EXG_MGR_ERR_NO_SUCH_REQUEST_ERROR;
}

/* -------------------------------------------------------------------------------------------
 * alp_exg_daemon_run_connection. This function shall return an error only if an unrecoverable error
 * occured (an error which prevent the daemon to continue its operation). Any error on the connection or in
 * the protocol should simply abort the current connection and free the entry. This will result in
 * a file descriptor error or hangup on the client side */
static alp_status_t alp_exg_daemon_run_connection (AlpExgDaemonCxnState *iCxnState)
{
	ssize_t theLen;
	alp_status_t theErr;
	
	// read the AlpExgMgrdMessageType, handle it depending on the opcode,
	switch (iCxnState->itsState)
	{
		case ALP_EXG_DAEMON_STATE_WAIT_CXN_MESSAGE:
			if (iCxnState->itsPollFD.revents & G_IO_IN)
			{
				//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: got G_IO_IN on socket %d (state = ALP_EXG_DAEMON_WAIT_CXN_MESSAGE)", iCxnState->itsPollFD.fd));
				theLen = read (iCxnState->itsPollFD.fd, iCxnState->itsTempBuf + iCxnState->itsTempBufUsed, sizeof (AlpExgMgrdMessageType) - iCxnState->itsTempBufUsed);
				if (theLen < 0)
				{
					ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: read error on socket %d, ABORTING.", iCxnState->itsPollFD.fd));
					goto exitTerminateEntry;
				}
				iCxnState->itsTempBufUsed += theLen;
				if (iCxnState->itsTempBufUsed == sizeof (AlpExgMgrdMessageType))
				{
					// We have the message in the first bytes of iCxnState->itsTempBuf
					AlpExgMgrdMessageType *theMessageP = (AlpExgMgrdMessageType *) iCxnState->itsTempBuf;
					
					// save the opcode. This is needed to know what to cleanup when the connection is closed.
					iCxnState->itsOpcode = theMessageP->itsOpcode;
					
					switch (theMessageP->itsOpcode)
					{
						case ALP_EXG_DAEMON_MSG_OPCODE_REGISTER_HANDLER :
						{
							ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: Received BUNDLE REGISTRATION for %s (%s) / %s / %s",
															theMessageP->params.register_handler_bundle.itsVerb, theMessageP->params.register_handler_bundle.itsLocalizedVerb, theMessageP->params.register_handler_bundle.itsMIMEType, theMessageP->params.register_handler_bundle.itsHandlerBundleName));
							theErr = alp_prv_exg_daemon_add_handler (theMessageP->params.register_handler_bundle.itsVerb,
																										theMessageP->params.register_handler_bundle.itsLocalizedVerb,
																										theMessageP->params.register_handler_bundle.itsMIMEType,
																										ALP_EXG_DAEMON_HANDLER_INVOKE_BY_BUNDLE,
																										theMessageP->params.register_handler_bundle.itsHandlerBundleName,
																										theMessageP->params.register_handler_bundle.itsLocalOnly,
																										theMessageP->params.register_handler_bundle.itsSerializeMe,
																										theMessageP->params.register_handler_bundle.itsMustBeUnique,
																										theMessageP->params.register_handler_bundle.itsAcceptGenericInvocation);
							//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_prv_exg_daemon_add_handler result = 0x%lx", theErr));
							// update the cache.
							if (theErr == 0)
							{
								theErr = prv_update_registration_cache();
								//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: prv_update_registration_cache result = 0x%lx", theErr));
							}
								
							((AlpExgMgrdMessageReplyType *) iCxnState->itsTempBuf)->itsReplyStatus = theErr;
							iCxnState->itsTempBufUsed = 0;
							iCxnState->itsPollFD.events = G_IO_OUT;
							iCxnState->itsState = ALP_EXG_DAEMON_STATE_SEND_FINAL_REPLY_STATUS;
						} break;
						
						case ALP_EXG_DAEMON_MSG_OPCODE_REGISTER_HANDLER_NOTIFICATION :
						{
							ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: Received NOTIFICATION REGISTRATION for %s (%s) / %s / %s",
															theMessageP->params.register_handler_notification.itsVerb, theMessageP->params.register_handler_notification.itsLocalizedVerb, theMessageP->params.register_handler_notification.itsMIMEType, theMessageP->params.register_handler_notification.itsHandlerNotificationType));
							((AlpExgMgrdMessageReplyType *) iCxnState->itsTempBuf)->itsReplyStatus = alp_prv_exg_daemon_add_handler (theMessageP->params.register_handler_notification.itsVerb,
																										theMessageP->params.register_handler_notification.itsLocalizedVerb,
																										theMessageP->params.register_handler_notification.itsMIMEType,
																										ALP_EXG_DAEMON_HANDLER_INVOKE_BY_NOTIFICATION,
																										theMessageP->params.register_handler_notification.itsHandlerNotificationType,
																										theMessageP->params.register_handler_notification.itsLocalOnly,
																										theMessageP->params.register_handler_notification.itsSerializeMe,
																										theMessageP->params.register_handler_notification.itsMustBeUnique,
																										theMessageP->params.register_handler_notification.itsAcceptGenericInvocation);
							iCxnState->itsTempBufUsed = 0;
							iCxnState->itsPollFD.events = G_IO_OUT;
							iCxnState->itsState = ALP_EXG_DAEMON_STATE_SEND_FINAL_REPLY_STATUS;
						} break;
						
						case ALP_EXG_DAEMON_MSG_OPCODE_UNREGISTER_HANDLER :
						{
							ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: Received UN-REGISTRATION for %s / %s / %s",
															theMessageP->params.unregister_handler.itsVerb, theMessageP->params.unregister_handler.itsMIMEType, theMessageP->params.unregister_handler.itsHandlerBundleName));
							theErr = alp_prv_exg_daemon_remove_handler (theMessageP->params.unregister_handler.itsVerb,
																										theMessageP->params.unregister_handler.itsMIMEType,
																										theMessageP->params.unregister_handler.itsHandlerBundleName);
							//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_prv_exg_daemon_remove_handler result = 0x%lx", theErr));
							// update the cache.
							if (theErr == 0)
							{
								theErr = prv_update_registration_cache();
							//	ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: prv_update_registration_cache result = 0x%lx", theErr));
							}
							
							((AlpExgMgrdMessageReplyType *) iCxnState->itsTempBuf)->itsReplyStatus = theErr;
							iCxnState->itsTempBufUsed = 0;
							iCxnState->itsPollFD.events = G_IO_OUT;
							iCxnState->itsState = ALP_EXG_DAEMON_STATE_SEND_FINAL_REPLY_STATUS;
						} break;
						
						case ALP_EXG_DAEMON_MSG_OPCODE_UNREGISTER_HANDLER_NOTIFICATION :
						{
							ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: Received NOTIFICATION UN-REGISTRATION for %s / %s / %s",
															theMessageP->params.unregister_handler_notification.itsVerb, theMessageP->params.unregister_handler_notification.itsMIMEType, theMessageP->params.unregister_handler_notification.itsHandlerNotificationType));
							((AlpExgMgrdMessageReplyType *) iCxnState->itsTempBuf)->itsReplyStatus = alp_prv_exg_daemon_remove_handler_notification (theMessageP->params.unregister_handler_notification.itsVerb,
																										theMessageP->params.unregister_handler_notification.itsMIMEType,
																										theMessageP->params.unregister_handler_notification.itsHandlerNotificationType);
							iCxnState->itsTempBufUsed = 0;
							iCxnState->itsPollFD.events = G_IO_OUT;
							iCxnState->itsState = ALP_EXG_DAEMON_STATE_SEND_FINAL_REPLY_STATUS;
						} break;
						
						case ALP_EXG_DAEMON_MSG_OPCODE_CHECK_HANDLER_REGISTRATION :
						{
							((AlpExgMgrdMessageReplyType *) iCxnState->itsTempBuf)->itsReplyStatus = alp_prv_exg_daemon_check_handler_registration (theMessageP->params.check_handler_registration.itsVerb,
																										theMessageP->params.check_handler_registration.itsMIMEType);
							iCxnState->itsTempBufUsed = 0;
							iCxnState->itsPollFD.events = G_IO_OUT;
							iCxnState->itsState = ALP_EXG_DAEMON_STATE_SEND_FINAL_REPLY_STATUS;
						} break;
						
						case ALP_EXG_DAEMON_MSG_OPCODE_QUERY_GENERIC_VERBS :
						{
							ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: Received ALP_EXG_DAEMON_MSG_OPCODE_QUERY_GENERIC_VERBS for %s",
															theMessageP->params.query_generic_verbs.itsMIMEType));
							((AlpExgMgrdMessageReplyType *) iCxnState->itsTempBuf)->itsReplyStatus = alp_prv_exg_daemon_query_generic_verbs (theMessageP->params.query_generic_verbs.itsMIMEType,
																										&iCxnState->params.queryGenericVerbsCxn.itsDataSize,
																										&iCxnState->params.queryGenericVerbsCxn.itsData);
							iCxnState->itsTempBufUsed = 0;
							iCxnState->itsPollFD.events = G_IO_OUT;
							if (((AlpExgMgrdMessageReplyType *) iCxnState->itsTempBuf)->itsReplyStatus != 0)
								iCxnState->itsState = ALP_EXG_DAEMON_STATE_SEND_FINAL_REPLY_STATUS;
							else
								iCxnState->itsState = ALP_EXG_DAEMON_STATE_SEND_GENERIC_VERBS_REPLY_STATUS;
						} break;
												
						case ALP_EXG_DAEMON_MSG_OPCODE_RECEIVE_REQUEST :
						{
							AlpExgMgrdDispatchInfoType *theDispatchInfoP;
							//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: Received ALP_EXG_DAEMON_MSG_OPCODE_RECEIVE_REQUEST connection"));
							
							// The handler is responding to the invocation we made and is passing back the request id in the message.
							// Find the request in the pending dispatch list.
							theDispatchInfoP = g_alp_exg_daemon_dispatch_info_list_head;
							while (theDispatchInfoP)
							{
								// itsDispatchInProgress was set when the handler was invoked.
								if (theDispatchInfoP->itsDispatchInProgress == 1 && iCxnState->params.handlerCxn.itsDispatchInfoP == NULL && 
									theDispatchInfoP->itsRequest == (AlpExgRequest) theMessageP->params.receive_request.itsRequestId)
									{
										//int theGuid;
										//theErr = alp_exg_request_get_int_parameter (theDispatchInfoP->itsRequest, ALP_EXGMGR_REQUEST_GLOBAL_UID_PARAM_NAME, &theGuid);
										//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: Found pending dispatch request GUID = %ld", theGuid));
										
										// found the dispatch entry waiting for this connection
										// Associate it with this connection.
										//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: Found pending dispatch request"));
										iCxnState->params.handlerCxn.itsDispatchInfoP = theDispatchInfoP;
										
										// clear the timeout source waiting for this handler connection
										if (theDispatchInfoP->itsTimeoutSourceId != 0)
										{
											g_source_remove (theDispatchInfoP->itsTimeoutSourceId);
											theDispatchInfoP->itsTimeoutSourceId = 0;
										}
										
										// Set the default error that will be returned to transport if communication between handler and daemon fails
										iCxnState->params.handlerCxn.itsReplyStatus = ALP_EXG_MGR_ERR_HANDLER_COMMUNICATION_ERROR;
										((AlpExgMgrdMessageReplyType *) iCxnState->itsTempBuf)->itsReplyStatus = 0;	// store reply status
										iCxnState->itsTempBufUsed = 0;
										iCxnState->itsPollFD.events = G_IO_OUT;
										iCxnState->itsState = ALP_EXG_DAEMON_STATE_SEND_RCV_REQ_REPLY_STATUS;
										break;
									}
								theDispatchInfoP = theDispatchInfoP->itsNext;
								if (theDispatchInfoP == NULL)
								{
									// no pending dispatch request. It may have timed out...
									ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: DID NOT FOUND pending dispatch request"));
									((AlpExgMgrdMessageReplyType *) iCxnState->itsTempBuf)->itsReplyStatus = ALP_EXG_MGR_ERR_NO_SUCH_REQUEST_ERROR;
									iCxnState->itsTempBufUsed = 0;
									iCxnState->itsPollFD.events = G_IO_OUT;
									iCxnState->itsState = ALP_EXG_DAEMON_STATE_SEND_FINAL_REPLY_STATUS;
								}
							}
						} break;
						
						case ALP_EXG_DAEMON_MSG_OPCODE_IS_ALIVE:
						{
							//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: Received ALP_EXG_DAEMON_MSG_OPCODE_IS_ALIVE"));
							((AlpExgMgrdMessageReplyType *) iCxnState->itsTempBuf)->itsReplyStatus = 0;
							iCxnState->itsTempBufUsed = 0;
							iCxnState->itsPollFD.events = G_IO_OUT;
							iCxnState->itsState = ALP_EXG_DAEMON_STATE_SEND_FINAL_REPLY_STATUS;
						} break;
						
						case ALP_EXG_DAEMON_MSG_OPCODE_GET_GUID:
						{
							// return a system wide unique uid to put in a request and track it across processes.
							((AlpExgMgrdMessageReplyType *) iCxnState->itsTempBuf)->itsReplyStatus = 0;
							((AlpExgMgrdMessageReplyType *) iCxnState->itsTempBuf)->params.get_request_guid.itsRequestGuid = gNextGuidValue++;
							iCxnState->itsTempBufUsed = 0;
							iCxnState->itsPollFD.events = G_IO_OUT;
							iCxnState->itsState = ALP_EXG_DAEMON_STATE_SEND_FINAL_REPLY_STATUS;						
						} break;

						case ALP_EXG_DAEMON_MSG_OPCODE_CANCEL_REQUEST:
						{
							ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: Received ALP_EXG_DAEMON_MSG_OPCODE_CANCEL_REQUEST for guid = %ld", theMessageP->params.cancel_request.itsGuid));
							((AlpExgMgrdMessageReplyType *) iCxnState->itsTempBuf)->itsReplyStatus = alp_prv_exg_daemon_cancel_request (theMessageP->params.cancel_request.itsGuid);
							iCxnState->itsTempBufUsed = 0;
							iCxnState->itsPollFD.events = G_IO_OUT;
							iCxnState->itsState = ALP_EXG_DAEMON_STATE_SEND_FINAL_REPLY_STATUS;
						} break;
						
						default :
							// unknown opcode : clean and exit (SHOUDLD WE RETURN AN ERROR TO CLIENT, LIKE UNKNOWN OPCODE ?)
							ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: current state invalid"));
							goto exitTerminateEntry;
							break;
					}
				}
			}
			break;
			
		case ALP_EXG_DAEMON_STATE_SEND_RCV_REQ_REPLY_STATUS:
		case ALP_EXG_DAEMON_STATE_SEND_GENERIC_VERBS_REPLY_STATUS:
		case ALP_EXG_DAEMON_STATE_SEND_FINAL_REPLY_STATUS:
			if (iCxnState->itsPollFD.revents & G_IO_OUT)
			{
				//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: got G_IO_OUT on socket %d (state = ALP_EXG_DAEMON_STATE_SEND_XXX_REPLY_STATUS)", iCxnState->itsPollFD.fd));
				theLen = write (iCxnState->itsPollFD.fd, iCxnState->itsTempBuf + iCxnState->itsTempBufUsed, sizeof (AlpExgMgrdMessageReplyType) - iCxnState->itsTempBufUsed);
				if (theLen < 0)
				{
					ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: write error on socket %d, ABORTING.", iCxnState->itsPollFD.fd));
					goto exitTerminateEntry;
				}
				iCxnState->itsTempBufUsed += theLen;
				if (iCxnState->itsTempBufUsed == sizeof (AlpExgMgrdMessageReplyType))
				{
					if (iCxnState->itsState == ALP_EXG_DAEMON_STATE_SEND_RCV_REQ_REPLY_STATUS)
					{
						// continue with sending the info len to the handler.
						alp_ext_request_prv_get_info (iCxnState->params.handlerCxn.itsDispatchInfoP->itsRequest, NULL, (size_t *) iCxnState->itsTempBuf);
						iCxnState->itsTempBufUsed = 0;
						iCxnState->itsPollFD.events = G_IO_OUT;
						iCxnState->itsState = ALP_EXG_DAEMON_STATE_SEND_REQUEST_DATA_SIZE;
					}
					else if (iCxnState->itsState == ALP_EXG_DAEMON_STATE_SEND_GENERIC_VERBS_REPLY_STATUS)
					{
						// continue with sending the size of the data to come
						* (size_t *) iCxnState->itsTempBuf = iCxnState->params.queryGenericVerbsCxn.itsDataSize;
						iCxnState->itsTempBufUsed = 0;
						iCxnState->itsPollFD.events = G_IO_OUT;
						iCxnState->itsState = ALP_EXG_DAEMON_STATE_SEND_GENERIC_VERBS_SIZE;
					}
					else	// this was a final status, terminate entry
					{
						// make sure all data is sent before the socket is closed from this side.
						shutdown (iCxnState->itsPollFD.fd, SHUT_RDWR);
						goto exitTerminateEntry;
					}
				}
			} break;

		case ALP_EXG_DAEMON_STATE_SEND_GENERIC_VERBS_SIZE:
			if (iCxnState->itsPollFD.revents & G_IO_OUT)
			{
				//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: got G_IO_OUT on socket %d (state = ALP_EXG_DAEMON_STATE_SEND_GENERIC_VERBS_SIZE)", iCxnState->itsPollFD.fd));
				theLen = write (iCxnState->itsPollFD.fd, iCxnState->itsTempBuf + iCxnState->itsTempBufUsed, sizeof (size_t) - iCxnState->itsTempBufUsed);
				if (theLen < 0)
				{
					ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: write error on socket %d, ABORTING.", iCxnState->itsPollFD.fd));
					goto exitTerminateEntry;
				}
				iCxnState->itsTempBufUsed += theLen;
				if (iCxnState->itsTempBufUsed == sizeof (size_t))
				{
					// continue with sending the verbs and descriptions data
					iCxnState->itsTempBufUsed = 0;
					iCxnState->itsPollFD.events = G_IO_OUT;
					iCxnState->itsState = ALP_EXG_DAEMON_STATE_SEND_GENERIC_VERBS;
				}
			} break;
				
		case ALP_EXG_DAEMON_STATE_SEND_GENERIC_VERBS:
			if (iCxnState->itsPollFD.revents & G_IO_OUT)
			{
				//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: got G_IO_OUT on socket %d (state = ALP_EXG_DAEMON_STATE_SEND_GENERIC_VERBS)", iCxnState->itsPollFD.fd));
				theLen = write (iCxnState->itsPollFD.fd, iCxnState->params.queryGenericVerbsCxn.itsData + iCxnState->itsTempBufUsed, iCxnState->params.queryGenericVerbsCxn.itsDataSize - iCxnState->itsTempBufUsed);
				if (theLen < 0)
				{
					ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: write error on socket %d, ABORTING.", iCxnState->itsPollFD.fd));
					goto exitTerminateEntry;
				}
				iCxnState->itsTempBufUsed += theLen;
				if (iCxnState->itsTempBufUsed == iCxnState->params.queryGenericVerbsCxn.itsDataSize)
				{
					shutdown (iCxnState->itsPollFD.fd, SHUT_RDWR);
					goto exitTerminateEntry;
				}
			} break;		
		
		case ALP_EXG_DAEMON_STATE_SEND_REQUEST_DATA_SIZE:
			if (iCxnState->itsPollFD.revents & G_IO_OUT)
			{
				//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: got G_IO_OUT on socket %d (state = ALP_EXG_DAEMON_STATE_SEND_REQUEST_DATA_SIZE)", iCxnState->itsPollFD.fd));
				theLen = write (iCxnState->itsPollFD.fd, iCxnState->itsTempBuf + iCxnState->itsTempBufUsed, sizeof (size_t) - iCxnState->itsTempBufUsed);
				if (theLen < 0)
				{
					ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: write error on socket %d, ABORTING.", iCxnState->itsPollFD.fd));
					goto exitTerminateEntry;
				}
				iCxnState->itsTempBufUsed += theLen;
				if (iCxnState->itsTempBufUsed == sizeof (size_t))
				{
					// continue with sending the info data to the handler.
					iCxnState->itsTempBufUsed = 0;
					iCxnState->itsPollFD.events = G_IO_OUT;
					iCxnState->itsState = ALP_EXG_DAEMON_STATE_SEND_REQUEST_DATA;
				}
			} break;
		
		case ALP_EXG_DAEMON_STATE_SEND_REQUEST_DATA:
			if (iCxnState->itsPollFD.revents & G_IO_OUT)
			{
				char *theInfo;
				size_t theInfoLen;
				alp_ext_request_prv_get_info (iCxnState->params.handlerCxn.itsDispatchInfoP->itsRequest, &theInfo, &theInfoLen);
				//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: got G_IO_OUT on socket %d (state = ALP_EXG_DAEMON_STATE_SEND_REQUEST_DATA)", iCxnState->itsPollFD.fd));
				theLen = write (iCxnState->itsPollFD.fd, theInfo + iCxnState->itsTempBufUsed, theInfoLen - iCxnState->itsTempBufUsed);
				if (theLen < 0)
				{
					ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: write error on socket %d, ABORTING.", iCxnState->itsPollFD.fd));
					goto exitTerminateEntry;
				}
				iCxnState->itsTempBufUsed += theLen;
				if (iCxnState->itsTempBufUsed == theInfoLen)
				{
					// continue with sending the FD present indication to the handler.
					// See if there is a dataFd. Remember alp_exg_request_get_data_fd gives us ownership of the fd, so give it back immediately
					int theDataFd;
					(void) alp_exg_request_get_data_fd (iCxnState->params.handlerCxn.itsDispatchInfoP->itsRequest, &theDataFd);
					* (int *) iCxnState->itsTempBuf = (theDataFd == -1 ? 0 : 1);
					(void) alp_exg_request_set_data_fd (iCxnState->params.handlerCxn.itsDispatchInfoP->itsRequest, theDataFd);	// it's OK if fd is -1
					iCxnState->itsTempBufUsed = 0;
					iCxnState->itsPollFD.events = G_IO_OUT;
					iCxnState->itsState = ALP_EXG_DAEMON_STATE_SEND_REQUEST_FD_PRESENT;
				}
			} break;
		
		case ALP_EXG_DAEMON_STATE_SEND_REQUEST_FD_PRESENT:
			if (iCxnState->itsPollFD.revents & G_IO_OUT)
			{
				//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: got G_IO_OUT on socket %d (state = ALP_EXG_DAEMON_STATE_SEND_REQUEST_FD_PRESENT)", iCxnState->itsPollFD.fd));
				theLen = write (iCxnState->itsPollFD.fd, iCxnState->itsTempBuf + iCxnState->itsTempBufUsed, sizeof (int) - iCxnState->itsTempBufUsed);
				if (theLen < 0)
				{
					ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: write error on socket %d, ABORTING.", iCxnState->itsPollFD.fd));
					goto exitTerminateEntry;
				}
				iCxnState->itsTempBufUsed += theLen;
				if (iCxnState->itsTempBufUsed == sizeof (int))
				{
					// Send the data FD if it exists
					int theDataFd;
					(void) alp_exg_request_get_data_fd(iCxnState->params.handlerCxn.itsDispatchInfoP->itsRequest, &theDataFd);
					if (theDataFd != -1)
					{
						if (alp_exg_prv_write_fd (iCxnState->itsPollFD.fd, iCxnState->itsTempBuf, 1, theDataFd) == -1)
						{
							ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: write_fd failed with error (0x%08lX).", errno));
							close (theDataFd);
							goto exitTerminateEntry;
						}
				
						// a copy of the dataFd now belongs to the handler side. Close ours here.
						close (theDataFd);
					}
					
					// Wait for the handler to send back a reply status, and a possible result request.
					iCxnState->itsTempBufUsed = 0;
					iCxnState->itsPollFD.events = G_IO_IN;
					iCxnState->itsState = ALP_EXG_DAEMON_STATE_WAIT_HANDLER_REPLY;
				}
			} break;
		
		case ALP_EXG_DAEMON_STATE_WAIT_HANDLER_REPLY:
			if (iCxnState->itsPollFD.revents & G_IO_IN)
			{
				// wait for the handler execution status.
				//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: got G_IO_IN on socket %d (state = ALP_EXG_DAEMON_STATE_WAIT_HANDLER_REPLY)", iCxnState->itsPollFD.fd));
				theLen = read (iCxnState->itsPollFD.fd, iCxnState->itsTempBuf + iCxnState->itsTempBufUsed, sizeof (alp_status_t) - iCxnState->itsTempBufUsed);
				if (theLen < 0)
				{
					ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: read error on socket %d, ABORTING.", iCxnState->itsPollFD.fd));
					goto exitTerminateEntry;
				}
				iCxnState->itsTempBufUsed += theLen;
				if (iCxnState->itsTempBufUsed == sizeof (alp_status_t))
				{
					iCxnState->itsTempBufUsed = 0;
					// if the handler returned a zero status, go to next state waiting for data size, else set status and terminate dispatch.
					if (* (alp_status_t *) iCxnState->itsTempBuf == 0)
						iCxnState->itsState = ALP_EXG_DAEMON_STATE_WAIT_REQUEST_DATA_SIZE;
					else
					{
						// Note : the special ALP_EXG_MGR_ERR_HANDLER_DID_NOT_CONSUME_ERROR error case is treated in exitTerminateEntry
						iCxnState->params.handlerCxn.itsReplyStatus = * (alp_status_t *) iCxnState->itsTempBuf;
						goto exitTerminateEntry;
					}
				}
				// dont test hangup or error if we had data to read
				if (theLen > 0)
					return 0;
			} break;
				
		case ALP_EXG_DAEMON_STATE_WAIT_REQUEST_DATA_SIZE:
			if (iCxnState->itsPollFD.revents & G_IO_IN)
			{
				//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: got G_IO_IN on socket %d (state = ALP_EXG_DAEMON_STATE_WAIT_REQUEST_DATA_SIZE)", iCxnState->itsPollFD.fd));
				theLen = read (iCxnState->itsPollFD.fd, iCxnState->itsTempBuf + iCxnState->itsTempBufUsed, sizeof (iCxnState->params.handlerCxn.itsInfoLen2) - iCxnState->itsTempBufUsed);
				if (theLen < 0)
				{
					ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: read error on socket %d, ABORTING.", iCxnState->itsPollFD.fd));
					goto exitTerminateEntry;
				}
				iCxnState->itsTempBufUsed += theLen;
				if (iCxnState->itsTempBufUsed == sizeof (iCxnState->params.handlerCxn.itsInfoLen2))
				{
					// We have the data size in the first bytes of iCxnState->itsTempBuf
					memcpy (&iCxnState->params.handlerCxn.itsInfoLen2, iCxnState->itsTempBuf, sizeof (iCxnState->params.handlerCxn.itsInfoLen2));
					// make sure the request we build can accept this much info, and get the pointer to the info buffer
					theErr = alp_exg_request_prv_alloc_init_info (iCxnState->params.handlerCxn.itsDispatchInfoP->itsRequest, iCxnState->params.handlerCxn.itsInfoLen2);
					if (theErr)
					{
						iCxnState->params.handlerCxn.itsReplyStatus = theErr;
						goto exitTerminateEntry;
					}
					alp_ext_request_prv_get_info(iCxnState->params.handlerCxn.itsDispatchInfoP->itsRequest, &iCxnState->params.handlerCxn.itsInfoP, NULL);
					iCxnState->itsTempBufUsed = 0;
					iCxnState->itsState = ALP_EXG_DAEMON_STATE_WAIT_REQUEST_DATA;
				}
				// dont test hangup or error if we had data to read
				if (theLen > 0)
					return 0;
			}
			break;
			
		case ALP_EXG_DAEMON_STATE_WAIT_REQUEST_DATA:
			if (iCxnState->itsPollFD.revents & G_IO_IN)
			{
				// read the info data
				//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: got G_IO_IN on socket %d (state = ALP_EXG_DAEMON_STATE_WAIT_REQUEST_DATA)", iCxnState->itsPollFD.fd));
				theLen = read (iCxnState->itsPollFD.fd, iCxnState->params.handlerCxn.itsInfoP + iCxnState->itsTempBufUsed, iCxnState->params.handlerCxn.itsInfoLen2 - iCxnState->itsTempBufUsed);
				if (theLen < 0)
				{
					ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: read error on socket %d, ABORTING.", iCxnState->itsPollFD.fd));
					goto exitTerminateEntry;
				}
				iCxnState->itsTempBufUsed += theLen;
				if (iCxnState->itsTempBufUsed == iCxnState->params.handlerCxn.itsInfoLen2)
				{
					// we have all the data in the request info field, set the len.
					alp_exg_request_prv_set_info (iCxnState->params.handlerCxn.itsDispatchInfoP->itsRequest, iCxnState->params.handlerCxn.itsInfoP, iCxnState->params.handlerCxn.itsInfoLen2);
					iCxnState->itsTempBufUsed = 0;
					iCxnState->itsState = ALP_EXG_DAEMON_STATE_WAIT_REQUEST_FD_PRESENT;
					// it is possible that there is no data at all so theLen would be zero. dont test hangup when going to next state
					return 0;
				}
				// dont test hangup or error if we had data to read
				if (theLen > 0)
					return 0;
			}
			break;
	
		case ALP_EXG_DAEMON_STATE_WAIT_REQUEST_FD_PRESENT:
			if (iCxnState->itsPollFD.revents & G_IO_IN)
			{
				//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: got G_IO_IN on socket %d (state = ALP_EXG_DAEMON_STATE_WAIT_REQUEST_FD_PRESENT)", iCxnState->itsPollFD.fd));
				theLen = read (iCxnState->itsPollFD.fd, iCxnState->itsTempBuf + iCxnState->itsTempBufUsed, sizeof (int) - iCxnState->itsTempBufUsed);
				if (theLen < 0)
				{
					ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: read error on socket %d, ABORTING.", iCxnState->itsPollFD.fd));
					goto exitTerminateEntry;
				}
				iCxnState->itsTempBufUsed += theLen;
				if (iCxnState->itsTempBufUsed == sizeof (int))
				{
					// We have the FdPresent int in the first bytes of iCxnState->itsTempBuf
					if (*(int *) iCxnState->itsTempBuf == 0)
					{
						// no fd, we're done. Send the result back to transport
						iCxnState->params.handlerCxn.itsReplyStatus = 0;
						goto exitTerminateEntry;
					}
					else
					{
						// go get the fd.
						iCxnState->itsState = ALP_EXG_DAEMON_STATE_WAIT_REQUEST_FD;
					}
				// dont test hangup or error if we had data to read
				if (theLen > 0)
					return 0;
				}
			}
			break;
			
		case ALP_EXG_DAEMON_STATE_WAIT_REQUEST_FD:
			if (iCxnState->itsPollFD.revents & G_IO_IN)
			{
				int theFd;
				//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: got G_IO_IN on socket %d (state = ALP_EXG_DAEMON_STATE_WAIT_REQUEST_FD)", iCxnState->itsPollFD.fd));
				// alp_exg_prv_read_fd must return 1 byte of data read
				if (alp_exg_prv_read_fd (iCxnState->itsPollFD.fd, iCxnState->itsTempBuf, 1, &theFd) != 1)
				{
					ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: read_fd error on socket %d, ABORTING.", iCxnState->itsPollFD.fd));
					goto exitTerminateEntry;
				}
				
				// Give dataFd to the request, and we're done. Send the result back to transport
				alp_exg_request_set_data_fd (iCxnState->params.handlerCxn.itsDispatchInfoP->itsRequest, theFd);
				iCxnState->params.handlerCxn.itsReplyStatus = 0;
				goto exitTerminateEntry;
			}
			break;
			

		default:
			ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: current state invalid"));
			goto exitTerminateEntry;
			break;
	}
	
	// check for connection hangup or error. The connection is always closed by out side when done,
	// so this is always an error
	if (iCxnState->itsPollFD.revents & G_IO_ERR)
	{
		ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: got G_IO_ERR on socket %d", iCxnState->itsPollFD.fd));
		goto exitTerminateEntry;
	}
	
	// ((iCxnState->itsPollFD.revents & G_IO_IN) == 0) was added to ignore HUP while we have data to read. This was done to make sure we
	// receive all the data before we eventually treat the HUP.
	// However, it turns out that we always get G_IO_IN with G_IO_HUP, so that HUP would never be treated when using these two conditions.
	// The second condition is removed for now. Let's see if we happen to detect HUP before we had a chance to read all the data....
	if (iCxnState->itsPollFD.revents & G_IO_HUP /* && ((iCxnState->itsPollFD.revents & G_IO_IN) == 0)*/)
	{
		ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: got G_IO_HUP on socket %d", iCxnState->itsPollFD.fd));
		goto exitTerminateEntry;
	}
	
	// return error only if fatal unrecoverable error encoutered.
	return 0;

exitTerminateEntry:

	if (iCxnState->itsState != ALP_EXG_DAEMON_STATE_WAIT_CXN_MESSAGE)
	{
		// If this is a handler connection and we have already associated the pending dispatch request, send the result back to transport.
		// and free the dispatch entry.
		if (iCxnState->itsOpcode == ALP_EXG_DAEMON_MSG_OPCODE_RECEIVE_REQUEST)
		{
			if (iCxnState->params.handlerCxn.itsDispatchInfoP != NULL)
			{
				int doDispatchAgain = 0;
				// If handler did not consume request, and handler is not unique (itsLastHandlerInfoP != NULL), reinitialize request
				// and leave it in the queue to be dispatched again to next handler.
				if (iCxnState->params.handlerCxn.itsReplyStatus == ALP_EXG_MGR_ERR_HANDLER_DID_NOT_CONSUME_ERROR &&
					iCxnState->params.handlerCxn.itsDispatchInfoP->itsLastHandlerInfoP != NULL)
				{
					int theFd;
					//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: Handler did not consume request. Leaving request in queue for dispatch to next handler."));
					iCxnState->params.handlerCxn.itsDispatchInfoP->itsDispatchInProgress = 0;
					// if there is a tmp file, open it and give the fd to the request (note that is there was a non consumed fd in the request,
					// it would be closed when replaced by the one we set)
					if (iCxnState->params.handlerCxn.itsDispatchInfoP->itsTmpDatafile != NULL)
					{
						theFd = open (iCxnState->params.handlerCxn.itsDispatchInfoP->itsTmpDatafile, O_RDONLY);
						if (theFd >= 0)
						{
							iCxnState->params.handlerCxn.itsReplyStatus = alp_exg_request_set_data_fd (iCxnState->params.handlerCxn.itsDispatchInfoP->itsRequest, theFd);
							if (iCxnState->params.handlerCxn.itsReplyStatus == 0)
								doDispatchAgain = 1;
							else
								close (theFd);
						}
						else
							iCxnState->params.handlerCxn.itsReplyStatus = ALP_EXGMGR_ERR_CANNOT_OPEN_TMP_FILE_ERROR;
					}
					else
					{
						// There is no data with this request, we can redispatch directly.
						doDispatchAgain = 1;
					}
				}

				// terminate request if we don't redispatch
				if (doDispatchAgain == 0)
				{
					//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: Sending dispatch result back to transport, replystatus = 0x%lx", iCxnState->params.handlerCxn.itsReplyStatus));
					// Send the handler result status and modified request back to the transport.
					g_alp_exg_daemon_transports[iCxnState->params.handlerCxn.itsDispatchInfoP->itsTransportId].itsTransportDispatchResultFunc (iCxnState->params.handlerCxn.itsDispatchInfoP->itsRequest, iCxnState->params.handlerCxn.itsReplyStatus);
					// and remove the whole dispatch entry from its queue
					alp_prv_exg_daemon_dequeue_dispatch_entry (iCxnState->params.handlerCxn.itsDispatchInfoP);
					free (iCxnState->params.handlerCxn.itsDispatchInfoP);
				}
			}
			theErr = alp_prv_exg_daemon_run_dispatcher();
			if (theErr)
			{
				ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: cannot re-initiate dispatcher 0x%lx", theErr));
			}
		}
		else if (iCxnState->itsOpcode == ALP_EXG_DAEMON_MSG_OPCODE_QUERY_GENERIC_VERBS)
		{
			if (iCxnState->params.queryGenericVerbsCxn.itsData)
				free (iCxnState->params.queryGenericVerbsCxn.itsData);
		}
	}

	// We're done with this connection
	//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_run_connection: Terminating connection socket %d", iCxnState->itsPollFD.fd));
	g_source_remove_poll (g_alp_exg_daemon_event_source, &iCxnState->itsPollFD);
	close (iCxnState->itsPollFD.fd);
	iCxnState->isUsed = 0;

	return 0;
}

/* -------------------------------------------------------------------------------------------
 * alp_exg_daemon_source_func_prepare */
static gboolean alp_exg_daemon_source_func_prepare (GSource *source, gint *timeout)
{
	//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_source_func_prepare: called"));

	*timeout = -1;

	return FALSE;
}

/* -------------------------------------------------------------------------------------------
 * alp_exg_daemon_source_func_check */
static gboolean alp_exg_daemon_source_func_check (GSource *source)
{
	int gotRevent = 0;
	int theIndex;
	
	// check listener : this one gets connection from the exgmgr library, either from an exgmgr client or handler
	if (g_alp_exg_daemon_GPollFd.revents != 0)
	{
		gotRevent = 1;
		/*
		if (g_alp_exg_daemon_GPollFd.revents & G_IO_IN)
			ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_source_func_check: got G_IO_IN on listener"));
		if (g_alp_exg_daemon_GPollFd.revents & G_IO_OUT)
			ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_source_func_check: got G_IO_OUT on listener"));
		if (g_alp_exg_daemon_GPollFd.revents & (G_IO_HUP | G_IO_ERR))
			ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_source_func_check: got G_IO_HUP or G_IO_ERR on listener"));
		*/
	}
	
	// check events for each connection from client or handler
	for (theIndex = 0; theIndex < ALP_EXG_DAEMON_MAX_CONNECTIONS; theIndex++)
		if (g_alp_exg_daemon_cxn_state[theIndex].isUsed != 0)
			if (g_alp_exg_daemon_cxn_state[theIndex].itsPollFD.revents != 0)
			{
				gotRevent = 1;
				/*
				if (g_alp_exg_daemon_cxn_state[theIndex].itsPollFD.revents & G_IO_IN)
					ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_source_func_check: got G_IO_IN on cxn %d socket %d", theIndex, g_alp_exg_daemon_cxn_state[theIndex].itsPollFD.fd));
				if (g_alp_exg_daemon_cxn_state[theIndex].itsPollFD.revents & G_IO_OUT)
					ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_source_func_check: got G_IO_OUT on cxn %d socket %d", theIndex, g_alp_exg_daemon_cxn_state[theIndex].itsPollFD.fd));
				if (g_alp_exg_daemon_cxn_state[theIndex].itsPollFD.revents & (G_IO_HUP | G_IO_ERR))
					ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_source_func_check: got G_IO_HUP or G_IO_ERR on cxn %d socket %d", theIndex, g_alp_exg_daemon_cxn_state[theIndex].itsPollFD.fd));
				*/
			}

	if (gotRevent)
		return TRUE;

	// no trace was output in this case.
	//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_source_func_check: called but no event detected"));	
	return FALSE;
}

/* -------------------------------------------------------------------------------------------
 * alp_exg_daemon_source_func_dispatch */
static gboolean alp_exg_daemon_source_func_dispatch (GSource *source, GSourceFunc callback, gpointer user_data)
{	
	alp_status_t theErr;
	int theIndex;
	int32_t flags = 0;
	
	//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_source_func_dispatch: called"));
	
	// check the listener. This receives connections from the handler or client through exgmgr lib.
	if (g_alp_exg_daemon_GPollFd.revents & G_IO_IN)
	{
		//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_source_func_dispatch: INCOMING CXN detected on listener"));
		
		for (theIndex = 0; theIndex < ALP_EXG_DAEMON_MAX_CONNECTIONS; theIndex++)
			if (g_alp_exg_daemon_cxn_state[theIndex].isUsed == 0)
			{
				//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_source_func_dispatch: allocate CxnState entry %d", theIndex));
				
				// accept the connection and set it non blocking
				g_alp_exg_daemon_cxn_state[theIndex].itsPollFD.fd = accept (g_alp_exg_daemon_GPollFd.fd, NULL, NULL);
				if (g_alp_exg_daemon_cxn_state[theIndex].itsPollFD.fd != -1)
					if ((flags = fcntl(g_alp_exg_daemon_cxn_state[theIndex].itsPollFD.fd, F_GETFL, 0)) == -1)
					{
						ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_source_func_dispatch: fcntl failed with error (%lu).", (uint32_t) errno));
						close (g_alp_exg_daemon_cxn_state[theIndex].itsPollFD.fd);
						g_alp_exg_daemon_cxn_state[theIndex].itsPollFD.fd = -1;
						break;
					}
				if (g_alp_exg_daemon_cxn_state[theIndex].itsPollFD.fd != -1)
					if (fcntl(g_alp_exg_daemon_cxn_state[theIndex].itsPollFD.fd, F_SETFL, (flags | O_NONBLOCK)) == -1)
					{
						ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_source_func_dispatch: fcntl failed with error (%lu).", (uint32_t) errno));
						close (g_alp_exg_daemon_cxn_state[theIndex].itsPollFD.fd);
						g_alp_exg_daemon_cxn_state[theIndex].itsPollFD.fd = -1;
						break;
					}

				// OK, connection accepted and configured,.
				if (g_alp_exg_daemon_cxn_state[theIndex].itsPollFD.fd != -1)
				{
					// Read the connection message sent by the client or handler, and determine what this connection is for...
					// The message is an AlpExgMgrdMessageType and the itsOpcode field tells us what to do
					// the itsState field will be set to reflect this.
					g_alp_exg_daemon_cxn_state[theIndex].isUsed = 1;
					g_alp_exg_daemon_cxn_state[theIndex].itsState = ALP_EXG_DAEMON_STATE_WAIT_CXN_MESSAGE;
					g_alp_exg_daemon_cxn_state[theIndex].itsTempBufUsed = 0;
					g_alp_exg_daemon_cxn_state[theIndex].itsOpcode = 0;
					memset (&g_alp_exg_daemon_cxn_state[theIndex].params, 0, sizeof(g_alp_exg_daemon_cxn_state[theIndex].params));
					g_alp_exg_daemon_cxn_state[theIndex].itsPollFD.events = G_IO_IN;
					g_alp_exg_daemon_cxn_state[theIndex].itsPollFD.revents = 0;	// reset this because this entry may be processed before the next poll.
					g_source_add_poll (g_alp_exg_daemon_event_source, &g_alp_exg_daemon_cxn_state[theIndex].itsPollFD);
				}
				else
				{
					// cannot accept.
					ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_source_func_dispatch: ACCEPT FAILED"));
					break;
				}
				break;
			}
	}

	// error or hangup on the listener : not good ....
	if (g_alp_exg_daemon_GPollFd.revents & (G_IO_HUP | G_IO_ERR))
	{
		ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_source_func_dispatch: HANGUP or ERROR on listener, ABORTING daemon"));

		// TODO: Listener dead : daemon must be terminated. How should this be handled with respect to the user?
	}

	// check the connections.
	for (theIndex = 0; theIndex < ALP_EXG_DAEMON_MAX_CONNECTIONS; theIndex++)
		if (g_alp_exg_daemon_cxn_state[theIndex].isUsed != 0)
			if (g_alp_exg_daemon_cxn_state[theIndex].itsPollFD.revents != 0)
			{
				//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_source_func_dispatch: got FD EVENT on cxn %d socket %d", theIndex, g_alp_exg_daemon_cxn_state[theIndex].itsPollFD.fd));
				theErr = alp_exg_daemon_run_connection (g_alp_exg_daemon_cxn_state + theIndex);
				if (theErr)
				{
					// use itsState to determine what this connection is doing. Error handling depends on the current state.
					
					// state is ALP_EXG_DAEMON_WAIT_CXN_MESSAGE -> close connection and free entry. If this is from a handler, the request will remain pending
					//											   and abord on timeout.
					
					// See CxnState.itsOpcode to determine what this connection is doing.
					
					
					
					
					// state is ALP_EXG_DAEMON_MSG_OPCODE_REGISTER_HANDLER or ALP_EXG_DAEMON_MSG_OPCODE_UNREGISTER_HANDLER -> close connection and free entry
					//											   Client will detect error.
				}
			}

	return TRUE;
}

/* -------------------------------------------------------------------------------------------
 * alp_exg_daemon_source_func_finalize */
static void alp_exg_daemon_source_func_finalize (GSource *source)
{
	//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: alp_exg_daemon_source_func_finalize: called"));
}

/* -------------------------------------------------------------------------------------------
 *  */
int alp_exg_mgr_daemon (int argc, char *argv[])
{
	struct sigaction theSigAction;
	int theIndex;
	char thePidBuf[10];
	alp_status_t theErr;
	struct sockaddr_un	theAddr;
	AlpAppContext *theAlpAppContext;
	guint theArrayIndex;
	int daemonReleaseDone = 0;
	GArray *theArray = NULL;
	mode_t theMask;

	ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : Starting..."));
	
	alp_daemon_start (0, 1);

	gNextGuidValue = EXG_MGR_FIRST_GUID;

	theErr = alp_bundle_init ();
	if (theErr)
	{
		ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : alp_bundle_init returned error = 0x%lx", theErr));
		goto exitErr;
	}
	
	// Enumerate transports
	g_alp_exg_daemon_transports_count = 0;
	g_alp_exg_daemon_transports = NULL;
	theArray = alp_bundle_search(NULL, 0, NULL, NULL, "exchange");
	if (theArray == NULL)
	{
		ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : alp_bundle_search returned NULL array"));
		theErr = ALP_EXGMGR_ERR_OUT_OF_MEMORY_ERROR;
		goto fini_bundle;
	}
	ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : alp_bundle_search found %d transports", theArray->len));
	
	// if no transports, exit
	if (theArray->len == 0)
	{
		ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : No transport found, exiting."));
		g_array_free (theArray, 1);
		theErr = ALP_EXG_MGR_ERR_NO_SUCH_TRANSPORT_ERROR;
		goto fini_bundle;
	}
	
	//	Allocate the transport table and clear it
	g_alp_exg_daemon_transports = (AlpExgMgrdTransportInfoType *) malloc (theArray->len * sizeof (AlpExgMgrdTransportInfoType));
	if (g_alp_exg_daemon_transports == NULL)
	{
		ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : cannot allocate AlpExgMgrdTransportInfoType table"));
		g_array_free (theArray, 1);
		theErr = ALP_EXGMGR_ERR_OUT_OF_MEMORY_ERROR;
		goto fini_bundle;
	}
	memset(g_alp_exg_daemon_transports, 0, theArray->len * sizeof (AlpExgMgrdTransportInfoType));
	
	// load each transport
	for (theArrayIndex = 0; theArrayIndex < theArray->len; theArrayIndex++)
	{
		gchar *theLibNameP;
		gchar *theFullPath = NULL;

		// Get bundle index
		AlpBundle theBundle = g_array_index(theArray, AlpBundle, theArrayIndex);
		if (alp_bundle_is_NULL (theBundle))
		{
			ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : Found a NULL AlpBundle in the array returned by alp_bundle_search()"));
			g_array_free (theArray, 1);
			theErr = ALP_EXGMGR_ERR_DAEMON_ERROR;
			goto cleanTransportTable;
		}

		// open it. Keep it open until the daemon exit.
		g_alp_exg_daemon_transports[g_alp_exg_daemon_transports_count].itsBundleRef = alp_bundle_open (theBundle);
		if (g_alp_exg_daemon_transports[g_alp_exg_daemon_transports_count].itsBundleRef == 0)
		{
			ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : cannot open Bundle"));
			g_array_free (theArray, 1);
			theErr = ALP_EXGMGR_ERR_DAEMON_ERROR;
			goto cleanTransportTable;
		}
	
		// Find the library name from the bundle Manifest, and build the full path
		theLibNameP = alp_bundle_property_value (theBundle, "exchange", 0, "library");
		//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : Library name = %s", theLibNameP));
		if (theLibNameP == NULL)
		{
			ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : cannot get bundle property 'library'"));
			g_array_free (theArray, 1);
			theErr = ALP_EXGMGR_ERR_DAEMON_ERROR;
			goto cleanTransportTable;
		}
		theFullPath = alp_bundle_ref_ro_pathname(g_alp_exg_daemon_transports[g_alp_exg_daemon_transports_count].itsBundleRef, theLibNameP);
		g_free (theLibNameP);
		if (theFullPath == NULL)
		{
			ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : cannot get transport fullpath"));
			g_array_free (theArray, 1);
			theErr = ALP_EXGMGR_ERR_DAEMON_ERROR;
			goto cleanTransportTable;
		}
		
		//ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : Library Path = %s", theFullPath));
		// Open the lib, init all fields
		g_alp_exg_daemon_transports[g_alp_exg_daemon_transports_count].itsLibRef = dlopen (theFullPath, RTLD_LAZY);
		if (g_alp_exg_daemon_transports[g_alp_exg_daemon_transports_count].itsLibRef == 0)
		{
			ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : CANNOT DLOPEN LIBRARY : %s", theFullPath));
			g_free (theFullPath);
			g_array_free (theArray, 1);
			theErr = ALP_EXGMGR_ERR_DAEMON_ERROR;
			goto cleanTransportTable;
		}
		g_free (theFullPath);
		
		// Get the transport internal name
		g_alp_exg_daemon_transports[g_alp_exg_daemon_transports_count].itsNameId = alp_bundle_property_value (theBundle, "exchange", 0, "nameid");
		if (g_alp_exg_daemon_transports[g_alp_exg_daemon_transports_count].itsNameId == NULL)
		{
			ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : cannot get bundle property 'nameid'"));
			g_array_free (theArray, 1);
			theErr = ALP_EXGMGR_ERR_DAEMON_ERROR;
			goto cleanTransportTable;
		}
		ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : Found Transport NameId = %s", g_alp_exg_daemon_transports[g_alp_exg_daemon_transports_count].itsNameId));

		// Count this lib as loaded
		g_alp_exg_daemon_transports_count++;
	}
	
	// Don't need this array anymore
	g_array_free (theArray, 1);

	// Get the functions pointers for each transport
	for (theIndex = 0; theIndex < g_alp_exg_daemon_transports_count; theIndex++)
	{
		if ((g_alp_exg_daemon_transports[theIndex].itsTransportInitFunc = (alp_exg_transport_init_func) dlsym (g_alp_exg_daemon_transports[theIndex].itsLibRef, "alp_exg_transport_init")) == NULL)
		{
			ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : cannot get transport alp_exg_transport_init function ptr"));
			theErr = ALP_EXGMGR_ERR_DAEMON_ERROR;
			goto cleanTransportTable;
		}
		if ((g_alp_exg_daemon_transports[theIndex].itsTransportDeinitFunc = (alp_exg_transport_deinit_func) dlsym (g_alp_exg_daemon_transports[theIndex].itsLibRef, "alp_exg_transport_deinit")) == NULL)
		{
			ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : cannot get transport alp_exg_transport_deinit function ptr"));
			theErr = ALP_EXGMGR_ERR_DAEMON_ERROR;
			goto cleanTransportTable;
		}
		if ((g_alp_exg_daemon_transports[theIndex].itsTransportDispatchResultFunc = (alp_exg_transport_dispatch_result_func) dlsym (g_alp_exg_daemon_transports[theIndex].itsLibRef, "alp_exg_transport_dispatch_result")) == NULL)
		{
			ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : cannot get transport alp_exg_transport_dispatch_result function ptr"));
			theErr = ALP_EXGMGR_ERR_DAEMON_ERROR;
			goto cleanTransportTable;
		}
		if ((g_alp_exg_daemon_transports[theIndex].itsTransportCancelFunc = (alp_exg_transport_cancel_func) dlsym (g_alp_exg_daemon_transports[theIndex].itsLibRef, "alp_exg_transport_cancel")) == NULL)
		{
			ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : cannot get transport alp_exg_transport_cancel function ptr"));
			theErr = ALP_EXGMGR_ERR_DAEMON_ERROR;
			goto cleanTransportTable;
		}
	}

	// Init GLib main loop
	g_alp_exg_daemon_main_loop = g_main_loop_new(NULL, 0);
	if (g_alp_exg_daemon_main_loop == NULL)
	{
		ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : Cannot create g_alp_exg_daemon_main_loop"));
		theErr = ALP_EXGMGR_ERR_DAEMON_ERROR;
		goto cleanTransportTable;
	}

	// Install a signal handler for SIGQUIT
	theSigAction.sa_handler = alp_prv_exg_daemon_SIGTERM_Handler;
	theSigAction.sa_flags = 0;
	theSigAction.sa_restorer = NULL;
	theSigAction.sa_mask.__val[0] = 0;
	sigaction (SIGTERM, &theSigAction, NULL);
	
	// init all incoming connections
	memset (g_alp_exg_daemon_cxn_state, 0, sizeof (g_alp_exg_daemon_cxn_state));

	// init the list of registered handlers
	g_alp_exg_daemon_handler_info_list_head = NULL;
	g_alp_exg_daemon_handler_info_list_tail = NULL;

	// check if the exgmgr cache folder exist. If it exists, then the cache file MUST exist or we exit to force a clean reboot
	if (access (PRV_HANDLER_CONFIG_DIR, R_OK) == 0)
	{
		// read registration cache and load the registered handler table.
		theErr = prv_load_handlers_config (PRV_HANDLER_CONFIG_FILE);
		if (theErr)
		{
			ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : Cannot load registered handlers cache data"));
			goto cleanTransportTable;
		}
	}

	// init the list of incoming request being dispatched
	g_alp_exg_daemon_dispatch_info_list_head = NULL;
	g_alp_exg_daemon_dispatch_info_list_tail = NULL;
	
	// initialize transports plug-in This will make them install
	// their Glib event sources along with their event handlers.
	for (theIndex = 0; theIndex < g_alp_exg_daemon_transports_count; theIndex++)
	{
		theErr = g_alp_exg_daemon_transports[theIndex].itsTransportInitFunc(alp_prv_exg_transport_event_notify, theIndex);
		if (theErr)
		{
			ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : Cannot Initialize transport %s", g_alp_exg_daemon_transports[theIndex].itsNameId));
			goto cleanDeinitTransports;
		}
		g_alp_exg_daemon_transports[theIndex].itsInited = 1;
	}

	// Initialize the listener to accept connections from handlers (AFUNIX sockets)
	if ((g_alp_exg_daemon_socket_listener = socket(AF_UNIX, SOCK_STREAM, 0)) < 0)
	{
		ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: Cannot create daemon socket, error (%ld).", (int) errno));
		theErr = ALP_EXGMGR_ERR_DAEMON_ERROR;
		goto cleanDeinitTransports;
	}
				
	// Bind socket
	memset(&theAddr, 0, sizeof(theAddr));
	theAddr.sun_family = AF_UNIX;
	strcpy (theAddr.sun_path, ALP_EXG_DAEMON_SOCKET_NAME);
	unlink(theAddr.sun_path);
	
	if (bind(g_alp_exg_daemon_socket_listener, (struct sockaddr*) &theAddr, strlen(theAddr.sun_path) + sizeof(theAddr.sun_family)) < 0)
	{
		ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: daemon socket bind failed with error (%ld).", (int) errno));
		theErr = ALP_EXGMGR_ERR_DAEMON_ERROR;
		goto cleanListenerExit;
	}

	// let everybody connect to this socket
	theMask = umask(0);
	chmod(ALP_EXG_DAEMON_SOCKET_NAME, S_IRWXU | S_IRWXG | S_IRWXO);
	umask(theMask);  // restore original umask
	
	// Listen the socket
	if (listen(g_alp_exg_daemon_socket_listener, 5) < 0)
	{
		ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: damon socket listen failed with error (%ld).", (int) errno));
		theErr = ALP_EXGMGR_ERR_DAEMON_ERROR;
		goto cleanListenerExit;
	}

	// Install GLib callbacks to let us know when something happens on this socket.
	g_alp_exg_daemon_event_source_funcs.prepare = alp_exg_daemon_source_func_prepare;
	g_alp_exg_daemon_event_source_funcs.check = alp_exg_daemon_source_func_check;
	g_alp_exg_daemon_event_source_funcs.dispatch = alp_exg_daemon_source_func_dispatch;
	g_alp_exg_daemon_event_source_funcs.finalize = alp_exg_daemon_source_func_finalize;
	g_alp_exg_daemon_event_source_funcs.closure_callback = NULL;
	g_alp_exg_daemon_event_source_funcs.closure_marshal = NULL;
	g_alp_exg_daemon_event_source = g_source_new (&g_alp_exg_daemon_event_source_funcs, sizeof (GSource));
	if (g_alp_exg_daemon_event_source == NULL)
	{
		ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: g_source_new failed with error (%ld).", (int) errno));
		theErr = ALP_EXGMGR_ERR_DAEMON_ERROR;
		goto cleanListenerExit;
	}
	// Add the listener fd to the GSource (cannot fail)
	g_alp_exg_daemon_GPollFd.fd = g_alp_exg_daemon_socket_listener;
	g_alp_exg_daemon_GPollFd.events = G_IO_IN;
	g_source_add_poll (g_alp_exg_daemon_event_source, &g_alp_exg_daemon_GPollFd);

	// Add the source to the default context
	g_alp_exg_daemon_event_source_id = g_source_attach (g_alp_exg_daemon_event_source, NULL);
	if (g_alp_exg_daemon_event_source_id <= 0)
	{
		ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: g_source_attach failed with error (%ld).", (int) errno));
		theErr = ALP_EXGMGR_ERR_DAEMON_ERROR;
		goto cleanPollFd;
	}

	// register with the app server
	theAlpAppContext = alp_app_setup ("com.access.alpexgmgrd", ALP_APP_LAUNCH_FLAGS_DEFAULT, NULL);
	if (theAlpAppContext == NULL)
	{
		ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon: CANNOT CONNECT TO APP SERVER"));
		theErr = ALP_EXGMGR_ERR_DAEMON_ERROR;
		goto cleanPollFd;
	}

	// init done, let boot script continue
	alp_daemon_release (0);
	daemonReleaseDone = 1;

	// Run. We'll get out of here when we receive a SIGTERM.
	g_main_loop_run (g_alp_exg_daemon_main_loop);

	// disconnect from app server
	if (theAlpAppContext)
		alp_app_destroy (theAlpAppContext, 0);
	
	// terminate pending dispatch requests
	alp_prv_exg_daemon_terminate_dispatcher ();
	
	// terminate all connections with handlers (they will detect a connection hangup and abort) and return abort result to transports.
	for (theIndex = 0; theIndex < ALP_EXG_DAEMON_MAX_CONNECTIONS; theIndex++)
		if (g_alp_exg_daemon_cxn_state[theIndex].isUsed != 0)
		{
		// TODO: not very important. When the daemon exits, all handlers will get a connection broken error
		// and abort gracefully.
		}

cleanPollFd:
	// remove the listener pollfd
	g_source_remove_poll (g_alp_exg_daemon_event_source, &g_alp_exg_daemon_GPollFd);

cleanGSourceExit:
	g_source_destroy (g_alp_exg_daemon_event_source);

cleanListenerExit:
	close (g_alp_exg_daemon_socket_listener);

cleanDeinitTransports:
	for (theIndex = 0; theIndex < g_alp_exg_daemon_transports_count; theIndex++)
		if (g_alp_exg_daemon_transports[theIndex].itsInited == 1)
			(void) g_alp_exg_daemon_transports[theIndex].itsTransportDeinitFunc ();

cleanTransportTable:
	if (g_alp_exg_daemon_transports != NULL)
	{
		for (theIndex = 0; theIndex < g_alp_exg_daemon_transports_count; theIndex++)
		{
			if (g_alp_exg_daemon_transports[theIndex].itsNameId)
				g_free (g_alp_exg_daemon_transports[theIndex].itsNameId);
			
			if (g_alp_exg_daemon_transports[theIndex].itsLibRef != 0)
				(void) dlclose (g_alp_exg_daemon_transports[theIndex].itsLibRef);	

			if (g_alp_exg_daemon_transports[theIndex].itsBundleRef != 0)
				alp_bundle_close (g_alp_exg_daemon_transports[theIndex].itsBundleRef);
		}
		free (g_alp_exg_daemon_transports);
	}

fini_bundle:
	alp_bundle_fini();

exitErr:
	if (daemonReleaseDone == 0)
		alp_daemon_release (theErr);
	
	ALP_TM(TL(ALP_CLASS_EXCHANGE, "ExgMgr Daemon : Exiting with error = 0x%lx...", theErr));
	return theErr;
}

/* -------------------------------------------------------------------------------------------
 *  */
int main (int argc, char *argv[])
{
	return alp_exg_mgr_daemon (argc, argv);
}
