/*! @file exgtransport.h 
 *
 * @brief Exchange Manager reserved transport plug-in headers
 *
*/
#ifndef EXGTRANSPORT_H_
#define EXGTRANSPORT_H_

#include <hiker/exgmgr.h>

#ifdef __cplusplus
 extern "C"
 {
#endif

// Prototypes for functions each transport library must implement

#define ALP_EXG_TRANSPORTS_MAX_COUNT 5
#define ALP_EXG_TRANSPORT_NAME_MAX_LEN 30	// Transport nameId max len

/*! @brief Transport events that can be notified back to the daemon */
//@{
#define ALP_EXG_TRANSPORT_EVENT_INCOMING_REQUEST 1	//!< transport has an incoming ready to dispatch
#define ALP_EXG_TRANSPORT_EVENT_FATAL_ERROR 2 		//!< transport encountered an unrecoverable error
//@}

typedef alp_status_t (*alp_exg_transport_event_notify_func) (int iTransportId, int iTransportEvent, void *iTransportEventData, size_t iTransportEventDataSize);
typedef alp_status_t (*alp_exg_transport_init_func) (alp_exg_transport_event_notify_func iTransportEventNotifyFunc, int iTransportId);
typedef alp_status_t (*alp_exg_transport_deinit_func) ();
typedef alp_status_t (*alp_exg_transport_dispatch_result_func) (AlpExgRequest iRequest, alp_status_t iHandlerErr);
typedef alp_status_t (*alp_exg_transport_execute_request_func) (AlpExgRequest iRequest, GPollFD *iPollFd, int *iState);
typedef alp_status_t (*alp_exg_transport_ask_params_func) (char **oTransportParams);
typedef alp_status_t (*alp_exg_transport_cancel_func) (int iGuid);

// Each asynchronous execute request maintains this:
typedef struct _AlpExgClientExecuteData
{
	int isUsed;
	alp_exg_execute_complete_func itsExecuteCompleteFunc;
	AlpExgRequest itsRequestP;
	alp_exg_transport_execute_request_func itsTransportExecuteRequestFunc;
	int itsState;
	GPollFD itsPollFd;
} AlpExgClientExecuteData;

// This is the global GSource used to manage callbacks when executing asynchronous request.
// The EXG_MAX_ASYNC_EXEC_COUNT limit is a per process limit. 
#define EXG_MAX_ASYNC_EXEC_COUNT 3
typedef struct _AlpExgClientExecGSource
{
	GSource itsGSource;
	AlpExgClientExecuteData itsExecuteData[EXG_MAX_ASYNC_EXEC_COUNT];	// table
} AlpExgClientExecGSource;

// state returned in AlpExgClientExecuteData.itsState when execution is complete
#define ALP_EXG_TRANSPORT_EXECUTE_INIT_STATE 0
#define ALP_EXG_TRANSPORT_EXECUTE_COMPLETE_STATE 0xffff

// Common structure that can be used by transports to send result or progress info while waiting for request execution
#define ALP_EXG_MSG_RESULT 1
#define ALP_EXG_MSG_PROGRESS_PREPARING 2
#define ALP_EXG_MSG_PROGRESS_CONNECTING 3
#define ALP_EXG_MSG_PROGRESS_SENDING 4
#define ALP_EXG_MSG_PROGRESS_DISCONNECTING 5
#define ALP_EXG_MSG_PROGRESS_ABORTING 6

#define ALP_EXG_MSG_RX_BT_CONNECTING 10		// bdaddr
#define ALP_EXG_MSG_RX_IR_CONNECTING 11
#define ALP_EXG_MSG_RX_BT_INIT_OBJECT 12	// bdaddr, null
#define ALP_EXG_MSG_RX_IR_INIT_OBJECT 13
#define ALP_EXG_MSG_RX_BT_FILENAME 14			// bdaddr, filename
#define ALP_EXG_MSG_RX_IR_FILENAME 15		// filename
#define ALP_EXG_MSG_RX_BT_RECEIVING_INFO 16	// bdaddr, description
#define ALP_EXG_MSG_RX_IR_RECEIVING_INFO 17	// description
#define ALP_EXG_MSG_RX_BT_RECEIVING_INFO_NODESC 18	// bdaddr, null
#define ALP_EXG_MSG_RX_IR_RECEIVING_INFO_NODESC 19

#define ALP_EXG_RCV_PROGRESS_NOTIF_TYPE "exg_obex_rcv_progress"

#define ALP_EXG_MSG_PROGRESS_INFO_MAX_LEN 50

typedef struct _AlpExgWaitExecuteResultMsg
{
	int itsMsgType;
	union
	{
		struct _result
		{
			alp_status_t itsReplyStatus;
		} result;
		
		struct _progress
		{
			char itsString1[ALP_EXG_MSG_PROGRESS_INFO_MAX_LEN+1];
			char itsString2[ALP_EXG_MSG_PROGRESS_INFO_MAX_LEN+1];
		} progress;
	} params;
} AlpExgWaitExecuteResultMsg;

/*! @brief Prototype of the daemon function to be called back when the transport has an event to notify
 
 @param[in] iTransportId our transport Id, given by daemon at init time
 @param[in] iTransportEvent 
 @param[in] iTransportEventData
 @param[in] iTransportEventDataSize the number of bytes passed in iTransportEventData

 @return
	#ALP_EXGMGR_ERR_NO_ERROR - success<br>
*/
alp_status_t alp_exg_transport_event_notify (int iTransportId, int iTransportEvent, void *iTransportEventData, size_t iTransportEventDataSize);

/*! @brief Initialize the transport

 This function runs in the context of the daemon. It must install its event sources in the
 GTK event loop, in order to be called back as necessary.
 The function callback passed must be called when an incoming request has been formed and is
 ready to be dispatched, or an unrecoverable error occurs in the transport.
 If this function returns an error, alp_exg_transport_deinit WILL NOT be called.
 
 @param[in] iRequestReadyFuncP daemon callback to dispatch an incoming request
 @param[in] iTransportId transport Id assigned by daemon, to be passed back with incoming request

 @return
	#ALP_EXGMGR_ERR_NO_ERROR - success<br>
*/
alp_status_t alp_exg_transport_init (alp_exg_transport_event_notify_func iTransportEventNotifyFunc, int iTransportId);

/*! @brief Uninitialize the transport

 This function runs in the context of the daemon. It is called right before closing the plug-in, and only if
 alp_exg_transport_init returned no error.
 
 @return
	#ALP_EXGMGR_ERR_NO_ERROR - success<br>
*/
alp_status_t alp_exg_transport_deinit ();

// function called when a request has been dispatched and result is available
/*! @brief Receive the request dispatch result

 This function runs in the context of the daemon. It is called when the handler has completed
 execution and returned the result. The transport gets the handler result and may send
 the request back to the initiator.
 
 @param[in] iRequest the request that must be transmitted
 @param[in] iHandlerErr the result of the handler execution

 @return
	#ALP_EXGMGR_ERR_NO_ERROR - success<br>
*/
alp_status_t alp_exg_transport_dispatch_result (AlpExgRequest iRequest, alp_status_t iHandlerErr);

/*! @brief Transmit request to destination

 This function runs in the context of the request initiator
 
 @param[in] iRequest the request that must be transmitted

 @return
	#ALP_EXGMGR_ERR_NO_ERROR - success<br>
*/
alp_status_t alp_exg_transport_execute_request (AlpExgRequest iRequest, GPollFD *iPollFd, int *iState);

/*! @brief Builds a full transport string

 This function builds a full transport parameter string, using UI if necessary to ask for missing parameters.
 
 @param[out] oTransportParams a pointer to a string that will be allocated by this function and returned. Caller must free it when done.

 @return
	#ALP_EXGMGR_ERR_NO_ERROR - success<br>
*/
alp_status_t alp_exg_transport_ask_params (char **oTransportParams);

/*! @brief Cancel a request executing in the transport

 This function is only usefull for transports that get requests from the client side and directly execute or send them
 to a remote destination. Its purpose is to cancel the currently executing request.
 
 @param[in] iGuid is a unique id that is contained in the request to be canceled.

 @return
	#ALP_EXGMGR_ERR_NO_ERROR - success<br>
	#ALP_EXG_MGR_ERR_NO_SUCH_REQUEST_ERROR - transport is not currectly executing a request with tihs guid
*/
alp_status_t alp_exg_transport_cancel (int iGuid);

#ifdef __cplusplus
 }
#endif

#endif /*EXGTRANSPORT_H_*/
