/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
#include <hiker/notifymgr.h>
#include <hiker/prv/notifymgr_prv.h>
#include "notify_mgr_server_prv.h"

// notification handler timeout
#define ALP_NOTIFY_WAIT_TIMEOUT 30

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>

// broadcast entries keeps track on what appIDs are currently have active notifications
// this prevents multiple notifications to the same appID
static AlpNotifyBroadcastEntry * sBroadcastList = 0;

// mutex
static pthread_mutex_t sMutex = PTHREAD_MUTEX_INITIALIZER;

// allocate a new broadcast entry
static AlpNotifyBroadcastEntry * alp_prv_notify_new(const char * appID, int pid)
{
	AlpNotifyBroadcastEntry * entry;

	entry = (AlpNotifyBroadcastEntry *)malloc(sizeof(AlpNotifyBroadcastEntry));
	if (!entry) return 0;

	entry->appID = malloc(strlen(appID)+1);
	if (!entry->appID) {
		free (entry);
		return 0;
	}

	entry->pid = pid;
	time(&entry->seconds);
	strcpy(entry->appID,appID);
	entry->next = 0;

	return entry;
}

// check all active notifications. make sure they aren't dead and haven't been running too long.
// if they are then perform some cleanup.
void alp_notify_cleanup(void)
{
	AlpNotifyBroadcastEntry * entry, * previous, * tmp;
	time_t current;

#ifdef ALP_NOTIFY_DEBUG
	printf("alp_notify_cleanup\n");
#endif
	// lock critical section
	pthread_mutex_lock(&sMutex);

	// get current time
	time(&current);

	// search all the entry
	entry = sBroadcastList;
	previous = 0;
	while (entry) {

		// check if pid exists
		if ((entry->pid != -1) && (kill(entry->pid,0) == -1)) {

#ifdef ALP_NOTIFY_DEBUG
			printf("alp_notify_cleanup: process does not exist: pid=%d\n",entry->pid);
#endif
			// remove any registered notifications
			alp_notify_register_remove_all(entry->appID,1);

			// check if the first item should be deleted
			tmp = entry;
			if (!previous) {
				entry = sBroadcastList = sBroadcastList->next;
			} else {
				previous->next = entry->next;
				entry = entry->next;
			}

			// free entry
			free(tmp->appID);
			free(tmp);

		} else {

			// check if the notification is taking too long
			if ((entry->pid != -1) && (current - entry->seconds > ALP_NOTIFY_WAIT_TIMEOUT)) {
#ifdef ALP_NOTIFY_DEBUG
				printf("alp_notify_cleanup: too long: pid=%d\n",entry->pid);
#endif
				// remove any registered notifications
				alp_notify_register_remove_all(entry->appID,1);

				// kill the process (it may be hung)
				kill(entry->pid,9);

				// check if the first item should be deleted
				tmp = entry;
				if (!previous) {
					entry = sBroadcastList = sBroadcastList->next;
				} else {
					previous->next = entry->next;
					entry = entry->next;
				}

				// free entry
				free(tmp->appID);
				free(tmp);

			} else {

				// goto next entry
				previous = entry;
				entry = entry->next;
			}
		}
	}

	// unlock critical section
	pthread_mutex_unlock(&sMutex);
}

// add a broadcast entry
int alp_notify_add(const char * appID, int pid)
{
	AlpNotifyBroadcastEntry * tmp, * new;

#ifdef ALP_NOTIFY_DEBUG
	printf("alp_notify_add appID=%s pid=%d\n",appID,pid);
#endif
	// allocate new notification
	new = alp_prv_notify_new(appID,pid);
	if (!new) return ALP_NOTIFY_ERR_OUT_OF_MEMORY;

	// lock critical section
	pthread_mutex_lock(&sMutex);

	// check if the list is empty
	if (!sBroadcastList) {

		// if empty just add one item
		sBroadcastList = new;

	} else {

		// otherwise add at the beginning of the list
		tmp = sBroadcastList;
		sBroadcastList = new;
		new->next = tmp;
	}

	// unlock critical section
	pthread_mutex_unlock(&sMutex);

	return 0;
}

// remove a broadcast entry
void alp_notify_remove(const char * appID)
{
	AlpNotifyBroadcastEntry * entry, * previous, * tmp;

#ifdef ALP_NOTIFY_DEBUG
	printf("alp_notify_remove appID=%s\n",appID);
#endif
	// lock critical section
	pthread_mutex_lock(&sMutex);

	// search all the entry
	entry = sBroadcastList;
	previous = 0;
	while (entry) {

		// check appID matches
		if (strcmp(appID,entry->appID) == 0) {

			// check if the first item should be deleted
			tmp = entry;
			if (!previous) {
				entry = sBroadcastList = sBroadcastList->next;
			} else {
				previous->next = entry->next;
				entry = entry->next;
			}

			// free entry
			free(tmp->appID);
			free(tmp);

		} else {

			// goto next entry
			previous = entry;
			entry = entry->next;
		}
	}

	// unlock critical section
	pthread_mutex_unlock(&sMutex);
}

// get the pid for a given appID
int alp_notify_get_pid(const char * appID)
{
	AlpNotifyBroadcastEntry * entry;

#ifdef ALP_NOTIFY_DEBUG
	printf("alp_notify_get_pid appID=%s\n",appID);
#endif
	// lock critical section
	pthread_mutex_lock(&sMutex);

	// search existing connections
	entry = sBroadcastList;
	while (entry) {

		// check for a match
		if (strcmp(entry->appID,appID) == 0) {
			pthread_mutex_unlock(&sMutex);
			return entry->pid;
		}

		// otherwise go to the next item
		entry = entry->next;
	}

	// unlock critical section
	pthread_mutex_unlock(&sMutex);

	return 0;
}
