/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/**
 * @file notify_server_database.c
 *
 * @brief Notify Manager server database implementation
 *
 *****************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>

#include <glib.h>
#include <glib/gstdio.h>

#include <hiker/syspaths.h>
#include <hiker/notifymgr.h>

#include <hiker/prv/notifymgr_prv.h>
#include <hiker/pdk/notifymgr_pdk.h>

#include "notify_mgr_server_prv.h"

#define ALP_NOTIFY_DATABASE_DIR  ALP_SYSPATH_SYSTEM_CACHE "notifymgr"
#define ALP_NOTIFY_DATABASE_FILE "notify_mgr.db"
#define ALP_NOTIFY_DATABASE_MODE (S_ISGID|S_ISVTX|S_IRWXU|S_IRWXG)

// ----------------------------------------------------------------------------
// sqlite3 specific code
// ----------------------------------------------------------------------------

#include <sqlite3.h>

// callback to handle each row from "SELECT ALL * FROM register"
// where each row is a persistent registration
static int alp_prv_notify_sqlite_restore_callback(void * ref, int ncol, char ** cols, char ** name)
{
	AlpNotifyCmdRegister request;

#ifdef ALP_NOTIFY_DEBUG
	printf("alp_prv_notify_sqlite_restore appID=%s notifyType=%s priority=%d flags=%d\n",*cols,*(cols+1),atoi(*(cols+2)),atoi(*(cols+3)));
#endif
	// fill out the message header
	memset(&request,0,sizeof(AlpNotifyCmdRegister));
	request.cmd = ALP_NOTIFY_CMD_REGISTER;
	request.size = sizeof(AlpNotifyCmdRegister);

	// fill out the message body
	strcpy(request.notifyType,*(cols+1));
	request.callback = 0;
    request.userData = 0;
	request.priority = atoi(*(cols+2));
	request.flags = atoi(*(cols+3)) | ALP_NOTIFY_REGISTER_FLAG_DONTSAVE | ALP_NOTIFY_REGISTER_FORCE;
	request.pid = 1;
	strcpy(request.appID,*cols);
	request.zero = 0;

	// register the app for the notification
	alp_notify_register_add(&request, 0);

	// should always return zero from the callback
	return 0;
}

// initialize the database
// create the register table
static int alp_prv_notify_sqlite_init(void)
{
	sqlite3 * db = 0;
	int err;
	char * mesg;

#ifdef ALP_NOTIFY_DEBUG
	printf("alp_prv_notify_sqlite_init\n");
#endif
	char *dbname = ALP_NOTIFY_DATABASE_DIR "/" ALP_NOTIFY_DATABASE_FILE;
    char *directory = ALP_NOTIFY_DATABASE_DIR;

    /* make sure database directory exists */
    if (!g_file_test (directory, G_FILE_TEST_IS_DIR)) {
#ifdef ALP_NOTIFY_DEBUG
        printf("alp_prv_notify_sqlite_init: "
               "directory %s does not exit or is not directory\n",
                directory);
#endif
        /* check if exists and if so, attempt to remove non-directory */
        if (g_file_test (directory, G_FILE_TEST_EXISTS)) {
            if (g_remove (directory) < 0) {
#ifdef ALP_NOTIFY_DEBUG
                printf("alp_prv_notify_sqlite_init: "
                       "unable to remove %s: %d %s\n",
                       directory, errno, strerror(errno));
#endif
                err = errno;
                goto Done;
            }
        }

        /* create directory */
        if (g_mkdir_with_parents (directory, ALP_NOTIFY_DATABASE_MODE) < 0) {
            /* unable to create directory */
#ifdef ALP_NOTIFY_DEBUG
            printf("alp_prv_notify_sqlite_init: "
                   "unable to create %s: %d %s\n",
                   directory, errno, strerror(errno));
#endif
            err = errno;
            goto Done;
        }
    }

	// open the database
	err = sqlite3_open(dbname,&db);
	if (err != SQLITE_OK) {
#ifdef ALP_NOTIFY_DEBUG
		printf("alp_prv_notify_sqlite_init: sqlite3_open err=%d\n",err);
#endif
		goto Done;
	}

	// create the table
	err = sqlite3_exec(db,"CREATE TABLE register (appID TEXT, notifyType TEXT, priority INT, flags INT)",0,0,&mesg);
	if (err != SQLITE_OK) {
#ifdef ALP_NOTIFY_DEBUG
		printf("alp_prv_notify_sqlite_init: sqlite3_exec err=%d\n",err);
#endif
		if (mesg) {
#ifdef ALP_NOTIFY_DEBUG
			printf("alp_prv_notify_sqlite_init: sqlite3_exec: %s\n",mesg);
#endif
			sqlite3_free(mesg);
		}
		goto Done;
	}
Done:
	// close the database
	if (db) sqlite3_close(db);

	// return status
	return err == SQLITE_OK ? ALP_OK : err;
}

// add a new registration to the database
// insert a row into the database which contains the registration information
static int alp_prv_notify_sqlite_add(const char * appID, const char * notifyType, int priority, int flags)
{
	sqlite3 * db = 0;
	int err;
	char * mesg;
	char str[1000];

#ifdef ALP_NOTIFY_DEBUG
	printf("alp_prv_notify_sqlite_add\n");
#endif
	char *dbname = ALP_NOTIFY_DATABASE_DIR "/" ALP_NOTIFY_DATABASE_FILE;

	// open the database
	err = sqlite3_open(dbname,&db);
	if (err != SQLITE_OK) {
#ifdef ALP_NOTIFY_DEBUG
		printf("alp_prv_notify_sqlite_add: sqlite3_open err=%d\n",err);
#endif
		goto Done;
	}

	// create the sql string
	sprintf(str,"INSERT INTO register VALUES ('%s','%s',%d,%d)",appID,notifyType,priority,flags);

	// add a row 
	err = sqlite3_exec(db,str,0,0,&mesg);
	if (err != SQLITE_OK) {
#ifdef ALP_NOTIFY_DEBUG
		printf("alp_prv_notify_sqlite_add: sqlite3_exec err=%d\n",err);
#endif
		if (mesg) {
#ifdef ALP_NOTIFY_DEBUG
			printf("alp_prv_notify_sqlite_add: sqlite3_exec: %s\n",mesg);
#endif
			sqlite3_free(mesg);
		}
		goto Done;
	}
Done:
	// close the database
	if (db) sqlite3_close(db);

	// return status
	return err == SQLITE_OK ? ALP_OK : err;
}

// remove a registration from the database
// delete a row from the database which contains the registration information
static int alp_prv_notify_sqlite_remove(const char * appID, const char * notifyType)
{
	sqlite3 * db = 0;
	int err;
	char * mesg;
	char str[1000];

#ifdef ALP_NOTIFY_DEBUG
	printf("alp_prv_notify_sqlite_remove\n");
#endif
	char *dbname = ALP_NOTIFY_DATABASE_DIR "/" ALP_NOTIFY_DATABASE_FILE;

	// open the database
	err = sqlite3_open(dbname,&db);
	if (err != SQLITE_OK) {
#ifdef ALP_NOTIFY_DEBUG
		printf("alp_prv_notify_sqlite_remove: sqlite3_open err=%d\n",err);
#endif
		goto Done;
	}

	// create the sql string
	sprintf(str,"DELETE FROM register WHERE (appID == '%s' AND notifyType == '%s')",appID,notifyType);

	// remove the row
	err = sqlite3_exec(db,str,0,0,&mesg);
	if (err != SQLITE_OK) {
#ifdef ALP_NOTIFY_DEBUG
		printf("alp_prv_notify_sqlite_remove: sqlite3_exec err=%d\n",err);
#endif
		if (mesg) {
#ifdef ALP_NOTIFY_DEBUG
			printf("alp_prv_notify_sqlite_remove: sqlite3_exec: %s\n",mesg);
#endif
			sqlite3_free(mesg);
		}
		goto Done;
	}
Done:
	// close the database
	if (db) sqlite3_close(db);

	// return status
	return err == SQLITE_OK ? ALP_OK : err;
}

// restore all previously save registrations
// select all rows and process each one in the callback
static int alp_prv_notify_sqlite_restore(void)
{
	sqlite3 * db = 0;
	int err;
	char * mesg;

#ifdef ALP_NOTIFY_DEBUG
	printf("alp_prv_notify_sqlite_restore\n");
#endif
	char *dbname = ALP_NOTIFY_DATABASE_DIR "/" ALP_NOTIFY_DATABASE_FILE;

	// open the database
	err = sqlite3_open(dbname,&db);
	if (err != SQLITE_OK) {
#ifdef ALP_NOTIFY_DEBUG
		printf("alp_prv_notify_sqlite_restore: sqlite3_open err=%d\n",err);
#endif
		goto Done;
	}

	// dump all the rows
	err = sqlite3_exec(db,"SELECT ALL * FROM register",alp_prv_notify_sqlite_restore_callback,0,&mesg);
	if (err != SQLITE_OK) {
#ifdef ALP_NOTIFY_DEBUG
		printf("alp_prv_notify_sqlite_restore: sqlite3_exec err=%d\n",err);
#endif
		if (mesg) {
#ifdef ALP_NOTIFY_DEBUG
			printf("alp_prv_notify_sqlite_restore: sqlite3_exec: %s\n",mesg);
#endif
			sqlite3_free(mesg);
		}
		goto Done;
	}
Done:
	// close the database
	if (db) sqlite3_close(db);

	// return status
	return err == SQLITE_OK ? ALP_OK : err;
}

// ----------------------------------------------------------------------------
// end of sqlite3 specific code
// ----------------------------------------------------------------------------

// initialize the database
int alp_notify_database_init(void)
{
#ifdef ALP_NOTIFY_DEBUG
	printf("alp_notify_database_init\n");
#endif

	// initialize the database
	return alp_prv_notify_sqlite_init();
}

// add a registration
int alp_notify_database_add(AlpNotifyRegisterEntry * entry)
{
#ifdef ALP_NOTIFY_DEBUG
	printf("alp_notify_database_add\n");
#endif

	// don't save server registrations
	if (entry->callback) return ALP_OK;

	// add the registration to the database
	return alp_prv_notify_sqlite_add(entry->appID, entry->notifyType, entry->priority, entry->flags);
}

// remove a registration
int alp_notify_database_remove(AlpNotifyRegisterEntry * entry)
{
#ifdef ALP_NOTIFY_DEBUG
	printf("alp_notify_database_remove\n");
#endif

	// remove the registration from the database
	return alp_prv_notify_sqlite_remove(entry->appID, entry->notifyType);
}

// restore all previously save registrations
int alp_notify_database_restore(void)
{
#ifdef ALP_NOTIFY_DEBUG
	printf("alp_notify_database_restore\n");
#endif

	return alp_prv_notify_sqlite_restore();
}
