/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
#include <hiker/notifymgr.h>
#include <hiker/prv/notifymgr_prv.h>
#include "notify_mgr_server_prv.h"

#include <hiker/ipc.h>
#include <hiker/pdk/sysutils_pdk.h>
#include <hiker/fail.h>

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>

static int sNotifyTimer = 0;
static AlpNotifyConnectionEntry * sConnectList = 0;

// allocate new connection entry
static AlpNotifyConnectionEntry * alp_prv_notify_connect_new(const char * appID, AlpConnection * connect)
{
	AlpNotifyConnectionEntry * entry;

#ifdef ALP_NOTIFY_DEBUG
	printf("alp_prv_notify_connect_new\n");
#endif
	entry = (AlpNotifyConnectionEntry *)malloc(sizeof(AlpNotifyConnectionEntry));
	if (!entry) return 0;

	entry->appID = malloc(strlen(appID)+1);
	if (!entry->appID) {
		free(entry);
		return 0;
	}

	entry->connect =  connect;
	strcpy(entry->appID,appID);
	entry->next = 0;

	return entry;
}

// add new connection
static int alp_prv_notify_connect_add(AlpNotifyCmdCallback * request, AlpConnection * connect)
{
	AlpNotifyConnectionEntry * tmp, * new;

#ifdef ALP_NOTIFY_DEBUG
	printf("alp_prv_notify_connect_add appID=%s connect=%x\n",request->appID,(int)connect);
#endif
	// check packet size
	if (request->size != sizeof(AlpNotifyCmdCallback))
		return ALP_NOTIFY_ERR_INVALID_CMD;

	// check packet data
	if (request->zero)
		return ALP_NOTIFY_ERR_INVALID_CMD;
	if (strlen(request->appID) > ALP_NOTIFY_MAX_APPID_SIZE)
		return ALP_NOTIFY_ERR_INVALID_APPID;

	// allocate a new connection entry
	new = alp_prv_notify_connect_new(request->appID,connect);
	if (!new) return ALP_NOTIFY_ERR_OUT_OF_MEMORY;

	// check if the list is empty
	if (!sConnectList) {

		// if empty just add one item
		sConnectList = new;

	} else {

		// otherwise add at the beginning of the list
		tmp = sConnectList;
		sConnectList = new;
		new->next = tmp;
	}

	return 0;
}

// remove connection
static void alp_prv_notify_connect_remove(AlpConnection * connect)
{
	AlpNotifyConnectionEntry * entry, * previous, * tmp;

#ifdef ALP_NOTIFY_DEBUG
	printf("alp_prv_notify_connect_remove connect=%x\n",(int)connect);
#endif
	// check if no connections
	if (!sConnectList) return;

	// search all the entry
	entry = sConnectList;
	previous = 0;
	while (entry) {

		// check if the AlpConnection matches
		if (entry->connect == connect) {

			// check if the first item should be deleted
			tmp = entry;
			if (!previous) {
				entry = sConnectList = sConnectList->next;
			} else {
				previous->next = entry->next;
				entry = entry->next;
			}

			// free entry
			free(tmp->appID);
			free(tmp);

		} else {

			// goto next entry
			previous = entry;
			entry = entry->next;
		}
	}
}

// get the connection for a given appID
static AlpConnection * alp_prv_notify_connect_get_connect(const char * appID)
{
	AlpNotifyConnectionEntry * entry;

#ifdef ALP_NOTIFY_DEBUG
	printf("alp_prv_notify_connect_get_connect appID=%s\n",appID);
#endif
	// check if no connections
	if (!sConnectList)
		return 0;

	// search existing connections
	entry = sConnectList;
	for (;;) {
		// check for a match
		if (!strcmp(entry->appID,appID)) {
#ifdef ALP_NOTIFY_DEBUG
	printf("alp_prv_notify_connect_get_connect found appID=%s connect=%x\n",entry->appID,(int)entry->connect);
#endif
			return entry->connect;
		}

		// check if at the end of the list
		if (!entry->next) break;

		// otherwise go to the next item
		entry = entry->next;
	}

	// not found
	return 0;
}

// get the appID for a given connection
static char * alp_prv_notify_connect_get_appID(AlpConnection * connect)
{
	AlpNotifyConnectionEntry * entry;

#ifdef ALP_NOTIFY_DEBUG
	printf("alp_prv_notify_connect_get_appID connect=%x\n",(int)connect);
#endif
	// search existing connections
	entry = sConnectList;
	while (entry) {
		// check for a match
		if (entry->connect == connect) {
#ifdef ALP_NOTIFY_DEBUG
	printf("alp_prv_notify_connect_get_appID found appID=%s connect=%x\n",entry->appID,(int)entry->connect);
#endif
			return entry->appID;
		}

		// otherwise go to the next item
		entry = entry->next;
	}

	// not found
	return 0;
}

// collection of periodic functions
static gint alp_prv_notify_server_timer(gpointer data)
{
#ifdef ALP_NOTIFY_DEBUG
	printf("alp_prv_notify_server_timer\n");
#endif
	// cleanup
	alp_notify_cleanup();

	// process pending broadcasts
	alp_notify_broadcast_do_pending();
	alp_notify_broadcast_cleanup();

	// keep timer going only if there are pending broadcasts
	sNotifyTimer = alp_notify_broadcast_get_pending() > 0; 
	return sNotifyTimer;
}

// start the timer to do periodic functions
void alp_notify_ipc_start_timer(void)
{
	if (!sNotifyTimer) {
		sNotifyTimer = 1;
		g_timeout_add(200,alp_prv_notify_server_timer,0);
	}
}

// parse incoming messages
static int alp_prv_notify_server_mesg(char * buf, int length, AlpConnection * connect)
{
	int status;

#ifdef ALP_NOTIFY_DEBUG
	printf("alp_prv_notify_server_mesg: length=%d\n",length);
#endif

	switch(*buf) {
		case ALP_NOTIFY_CMD_REGISTER:
                	ALP_FAIL_NON_FATAL_IF(length != sizeof(AlpNotifyCmdRegister),"bad AlpNotifyCmdRegister message\n");
			if (length != sizeof(AlpNotifyCmdRegister))
				return ALP_NOTIFY_ERR_INVALID_CMD;
			status = alp_notify_register_add((AlpNotifyCmdRegister *)buf,alp_ipc_connection_get_peer_pid(connect));
		break;
		case ALP_NOTIFY_CMD_UNREGISTER:
                	ALP_FAIL_NON_FATAL_IF(length != sizeof(AlpNotifyCmdUnregister),"bad AlpNotifyCmdUnregister message\n");
			if (length != sizeof(AlpNotifyCmdUnregister))
				return ALP_NOTIFY_ERR_INVALID_CMD;
			status = alp_notify_register_remove((AlpNotifyCmdUnregister *)buf);
		break;
		case ALP_NOTIFY_CMD_BROADCAST:
                	ALP_FAIL_NON_FATAL_IF(length != sizeof(AlpNotifyCmdBroadcast),"bad AlpNotifyCmdBroadcast message\n");
			if (length != sizeof(AlpNotifyCmdBroadcast))
				return ALP_NOTIFY_ERR_INVALID_CMD;
			status = alp_notify_broadcast_add((AlpNotifyCmdBroadcast *)buf,connect);

			// start timer function if there isn't already one present
			alp_notify_ipc_start_timer();
		break;
		case ALP_NOTIFY_CMD_NOTIFY_DONE:
                	ALP_FAIL_NON_FATAL_IF(length != sizeof(AlpNotifyCmdNotificationDone),"bad AlpNotifyCmdNotificationDone message\n");
			if (length != sizeof(AlpNotifyCmdNotificationDone))
				return ALP_NOTIFY_ERR_INVALID_CMD;
			status = alp_notify_broadcast_done((AlpNotifyCmdNotificationDone *)buf);
		break;
		case ALP_NOTIFY_CMD_CALLBACK:
                	ALP_FAIL_NON_FATAL_IF(length != sizeof(AlpNotifyCmdCallback),"bad AlpNotifyCmdCallback message\n");
			if (length != sizeof(AlpNotifyCmdCallback))
				return ALP_NOTIFY_ERR_INVALID_CMD;
			status = alp_prv_notify_connect_add((AlpNotifyCmdCallback *)buf, connect);
		break;
		default:
                	ALP_FAIL_NON_FATAL_IF(true,"unknown message\n");
#ifdef ALP_NOTIFY_DEBUG
			printf("alp_prv_notify_server_mesg: unknown command\n");
#endif
			status = ALP_NOTIFY_ERR_INVALID_CMD;
		break;
	}

	return status;
}

// receive callback function
static void alp_prv_notify_server_receive(AlpConnection* connect, AlpMessage * message, gpointer cbData)
{
	guchar * buf;
	gint length, status;

#ifdef ALP_NOTIFY_DEBUG
	printf("alp_prv_notify_server_receive %x\n",(int)connect);
#endif
	// unpack the message from the client
	alp_ipc_message_unpack_start(message);
	alp_ipc_message_unpack_byte_array(message,&buf,&length);
	alp_ipc_message_unpack_end(message);
	ALP_FAIL_NON_FATAL_IF(buf == 0,"unpacked message buffer is null");
	ALP_FAIL_NON_FATAL_IF(length == 0,"unpacked message length is zero\n");

	// process the message
	if (buf && length)
		status = alp_prv_notify_server_mesg(buf,length,connect);

	// pack a status message
	alp_ipc_message_response_pack_start(message, 0, message);
	alp_ipc_message_pack_int32(message,status);
	alp_ipc_message_pack_end(message);

	// send the status back to the client
	alp_ipc_connection_send(connect,message);

	// deallocate the message
	alp_ipc_message_destroy(message);

#if 1
	// check if notification is done or a new broadcast
	if (buf && (*buf == ALP_NOTIFY_CMD_NOTIFY_DONE || *buf == ALP_NOTIFY_CMD_BROADCAST)) {
		// do the next broadcast if possible
		alp_notify_broadcast_do_pending();
		alp_notify_broadcast_cleanup();
	}
#endif

	// free the buffer from unpack_byte_array
	if (buf) g_free(buf);
}

// disconnect callback function
static void alp_prv_notify_server_disconnect(AlpConnection * connect, gpointer cbData)
{
	char * appID;
#ifdef ALP_NOTIFY_DEBUG
	printf("alp_prv_notify_server_disconnect %x\n",(int)connect);
#endif
	// get the appID
	appID = alp_prv_notify_connect_get_appID(connect);
	if (appID) {

		// remove any registered notification
		alp_notify_register_remove_all(appID,1);

		// remove from active broadcasts list
		alp_notify_remove(appID);

		// remove from active server list
		alp_prv_notify_connect_remove(connect);
	}

	// cancel any sync broadcasts
	alp_notify_broadcast_cancel_sync(connect);
}

// connect callback function
static void alp_prv_notify_server_connect(AlpConnection * connect, gpointer cbData)
{
	alp_status_t err;

#ifdef ALP_NOTIFY_DEBUG
	printf("alp_prv_notify_server_connect %x\n",(int)connect);
#endif
	err = alp_ipc_connection_register_receive_callback(connect,alp_prv_notify_server_receive,connect);
}

// initialize the ipc
void alp_notify_ipc_init(void)
{
	AlpChannel * channel;
	alp_status_t err;

	// create the alp ipc channel
	channel = alp_ipc_channel_create(ALP_NOTIFY_MGR_SERVER_CHANNEL,
							   ALP_IPC_CHANNEL_ACCESS_ALL);

	// register the connection callback
	err = alp_ipc_channel_register_connect_callback(channel,alp_prv_notify_server_connect,NULL);

	// register the disconnection callback
	err = alp_ipc_channel_register_disconnect_callback(channel,alp_prv_notify_server_disconnect,NULL);
}

// send an ipc message
int alp_notify_ipc_send(const char * appID, void * buf, int length)
{
	AlpConnection * connect;
	AlpMessage * message;
	alp_status_t err;

#ifdef ALP_NOTIFY_DEBUG
	printf("alp_notify_ipc_send %s\n",appID);
#endif
	// get the connection
	connect = alp_prv_notify_connect_get_connect(appID);
	if (!connect) return -1;

	// pack the message
	message = alp_ipc_message_create();
	alp_ipc_message_pack_start(message,0);
	alp_ipc_message_pack_byte_array(message,buf,length);
	alp_ipc_message_pack_end(message);

	// send the message to the client
	if ((err = alp_ipc_connection_send(connect,message)) != ALP_OK) {
#ifdef ALP_NOTIFY_DEBUG
		printf("alp_notify_ipc_send: alp_ipc_connection_send err=%x\n",err);
#endif
		goto Done;
	}

Done:
	// deallocate the message
	alp_ipc_message_destroy(message);

	// return status
	return err;
}
