/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/* SecurityObject.c
**
**
** Security Object Implementation
**
** This file contains functions and data structures that implement the
** Security Object interface.
**
*/

/* Macro Definitions */
#define ALP_PRV_SPF_SO_BREAK_IF_NULL( object, type, err ) if( object == (type *) NULL ) { err = ALP_PRV_SPF_SO_E_NULL_PARAMETER; break; }

#ifndef DEBUG
#define DEBUG 1
#endif

#ifdef DEBUG
#define d(o, msg) if( o->debug != 0 ) { fputs( msg, stderr ); }
#else
#define d(o, msg)
#endif

/* File Includes */
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <netinet/in.h>
#include "SecurityObject.h"
#include "SecurityKernel.h"
#include <glib.h>
#include "hiker/prv/lsm_cmd.h"


/* Typedefs, etc. */
typedef struct {
  SecurityObject *object;
  AlpPrvSpfSoSecurityPolicy *current_policy;
  AlpPrvSpfSoSecurityAssertion *current_assertion;
  int security_fd;
} KernelCommunicationContext;

/* Static Function Prototypes */

static AlpPrvSpfSoErr alp_prv_spf_so_kernel_assertion_write_open( void );
static void alp_prv_spf_so_kernel_assertion_write_policy( gpointer data, gpointer user_data );
static void alp_prv_spf_so_kernel_assertion_write_assertion( gpointer data, gpointer user_data );
static AlpPrvSpfSoErr alp_prv_spf_so_kernel_assertion_write_close( void );
static int alp_prv_spf_so_find_string_in_list( char *list[], char *trial );
static int alp_prv_spf_so_find_element_in_list( char *trial );

/* Global Variables */

static KernelCommunicationContext kcc;

/* Function Definitions */

AlpPrvSpfSoErr alp_prv_spf_get_user_group_from_pid( SecurityObject *object, unsigned int pid, unsigned int *out_user, unsigned int *out_group ) {
  AlpPrvSpfSoErr err = ALP_PRV_SPF_SO_E_NOERR;
  Alp_lsmCmd_unit cmd_unit;
  int results;

  do {
    ALP_PRV_SPF_SO_BREAK_IF_NULL( object, SecurityObject, err);
    ALP_PRV_SPF_SO_BREAK_IF_NULL( out_user, unsigned int, err );
    ALP_PRV_SPF_SO_BREAK_IF_NULL( out_group, unsigned int, err );

    /* This code queries the SPF through the /dev/security interface */
    /* asking it to return the uid and gid of the given pid.  */
 	/* set up command structure */

    cmd_unit.command 	= ALP_LSMCMD_PID_STAT;
    cmd_unit.subject_id = 0;
    cmd_unit.object_id1 = 0;
    cmd_unit.object_id2 = 0;
    cmd_unit.object_id3 = 0;
    cmd_unit.process_id = pid;
    cmd_unit.bit_flag	= 0;
    cmd_unit.status     = 0;

    /* open interface */
    if( ( err = alp_prv_spf_so_kernel_assertion_write_open( ) ) !=
	ALP_PRV_SPF_SO_E_NOERR ) {
      break;
    }

    /* write command */
    results = write(
		kcc.security_fd,
		&cmd_unit,
		sizeof(Alp_lsmCmd_unit));

    /* Close interface */
    if( ( err = alp_prv_spf_so_kernel_assertion_write_close( ) ) !=
	ALP_PRV_SPF_SO_E_NOERR ) {
      break;
    }

    /* check for successful write */
    if ((results != sizeof(Alp_lsmCmd_unit))
      ||(cmd_unit.status != 0))
    {
      err = ALP_PRV_SPF_SO_E_FILE;
      break;
    }

    /* parse return values */
    *out_user	= cmd_unit.object_id2;
    *out_group	= cmd_unit.object_id3;
    return 0;

  } while( 0 );

  return( err );
}

AlpPrvSpfSoErr alp_prv_spf_so_kernel_load_meta()
{
  AlpPrvSpfSoErr err = ALP_PRV_SPF_SO_E_NOERR;
  Alp_lsmCmd_unit cmd_unit;
  int results;

  do {
    /* This code sends a ALP_LSMCMD_LSM_LOAD_META to
       initialize the SPF-LSM. Failure indicates a need
       for a kernel initialize */

    cmd_unit.command 	= ALP_LSMCMD_LSM_LOAD_META;
    cmd_unit.subject_id = 0;
    cmd_unit.object_id1 = 0;
    cmd_unit.object_id2 = 0;
    cmd_unit.object_id3 = 0;
    cmd_unit.process_id = 0;
    cmd_unit.bit_flag	= 0;
    cmd_unit.status     = 0;

    /* open interface */
    if( ( err = alp_prv_spf_so_kernel_assertion_write_open( ) ) !=
	ALP_PRV_SPF_SO_E_NOERR ) {
      break;
    }

    /* write command */
    results = write(
		kcc.security_fd,
		&cmd_unit,
		sizeof(Alp_lsmCmd_unit));

    /* Close interface */
    if( ( err = alp_prv_spf_so_kernel_assertion_write_close( ) ) !=
	ALP_PRV_SPF_SO_E_NOERR ) {
      break;
    }

    /* check for successful write */
    if (results != sizeof(Alp_lsmCmd_unit))
    {
      err = ALP_PRV_SPF_SO_E_FILE;
      break;
    }

    /* check for successful LSM results */
    if (cmd_unit.status != 0)
    {
      err = cmd_unit.status;
      break;
    }
    else return 0;

  } while (0);
  return( err );
}

AlpPrvSpfSoErr alp_prv_spf_so_kernel_dump_meta()
{
  AlpPrvSpfSoErr err = ALP_PRV_SPF_SO_E_NOERR;
  Alp_lsmCmd_unit cmd_unit;
  int results;

  do {
    /* This code sends a ALP_LSMCMD_LSM_DUMP_META to
       save kernel Meta Data to the SPF-LSM init file */

    cmd_unit.command 	= ALP_LSMCMD_LSM_DUMP_META;
    cmd_unit.subject_id = 0;
    cmd_unit.object_id1 = 0;
    cmd_unit.object_id2 = 0;
    cmd_unit.object_id3 = 0;
    cmd_unit.process_id = 0;
    cmd_unit.bit_flag	= 0;
    cmd_unit.status     = 0;

    /* open interface */
    if( ( err = alp_prv_spf_so_kernel_assertion_write_open( ) ) !=
	ALP_PRV_SPF_SO_E_NOERR ) {
      break;
    }

    /* write command */
    results = write(
		kcc.security_fd,
		&cmd_unit,
		sizeof(Alp_lsmCmd_unit));

    /* Close interface */
    if( ( err = alp_prv_spf_so_kernel_assertion_write_close( ) ) !=
	ALP_PRV_SPF_SO_E_NOERR ) {
      break;
    }

    /* check for successful write */
    if (results != sizeof(Alp_lsmCmd_unit))
    {
      err = ALP_PRV_SPF_SO_E_FILE;
      break;
    }

    /* check for successful LSM init */
    if (cmd_unit.status != 0)
    {
      err = cmd_unit.status;
      break;
    }
    else return 0;

  } while (0);
  return( err );
}


AlpPrvSpfSoErr alp_prv_spf_so_kernel_lsm_enable()
{
  AlpPrvSpfSoErr err = ALP_PRV_SPF_SO_E_NOERR;
  Alp_lsmCmd_unit cmd_unit;
  int results;

  do {
    /* This code sends a ALP_LSMCMD_LSM_ENABLE to
       activate LSM enforcement (off by default) */

    cmd_unit.command 	= ALP_LSMCMD_LSM_ENABLE;
    cmd_unit.subject_id = 0;
    cmd_unit.object_id1 = 0;
    cmd_unit.object_id2 = 0;
    cmd_unit.object_id3 = 0;
    cmd_unit.process_id = 0;
    cmd_unit.bit_flag	= 0;
    cmd_unit.status     = 0;

    /* open interface */
    if( ( err = alp_prv_spf_so_kernel_assertion_write_open( ) ) !=
	ALP_PRV_SPF_SO_E_NOERR ) {
      break;
    }

    /* write command */
    results = write(
		kcc.security_fd,
		&cmd_unit,
		sizeof(Alp_lsmCmd_unit));

    /* Close interface */
    if( ( err = alp_prv_spf_so_kernel_assertion_write_close( ) ) !=
	ALP_PRV_SPF_SO_E_NOERR ) {
      break;
    }

    /* check for successful write */
    if (results != sizeof(Alp_lsmCmd_unit))
    {
      err = ALP_PRV_SPF_SO_E_FILE;
      break;
    }

    /* check for successful LSM results */
    if (cmd_unit.status != 0)
    {
      err = cmd_unit.status;
      break;
    }
    else return 0;

  } while (0);
  return( err );
}

static AlpPrvSpfSoErr alp_prv_spf_so_kernel_assertion_write_open(  ) {
  AlpPrvSpfSoErr err = ALP_PRV_SPF_SO_E_NOERR;

  do {
    /* todo: Put code here to open the /dev/security device. Then put the */
    /* file descriptor in the kcc context                                 */
    if( ( kcc.security_fd = open(SPF_LSM_DEV_IN, O_WRONLY ) ) < 0 ) {
      err = ALP_PRV_SPF_SO_E_FILE;
      break;
    }
  } while( 0 );

  return( err );
}


AlpPrvSpfSoErr alp_prv_spf_so_kernel_initialize( SecurityObject *object ) {
  AlpPrvSpfSoErr err = ALP_PRV_SPF_SO_E_NOERR;

  do {
    ALP_PRV_SPF_SO_BREAK_IF_NULL( object, SecurityObject, err);

    /* This code goes through all the policies and all the assertions */
    /* and checks to see if the assertion has been marked as a one    */
    /* to be handed off to the kernel. If so, the static function     */
    /* alp_prv_spf_so_kernel_assertion_write() is called.             */

    kcc.object = object;
    kcc.current_policy = (AlpPrvSpfSoSecurityPolicy *) NULL;
    kcc.current_assertion = (AlpPrvSpfSoSecurityAssertion *) NULL;

    if( ( err = alp_prv_spf_so_kernel_assertion_write_open( ) ) !=
	ALP_PRV_SPF_SO_E_NOERR ) {
      break;
    }

    g_ptr_array_foreach( object->policies, alp_prv_spf_so_kernel_assertion_write_policy, (gpointer) &kcc );

    if( ( err = alp_prv_spf_so_kernel_assertion_write_close( ) ) !=
	ALP_PRV_SPF_SO_E_NOERR ) {
      break;
    }
  } while( 0 );

  return( err );
}

static void alp_prv_spf_so_kernel_assertion_write_policy( gpointer data, gpointer user_data ) {
  AlpPrvSpfSoSecurityPolicy *current_policy = (AlpPrvSpfSoSecurityPolicy *) data;
  Alp_lsmCmd_unit cmd_unit; 	/* LSM Command structure */
  int status;

  kcc.current_policy = current_policy;
  memset(&cmd_unit, 0, sizeof(Alp_lsmCmd_unit));

  /* assemble Command Unit */
  cmd_unit.command    = ALP_LSMCMD_SET_CREATE;
  cmd_unit.subject_id = (kcc.current_policy)->gid;
  cmd_unit.object_id1 = (kcc.current_policy)->sflag;
  cmd_unit.object_id2 = 0;
  cmd_unit.object_id3 = 0;
  cmd_unit.process_id = 0;
  cmd_unit.bit_flag   = (kcc.current_policy)->bflag;
  cmd_unit.status     = 0;

  /* push the Command to the Kernel Module */
  /* write command */
  status = write( kcc.security_fd, &cmd_unit, sizeof(Alp_lsmCmd_unit));

  /* check for successful write */
  if (status != sizeof(Alp_lsmCmd_unit))
  {
    return;
  }

  g_ptr_array_foreach( current_policy->assertions, alp_prv_spf_so_kernel_assertion_write_assertion, (gpointer) NULL );

}

static void alp_prv_spf_so_kernel_assertion_write_assertion( gpointer data, gpointer user_data ) {
  AlpPrvSpfSoSecurityAssertion *current_assertion = (AlpPrvSpfSoSecurityAssertion *) data;
  AlpPrvSpfSoSecurityAssertion *perm_assertion = NULL;
  kcc.current_assertion = current_assertion;
  Alp_lsmCmd_unit cmd_unit; 	/* LSM Command structure */
  int status;
  gchar* perm_str = NULL;


  /* todo: insert code here to write an assertion policy, key and value */
  /* to the SPF by way of the /dev/security interface                   */
  /* Assumes ACL Set commands only */
  /* Assumes ACL Set = current_policy->id already exists */
  memset(&cmd_unit, 0, sizeof(Alp_lsmCmd_unit));

  /* Command and other fields vary according to Data type */
  switch (current_assertion->type)
  {
  case ALP_PRV_SPF_SO_TYPE_PATH:
  {
#ifdef DEBUG
  if( kcc.object->debug != 0 ) {
    fprintf( kcc.object->debug_fp, "%%SPFSS-I-KPUSH; pushing assertion (%s,%s) to kernel\n", current_assertion->key, current_assertion->value );
  }
#endif
    struct stat stat_buf;		/* stat buffer */
    memset(&stat_buf, 0, sizeof(struct stat));
    status = stat(current_assertion->value, &stat_buf);
    if (status != 0)
    {
#ifdef DEBUG
      fprintf( kcc.object->debug_fp, "%%SPFSS-F-KSTATF; File Stat call failed with filename %s\n", current_assertion->value);
      fprintf( kcc.object->debug_fp, "%%SPFSS-F-KSTATS; File Stat call failed with status %d\n", status);
#endif
      return;
    }

    /* assemble Command Unit */
    cmd_unit.command    = ALP_LSMCMD_ACL_INODE_ADD;
    cmd_unit.subject_id = (kcc.current_policy)->gid;
    cmd_unit.object_id1 = stat_buf.st_dev;
    cmd_unit.object_id2 = stat_buf.st_ino;
    cmd_unit.object_id3 = 0;
    cmd_unit.process_id = 0;
    cmd_unit.status     = 0;
  }
  break;

  case ALP_PRV_SPF_SO_TYPE_ADDRESS:
  {
    unsigned long ub1, ub2, ub3, ub4, bitmask, min_port, max_port, addr;
#ifdef DEBUG
  if( kcc.object->debug != 0 ) {
    fprintf( kcc.object->debug_fp, "%%SPFSS-I-KPUSH; pushing assertion (%s,%s) to kernel\n", current_assertion->key, current_assertion->value );
  }
#endif

    /* were assuming a value format of "w.x.y.z/bits[pmin:pmax]
       where w,x,y,z,bits,pmin,pmax are scanable integers */
    status = sscanf( current_assertion->value, "%lu.%lu.%lu.%lu/%lu[%lu-%lu]",
				   	&ub1, &ub2, &ub3, &ub4, &bitmask, &min_port, &max_port);
    if (status != 7 )
    {
#ifdef DEBUG
      fprintf( kcc.object->debug_fp, "%%SPFSS-F-APARSE; Could not parse Address attributes from value\n");
#endif
      return;
    }
    /* assemble IP address */
    addr = (ub1 << 24) + (ub2 << 16) + (ub3 << 8) + ub4;

    /* drop off insignificent bits */
    addr = addr >> (32 - bitmask);
    addr = addr << (32 - bitmask);

     /* assemble command unit */
     cmd_unit.command    = ALP_LSMCMD_ACL_NET_ADD;
     cmd_unit.subject_id = (kcc.current_policy)->gid;
     cmd_unit.object_id1 = htonl(addr);
     cmd_unit.object_id2 = min_port;
     cmd_unit.object_id3 = max_port;
     cmd_unit.process_id = 0;
     cmd_unit.status     = 0;
  }
  break;

  default:
    return;
  } /* end switch */

  /* look up associated permission entry */
  perm_str = g_strconcat(current_assertion->key, ".permission", NULL);
#ifdef DEBUG
  fprintf( kcc.object->debug_fp, "%%SPFSS-I-LOOK; Looking for Permission assertion %s\n", perm_str);
#endif
  status = alp_prv_spf_so_policy_assertion_find( kcc.current_policy, perm_str, &perm_assertion);
  g_free(perm_str);

  if (status != 0)
  {
#ifdef DEBUG
      fprintf( kcc.object->debug_fp, "%%SPFSS-F-LOOK; Could not find associated permission attributes\n");
#endif
	  return;
  }
    /* Need to scan perm_assertion->key to decide on the bit_flag value */
    /* bit_flag must be ONE of the following values
	ALP_LSMACL_BFLAG_DENY
	ALP_LSMACL_BFLAG_PERMIT
	ALP_LSMACL_BFLAG_ASK */

#ifdef DEBUG
      fprintf( kcc.object->debug_fp, "%%SPFSS-I-PERMAV; Permission Attribute value=%s\n", perm_assertion->value);
#endif
    status = alp_prv_spf_so_find_element_in_list(perm_assertion->value);

  switch (status)
  {
  case 0:
    cmd_unit.bit_flag = ALP_LSMACL_BFLAG_DENY;
  break;
		
  case 1:
    cmd_unit.bit_flag = ALP_LSMACL_BFLAG_PERMIT;
  break;
		
  case 2:
    cmd_unit.bit_flag = ALP_LSMACL_BFLAG_ASK;
  break;
  } /* end switch */

  /* push the Command to the Kernel Module */
  /* write command */
  status = write( kcc.security_fd, &cmd_unit, sizeof(Alp_lsmCmd_unit));

  /* check for successful write */
  if (status != sizeof(Alp_lsmCmd_unit))
  {
    return;
  }

  /* parse status */
  status = cmd_unit.status;

  if (status != 0)
  {
#ifdef DEBUG
    fprintf( kcc.object->debug_fp, "%%SPFSS-F-LSMFAIL; SPF_LSM failed with status of %d\n", status);
#endif
  }
  return;

}

static AlpPrvSpfSoErr alp_prv_spf_so_kernel_assertion_write_close( ) {
  AlpPrvSpfSoErr err = ALP_PRV_SPF_SO_E_NOERR;

  do {
    /* todo: Put code here to close the /dev/security device */
	close(kcc.security_fd);
  } while( 0 );

  return( err );
}

static int alp_prv_spf_so_find_string_in_list( char *list[], char *trial ){
  int i;

  if( trial == NULL ) {
    i = -1;
  } else {
    for( i=0; list[i] != NULL; i++ ) {

      if( !strcmp(list[i], trial) )
	break;
    }

    if( list[i] == NULL ) {
      i = -1;
    }
  }
  return( i );
}

static int alp_prv_spf_so_find_element_in_list( char *trial ) {
  static char *so_elements[] = {
    "deny",
    "allow",
    "ask",
    NULL
  };

  return( alp_prv_spf_so_find_string_in_list( so_elements, trial ) );
}
