/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/* SecurityObject.c
**
**
** Security Object Implementation
**
** This file contains functions and data structures that implement the
** Security Object interface.
**
*/

/* Macro Definitions */
#define ALP_PRV_SPF_SO_BREAK_IF_NULL( object, type, err ) if( object == (type *) NULL ) { err = ALP_PRV_SPF_SO_E_NULL_PARAMETER; break; }

#define ALP_PRV_SPF_SO_ELEMENT_UNKNOWN      -1
#define ALP_PRV_SPF_SO_ELEMENT_SECURITY     0
#define ALP_PRV_SPF_SO_ELEMENT_POLICIES     1
#define ALP_PRV_SPF_SO_ELEMENT_POLICY       2
#define ALP_PRV_SPF_SO_ELEMENT_ASSERTION    3
#define ALP_PRV_SPF_SO_ELEMENT_CERTREF      4
#define ALP_PRV_SPF_SO_ELEMENT_PACKAGES     5
#define ALP_PRV_SPF_SO_ELEMENT_PACKAGE      6
#define ALP_PRV_SPF_SO_ELEMENT_CERTIFICATES 7
#define ALP_PRV_SPF_SO_ELEMENT_CERTIFICATE  8
#define ALP_PRV_SPF_SO_ELEMENT_PKCS5        9

#ifndef DEBUG
#define DEBUG 1
#endif

#ifdef DEBUG
#define d(o, msg) if( o->debug != 0 ) { fprintf( o->debug_fp, "pid = %d : ", getpid() ); fputs( msg, o->debug_fp ); }
#define m(o, abbrev, sev, msg ) if( o->debug != 0 ) {  fprintf( o->debug_fp, "pid = %d : ", getpid() ); fprintf(o->debug_fp ,"%%SPFSO-%s-%s; %s.\n", #sev, #abbrev, msg ); }
#else
#define d(o, msg)
#define m(o, abbrev, sev, msg )
#endif

/* File Includes */
#include "SecurityObject.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <expat.h>
#include <string.h>
#include <openssl/bio.h>
#include <openssl/evp.h>
#include <openssl/buffer.h>
#include "SecurityPKCS5Support.h"

/* Typedefs, Structs, Unions, Enums, etc. */

typedef struct {
  SecurityObject *object;
  AlpPrvSpfSoSecurityPolicy *current_policy;
  AlpPrvSpfSoSecurityAssertion *current_assertion;
  AlpPrvSpfSoCertificateReference *current_certificate;
  AlpPrvSpfSoPackage *current_package;
  AlpPrvSpfSuPKCS5 *current_pkcs5;
  GString *current_string;
  unsigned int in_element;
} _ParseData;

typedef struct {
  char *attribute;
  void *put_here;
} _AttThere;

typedef struct {
  unsigned int policy_id;
  char *policy_name;
  AlpPrvSpfSoSecurityPolicy *policy;
} _policyData;

/* Static Function Prototypes */
static AlpPrvSpfSoErr alp_prv_spf_so_parse( SecurityObject *object, char *data, unsigned int length );
static unsigned int _g_ptr_array_count( GPtrArray *array );
static void alp_prv_spf_so_count_g_ptr_array_members( gpointer data, gpointer user_data );
static void alp_prv_spf_so_policy_file_out( gpointer data, gpointer user_data );
static void alp_prv_spf_so_assertion_file_out( gpointer data, gpointer user_data );
static void alp_prv_spf_so_certref_in_policy_file_out( gpointer data, gpointer user_data );
static void alp_prv_spf_so_package_file_out( gpointer data, gpointer user_data );
static void alp_prv_spf_so_certificate_file_out( gpointer data, gpointer user_data );
static void alp_prv_spf_so_find_certificate_in_list( gpointer data, gpointer user_data );
static AlpPrvSpfSoErr alp_prv_spf_so_version_integer_to_string( SecurityObject *object, char *buffer, unsigned int * length );
static gint alp_prv_spf_so_policies_sort( gconstpointer a, gconstpointer b );
static gint alp_prv_spf_so_assertions_sort( gconstpointer a, gconstpointer b );
static gint alp_prv_spf_so_certificates_sort( gconstpointer a, gconstpointer b );
static gint alp_prv_spf_so_packages_sort( gconstpointer a, gconstpointer b );
static void alp_prv_spf_so_start_handler( void *userData, const XML_Char *name, const XML_Char **atts );
static void alp_prv_spf_so_end_handler( void *userData, const XML_Char *name );
static void alp_prv_spf_so_character_handler( void *userData, const XML_Char *s, int len);
static int alp_prv_spf_so_find_string_in_list( char *list[], char *trial );
static int alp_prv_spf_so_find_element_in_list( char *trial );
static void alp_prv_spf_so_find_attributes( unsigned int element, char *atts[], unsigned int target_size, char *target[] );
static unsigned int alp_prv_spf_so_find_type_in_list( char *trial );
static char * alp_prv_spf_so_new_copy( char *original );
static char * alp_prv_spf_so_new_copy_with_length( char *o, unsigned int l );
static void alp_prv_spf_so_find_policy_by_id_func( gpointer data, gpointer user_data );
static void alp_prv_spf_so_find_policy_by_name_func( gpointer data, gpointer user_data );
static void alp_prv_spf_so_find_policy_by_cert_ref_func( gpointer data, gpointer user_data );
static void alp_prv_spf_so_rfc1113_encode( char *input, unsigned int input_length, char *output, unsigned int output_max_length, unsigned int *output_bytes_written );
static void alp_prv_decode_hash( char *input, char *output );

/* Variable Declarations */

/* Function Definitions */
AlpPrvSpfSoErr alp_prv_spf_so_initialize( SecurityObject *object ) {
  AlpPrvSpfSoErr err = ALP_PRV_SPF_SO_E_NOERR;

  do {

    ALP_PRV_SPF_SO_BREAK_IF_NULL( object, SecurityObject, err )

    object->version          = ALP_PRV_SPF_SO_DEFAULT_VERSION;
    object->debug            = 1;
    object->debug_fp         = stderr;
    object->config_path      = ALP_PRV_SPF_SO_DEFAULT_CONFIG;
    object->policies         = g_ptr_array_new();
    object->policy_base      = ALP_PRV_SPF_SO_DEFAULT_POLICY_BASE;
    object->policy_default   = ALP_PRV_SPF_SO_DEFAULT_POLICY_DEFAULT;
    object->policy_gid_min   = ALP_PRV_SPF_SO_DEFAULT_POLICY_GID_MIN;
    object->policy_gid_next  = ALP_PRV_SPF_SO_DEFAULT_POLICY_GID_NEXT;
    object->policy_id_min    = ALP_PRV_SPF_SO_DEFAULT_POLICY_ID_MIN;
    object->policy_id_next   = ALP_PRV_SPF_SO_DEFAULT_POLICY_ID_NEXT;

    object->packages         = g_ptr_array_new();
    object->package_uid_min  = ALP_PRV_SPF_SO_DEFAULT_PACKAGE_UID_MIN;
    object->package_uid_next = ALP_PRV_SPF_SO_DEFAULT_PACKAGE_UID_NEXT;
    object->package_id_min   = ALP_PRV_SPF_SO_DEFAULT_PACKAGE_ID_MIN;
    object->package_id_next  = ALP_PRV_SPF_SO_DEFAULT_PACKAGE_ID_NEXT;

    object->certificates     = g_ptr_array_new();

    object->attn_queue_depth = 0;
    object->attn_queue_next  = 31337;

    object->pin = malloc( sizeof( AlpPrvSpfSuPKCS5 ) );
    alp_prv_spf_pkcs5_init_default( & object->pin, "pin" );
    object->puk = malloc( sizeof( AlpPrvSpfSuPKCS5 ) );
    alp_prv_spf_pkcs5_init_default( & object->puk, "puk" );

  } while( 0 );

  return( err );
}

AlpPrvSpfSoErr alp_prv_spf_so_uninitialize( SecurityObject *object ) {
  AlpPrvSpfSoErr err = ALP_PRV_SPF_SO_E_NOERR;

  m(object, BUNINIT, D, "entering alp_prv_spf_so_uninitialize()");

  do {
    ALP_PRV_SPF_SO_BREAK_IF_NULL( object, SecurityObject, err );

    g_ptr_array_free( object->policies, TRUE );
    g_ptr_array_free( object->packages, TRUE );
    g_ptr_array_free( object->certificates, TRUE );

  } while( 0 );

  return( err );
}

AlpPrvSpfSoErr alp_prv_spf_so_read( SecurityObject *object, char *path ) {
  AlpPrvSpfSoErr err = ALP_PRV_SPF_SO_E_NOERR;
  struct stat buf;
  int filesize;
  int bytes_read;
  char *file = NULL;
  FILE *fp;

  m(object, EREAD, D, "entering alp_prv_spf_so_read()");

  do {
    ALP_PRV_SPF_SO_BREAK_IF_NULL( object, SecurityObject, err );
    ALP_PRV_SPF_SO_BREAK_IF_NULL( path, char, err );

    if( stat( path, & buf ) != 0 ) {
      err = ALP_PRV_SPF_SO_E_FILE;
      break;
    }

    filesize = (unsigned int) buf.st_size;
    if( ( file = malloc( filesize + 1 ) ) == NULL ) {
      err = ALP_PRV_SPF_SO_E_MALLOC;
      break;
    }

    if( ( fp = fopen( path, "r" ) ) == (FILE *) NULL ) {
      err = ALP_PRV_SPF_SO_E_FILE;
      break;
    }

    if( ( bytes_read = fread( file, (size_t) 1, (size_t) filesize, fp ) ) !=
	filesize ) {
      err = ALP_PRV_SPF_SO_E_FILE;
      break;
    }

    fclose( fp );

    file[ filesize ] = '\0';

    if( ( err = alp_prv_spf_so_parse( object, file, filesize ) ) != ALP_PRV_SPF_SO_E_NOERR ) {
      break;
    }
    
  } while( 0 );

  if( file != NULL ) {
    free( file );
  }

  return( err );
}

AlpPrvSpfSoErr alp_prv_spf_so_write( SecurityObject *object, char *path ) {
  AlpPrvSpfSoErr err = ALP_PRV_SPF_SO_E_NOERR;
  FILE *fp = NULL;
  char vers[ 16 ];
  unsigned int vers_length = 16;

  m( object, EWRITE, D, "entering alp_prv_spf_so_write()");
  do {
    ALP_PRV_SPF_SO_BREAK_IF_NULL( object, SecurityObject, err );
    ALP_PRV_SPF_SO_BREAK_IF_NULL( path, char, err );

    /* todo: if file exists, rename existing file as backup file */

    m( object, EWRITE, D, "opening file");

    if( ( fp = fopen( path, "w" ) ) == (FILE *) NULL ) {
      err = ALP_PRV_SPF_SO_E_FILE;
      break;
    }

    /* convert the version integer into a printable string */
    if( ( err = alp_prv_spf_so_version_integer_to_string( object, vers, & vers_length ) ) !=
	ALP_PRV_SPF_SO_E_NOERR ) {
      break;
    }

    /* write the header */

    m( object, EWRITE, D, "writing header");

    fprintf( fp, "<?xml version=\"1.0\"?>\n" );
    fprintf( fp, "<security version=\"%*.*s\">\n", vers_length, vers_length, vers );
    
    /* write the PKCS5 information */
    fprintf( fp, " " );
    alp_prv_spf_pkcs5_fput( object->pin, fp );
    fprintf( fp, "\n" );

    fprintf( fp, " " );
    alp_prv_spf_pkcs5_fput( object->puk, fp );
    fprintf( fp, "\n" );
    
    /* write certificates */

    m( object, EWRITE, D, "writing certificates");

    if( _g_ptr_array_count( object->certificates ) == 0 ) {
      fprintf( fp, " <certificates/>\n" );
    } else {
      fprintf( fp, " <certificates>\n" );
      g_ptr_array_foreach( object->certificates, alp_prv_spf_so_certificate_file_out,
			   (gpointer) fp );
      fprintf( fp, " </certificates>\n" );
    }

    /* write policies */

    m( object, EWRITE, D, "writing policies");

    fprintf( fp, " <policies base=\"%s\" default=\"%s\" minid=\"%d\" nextid=\"%d\" mingid=\"%d\" nextgid=\"%d\"", object->policy_base, object->policy_default, object->policy_id_min, object->policy_id_next, object->policy_gid_min, object->policy_gid_next );

    if( _g_ptr_array_count( object->policies ) == 0 ) {
      fprintf( fp, "/>\n" );
    } else {
      fprintf( fp, ">\n" );
      g_ptr_array_foreach( object->policies, alp_prv_spf_so_policy_file_out,
			   (gpointer) fp );
      fprintf( fp, " </policies>\n" );
    }

    /* write package references */

    m( object, EWRITE, D, "writing packages");

    fprintf( fp, " <packages minid=\"%d\" nextid=\"%d\" minuid=\"%d\" nextuid=\"%d\"", object->package_id_min, object->package_id_next, object->package_uid_min, object->package_uid_next );

    if( _g_ptr_array_count( object->packages ) == 0 ) {
      fprintf( fp, "/>\n" );
    } else {
      fprintf( fp, ">\n" );
      g_ptr_array_foreach( object->packages, alp_prv_spf_so_package_file_out,
			   (gpointer) fp );
      fprintf( fp, " </packages>\n" );
    }

    fprintf( fp, "</security>\n");

    /* todo: unlink backup file */

  } while( 0 );

  if( fp != (FILE *)NULL ) {
    fclose( fp );
  }
 
  return( err );
}

AlpPrvSpfSoErr alp_prv_spf_so_package_add( SecurityObject *object, AlpPrvSpfSoPackage *package) {
  AlpPrvSpfSoErr err = ALP_PRV_SPF_SO_E_NOERR;
  unsigned int increment_uid_next = 0;

  do {
    ALP_PRV_SPF_SO_BREAK_IF_NULL( object, SecurityObject, err );
    ALP_PRV_SPF_SO_BREAK_IF_NULL( package, AlpPrvSpfSoPackage, err );

    if( package->uid < object->package_uid_min ) {
      package->uid = object->package_uid_next;
      increment_uid_next = 1;
    } else {
      /* if specified uid is already in use, return an error */
      if( package->uid < object->package_uid_next ) {
	err = ALP_PRV_SPF_SO_E_UID_IN_USE;
	break;
      } else {
	object->package_uid_next = package->uid + 1;
      }
    }

    if( package->id < object->package_id_min ) {
      package->id = object->package_id_next;
      (object->package_id_next)++;
    } else {
      /* if specified id is already in use, return an error */
      if( package->id < object->package_id_next ) {
	err = ALP_PRV_SPF_SO_E_ID_IN_USE;
	break;
      } else {
	object->package_id_next = package->id + 1;
      }      
    }

    if( increment_uid_next == 1 ) {
      (object->package_uid_next)++;
    }

    g_ptr_array_add( object->packages, (gpointer) package );
    g_ptr_array_sort( object->packages, alp_prv_spf_so_packages_sort );

    /* todo : set dirty flag and write policy file */
  } while( 0 );

  return( err );
}

AlpPrvSpfSoErr alp_prv_spf_so_policy_initialize( AlpPrvSpfSoSecurityPolicy *policy ) {
  AlpPrvSpfSoErr err = ALP_PRV_SPF_SO_E_NOERR;

  do {
    ALP_PRV_SPF_SO_BREAK_IF_NULL( policy, AlpPrvSpfSoSecurityPolicy, err );

    policy->assertions = g_ptr_array_new();
    policy->certificates = g_ptr_array_new();
    policy->sflag = 0;
    policy->bflag = 0;
  } while( 0 );

  return( err );
}

AlpPrvSpfSoErr alp_prv_spf_so_policy_uninitialize( AlpPrvSpfSoSecurityPolicy *policy ) {
  AlpPrvSpfSoErr err = ALP_PRV_SPF_SO_E_NOERR;

  do {
    ALP_PRV_SPF_SO_BREAK_IF_NULL( policy, AlpPrvSpfSoSecurityPolicy, err );

    g_ptr_array_free( policy->assertions, TRUE );
    g_ptr_array_free( policy->certificates, TRUE );
  } while( 0 );

  return( err );
}

AlpPrvSpfSoErr alp_prv_spf_so_policy_write( AlpPrvSpfSoSecurityPolicy *policy , FILE *fp, char *indent ) {
  AlpPrvSpfSoErr err = ALP_PRV_SPF_SO_E_NOERR;
  unsigned int name_length;
  unsigned int pname_length;
  unsigned int certref_size;
  unsigned int assertion_size;

  do {
    ALP_PRV_SPF_SO_BREAK_IF_NULL( policy, AlpPrvSpfSoSecurityPolicy, err );
    ALP_PRV_SPF_SO_BREAK_IF_NULL( fp, FILE, err );

    if( indent == NULL ) {
      indent = "";
    }

    name_length = strlen( policy->name );

    fputs( indent, fp );
    fprintf( fp, "<policy id=\"%d\" gid=\"%d\" name=\"%*.*s\"", policy->id,
	     policy->gid, name_length, name_length, policy->name );

    if( policy->base != (AlpPrvSpfSoSecurityPolicy *)NULL ) {
      pname_length = strlen( (policy->base)->name );
      fprintf( fp, " base=\"%*.*s\"", pname_length, pname_length, 
	       (policy->base)->name );
    }

    if( ( policy->sgids != NULL ) && ( policy->sgids[0] != '\0' ) ) {
      fprintf( fp, " sgids=\"%s\"", policy->sgids );
    }

    if( policy->bflag != 0 ) {
      fprintf( fp, " bflag=\"%d\"", policy->bflag );
    }

    if( policy->sflag != 0 ) {
      fprintf( fp, " sflag=\"%d\"", policy->sflag );
    }

    assertion_size = _g_ptr_array_count( policy->assertions );
    certref_size = _g_ptr_array_count( policy->certificates );

    if( ( assertion_size > 0 ) || ( certref_size > 0 ) ) {
      fprintf( fp, ">\n" );

      if( certref_size > 0 ) {
	g_ptr_array_foreach( policy->certificates, alp_prv_spf_so_certref_in_policy_file_out, fp );
      }

      if( assertion_size > 0 ) {
	g_ptr_array_foreach( policy->assertions, alp_prv_spf_so_assertion_file_out, fp );
      }

      fputs( indent, fp );
      fprintf( fp, "</policy>\n" );
    } else {
      fprintf( fp, "/>\n" );
    }

  } while( 0 );

  return( err );
}

AlpPrvSpfSoErr alp_prv_spf_so_policy_assertion_find (
                               AlpPrvSpfSoSecurityPolicy *policy, 
                               char *key, 
			       AlpPrvSpfSoSecurityAssertion **out_assertion) {

  AlpPrvSpfSoErr err = ALP_PRV_SPF_SO_E_NOERR;
  unsigned int index;
  AlpPrvSpfSoSecurityAssertion *current;

  if ((policy == NULL) || (key == NULL) || (out_assertion == NULL)) {
      return(ALP_PRV_SPF_SO_E_NULL_PARAMETER);
  }

  err = ALP_PRV_SPF_SO_E_NOT_FOUND;
  for (index = 0; index < policy->assertions->len; index++) {
      current = (AlpPrvSpfSoSecurityAssertion *)policy->assertions->pdata[index];
      if (strcmp(current->key, key) == 0) {
          *out_assertion = current;
          err = ALP_PRV_SPF_SO_E_NOERR;
          break;
      }
  }

  if (err == ALP_PRV_SPF_SO_E_NOT_FOUND) {
      *out_assertion = NULL;
      err = ALP_PRV_SPF_SO_E_UNKNOWN_ASSERTION;
  }

  return (err);  
}


AlpPrvSpfSoErr alp_prv_spf_so_assertion_write( AlpPrvSpfSoSecurityAssertion *assertion, FILE *fp, char *indent ) {
  AlpPrvSpfSoErr err = ALP_PRV_SPF_SO_E_NOERR;
  unsigned int id_length;
  unsigned int value_length;
  static char *types[] = {
    "string",
    "boolean",
    "integer",
    "address",
    "path",
    "url"
  };

  do {
    ALP_PRV_SPF_SO_BREAK_IF_NULL( assertion, AlpPrvSpfSoSecurityAssertion, err );
    ALP_PRV_SPF_SO_BREAK_IF_NULL( fp, FILE, err );

    if( indent == NULL ) {
      indent = "";
    }

    id_length = strlen( assertion->key );
    value_length = strlen( assertion->value );

    fputs( indent, fp );
    fprintf( fp, "<assertion id=\"%*.*s\"", id_length, id_length,
	     assertion->key );
    
    if( ( assertion->type != ALP_PRV_SPF_SO_TYPE_STRING ) &&
	( assertion->type <= ALP_PRV_SPF_SO_TYPE_MAX ) ) {
      fprintf( fp, " type=\"%s\"", types[ assertion->type ] );
    }

    if( assertion->kernel != 0 ) {
      fputs( " kernel=\"true\"", fp );
    }

    fprintf( fp, ">%*.*s</assertion>\n", value_length, value_length,
	     assertion->value );
  } while( 0 );

  return( err );
}


AlpPrvSpfSoErr alp_prv_spf_so_policy_find_by_id (
                                        SecurityObject *object, 
		                        unsigned int id, 
					AlpPrvSpfSoSecurityPolicy **policy) {
  _policyData out_policy;
  GPtrArray *array; 
  AlpPrvSpfSoErr err = ALP_PRV_SPF_SO_E_NOERR;

  if ((object == NULL) || (policy == NULL)) {
      return (ALP_PRV_SPF_SO_E_NULL_PARAMETER);
  }

  array = object->policies;

  out_policy.policy_id = id;
  out_policy.policy = NULL;

  g_ptr_array_foreach(array, alp_prv_spf_so_find_policy_by_id_func, 
                      (gpointer)&out_policy);

  *policy = out_policy.policy;

  if (*policy == (AlpPrvSpfSoSecurityPolicy *)NULL) {
      err = ALP_PRV_SPF_SO_E_NOT_FOUND;
  } 

  return (err);
}


AlpPrvSpfSoErr alp_prv_spf_so_policy_find_by_gid (
                                        SecurityObject *object, 
                                        unsigned int gid, 
				        AlpPrvSpfSoSecurityPolicy **policy) {
  AlpPrvSpfSoErr err = ALP_PRV_SPF_SO_E_NOT_FOUND;
  unsigned int index = 0;
  AlpPrvSpfSoSecurityPolicy *current = NULL;

  if ((object == NULL) || (policy == NULL)) {
      return (ALP_PRV_SPF_SO_E_NULL_PARAMETER);
  }

  err = ALP_PRV_SPF_SO_E_NOT_FOUND;
  for (index = 0; index < object->policies->len; index++) {
      current = (AlpPrvSpfSoSecurityPolicy *)object->policies->pdata[index];
      if (current->gid == gid) {
          err = ALP_PRV_SPF_SO_E_NOERR;
          *policy = current;
          break;
      }
  }

  /*
   * if policy is not found then check for the "default"
   */
  if (err == ALP_PRV_SPF_SO_E_NOT_FOUND) {
      err = alp_prv_spf_so_policy_find_by_name(object, 
		                               object->policy_default, 
					       policy);
  }

  return (err);
}


AlpPrvSpfSoErr alp_prv_spf_so_policy_find_by_name(
		                          SecurityObject *object, 
		                          char *name, 
					  AlpPrvSpfSoSecurityPolicy **policy) {
  _policyData out_policy;
  GPtrArray *array; 
  AlpPrvSpfSoErr err = ALP_PRV_SPF_SO_E_NOERR;

  if ((object == NULL) || (name == NULL) || (policy == NULL)) {
      return (ALP_PRV_SPF_SO_E_NULL_PARAMETER);
  }

  array = object->policies;

  out_policy.policy_name = name;
  out_policy.policy = NULL;

  g_ptr_array_foreach(array, alp_prv_spf_so_find_policy_by_name_func, 
                      (gpointer)&out_policy);

  *policy = out_policy.policy;

  if (*policy == (AlpPrvSpfSoSecurityPolicy *)NULL) {
      err = ALP_PRV_SPF_SO_E_NOT_FOUND;
  } 

  return (err);
}


AlpPrvSpfSoErr alp_prv_spf_so_policy_find_by_certificate_reference( SecurityObject *object, AlpPrvSpfSoCertificateReference *certificate, AlpPrvSpfSoSecurityPolicy **policy ) {
  AlpPrvSpfSoErr err = ALP_PRV_SPF_SO_E_NOERR;
  unsigned int i = 0;
  unsigned int j = 0;
  AlpPrvSpfSoSecurityPolicy *current_policy;
  AlpPrvSpfSoCertificateReference *current_cert;
  GPtrArray *policies;

  do{
    ALP_PRV_SPF_SO_BREAK_IF_NULL( object, SecurityObject, err );
    ALP_PRV_SPF_SO_BREAK_IF_NULL( certificate, AlpPrvSpfSoCertificateReference, err );
    ALP_PRV_SPF_SO_BREAK_IF_NULL( policy, AlpPrvSpfSoSecurityPolicy *, err );

    *policy = (AlpPrvSpfSoSecurityPolicy *) 0xDEADBEEF;

    /* iterate through policies */
    policies = object->policies;
    for( i = 0; i < policies->len; i++ ) {
      current_policy = (AlpPrvSpfSoSecurityPolicy *) policies->pdata[i];

      /* iterate through cert references */
      for( j = 0; j < current_policy->certificates->len; j++ ) {
	current_cert = (AlpPrvSpfSoCertificateReference *) current_policy->certificates->pdata[j];
	/* is the issuer and serial number the same? */
	if( !strcmp( certificate->serial, current_cert->serial ) ) {
	  if( !strcmp( certificate->issuer, current_cert->issuer ) ) {
	    /* yeah! we found it! */
	    *policy = current_policy;
	    break;
	  }
	}
      }

      if( j < current_policy->certificates->len ) {
	break;
      }
    }

    if( j >= policies->len ) {
      /* uh oh, we didn't find it */
      err = ALP_PRV_SPF_SO_E_NOT_FOUND;
      break;
    }
  } while( 0 );

  return( err );
}

AlpPrvSpfSoErr alp_prv_spf_so_certificate_populate( AlpPrvSpfSoCertificateReference *certificate ) {
  AlpPrvSpfSoErr err = ALP_PRV_SPF_SO_E_NOERR;

  do{
    ALP_PRV_SPF_SO_BREAK_IF_NULL( certificate, AlpPrvSpfSoCertificateReference, err );
    
  } while( 0 );

  return( err );
}

AlpPrvSpfSoErr alp_prv_spf_so_package_find_by_id( SecurityObject *object, unsigned int packageid, AlpPrvSpfSoPackage **out_package ) {
  AlpPrvSpfSoErr err = ALP_PRV_SPF_SO_E_NOERR;
  unsigned int i;
  AlpPrvSpfSoPackage *current;

  do { 
    ALP_PRV_SPF_SO_BREAK_IF_NULL( object, SecurityObject , err );
    ALP_PRV_SPF_SO_BREAK_IF_NULL( out_package, AlpPrvSpfSoPackage *, err );

    for( i = 0; i < object->packages->len; i++ ) {
      current = (AlpPrvSpfSoPackage *) object->packages->pdata[ i ];
      if( packageid == current->id ) {
	break;
      }
    }

    if( i >= object->packages->len ) {
      err = ALP_PRV_SPF_SO_E_UNKNOWN_PACKAGE;
      break;
    }
    
    *out_package = current;

  } while( 0 );

  return( err );
}

AlpPrvSpfSoErr alp_prv_spf_so_package_find_by_uid( SecurityObject *object, unsigned int uid, AlpPrvSpfSoPackage **out_package ) {
  AlpPrvSpfSoErr err = ALP_PRV_SPF_SO_E_NOERR;
  unsigned int i;
  AlpPrvSpfSoPackage *current;

  if ((object == NULL) || (out_package == NULL)) {
      return (ALP_PRV_SPF_SO_E_NULL_PARAMETER);
  }

  *out_package = NULL;

  err = ALP_PRV_SPF_SO_E_UNKNOWN_PACKAGE;
  for (i = 0; i < object->packages->len; i++) {
      current = (AlpPrvSpfSoPackage *)object->packages->pdata[i];
      if (uid == current->uid) {
	  *out_package = current;
	  err = ALP_PRV_SPF_SO_E_NOERR;
	  break;
      }
  }

  return (err);
}


AlpPrvSpfSoErr alp_prv_spf_so_package_find_by_package( SecurityObject *object, char *package, AlpPrvSpfSoPackage **out_package ) {
  AlpPrvSpfSoErr err = ALP_PRV_SPF_SO_E_NOERR;
  unsigned int i;
  AlpPrvSpfSoPackage *current;

  do { 
    ALP_PRV_SPF_SO_BREAK_IF_NULL( object, SecurityObject , err );
    ALP_PRV_SPF_SO_BREAK_IF_NULL( package, char , err );
    ALP_PRV_SPF_SO_BREAK_IF_NULL( out_package, AlpPrvSpfSoPackage *, err );

    for( i = 0; i < object->packages->len; i++ ) {
      current = (AlpPrvSpfSoPackage *) object->packages->pdata[ i ];
      if( !strcmp( package, current->package ) ) {
	break;
      }
    }

    if( i >= object->packages->len ) {
      err = ALP_PRV_SPF_SO_E_UNKNOWN_PACKAGE;
      break;
    }
    
    *out_package = current;

  } while( 0 );

  return( err );
}

static AlpPrvSpfSoErr alp_prv_spf_so_parse( SecurityObject *object, char *data, unsigned int length ) {
  AlpPrvSpfSoErr err = ALP_PRV_SPF_SO_E_NOERR;
  XML_Parser parser;
  _ParseData pd;

  do {
    ALP_PRV_SPF_SO_BREAK_IF_NULL( object, SecurityObject, err );
    ALP_PRV_SPF_SO_BREAK_IF_NULL( data, char, err );

    /* todo: put XML parser here */

    parser = XML_ParserCreate( "UTF-8" );

    pd.object = object;
    pd.current_policy = (AlpPrvSpfSoSecurityPolicy *) NULL;
    pd.current_assertion = (AlpPrvSpfSoSecurityAssertion *) NULL;
    pd.current_string = g_string_sized_new(2048);
    pd.in_element = ALP_PRV_SPF_SO_ELEMENT_UNKNOWN;

    XML_SetUserData( parser, &pd );
    XML_SetElementHandler( parser, alp_prv_spf_so_start_handler, alp_prv_spf_so_end_handler );
    XML_SetCharacterDataHandler( parser, alp_prv_spf_so_character_handler );
    XML_Parse( parser, data, length, 0 == 0 );

  } while( 0 );

  return( err );
}

static unsigned int _g_ptr_array_count( GPtrArray *array ) {
  unsigned int rv = 0;

  if( array != (GPtrArray *)NULL ) {

    g_ptr_array_foreach( array, alp_prv_spf_so_count_g_ptr_array_members,
			 (gpointer) &rv );

  }

  return( rv );
}

static void alp_prv_spf_so_count_g_ptr_array_members( gpointer data, gpointer user_data) {
  unsigned int *current;

  if( user_data != (gpointer) NULL ) {
    current = (unsigned int *) user_data;
    (*current)++;
  }
}

static void alp_prv_spf_so_policy_file_out( gpointer data, gpointer user_data ) {
  FILE *fp = (FILE *) user_data;
  AlpPrvSpfSoSecurityPolicy *policy = (AlpPrvSpfSoSecurityPolicy *) data;

  if( ( data != (gpointer) NULL ) && ( user_data != (gpointer) NULL ) ) {
   alp_prv_spf_so_policy_write( policy, fp, "  " );
  }
}

static void alp_prv_spf_so_assertion_file_out( gpointer data, gpointer user_data ) {
  FILE *fp = (FILE *)user_data;
  AlpPrvSpfSoSecurityAssertion *assertion = (AlpPrvSpfSoSecurityAssertion *) data;

  if( ( data != (gpointer) NULL ) && ( user_data != (gpointer) NULL ) ) {
   alp_prv_spf_so_assertion_write( assertion, fp, "   " );
  }
}

static void alp_prv_spf_so_certref_in_policy_file_out( gpointer data, gpointer user_data ) {
  FILE *fp = (FILE *)user_data;
  AlpPrvSpfSoCertificateReference *certificate = (AlpPrvSpfSoCertificateReference *)data;
  unsigned int issuer_length;
  unsigned int serial_length;

  if( ( data != (gpointer) NULL ) && ( user_data != (gpointer) NULL ) ) {
    issuer_length = strlen( certificate->issuer );
    serial_length = strlen( certificate->serial );
    fprintf( fp, "   <certref issuer=\"%*.*s\" serial=\"%*.*s\"/>\n",
	     issuer_length, issuer_length, certificate->issuer,
	     serial_length, serial_length, certificate->serial );
  }
}

static void alp_prv_spf_so_package_file_out( gpointer data, gpointer user_data ) {
  FILE *fp = (FILE *) user_data;
  AlpPrvSpfSoPackage *package = (AlpPrvSpfSoPackage *)data;
  char hashbuffer[32];
  unsigned int bytes_out;

  alp_prv_spf_so_rfc1113_encode( package->hash, 20, hashbuffer, 31, &bytes_out );
  fprintf( fp, "  <package id=\"%d\" package=\"%s\" refcount=\"%d\" uid=\"%d\" policy=\"%d\" digest=\"%*.*s\"/>\n", package->id, package->package, package->ref_count, package->uid, (package->policy)->id, bytes_out, bytes_out, hashbuffer );
}

static void alp_prv_spf_so_certificate_file_out( gpointer data, gpointer user_data ) {
  FILE *fp = (FILE *) user_data;
  AlpPrvSpfSoCertificateReference *cert = (AlpPrvSpfSoCertificateReference *)data;
  
  if( cert == NULL ) {
    fprintf( fp, "  <certificate/>\n" );
  } else {
    fprintf( fp, "  <certificate issuer=\"%s\" serial=\"%s\">%s</certificate>\n", cert->issuer, cert->serial, cert->encoding);
  }
}

static void alp_prv_spf_so_find_certificate_in_list( gpointer data, gpointer user_data ) {
  void **params =  (void **) user_data;
  AlpPrvSpfSoCertificateReference **out_certificate = (AlpPrvSpfSoCertificateReference **)(params[0]);
  char *issuer = (char *)(params[1]);
  char *serial = (char *)(params[2]);

  if( *out_certificate != (AlpPrvSpfSoCertificateReference *)NULL ) {
    
  }
}

static AlpPrvSpfSoErr alp_prv_spf_so_version_integer_to_string( SecurityObject *object, char *buffer, unsigned int *length ) {

  unsigned int real_length;
  static char *default_version_string = "1.0.0.0";

  real_length = MIN( *length, strlen( default_version_string ) );
  *length = real_length;

  /* todo: actually print the version number */
  strncpy(buffer, default_version_string, real_length );

  buffer[ real_length ] = '\0';

  return( ALP_PRV_SPF_SO_E_NOERR );
}


static gint alp_prv_spf_so_policies_sort( gconstpointer aaa, gconstpointer bbb ) {
  AlpPrvSpfSoSecurityPolicy **aa = (AlpPrvSpfSoSecurityPolicy **)aaa;
  AlpPrvSpfSoSecurityPolicy **bb = (AlpPrvSpfSoSecurityPolicy **)bbb;
  AlpPrvSpfSoSecurityPolicy *a = *aa;
  AlpPrvSpfSoSecurityPolicy *b = *bb;

  return( (gint) (a->id - b->id ));
}

static gint alp_prv_spf_so_assertions_sort( gconstpointer aaa, gconstpointer bbb ) {
  AlpPrvSpfSoSecurityAssertion **aa = (AlpPrvSpfSoSecurityAssertion **)aaa;
  AlpPrvSpfSoSecurityAssertion **bb = (AlpPrvSpfSoSecurityAssertion **)bbb;
  AlpPrvSpfSoSecurityAssertion *a = *aa;
  AlpPrvSpfSoSecurityAssertion *b = *bb;

  return( (gint) strcmp( a->key, b->key) );
}

static gint alp_prv_spf_so_certificates_sort( gconstpointer aaa, gconstpointer bbb ) {
  AlpPrvSpfSoCertificateReference **aa = (AlpPrvSpfSoCertificateReference **)aaa;
  AlpPrvSpfSoCertificateReference **bb = (AlpPrvSpfSoCertificateReference **)bbb;
  AlpPrvSpfSoCertificateReference *a = *aa;
  AlpPrvSpfSoCertificateReference *b = *bb;
  gint rv;
  
  rv = strcmp( a->issuer, b->issuer );

  if( rv == 0 ) {
    rv = strcmp( a->serial, b->serial );
  }
  
  return( rv );
}

static gint alp_prv_spf_so_packages_sort( gconstpointer aaa, gconstpointer bbb ) {
  AlpPrvSpfSoPackage **aa = (AlpPrvSpfSoPackage **)aaa;
  AlpPrvSpfSoPackage **bb = (AlpPrvSpfSoPackage **)bbb;
  AlpPrvSpfSoPackage *a = *aa;
  AlpPrvSpfSoPackage *b = *bb;

  return( (gint) (a->id - b->id ));
}

static void alp_prv_spf_so_start_handler( void *userData, const XML_Char *name, const XML_Char **atts ) {
  _ParseData *pd = (_ParseData *)userData;
  GString *current_string = pd->current_string;
  int element;
  char *values[10];
  int i;

  element = alp_prv_spf_so_find_element_in_list( (char *)name );

  for( i=0; i<10; i++ ) {
    values[i] = NULL;
  }

  if( (element >= 0) && (element < 10 ) ) {
    alp_prv_spf_so_find_attributes( element, (char **)atts,  10, values );
  }

  switch( element ) {
  case ALP_PRV_SPF_SO_ELEMENT_SECURITY:
    /* todo: actually read this value */
    break;
  case ALP_PRV_SPF_SO_ELEMENT_POLICIES:
    pd->object->policy_base = alp_prv_spf_so_new_copy( values[0] );
    pd->object->policy_default = alp_prv_spf_so_new_copy( values[1] );
    pd->object->policy_id_min = atoi( values[2] );
    pd->object->policy_id_next = atoi( values[3] );
    pd->object->policy_gid_min = atoi( values[4] );
    pd->object->policy_gid_next = atoi( values[5] );
    break;
  case ALP_PRV_SPF_SO_ELEMENT_POLICY:
    pd->current_policy = malloc( sizeof( AlpPrvSpfSoSecurityPolicy ) );
   alp_prv_spf_so_policy_initialize( pd->current_policy );
    pd->current_policy->id = atoi( values[0] );
    pd->current_policy->gid = atoi( values[1] );
    pd->current_policy->name = alp_prv_spf_so_new_copy( values[2] );
    if( values[3] != NULL ) {
     alp_prv_spf_so_policy_find_by_name( pd->object, values[3], &( pd->current_policy->base ) );
    }
    if( values[4] != NULL ) {
      pd->current_policy->sgids = g_strdup( values[4] );
    } else {
      pd->current_policy->sgids = "";
    }
    if( values[5] != NULL ) {
      pd->current_policy->bflag = atoi( values[5] );
    }
    if( values[6] != NULL ) {
      pd->current_policy->sflag = atoi( values[6] );
    }
    break;
  case ALP_PRV_SPF_SO_ELEMENT_ASSERTION:
    pd->current_assertion = malloc( sizeof( AlpPrvSpfSoSecurityAssertion ) );
    pd->current_assertion->key = alp_prv_spf_so_new_copy( values[0] );
    pd->current_assertion->type = alp_prv_spf_so_find_type_in_list( values[1] );
    if( ( values[2] != NULL ) && ( !strncmp("true",values[2],4) ) ) {
      pd->current_assertion->kernel = 1;
    } else {
      pd->current_assertion->kernel = 0;
    }

    break;
  case ALP_PRV_SPF_SO_ELEMENT_CERTREF:
    pd->current_certificate = malloc( sizeof( AlpPrvSpfSoCertificateReference ) );
    pd->current_certificate->issuer = alp_prv_spf_so_new_copy( values[0] );
    pd->current_certificate->serial = alp_prv_spf_so_new_copy( values[1] );
    g_ptr_array_add( pd->current_policy->certificates, pd->current_certificate );
    break;
  case ALP_PRV_SPF_SO_ELEMENT_PACKAGES:
    pd->object->package_id_min   = atoi( values[0] );
    pd->object->package_id_next  = atoi( values[1] );
    pd->object->package_uid_min  = atoi( values[2] );
    pd->object->package_uid_next = atoi( values[3] );
    break;
  case ALP_PRV_SPF_SO_ELEMENT_PACKAGE:
    pd->current_package = malloc( sizeof( AlpPrvSpfSoPackage ) );
    pd->current_package->id = atoi( values[0] );
    pd->current_package->package = alp_prv_spf_so_new_copy( values[1] );
    pd->current_package->ref_count = atoi( values[2] );
    pd->current_package->uid= atoi( values[3] );
    if( ( NULL == values[5] ) || ( '\0' == values[5][0] ) ) {
      memset( pd->current_package->hash, 0, 20 );
    } else {
      alp_prv_decode_hash( values[5], pd->current_package->hash );
    }

   alp_prv_spf_so_policy_find_by_id( pd->object, atoi( values[4] ), &(pd->current_package->policy) );
    g_ptr_array_add( pd->object->packages, pd->current_package );
    break;
  case ALP_PRV_SPF_SO_ELEMENT_CERTIFICATES:
    break;
  case ALP_PRV_SPF_SO_ELEMENT_CERTIFICATE:
    pd->current_certificate = malloc( sizeof( AlpPrvSpfSoCertificateReference ) );
    pd->current_certificate->issuer = alp_prv_spf_so_new_copy( values[0] );
    pd->current_certificate->serial = alp_prv_spf_so_new_copy( values[1] );
    break;
  case ALP_PRV_SPF_SO_ELEMENT_PKCS5:
    if( !strcmp( "puk", values[0] ) ) {
      pd->current_pkcs5 = pd->object->puk;
    } else {
      pd->current_pkcs5 = pd->object->pin;
    }

    fprintf( pd->object->debug_fp, "%%SPF-I-PKCS5P; %s %s %s %s\n", values[0], values[1], values[2], values[3] );
    fflush( pd->object->debug_fp );
    

    unsigned int t,p;
    
    t = strlen( values[3] );
    pd->current_pkcs5->salt = malloc( t );
    alp_prv_spf_rfc1113_decode( values[3], t, pd->current_pkcs5->salt, t, &p );
    pd->current_pkcs5->salt_length = p;
    pd->current_pkcs5->iteration = atoi( values[2] );
    pd->current_pkcs5->active = strcmp( "true", values[1] );
    pd->current_pkcs5->id = strdup( values[0] );

    break;
  }

  g_string_erase( pd->current_string, 0, current_string->len );

}

static void alp_prv_spf_so_end_handler( void *userData, const XML_Char *name ){
  _ParseData *pd = (_ParseData *)userData;
  GString *current_string = pd->current_string;
  int element;

  element = alp_prv_spf_so_find_element_in_list( (char *) name );
  switch( element ) {
  case ALP_PRV_SPF_SO_ELEMENT_CERTIFICATE:
    pd->current_certificate->encoding = alp_prv_spf_so_new_copy_with_length( current_string->str, current_string->len );
    g_ptr_array_add( pd->object->certificates, pd->current_certificate );
    break;
  case ALP_PRV_SPF_SO_ELEMENT_ASSERTION:
    pd->current_assertion->value = alp_prv_spf_so_new_copy_with_length( current_string->str, current_string->len );
    g_ptr_array_add( pd->current_policy->assertions, pd->current_assertion );
    break;
  case ALP_PRV_SPF_SO_ELEMENT_POLICY:
    g_ptr_array_add( pd->object->policies, pd->current_policy );
    break;
  case ALP_PRV_SPF_SO_ELEMENT_PKCS5:
    alp_prv_spf_rfc1113_decode( current_string->str, current_string->len, pd->current_pkcs5->phrase, 20, NULL );
    alp_prv_spf_pkcs5_fput( pd->current_pkcs5,  pd->object->debug_fp );
    fprintf(pd->object->debug_fp, "\n");
    
    break;
  }

  g_string_erase( pd->current_string, 0, current_string->len );
}

static void alp_prv_spf_so_character_handler( void *userData, const XML_Char *s, int len) {
  _ParseData *pd = (_ParseData *)userData;
  GString *current_string = pd->current_string;
  
  g_string_append_len( current_string, s, len );
}

static int alp_prv_spf_so_find_string_in_list( char *list[], char *trial ){
  int i;

  if( trial == NULL ) {
    i = -1;
  } else {
    for( i=0; list[i] != NULL; i++ ) {

      if( !strcmp(list[i], trial) )
	break;
    }

    if( list[i] == NULL ) {
      i = -1;
    }
  }
  return( i );
}

static int alp_prv_spf_so_find_element_in_list( char *trial ) {
  static char *so_elements[] = {
    "security",
    "policies",
    "policy",
    "assertion",
    "certref",
    "packages",
    "package",
    "certificates",
    "certificate",
    "pkcs5",
    NULL
  };

  return( alp_prv_spf_so_find_string_in_list( so_elements, trial ) );
}


static void alp_prv_spf_so_find_attributes( unsigned int element, char *atts[],  unsigned int target_size, char *target[] ) {
  char *att;
  char *value;
  char *current;
  unsigned int att_count, current_count, i;

  static unsigned int so_attributes_offsets[] = {
    0, 2, 9, 17, 21, 24, 29, 36, 37, 40
  };

  static char *so_attributes[] = {
    "version",
    NULL,
    "base",
    "default",
    "minid",
    "nextid",
    "mingid",
    "nextgid",
    NULL,
    "id",
    "gid",
    "name",
    "base",
    "sgids",
    "bflag",
    "sflag",
    NULL,
    "id",
    "type",
    "kernel",
    NULL,
    "issuer",
    "serial",
    NULL,
    "minid",
    "nextid",
    "minuid",
    "nextuid",
    NULL,
    "id",
    "package",
    "refcount",
    "uid",
    "policy",
    "digest",
    NULL,
    NULL,
    "issuer",
    "serial",
    NULL,
    "id",
    "active",
    "iteration",
    "salt",
    NULL
  };

  for( i=0; i<target_size; i++ ) {
    target[i] = NULL;
  }

  for( att_count = 0; atts[ att_count ] != NULL; att_count += 2 ) {
    att = atts[ att_count ];
    value = atts[ att_count + 1];
    current_count = so_attributes_offsets[ element ];

    for( i = 0; so_attributes[ i + current_count ] != NULL; i++ ) {
      if( !strcmp( att, so_attributes[i + current_count ] ) ) {
	target[i] = alp_prv_spf_so_new_copy( value );
	break;
      }
    }
  }
}

static unsigned int alp_prv_spf_so_find_type_in_list( char *trial ) {
  int rv = 0;
  static char *types[] = {
    "string",
    "boolean",
    "integer",
    "address",
    "path",
    "url",
    NULL
  };

  rv = alp_prv_spf_so_find_string_in_list( types, trial );

  if( rv < 0 ) { rv = 0; }

  return( (unsigned int) rv );
}

static char * alp_prv_spf_so_new_copy( char *original ) {
  return( alp_prv_spf_so_new_copy_with_length( original, strlen( original ) ) );
}

static char * alp_prv_spf_so_new_copy_with_length( char *original, unsigned int length ) {
  char *rv = malloc( length + 1 );
  strncpy( rv, original, length );
  rv[length] = '\0';
  return( rv );  
}
  
static void alp_prv_spf_so_find_policy_by_id_func (gpointer data, 
		                                   gpointer user_data) {
  AlpPrvSpfSoSecurityPolicy *policy = (AlpPrvSpfSoSecurityPolicy *)data;
  _policyData *out_policy;

  out_policy = (_policyData *)user_data;
  if (out_policy->policy ==  NULL) {
      if (policy->id == out_policy->policy_id) {
          out_policy->policy = policy;
      }
  }
}

static void alp_prv_spf_so_find_policy_by_name_func(gpointer data, 
		                                    gpointer user_data) {
  AlpPrvSpfSoSecurityPolicy *policy = (AlpPrvSpfSoSecurityPolicy *)data;
  _policyData *out_policy;

  out_policy = (_policyData *)user_data;
  if (out_policy->policy == NULL) {
      if (strcmp(policy->name, out_policy->policy_name) == 0) {
          out_policy->policy = policy;
      }
  }
}

static void alp_prv_spf_so_find_policy_by_cert_ref_func( gpointer data, gpointer user_data ) {

}

static void alp_prv_spf_so_rfc1113_encode( char *input, unsigned int input_length, char *output, unsigned int output_max_length, unsigned int *output_bytes_written ) {
#if 1
  unsigned int out_len = ( input_length * 4 ) / 3;
  unsigned int i;
  for( i = 0; ( i < out_len ) && ( i < output_max_length ); i++ ) {
    output[i] = 'A';
  }
  if( output_bytes_written != NULL ) {
    *output_bytes_written = i;
  }
#else
  BIO *bio, *b64, *mem;
  BUF_MEM *buf;
  unsigned int bytes_to_write;

  b64 = BIO_new( BIO_f_base64() );
  mem = BIO_new( BIO_s_mem() );
  bio = BIO_push( b64, mem );

  BIO_write( b64, input, input_length );
  BIO_flush( b64 );
  BIO_get_mem_ptr( b64, &buf );

  memset(output, '*', output_max_length);

  bytes_to_write = MIN( (unsigned int) (buf->length - 1), output_max_length );
  memcpy( output, buf->data, bytes_to_write );


  if( output_bytes_written != NULL ) {
    *output_bytes_written = bytes_to_write;
  }

  BIO_free_all( bio );
#endif
}

AlpPrvSpfSoErr alp_prv_spf_so_queue_add( SecurityObject *object, AlpPrvSpfSoSecurityAssertion *assertion, AlpPrvSpfSoSecurityPolicy *policy, unsigned int *id ) {
  AlpPrvSpfSoErr rv = ALP_PRV_SPF_SO_E_NOERR;

  do {
    if( object->attn_queue_depth >= ALP_PRV_SPF_SO_ATTN_QUEUE_SIZE ) {
      rv = ALP_PRV_SPF_SO_E_FULL_QUEUE;
      break;
    }

    object->attn_queue[ object->attn_queue_depth ].id =
      object->attn_queue_next++;
    object->attn_queue[ object->attn_queue_depth ].policy = policy;
    object->attn_queue[ object->attn_queue_depth ].assertion = assertion;
    object->attn_queue[ object->attn_queue_depth ].response = NULL;

    *id = object->attn_queue[ object->attn_queue_depth ].id;
    
    object->attn_queue_depth++;

  } while ( 0 );

  return( rv );
}

AlpPrvSpfSoErr alp_prv_spf_so_queue_remove (SecurityObject *object, 
		                            unsigned int id) {
  AlpPrvSpfSoErr err = ALP_PRV_SPF_SO_E_NOT_FOUND;
  AlpPrvSpfSoAttn *attn, *current;
  unsigned int i;

  do {
    err = alp_prv_spf_so_queue_find(object, id, & attn);
    if (err != ALP_PRV_SPF_SO_E_NOERR) {
        break;
    }

    for (i = 0; i < (object->attn_queue_depth - 1); i++) {
        current = &(object->attn_queue[i]);

        if (current >= attn) {
	    object->attn_queue[i].id = object->attn_queue[i + 1].id;
	    object->attn_queue[i].policy = object->attn_queue[i + 1].policy;
	    object->attn_queue[i].assertion = object->attn_queue[i + 1].assertion;
	    object->attn_queue[i].response = object->attn_queue[i + 1].response;
        }
    }

    object->attn_queue[i].id = 0;
    object->attn_queue[i].policy = NULL;
    object->attn_queue[i].assertion = NULL;
    object->attn_queue[i].response = NULL;

    object->attn_queue_depth--;
  } while (0);

  return (err);    
}


AlpPrvSpfSoErr alp_prv_spf_so_queue_find (SecurityObject *object, 
		                          unsigned int id, 
					  AlpPrvSpfSoAttn **attn) {
  AlpPrvSpfSoErr err = ALP_PRV_SPF_SO_E_NOT_FOUND;
  unsigned int i;

  do {

    for (i = 0; i < object->attn_queue_depth; i++) {
        if (id == object->attn_queue[i].id) {
	    *attn = &(object->attn_queue[i]);
	    err = ALP_PRV_SPF_SO_E_NOERR;
	    break;
        }
    }

  } while (0);

  return (err);  
}

gboolean alp_prv_spf_so_package_find_by_hash( SecurityObject *object, char *digest, AlpPrvSpfSoPackage **out_package ) {
  gboolean found = FALSE;
  unsigned int i;
  AlpPrvSpfSoPackage *current;

  do {
    for( i=0; i < object->packages->len; i++ ) {
      current = (AlpPrvSpfSoPackage *) object->packages->pdata[i];
      if( 0 == memcmp( digest, current->hash, 20 ) ) {
	*out_package = current;
	found = TRUE;
	break;
      }
    }
  } while( 0 );

  return( found );
}

static void alp_prv_decode_hash( char *input, char *output ) {
  BIO *mem, *b64, *bio;
  
  mem = BIO_new_mem_buf( input, strlen( input ) );
  b64 = BIO_new( BIO_f_base64() );
  BIO_set_flags( b64, BIO_FLAGS_BASE64_NO_NL );
  bio = BIO_push( b64, mem );
  BIO_read( bio, output, 20 );
  
  BIO_free_all(mem);
}
