/********************************************************************
 *
 * Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the entire contents of this file.
 *
 * The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.
 *
 * Portions created by ACCESS Systems Americas, Inc. are Copyright � 2006. All
 * Rights Reserved.
 *
 * Contributor(s): none.
 *
 * ********************************************************************/
#define DBUS_API_SUBJECT_TO_CHANGE

#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-bindings.h>
#include <dbus/dbus-glib-lowlevel.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <pthread.h>
#include <errno.h>
#include <sys/types.h>
#include <unistd.h>
#include <time.h>
#include <signal.h>
#include <pwd.h>
#include <grp.h>
#include <sys/stat.h>
#include <hiker/pdk/sysutils_pdk.h>
#include <hiker/types.h>
#include <hiker/bundlemgr.h>
#include <hiker/attnmgr.h>
#include <hiker/security.h>
#include "SecurityService.h"
#include "SecurityObject.h"
#include "SecurityUpdate.h"
#include "SecurityKernel.h"
#include "SecurityManifest.h"
#include "SecurityUtilities.h"
#include <openssl/sha.h>
#include "SecurityPKCS5Support.h"

static gboolean alp_prv_spf_pbkdf1 (AlpPrvSpfSuPKCS5 *context,
                                    char *phrase,
                                    char *outdata);

char *default_salt = "bXNoIHdhcyBoZXJl";


/*!
** @brief  This   function  initializes  a   AlpPrvSpfSuPKCS5  context  data
** structure with default parameters. The  only parameter that we don't have
** a default for is the "id" parameter.
** 
** @param[in/out]  context  -   this  is  a  pointer  to   a  pointer  to  a
** AlpPrvSpfSuPKCS5 data  structure. If it  is null, then the  function will
** allocate  a data structure  on the  caller's behalf,  and will  write the
** address of  the new data  structure into the  pointer pointed to  by this
** parameter. If the value is non-null, the function assumes it points to an
** AlpPrvSpfSuPKCS5 data structure and initializes it.
** 
** @param[in] id -  the "id" of the context. Typical  values include PIN and
** PUK.
** 
** @par Example
** Here's a quick  example of using this function.  Function foo allows this
** init  call to  allocate  the  context while  function  bar allocates  the
** context itself.
** 
** @code
** #include <SecurityUtilities.c>
** 
** void foo() {
**   AlpPrvSpfSuPKCS5 *context = NULL;
** 
**   alp_prv_spf_su_pkcs5_init_default( &context, "foo" );
** }
** 
** void bar() {
**   AlpPrvSpfSuPKCS5 *context = malloc( sizeof( AlpPrvSpfSuPKCS5 ) );
** 
**   alp_prv_spf_su_pkcs5_init_default( &context, "bar" );
** }
** @endcode
** 
*/

void alp_prv_spf_pkcs5_init_default (AlpPrvSpfSuPKCS5 **context, char *id) {
  alp_prv_spf_pkcs5_init(context, NULL, 0, id);
}

void alp_prv_spf_pkcs5_init (AlpPrvSpfSuPKCS5 **context, char *salt, 
		             unsigned int iteration, char *id) {
  unsigned int salt_length = 0;

  do {
    /* allocate the context if needed */
    if (NULL == *context) {
        *context = (AlpPrvSpfSuPKCS5 *)malloc(sizeof(AlpPrvSpfSuPKCS5));
    }

    /*
     * Out of resources bail
     */
    if (NULL == *context) {
        break;
    }

    /* 
     * generate a random salt if needed 
     */
    if (NULL == salt) {
        /* 
 	 * TODO : actually read 12 bytes from /dev/urandom 
	 */
        salt = default_salt;
    }

    /* 
     * set the default iteration count 
     */ 
    if (0 == iteration) {
        iteration = DEFAULT_ITERATION_COUNT;
    }

    /* 
     * un-rfc1113-ify the salt 
     */
    (*context)->salt = malloc(MAX_PKCS5_SALT_LEN);
    if ((*context)->salt == NULL) {
	break;
    }

    salt_length = strlen(salt);

    memset((*context)->salt, '\0', MAX_PKCS5_SALT_LEN);
    alp_prv_spf_rfc1113_decode(salt, salt_length, 
                               (*context)->salt, MAX_PKCS5_SALT_LEN - 1, 
                               &((*context)->salt_length));

    (*context)->iteration = iteration;
    (*context)->active = 0;
    (*context)->id = strdup(id);

  } while(0);

  return;
}


void alp_prv_spf_pkcs5_fput (AlpPrvSpfSuPKCS5 *context, FILE *fp) {
  char *active;
  char *salt = NULL;
  char *phrase = NULL;
  unsigned int salt_length;
  unsigned int salt_bytes_written = 0;
  unsigned int phrase_bytes_written = 0;

  do {

    /* 
     * convert binary values to strings 
     */
    if (0 == context->active) {
        active = "true";
    } else {
        active = "false";
    }

    salt_bytes_written = 0;

    salt = malloc(MAX_PKCS5_SALT_LEN);
    if (salt == NULL) {
	break;
    }

    memset(salt, '\0', MAX_PKCS5_SALT_LEN);
    alp_prv_spf_rfc1113_encode(context->salt, context->salt_length, 
		               salt, MAX_PKCS5_SALT_LEN - 1, 
		               &salt_bytes_written);
    
    phrase = malloc(MAX_PKCS5_PHRASE_LEN);
    if (phrase == NULL) {
	break;
    }

    memset(phrase, '\0', MAX_PKCS5_PHRASE_LEN);
    alp_prv_spf_rfc1113_encode(context->phrase, SHA1_HASH_SIZE, 
		               phrase, MAX_PKCS5_PHRASE_LEN - 1, 
				  &phrase_bytes_written);

    fprintf(fp, "<pkcs5 id=\"%s\" active=\"%s\" iteration=\"%d\" salt=\"%*.*s\">%*.*s</pkcs5>", context->id, active, context->iteration, salt_bytes_written, salt_bytes_written, salt, phrase_bytes_written, phrase_bytes_written, phrase);

  } while( 0 );

  if (NULL != salt) {
      free(salt);
  }

  if (NULL != phrase) {
      free(phrase);
  }
}


static void alp_prv_spf_pkcs5_phrase_set (AlpPrvSpfSuPKCS5 *context, 
		                   char *phrase) {
  alp_prv_spf_pbkdf1(context, phrase, context->phrase);
}


static void alp_prv_spf_pkcs5_phrase_test (AlpPrvSpfSuPKCS5 *context, 
		                    char *phrase, 
                                    unsigned int *success) {
  char trial[SHA1_HASH_SIZE];
  unsigned char s;
  unsigned int i;
  gboolean rv = FALSE;

  rv = alp_prv_spf_pbkdf1(context, phrase, trial);
  if (rv == FALSE) {
      *success = 0;
      return;
  }

  /* 
   * Use the SecurityTechnique.Com trick to see if the two hashes are the
   * same. See http://www.securitytechnique.com/1/8 for more info.        
   */

  s = 0;

  for (i = 0; i < SHA1_HASH_SIZE; i++) {
    s |= context->phrase[i] ^ trial[i];
  }

  *success = (unsigned int)s;
}

/*
 * check !!!!!!!!
 */
static gboolean alp_prv_spf_pbkdf1 (AlpPrvSpfSuPKCS5 *context, 
		                    char *phrase, 
				    char *outdata) {
  char *initial;
  char output[SHA1_HASH_SIZE], input[SHA1_HASH_SIZE];
  unsigned int salt_length = context->salt_length;
  unsigned int phrase_length = strlen(phrase);
  unsigned int i;
  gboolean rv = TRUE;

  initial = malloc(salt_length + phrase_length);
  if (initial == NULL) {
      return FALSE;
  }

  memcpy(initial, phrase, phrase_length);
  memcpy(initial + phrase_length, context->salt, salt_length);

  alp_prv_spf_sha1(initial, salt_length + phrase_length, input);

  for (i = 1; i < context->iteration; i++) {
      alp_prv_spf_sha1(input, SHA1_HASH_SIZE, output);
      memcpy(input, output, SHA1_HASH_SIZE);
  }

  memcpy(outdata, output, SHA1_HASH_SIZE);

  return rv;
}


gboolean alp_prv_spf_pkcs5_set_phrase (SecurityObject *obj, 
		                       unsigned int which, 
                                       char *phrase, 
                                       GError **error ) {
  gboolean rv = TRUE;

  fprintf(obj->debug_fp, "; which = %d, new phrase = (%s)\n; ", which, phrase);

  /* Hmm.. right now we're only allowing the PUK to be set by a policy 
   * update. So.. we'll simply fail silently when we try to set it 
   *
   * TODO: check to see that this request is coming from the passwd app 
   *
   */

  if (which == ALP_SPF_PKCS5_PIN) {
      alp_prv_spf_pkcs5_fput(obj->pin, obj->debug_fp);
      alp_prv_spf_pkcs5_phrase_set(obj->pin, phrase);

      fprintf(obj->debug_fp, "\n; ");

      alp_prv_spf_pkcs5_fput(obj->pin, obj->debug_fp);

      fprintf(obj->debug_fp, "\n");
  }

  alp_prv_spf_so_write(obj, obj->config_path);

  fflush(obj->debug_fp);

  return(rv);
}

gboolean alp_prv_spf_pkcs5_check_phrase (SecurityObject *obj, 
		                                  unsigned int which, 
						  char *trial_phrase, 
						  unsigned int *correct, 
						  GError **error) {
  gboolean rv = TRUE;
  AlpPrvSpfSuPKCS5 *pkcs5;

  /* todo: figure out if anyone is allowed to do this. If not, then we
   * need to check who's calling. If it's not the passwd app, then return
   * a suitable error.
   */

  if (which == ALP_SPF_PKCS5_PIN) {
      pkcs5 = obj->pin;
  } else {
      pkcs5 = obj->puk;
  }

  fprintf(obj->debug_fp, "%%SPFSS-I-PKCS5B; trial = (%s)\n; ", trial_phrase);

  alp_prv_spf_pkcs5_fput(pkcs5, obj->debug_fp);
  alp_prv_spf_pkcs5_phrase_test(pkcs5, trial_phrase, correct);

  fprintf(obj->debug_fp, "%%SPFSS-I-PKCS5A; \n; ");
  alp_prv_spf_pkcs5_fput(pkcs5, obj->debug_fp);

  fflush(obj->debug_fp);

  return (rv);
}


gboolean alp_prv_spf_pkcs5_set_parameters (SecurityObject *obj, 
		                                    unsigned int which, 
						    unsigned int iteration, 
						    unsigned int active, 
						    char *salt, 
						    GError **error) {
  gboolean rv = TRUE;
  AlpPrvSpfSuPKCS5 *pkcs5;
  void *old_salt;


  do {
    if (ALP_SPF_PKCS5_PIN == which) {
        pkcs5 = obj->pin;
    } else {
        pkcs5 = obj->puk;
    }

    pkcs5->iteration = iteration;
    pkcs5->active = active;

    /*
     * save the old salt. needed to revert if malloc fails
     */
    old_salt = pkcs5->salt;

    pkcs5->salt = malloc(MAX_PKCS5_SALT_LEN);
    if (pkcs5->salt == NULL) {
	pkcs5->salt = old_salt;
	rv = FALSE;
	break;
    }

    /*
     * Now safely old salt can be removed
     */
    free(old_salt);

    alp_prv_spf_rfc1113_decode(salt, strlen(salt), 
		               pkcs5->salt, MAX_PKCS5_SALT_LEN, 
                               &pkcs5->salt_length);

    alp_prv_spf_so_write(obj, obj->config_path);

  } while (0);

  return (rv);
}


gboolean alp_prv_spf_pkcs5_get_parameters (SecurityObject *obj, 
		                                    unsigned int which, 
						    unsigned int *iteration, 
						    unsigned int *active, 
						    char **salt, 
						    GError **error) {
  gboolean rv = TRUE;
  AlpPrvSpfSuPKCS5 *pkcs5;
  char *local_salt = NULL;
  int bytes_written = 0;


  do {
    if (ALP_SPF_PKCS5_PIN == which) {
        pkcs5 = obj->pin;
    } else {
        pkcs5 = obj->puk;
    }

    fprintf(obj->debug_fp, "\nGET %d\n", pkcs5->salt_length);
    fflush(obj->debug_fp);

    *iteration = pkcs5->iteration;
    *active = pkcs5->active;

    local_salt = malloc(MAX_PKCS5_SALT_LEN);
    if (local_salt == NULL) {
	rv = FALSE;
	break;
    }

    memset(local_salt, '\0', MAX_PKCS5_SALT_LEN);
    alp_prv_spf_rfc1113_encode(pkcs5->salt, pkcs5->salt_length, 
		               local_salt, MAX_PKCS5_SALT_LEN, 
                               &bytes_written);

    *salt = strdup(local_salt);
    free(local_salt);

  } while (0);

  return (rv);
}
