/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright � 2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/* SecurityQuery.c
**
**
** $Id: //dev/alp/platform/main/servers/security/SecurityQuery.c#61 $
**
** This file contains the "Query implementation" of the SPF security service.
*/

#define OUT_ASSERTION_STR_LEN    128
#define DEFAULT_POLICY_NAME      "default" 

#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-bindings.h>
#include <dbus/dbus-glib-lowlevel.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <pthread.h>
#include <errno.h>
#include <sys/types.h>
#include <unistd.h>
#include <time.h>
#include <signal.h>
#include <pwd.h>
#include <grp.h>
#include <sys/stat.h>
#include <hiker/pdk/sysutils_pdk.h>
#include <hiker/types.h>
#include <hiker/bundlemgr.h>
#include <hiker/attnmgr.h>
#include <hiker/security.h>
#include "SecurityService.h"
#include "SecurityObject.h"
#include "SecurityUpdate.h"
#include "SecurityKernel.h"
#include "SecurityManifest.h"
#include "SecurityUtilities.h"
#include "SecurityPKCS5Support.h"
#include "SecurityUi.h"
#include "SecurityQuery.h"
#include <openssl/sha.h>

/* 
 * Static Function Prototypes 
 */
static gboolean alp_prv_spf_get_assertion_object (
                                SecurityObject *obj,
                                unsigned int policy_id,
                                char *key,
                                AlpPrvSpfSoSecurityAssertion **out_assertion,
                                GError **error);

static gboolean alp_prv_spf_get_assertion_object (
                                SecurityObject *obj, 
                                unsigned int policy_id, 
                                char *key, 
                                AlpPrvSpfSoSecurityAssertion **out_assertion,
                                GError **error) {

  gboolean rv = TRUE;
  App *app;
  GQuark SSQ;
  AlpPrvSpfSoSecurityPolicy *policy_structure = NULL;
  AlpPrvSpfSoSecurityAssertion *assertion_structure = NULL; 
  AlpPrvSpfSoErr err = ALP_PRV_SPF_SO_E_NOERR;

  app = obj->app;
  SSQ = obj->SSQ; 

  do {
    /* 
     * Find the policy 
     */
    err = alp_prv_spf_so_policy_find_by_id(obj, policy_id, &policy_structure);
    if (err != ALP_PRV_SPF_SO_E_NOERR) {
        /* 
         * If you can't find the specified policy, select the default policy 
         */
         err = alp_prv_spf_so_policy_find_by_name(obj, DEFAULT_POLICY_NAME, 
                                                  &policy_structure);
         if (err != ALP_PRV_SPF_SO_E_NOERR) {
	     /* 
              * If we can't find the default policy, bail 
              */
	     rv = FALSE;
	     g_set_error(error, SSQ, ALP_PRV_SPF_SO_E_UNKNOWN_POLICY, 
                         "can't find default policy" ); 
             break;
         }
    }

    /* 
     * Now find the assertion in the policy 
     */
    rv = TRUE;
    err = alp_prv_spf_so_policy_assertion_find(policy_structure, key,
                                               &assertion_structure);

    if (err != ALP_PRV_SPF_SO_E_NOERR) {
        /* 
         * If we're the default policy and we can't find the assertion, 
         * return an error. Otherwise, look in the parent policy (base)
         */
         if (policy_structure->base != (AlpPrvSpfSoSecurityPolicy *)NULL) {
	     rv = alp_prv_spf_get_assertion_object(
                                                    obj, 
                                                    policy_structure->base->id,
                                                    key, 
                                                    out_assertion, 
                                                    error);
         } else {
             rv = FALSE;
	     g_set_error(error, SSQ, ALP_PRV_SPF_SO_E_UNKNOWN_POLICY,
		       "can't find assertion %s in policy %d", key, policy_id); 
	     break;
         }
    }

  } while (0);

  *out_assertion = assertion_structure;

  return (rv);
}


gboolean alp_prv_spf_query_get_assertion_string (SecurityObject *obj, 
                                                 uint policy_id, 
                                                 char *assertion, 
                                                 unsigned int subject_pid, 
                                                 unsigned int client_pid, 
                                                 gchar **out_assertion_value, 
                                                 GError **error) {
  gboolean rv = TRUE;
  App *app;
  GQuark SSQ;
  AlpPrvSpfSoSecurityAssertion *assertion_structure = NULL;
  AlpPrvSpfSoSecurityPolicy *policy_object = NULL;
  unsigned int policy_id_to_use = 0;
  AlpPrvSpfSoErr err = ALP_PRV_SPF_SO_E_NOERR;
  char assertion_value[OUT_ASSERTION_STR_LEN];
  unsigned int out_id;

  if ((obj == NULL) || (assertion == NULL) || (out_assertion_value == NULL) ||
      (error == NULL)) {
      return(FALSE);
  }

  app = obj->app;
  SSQ = obj->SSQ; 

  policy_id_to_use = policy_id;
  do {
    /* 
     * if policyid is zero, find it from the client or subject PID 
     */
    if (policy_id_to_use == 0) {
        err = alp_prv_get_policy_and_package_from_pids(obj, client_pid, 
                                                       subject_pid, 
                                                       &policy_object, NULL);

        if (err != ALP_PRV_SPF_SO_E_NOERR) {
	    rv = FALSE;

	    g_set_error(error, SSQ, err, 
              "problem finding policy from client pid(%d) and subject pid (%d)",
              client_pid, subject_pid);

	    break;
        }

        policy_id_to_use = policy_object->id;
    }

    err = alp_prv_spf_so_policy_find_by_id(obj, policy_id_to_use, 
                                           &policy_object);
    if (err != ALP_PRV_SPF_SO_E_NOERR) {
	rv = FALSE;
        g_set_error(error, SSQ, err, 
                    "Policy %d failed to produce a policy object.", 
		    policy_id_to_use);
        break;
    }

    rv = alp_prv_spf_get_assertion_object(obj, policy_id_to_use,
                                          assertion,
					  &assertion_structure,
					  error);

    if (rv == FALSE) { 
        break;
    }


    /*
     * Check if it is ask, if yes then it requires a ask dialog launch
     */
    if (assertion_structure != NULL) {
        if (strcmp("ask", assertion_structure->value) == 0) {
	    out_id = 0;
	    err = alp_prv_spf_so_queue_add(obj, assertion_structure,
                                           policy_object,  &out_id);

	    if (err != ALP_PRV_SPF_SO_E_NOERR) {
	        rv = FALSE;
	        g_set_error(error, SSQ, ALP_PRV_SPF_SO_E_FULL_QUEUE, 
                            "yikes, the queue is full");
	        break;
	    }

            rv = alp_prv_spf_ask_the_user(obj, app, assertion_structure, 
			                  policy_object, out_id);
	    if (rv == FALSE) {
	        g_set_error(error, SSQ, ALP_PRV_SPF_SO_E_UNSPECIFIED, 
                            "UNSPECIFIED ERROR." );
	        break;
	    }

	    sprintf(assertion_value, "ask:%d", out_id);

	    *out_assertion_value = g_strdup(assertion_value);
        } else {
	    *out_assertion_value = g_strdup(assertion_structure->value);
        }
    } 

  } while (0);

  return (rv);
}


gboolean alp_prv_spf_query_get_assertion_type (SecurityObject *obj,
		                              unsigned int policy_id, 
    				              char *assertion, 
				              unsigned int subject_pid, 
				              unsigned int client_pid, 
				              unsigned int *out_assertion_type, 
				              GError **error) {
  App *app;
  GQuark SSQ;
  gboolean rv = TRUE;
  AlpPrvSpfSoSecurityAssertion *assertion_structure;
  AlpPrvSpfSoSecurityPolicy *policy_object;
  unsigned int policy_id_to_use = 0;
  AlpPrvSpfSoErr err = ALP_PRV_SPF_SO_E_NOERR;

  if ((obj == NULL) || (assertion == NULL) || (out_assertion_type == NULL) ||
      (error == NULL)) {
      return(FALSE);
  }

  app = obj->app;
  SSQ = obj->SSQ; 

  do {
    /* 
     * if policyid is zero, find it from the client or subject PID 
     */
    if (policy_id_to_use == 0) {
        err = alp_prv_get_policy_and_package_from_pids(obj, client_pid,
                                                       subject_pid,
                                                       &policy_object, NULL);

        if (err != ALP_PRV_SPF_SO_E_NOERR) {
            rv = FALSE;

            g_set_error(error, SSQ, err,
              "problem finding policy from client pid(%d) and subject pid (%d)",
              client_pid, subject_pid);

            break;
        }

        policy_id_to_use = policy_object->id;
    }

    err = alp_prv_spf_so_policy_find_by_id(obj, policy_id_to_use,
                                           &policy_object);
    if (err != ALP_PRV_SPF_SO_E_NOERR) {
        rv = FALSE;
        g_set_error(error, SSQ, err,
                    "Policy %d failed to produce a policy object.",
                    policy_id_to_use);
        break;
    }

    rv = alp_prv_spf_get_assertion_object(obj, policy_id_to_use,
                                                       assertion,
                                                       &assertion_structure,
                                                       error);

    if (rv == FALSE) {
        break;
    }

    if (assertion_structure != NULL) {
        *out_assertion_type = assertion_structure->type;
        rv = TRUE;
        break;
    }

    /*
     * Could not find assertion. Return false
     */

    rv = FALSE;

  } while (0);

  return (rv);
}


gboolean alp_prv_spf_query_complete (SecurityObject *obj, 
                                     unsigned int id, 
                                     gchar **out_assertion_value, 
                                     unsigned int *out_exit_okay, 
                                     GError **error) {
  App *app;
  GQuark SSQ;
  AlpPrvSpfSoAttn *attn = NULL;
  gboolean rv = TRUE;
  AlpPrvSpfSoErr err = ALP_PRV_SPF_SO_E_NOERR;


  if ((obj == NULL) || (out_assertion_value == NULL) || 
      (out_exit_okay == NULL) || (error == NULL)) {
      return (FALSE);
  }

  app = obj->app;
  SSQ = obj->SSQ; 

  do { 

    err = alp_prv_spf_so_queue_find(obj, id, &attn);
    if ((err != ALP_PRV_SPF_SO_E_NOERR) || (attn == NULL)) {
        rv = FALSE;
        g_set_error(error, SSQ, ALP_PRV_SPF_SO_E_NOT_FOUND, 
                    "can't find id %d", id );
        break;
    } 

    if (attn->response == NULL) {
        *out_exit_okay = 0;
    } else {
        *out_exit_okay = 1;
        *out_assertion_value = g_strdup(attn->response);
        alp_prv_spf_so_queue_remove(obj, id);
    }

  } while (0);

  return (rv);
}
