/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/* SecurityUtilities.c
**
**
** @file SecurityUtilites.c
** @brief Utility Functions for the SPF Security Service
**
** This file implements  the interface to a number  of utility functions for
** use by the  SPF. Many of these functions are here  to mask the complexity
** of using the OpenSSL C API.
**
** $Id:$
*/

/* Macro Definitions */

/* File Includes */
#include "SecurityUtilities.h"
#include <stdio.h>
#include <string.h>
#include <openssl/bio.h>
#include <openssl/evp.h>
#include <openssl/buffer.h>
#include <openssl/sha.h>

/* Typedefs, Structs, Unions, Enums, etc. */

/* Global Variable Declarations */

/* Function Definitions */

/*!
** @brief This  function performs  an rfc1113 encoding.   That is,  it takes
** binary  data of  a  fixed length  and  converts it  into ASCII  printable
** data. Some  people refer  to this as  "BASE64 encoding" thought  the term
** RFC1113 encoding may be more accurate as RFC1113 specifies one particular
** BASE64  encoding that  seems to  be in  near-universal use.  The function
** takes  a pointer  to  binary data  you  want to  convert,  along with  an
** unsigned int indicating the length of  the input. It also takes a pointer
** to  where  you  want to  put  the  converted  data  and an  unsigned  int
** indicating the size of the output buffer. The function will write at most
** output_length  bytes to  the output  buffer. If  the output_bytes_written
** parameter is  non-null, the  function will write  the number of  bytes it
** wrote to the unsigned int pointed to by this pointer.
** 
** @param[in] input - This is a  char pointer to the beginning of the binary
** data you want to encode.
** 
** @param[in] input_length -  This is the length of  the binary data pointed
** to by the "input" parameter. Remember,  we may want to encode null bytes,
** so we can't use traditional null-terminated C "strings" here.
** 
** @param[out] output - This is a pointer to the output buffer. Encoded data
** will be written here.
** 
** @param[in] output_max_length - This is the size of the output buffer. The
** function  will write  no more  than this  number of  bytes to  the output
** buffer.
** 
** @param[out]  output_bytes_written -  This  is a  pointer  to an  unsigned
** integer. Assuming this parameter is non-null, the function will write the
** number  of bytes  it wrote  to the  output buffer  into the  unsigned int
** pointed to by this parameter.
*/

void alp_prv_spf_rfc1113_encode( char *input, unsigned int input_length, char *output, unsigned int output_max_length, unsigned int *output_bytes_written ) {
  unsigned int bytes_to_write;

  BIO *mem = NULL;
  BIO *b64 = NULL;
  BIO *bio = NULL;
  BUF_MEM *buf = NULL;

  do {

    /* check the input parameters */

    if( NULL == input ) break;
    if( NULL == output ) break;
    if( 0 == output_max_length ) break;

    /* create the SSL BIO structures */

    b64 = BIO_new( BIO_f_base64() );
    mem = BIO_new( BIO_s_mem() );
    bio = BIO_push( b64, mem );

    /* push binary data into the BIO filter chain */

    BIO_write( b64, input, input_length );
    BIO_flush( b64 );

    /* now get a reference to where OpenSSL put the output */
    BIO_get_mem_ptr( b64, & buf );

    /* Figure out how many bytes to write to the output buffer */
    bytes_to_write = buf->length - 1;
    if( bytes_to_write > output_max_length ) {
      bytes_to_write = output_max_length;
    }

    /* Write the bytes to the output buffer */
    memcpy( output, buf->data, bytes_to_write );

    /* if output_bytes_written is non-null, write how many bytes we wrote */
    if( NULL != output_bytes_written ) {
      *output_bytes_written = bytes_to_write;
    }

  } while( 0 );

  /* todo: free data structures */
  /* Strangely, I encountered some errors when trying to free the BIO */
  /* data structures. I'm going to defer this until I have time to    */
  /* research this a bit more. But it does need to be done.           */
  /* if( NULL != bio ) BIO_free_all(bio); */
  /* if( NULL != b64 ) BIO_free_all(b64); */
  /* if( NULL != mem ) BIO_free_all(mem); */
}

/*!
** @brief This  function performs  an rfc1113 decoding.   That is,  it takes
** RFC1113 encoded  ASCII data  and converts it  back into binary  data. The
** function takes a pointer to ASCII data you want to convert, along with an
** unsigned int indicating the length of  the input. It also takes a pointer
** to  where you want  to put  the binary  output data  and an  unsigned int
** indicating the size of the output buffer. The function will write at most
** output_length  bytes to  the output  buffer. If  the output_bytes_written
** parameter is  non-null, the  function will write  the number of  bytes it
** wrote to the unsigned int pointed to by this pointer.
** 
** @param[in] input - This is a  char pointer to the beginning of the ASCII
** data you want to decode.
** 
** @param[in] input_length -  This is the length of  the ASCII data pointed
** to by the "input" parameter.
** 
** @param[out] output - This is a pointer to the output buffer. Encoded data
** will be written here.
** 
** @param[in] output_max_length - This is the size of the output buffer. The
** function  will write  no more  than this  number of  bytes to  the output
** buffer.
** 
** @param[out]  output_bytes_written -  This  is a  pointer  to an  unsigned
** integer. Assuming this parameter is non-null, the function will write the
** number  of bytes  it wrote  to the  output buffer  into the  unsigned int
** pointed to by this parameter.
*/

void alp_prv_spf_rfc1113_decode( char *input, unsigned int input_length, char *output, unsigned int output_max_length, unsigned int *output_bytes_written ) {
  int bytes_read;

  BIO *mem = NULL;
  BIO *b64 = NULL;
  BIO *bio = NULL;

  do {

    /* check the input parameters */

    if( NULL == input ) break;
    if( NULL == output ) break;
    if( 0 == output_max_length ) break;

    /* create the SSL BIO structures */

    mem = BIO_new_mem_buf( input, input_length );
    b64 = BIO_new( BIO_f_base64() );
    BIO_set_flags( b64, BIO_FLAGS_BASE64_NO_NL );
    bio = BIO_push( b64, mem );

    /* pull binary data from BIO filter chain */
    bytes_read = BIO_read( bio, output, output_max_length );

    if( bytes_read < 0 ) {
      bytes_read = 0;
    }

    if( NULL != output_bytes_written ) {
      *output_bytes_written = (unsigned int) bytes_read;
    }
  } while( 0 );

  /* todo: free data structures */
  /* Strangely, I encountered some errors when trying to free the BIO */
  /* data structures. I'm going to defer this until I have time to    */
  /* research this a bit more. But it does need to be done.           */
  /* if( NULL != bio ) BIO_free_all(bio); */
  /* if( NULL != b64 ) BIO_free_all(b64); */
  /* if( NULL != mem ) BIO_free_all(mem); */

}

/*!
** @brief This is a simple convenience function that performs a SHA1 hash of
** a block of  data. SHA1 hashes are 20 bytes long,  so the function assumes
** that the output parameter points to (at least) a 20 byte buffer.
** 
** @param[in] input - a pointer to the data to be hashed
** 
** @param[in] input_length - the length of the input data
** 
** @param[out] output - a pointer to  a 20 byte output buffer into which the
** digest will be placed.
*/

void alp_prv_spf_sha1( char *input, unsigned int input_length, unsigned char *output ) {
  do {

    /* check the input parameters */

    if( NULL == input ) break;
    if( 0 == input_length ) break;
    if( NULL == output ) break;

    /* Now call the SHA1 function */

    SHA1( (const unsigned char *) input, (unsigned long) input_length, (unsigned char *) output );

  } while( 0 );
}
