/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/* SecurityManifest.c
**
**
** Security Object Implementation
**
** This file contains functions and data structures that implement the
** Security Object interface.
**
*/

#define CDATA_START               TRUE
#define CDATA_END                 FALSE
#define CDATA_NOT_FOUND           FALSE

#define SECURITY_ID_TAG       0
#define SECURITY_NAME_OFFSET  1
#define CERTISSUER_TAG         0
#define CERTISSUER_OFFSET      1
#define CERTSERIAL_TAG        2
#define CERTSERIAL_OFFSET     3

#define ALP_PRV_SPF_XML_ELEMENT_UNKNOWN      -1
#define ALP_PRV_SPF_XML_ELEMENT_SECURITY      0
#define ALP_PRV_SPF_XML_ELEMENT_CERTREF       1

#define ALP_PRV_SPF_TMP_XML_PATH "/var/opt/alp/data/lib/security/"


#include <stdio.h>
#include <glib.h>
#include <expat.h>
#include <string.h>
#include "SecurityXml.h"


/* Typedefs, Structs, Unions, Enums, etc. */

typedef struct {
    SecurityObject* _object;
    AlpPrvSpfXmlParse *_xmlParse; 
    gboolean _cdata;
} _ParseData;


/*
 *Interested ELEMENTS in the XML based update message
 */
static char *spf_xml_elements[] = {
    "security",
    "certref",
};


/************************* Static Function Prototypes **********************/

static void alp_prv_spf_xml_start_handler (void *userData,
                                           const XML_Char *name,
                                           const XML_Char **atts);

static void alp_prv_spf_xml_end_handler (void *userData, 
		                         const XML_Char *name);

static void alp_prv_spf_xml_character_handler (void *userData,
                                               const XML_Char *s, 
					       int len);

static void alp_prv_spf_xml_cdata_start_handler (void *userData);

static void alp_prv_spf_xml_cdata_end_handler (void *userData);

static int alp_prv_spf_find_string_in_xml_elements (char *trial);

/**************************************************************************/

AlpPrvSpfSoErr alp_prv_spf_xml_parse (SecurityObject *object, 
		                      const char *xmlData,
                                      AlpPrvSpfXmlParse *xmlParse,
				      unsigned int len) {

    AlpPrvSpfSoErr err = ALP_PRV_SPF_SO_E_NOERR;
    XML_Parser parser;
    _ParseData pd;
    enum XML_Error xml_error = 0;
    XML_LChar *xml_error_string = (XML_LChar *) "";
    long byte_index;
    int line_number;
    int column_number;

    do {
      if ((object == NULL) || (xmlParse == NULL)) {
          err = ALP_PRV_SPF_SO_E_NULL_PARAMETER;
	  break;
      }

      parser = XML_ParserCreate( "UTF-8" );

      pd._object         = object;
      pd._cdata          = CDATA_NOT_FOUND;
      pd._xmlParse       = xmlParse;
    

      XML_SetUserData(parser, &pd);

      XML_SetElementHandler(parser, 
                            alp_prv_spf_xml_start_handler, 
                            alp_prv_spf_xml_end_handler );

      XML_SetCharacterDataHandler(parser, alp_prv_spf_xml_character_handler);

    /*
     * SMIME MSG embedded in the CDATA section 
     */
      XML_SetCdataSectionHandler(parser, 
                                 alp_prv_spf_xml_cdata_start_handler, 
                                 alp_prv_spf_xml_cdata_end_handler );

      XML_Parse(parser, xmlData, len, 0 == 0);
      if (0 != (int) xml_error) {
          xml_error_string = (XML_LChar *)XML_ErrorString((int)xml_error);

          byte_index    = XML_GetCurrentByteIndex(parser);
          line_number   = XML_GetCurrentLineNumber(parser);
          column_number = XML_GetCurrentColumnNumber(parser);

          err = ALP_PRV_SPF_SO_E_PARSE;

          fprintf(object->debug_fp, "%%SPFSM-F-PARSE;"
                  "While parsing we received an error.\n;"
                  "Error %d (%s) at byte %ld, line %d, column %d.\n", 
                   xml_error, xml_error_string, 
                   byte_index, line_number, column_number );

          fflush( object->debug_fp );
          break;  	   
      }

      err = ALP_PRV_SPF_SO_E_NOERR;

    } while (0);

    return (err);
}


/*************************** Private methods ******************************/

static void alp_prv_spf_xml_certref_get_certificate (gpointer data,
                                                     gpointer user_data) {
    AlpPrvSpfCertificateReference *certref=
                             (AlpPrvSpfCertificateReference *)user_data;
    AlpPrvSpfCertificateReference *certificate =
                             (AlpPrvSpfCertificateReference *)data;

    if ((data != (gpointer)NULL) && (user_data != (gpointer)NULL)) {
        if ((strcmp(certref->_serial, certificate->_serial) == 0) &&
            (strcmp(certref->_issuer, certificate->_issuer) == 0)) {
            certref->_encoding = strdup(certificate->_encoding);
        }
    }
}

static void alp_prv_spf_xml_start_handler (void *userData, 
                                           const XML_Char *name, 
                                           const XML_Char **atts) {
    _ParseData *pd = (_ParseData *)userData;
    int element;
    int numElements = (sizeof(spf_xml_elements)/sizeof(char *));

    element = alp_prv_spf_find_string_in_xml_elements((char *)name);

    /*
     * if the element == -1, then the element was not found. Make a immediate
     * exit
     */
    do {
      if (element < 0) {
          break; 
      }
        
      switch (element) {
          case ALP_PRV_SPF_XML_ELEMENT_SECURITY:
              pd->_xmlParse->_name = NULL;
              if (strcmp(atts[SECURITY_ID_TAG], "id") == 0) {
                  pd->_xmlParse->_name = strdup(atts[SECURITY_NAME_OFFSET]);
              }
  	    break;

          case ALP_PRV_SPF_XML_ELEMENT_CERTREF:
              if (strcmp(atts[CERTISSUER_TAG], "issuer") == 0) {
                  pd->_xmlParse->_certref._issuer = 
			                      strdup(atts[CERTISSUER_OFFSET]);
              }

              if (strcmp(atts[CERTSERIAL_TAG], "serial") == 0) {
                  pd->_xmlParse->_certref._serial = 
			                      strdup(atts[CERTSERIAL_OFFSET]);
              }
  
              pd->_xmlParse->_certref._encoding = NULL;
              g_ptr_array_foreach(pd->_object->certificates, 
                                  alp_prv_spf_xml_certref_get_certificate,
                                  (gpointer)&pd->_xmlParse->_certref);
            
              break;
          
           default:
              /*
               * TODO: throw a DEBUG message, UNKOWN ELEMENT in the
               *       update message
               */
              break;
      }

    } while (0);

    return;
}

static void alp_prv_spf_xml_end_handler (void *userData, const XML_Char *name){
    _ParseData *pd = (_ParseData *)userData;

    return;
}


static void alp_prv_spf_xml_character_handler (void *userData, 
		                              const XML_Char *s, 
					      int len) {
    _ParseData *pd = (_ParseData *)userData;

    if (pd->_cdata) {
        /* 
	 * If cdata Flag is to TRUE when the stream currently points
	 * to signed SMIME msg . End of the CDATA section is 
	 * signalled by setting cdata Flag to FALSE in the cdata_end_handler
	 */
         strncat(pd->_xmlParse->_smimeMsg, s, len);
    }
}


static void alp_prv_spf_xml_cdata_start_handler (void *userData) {
     _ParseData *pd = (_ParseData *)userData;

    /* 
     * cdata Flag is set to TRUE. We have hit the CDATA section of the 
     * stream which contains smime msg which is handled by 
     * alp_prv_spf_xml_character_handler.
     *
     * XML parser call sequence when it hits CDATA section is
     *      alp_prv_spf_xml_cdata_start_handler()
     *      alp_prv_spf_xml_character_handler()
     *      alp_prv_spf_xml_cdata_end_handler() 
     */
    
    pd->_cdata = CDATA_START;
}


static void alp_prv_spf_xml_cdata_end_handler (void *userData) {
    _ParseData *pd = (_ParseData *)userData;
    /*
     * End of CDATA section encountered. Flag this so that character handler
     * can stop writing to the new policy file
     */
    pd->_cdata = CDATA_END;
}


static int alp_prv_spf_find_string_in_xml_elements (char *trial) {
    int i = -1;
    int numElements = (sizeof(spf_xml_elements)/sizeof(char *));

    if (trial == NULL) {
        return (i); 
    }

    for (i=0; i < numElements; i++) {
        if (!strcmp(spf_xml_elements[i], trial)) {
            break;
        }
    }

    if (i == numElements) {
        i = -1;
    }

    return (i);
}


void alp_prv_spf_cleanup_xml_parse (AlpPrvSpfXmlParse *xmlParse) {

    if (xmlParse == NULL) {
        return;
    }

    if (xmlParse->_smimeMsg != NULL) {
        free(xmlParse->_smimeMsg);
	xmlParse->_smimeMsg = NULL;
    }

    if (xmlParse->_name != NULL) {
        free(xmlParse->_name);
	xmlParse->_name = NULL;
    }

    if (xmlParse->_certref._issuer != NULL) {
        free(xmlParse->_certref._issuer);
	xmlParse->_certref._issuer = NULL;
    }

    if (xmlParse->_certref._serial != NULL) {
        free(xmlParse->_certref._serial);
	xmlParse->_certref._serial = NULL;
    }

    if (xmlParse->_certref._encoding != NULL) {
        free(xmlParse->_certref._encoding);
	xmlParse->_certref._encoding = NULL;
    }
}

