/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/*!
 *    @file volumesvcs_rpc.c implements the interface to the client
 *    @addtogroup ALP_VOLUMESVCS Volume Services
 *    @{
 *    This functions entirely thorugh callbacks, the only extern routine is
 *    prv_init_rpc
 */

#include <string.h>
#include <sys/types.h>
#include <hiker/fail.h>
#include <hiker/ipc.h>
#include "volumesvcs.h"
#include <errno.h>

extern bool prv_mount( prv_fsinfo_t *entry );
extern bool prv_umount( prv_fsinfo_t *entry );
extern bool prv_bind_mount( gchar *name, gchar *path );
extern bool prv_rename( prv_fsinfo_t *entry, const gchar *name );
extern bool prv_makefs( prv_fsinfo_t *entry, const gchar *name, const gchar *fstype);
extern void prv_dump(int trace_level);
extern void prv_device_add(prv_fsinfo_t *entry);
extern void prv_device_remove(prv_fsinfo_t *entry);
extern prv_fsinfo_t *prv_entry_match( const gchar *udi, const bool allocate );
extern GArray *gfsinfo;
extern GArray *gslots;
extern GPtrArray *gfstypes;
static AlpChannel* sch;
static GStaticMutex suser_lock= G_STATIC_MUTEX_INIT;

static guint user_cmd_range[2]={ALP_VOLUMESVCS_USER_CMD_MIN,ALP_VOLUMESVCS_USER_CMD_MAX};
static guint udev_cmd_range[2]={ALP_VOLUMESVCS_UDEV_CMD_MIN,ALP_VOLUMESVCS_UDEV_CMD_MAX};

static alp_status_t pack_entry( AlpMessage *rsp, prv_fsinfo_t *e )
{
	alp_status_t as = ALP_STATUS_OK;
	
	PACKSTR( as, rsp, e->pub.udi, error );
	PACKSTR( as, rsp, e->pub.dev_path, error );
	PACKSTR( as, rsp, e->pub.fs_path, error );
	PACKSTR( as, rsp, e->pub.name, error );
	PACKSTR( as, rsp, e->pub.slot, error );
	PACKSTR( as, rsp, e->pub.vendor_model, error );
	PACKSTR( as, rsp, e->pub.uuid, error );
	PACKSTR( as, rsp, e->pub.card_type, error );
	PACKBYTES( as, rsp, (char*)&e->pub.total_space, sizeof(e->pub.total_space)+
			sizeof(e->pub.used_space)+sizeof(e->pub.total_files)+
			sizeof(e->pub.used_files)+sizeof(e->pub.flags)+
			sizeof(e->pub.index), error);
error:
	return as;
}

static void msg_recv_cb( AlpConnection *con, AlpMessage *msg, gpointer cr)
{
	alp_status_t as = ALP_STATUS_OK;
	alp_status_t as2send = ALP_STATUS_OK;
	guint32 cmd;
	gchar *value=NULL;
	gchar *name=NULL;
	gchar *path=NULL;
	gchar *fstype=NULL;
	gchar *udi=NULL;
	AlpMessage *rsp=NULL;
	prv_fsinfo_t *e;
	prv_fsinfo_t entry;
	prv_slot_t *sp;
	guint i,j;
	int cnt;
	guint *cmd_range=(guint*)cr;

	bzero((char*)&entry, sizeof(entry));
	/*
	 * its unlikely that multiple callers will get in at the same time, but makefs
	 * does require waiting for new (virtual since its just reprobing the device)
	 * insert events to update things so it is possilbe.  prevent all
	 * the complications by just serializing user requests
	 */
	g_static_mutex_lock( &suser_lock );

	UNPACKSTART(as, msg, error);
	cmd = alp_ipc_message_get_message_ID(msg);
	if (cmd < cmd_range[0] || cmd > cmd_range[1]) {
		ALP_VSTL( 0, "IPC cmd %x, out of range %x:%x",cmd,cmd_range[0],cmd_range[1]);
		goto done;  // ignore it
	}
	ALP_VSTL( 1, "**** NEW IPC MSG RECV: %x *****",cmd);
	rsp = alp_ipc_message_create();
	switch (cmd) {
		/*
		 * used only to generate an exit for use with valgrind, note only root
		 * procs can send on this channel.
		 */
		case ALP_VOLUMESVCS_CTRL_EXIT:
			exit(0);
		/*
		 * udev events
		 */
		case ALP_VOLUMESVCS_UDEV_ADD:
			for (i=0; i< ALP_VOLUMESVCS_UDEV_CODES; i++) {
				/*
				 * sending the env var encoding doesn't buy us much but its cheap
				 * and might ease things since then the env vars don't have to be
				 * in order just be UDEV_CODES of them)
				 */
				UNPACKU32(as, msg, &j, udev_error );
				UNPACKSTR(as, msg, &value, udev_error);
				if ( (value != NULL) && (strlen(value) == 0) )
					value = NULL;
				ALP_VSTL(2, "unpacking of code %d value %s",j,value);
				switch (j) {
					case ALP_VOLUMESVCS_DEVNAME_CODE:
						entry.pub.dev_path = value; break;
					case ALP_VOLUMESVCS_DEVPATH_CODE:
						entry.sys_dev_path = value; break;
					case ALP_VOLUMESVCS_PHYSBUS_CODE:
						entry.pub.card_type = value; break;
					case ALP_VOLUMESVCS_MAJOR_CODE:
						if (value)
							entry.major = atoi(value); break;
					case ALP_VOLUMESVCS_MINOR_CODE:
						if (value)
							entry.minor = atoi(value); break;
					case ALP_VOLUMESVCS_ID_FS_LABEL_CODE:
						entry.pub.name = value; break;
					case ALP_VOLUMESVCS_ID_FS_TYPE_CODE:
						entry.fstype = value; break;
					case ALP_VOLUMESVCS_ID_FS_UUID_CODE:
						entry.pub.uuid = value; break;
					default:
						goto udev_error;
				}
			}
			/*
			 * The memory allocated by UNPACKSTR ends up in the (copy of) entry
			 * that prv_device_add adds to the gfsinfo GArray.  When that element is
			 * removed the memory for all these strings is freed.  This is in contrast
			 * to user requests that are freed after processing.
			 */
			prv_device_add(&entry);
			goto done; // no repyl to send

		case ALP_VOLUMESVCS_UDEV_REMOVE:
			UNPACKSTR(as, msg, &value, udev_error);
			for (i=0; i<gfsinfo->len; i++) {
				e = &g_array_index( gfsinfo, prv_fsinfo_t, i);
				if ( !(e->pub.flags & ALP_VOLUMESVCS_FLAGS_BINDMOUNT) &&
						strcmp(e->sys_dev_path, value) == 0) {
					prv_device_remove(e);
					goto done;
				}
			}
udev_error:
		ALP_VSTL( 2, "alp error %x, errno %d on cmd %x",as, errno,cmd);
		goto done; // no repyl to send

		/*
		 * user requests
		 */
		case ALP_VOLUMESVCS_CMD_MOUNT:
			UNPACKSTR(as, msg, &udi, error);
			if ( (e = prv_entry_match ( udi, FALSE )) != NULL) {
				if (!prv_mount( e ))
					as2send = ALP_STATUS_VOLUMESVCS_INVALID_PARAMETER;
			} else
				as2send = ALP_STATUS_VOLUMESVCS_NOT_FOUND;
			PACKRSPSTART(as, rsp, as2send, msg, error);
			break;

		case ALP_VOLUMESVCS_CMD_UMOUNT:
			UNPACKSTR(as, msg, &udi, error);
			if ( (e = prv_entry_match ( udi, FALSE )) != NULL) {
				if (!prv_umount( e )) {
					if (errno == EBUSY)
						as2send = ALP_STATUS_VOLUMESVCS_DEVICE_BUSY;
					else
						as2send = ALP_STATUS_VOLUMESVCS_INVALID_PARAMETER;
				}
			} else
				as2send = ALP_STATUS_VOLUMESVCS_NOT_FOUND;
			PACKRSPSTART(as, rsp, as2send, msg, error);
			break;

		case ALP_VOLUMESVCS_CMD_BIND:
			UNPACKSTR( as, msg, &name, error);
			UNPACKSTR( as, msg, &path, error);
			if (!prv_bind_mount( name, path ))
				as2send = ALP_STATUS_VOLUMESVCS_INVALID_PARAMETER;
			PACKRSPSTART(as, rsp, as2send, msg, error);
			break;

		case ALP_VOLUMESVCS_CMD_RENAME:
			UNPACKSTR( as, msg, &udi, error);
			UNPACKSTR( as, msg, &name, error);
			if ( (e = prv_entry_match ( udi, FALSE )) != NULL) {
				if (!prv_rename( e, name )) {
					if (errno == EBUSY)
						as2send = ALP_STATUS_VOLUMESVCS_DEVICE_BUSY;
					else
						as2send = ALP_STATUS_VOLUMESVCS_INVALID_PARAMETER;
				}
			} else
				as2send = ALP_STATUS_VOLUMESVCS_NOT_FOUND;
			PACKRSPSTART(as, rsp, as2send, msg, error);
			break;

		case ALP_VOLUMESVCS_CMD_MAKEFS:
			UNPACKSTR( as, msg, &udi, error);
			UNPACKSTR( as, msg, &name, error);
			UNPACKSTR( as, msg, &fstype, error);
			for (i=0; i<gfstypes->len; i++) {
				if (!strcasecmp(g_ptr_array_index( gfstypes, i), fstype ))
					break;
			}
			if (i == gfstypes->len )
				as2send = ALP_STATUS_VOLUMESVCS_INVALID_PARAMETER;
			else if ( (e = prv_entry_match ( udi, FALSE )) == NULL)
				as2send = ALP_STATUS_VOLUMESVCS_NOT_FOUND;
			else if ( e->pub.flags & ALP_VOLUMESVCS_FLAGS_BINDMOUNT )
				as2send = ALP_STATUS_VOLUMESVCS_INVALID_PARAMETER;
			else if (!prv_makefs( e, name, fstype )) {
					if (errno == EBUSY)
						as2send = ALP_STATUS_VOLUMESVCS_DEVICE_BUSY;
					else
						as2send = ALP_STATUS_VOLUMESVCS_INVALID_PARAMETER;
				}
			PACKRSPSTART(as, rsp, as2send, msg, error);
			break;

		case ALP_VOLUMESVCS_CMD_FSINFO:
			PACKRSPSTART( as, rsp, as2send, msg, error);
			/*
			 * slot names that are defined but not referenced by an inserted card
			 * must be included as well, so fake an entry for those.  We also do
			 * not return hidden entries (e.g. devcies that have children).
			 */
			prv_dump(3);
			cnt = 0;
			for (i=0; i<gslots->len; i++) {
				sp = &g_array_index( gslots, prv_slot_t, i);
				if (sp->refcnt == 0) {
					cnt++;
				}
			}
			for (i=0; i<gfsinfo->len; i++) {
				e = &g_array_index( gfsinfo, prv_fsinfo_t, i);
				if (! (e->pub.flags & ALP_VOLUMESVCS_FLAGS_HIDE))
					cnt++;
			}
			ALP_VSTL(2, "returning %d entries",cnt);
			PACKU32( as, rsp, cnt, fsinfo_error );
			for (i=0; i<gfsinfo->len; i++) {
				e = &g_array_index( gfsinfo, prv_fsinfo_t, i);
				if (! (e->pub.flags & ALP_VOLUMESVCS_FLAGS_HIDE))
					if ((as = pack_entry( rsp, e)) != ALP_STATUS_OK)
						goto fsinfo_error;
			}
			bzero( &entry, sizeof(entry));
			for (i=0; i<gslots->len; i++) {
				sp = &g_array_index( gslots, prv_slot_t, i);
				if (sp->refcnt == 0) {
					entry.pub.slot = sp->name;
					entry.pub.index = i+gfsinfo->len;
					if ((as = pack_entry( rsp, &entry)) != ALP_STATUS_OK)
						goto fsinfo_error;
				}
			}
fsinfo_error:
			if (as != ALP_STATUS_OK)
				goto error;
			break;

		case ALP_VOLUMESVCS_CMD_FSTYPES:
			PACKRSPSTART(as, rsp, as2send, msg, error);
			PACKU32( as, rsp, gfstypes->len, error );
			for (i=0; i<gfstypes->len; i++) {
				PACKSTR( as, rsp, g_ptr_array_index( gfstypes, i), error );
			}
			break;

		default:
			ALP_VSTL(1, "unknown cmd (%x)",cmd);
			PACKRSPSTART(as, rsp, ALP_STATUS_VOLUMESVCS_INVALID_PARAMETER, msg, error);
		}
	PACKEND( as, rsp, error );
	as = alp_ipc_connection_send( con, rsp );
	ALP_VSTL( 3, "Response send, %s",as == ALP_STATUS_OK?"success":"failure");
	goto done;

error:
	/*
	 * something has gone wrong, in packing or unpacking
	 */
	if (as2send == ALP_STATUS_OK ) {
		ALP_FAIL_FATAL_IF( as == ALP_STATUS_OK,
				"Internal error, at error return but as == as2send == ALP_STATUS_OK");
		as2send = as;
	}
	/*
	 * start packing this message over, if these pack calls fail there is nothing we can do,
	 * our caller is likely to hang since we can't send a reply
	 */
	PACKRSPSTART(as, rsp, as2send, msg, done);
	PACKEND( as, rsp, done );
	as = alp_ipc_connection_send( con, rsp );
	ALP_VSTL( 3, "Returned status for cmd %x is: %x",cmd, as2send);

done:
	if (rsp)
		alp_ipc_message_destroy(rsp);
	if (msg)
		alp_ipc_message_destroy(msg);
	g_static_mutex_unlock( &suser_lock );

	/*
	 * g_free is happy with a NULL input, and we only free strings allocated to
	 * pass as direct args, all those that went into a fsinfo_t entry will be
	 * freeed when that entry is freed
	 */
	g_free( udi );
	g_free( name );
	g_free( path );
	g_free( fstype );
}

static void connect_cb( AlpConnection *con, gpointer cmd_range)
{
	alp_status_t as;

	as = alp_ipc_connection_register_receive_callback( con, msg_recv_cb, cmd_range);
	ALP_VSTL(3, "New connection received, rcv callback %s",
			(as==ALP_STATUS_OK?"registered":"did NOT register"));
}

void prv_init_rpc()
{
	alp_status_t as;

	/*
	 * create the user request channel
	 */
	sch=alp_ipc_channel_create( ALP_VOLUMESVCS_SERVER_CHANNEL,
			ALP_IPC_CHANNEL_ACCESS_ALL );
	ALP_FAIL_FATAL_IF( sch == NULL, "Could not create volumesvcs server channel (%s)",
			ALP_VOLUMESVCS_SERVER_CHANNEL);
	as = alp_ipc_channel_register_connect_callback( sch, connect_cb, user_cmd_range);
	ALP_FAIL_FATAL_IF( as != ALP_STATUS_OK,
			"Could not register connection callback, alp_status: %x",as);

	/*
	 * create the udev device insert/remove channel
	 */
	sch=alp_ipc_channel_create( ALP_VOLUMESVCS_UDEV_CHANNEL,
			ALP_IPC_CHANNEL_ACCESS_OWNER );
	ALP_FAIL_FATAL_IF( sch == NULL, "Could not create volumesvcs server channel (%s)",
			ALP_VOLUMESVCS_UDEV_CHANNEL);
	as = alp_ipc_channel_register_connect_callback( sch, connect_cb, udev_cmd_range);
	ALP_FAIL_FATAL_IF( as != ALP_STATUS_OK,
			"Could not register connection callback, alp_status: %x",as);

}

