/********************************************************************

Copyright 2006, ACCESS Systems Americas, Inc. All rights reserved.

The contents of this file are subject to the Mozilla Public License Version
1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
for the specific language governing rights and limitations under the
License.

The Original Code is the entire contents of this file.

The Initial Developer of the Original Code is ACCESS Systems Americas, Inc.

Portions created by ACCESS Systems Americas, Inc. are Copyright  2006. All
Rights Reserved.

Contributor(s): none.

********************************************************************/
/*!
 *   @file volumesvcsd.c implements the main volume services daemon
 *   @addtogroup ALP_VOLUMESVCS Volume Services
 *  @{
*/

#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mount.h>
#include <sys/time.h>
#include <time.h>
#include <fcntl.h>
#include <sys/vfs.h>
#include <errno.h>
#include <sched.h>
#include <getopt.h>
#include "volumesvcs.h"
#include <hiker/fail.h>
#include <hiker/notifymgr.h>
#include <hiker/global_settings.h>
#include <hiker/pdk/sysutils_pdk.h>
#include <sys/ioctl.h>
/*
 * the following prevent us from having to include linux/types.h and linux/msdos_fs.h
 */
#define FAT_IOCTL_SET_VOLUME_LABEL 0x40047215
#define MSDOS_NAME 11

static char* smountroot;
static int srootdev = -1; // major device number of root dev to ignore
int gtracelevel=TRACELEVEL;
/*
 * for the most part this is all serialized (enforced by a mutex that serializes
 * user level calls), so only one can be going on at once.  This is all IPC driven, either
 * from device inserts or user requests.  Those callbacks can not be fired until the current
 * one returns back to the glib main loop.  So we are safe from the gfsinfo struct changing
 * out from under us, as long as we do not let glib iterate.  When we do, we must assume
 * things might have changed.
 */
GArray *gfsinfo;       // internal version of the array of prv_fsinfo_t structs
GPtrArray *gfstypes;   // array of fstype strings the kernel supports
GArray *gslots;

extern void prv_init_rpc();
prv_fsinfo_t *prv_entry_match( const gchar *udi, bool allocate );

/*
 * as the system peformance goes up, this timeout values can go down.  it should
 * ideallly only be about 1 sec.  It will slow down the issuing of a notification
 * for unformated volumes only.
 */
#define EVENT_TIMEOUT 2000


/*! @brief dump out a trace of our current view of the mounted volumes & devices
 *
 *  @return - none
 */ 
void prv_dump(int level)
{
	guint i;
	prv_fsinfo_t *e;

	for (i=0; i<gfsinfo->len; i++) {
		e = &g_array_index( gfsinfo, prv_fsinfo_t, i);
		ALP_VSTL( level, "%d - index: %d udi: %s",i, e->pub.index,e->pub.udi);
		ALP_VSTL( level, "%d -         uuid: %s",i, e->pub.uuid);
		ALP_VSTL( level, "%d -     dev_path: %s",i, e->pub.dev_path);
		ALP_VSTL( level, "%d -   card_types: %s",i, e->pub.card_type);
		ALP_VSTL( level, "%d -      fs_path: %s",i, e->pub.fs_path);
		ALP_VSTL( level, "%d -         name: %s",i, e->pub.name);
		ALP_VSTL( level, "%d -         slot: %s",i, e->pub.slot);
		ALP_VSTL( level, "%d - vendor_model: %s",i, e->pub.vendor_model);
		ALP_VSTL( level, "%d -  total_space: %llu",i, e->pub.total_space);
		ALP_VSTL( level, "%d -   used_space: %llu",i, e->pub.used_space);
		ALP_VSTL( level, "%d -  total_files: %llu",i, e->pub.total_files);
		ALP_VSTL( level, "%d -   used_files: %llu",i, e->pub.used_files);
		ALP_VSTL( level, "%d -        flags: %s %s %s %s %s %s",i,
				e->pub.flags & ALP_VOLUMESVCS_FLAGS_PARTITION?"Volume":" ",
				e->pub.flags & ALP_VOLUMESVCS_FLAGS_BINDMOUNT?"BindMount":" ",
				e->pub.flags & ALP_VOLUMESVCS_FLAGS_NOTIFY_PENDING?"NotifyPending":" ",
				e->pub.flags & ALP_VOLUMESVCS_FLAGS_NONOTIFY?"NoMotify":" ",
				e->pub.flags & ALP_VOLUMESVCS_FLAGS_HIDE?"Hide":" ",
				e->pub.flags & ALP_VOLUMESVCS_FLAGS_QUIET_KIDS?"QuietKids":" ");
		ALP_VSTL( level, "%d -        major: %d",i, e->major);
		ALP_VSTL( level, "%d -        minor: %d",i, e->minor);
		ALP_VSTL( level, "%d -       fstype: %s",i, e->fstype);
		ALP_VSTL( level, "%d -   parent udi: %s",i, e->parent_udi);
	}
}

static void prv_newudi(prv_fsinfo_t *entry)
{
	char *udi;
	struct timeval tv;

	// if we did not get a UUID manufacture one
	if (entry->pub.uuid == NULL) {
		entry->pub.uuid = g_malloc(22); // two 10 digit ints plus dash plus null
		MEMCHECK(entry->pub.uuid);
		gettimeofday( &tv, NULL );
		sprintf( entry->pub.uuid, "%d-%d", (int)(tv.tv_sec), (int)(tv.tv_usec));
	}
	// use rootname + uuid + null
	udi = g_malloc(strlen(ACCESS_UDI_ROOT) + strlen(entry->pub.uuid) + 1 );
	MEMCHECK(udi);
	sprintf( udi, "%s%s", ACCESS_UDI_ROOT, entry->pub.uuid );
	entry->pub.udi = udi;
	ALP_VSTL(2,"constructed udi: %s",udi);
}

static prv_fsinfo_t *prv_find_parent( prv_fsinfo_t *entry )
{
	guint i;
	prv_fsinfo_t *e;

	for (i=0; i<gfsinfo->len; i++) {
		e = &g_array_index( gfsinfo, prv_fsinfo_t, i);
		if (AisBsParent(e, entry))
			return e;
	}
	return NULL;
	   
}

/*! @brief update the statfs info for a card or volume
 *  @param[in] entry - pointer to the fsinfo struct
 *
 *  @return - none
 */
static void prv_update_fsinfo( prv_fsinfo_t *entry)
{
	struct statfs sfs;
	char *p, *s, sz[20];
	int len,fd;
	GError *ignored;
	prv_fsinfo_t *parent;


	ALP_VSTL( 2, "updating fsinfo info for udi: %s, fs_path: %s, uuid:%s",
			entry->pub.udi, entry->pub.fs_path, entry->pub.uuid);
	if (entry->pub.fs_path != NULL && !statfs( entry->pub.fs_path, &sfs)) {
		entry->pub.total_space = sfs.f_blocks * sfs.f_bsize;
		entry->pub.used_space = entry->pub.total_space - sfs.f_bfree * sfs.f_bsize;
		entry->pub.total_files = sfs.f_files;
		entry->pub.used_files = entry->pub.total_files - sfs.f_ffree;
	} else {
		/*
		 * dig out the size of the device (not just this partitoin) from /sys
		 * if anything goes wrong, just leave it empty
		 */
		if (entry->parent_udi)
			parent = prv_entry_match( entry->parent_udi, FALSE);
		else
			parent = entry;
		p = parent->pub.dev_path + 5; // skip past /dev/
		s = g_malloc(strlen(p)+strlen("/sys/block//size") + 1);
		MEMCHECK(s);
		sprintf( s, "/sys/block/%s/size", p);
		fd = open(s, O_RDONLY);
		if ( fd >= 0) {
			len = read( fd, sz, 20);
			if (len > 0) {
				sscanf( sz, "%lld", &entry->pub.total_space );
				entry->pub.total_space *= 512; // its a count of 512 byte blocks
			}
		}
		close(fd);
		g_free(s);
	}

}

/*! @brief mount a device entry turning it into a volume entry
 *  @param[in] entry - prv_fsinfo_t for the target device/volume
 *
 *  @return - success or failure of mount
 */
bool prv_mount( prv_fsinfo_t *entry )
{
	char *p;
	char *mdir=NULL;
    char *fstype=NULL;
	struct stat statbuf;
	bool r;
	int f;
	int options;
	int unused;
	alp_status_t ae;

	ALP_VSTL( 2, "attempting to mount udi: %s, uuid:%s", entry->pub.udi, entry->pub.uuid);
	fstype = entry->fstype;
	if ( !(entry->pub.flags && ALP_VOLUMESVCS_FLAGS_BINDMOUNT) &&
			((fstype == NULL) || strlen(fstype) == 0)) {
		ALP_VSTL(0,"Could not discern fstype for udi: %s, mdir: %s, name: %s, fstype: %s",
				entry->pub.udi, mdir, entry->pub.name, fstype);
		return false;
	}
	if (entry->pub.name && strlen(entry->pub.name) > 0) {
		mdir = entry->pub.name;
	} else if (entry->pub.uuid && strlen(entry->pub.uuid) > 0) {
		mdir = entry->pub.uuid;
		ALP_VSTL(1,"Using uuid for mount point for udi: %s, mdir: %s, name: %s, slot: %s",
				entry->pub.udi, mdir, entry->pub.name, entry->pub.slot);
	} else if (entry->pub.slot && strlen(entry->pub.slot) > 0) {
		mdir = entry->pub.slot;
		ALP_VSTL(1,"Using slot for mount point for udi: %s, mdir: %s, name: %s, slot: %s",
				entry->pub.udi, mdir, entry->pub.name, entry->pub.slot);
	} else {
		ALP_VSTL(0,"Could not discern mount point for udi: %s, mdir: %s, name: %s, slot: %s",
				entry->pub.udi, mdir, entry->pub.name, entry->pub.slot);
		return FALSE; // should not happen
	}
	/*
	 * make a path out of the name or slot name we used (allow for a _<n> if we need it)
	 */
	p = g_malloc( strlen(smountroot) + 1 + strlen(mdir) + 5);
	MEMCHECK(p);
	sprintf( p, "%s/%s", smountroot, mdir);
	f = 1;
	while ( g_file_test(p,G_FILE_TEST_IS_DIR) ) {
		sprintf( p, "%s/%s_%d", smountroot, mdir, f++);
		if (f > 100) {
			ALP_VSTL(0, "Could not create mount point with 100 versions");
			g_free(p);
			return FALSE; // REALLY should not happen
		}
	}
	mkdir(p, 0777);
	options = MOUNT_OPTIONS;
	if (entry->pub.flags & ALP_VOLUMESVCS_FLAGS_BINDMOUNT)
		options += MS_BIND;
	ALP_VSTL( 2, "calling mount( %s, %s, %s, %0X, NULL)",entry->pub.dev_path, p, fstype, options);
	if (!mount(entry->pub.dev_path, p, fstype, options, NULL)) {
		entry->pub.fs_path = p;
		prv_update_fsinfo( entry );
		p = g_malloc( strlen(entry->pub.fs_path)+strlen(ALP_VOLUMESVCS_HIDEVOL_PATH)+1);
		MEMCHECK(p);
		sprintf(p, "%s%s", entry->pub.fs_path, ALP_VOLUMESVCS_HIDEVOL_PATH);
		if (!stat( p, &statbuf))
			entry->pub.flags |= ALP_VOLUMESVCS_FLAGS_HIDDEN;
		g_free(p);
		r = TRUE;
		ALP_VSTL(2,"Volume %s mounted on %s",entry->pub.udi, entry->pub.fs_path);
		if ( !(entry->pub.flags & ALP_VOLUMESVCS_FLAGS_NONOTIFY) ) {
			ALP_VSTL(1, "calling alp_notify_brodcast %s for %s",ALP_NOTIFY_EVENT_VOLUME_ADDED,
					entry->pub.udi);
			alp_notify_broadcast(ALP_VOLUMESVCS_APPID, ALP_NOTIFY_EVENT_VOLUME_ADDED,
					(void*)entry->pub.udi, strlen(entry->pub.udi)+1,
					ALP_NOTIFY_QUEUE_DEFAULT, ALP_NOTIFY_PRIORITY_NORMAL);
		}
	} else {
		ALP_VSTL(1,"Volume %s FAILED to mount on %s, errno:%d",entry->pub.udi, p, errno);
		g_free(p);
		r = FALSE;
	}
	return r;
}


/*! @brief umount a volume turning it back into a device, which usually was just ejected
 *  @param[in] entry - the volume
 *
 *  @return - none, always success
 */
bool prv_umount( prv_fsinfo_t *entry )
{
	ALP_VSTL(2,"UNmounting udi: %s from: %s",entry->pub.udi, entry->pub.fs_path);
	/*
	 * unless its busy we ingore all other errors as there isn't anything we can
	 * do about them
	 */
	if (umount2( entry->pub.fs_path, UMOUNT_OPTIONS ) <0) {
		ALP_VSTL(0, "umount of '%s' failed with error: %d",entry->pub.fs_path, errno);
		if (errno == EBUSY)
			return false;
	}
	if (rmdir( entry->pub.fs_path ) < 0) {
		ALP_VSTL(0, "rmdir of '%s' failed with error: %d",entry->pub.fs_path, errno);
	}
	g_free( entry->pub.fs_path);
	entry->pub.fs_path = NULL;
	entry->pub.total_space = 0;
	entry->pub.used_space = 0;
	entry->pub.total_files = 0;
	entry->pub.used_files = 0;
	if ( !(entry->pub.flags & ALP_VOLUMESVCS_FLAGS_NONOTIFY)) {
		ALP_VSTL(1, "calling alp_notify_brodcast %s for %s",ALP_NOTIFY_EVENT_VOLUME_REMOVED,
				entry->pub.udi);
		alp_notify_broadcast( ALP_VOLUMESVCS_APPID, ALP_NOTIFY_EVENT_VOLUME_REMOVED,
				(void*)entry->pub.udi, strlen(entry->pub.udi)+1,
				ALP_NOTIFY_QUEUE_DEFAULT, ALP_NOTIFY_PRIORITY_NORMAL);
	}
	return true;
}

/*! @brief rename a volume
 *  @param[in] entry - prv_fsinfo_t for the target device/volume
 *  @param[in] name - what to rename it too
 *
 *  @return - success or failure of mount
 */
bool prv_rename( prv_fsinfo_t *entry, gchar *name )
{
	int fd;
	char cmd[50];
	prv_fsinfo_t *parent;
	bool r=false;
	char *tmpname;

	ALP_VSTL(2,"rename udi: %s to: %s",entry->pub.udi, name);
	/*
	 * it has to be mounted to rename it, but we want it to be unmounted so a
	 * volume removed is announced.  if its already unmounted that has already
	 * happened.
	 */
	if (entry->pub.fs_path)
		if (!prv_umount( entry )) {
			return false;
		}
	if (entry->pub.flags & ALP_VOLUMESVCS_FLAGS_BINDMOUNT) {
		tmpname = entry->pub.name;
		entry->pub.name = strdup(name);
		MEMCHECK(entry->pub.name);
		if (prv_mount( entry )) {
			// success, discard the old ones
			g_free(tmpname);
			r = true;
		} else {
			// failure, put the old ones back
			g_free(entry->pub.name);
			entry->pub.name = tmpname;
			r = false;
		}
	} else {
		/*
		 * its a real vfat volume on some media, remount it to do the rename but don't
		 * announce that to anyone.  note we presume vfat and an 11 char name max
		 * FIXME This code should be restructued as its ugly -jrg.
		 */
		if (strlen(name) > MSDOS_NAME) {
			r = false;
		} else {
			entry->pub.flags |= ALP_VOLUMESVCS_FLAGS_NONOTIFY+ALP_VOLUMESVCS_FLAGS_HIDE;
			if (!prv_mount( entry )) {
				r = false;
				entry->pub.flags &= ALP_VOLUMESVCS_FLAGS_NONOTIFY+ALP_VOLUMESVCS_FLAGS_HIDE;
			} else {
				ALP_VSTL(2, "opening '%s' for rename ioctl",entry->pub.fs_path);
				fd = open(entry->pub.fs_path, O_RDONLY);
				if (fd < 0) {
					ALP_VSTL(1, "open of '%s' for rename of volume failed with error %d",
							entry->pub.fs_path,errno);
					r = false;
					entry->pub.flags &= ALP_VOLUMESVCS_FLAGS_NONOTIFY+ALP_VOLUMESVCS_FLAGS_HIDE;
				} else {
					ALP_VSTL(2, "calling rename ioctl on '%s'", name);
					if (ioctl( fd, 0 /*make it compile for now *//*FAT_IOCTL_SET_VOLUME_LABEL*/, name) < 0) {
						ALP_VSTL(2, "rename ioctl failed with error %d",errno );
						r = false;
						entry->pub.flags &= ALP_VOLUMESVCS_FLAGS_NONOTIFY+
								ALP_VOLUMESVCS_FLAGS_HIDE;
						close(fd);
					} else {
						close(fd);
						/*
						 * the partprobe will cause removal and addition events (if its umounted),
						 * the NONOTIFY prevents the first from being announced, the second will be
						 * seen normally as a new volume and get mounted and be announced
						 */
						if (! prv_umount( entry )) {
							r = false;
							entry->pub.flags &= ALP_VOLUMESVCS_FLAGS_NONOTIFY+
									ALP_VOLUMESVCS_FLAGS_HIDE;
						} else {
							parent = prv_entry_match( entry->parent_udi, FALSE);
							ALP_FAIL_FATAL_IF( parent == NULL, "did not find parent for entry %s",
									entry->pub.udi );
							sprintf(cmd, "partprobe %s", parent->pub.dev_path);
							ALP_VSTL(2, "executing '%s'", cmd);
							if ((r=system(cmd)) < 0) {
								/*
								 * if the partprobe fails we won't see the new name, but it did get
								 * renamed so its not clear what to do at this point. same for the
								 * umount above sicne if we can not unmount the partproe won't do
								 * anything
								 */
								ALP_VSTL(0,"partprobe failed with error:%d",r);
								entry->pub.flags &= ALP_VOLUMESVCS_FLAGS_NONOTIFY+
										ALP_VOLUMESVCS_FLAGS_HIDE;
							}
						r = true;
						}
					} 
				}
			}
		}
	}
	return r;
}

/*! @brief, free the contents of an entry
 * @param[in] entry - pointer to entry to empty out
 *
 * @return - none
 */
void prv_empty_entry( prv_fsinfo_t *entry)
{
	entry->slotp = NULL;
	g_free( entry->pub.udi );
	entry->pub.udi = NULL;
	g_free( entry->pub.dev_path );
	entry->pub.dev_path = NULL;
	g_free( entry->pub.vendor_model );
	entry->pub.vendor_model = NULL;
	g_free( entry->pub.uuid );
	entry->pub.uuid = NULL;
	g_free( entry->pub.card_type );
	entry->pub.card_type = NULL;
	g_free(entry->pub.slot);
	entry->pub.slot = NULL;
	entry->pub.total_files = 0;
	entry->pub.used_files = 0;
	entry->pub.total_space = 0;
	entry->pub.used_space = 0;
	entry->pub.flags = 0;
	g_free( entry->fstype);
}

/*! @brief remove an entry from the internal table,
 *  @param[in] entry - pointer to entry to remove
 *
 *  @return - none
 */
void prv_entry_remove( prv_fsinfo_t *entry )
{
	guint i;
	prv_fsinfo_t *e;

	ALP_VSTL(2,"removing entry %d udi; %s",entry->pub.index,entry->pub.udi);
	prv_dump(3);
	if (entry->slotp != NULL) {
		entry->slotp->refcnt--;
		ALP_FAIL_FATAL_IF( entry->slotp->refcnt < 0, "remove of entry udi %d casued slot refcnt <0",
				entry->pub.udi);
	}
	prv_empty_entry( entry );
	g_array_remove_index_fast( gfsinfo, entry->pub.index);
	/*
	 * reset the index fields
	 */
	for (i=0; i<gfsinfo->len; i++) {
		e = &g_array_index( gfsinfo, prv_fsinfo_t, i);
		e->pub.index = i;
	}
}

/*! @brief see if we have an existing entry that matches the passed udi
 *  @param[in]  udi - udi to match against
 *  @param[in]  allocate - if true allocate an entry if no match is found
 *
 *  @return - pointer to matching prv_fsinfo_t entry or NULL
 */ 
prv_fsinfo_t *prv_entry_match( const gchar *udi, bool allocate )
{
	guint i;
	prv_fsinfo_t *entry;
	prv_fsinfo_t e;
	char *property;
	char *m,*v;
	char *parent_udi;
	prv_slot_t *sp;
  
	ALP_VSTL(2,"Attempting to match udi: %s",udi);
	for (i=0; i<gfsinfo->len; i++) {
		entry = &g_array_index( gfsinfo, prv_fsinfo_t, i);
		ALP_FAIL_FATAL_IF( entry->pub.index != i, "index is off, entry %d is set to %d",
				i, entry->pub.index);
		if (entry->pub.udi && !strcmp(entry->pub.udi, udi)) {
			return entry;
		}
	}
	return NULL;
}

bool prv_bind_mount( gchar *name, gchar *dir_path)
{
	prv_fsinfo_t *entry;
	prv_fsinfo_t e;
	bool r;

	/*
	 * we allocate an entry by hand here instead of in device_add
	 */
	bzero( (char*)&e, sizeof(e));
	g_array_append_vals( gfsinfo, &e, 1 );
	entry = &g_array_index( gfsinfo, prv_fsinfo_t, gfsinfo->len-1);
	entry->pub.index = gfsinfo->len - 1;
	prv_newudi(entry);
	ALP_VSTL(2,"bind_mount of %s on %s: udi: %s",name, dir_path, entry->pub.udi);
	entry->pub.dev_path = strdup(dir_path);
	MEMCHECK(entry->pub.dev_path);
	entry->pub.name = strdup(name);
	MEMCHECK(entry->pub.name);
	entry->pub.slot = strdup("Internal");
	MEMCHECK(entry->pub.slot);
	entry->pub.card_type = strdup("Internal");
	MEMCHECK(entry->pub.card_type);
	entry->pub.flags |= ALP_VOLUMESVCS_FLAGS_BINDMOUNT;
	r = prv_mount( entry );
	ALP_VSTL(2, "fsinfo table at end of bind mount for device %s",entry->pub.udi);
	prv_dump(2);
	return r;
}

/*
 * this timer is used to wait
 */
gboolean prv_timeout_cb( gpointer flag )
{
	*(int*)flag = 1;
	return FALSE; // this auto unregisters
}
/*! partiiton and create a file system on this device, using the entire device
 * No provisions are made for creating multiple partitions or using one that
 * already exists.
 *  @param[in] entry - pointer to the thing to makefs on, might a volume or a device
 *                    e.g. /dev/mmcblk0p1 or /dev/mmcblk0
 *  @param[in] name - the label for the volume
 *  @param[in] fstype - currently either fat32 or fat16 (case insensitive)
 *
 *  @return - TRUE it all worked, FALSE it didn't
 *
 * beware that the commands executed here will both casue udev events and thus
 * to get those callbacks we have to let the glib main loop run and anytime
 * that happens we assume the gfsinfo array might have changed
 *
 */
bool prv_makefs( prv_fsinfo_t *entry, const gchar *name, const gchar *fstype )
{
	char cmd[100];
	char *dev;
	char *parent;
	bool r;
	bool unmounted;
	int fats;
	int timedout;
	guint i;
	prv_fsinfo_t *e, *parent_entry;

	ALP_VSTL(2, "makefs on %s",entry->pub.udi);
    if (entry->pub.flags & ALP_VOLUMESVCS_FLAGS_PARTITION) {
		// target was a partition maybe mounted maybe not
		parent = strdup( entry->parent_udi );
	} else {
		// target was the parent device (unformated)
		parent = strdup( entry->pub.udi );
	}
	MEMCHECK(parent);
	parent_entry = prv_entry_match( parent, FALSE);
	ALP_FAIL_FATAL_IF( parent_entry == NULL, "makefs found missing parent device");
	dev = strdup( parent_entry->pub.dev_path );
	MEMCHECK(dev);
	parent_entry->pub.flags |= ALP_VOLUMESVCS_FLAGS_QUIET_KIDS;
	/*
	 * find all sibling partitions and make sure all are unmounted (which includes us)
	 * if the target was a raw device there can't be any volumes to unmount
	 */
	for (i=0; i<gfsinfo->len; i++) {
		e = &g_array_index( gfsinfo, prv_fsinfo_t, i);
		ALP_VSTL(3, "checking if %s is part of this makefs and is mounted", e->pub.udi);
		if ( (e->pub.flags & ALP_VOLUMESVCS_FLAGS_PARTITION) && e->parent_udi &&
				e->pub.fs_path && !strcmp(e->parent_udi, parent) ) {
			if (prv_umount( e )) {
				e->pub.flags |= ALP_VOLUMESVCS_FLAGS_NONOTIFY;
			} else {
				/*
				 * this is undesireable as it might leave some volumes mounted and others
				 * not mounted, might want to wait instead and retry. and cleanup better
				 */
				return false;
			}
		}
	}
	sprintf(cmd, "/usr/sbin/parted -s %s mklabel msdos mkpart primary 0%% 100%%", dev);
	ALP_VSTL(2, "executing '%s'", cmd);
	if (system(cmd) < 0) {
		ALP_VSTL(2, "parted returned error %d",errno);
		parent_entry->pub.flags &= ~ALP_VOLUMESVCS_FLAGS_QUIET_KIDS;
		r = false;
	} else {
		/*
		 * this partition create will cause the current partion(s) if present, to be removed
		 * and a new one to be added.  so we want to wait for a device add with  a parent udi
		 * matching the one we saved, but is new (i.e. does not have the no-notify flag on).
		 */
		timedout = 0;
		g_timeout_add( EVENT_TIMEOUT, prv_timeout_cb, &timedout );
		while (!timedout) {
			for (i=0; i<gfsinfo->len; i++) {
				e = &g_array_index( gfsinfo, prv_fsinfo_t, i);
				ALP_VSTL(2, "checking if %s is the new partition", e->pub.udi);
				if ((e->pub.flags & ALP_VOLUMESVCS_FLAGS_PARTITION) &&
						!(e->pub.flags & ALP_VOLUMESVCS_FLAGS_NONOTIFY) &&
						!strcmp( e->parent_udi, parent)) {
					break;
				}
				e = NULL;
			}
			if (e)
				break;
			else {
				/*
				 * wait for a callbck (or some glib event) to fire
				 */
				ALP_VSTL(2," waiting for a udev event");
				g_main_context_iteration( NULL, TRUE );
			}
		}
		if (timedout && e == NULL) {
			ALP_VSTL(0,"timed out wating for a device partition to be added by udev");
			r = false;
		} else {
			/*
			 * When we probe the newly formated partition udev again will do a remove
			 * and add as above, so prevent the remove event being broadcast (but
			 * don't stop the remove from happeneing) and this time allow the add
			 * to be handeled normally.  When the new partion gets mounted that will
			 * generate a volume add event.  we did let glib iterate, so we have to look up
			 * the parent entry again
			 */
			e->pub.flags |= ALP_VOLUMESVCS_FLAGS_NONOTIFY;
			parent_entry = prv_entry_match( parent, FALSE);
			if (parent_entry == NULL) {
				/*
				 * the card was ejected while we let the glib main loop interate to wait
				 * just return an error.
				 */
				r = false;
			} else {
				parent_entry->pub.flags &= ~ALP_VOLUMESVCS_FLAGS_QUIET_KIDS;
				if (!strcasecmp( fstype, "fat16" ))
					fats = 16;
				else
					fats = 32;
				sprintf(cmd, "mkdosfs -n %s -F %d %s", name, fats, e->pub.dev_path);
				ALP_VSTL(2, "executing '%s'", cmd);
				if (system(cmd) < 0) {
					ALP_VSTL(2, "mkdosfs returned error: %d", errno);
					r = false;
				} else {
					sprintf(cmd, "partprobe %s", dev);
					ALP_VSTL(2, "executing '%s'", cmd);
					system(cmd);
					r = true;
				}
			}
		}
	}
	g_free(dev);
	g_free(parent);
	return r;
}

/*! @brief callback for time out waiting for udev to add a partition entry after it has
 *        already told us about a device (the parent) entry.
 *        
 *  @param[in] - data, the udi in quesiton
 *  @return - none
 */
static gboolean prv_device_check_cb( gpointer data )
{
	char *notify;
	char *udi = (char*)data;
	prv_fsinfo_t *e;
	guint j;

	ALP_VSTL(2, "device check timeout for pending notifys for %s",udi);
	for (j=0; j<gfsinfo->len; j++) {
		e = &g_array_index( gfsinfo, prv_fsinfo_t, j);
		if ( !strcmp( e->pub.udi, udi )) {
			if ( e->pub.flags & ALP_VOLUMESVCS_FLAGS_NOTIFY_PENDING ) {
				/*
				 * only devices need this now, but no harm in checking for either, also we
				 * can't get here if NOTIFY_PENDING is not on so that is a bit redundant
				 */
				prv_dump(3);
				e->pub.flags &= ~ALP_VOLUMESVCS_FLAGS_NOTIFY_PENDING;
				if (e->pub.flags & ALP_VOLUMESVCS_FLAGS_PARTITION)
					notify = ALP_NOTIFY_EVENT_VOLUME_ADDED;
				else
					notify = ALP_NOTIFY_EVENT_DEVICE_ADDED;
				ALP_VSTL(1, "delayed call of alp_notify_brodcast %s for %s", notify, e->pub.udi);
				alp_notify_broadcast(ALP_VOLUMESVCS_APPID, notify,
						(void*)e->pub.udi, strlen(e->pub.udi)+1, ALP_NOTIFY_QUEUE_DEFAULT,
						ALP_NOTIFY_PRIORITY_NORMAL );
				break;
			}
		}
	}
	g_free( data );
	return FALSE; // turns off the timer, we always use it as a single shot
}

/*! @brief addition of a device
 *  @param[in] entry -  pointer to a partially filled in block but no udi
 *
 *  @return - none
 */ 
void prv_device_add( prv_fsinfo_t *new_entry)
{
	prv_fsinfo_t *e, *entry;
	prv_fsinfo_t *parent_entry;
	char cmd[250];
	char *category;
	char *notify;
	guint j,i;
	prv_slot_t *sp;
	struct stat statbuf;

	/*
	 * udev does not always provide major and minor dev numbres so check that
	 */
	if (new_entry->major <= 0) {
	    if (stat(new_entry->pub.dev_path, &statbuf) == 0) {
	    	new_entry->major = major(statbuf.st_rdev);
	    	new_entry->minor = minor(statbuf.st_rdev);
	    } else {
			ALP_VSTL(0, "skipping device %s as the major devnum could not be determined",new_entry->sys_dev_path);
			prv_empty_entry( new_entry );
			return;
		}
	}
	if ( (srootdev > 0) && (new_entry->major == srootdev) ) {
		ALP_VSTL(1, "skipping device %s since it is our root fs", new_entry->sys_dev_path);
		prv_empty_entry( new_entry );
		return;
	}
	/*
	 * make a udi (HAL used to give us these so now we just fake them), could use
	 * someting else as an ID but why chnage the API
	 */
	g_array_append_vals( gfsinfo, new_entry, 1 );
	entry = &g_array_index( gfsinfo, prv_fsinfo_t, gfsinfo->len-1);
	prv_newudi(entry);
	entry->pub.index = gfsinfo->len - 1;
	ALP_VSTL(1, "device add for %s, udi: %s", entry->sys_dev_path, entry->pub.udi);

	/*
	 * see if this new entry matches a slot name. only compare on the length of the
	 * requested match so /dev/mmcblk matches /dev/mmcblk0p1.  Make sure
	 * it has a name too. prv_newudi guarantees a UUID so use that if we have to
	 * FIXME This will have to allow regular expressions for /dev/sd..
	 */
	entry->slotp = NULL;
	for (i=0; i<gslots->len; i++) {
		sp = &g_array_index( gslots, prv_slot_t, i);
		if (!strncmp(entry->pub.dev_path, sp->dev_path, strlen(sp->dev_path))) {
			ALP_VSTL(2,"found a match for name: %s, udi: %s",sp->name, entry->pub.udi);
			entry->pub.slot = strdup( sp->name );
			MEMCHECK(entry->pub.slot);
			entry->slotp = sp;
			sp->refcnt++;
			break;
		}
	}
	if (entry->pub.name == NULL)
		entry->pub.name = strdup( entry->pub.uuid );
	/*
	 * whole disks are always minor 0, partions with or without an FS are > 0
	 */
	if (entry->minor != 0) {
		entry->pub.flags |= ALP_VOLUMESVCS_FLAGS_PARTITION;
		/*
		 * if there is a file system on it, mount it otherwise announce a device add
		 * unless its parent has asked to keep things quiet (i.e. a makefs is occuring)
		 */
		notify = NULL;
		parent_entry = prv_find_parent(entry);
		entry->parent_udi = strdup(parent_entry->pub.udi);
		ALP_FAIL_FATAL_IF( parent_entry == NULL, "partition %s has no parent",entry->pub.udi);
		entry->pub.total_space = parent_entry->pub.total_space;
		if (!(parent_entry->pub.flags & ALP_VOLUMESVCS_FLAGS_QUIET_KIDS) ) {
			if ((entry->fstype != NULL) && prv_mount( entry ))
				notify = NULL; // it sends the notify it if the mount works
			else
				notify = ALP_NOTIFY_EVENT_DEVICE_ADDED;
		}
		if (notify) {
			ALP_VSTL(1, "calling alp_notify_brodcast %s for %s", notify, entry->pub.udi);
			alp_notify_broadcast(ALP_VOLUMESVCS_APPID, notify,
					(void*)entry->pub.udi, strlen(entry->pub.udi)+1, ALP_NOTIFY_QUEUE_DEFAULT,
					ALP_NOTIFY_PRIORITY_NORMAL);
		}
		/*
		 * now that we have a child, no notices about the parent need be sent.  its
		 * ok to adjust these even if its a makefs
		 */
		parent_entry->pub.flags &= ~ALP_VOLUMESVCS_FLAGS_NOTIFY_PENDING;
		parent_entry->pub.flags |= ALP_VOLUMESVCS_FLAGS_NONOTIFY | ALP_VOLUMESVCS_FLAGS_HIDE;
	} else {
		/*
		 * udev does not tell us ahead of time if there are partitions defined on a
		 * device and we do not want to send a notifcation here if subsequent partitions
		 * will be added, so just wait a bit and when the timer goes off if the code above
		 * hasn't hidden this entry, the timer callback will announce it.  The exception
		 * is initizliaztion where we purposefully add the volumes first so we will see
		 * them here and not have to use the timeout.
		 */
		char *p;
		prv_update_fsinfo(entry); // just updates the total device size
		entry->pub.flags |= ALP_VOLUMESVCS_FLAGS_NOTIFY_PENDING;
		p = strdup( entry->pub.udi ); // prv_device_check_cd will free it
		MEMCHECK(p);
		g_timeout_add( EVENT_TIMEOUT, prv_device_check_cb, p );
		ALP_VSTL(2,"added timeout for pending notify on %s",entry->pub.udi);
	}
	ALP_VSTL(2, "fsinfo table at end of device added for device %s",entry->pub.udi);
	prv_dump(2);
}

/*! @brief callback for the removal of a device
 *  @param[in] entry - pointer to entry to remove
 *
 *  @return - none
 */ 
void prv_device_remove( prv_fsinfo_t *entry)
{

	ALP_VSTL(1, "device remove for %s", entry->pub.udi);
	if (entry->pub.fs_path) {
		(void)prv_umount( entry );
	} else if ( !(entry->pub.flags & ALP_VOLUMESVCS_FLAGS_NONOTIFY)) {
		ALP_VSTL(1, "calling alp_notify_brodcast %s for %s",
				ALP_NOTIFY_EVENT_DEVICE_REMOVED, entry->pub.udi);
		alp_notify_broadcast(ALP_VOLUMESVCS_APPID, ALP_NOTIFY_EVENT_DEVICE_REMOVED,
				(void*)entry->pub.udi, strlen(entry->pub.udi)+1, ALP_NOTIFY_QUEUE_DEFAULT,
				ALP_NOTIFY_PRIORITY_NORMAL);
	}
	prv_entry_remove( entry );
	ALP_VSTL(2, "\nfsinfo table at end of device removed for device %s",entry->pub.udi);
	prv_dump(2);
}


/*! @brief allocate some global structions and pull out our global settings
 *         this is only called once in the lifetime of the server
 *
 *  @return - success, all errors are fatal
 */ 
static void prv_init_system()
{
    prv_slot_t slot;
	prv_slot_t *sp;
	char dev_path[200];
	char key[200];
	char val[200];
	int n;
	alp_status_t status;
	AlpGlobalSettingsContext *context;


	gfsinfo = g_array_sized_new( TRUE, TRUE, sizeof(prv_fsinfo_t), 10);
	gslots = g_array_sized_new( TRUE, TRUE, sizeof(prv_slot_t), 4);
	gfstypes = g_ptr_array_sized_new(2);
	smountroot = MOUNTROOT;
	/*
	 * these are filesystems that we will FORMAT for, not just that we will mount.
	 * also /proc/filesystems has more entries and vfat is the actual linux fs name
	 * but we wanted to maintain the old fatnames
	 */ 
	g_ptr_array_add( gfstypes, "FAT16");
	g_ptr_array_add( gfstypes, "FAT32");
	/*
	 * pull out the slot name info from global settings.  
	 */
    bzero(&slot, sizeof(slot));
	ALP_FAIL_FATAL_IF(( alp_global_settings_open(&context) != ALP_STATUS_OK ),
			"could not open global settings");
	n = 0;
	status = ALP_STATUS_OK;
	while (status == ALP_STATUS_OK) {
		sprintf(key, "%s%d%s",ALP_SYSTEM_SLOTS_PREFIX,n,ALP_SYSTEM_SLOTS_NAME);
		status = alp_global_settings_get_key_string_value(context, key, val, sizeof(val));
		if (status == ALP_STATUS_OK) {
			g_array_append_vals( gslots, &slot, 1 );
			sp = &g_array_index( gslots, prv_slot_t, n);
			sp->name = strdup(val);
			MEMCHECK(sp->name);
			sprintf(key, "%s%d%s",ALP_SYSTEM_SLOTS_PREFIX,n,ALP_SYSTEM_SLOTS_DEV_PATH);
			ALP_FAIL_FATAL_IF((alp_global_settings_get_key_string_value(context, key, val,
					sizeof(val)) != ALP_STATUS_OK),
					"currupted global settings slot names, %s not found", key);
			sp->dev_path = strdup(val);
			MEMCHECK(sp->dev_path);
			sp->refcnt = 0;
			n++;
		}
	}
	(void)alp_global_settings_close( context );

}



/*! main serer entry
 *  
 */ 
int main(int argc, char **argv)
{
	GMainLoop *loop;
	int i;
	guint j;
	prv_fsinfo_t *entry;
	char *notify;
	bool daemonize = TRUE;
	struct stat statbuf;
	struct option lopts[]={
		{"tracelevel", 1, NULL, 1},
		{"tl", 1, NULL, 1},
		{"nodaemon", 0, NULL, 2},
		{"nd", 0, NULL, 2},
		{"ignore", 1, NULL, 3},
		{0,0,0,0}
	};

	while ( (i = getopt_long( argc, argv, "", lopts, NULL)) > 0) {
		switch (i) {
			case 1:
				gtracelevel = atoi(optarg);
				if (gtracelevel <0 || gtracelevel >MAX_TRACELEVEL)
					gtracelevel = 0;
				break;
			case 2:
				daemonize = FALSE;
				break;
			case 3:
				if (stat(optarg, &statbuf) == 0) {
					srootdev = major(statbuf.st_rdev);
					ALP_VSTL(1, "ignorming device with major num %d",srootdev);
				} else {
					ALP_VSTL(1, "could not fstat root fs device to %s",optarg);
				}
				break;
			default:
				ALP_VSTL(0, "unknowncommandline arguments");
				printf("unknown commandline arguments\n");
				printf("usage: %s [--<tl, tracelevel> <0:%d>] [--<nodaemon, nd>] [--ignore <block dev path>",
						argv[0], MAX_TRACELEVEL);
				exit(0);
		}
	}

	if (daemonize) {
		ALP_VSTL(2, "calling alp_daemon_start( FALSE, TRUE)");
		alp_daemon_start( FALSE, FALSE);
	}
	loop = g_main_loop_new( NULL, FALSE);
	prv_init_system();
	ALP_VSTL(2, "init_sytsem complete");
	prv_init_rpc();
	ALP_VSTL(2, "init_rpc complete");

	umask(0);  // allow modes of rwxrwxrwx on the mount points we create

	/*
	 * we are open for business, indicate that by forking a new process, letting
	 * the parrent return, unless we have been instructed otherwise (debug).  
	 */
	if (daemonize) {
		alp_daemon_release( 0 );
    }	
	ALP_VSTL(1, "Entering main loop");
	g_main_loop_run(loop);
	ALP_VSTL(1, "Main loop exited");
	return 0;
}

